
/*****************************************************************************
 * FLT: An OpenFlight file loader
 *
 * Copyright (C) 2001  Michael M. Morrison   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * See the file EXCLUSIONS for further details.
 *
 *****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <math.h>
#include <string.h>
#include "flt.h"

int
fltReadRecordAttr( FltFile * flt, uint16 * type, uint16 * length )
{
	int result;
	uint16 ltype, llength;
	uint8 * bufPtr;

	assert( flt );
	assert( type );
	assert( length );

	if( fread( &ltype, 1, sizeof( uint16 ), flt->stdFile ) != sizeof(uint16) )
		// eof
		return 0;

	flt->byteOffset += sizeof( uint16 );

	// bail out on NULL type
	if( ltype == 0 ) {
		printf("FLT: Caught NULL node.. bailing out.\n");
		return 0;
	}

#if 0
	// hack -- for extra zeroes (doesn't seem to work in all cases)
	if( ltype == 0 ) {
		uint8 dummy;

		printf("FLT: Caught NULL node, applying Hack.\n");

		// read again
		if( fread( &dummy, 1, sizeof( uint8 ), flt->stdFile ) != sizeof(uint8) ) {
			printf("ONE.5\n");
			return 0;
		}

		flt->byteOffset += sizeof( uint16 );
	}
#endif

//	printf("BO llength: %x\n", flt->byteOffset );

	if( fread( &llength, 1, sizeof( uint16), flt->stdFile ) != sizeof(uint16) ) {
//		printf("TWO\n");
		return 0;
	}

//	printf("llength: %04x\n", llength );

	flt->byteOffset += sizeof( uint16 );

	ENDIAN_16( ltype );
	ENDIAN_16( llength );

	// for header
	if( llength)
		llength -= 4;

	*type = ltype;
	*length = llength;

	// read the block into memory
	bufPtr = FltBufferResize( flt->buffer, llength );

	if( llength ) {
		result = fread( bufPtr, 1, llength, flt->stdFile );
		if( result != llength )  {
			printf("FLT: Unexpected EOF: Type: %d Attempted: %d Actual: %d Off: %x\n",
											ltype, llength, result, flt->byteOffset );
		printf("FLT: %04x %04x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x\n",
										ltype & 0xffff,
										llength & 0xffff,
										bufPtr[0]&0xff,
										bufPtr[1]&0xff,
										bufPtr[2]&0xff,
										bufPtr[3]&0xff,
										bufPtr[4]&0xff,
										bufPtr[5]&0xff,
										bufPtr[6]&0xff,
										bufPtr[7]&0xff,
										bufPtr[8]&0xff,
										bufPtr[9]&0xff );
			return 0;
		}
	}

	flt->byteOffset += llength;

	FltBufferResetWithLength( flt->buffer, llength );

	return 1;
}

void
fltSkipRecord( FltFile * flt, uint16 length )
{
	assert( flt );

//	return fseek( flt->stdFile, length, SEEK_CUR ) + 1;
}

FltFile *
fltOpen( const char * name )
{
	FltFile * f;

	assert( name );

	f = calloc( 1, sizeof( FltFile ) );
	if(!f) {
		printf("unable to malloc..\n");
		return 0;
	}

	f->stdFile = fopen( name, "rb" );
	if( !f->stdFile ) {
		printf("unable to find file: %s\n", name );
		free( f );
		return 0;
	}

	strcpy( f->fileName, name );

	FltBufferAlloc( f );
	FltStackAlloc( f );

	return f;
}

void
fltClose( FltFile * file )
{
	assert( file );

	if( file->stdFile )
		fclose( file->stdFile );

	file->stdFile = 0;
}

void
fltFileFree( FltFile * flt )
{
	unsigned int i;

	//
	// Free all resources
	//

	if( flt->stdFile )
		fltClose( flt );

	// rundown tree, freeing all nodes as we go
	for(i=0;i<flt->numNodes;i++) {
		FltNode * node = flt->allNodes[i];

		// free nodes with special attrs
		switch( node->type ) {

			case FLTRECORD_VERTEXLIST: {
				FltVertexList * l = (FltVertexList *)node;
				free( l->list );
				free( l->indexList );
				break;
			}

			case FLTRECORD_VERTEXPALETTE: {
				FltVertexPalette * p = (FltVertexPalette *)node;
				free( p->verts );
				break;
			}

			case FLTRECORD_SWITCH: {
				FltSwitch * s = (FltSwitch *)node;
				free( s->masks );
				break;
			}

		}

		if( node->child )
			free( node->child );
		free( node );
	}

	if( flt->materialPalette ) {
		if( flt->materialPalette->material )
			free( flt->materialPalette->material );
		free( flt->materialPalette );
	}

	FltBufferFree( flt );
	FltStackFree( flt );

	if( flt->allNodes )
		free( flt->allNodes );

	free( flt );
}
