
/*****************************************************************************
 * FLT: An OpenFlight file loader
 *
 * Copyright (C) 2001  Michael M. Morrison   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * See the file EXCLUSIONS for further details.
 *
 *****************************************************************************/

/*
 * OpenFlight Parser
 * 
 * Author: Mike Morrison
 *         mmorrison@kqcorp.com
 */

#ifndef _FLT_H_
#define _FLT_H_

#ifdef __cplusplus
extern "C" {
#endif

#include <stdio.h>

typedef unsigned char uint8;
typedef unsigned short uint16;
typedef unsigned int uint32;
typedef signed char int8;
typedef signed short int16;
typedef signed int int32;
typedef float real32;
typedef double real64;

typedef enum {
	FLT_TYPE_NULL	= 0x00,
	FLT_UNKNOWN	= 0x01,
	FLT_UINT8 	= 0x02,
	FLT_UINT16	= 0x04,
	FLT_UINT32	= 0x08,
	FLT_INT8		= 0x10,
	FLT_INT16		= 0x20,
	FLT_INT32		= 0x40,
	FLT_REAL32	= 0x80,
	FLT_REAL64	= 0x0100,
	FLT_ARRAY		= 0x0200,
	FLT_POINTER	= 0x0400,
	FLT_POINTER2	= 0x0800,
	FLT_CHAR	= 0x1000
} FltType;

#ifndef MAX_PATHLEN
#define MAX_PATHLEN 256
#endif

/* macros for endian swapping ptr to ptr */
#define _ENDIAN_16( a, b )	((uint8*)(a))[0] = ((uint8*)(b))[1], \
														((uint8*)(a))[1] = ((uint8*)(b))[0];

#define _ENDIAN_32( a, b )	((uint8*)(a))[0] = ((uint8*)(b))[3], \
														((uint8*)(a))[1] = ((uint8*)(b))[2], \
														((uint8*)(a))[2] = ((uint8*)(b))[1], \
														((uint8*)(a))[3] = ((uint8*)(b))[0];

#define _ENDIAN_64( a, b )	((uint8*)(a))[0] = ((uint8*)(b))[7], \
														((uint8*)(a))[1] = ((uint8*)(b))[6], \
														((uint8*)(a))[2] = ((uint8*)(b))[5], \
														((uint8*)(a))[3] = ((uint8*)(b))[4], \
														((uint8*)(a))[4] = ((uint8*)(b))[3], \
														((uint8*)(a))[5] = ((uint8*)(b))[2], \
														((uint8*)(a))[6] = ((uint8*)(b))[1], \
														((uint8*)(a))[7] = ((uint8*)(b))[0];

/////// taken from SGL sgldb/sgldb.h

#if defined(linux)
#include <endian.h>
#endif

// are there no WIN32 big-endian targets, or vxworks little endian?

#if defined(WIN32) || (defined(sgi) && defined(unix) && defined(_MIPSEL)) || (defined(sun) && defined(unix) && !defined(_BIG_ENDIAN)) || (defined(__BYTE_ORDER) && (__BYTE_ORDER == __LITTLE_ENDIAN)) || (defined(__APPLE__) && defined(__LITTLE_ENDIAN__))
#  define FLTLIB_LITTLE_ENDIAN
#elif (defined(sgi) && defined(unix) && defined(_MIPSEB)) || (defined(sun) && defined(unix) && defined(_BIG_ENDIAN)) || defined(vxw) || (defined(__BYTE_ORDER) && (__BYTE_ORDER == __BIG_ENDIAN)) || ( defined(__APPLE__) && defined(__BIG_ENDIAN__))
#  define FLTLIB_BIG_ENDIAN
#else
#  error unknown endian type
#endif

/////////////////////

#ifdef FLTLIB_LITTLE_ENDIAN
#define ENDIAN_16( d ) { uint16 tmp; _ENDIAN_16( &tmp, &(d) ); (d) = tmp; }
#define ENDIAN_32( d ) { uint32 tmp; _ENDIAN_32( &tmp, &(d) ); (d) = tmp; }
#define ENDIAN_32r( d ) { real32 tmp; _ENDIAN_32( &tmp, &(d) ); (d) = tmp; }
#define ENDIAN_64( d ) { real64 tmp; _ENDIAN_64( &tmp, &(d) ); (d) = tmp; }
#define ENDIAN_64r( d ) { real64 tmp; _ENDIAN_64( &tmp, &(d) ); (d) = tmp; }
#else
#define ENDIAN_16( d ) 
#define ENDIAN_32( d )
#define ENDIAN_32r( d )
#define ENDIAN_64( d )
#define ENDIAN_64r( d )
#endif

typedef struct _FltNode {
	uint32 type;
	uint32 length;
	uint32 treeDepth;
	uint32 byteOffset;
	struct _FltNode * parent;
	struct _FltNode * prev, * next;
	struct _FltNode ** child;
	struct _FltNode * attr;
	uint32 numChildren;
} FltNode;

#define FLTRECORD_HEADER 1

typedef struct _FltHeader {
	FltNode node;
	char		ID[8];
	uint32	formatRevision;
	uint32	editRevision;
	char		dateTime[32];
	uint16	nextGroupNodeID;
	uint16	nextLODNodeID;
	uint16	nextObjectNodeID;
	uint16	nextFaceNodeID;
	uint16	unitMultiplier;
	uint8		coordUnits;
	uint8		setTexWhiteOnNewFaces;
	uint32	flags;
	uint32	reserved0[6];
	uint32	projectionType;
	uint32	reserved1[7];
	uint16	nextDOFNodeID;
	uint16	vertexStorageLength;
	uint32	databaseOrigin;
	real64	swDBCoordX;
	real64	swDBCoordY;
	real64	deltaXPlace;
	real64	deltaYPlace;
	uint16	nextSoundNodeID;
	uint16	nextPathNodeID;
	uint32	reserved2[2];
	uint16	nextClipNodeID;
	uint16	nextTextNodeID;
	uint16	nextBSPNodeID;
	uint16	nextSwitchNodeID;
	uint32	reserved3;
	real64	swDBLatitude;
	real64	swDBLongitude;
	real64	neDBLatitude;
	real64	neDBLongitude;
	real64	originDBLatitude;
	real64	originDBLongitude;
	real64	lambertUpperLatitude;
	real64	lambertLowerLatitude;
	uint16	nextLightSourceNodeID;
	uint16	nextLightPointNodeID;
	uint16	nextRoadNodeID;
	uint16	nextCATNodeID;
	uint16	reserved4;
	uint16	reserved5;
	uint16	reserved6;
	uint16	reserved7;
	uint32	earthEllipsoidModel;
	uint16	nextAdaptiveNodeID;
	uint16	nextCurveNodeID;
	uint16	reserved8;
	real64	deltaZPlace;
	real64	dbRadius;
	uint16	nextMeshNodeID;
	uint16	reserved9;
} FltHeader;

typedef enum {
	FVHAS_NORMAL =	(1<<0),
	FVHAS_COLOR = 	(1<<1),
	FVHAS_TEXTURE = (1<<2)
} FltVertexFlags;

#define FVSTART_HARD_EDGE (1<<(31-0))
#define FVNORMAL_FROZEN		(1<<(31-1))
#define FVNO_COLOR				(1<<(31-2))
#define FVPACKED_COLOR		(1<<(31-3))

#define FLTRECORD_VERTEXC 68			// with color
#define FLTRECORD_VERTEXCN 69			// with color + normal
#define FLTRECORD_VERTEXCNUV 70		// with color + normal + uv tex
#define FLTRECORD_VERTEXCUV 71		// with color + uv tex

typedef struct _FltVertex {
	FltNode node;
	uint16 colorNameIndex;
	uint16 flags;
	real64 x, y, z;
	real32 i, j, k;
	real32 u, v;
	uint32 packedColor;
	uint32 colorIndex;
	FltVertexFlags localFlags;
	uint32 indx;
} FltVertex;

#define FLTRECORD_SWITCH 96

typedef struct _FltSwitch {
	FltNode node;
	char ID[8];
	uint32 reserved0;
	uint32 currentMask;
	uint32 numUInt32sPerMask;
	uint32 numMasks;
	uint32 * masks; // = numUInt32sPerMask * numMasks * 4
} FltSwitch;

#define FLTRECORD_COMMENT 31
#define FLTRECORD_COMMENT_MAX_LEN 256

typedef struct _FltComment {
	FltNode node;
	char text[FLTRECORD_COMMENT_MAX_LEN];
} FltComment;

#define FLTRECORD_LONGID 33
#define FLTRECORD_LONGID_MAX_LEN 64

typedef struct _FltLongID {
	FltNode node;
	char text[FLTRECORD_LONGID_MAX_LEN];
} FltLongID;

#define FLTRECORD_DOF 14

#define FLTDOF_LIMITX	(1<<(31-0))
#define FLTDOF_LIMITY	(1<<(31-1))
#define FLTDOF_LIMITZ	(1<<(31-2))
#define FLTDOF_LIMITPITCH	(1<<(31-3))
#define FLTDOF_LIMITROLL	(1<<(31-4))
#define FLTDOF_LIMITYAW	(1<<(31-5))
#define FLTDOF_LIMITSCALEX	(1<<(31-6))
#define FLTDOF_LIMITSCALEY	(1<<(31-7))
#define FLTDOF_LIMITSCALEZ	(1<<(31-8))

typedef struct _FltDOF {
	FltNode node;
	char ID[8];
	uint32 reserved0;

	real64 localOriginX;
	real64 localOriginY;
	real64 localOriginZ;

	real64 localPointX;
	real64 localPointY;
	real64 localPointZ;

	real64 localPlanePointX;
	real64 localPlanePointY;
	real64 localPlanePointZ;

	real64 localMinZ;
	real64 localMaxZ;
	real64 localCurZ;
	real64 localIncZ;

	real64 localMinY;
	real64 localMaxY;
	real64 localCurY;
	real64 localIncY;

	real64 localMinX;
	real64 localMaxX;
	real64 localCurX;
	real64 localIncX;

	real64 localMinPitch;		// about X
	real64 localMaxPitch;
	real64 localCurPitch;
	real64 localIncPitch;

	real64 localMinRoll;		// about Y
	real64 localMaxRoll;
	real64 localCurRoll;
	real64 localIncRoll;

	real64 localMinYaw;		// about Z
	real64 localMaxYaw;
	real64 localCurYaw;
	real64 localIncYaw;

	real64 localMinScaleZ;
	real64 localMaxScaleZ;
	real64 localCurScaleZ;
	real64 localIncScaleZ;

	real64 localMinScaleY;
	real64 localMaxScaleY;
	real64 localCurScaleY;
	real64 localIncScaleY;

	real64 localMinScaleX;
	real64 localMaxScaleX;
	real64 localCurScaleX;
	real64 localIncScaleX;

	uint32 flags;
} FltDOF;

#define FLTLOD_USEPREVSLANT (1<<(31-0))
#define FLTLOD_FREEZECENTER (1<<(31-2))

#define FLTRECORD_LOD 73

typedef struct _FltLOD {
	FltNode node;
	char ID[8];
	uint32 reserved0;
	real64 switchInDistance;
	real64 switchOutDistance;
	uint16 specialEffectID1;
	uint16 specialEffectID2;
	uint32 flags;
	real64 centerX;
	real64 centerY;
	real64 centerZ;
	real64 transitionRange;
} FltLOD;

#define FLTRECORD_VERTEXPALETTE 67

typedef struct _FltVertexPalette {
	FltNode node;
	uint32 numVerts;
	FltVertex ** verts;
} FltVertexPalette;

#define FLTRECORD_VERTEXLIST 72

typedef struct _FltVertexList {
	FltNode node;
	uint32 numVerts;
	FltVertex ** list;
	uint32 *indexList;
} FltVertexList;

#define FLTGROUP_FORWARD_ANIM (1<<(31-1))
#define FLTGROUP_SWING_ANIM (1<<(31-2))
#define FLTGROUP_BOUNDINGBOX_FOLLOWS (1<<(31-3))
#define FLTGROUP_FREEZE_BOUNDINGBOX (1<<(31-4))
#define FLTGROUP_DEFAULT_PARENT (1<<(31-5))

#define FLTRECORD_GROUP 2

typedef struct _FltGroup {
	FltNode node;
	char ID[8];
	uint16 relativePriority;
	uint16 reserved0;
	uint32 flags;
	uint16 specialEffectID1;
	uint16 specialEffectID2;
	uint16 significance;
	uint8  layerCode;
	uint8  reserved1;
	uint32 reserved2;
} FltGroup;

#define FLTOBJECT_NODISPLAYDAY (1<<(31-0))
#define FLTOBJECT_NODISPLAYDUSK (1<<(31-1))
#define FLTOBJECT_NODISPLAYNIGHT (1<<(31-2))
#define FLTOBJECT_NOILLUMINATION (1<<(31-3))
#define FLTOBJECT_FLATSHADED (1<<(31-4))
#define FLTOBJECT_GROUPSSHADOWOBJECT (1<<(31-5))

#define FLTRECORD_OBJECT 4

typedef struct _FltObject {
	FltNode node;
	char ID[8];
	uint32 flags;
	uint16 relativePriority;
	uint16 transparency;
	uint16 specialEffectID1;
	uint16 specialEffectID2;
	uint16 significance;
	uint16 reserved0;
} FltObject;

#define FLTFACEDT_DRAWSOLIDBACKFACE 0
#define FLTFACEDT_DRAWSOLIDNOBACKFACE 1
#define FLTFACEDT_DRAWWIREFRAME 2
#define FLTFACEDT_DRAWWIREFRAMECLOSE 3
#define FLTFACEDT_SURROUNDWIREALTCOLOR 4
#define FLTFACEDT_OMNIDIRLIGHT 8
#define FLTFACEDT_UNIDIRLIGHT 9
#define FLTFACEDT_BIDIRLIGHT 10

#define FLTFACEBB_FIXEDNOALPHA 0
#define FLTFACEBB_FIXEDALPHA 1
#define FLTFACEBB_AXIALROTATE 2
#define FLTFACEBB_POINTROTATE 3

#define FLTFACEMF_TERRAIN (1<<(31-0))
#define FLTFACEMF_NOCOLOR (1<<(31-1))
#define FLTFACEMF_NOALTCOLOR (1<<(31-2))
#define FLTFACEMF_PACKEDCOLOR (1<<(31-3))
#define FLTFACEMF_TERRAINCULTURECUTOUT (1<<(31-4))
#define FLTFACEMF_HIDDEN (1<<(31-5))

#define FLTFACELM_FCNOTILLUM 0
#define FLTFACELM_VCNOTILLUM 1
#define FLTFACELM_FCVN 2
#define FLTFACELM_VCVN 3

#define FLTRECORD_FACE 5

typedef struct _FltFace {
	FltNode node;
	char ID[8];
	uint32 irColorCode;
	uint16 relativePriority;
	uint8  drawType;
	uint8  textureWhite;
	uint16 colorNameIndex;
	uint16 alternateColorNameIndex;
	uint8  reserved0;
	uint8  billboardFlags;
	int16  detailTexturePatternIndex;
	int16  texturePatternIndex;
	int16  materialIndex;
	uint16 surfaceMaterialCode;
	uint16 featureID;
	uint32 irMaterialCode;
	uint16 transparency;
	uint8  LODGenerationControl;
	uint8  lineStyleIndex;
	uint32 miscFlags;
	uint8  lightMode;
	uint8  reserved1;
	uint16 reserved2;
	uint32 reserved3;
	uint32 packedColorPrimary;
	uint32 packedColorAlternate;
	uint16 textureMappingIndex;
	uint16 reserved4;
	uint32 primaryColorIndex;
	uint32 alternateColorIndex;
	uint16 reserved5;
	uint16 reserved6;
} FltFace;

#define FLTRECORD_TEXTURE 64

typedef struct _FltTexture {
	FltNode node;
	char ID[200];
	uint32 index;
	uint32 xloc;
	uint32 yloc;
} FltTexture;

#define FLTRECORD_COLORPALETTE 32

typedef struct _FltColorPalette {
	FltNode node;
	uint32 color[1024];
} FltColorPalette;

#define FLTRECORD_MATRIX 49

typedef struct _FltMatrix {
	FltNode node;
	real32 matrix[16]; // single precision, row major
} FltMatrix;

#define FLTRECORD_GENERALMATRIX 49

typedef struct _FltGeneralMatrix {
	FltNode node;
	real32 matrix[16]; // single precision, row major
} FltGeneralMatrix;

#define FLTRECORD_INSTANCEDEFINITION 62

typedef struct _FltInstanceDefinition {
	FltNode node;
	uint16 reserved0;
	uint16 instance;
} FltInstanceDefinition;

#define FLTRECORD_INSTANCEREFERENCE 61

typedef struct _FltInstanceReference {
	FltNode node;
	uint16 reserved0;
	uint16 instance;
} FltInstanceReference;

#define FLTRECORD_TRANSLATE 78

typedef struct _FltTranslate {
	FltNode node;
	uint32 reserved0;
	real64 fromX, fromY, fromZ;
	real64 deltaX, deltaY, deltaZ;
} FltTranslate;

#define FLTRECORD_SCALE 79

typedef struct _FltScale {
	FltNode node;
	uint32 reserved0;
	real64 centerX, centerY, centerZ;
	real64 scaleX, scaleY, scaleZ;
} FltScale;

#define FLTRECORD_ROTATEABOUTPOINT 80

typedef struct _FltRotateAboutPoint {
	FltNode node;
	uint32 reserved0;
	real64 centerX, centerY, centerZ;
	real32 rotI, rotJ, rotK;
	real32 rotAngle;
} FltRotateAboutPoint;

#define FLTRECORD_BSP 55

typedef struct _FltBSP {
	FltNode node;
	char ID[8];
	uint32 reserved0;
	real64 coefA, coefB, coefC, coefD;
} FltBSP;

#define FLTRECORD_MATERIAL 113

typedef struct _FltMaterial {
	FltNode node;
	char ID[12];		// out of order, but matches everything else
	uint32 index;
	uint32 flags;
	real32 ambientRed;
	real32 ambientGreen;
	real32 ambientBlue;
	real32 diffuseRed;
	real32 diffuseGreen;
	real32 diffuseBlue;
	real32 specularRed;
	real32 specularGreen;
	real32 specularBlue;
	real32 emissiveRed;
	real32 emissiveGreen;
	real32 emissiveBlue;
	real32 shininess;		// 0->128.0
	real32 alpha;				// 0->1.0 1.0 = opaque
	uint32 spare;
} FltMaterial;

typedef struct _FltMaterialPalette {
	FltNode node;
	FltMaterial ** material;
	uint32 numMaterials;
} FltMaterialPalette;

#define FLTRECORD_EXTERNALREFERENCE 63

// flags bits
#define FLTEXTREF_COLORPALETTE_OVERRIDE 0
#define FLTEXTREF_MATERIALPALETTE_OVERRIDE 1
#define FLTEXTREF_TEXTUREPALETTE_OVERRIDE 2
#define FLTEXTREF_LINESTYLEPALETTE_OVERRIDE 3
#define FLTEXTREF_SOUNDPALETTE_OVERRIDE 4
#define FLTEXTREF_LIGHTSOURCEPALETTE_OVERRIDE 5

typedef struct _FltExternalReference {
	FltNode node;
	char path[200];		// path to extref filename<node name>
	uint8 reserved0;
	uint8 reserved1;
	uint16 reserved2;
	uint32 flags;
	uint16 reserved3;
} FltExternalReference;

typedef struct _FltBuffer {
	uint8 *buffer;
	uint8 *curPtr;
	uint32 bytesRead;
	uint32 bytesRemaining;
	uint32 readLength; // length of last read to buffer
	uint32 bufferSize;	// current total size of buffer
} FltBuffer;

typedef struct _FltStack {
	void ** stack;
	uint32 stackPtr;
	uint32 stackSize;
} FltStack;

typedef struct {
	char * head, * tail;
	int last;
} MStrtokCtxt;

typedef struct _FltFile {
	FltNode node;
	FILE * stdFile;
	FltHeader * header;
	FltVertexPalette * vertPalette;
	FltColorPalette * colorPalette;
	FltMaterialPalette * materialPalette;
	FltBuffer * buffer;
	FltStack * stack;
	FltNode * lastNode;
	FltNode * lastParent;
	FltNode ** allNodes;
	uint32 numNodes;
	uint32 byteOffset;
	uint32 treeDepth;
	char fileName[MAX_PATHLEN];

	// for path building
	struct {
		char searchPath[MAX_PATHLEN*5];
		char evPath[ MAX_PATHLEN*5 ];
		int first;
		MStrtokCtxt stc;
	} pathInfo;
} FltFile;

// add nodes that are attributes here

#define FLTNODE_ISATTRIBUTE(xn) ( (xn) == FLTRECORD_COMMENT || \
																	(xn) == FLTRECORD_MATRIX  || \
																	(xn) == FLTRECORD_LONGID )

// add nodes that can be a parent node here

#define FLTNODE_CANPARENT(xn) ( (xn) == FLTRECORD_HEADER || \
																(xn) == FLTRECORD_SWITCH || \
																(xn) == FLTRECORD_DOF || \
																(xn) == FLTRECORD_LOD || \
																(xn) == FLTRECORD_GROUP || \
																(xn) == FLTRECORD_OBJECT || \
																(xn) == FLTRECORD_BSP || \
																(xn) == FLTRECORD_INSTANCEDEFINITION || \
																(xn) == FLTRECORD_FACE )

// add nodes that contain an 'ID' or are named.

#define FLTNODE_HASID(xn) ( (xn) == FLTRECORD_HEADER || \
														(xn) == FLTRECORD_SWITCH || \
														(xn) == FLTRECORD_DOF || \
														(xn) == FLTRECORD_LOD || \
														(xn) == FLTRECORD_GROUP || \
														(xn) == FLTRECORD_OBJECT || \
														(xn) == FLTRECORD_MATERIAL || \
														(xn) == FLTRECORD_BSP || \
														(xn) == FLTRECORD_FACE )

#define FLT_GETPARENT(f) ( FltStackGetCurrent(f->stack) )

#define FLTRECORDFUNC_ARGLIST FltFile * flt, uint16 recLength
#define FLTPOSTPROCESSFUNC_ARGLIST FltFile * flt, FltNode * node
#define FLTRECORDUSERCALLBACK_ARGLIST FltFile *flt, void *recData, void *userData

typedef void * (*FltRecordFunc)( FLTRECORDFUNC_ARGLIST );
typedef void * (*FltPostProcessFunc)( FLTPOSTPROCESSFUNC_ARGLIST );
typedef void * (*FltRecordUserCallback)( FLTRECORDUSERCALLBACK_ARGLIST );

typedef struct _FltRecordEntryName {
	char * type;
	char * name;
	uint32 offset;
} FltRecordEntryName;

typedef struct _FltRecord {
	uint16 opcode;
	char * name;
	FltRecordFunc func;
	FltRecordEntryName * entry;
	FltPostProcessFunc postProcessFunc;
	FltRecordUserCallback userFunc;
	void * userData;
} FltRecord;

typedef enum {
	FLTATTR_MIN_POINT = 0,
	FLTATTR_MIN_BILINEAR = 1,
	FLTATTR_MIN_MIPMAP = 2, // obsolete
	FLTATTR_MIN_MIPMAP_POINT = 3,
	FLTATTR_MIN_MIPMAP_LINEAR = 4,
	FLTATTR_MIN_MIPMAP_BILINEAR = 5,
	FLTATTR_MIN_MIPMAP_TRILINEAR = 6,
	FLTATTR_MIN_NONE = 7
	// todo: add others
} FltAttrMinification;

typedef enum {
	FLTATTR_MAG_POINT = 0,
	FLTATTR_MAG_BILINEAR = 1,
	FLTATTR_MAG_NONE = 2,
	FLTATTR_MAG_BICUBIC = 3,
	FLTATTR_MAG_SHARPEN = 4,
	FLTATTR_MAG_ADD_DETAIL = 5
	// todo: add others
} FltAttrMagnification;

typedef enum {
	FLTATTR_REP_REPEAT = 0,
	FLTATTR_REP_CLAMP = 1
} FltAttrRepetition;

typedef enum {
	FLTATTR_ENV_MODULATE = 0,
	FLTATTR_ENV_BLEND = 1,
	FLTATTR_ENV_DECAL = 2,
	FLTATTR_ENV_COLOR = 3
} FltAttrEnvironment;

typedef struct _FltTxAttributes {
	uint32 uTexels;
	uint32 vTexels;
	uint32 realWorldU;
	uint32 realWorldV;
	uint32 upVectorX;
	uint32 upVectorY;
	uint32 fileFormat;
	uint32 minificationFilter;
	uint32 magnificationFilter;
	uint32 repetitionType;
	uint32 repetitionU;
	uint32 repetitionV;
	uint32 modifyFlag;
	uint32 xPivot;
	uint32 yPivot;
	uint32 environmentType;
	// todo: add others
} FltTxAttributes;

// public
extern FltFile * fltOpen( const char * name );
extern void fltClose( FltFile *theFile );
extern void fltFileFree( FltFile *theFile );
extern int fltParse( FltFile *theFile, uint32 skip );
extern int fltRegisterRecordUserCallback( uint16 record, FltRecordUserCallback, 
																														void * userData );
extern FltRecord * fltRecordGetDefinition( uint16 record );
extern char * fltNodeName( FltNode * node );
extern char * fltSafeNodeName( FltNode * node );
extern void fltDumpNode( FltNode * node );
extern FltType fltStringToType( char * string );
extern void fltRecordEntryNameToString( char * string, FltNode * node, 
																										FltRecordEntryName * e );
extern uint32 fltLookupColor( FltFile * flt, uint32 colorIndex, real32 *r, 
																						real32 *g, real32 *b, real32 *a );
extern FltTexture * fltLookupTexture( FltFile * flt, uint16 indx );
extern FltMaterial * fltLookupMaterial( FltFile * flt, uint16 indx );
extern FltInstanceDefinition * fltLookupInstance( FltFile * flt, uint16 indx );
extern FltTxAttributes * fltLoadAttributes( const char * file );
extern void fltFreeAttributes( FltTxAttributes * );
extern int fltFindFile ( FltFile *, char * inFileName, char *outFileName );
extern void fltAddSearchPath( FltFile * flt, char * path );

// private
extern int fltReadRecordAttr( FltFile * flt, uint16 * type, uint16 * length );
extern void fltSkipRecord( FltFile * flt, uint16 length );

extern void * FltBufferResize( FltBuffer *, uint32 size );
extern void FltBufferRewind( FltBuffer * );
extern void FltBufferResetWithLength( FltBuffer *, uint32 len );
extern FltBuffer * FltBufferAlloc( FltFile * );
extern void FltBufferFree( FltFile * flt );

extern void * FltStackAlloc( FltFile * );
extern void FltStackFree( FltFile * flt );
extern void FltStackPush( FltStack *, void * );
extern void * FltStackPop( FltStack * );
extern void * FltStackGetCurrent( FltStack * );

extern int fltReadBlock( FltFile * flt, void * data, uint32 length );
extern int fltSkipBlock( FltFile * flt, uint32 len );
extern uint32 fltReadUInt32( FltFile * flt );
extern int32 fltReadInt32( FltFile * flt );
extern uint16 fltReadUInt16( FltFile * flt );
extern int16 fltReadInt16( FltFile * flt );
extern uint8 fltReadUInt8( FltFile * flt );
extern int8 fltReadInt8( FltFile * flt );
extern real32 fltReadReal32( FltFile * flt );
extern real64 fltReadReal64( FltFile * flt );

extern int fltFinite64( real64 value );

// MISC macros (DANGER -- MAY ASSUME SIZEOF(POINTER) == 32 bits!

#ifndef offsetof
#define offsetof( structPointer, element ) \
													(uint32)&( ((structPointer)((void *)0))->element )
#endif

#define elementat( structPointer, theStructPtr, elemType, element ) \
		*(elemType *)( (char *)theStructPtr + \
										(uint32)offsetof( structPointer, element ) )
																											

#ifdef __cplusplus
}
#endif

#endif /* _FLT_H_ */
