
/*****************************************************************************
 * FLT: An OpenFlight file loader
 *
 * Copyright (C) 2001  Michael M. Morrison   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * See the file EXCLUSIONS for further details.
 *
 *****************************************************************************/

#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "flt.h"
#include "recnames.h"

static int debug = 0;

FltNode *
fltNewNode( FltFile * flt, uint32 size )
{
	FltNode * node;

	if( (node = calloc( 1, size )) ) {

		flt->allNodes = realloc( flt->allNodes, (1+flt->numNodes) * 
																									sizeof( FltNode * ) );
		if( flt->allNodes ) {
			flt->allNodes[ flt->numNodes ] = node;
			flt->numNodes++;

			return node;
		}
	}

	// need something better here.
	printf("Unable to malloc!\n");
	exit(0);

	// keep compiler quiet
	return 0;
}

char *
fltNodeName( FltNode * node )
{
	// check for nodes with id's
	if( FLTNODE_HASID( node->type ) ) {
		FltRecord * rec;
		FltNode * findNode;

		// first, check for "longid" field
		findNode = node->attr;

		while( findNode ) {
			if( findNode->type == FLTRECORD_LONGID ) {
				FltLongID * lid = (FltLongID *)findNode;
				return lid->text;
			}

			findNode = findNode->next;
		}

		// if no longid, return regular ID field
		if( (rec = fltRecordGetDefinition( (uint16)node->type )) ) {
			int x = 0;

			while( rec->entry[x].type ) {
				if( !strcmp( rec->entry[x].name, "ID" ) ) {
					return (char *)( (char *)node + rec->entry[x].offset );
				}
				x ++;
			}

		}

		return NULL;

	} else
		return NULL;
}

char *
fltSafeNodeName( FltNode * node )
{
	static char unnamed[] = {"Unnamed"};
	static char null[] = {"Null"};
	char * ret;

	if( node ) {
		ret = fltNodeName( node );

		if( ret )
			return ret;
		else
			return unnamed;
	} else
		return null;
}

void
fltNodeAddChild( FltNode * parent, FltNode * child )
{
	assert( parent );

	parent->child = realloc( parent->child, (1+parent->numChildren) *
																									sizeof( FltNode * ) );
	parent->child[ parent->numChildren ] = child;
	parent->numChildren++;
}

static void *
fltPushLevel( FLTRECORDFUNC_ARGLIST )
{
	assert( flt );

	FltStackPush( flt->stack, flt->lastParent );

	if( debug )
		printf("Pushing: '%s' is parent.\n", fltSafeNodeName( FLT_GETPARENT(flt) ));

	flt->lastNode = NULL;

	flt->treeDepth++;

	return NULL;
}

static void *
fltPopLevel( FLTRECORDFUNC_ARGLIST )
{
	assert( flt );

	// pop off last node
	// we set flt->lastNode here, so that the next node has this node
	// as its neighbor.

	flt->lastNode = FltStackPop( flt->stack );

	if( debug ) {
		if( FLT_GETPARENT(flt) == NULL )
			printf("Popping: At top level.\n" );
		else
			printf("Popping: '%s' is new parent.\t", 
															fltSafeNodeName( FLT_GETPARENT(flt) ) );

		printf("lastNode: %s\n", fltSafeNodeName( flt->lastNode ) );
	}

	flt->treeDepth--;

	return NULL;
}

static void *
fltHeader( FLTRECORDFUNC_ARGLIST )
{
	char path[MAX_PATHLEN];
	char * ptr;

	assert( flt );
	assert( flt->stdFile );

	flt->header = (FltHeader *) fltNewNode( flt, sizeof( FltHeader ) );

	fltReadBlock( flt, flt->header->ID, 8 );

	// byte swap important fields

	flt->header->formatRevision = fltReadUInt32( flt );
	flt->header->editRevision   = fltReadUInt32( flt );

	fltReadBlock( flt, flt->header->dateTime, 32 );

	if( debug )
		printf("FLT: %s: is a version %d flt file.\n", flt->fileName, 
																								flt->header->formatRevision );

	// automagically add the path to this file
	strcpy( path, flt->fileName );
	if( (ptr = strrchr( path, '/' )) ) 
		*ptr = 0;
	else if( (ptr = strrchr( path, '\\' )) )
		*ptr = 0;
	else
		path[0] = 0;
	
	fltAddSearchPath( flt, path ); 

	return flt->header;
}

static void *
fltExternalReference( FLTRECORDFUNC_ARGLIST )
{
	FltExternalReference * ext;

	assert( flt );
	assert( flt->stdFile );

	ext = (FltExternalReference *) fltNewNode( flt, sizeof(FltExternalReference));

	fltReadBlock( flt, ext->path, 200 );

	ext->reserved0 = fltReadUInt8( flt );
	ext->reserved1 = fltReadUInt8( flt );
	ext->reserved2 = fltReadUInt16( flt );

	ext->flags = fltReadUInt32( flt );

	ext->reserved3 = fltReadUInt16( flt );

	return ext;
}

static void *
fltMatrix( FLTRECORDFUNC_ARGLIST )
{
	FltMatrix * mat;
	int i;

	assert( flt );
	assert( flt->stdFile );

	mat = (FltMatrix *) fltNewNode( flt, sizeof(FltMatrix));

	// read matrix
	for(i=0;i<16;i++)
		mat->matrix[i] = fltReadReal32( flt );

	return mat;
}

static void *
fltExternalReferencePostProcess( FLTPOSTPROCESSFUNC_ARGLIST )
{
	FltExternalReference * extref = (FltExternalReference *)node;

	assert( flt );

	// go find the files, and reference (later)
	(void)(extref);

	return 0;
}

static void *
fltVertexPalette( FLTRECORDFUNC_ARGLIST )
{
	assert( flt );
	assert( flt->stdFile );

	flt->vertPalette = (FltVertexPalette *) 
																fltNewNode( flt, sizeof( FltVertexPalette ) );

	// may be needed?
	//len = fltReadUInt32( flt );

	(void)fltReadUInt32( flt );

	flt->vertPalette->numVerts = 0;

	return flt->vertPalette;
}

static void *
fltColorPalette( FLTRECORDFUNC_ARGLIST )
{
	uint32 i;

	assert( flt );
	assert( flt->stdFile );

	flt->colorPalette = (FltColorPalette *) 
																fltNewNode( flt, sizeof( FltColorPalette ) );

	// skip reserved block 
	fltSkipBlock( flt, 128 );

	for(i=0;i<1024;i++)
		flt->colorPalette->color[i] = fltReadUInt32( flt );

	return flt->colorPalette;
}

static void *
fltVertexPalettePostProcess( FLTPOSTPROCESSFUNC_ARGLIST )
{
	uint32 i;

	assert( flt );

	// adjust byte-offset of the record to anchor from VertexPalette, rather
	// than start of file

	for(i=0; i<flt->vertPalette->numVerts; i++) {
		FltVertex *v = flt->vertPalette->verts[i];

		v->node.byteOffset -= flt->vertPalette->node.byteOffset;
	}

	return 0;
}

static void *
fltSwitch( FLTRECORDFUNC_ARGLIST )
{
	uint32 i;
	FltSwitch * sw;

	assert( flt );
	assert( flt->stdFile );

	sw = (FltSwitch *) fltNewNode( flt, sizeof( FltSwitch ) );

	fltReadBlock( flt, sw->ID, 8 );

	sw->reserved0 = fltReadUInt32( flt );
	sw->currentMask = fltReadUInt32( flt );
	//
	// NOTE: The docs must be wrong.  They state that the order is
	// num32bitInts/mask first, then numMasks.
	// Also: The bit pattern appears to be inverted as well.
	//
	sw->numMasks = fltReadUInt32( flt );
	sw->numUInt32sPerMask = fltReadUInt32( flt );

	sw->masks = (uint32 *)calloc( 1, sw->numUInt32sPerMask * 
																				sw->numMasks * sizeof(uint32) );

//	printf("Switch: '%s' - numMasks: %d (%d/mask)\n", sw->ID, sw->numMasks,
//																										sw->numUInt32sPerMask);

	for(i=0;i<sw->numMasks*sw->numUInt32sPerMask;i++) {
		sw->masks[i] = fltReadUInt32( flt );
		//printf("\t%d = %08x\n", i, sw->masks[i] );
	}

	return sw;
}

static void *
fltComment( FLTRECORDFUNC_ARGLIST )
{
	FltComment * comment;

	assert( flt );
	assert( flt->stdFile );

	comment = (FltComment *) fltNewNode( flt, sizeof( FltComment ) );

	fltReadBlock( flt, comment->text, 
				(recLength > FLTRECORD_COMMENT_MAX_LEN) ? FLTRECORD_COMMENT_MAX_LEN-1 :
						recLength );
	// null term
	comment->text[FLTRECORD_COMMENT_MAX_LEN-1] = 0;

	return comment;
}

static void *
fltLongID( FLTRECORDFUNC_ARGLIST )
{
	FltLongID * longid;

	assert( flt );
	assert( flt->stdFile );

	longid = (FltLongID *) fltNewNode( flt, sizeof( FltLongID ) );

	fltReadBlock( flt, longid->text, 
				(recLength > FLTRECORD_LONGID_MAX_LEN) ? FLTRECORD_LONGID_MAX_LEN-1 :
						recLength );
	// null term
	longid->text[FLTRECORD_LONGID_MAX_LEN-1] = 0;

	return longid;
}

static void *
fltLOD( FLTRECORDFUNC_ARGLIST )
{
	FltLOD * lod;

	assert( flt );
	assert( flt->stdFile );

	lod = (FltLOD *) fltNewNode( flt, sizeof( FltLOD ) );

	fltReadBlock( flt, lod->ID, 8 );

	lod->reserved0 = fltReadUInt32( flt );
	lod->switchInDistance = fltReadReal64( flt );
	lod->switchOutDistance = fltReadReal64( flt );
	lod->specialEffectID1 = fltReadUInt16( flt );
	lod->specialEffectID2 = fltReadUInt16( flt );
	lod->flags = fltReadUInt32( flt );
	lod->centerX = fltReadReal64( flt );
	lod->centerY = fltReadReal64( flt );
	lod->centerZ = fltReadReal64( flt );
	lod->transitionRange = fltReadReal64( flt );

	return lod;
}

void
fltAddVertexToPalette( FltVertexPalette * palette, FltVertex * vert )
{
	assert( palette );
	
	palette->verts = realloc( palette->verts, (palette->numVerts+1) * 
																									sizeof( FltVertex * ) );

	palette->verts[ palette->numVerts ] = vert;
	vert->indx = palette->numVerts;

	palette->numVerts++;
}

static void *
fltVertexWithColor( FLTRECORDFUNC_ARGLIST )
{
	FltVertex * vert;

	assert( flt );
	assert( flt->stdFile );

	vert = (FltVertex *) fltNewNode( flt, sizeof( FltVertex ) );

	vert->colorNameIndex = fltReadUInt16( flt );
	vert->flags = fltReadUInt16( flt );

	vert->x = fltReadReal64( flt );
	vert->y = fltReadReal64( flt );
	vert->z = fltReadReal64( flt );

	vert->packedColor = fltReadUInt32( flt );
	vert->colorIndex  = fltReadUInt32( flt );

	vert->localFlags = FVHAS_COLOR;

	fltAddVertexToPalette( flt->vertPalette, vert );

	return vert;
}

static void *
fltVertexWithColorNormal( FLTRECORDFUNC_ARGLIST )
{
	FltVertex * vert;

	assert( flt );
	assert( flt->stdFile );

	vert = (FltVertex *) fltNewNode( flt, sizeof( FltVertex ) );

	vert->colorNameIndex = fltReadUInt16( flt );
	vert->flags = fltReadUInt16( flt );

	vert->x = fltReadReal64( flt );
	vert->y = fltReadReal64( flt );
	vert->z = fltReadReal64( flt );

	vert->i = fltReadReal32( flt );
	vert->j = fltReadReal32( flt );
	vert->k = fltReadReal32( flt );

	vert->packedColor = fltReadUInt32( flt );
	vert->colorIndex  = fltReadUInt32( flt );

	vert->localFlags = FVHAS_COLOR | FVHAS_NORMAL;

	fltAddVertexToPalette( flt->vertPalette, vert );

	return vert;
}

static void *
fltVertexWithColorTexture( FLTRECORDFUNC_ARGLIST )
{
	FltVertex * vert;

	assert( flt );
	assert( flt->stdFile );

	vert = (FltVertex *) fltNewNode( flt, sizeof( FltVertex ) );

	vert->colorNameIndex = fltReadUInt16( flt );
	vert->flags = fltReadUInt16( flt );

	vert->x = fltReadReal64( flt );
	vert->y = fltReadReal64( flt );
	vert->z = fltReadReal64( flt );

	vert->u = fltReadReal32( flt );
	vert->v = fltReadReal32( flt );

//	printf("CTUV: %f,%f\n", vert->u, vert->v );

	vert->packedColor = fltReadUInt32( flt );
	vert->colorIndex  = fltReadUInt32( flt );

	vert->localFlags = FVHAS_COLOR | FVHAS_TEXTURE;

	fltAddVertexToPalette( flt->vertPalette, vert );

	return vert;
}

static void *
fltVertexWithColorNormalTexture( FLTRECORDFUNC_ARGLIST )
{
	FltVertex * vert;

	assert( flt );
	assert( flt->stdFile );

	vert = (FltVertex *) fltNewNode( flt, sizeof( FltVertex ) );

	vert->colorNameIndex = fltReadUInt16( flt );
	vert->flags = fltReadUInt16( flt );

	vert->x = fltReadReal64( flt );
	vert->y = fltReadReal64( flt );
	vert->z = fltReadReal64( flt );

	vert->i = fltReadReal32( flt );
	vert->j = fltReadReal32( flt );
	vert->k = fltReadReal32( flt );

//	printf("CNTN: %f,%f,%f\n", vert->i, vert->j, vert->k );

	vert->u = fltReadReal32( flt );
	vert->v = fltReadReal32( flt );

//	printf("CNTUV: %f,%f\n", vert->u, vert->v );

	vert->packedColor = fltReadUInt32( flt );
	vert->colorIndex  = fltReadUInt32( flt );

	vert->localFlags = FVHAS_COLOR | FVHAS_TEXTURE | FVHAS_NORMAL;

	fltAddVertexToPalette( flt->vertPalette, vert );

	return vert;
}

static void *
fltVertexList( FLTRECORDFUNC_ARGLIST )
{
	FltVertexList * list;
	uint32 numVerts = recLength/4;
	uint32 i;

	assert( flt );
	assert( flt->stdFile );

	list = (FltVertexList *) fltNewNode( flt, sizeof( FltVertexList ) );

	list->numVerts = numVerts;
	list->list = (FltVertex **) calloc( 1, (sizeof(FltVertex *) * numVerts) );
	list->indexList = (uint32 *) calloc( 1, (sizeof(uint32 *) * numVerts) );

	// read in the ints, we later convert them to pointers
	for(i=0;i<numVerts;i++)
		list->list[i] = (FltVertex *)fltReadUInt32( flt );

	return list;
}

static int
vertFunc( const void * xkey, const void * member )
{
	FltVertex * v = *((FltVertex **)member);
	uint32 off = *(uint32 *)xkey;

	if( off > v->node.byteOffset )
		return 1;
	else if( off < v->node.byteOffset )
		return -1;
	else
		return 0;
}

static void *
findVert( FltFile * flt, uint32 *offset )
{
	FltVertex ** vert;

	vert = (FltVertex **) bsearch( offset,
																flt->vertPalette->verts,
																flt->vertPalette->numVerts,
																sizeof( FltVertex * ),
																vertFunc );

	if( vert ) {
		*offset = (*vert)->indx;
		return *vert;
	} else	
		return 0;
}

static void *
fltVertexListPostProcess( FLTPOSTPROCESSFUNC_ARGLIST )
{
	uint32 i, offset;
	FltVertex * idx;
	FltVertexList * list = (FltVertexList *)node;

	assert( flt );

	// connect all "byte offsets" from vertex palette into this list
	// we assume that the PostProcessFunc from vertexPalette has already
	// been executed.

	for(i=0;i<list->numVerts;i++) {
		offset = (uint32)list->list[i];
		idx = findVert( flt, &offset );
		if( idx ) {
			if( debug )
				printf("FLT: vertex record at offset: %p found (%p)\n", list->list[i],
																																	idx );
			list->list[i] = idx;
			list->indexList[i] = offset;
		} else 
			printf("FLT: vertex record at offset: %p not found!\n", list->list[i] );
	}

	return 0;
}


static void *
fltGroup( FLTRECORDFUNC_ARGLIST )
{
	FltGroup * group;

	assert( flt );
	assert( flt->stdFile );

	group = (FltGroup *) fltNewNode( flt, sizeof( FltGroup ) );

	fltReadBlock( flt, group->ID, 8 * sizeof(char) );

	group->relativePriority = fltReadUInt16( flt );
	group->reserved0 = fltReadUInt16( flt );
	group->flags = fltReadUInt32( flt );
	group->specialEffectID1 = fltReadUInt16( flt );
	group->specialEffectID2 = fltReadUInt16( flt );
	group->significance = fltReadUInt16( flt );
	group->layerCode = fltReadUInt8( flt );
	group->reserved1 = fltReadUInt8( flt );
	group->reserved2 = fltReadUInt32( flt );

	return group;
}

static void *
fltInstanceDefinition( FLTRECORDFUNC_ARGLIST )
{
	FltInstanceDefinition * instDef;

	assert( flt );
	assert( flt->stdFile );

	instDef = (FltInstanceDefinition *) 
													fltNewNode( flt, sizeof( FltInstanceDefinition ) );

	instDef->reserved0 = fltReadUInt16( flt );
	instDef->instance  = fltReadUInt16( flt );

	return instDef;
}

static void *
fltInstanceReference( FLTRECORDFUNC_ARGLIST )
{
	FltInstanceReference * instRef;

	assert( flt );
	assert( flt->stdFile );

	instRef = (FltInstanceReference *) 
													fltNewNode( flt, sizeof( FltInstanceReference ) );

	instRef->reserved0 = fltReadUInt16( flt );
	instRef->instance  = fltReadUInt16( flt );

	return instRef;
}

static void *
fltDOF( FLTRECORDFUNC_ARGLIST )
{
	FltDOF * dof;

	assert( flt );
	assert( flt->stdFile );

	dof = (FltDOF *) fltNewNode( flt, sizeof( FltDOF ) );

	fltReadBlock( flt, dof->ID, 8 * sizeof(char) );

	dof->reserved0 = fltReadUInt32( flt );

	dof->localOriginX = fltReadReal64( flt );
	dof->localOriginY = fltReadReal64( flt );
	dof->localOriginZ = fltReadReal64( flt );

	dof->localPointX = fltReadReal64( flt );
	dof->localPointY = fltReadReal64( flt );
	dof->localPointZ = fltReadReal64( flt );
	
	dof->localPlanePointX = fltReadReal64( flt );
	dof->localPlanePointY = fltReadReal64( flt );
	dof->localPlanePointZ = fltReadReal64( flt );
	
	dof->localMinZ = fltReadReal64( flt );
	dof->localMaxZ = fltReadReal64( flt );
	dof->localCurZ = fltReadReal64( flt );
	dof->localIncZ = fltReadReal64( flt );

	dof->localMinY = fltReadReal64( flt );
	dof->localMaxY = fltReadReal64( flt );
	dof->localCurY = fltReadReal64( flt );
	dof->localIncY = fltReadReal64( flt );

	dof->localMinX = fltReadReal64( flt );
	dof->localMaxX = fltReadReal64( flt );
	dof->localCurX = fltReadReal64( flt );
	dof->localIncX = fltReadReal64( flt );

	dof->localMinPitch = fltReadReal64( flt );
	dof->localMaxPitch = fltReadReal64( flt );
	dof->localCurPitch = fltReadReal64( flt );
	dof->localIncPitch = fltReadReal64( flt );

	dof->localMinRoll = fltReadReal64( flt );
	dof->localMaxRoll = fltReadReal64( flt );
	dof->localCurRoll = fltReadReal64( flt );
	dof->localIncRoll = fltReadReal64( flt );

	dof->localMinYaw = fltReadReal64( flt );
	dof->localMaxYaw = fltReadReal64( flt );
	dof->localCurYaw = fltReadReal64( flt );
	dof->localIncYaw = fltReadReal64( flt );

	dof->localMinScaleZ = fltReadReal64( flt );
	dof->localMaxScaleZ = fltReadReal64( flt );
	dof->localCurScaleZ = fltReadReal64( flt );
	dof->localIncScaleZ = fltReadReal64( flt );

	dof->localMinScaleY = fltReadReal64( flt );
	dof->localMaxScaleY = fltReadReal64( flt );
	dof->localCurScaleY = fltReadReal64( flt );
	dof->localIncScaleY = fltReadReal64( flt );

	dof->localMinScaleX = fltReadReal64( flt );
	dof->localMaxScaleX = fltReadReal64( flt );
	dof->localCurScaleX = fltReadReal64( flt );
	dof->localIncScaleX = fltReadReal64( flt );

	dof->flags = fltReadUInt32( flt );

	return dof;
}


static void *
fltObject( FLTRECORDFUNC_ARGLIST )
{
	FltObject * obj;

	assert( flt );
	assert( flt->stdFile );

	obj = (FltObject *) fltNewNode( flt, sizeof( FltObject ) );

	fltReadBlock( flt, obj->ID, 8 * sizeof(char) );

	obj->flags = fltReadUInt32( flt );
	obj->relativePriority = fltReadUInt16( flt );
	obj->transparency = fltReadUInt16( flt );
	obj->specialEffectID1 = fltReadUInt16( flt );
	obj->specialEffectID2 = fltReadUInt16( flt );
	obj->significance = fltReadUInt16( flt );
	obj->reserved0 = fltReadUInt16( flt );

	return obj;
}

void
fltAddMaterialToPalette( FltFile * flt, FltMaterial * mat )
{
	assert( flt );

	if( flt->materialPalette == 0 )
		flt->materialPalette = calloc( 1, sizeof( FltMaterialPalette ) ); 
	
	flt->materialPalette->material = realloc( flt->materialPalette->material, 
															(flt->materialPalette->numMaterials+1) * 
																									sizeof( FltMaterial * ) );

	flt->materialPalette->material[ flt->materialPalette->numMaterials++ ] = mat;
}

static void *
fltMaterial( FLTRECORDFUNC_ARGLIST )
{
	FltMaterial * mat;

	assert( flt );
	assert( flt->stdFile );

	mat = (FltMaterial *) fltNewNode( flt, sizeof( FltMaterial ) );

	mat->index = fltReadUInt32( flt );
	fltReadBlock( flt, mat->ID, 12 * sizeof(char) );
	mat->flags = fltReadUInt32( flt );

	mat->ambientRed = fltReadReal32( flt );
	mat->ambientGreen = fltReadReal32( flt );
	mat->ambientBlue = fltReadReal32( flt );

	mat->diffuseRed = fltReadReal32( flt );
	mat->diffuseGreen = fltReadReal32( flt );
	mat->diffuseBlue = fltReadReal32( flt );

	mat->specularRed = fltReadReal32( flt );
	mat->specularGreen = fltReadReal32( flt );
	mat->specularBlue = fltReadReal32( flt );

	mat->emissiveRed = fltReadReal32( flt );
	mat->emissiveGreen = fltReadReal32( flt );
	mat->emissiveBlue = fltReadReal32( flt );

	mat->shininess = fltReadReal32( flt );
	mat->alpha = fltReadReal32( flt );

	fltAddMaterialToPalette( flt, mat );

	return mat;
}

static void *
fltTexture( FLTRECORDFUNC_ARGLIST )
{
	FltTexture * tex;

	assert( flt );
	assert( flt->stdFile );

	tex = (FltTexture *) fltNewNode( flt, sizeof( FltTexture ) );

	fltReadBlock( flt, tex->ID, 200 * sizeof(char) );

#if 0
	{
		int i;
		printf("TEXTURE: %s\n", tex->ID ); 
		for(i=0;i<200;i++)
			printf("%02x ", tex->ID[i]&0xff );
		printf("\n");
	}
#endif

	tex->index = fltReadUInt32( flt );
	tex->xloc = fltReadUInt32( flt );
	tex->yloc = fltReadUInt32( flt );

	return tex;
}

static void *
fltBSP( FLTRECORDFUNC_ARGLIST )
{
	FltBSP * bsp;

	assert( flt );
	assert( flt->stdFile );

	bsp = (FltBSP *) fltNewNode( flt, sizeof( FltBSP ) );

	fltReadBlock( flt, bsp->ID, 8 * sizeof(char) );

	bsp->reserved0 = fltReadUInt32( flt );
	bsp->coefA = fltReadReal64( flt );
	bsp->coefB = fltReadReal64( flt );
	bsp->coefC = fltReadReal64( flt );
	bsp->coefD = fltReadReal64( flt );

	return bsp;
}

static void *
fltFace( FLTRECORDFUNC_ARGLIST )
{
	FltFace * face;

	assert( flt );
	assert( flt->stdFile );

	face = (FltFace *) fltNewNode( flt, sizeof( FltFace ) );

	fltReadBlock( flt, face->ID, 8 * sizeof(char) );

	face->irColorCode = fltReadUInt32( flt );
	face->relativePriority = fltReadUInt16( flt );
	face->drawType = fltReadUInt8( flt );
	face->textureWhite = fltReadUInt8( flt );
	face->colorNameIndex = fltReadUInt16( flt );
	face->alternateColorNameIndex = fltReadUInt16( flt );
	face->reserved0 = fltReadUInt8( flt );
	face->billboardFlags = fltReadUInt8( flt );
	face->detailTexturePatternIndex = fltReadInt16( flt );
	face->texturePatternIndex = fltReadInt16( flt );
	face->materialIndex = fltReadInt16( flt );
	face->surfaceMaterialCode = fltReadUInt16( flt );
	face->featureID = fltReadUInt16( flt );
	face->irMaterialCode = fltReadUInt32( flt );
	face->transparency = fltReadUInt16( flt );
	face->LODGenerationControl = fltReadUInt8( flt );
	face->lineStyleIndex = fltReadUInt8( flt );
	face->miscFlags = fltReadUInt32( flt );
	face->lightMode = fltReadUInt8( flt );
	face->reserved1 = fltReadUInt8( flt );
	face->reserved2 = fltReadUInt16( flt );
	face->reserved3 = fltReadUInt32( flt );
	face->packedColorPrimary = fltReadUInt32( flt );
	face->packedColorAlternate = fltReadUInt32( flt );
	face->textureMappingIndex = fltReadUInt16( flt );
	face->reserved4 = fltReadUInt16( flt );
	face->primaryColorIndex = fltReadUInt32( flt );
	face->alternateColorIndex = fltReadUInt32( flt );
	face->reserved5 = fltReadUInt16( flt );
	face->reserved6 = fltReadUInt16( flt );

	return face;
}


// record types
static FltRecord FlightRecords[] = {
	{ 1, "Header", fltHeader, FltHeaderEntryName }, 
	{ 2, "Group", fltGroup, FltGroupEntryName },
	{ 4, "Object", fltObject, FltObjectEntryName }, 
	{ 5, "Face", fltFace, FltFaceEntryName }, 
	{ 10, "Push Level", fltPushLevel }, 
	{ 11, "Pop Level", fltPopLevel }, 
	{ 14, "Degree of Freedom", fltDOF, FltDOFEntryName }, 
	{ 19, "Push Subface", NULL }, 
	{ 20, "Pop Subface", NULL }, 
	{ 21, "Push Extension", NULL }, 
	{ 22, "Pop Extension", NULL }, 
	{ 23, "Continuation", NULL }, 
	{ 31, "Text Comment", fltComment, FltCommentEntryName }, 
	{ 32, "Color Palette", fltColorPalette, FltColorPaletteEntryName }, 
	{ 33, "Long ID", fltLongID, FltLongIDEntryName }, 
	{ 49, "Transformation Matrix", fltMatrix, FltMatrixEntryName }, 
	{ 50, "Vector", NULL }, 
	{ 52, "MultiTexture", NULL }, 
	{ 53, "UV List", NULL }, 
	{ 55, "Binary Separating Plane", fltBSP, FltBSPEntryName }, 
	{ 60, "Replicate", NULL }, 
	{ 61, "Instance Reference", fltInstanceReference, 
															FltInstanceReferenceEntryName },
	{ 62, "Instance Definition", fltInstanceDefinition,
															 FltInstanceDefinitionEntryName }, 
	{ 63, "External Reference", fltExternalReference, 
																					FltExternalReferenceEntryName,
																					fltExternalReferencePostProcess }, 
	{ 64, "Texture Reference", fltTexture, FltTextureEntryName }, 
	{ 67, "Vertex Palette", fltVertexPalette, FltVertexPaletteEntryName, 
																									fltVertexPalettePostProcess}, 
	{ 68, "Vertex with Color", fltVertexWithColor, FltVertexEntryName }, 
	{ 69, "Vertex with Color and Normal", fltVertexWithColorNormal, 
																												FltVertexEntryName }, 
	{ 70, "Vertex with Color, Normal and UV", fltVertexWithColorNormalTexture,
																											FltVertexEntryName }, 
	{ 71, "Vertex with Color and UV", fltVertexWithColorTexture,
																											FltVertexEntryName }, 
	{ 72, "Vertex List", fltVertexList, FltVertexListEntryName,
																										fltVertexListPostProcess},
	{ 73, "Level of Detail", fltLOD, FltLODEntryName }, 
	{ 74, "Bounding Box", NULL }, 
	{ 76, "Rotate about Edge", NULL }, 
	{ 78, "Translate", NULL }, 
	{ 79, "Scale (Nonuniform)", NULL }, 
	{ 80, "Rotate about Point", NULL }, 
	{ 81, "Rotate and/or Scale to", NULL }, 
	{ 82, "Put Transform", NULL }, 
	{ 83, "Eyepoint and Trackplane", NULL }, 
	{ 84, "Mesh", NULL }, 
	{ 85, "Local Vertex Pool", NULL }, 
	{ 86, "Mesh Primitive", NULL }, 
	{ 87, "Road Segment", NULL }, 
	{ 88, "Road Zone", NULL }, 
	{ 89, "Morph Vertex List", NULL }, 
	{ 90, "Behavior (Linkage) Palette", NULL }, 
	{ 91, "Sound", NULL },
	{ 92, "Road Path", NULL }, 
	{ 93, "Sound Palette", NULL }, 
	{ 94, "General Matrix", NULL }, 
	{ 95, "Text", NULL },
	{ 96, "Switch", fltSwitch, FltSwitchEntryName }, 
	{ 97, "Line Style", NULL }, 
	{ 98, "Clip Region", NULL }, 
	{ 100, "Extension", NULL }, 
	{ 101, "Light Source", NULL }, 
	{ 102, "Light Source Palette", NULL }, 
	{ 103, "Reserved", NULL }, 
	{ 104, "Reserved", NULL }, 
	{ 105, "Bounding Sphere", NULL }, 
	{ 106, "Bounding Cylinder", NULL }, 
	{ 107, "Reserved", NULL }, 
	{ 108, "Bounding Volume Center", NULL }, 
	{ 109, "Bounding Volume Orientation", NULL }, 
	{ 111, "Light Point", NULL }, 
	{ 112, "Texture Mapping Palette", NULL }, 
	{ 113, "Material Palette", fltMaterial, FltMaterialEntryName }, 
	{ 114, "Color Name Palette", NULL }, 
	{ 115, "Continuously Adaptive", NULL }, 
	{ 116, "CAT Data", NULL }, 
	{ 117, "Reserved", NULL }, 
	{ 118, "Reserved", NULL }, 
	{ 119, "Reserved", NULL },
	{ 120, "Reserved", NULL }, 
	{ 121, "Reserved", NULL }, 
	{ 122, "Push Attribute", NULL }, 
	{ 123, "Pop Attribute", NULL }, 
	{ 124, "Reserved", NULL }, 
	{ 125, "Adaptive Attribute", NULL },
	{ 126, "Curve Node", NULL }
};
#define FLTRECORDS_NUM ( sizeof( FlightRecords ) / sizeof( FltRecord ) )

// obsolete
static FltRecord ObsoleteRecords[] = {
	{ 3, "Level of Detail", NULL }, 
	{ 6, "Vertex with ID", NULL }, 
	{ 7, "ShortVertex", NULL }, 
	{ 8, "Vertex with Color", NULL }, 
	{ 9, "Vertex with Color and Normal", NULL }, 
	{ 12, "Translate", NULL }, 
	{ 13, "Degree of Freedom", NULL }, 
	{ 16, "Instance Reference", NULL }, 
	{ 17, "Instance Definition", NULL }, 
	{ 40, "Translate", NULL }, 
	{ 41, "Rotate about Point", NULL }, 
	{ 42, "Rotate about Edge", NULL }, 
	{ 43, "Scale", NULL }, 
	{ 44, "Translate", NULL }, 
	{ 45, "Scale (Nonuniform)", NULL }, 
	{ 46, "Rotate about Point", NULL }, 
	{ 47, "Rotate and/or Scale to Point", NULL }, 
	{ 48, "Put Transform", NULL }, 
	{ 51, "Bounding Box", NULL }, 
	{ 65, "Eyepoint Palette", NULL }, 
	{ 66, "Material Palette", NULL }, 
	{ 77, "Scale", NULL }, 
	{ 110, "Histogram Bounding Volume", NULL }, 
};
#define OBSOLETERECORDS_NUM ( sizeof( ObsoleteRecords ) / sizeof( FltRecord ) )

static int
compareFunc( const void * xkey, const void * member )
{
	FltRecord * rec = (FltRecord *)member;
	uint16 key = *(uint16 *)xkey;

	if( key > rec->opcode )
		return 1;
	else if( key < rec->opcode )
		return -1;
	else
		return 0;
}

int
fltRegisterRecordUserCallback( uint16 record, FltRecordUserCallback cb, 
																														void * userData )
{
	FltRecord * rec;

	rec = (FltRecord *) bsearch(	&record, 
																FlightRecords, 
																FLTRECORDS_NUM,
																sizeof( FltRecord ),
																compareFunc );

	if( rec ) {
		rec->userFunc = cb;
		rec->userData = userData;
		return 1;
	} else {
		printf("fltRegisterRecordUserCallback: unknown record type: %d\n", record);
		return 0;
	}

}

FltRecord *
fltRecordGetDefinition( uint16 type )
{
	FltRecord * rec;

	rec = (FltRecord *) bsearch(	&type, 
																FlightRecords, 
																FLTRECORDS_NUM,
																sizeof( FltRecord ),
																compareFunc );

	// may be null
	return rec;
}

static void
fltPostProcess( FltFile * flt )
{
	uint32 i;
	FltRecord * rec;

	if( debug )
		printf("Post Processing Begin\n");

	for(i=0;i<flt->numNodes;i++) {
		rec = fltRecordGetDefinition( (uint16)flt->allNodes[i]->type );
		if( rec && rec->postProcessFunc )
			rec->postProcessFunc( flt, flt->allNodes[i] );
	}

	if( debug )
		printf("Post Processing End\n");
}

int
fltParse( FltFile * flt, uint32 skip )
{
	uint16 length, type;
	FltRecord * rec;

	while( fltReadRecordAttr( flt, &type, &length ) ) {

		rec = (FltRecord *) bsearch(	&type, 
																	FlightRecords, 
																	FLTRECORDS_NUM,
																	sizeof( FltRecord ),
																	compareFunc );

		if( rec ) { 
			if( rec->func ) {
				FltNode * node;

				if( debug )
					printf("supported node type: %s len: %d\n", rec->name, length );

				// if we are skipping, then continue
				if( skip ) {
					fltSkipRecord( flt, length );
					continue;
				}

				// call the internal function for this record type
				node = (FltNode *) rec->func( flt, length );

				// set state on the returned node
				if( node ) {
					node->type = type;
					node->length = length;
					node->parent = FLT_GETPARENT( flt );

					// note: we want the byte-offset for the start of record
					node->byteOffset = flt->byteOffset - length - 4;

					if( FLTNODE_ISATTRIBUTE( node->type ) ) {

						//
						// is an attribute/ancillary record
						//

						// if lastNode available, apply attr

						if( flt->lastNode ) {

							// set up cross linking

							node->prev = NULL;
							node->next = flt->lastNode->attr;

							if( node->next )
								node->next->prev = node;

							flt->lastNode->attr = node;

							node->treeDepth = flt->treeDepth;

						} else {

							if( debug )
								printf("hanging attribute of type \"%s\" length: %d, not"
												" attached to parent node..\n", rec->name, length);

						}

						// 
						// we do not reset flt->lastNode here so that we can catch all
						// additional attributes that follow
						//

					} else {

						//
						// was not an attribute
						//

						// if node has a parent, add this one as a child
						if( node->parent )
							fltNodeAddChild( node->parent, node );

						// set up cross linking
						// popping sets lastNode to be the parent node

						node->next = NULL;
						node->prev = flt->lastNode;

						if( node->prev )
							node->prev->next = node;

						if( FLTNODE_CANPARENT( node->type ) )
							flt->lastParent = node;

						node->treeDepth = flt->treeDepth;

						// can be null, which is OK
						flt->lastNode = node;
					}
				}

				// call user func, if available
				if( rec->userFunc )
					rec->userFunc( flt, (void *)node, rec->userData );

			} else {
				if( debug )
					printf("unsupported node of type \"%s\" length: %d\n", rec->name, 
																																			length);
				fltSkipRecord( flt, length );
			}
		} else { 
			rec = (FltRecord *) bsearch(	&type, 
																		ObsoleteRecords, 
																		OBSOLETERECORDS_NUM,
																		sizeof( FltRecord ),
																		compareFunc );

			if( rec )
				printf("obsolete node of type \"%s\" found.\n", rec->name ); 
			else
				printf("Found undefined node: %d, length: %d\n", type, length );

			fltSkipRecord( flt, length );
		}

	}

	// run post-processing step
	fltPostProcess( flt );

	return 1;
}
