
/*****************************************************************************
 * FLT: An OpenFlight file loader
 *
 * Copyright (C) 2001  Michael M. Morrison   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * See the file EXCLUSIONS for further details.
 *
 *****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <math.h>
#include <sys/stat.h>
#include "flt.h"

FltType
fltStringToType( char * string )
{
	FltType t = FLT_TYPE_NULL;

	if( strchr( string, '[' ) )
		t |= FLT_ARRAY;
	
	if( strstr( string, "**" ) )
		t |= FLT_POINTER2;
	else if( strchr( string, '*' ) )
		t |= FLT_POINTER;
		
	if( !strncmp( string, "char", strlen( "char" ) ) )
		return t | FLT_CHAR;

	if( !strncmp( string, "uint8", strlen( "uint8" ) ) )
		return t | FLT_UINT8;

	if( !strncmp( string, "uint16", strlen( "uint16" ) ) )
		return t | FLT_UINT16;

	if( !strncmp( string, "uint32", strlen( "uint32 " ) ) )
		return t | FLT_UINT32;

	if( !strncmp( string, "int8", strlen( "int8" ) ) )
		return t | FLT_INT8;

	if( !strncmp( string, "int16", strlen( "int16" ) ) )
		return t | FLT_INT16;

	if( !strncmp( string, "int32", strlen( "int32 " ) ) )
		return t | FLT_INT32;

	if( !strncmp( string, "real32", strlen( "real32" ) ) )
		return t | FLT_REAL32;

	if( !strncmp( string, "real64", strlen( "real64 " ) ) )
		return t | FLT_REAL64;

	return t | FLT_UNKNOWN;
}

void
fltRecordEntryNameToString( char * string, FltNode * node, 
																										FltRecordEntryName * e )
{
	// null string
	string[0] = 0;

	switch( fltStringToType( e->type ) ) {
		case FLT_UINT8:
			sprintf( string, "%s: (%s) %02x", e->name, e->type,
																			*((uint8 *) ((char *)node + 
																							e->offset ) ) & 0xff );
			break;

		case FLT_UINT16:
			sprintf( string, "%s: (%s) %04x", e->name, e->type,
																			*((uint16 *) ((char *)node + 
																							e->offset ) ) & 0xffff);
			break;

		case FLT_UINT32:
			sprintf( string, "%s: (%s) %08x", e->name, e->type,
																			*((uint32 *) ((char *)node + 
																										e->offset ) ) );
			break;

		case FLT_INT8:
			sprintf( string, "%s: (%s) %02x", e->name, e->type,
																			*((int8 *) ((char *)node + 
																							e->offset ) ) & 0xff );
			break;

		case FLT_INT16:
			sprintf( string, "%s: (%s) %04x", e->name, e->type,
																			*((int16 *) ((char *)node + 
																							e->offset ) ) & 0xffff );
			break;
		case FLT_INT32:
			sprintf( string, "%s: (%s) %08x", e->name, e->type,
																			*((int32 *) ((char *)node + 
																										e->offset ) ) );
			break;

		case FLT_REAL64:
			sprintf( string, "%s: (%s) %f", e->name, e->type,
																			*((real64 *) ((char *)node + 
																										e->offset ) ) );
			break;

		case FLT_REAL32:
			sprintf( string, "%s: (%s) %f", e->name, e->type,
																			*((real32 *) ((char *)node + 
																										e->offset ) ) );
			break;

		case FLT_CHAR | FLT_ARRAY:
			sprintf( string, "%s: (%s) \"%s\"", e->name, e->type,
																			((char *) ((char *)node + 
																										e->offset ) ) );
			break;

		default:
			sprintf( string, "%s: (%s) Text Output Not Supported", e->name, e->type);
			break;
	}
}

void
fltDumpNode( FltNode * node )
{
	char string[300];
	FltRecord * rec;
	int x = 0;

	if( (rec = fltRecordGetDefinition( (uint16) node->type )) ) {
		while( rec->entry[x].type ) {
			fltRecordEntryNameToString( string, node, &rec->entry[x] ); 
			printf("%s\n", string);
			x ++;
		}
	}
}

uint32
fltLookupColor( FltFile * flt, uint32 colorIndex, real32 *r, real32 *g, 
																									real32 *b, real32 *a )
{
	uint32 idx = colorIndex / 128;
	real32 intensity = (float)(colorIndex % 128) / 128.0f;

	if( !flt->colorPalette )
		return 0;

	if( idx > 1023 ) {
		// return white

		*r = 1.0;
		*g = 1.0;
		*b = 1.0;
		*a = 1.0;
		
		return 1;
	}

	*r = (real32)( flt->colorPalette->color[ idx ] & 0x000000ff ) / 255.0f;
	*g = (real32)((flt->colorPalette->color[ idx ] >> 8) & 0x000000ff) / 255.0f;
	*b = (real32)((flt->colorPalette->color[ idx ] >> 16) & 0x000000ff) / 255.0f;
	*a = (real32)((flt->colorPalette->color[ idx ] >> 24) & 0x000000ff) / 255.0f;

	*r = *r * intensity;
	*g = *g * intensity;
	*b = *b * intensity;
	*a = *a * intensity;

	return 1;
}

//
// Find the FltTexture Record associated with an index
//
FltTexture *
fltLookupTexture( FltFile * flt, uint16 idx )
{
  uint32 i;

	for (i=0;i<flt->numNodes;i++) {

		if (flt->allNodes[i]->type == FLTRECORD_TEXTURE) {
			if (((FltTexture *)flt->allNodes[i])->index == idx)
				return (FltTexture *)flt->allNodes[i];
		}

	}

	return 0;
}

//
// Find the FltMaterial Record associated with an index
//
FltMaterial *
fltLookupMaterial( FltFile * flt, uint16 idx )
{
  uint32 i;

	for (i=0;i<flt->numNodes;i++) {

		if (flt->allNodes[i]->type == FLTRECORD_MATERIAL) {
			if (((FltMaterial *)flt->allNodes[i])->index == idx)
				return (FltMaterial *)flt->allNodes[i];
		}

	}

	return 0;
}

//
// Find the FltInstance Record associated with an index
//
FltInstanceDefinition *
fltLookupInstance( FltFile * flt, uint16 idx )
{
  uint32 i;

	for (i=0;i<flt->numNodes;i++) {

		if (flt->allNodes[i]->type == FLTRECORD_INSTANCEDEFINITION) {
			if (((FltInstanceDefinition *)flt->allNodes[i])->instance == idx)
				return (FltInstanceDefinition *)flt->allNodes[i];
		}

	}

	return 0;
}

static uint32
attrReadUInt32( FILE * f, uint32 * ret )
{
	uint32 uint;

	if( fread( &uint, 1, sizeof( uint32 ), f ) != sizeof( uint32 ) )
		return 0;

	ENDIAN_32( uint );

	*ret = uint;

	return 1;
}

FltTxAttributes *
fltLoadAttributes( const char * file )
{
	FILE * f;
	FltTxAttributes * attrs;

	if(!( f = fopen( file, "rb" ) ))
		return (FltTxAttributes *)NULL;

	if(!( attrs = malloc( sizeof( FltTxAttributes ) ) )) {
		fclose( f );
		return (FltTxAttributes *)NULL;
	}

	// read first set of attrs from file
	// todo: check for errors
	attrReadUInt32( f, &attrs->uTexels );		
	attrReadUInt32( f, &attrs->vTexels );		
	attrReadUInt32( f, &attrs->realWorldU );		
	attrReadUInt32( f, &attrs->realWorldV );		
	attrReadUInt32( f, &attrs->upVectorX );		
	attrReadUInt32( f, &attrs->upVectorY );		
	attrReadUInt32( f, &attrs->fileFormat );		
	attrReadUInt32( f, &attrs->minificationFilter );		
	attrReadUInt32( f, &attrs->magnificationFilter );		
	attrReadUInt32( f, &attrs->repetitionType );		
	attrReadUInt32( f, &attrs->repetitionU );		
	attrReadUInt32( f, &attrs->repetitionV );		
	attrReadUInt32( f, &attrs->modifyFlag );		
	attrReadUInt32( f, &attrs->xPivot );		
	attrReadUInt32( f, &attrs->yPivot );		
	attrReadUInt32( f, &attrs->environmentType );		

	fclose( f );
	return attrs;
}

void
fltFreeAttributes( FltTxAttributes * attrs )
{
	free( attrs );
}

static char *
mstrtok( MStrtokCtxt * ctxt, char * initstr, char * sep )
{
	int i, sepLen, eol = 0, found = 0;

	assert( ctxt );
	assert( sep );

	sepLen = strlen( sep );

	if( initstr ) {

		// reset everything
		ctxt->head = initstr;
		ctxt->tail = initstr;
		ctxt->last = 0;

	} else {

		// index curPtr
		if( ctxt->last )
			return 0;

		ctxt->tail++;
		ctxt->head = ctxt->tail;
	}


	// search for next separator, using tail
	while( 1 ) {
		if( *ctxt->tail ) {

			for(i=0;i<sepLen;i++) {
				if( *ctxt->tail == sep[i] ) {
					found++;
					break;
				}
			}

			if( found )
				break;

		} else {

			eol++;
			break;

		}

		ctxt->tail++;
	}

	if( eol ) {
		ctxt->last = 1;
		return ctxt->head;
	}

	// null term, and return
	if( found ) {
		*ctxt->tail = 0;
		return ctxt->head;
	}

	// sep not found
	ctxt->last = 1;
	return 0;
}

//
// helper func for fltFindFile
//

static int
buildPath( FltFile * flt, char * buf, char * filePath, char * envPath )
{
  char * ptr = 0;

  // if no envpath, nothing to do
  if( envPath == 0 ) {
    flt->pathInfo.first = 0;
    return 0;
  }

  if( flt->pathInfo.first == 0 ) {
    // copy strings since mstrtok changes them
    strcpy( flt->pathInfo.evPath, envPath );

    if(!( ptr = mstrtok( &flt->pathInfo.stc, flt->pathInfo.evPath, ";" ) ))
      // only 1 listed
      ptr = flt->pathInfo.evPath;

    flt->pathInfo.first = 1;

  } else
    ptr = mstrtok( &flt->pathInfo.stc, 0, ";" );

  if( ptr ) {

    strcpy( buf, ptr );
    strcat( buf, "/" );

		// check dos and unix sep
    if( strrchr( filePath, '/' ) )
      strcat( buf, strrchr( filePath, '/' ) + 1 );
		else if( strrchr( filePath, '\\' ) )
      strcat( buf, strrchr( filePath, '\\' ) + 1 );
    else
      strcat( buf, filePath );

  } else {

    flt->pathInfo.first = 0;
    return 0;

  }

  return 1;
}

void
fltAddSearchPath( FltFile * flt, char * path )
{
	if( flt->pathInfo.searchPath[0] )
		strcat( flt->pathInfo.searchPath, ";" );
	else {
		char * txtpath;

		// automatically add TXTPATH as well
		txtpath = getenv( "TXTPATH" );

		if( txtpath ) {
			strcat( flt->pathInfo.searchPath, txtpath );
			strcat( flt->pathInfo.searchPath, ";" );
		}
	}

	strcat( flt->pathInfo.searchPath, path );
}

//
// Find a file based on TXTPATH variable, separated by ';' (due to Creator
// for windows drive letters).
//

int
fltFindFile ( FltFile * flt, char * inFileName, char *outFileName )
{
  struct stat st;
  int result;
  char tmpBuf[MAX_PATHLEN];
  char * txtpath;

	//printf("looking for: %s\n", inFileName );

  result = stat( inFileName, &st );

  if( result < 0 ) {
    txtpath = flt->pathInfo.searchPath;
		//printf("search path: %s\n", txtpath );
    if( txtpath[0] ) {
      while( buildPath( flt, tmpBuf, inFileName, txtpath ) ) {
        if( stat( tmpBuf, &st ) < 0 ) {
          //printf("FLT: Unable to find texture in path: \"%s\"\n", tmpBuf );
          continue;
        } else {
          //printf("FLT: Texture located: \"%s\"\n", tmpBuf );
          strcpy( outFileName, tmpBuf );
          // reset buildpath
          buildPath( flt, 0, 0, 0 );
          return 1;
        }
      }
    }
  } else {
    strcpy( outFileName, inFileName );
    return 1;
  }

  return 0;
} 

//
// check for a finite value (ie: not NAN or INF), as stored in the flight file
//
int
fltFinite64( real64 value )
{
	// IEEE 754 values that will be stored in the flight file
	static unsigned char 
		posInf[] = { 0x7f, 0xf0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 };
	static unsigned char 
		negInf[] = { 0xff, 0xf0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 };
	static unsigned char 
		posNan[] = { 0x7f, 0xf8, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 };
	static unsigned char 
		negNan[] = { 0xff, 0xf8, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 };
	unsigned char * valPtr;

	// if little endian, convert to big
	ENDIAN_64r( value );

	valPtr = (unsigned char *)&value;

	if(	!memcmp( negNan, valPtr, 8 ) || !memcmp( posNan, valPtr, 8 ) ||
			!memcmp( negInf, valPtr, 8 ) || !memcmp( posInf, valPtr, 8 )  )
		return 0;
	else
		return 1;
}
