/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglAnimation.cpp
 *   Author: Scott McMillan
 *  Created: 29 May 2001
 *  Summary: abstract base for all animation class types
 *****************************************************************************/

#include <sglAnimation.hpp>

//----------------------------------------------------------------------------
sglAnimation::sglAnimation()
      : sglGroup(),
        m_mode(eCYCLE),
        m_state(eSTOP),
        m_state_dirty_flag(false),
        m_range_flag(false),
        m_first(0),
        m_last(0),
        m_reverse_flag(false),
        m_speed(1.f),
        m_num_cycles(0)
{
}

//----------------------------------------------------------------------------
sglAnimation::~sglAnimation()
{
}

//----------------------------------------------------------------------------
bool sglAnimation::setState(StateEnum state)
{
   bool flag = false;

   if (state != m_state || state == eSTART)
   {
      // only allow valid state changes:
      // STOP->START, START/RESUME->STOP/PAUSE, PAUSE->RESUME/STOP
      if (m_state == eSTART && state == eSTART)
      {
         flag = true;
      }
      else if (m_state == eSTOP && state == eSTART)
      {
         flag = true;
      }
      else if ((m_state == eSTART || m_state == eRESUME) &&
               (state == eSTOP || state == ePAUSE))
      {
         flag = true;
      }
      else if ((m_state == ePAUSE) &&
               (state == eRESUME || state == eSTOP))
      {
         flag = true;
      }

      if (flag)
      {
         m_state = state;
         m_state_dirty_flag = true; // need a separate local dirty flag
         dirty(eANIMATION);         // need to notify all ancestor nodes
      }
   }

   return flag;
}

//----------------------------------------------------------------------------
void sglAnimation::setRange(int first_node, int last_node)
{
   if (getNumChildren() == 0 || (first_node == 0 && last_node < 0))
   {
      m_range_flag = false;
   }
   else
   {
      int last_index = (int)getNumChildren() - 1;
      if (last_node  < 0 || last_node  > last_index)  last_node  = last_index;
      if (first_node < 0 || first_node > last_index)  first_node = last_index;

      // exchange values
      if (last_node < first_node)
      {
         int tmp = last_node;
         last_node = first_node;
         first_node = tmp;
      }

      m_range_flag = true;
      m_first = (unsigned int)first_node;  // HACK cast
      m_last  = (unsigned int)last_node;   // HACK cast
   }

   computeValues();
}

//----------------------------------------------------------------------------
void sglAnimation::getRange(unsigned int &first_node,
                            unsigned int &last_node) const
{
   if (m_range_flag)
   {
      first_node = m_first;
      last_node = m_last;
   }
   else
   {
      first_node = 0;
      last_node = getNumChildren() - 1;
   }
}

//----------------------------------------------------------------------------
void sglAnimation::setSpeed(float speed)
{
   if (speed != ((m_reverse_flag ? -1.f : 1.f)*m_speed))
   {
      if (speed < 0.f)
      {
         speed *= -1.f;
         m_reverse_flag = true;
      }
      else
      {
         m_reverse_flag = false;
      }

      m_speed = speed;
      computeValues();
   }
}

//----------------------------------------------------------------------------
float sglAnimation::getSpeed() const
{
   if (m_reverse_flag)
   {
      return -m_speed;
   }
   else
   {
      return m_speed;
   }
}

//----------------------------------------------------------------------------
void sglAnimation::setNumCycles(unsigned int num_cycles)
{
   if (m_num_cycles != num_cycles)
   {
      m_num_cycles = num_cycles;
      computeValues();
   }
}

//----------------------------------------------------------------------------
void sglAnimation::copyTo(sglAnimation *dst, unsigned int mode) const
{
   sglGroup::copyTo(dst, mode);

   dst->setMode(m_mode);
   dst->setState(m_state);

   if (m_range_flag)
   {
      unsigned int first_node, last_node;
      getRange(first_node, last_node);
      dst->setRange((int)first_node, (int)last_node);
   }

   dst->setSpeed(getSpeed());
   dst->setNumCycles(m_num_cycles);
}

//----------------------------------------------------------------------------
void sglAnimation::printInfo(ostream &ostrm, const char *indent_string) const
{
   sglGroup::printInfo(ostrm, indent_string);
   ostrm << indent_string << "mode =  "
         << ((m_mode == eSWING) ? "SWING" : "CYCLE") << endl;
   ostrm << indent_string << "state = ";
   switch (m_state)
   {
      case eSTOP:
         ostrm << "STOP" << endl;
         break;
      case eSTART:
         ostrm << "START" << endl;
         break;
      case ePAUSE:
         ostrm << "PAUSE" << endl;
         break;
      case eRESUME:
         ostrm << "RESUME" << endl;
         break;
   }
   if (m_range_flag)
      ostrm << indent_string << "range = [" << m_first << ".."
            << m_last << "]" << endl;
   ostrm << indent_string << "speed = "
         << m_speed*(m_reverse_flag ? -1.f : 1.f) << endl;
   ostrm << indent_string << "cycles = " << m_num_cycles << endl;
}
