/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglAnimation.hpp
 *  Created: 29 May 2001
 *  Summary: abstract base class for all animation classes.
 *****************************************************************************/

#ifndef __SGL_ANIMATION_HPP
#define __SGL_ANIMATION_HPP

#include <sgl.h>
#include <sglGroup.hpp>

//----------------------------------------------------------------------------

/**

Abstract base class for sglTimedAnimation and sglKeyFrameAnimation nodes that
will select the child to be traversed based on frame time (real time) or frame
count and whether or not the animation is active or not.  This node is a
subclass of sglGroup but the order (index) of the children is important and
determines the order that they are displayed in the animation.  Functions in
the subclasses determine the duration that each child is displayed.

The speed of the animation can be changed with a call to setSpeed(..) where
negative values reverse the animation.  Speeds with magnitude between zero and
one slow the animation down (0.5 corresponds to half speed), and magnitudes
greater than one speed the animation up (2.0 corresponds to double speed).
Specifying a speed of zero disables the animation.

The subset of the children can be animated by specifying the indices of the
first and last child in a call to setRange(..).  Passing no arguments to this
function reverts the animation to using all the children.  The number of times
the animation is performed can also be set by calling setNumCycles(..).
Passing a zero to this function causes the animation to repeat endlessly when
activated.

The animation can be set to cycle (mode = eCYCLE) in ascending order or cycle
(mode = eSWING) back and forth by calling the setMode(..) member function.
A complete cycle in eCYCLE mode consists of animation from the first to last
child (either from the complete set of child or the limits specified by a call
to setRange()).  A complete cycle in eSWING mode consists of a cycle up and
back.  Note that the last child is not displayed twice, nor is the first if the
number of cycles is specified as greater than 1.

The state (control mode) of the system can be one of four: eSTOP (the initial
state, and discontinues any current animation), eSTART (begins the animation),
ePAUSE (pauses the animation), and eRESUME (continues the animation after a
pause).  While the state is eSTOP, nothing is shown.  Likewise, when the
prescribed number of cycles is finished no children will be selected for
display.  While the animation is running or when the animation ends (provided
the number of cycles is not set to zero), it is still in the eSTART state the
animation can be restarted by calling setState(sglAnimation::eSTART).

The subclasses of this node depend on either the current frame count or frame
time that is stored in the sglScene node at the root of the scene graph
containing this node.  The frame count is incremented and the current frame
time is determined each time sglScene::preDraw() is called.  The different
traversals use these values to realize the correct behaviour for these nodes so
that it is necessary for the traversals to start at the sglScene node.  If the
state of this node has changed, it will be marked as dirty and the preDraw()
traversal will set some member variables internal to this node.  If this node
is a member of more than one scene graph then proper behaviour is not
guaranteed.

<b>Special behaviour for intersect and pick traversals</b>

The default behaviour of the intersect and pick traversals of the subclasses of
this node is to test against the current child to be displayed based on the
current frame count stored in the sglScene node at the root of the scene graph
containing this node.  The behaviour of the intersect and pick traversals of
these nodes can be modified by adding sglProbe::eANIMATION_ALL or
sglProbe::eANIMATION_NONE to the sglProbe mode.  The former behaves as a normal
sglGroup node testing against all the children, while the latter returns
without testing against any children.

 */

//----------------------------------------------------------------------------

class SGL_DLL_API sglAnimation : public sglGroup
{
public:
   /** Controls the behaviour of the animation */
   enum ModeEnum
   {
      /// forward repeatedly
      eCYCLE,
      /// back and forth
      eSWING
   };

   /** The current state of the animation */
   enum StateEnum
   {
      /// begin at child 0
      eSTART,
      /// stop at current child
      ePAUSE,
      /// continue at current child
      eRESUME,
      /// stop and reset to child 0
      eSTOP
   };

public:
   /// default constructor
   sglAnimation();
   /// virtual destructor
   virtual ~sglAnimation();

   /** Set the animation mode for this node to either eCYCLE (the default) or
       eSWING.  When in eCYCLE mode, the children will be displayed in order
       from lowest index to highest, and repeat (if required) from lowest to
       highest.  When in eSWING mode, the children will be displayed in order
       from lowest to highest and then reverse from highest to lowest (this
       entire cycle will be repeated if so configured.  In swing mode the
       highest will only be displayed for its allotted time once before
       reversing direction.
       @param mode Either eCYCLE or eSWING.
   */
   void setMode(ModeEnum mode) { m_mode = mode; }

   /** Return the current animation mode.
       @return ModeEnum the current mode: either eCYCLE or eSWING.
   */
   ModeEnum getMode() const { return m_mode; }

   /** Change the state of the animation node.  Any valid change here causes
       sets the ANIMATION dirty bit so that preDraw will do something to set up
       the animation.  The current valid state changes are as follows:
       eSTART->eSTOP/ePAUSE, ePAUSE->eRESUME/eSTOP, eRESUME->eSTOP/ePAUSE,
       eSTOP->eSTART, and eSTART->eSTART.  Note that is eRESUME is specified it
       will be changed to eSTART automatically during the preDraw traversal.
       @param state One of eSTART, ePAUSE, eRESUME, eSTOP
       @return true if the state change was valid; otherwise, false.
   */
   bool setState(StateEnum state);

   /** Get the current state of the node.
       @return state One of eSTART, ePAUSE, eRESUME, eSTOP.
   */
   StateEnum getState() const { return m_state; }

   /** This sets the range of nodes to display (by child index number).  By
       passing no arguments, then the system reverts to the default of showing
       all children.  If last_node is less than first_node, then they will be
       exchanged (reverse animations should be specified with a negative
       speed.  Note that if the number of children of this node changes
       after a range is set and causes one or both of the range values to be
       invalid, it may lead to unspecified behaviour.
       @param first_node Index of first child to be displayed.  If a number
         less than zero or greater then the highest child index
         (getNumChildren-1) is specified, then it will be set to the last
         child.
       @param last_node Index of last child to be displayed.  If a number
         less than zero or greater than the highest child index is specified,
         then it will be set to the last child.
   */
   void setRange(int first_node = 0, int last_node = -1);

   /** Get the current range of children to be displayed.  If the child list is
       empty then first_node will be set to zero and last_node will be set to
       MAX_UINT.
       @param first_node Index of first child to be displayed.
       @param last_node Index of last child to be displayed.
   */
   void getRange(unsigned int &first_node, unsigned int &last_node) const;

   /** Set the speed of the animation relative to the durations set by the
       subclasses (the time that each child is shown, its duration, is divided
       by this value).  A speed of zero disables the animation, a positive
       speed corresponds to a forward animation, and a negative speed reverses
       the animation.  A speed with magnitude less than 1.0 will slow the
       animation down, while a magnitude greater than 1.0 will increase the
       speed.
       @param speed Speed of the animation (default is 1.0f).
    */
   void setSpeed(float speed);

   /** Get the current speed.
       @return current speed.
    */
   float getSpeed() const;

   /** Set the number cycles this animations is supposed to perform.  For the
       eCYCLE mode, a period consists of (child 0 ... child n-1).  For the
       eSWING mode, a period consists of (child 0 ... child n-1 ... child 1).
       If the number cycles is set to zero, then the animation node will cycle
       continuously.  After the specified number of cycles, the state will
       remain in eSTART/eRESUME (whichever it was in).  And the user will have
       to change the state to eSTOP before typing start again.
       @param num_cycles Number of cycles to perform.
   */
   void setNumCycles(unsigned int num_cycles);

   /** Get the current number of cycles.
       @return number of cycles.
    */
   unsigned int getNumCycles() const { return m_num_cycles; }

   // see sglObject::printInfo() for details
   virtual void printInfo(ostream &ostrm, const char *indent_string) const;

protected:
   virtual void computeValues() = 0;

   void copyTo(sglAnimation *dst, unsigned int mode) const;

private:    // not implemented
   sglAnimation(const sglAnimation &);
   sglAnimation &operator=(const sglAnimation &);

protected:
   ModeEnum     m_mode;
   StateEnum    m_state;
   bool         m_state_dirty_flag;  // need separate dirty flag

   bool         m_range_flag;
   unsigned int m_first;
   unsigned int m_last;

   bool         m_reverse_flag;
   float        m_speed;
   unsigned int m_num_cycles;
};

#endif
