/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglBezierCurveSet.hpp
 *   Author: Tom Stimson
 *  Created: 13 Feb 2000
 *  Summary:
 *****************************************************************************/

#ifndef __SGL_BEZIERCURVESET_HPP
#define __SGL_BEZIERCURVESET_HPP

#include <sgl.h>
#include <sglGeoSet.hpp>

//----------------------------------------------------------------------------

/**

This class encapsulates OpenGL's interface to the bezier curve mechanism.
I need Tom to document this one.

Note: The BindEnum's for BezierCurves have different meanings than for other
sglGeoSets.
\begin{description}
\item[sglGeoSet::ePER_VERTEX] A different set of control points are applied to
each primitive.  This requires order*num_prim control points.
\item[sglGeoSet::ePER_PRIM]   The same set of control points is applied to each
primitive. The range applied is also the same (min-max).  This will create the
effect of repeating the values on each primitive.  This requires order control
points.
\item[sglGeoSet::eOVERALL]    The same set of control points is applied to each
primitive.  The range is streched, and applied linearly across the primitives.
This will have the effect of streching the values across all of the
primitives.  This requires order control points.
\end{description}

{\bf Example:}

\begin{verbatim}
   sglBezierCurveSet *gset = new sglBezierCurveSet();

   sglVec3f *vertex = new sglVec3f[4];
   vertex[0].set(-10.0f, 0.0f, -10.0f);
   vertex[1].set(-08.0f, 7.0f,  10.0f);
   vertex[2].set( 08.0f, 0.0f, -10.0f);
   vertex[3].set( 10.0f, 0.0f,  10.0f);

   gset->setCoordList(vertex);
   gset->setCoordOrder(4);

   sglVec4f *color = new sglVec4f(1.0f, 0, 0, 1.0f);
   gset->setColorList(sglGeoSet::eOVERALL, color);
   gset->setColorOrder(1);

   gset->setNumPrims(1);
   gset->setNumSegments(10);
\end{verbatim}
*/

class SGL_DLL_API sglBezierCurveSet : public sglGeoSet
{
public:
   /// default constructor
   sglBezierCurveSet();
   /// virtual destructor
   virtual ~sglBezierCurveSet();

   ///
   void setTexCoordBind(BindEnum bind);
   ///
   BindEnum getTexCoordBind() const;

   /** order is the degree of the curve plus one
       (should == the number of control points/prim)
   */
   void setCoordOrder(GLint order = 4);
   ///
   void setTexCoordOrder(GLint order = 4);
   ///
   void setNormalOrder(GLint order = 4);
   ///
   void setColorOrder(GLint order = 4);

   ///
   GLint getCoordOrder() {return m_vorder;}
   ///
   GLint getTexCoordOrder() {return m_torder;}
   ///
   GLint getNormalOrder() {return m_norder;}
   ///
   GLint getColorOrder() {return m_corder;}

   /// get/set the range of the parametric value.
   void setRange(float min = 0, float max = 1.0f);
   ///
   void getRange(float &min, float &max);

   /// get/set the number of segments to be drawn for each curve
   void setNumSegments(unsigned int num);
   ///
   unsigned int getNumSegments() { return m_segments; }

   /// draw the geometry of the bezier curve.
   virtual void drawGeometry(const vector<sglTexCoords> &tex_coords) const;

   /// retrieve the primitive type
   sglStats::PrimEnum getPrimType() const {return sglStats::eBEZIER_CURVES;}
   ///
   GLenum getGLPrimType() const {return GL_LINE_STRIP;}
   ///
   unsigned int getPrimSize() const {return m_segments+1;}
   ///
   AAEnum getAAType() const {return eLINE_AA;}

protected:
   virtual bool computeBounds();
   virtual void updateDrawFunc();

private:  // not implemented
   sglBezierCurveSet(const sglBezierCurveSet &);
   sglBezierCurveSet &operator=(const sglBezierCurveSet &);

   typedef void (*Map1Func)(float, float, int, int, const void*);
   static void invalidFunc(float, float, int, int, const void*);
   static void applyVerts3f(float, float, int, int, const void*);
   static void applyVerts3d(float, float, int, int, const void*);
   static void applyVerts4f(float, float, int, int, const void*);
   static void applyVerts4d(float, float, int, int, const void*);
   static void applyNorms3f(float, float, int, int, const void*);
   static void applyNorms3d(float, float, int, int, const void*);
   static void applyColors4f(float, float, int, int, const void*);
   static void applyColors4d(float, float, int, int, const void*);
   static void applyTexCoords1f(float, float, int, int, const void*);
   static void applyTexCoords1d(float, float, int, int, const void*);
   static void applyTexCoords2f(float, float, int, int, const void*);
   static void applyTexCoords2d(float, float, int, int, const void*);
   static void applyTexCoords3f(float, float, int, int, const void*);
   static void applyTexCoords3d(float, float, int, int, const void*);
   static void applyTexCoords4f(float, float, int, int, const void*);
   static void applyTexCoords4d(float, float, int, int, const void*);

protected:
   GLint        m_vorder;
   GLint        m_norder;
   GLint        m_corder;
   GLint        m_torder;

   float        m_min;
   float        m_max;

   unsigned int m_segments;

   static Map1Func s_vfunc_table[24];
   static Map1Func s_nfunc_table[8];
   static Map1Func s_cfunc_table[16];
   static Map1Func s_tfunc_table[32];
};

#endif
