/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglBezierSurfaceSet.hpp
 *   Author: Tom Stimson
 *  Created: 30 April 2000
 *  Summary:
 *****************************************************************************/

#ifndef __SGL_BEZIERSURFACESET_HPP
#define __SGL_BEZIERSURFACESET_HPP

#include <sgl.h>
#include <sglGeoSet.hpp>

//----------------------------------------------------------------------------

/**

This class encapsulates OpenGL's interface to the bezier (nurbs) surface
mechanism.  I need Tom to document this one.

Note: The BindEnum's for BezierSurfaces have different meanings than for other
sglGeoSets.
\begin{description}
\item[sglGeoSet::ePER_VERTEX] A different set of control points are applied to
each primitive.  This requires order_u*order_v*num_prim control points.

\item[sglGeoSet::ePER_PRIM] The same set of control points is applied to each
primitive. The range applied is also the same (min-max).  This will create the
effect of repeating the values on each primitive.  This requires
order_u*order_v control points.

\item[sglGeoSet::eOVERALL] The same set of control points is applied to each
primitive.  The range (along the u dimension) is streched, and applied linearly
across the primitives.  This will have the effect of streching the values
across all of the primitives.  This requires order_u*order_v control points.
\end{description}

{\bf Example:}
\begin{verbatim}
   sglBezierSurfaceSet *gset = new sglBezierSurfaceSet();

   sglVec3f *vertex = new sglVec3f[16];
   vertex[ 0].set(-10.0f,  0.0f, -10.0f);
   vertex[ 1].set(-05.0f,  7.0f, -10.0f);
   vertex[ 2].set( 05.0f, -7.0f, -10.0f);
   vertex[ 3].set( 10.0f,  0.0f, -10.0f);
   vertex[ 4].set(-10.0f,  0.0f, -05.0f);
   vertex[ 5].set(-05.0f,  7.0f, -05.0f);
   vertex[ 6].set( 05.0f, -7.0f, -05.0f);
   vertex[ 7].set( 10.0f,  0.0f, -05.0f);
   vertex[ 8].set(-10.0f,  0.0f,  05.0f);
   vertex[ 9].set(-05.0f,  7.0f,  05.0f);
   vertex[10].set( 05.0f, -7.0f,  05.0f);
   vertex[11].set( 10.0f,  0.0f,  05.0f);
   vertex[12].set(-10.0f,  0.0f,  10.0f);
   vertex[13].set(-05.0f,  7.0f,  10.0f);
   vertex[14].set( 05.0f, -7.0f,  10.0f);
   vertex[15].set( 10.0f,  0.0f,  10.0f);

   gset->setCoordList(vertex);
   gset->setCoordOrder(4, 4);

   sglVec4f *color = new sglVec4f(1.0f, 0, 0, 1.0f);
   gset->setColorList(SGL_OVERALL, color);
   gset->setColorOrder(1, 1);

   gset->setNumPrims(1);
   gset->setNumSegments(10, 10);
\end{verbatim}
  */

class SGL_DLL_API sglBezierSurfaceSet : public sglGeoSet
{
public:
   /// default constructor
   sglBezierSurfaceSet();
   /// virtual destructor
   virtual ~sglBezierSurfaceSet();

   ///
   void setTexCoordBind(BindEnum bind);
   ///
   BindEnum getTexCoordBind() const;

   /** order is the degree of the surface plus one
       (should == the number of control points/prim)
   */
   void setCoordOrder(GLint order_u = 4, GLint order_v = 4);
   ///
   void setTexCoordOrder(GLint order_u = 4, GLint order_v = 4);
   ///
   void setNormalOrder(GLint order_u = 4, GLint order_v = 4);
   ///
   void setColorOrder(GLint order_u = 4, GLint order_v = 4);

   ///
   void getCoordOrder(GLint &order_u, GLint &order_v)
      { order_u = m_vorder_u; order_v = m_vorder_v; }
   ///
   void getTexCoordOrder(GLint &order_u, GLint &order_v)
      { order_u = m_torder_u; order_v = m_torder_v; }
   ///
   void getNormalOrder(GLint &order_u, GLint &order_v)
      { order_u = m_norder_u; order_v = m_norder_v; }
   ///
   void getColorOrder(GLint &order_u, GLint &order_v)
      { order_u = m_corder_u; order_v = m_corder_v; }

   /// get/set the range of the parametric value in the u and v dimensions.
   void setRange(float min_u = 0, float max_u = 1.0f,
                 float min_v = 0, float max_v = 1.0f);
   ///
   void getRange(float &min_u, float &max_u, float &min_v, float &max_v);

   /** get/set the number of segments to be drawn for each surface
       in the u and v dimensions.
   */
   void setNumSegments(unsigned int num_u, unsigned int num_v);
   ///
   void setNumSegments(unsigned int &num_u, unsigned int &num_v)
      { num_u = m_segments_u; num_v = m_segments_v; }

   /// Add this drawable's stats to the given debug_stats.
   virtual void addStats(sglStats &stats) const;

   /// draw the geometry of the bezier surface.
   virtual void drawGeometry(const vector<sglTexCoords> &tex_coords) const;

   /// retrieve the primitive type
   sglStats::PrimEnum getPrimType() const {return sglStats::eBEZIER_SURFACES;}
   ///
   GLenum getGLPrimType() const {return GL_QUAD_STRIP;}
   ///
   unsigned int getPrimSize() const {return (m_segments_u+1)*(m_segments_v+1);}
   ///
   AAEnum getAAType() const {return ePOLYGON_AA;}

protected:
   virtual bool computeBounds();
   virtual void updateDrawFunc();

private:  // not implemented
   sglBezierSurfaceSet(const sglBezierSurfaceSet &);
   sglBezierSurfaceSet &operator=(const sglBezierSurfaceSet &);

   typedef void (*Map2Func)(float, float, int, int,
                            float, float, int, int, const void*);
   static void invalidFunc(float, float, int, int,
                           float, float, int, int, const void*);
   static void applyVerts3f(float, float, int, int,
                            float, float, int, int, const void*);
   static void applyVerts3d(float, float, int, int,
                            float, float, int, int, const void*);
   static void applyVerts4f(float, float, int, int,
                            float, float, int, int, const void*);
   static void applyVerts4d(float, float, int, int,
                            float, float, int, int, const void*);
   static void applyNorms3f(float, float, int, int,
                            float, float, int, int, const void*);
   static void applyNorms3d(float, float, int, int,
                            float, float, int, int, const void*);
   static void applyColors4f(float, float, int, int,
                             float, float, int, int, const void*);
   static void applyColors4d(float, float, int, int,
                             float, float, int, int, const void*);
   static void applyTexCoords1f(float, float, int, int,
                                float, float, int, int, const void*);
   static void applyTexCoords1d(float, float, int, int,
                                float, float, int, int, const void*);
   static void applyTexCoords2f(float, float, int, int,
                                float, float, int, int, const void*);
   static void applyTexCoords2d(float, float, int, int,
                                float, float, int, int, const void*);
   static void applyTexCoords3f(float, float, int, int,
                                float, float, int, int, const void*);
   static void applyTexCoords3d(float, float, int, int,
                                float, float, int, int, const void*);
   static void applyTexCoords4f(float, float, int, int,
                                float, float, int, int, const void*);
   static void applyTexCoords4d(float, float, int, int,
                                float, float, int, int, const void*);

protected:
   GLint        m_vorder_u;
   GLint        m_vorder_v;
   GLint        m_norder_u;
   GLint        m_norder_v;
   GLint        m_corder_u;
   GLint        m_corder_v;
   GLint        m_torder_u;
   GLint        m_torder_v;

   float        m_min_u;
   float        m_min_v;
   float        m_max_u;
   float        m_max_v;

   unsigned int m_segments_u;
   unsigned int m_segments_v;

   static Map2Func s_vfunc_table[24];
   static Map2Func s_nfunc_table[8];
   static Map2Func s_cfunc_table[16];
   static Map2Func s_tfunc_table[32];
};

#endif
