/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglBillboard.cpp
 *  Created: 2 December 1998
 *  Summary: modify modelview matrix to affect some specified orientation
 *         : behaviour before traversing the children
 *****************************************************************************/

#include <sgl.h>
#include <sglBillboard.hpp>
#include <sglCull.hpp>

//----------------------------------------------------------------------------
sglBillboard::sglBillboard() : sglGroup(),
   m_mode(ePOINT_ROT_EYE)
{
   m_pos.set(0., 0., 0.);
   m_axis.set(0., 0., 1.);
}

//----------------------------------------------------------------------------
sglBillboard::~sglBillboard()
{
}

//----------------------------------------------------------------------------
void sglBillboard::setPos(const sglVec3d &p)
{
   if (!(m_pos == p))
   {
      m_pos = p;
      dirty(eBOUND);
   }
}

//----------------------------------------------------------------------------
void sglBillboard::setAxis(const sglVec3d &v)
{
   if (!(m_axis == v))
   {
      if (m_axis.normalize(v) == 0.)
      {
         m_axis.set(0., 0., 1.);
      }
   }
}


//----------------------------------------------------------------------------
void sglBillboard::setMode(ModeEnum mode)
{
   if (m_mode != mode)
   {
      m_mode = mode;
      dirty(eBOUND);
   }
}

//----------------------------------------------------------------------------
void sglBillboard::computeBound()
{
   sglGroup::computeBound();
   if (!m_sphere.isEmpty() && !m_sphere.isInfinite())
   {
      sglVec3d center = m_sphere.getCenter();
      center += m_pos;
      m_sphere.setCenter(center);
   }
}

//----------------------------------------------------------------------------
void sglBillboard::computeOrientation(const sglMat4d &modelview,
                                      sglMat4d &new_matrix) const
{
   switch(m_mode)
   {
      case eAXIAL_ROT:  // the outboard Y axis about the inboard m_axis
      {
         GLint viewport[4]; // 0 - origin_x, 1 - origin_y, 2 - width, 3 - ht.
         double dx, dy, dz;
         sglVec3d o, x, z, xx, zz;;

         sglMat4d m;
         sglVec4d in, out;

         glGetIntegerv (GL_VIEWPORT, viewport);
         m.inverse(modelview);

         // get the origin (in taken from Mesa 3.4 src-glu/project.c)
         in[0] = (viewport[0]) * 2.0 / viewport[2] - 1.0;
         in[1] = (viewport[1]) * 2.0 / viewport[3] - 1.0;
         in[2] =-1.0;
         in[3] = 1.0;

         mul(in, m, out);
         o[0] = out[0]/out[3];
         o[1] = out[1]/out[3];
         o[2] = out[2]/out[3];
         //cout << "o = " << o[0] << ' ' << o[1] << ' ' << o[2] << endl;

         // get a point along the z axis
         // (in taken from Mesa 3.4 src-glu/project.c)
         in[0] = viewport[0] * 2.0 / viewport[2] - 1.0;
         in[1] = viewport[1] * 2.0 / viewport[3] - 1.0;
         in[2] = 1.0;
         in[3] = 1.0;

         mul(in, m, out);
         dx = out[0]/out[3];
         dy = out[1]/out[3];
         dz = out[2]/out[3];
         //cout << "dz = " << dx << ' ' << dy << ' ' << dz << endl;

         z[0] = dx - o[0];
         z[1] = dy - o[1];
         z[2] = dz - o[2];

         x.cross(m_axis, z);
         if (xx.normalize(x) == 0.0)
         {
            // view axis aligns with rotation axis -- ick;
            // get point along x axis instead
            // (in taken from Mesa 3.4 src-glu/project.c)
            in[0] = (1.0 - viewport[0]) * 2 / viewport[2] - 1.0;
            in[1] = viewport[1] * 2.0 / viewport[3] - 1.0;
            in[2] =-1.0;
            in[3] = 1.0;

            mul(in, m, out);
            dx = out[0]/out[3];
            dy = out[1]/out[3];
            dz = out[2]/out[3];
            //cout << "dx = " << dx << ' ' << dy << ' ' << dz << endl;

            x[0] = dx - o[0];
            x[1] = dy - o[1];
            x[2] = dz - o[2];

            z.cross(x, m_axis);
            x.cross(m_axis, z);
            xx.normalize(x);
            zz.normalize(z);
         }
         else
         {
            z.cross(xx, m_axis);
            zz.normalize(z);
         }

         for (unsigned int i=0; i<3; i++)
         {
            m[0][i] = xx[i];
            m[1][i] = m_axis[i];
            m[2][i] = zz[i];
            m[3][i] = m[i][3] = 0.0;
         }
         m[3][3] = 1.;

         new_matrix.mul(m, modelview);
         for (unsigned int ix=0; ix<3; ix++)
            for (unsigned int iy=0; iy<3; iy++)
               new_matrix[3][ix] += modelview[iy][ix]*m_pos[iy];

         break;
      }

      case ePOINT_ROT_EYE:
      {
         // have to do the following to preserve any scaling
         double factor[3];
         int ix, iy;
         for (ix=0; ix<3; ix++)
         {
            factor[ix] = sqrt(modelview[ix][0]*modelview[ix][0] +
                              modelview[ix][1]*modelview[ix][1] +
                              modelview[ix][2]*modelview[ix][2]);
         }
         for (ix=0; ix<3; ix++)
         {
            for (iy=0; iy<3; iy++)
               new_matrix[ix][iy] = 0.0;
            new_matrix[ix][ix] = factor[ix];
            new_matrix[3][ix] = modelview[3][ix];
            for (iy=0; iy<3; iy++)
               new_matrix[3][ix] += modelview[iy][ix]*m_pos[iy];
            new_matrix[ix][3] = modelview[ix][3];
         }
         new_matrix[3][3] = modelview[3][3];
         break;
      }
      case ePOINT_ROT_WORLD:
         // NOT IMPLEMENTED YET
         break;
   }
}

//----------------------------------------------------------------------------
void sglBillboard::computeOrientation(const sglMat4f &modelview,
                                      sglMat4f &new_matrix) const
{
   unsigned int i, j;
   sglMat4d mv, nm;

   for (i=0; i<4; i++)
      for (j=0; j<4; j++)
         mv[i][j] = modelview[i][j];

   computeOrientation(mv, nm);

   for (i=0; i<4; i++)
      for (j=0; j<4; j++)
         new_matrix[i][j] = nm[i][j];
}

//----------------------------------------------------------------------------
sglBound::IntersectResultEnum sglBillboard::intersect(
   sglIntersectf &segset) const
{
   if (!m_child_list.empty() && segset.getMode()&sglProbe::eBILLBOARD_VIEW)
   {
      sglSegmentf seg(segset.getSegment()); // copy the current segment

      // test against the current bounding sphere
      sglBound::IntersectResultEnum result = m_sphere.intersect(seg);
      if (result == sglBound::eOUTSIDE)
      {
         return result;
      }

      // save the original view matrix
      sglMat4f old_matrix = segset.getModelViewMatrix();

      // compute the new view matrix
      computeOrientation(old_matrix, segset.getModelViewMatrix());

      // store the original segment
      seg = segset.getSegment();

      // transform the original segment
      sglMat4f inv_mat;
      inv_mat.inverseAffine(segset.getModelViewMatrix());
      segset.mul(old_matrix);
      segset.mul(inv_mat);

      // compute a new view frustum
      sglFrustumf old_view_frustum;
      if (segset.getMode()&(sglProbe::eUNSCALE_VIEW|
                            sglProbe::eUNPROJECT_VIEW))
      {
         old_view_frustum = segset.getViewFrustum();
         segset.getViewFrustum().mul(segset.getOrigViewFrustum(), inv_mat);
      }

      // traverse the children
      result = sglBound::eOUTSIDE;
      for (vector<sglNode*>::const_iterator elt = m_child_list.begin();
           elt != m_child_list.end(); ++elt)
      {
         if ((*elt)->intersect(segset) == sglBound::eINTERSECT)
         {
            result = sglBound::eINTERSECT;
         }
      }

      // restore the original pos and dir (leave len intact)
      seg.setLen(segset.getSegment().getLen());
      segset.setSegment(seg);
      segset.setModelViewMatrix(old_matrix);

      // restore the old view frustum
      if (segset.getMode()&(sglProbe::eUNSCALE_VIEW|
                            sglProbe::eUNPROJECT_VIEW))
      {
         segset.getViewFrustum() = old_view_frustum;
      }

      if (result == sglBound::eINTERSECT &&
          segset.getMode() & sglProbe::ePATH)
      {
         segset.getPath().push_front((sglNode*)this);
      }
      return result;
   }
   else
   {
      return sglBound::eOUTSIDE;
   }
}

//----------------------------------------------------------------------------
sglBound::IntersectResultEnum sglBillboard::intersect(
   sglIntersectd &segset) const
{
   if (!m_child_list.empty() && segset.getMode()&sglProbe::eBILLBOARD_VIEW)
   {
      sglSegmentd seg(segset.getSegment()); // copy the current segment

      // test against the current bounding sphere
      sglBound::IntersectResultEnum result = m_sphere.intersect(seg);
      if (result == sglBound::eOUTSIDE)
      {
         return result;
      }

      // save the original view matrix
      sglMat4d old_matrix = segset.getModelViewMatrix();

      // compute the new view matrix
      computeOrientation(old_matrix, segset.getModelViewMatrix());

      // store the original segment
      seg = segset.getSegment();

      // transform the original segment
      sglMat4d inv_mat;
      inv_mat.inverseAffine(segset.getModelViewMatrix());
      segset.mul(old_matrix);
      segset.mul(inv_mat);

      // compute a new view frustum
      sglFrustumd old_view_frustum;
      if (segset.getMode()&(sglProbe::eUNSCALE_VIEW|
                            sglProbe::eUNPROJECT_VIEW))
      {
         old_view_frustum = segset.getViewFrustum();
         segset.getViewFrustum().mul(segset.getOrigViewFrustum(), inv_mat);
      }

      // traverse the children
      result = sglBound::eOUTSIDE;
      for (vector<sglNode*>::const_iterator elt = m_child_list.begin();
           elt != m_child_list.end(); ++elt)
      {
         if ((*elt)->intersect(segset) == sglBound::eINTERSECT)
         {
            result = sglBound::eINTERSECT;
         }
      }

      // restore the original pos and dir (leave len intact)
      seg.setLen(segset.getSegment().getLen());
      segset.setSegment(seg);
      segset.setModelViewMatrix(old_matrix);

      // restore the old view frustum
      if (segset.getMode()&(sglProbe::eUNSCALE_VIEW|
                            sglProbe::eUNPROJECT_VIEW))
      {
         segset.getViewFrustum() = old_view_frustum;
      }

      if (result == sglBound::eINTERSECT &&
          segset.getMode() & sglProbe::ePATH)
      {
         segset.getPath().push_front((sglNode*)this);
      }
      return result;
   }
   else
   {
      return sglBound::eOUTSIDE;
   }
}

//----------------------------------------------------------------------------
void sglBillboard::pick(sglPickf &pick_state, unsigned int cull_flags) const
{
   if (!(pick_state.getMode()&sglProbe::eBILLBOARD_VIEW) ||
       m_child_list.empty() ||
       (cull_flags &&
        (m_sphere.intersect(pick_state.getPickFrustum(),
                            cull_flags) == sglBound::eOUTSIDE)))
   {
      return; // select nothing
   }
   else
   {
      // HACK - I don't know if this is right.  Must verify.

      //save the original view matrix
      sglMat4f old_matrix = pick_state.getModelViewMatrix();

      // compute a new matrix
      sglMat4f new_mat;
      computeOrientation(old_matrix, new_mat);

      // compute a new frustum - INEFFICIENT?
      sglFrustumf old_frustum = pick_state.getPickFrustum();
      sglMat4f inv_mat;
      inv_mat.inverseAffine(new_mat);

      pick_state.getPickFrustum().mul(pick_state.getOrigPickFrustum(), inv_mat);
      pick_state.setModelViewMatrix(new_mat);

      // compute a new view frustum
      sglFrustumf old_view_frustum;
      if (pick_state.getMode()&(sglProbe::eUNSCALE_VIEW|
                                sglProbe::eUNPROJECT_VIEW))
      {
         old_view_frustum = pick_state.getViewFrustum();
         pick_state.getViewFrustum().mul(pick_state.getOrigViewFrustum(),
                                         inv_mat);
      }

      // if path is requested store it ahead of time (different from
      // intersect), in case something beneath here is picked successfully.
      if (pick_state.getMode() & sglProbe::ePATH)
      {
         pick_state.getPath().push_back((sglNode*)this);
      }

      // --------------------------------------------------

      for (vector<sglNode*>::const_iterator elt = m_child_list.begin();
           elt != m_child_list.end(); ++elt)
      {
         (*elt)->pick(pick_state, cull_flags);
      }

      // --------------------------------------------------
      // restore old state
      pick_state.setModelViewMatrix(old_matrix);

      // restore the old view frustum
      if (pick_state.getMode()&(sglProbe::eUNSCALE_VIEW|
                                sglProbe::eUNPROJECT_VIEW))
      {
         pick_state.getViewFrustum() = old_view_frustum;
      }

      if (pick_state.getMode() & sglProbe::ePATH)
      {
         pick_state.getPath().pop_back();
      }

      pick_state.getPickFrustum() = old_frustum;
   }
}

//----------------------------------------------------------------------------
void sglBillboard::pick(sglPickd &pick_state, unsigned int cull_flags) const
{
   if (!(pick_state.getMode()&sglProbe::eBILLBOARD_VIEW) ||
       m_child_list.empty() ||
       (cull_flags &&
        (m_sphere.intersect(pick_state.getPickFrustum(),
                            cull_flags) == sglBound::eOUTSIDE)))
   {
      return; // select nothing
   }
   else
   {
      // HACK - I don't know if this is right.  Must verify.

      //save the original view matrix
      sglMat4d old_matrix = pick_state.getModelViewMatrix();

      // compute a new matrix
      sglMat4f new_mat;
      computeOrientation(old_matrix, new_mat);

      // compute a new pick frustum - INEFFICIENT?
      sglFrustumd old_frustum = pick_state.getPickFrustum();
      sglMat4d inv_mat;
      inv_mat.inverseAffine(new_mat);

      pick_state.getPickFrustum().mul(pick_state.getOrigPickFrustum(), inv_mat);
      pick_state.setModelViewMatrix(new_mat);

      // compute a new view frustum
      sglFrustumd old_view_frustum;
      if (pick_state.getMode()&(sglProbe::eUNSCALE_VIEW|
                                sglProbe::eUNPROJECT_VIEW))
      {
         old_view_frustum = pick_state.getViewFrustum();
         pick_state.getViewFrustum().mul(pick_state.getOrigViewFrustum(),
                                         inv_mat);
      }

      // if path is requested store it ahead of time (different from
      // intersect), in case something beneath here is picked successfully.
      if (pick_state.getMode() & sglProbe::ePATH)
      {
         pick_state.getPath().push_back((sglNode*)this);
      }

      // --------------------------------------------------

      for (vector<sglNode*>::const_iterator elt = m_child_list.begin();
           elt != m_child_list.end(); ++elt)
      {
         (*elt)->pick(pick_state, cull_flags);
      }

      // --------------------------------------------------
      // restore old state
      pick_state.setModelViewMatrix(old_matrix);

      // restore the old view frustum
      if (pick_state.getMode()&(sglProbe::eUNSCALE_VIEW|
                                sglProbe::eUNPROJECT_VIEW))
      {
         pick_state.getViewFrustum() = old_view_frustum;
      }

      if (pick_state.getMode() & sglProbe::ePATH)
      {
         pick_state.getPath().pop_back();
      }

      pick_state.getPickFrustum() = old_frustum;
   }
}

//----------------------------------------------------------------------------
void sglBillboard::cull(sglCull<float> &trav_state,
                        unsigned int cull_flags) const
{
   if (m_child_list.empty() ||
       (cull_flags &&
        (m_sphere.intersect(trav_state.getFrustum(),
                            cull_flags) == sglBound::eOUTSIDE)))
   {
      return; // draw nothing
   }
   else
   {
      // compute a new matrix
      sglMat4f new_mat;
      computeOrientation(trav_state.getModelViewMatrix(), new_mat);

      // compute a new frustum - VERY INEFFICIENT
      sglFrustumf old_frustum = trav_state.getFrustum();
      sglMat4f inv_mat;
      inv_mat.inverseAffine(new_mat);
      trav_state.getFrustum().mul(trav_state.getOrigFrustum(), inv_mat);

      trav_state.pushModelViewMatrix(new_mat, false, false);

      // --------------------------------------------------

      for (vector<sglNode*>::const_iterator elt = m_child_list.begin();
           elt != m_child_list.end(); ++elt)
      {
         (*elt)->cull(trav_state, cull_flags);
      }

      // --------------------------------------------------
      // restore old modelview matrix
      trav_state.popModelViewMatrix();
      trav_state.getFrustum() = old_frustum;
   }
}

//----------------------------------------------------------------------------
void sglBillboard::cull(sglCull<double> &trav_state,
                        unsigned int cull_flags) const
{
   if (m_child_list.empty() ||
       (cull_flags &&
        (m_sphere.intersect(trav_state.getFrustum(),
                            cull_flags) == sglBound::eOUTSIDE)))
   {
      return; // draw nothing
   }
   else
   {
      // compute a new matrix
      sglMat4d new_mat;
      computeOrientation(trav_state.getModelViewMatrix(), new_mat);

      // compute a new frustum - VERY INEFFICIENT
      sglFrustumd old_frustum = trav_state.getFrustum();
      sglMat4d inv_mat;
      inv_mat.inverseAffine(new_mat);
      trav_state.getFrustum().mul(trav_state.getOrigFrustum(), inv_mat);

      trav_state.pushModelViewMatrix(new_mat, false, false);

      // --------------------------------------------------

      for (vector<sglNode*>::const_iterator elt = m_child_list.begin();
           elt != m_child_list.end(); ++elt)
      {
         (*elt)->cull(trav_state, cull_flags);
      }

      // --------------------------------------------------
      // restore old modelview matrix
      trav_state.popModelViewMatrix();
      trav_state.getFrustum() = old_frustum;
   }
}

//----------------------------------------------------------------------------
sglNode *sglBillboard::clone(unsigned int mode) const
{
   sglBillboard *root = new sglBillboard;
   sglBillboard::copyTo(root, mode);
   return root;
}

//----------------------------------------------------------------------------
void sglBillboard::copyTo(sglBillboard *dst, unsigned int mode) const
{
   sglGroup::copyTo(dst, mode);

   // transfer stuff
   dst->setPos(m_pos);
   dst->setAxis(m_axis);
   dst->setMode(m_mode);
}

//----------------------------------------------------------------------------
void sglBillboard::printInfo(ostream &ostrm, const char *indent_string) const
{
    sglGroup::printInfo(ostrm, indent_string);

    ostrm << indent_string << "sglBillboard: mode = ";
    switch (m_mode)
    {
       case eAXIAL_ROT:
          ostrm << "eAXIAL_ROT" << endl;
          break;
       case ePOINT_ROT_EYE:
          ostrm << "ePOINT_ROT_EYE" << endl;
          break;
       case ePOINT_ROT_WORLD:
          ostrm << "ePOINT_ROT_WORLD" << endl;
          break;
    }
    ostrm << indent_string << "              axis = "
          << m_axis[0] << ' ' << m_axis[1] << ' ' << m_axis[2] << endl;
    ostrm << indent_string << "              pos  = "
          << m_pos[0] << ' ' << m_pos[1] << ' ' << m_pos[2] << endl;
}
