/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglBillboard.hpp
 *  Created: 2 December 1998
 *  Summary: transform the model view matrix to effect some sort of orientation
 *         : behaviour before traversing the children
 *****************************************************************************/

#ifndef __SGL_BILLBOARD_HPP
#define __SGL_BILLBOARD_HPP

#include <sgl.h>
#include <sglVector.hpp>
#include <sglMatrix.hpp>
#include <sglGroup.hpp>

//----------------------------------------------------------------------------

/**

The sglBillboard node modifies traversal modelview matrices to effect one of
three (currently two) orientation behaviours as a function of the view
direction before traversing the children.  This matrix can be modified by
setting one of the following modes:

 -# sglBillboard::ePOINT_ROT_EYE
In this mode, the modelview matrix is modified so that the origin of the
childrens' coordinate systems spins about a point in this node as specified by
the setPos() member function.  The childrens' local coordinate systems
will spin such that the Y-axes always stay upright on the screen and the Z-axes
point back to the user (out of the screen).  The default point of rotation
is this node's origin.
<p>
 -# sglBillboard::eAXIAL_ROT
In this mode, the modelview matrix is modified so that each child's Y-axis
system rotates about this node's axis as specified by the setAxis() member
function.  The childrens' coordinate systems rotate so that the childrens'
local Z-axes point back to the viewer as closely as possible.  As in the
ePOINT_ROT_EYE mode, the child's origin will rotate about the a point in the
sglBillboard's coordinate system as specified by the setPos() member
function.  The default axis of rotation is the sglBillboard's +Z-axis.
<p>
 -# sglBillboard::ePOINT_ROT_WORLD
Not implemented yet.

The default mode is ePOINT_ROT_EYE.

<b>Special behaviour for intersect and pick traversals</b>

In the default case, the intersect and pick traversals stop at sglBillboard
nodes.  If the user wishes to traverse these nodes,  the user must
set the modelview matrix in the sglProbe object (actually sglTraverser base
class) and add the sglProbe::eBILLBOARD_VIEW flag to its mode.

<b>Example:</b>
<pre>
   sglBillboard *billboard_node = new sglBillboard;

   sglVec3d pos(3.0, 0.0, 0.0);  // spin about a point along the x axis
   billboard_node->setPos(pos);
   billboard_node->setMode(sglBillboard::ePOINT_ROT_EYE);

   // or

   sglVec3d pos(3.0, 0.0, 0.0);    // origin at a point along the x axis
   billboard_node->setPos(pos);
   sglVec3d axis(1.0, 1.0, 0.0);   // will be normalized automatically
   billboard_node->setAxis(axis);  // rotate around the (1, 1, 0) axis
   billboard_node->setMode(sglBillboard::eAXIAL_ROT);
</pre>

\todo implement ePOINT_ROT_WORLD (what is the correct behaviour?).
\todo add a setMode() to switch between Z_UP/Y_UP (or whatever).  Possibly a
setChildAxis function to be completely generic?
\todo Check the computational complexity of eAXIAL_ROT and see if it can be
reduced.

*/

class SGL_DLL_API sglBillboard : public sglGroup
{
public:
   ///
   enum ModeEnum
   {
      ePOINT_ROT_EYE,
      ePOINT_ROT_WORLD,  // not yet implemented
      eAXIAL_ROT
   };

public:
   /// default constructor
   sglBillboard();
   /// virtual destructor
   virtual ~sglBillboard();

   /** Get the current point of spin for all modes.
       @return the current point of rotation.
   */
   const sglVec3d &getPos() const { return m_pos; }

   /** Set the point of spin for all modes.  The children's origins are
       translated to this position before rotation.
       @param p Point of spin with respect to this node's coordinate
       system.
   */
   void setPos(const sglVec3d &p);

   /** Get the current axis of rotation for the eAXIAL_ROT mode.
       @return axis of rotation with respect to sglBillboard's coordinate
       system.
   */
   const sglVec3d &getAxis() const { return m_axis; }

   /** Set the axis of rotation for the eAXIAL_ROT mode.
       @param v Axis of rotation with respect to sglBillboard's coordinate
       system.  This vector will be normalized.  Specifying a vector of length
       zero will result in the default of the Z-axis.
    */
   void setAxis(const sglVec3d &v);

   /** Specify the behaviour mode.
       @param mode Either ePOINT_ROT_EYE (the default) or eAXIAL_ROT.
   */
   void setMode(ModeEnum mode);

   /** Query the current behaviour mode.
       @return the current mode (either ePOINT_ROT_EYE or eAXIAL_ROT).
   */
   ModeEnum getMode() { return m_mode; }

   // see sglNode::intersect(f) for details.
   virtual sglBound::IntersectResultEnum intersect(
      sglIntersectf &isector) const;

   // see sglNode::intersect(d) for details.
   virtual sglBound::IntersectResultEnum intersect(
      sglIntersectd &isector) const;

   // see sglNode::pick(f) for details.
   virtual void pick(sglPickf &pick_state, unsigned int cull_flags) const;

   // see sglNode::pick(d) for details.
   virtual void pick(sglPickd &pick_state, unsigned int cull_flags) const;

   // see sglNode::clone() for details.
   virtual sglNode *clone(unsigned int mode) const;

   // see sglObject::printInfo() for details.
   virtual void printInfo(ostream &ostrm, const char *indent_string) const;

protected:
   void computeOrientation(const sglMat4f &current_mat,
                           sglMat4f &new_mat) const;
   void computeOrientation(const sglMat4d &current_mat,
                           sglMat4d &new_mat) const;

   // compute the bounding sphere
   virtual void computeBound();

   // sglNode::cull(f) for details
   virtual void cull(sglCull<float> &trav_state,
                     unsigned int cull_flags) const;

   // sglNode::cull(d) for details
   virtual void cull(sglCull<double> &trav_state,
                     unsigned int cull_flags) const;

   void copyTo(sglBillboard *dst, unsigned int mode) const;

private:  // not implemented
   sglBillboard(const sglBillboard &);
   sglBillboard &operator=(const sglBillboard &);

protected:
   ModeEnum m_mode;
   sglVec3d m_pos;
   sglVec3d m_axis;
};

#endif
