/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglBitMaskSwitch.cpp
 *   Author: Scott McMillan
 *  Created: 18 May 2001
 *  Summary: selects which child (or none or all children) to be traversed
 *****************************************************************************/

#include <sgl.h>
#include <sglBitMaskSwitch.hpp>
#include <sglCull.hpp>
#include <sglOStream.hpp>

//----------------------------------------------------------------------------
sglBitMaskSwitch::sglBitMaskSwitch() : sglGroup(),
   m_switch_mask(~0x0U) // defaults to a sglGroup
{
}

//----------------------------------------------------------------------------
sglBitMaskSwitch::~sglBitMaskSwitch()
{
}

//----------------------------------------------------------------------------
sglBound::IntersectResultEnum sglBitMaskSwitch::intersect(
   sglIntersectf &isector) const
{
   if ((isector.getMode()&sglProbe::eSWITCH_NONE) ||
       (m_switch_mask == 0) || m_child_list.empty())
   {
      return sglBound::eOUTSIDE;
   }
   else if ((isector.getMode()&sglProbe::eSWITCH_ALL) ||
            (m_switch_mask == ~0x0U))
   {
      return sglGroup::intersect(isector);
   }
   else // default: current children selected by switch only
   {
      sglSegmentf seg(isector.getSegment());
      sglBound::IntersectResultEnum result = m_sphere.intersect(seg);
      if (result == sglBound::eOUTSIDE)
      {
         return result;
      }

      result = sglBound::eOUTSIDE;
      unsigned int mask = 0x1;
      for (vector<sglNode*>::const_iterator elt = m_child_list.begin();
           (elt != m_child_list.end()) &&
              (mask <= m_switch_mask) && (mask > 0);
           ++elt)
      {
         if (mask & m_switch_mask)
         {
            if ((*elt)->intersect(isector) == sglBound::eINTERSECT)
            {
               result = sglBound::eINTERSECT;
            }
         }
         mask = mask << 1;
      }

      if (result == sglBound::eINTERSECT &&
          isector.getMode() & sglProbe::ePATH)
      {
         isector.getPath().push_front((sglNode*)this);
      }
      return result;
   }
}

//----------------------------------------------------------------------------
sglBound::IntersectResultEnum sglBitMaskSwitch::intersect(
   sglIntersectd &isector) const
{
   sglPrintDebug() << "sglBitMaskSwitch::intersect(sglIntersectd&) begin "
                   << this << endl;

   if ((isector.getMode()&sglProbe::eSWITCH_NONE) ||
       (m_switch_mask == 0) || m_child_list.empty())
   {
      return sglBound::eOUTSIDE;
   }
   else if ((isector.getMode()&sglProbe::eSWITCH_ALL) ||
            (m_switch_mask == ~0x0U))
   {
      return sglGroup::intersect(isector);
   }
   else // default: current children selected by switch only
   {
      sglSegmentd seg(isector.getSegment());
      sglBound::IntersectResultEnum result = m_sphere.intersect(seg);
      if (result == sglBound::eOUTSIDE)
      {
         return result;
      }

      result = sglBound::eOUTSIDE;
      unsigned int mask = 0x1;
      for (vector<sglNode*>::const_iterator elt = m_child_list.begin();
           (elt != m_child_list.end()) &&
              (mask <= m_switch_mask) && (mask > 0);
           ++elt)
      {
         if (mask & m_switch_mask)
         {
            if ((*elt)->intersect(isector) == sglBound::eINTERSECT)
            {
               result = sglBound::eINTERSECT;
            }
         }
         mask = mask << 1;
      }

      if (result == sglBound::eINTERSECT &&
          isector.getMode() & sglProbe::ePATH)
      {
         isector.getPath().push_front((sglNode*)this);
      }
      return result;
   }
}

//----------------------------------------------------------------------------
void sglBitMaskSwitch::pick(sglPickf &pick_state,
                            unsigned int cull_flags) const
{
   if ((pick_state.getMode()&sglProbe::eSWITCH_NONE) ||
       (m_switch_mask == 0) || m_child_list.empty())
   {
      return;  // select nothing
   }
   else if ((pick_state.getMode()&sglProbe::eSWITCH_ALL) ||
            (m_switch_mask == ~0x0U))
   {
      sglGroup::pick(pick_state, cull_flags);
   }
   else // default: current children selected by switch only
   {
      if (cull_flags && (m_sphere.intersect(pick_state.getPickFrustum(),
                                            cull_flags) == sglBound::eOUTSIDE))
      {
         return; // select nothing
      }

      unsigned int mask = 0x1;
      for (vector<sglNode*>::const_iterator elt = m_child_list.begin();
           (elt != m_child_list.end()) &&
              (mask <= m_switch_mask) && (mask > 0);
           ++elt)
      {
         if (mask & m_switch_mask)
         {
            // if path is requested store it ahead of time (different from
            // intersect), in case something beneath here is picked
            // successfully.
            if (pick_state.getMode() & sglProbe::ePATH)
            {
               pick_state.getPath().push_back((sglNode*)this);
            }

            (*elt)->pick(pick_state, cull_flags);

            if (pick_state.getMode() & sglProbe::ePATH)
            {
               pick_state.getPath().pop_back();
            }
         }
         mask = mask << 1;
      }
   }
}

//----------------------------------------------------------------------------
void sglBitMaskSwitch::pick(sglPickd &pick_state, unsigned int cull_flags) const
{
   if ((pick_state.getMode()&sglProbe::eSWITCH_NONE) ||
       (m_switch_mask == 0) || m_child_list.empty())
   {
      return;  // select nothing
   }
   else if ((pick_state.getMode()&sglProbe::eSWITCH_ALL) ||
            (m_switch_mask == ~0x0U))
   {
      sglGroup::pick(pick_state, cull_flags);
   }
   else // default: current children selected by switch only
   {
      if (cull_flags && (m_sphere.intersect(pick_state.getPickFrustum(),
                                            cull_flags) == sglBound::eOUTSIDE))
      {
         return; // select nothing
      }

      unsigned int mask = 0x1;
      for (vector<sglNode*>::const_iterator elt = m_child_list.begin();
           (elt != m_child_list.end()) &&
              (mask <= m_switch_mask) && (mask > 0);
           ++elt)
      {
         if (mask & m_switch_mask)
         {
            // if path is requested store it ahead of time (different from
            // intersect), in case something beneath here is picked
            // successfully.
            if (pick_state.getMode() & sglProbe::ePATH)
            {
               pick_state.getPath().push_back((sglNode*)this);
            }

            (*elt)->pick(pick_state, cull_flags);

            if (pick_state.getMode() & sglProbe::ePATH)
            {
               pick_state.getPath().pop_back();
            }
         }
         mask = mask << 1;
      }
   }
}

//----------------------------------------------------------------------------
void sglBitMaskSwitch::cull(sglCull<float> &trav_state,
                            unsigned int cull_flags,
                            unsigned int switch_mask) const
{
   if (switch_mask == 0 || m_child_list.empty())
   {
      return; // draw nothing
   }
   else if (switch_mask == ~0x0U)
   {
      sglGroup::cull(trav_state, cull_flags);
   }
   else
   {
      if (cull_flags && (m_sphere.intersect(trav_state.getFrustum(),
                                            cull_flags) == sglBound::eOUTSIDE))
      {
         return; // draw nothing
      }

      unsigned int mask = 0x1;
      for (vector<sglNode*>::const_iterator elt = m_child_list.begin();
           (elt != m_child_list.end()) &&
              (mask <= m_switch_mask) && (mask > 0);
           ++elt)
      {
         if (mask & m_switch_mask)
         {
            (*elt)->cull(trav_state, cull_flags);
         }
         mask = mask << 1;
      }
   }
}

//----------------------------------------------------------------------------
void sglBitMaskSwitch::cull(sglCull<double> &trav_state,
                            unsigned int cull_flags,
                            unsigned int switch_mask) const
{
   if (switch_mask == 0 || m_child_list.empty())
   {
      return; // draw nothing
   }
   else if (switch_mask == ~0x0U)
   {
      sglGroup::cull(trav_state, cull_flags);
   }
   else
   {
      if (cull_flags && (m_sphere.intersect(trav_state.getFrustum(),
                                            cull_flags) == sglBound::eOUTSIDE))
      {
         return; // draw nothing
      }

      unsigned int mask = 0x1;
      for (vector<sglNode*>::const_iterator elt = m_child_list.begin();
           (elt != m_child_list.end()) &&
              (mask <= m_switch_mask) && (mask > 0);
           ++elt)
      {
         if (mask & m_switch_mask)
         {
            (*elt)->cull(trav_state, cull_flags);
         }
         mask = mask << 1;
      }
   }
}

//----------------------------------------------------------------------------
sglNode *sglBitMaskSwitch::clone(unsigned int mode) const
{
   sglBitMaskSwitch *root = new sglBitMaskSwitch;
   sglBitMaskSwitch::copyTo(root, mode);
   return root;
}

//----------------------------------------------------------------------------
void sglBitMaskSwitch::copyTo(sglBitMaskSwitch *dst, unsigned int mode) const
{
   sglGroup::copyTo(dst, mode);

   // transfer stuff
   dst->setMask(m_switch_mask);
}

//----------------------------------------------------------------------------
void sglBitMaskSwitch::printInfo(ostream &ostrm,
                                 const char *indent_string) const
{
    sglGroup::printInfo(ostrm, indent_string);

    ostrm << indent_string << "sglBitMaskSwitch val = " << hex
          << m_switch_mask << dec << endl;
}
