/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglBitMaskSwitch.hpp
 *   Author: Scott McMillan
 *  Created: 18 May 2001
 *  Summary: select (or not) the children to be traversed
 *****************************************************************************/

#ifndef __SGL_BIT_MASK_SWITCH_HPP
#define __SGL_BIT_MASK_SWITCH_HPP

#include <sgl.h>
#include <sglGroup.hpp>

//----------------------------------------------------------------------------

/**

This will select any combination of up to 32 children to be traversed.  This is
done according to the values of the bits in this node's mask where one bit
corresponds for each child (i.e., bit 0 toggles child 0, bit 1 toggles child 1,
etc.).  As with sglSwitch, the order that the children are added to the
node is significant.

The switch mask is an unsigned integer which means it can be used to select
one of up to 32 children.  If this node has less than 32 children some of the
bits are out of range and will be ignored.

<b>Special behaviour for intersect and pick traversals</b>

In addition to this behaviour, the intersect and pick traversals of this node
type can be overridden by adding sglProbe::eSWITCH_ALL or
sglProbe::eSWITCH_NONE to the mode.  The former behaves as if mask 0xffffffff
(~0x0) has been set on this node, and the latter corresponds to a mask of 0x0.

<b>Example:</b>

<pre>
   sglBitMaskSwitch *switch_node = new sglBitMaskSwitch;
   sglGeode *geode0 = new sglGeode;
   sglGeode *geode1 = new sglGeode;

   switch_node->addChild(geode0);
   switch_node->addChild(geode1);

   switch_node->setMask(0x0);     // select neither
   switch_node->setMask(0x1);     // select geode0
   switch_node->setMask(0x2);     // select geode1
   switch_node->setMask(0x3);     // select both
   switch_node->setMask(0x4);     // selects neither (no corresponding child)
   switch_node->setMask(~0x0);    // select all children (the default)
</pre>

 */

//----------------------------------------------------------------------------

class SGL_DLL_API sglBitMaskSwitch : public sglGroup
{
public:
   /// default constructor
   sglBitMaskSwitch();
   /// destructor
   virtual ~sglBitMaskSwitch();

   /** Set the bit mask for selecting children.
       @mask The mask where each bit corresponds to one of up to 32 children.
   */
   void setMask(unsigned int mask) { m_switch_mask = mask; }

   /** Get the current bit mask.
       @returns mask The unsigned int bit mask.
   */
   unsigned int getMask()          { return m_switch_mask; }

   // See sglNode::intersect(f) for details.
   virtual sglBound::IntersectResultEnum intersect(
      sglIntersectf &isector) const;

   // See sglNode::intersect(d) for details.
   virtual sglBound::IntersectResultEnum intersect(
      sglIntersectd &isector) const;

   // See sglNode::pick(f) for details.
   virtual void pick(sglPickf &pick_state, unsigned int cull_flags) const;

   // See sglNode::pick(d) for details.
   virtual void pick(sglPickd &pick_state, unsigned int cull_flags) const;

   // See sglNode::clone() for details.
   virtual sglNode *clone(unsigned int mode) const;

   // See sglObject::printInfo() for details.
   virtual void printInfo(ostream &ostrm, const char *indent_string) const;

protected:
   // See sglNode::cull(f) for details
   virtual void cull(sglCull<float> &trav_state,
                     unsigned int cull_flags) const
      {
         cull(trav_state, cull_flags, m_switch_mask);
      }

   // See sglNode::cull(d) for details
   virtual void cull(sglCull<double> &trav_state,
                     unsigned int cull_flags) const
      {
         cull(trav_state, cull_flags, m_switch_mask);
      }

   // reentrant versions of the cull functions for multithreading
   void cull(sglCull<float> &trav_state,
             unsigned int cull_flags, unsigned int switch_mask) const;
   void cull(sglCull<double> &trav_state,
             unsigned int cull_flags, unsigned int switch_mask) const;

   void copyTo(sglBitMaskSwitch *dst, unsigned int mode) const;

private:
   sglBitMaskSwitch(const sglBitMaskSwitch &);            // not implemented
   sglBitMaskSwitch &operator=(const sglBitMaskSwitch &); // not implemented

protected:
   unsigned int m_switch_mask;
};

#endif
