/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglBitmapFont.cpp
 *   Author: Tom Stimson
 *  Created: 20 April 2000
 *****************************************************************************/

#include <sglBitmapFontWGL.hpp>
#include <sglBitmapFont.hpp>
#include <sglOStream.hpp>

sglBitmapFont *sglBitmapFont::s_default_font = NULL;


//----------------------------------------------------------------------------
void sglBitmapFont::setDefaultFont(sglBitmapFont* font)
{
   if (s_default_font)
   {
      s_default_font->deref();
      if (s_default_font->getRefCount() == 0)
         delete s_default_font;
   }
   if (font)
      font->ref();
   s_default_font = font;
}

//----------------------------------------------------------------------------
unsigned int sglBitmapFont::getStringWidth(const string &str) const
{
   unsigned int width = 0;
   for (string::const_iterator it = str.begin(); it != str.end(); ++it)
   {
      width += getCharWidth(*it);
   }
   return width;
}


#if defined(WIN32)
//----------------------------------------------------------------------------
sglBitmapFontWGL::sglBitmapFontWGL(HFONT font)
   : m_font(font),
     m_list_base(0),
     m_height(0),
     m_descent(0)
{
   m_list_base = glGenLists(128);
   if (m_list_base == 0)
   {
      sglPrintWarning() << "glGenLists failed during font creation: " << endl;
      return;
   }

   HDC hdc = wglGetCurrentDC();

   HFONT old_font = (HFONT)SelectObject(hdc, m_font);
   if (old_font)
   {
      TEXTMETRIC metrics;
      if (GetTextMetrics(hdc, &metrics) == TRUE)
      {
         m_height = metrics.tmHeight - metrics.tmInternalLeading;
         m_descent = metrics.tmDescent;

         ABC abcWidths[128];
         if (GetCharABCWidths(hdc, 0, 127, abcWidths) != 0)
         {
            for (int ii=0; ii<128; ++ii)
            {
               m_widths[ii] = abcWidths[ii].abcA
                  + abcWidths[ii].abcB + abcWidths[ii].abcC;
            }
         }
         else if(GetCharWidth(hdc, 0, 127, (LPINT)m_widths) == 0)
         {
            sglPrintWarning()
               << "GetCharWidth() failed: "
               << sglOStream::formatWIN32Error(GetLastError()) << endl;
            for (int ii=0; ii<128; ++ii)
            {
               m_widths[ii] = 1;
            }
         }
      }
      else
      {
         sglPrintWarning()
            << "GetTextMetrics() failed: "
            << sglOStream::formatWIN32Error(GetLastError()) << endl;
      }

      if (wglUseFontBitmaps(hdc, 0, 128, m_list_base) != TRUE)
      {
         sglPrintWarning()
            << "wglUseFontBitmaps() failed: "
            << sglOStream::formatWIN32Error(GetLastError()) << endl;
      }

      SelectObject(hdc, old_font);
   }
   else
   {
      sglPrintWarning()
         << "Unable to SelectObject during font creation: "
         << sglOStream::formatWIN32Error(GetLastError()) << endl;
   }
}

//----------------------------------------------------------------------------
sglBitmapFontWGL::~sglBitmapFontWGL()
{
   glDeleteLists(m_list_base, 128);
}

//----------------------------------------------------------------------------
sglBitmapFont* sglBitmapFontWGL::clone() const
{
   return new sglBitmapFontWGL(m_font);
}

//----------------------------------------------------------------------------
void sglBitmapFontWGL::rasterString(const char *str, int size) const
{
   glListBase(m_list_base);
   if (size < 0)
      glCallLists(strlen(str), GL_BYTE, (const GLvoid*)str);
   else if (size > 0)
      glCallLists(size, GL_BYTE, (const GLvoid*)str);
}

//----------------------------------------------------------------------------
unsigned int sglBitmapFontWGL::getHeight() const
{
   return m_height;
}

//----------------------------------------------------------------------------
unsigned int sglBitmapFontWGL::getDescent() const
{
   return m_descent;
}

//----------------------------------------------------------------------------
unsigned int sglBitmapFontWGL::getCharWidth(char c) const
{
   if (c < 0 || c > 128)
      return 0;
   else
      return m_widths[c];
}
#endif

#if defined( MAC_OSX )

// add mac code here

#elif !defined(WIN32)
# include <GL/glx.h>
//----------------------------------------------------------------------------
sglBitmapFontGLX::sglBitmapFontGLX(Font font)
   : m_font(font),
     m_list_base(0),
     m_font_info(NULL)
{
   Display *dpy = glXGetCurrentDisplay();
   m_font_info = XQueryFont(dpy, m_font);
   if (m_font_info == NULL)
   {
      sglPrintWarning() << "XQueryFont() failed during font creation" << endl;
      return;
   }

   m_list_base = glGenLists(128);
   if (m_list_base == 0)
   {
      sglPrintWarning() << "glGenLists failed during font creation" << endl;
      return;
   }

   glXUseXFont(m_font, 0, 128, m_list_base);
   glFinish();
}

//----------------------------------------------------------------------------
sglBitmapFontGLX::~sglBitmapFontGLX()
{
   if (m_list_base)
      glDeleteLists(m_list_base, 128);
   if (m_font_info)
      XFreeFontInfo(NULL, m_font_info, 1);
}

//----------------------------------------------------------------------------
sglBitmapFont* sglBitmapFontGLX::clone() const
{
   return new sglBitmapFontGLX(m_font);
}

//----------------------------------------------------------------------------
void sglBitmapFontGLX::rasterString(const char *str, int size) const
{
   glListBase(m_list_base);
   if (size < 0)
      glCallLists(strlen(str), GL_BYTE, (const GLvoid*)str);
   else if (size > 0)
      glCallLists(size, GL_BYTE, (const GLvoid*)str);
}

//----------------------------------------------------------------------------
unsigned int sglBitmapFontGLX::getHeight() const
{
   if (m_font_info)
      return m_font_info->ascent + m_font_info->descent;
   else
      return 12; // HACK Not sure what to do here.
}

//----------------------------------------------------------------------------
unsigned int sglBitmapFontGLX::getDescent() const
{
   if (m_font_info)
      return m_font_info->descent;
   else
      return 2; // HACK Not sure what to do here.
}

//----------------------------------------------------------------------------
unsigned int sglBitmapFontGLX::getCharWidth(char c) const
{
   if (m_font_info)
      return XTextWidth(m_font_info, &c, 1);
   else
      return 8; // HACK Not sure what to do here.
}

//----------------------------------------------------------------------------
unsigned int sglBitmapFontGLX::getStringWidth(const string &str) const
{
   if (m_font_info)
      return XTextWidth(m_font_info, str.c_str(), str.size());
   else
      return 8*str.size(); // HACK Not sure what to do here.
}
#endif


//----------------------------------------------------------------------------
// This was generated from the output of calling showfont on the 8x13 font
// that comes with X.  It had this copyright message:
//   COPYRIGHT       Public domain font.  Share and enjoy.
// Note this actually contains 256 characters, but we only use the first 128.
static GLubyte sglBitmapFont8x13Letters[][13] = {
   {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00},
   {0x00, 0x00, 0x00, 0x10, 0x38, 0x7c, 0xfe, 0x7c, 0x38, 0x10, 0x00, 0x00, 0x00},
   {0x55, 0xaa, 0x55, 0xaa, 0x55, 0xaa, 0x55, 0xaa, 0x55, 0xaa, 0x55, 0xaa, 0x00},
   {0x08, 0x08, 0x08, 0x3e, 0x88, 0x88, 0xf8, 0x88, 0x88, 0x00, 0x00, 0x00, 0x00},
   {0x10, 0x10, 0x1c, 0x10, 0x9e, 0x80, 0xe0, 0x80, 0xf0, 0x00, 0x00, 0x00, 0x00},
   {0x22, 0x22, 0x3c, 0x22, 0x3c, 0x78, 0x80, 0x80, 0x78, 0x00, 0x00, 0x00, 0x00},
   {0x20, 0x20, 0x3c, 0x20, 0x3e, 0xf8, 0x80, 0x80, 0x80, 0x00, 0x00, 0x00, 0x00},
   {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x38, 0x44, 0x44, 0x38, 0x00, 0x00},
   {0x00, 0x00, 0xfe, 0x10, 0x10, 0xfe, 0x10, 0x10, 0x00, 0x00, 0x00, 0x00, 0x00},
   {0x3e, 0x20, 0x20, 0x20, 0x88, 0x98, 0xa8, 0xc8, 0x88, 0x00, 0x00, 0x00, 0x00},
   {0x08, 0x08, 0x08, 0x08, 0x3e, 0x20, 0x50, 0x88, 0x88, 0x00, 0x00, 0x00, 0x00},
   {0x00, 0x00, 0x00, 0x00, 0x00, 0xf0, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10},
   {0x10, 0x10, 0x10, 0x10, 0x10, 0xf0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00},
   {0x10, 0x10, 0x10, 0x10, 0x10, 0x1f, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00},
   {0x00, 0x00, 0x00, 0x00, 0x00, 0x1f, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10},
   {0x10, 0x10, 0x10, 0x10, 0x10, 0xff, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10},
   {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xff, 0x00, 0x00, 0x00},
   {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xff, 0x00, 0x00, 0x00, 0x00, 0x00},
   {0x00, 0x00, 0x00, 0x00, 0x00, 0xff, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00},
   {0x00, 0x00, 0x00, 0xff, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00},
   {0x00, 0xff, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00},
   {0x10, 0x10, 0x10, 0x10, 0x10, 0x1f, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10},
   {0x10, 0x10, 0x10, 0x10, 0x10, 0xf0, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10},
   {0x00, 0x00, 0x00, 0x00, 0x00, 0xff, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10},
   {0x10, 0x10, 0x10, 0x10, 0x10, 0xff, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00},
   {0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10},
   {0x00, 0x00, 0xfe, 0x02, 0x08, 0x20, 0x80, 0x20, 0x08, 0x02, 0x00, 0x00, 0x00},
   {0x00, 0x00, 0xfe, 0x80, 0x20, 0x08, 0x02, 0x08, 0x20, 0x80, 0x00, 0x00, 0x00},
   {0x00, 0x00, 0x44, 0x24, 0x24, 0x24, 0x24, 0x7e, 0x00, 0x00, 0x00, 0x00, 0x00},
   {0x00, 0x00, 0x80, 0x40, 0xfe, 0x10, 0xfe, 0x04, 0x02, 0x00, 0x00, 0x00, 0x00},
   {0x00, 0x00, 0xdc, 0x62, 0x20, 0x20, 0x20, 0x70, 0x20, 0x22, 0x1c, 0x00, 0x00},
   {0x00, 0x00, 0x00, 0x00, 0x00, 0x10, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00},
   {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00},
   {0x00, 0x00, 0x10, 0x00, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x00, 0x00},
   {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x24, 0x24, 0x24, 0x00, 0x00},
   {0x00, 0x00, 0x00, 0x24, 0x24, 0x7e, 0x24, 0x7e, 0x24, 0x24, 0x00, 0x00, 0x00},
   {0x00, 0x00, 0x00, 0x10, 0x78, 0x14, 0x38, 0x50, 0x3c, 0x10, 0x00, 0x00, 0x00},
   {0x00, 0x00, 0x44, 0x2a, 0x24, 0x10, 0x08, 0x08, 0x24, 0x52, 0x22, 0x00, 0x00},
   {0x00, 0x00, 0x3a, 0x44, 0x4a, 0x30, 0x48, 0x48, 0x30, 0x00, 0x00, 0x00, 0x00},
   {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x40, 0x30, 0x38, 0x00, 0x00},
   {0x00, 0x00, 0x04, 0x08, 0x08, 0x10, 0x10, 0x10, 0x08, 0x08, 0x04, 0x00, 0x00},
   {0x00, 0x00, 0x20, 0x10, 0x10, 0x08, 0x08, 0x08, 0x10, 0x10, 0x20, 0x00, 0x00},
   {0x00, 0x00, 0x00, 0x00, 0x24, 0x18, 0x7e, 0x18, 0x24, 0x00, 0x00, 0x00, 0x00},
   {0x00, 0x00, 0x00, 0x00, 0x10, 0x10, 0x7c, 0x10, 0x10, 0x00, 0x00, 0x00, 0x00},
   {0x00, 0x40, 0x30, 0x38, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00},
   {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x7e, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00},
   {0x00, 0x10, 0x38, 0x10, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00},
   {0x00, 0x00, 0x80, 0x80, 0x40, 0x20, 0x10, 0x08, 0x04, 0x02, 0x02, 0x00, 0x00},
   {0x00, 0x00, 0x18, 0x24, 0x42, 0x42, 0x42, 0x42, 0x42, 0x24, 0x18, 0x00, 0x00},
   {0x00, 0x00, 0x7c, 0x10, 0x10, 0x10, 0x10, 0x10, 0x50, 0x30, 0x10, 0x00, 0x00},
   {0x00, 0x00, 0x7e, 0x40, 0x20, 0x18, 0x04, 0x02, 0x42, 0x42, 0x3c, 0x00, 0x00},
   {0x00, 0x00, 0x3c, 0x42, 0x02, 0x02, 0x1c, 0x08, 0x04, 0x02, 0x7e, 0x00, 0x00},
   {0x00, 0x00, 0x04, 0x04, 0x7e, 0x44, 0x44, 0x24, 0x14, 0x0c, 0x04, 0x00, 0x00},
   {0x00, 0x00, 0x3c, 0x42, 0x02, 0x02, 0x62, 0x5c, 0x40, 0x40, 0x7e, 0x00, 0x00},
   {0x00, 0x00, 0x3c, 0x42, 0x42, 0x62, 0x5c, 0x40, 0x40, 0x20, 0x1c, 0x00, 0x00},
   {0x00, 0x00, 0x20, 0x20, 0x10, 0x10, 0x08, 0x08, 0x04, 0x02, 0x7e, 0x00, 0x00},
   {0x00, 0x00, 0x3c, 0x42, 0x42, 0x42, 0x3c, 0x42, 0x42, 0x42, 0x3c, 0x00, 0x00},
   {0x00, 0x00, 0x38, 0x04, 0x02, 0x02, 0x3a, 0x46, 0x42, 0x42, 0x3c, 0x00, 0x00},
   {0x00, 0x10, 0x38, 0x10, 0x00, 0x00, 0x10, 0x38, 0x10, 0x00, 0x00, 0x00, 0x00},
   {0x00, 0x40, 0x30, 0x38, 0x00, 0x00, 0x10, 0x38, 0x10, 0x00, 0x00, 0x00, 0x00},
   {0x00, 0x00, 0x02, 0x04, 0x08, 0x10, 0x20, 0x10, 0x08, 0x04, 0x02, 0x00, 0x00},
   {0x00, 0x00, 0x00, 0x00, 0x7e, 0x00, 0x00, 0x7e, 0x00, 0x00, 0x00, 0x00, 0x00},
   {0x00, 0x00, 0x40, 0x20, 0x10, 0x08, 0x04, 0x08, 0x10, 0x20, 0x40, 0x00, 0x00},
   {0x00, 0x00, 0x08, 0x00, 0x08, 0x08, 0x04, 0x02, 0x42, 0x42, 0x3c, 0x00, 0x00},
   {0x00, 0x00, 0x3c, 0x40, 0x4a, 0x56, 0x52, 0x4e, 0x42, 0x42, 0x3c, 0x00, 0x00},
   {0x00, 0x00, 0x42, 0x42, 0x42, 0x7e, 0x42, 0x42, 0x42, 0x24, 0x18, 0x00, 0x00},
   {0x00, 0x00, 0xfc, 0x42, 0x42, 0x42, 0x7c, 0x42, 0x42, 0x42, 0xfc, 0x00, 0x00},
   {0x00, 0x00, 0x3c, 0x42, 0x40, 0x40, 0x40, 0x40, 0x40, 0x42, 0x3c, 0x00, 0x00},
   {0x00, 0x00, 0xfc, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0xfc, 0x00, 0x00},
   {0x00, 0x00, 0x7e, 0x40, 0x40, 0x40, 0x78, 0x40, 0x40, 0x40, 0x7e, 0x00, 0x00},
   {0x00, 0x00, 0x40, 0x40, 0x40, 0x40, 0x78, 0x40, 0x40, 0x40, 0x7e, 0x00, 0x00},
   {0x00, 0x00, 0x3a, 0x46, 0x42, 0x4e, 0x40, 0x40, 0x40, 0x42, 0x3c, 0x00, 0x00},
   {0x00, 0x00, 0x42, 0x42, 0x42, 0x42, 0x7e, 0x42, 0x42, 0x42, 0x42, 0x00, 0x00},
   {0x00, 0x00, 0x7c, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x7c, 0x00, 0x00},
   {0x00, 0x00, 0x38, 0x44, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x1e, 0x00, 0x00},
   {0x00, 0x00, 0x42, 0x44, 0x48, 0x50, 0x60, 0x50, 0x48, 0x44, 0x42, 0x00, 0x00},
   {0x00, 0x00, 0x7e, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x00, 0x00},
   {0x00, 0x00, 0x82, 0x82, 0x82, 0x92, 0x92, 0xaa, 0xc6, 0x82, 0x82, 0x00, 0x00},
   {0x00, 0x00, 0x42, 0x42, 0x42, 0x46, 0x4a, 0x52, 0x62, 0x42, 0x42, 0x00, 0x00},
   {0x00, 0x00, 0x3c, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3c, 0x00, 0x00},
   {0x00, 0x00, 0x40, 0x40, 0x40, 0x40, 0x7c, 0x42, 0x42, 0x42, 0x7c, 0x00, 0x00},
   {0x00, 0x02, 0x3c, 0x4a, 0x52, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3c, 0x00, 0x00},
   {0x00, 0x00, 0x42, 0x44, 0x48, 0x50, 0x7c, 0x42, 0x42, 0x42, 0x7c, 0x00, 0x00},
   {0x00, 0x00, 0x3c, 0x42, 0x02, 0x02, 0x3c, 0x40, 0x40, 0x42, 0x3c, 0x00, 0x00},
   {0x00, 0x00, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0xfe, 0x00, 0x00},
   {0x00, 0x00, 0x3c, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x00, 0x00},
   {0x00, 0x00, 0x10, 0x28, 0x28, 0x28, 0x44, 0x44, 0x44, 0x82, 0x82, 0x00, 0x00},
   {0x00, 0x00, 0x44, 0xaa, 0x92, 0x92, 0x92, 0x82, 0x82, 0x82, 0x82, 0x00, 0x00},
   {0x00, 0x00, 0x82, 0x82, 0x44, 0x28, 0x10, 0x28, 0x44, 0x82, 0x82, 0x00, 0x00},
   {0x00, 0x00, 0x10, 0x10, 0x10, 0x10, 0x10, 0x28, 0x44, 0x82, 0x82, 0x00, 0x00},
   {0x00, 0x00, 0x7e, 0x40, 0x40, 0x20, 0x10, 0x08, 0x04, 0x02, 0x7e, 0x00, 0x00},
   {0x00, 0x00, 0x3c, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x3c, 0x00, 0x00},
   {0x00, 0x00, 0x02, 0x02, 0x04, 0x08, 0x10, 0x20, 0x40, 0x80, 0x80, 0x00, 0x00},
   {0x00, 0x00, 0x78, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x78, 0x00, 0x00},
   {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x44, 0x28, 0x10, 0x00, 0x00},
   {0x00, 0xfe, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00},
   {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x04, 0x18, 0x38, 0x00, 0x00},
   {0x00, 0x00, 0x3a, 0x46, 0x42, 0x3e, 0x02, 0x3c, 0x00, 0x00, 0x00, 0x00, 0x00},
   {0x00, 0x00, 0x5c, 0x62, 0x42, 0x42, 0x62, 0x5c, 0x40, 0x40, 0x40, 0x00, 0x00},
   {0x00, 0x00, 0x3c, 0x42, 0x40, 0x40, 0x42, 0x3c, 0x00, 0x00, 0x00, 0x00, 0x00},
   {0x00, 0x00, 0x3a, 0x46, 0x42, 0x42, 0x46, 0x3a, 0x02, 0x02, 0x02, 0x00, 0x00},
   {0x00, 0x00, 0x3c, 0x42, 0x40, 0x7e, 0x42, 0x3c, 0x00, 0x00, 0x00, 0x00, 0x00},
   {0x00, 0x00, 0x20, 0x20, 0x20, 0x20, 0x7c, 0x20, 0x20, 0x22, 0x1c, 0x00, 0x00},
   {0x3c, 0x42, 0x3c, 0x40, 0x38, 0x44, 0x44, 0x3a, 0x00, 0x00, 0x00, 0x00, 0x00},
   {0x00, 0x00, 0x42, 0x42, 0x42, 0x42, 0x62, 0x5c, 0x40, 0x40, 0x40, 0x00, 0x00},
   {0x00, 0x00, 0x7c, 0x10, 0x10, 0x10, 0x10, 0x30, 0x00, 0x10, 0x00, 0x00, 0x00},
   {0x38, 0x44, 0x44, 0x04, 0x04, 0x04, 0x04, 0x0c, 0x00, 0x04, 0x00, 0x00, 0x00},
   {0x00, 0x00, 0x42, 0x44, 0x48, 0x70, 0x48, 0x44, 0x40, 0x40, 0x40, 0x00, 0x00},
   {0x00, 0x00, 0x7c, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x30, 0x00, 0x00},
   {0x00, 0x00, 0x82, 0x92, 0x92, 0x92, 0x92, 0xec, 0x00, 0x00, 0x00, 0x00, 0x00},
   {0x00, 0x00, 0x42, 0x42, 0x42, 0x42, 0x62, 0x5c, 0x00, 0x00, 0x00, 0x00, 0x00},
   {0x00, 0x00, 0x3c, 0x42, 0x42, 0x42, 0x42, 0x3c, 0x00, 0x00, 0x00, 0x00, 0x00},
   {0x40, 0x40, 0x40, 0x5c, 0x62, 0x42, 0x62, 0x5c, 0x00, 0x00, 0x00, 0x00, 0x00},
   {0x02, 0x02, 0x02, 0x3a, 0x46, 0x42, 0x46, 0x3a, 0x00, 0x00, 0x00, 0x00, 0x00},
   {0x00, 0x00, 0x20, 0x20, 0x20, 0x20, 0x22, 0x5c, 0x00, 0x00, 0x00, 0x00, 0x00},
   {0x00, 0x00, 0x3c, 0x42, 0x0c, 0x30, 0x42, 0x3c, 0x00, 0x00, 0x00, 0x00, 0x00},
   {0x00, 0x00, 0x1c, 0x22, 0x20, 0x20, 0x20, 0x7c, 0x20, 0x20, 0x00, 0x00, 0x00},
   {0x00, 0x00, 0x3a, 0x44, 0x44, 0x44, 0x44, 0x44, 0x00, 0x00, 0x00, 0x00, 0x00},
   {0x00, 0x00, 0x10, 0x28, 0x28, 0x44, 0x44, 0x44, 0x00, 0x00, 0x00, 0x00, 0x00},
   {0x00, 0x00, 0x44, 0xaa, 0x92, 0x92, 0x82, 0x82, 0x00, 0x00, 0x00, 0x00, 0x00},
   {0x00, 0x00, 0x42, 0x24, 0x18, 0x18, 0x24, 0x42, 0x00, 0x00, 0x00, 0x00, 0x00},
   {0x3c, 0x42, 0x02, 0x3a, 0x46, 0x42, 0x42, 0x42, 0x00, 0x00, 0x00, 0x00, 0x00},
   {0x00, 0x00, 0x7e, 0x20, 0x10, 0x08, 0x04, 0x7e, 0x00, 0x00, 0x00, 0x00, 0x00},
   {0x00, 0x00, 0x0e, 0x10, 0x10, 0x08, 0x30, 0x08, 0x10, 0x10, 0x0e, 0x00, 0x00},
   {0x00, 0x00, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x00, 0x00},
   {0x00, 0x00, 0x70, 0x08, 0x08, 0x10, 0x0c, 0x10, 0x08, 0x08, 0x70, 0x00, 0x00},
   {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x48, 0x54, 0x24, 0x00, 0x00},
   {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00},
   {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00},
   {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00},
   {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00},
   {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00},
   {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00},
   {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00},
   {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00},
   {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00},
   {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00},
   {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00},
   {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00},
   {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00},
   {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00},
   {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00},
   {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00},
   {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00},
   {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00},
   {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00},
   {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00},
   {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00},
   {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00},
   {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00},
   {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00},
   {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00},
   {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00},
   {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00},
   {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00},
   {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00},
   {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00},
   {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00},
   {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00},
   {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00},
   {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00},
   {0x00, 0x00, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x00, 0x10, 0x00, 0x00},
   {0x00, 0x00, 0x00, 0x10, 0x38, 0x54, 0x50, 0x50, 0x54, 0x38, 0x10, 0x00, 0x00},
   {0x00, 0x00, 0xdc, 0x62, 0x20, 0x20, 0x20, 0x70, 0x20, 0x22, 0x1c, 0x00, 0x00},
   {0x00, 0x00, 0x00, 0x42, 0x3c, 0x24, 0x24, 0x3c, 0x42, 0x00, 0x00, 0x00, 0x00},
   {0x00, 0x00, 0x10, 0x10, 0x7c, 0x10, 0x7c, 0x28, 0x44, 0x82, 0x82, 0x00, 0x00},
   {0x00, 0x00, 0x10, 0x10, 0x10, 0x10, 0x00, 0x10, 0x10, 0x10, 0x10, 0x00, 0x00},
   {0x00, 0x00, 0x18, 0x24, 0x04, 0x18, 0x24, 0x24, 0x18, 0x20, 0x24, 0x18, 0x00},
   {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x6c, 0x00, 0x00},
   {0x00, 0x00, 0x00, 0x38, 0x44, 0x92, 0xaa, 0xa2, 0xaa, 0x92, 0x44, 0x38, 0x00},
   {0x00, 0x00, 0x00, 0x00, 0x7c, 0x00, 0x3c, 0x44, 0x3c, 0x04, 0x38, 0x00, 0x00},
   {0x00, 0x00, 0x00, 0x12, 0x24, 0x48, 0x90, 0x48, 0x24, 0x12, 0x00, 0x00, 0x00},
   {0x00, 0x00, 0x00, 0x02, 0x02, 0x02, 0x7e, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00},
   {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x7e, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00},
   {0x00, 0x00, 0x00, 0x38, 0x44, 0xaa, 0xb2, 0xaa, 0xaa, 0x92, 0x44, 0x38, 0x00},
   {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x7e, 0x00, 0x00},
   {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x18, 0x24, 0x24, 0x18, 0x00, 0x00},
   {0x00, 0x00, 0x00, 0x7c, 0x00, 0x10, 0x10, 0x7c, 0x10, 0x10, 0x00, 0x00, 0x00},
   {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x78, 0x40, 0x30, 0x08, 0x48, 0x30, 0x00},
   {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x30, 0x48, 0x08, 0x10, 0x48, 0x30, 0x00},
   {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x10, 0x08, 0x00},
   {0x00, 0x40, 0x5a, 0x66, 0x42, 0x42, 0x42, 0x42, 0x00, 0x00, 0x00, 0x00, 0x00},
   {0x00, 0x00, 0x14, 0x14, 0x14, 0x14, 0x34, 0x74, 0x74, 0x74, 0x3e, 0x00, 0x00},
   {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x18, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00},
   {0x18, 0x08, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00},
   {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x70, 0x20, 0x20, 0x20, 0x60, 0x20, 0x00},
   {0x00, 0x00, 0x00, 0x00, 0x00, 0x78, 0x00, 0x30, 0x48, 0x48, 0x30, 0x00, 0x00},
   {0x00, 0x00, 0x00, 0x90, 0x48, 0x24, 0x12, 0x24, 0x48, 0x90, 0x00, 0x00, 0x00},
   {0x00, 0x00, 0x06, 0x1a, 0x12, 0x0a, 0xe6, 0x42, 0x40, 0x40, 0xc0, 0x40, 0x00},
   {0x00, 0x00, 0x1e, 0x10, 0x0c, 0x02, 0xf2, 0x4c, 0x40, 0x40, 0xc0, 0x40, 0x00},
   {0x00, 0x00, 0x06, 0x1a, 0x12, 0x0a, 0x66, 0x92, 0x10, 0x20, 0x90, 0x60, 0x00},
   {0x00, 0x00, 0x3c, 0x42, 0x42, 0x40, 0x20, 0x10, 0x10, 0x00, 0x10, 0x00, 0x00},
   {0x00, 0x00, 0x42, 0x42, 0x7e, 0x42, 0x42, 0x24, 0x18, 0x00, 0x08, 0x10, 0x00},
   {0x00, 0x00, 0x42, 0x42, 0x7e, 0x42, 0x42, 0x24, 0x18, 0x00, 0x10, 0x08, 0x00},
   {0x00, 0x00, 0x42, 0x42, 0x7e, 0x42, 0x42, 0x24, 0x18, 0x00, 0x24, 0x18, 0x00},
   {0x00, 0x00, 0x42, 0x42, 0x7e, 0x42, 0x42, 0x24, 0x18, 0x00, 0x28, 0x14, 0x00},
   {0x00, 0x00, 0x42, 0x42, 0x7e, 0x42, 0x42, 0x24, 0x18, 0x00, 0x24, 0x24, 0x00},
   {0x00, 0x00, 0x42, 0x42, 0x7e, 0x42, 0x42, 0x24, 0x18, 0x18, 0x24, 0x18, 0x00},
   {0x00, 0x00, 0x9e, 0x90, 0x90, 0xf0, 0x9c, 0x90, 0x90, 0x90, 0x6e, 0x00, 0x00},
   {0x10, 0x08, 0x3c, 0x42, 0x40, 0x40, 0x40, 0x40, 0x40, 0x42, 0x3c, 0x00, 0x00},
   {0x00, 0x00, 0x7e, 0x40, 0x40, 0x78, 0x40, 0x40, 0x7e, 0x00, 0x08, 0x10, 0x00},
   {0x00, 0x00, 0x7e, 0x40, 0x40, 0x78, 0x40, 0x40, 0x7e, 0x00, 0x10, 0x08, 0x00},
   {0x00, 0x00, 0x7e, 0x40, 0x40, 0x78, 0x40, 0x40, 0x7e, 0x00, 0x24, 0x18, 0x00},
   {0x00, 0x00, 0x7e, 0x40, 0x40, 0x78, 0x40, 0x40, 0x7e, 0x00, 0x24, 0x24, 0x00},
   {0x00, 0x00, 0x7c, 0x10, 0x10, 0x10, 0x10, 0x10, 0x7c, 0x00, 0x08, 0x10, 0x00},
   {0x00, 0x00, 0x7c, 0x10, 0x10, 0x10, 0x10, 0x10, 0x7c, 0x00, 0x10, 0x08, 0x00},
   {0x00, 0x00, 0x7c, 0x10, 0x10, 0x10, 0x10, 0x10, 0x7c, 0x00, 0x24, 0x18, 0x00},
   {0x00, 0x00, 0x7c, 0x10, 0x10, 0x10, 0x10, 0x10, 0x7c, 0x00, 0x28, 0x28, 0x00},
   {0x00, 0x00, 0xfc, 0x42, 0x42, 0x42, 0xe2, 0x42, 0x42, 0x42, 0xfc, 0x00, 0x00},
   {0x00, 0x00, 0x82, 0x86, 0x8a, 0x92, 0xa2, 0xc2, 0x82, 0x00, 0x28, 0x14, 0x00},
   {0x00, 0x00, 0x7c, 0x82, 0x82, 0x82, 0x82, 0x82, 0x7c, 0x00, 0x08, 0x10, 0x00},
   {0x00, 0x00, 0x7c, 0x82, 0x82, 0x82, 0x82, 0x82, 0x7c, 0x00, 0x10, 0x08, 0x00},
   {0x00, 0x00, 0x7c, 0x82, 0x82, 0x82, 0x82, 0x82, 0x7c, 0x00, 0x24, 0x18, 0x00},
   {0x00, 0x00, 0x7c, 0x82, 0x82, 0x82, 0x82, 0x82, 0x7c, 0x00, 0x28, 0x14, 0x00},
   {0x00, 0x00, 0x7c, 0x82, 0x82, 0x82, 0x82, 0x82, 0x7c, 0x00, 0x28, 0x28, 0x00},
   {0x00, 0x00, 0x00, 0x42, 0x24, 0x18, 0x18, 0x24, 0x42, 0x00, 0x00, 0x00, 0x00},
   {0x00, 0x40, 0x3c, 0x62, 0x52, 0x52, 0x52, 0x4a, 0x4a, 0x46, 0x3c, 0x02, 0x00},
   {0x00, 0x00, 0x3c, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x00, 0x08, 0x10, 0x00},
   {0x00, 0x00, 0x3c, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x00, 0x10, 0x08, 0x00},
   {0x00, 0x00, 0x3c, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x00, 0x24, 0x18, 0x00},
   {0x00, 0x00, 0x3c, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x00, 0x24, 0x24, 0x00},
   {0x00, 0x00, 0x10, 0x10, 0x10, 0x10, 0x28, 0x44, 0x44, 0x00, 0x10, 0x08, 0x00},
   {0x00, 0x00, 0x40, 0x40, 0x40, 0x7c, 0x42, 0x42, 0x42, 0x7c, 0x40, 0x00, 0x00},
   {0x00, 0x40, 0x5c, 0x62, 0x42, 0x42, 0x7c, 0x42, 0x42, 0x3c, 0x00, 0x00, 0x00},
   {0x00, 0x00, 0x3a, 0x46, 0x42, 0x3e, 0x02, 0x3c, 0x00, 0x00, 0x08, 0x10, 0x00},
   {0x00, 0x00, 0x3a, 0x46, 0x42, 0x3e, 0x02, 0x3c, 0x00, 0x00, 0x10, 0x08, 0x00},
   {0x00, 0x00, 0x3a, 0x46, 0x42, 0x3e, 0x02, 0x3c, 0x00, 0x00, 0x24, 0x18, 0x00},
   {0x00, 0x00, 0x3a, 0x46, 0x42, 0x3e, 0x02, 0x3c, 0x00, 0x00, 0x28, 0x14, 0x00},
   {0x00, 0x00, 0x3a, 0x46, 0x42, 0x3e, 0x02, 0x3c, 0x00, 0x00, 0x24, 0x24, 0x00},
   {0x00, 0x00, 0x3a, 0x46, 0x42, 0x3e, 0x02, 0x3c, 0x00, 0x18, 0x24, 0x18, 0x00},
   {0x00, 0x00, 0x6c, 0x92, 0x90, 0x7c, 0x12, 0x6c, 0x00, 0x00, 0x00, 0x00, 0x00},
   {0x10, 0x08, 0x3c, 0x42, 0x40, 0x40, 0x42, 0x3c, 0x00, 0x00, 0x00, 0x00, 0x00},
   {0x00, 0x00, 0x3c, 0x42, 0x40, 0x7e, 0x42, 0x3c, 0x00, 0x00, 0x08, 0x10, 0x00},
   {0x00, 0x00, 0x3c, 0x42, 0x40, 0x7e, 0x42, 0x3c, 0x00, 0x00, 0x10, 0x08, 0x00},
   {0x00, 0x00, 0x3c, 0x42, 0x40, 0x7e, 0x42, 0x3c, 0x00, 0x00, 0x24, 0x18, 0x00},
   {0x00, 0x00, 0x3c, 0x42, 0x40, 0x7e, 0x42, 0x3c, 0x00, 0x00, 0x24, 0x24, 0x00},
   {0x00, 0x00, 0x7c, 0x10, 0x10, 0x10, 0x10, 0x30, 0x00, 0x00, 0x10, 0x20, 0x00},
   {0x00, 0x00, 0x7c, 0x10, 0x10, 0x10, 0x10, 0x30, 0x00, 0x00, 0x20, 0x10, 0x00},
   {0x00, 0x00, 0x7c, 0x10, 0x10, 0x10, 0x10, 0x30, 0x00, 0x00, 0x48, 0x30, 0x00},
   {0x00, 0x00, 0x7c, 0x10, 0x10, 0x10, 0x10, 0x30, 0x00, 0x00, 0x28, 0x28, 0x00},
   {0x00, 0x00, 0x3c, 0x42, 0x42, 0x42, 0x42, 0x3c, 0x04, 0x28, 0x18, 0x24, 0x00},
   {0x00, 0x00, 0x42, 0x42, 0x42, 0x42, 0x62, 0x5c, 0x00, 0x00, 0x28, 0x14, 0x00},
   {0x00, 0x00, 0x3c, 0x42, 0x42, 0x42, 0x42, 0x3c, 0x00, 0x00, 0x08, 0x10, 0x00},
   {0x00, 0x00, 0x3c, 0x42, 0x42, 0x42, 0x42, 0x3c, 0x00, 0x00, 0x10, 0x08, 0x00},
   {0x00, 0x00, 0x3c, 0x42, 0x42, 0x42, 0x42, 0x3c, 0x00, 0x00, 0x24, 0x18, 0x00},
   {0x00, 0x00, 0x3c, 0x42, 0x42, 0x42, 0x42, 0x3c, 0x00, 0x00, 0x28, 0x14, 0x00},
   {0x00, 0x00, 0x3c, 0x42, 0x42, 0x42, 0x42, 0x3c, 0x00, 0x00, 0x24, 0x24, 0x00},
   {0x00, 0x00, 0x00, 0x10, 0x10, 0x00, 0x7c, 0x00, 0x10, 0x10, 0x00, 0x00, 0x00},
   {0x00, 0x40, 0x3c, 0x62, 0x52, 0x4a, 0x46, 0x3c, 0x02, 0x00, 0x00, 0x00, 0x00},
   {0x00, 0x00, 0x3a, 0x44, 0x44, 0x44, 0x44, 0x44, 0x00, 0x00, 0x08, 0x10, 0x00},
   {0x00, 0x00, 0x3a, 0x44, 0x44, 0x44, 0x44, 0x44, 0x00, 0x00, 0x10, 0x08, 0x00},
   {0x00, 0x00, 0x3a, 0x44, 0x44, 0x44, 0x44, 0x44, 0x00, 0x00, 0x24, 0x18, 0x00},
   {0x00, 0x00, 0x3a, 0x44, 0x44, 0x44, 0x44, 0x44, 0x00, 0x00, 0x24, 0x24, 0x00},
   {0x3c, 0x42, 0x02, 0x3a, 0x46, 0x42, 0x42, 0x42, 0x00, 0x00, 0x10, 0x08, 0x00},
   {0x40, 0x40, 0x5c, 0x62, 0x42, 0x42, 0x62, 0x5c, 0x40, 0x40, 0x00, 0x00, 0x00},
   {0x3c, 0x42, 0x02, 0x3a, 0x46, 0x42, 0x42, 0x42, 0x00, 0x00, 0x24, 0x24, 0x00}
};


//----------------------------------------------------------------------------
sglBitmapFont8x13::sglBitmapFont8x13()
   : m_list_base(0)
{
   m_list_base = glGenLists(128);
   if (m_list_base == 0)
   {
      sglPrintWarning() << "glGenLists failed during font creation" << endl;
      return;
   }

   glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
   glPixelStorei(GL_UNPACK_ROW_LENGTH, 0);
   glPixelStorei(GL_UNPACK_SKIP_ROWS, 0);
   glPixelStorei(GL_UNPACK_SKIP_PIXELS, 0);

   for (int i=0; i<128; ++i)
   {
      glNewList(m_list_base+i, GL_COMPILE);
      glBitmap(8, 13, 0.0, 2.0, 10.0, 0.0, sglBitmapFont8x13Letters[i]);
      glEndList();
   }
}

//----------------------------------------------------------------------------
sglBitmapFont8x13::~sglBitmapFont8x13()
{
   glDeleteLists(m_list_base, 128);
}

//----------------------------------------------------------------------------
sglBitmapFont* sglBitmapFont8x13::clone() const
{
   return new sglBitmapFont8x13();
}

//----------------------------------------------------------------------------
void sglBitmapFont8x13::rasterString(const char *str, int len) const
{
   glListBase(m_list_base);
   if (len < 0)
      glCallLists(strlen(str), GL_BYTE, (const GLvoid*)str);
   else if (len > 0)
      glCallLists(len, GL_BYTE, (const GLvoid*)str);
}

//----------------------------------------------------------------------------
unsigned int sglBitmapFont8x13::getHeight() const
{
   return 13;
}

//----------------------------------------------------------------------------
unsigned int sglBitmapFont8x13::getDescent() const
{
   return 2;
}

//----------------------------------------------------------------------------
unsigned int sglBitmapFont8x13::getCharWidth(char) const
{
   return 8;
}
