/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglBitmapFont.hpp
 *   Author: Tom Stimson
 *  Created: 20 April 2000
 *  Summary: Class to render text strings.
 *****************************************************************************/

#ifndef __SGL_BITMAP_FONT_HPP
#define __SGL_BITMAP_FONT_HPP

#include <sgl.h>
#include <sglObject.hpp>
#include <string>

//============================================================================

/**

sglBitmapFont is the base class for platform-specific bitmapped font classes:
sglBitmapFontGLX (X windows systems), sglBitmapFontWGL (Microsoft windows
systems), and sglBitmapFont8x13 (any system).  The constructor takes a font
specification whose details are platform specific.
These classes also implement a rasterString() member function that renders a
string at the current raster position.  Characters should be limited to the
ASCII range 0-127.

{\bf Example:}

\begin{verbatim}
   // a subclass of sglBitmapFont specific to X11
   sglBitmapFontGLX *fnt = NULL;

   Display *dpy = glXGetCurrentDisplay();
   XFontStruct *font = XLoadQueryFont(dpy, "fixed");
   if (font)
   {
      sglBitmapFontGLX *fnt = new sglBitmapFontGLX(font->fid);
      fnt->rasterString("This is rendered at current raster position.");
   }
\end{verbatim}

*/

class SGL_DLL_API sglBitmapFont : public sglObject
{
public:
   /// default constructor
   sglBitmapFont() {};
   /// destructor
   virtual ~sglBitmapFont() {};

   /** Create a copy of this font.  The copy will have display lists
       that are valid for the current context.
       @return a copy of this font
   */
   virtual sglBitmapFont* clone() const = 0;

   /** Renders a bitmap string at the current raster position.
       Characters should be limited to the ASCII range 0-127.
       @param str the string to render
   */
   void rasterString(const string &str) const
        { rasterString(str.c_str(), str.size()); }

   /** Renders a bitmap string at the current raster position.
       Characters should be limited to the ASCII range 0-127.
       @param str the string to render
       @param len the length of the string.  If len is less than zero,
              strlen() will be used to determine the length.
   */
   virtual void rasterString(const char *str, int len = -1) const = 0;

   /** Gets the height of the font in pixels.
       @return the height of the font in pixels
   */
   virtual unsigned int getHeight() const = 0;

   /** Gets the extent of the font below the baseline in pixels.
       @return the extent of the font below the baseline in pixels.
   */
   virtual unsigned int getDescent() const = 0;

   /** Gets the width of the given character.
       @param c the character whose width will be returned
       @return the width of the given character
   */
   virtual unsigned int getCharWidth(char c) const = 0;

   /** Gets the width of the given string.
       @param str the string whose width will be returned
       @return the width of the given string
   */
   virtual unsigned int getStringWidth(const string &str) const;

   /** Sets the default font to be used by sgl.
       @see sglText
       @param font the new default font
   */
   static void setDefaultFont(sglBitmapFont* font);

   /** Gets the default font used by sgl.
       @see sglText
       @return the default font used by sgl
   */
   static sglBitmapFont* getDefaultFont() { return s_default_font; }

private:  // not implemented
   sglBitmapFont(const sglBitmapFont &);
   sglBitmapFont &operator=(const sglBitmapFont &);

private:
   static sglBitmapFont *s_default_font;
};


//============================================================================

/** A bitmap font that uses a hardcoded 8x13 font.
 */

class SGL_DLL_API sglBitmapFont8x13 : public sglBitmapFont
{
public:
   /** Construct a font.
       NOTE: A gl context should be current when this method is called.
   */
   sglBitmapFont8x13();

   /// NOTE: A gl context should be current when this method is called.
   ~sglBitmapFont8x13();

   /** Create a copy of this font.  The copy will have display lists
       that are valid for the current context.
       @return a copy of this font
   */
   virtual sglBitmapFont* clone() const;

   /** Renders a bitmap string at the current raster position.
       Characters should be limited to the ASCII range 0-127.
       @param str the string to render
       @param len the length of the string.  If len is less than zero,
              strlen() will be used to determine the length.
   */
   virtual void rasterString(const char *str, int len = -1) const;

   /** Gets the height of the font in pixels.
       @return the height of the font in pixels
   */
   virtual unsigned int getHeight() const;

   /** Gets the extent of the font below the baseline in pixels.
       @return the extent of the font below the baseline in pixels.
   */
   virtual unsigned int getDescent() const;

   /** Gets the width of the given character.
       @param c the character whose width will be returned
       @return the width of the given character
   */
   virtual unsigned int getCharWidth(char c) const;

private:  // not implemented
   sglBitmapFont8x13(const sglBitmapFont8x13 &);
   sglBitmapFont8x13 &operator=(const sglBitmapFont8x13 &);

private:
   GLuint  m_list_base;
};


//============================================================================

#if !defined(WIN32) && !defined( MAC_OSX )
# include <X11/Xlib.h>

/// Font that uses glx to rasterize an X font.

class SGL_DLL_API sglBitmapFontGLX : public sglBitmapFont
{
public:
   /** Construct a font from the given X font.
       NOTE: A gl context should be current when this method is called.
       @param font the X font
   */
   sglBitmapFontGLX(Font font);

   /// NOTE: A gl context should be current when this method is called.
   ~sglBitmapFontGLX();

   /** Create a copy of this font.  The copy will have display lists
       that are valid for the current context.
       @return a copy of this font
   */
   virtual sglBitmapFont* clone() const;

   /** Renders a bitmap string at the current raster position.
       Characters should be limited to the ASCII range 0-127.
       @param str the string to render
       @param len the length of the string.  If len is less than zero,
              strlen() will be used to determine the length.
   */
   virtual void rasterString(const char *str, int len = -1) const;

   /** Gets the height of the font in pixels.
       @return the height of the font in pixels
   */
   virtual unsigned int getHeight() const;

   /** Gets the extent of the font below the baseline in pixels.
       @return the extent of the font below the baseline in pixels.
   */
   virtual unsigned int getDescent() const;

   /** Gets the width of the given character.
       @param c the character whose width will be returned
       @return the width of the given character
   */
   virtual unsigned int getCharWidth(char c) const;

   /** Gets the width of the given string.
       @param str the string whose width will be returned
       @return the width of the given string
   */
   virtual unsigned int getStringWidth(const string &str) const;

private:  // not implemented
   sglBitmapFontGLX();
   sglBitmapFontGLX(const sglBitmapFontGLX &);
   sglBitmapFontGLX &operator=(const sglBitmapFontGLX &);

private:
   Font         m_font;
   GLuint       m_list_base;
   XFontStruct *m_font_info;
};
#endif


#endif
