/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglCamera.hpp
 *   Author: Scott McMillan
 *  Created: 19 December 2000
 *  Summary: abstract base class for cameras (subclassed to perspective and
 *           orthographic cameras.
 *****************************************************************************/

#ifndef __SGL_CAMERA_HPP
#define __SGL_CAMERA_HPP

#include <sgl.h>
#include <sglObject.hpp>
#include <sglFrustum.hpp>

//----------------------------------------------------------------------------

/**

This is the abstract base class for both sglPerspectiveCamera and
sglOrthographicCamera classes.  These classes are used to specify the frustum
that is to be used for culling in SGL and setting up the projection for
rendering in OpenGL.  As such the class contains the sglCamera::getFrustum()
method for obtaining a reference to the internal sglFrustum class (only double
precision for now), so that it can be passed into the sglScene::cull(...)
function.  It also contains the pure virtual sglCamera::applyProjection() which
will set OpenGL's projection matrix to correspond to the internal frustum.

In normal practice, an sglCamera is assigned to an sglViewPlatform to define a
position and orientation of the eye point in the 3D scene. See
sglViewPlatform's code example for an illustration of the use of these
functions.

 */

class SGL_DLL_API sglCamera : public sglObject
{
public:
   /// default constructor
   sglCamera();

   /// virtual destructor
   virtual ~sglCamera();

   /** Query the current view frustum.  Currently only a double precision
       frustum is available (OpenGL only takes doubles for the glFrustum
       commands.
       @return A reference to the current frustum.
   */
   const sglFrustumd &getFrustum() const { return m_frustum; }

   /** Setup the OpenGL projection matrix to match the specified frustum.
       This function leaves the glMatrixMode(GL_PROJECTION) in effect (current)
       upon return.
   */
   virtual void applyProjection() const = 0;

   // See sglObject::printInfo() for details.
   virtual void printInfo(ostream &ostrm, const char *indent_string) const;

protected:
   sglCamera(const sglCamera &);            // not implemented
   sglCamera &operator=(const sglCamera &); // not implemented

protected:
   double m_near;
   double m_far;

   sglFrustumd m_frustum;
};

#endif
