/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglCull.cpp
 *  Created: 19 December 2000
 *****************************************************************************/

#include <sglCull.hpp>
#include <sglLightProcessor.hpp>

#include <algorithm>

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------

bool sglRangeObject::smallerRange(const sglRangeObjectPtr &x,
                                  const sglRangeObjectPtr &y)
{
   return (x->m_range < y->m_range);
}

//----------------------------------------------------------------------------
bool sglRangeObject::greaterRange(const sglRangeObjectPtr &x,
                                  const sglRangeObjectPtr &y)
{
   return (x->m_range > y->m_range);
}

//----------------------------------------------------------------------------
void sglRangeObject::draw(sglCurrState *curr_state) const
{
   curr_state->apply(m_fullstate);
   curr_state->draw(m_drawable);
}

//----------------------------------------------------------------------------
void sglRangeObject::statsDraw(sglCurrState *curr_state,
                               sglStats& stats) const
{
   m_drawable->addStats(stats);
   m_fullstate->addStats(stats);

   curr_state->statsApply(m_fullstate, stats);
   curr_state->draw(m_drawable);
}

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------

void sglOcclusionObject::draw(sglCurrState *curr_state) const
{
   GLboolean result = GL_FALSE;
   glPushAttrib(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
   glDepthMask(GL_FALSE);
   glColorMask(GL_FALSE,GL_FALSE,GL_FALSE,GL_FALSE);
   glColor4f(0,0,1,1);
   glEnable(sgl::s_occlusion_enable_flag);
   m_drawable->getPrevBound().drawOcclusionBox();
   glGetBooleanv(sgl::s_occlusion_result_flag, &result);
   glDisable(sgl::s_occlusion_enable_flag);
   glPopAttrib();
   if (result != GL_FALSE)
   {
      curr_state->apply(m_fullstate);
      curr_state->draw(m_drawable);
   }
}

//----------------------------------------------------------------------------
void sglOcclusionObject::statsDraw(sglCurrState *curr_state,
                                   sglStats& stats) const
{
   GLboolean result = GL_FALSE;
   glPushAttrib(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
   glDepthMask(GL_FALSE);
   glColorMask(GL_FALSE,GL_FALSE,GL_FALSE,GL_FALSE);
   glColor4f(0,0,1,1);
   glEnable(sgl::s_occlusion_enable_flag);
   m_drawable->getPrevBound().drawOcclusionBox();
   glGetBooleanv(sgl::s_occlusion_result_flag, &result);
   glDisable(sgl::s_occlusion_enable_flag);
   glPopAttrib();
   if (result == GL_FALSE)
      stats.addOccludedGSets(1);
   else
   {
      m_drawable->addStats(stats);
      m_fullstate->addStats(stats);

      curr_state->statsApply(m_fullstate, stats);
      curr_state->draw(m_drawable);
   }
}

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------

template <class T>
void sglTransformStatelet<T>::apply(sglCurrState*) const
{
   sglLoadMatrix(m_matrix);
   if (m_normalize_flag)
   {
      glDisable(sgl::s_rescale_normal);
      glEnable(GL_NORMALIZE);
   }
   else if (m_rescale_flag)
   {
      glDisable(GL_NORMALIZE);
      glEnable(sgl::s_rescale_normal);
   }
   else
   {
      glDisable(sgl::s_rescale_normal);
      glDisable(GL_NORMALIZE);
   }
}

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------

void sglCullDrawablePool::setSortOrder(
   const vector<unsigned int> &sort_order)
{
   // There is no point to sort on transparency or occlusion culling,
   // since we handle that case explicitly.
   m_sort_order.clear();
   for (vector<unsigned int>::const_iterator ii = sort_order.begin();
        ii != sort_order.end(); ++ii)
   {
      if (*ii != sglStatelet::eTRANSPARENCY &&
          *ii != sglStatelet::eOCCLUSION_CULL)
         m_sort_order.push_back(*ii);
   }

   delete m_opaque_bin;

   if (m_sort_order.empty())
   {
      m_opaque_bin = new sglUnSortedDrawablePool();
   }
   else
   {
      m_sort_order.push_back(UINT_MAX);
      m_opaque_bin = NULL;
   }
}

//----------------------------------------------------------------------------
void sglCullDrawablePool::setupSort(
   const vector<const sglStatelet*> &statelets)
{
   if (m_opaque_bin == NULL)
   {
      if (statelets[*(m_sort_order.begin())]->getSorted())
      {
         sglSortedDrawablePool *pool = new sglSortedDrawablePool();
         pool->setSources(&m_drawable_pool_sources);
         pool->setSortKey(m_sort_order.begin());
         m_opaque_bin = pool;
      }
      else
      {
         sglIndexedDrawablePool *pool = new sglIndexedDrawablePool();
         pool->setSources(&m_drawable_pool_sources);
         pool->setSortKey(m_sort_order.begin());
         m_opaque_bin = pool;
      }
   }
}

//----------------------------------------------------------------------------
void sglCullDrawablePool::clear()
{
   m_opaque_bin->clear();
   m_occlusion_bin.clear();
   m_occlusion_sorted_flag = true;
   m_transparency_bin.clear();
   m_transparency_sorted_flag = true;
   m_range_pool.reset();
   m_occlusion_pool.reset();
   m_drawable_pair_pool.reset();
}

//----------------------------------------------------------------------------
void sglCullDrawablePool::insert(sglDrawItem *d_item)
{
   m_opaque_bin->insert(d_item);
}

//----------------------------------------------------------------------------
void sglCullDrawablePool::sortBins()
{
   if (!m_occlusion_sorted_flag)
   {
      // sort the m_occlusion_bin front to back
      sort(m_occlusion_bin.begin(), m_occlusion_bin.end(),
           sglRangeObject::smallerRange);
      m_occlusion_sorted_flag = true;
   }

   if (!m_transparency_sorted_flag)
   {
      // sort the m_transparency_bin back to front
      sort(m_transparency_bin.begin(), m_transparency_bin.end(),
           sglRangeObject::greaterRange);
      m_transparency_sorted_flag = true;
   }
}

//----------------------------------------------------------------------------
void sglCullDrawablePool::draw(sglCurrState *curr_state) const
{
   m_opaque_bin->draw(curr_state);

   GLboolean result;
   if (!m_occlusion_bin.empty() && sgl::s_occlusion_capable)
      // Clear the occlusion state.
      glGetBooleanv(sgl::s_occlusion_result_flag, &result);
   for (vector<sglRangeObjectPtr>::const_iterator oelt =
           m_occlusion_bin.begin();
        oelt != m_occlusion_bin.end();
        ++oelt)
   {
      (*oelt)->draw(curr_state);
   }

   for (vector<sglRangeObjectPtr>::const_iterator telt =
           m_transparency_bin.begin();
        telt != m_transparency_bin.end();
        ++telt)
   {
      (*telt)->draw(curr_state);
   }
}

//----------------------------------------------------------------------------
void sglCullDrawablePool::statsDraw(sglCurrState *curr_state,
                                         sglStats& stats) const
{
   // Add stats for occlusion tested and transparent objects.
   stats.addOcclusionEnabledGSets(m_occlusion_pool.currSize());
   stats.addTransparencyEnabledGSets(m_transparency_bin.size());
   stats.addTotalGSets(m_occlusion_pool.currSize());
   stats.addTotalGSets(m_range_pool.currSize());

   m_opaque_bin->statsDraw(curr_state, stats);

   GLboolean result;
   if (!m_occlusion_bin.empty() && sgl::s_occlusion_capable)
      // Clear the occlusion state.
      glGetBooleanv(sgl::s_occlusion_result_flag, &result);
   for (vector<sglRangeObjectPtr>::const_iterator oelt =
           m_occlusion_bin.begin();
        oelt != m_occlusion_bin.end();
        oelt++)
   {
      (*oelt)->statsDraw(curr_state, stats);
   }

   for (vector<sglRangeObjectPtr>::const_iterator telt =
           m_transparency_bin.begin();
        telt != m_transparency_bin.end();
        telt++)
   {
      (*telt)->statsDraw(curr_state, stats);
   }
}

//----------------------------------------------------------------------------
const sglFullState* sglCullDrawablePool::getFirstState() const
{
   const sglFullState *first_state = m_opaque_bin->getFirstState();
   if (!first_state &&
       m_occlusion_bin.begin() != m_occlusion_bin.end())
      first_state = (*m_occlusion_bin.begin())->getState();
   if (!first_state &&
       m_transparency_bin.begin() != m_transparency_bin.end())
      first_state = (*m_transparency_bin.begin())->getState();
   return first_state;
}

//----------------------------------------------------------------------------

//----------------------------------------------------------------------------
template <class T>
sglCull<T>::sglCull()
      : m_state_mask(0),
        m_viewport_x(0),
        m_viewport_y(0),
        m_orig_frustum(),
        m_current_frustum(),
        m_current_matrix_statelet(),
        m_num_matrices(0),
        m_statelet(sgl::s_num_statelet_types, (const sglStatelet*)NULL),
        m_override(sgl::s_num_statelet_types, false),
        m_curr_state(),
        m_drawable_pool_sources(),
        m_default_bin(m_drawable_pool_sources),
        m_transform_pool(100, 100),
        m_state_pool(0, 100),
        m_light_processor(new sglDefaultLightProcessor())
{
   if (!sgl::s_initialized)
   {
      sglPrintWarning() << "sglCull::sglCull() error: "
                        << "called before sgl::initialize().  "
                        << "The cull traversal will not work if there is more"
                        << " than one texture unit" << endl;
   }

   m_light_processor->ref();

   vector<unsigned int> default_sort_order;
   default_sort_order.push_back(sglStatelet::eTRANSFORM_MATRIX);
   setSortOrder(default_sort_order);
}

//----------------------------------------------------------------------------
template <class T>
sglCull<T>::~sglCull()
{
   m_light_processor->deref();
   if (m_light_processor->getRefCount() == 0)
      delete m_light_processor; 
}

//----------------------------------------------------------------------------
template <class T>
void sglCull<T>::initialize(
   float lod_scale,
   unsigned int discriminator_mask,
   unsigned int state_mask,
   const sglMat4<T>  &view_matrix,
   const sglFrustum<T> &frustum,
   unsigned int viewport_x, unsigned int viewport_y,
   const vector<const sglStatelet*> &default_statelets,
   const deque<const sglStatelet*> &override_statelets,
   unsigned int frame_count,
   const sglTimespec &frame_time)
{
   // clear bins and reset pools
   m_light_processor->reset();
   m_transform_pool.reset();
   m_state_pool.reset();
   m_drawable_pool_stack.clear();
   m_drawable_pool_sources.reset();
   m_default_state_stack.clear();
   m_override_state_stack.clear();

   m_lod_scale = lod_scale;
   m_lod_offset = 0.f;
   m_discriminator_mask = discriminator_mask;
   m_state_mask = state_mask;

   m_viewport_x = viewport_x;
   m_viewport_y = viewport_y;

   m_orig_frustum = frustum;
   sglMat4<T> inv_view_mat;
   inv_view_mat.inverseAffine(view_matrix);
   m_current_frustum.mul(frustum, inv_view_mat);

   m_num_matrices = 0;

   m_frame_count = frame_count;
   m_frame_time  = frame_time;

   m_current_matrix_statelet.clear();
   sglTransformStatelet<T>* trans = m_transform_pool.newObject();
   trans->setMatrix(view_matrix);
   trans->setRescale(false);
   trans->setNormalize(false);
   trans->setIndex(m_num_matrices++);
   m_current_matrix_statelet.push_back(trans);
   m_statelet[sglStatelet::eTRANSFORM_MATRIX] = trans;

   for (unsigned int kk = 0; kk < sgl::s_num_statelet_types; ++kk)
      m_override[kk] = false;
   for (deque<const sglStatelet*>::const_iterator ii =
           override_statelets.begin();
        ii != override_statelets.end(); ++ii)
      pushOverrideStatelet(*ii);
   for (vector<const sglStatelet*>::const_iterator jj =
           default_statelets.begin();
        jj != default_statelets.end(); ++jj)
      pushDefaultStatelet(*jj);

   // these must be done after piecing together m_statelet
   m_default_bin.setupSort(m_statelet);
   m_default_bin.clear();
}

//----------------------------------------------------------------------------
template <class T>
void sglCull<T>::pushDefaultStatelet(const sglStatelet *state)
{
   unsigned int statelet_type = state->getType();
   if (statelet_type < sgl::s_num_statelet_types && !m_override[statelet_type])
   {
      m_default_state_stack.push_back(m_statelet[statelet_type]);
      m_statelet[statelet_type] = state;
   }
   else
      m_default_state_stack.push_back(NULL);
}

//----------------------------------------------------------------------------
template <class T>
void sglCull<T>::popDefaultStatelet()
{
   const sglStatelet *state = m_default_state_stack.back();
   if (state)
   {
      m_statelet[state->getType()] = state;
   }
   m_default_state_stack.pop_back();
}

//----------------------------------------------------------------------------
template <class T>
void sglCull<T>::pushOverrideStatelet(const sglStatelet *state)
{
   unsigned int statelet_type = state->getType();
   if (statelet_type < sgl::s_num_statelet_types && !m_override[statelet_type])
   {
      m_override_state_stack.push_back(m_statelet[statelet_type]);
      m_statelet[statelet_type] = state;
      m_override[statelet_type] = true;
   }
   else
      m_override_state_stack.push_back(NULL);
}

//----------------------------------------------------------------------------
template <class T>
void sglCull<T>::popOverrideStatelet()
{
   const sglStatelet *state = m_override_state_stack.back();
   if (state)
   {
      unsigned int statelet_type = state->getType();
      m_statelet[statelet_type] = state;
      m_override[statelet_type] = false;
   }
   m_override_state_stack.pop_back();
}

//----------------------------------------------------------------------------
template <class T>
sglDrawablePool* sglCull<T>::pushDrawablePool()
{
   sglUnSortedDrawablePool *d_pool =
      m_drawable_pool_sources.m_unsorted_dp_pool.newObject();
   d_pool->clear();

   m_drawable_pool_stack.push_back(d_pool);
   return d_pool;
}

//----------------------------------------------------------------------------
template <class T>
void sglCull<T>::popDrawablePool()
{
   sglDrawablePool *d_pool = m_drawable_pool_stack.back();
   m_drawable_pool_stack.pop_back();

   if (d_pool->getFirstState())
   {
      if (m_drawable_pool_stack.empty())
         m_default_bin.insert(d_pool);
      else
         m_drawable_pool_stack.back()->insert(d_pool);
   }
}

//----------------------------------------------------------------------------
template <class T>
bool sglCull<T>::initState()
{
   const sglFullState *first_state = m_default_bin.getFirstState();
   if (first_state)
   {
      m_light_processor->apply();

      m_curr_state.force(first_state);
      return true;
   }
   else
      return false;
}

//----------------------------------------------------------------------------
template <class T>
void sglCull<T>::setLightProcessor(sglLightProcessor *light_processor)
{
   sglLightProcessor *old_processor = m_light_processor;

   if (light_processor == NULL)
      m_light_processor = new sglDefaultLightProcessor();
   else
      m_light_processor = light_processor;

   m_light_processor->ref();
   old_processor->deref();
   if (old_processor->getRefCount() == 0)
      delete old_processor;
}

//----------------------------------------------------------------------------
template <class T>
void sglCull<T>::draw()
{
   glMatrixMode(GL_MODELVIEW);

   // save GL state
   glPushAttrib(GL_ALL_ATTRIB_BITS); // VERY COSTLY?

   // HACK - SGIs and old versions of Mesa misspelled the following token
   // as GL_CLIENT_ALL_ATTRIB_BITS, so we are just going to use ~0x0 for
   // now.
   glPushClientAttrib(~0x0u); //GL_ALL_CLIENT_ATTRIB_BITS);
   glMatrixMode(GL_MODELVIEW);
   glPushMatrix();

   if (initState())
   {
      m_default_bin.sortBins();
      m_default_bin.draw(&m_curr_state);
   }

   // restore GL state
   glPopMatrix();
   glPopAttrib();
   glPopClientAttrib();
}

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------

template <class T>
void sglStatsCull<T>::draw()
{
   glMatrixMode(GL_MODELVIEW);

   // save GL state
   glPushAttrib(GL_ALL_ATTRIB_BITS); // VERY COSTLY?

   // HACK - SGIs and old versions of Mesa misspelled the following token
   // as GL_CLIENT_ALL_ATTRIB_BITS, so we are just going to use ~0x0 for
   // now.
   glPushClientAttrib(~0x0u); //GL_ALL_CLIENT_ATTRIB_BITS);
   glMatrixMode(GL_MODELVIEW);
   glPushMatrix();

   m_stats.addFrames(1);

   if (initState())
   {
      // Add stats for initial state setup.
      m_stats.addStateChanges(sgl::s_num_statelet_types);
      m_stats.addTransforms(m_num_matrices);

      unsigned int num_lights = m_light_processor->getNumLights();
      if ((GLint)num_lights > sgl::s_max_lights)
         m_stats.addNumLights(sgl::s_max_lights, num_lights);
      else
         m_stats.addNumLights(num_lights, num_lights);

      m_default_bin.sortBins();
      m_default_bin.statsDraw(&m_curr_state, m_stats);
   }

   // restore GL state
   glPopMatrix();
   glPopAttrib();
   glPopClientAttrib();
}


template class SGL_DLL_API sglTransformStatelet<double>;
template class SGL_DLL_API sglTransformStatelet<float>;
template class SGL_DLL_API sglCull<double>;
template class SGL_DLL_API sglCull<float>;
template class SGL_DLL_API sglStatsCull<double>;
template class SGL_DLL_API sglStatsCull<float>;
