/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglCurrState.cpp
 *   Author: Tom Stimson
 *  Created: 9 July 2000
 *****************************************************************************/

#include <sglCurrState.hpp>
#include <sglTexture.hpp>


//----------------------------------------------------------------------------
void sglFullState::addStats(sglStats& stats) const
{
   const sglTexture *tex =
      static_cast<const sglTexture*>(m_statelet[sglStatelet::eTEXTURE]);
   if (tex->isValid())
   {
      stats.addTexturedGSets(1);
      stats.addTexture(tex);
   }
   const sglLighting *li =
      static_cast<const sglLighting*>(m_statelet[sglStatelet::eLIGHTING]);
   if (li->isEnabled())
   {
      stats.addLitGSets(1);
   }

   if (m_dlist)
   {
      stats.addDListedGSets(1);
   }
}

//----------------------------------------------------------------------------
void sglCurrState::checkDepthMask()
{
   if (static_cast<const sglDepthMask*>(
      m_state.m_statelet[sglStatelet::eDEPTH_MASK])->isEnabled() &&
       static_cast<const sglTransparency*>(
          m_state.m_statelet[
             sglStatelet::eTRANSPARENCY])->isEnabled() == false)
   {
      glDepthMask(GL_TRUE);
   }
   else
   {
      glDepthMask(GL_FALSE);
   }
}

//----------------------------------------------------------------------------
void sglCurrState::checkBlend()
{
   if (static_cast<const sglBlending*>(
      m_state.m_statelet[sglStatelet::eBLENDING])->isEnabled() ||
       static_cast<const sglTransparency*>(
          m_state.m_statelet[sglStatelet::eTRANSPARENCY])->isEnabled() ||
       m_point_smooth || m_line_smooth)
   {
      glEnable(GL_BLEND);
   }
   else
   {
      glDisable(GL_BLEND);
   }
}

//----------------------------------------------------------------------------
void sglCurrState::checkPointSmooth(bool point_type)
{
   if (point_type && static_cast<const sglAntiAliasPoint*>(
      m_state.m_statelet[sglStatelet::eANTI_ALIAS_POINT])->isEnabled())
   {
      if (!m_point_smooth)
      {
         m_point_smooth = true;
         glEnable(GL_POINT_SMOOTH);
         checkBlend();
      }
   }
   else if (m_point_smooth)
   {
      m_point_smooth = false;
      checkBlend();
      glDisable(GL_POINT_SMOOTH);
   }
}

//----------------------------------------------------------------------------
void sglCurrState::checkLineSmooth(bool line_type)
{
   if (line_type && static_cast<const sglAntiAliasLine*>(
      m_state.m_statelet[sglStatelet::eANTI_ALIAS_LINE])->isEnabled())
   {
      if (!m_line_smooth)
      {
         m_line_smooth = true;
         glEnable(GL_LINE_SMOOTH);
         checkBlend();
      }
   }
   else if (m_line_smooth)
   {
      m_line_smooth = false;
      checkBlend();
      glDisable(GL_LINE_SMOOTH);
   }
}

//----------------------------------------------------------------------------
void sglCurrState::checkMultisample(bool line_type, bool fill_type)
{
   if ((line_type && static_cast<const sglAntiAliasLine*>(
      m_state.m_statelet[sglStatelet::eANTI_ALIAS_LINE])->isEnabled()) ||
      (fill_type && static_cast<const sglAntiAliasPolygon*>(
         m_state.m_statelet[sglStatelet::eANTI_ALIAS_POLYGON])->isEnabled()))
   {
      if (!m_multisample)
      {
         m_multisample = true;
         glEnable((GLenum)0x809D); // GL_MULTISAMPLE_EXT
      }
   }
   else if (m_multisample)
   {
      m_multisample = false;
      glDisable((GLenum)0x809D); // GL_MULTISAMPLE_EXT
   }
}

//----------------------------------------------------------------------------
void sglCurrState::drawGeometry(const sglDrawable* drawable)
{
   glColor4f(1.0f, 1.0f, 1.0f, 1.0f);

   // draw the geometry
   if (m_state.m_dlist)
      glCallList(m_state.m_dlist);
   else
      drawable->drawImmediate(*m_state.m_tex_coords);
}

//----------------------------------------------------------------------------
void sglCurrState::apply(const sglFullState *new_state)
{
   vector<const sglStatelet*>::iterator curr_statelet
      = m_state.m_statelet.begin();
   vector<const sglStatelet*>::const_iterator new_statelet
      = new_state->m_statelet.begin();
   vector<const sglStatelet*>::iterator last
      = m_state.m_statelet.end();
   while (curr_statelet != last)
   {
      if (!(**curr_statelet == **new_statelet))
      {
         *curr_statelet = *new_statelet;
         (*curr_statelet)->apply(this);
      }
      ++curr_statelet;
      ++new_statelet;
   }
   m_state.m_tex_coords = new_state->m_tex_coords;
   m_state.m_dlist = new_state->m_dlist;
}

//----------------------------------------------------------------------------
void sglCurrState::statsApply(const sglFullState *new_state,
                              sglStats& stats)
{
   vector<const sglStatelet*>::iterator curr_statelet
      = m_state.m_statelet.begin();
   vector<const sglStatelet*>::const_iterator new_statelet
      = new_state->m_statelet.begin();
   vector<const sglStatelet*>::iterator last
      = m_state.m_statelet.end();
   while (curr_statelet != last)
   {
      if (!(**curr_statelet == **new_statelet))
      {
         *curr_statelet = *new_statelet;
         (*curr_statelet)->apply(this);
         stats.addStateChanges(1);
      }
      ++curr_statelet;
      ++new_statelet;
   }
   m_state.m_tex_coords = new_state->m_tex_coords;
   m_state.m_dlist = new_state->m_dlist;
}

//----------------------------------------------------------------------------
void sglCurrState::force(const sglFullState *new_state)
{
   if (sgl::s_multisample_capable)
      glDisable((GLenum)0x809D); // GL_MULTISAMPLE_EXT
   glDisable(GL_LINE_SMOOTH);
   glDisable(GL_POINT_SMOOTH);
   m_line_smooth = false;
   m_point_smooth = false;
   m_multisample = false;

   m_state.m_statelet = new_state->m_statelet;
   vector<const sglStatelet*>::iterator end
      = m_state.m_statelet.end();
   for (vector<const sglStatelet*>::iterator it
           = m_state.m_statelet.begin(); it != end; ++it)
   {
      (*it)->apply(this);
   }
   m_state.m_tex_coords = new_state->m_tex_coords;
   m_state.m_dlist = new_state->m_dlist;
}

//----------------------------------------------------------------------------
void sglCurrState::draw(const sglDrawable *drawable)
{
   bool point_type = false;
   bool line_type = false;
   bool fill_type = false;

   switch (drawable->getAAType())
   {
      case sglDrawable::ePOINT_AA:
         point_type = true;
         break;
      case sglDrawable::eLINE_AA:
         line_type = true;
         break;
      case sglDrawable::ePOLYGON_AA:
         switch (static_cast<const sglPolyFrontStyle*>(
            m_state.m_statelet[sglStatelet::ePOLY_FRONT_STYLE])->getStyle())
         {
            case sglPolyStyle::ePOINT:
               point_type = true;
               break;
            case sglPolyStyle::eLINE:
               line_type = true;
               break;
            case sglPolyStyle::eFILLED:
               fill_type = true;
               break;
         }
         switch (static_cast<const sglPolyBackStyle*>(
            m_state.m_statelet[sglStatelet::ePOLY_BACK_STYLE])->getStyle())
         {
            case sglPolyStyle::ePOINT:
               point_type = true;
               break;
            case sglPolyStyle::eLINE:
               line_type = true;
               break;
            case sglPolyStyle::eFILLED:
               fill_type = true;
               break;
         }
         break;
      default:
         break;
   }

   checkPointSmooth(point_type);

   if (sgl::s_multisample_capable)
   {
      checkMultisample(line_type, fill_type);
      drawGeometry(drawable);
   }
   else
   {
      checkLineSmooth(line_type);
      drawGeometry(drawable);

      if (fill_type && static_cast<const sglAntiAliasPolygon*>(
          m_state.m_statelet[sglStatelet::eANTI_ALIAS_POLYGON])->isEnabled())
      {
         GLenum fill_sides = GL_FRONT_AND_BACK;
         if (static_cast<const sglPolyFrontStyle*>(
            m_state.m_statelet[sglStatelet::ePOLY_FRONT_STYLE])->getStyle()
             == sglPolyStyle::eFILLED)
         {
            fill_sides = GL_FRONT;
         }
         if (static_cast<const sglPolyBackStyle*>(
            m_state.m_statelet[sglStatelet::ePOLY_BACK_STYLE])->getStyle()
             == sglPolyStyle::eFILLED)
         {
            if (fill_sides == GL_FRONT)
               fill_sides = GL_FRONT_AND_BACK;
            else
               fill_sides = GL_BACK;
         }

         bool cull_changed = false;
         if (fill_sides == GL_FRONT)
         {
            sglCullMode::CullModeEnum cull_mode =
               static_cast<const sglCullMode*>(
                  m_state.m_statelet[sglStatelet::eCULL_MODE])->getCullMode();

            if (cull_mode == sglCullMode::eOFF)
            {
               glCullFace(GL_BACK);
               glEnable(GL_CULL_FACE);
               cull_changed = true;
            }
            else if (cull_mode == sglCullMode::eFRONT)
            {
               glCullFace(GL_FRONT_AND_BACK);
               glEnable(GL_CULL_FACE);
               cull_changed = true;
            }
         }
         else if (fill_sides == GL_BACK)
         {
            sglCullMode::CullModeEnum cull_mode =
               static_cast<const sglCullMode*>(
                  m_state.m_statelet[sglStatelet::eCULL_MODE])->getCullMode();

            if (cull_mode == sglCullMode::eOFF)
            {
               glCullFace(GL_FRONT);
               glEnable(GL_CULL_FACE);
               cull_changed = true;
            }
            else if (cull_mode == sglCullMode::eBACK)
            {
               glCullFace(GL_FRONT_AND_BACK);
               glEnable(GL_CULL_FACE);
               cull_changed = true;
            }
         }

         if (!m_line_smooth)
         {
            m_line_smooth = true;
            glEnable(GL_LINE_SMOOTH);
            checkBlend();
         }

         glPushAttrib(GL_DEPTH_BUFFER_BIT);
         glDepthFunc(GL_LEQUAL);
         glDepthMask(GL_FALSE);
         glPolygonMode(fill_sides, GL_LINE);
         drawGeometry(drawable);
         glPolygonMode(fill_sides, GL_FILL);
         glPopAttrib();

         if (cull_changed)
         {
            m_state.m_statelet[sglStatelet::eCULL_MODE]->apply(this);
         }
      }
   }

   if (static_cast<const sglHighlight*>(
      m_state.m_statelet[sglStatelet::eHIGHLIGHT])->isEnabled())
   {
      glColor3f(1.0f,1.0f,1.0f);
      drawable->getPrevBound().drawImmediate();
   }
}

