/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglDataPtr.hpp
 *   Author: Tom Stimson
 *  Created: 6 April 2002
 *  Summary: Reference counted container for data pointers.
 *****************************************************************************/

#ifndef __SGL_DATA_PTR_HPP
#define __SGL_DATA_PTR_HPP

#include <sglObject.hpp>

/** Abstract base class for reference counted containers of data pointers.
*/
class SGL_DLL_API sglDataPtr : public sglObject
{
public:
   /** Enumeration of all data types supported by SGL. */
   enum DataTypeEnum
   {
      eBYTE,
      eUNSIGNED_BYTE,
      eSHORT,
      eUNSIGNED_SHORT,
      eINT,
      eUNSIGNED_INT,
      eFLOAT,
      eDOUBLE
   };

public:
   /** Constructor. */
   sglDataPtr() {}

   /** Destructor. */
   virtual ~sglDataPtr() {}

   /** Get the data pointer.
       @return Pointer to the data.
   */
   virtual void *getData() const = 0;

   /** Get the size of a data type.
       @param type One of the data types supported by SGL
       @return number of bytes requred for the given type
   */
   static int getSizeOfType(DataTypeEnum type)
      { return s_size[type]; }

   /** Get the GLenum equivalent for the given data type.
       @param type One of the data types supported by SGL
       @return The GLenum equivalent for the given type
   */
   static GLenum getGLType(DataTypeEnum type)
      { return s_gltype[type]; }

private:
   // Not Implemented.
   sglDataPtr(const sglDataPtr &);
   sglDataPtr &operator=(const sglDataPtr &);

private:
   static int    s_size[8];
   static GLenum s_gltype[8];
};


/** Reference counted container for a data pointer.
    Upon destruction the data pointer will be deleted.
*/
template <class T>
class SGL_DLL_API sglDeleteableDataPtr : public sglDataPtr
{
public:
   /** Constructor. */
   sglDeleteableDataPtr(T *ptr) : m_data(ptr) {}

   /** Destructor. */
   virtual ~sglDeleteableDataPtr() { delete[] m_data; }

   // See sglDataPtr::getData() for details.
   virtual void *getData() const { return m_data; }

private:
   // Not Implemented.
   sglDeleteableDataPtr(const sglDeleteableDataPtr &);
   sglDeleteableDataPtr &operator=(const sglDeleteableDataPtr &);

private:
   T *m_data;
};


/** Reference counted container for a data pointer.
    Upon destruction the data pointer will be freed.
*/
class SGL_DLL_API sglFreeableDataPtr : public sglDataPtr
{
public:
   /** Constructor. */
   sglFreeableDataPtr(void *ptr) : m_data(ptr) {}

   /** Destructor. */
   virtual ~sglFreeableDataPtr() { if (m_data) free(m_data); }

   // See sglDataPtr::getData() for details.
   virtual void *getData() const { return m_data; }

private:
   // Not Implemented.
   sglFreeableDataPtr(const sglFreeableDataPtr &);
   sglFreeableDataPtr &operator=(const sglFreeableDataPtr &);

private:
   void *m_data;
};


/** Reference counted container for a data pointer.
    Upon destruction the data pointer will NOT be deleted or freed.
    It is the user's responsibility to manage this memory.
*/
class SGL_DLL_API sglNoRefDataPtr : public sglDataPtr
{
public:
   /** Constructor. */
   sglNoRefDataPtr(void *ptr) : m_data(ptr) {}

   /** Destructor. */
   virtual ~sglNoRefDataPtr() {}

   // See sglDataPtr::getData() for details.
   virtual void *getData() const { return m_data; }

private:
   // Not Implemented.
   sglNoRefDataPtr(const sglNoRefDataPtr &);
   sglNoRefDataPtr &operator=(const sglNoRefDataPtr &);

private:
   void *m_data;
};


#endif
