/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglDefaultStateNode.hpp
 *   Author: Tom Stimson
 *  Created: 21 July 2000
 *  Summary: change the default state of all nodes below this one
 *****************************************************************************/

#ifndef __SGL_DEFAULTSTATENODE_HPP
#define __SGL_DEFAULTSTATENODE_HPP

#include <sgl.h>
#include <sglGroup.hpp>

//----------------------------------------------------------------------------

/**

An sglDefaultStateNode is a container of sglStatelets that affect the default
state of all geometry below it in the graph.  The statelets in this node take
precedence over statelets in sglDefaultStateNodes higher in the tree, but not
those lower in the tree.  They also have a lower priority than statelets in the
sglDrawable and sglOverrideStateNode nodes (above or below).

<b>Example:</b>

<pre>
   sglScene *root = new sglScene;

   sglCullMode *cull_mode = new sglCullMode;
   cull_mode->setCullMode(sglCullMode::OFF);

   vector<sglStatelet*> default_statelets;
   default_statelets.push_back(cull_mode);

   // sets the default cull mode to OFF for all nodes below this one.
   sglDefaultStateNode *default_node = new sglDefaultStateNode;
   default_node->setStatelets(default_statelets);

   root->addChild(def);
</pre>

 */

class SGL_DLL_API sglDefaultStateNode : public sglGroup
{
public:
   /// default constructor
   sglDefaultStateNode();
   /// virtual destructor
   virtual ~sglDefaultStateNode();

   /** Set the current set of default statelets from the vector provided.  The
       new statelets' reference counts are incremented, while the previous
       statelets are decremented, and if their reference counts go to zero they
       are deleted.
       @param statelets Vector of new statelets (that are dynamically
         allocated).
   */
   void setStatelets(const vector<sglStatelet*> &statelets);

   /** Get a reference to the current set of default statelets.
       @return A reference to the internal list of statelets.
   */
   const vector<sglStatelet*> &getStatelets() const {return m_statelets;}

   // See sglNode::clone() for details.
   virtual sglNode *clone(unsigned int mode) const;

   // See sglObject::printInfo() for details.
   virtual void printInfo(ostream &ostrm, const char *indent_string) const;

protected:
   // See sglNode::cull(f) for details
   virtual void cull(sglCull<float> &trav_state,
                     unsigned int cull_flags) const;

   // See sglNode::cull(d) for details
   virtual void cull(sglCull<double> &trav_state,
                     unsigned int cull_flags) const;

   void copyTo(sglDefaultStateNode *dst, unsigned int mode) const;

private:  // not implemented
   sglDefaultStateNode(const sglDefaultStateNode &);
   sglDefaultStateNode &operator=(const sglDefaultStateNode &);

protected:
   vector<sglStatelet*>  m_statelets;
};

#endif
