/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglDirectionalLight.hpp
 *   Author: Scott McMillan
 *  Created: 9 December 2000
 *  Summary: encapsulates an infinite light source
 *****************************************************************************/

#ifndef __SGL_DIRECTIONAL_LIGHT_HPP
#define __SGL_DIRECTIONAL_LIGHT_HPP

#include <sgl.h>
#include <sglLight.hpp>

//----------------------------------------------------------------------------

/**

sglDirectionalLight is a subclass of sglLight, and defines light that
comes from an infinite distance away in a specified direction.  The direction
can be set by the user by calling sglDirectionalLight::setDirection(..) with
a vector in the direction that the light comes from.  This direction will be
defined relative to this node's coordinate system, so that any rotations that
occur in the scene graph above this node will affect this direction.

The bounding sphere for this light is set to infinite by default.

<b>Example:</b>

<pre>
   sglDirectionalLight *dlite = new sglDirectionalLight;
   dlite->setDiffuse(1.0f, 0.5f, 0.5f, 1.0f);    // reddish light
   dlite->setDirection(sglVec3f(0., 0., 1.));    // shines along -Z axis
</pre>

*/

class SGL_DLL_API sglDirectionalLight : public sglLight
{
public:
   /// default constructor
   sglDirectionalLight();
   /// virtual destructor
   virtual ~sglDirectionalLight();

   /** Set the direction of the light.  The default value is (0,0,1).
       @param dir Direction vector (does not need to be normalized).
   */
   void setDirection(const sglVec3f &dir);

   /** Get the current direction vector.
       @param dir Reference to vector that will be set to the current
       direction.
   */
   void getDirection(sglVec3f &dir) const;

   // See sglLight::setup() for details.
   virtual void setup(GLenum gl_light_number) const;

   // See sglNode::clone() for details.
   virtual sglNode *clone(unsigned int mode) const;

   // See sglObject::printInfo() for details.
   virtual void printInfo(ostream &ostrm, const char *indent_string) const;

protected:
   virtual void computeBound();

   void copyTo(sglDirectionalLight *dst, unsigned int mode) const;

private: // methods - not implemented
   sglDirectionalLight(const sglLight &);
   sglDirectionalLight &operator=(const sglDirectionalLight &);

protected: // variables
   sglVec4f m_direction;  // 4th element is always zero
};

#endif
