/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglDiscriminator.hpp
 *  Created: 21 February 1999
 *  Summary: use a discriminator mask to select/deselect the entire subtree
 *****************************************************************************/

#ifndef __SGL_DISCRIMINATOR_HPP
#define __SGL_DISCRIMINATOR_HPP

#include <sgl.h>
#include <sglGroup.hpp>

//----------------------------------------------------------------------------

/**

This is a subclass of sglGroup contains a discriminator bit mask.
If the discriminator mask ANDed with the mask passed in with any of the
traversals returns a true (non-zero) value, then the children
are traversed.  If this test returns false (zero), the traversal down
this branch of the scene graph is halted at this point.

The discriminator mask is currently implemented as an unsigned int.
On most systems this corresponds to a 32-bit quantity which allows for
up to 32 unique masks.

<b>Example:</b>

<pre>
   sglGeode *geode = new sglGeode;
   sglDiscriminator *disc_node = new sglDiscriminator;
   disc_node->setMask(0x01);
   disc_node->addChild(geode);
</pre>

*/

class SGL_DLL_API sglDiscriminator : public sglGroup
{
public:
   /// default constructor
   sglDiscriminator();
   /// virtual destructor
   virtual ~sglDiscriminator();

   /** Set the discriminator mask for this node.
       @param mask The discriminator mask.
   */
   void setMask(unsigned int mask) { m_mask = mask; }

   /** Get the current discriminator mask.
       @returns the current discriminator mask.
   */
   unsigned int getMask() const { return m_mask; }

   // see sglNode::intersect(f) for details.
   virtual sglBound::IntersectResultEnum intersect(
      sglIntersectf &isector) const;

   // see sglNode::intersect(d) for details.
   virtual sglBound::IntersectResultEnum intersect(
      sglIntersectd &isector) const;

   // see sglNode::pick(f) for details.
   virtual void pick(sglPickf &pick_state, unsigned int cull_flags) const;

   // see sglNode::pick(d) for details.
   virtual void pick(sglPickd &pick_state, unsigned int cull_flags) const;

   // See sglNode::clone() for details.
   virtual sglNode *clone(unsigned int mode) const;

   // See sglObject::printInfo() for details.
   virtual void printInfo(ostream &ostrm, const char *indent_string) const;

protected:
   // compute the bounding sphere NOT YET IMPLEMENTED (using sglGroup)
   //virtual void computeBound();

   // sglNode::cull(f) for details
   virtual void cull(sglCull<float> &trav_state,
                     unsigned int cull_flags) const;

   // sglNode::cull(d) for details
   virtual void cull(sglCull<double> &trav_state,
                     unsigned int cull_flags) const;

   void copyTo(sglDiscriminator *dst, unsigned int mode) const;

private:  // not implemented
   sglDiscriminator(const sglDiscriminator &);
   sglDiscriminator &operator=(const sglDiscriminator &);

protected:
   unsigned int m_mask;
};

#endif
