/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglDistanceSwitch.cpp
 *  Created: 24 January 2002
 *****************************************************************************/

#include <sgl.h>
#include <sglVector.hpp>
#include <sglMatrix.hpp>

#include <sglDistanceSwitch.hpp>
#include <sglCull.hpp>

float sglDistanceSwitch::s_range_scale = 1.0f;

//----------------------------------------------------------------------------
sglDistanceSwitch::sglDistanceSwitch()
      : sglGroup(),
        m_center(0.0f, 0.0f, 0.0f),
        m_range(0.f),
        m_lod_state(NULL)
{
}

//----------------------------------------------------------------------------
sglDistanceSwitch::~sglDistanceSwitch()
{
   if (m_lod_state)
   {
      m_lod_state->deref();
      if (m_lod_state->getRefCount() == 0)
      {
         delete m_lod_state;
      }
      m_lod_state = NULL;
   }
}

//----------------------------------------------------------------------------
void sglDistanceSwitch::setRange(float range)
{
   if (range > 0.0f)
      m_range = range;
   else
      m_range = 0.f;
}

//----------------------------------------------------------------------------
void sglDistanceSwitch::setLODState(sglLODState *lod_state)
{
   if (m_lod_state != lod_state)
   {
      if (m_lod_state)
      {
         m_lod_state->deref();
         if (m_lod_state->getRefCount() == 0)
         {
            delete m_lod_state;
         }
         m_lod_state = NULL;
      }

      if (lod_state)
      {
         m_lod_state = lod_state;
         m_lod_state->ref();
      }
   }
}

//----------------------------------------------------------------------------
sglBound::IntersectResultEnum sglDistanceSwitch::intersect(
   sglIntersectf &segset) const
{
   if (segset.getMode()&sglProbe::eLOD_NONE || m_child_list.empty())
   {
      return sglBound::eOUTSIDE;
   }
   else if (segset.getMode()&sglProbe::eLOD_ALL)
   {
      return sglGroup::intersect(segset);
   }
   else if (segset.getMode()&sglProbe::eLOD_VIEW)
   {
      if (m_range > computeRange(segset.getLODScale(),
                                 segset.getLODOffset(),
                                 segset.getModelViewMatrix()))
      {
         return sglGroup::intersect(segset);
      }
      else
      {
         return sglBound::eOUTSIDE;
      }
   }
   else // default: intersect against range 0 only
   {
      return sglGroup::intersect(segset);
   }
}

//----------------------------------------------------------------------------
sglBound::IntersectResultEnum sglDistanceSwitch::intersect(
   sglIntersectd &segset) const
{
   if (segset.getMode()&sglProbe::eLOD_NONE || m_child_list.empty())
   {
      return sglBound::eOUTSIDE;
   }
   else if (segset.getMode()&sglProbe::eLOD_ALL)
   {
      return sglGroup::intersect(segset);
   }
   else if (segset.getMode()&sglProbe::eLOD_VIEW)
   {
      if (m_range > computeRange(segset.getLODScale(),
                                 segset.getLODOffset(),
                                 segset.getModelViewMatrix()))
      {
         return sglGroup::intersect(segset);
      }
      else
      {
         return sglBound::eOUTSIDE;
      }
   }
   else // default: intersect against range 0 only
   {
      return sglGroup::intersect(segset);
   }
}

//----------------------------------------------------------------------------
void sglDistanceSwitch::pick(sglPickf &pick_state,
                             unsigned int cull_flags) const
{
   if (pick_state.getMode()&sglProbe::eLOD_NONE || m_child_list.empty())
   {
      return; // select nothing
   }
   else if (pick_state.getMode()&sglProbe::eLOD_ALL)
   {
      sglGroup::pick(pick_state, cull_flags);
      return;
   }
   else if (pick_state.getMode()&sglProbe::eLOD_VIEW)
   {
      if (m_range > computeRange(pick_state.getLODScale(),
                                 pick_state.getLODOffset(),
                                 pick_state.getModelViewMatrix()))
      {
         sglGroup::pick(pick_state, cull_flags);
      }
      return;
   }
   else // default: pick against range 0 only
   {
      sglGroup::pick(pick_state, cull_flags);
      return;
   }
}

//----------------------------------------------------------------------------
void sglDistanceSwitch::pick(sglPickd &pick_state,
                             unsigned int cull_flags) const
{
   if (pick_state.getMode()&sglProbe::eLOD_NONE || m_child_list.empty())
   {
      return;  // select nothing
   }
   else if (pick_state.getMode()&sglProbe::eLOD_ALL)
   {
      sglGroup::pick(pick_state, cull_flags);
      return;
   }
   else if (pick_state.getMode()&sglProbe::eLOD_VIEW)
   {
      if (m_range > computeRange(pick_state.getLODScale(),
                                 pick_state.getLODOffset(),
                                 pick_state.getModelViewMatrix()))
      {
         sglGroup::pick(pick_state, cull_flags);
      }
      return;
   }
   else // default: pick against range 0 only
   {
      sglGroup::pick(pick_state, cull_flags);
      return;
   }
}

//----------------------------------------------------------------------------
void sglDistanceSwitch::cull(sglCull<float> &trav_state,
                             unsigned int cull_flags) const
{
   if (!m_child_list.empty() &&
       m_range > computeRange(trav_state.getLODScale(),
                              trav_state.getLODOffset(),
                              trav_state.getModelViewMatrix()))
   {
      sglGroup::cull(trav_state, cull_flags);
   }
}

//----------------------------------------------------------------------------
void sglDistanceSwitch::cull(sglCull<double> &trav_state,
                             unsigned int cull_flags) const
{
   if (!m_child_list.empty() &&
       m_range > computeRange(trav_state.getLODScale(),
                              trav_state.getLODOffset(),
                              trav_state.getModelViewMatrix()))
   {
      sglGroup::cull(trav_state, cull_flags);
   }
}

//----------------------------------------------------------------------------
float sglDistanceSwitch::computeRange(float trav_lod_scale,
                                      float trav_lod_offset,
                                      const sglMat4f &view_matrix) const
{
   sglVec3d offset;
   for (unsigned int j=0; j<3; j++)
   {
      offset[j] =
         view_matrix[0][j]*m_center[0] +
         view_matrix[1][j]*m_center[1] +
         view_matrix[2][j]*m_center[2] +
         view_matrix[3][j];
   }

   // HACK - not really sure where the FOV-based LOD scale (modified by any
   // sglLODStateNodes along the way) is supposed to be applied.  Multiplying
   // the range scale portion for now.  Not really sure about the offset either

   float range;
   if (m_lod_state)
   {
      range = trav_lod_scale *
         (s_range_scale * (m_lod_state->getRangeScale() *
                           sqrt(offset[0]*offset[0] +
                                offset[1]*offset[1] +
                                offset[2]*offset[2]) +
                           m_lod_state->getRangeOffset())) +
         trav_lod_offset;
   }
   else
   {
      range = trav_lod_scale *
         (s_range_scale * sqrt(offset[0]*offset[0] +
                               offset[1]*offset[1] +
                               offset[2]*offset[2])) +
         trav_lod_offset;
   }

   return range;
}

//----------------------------------------------------------------------------
float sglDistanceSwitch::computeRange(float trav_lod_scale,
                                      float trav_lod_offset,
                                      const sglMat4d &view_matrix) const
{
   sglVec3d offset;
   for (unsigned int j=0; j<3; j++)
   {
      offset[j] =
         view_matrix[0][j]*m_center[0] +
         view_matrix[1][j]*m_center[1] +
         view_matrix[2][j]*m_center[2] +
         view_matrix[3][j];
   }

   // HACK - not really sure where the FOV-based LOD scale (modified by any
   // sglLODStateNodes along the way) is supposed to be applied.  Multiplying
   // the range scale portion for now.  Not really sure about the offset either

   float range;
   if (m_lod_state)
   {
      range = trav_lod_scale *
         (s_range_scale * (m_lod_state->getRangeScale() *
                           sqrt(offset[0]*offset[0] +
                                offset[1]*offset[1] +
                                offset[2]*offset[2]) +
                           m_lod_state->getRangeOffset())) +
         trav_lod_offset;
   }
   else
   {
      range = trav_lod_scale *
         (s_range_scale * sqrt(offset[0]*offset[0] +
                               offset[1]*offset[1] +
                               offset[2]*offset[2])) +
         trav_lod_offset;
   }

   return range;
}

//----------------------------------------------------------------------------
sglNode *sglDistanceSwitch::clone(unsigned int mode) const
{
   sglDistanceSwitch *root = new sglDistanceSwitch;
   sglDistanceSwitch::copyTo(root, mode);
   return root;
}

//----------------------------------------------------------------------------
void sglDistanceSwitch::copyTo(sglDistanceSwitch *dst, unsigned int mode) const
{
   sglGroup::copyTo(dst, mode);

   // transfer stuff
   dst->setRange(m_range);
   dst->setCenter(m_center);
   dst->setLODState(m_lod_state);
}

//----------------------------------------------------------------------------
void sglDistanceSwitch::printInfo(ostream &ostrm, const char *indent_string) const
{
   sglGroup::printInfo(ostrm, indent_string);
   ostrm << indent_string << "\t range:  " << m_range << endl;
   ostrm << indent_string << "\t center: "
         << m_center[0] << ", " << m_center[1] << ", " << m_center[2] << endl;
   if (m_lod_state)
   {
      m_lod_state->printInfo(ostrm, indent_string);
   }
}
