/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglDistanceSwitch.hpp
 *  Created: 24 January 2002
 *  Summary: select (or not) all the children to be traversed based on a single
 *           range value (a very specialized LOD node).
 *****************************************************************************/

#ifndef __SGL_DISTANCE_SWITCH_HPP
#define __SGL_DISTANCE_SWITCH_HPP

#include <sgl.h>
#include <vector>

#include <sglVector.hpp>
#include <sglGroup.hpp>
#include <sglLODState.hpp>

//----------------------------------------------------------------------------

/**

This subclass of sglGroup will select all or none of its children to be
traversed based on a function of its range from the eye point (the "origin" of
the view frustum) and what is defined as the center of this node.  This acts
like a simplified LOD node where there is a single range threshold.

After the children have been added, the user specifies the range and specifies
the center point which is a reference point relative to the origin of the local
coordinate system of this node used to compute the range to the viewpoint.
This range value can also be scaled and biased to change which child is
selected.  Unlike LOD, the order that the children are added to the node is NOT
significant.

The children are selected for further traversal according to the following
table:

<pre>
  Range Value, x    Child selected:
 -----------------  ---------------
   x <= distance    all children
   x >  distance    none
</pre>

This node is designed to make a specific construct more efficient.  With the
sglLOD node you might have the following subgraph to switch a set of children
with a single range (i.e., turn them off if they are too far away):
<pre>
      sglLOD (range[0] = 0, range[1] = view_distance)
         |
     sglGroup
     /   |   \
 Node1 Node2 Node3
</pre>

This can be replaced with a more efficient construct as follows:
<pre>
 sglSwitchDistance (range = view_distance)
      /  |  \
     /   |   \
 Node1 Node2 Node3
</pre>

<b>Scale and Bias</b>

See sglLOD documentation for details.

<b>Special behaviour for intersect and pick traversals</b>

This node uses the same flags as sglLOD.  See sglLOD documentation for
details.  Note that the default behaviour for sglLOD (to intersect against
range zero only) is equivalent to using the eLOD_ALL flag in this node.

<b>Example:</b>

<pre>
   sglDistanceSwitch *ds_node = new sglDistanceSwitch;
   ds_node->setLODState(new sglLODState);

   sglGeode *geode1 = new sglGeode;
   sglGeode *geode2 = new sglGeode;
   sglGeode *geode3 = new sglGeode;

   ds_node->addChild(geode1);     // finest resolution
   ds_node->addChild(geode2);
   ds_node->addChild(geode3);     // coarsest resolution

   ds_node->setRange(50.0f);      // deselect all above 50.

   // double computed range for all sglDistanceSwitch nodes: has effect of
   // halving the switch range.
   sglDistanceSwitch::setScaleFactor(2.0f);

   // halve the computed range and add a bias for this LOD only: has effect
   // of selecting finer resolution children
   sglLODState *lod_state = ds_node->getLODState();
   if (lod_state)
   {
      lod_state->setRangeScale(0.5f);
      lod_state->setRangeOffset(-2.0f);
   }
</pre>

*/

class SGL_DLL_API sglDistanceSwitch : public sglGroup
{
public:
   /// default constructor
   sglDistanceSwitch();
   /// virtual destructor
   virtual ~sglDistanceSwitch();

   /** Set a range value.  Distances computed outside this range will result in
       all children being culled.
       @param range The distance switch range.  Negative values is equivalent
       to zero (the default) and result in children being culled all the time.
   */
   void setRange(float range);

   /** Get a range value.
       @returns range value.
   */
   float getRange() const { return m_range; }

   /** Set the reference position for computing a range to.  This defaults to
       the origin of this node's local coordinate system.
       @param center The reference position.
   */
   void setCenter(const sglVec3f &center) { m_center = center; }

   /** Get the current reference position.
       @returns The current reference position.
    */
   const sglVec3f &getCenter() const { return m_center; }

   /** Set a global scale factor used to affect transitions of groups of ALL
       SwitchDistance nodes (this is a separate factor from sglLOD -- should
       they share the same value?).
       @param scale The scale factor (should be positive).
   */
   static void setScaleFactor(float scale) { s_range_scale = scale; }

   /** Get the current global scale factor.
       @returns The current scale factor.
   */
   static float getScaleFactor() { return s_range_scale; }

   /** Specify a scale and bias to affect transitions of this SwitchDistance
       node.  By using the same sglLODState object for more than one
       sglDistanceSwitch (and sglLOD) node, a group of sglDistanceSwitch's
       range calculation can be affected simultaneously.  These objects are
       reference counted with this function.  New objects are
       ref()-ed while previous objects are deref()-ed.  If an sglLODState
       object's reference count goes to zero when it is replaced by this
       function, it will be deleted.
       @param lod_scale The new sglLODState object.
       @see sglLODState
   */
   void setLODState(sglLODState *lod_state);

   /** Get a pointer to the current sglLODState object.
       @return Pointer to the current sglLODState object.  NULL if no object
       has be assigned.
   */
   sglLODState *getLODState() const { return m_lod_state; }

   // see sglNode::intersect(f) for details.
   virtual sglBound::IntersectResultEnum intersect(
      sglIntersectf &isector) const;

   // see sglNode::intersect(d) for details.
   virtual sglBound::IntersectResultEnum intersect(
      sglIntersectd &isector) const;

   // see sglNode::pick(f) for details.
   virtual void pick(sglPickf &pick_state, unsigned int cull_flags) const;

   // see sglNode::pick(d) for details.
   virtual void pick(sglPickd &pick_state, unsigned int cull_flags) const;

   // See sglNode::clone() for details.
   virtual sglNode *clone(unsigned int mode) const;

   // See sglObject::printInfo() for details.
   virtual void printInfo(ostream &ostrm, const char *indent_string) const;

protected:
   // See sglNode::cull(f) for details
   virtual void cull(sglCull<float> &trav_state,
                     unsigned int cull_flags) const;

   // See sglNode::cull(d) for details
   virtual void cull(sglCull<double> &trav_state,
                     unsigned int cull_flags) const;

   // If a user wants to subclass this node and replace the equation used for
   // determining the range used in child selection, the following function
   // should be reimplemented in the subclass (one each for single and double
   // precision traversals.
   virtual float computeRange(float trav_lod_scale,
                              float trav_lod_offset,
                              const sglMat4f &view_matrix) const;
   virtual float computeRange(float trav_lod_scale,
                              float trav_lod_offset,
                              const sglMat4d &view_matrix) const;

   void copyTo(sglDistanceSwitch *dst, unsigned int mode) const;

private: // not implemented
   sglDistanceSwitch(const sglDistanceSwitch &);
   sglDistanceSwitch &operator=(const sglDistanceSwitch &);

protected:
   static float s_range_scale;

   sglVec3f     m_center;
   float        m_range;

   sglLODState *m_lod_state;
};

#endif
