/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglDrawable.cpp (was sglGeometry before merging with sglDrawable).
 *  Created: 3 January 2000
 *  Summary: abstract base class for sglGeoSet and sglInterleavedGeometry
 *****************************************************************************/

#include <sglGeode.hpp>
#include <sglDrawable.hpp>
#include <sglOStream.hpp>
#include <sglIntersect.hpp>

//----------------------------------------------------------------------------
sglDrawable::StateElement::StateElement(unsigned int mask)
   : m_mask(mask), m_statelets(),
     m_tex_coords(), m_dlist(0), m_dirty(true)
{
}

//----------------------------------------------------------------------------
sglDrawable::StateElement::StateElement(unsigned int mask,
                                        const vector<sglStatelet*> &statelets,
                                        const vector<sglTexCoords> &tex_coords)
   : m_mask(mask), m_statelets(statelets),
     m_tex_coords(tex_coords), m_dlist(0), m_dirty(true)
{
   vector<sglStatelet*>::iterator end = m_statelets.end();
   for (vector<sglStatelet*>::iterator ii = m_statelets.begin();
        ii != end; ++ii)
   {
      (*ii)->ref();
   }
}

//----------------------------------------------------------------------------
sglDrawable::StateElement::StateElement(unsigned int mask,
                                        const vector<sglStatelet*> &statelets,
                                        const sglTexCoords &tex_coords)
   : m_mask(mask), m_statelets(statelets),
     m_tex_coords(), m_dlist(0), m_dirty(true)
{
   vector<sglStatelet*>::iterator end = m_statelets.end();
   for (vector<sglStatelet*>::iterator ii = m_statelets.begin();
        ii != end; ++ii)
   {
      (*ii)->ref();
   }

   if (tex_coords.isValid())
      m_tex_coords.push_back(tex_coords);
}

//----------------------------------------------------------------------------
sglDrawable::StateElement::~StateElement()
{
   vector<sglStatelet*>::iterator end = m_statelets.end();
   for (vector<sglStatelet*>::iterator ii = m_statelets.begin();
        ii != end; ++ii)
   {
      (*ii)->deref();
      if ((*ii)->getRefCount() == 0)
         delete *ii;
   }

   if (m_dlist)
      sgl::releaseDisplayList(m_dlist);
}

//----------------------------------------------------------------------------
void sglDrawable::StateElement::setStatelets(
   const vector<sglStatelet*> &statelets)
{
   vector<sglStatelet*>::const_iterator end = statelets.end();
   vector<sglStatelet*>::const_iterator ii = statelets.begin();
   for (; ii != end; ++ii)
   {
      (*ii)->ref();
   }
   end = m_statelets.end();
   for (ii = m_statelets.begin(); ii != end; ++ii)
   {
      (*ii)->deref();
      if ((*ii)->getRefCount() == 0)
         delete *ii;
   }
   m_statelets = statelets;
}

//----------------------------------------------------------------------------
void sglDrawable::StateElement::clearStatelets()
{
   vector<sglStatelet*>::iterator end = m_statelets.end();
   for (vector<sglStatelet*>::iterator ii = m_statelets.begin();
        ii != end; ++ii)
   {
      (*ii)->deref();
      if ((*ii)->getRefCount() == 0)
         delete *ii;
   }
   m_statelets.clear();
}

//----------------------------------------------------------------------------
void sglDrawable::StateElement::setTexCoordList(
   const vector<sglTexCoords> &tex_coords)
{
   m_tex_coords = tex_coords;
   m_dirty = true;
}

//----------------------------------------------------------------------------
void sglDrawable::StateElement::setTexCoordList(
   const sglTexCoords &tex_coords)
{
   m_tex_coords.clear();
   if (tex_coords.isValid())
      m_tex_coords.push_back(tex_coords);
   m_dirty = true;
}

//----------------------------------------------------------------------------
void sglDrawable::StateElement::clearTexCoordList()
{
   m_tex_coords.clear();
   m_dirty = true;
}

//----------------------------------------------------------------------------
void sglDrawable::StateElement::setDisplayList(GLuint dlist)
{
   m_dlist = dlist;
   m_dirty = true;
}

//----------------------------------------------------------------------------
void sglDrawable::StateElement::printInfo(
   ostream &ostrm, const char *indent_string) const
{
   ostrm << indent_string << "mask = 0x" << hex << m_mask << dec
         << ", texcoords =";
   if (m_tex_coords.empty())
      ostrm << " NULL";
   else
   {
      for (vector<sglTexCoords>::const_iterator tt = m_tex_coords.begin();
           tt != m_tex_coords.end(); ++tt)
      {
         tt->printInfo(ostrm, indent_string);
      }
   }
   ostrm << ", dlist = " << m_dlist
         << ", " << m_statelets.size() << " statelet"
         << ((m_statelets.size() != 1) ? "s = " : " = ") << endl;
   vector<sglStatelet*>::const_iterator end = m_statelets.end();
   for (vector<sglStatelet*>::const_iterator ii = m_statelets.begin();
        ii != end; ++ii)
   {
      ostrm << indent_string << "................................"
            << endl;
      (*ii)->printInfo(ostrm, indent_string);
      ostrm << indent_string << "................................"
            << endl;
   }
}

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------

//----------------------------------------------------------------------------
sglDrawable::sglDrawable()
      : sglObject(),
        m_bbox(),
        m_state_list(),
        m_use_dlist(true),
        m_draw_func(NULL),
        m_bound_mode(sglBound::eDYNAMIC),
        m_dirty(eBOUND|eDLIST)
{
   // set up the first state element -- the default state (empty)
   m_state_list.push_back(new StateElement(0));
}

//----------------------------------------------------------------------------
sglDrawable::~sglDrawable()
{
   // clean up any reference counted objects contained in this class
   while (m_state_list.size())
   {
      StateElement *state = m_state_list.back();
      m_state_list.pop_back();
      delete state;
   }

   // dereference but do not delete any parent geodes.
   while (m_parent_list.size())
   {
      sglGeode *geode = m_parent_list.back();
      geode->removeGeometry(this);
   }
}

//----------------------------------------------------------------------------
sglGeode *sglDrawable::getParent(unsigned int i)
{
   if (i < getNumParents())
   {
      return m_parent_list[i];
   }
   return 0;
}

//----------------------------------------------------------------------------
void sglDrawable::addParent(sglGeode *parent)
{
   if (parent)
   {
      m_parent_list.push_back(parent);
      parent->dirty(m_dirty|eBOUND);
   }
}

//----------------------------------------------------------------------------
void sglDrawable::removeParent(sglGeode *parent)
{
   if (parent)
   {
      vector<sglGeode*>::iterator elt;

      for (elt=m_parent_list.begin();
           elt!=m_parent_list.end() && (*elt != parent);
           ++elt) ;
      {
         if (*elt == parent)
         {
            m_parent_list.erase(elt);
         }
      }
      parent->dirty(eBOUND);
   }
}

//----------------------------------------------------------------------------
sglDrawable::StateElement* sglDrawable::getState(unsigned int index)
{
   StateElement *state = NULL;
   if (index < m_state_list.size())
   {
      state = m_state_list[index];
   }

   return state;
}

//----------------------------------------------------------------------------
sglDrawable::StateElement *sglDrawable::getStateByMask(unsigned int mask) const
{
   // Select the proper state element.  Defaults to the first element -
   // mask == 0 - when no mask matches
   StateElement *state = m_state_list[0];

   if (mask)
   {
       for (vector<StateElement*>::const_iterator elt=m_state_list.begin();
            elt != m_state_list.end();
            ++elt)
       {
           if (mask & ((*elt)->getMask()))
           {
               state = *elt;
               break;
           }
       }
   }

   return state;
}

//----------------------------------------------------------------------------
sglDrawable::StateElement *sglDrawable::findState(unsigned int mask) const
{
   StateElement *state = NULL;
   vector<StateElement*>::const_iterator elt = m_state_list.begin();
   for (; elt != m_state_list.end(); ++elt)
   {
      if (mask == ((*elt)->getMask()))
      {
         state = *elt;
         break;
      }
   }
   return state;
}

//----------------------------------------------------------------------------
void sglDrawable::setState(unsigned int mask,
                           const vector<sglStatelet*> &statelets)
{
   sglTexCoords tmp;
   setState(mask, statelets, tmp);
}

//----------------------------------------------------------------------------
void sglDrawable::setState(unsigned int mask,
                           const vector<sglStatelet*> &statelets,
                           sglVec2f *tex_coords)
{
   sglTexCoords tmp(new sglNoRefDataPtr(tex_coords));
   setState(mask, statelets, tmp);
}

//----------------------------------------------------------------------------
void sglDrawable::setState(unsigned int mask,
                           const vector<sglStatelet*> &statelets,
                           const sglTexCoords &tex_coords)
{
   // if mask already exists in state list then replace the entry
   StateElement *state = findState(mask);

   // if match was not found in list, make a new one and push it on back of the
   // state list
   if (state == NULL)
   {
      m_state_list.push_back(new StateElement(mask, statelets, tex_coords));
      dirty(eDLIST, false);
   }
   else // change the values of the existing state element
   {
      state->setStatelets(statelets);
      if (tex_coords.isValid() ||
          !state->getTexCoordList().empty())
      {
         state->setTexCoordList(tex_coords);
         dirty(eDLIST, false);
      }
   }
}

//----------------------------------------------------------------------------
void sglDrawable::setState(unsigned int mask,
                           const vector<sglStatelet*> &statelets,
                           const vector<sglTexCoords> &tex_coords)
{
   // if mask already exists in state list then replace the entry
   StateElement *state = findState(mask);

   // if match was not found in list, make a new one and push it on back of the
   // state list
   if (state == NULL)
   {
      m_state_list.push_back(new StateElement(mask, statelets, tex_coords));
      dirty(eDLIST, false);
   }
   else // change the values of the existing state element
   {
      state->setStatelets(statelets);
      if (!tex_coords.empty() || !state->getTexCoordList().empty())
      {
         state->setTexCoordList(tex_coords);
         dirty(eDLIST, false);
      }
   }
}

//----------------------------------------------------------------------------
void sglDrawable::setTexCoordList(sglVec2f *tex_coords)
{
   sglTexCoords tmp(new sglNoRefDataPtr(tex_coords));
   m_state_list[0]->setTexCoordList(tmp);
   dirty(eDLIST, false);
}

//----------------------------------------------------------------------------
void sglDrawable::setTexCoordList(const sglTexCoords &tex_coords)
{
   if (tex_coords.isValid() ||
       !m_state_list[0]->getTexCoordList().empty())
   {
      m_state_list[0]->setTexCoordList(tex_coords);
      dirty(eDLIST, false);
   }
}

//----------------------------------------------------------------------------
void sglDrawable::setTexCoordList(const vector<sglTexCoords> &tex_coords)
{
   if (!tex_coords.empty() || !m_state_list[0]->getTexCoordList().empty())
   {
      m_state_list[0]->setTexCoordList(tex_coords);
      dirty(eDLIST, false);
   }
}

//----------------------------------------------------------------------------
bool sglDrawable::removeState(unsigned int mask)
{
   if (mask == 0)
   {
      m_state_list[0]->clearStatelets();

      if (!m_state_list[0]->getTexCoordList().empty())
      {
         m_state_list[0]->clearTexCoordList();
         dirty(eDLIST, false);
      }

      return true;  // cannot remove the 0th state but return true anyway
   }

   // find state element matching mask
   StateElement *state = NULL;
   vector<StateElement*>::iterator elt = m_state_list.begin();
   for (; elt != m_state_list.end(); ++elt)
   {
      if (mask == ((*elt)->getMask()))
      {
         state = *elt;
         break;
      }
   }

   if (state)
   {
      m_state_list.erase(elt);
      sgl::releaseDisplayList(state->getDisplayList());
      state->setDisplayList(0);

      delete state;
      return true;
   }
   else
   {
      return false;  // matching state not found
   }
}

//----------------------------------------------------------------------------
void sglDrawable::enableDisplayList(bool on)
{
   if (on == m_use_dlist)
      return;

   m_use_dlist = on;
   if (m_use_dlist)
   {
      dirty(eDLIST, true);
   }
   else
   {
      vector<StateElement*>::iterator elt = m_state_list.begin();
      for (; elt != m_state_list.end(); ++elt)
      {
         sgl::releaseDisplayList((*elt)->getDisplayList());
         (*elt)->setDisplayList(0);
      }
      dirty(eDLIST, false);
   }
}

//----------------------------------------------------------------------------
void sglDrawable::setBound(const sglBoxBound *bbox, sglBound::ModeEnum mode)
{
   // warning, this will compute upwards whether computeBounds is
   // successful or not.  This may or may not be the desired behaviour.
   dirty(eBOUND, false);

   // if sphere is NULL compute the bounds for this node
   if (bbox == NULL)
   {
      if (mode == sglBound::eSTATIC)
      {
         m_bound_mode = sglBound::eDYNAMIC;
         computeBounds(); // compute bound immediately for this node

         // warning: if computeBounds returned false at this point that means
         // there wasn't enough info in the geoset to compute a bbox and it
         // returned an empty one.

         m_dirty &= ~eBOUND;
      }
   }
   else
   {
      m_bbox = *bbox;
      m_dirty &= ~eBOUND;
   }

   m_bound_mode = mode;
}

//----------------------------------------------------------------------------
sglBound::IntersectResultEnum sglDrawable::intersectBBox(
   sglIntersect<float> &segset)
{
   sglSegmentf seg(segset.getSegment());
   sglBound::IntersectResultEnum result = m_bbox.intersect(seg);

   if (result == sglBound::eINTERSECT)
   {
      segset.setGeometry(this);
      segset.setSegment(seg);
      segset.setResultEnum(result);
      return result;
   }

   return sglBound::eOUTSIDE;
}

//----------------------------------------------------------------------------
sglBound::IntersectResultEnum sglDrawable::intersectBBox(
   sglIntersect<double> &segset)
{
   sglSegmentd seg(segset.getSegment());
   sglBound::IntersectResultEnum result = m_bbox.intersect(seg);

   if (result == sglBound::eINTERSECT)
   {
      segset.setGeometry(this);
      segset.setSegment(seg);
      segset.setResultEnum(result);
      return result;
   }

   return sglBound::eOUTSIDE;
}

//----------------------------------------------------------------------------
const sglBoxBound &sglDrawable::getBound()
{
   computeBoundsRecursiveDown();
   return m_bbox;
}

//----------------------------------------------------------------------------
void sglDrawable::dirty(unsigned char mask, bool all)
{
   if (!m_use_dlist)
      mask &= ~eDLIST;

   if ((m_dirty|mask) != m_dirty)
   {
      m_dirty |= mask;

      for (vector<sglGeode*>::iterator ii = m_parent_list.begin();
           ii != m_parent_list.end(); ++ii)
      {
         (*ii)->dirty(m_dirty);
      }
   }

   if (all && (mask&eDLIST))
   {
      vector<StateElement*>::iterator elt = m_state_list.begin();
      for (; elt != m_state_list.end(); ++elt)
      {
         (*elt)->setDirty(true);
      }
   }
}

//----------------------------------------------------------------------------
void sglDrawable::computeBoundsRecursiveDown()
{
   if ((m_dirty&eBOUND) && m_bound_mode != sglBound::eSTATIC)
   {
      computeBounds();
   }
}

//----------------------------------------------------------------------------
void sglDrawable::computeNoBounds()
{
   m_dirty &= ~eBOUND;
}

//----------------------------------------------------------------------------
void sglDrawable::preDraw()
{
   // Call computeBounds before drawGeometry (MonoIndexed types compute
   // some required values in computeBounds).
   if (m_dirty&eBOUND)
   {
      if (m_bound_mode == sglBound::eSTATIC)
      {
         computeNoBounds();
      }
      else
      {
         computeBounds();
      }
   }

   if ((m_dirty&eDLIST) && m_use_dlist)
   {
      for (vector<StateElement*>::iterator elt = m_state_list.begin();
           elt != m_state_list.end(); ++elt)
      {
         GLuint dlist = (*elt)->getDisplayList();
         if (!dlist)
         {
            dlist = sgl::getDisplayList();
            (*elt)->setDisplayList(dlist);
         }
         if (dlist && (*elt)->getDirty())
         {
            sglPrintDebug() << "sglDrawable dlist = " << dlist << endl;

            glNewList(dlist, GL_COMPILE);
            if (!m_bbox.isEmpty())
               drawImmediate((*elt)->getTexCoordList());
            glEndList();

            (*elt)->setDirty(false);
         }
      }
      m_dirty &= ~eDLIST;
   }

   sglPrintDebug() << "sglDrawable::preDraw()" << endl;
}

//----------------------------------------------------------------------------
bool sglDrawable::cull(const sglFrustum<float> &frustum,
                       unsigned int cull_flags) const
{
   if (cull_flags)
   {
      sglBound::IntersectResultEnum result = m_bbox.intersect(frustum,
                                                              cull_flags);

      if (result == sglBound::eOUTSIDE)
      {
         return false; // draw nothing
      }
   }

   return true;
}

//----------------------------------------------------------------------------
bool sglDrawable::cull(const sglFrustum<double> &frustum,
                       unsigned int cull_flags) const
{
   if (cull_flags)
   {
      sglBound::IntersectResultEnum result = m_bbox.intersect(frustum,
                                                              cull_flags);

      if (result == sglBound::eOUTSIDE)
      {
         return false; // draw nothing
      }
   }

   return true;
}
#include <typeinfo>
//----------------------------------------------------------------------------
void sglDrawable::drawImmediate(const vector<sglTexCoords> &tex_coords) const
{
   if (m_draw_func)
      (this->*m_draw_func)(tex_coords);
   else
      sglPrintWarning() << "No draw func." << typeid(*this).name() << endl;
}

//----------------------------------------------------------------------------
void sglDrawable::printInfo(ostream &ostrm, const char *indent_string) const
{
   sglObject::printInfo(ostrm, indent_string);
   ostrm << indent_string << "Bounding box: " << m_bbox << endl;
   ostrm << indent_string << "# parents = " << getNumParents()
         << ", parent geodes:" << endl;

   for (vector<sglGeode*>::const_iterator elt = m_parent_list.begin();
        elt != m_parent_list.end();
        ++elt)
   {
      ostrm << indent_string << "        " << (*elt) << endl;
   }

   unsigned int num_states = m_state_list.size();
   ostrm << indent_string << "Number of states: " << num_states << endl;

   // states output by sgluOutputState (called by sgluOutputGeometry)
}
