/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglDrawable.hpp
 *  Created: 18 August 1999
 *  Summary: Base class for all sglObjects that can be collected by the cull
 *           traversal (sglCull) for rendering.
 *****************************************************************************/

#ifndef __SGL_DRAWABLE_HPP
#define __SGL_DRAWABLE_HPP

#include <sgl.h>
#include <sglObject.hpp>

#include <sglBoxBound.hpp>
#include <sglFrustum.hpp>

#include <sglIntersect.hpp>
#include <sglTexCoords.hpp>
#include <sglStatelet.hpp>
#include <sglStats.hpp>
#include <sglOStream.hpp>

#include <vector>

// foward reference this stuff
class sglGeode;
template <class T> class sglIntersect;

// ===========================================================================

/**

sglDrawable is the abstract base class for all objects that can be added to the
sglGeode class and collected by the cull traversal for rendering (currently
sglGeoSet, sglInterleavedGeometry and an internal sglText::sglTextSet class).
It defines the interface (virtual functions) expected by the sglCull class.
Its subclasses contain the information need to define various types of 3D
objects.  This class also contains "states", collections of statelets, that
further define the appearance of the 3D objects.  You will need to understand
this class if you want to extend SGL's geometry nodes or create new types of
renderables (that is, drawables).


{\bf Parent Lists}

Just as sglNodes can be added as children to multiple sglGroups, sglDrawable
objects can be added to multiple sglGeodes.  The sglGeodes are, in
effect, the parents of these objects.  These objects, therefore, maintain
a list of pointers to their parent sglGeodes.  The user can query
the number of parents a geode has, as well as get pointers to these sglGeodes.

{\bf Axis-Aligned Bounding Boxes}

The group nodes (in the previous chapter) and sglGeode all have bounding
spheres that encompass the extent in 3D space that all of its descendents
lie within.  A bounding sphere is used because its shape is invariant
in the presence of affine transformations (is this really true).
At this point in the scene graph (below sglGeode) there are no more
transformations, and the extent of the geometry is easily represented by an
axis-aligned bounding box.  By default, this bounding box is automatically
computed (by the subclasses).  As in the sglNode classes, the user can override
this behaviour by calling setBound with the sglBound::eSTATIC mode.
The bound is either user-specified (if an sglBoxBound is provided) or computed
from the current vertex data.

{\bf Disabling Display Lists}

As in sglText, the user can enable or disable the use of display lists
to contain the geometry data.  The same general rule of thumb applies:
if the geometry data does not change often, then the display list generation
should be enabled (the default).  Otherwise, it is probably more efficient
to disable the use of display lists as they won't mark the tree as dirty
when the geometry changes (a dirtied tree is cleaned in preDraw() by traversing
all the nodes with dirty bits set).

{\bf State}

Along with geometry data which is managed by the subclasses, OpenGL state
information must also be associated with this to affect the appearance
of the geometry.  In SGL, state consists of a set (STL vector) of
sglStatelets (one of each) and, optionally a set of texture coordinates (see
State and Texture Coordinates below).  SGL is capable of associating multiple
different states with a given geometry using the setState(...) member function.
The first parameter of this function is a unique bit flag (index) that
identifies the state.

By default every sglDrawable has a default (empty) state associate with
index 0.  The user may override this, or add additional states with
different indices.  Usually a given state is associated with a single
bit of the index (e.g., 0x01, 0x02, 0x04, etc.); however, there is nothing
to stop the user from setting the state with an index with multiple bits
set (e.g., 0x03, 0x11, etc.).  If setState is called with the same
index then the vector of statelets and texcoords replace the existing ones.
<p>The purpose of the index is to aid in selection of a single state during
the cull traversal.   The sglScene::cull(...) function takes
a state mask which is an unsigned int.  When an sglDrawable passes
the cull test (is selected for drawing),  it is place on a render
list along with a particular state.  If no states have been defined,
the default state (0) is placed in the render list along with the sglDrawable.
Otherwise the list of state's indices are examined in the order they were
added to the sglDrawable until the sglCull's state mask ANDed
with the this evaluates to non-zero.  The state corresponding to this
index will be the one put in the render list.  If none match, then
the default state (index 0) will be used.

Note: when multitexture support is implemented
then the behaviour described above may be modified.  That is, more
than one state may be selected and rendered.  Multipass techniques
might be accomplished this way as well.

{\bf State and Texture Coordinates}

Texture coordinates used to be one of the attribute arrays handled by
sglGeoSet.  Because each state (group of statelets) set in the call to setState
can have a unique sglTexture, a unique texture coordinate array can be assigned
with each state in the same call.  If the state contains no sglTexture
or an sglTexGen statelet is also present, then this parameter can be given
as NULL (see the example below).  Note that the other subclass of
sglDrawable, sglInterleavedGeometry, has texture coordinates built in to
the interleaved arrays, and any texture coordinate array passed into the
setState() function is ignored.

{\bf Example:}
\begin{verbatim}
   sglGeode *geode  = new sglGeode;
   sglLineSet *geom = new sglLineSet;         // a subclass of sglGeoSet
   geom->enableDisplayList(true);             // the default anyway

   geode->addGeometry(geom);
   unsigned int num = geom->getNumParents();  // returns 1
   sglGeode *geode1 = geom->getParent(0);     // returns geode

   sglBoxBound bbox;
   bbox.setMin(-10.0f, -10.0f, -10.0f);
   bbox.setMax( 10.0f,  10.0f,  10.0f);
   geom->setBound(&bbox, sglBound::eSTATIC);  // make bounds static

   geom->setBound(NULL, sglBound::eDYNAMIC);  // enable automatic computation

   // set a state - see statelets.html for more details
   //     transparency is enabled, and front
   //     face polygon style is GL_LINE.
   vector<sglStatelet*> statelets;
   statelets.push_back(new sglTransparency(true));
   statelets.push_back(new sglPolyFrontStyle(sglPolyStyle::eLINE));

   geom->setState(0x01, statelets);    // no texcoords</pre>
\end{verbatim}
*/

class SGL_DLL_API sglDrawable : public sglObject
{
public:
   class SGL_DLL_API StateElement
   {
   public:
      StateElement(unsigned int mask);
      StateElement(unsigned int mask,
                   const vector<sglStatelet*> &statelets,
                   const vector<sglTexCoords> &tex_coords);
      StateElement(unsigned int mask,
                   const vector<sglStatelet*> &statelets,
                   const sglTexCoords &tex_coords);
      ~StateElement();

      unsigned int getMask() const { return m_mask; }

      const vector<sglStatelet*>& getStatelets() const { return m_statelets; }

      const sglTexCoords *getTexCoords() const
         { return (m_tex_coords.empty() ? NULL : &m_tex_coords[0]); }

      const vector<sglTexCoords>& getTexCoordList() const
         { return m_tex_coords; }
      const vector<sglTexCoords>* getTexCoordListPtr() const
         { return &m_tex_coords; }

      GLuint getDisplayList() const { return m_dlist; }

      bool getDirty() const { return m_dirty; }

      void printInfo(ostream &ostrm, const char *indent_string) const;

   private:
      // Not implemented.
      StateElement();
      StateElement(const StateElement &);
      StateElement &operator=(const StateElement &);

      friend class sglDrawable;

      void setStatelets(const vector<sglStatelet*> &statelets);
      void clearStatelets();
      void setTexCoordList(const vector<sglTexCoords> &tex_coords);
      void setTexCoordList(const sglTexCoords &tex_coords);
      void clearTexCoordList();
      void setDisplayList(GLuint dlist);
      void setDirty(bool dirty) { m_dirty = dirty; }

   private:
      unsigned int          m_mask;
      vector<sglStatelet*>  m_statelets;
      vector<sglTexCoords>  m_tex_coords;
      GLuint                m_dlist;
      bool                  m_dirty;

      // tex coord bounds could go here if I ever want to support clip texture
   };

public:
   /// default constructor
   sglDrawable();
   /// virtual destructor
   virtual ~sglDrawable();

   /// control whether to create a display list for this (default: true)
   void enableDisplayList(bool on);
   ///
   bool getDisplayListFlag() const { return m_use_dlist; }

   ///
   unsigned int getNumParents() const { return m_parent_list.size(); }
   ///
   sglGeode* getParent(unsigned int i);

   /// set/get the states
   unsigned int getNumStates() { return m_state_list.size(); }
   ///
   StateElement *getState(unsigned int index = 0);
   ///
   StateElement *getStateByMask(unsigned int mask = 0) const;

   ///
   StateElement *findState(unsigned int mask) const;

   ///
   void setState(unsigned int mask,
                 const vector<sglStatelet*> &statelets);
   ///
   void setState(unsigned int mask,
                 const vector<sglStatelet*> &statelets,
                 sglVec2f *tex_coords);
   ///
   void setState(unsigned int mask,
                 const vector<sglStatelet*> &statelets,
                 const sglTexCoords &tex_coords);
   ///
   void setState(unsigned int mask,
                 const vector<sglStatelet*> &statelets,
                 const vector<sglTexCoords> &tex_coords);

   /// Assumes state mask 0
   void setTexCoordList(sglVec2f* tex_coords);

   /// Assumes state mask 0
   void setTexCoordList(const sglTexCoords &tex_coords);

   /// Assumes state mask 0
   void setTexCoordList(const vector<sglTexCoords> &tex_coords_list);

   /** returns true if corresponding state was found AND deleted.  Note you
       cannot remove the StateElement with zero mask, but calling
       removeState(0) is equivalent calling setState(0, NULL, NULL).  Beware
       this doesn't delete tex_coords, but will delete statelets if their
       reference count goes to zero
   */
   bool removeState(unsigned int mask);

   /** If bbox = NULL, the bounds of this node are computed,
       else they are set from bbox.
       If mode is sglBound::eSTATIC, bounds are not set to reflect further
         changes (if bbox is NULL the bounding box is computed immediately),
       else if mode is sglBound::eDYNAMIC, bounds are recomputed as necessary.
   */
   void setBound(const sglBoxBound *bbox, sglBound::ModeEnum mode);

   /// Returns the bounding volume (recalculating if necessary).
   virtual const sglBoxBound &getBound();
   ///
   sglBound::ModeEnum getBoundMode() {return m_bound_mode;}

   /// Returns the bounding volume as it was last computed.
   const sglBoxBound &getPrevBound() const { return m_bbox; }

   /** Add this drawable's stats to the given sglStats.
       @param stats The sglStats object to which to add the statistics.
   */
   virtual void addStats(sglStats &stats) const = 0;

   /// pre-draw traversal to clean the tree
   void preDraw();

   /// intersect against the geoset (bbox only or primitives)
   sglBound::IntersectResultEnum intersectBBox(sglIntersectf &segset);
   ///
   sglBound::IntersectResultEnum intersectBBox(sglIntersectd &segset);

   /// default behaviour for unimplemented versions
   virtual sglBound::IntersectResultEnum intersectPrim(sglIntersectf &)
      { return sglBound::eOUTSIDE; }
   ///
   virtual sglBound::IntersectResultEnum intersectPrim(sglIntersectd &)
      { return sglBound::eOUTSIDE; }

   ///
   virtual bool isValid() const { return true; }

   /// Anti-Aliasing types.
   enum AAEnum {ePOINT_AA, eLINE_AA, ePOLYGON_AA, eNO_AA};
   ///
   virtual AAEnum getAAType() const { return eNO_AA; }

   ///
   virtual void printInfo(ostream &ostrm, const char *indent_string) const;

   ///
   void drawImmediate(const vector<sglTexCoords> &tex_coords) const;

   typedef void (sglDrawable::*DrawFunc)(
      const vector<sglTexCoords> &tex_coords) const;

protected:
   friend class sglGeode;

   /** Compute the bounding box for this geometry and fill in the member
       variable, m_bbox, in this class.
       @return true if successful (all information is available for computing a
               bounding box), otherwise false.
   */
   virtual bool computeBounds() = 0;

   virtual void computeNoBounds();
   virtual void computeBoundsRecursiveDown();

   void dirty(unsigned char mask, bool all);

   void addParent(sglGeode *parent);
   void removeParent(sglGeode *parent);

   bool cull(const sglFrustum<float> &frustum, unsigned int cull_flags) const;
   bool cull(const sglFrustum<double> &frustum, unsigned int cull_flags) const;

private:  // not implemented
   sglDrawable(const sglDrawable &);
   sglDrawable &operator=(const sglDrawable &);

protected:
   sglBoxBound m_bbox;

   // the first element of the following list is set in the ctor, to
   // something with mask==0, empty statelet list, and NULL tex coords.
   vector<StateElement *> m_state_list;
   bool m_use_dlist;

   DrawFunc m_draw_func;

   sglBound::ModeEnum m_bound_mode;

   enum DirtyEnum { eBOUND = 0x01, eDLIST = 0x02 };
   unsigned char m_dirty;

   vector<sglGeode*> m_parent_list;
};

#endif
