/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglDrawablePool.cpp
 *   Author: Tom Stimson
 *  Created: 9 July 2000
 *****************************************************************************/

#include <sglDrawablePool.hpp>

#include <sglDrawable.hpp>
#include <sglStats.hpp>

//----------------------------------------------------------------------------
void sglDrawablePair::draw(sglCurrState *curr_state) const
{
   curr_state->apply(m_fullstate);
   curr_state->draw(m_drawable);
}

//----------------------------------------------------------------------------
void sglDrawablePair::statsDraw(sglCurrState *curr_state,
                                sglStats& stats) const
{
   curr_state->statsApply(m_fullstate, stats);
   curr_state->draw(m_drawable);
   stats.addTotalGSets(1);
   m_drawable->addStats(stats);
   m_fullstate->addStats(stats);
}

//----------------------------------------------------------------------------
const sglFullState* sglDrawablePair::getFirstState() const
{
   return m_fullstate;
}

//----------------------------------------------------------------------------
void sglUnSortedDrawablePool::draw(sglCurrState *curr_state) const
{
   vector<sglDrawItem*>::const_iterator end = m_list.end();
   for (vector<sglDrawItem*>::const_iterator ii = m_list.begin();
        ii != end; ++ii)
   {
      (*ii)->draw(curr_state);
   }
}

//----------------------------------------------------------------------------
void sglUnSortedDrawablePool::statsDraw(sglCurrState *curr_state,
                                        sglStats& stats) const
{
   vector<sglDrawItem*>::const_iterator end = m_list.end();
   for (vector<sglDrawItem*>::const_iterator ii = m_list.begin();
        ii != end; ++ii)
   {
      (*ii)->statsDraw(curr_state, stats);
   }
}

//----------------------------------------------------------------------------
const sglFullState* sglUnSortedDrawablePool::getFirstState() const
{
   vector<sglDrawItem*>::const_iterator end = m_list.end();
   for (vector<sglDrawItem*>::const_iterator ii = m_list.begin();
        ii != end; ++ii)
   {
      const sglFullState* fs = (*ii)->getFirstState();
      if (fs)
         return fs;
   }
   return NULL;
}

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
//----------------------------------------------------------------------------

// HACK Any function (including the construtor and destructor) that touches
// the map cannot be in the header file.  Otherwise a buggy STL implementation
// on WIN32 systems will can cause the program to crash.
//----------------------------------------------------------------------------
sglSortedDrawablePool::sglSortedDrawablePool()
{
}

//----------------------------------------------------------------------------
sglSortedDrawablePool::~sglSortedDrawablePool()
{
}

//----------------------------------------------------------------------------
void sglSortedDrawablePool::clear()
{
    m_map.clear();
}

//----------------------------------------------------------------------------
void sglSortedDrawablePool::insert(sglDrawItem *d_item)
{
   const sglFullState* fullstate = d_item->getFirstState();
   const sglStatelet* statelet = fullstate->m_statelet[*m_sort_key];
   const void *key = statelet->getSortKey();

   map<const void*,sglDrawablePool*>::iterator ii = m_map.lower_bound(key);
   if (ii == m_map.end() || key != ii->first)
   {
      vector<unsigned int>::iterator next_key = m_sort_key;
      ++next_key;
      if (*next_key == UINT_MAX)
      {
         sglUnSortedDrawablePool* pool =
            m_sources->m_unsorted_dp_pool.newObject();
         pool->clear();
         ii = m_map.insert(ii, pair<const void*,sglDrawablePool*>(key, pool));
      }
      else
      {
         if (fullstate->m_statelet[*next_key]->getSorted())
         {
            sglSortedDrawablePool *pool =
               m_sources->m_sorted_dp_pool.newObject();
            pool->clear();
            pool->setSources(m_sources);
            pool->setSortKey(next_key);
            ii =
               m_map.insert(ii, pair<const void*,sglDrawablePool*>(key, pool));
         }
         else
         {
            sglIndexedDrawablePool *pool =
               m_sources->m_indexed_dp_pool.newObject();
            pool->clear();
            pool->setSources(m_sources);
            pool->setSortKey(next_key);
            ii =
               m_map.insert(ii, pair<const void*,sglDrawablePool*>(key, pool));
         }
      }
   }

   ii->second->insert(d_item);
}

//----------------------------------------------------------------------------
void sglSortedDrawablePool::draw(sglCurrState *curr_state) const
{
   map<const void*,sglDrawablePool*>::const_iterator end = m_map.end();
   for (map<const void*,sglDrawablePool*>::const_iterator ii = m_map.begin();
        ii != end; ++ii)
   {
      ii->second->draw(curr_state);
   }
}

//----------------------------------------------------------------------------
void sglSortedDrawablePool::statsDraw(sglCurrState *curr_state,
                                      sglStats& stats) const
{
   map<const void*,sglDrawablePool*>::const_iterator end = m_map.end();
   for (map<const void*,sglDrawablePool*>::const_iterator ii = m_map.begin();
        ii != end; ++ii)
   {
      ii->second->statsDraw(curr_state, stats);
   }
}

//----------------------------------------------------------------------------
const sglFullState* sglSortedDrawablePool::getFirstState() const
{
   map<const void*,sglDrawablePool*>::const_iterator end = m_map.end();
   for (map<const void*,sglDrawablePool*>::const_iterator ii = m_map.begin();
        ii != end; ++ii)
   {
      const sglFullState* fs = ii->second->getFirstState();
      if (fs)
         return fs;
   }
   return NULL;
}

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
//----------------------------------------------------------------------------

//----------------------------------------------------------------------------
void sglIndexedDrawablePool::insert(sglDrawItem *d_item)
{
   const sglFullState* fullstate = d_item->getFirstState();
   const sglStatelet* statelet = fullstate->m_statelet[*m_sort_key];
   unsigned int indx = statelet->getIndex();

   if (indx >= m_list.size())
   {
      vector<unsigned int>::iterator next_key = m_sort_key;
      ++next_key;
      if (*next_key == UINT_MAX)
      {
         while (indx >= m_list.size())
         {
            sglUnSortedDrawablePool* pool =
               m_sources->m_unsorted_dp_pool.newObject();
            pool->clear();
            m_list.push_back(pool);
         }
      }
      else
      {
         if (fullstate->m_statelet[*next_key]->getSorted())
         {
            while (indx >= m_list.size())
            {
               sglSortedDrawablePool* pool =
                  m_sources->m_sorted_dp_pool.newObject();
               pool->clear();
               pool->setSources(m_sources);
               pool->setSortKey(next_key);
               m_list.push_back(pool);
            }
         }
         else
         {
            while (indx >= m_list.size())
            {
               sglIndexedDrawablePool* pool =
                  m_sources->m_indexed_dp_pool.newObject();
               pool->clear();
               pool->setSources(m_sources);
               pool->setSortKey(next_key);
               m_list.push_back(pool);
            }
         }
      }
   }

   m_list[indx]->insert(d_item);
}

//----------------------------------------------------------------------------
void sglIndexedDrawablePool::draw(sglCurrState *curr_state) const
{
   vector<sglDrawablePool*>::const_iterator end = m_list.end();
   for (vector<sglDrawablePool*>::const_iterator ii = m_list.begin();
        ii != end; ++ii)
   {
      (*ii)->draw(curr_state);
   }
}

//----------------------------------------------------------------------------
void sglIndexedDrawablePool::statsDraw(sglCurrState *curr_state,
                                       sglStats& stats) const
{
   vector<sglDrawablePool*>::const_iterator end = m_list.end();
   for (vector<sglDrawablePool*>::const_iterator ii = m_list.begin();
        ii != end; ++ii)
   {
      (*ii)->statsDraw(curr_state, stats);
   }
}

//----------------------------------------------------------------------------
const sglFullState* sglIndexedDrawablePool::getFirstState() const
{
   vector<sglDrawablePool*>::const_iterator end = m_list.end();
   for (vector<sglDrawablePool*>::const_iterator ii = m_list.begin();
        ii != end; ++ii)
   {
      const sglFullState* fs = (*ii)->getFirstState();
      if (fs)
         return fs;
   }
   return NULL;
}
