/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglFog.hpp
 *   Author: Scott McMillan and Michael Morrison.
 *  Created: 28 November 2000
 *  Summary: sglStatelet to encapsulate fog functionality.
 *****************************************************************************/

#ifndef __SGL_FOG_HPP
#define __SGL_FOG_HPP

#include <sgl.h>
#include <sglStatelet.hpp>
#include <sglVector.hpp>

//----------------------------------------------------------------------------

/**

This statelet encapsulates the OpenGL glFog*(..) and glEnable(GL_FOG) functions
that control OpenGL fog characteristics.  With this enabled, objects that are
farther from the viewpoint begin to fade into the fog color.  This statelet
also supports the "eye plane absolute" and "eye radial" modes of the
GL_NV_fog_distance extension.

<b>Example:</b>

<pre>
   sglFog *fog = new sglFog;
   fog->setMode(sglFog::eLINEAR);
   fog->setColor(0.f, 0.f, 0.f, 1.0f);
   fog->setExtent(0.f, 100.f);
   fog->setDensity(0.025f);

   if (sgl::s_fog_distance_capable)
   {
      fog->setDistanceEquation(sglFog::eEYE_RADIAL);
   }
</pre>

 */

class SGL_DLL_API sglFog : public sglStatelet
{
public:
   /// Valid sglFog modes:
   enum ModeEnum
   {
      /// disable fog (the default)
      eOFF    = 0,

      /// exponential fog
      eEXP    = GL_EXP,

      /// exponential squared fog
      eEXP2   = GL_EXP2,

      /// linear fog.
      eLINEAR = GL_LINEAR
   };

   /// Distance equations for systems with the NVIDIA extensions
   enum DistanceEqEnum
   {
      /// perpendicular distance from the near clip plane
      eEYE_PLANE_ABSOLUTE = GL_EYE_PLANE_ABSOLUTE_NV,

      /// real distance from the view point
      eEYE_RADIAL = GL_EYE_RADIAL_NV
   };


public:
   /// Default constructor.  Initializes a disabled fog statelet (i.e. no fog)
   sglFog();

   /// Virtual destructor
   virtual ~sglFog();

   /** Set the fog mode.
       @param mode One of {eOFF, eEXP, eEXP2, or eLINEAR}.  Default: eOFF
   */
   void setMode(ModeEnum mode);

   /** Get the current fog mode.
       @return One of {eOFF, eEXP, eEXP2, or eLINEAR}.
   */
   ModeEnum getMode() const { return m_mode; }

   /** Set the fog distance calculation equation.
       @param mode One of {eEYE_PLANE_ABSOLUTE, eEYE_RADIAL}.
         Default: eEYE_PLANE_ABSOLUTE
   */
   void setDistanceEquation(DistanceEqEnum eq);

   /** Get the current fog distance calc. eq.
       @return One of {eEYE_PLANE_ABSOLUTE, eEYE_RADIAL}
   */
   DistanceEqEnum getDistanceEquation() const { return m_distance_equation; }

   /** Set the extent of the fog (distances from the viewer).  This is used in
       eLinear mode only.  Only values greater than or equal to zero are
       accepted and end must be greater than start.
       @param start The closest distance of the fog. Default: 0.
       @param end The furthest distance of the fog. Default 1.
   */
   void setExtent(float start, float end);

   /** Get the linear fog extents.
       @param start On return contains the near distance.
       @param end On return contains the far distance.
   */
   void getExtent(float &start, float &end) const
      { start = m_start; end = m_end; }

   /** Set the fog density parameter for exponential modes.
       @param density Exponent factor in the exponential fog
       functions. Default: 1
   */
   void setDensity(float density);

   /** Get the density.
       @return The density (i.e. exponent factor) of the fog functions.
   */
   float getDensity() const { return m_density; }

   /** Set the fog color.
       @param color Vector containing a 4 component RGBA color.
              Default: 0, 0, 0, 0.
   */
   void setColor(const sglVec4f &color) { m_color = color; }

   /** Set the fog color using 4 floating point values (from 0...1).
       @param r Red color component.
       @param g Green color component.
       @param b Blue color component.
       @param a Alpha color component.
   */
   void setColor(float r, float g, float b, float a) { m_color.set(r,g,b,a); }

   /** Get the fog color.
       @return A reference to the vector containing the 4 component RGBA color.
   */
   const sglVec4f &getColor() const { return m_color; }

   // See sglStatelet::apply() for details.
   void apply(sglCurrState*) const;

   // See sglStatelet::getSorted() for details.
   virtual bool getSorted() const { return true; }

   // See sglObject::printInfo() for details.
   virtual void printInfo(ostream &ostrm, const char *indent_string) const;

protected:

private: // methods - not implemented
   sglFog(const sglFog &);
   sglFog &operator=(const sglFog &);

   void setIndex() { m_index = ((m_mode == eOFF) ? 0 : UINT_MAX); }

protected: // variables
   ModeEnum m_mode;

   float    m_start;
   float    m_end;
   float    m_density;
   DistanceEqEnum m_distance_equation;

   sglVec4f m_color;
};

#endif
