/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglFrustum.hpp
 *  Created: 17 November 1998
 *  Summary: a polytope with 6 planes that defines a truncated pyramid
 *****************************************************************************/

#ifndef __SGL_FRUSTUM_HPP
#define __SGL_FRUSTUM_HPP

#include <sgl.h>
#include <sglPolytope.hpp>

//----------------------------------------------------------------------------
// convenience types:
//    template class 'should' only be float or double

template <class T> class sglFrustum;

typedef sglFrustum<float>  sglFrustumf;
typedef sglFrustum<double> sglFrustumd;

//----------------------------------------------------------------------------

/**

The sglFrustum class is a subclass of sglPolytope that has six planes that
usually either defines a truncated pyramid corresponding to the volume "seen"
with a perspective view or a rectangular solid volume corresponding to an
orthographic view.

*/

template <class T>
class sglFrustum : public sglPolytope<T>
{
public:
   /// Indices to access the individual planes of the polytope
   enum { eLEFT = 0,  eUP = 1, eRIGHT = 2, eDOWN = 3, eFRONT = 4, eBACK = 5,
          eNUM_PLANES = 6};

public:
   /// Default constructor
   sglFrustum() : sglPolytope<T>(), m_ortho_flag(true), m_fov_y((T)-1) {}

   /// Copy constructor
   template <class S>
   sglFrustum(const sglFrustum<S> &rhs)
           : sglPolytope<T>(rhs),
             m_ortho_flag(rhs.getOrthoFlag()),
             m_fov_y((T)rhs.getFOVY())
        {}

   /// Destructor
   virtual ~sglFrustum() {}

   /// assignment operator
   template <class S>
   sglFrustum &operator=(const sglFrustum<S> &rhs)
      {
         if (this != (sglPolytope<T>*)&rhs)
         {
             sglPolytope<T>::operator=(rhs);
             m_ortho_flag = rhs.getOrthoFlag();
             m_fov_y = rhs.getFOVY();
         }
         return *this;
      }

   /** Query whether or not this is orthographic
       @return True if orthographic, else false.
   */
   bool getOrthoFlag() const { return m_ortho_flag; }

   /** Query the aspect ratio of the frustum: width/height
       @return The current aspect ratio.
   */
   double getAspect() const { return m_aspect; }

   /** Query the current vertical field of view (FOV) in radians.
       @return The vertical FOV angle in radians.  If this is
         not a perspective frustum, -1 is returned.
   */
   double getFOVY() const { return m_fov_y; }

   // HACK is the following DEPRECATED?
   /** This computes a frustum/polytope based on an OpenGL format projection
       matrix.
       @param projection_matrix The project matrix (as OpenGL would expect
         it).
       @return true if it is not a degenerate matrix, else false.
   */
   bool build(sglMat4<T>& projection_matrix);

   /** Compute a perspective frustum (from arguments to gluPerspective)
       @param fovY Vertical field of view in radians.
       @param aspect width/height ratio of view.
       @param near_clip Distance from view point to near clipping plane.
       @param far_clip Distance from view point to far clipping plane.
   */
   bool buildPerspective(double fovY,  // radians
                         double aspect,
                         double near_clip, double far_clip);

   /** Compute a perspective frustum (from arguments to glFrustum).  The volume
       is defined by the parameters (left, bottom, -near) and
       (right, top, -near)
       @param near_left Left coordinate of near clipping plane.
       @param near_right Right coordinate of near clipping plane.
       @param near_top Top coordinate of near clipping plane.
       @param near_bottom Bottom coordinate of near clipping plane.
       @param near_clip Distance from view point to near clipping plane.
       @param far_clip Distance from view point to far clipping plane.
   */
   bool buildPerspective(double near_left, double near_right,
                         double near_bottom, double near_top,
                         double near_clip, double far_clip);

   /** Compute an orthographic view volume (from arguments to glOrtho)
       @param left Coordinate of left clipping plane.
       @param right Coordinate of right clipping plane.
       @param top Coordinate of top clipping plane.
       @param bottom Coordinate of bottom clipping plane.
       @param near_clip Distance from view point to near clipping plane.
       @param far_clip Distance from view point to far clipping plane.
   */
   bool buildOrtho(double left, double right, double bottom, double top,
                   double near_clip, double far_clip);

#ifdef SGL_FRUSTUM_DEBUG
   // draw the frustum
   void drawImmediate() const;

private:
   sglVec3f m_vertices[8];  // for drawing the frustum:
                            //  4      5
                            //    0  1
                            //    3  2
                            //  7      6
#endif

private:
    bool buildPerspectiveInternal(double near_left, double near_right,
                                  double near_bottom, double near_top,
                                  double near_clip, double far_clip);

private:
    bool   m_ortho_flag;
    double m_aspect;
    double m_fov_y;
};

//----------------------------------------------------------------------------
template <class T>
inline bool sglFrustum<T>::build(sglMat4<T>& projectionMatrix)
{
   // in case something goes wrong
   m_num_planes = 0;

   sglVec3<T> p1, p2, p3;
   sglVec4<T> r1, r2, r3;
   // get the inverse of the projection matrix
   sglMat4<T> inv;
   inv.inverse(projectionMatrix);

   for (int i=0; i<sglFrustumf::eNUM_PLANES; i++)
   {
      switch(i)
      {
         case sglFrustum::eLEFT:
            p1.set(-1.f,  1.f,  1.f);
            p2.set(-1.f, -1.f,  1.f);
            p3.set(-1.f, -1.f, -1.f);
            break;
         case sglFrustum::eUP:
            p1.set(-1.f,  1.f,  1.f);
            p2.set(-1.f,  1.f, -1.f);
            p3.set( 1.f,  1.f, -1.f);
            break;
         case sglFrustum::eRIGHT:
            p1.set( 1.f,  1.f, -1.f);
            p2.set( 1.f, -1.f, -1.f);
            p3.set( 1.f, -1.f,  1.f);
            break;
         case sglFrustum::eDOWN:
            p1.set(-1.f, -1.f, -1.f);
            p2.set(-1.f, -1.f,  1.f);
            p3.set( 1.f, -1.f,  1.f);
            break;
         case sglFrustum::eFRONT:
            p1.set(-1.f,  1.f, -1.f);
            p2.set(-1.f, -1.f, -1.f);
            p3.set( 1.f, -1.f, -1.f);
            break;
         case sglFrustum::eBACK:
            p1.set( 1.f,  1.f,  1.f);
            p2.set( 1.f, -1.f,  1.f);
            p3.set(-1.f, -1.f,  1.f);
            break;
      }

      ::mul(p1,(T)1.0,inv, r1);
      ::mul(p2,(T)1.0,inv, r2);
      ::mul(p3,(T)1.0,inv, r3);

      if (r1[3] == 0.0 || r2[3] == 0.0 || r3[3] == 0.0)
      {
         return false;
      }

      r1 *= (1.0/r1[3]);
      r2 *= (1.0/r2[3]);
      r3 *= (1.0/r3[3]);

      p1.set(r1[0],r1[1],r1[2]);
      p2.set(r2[0],r2[1],r2[2]);
      p3.set(r3[0],r3[1],r3[2]);

      if (!m_plane[i].build(p1,p2,p3))
      {
         return false;
      }
   }

#ifdef SGL_FRUSTUM_DEBUG
   // build the frustum geometry
   //  7      6
   //    3  2
   //    0  1
   //  4      5
   m_vertices[0].set(-1.f,-1.f,-1.f);   // front
   m_vertices[1].set( 1.f,-1.f,-1.f);
   m_vertices[2].set( 1.f, 1.f,-1.f);
   m_vertices[3].set(-1.f, 1.f,-1.f);

   m_vertices[4].set(-1.f,-1.f, 1.f);   // back
   m_vertices[5].set( 1.f,-1.f, 1.f);
   m_vertices[6].set( 1.f, 1.f, 1.f);
   m_vertices[7].set(-1.f, 1.f, 1.f);

   sglVec4f vertices[8];
   for (unsigned int ix=0; ix<8; ix++)
   {
      ::mul(m_vertices[ix], (T)1.0, inv, vertices[ix]);
      vertices[ix] *= (1.0/vertices[ix][3]);
      m_vertices[ix].set(vertices[ix][0], vertices[ix][1], vertices[ix][2]);
   }
#endif

   m_num_planes = 6;
   return true;
}


//----------------------------------------------------------------------------
template <class T>
inline bool sglFrustum<T>::buildPerspective(double near_left,
                                            double near_right,
                                            double near_bottom,
                                            double near_top,
                                            double near_clip, double far_clip)
{
   if (near_clip <= 0.0)
   {
       return false;
   }

   m_ortho_flag = false;
   m_aspect = (near_right - near_left)/(near_top - near_bottom);
   m_fov_y = atan2(near_top, near_clip) - atan2(near_bottom, near_clip);

   return buildPerspectiveInternal(near_left, near_right,
                                   near_bottom, near_top,
                                   near_clip, far_clip);
}


//----------------------------------------------------------------------------
template <class T>
inline bool sglFrustum<T>::buildPerspective(double fovy, double aspect,
                                            double near_clip, double far_clip)
{
   m_ortho_flag = false;
   m_aspect = aspect;
   m_fov_y = fovy;

   // ============== from Mesa 3.0 distribution =======================
   // ============== Copyright 1998, Brian Paul =======================
   GLdouble near_top, near_bottom, near_left, near_right;

   near_top    = near_clip * tan( fovy * 0.5 );
   near_bottom = -near_top;

   near_left   = near_bottom * aspect;
   near_right  = near_top * aspect;
   // =================================================================

   return buildPerspectiveInternal(near_left, near_right,
                                   near_bottom, near_top,
                                   near_clip, far_clip);
}

//----------------------------------------------------------------------------
template <class T>
inline bool sglFrustum<T>::buildPerspectiveInternal(double near_left,
                                                    double near_right,
                                                    double near_bottom,
                                                    double near_top,
                                                    double near_clip,
                                                    double far_clip)
{
   // in case something goes wrong
   m_num_planes = 0;

   sglVec3<T> p1, p2, p3;
   GLdouble far_top, far_bottom, far_left, far_right;
   GLdouble factor = far_clip/near_clip;

   far_top    = near_top*factor;
   far_bottom = near_bottom*factor;

   far_left   = near_left*factor;
   far_right  = near_right*factor;

   for (int i=0; i<sglFrustumf::eNUM_PLANES; i++)
   {
      switch(i)
      {
         case sglFrustum::eLEFT:
            p1.set(far_left,  far_top,     -far_clip);
            p2.set(far_left,  far_bottom,  -far_clip);
            p3.set(near_left, near_bottom, -near_clip);
            break;
         case sglFrustum::eUP:
            p1.set(far_left,   far_top,  -far_clip);
            p2.set(near_left,  near_top, -near_clip);
            p3.set(near_right, near_top, -near_clip);
            break;
         case sglFrustum::eRIGHT:
            p1.set(near_right, near_top,    -near_clip);
            p2.set(near_right, near_bottom, -near_clip);
            p3.set(far_right,  far_bottom,  -far_clip);
            break;
         case sglFrustum::eDOWN:
            p1.set(near_left, near_bottom, -near_clip);
            p2.set(far_left,  far_bottom,  -far_clip);
            p3.set(far_right, far_bottom,  -far_clip);
            break;
         case sglFrustum::eFRONT:
            p1.set(near_left,  near_top,    -near_clip);
            p2.set(near_left,  near_bottom, -near_clip);
            p3.set(near_right, near_bottom, -near_clip);
            break;
         case sglFrustum::eBACK:
            p1.set(far_right, far_top,    -far_clip);
            p2.set(far_right, far_bottom, -far_clip);
            p3.set(far_left,  far_bottom, -far_clip);
            break;
      }

      if (!m_plane[i].build(p1,p2,p3))
      {
         return false;
      }
   }

#ifdef SGL_FRUSTUM_DEBUG
   // build the frustum geometry
   //  7      6
   //    3  2
   //    0  1
   //  4      5
   m_vertices[0].set(near_left,  near_bottom, -near_clip);
   m_vertices[1].set(near_right, near_bottom, -near_clip);
   m_vertices[2].set(near_right, near_top,    -near_clip);
   m_vertices[3].set(near_left,  near_top,    -near_clip);

   m_vertices[4].set(far_left,  far_bottom, -far_clip);;
   m_vertices[5].set(far_right, far_bottom, -far_clip);
   m_vertices[6].set(far_right, far_top,    -far_clip);
   m_vertices[7].set(far_left,  far_top,    -far_clip);
#endif

   m_num_planes = 6;
   return true;
}

//----------------------------------------------------------------------------
template <class T>
inline bool sglFrustum<T>::buildOrtho(double left, double right,
                                      double bottom, double top,
                                      double near_clip, double far_clip)
{
   // in case something goes wrong
   m_num_planes = 0;

   m_ortho_flag = true;
   m_aspect = (right - left)/(top - bottom);
   m_fov_y = -1.0;

   sglVec3<T> p1, p2, p3;

   for (int i=0; i<sglFrustumf::eNUM_PLANES; i++)
   {
      switch(i)
      {
         case sglFrustum::eLEFT:
            p1.set(left, top,    -far_clip);
            p2.set(left, bottom, -far_clip);
            p3.set(left, bottom, -near_clip);
            break;
         case sglFrustum::eUP:
            p1.set(left,  top, -far_clip);
            p2.set(left,  top, -near_clip);
            p3.set(right, top, -near_clip);
            break;
         case sglFrustum::eRIGHT:
            p1.set(right, top,    -near_clip);
            p2.set(right, bottom, -near_clip);
            p3.set(right, bottom, -far_clip);
            break;
         case sglFrustum::eDOWN:
            p1.set(left,  bottom, -near_clip);
            p2.set(left,  bottom, -far_clip);
            p3.set(right, bottom, -far_clip);
            break;
         case sglFrustum::eFRONT:
            p1.set(left,  top,    -near_clip);
            p2.set(left,  bottom, -near_clip);
            p3.set(right, bottom, -near_clip);
            break;
         case sglFrustum::eBACK:
            p1.set(right, top,    -far_clip);
            p2.set(right, bottom, -far_clip);
            p3.set(left,  bottom, -far_clip);
            break;
      }

      if (!m_plane[i].build(p1,p2,p3))
      {
         return false;
      }
   }

#ifdef SGL_FRUSTUM_DEBUG
   // build the frustum geometry
   //  7      6
   //    3  2
   //    0  1
   //  4      5
   m_vertices[0].set(left,  bottom, -near_clip);;
   m_vertices[1].set(right, bottom, -near_clip);
   m_vertices[2].set(right, top,    -near_clip);
   m_vertices[3].set(left,  top,    -near_clip);

   m_vertices[4].set(left,  bottom, -far_clip);;
   m_vertices[5].set(right, bottom, -far_clip);
   m_vertices[6].set(right, top,    -far_clip);
   m_vertices[7].set(left,  top,    -far_clip);
#endif

   m_num_planes = 6;
   return true;
}

#ifdef SGL_FRUSTUM_DEBUG
//----------------------------------------------------------------------------
template <class T>
inline void sglFrustum<T>::drawImmediate() const
{
   //  7      6
   //    3  2
   //    0  1
   //  4      5
   glColor3f(1.f, 1.f, 1.f);

   glBegin(GL_LINE_STRIP);
   glVertex3f(m_vertices[0][0], m_vertices[0][1], m_vertices[0][2]);
   glVertex3f(m_vertices[1][0], m_vertices[1][1], m_vertices[1][2]);
   glVertex3f(m_vertices[2][0], m_vertices[2][1], m_vertices[2][2]);
   glVertex3f(m_vertices[3][0], m_vertices[3][1], m_vertices[3][2]);
   glVertex3f(m_vertices[0][0], m_vertices[0][1], m_vertices[0][2]);
   glVertex3f(m_vertices[4][0], m_vertices[4][1], m_vertices[4][2]);
   glVertex3f(m_vertices[5][0], m_vertices[5][1], m_vertices[5][2]);
   glVertex3f(m_vertices[6][0], m_vertices[6][1], m_vertices[6][2]);
   glVertex3f(m_vertices[7][0], m_vertices[7][1], m_vertices[7][2]);
   glVertex3f(m_vertices[4][0], m_vertices[4][1], m_vertices[4][2]);
   glEnd();

   glBegin(GL_LINES);
   glVertex3f(m_vertices[1][0], m_vertices[1][1], m_vertices[1][2]);
   glVertex3f(m_vertices[5][0], m_vertices[5][1], m_vertices[5][2]);
   glVertex3f(m_vertices[2][0], m_vertices[2][1], m_vertices[2][2]);
   glVertex3f(m_vertices[6][0], m_vertices[6][1], m_vertices[6][2]);
   glVertex3f(m_vertices[3][0], m_vertices[3][1], m_vertices[3][2]);
   glVertex3f(m_vertices[7][0], m_vertices[7][1], m_vertices[7][2]);
   glEnd();
}
#endif // SGL_FRUSTUM_DEBUG

#endif
