/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglGeoSet.hpp
 *  Created: 22 October 1997
 *  Summary: Contains vertex, color, and normal data for geometry (texcoords
 *           have been moved to sglDrawable with setting of state).  Specific
 *           primitive types are defined in the subclasses
 *****************************************************************************/

#ifndef __SGL_GEOSET_HPP
#define __SGL_GEOSET_HPP

#include <sgl.h>
#include <sglVector.hpp>
#include <sglDrawable.hpp>
#include <sglGeoSetDataPtr.hpp>

//============================================================================

/**

This abstract class contains the vertex, normal, and color arrays for a single
type of primitive (e.g. points, lines, triangles, etc.) that is determined
when instantiating one of the concrete subclasses (e.g. sglPointSet,
sglLineSet, sglTriangleSet, etc.).  Every sglGeoSet must, at a minimum, have an
array of vertices assigned using the setCoordList() member function, and
the number of primitives set using the setNumPrims() member function.
Optionally, normal and vertex color attributes can be also be defined.
The first parameter to the setColorList() or setNormalList() member functions
define the way that these arrays can be bound to the vertices as follows:

\begin{enumerate}
\item sglGeoSet::ePER_VERTEX - like the coordinate vertex array, this array
must have the same number of entries and each element is applied with a
corresponding vertex.

\item sglGeoSet::ePER_PRIM - the array contains as many entries as the number
of primitives, one for each.

\item sglGeoSet::eOVERALL - the array contains one element and applies to for
all vertices in the coordinate list.
\end{enumerate}

Passing in a NULL pointer for the attribute list has the effect of disabling
that attribute.  The getColorBind() and getNormalBind() functions
return what mode the particular attribute is currently set to.  If
the attribute is unused (disabled) then the SGL_OFF token is returned.

{\bf Concrete subclasses: sglPointSet, sglLineSet, sglTriangleSet, sglQuadSet}

sglPointSet, sglLineSet, sglTriangleSet, and sglQuadSet are the only immediate
concrete subclasses of sglGeoSet and allow the rendering of individual
points, line segments, triangles, and quads.  They imply that the
primitive sizes are 1, 2, 3, and 4 respectively.  The number of primitives,
which is set by calling the sglGeoSet::setNumPrims() function, along with
the classes' implied primitive size determine the required sizes of the
geometry attributes.  For example, an sglLineSet with the number of
primitives set to 9, means that the vertex array must be at least 18 elements
long.  Note, no error checking is done by SGL so if the user makes
a mistake, access out of array bounds can result in undefined behaviour.

{\bf Example:}
\begin{verbatim}
   sglVec3f *vertex = new sglVec4f[6];      // vertices for two triangles
   vertex[0].set( 0.0f, 0.0f, 0.0f);
   vertex[1].set( 5.0f, 0.0f, 0.0f);
   vertex[2].set( 0.0f, 5.0f, 0.0f);
   vertex[3].set( 0.0f, 0.0f, 0.0f);
   vertex[4].set( 0.0f,-5.0f, 0.0f);
   vertex[5].set(-5.0f, 0.0f, 0.0f);

   sglVec4f *color = new sglVec4f[2];
   color[0].set(1.0f, 0.0f, 0.0f, 1.0f);   // first triangle is red
   color[1].set(0.0f, 1.0f, 0.0f, 1.0f);   // second triangle is green

   sglVec3f *normal = new sglVec3f[1];     // one normal for all vertices
   normal[0].set(0.0f,0.0f,1.0f);

   sglTriangleSet *tri_geom = new sglTriangleSet;
   tri_geom->setNumPrims(2);
   tri_geom->setCoorList(vertex);          // sglGeoSet::ePER_VERTEX is implied
   tri_geom->setColorList(sglGeoSet::ePER_PRIM, color);
   tri_geom->setNormalList(sglGeoSet::eOVERALL, normal);
\end{verbatim}
*/

class SGL_DLL_API sglGeoSet : public sglDrawable
{
public:
   ///
   enum BindEnum {eOFF,
                  eOVERALL,
                  ePER_PRIM,
                  ePER_VERTEX};

public:
   /// default constructor
   sglGeoSet();
   /// virtual destructor
   virtual ~sglGeoSet();

   /// retrieve the type of primitive
   virtual sglStats::PrimEnum getPrimType() const = 0;
   ///
   virtual GLenum getGLPrimType() const = 0;
   ///
   virtual unsigned int getPrimSize() const = 0;

   /// get/set the number of primitives in this geoset
   unsigned int getNumPrims() const {return m_num_prims;}
   ///
   void setNumPrims(unsigned int num);

   /// Set the array of vertices.  Deprecated.
   void setCoordList(sglVec3f *alist)
      { setCoordList(new sglNoRefDataPtr(alist)); }

   /** Set the array of vertices.  This will increment the reference count
       on the given data pointer, and decrement the reference count of
       the previous data pointer.  If the previous data pointer's refernce
       count goes to zero, it will be deleted.
       @param data Pointer to the list of vertices.
       @param type Data type of vertices [eSHORT, eINT, eFLOAT, eDOUBLE].
       @param size Number of values per vertex [2,3,4].
       @param offset Number of bytes in the array before the first vertex.
       @param stride Number of bytes from the start of one vertex to the start
                     of the next vertex.  If zero, the vertices are assumed
                     to be tightly packed (i.e. stride == size * sizeof(type)).
                     Stride must be non-negative.
   */
   void setCoordList(sglDataPtr *data,
                     sglDataPtr::DataTypeEnum type = sglDataPtr::eFLOAT,
                     int size = 3,
                     int offset = 0,
                     int stride = 0);

   /// Set the array of normals.  Deprecated.
   void setNormalList(BindEnum bind, sglVec3f *alist)
      { setNormalList(bind, new sglNoRefDataPtr(alist)); }

   /** Set the array of normals.  This will increment the reference count
       on the given data pointer, and decrement the reference count of
       the previous data pointer.  If the previous data pointer's refernce
       count goes to zero, it will be deleted.
       Normals must have three values per normal.
       @param bind One of [eOVERALL, ePER_PRIM, ePER_VERTEX].
       @param data Pointer to the list of normals.
       @param type Data type of normals [eBYTE, eSHORT, eINT, eFLOAT, eDOUBLE].
       @param offset Number of bytes in the array before the first normal.
       @param stride Number of bytes from the start of one normal to the start
                     of the next normal.  If zero, the normals are assumed
                     to be tightly packed (i.e. stride == 3 * sizeof(type)).
                     Stride must be non-negative.
   */
   void setNormalList(BindEnum bind,
                      sglDataPtr *data,
                      sglDataPtr::DataTypeEnum type = sglDataPtr::eFLOAT,
                      int offset = 0,
                      int stride = 0);

   /// Set the array of colors.  Deprecated.
   void setColorList(BindEnum bind, sglVec4f *alist)
      { setColorList(bind, new sglNoRefDataPtr(alist)); }

   /** Set the array of colors.  This will increment the reference count
       on the given data pointer, and decrement the reference count of
       the previous data pointer.  If the previous data pointer's refernce
       count goes to zero, it will be deleted.
       @param bind One of [eOVERALL, ePER_PRIM, ePER_VERTEX].
       @param data Pointer to the list of colors.
       @param type Data type of colors (any).
       @param size Number of values per color [3,4].
       @param offset Number of bytes in the array before the first color.
       @param stride Number of bytes from the start of one color to the start
                     of the next color.  If zero, the colors are assumed
                     to be tightly packed (i.e. stride == size * sizeof(type)).
                     Stride must be non-negative.
   */
   void setColorList(BindEnum bind,
                     sglDataPtr *data,
                     sglDataPtr::DataTypeEnum type = sglDataPtr::eFLOAT,
                     int size = 4,
                     int offset = 0,
                     int stride = 0);

   /// get the current normal and color binding state
   BindEnum getNormalBind() const;
   ///
   BindEnum getColorBind() const;

   /// get the current vertex, normal, color attribute lists
   sglDataPtr *getCoordList() const { return m_vertices.getDataPtr(); }
   ///
   sglDataPtr *getNormalList() const { return m_normals.getDataPtr(); }
   ///
   sglDataPtr *getColorList() const { return m_colors.getDataPtr(); }

   /// get the current vertex, normal, color data types
   sglDataPtr::DataTypeEnum getCoordType() const
      { return m_vertices.getDataType(); }
   ///
   sglDataPtr::DataTypeEnum getNormalType() const
      { return m_normals.getDataType(); }
   ///
   sglDataPtr::DataTypeEnum getColorType() const
      { return m_colors.getDataType(); }

   /// get the number of values per vertex coordinate and color
   int getCoordSize() const { return m_vertices.getSize(); }
   ///
   int getColorSize() const { return m_colors.getSize(); }

   /// get the number of bytes in the array before the first coordinate value
   int getCoordOffset() const { return m_vertices.getOffset(); }
   ///
   int getNormalOffset() const { return m_normals.getOffset(); }
   ///
   int getColorOffset() const { return m_colors.getOffset(); }

   /// get the number of bytes from the start of one vertex to the next
   int getCoordStride() const { return m_vertices.getStride(); }
   ///
   int getNormalStride() const { return m_normals.getStride(); }
   ///
   int getColorStride() const { return m_colors.getStride(); }

   ///
   virtual bool isValid() const;

   /// Add this drawable's stats to the given debug_stats.
   virtual void addStats(sglStats &stats) const;

   ///
   virtual void printInfo(ostream &ostrm, const char *indent_string) const;

protected:
   virtual bool computeBounds();
   virtual void updateDrawFunc();

   enum NormalEnum
   {
      eNORMALS_NONE         = 0x00000,
      eNORMALS_OVERALL      = 0x00001,
      eNORMALS_PER_PRIM     = 0x00002,
      eNORMALS_PER_VERTEX   = 0x00003
   };

   enum ColorEnum
   {
      eCOLORS_NONE          = 0x00000,
      eCOLORS_OVERALL       = 0x00004,
      eCOLORS_PER_PRIM      = 0x00008,
      eCOLORS_PER_VERTEX    = 0x0000c
   };

   enum ModeEnum
   {
      eNORMALS              = 0x00003,
      eCOLORS               = 0x0000c,
      eMULTI_TEXTURE        = 0x00010,
      ePREFETCH             = 0x00020
   };

   // For bezier classes.
   enum TexCoordsEnum
   {
      eTEXCOORDS_OVERALL    = 0x01000,
      eTEXCOORDS_PER_PRIM   = 0x02000,
      eTEXCOORDS_PER_VERTEX = 0x04000,
      eTEXCOORDS            = 0x07000
   };

private:  // not implemented
   sglGeoSet(const sglGeoSet &);
   sglGeoSet &operator=(const sglGeoSet &);

#ifdef INLINE_BOXBOUND
   typedef void (*BBFunc)(
      sglBoxBound &bbox, void *ptr, int vert_inc, int num_verts);

   static void buildBoxBound2s(
      sglBoxBound &bbox, void *ptr, int vert_inc, int num_verts);
   static void buildBoxBound2i(
      sglBoxBound &bbox, void *ptr, int vert_inc, int num_verts);
   static void buildBoxBound2f(
      sglBoxBound &bbox, void *ptr, int vert_inc, int num_verts);
   static void buildBoxBound2d(
      sglBoxBound &bbox, void *ptr, int vert_inc, int num_verts);
   static void buildBoxBound3s(
      sglBoxBound &bbox, void *ptr, int vert_inc, int num_verts);
   static void buildBoxBound3i(
      sglBoxBound &bbox, void *ptr, int vert_inc, int num_verts);
   static void buildBoxBound3f(
      sglBoxBound &bbox, void *ptr, int vert_inc, int num_verts);
   static void buildBoxBound3d(
      sglBoxBound &bbox, void *ptr, int vert_inc, int num_verts);
   static void buildBoxBound4s(
      sglBoxBound &bbox, void *ptr, int vert_inc, int num_verts);
   static void buildBoxBound4i(
      sglBoxBound &bbox, void *ptr, int vert_inc, int num_verts);
   static void buildBoxBound4f(
      sglBoxBound &bbox, void *ptr, int vert_inc, int num_verts);
   static void buildBoxBound4d(
      sglBoxBound &bbox, void *ptr, int vert_inc, int num_verts);
#endif

   void drawVT(const vector<sglTexCoords> &tex_coords) const;
   void drawVNoT(const vector<sglTexCoords> &tex_coords) const;
   void drawVNpT(const vector<sglTexCoords> &tex_coords) const;
   void drawVNvT(const vector<sglTexCoords> &tex_coords) const;
   void drawVCoT(const vector<sglTexCoords> &tex_coords) const;
   void drawVCoNoT(const vector<sglTexCoords> &tex_coords) const;
   void drawVCoNpT(const vector<sglTexCoords> &tex_coords) const;
   void drawVCoNvT(const vector<sglTexCoords> &tex_coords) const;
   void drawVCpT(const vector<sglTexCoords> &tex_coords) const;
   void drawVCpNoT(const vector<sglTexCoords> &tex_coords) const;
   void drawVCpNpT(const vector<sglTexCoords> &tex_coords) const;
   void drawVCpNvT(const vector<sglTexCoords> &tex_coords) const;
   void drawVCvT(const vector<sglTexCoords> &tex_coords) const;
   void drawVCvNoT(const vector<sglTexCoords> &tex_coords) const;
   void drawVCvNpT(const vector<sglTexCoords> &tex_coords) const;
   void drawVCvNvT(const vector<sglTexCoords> &tex_coords) const;

#ifndef SGL_NO_VERTEX_ARRAYS
   void drawVTm(const vector<sglTexCoords> &tex_coords) const;
   void drawVNoTm(const vector<sglTexCoords> &tex_coords) const;
   void drawVNpTm(const vector<sglTexCoords> &tex_coords) const;
   void drawVNvTm(const vector<sglTexCoords> &tex_coords) const;
   void drawVCoTm(const vector<sglTexCoords> &tex_coords) const;
   void drawVCoNoTm(const vector<sglTexCoords> &tex_coords) const;
   void drawVCoNpTm(const vector<sglTexCoords> &tex_coords) const;
   void drawVCoNvTm(const vector<sglTexCoords> &tex_coords) const;
   void drawVCpTm(const vector<sglTexCoords> &tex_coords) const;
   void drawVCpNoTm(const vector<sglTexCoords> &tex_coords) const;
   void drawVCpNpTm(const vector<sglTexCoords> &tex_coords) const;
   void drawVCpNvTm(const vector<sglTexCoords> &tex_coords) const;
   void drawVCvTm(const vector<sglTexCoords> &tex_coords) const;
   void drawVCvNoTm(const vector<sglTexCoords> &tex_coords) const;
   void drawVCvNpTm(const vector<sglTexCoords> &tex_coords) const;
   void drawVCvNvTm(const vector<sglTexCoords> &tex_coords) const;
#endif

   class Initializer;
   friend class Initializer;
   static void setupFuncTable();

protected:
   unsigned int m_mode; // encodes everything about this geoset

   unsigned int m_num_prims;

   sglVertexPtr m_vertices;
   sglNormalPtr m_normals;
   sglColorPtr  m_colors;

   enum BoundsEnum {eNUMPRIMS  = 0x01,
                    eLENGTHS   = 0x02,
                    eVERTICES  = 0x04,
                    eVINDICES  = 0x08};

   unsigned int m_compute_bounds;

private:
   static DrawFunc s_func_table[(eMULTI_TEXTURE|eCOLORS|eNORMALS)+1];
#ifdef INLINE_BOXBOUND
   static BBFunc s_bb_func_table[24];
#endif
   static Initializer sInitializer;
};

#endif
