/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglGeoStripSet.hpp
 *  Created: 28 August 1999
 *  Summary: Subclass of geoset that has primitives of varying lengths.
 *****************************************************************************/

#ifndef __SGL_GEO_STRIP_SET_HPP
#define __SGL_GEO_STRIP_SET_HPP

#include <sgl.h>
#include <sglGeoSet.hpp>

//============================================================================

/**

Unlike the sglGeoSet class, sglGeoStripSet is capable of fully specifying those
primitives whose lengths are variable: sglLineStripSet, sglTriangleFanSet,
sglTriangleStripSet, sglQuadStripSet or sglPolygonSet.  This class extends
the sglGeoSet class to include a lengths array with a number of elements equal
to the number of primitives set by the setNumPrims() function.  Each
element of the lengths array specifies how many vertices are used for each
primitive (the sum of all the elements in the lengths array defines the
required size of the coordinate array and any other attribute array that has an
sglGeoSet::ePER_VERTEX binding).

{\bf Concrete subclasses: sglLineStripSet, sglTriangleFanSet,
sglTriangleStripSet, sglQuadStripSet, sglPolygonSet}

These are the concrete subclasses of sglGeoStripSet that can be instantiated
by the user.  The example below illustrates the creation of an
sglTriangleStripSet with two strips of lengths 6 and 4 with per primitive
normal attributes, and overall color attribute.

{\bf Example:}
\begin{verbatim}
   unsigned int *lengths = new lengths[2];
   lengths[0] = 6;                   // implies use of vertices[0..5]
   lengths[1] = 4;                   // implies use of vertices[6..9]

   sglVec3f *vertex = new sglVec4f[10];
   vertex[0].set(0.0f, 0.0f, 0.0f);  // 1st primitive: 4 triangles in a strip
   vertex[1].set(5.0f, 0.0f, 0.0f);
   vertex[2].set(0.0f, 4.0f, 0.0f);
   vertex[3].set(5.0f, 4.0f, 0.0f);
   vertex[4].set(0.0f, 8.0f, 0.0f);
   vertex[5].set(5.0f, 8.0f, 0.0f);

   vertex[6].set(5.0f, 0.0f, 0.0f);  // 2nd primitive: 2 triangles in a strip
   vertex[7].set(0.0f, 0.0f, 0.0f);
   vertex[8].set(5.0f, 0.0f, 4.0f);
   vertex[9].set(0.0f, 0.0f, 4.0f);

   sglVec4f *color = new sglVec4f[1];
   color[0].set(1.0f, 0.0f, 0.0f, 1.0f);   // both strips red

   sglVec3f *normal = new sglVec3f[2];     // one normal for each primitive
   normal[0].set(0.0f,0.0f,1.0f);
   normal[1].set(0.0f,1.0f,0.0f);

   sglTriangleStripSet *tstrip = new sglTriangleStripSet;
   tstrip->setNumPrims(2);
   tstrip->setPrimLengths(lengths)
   tstrip->setCoordList(vertex);           // sglGeoSet::ePER_VERTEX is implied
   tstrip->setColorList(sglGeoSet::eOVERALL, color);
   tstrip->setNormalList(sglGeoSet::ePER_PRIM, normal);
\end{verbatim}

 */

class SGL_DLL_API sglGeoStripSet : public sglGeoSet
{
public:
   /// default constructor
   sglGeoStripSet();
   /// virtual destructor
   virtual ~sglGeoStripSet();

   /// Not very useful for variable sized primitives.
   virtual unsigned int getPrimSize() const {return 0;}

   /// Get the lengths array (for strip and poly primitives only).
   sglDataPtr *getPrimLengths() const {return m_lengths.getDataPtr();}

   /// Set the array of lengths.  Deprecated.
   void setPrimLengths(sglIndexType *alist)
      { setPrimLengths(new sglNoRefDataPtr(alist)); }

   /** Set the array of lengths.  This will increment the reference count
       on the given data pointer, and decrement the reference count of
       the previous data pointer.  If the previous data pointer's refernce
       count goes to zero, it will be deleted.
       The lengths must be of type sglIndexType.
       @param data Pointer to the list of lengths.
       @param offset Number of bytes in the array before the first length.
       @param stride Number of bytes from the start of one length to the start
                     of the next length.  If zero, the lengths are assumed
                     to be tightly packed (i.e. stride == sizeof(type)).
                     Stride must be non-negative.
   */
   void setPrimLengths(sglDataPtr *data,
                       int offset = 0,
                       int stride = 0);

   /// Get the number of bytes in the array before the first length.
   int getPrimLengthsOffset() const { return m_lengths.getOffset(); }

   /// Get the number of bytes from the start of one length to the next.
   int getPrimLengthsStride() const { return m_lengths.getStride(); }

   /// Add this drawable's stats to the given sglStats.
   virtual void addStats(sglStats &stats) const;

protected:
   virtual bool computeBounds();
   virtual void updateDrawFunc();

private:  // not implemented
   sglGeoStripSet(const sglGeoStripSet &);
   sglGeoStripSet &operator=(const sglGeoStripSet &);

   void drawVT(const vector<sglTexCoords> &tex_coords) const;
   void drawVNoT(const vector<sglTexCoords> &tex_coords) const;
   void drawVNpT(const vector<sglTexCoords> &tex_coords) const;
   void drawVNvT(const vector<sglTexCoords> &tex_coords) const;
   void drawVCoT(const vector<sglTexCoords> &tex_coords) const;
   void drawVCoNoT(const vector<sglTexCoords> &tex_coords) const;
   void drawVCoNpT(const vector<sglTexCoords> &tex_coords) const;
   void drawVCoNvT(const vector<sglTexCoords> &tex_coords) const;
   void drawVCpT(const vector<sglTexCoords> &tex_coords) const;
   void drawVCpNoT(const vector<sglTexCoords> &tex_coords) const;
   void drawVCpNpT(const vector<sglTexCoords> &tex_coords) const;
   void drawVCpNvT(const vector<sglTexCoords> &tex_coords) const;
   void drawVCvT(const vector<sglTexCoords> &tex_coords) const;
   void drawVCvNoT(const vector<sglTexCoords> &tex_coords) const;
   void drawVCvNpT(const vector<sglTexCoords> &tex_coords) const;
   void drawVCvNvT(const vector<sglTexCoords> &tex_coords) const;

#ifndef SGL_NO_VERTEX_ARRAYS
   void drawVTm(const vector<sglTexCoords> &tex_coords) const;
   void drawVNoTm(const vector<sglTexCoords> &tex_coords) const;
   void drawVNpTm(const vector<sglTexCoords> &tex_coords) const;
   void drawVNvTm(const vector<sglTexCoords> &tex_coords) const;
   void drawVCoTm(const vector<sglTexCoords> &tex_coords) const;
   void drawVCoNoTm(const vector<sglTexCoords> &tex_coords) const;
   void drawVCoNpTm(const vector<sglTexCoords> &tex_coords) const;
   void drawVCoNvTm(const vector<sglTexCoords> &tex_coords) const;
   void drawVCpTm(const vector<sglTexCoords> &tex_coords) const;
   void drawVCpNoTm(const vector<sglTexCoords> &tex_coords) const;
   void drawVCpNpTm(const vector<sglTexCoords> &tex_coords) const;
   void drawVCpNvTm(const vector<sglTexCoords> &tex_coords) const;
   void drawVCvTm(const vector<sglTexCoords> &tex_coords) const;
   void drawVCvNoTm(const vector<sglTexCoords> &tex_coords) const;
   void drawVCvNpTm(const vector<sglTexCoords> &tex_coords) const;
   void drawVCvNvTm(const vector<sglTexCoords> &tex_coords) const;
#endif

   class Initializer;
   friend class Initializer;
   static void setupFuncTable();

protected:
   sglIndexPtr m_lengths;    // List of strip lengths

private:
   static DrawFunc s_func_table[(eMULTI_TEXTURE|eCOLORS|eNORMALS)+1];
   static Initializer sInitializer;
};

#endif
