/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglGeode.cpp
 *  Created: 12 August 1997
 *  Summary: scenegraph leaf nodes use to hold a set of sglGeometries
 *****************************************************************************/

#include <sgl.h>
#include <sglGeode.hpp>
#include <sglDrawable.hpp>
#include <sglSphereBound.hpp>
#include <sglBoxBound.hpp>
#include <sglCull.hpp>
#include <sglOStream.hpp>

//============================================================================
// class sglGeode : public sglNode
//============================================================================

//----------------------------------------------------------------------------
sglGeode::sglGeode()
      : sglNode(),
        m_geometry_list()
{
}

//----------------------------------------------------------------------------
sglGeode::~sglGeode()
{
   // traverse list and deref (and delete?) everything
   while (m_geometry_list.size())
   {
      sglDrawable *geometry = m_geometry_list.back();
      m_geometry_list.pop_back();
      geometry->deref();

      geometry->removeParent(this);

      if (geometry->getRefCount() == 0)
      {
         delete geometry;
      }
   }
}

//----------------------------------------------------------------------------
bool sglGeode::addGeometry(sglDrawable *geometry)
{
   if (geometry)
   {
      m_geometry_list.push_back(geometry);
      geometry->ref();
      geometry->addParent(this);

      return true;
   }
   else
   {
      sglPrintWarning()
         << "sglGroup::addGeometry error: trying to add a NULL geometry.\n";
      return false;
   }
}

//----------------------------------------------------------------------------
bool sglGeode::insertGeometry(unsigned int index, sglDrawable *geometry)
{
   if (geometry)
   {
      if (index < getNumGeometries())
      {
         m_geometry_list.insert(m_geometry_list.begin()+index, geometry);
         geometry->ref();
         geometry->addParent(this);

         return true;
      }
      else if (index == getNumGeometries())
      {
         m_geometry_list.push_back(geometry);
         geometry->ref();
         geometry->addParent(this);

         return true;
      }
      else
      {
         sglPrintWarning() << "sglGroup::insertGeometry error: "
              << "trying to insert with invalid index = " << index << endl;
         return false;
      }
   }
   else
   {
      sglPrintWarning() << "sglGroup::insertGeometry error: "
           << "trying to insert a NULL geometry.\n";
      return false;
   }
}

//----------------------------------------------------------------------------
sglDrawable *sglGeode::getGeometry(unsigned int index)
{
   if (index < getNumGeometries())
   {
      return m_geometry_list[index];
   }
   else
   {
      return NULL;
   }
}

//----------------------------------------------------------------------------
int sglGeode::searchGeometries(sglDrawable *geometry)
{
   if (geometry)
   {
      int i=0;
      for (vector<sglDrawable*>::iterator elt=m_geometry_list.begin();
           elt != m_geometry_list.end();
           ++elt, ++i)
      {
         if (*elt == geometry)
         {
            return i;
         }
      }
   }

   return -1;
}

//----------------------------------------------------------------------------
bool sglGeode::removeGeometry(sglDrawable *geometry)
{
   if (geometry)
   {
      int index = searchGeometries(geometry);
      if (index >= 0)
      {
         m_geometry_list[index]->removeParent(this);

         // ----------------------------------------------------------
         // NOTE: m_geometry_list[index]  = *(m_geometry_list.begin()+index)
         // ----------------------------------------------------------

         m_geometry_list.erase(m_geometry_list.begin()+index);
         geometry->deref();

         return true;
      }
      else
      {
         sglPrintWarning() << "sglGroup::removeGeometry error: "
              << "could not find referenced geometry.\n";
         return false;
      }
   }
   else
   {
      sglPrintWarning() << "sglGroup::removeGeometry error: "
           << "trying to remove a NULL geometry.\n";
      return false;
   }
}


//----------------------------------------------------------------------------
void sglGeode::computeBound()
{
   m_dirty &= ~eBOUND;

   if (getNumGeometries())
   {
      // get first bbox;
      sglBoxBound bbox(m_geometry_list[0]->getBound());

      // don't clear this dirty bit if child's bit is dirty; can happen during
      // a getBound() query.
      m_dirty |= ((m_geometry_list[0]->m_dirty)&eBOUND);

      // expand box to include each geode
      for (vector<sglDrawable *>::const_iterator elt =
              m_geometry_list.begin() + 1;
           elt != m_geometry_list.end();
           ++elt)
      {
         // add another child's box
         bbox.expandToInclude((*elt)->getBound());

         // don't clear this dirty bit if child's bit is dirty; can happen
         // during a getBound() query.
         m_dirty |= (((*elt)->m_dirty)&eBOUND);
      }

      m_sphere.computeFromBox(bbox);
   }
   else
   {
      // mark sphere as empty
      m_sphere.makeEmpty();
   }
}

//----------------------------------------------------------------------------
void sglGeode::computeBoundsRecursiveDown()
{
   if ((m_dirty&eBOUND) && m_bound_mode != sglBound::eSTATIC)
   {
      for (vector<sglDrawable*>::iterator ii = m_geometry_list.begin();
           ii != m_geometry_list.end(); ++ii)
      {
         if ((*ii)->m_dirty&eBOUND)
            (*ii)->computeBoundsRecursiveDown();
      }

      computeBound();
   }
}

//----------------------------------------------------------------------------
sglBound::IntersectResultEnum sglGeode::intersect(sglIntersectf &segset) const
{
   sglSegmentf seg(segset.getSegment());
   sglBound::IntersectResultEnum result = m_sphere.intersect(seg);
   if (result == sglBound::eOUTSIDE)
   {
      return result;
   }

   // else result == INSIDE, INTERSECT
   if (segset.getMode()&sglProbe::ePRIMITIVE)
   {
      result = sglBound::eOUTSIDE;
      for (vector<sglDrawable *>::const_iterator elt=m_geometry_list.begin();
           elt != m_geometry_list.end();
           ++elt)
      {
         sglBound::IntersectResultEnum hit = (*elt)->intersectPrim(segset);
         if (hit == sglBound::eINTERSECT)
         {
            if (segset.getMode() & sglProbe::eGEODE)
            {
               segset.setGeode((sglGeode*)this);
            }

            if (segset.getMode() & sglProbe::eGEOMETRY)
            {
               segset.setGeometry(*elt);
            }

            result = sglBound::eINTERSECT;
            segset.setResultEnum(result);
         }
      }
   }

   else if (segset.getMode()&sglProbe::eGEOMETRY)
   {
      result = sglBound::eOUTSIDE;
      for (vector<sglDrawable *>::const_iterator elt=m_geometry_list.begin();
           elt != m_geometry_list.end();
           ++elt)
      {
         if ((*elt)->intersectBBox(segset) == sglBound::eINTERSECT)
         {
            // any sglBound::eINSIDE is same as no intersection in this case
            if (segset.getMode() & sglProbe::eGEODE)
            {
               segset.setGeode((sglGeode*)this);
            }

            result = sglBound::eINTERSECT;
            segset.setGeometry(*elt);
            segset.setResultEnum(result);
         }
      }
   }

   else // segset.getMode()&sglProbe::eGEODE
   {
      if (result == sglBound::eINTERSECT)
      {
         segset.setGeode((sglGeode*)this);
         segset.setSegment(seg);
         segset.setResultEnum(result);
      }
   }

   if (result == sglBound::eINTERSECT && segset.getMode()&sglProbe::eTRANSFORM)
   {
      segset.setTransformMatrix(segset.getModelViewMatrix());
   }

   if (result == sglBound::eINTERSECT && segset.getMode()&sglProbe::ePATH)
   {
      segset.getPath().clear();
      segset.getPath().push_front((sglNode*)this);
   }

   return result;
}

//----------------------------------------------------------------------------
sglBound::IntersectResultEnum sglGeode::intersect(sglIntersectd &segset) const
{
   sglSegmentd seg(segset.getSegment());
   sglBound::IntersectResultEnum result = m_sphere.intersect(seg);
   if (result == sglBound::eOUTSIDE)
   {
      return result;
   }

   // else result == INSIDE, INTERSECT
   if (segset.getMode()&sglProbe::ePRIMITIVE)
   {
      result = sglBound::eOUTSIDE;
      for (vector<sglDrawable *>::const_iterator elt=m_geometry_list.begin();
           elt != m_geometry_list.end();
           ++elt)
      {
         sglBound::IntersectResultEnum hit = (*elt)->intersectPrim(segset);
         if (hit == sglBound::eINTERSECT)
         {
            if (segset.getMode() & sglProbe::eGEODE)
            {
               segset.setGeode((sglGeode*)this);
            }

            if (segset.getMode() & sglProbe::eGEOMETRY)
            {
               segset.setGeometry(*elt);
            }

            result = sglBound::eINTERSECT;
            segset.setResultEnum(result);
         }
      }
   }

   else if (segset.getMode()&sglProbe::eGEOMETRY)
   {
      result = sglBound::eOUTSIDE;
      for (vector<sglDrawable *>::const_iterator elt=m_geometry_list.begin();
           elt != m_geometry_list.end();
           ++elt)
      {
         if ((*elt)->intersectBBox(segset) == sglBound::eINTERSECT)
         {
            // any sglBound::eINSIDE is same as no intersection in this case
            if (segset.getMode() & sglProbe::eGEODE)
            {
               segset.setGeode((sglGeode*)this);
            }

            result = sglBound::eINTERSECT;
            segset.setGeometry(*elt);
            segset.setResultEnum(result);
         }
      }
   }

   else // segset.getMode()&sglProbe::eGEODE
   {
      if (result == sglBound::eINTERSECT)
      {
         segset.setGeode((sglGeode*)this);
         segset.setSegment(seg);
         segset.setResultEnum(result);
      }
   }

   if (result == sglBound::eINTERSECT && segset.getMode()&sglProbe::eTRANSFORM)
   {
      segset.setTransformMatrix(segset.getModelViewMatrix());
   }

   if (result == sglBound::eINTERSECT && segset.getMode() & sglProbe::ePATH)
   {
      segset.getPath().clear();
      segset.getPath().push_front((sglNode*)this);
   }

   return result;
}

//----------------------------------------------------------------------------
void sglGeode::pick(sglPickf &pick_state,
                    unsigned int cull_flags) const
{
   // HACK If cull_flags == 0, then I don't for sure if this sphere is actually
   // inside the frustum or if the user passed in 0 for cull flags.  Should I
   // reset them here?
   // cull_flags = ~0x0;

   sglBound::IntersectResultEnum result =
      m_sphere.intersect(pick_state.getPickFrustum(), cull_flags);

   if (result == sglBound::eOUTSIDE)
   {
      return; // select nothing
   }

   if (pick_state.getMode()&sglProbe::ePATH)
   {
      pick_state.getPath().push_back((sglNode*)this);
   }

   if (pick_state.getMode() & sglProbe::eGEOMETRY)
   {
      result = sglBound::eOUTSIDE;
      for (vector<sglDrawable *>::const_iterator elt=m_geometry_list.begin();
           elt != m_geometry_list.end();
           ++elt)
      {
         if ((*elt)->cull(pick_state.getPickFrustum(), cull_flags))
         {
            pick_state.addHit(this, *elt); // sglGeode only copied if mode set

            result = sglBound::eINTERSECT;
         }
      }
   }
   else if (pick_state.getMode()&sglProbe::eGEODE)
   {
      pick_state.addHit(this, NULL);
   }

   if (pick_state.getMode()&sglProbe::ePATH)
   {
      pick_state.getPath().pop_back();
   }
}


//----------------------------------------------------------------------------
void sglGeode::pick(sglPickd &pick_state,
                    unsigned int cull_flags) const
{
   // HACK If cull_flags == 0, then I don't for sure if this sphere is actually
   // inside the frustum or if the user passed in 0 for cull flags.  Should I
   // reset them here?
   // cull_flags = ~0x0;

   sglBound::IntersectResultEnum result =
      m_sphere.intersect(pick_state.getPickFrustum(), cull_flags);

   if (result == sglBound::eOUTSIDE)
   {
      return; // select nothing
   }

   if (pick_state.getMode()&sglProbe::ePATH)
   {
      pick_state.getPath().push_back((sglGeode*)this);
   }

   if (pick_state.getMode() & sglProbe::eGEOMETRY)
   {
      result = sglBound::eOUTSIDE;
      for (vector<sglDrawable *>::const_iterator elt=m_geometry_list.begin();
           elt != m_geometry_list.end();
           ++elt)
      {
         if ((*elt)->cull(pick_state.getPickFrustum(), cull_flags))
         {
            pick_state.addHit(this, *elt); // sglGeode only copied if mode set
            result = sglBound::eINTERSECT;
         }
      }
   }
   else if (pick_state.getMode()&sglProbe::eGEODE)
   {
      pick_state.addHit(this, NULL);
   }

   if (pick_state.getMode()&sglProbe::ePATH)
   {
      pick_state.getPath().pop_back();
   }
}


//----------------------------------------------------------------------------
void sglGeode::preDraw(const PreDrawStruct &trav_state)
{
   for (vector<sglDrawable*>::iterator ii = m_geometry_list.begin();
        ii != m_geometry_list.end(); ++ii)
   {
      if ((*ii)->m_dirty)
         (*ii)->preDraw();
   }

   sglNode::preDraw(trav_state);
}

//----------------------------------------------------------------------------
void sglGeode::cull(sglCull<float> &trav_state,
                    unsigned int cull_flags) const
{
   if (m_geometry_list.empty() ||
       (cull_flags && (m_sphere.intersect(trav_state.getFrustum(),
                                          cull_flags) == sglBound::eOUTSIDE)))
   {
      return; // draw nothing
   }
   else
   {
      for (vector<sglDrawable *>::const_iterator elt=m_geometry_list.begin();
           elt != m_geometry_list.end();
           ++elt)
      {
         if ((*elt)->cull(trav_state.getFrustum(), cull_flags))
            trav_state.addDrawable(
               *elt, (*elt)->getStateByMask(trav_state.getStateMask()));
      }
   }
}


//----------------------------------------------------------------------------
void sglGeode::cull(sglCull<double> &trav_state,
                    unsigned int cull_flags) const
{
   if (m_geometry_list.empty() ||
       (cull_flags && (m_sphere.intersect(trav_state.getFrustum(),
                                          cull_flags) == sglBound::eOUTSIDE)))
   {
      return; // draw nothing
   }
   else
   {
      for (vector<sglDrawable *>::const_iterator elt=m_geometry_list.begin();
           elt != m_geometry_list.end();
           ++elt)
      {
         if ((*elt)->cull(trav_state.getFrustum(), cull_flags))
            trav_state.addDrawable(
               *elt, (*elt)->getStateByMask(trav_state.getStateMask()));
      }
   }
}


//----------------------------------------------------------------------------
sglNode *sglGeode::clone(unsigned int mode) const
{
   sglGeode *root = new sglGeode;
   sglGeode::copyTo(root, mode);
   return root;
}

//----------------------------------------------------------------------------
void sglGeode::copyTo(sglGeode *dst, unsigned int mode) const
{
   sglNode::copyTo(dst, mode);

   // transfer stuff
   if (mode & sglObject::eCLONE_NODES_RECURSIVE)
   {
      for (vector<sglDrawable*>::const_iterator elt=m_geometry_list.begin();
           elt != m_geometry_list.end();
           ++elt)
      {
         //if (mode & sglObject::eCLONE_GEOMETRY)
         //   dst->addGeometry((*elt)->clone(mode));
         //else
            dst->addGeometry(*elt);

            // HACK does not account for duplicated geometry
      }
   }
}

//----------------------------------------------------------------------------
void sglGeode::printInfo(ostream &ostrm, const char *indent_string) const
{
   sglNode::printInfo(ostrm, indent_string);

   ostrm << indent_string << "sglGeode, " << getNumGeometries()
         << " geometries" << endl;

   int i=0;
   for (vector<sglDrawable *>::const_iterator elt=m_geometry_list.begin();
        elt != m_geometry_list.end();
        ++elt, ++i)
   {
      ostrm << indent_string << "    geometry " << i << ": " << *elt << endl;
   }
}
