/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglGeode.hpp
 *  Created: 11 August 1997
 *  Summary: scenegraph leaf node to hold a set of sglGeoSet's.
 *****************************************************************************/

#ifndef __SGL_GEODE_HPP
#define __SGL_GEODE_HPP

#include <sgl.h>
#include <vector>

#include <sglNode.hpp>
#include <sglGroup.hpp>  // need this include for Windoze.

class sglDrawable;

//============================================================================

/**

The sglGeode node is a container for a set of 3D geometry classes.  It is also
considered a leaf node, even though it has children of a different sort:
sglDrawable and its subclasses that encapsulate the geometry and
graphics state information that specify the 3D objects that are rendered.

<b>Example:</b>

<pre>
   sglGeode       *geode = new sglGeode;

   // subclasses of sglDrawable:
   sglLineSet     *geom1 = new sglLineSet;
   sglTriangleSet *geom2 = new sglTriangleSet;

   geode->addGeometry(geom1);
   geode->addGeometry(geom2);

   unsigned int num = geode->getNumGeometries(); // returns 2
   sglDrawable* geom = geode->getGeometry(0);    // returns geom1
   geom = geode->getGeometry(2);                 // returns NULL

   geode->removeGeometry(geom2);                 // returns true;
   geode->insertGeometry(0, geom2);

   geom = geode->getGeometry(0);                 // returns geom2
</pre>
 */

class SGL_DLL_API sglGeode : public sglNode
{
public:
   /// default constructor
   sglGeode();
   /// virtual destructor
   virtual ~sglGeode();

   /** Get the number of sglDrawable nodes.
       @return The current number of sglDrawable objects contained in this
       class.
   */
   unsigned int getNumGeometries() const { return m_geometry_list.size(); }

   /** Add an sglDrawable to the end of the list.
       @param geometry Pointer to an sglDrawable object.
       @return true if geometry parameter is non-NULL.
   */
   bool addGeometry(sglDrawable *geometry);

   /** Insert an sglDrawable in a specific location of the list
       @param index position in vector to place the new geometry.
       @param geometry Pointer to the new geometry.
       @return true if geometry pointer is non-NULL, and the index is not out
         of range (greater than the number of geometries currently in the list.
   */
   bool insertGeometry(unsigned int index, sglDrawable *geometry);

   /** Remove an sglDrawable specified by its pointer from the list.
       @param geometry Pointer to the geometry to be removed.
       @return true if geometry parameter is non-NULL and the pointer is found
         in the current list (and is removed).
   */
   bool removeGeometry(sglDrawable *geometry);

   /** Get a pointer to a Geometry in the list at the position specified.
       @param index The position in the list (zero is the first position).
       @return Pointer the the geometry at that position, or NULL if index is
         out of range of the current index.
   */
   sglDrawable *getGeometry(unsigned int index);

   // See sglNode::intersect(f) for details.
   virtual sglBound::IntersectResultEnum intersect(
      sglIntersectf &isector) const;

   // See sglNode::intersect(d) for details.
   virtual sglBound::IntersectResultEnum intersect(
      sglIntersectd &segset) const;

   // See sglNode::pick(f) for details.
   virtual void pick(sglPickf &pick_state, unsigned int cull_flags) const;

   // See sglNode::pick(d) for details.
   virtual void pick(sglPickd &pick_state, unsigned int cull_flags) const;

   // See sglNode::preDraw() for details.
   virtual void preDraw(const PreDrawStruct &trav_state);

   // See sglNode::clone() for details.
   virtual sglNode *clone(unsigned int mode) const;

   // See sglObject::printInfo() for details.
   virtual void printInfo(ostream &ostrm, const char *indent_string) const;

protected:
   // compute the bounding sphere
   virtual void computeBound();
   virtual void computeBoundsRecursiveDown();

   // See sglNode::cull(f) for details
   virtual void cull(sglCull<float> &trav_state,
                     unsigned int cull_flags) const;

   // See sglNode::cull(d) for details
   virtual void cull(sglCull<double> &trav_state,
                     unsigned int cull_flags) const;

   void copyTo(sglGeode *dst, unsigned int mode) const;

private:
   int searchGeometries(sglDrawable *geometry);

   // not implemented
   sglGeode(const sglGeode &);
   sglGeode &operator=(const sglGeode &);

protected:
   vector<sglDrawable *> m_geometry_list;
};

#endif
