/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglGroup.cpp
 *  Created: 30 December 1997
 *  Summary: Base class for all internal scene graph node types
 *****************************************************************************/

#include <sgl.h>
#include <sglNode.hpp>
#include <sglGroup.hpp>
#include <sglCull.hpp>
#include <sglOStream.hpp>

#include <algorithm>

//----------------------------------------------------------------------------
sglGroup::sglGroup() : sglNode()
{
}

//----------------------------------------------------------------------------
sglGroup::~sglGroup()
{
   // pop any children off the list and decrement their references
   // and remove them if their references go to zero.
   while (m_child_list.size())
   {
      sglNode *child = m_child_list.back();
      m_child_list.pop_back();
      child->deref();

      child->removeParent(this);

      if (child->getRefCount() == 0)
      {
         delete child;
      }
   }
}

//----------------------------------------------------------------------------
bool sglGroup::addChild(sglNode *child)
{
   if (child)
   {
      if (searchChild(child) == -1)
      {
         child->ref();
         child->addParent(this);
      }
      m_child_list.push_back(child);
      return true;
   }
   else
   {
      sglPrintWarning()
         << "sglGroup::addChild error: trying to add a NULL child.\n";
      return false;
   }
}

//----------------------------------------------------------------------------
bool sglGroup::removeChild(sglNode *child)
{
   bool success = false;
   if (child)
   {
      int index = m_child_list.size()-1;
      while(index >= 0 && m_child_list[index] != child)
         --index;

      if (index >= 0)
      {
         // index was found via backward search, searchChild uses a forward
         // search; if they match, then child is only in the list once
         if (searchChild(child) == index)
         {
            child->removeParent(this);
            child->deref();
         }

         // ----------------------------------------------------------
         // NOTE: m_child_list[index]  = *(m_child_list.begin()+index)
         // ----------------------------------------------------------
         m_child_list.erase(m_child_list.begin()+index);

         success = true;
      }
      else
      {
         sglPrintWarning() << "sglGroup::removeChild error: "
                           << "could not find referenced child.\n";
      }
   }
   else
   {
      sglPrintWarning() << "sglGroup::removeChild error: "
                        << "trying to remove a NULL child.\n";
   }
   return success;
}

//----------------------------------------------------------------------------
bool sglGroup::insertChild(unsigned int index, sglNode *child)
{
   bool success = false;
   if (child)
   {
      if (index < getNumChildren())
      {
         if (searchChild(child) == -1)
         {
            child->ref();
            child->addParent(this);
         }
         m_child_list.insert(m_child_list.begin()+index, child);
         success = true;
      }
      else if (index == getNumChildren())
      {
         if (searchChild(child) == -1)
         {
            child->ref();
            child->addParent(this);
         }
         m_child_list.push_back(child);
         success = true;
      }
      else
      {
         sglPrintWarning() << "sglGroup::insertChild error: "
              << "trying to insert with invalid index = " << index << endl;
      }
   }
   else
   {
      sglPrintWarning() << "sglGroup::insertChild error: "
           << "trying to insert a NULL child.\n";
   }

   return success;
}

//----------------------------------------------------------------------------
sglNode* sglGroup::replaceChild(unsigned int index, sglNode *child)
{
   sglNode *old = NULL;

   if (child)
   {
      if (index < getNumChildren())
      {
         old = m_child_list[index];
         if (old != child)
         {
            if (searchChild(child) == -1)
            {
               child->ref();
               child->addParent(this);
            }
            m_child_list[index] = child;

            if (searchChild(old) == -1)
            {
               old->removeParent(this);
               old->deref();
            }
         }
      }
      else
      {
         sglPrintWarning()
            << "sglGroup::replaceChild error: "
            << "trying to replace with invalid index = " << index << endl;
      }
   }
   else
   {
      sglPrintWarning() << "sglGroup::replaceChild error: "
                        << "trying to replace a child with a NULL child.\n";
   }

   return old;
}

//----------------------------------------------------------------------------
sglNode* sglGroup::removeChild(unsigned int index)
{
   sglNode *child = NULL;

   if (index < getNumChildren())
   {
      // ----------------------------------------------------------
      // NOTE: m_child_list[index]  = *(m_child_list.begin()+index)
      // ----------------------------------------------------------
      child = m_child_list[index];
      m_child_list.erase(m_child_list.begin()+index);

      if (searchChild(child) == -1)
      {
         child->removeParent(this);
         child->deref();
      }
   }
   else
   {
      sglPrintWarning() << "sglGroup::replaceChild error: "
                        << "trying to replace with invalid index = " << index
                        << endl;
   }

   return child;
}

//----------------------------------------------------------------------------
sglNode* sglGroup::getChild(unsigned int index)
{
   sglNode *child = NULL;
   if (index < getNumChildren())
   {
      child = m_child_list[index];
   }

   return child;
}

//----------------------------------------------------------------------------
int sglGroup::searchChild(sglNode *child, unsigned int start) const
{
   if (start >= m_child_list.size() || child == NULL) return -1;

   vector<sglNode*>::const_iterator it =
      find(m_child_list.begin()+start, m_child_list.end(), child);

   if (it == m_child_list.end())
      return -1;
   else
      return (it - m_child_list.begin());
}

//----------------------------------------------------------------------------
void sglGroup::computeBound()
{
   m_dirty &= ~eBOUND;

   if (getNumChildren())
   {
      // initialize with first child
      m_sphere = m_child_list[0]->m_sphere;

      // don't clear this dirty bit if child's bit is dirty; can happen during
      // a getBound() query.
      m_dirty |= ((m_child_list[0]->m_dirty)&eBOUND);

      // how many children
      for (vector<sglNode*>::iterator elt=m_child_list.begin() + 1;
           elt != m_child_list.end();
           ++elt)
      {
         m_sphere.expandToInclude((*elt)->m_sphere);

         // don't clear this dirty bit if child's bit is dirty; can happen
         // during a getBound() query.
         m_dirty |= (((*elt)->m_dirty)&eBOUND);
      }
   }
   else
   {
      // mark sphere as empty
      m_sphere.makeEmpty();
   }
}

//----------------------------------------------------------------------------
void sglGroup::computeBoundsRecursiveDown()
{
   if ((m_dirty&eBOUND) && m_bound_mode != sglBound::eSTATIC)
   {
      for (vector<sglNode*>::iterator ii = m_child_list.begin();
           ii != m_child_list.end(); ++ii)
      {
         if ((*ii)->m_dirty&eBOUND)
            (*ii)->computeBoundsRecursiveDown();
      }
      computeBound();
   }
}

//----------------------------------------------------------------------------
sglBound::IntersectResultEnum sglGroup::intersect(sglIntersectf &segset) const
{
   if (m_child_list.empty())
   {
      return sglBound::eOUTSIDE;
   }

   sglSegmentf seg(segset.getSegment());
   sglBound::IntersectResultEnum result = m_sphere.intersect(seg);
   if (result == sglBound::eOUTSIDE)
   {
      return result;
   }

   result = sglBound::eOUTSIDE;
   for (vector<sglNode*>::const_iterator elt = m_child_list.begin();
        elt != m_child_list.end(); ++elt)
   {
      if ((*elt)->intersect(segset) == sglBound::eINTERSECT)
      {
         result = sglBound::eINTERSECT;
      }
   }

   if (result == sglBound::eINTERSECT &&
       segset.getMode() & sglProbe::ePATH)
   {
      segset.getPath().push_front((sglNode*)this);
   }
   return result;
}

//----------------------------------------------------------------------------
sglBound::IntersectResultEnum sglGroup::intersect(sglIntersectd &segset) const
{
   if (m_child_list.empty())
   {
      return sglBound::eOUTSIDE;
   }

   sglSegmentd seg(segset.getSegment());
   sglBound::IntersectResultEnum result = m_sphere.intersect(seg);
   if (result == sglBound::eOUTSIDE)
   {
      return result;
   }

   result = sglBound::eOUTSIDE;
   for (vector<sglNode*>::const_iterator elt = m_child_list.begin();
        elt != m_child_list.end(); ++elt)
   {
      if ((*elt)->intersect(segset) == sglBound::eINTERSECT)
      {
         result = sglBound::eINTERSECT;
      }
   }

   if (result == sglBound::eINTERSECT &&
       segset.getMode() & sglProbe::ePATH)
   {
      segset.getPath().push_front((sglNode*)this);
   }
   return result;
}

//----------------------------------------------------------------------------
void sglGroup::pick(sglPickf &pick_state, unsigned int cull_flags) const
{
   if (m_child_list.empty() ||
       (cull_flags && (m_sphere.intersect(pick_state.getPickFrustum(),
                                          cull_flags) == sglBound::eOUTSIDE)))
   {
      return; // select nothing
   }

   // if path is requested store it ahead of time (different from intersect),
   // in case something beneath here is picked successfully.
   if (pick_state.getMode() & sglProbe::ePATH)
   {
      pick_state.getPath().push_back((sglNode*)this);
   }

   for (vector<sglNode*>::const_iterator elt = m_child_list.begin();
        elt != m_child_list.end(); ++elt)
   {
      (*elt)->pick(pick_state, cull_flags);
   }

   if (pick_state.getMode() & sglProbe::ePATH)
   {
      pick_state.getPath().pop_back();
   }
}

//----------------------------------------------------------------------------
void sglGroup::pick(sglPickd &pick_state, unsigned int cull_flags) const
{
   if (m_child_list.empty() ||
       (cull_flags && (m_sphere.intersect(pick_state.getPickFrustum(),
                                          cull_flags) == sglBound::eOUTSIDE)))
   {
      return; // select nothing
   }

   // if path is requested store it ahead of time (different from intersect),
   // in case something beneath here is picked successfully.
   if (pick_state.getMode() & sglProbe::ePATH)
   {
      pick_state.getPath().push_back((sglNode*)this);
   }

   for (vector<sglNode*>::const_iterator elt = m_child_list.begin();
        elt != m_child_list.end(); ++elt)
   {
      (*elt)->pick(pick_state, cull_flags);
   }

   if (pick_state.getMode() & sglProbe::ePATH)
   {
      pick_state.getPath().pop_back();
   }
}

//----------------------------------------------------------------------------
void sglGroup::preDraw(const PreDrawStruct &trav_state)
{
   for (vector<sglNode*>::iterator ii = m_child_list.begin();
        ii != m_child_list.end(); ++ii)
   {
      if ((*ii)->m_dirty)
         (*ii)->preDraw(trav_state);
   }

   sglNode::preDraw(trav_state);
}

//----------------------------------------------------------------------------
void sglGroup::cull(sglCull<float> &trav_state,
                    unsigned int cull_flags) const
{
   if (m_child_list.empty() ||
       (cull_flags && (m_sphere.intersect(trav_state.getFrustum(),
                                          cull_flags) == sglBound::eOUTSIDE)))
   {
      return; // draw nothing
   }

   for (vector<sglNode*>::const_iterator elt = m_child_list.begin();
        elt != m_child_list.end(); ++elt)
   {
      (*elt)->cull(trav_state, cull_flags);
   }
}

//----------------------------------------------------------------------------
void sglGroup::cull(sglCull<double> &trav_state,
                    unsigned int cull_flags) const
{
   if (m_child_list.empty() ||
       (cull_flags && (m_sphere.intersect(trav_state.getFrustum(),
                                          cull_flags) == sglBound::eOUTSIDE)))
   {
      return; // draw nothing
   }

   for (vector<sglNode*>::const_iterator elt = m_child_list.begin();
        elt != m_child_list.end(); ++elt)
   {
      (*elt)->cull(trav_state, cull_flags);
   }
}

//----------------------------------------------------------------------------
sglNode *sglGroup::clone(unsigned int mode) const
{
   sglGroup *root = new sglGroup;
   sglGroup::copyTo(root, mode);

   return root;
}

//----------------------------------------------------------------------------
void sglGroup::copyTo(sglGroup *dst, unsigned int mode) const
{
   sglNode::copyTo(dst, mode);

   // transfer stuff
   if (mode & sglObject::eCLONE_NODES_RECURSIVE)
   {
      int index = 0;
      for (vector<sglNode*>::const_iterator elt = m_child_list.begin();
           elt != m_child_list.end();
           ++elt, ++index)
      {
         int new_index;
         // account for duplicated children properly
         if ((new_index = searchChild(*elt)) != index)
         {
            if (new_index >= 0)
               dst->addChild(dst->getChild((unsigned int)new_index));
         }
         else
            dst->addChild((*elt)->clone(mode));
      }
   }
}

//----------------------------------------------------------------------------
void sglGroup::printInfo(ostream &ostrm, const char *indent_string) const
{
   sglNode::printInfo(ostrm, indent_string);

   ostrm << indent_string << "sglGroup: " << getNumChildren() << " children"
         << endl;
   ostrm << indent_string << "    child_list capacity = "
         << m_child_list.capacity() << endl;

   unsigned int i=0;
   for (vector<sglNode*>::const_iterator elt=m_child_list.begin();
        elt != m_child_list.end();
        ++elt, ++i)
   {
      ostrm << indent_string << "    child " << i << ": " << *elt
            << ", name = " << (*elt)->getName() << endl;
   }
}
