/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglGroup.hpp
 *  Created: 30 December 1997
 *  Summary: Base class for all scene graph internal node types
 *****************************************************************************/

#ifndef __SGL_GROUP_HPP
#define __SGL_GROUP_HPP

#include <sgl.h>
#include <vector>

#include <sglPolytope.hpp>
#include <sglNode.hpp>

//----------------------------------------------------------------------------

/**

The scene graph is a directed, acyclic graph (DAG).  As such, interior nodes of
the scene graph can have multiple children.  The sglGroup node is the base
class (concrete) for all such nodes.  It has the ability of managing
a list of child nodes (any subclass of sglNode).  There are functions
to get the number of children currently attached to this node, add new
children to the end of the list, insert children at a desired index, remove
children, get the nth child, and search for a given child.  For the
sake of subclasses like sglSwitch and sglLOD children can be added to the
list more than once.  The user may not add a NULL as a child, however.

When the traversal funtions get to an sglGroup node it will visit all of the
children.  Note that just because the child list is internally stored as an
ordered list of sglNode's, this order does not in any way determine the order
in which children get traversed or rendered (unless a subclass dictates
otherwise.

<b>Example:</b>

<pre>
   sglGroup *parent = new sglGroup;
   sglGroup *child1 = new sglGroup;
   unsigned int num;

   parent->addChild(child1);
   num = parent->getRefCount();           // returns 0
   num = parent->getNumChildren();        // returns 1
   num = child1->getRefCount();           // returns 1
   num = child1->getNumParents();         // returns 1

   parent->remove(child1);
   num = parent->getNumChildren();        // returns 0
   num = child1->getRefCount();           // returns 0
   num = child1->getNumParents();         // returns 0
</pre>

 */

//----------------------------------------------------------------------------

class SGL_DLL_API sglGroup : public sglNode
{
public:
   /// default constructor
   sglGroup();
   /// destructor
   virtual ~sglGroup();

   /** Get the number of child nodes in this group.
       @return number of children
   */
   unsigned int getNumChildren() const { return m_child_list.size(); }

   /** Add child to the end of the list.
       @param child Pointer to node to be added.
       @return false if child pointer is NULL; otherwise, true.
   */
   virtual bool addChild(sglNode *child);

   /** Remove the last instance of the child from the list
       @param child Pointer to the node to be removed.
       @return false if child pointer is NULL or not found in child list;
       otherwise, true.
   */
   virtual bool removeChild(sglNode *child);

   /** Insert a child at a specific index. Valid indices are in the range
       [0,getNumChildren()].
       @param index Position at which node is to be inserted.
       @param child Pointer to node to be inserted.
       @return false if child is NULL or index is out of range; otherwise,
       true.
   */
   virtual bool insertChild(unsigned int index, sglNode *child);

   /** Replace child at given index where the valid range is
       [0,getNumChildren()-1].
       @param index Position at which node is to be placed.
       @param child Pointer to node to be added.
       @return Pointer to the child previously at position index or NULL on
       error (NULL child pointer, index out of range).
   */
   virtual sglNode *replaceChild(unsigned int index, sglNode *child);

   /** Removes child at position index (valid indices:
       [0,getNumChildren()-1]).  Nodes after this in the list then shift down
       one index.
       @param index Index of child to be removed.
       @return The child previously at position index, or NULL on error.
   */
   virtual sglNode *removeChild(unsigned int index);

   /** Get a pointer to the child at the given index (valid indices:
       [0,getNumChildren()-1]).
       @param index Index of the desired child.
       @return The child at position index or NULL on error (if index is out of
       range.
    */
   sglNode* getChild(unsigned int index);

   /** Find the index of the first occurrence of a specified child at or after
       a particular index.
       @param child Pointer of child to search for.
       @param start The start index of the search.
       @return Index of the first child found, or -1 on error (child is NULL or
       start is >= getNumChildren()).
   */
   int searchChild(sglNode *child, unsigned int start = 0) const;

   // See sglNode::intersect(f) for details.
   virtual sglBound::IntersectResultEnum intersect(
      sglIntersectf &isector) const;

   // See sglNode::intersect(d) for details.
   virtual sglBound::IntersectResultEnum intersect(
      sglIntersectd &isector) const;

   // See sglNode::pick(f) for details.
   virtual void pick(sglPickf &pick_state, unsigned int cull_flags) const;

   // See sglNode::pick(d) for details.
   virtual void pick(sglPickd &pick_state, unsigned int cull_flags) const;

   // See sglNode::preDraw() for details.
   virtual void preDraw(const PreDrawStruct &trav_state);

   // See sglNode::clone() for details.
   virtual sglNode *clone(unsigned int mode) const;

   // See sglObject::printInfo() for details.
   virtual void printInfo(ostream &ostrm, const char *indent_string) const;

protected:
   // compute the bounding sphere
   virtual void computeBound();
   virtual void computeBoundsRecursiveDown();

   // See sglNode::cull(f) for details
   virtual void cull(sglCull<float> &trav_state,
                     unsigned int cull_flags) const;

   // See sglNode::cull(d) for details
   virtual void cull(sglCull<double> &trav_state,
                     unsigned int cull_flags) const;

   void copyTo(sglGroup *dst, unsigned int mode) const;

private:  // not implemented
   sglGroup(const sglGroup &);
   sglGroup &operator=(const sglGroup &);

protected:
    vector<sglNode*> m_child_list;
};

#endif
