/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglIndexedBitMaskSwitch.hpp
 *   Author: Michael Morrison
 *  Created: October 2001
 *  Summary: select (or not) the children to be traversed
 *****************************************************************************/

#ifndef __SGL_INDEXED_BIT_MASK_SWITCH_HPP
#define __SGL_INDEXED_BIT_MASK_SWITCH_HPP

#include <sgl.h>
#include <sglGroup.hpp>
#include <sglBitMaskSwitch.hpp>
#include <vector>

//----------------------------------------------------------------------------

/**

This is a subclass of sglBitMaskSwitch that adds a level of indirection to the
selection mechanism by associating an index with each set of up to 32 children
a nearly unlimited number of combinations of these 32 children can be selected
with an integer index.  The index is constructed as a vector of bit masks to be
used (see sglBitMaskSwitch for full details on what the bitmasks mean).

As with sglBitMaskSwitch, the order that the children are added to the node is
significant.  In addition, the order the masks are added is significant, as it
implicitly selects the index value for that bitmask.

Initially, the default index value and mask is undefined.  At least one mask
should be added using addMask(), and setIndex() should be used to set the
default index.  Otherwise, the default set of children selected will be the
same as the default of the sglBitMaskSwitch, which is to select all children.

<b>Special behaviour for intersect and pick traversals</b>

This behaviour is the same as for the sglBitMaskSwitch.

<b>Example:</b>

<pre>
   sglIndexedBitMaskSwitch *switch_node = new sglIndexedBitMaskSwitch;
   sglGeode *geode0 = new sglGeode;
   sglGeode *geode1 = new sglGeode;

   switch_node->addChild(geode0);
   switch_node->addChild(geode1);

   switch_node->addMask(0x0);     // select neither for index 0
   switch_node->addMask(0x2);     // select geode1 for index 1
   switch_node->addMask(0x1);     // select geode0 for index 2
   switch_node->addMask(0x3);     // select both for index 3
   switch_node->addMask(~0x0);    // select all children for index 4
   switch_node->setIndex(2);      // select zeroeth index (geode 0)
</pre>

\todo Do we need to add complete mask management (remove, insert, etc)?

 */

//----------------------------------------------------------------------------

class SGL_DLL_API sglIndexedBitMaskSwitch : public sglBitMaskSwitch
{
public:
   /// default constructor
   sglIndexedBitMaskSwitch();
   /// destructor
   virtual ~sglIndexedBitMaskSwitch();

   /** Set the index for selecting children.
       @index The index into the vector of masks.
   */
   void setIndex(unsigned int index)
   {
      if (index >= m_index.size())
         return;

      m_switch_index = index;
      setMask( m_index[ m_switch_index ] );
   }

   /** Get the current index.
       @returns The current index value (unsigned int).
   */
   unsigned int getIndex() const { return m_switch_index; }

   /** Add a new mask to the index vector.
       @mask The index into the vector of masks.
   */
   void addMask( unsigned int mask) { m_index.push_back( mask ); }

   // See sglNode::clone() for details.
   virtual sglNode *clone(unsigned int mode) const;

   // See sglObject::printInfo() for details.
   virtual void printInfo(ostream &ostrm, const char *indent_string) const;

protected:
   void copyTo(sglIndexedBitMaskSwitch *dst, unsigned int mode) const;

private:
   sglIndexedBitMaskSwitch(const sglIndexedBitMaskSwitch &);     // not imp
   sglIndexedBitMaskSwitch &operator=(const sglIndexedBitMaskSwitch &);

protected:
   unsigned int m_switch_index;
   vector<unsigned int> m_index;
};

#endif
