/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglIndexedGeoSet.hpp
 *  Created: 12 August 1997
 *  Summary: Approximates a pfGeoSet which contains geometry data and index
 *         : arrays for all indexed geosets, and pointers to the graphics state
 *         : associated with it (sglGeoState).
 *****************************************************************************/

#ifndef __SGL_INDEXED_GEOSET_HPP
#define __SGL_INDEXED_GEOSET_HPP

#include <sgl.h>
#include <sglGeoSet.hpp>

//============================================================================

/**

sglIndexedGeoSet is the abstract base class for all indexed geometry classes.
Sometimes the vertices in the geometry is used more than once.  In the concrete
subclasses of sglGeoSet and sglGeoStripSet, these vertices would have to be
present in the array once for each time the primitives reference it.  This can
lead to higher storage requirements.  This can potentially be reduced in the
sglIndexedGeoSet class (and its subclasses) by using indexing into the
attribute arrays.

Note that if one attribute array (coord, normal, color, and texcoord) is
indexed, then all of them have to be indexed.  The exception to this rule is if
the color or normal attributes are bound as sglGeoSet::eOVERALL (that is, no
index array is needed in this case).  Each attribute can have its own index
array or share them in any combination.

The index arrays consist of unsigned ints.  Should the indexed arrays be
unsigned shorts to reduce storage requirements? The original need for index
arrays that could index into attribute arrays larger than 2^16 may no longer be
there.

{\bf Concrete subclasses: sglIndexedPointSet, sglIndexedLineSet,
sglIndexedTriangleSet, sglIndexedQuadSet}

These are the concrete subclasses of sglIndexedGeoSet that implement the
point, line segment, triangle and quad primitives.

{\bf Example:}

\begin{verbatim}
   sglVec3f *vertex = new sglVec3f[6];
   vertex[0].set( 0.0f, 0.0f, 0.0f);
   vertex[1].set( 0.0f, 1.0f, 0.0f);
   vertex[2].set( 1.0f, 0.0f, 0.0f);
   vertex[3].set( 1.0f,-1.0f, 0.0f);
   vertex[4].set(-1.0f,-1.0f, 0.0f);
   vertex[5].set(-1.0f, 0.0f, 0.0f);

   sglVec4f *color = new sglVec4f[3];
   color[0].set( 1.0f, 0.0f, 0.0f, 1.0f);
   color[1].set( 0.0f, 1.0f, 0.0f, 1.0f);
   color[2].set( 0.0f, 0.0f, 1.0f, 1.0f);

   unsigned int *vindex = new unsigned int[10];
   vindex[0] = 0; vindex[1] = 1;   // indices for a five pointed star (sorta)
   vindex[2] = 0; vindex[3] = 2;
   vindex[4] = 0; vindex[5] = 3;
   vindex[6] = 0; vindex[7] = 4;
   vindex[8] = 0; vindex[9] = 5;

   unsigned int *cindex = new unsigned int[5];
   cindex[0] = 0;
   cindex[1] = 1;
   cindex[2] = 2;
   cindex[3] = 2;
   cindex[4] = 1;

   sglIndexedLineSet *line_gset = new sglIndexedLineSet;
   line_gset->setNumPrims(5);
   line_gset->setCoordList(vertex);
   line_gset->setCoordIndexList(vindex);
   line_gset->setColorList(sglGeoSet::ePER_PRIM, color);
   line_gset->setColorIndexList(cindex);
\end{verbatim} */

class SGL_DLL_API sglIndexedGeoSet : public sglGeoSet
{
public:
   /// default constructor
   sglIndexedGeoSet();
   /// virtual destructor
   virtual ~sglIndexedGeoSet();

   /// Set the array of vertex indices.  Deprecated.
   void setCoordIndexList(sglIndexType *alist)
      { setCoordIndexList(new sglNoRefDataPtr(alist)); }

   /** Set the array of vertex indices.
       This will increment the reference count
       on the given data pointer, and decrement the reference count of
       the previous data pointer.  If the previous data pointer's refernce
       count goes to zero, it will be deleted.
       The indices must be of type sglIndexType.
       @param data Pointer to the list of indices.
       @param offset Number of bytes in the array before the first index.
       @param stride Number of bytes from the start of one index to the start
                     of the next index.  If zero, the indices are assumed
                     to be tightly packed (i.e. stride == sizeof(type)).
                     Stride must be non-negative.
   */
   void setCoordIndexList(sglDataPtr *data,
                          int offset = 0,
                          int stride = 0);

   /// Set the array of normal indices.  Deprecated.
   void setNormalIndexList(sglIndexType *alist)
      { setNormalIndexList(new sglNoRefDataPtr(alist)); }

   /** Set the array of vertex indices.
       This will increment the reference count
       on the given data pointer, and decrement the reference count of
       the previous data pointer.  If the previous data pointer's refernce
       count goes to zero, it will be deleted.
       The indices must be of type sglIndexType.
       @param data Pointer to the list of indices.
       @param offset Number of bytes in the array before the first index.
       @param stride Number of bytes from the start of one index to the start
                     of the next index.  If zero, the indices are assumed
                     to be tightly packed (i.e. stride == sizeof(type)).
                     Stride must be non-negative.
   */
   void setNormalIndexList(sglDataPtr *data,
                           int offset = 0,
                           int stride = 0);

   /// Set the array of texture coordinate indices.  Deprecated.
   void setTexCoordIndexList(sglIndexType *alist)
      { setTexCoordIndexList(new sglNoRefDataPtr(alist)); }

   /** Set the array of texture coordinate indices.
       This will increment the reference count
       on the given data pointer, and decrement the reference count of
       the previous data pointer.  If the previous data pointer's refernce
       count goes to zero, it will be deleted.
       The indices must be of type sglIndexType.
       @param data Pointer to the list of indices.
       @param offset Number of bytes in the array before the first index.
       @param stride Number of bytes from the start of one index to the start
                     of the next index.  If zero, the indices are assumed
                     to be tightly packed (i.e. stride == sizeof(type)).
                     Stride must be non-negative.
   */
   void setTexCoordIndexList(sglDataPtr *data,
                             int offset = 0,
                             int stride = 0);

   ///Set the array of color indices.  Deprecated.
   void setColorIndexList(sglIndexType *alist)
      { setColorIndexList(new sglNoRefDataPtr(alist)); }

   /** Set the array of color indices.
       This will increment the reference count
       on the given data pointer, and decrement the reference count of
       the previous data pointer.  If the previous data pointer's refernce
       count goes to zero, it will be deleted.
       The indices must be of type sglIndexType.
       @param data Pointer to the list of indices.
       @param offset Number of bytes in the array before the first index.
       @param stride Number of bytes from the start of one index to the start
                     of the next index.  If zero, the indices are assumed
                     to be tightly packed (i.e. stride == sizeof(type)).
                     Stride must be non-negative.
   */
   void setColorIndexList(sglDataPtr *data,
                          int offset = 0,
                          int stride = 0);

   /// get the current vertex, normal, color, texcoord attribute lists
   sglDataPtr *getCoordIndexList() const {return m_vindex.getDataPtr();}
   ///
   sglDataPtr *getNormalIndexList() const {return m_nindex.getDataPtr();}
   ///
   sglDataPtr *getTexCoordIndexList() const {return m_tindex.getDataPtr();}
   ///
   sglDataPtr *getColorIndexList() const {return m_cindex.getDataPtr();}

   /// Get the number of bytes in the array before the first index.
   int getCoordIndexOffset() const { return m_vindex.getOffset(); }
   ///
   int getNormalIndexOffset() const { return m_nindex.getOffset(); }
   ///
   int getTexCoordIndexOffset() const { return m_tindex.getOffset(); }
   ///
   int getColorIndexOffset() const { return m_cindex.getOffset(); }

   /// Get the number of bytes from the start of one index to the next.
   int getCoordIndexStride() const { return m_vindex.getStride(); }
   ///
   int getNormalIndexStride() const { return m_nindex.getStride(); }
   ///
   int getTexCoordIndexStride() const { return m_tindex.getStride(); }
   ///
   int getColorIndexStride() const { return m_cindex.getStride(); }

protected:
   virtual bool computeBounds();

private:  // not implemented
   sglIndexedGeoSet(const sglIndexedGeoSet &);
   sglIndexedGeoSet &operator=(const sglIndexedGeoSet &);

protected:
   sglIndexPtr m_vindex; // List of coordinate indexes
   sglIndexPtr m_nindex; // List of normal indexes
   sglIndexPtr m_tindex; // List of texture coordinate indexes
   sglIndexPtr m_cindex; // List of color indexes
};

#endif
