/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglIndexedGeoStripSet.hpp
 *  Created: 28 August 1999
 *  Summary: GeoSet where vertex data is indexed via a single index list.
 *****************************************************************************/

#ifndef __SGL_INDEXED_GEO_STRIP_SET_HPP
#define __SGL_INDEXED_GEO_STRIP_SET_HPP

#include <sgl.h>
#include <sglIndexedGeoSet.hpp>

//============================================================================

/**

Just as sglGeoStripSet specializes the sglGeoSet class to add a lengths array
for specifying the size of variable length primitives, this class specializes
the sglIndexedGeoSet to add the same lengths array for the indexed versions of
the primitive types.

{\bf Concrete subclasses:
sglIndexedLineStripSet, sglIndexedTriangleFanSet, sglIndexedTriangleStripSet,
sglIndexedQuadStripSet, sglIndexedPolygonSet}

These are the concrete subclasses of sglIndexedGeoStripSet that implement
the line strip, triangle fan, triangle strip, quad strip and polygon
primitives.

{\bf Example:}
\begin{verbatim}
   unsigned int *lengths = new unsigned int[1];
   lengths[0] = 5;

   sglVec3f *vertex = new sglVec3f[5];
   vertex[0].set( 0.0f, 1.0f, 0.0f);
   vertex[1].set( 1.0f, 0.0f, 0.0f);
   vertex[2].set( 1.0f,-1.0f, 0.0f);
   vertex[3].set(-1.0f,-1.0f, 0.0f);
   vertex[4].set(-1.0f, 0.0f, 0.0f);

   sglVec4f *color = new sglVec4f[3];
   color[0].set( 1.0f, 0.0f, 0.0f, 1.0f);
   color[1].set( 0.0f, 1.0f, 0.0f, 1.0f);

   unsigned int *vindex = new unsigned int[5];
   vindex[0] = 0;          // indices for a pentagon (irregular)
   vindex[1] = 1;
   vindex[2] = 2;
   vindex[3] = 3;
   vindex[4] = 4;

   unsigned int *cindex = new unsigned int[5];
   cindex[0] = 0;
   cindex[1] = 1;
   cindex[2] = 0;
   cindex[3] = 1;
   cindex[4] = 0;

   sglIndexedPolygonSet *poly_set = new sglIndexedPolygonSet;
   poly_set->setNumPrims(1);
   poly_set->setPrimLengths(lengths);
   poly_set->setCoordList(vertex);
   poly_set->setCoordIndexList(vindex);
   poly_set->setColorList(sglGeoSet::ePER_VERTEX, color);
   poly_set->setColorIndexList(cindex);
\end{verbatim}
 */

class SGL_DLL_API sglIndexedGeoStripSet : public sglIndexedGeoSet
{
public:
   /// default constructor
   sglIndexedGeoStripSet();
   /// virtual destructor
   virtual ~sglIndexedGeoStripSet();

   /// Not very useful for variable sized primitives.
   virtual unsigned int getPrimSize() const {return 0;}

   /// Get the lengths array.
   sglDataPtr *getPrimLengths() const {return m_lengths.getDataPtr();}

   /// Set the array of lengths.  Deprecated.
   void setPrimLengths(sglIndexType *alist)
      { setPrimLengths(new sglNoRefDataPtr(alist)); }

   /** Set the array of lengths.  This will increment the reference count
       on the given data pointer, and decrement the reference count of
       the previous data pointer.  If the previous data pointer's refernce
       count goes to zero, it will be deleted.
       The lengths must be of type sglIndexType.
       @param data Pointer to the list of lengths.
       @param offset Number of bytes in the array before the first length.
       @param stride Number of bytes from the start of one length to the start
                     of the next length.  If zero, the lengths are assumed
                     to be tightly packed (i.e. stride == sizeof(type)).
                     Stride must be non-negative.
   */
   void setPrimLengths(sglDataPtr *data,
                       int offset = 0,
                       int stride = 0);

   /// Get the number of bytes in the array before the first length.
   int getPrimLengthsOffset() const { return m_lengths.getOffset(); }

   /// Get the number of bytes from the start of one length to the next.
   int getPrimLengthsStride() const { return m_lengths.getStride(); }

   /// Add this drawable's stats to the given sglStats.
   virtual void addStats(sglStats &stats) const;

protected:
   virtual bool computeBounds();
   virtual void updateDrawFunc();

private: // not implemented
   sglIndexedGeoStripSet(const sglIndexedGeoStripSet &);
   sglIndexedGeoStripSet &operator=(const sglIndexedGeoStripSet &);

   void drawVT(const vector<sglTexCoords> &tex_coords) const;
   void drawVNoT(const vector<sglTexCoords> &tex_coords) const;
   void drawVNpT(const vector<sglTexCoords> &tex_coords) const;
   void drawVNvT(const vector<sglTexCoords> &tex_coords) const;
   void drawVCoT(const vector<sglTexCoords> &tex_coords) const;
   void drawVCoNoT(const vector<sglTexCoords> &tex_coords) const;
   void drawVCoNpT(const vector<sglTexCoords> &tex_coords) const;
   void drawVCoNvT(const vector<sglTexCoords> &tex_coords) const;
   void drawVCpT(const vector<sglTexCoords> &tex_coords) const;
   void drawVCpNoT(const vector<sglTexCoords> &tex_coords) const;
   void drawVCpNpT(const vector<sglTexCoords> &tex_coords) const;
   void drawVCpNvT(const vector<sglTexCoords> &tex_coords) const;
   void drawVCvT(const vector<sglTexCoords> &tex_coords) const;
   void drawVCvNoT(const vector<sglTexCoords> &tex_coords) const;
   void drawVCvNpT(const vector<sglTexCoords> &tex_coords) const;
   void drawVCvNvT(const vector<sglTexCoords> &tex_coords) const;

   class Initializer;
   friend class Initializer;
   static void setupFuncTable();

protected:
   sglIndexPtr m_lengths; // List of strip lengths

private:
   static DrawFunc s_func_table[(eCOLORS|eNORMALS)+1];
   static Initializer sInitializer;
};

#endif
