/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglInterleavedGeometry.cpp
 *  Created: 21 March 2000
 *  Summary: implementing the interleaved vertex arrays
 *****************************************************************************/

#include <sglInterleavedGeometry.hpp>
#include <sglOStream.hpp>

//----------------------------------------------------------------------------
sglInterleavedGeometry::sglInterleavedGeometry()
      : sglDrawable(),
        m_prim_type(GL_POINT),
        m_num_prims(0),
        m_lengths(NULL),
        m_array(NULL),
        m_array_format(GL_V3F),
        m_array_stride(0),
        m_compute_bounds(0)
{
   m_draw_func = reinterpret_cast<DrawFunc>(
      &sglInterleavedGeometry::drawGeometry);
}

//----------------------------------------------------------------------------
sglInterleavedGeometry::~sglInterleavedGeometry()
{
}

//----------------------------------------------------------------------------
sglDrawable::AAEnum sglInterleavedGeometry::getAAType() const
{
   sglDrawable::AAEnum aa_type;

   switch (m_prim_type)
   {
      case GL_POINTS:
         aa_type = ePOINT_AA;
         break;
      case GL_LINES:
      case GL_LINE_STRIP:
      case GL_LINE_LOOP:
         aa_type = eLINE_AA;
         break;
      case GL_TRIANGLES:
      case GL_TRIANGLE_STRIP:
      case GL_TRIANGLE_FAN:
      case GL_QUADS:
      case GL_QUAD_STRIP:
      case GL_POLYGON:
         aa_type = ePOLYGON_AA;
         break;
      default:
         sglPrintWarning() << "sglInterleavedGeometry::getAAType error: "
                           << "unknown geometry type = 0x" << hex
                           << (int)m_prim_type << dec << endl;
         aa_type = ePOINT_AA;
         break;
   }

   return aa_type;
}

//----------------------------------------------------------------------------
void sglInterleavedGeometry::setGLPrimType(GLenum prim_type)
{
   switch (prim_type)
   {
      case GL_POINTS:
      case GL_LINES:
      case GL_LINE_STRIP:
      case GL_LINE_LOOP:
      case GL_TRIANGLES:
      case GL_TRIANGLE_STRIP:
      case GL_TRIANGLE_FAN:
      case GL_QUADS:
      case GL_QUAD_STRIP:
      case GL_POLYGON:
         m_prim_type = prim_type;
         m_compute_bounds |= ePRIM_TYPE;
         dirty(eBOUND|eDLIST, true);        // ???
         break;

      default:
         m_compute_bounds &= ~ePRIM_TYPE;
         sglPrintWarning() << "sglInterleavedGeometry::setGLPrimType error:"
                           << " unknown geometry type = 0x" << hex
                           << (int)prim_type << dec << endl;
         break;
   }
}

//----------------------------------------------------------------------------
sglStats::PrimEnum sglInterleavedGeometry::getPrimType() const
{
   sglStats::PrimEnum prim_type = sglStats::eINTERLEAVED_POINTS;
   switch (m_prim_type)
   {
      case GL_POINTS:
         prim_type = sglStats::eINTERLEAVED_POINTS;
         break;
      case GL_LINES:
         prim_type = sglStats::eINTERLEAVED_LINES;
         break;
      case GL_LINE_STRIP:
         prim_type = sglStats::eINTERLEAVED_LINESTRIPS;
         break;
      case GL_LINE_LOOP:
         prim_type = sglStats::eINTERLEAVED_LINELOOPS;
         break;
      case GL_TRIANGLES:
         prim_type = sglStats::eINTERLEAVED_TRIS;
         break;
      case GL_TRIANGLE_STRIP:
         prim_type = sglStats::eINTERLEAVED_TRISTRIPS;
         break;
      case GL_TRIANGLE_FAN:
         prim_type = sglStats::eINTERLEAVED_TRIFANS;
         break;
      case GL_QUADS:
         prim_type = sglStats::eINTERLEAVED_QUADS;
         break;
      case GL_QUAD_STRIP:
         prim_type = sglStats::eINTERLEAVED_QUADSTRIPS;
         break;
      case GL_POLYGON:
         prim_type = sglStats::eINTERLEAVED_POLYS;
         break;
      default:
         break;
   }
   return prim_type;
}


//----------------------------------------------------------------------------
unsigned int sglInterleavedGeometry::getPrimSize() const
{
   unsigned int prim_size = 0;
   switch (m_prim_type)
   {
      case GL_POINTS:
         prim_size = 1;
         break;
      case GL_LINES:
         prim_size = 2;
         break;
      case GL_TRIANGLES:
         prim_size = 3;
         break;
      case GL_QUADS:
         prim_size = 4;
         break;
      default:
         break;
   }
   return prim_size;
}

//----------------------------------------------------------------------------
void sglInterleavedGeometry::setNumPrims(unsigned int num)
{
   m_num_prims = num;
   if (num)
   {
      m_compute_bounds |= eNUMPRIMS;
   }
   else
   {
      m_compute_bounds &= ~eNUMPRIMS;
   }
   dirty(eBOUND|eDLIST, true);
}

//----------------------------------------------------------------------------
void sglInterleavedGeometry::setPrimLengths(unsigned int *lengths)
{
   m_lengths = lengths;
   if (lengths)
   {
      m_compute_bounds |= eLENGTHS;
   }
   else
   {
      m_compute_bounds &= ~eLENGTHS;
   }
   dirty(eBOUND|eDLIST, true);
}

//----------------------------------------------------------------------------
void sglInterleavedGeometry::setInterleavedArray(GLenum format,
                                                 void *array,
                                                 GLsizei stride)
{
   if (format == GL_V2F ||
       format == GL_V3F ||
       format == GL_C4UB_V2F ||
       format == GL_C4UB_V3F ||
       format == GL_C3F_V3F ||
       format == GL_N3F_V3F ||
       format == GL_C4F_N3F_V3F ||
       format == GL_T2F_V3F ||
       format == GL_T4F_V4F ||
       format == GL_T2F_C4UB_V3F ||
       format == GL_T2F_C3F_V3F ||
       format == GL_T2F_N3F_V3F ||
       format == GL_T2F_C4F_N3F_V3F ||
       format == GL_T4F_C4F_N3F_V4F)
   {
      m_compute_bounds |= eARRAY_FORMAT;
      m_array_format = format;
   }
   else
   {
      m_compute_bounds &= ~eARRAY_FORMAT;
      sglPrintWarning() << "sglInterleavedGeometry::setInterleavedArray error:"
                        << " invalid array format = 0x" << hex << (int)format
                        << dec << endl;
   }

   if (array == NULL)
   {
      m_compute_bounds &= ~eARRAY;
      sglPrintWarning() << "sglInterleavedGeometry::setInterleavedArray error:"
                        << " NULL array pointer" << endl;
   }
   else
   {
      m_compute_bounds |= eARRAY;
      m_array = array;
   }

   m_array_stride = stride;

   dirty(eBOUND|eDLIST, true);
}

//----------------------------------------------------------------------------
void sglInterleavedGeometry::getInterleavedArray(GLenum        *format,
                                                 void         **array,
                                                 GLsizei       *stride)
{
   *format = m_array_format;
   *array = m_array;
   *stride = m_array_stride;
}

//----------------------------------------------------------------------------
bool sglInterleavedGeometry::computeBounds()
{
   const unsigned int full_mask = (ePRIM_TYPE |
                                   eNUMPRIMS |
                                   eLENGTHS |
                                   eARRAY |
                                   eARRAY_FORMAT);

   if (m_compute_bounds&full_mask != full_mask)
      return false;

   char *ptr = (char*)m_array;
   unsigned int stride = m_array_stride;

   switch (m_array_format)
   {
      case  GL_V2F:
      {
         if (stride == 0)
            stride = sizeof(sglVec2f);

         sglVec2f *array = (sglVec2f*)ptr;
         m_bbox.setMin(array[0][0], array[0][1], 0.f);
         m_bbox.setMax(array[0][0], array[0][1], 0.f);

         for (unsigned int iy = 0; iy < m_num_prims; ++iy)
         {
            for (unsigned int ix = 0; ix < m_lengths[iy]; ++ix)
            {
               ptr += stride;
               array = (sglVec2f*)ptr;
               m_bbox.expandToInclude(sglVec3f(array[0][0], array[0][1], 0.f));
            }
         }
         break;
      }
      case  GL_V3F:
      {
         if (stride == 0)
            stride = sizeof(sglVec3f);

         sglVec3f *array = (sglVec3f*)ptr;
         m_bbox.setMin(array[0]);
         m_bbox.setMax(array[0]);

         for (unsigned int iy = 0; iy < m_num_prims; ++iy)
         {
            for (unsigned int ix = 0; ix < m_lengths[iy]; ++ix)
            {
               ptr += stride;
               array = (sglVec3f*)ptr;
               m_bbox.expandToInclude(array[0]);
            }
         }
         break;
      }
      case  GL_C4UB_V2F:
      {
         ptr += 4;    // step past C4ub
         if (stride == 0)
            stride = sizeof(sglC4ubV2fStruct);

         sglVec2f *array = (sglVec2f*)ptr;
         m_bbox.setMin(array[0][0], array[0][1], 0.f);
         m_bbox.setMax(array[0][0], array[0][1], 0.f);

         for (unsigned int iy = 0; iy < m_num_prims; ++iy)
         {
            for (unsigned int ix = 0; ix < m_lengths[iy]; ++ix)
            {
               ptr += stride;
               array = (sglVec2f*)ptr;
               m_bbox.expandToInclude(sglVec3f(array[0][0], array[0][1], 0.f));
            }
         }
         break;
      }
      case  GL_C4UB_V3F:
      {
         ptr += 4; // step past C4ub
         if (stride == 0)
            stride = sizeof(sglC4ubV3fStruct);

         sglVec3f *array = (sglVec3f*)ptr;
         m_bbox.setMin(array[0]);
         m_bbox.setMax(array[0]);

         for (unsigned int iy = 0; iy < m_num_prims; ++iy)
         {
            for (unsigned int ix = 0; ix < m_lengths[iy]; ++ix)
            {
               ptr += stride;
               array = (sglVec3f*)ptr;
               m_bbox.expandToInclude(array[0]);
            }
         }
         break;
      }
      case  GL_C3F_V3F:
      {
         ptr += sizeof(sglVec3f);  // step past C3F
         if (stride == 0)
            stride = sizeof(sglC3fV3fStruct);

         sglVec3f *array = (sglVec3f*)ptr;
         m_bbox.setMin(array[0]);
         m_bbox.setMax(array[0]);

         for (unsigned int iy = 0; iy < m_num_prims; ++iy)
         {
            for (unsigned int ix = 0; ix < m_lengths[iy]; ++ix)
            {
               ptr += stride;
               array = (sglVec3f*)ptr;
               m_bbox.expandToInclude(array[0]);
            }
         }
         break;
      }
      case  GL_N3F_V3F:
      {
         ptr += sizeof(sglVec3f);  // step past N3F
         if (stride == 0)
            stride = sizeof(sglN3fV3fStruct);

         sglVec3f *array = (sglVec3f*)ptr;
         m_bbox.setMin(array[0]);
         m_bbox.setMax(array[0]);

         for (unsigned int iy = 0; iy < m_num_prims; ++iy)
         {
            for (unsigned int ix = 0; ix < m_lengths[iy]; ++ix)
            {
               ptr += stride;
               array = (sglVec3f*)ptr;
               m_bbox.expandToInclude(array[0]);
            }
         }
         break;
      }
      case  GL_C4F_N3F_V3F:
      {
         ptr += sizeof(sglVec4f) + sizeof(sglVec3f);  // step past C4F_N3F
         if (stride == 0)
            stride = sizeof(sglC4fN3fV3fStruct);

         sglVec3f *array = (sglVec3f*)ptr;
         m_bbox.setMin(array[0]);
         m_bbox.setMax(array[0]);

         for (unsigned int iy = 0; iy < m_num_prims; ++iy)
         {
            for (unsigned int ix = 0; ix < m_lengths[iy]; ++ix)
            {
               ptr += stride;
               array = (sglVec3f*)ptr;
               m_bbox.expandToInclude(array[0]);
            }
         }
         break;
      }
      case  GL_T2F_V3F:
      {
         ptr += sizeof(sglVec2f);  // step past T2F
         if (stride == 0)
            stride = sizeof(sglT2fV3fStruct);

         sglVec3f *array = (sglVec3f*)ptr;
         m_bbox.setMin(array[0]);
         m_bbox.setMax(array[0]);

         for (unsigned int iy = 0; iy < m_num_prims; ++iy)
         {
            for (unsigned int ix = 0; ix < m_lengths[iy]; ++ix)
            {
               ptr += stride;
               array = (sglVec3f*)ptr;
               m_bbox.expandToInclude(array[0]);
            }
         }
         break;
      }
      case  GL_T4F_V4F:
      {
         ptr += sizeof(sglVec4f);  // step past T4F
         if (stride == 0)
            stride = sizeof(sglT4fV4fStruct);

         sglVec4f *array = (sglVec4f*)ptr;
         m_bbox.setMin(array[0][0], array[0][1], array[0][2]);
         m_bbox.setMax(array[0][0], array[0][1], array[0][2]);

         for (unsigned int iy = 0; iy < m_num_prims; ++iy)
         {
            for (unsigned int ix = 0; ix < m_lengths[iy]; ++ix)
            {
               ptr += stride;
               array = (sglVec4f*)ptr;
               m_bbox.expandToInclude(sglVec3f(array[0][0]/array[0][3],
                                               array[0][1]/array[0][3],
                                               array[0][2]/array[0][3]));
            }
         }
         break;
      }
      case  GL_T2F_C4UB_V3F:
      {
         ptr += (4 + sizeof(sglVec2f));  // step past T2F_C4UB
         if (stride == 0)
            stride = sizeof(sglT2fC4ubV3fStruct);

         sglVec3f *array = (sglVec3f*)ptr;
         m_bbox.setMin(array[0]);
         m_bbox.setMax(array[0]);

         for (unsigned int iy = 0; iy < m_num_prims; ++iy)
         {
            for (unsigned int ix = 0; ix < m_lengths[iy]; ++ix)
            {
               ptr += stride;
               array = (sglVec3f*)ptr;
               m_bbox.expandToInclude(array[0]);
            }
         }
         break;
      }
      case  GL_T2F_C3F_V3F:
      {
         ptr += (sizeof(sglVec3f) + sizeof(sglVec2f));  // step past T2F_C3F
         if (stride == 0)
            stride = sizeof(sglT2fC3fV3fStruct);

         sglVec3f *array = (sglVec3f*)ptr;
         m_bbox.setMin(array[0]);
         m_bbox.setMax(array[0]);

         for (unsigned int iy = 0; iy < m_num_prims; ++iy)
         {
            for (unsigned int ix = 0; ix < m_lengths[iy]; ++ix)
            {
               ptr += stride;
               array = (sglVec3f*)ptr;
               m_bbox.expandToInclude(array[0]);
            }
         }
         break;
      }
      case  GL_T2F_N3F_V3F:
      {
         ptr += (sizeof(sglVec3f) + sizeof(sglVec2f));  // step past T2F_N3F
         if (stride == 0)
            stride = sizeof(sglT2fN3fV3fStruct);

         sglVec3f *array = (sglVec3f*)ptr;
         m_bbox.setMin(array[0]);
         m_bbox.setMax(array[0]);

         for (unsigned int iy = 0; iy < m_num_prims; ++iy)
         {
            for (unsigned int ix = 0; ix < m_lengths[iy]; ++ix)
            {
               ptr += stride;
               array = (sglVec3f*)ptr;
               m_bbox.expandToInclude(array[0]);
            }
         }
         break;
      }
      case  GL_T2F_C4F_N3F_V3F:
      {
         ptr += (sizeof(sglVec2f) + sizeof(sglVec4f) +
                 sizeof(sglVec3f));  // step past T2F_C4F_N3F
         if (stride == 0)
            stride = sizeof(sglT2fC4fN3fV3fStruct);

         sglVec3f *array = (sglVec3f*)ptr;
         m_bbox.setMin(array[0]);
         m_bbox.setMax(array[0]);

         for (unsigned int iy = 0; iy < m_num_prims; ++iy)
         {
            for (unsigned int ix = 0; ix < m_lengths[iy]; ++ix)
            {
               ptr += stride;
               array = (sglVec3f*)ptr;
               m_bbox.expandToInclude(array[0]);
            }
         }
         break;
      }
      case  GL_T4F_C4F_N3F_V4F:
      {
         // HACK?  What do I do with the 4th component / divide?

         ptr += (2*sizeof(sglVec4f) +
                 sizeof(sglVec3f));  // step past T4F_C4F_N3F
         if (stride == 0)
            stride = sizeof(sglT4fC4fN3fV4fStruct);

         sglVec4f *array = (sglVec4f*)ptr;
         m_bbox.setMin(array[0][0], array[0][1], array[0][2]);
         m_bbox.setMax(array[0][0], array[0][1], array[0][2]);

         for (unsigned int iy = 0; iy < m_num_prims; ++iy)
         {
            for (unsigned int ix = 0; ix < m_lengths[iy]; ++ix)
            {
               ptr += stride;
               array = (sglVec4f*)ptr;
               m_bbox.expandToInclude(sglVec3f(array[0][0]/array[0][3],
                                               array[0][1]/array[0][3],
                                               array[0][2]/array[0][3]));
            }
         }
         break;
      }
      default:
      {
         // if m_compute_bounds == full_mask it should never get here
         m_bbox.makeEmpty();
         return false;
      }
   }

   m_dirty &= ~eBOUND;  // ??
   return true;
}

//----------------------------------------------------------------------------
bool sglInterleavedGeometry::isValid() const
{
   const unsigned int full_mask = (ePRIM_TYPE |
                                   eARRAY |
                                   eARRAY_FORMAT);

   if (m_compute_bounds&full_mask != full_mask)
      return false;
   else
      return true;
}

//----------------------------------------------------------------------------
void sglInterleavedGeometry::drawGeometry(const vector<sglTexCoords> &) const
{
   const unsigned int full_mask = (ePRIM_TYPE |
                                   eNUMPRIMS |
                                   eLENGTHS |
                                   eARRAY |
                                   eARRAY_FORMAT);

   if (m_compute_bounds&full_mask != full_mask)
      return;

   glPushClientAttrib(GL_CLIENT_VERTEX_ARRAY_BIT);   // is this necessary?

   unsigned int index = 0;
   for (unsigned int i=0; i<m_num_prims; ++i)
   {
      glInterleavedArrays(m_array_format, m_array_stride, m_array);
      glDrawArrays(m_prim_type, index, m_lengths[i]);
      index += m_lengths[i];
   }
   glPopClientAttrib();
}

//----------------------------------------------------------------------------
void sglInterleavedGeometry::addStats(sglStats &stats) const
{
   unsigned int i;
   sglStats::PrimEnum prim_type = getPrimType();

   stats.addGSets(prim_type, 1);

   for (i=0; i<m_num_prims; ++i)
      stats.addVertices(prim_type, m_lengths[i]);

   switch (m_prim_type)
   {
      case GL_POINTS:
      case GL_LINES:
         for (i=0; i<m_num_prims; ++i)
            stats.addPrims(prim_type, m_lengths[i]/getPrimSize());
         break;

      case GL_LINE_STRIP:
      case GL_LINE_LOOP:
         stats.addPrims(prim_type, m_num_prims);
         break;

      case GL_TRIANGLES:
         for (i=0; i<m_num_prims; ++i)
         {
            unsigned int num = m_lengths[i]/3;
            stats.addPrims(prim_type, (float)num);
            stats.addTriangles(prim_type, num);
         }
         break;

      case GL_TRIANGLE_STRIP:
      case GL_TRIANGLE_FAN:
         stats.addPrims(prim_type, m_num_prims);
         for (i=0; i<m_num_prims; ++i)
         {
            if (m_lengths[i] > 2)
               stats.addTriangles(prim_type, (m_lengths[i] - 2));
         }
         break;

      case GL_QUADS:
         for (i=0; i<m_num_prims; ++i)
         {
            unsigned int num = m_lengths[i]/4;
            stats.addPrims(prim_type, num);
            stats.addTriangles(prim_type, (num*2));
         }
         break;

      case GL_QUAD_STRIP:
         stats.addPrims(prim_type, m_num_prims);
         for (i=0; i<m_num_prims; ++i)
         {
            unsigned int num = (m_lengths[i]-2)/2;
            if (num > 0)
               stats.addTriangles(prim_type, (num*2));
         }
         break;

      case GL_POLYGON:
         stats.addPrims(prim_type, m_num_prims);
         for (i=0; i<m_num_prims; ++i)
         {
            if (m_lengths[i] > 2)
               stats.addTriangles(prim_type, (m_lengths[i]-2));
         }
         break;

      default:
         break;
   }

}

//----------------------------------------------------------------------------
void sglInterleavedGeometry::printInfo(ostream &ostrm,
                                       const char *indent_string) const
{
   sglDrawable::printInfo(ostrm, indent_string);
   ostrm << indent_string << "sglInterleavedGeometry prim_type: 0x" << hex
         << (int)m_prim_type << dec << "(";

   switch (m_prim_type)
   {
      case GL_POINTS:
         ostrm << "sglStats::eINTERLEAVED_POINTS)" << endl;
         break;
      case GL_LINES:
         ostrm << "sglStats::eINTERLEAVED_LINES)" << endl;
         break;
      case GL_LINE_STRIP:
         ostrm << "sglStats::eINTERLEAVED_LINESTRIPS)" << endl;
         break;
      case GL_LINE_LOOP:
         ostrm << "sglStats::eINTERLEAVED_LINELOOPS)" << endl;
         break;
      case GL_TRIANGLES:
         ostrm << "sglStats::eINTERLEAVED_TRIS)" << endl;
         break;
      case GL_TRIANGLE_STRIP:
         ostrm << "sglStats::eINTERLEAVED_TRISTRIPS)" << endl;
         break;
      case GL_TRIANGLE_FAN:
         ostrm << "sglStats::eINTERLEAVED_TRIFANS)" << endl;
         break;
      case GL_QUADS:
         ostrm << "sglStats::eINTERLEAVED_QUADS)" << endl;
         break;
      case GL_QUAD_STRIP:
         ostrm << "sglStats::eINTERLEAVED_QUADSTRIPS)" << endl;
         break;
      case GL_POLYGON:
         ostrm << "sglStats::eINTERLEAVED_POLYS)" << endl;
         break;
      default:
         ostrm << "unknown)" << endl;
         break;
   }

   ostrm << indent_string << "     num_prims: " << m_num_prims << endl;
   ostrm << indent_string << "     prim_lengths: ";
   for (unsigned int ix=0; ix<m_num_prims; ++ix)
      ostrm << m_lengths[ix] << ",";
   ostrm << endl;

   ostrm << indent_string << "     array: 0x" << hex
         << m_array << dec << endl;

   ostrm << indent_string << "     format: 0x" << hex
         << (int)m_array_format << dec << "(";

   switch (m_array_format)
   {
      case  GL_V2F:
         ostrm << "GL_V2F)" << endl;
         break;
      case  GL_V3F:
         ostrm << "GL_V3F)" << endl;
         break;
      case  GL_C4UB_V2F:
         ostrm << "GL_C4UB_V2F)" << endl;
         break;
      case  GL_C4UB_V3F:
         ostrm << "GL_C4UB_V3F)" << endl;
         break;
      case  GL_C3F_V3F:
         ostrm << "GL_C3F_V3F)" << endl;
         break;
      case  GL_N3F_V3F:
         ostrm << "GL_N3F_V3F)" << endl;
         break;
      case  GL_C4F_N3F_V3F:
         ostrm << "GL_C4F_N3F_V3F)" << endl;
         break;
      case  GL_T2F_V3F:
         ostrm << "GL_T2F_V3F)" << endl;
         break;
      case  GL_T4F_V4F:
         ostrm << "GL_T4F_V4F)" << endl;
         break;
      case  GL_T2F_C4UB_V3F:
         ostrm << "GL_T2F_C4UB_V3F)" << endl;
         break;
      case  GL_T2F_C3F_V3F:
         ostrm << "GL_T2F_C3F_V3F)" << endl;
         break;
      case  GL_T2F_N3F_V3F:
         ostrm << "GL_T2F_N3F_V3F)" << endl;
         break;
      case  GL_T2F_C4F_N3F_V3F:
         ostrm << "GL_T2F_C4F_N3F_V3F)" << endl;
         break;
      case  GL_T4F_C4F_N3F_V4F:
         ostrm << "GL_T4F_C4F_N3F_V4F)" << endl;
         break;
      default:
         ostrm << "unknown)" << endl;
         break;
   }

   ostrm << indent_string << "     stride: " << m_array_stride << endl;
}
