/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglInterleavedGeometry.hpp
 *  Created: 3 January 2000
 *  Summary: geometry primitives utilizing glInterleaveArray
 *****************************************************************************/

#ifndef __SGL_INTERLEAVED_GEOMETRY_HPP
#define __SGL_INTERLEAVED_GEOMETRY_HPP

#include <sgl.h>
#include <sglDrawable.hpp>

//============================================================================

/**

sglInterleavedGeometry encapsulates OpennGL's mechanism for packing
(interleaving if you will) the geometry data (including vertices, normals,
vertex colors, and texture coordinates) in a single array of data and render
the corresponding geometry.  The sglInterleavedGeometry class wraps this
functionality.  To use this class, the following must be specified:
\begin{itemize}
\item the interleaved array of data and the stride using
sglInterleavedGeometry::setInterleavedArray(..).

\item the primitive type by calling sglInterleavedGeometry::setGLPrimType(..).

\item the number of primitives by calling
sglInterleavedGeometry::setNumPrims(..).

\item the lengths array by calling sglInterleavedGeometry::setPrimLengths(..).
\end{itemize}

For the first step the format of the interleaved geometry must be one of
the 14 formats in the OpenGL specification.  When using this class with the
default stride of zero, the user can allocate an array of one of the helper
structs provided with this class.  The following is a list of these structs
and the corresponding OpenGL token that is needed in the call to
sglInterleavedGeometry::setInterleavedArray(..):

\begin{verbatim}
     OpenGL format         SGL struct name
     ------------------    ---------------------
     GL_V2F                sglVec2f
     GL_V3F                sglVec3f
     GL_C4UB_V2F           sglC4ubV2fStruct
     GL_C4UB_V3F           sglC4ubV3fStruct
     GL_C3F_V3F            sglC3fV3fStruct
     GL_N3F_V3F            sglN3fV3fStruct
     GL_C4F_N3F_V3F        sglC4fN3fV3fStruct
     GL_T2F_V3F            sglT2fV3fStruct
     GL_T4F_V4F            sglT4fV4fStruct
     GL_T2F_C4UB_V3F       sglT2fC4ubV3fStruct
     GL_T2F_C3F_V3F        sglT2fC3fV3fStruct
     GL_T2F_N3F_V3F        sglT2fN3fV3fStruct
     GL_T2F_C4F_N3F_V3F    sglT2fC4fN3fV3fStruct
     GL_T4F_C4F_N3F_V4F    sglT4fC4fN3fV4fStruct
\end{verbatim}

Once the array is allocated and the values are set the the
sglInterleavedGeometry::setInterleavedArray(...) is called to pass a pointer
to this array into the object.  Note: the array will be used directly by the
class; however, the array will not get destructed with the class is deleted
(this is the user's responsibility).  With the call to
setInterleavedArray(...), the user must also specify the corresponding OpenGL
format token, and can optionally specify the stride between elements in the
array (it defaults to zero).

The second step is to specify the primitive type from one of the following
OpenGL tokens: GL_POINTS, GL_LINES, GL_LINE_STRIP, GL_LINE_LOOP,
GL_TRIANGLES, GL_TRIANGLE_STRIP, GL_TRIANGLE_FAN, GL_QUADS, GL_QUAD_STRIP,
GL_POLYGON.

This class must support the strip, fan and polygon primitives which
require the lengths array.  Because of this all primitive types require a
lengths array.  For primitive types that are NOT strip, fan, or polygon types
a lengths array with a single element is all that is required and that value
is set to the number of vertices present in the interleaved array.  For this
setup the user must call sglInterleavedArrays::setNumPrims(1).

For strip, fan, and polygon types the number of primitives and the lengths
array behaves just like the equivalent sglGeoSet subclasses above.

There is no indexing supported for this class.

This class imposes some restrictions on the base sglDrawable class API.  While
multiple states can be associated with this geometry type.  The individual
texture coordinates in the StateElement struct will be ignored, as the texture
coordinates are stored in the respective interleaved array formats and cannot
be switched out.

{\bf Example:}
\begin{verbatim}
   sglVec3f color[3];
   color[0].set(1.0f, 0.6f, 0.6f);
   color[1].set(0.6f, 1.0f, 0.6f);
   color[2].set(0.6f, 0.6f, 1.0f);

   // interleaved 3 component color and vertices
   sglC3fV3fStruct *array = new sglC3fV3fStruct[6];

   unsigned int *lengths = new unsigned int[1];
   lengths[0] = 6;

   unsigned int index = 0;
   for (unsigned int ix=0; ix<3; ix++)
   {
      array[index].vertex.set(ix*8.0f - 8.0f, 0.0f, 8.0f);
      array[index].color = color[ix];
      index++;

      array[index].vertex.set(ix*8.0f - 8.0f, 0.0f, -8.0f);
      array[index].color = color[ix];
      index++;
   }

   sglInterleavedGeometry *gset = new sglInterleavedGeometry();
   gset->setNumPrims(1);
   gset->setPrimLengths(lengths);
   gset->setGLPrimType(GL_QUAD_STRIP);
   gset->setInterleavedArray(GL_C3F_V3F, array, 0);

   // This object is too small to want a display list.
   gset->enableDisplayList(false);
\end{verbatim}

\todo Implement pick and intersect functions.

*/

class SGL_DLL_API sglInterleavedGeometry : public sglDrawable
{
public:
   ///
   sglInterleavedGeometry();
   ///
   virtual ~sglInterleavedGeometry();

   ///
   virtual AAEnum getAAType() const;

   /** get/set the type of primitive
       prim_type is one of GL_LINES, GL_TRIANGLE_STRIP, etc
   */
   void setGLPrimType(GLenum prim_type);
   ///
   GLenum getGLPrimType() const { return m_prim_type; }
   ///
   virtual sglStats::PrimEnum getPrimType() const;
   ///
   virtual unsigned int getPrimSize() const;

   /** get/set the number of primitives in this geometry.  This has a slightly
       different meaning than in sglGeoSet and sglGeoStripSet.  This will be
       used in conjunction with the lengths array regardless of whether a strip
       primitive is specified or not.
   */
   unsigned int getNumPrims() const {return m_num_prims;}
   ///
   void setNumPrims(unsigned int num);

   /// get/set the lengths array (for strip and poly primitives only).
   unsigned int *getPrimLengths() const {return m_lengths;}
   ///
   void setPrimLengths(unsigned int *lengths);

   /** specify the geometry attribute array: format is one of the interleaved
       formats listed above...if a packed array of the one of the sgl types is
       used to organize the data then the stride is zero if every element is
       desired.
   */
   void setInterleavedArray(GLenum format,
                            void *array,
                            GLsizei stride = 0);

   ///
   void getInterleavedArray(GLenum *format,
                            void **array,
                            GLsizei *stride);

   ///
   virtual bool isValid() const;

   /// tex_coords not used here.
   virtual void drawGeometry(const vector<sglTexCoords> &tex_coords) const;

   ///
   virtual void addStats(sglStats &stats) const;

   ///
   virtual void printInfo(ostream &ostrm, const char *indent_string) const;

protected:
   virtual bool computeBounds();

private:   // not implemented
   sglInterleavedGeometry(const sglInterleavedGeometry &);
   sglInterleavedGeometry &operator=(const sglInterleavedGeometry &);

private:
   GLenum        m_prim_type;
   unsigned int  m_num_prims;
   unsigned int *m_lengths;       // List of strip lengths

   void         *m_array;         // the packed data (see structs below).
   GLenum        m_array_format;  // one of the fourteen types
   GLsizei       m_array_stride;

   enum BoundsEnum {ePRIM_TYPE    = 0x01,
                    eNUMPRIMS     = 0x02,
                    eLENGTHS      = 0x04,
                    eARRAY        = 0x08,
                    eARRAY_FORMAT = 0x10};
   unsigned int  m_compute_bounds;
};

//----------------------------------------------------------------------------
// structures to aid in packing the interleaved elements

//struct sglV2fStruct

//struct sglV3fStruct

struct sglC4ubV2fStruct
{
   unsigned char color[4];
   sglVec2f      vertex;
};

struct sglC4ubV3fStruct
{
   unsigned char color[4];
   sglVec3f      vertex;
};

struct sglC3fV3fStruct
{
   sglVec3f color;
   sglVec3f vertex;
};

struct sglN3fV3fStruct
{
   sglVec3f normal;
   sglVec3f vertex;
};

struct sglC4fN3fV3fStruct
{
   sglVec4f color;
   sglVec3f normal;
   sglVec3f vertex;
};

struct sglT2fV3fStruct
{
   sglVec2f tex_coord;
   sglVec3f vertex;
};

struct sglT4fV4fStruct
{
   sglVec4f tex_coord;
   sglVec4f vertex;
};

struct sglT2fC4ubV3fStruct
{
   sglVec2f      tex_coord;
   unsigned char color[4];
   sglVec3f      vertex;
};

struct sglT2fC3fV3fStruct
{
   sglVec2f tex_coord;
   sglVec3f color;
   sglVec3f vertex;
};

struct sglT2fN3fV3fStruct
{
   sglVec2f tex_coord;
   sglVec3f normal;
   sglVec3f vertex;
};

struct sglT2fC4fN3fV3fStruct
{
   sglVec2f tex_coord;
   sglVec4f color;
   sglVec3f normal;
   sglVec3f vertex;
};

struct sglT4fC4fN3fV4fStruct
{
   sglVec4f tex_coord;
   sglVec4f color;
   sglVec3f normal;
   sglVec4f vertex;
};

//----------------------------------------------------------------------------

#endif
