/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglIntersect.cpp
 *   Author: Scott McMillan
 *  Created: 26 March 2001
 *****************************************************************************/

#include <sglIntersect.hpp>


//----------------------------------------------------------------------------
template <class T>
sglIntersect<T>::sglIntersect()
      : sglProbeTemplate<T>(),
        m_result(),
        m_path()
{
}

//----------------------------------------------------------------------------
template <class T>
sglIntersect<T>::~sglIntersect()
{
   m_path.clear();
}

//----------------------------------------------------------------------------
template <class T>
void sglIntersect<T>::reset()
{
   m_result.reset();
   m_path.clear();
   sglProbeTemplate<T>::reset();
}

//----------------------------------------------------------------------------
template <class T>
void sglIntersect<T>::intersectTriangle(
   sglSegment<T> &seg,
   const sglVec3<T> &vertex0,
   const sglVec3<T> &vertex1,
   const sglVec3<T> &vertex2,
   sglBound::IntersectResultEnum &result)
{
   // ---------------------------------------------------------------------
   // The following is adapted from "An Efficient Ray/Polygon Intersection"
   // by Didier Badouel from "Graphics Gems", Academic Press, 1990

   // these normals should be stored in the geosets for future intersects
   sglVec3<T> normal, vec1, vec2;
   vec1.sub(vertex1, vertex0);
   vec2.sub(vertex2, vertex0);
   normal.cross(vec1, vec2);
   normal.normalize();

   T sface = normal.dot(seg.getDir());

   // if seg.dir dot face.normal is negative seg will go through front of
   // triangle face
   if (sface <= 0.0 && m_mode&sglProbe::eCULL_FRONT)
   {
      return;
   }
   else if (sface >= 0.0 && m_mode&sglProbe::eCULL_BACK)
   {
      return;
   }

   // if triangle and ray are parallel skip...
   else if (fabs(sface) < 1.0e-6)
   {
      return;
   }

   // if intersection is behind the segment (t<0) or farther than the
   // segment (t>seg.getLen()), skip...
   T d = -normal.dot(vertex0);  // should be stored
   T new_len = -(d + normal.dot(seg.getPos()))/(normal.dot(seg.getDir()));

   if (new_len <= 0 || new_len >= seg.getLen())
   {
      return;
   }

   // find the largest component of the normal
   unsigned int u, v;
   T nx = fabs(normal[0]);
   T ny = fabs(normal[1]);
   T nz = fabs(normal[2]);
   if (nx > ny && nx > nz)
   {
      u = 1;
      v = 2;
   }
   else if (ny > nx && ny > nz)
   {
      u = 2;
      v = 0;
   }
   else
   {
      u = 0;
      v = 1;
   }

   sglVec3<T> seg_endpt;
   seg_endpt.scale(seg.getDir(), new_len);
   seg_endpt += seg.getPos();

   T u0, u1, u2, v0, v1, v2;
   u0 = seg_endpt[u] - vertex0[u];
   v0 = seg_endpt[v] - vertex0[v];

   u1 = vertex1[u] - vertex0[u];
   v1 = vertex1[v] - vertex0[v];

   u2 = vertex2[u] - vertex0[u];
   v2 = vertex2[v] - vertex0[v];

   T alpha, beta;
   if (u1 == 0)
   {
      beta = u0/u2;
      if ((beta >= 0.)&&(beta <= 1.))
      {
         alpha = (v0 - beta*v2)/v1;
         if ((alpha >= 0.)&&(alpha+beta) <= 1.)
         {
            result = sglBound::eINTERSECT;
            seg.setLen(new_len);

            if (m_mode&sglProbe::eNORMAL)
               m_result.m_normal = normal;
         }
      }
   }
   else
   {
      beta = (v0*u1 - u0*v1)/(v2*u1 - u2*v1);
      if ((beta >= 0.)&&(beta <= 1.))
      {
         alpha = (u0 - beta*u2)/u1;
         if ((alpha >= 0)&&((alpha+beta) <= 1.))
         {
            result = sglBound::eINTERSECT;
            seg.setLen(new_len);

            if (m_mode&sglProbe::eNORMAL)
               m_result.m_normal = normal;
         }
      }
   }
}

//----------------------------------------------------------------------------
// template <class T> template <class S>
// void sglIntersect<T>::intersectTriangle2(
//    sglSegment<T> &seg,
//    const void *vtx0, const void *vtx1, const void *vtx2,
//    sglBound::IntersectResultEnum &result)
// {
//    const S *ptr = (const S*)vtx0;
//    sglVec3<T> v0(ptr[0], ptr[1], 0);
//    ptr = (const S*)vtx1;
//    sglVec3<T> v1(ptr[0], ptr[1], 0);
//    ptr = (const S*)vtx2;
//    sglVec3<T> v2(ptr[0], ptr[1], 0);

//    intersectTriangle(seg, v0, v1, v2, result);
// }

//----------------------------------------------------------------------------
template <class T>
void sglIntersect<T>::intersectTriangle2s(
   sglIntersect<T> &segset,
   sglSegment<T> &seg,
   const void *vtx0, const void *vtx1, const void *vtx2,
   sglBound::IntersectResultEnum &result)
{
   const short *ptr = (const short*)vtx0;
   sglVec3<T> v0(ptr[0], ptr[1], 0);
   ptr = (const short*)vtx1;
   sglVec3<T> v1(ptr[0], ptr[1], 0);
   ptr = (const short*)vtx2;
   sglVec3<T> v2(ptr[0], ptr[1], 0);

   segset.intersectTriangle(seg, v0, v1, v2, result);
}

//----------------------------------------------------------------------------
template <class T>
void sglIntersect<T>::intersectTriangle2i(
   sglIntersect<T> &segset,
   sglSegment<T> &seg,
   const void *vtx0, const void *vtx1, const void *vtx2,
   sglBound::IntersectResultEnum &result)
{
   const int *ptr = (const int*)vtx0;
   sglVec3<T> v0(ptr[0], ptr[1], 0);
   ptr = (const int*)vtx1;
   sglVec3<T> v1(ptr[0], ptr[1], 0);
   ptr = (const int*)vtx2;
   sglVec3<T> v2(ptr[0], ptr[1], 0);

   segset.intersectTriangle(seg, v0, v1, v2, result);
}

//----------------------------------------------------------------------------
template <class T>
void sglIntersect<T>::intersectTriangle2f(
   sglIntersect<T> &segset,
   sglSegment<T> &seg,
   const void *vtx0, const void *vtx1, const void *vtx2,
   sglBound::IntersectResultEnum &result)
{
   const float *ptr = (const float*)vtx0;
   sglVec3<T> v0(ptr[0], ptr[1], 0);
   ptr = (const float*)vtx1;
   sglVec3<T> v1(ptr[0], ptr[1], 0);
   ptr = (const float*)vtx2;
   sglVec3<T> v2(ptr[0], ptr[1], 0);

   segset.intersectTriangle(seg, v0, v1, v2, result);
}

//----------------------------------------------------------------------------
template <class T>
void sglIntersect<T>::intersectTriangle2d(
   sglIntersect<T> &segset,
   sglSegment<T> &seg,
   const void *vtx0, const void *vtx1, const void *vtx2,
   sglBound::IntersectResultEnum &result)
{
   const double *ptr = (const double*)vtx0;
   sglVec3<T> v0(ptr[0], ptr[1], 0);
   ptr = (const double*)vtx1;
   sglVec3<T> v1(ptr[0], ptr[1], 0);
   ptr = (const double*)vtx2;
   sglVec3<T> v2(ptr[0], ptr[1], 0);

   segset.intersectTriangle(seg, v0, v1, v2, result);
}

//----------------------------------------------------------------------------
// template <class T> template <class S>
// void sglIntersect<T>::intersectTriangle3(
//    sglSegment<T> &seg,
//    const void *vtx0, const void *vtx1, const void *vtx2,
//    sglBound::IntersectResultEnum &result)
// {
//    const S *ptr = (const S*)vtx0;
//    sglVec3<T> v0(ptr[0], ptr[1], ptr[2]);
//    ptr = (const S*)vtx1;
//    sglVec3<T> v1(ptr[0], ptr[1], ptr[2]);
//    ptr = (const S*)vtx2;
//    sglVec3<T> v2(ptr[0], ptr[1], ptr[2]);

//    intersectTriangle(seg, v0, v1, v2, result);
// }

// //----------------------------------------------------------------------------
// #ifdef sgi
// template <> template <>
// #else
// template <>
// #endif
// void sglIntersect<float>::intersectTriangle3<float>(
//    sglSegment<float> &seg,
//    const void *vtx0, const void *vtx1, const void *vtx2,
//    sglBound::IntersectResultEnum &result)
// {
//    intersectTriangle(seg,
//                      *(const sglVec3f*)vtx0,
//                      *(const sglVec3f*)vtx1,
//                      *(const sglVec3f*)vtx2,
//                      result);
// }

// //----------------------------------------------------------------------------
// #ifdef sgi
// template <> template <>
// #else
// template <>
// #endif
// void sglIntersect<double>::intersectTriangle3<double>(
//    sglSegment<double> &seg,
//    const void *vtx0, const void *vtx1, const void *vtx2,
//    sglBound::IntersectResultEnum &result)
// {
//    intersectTriangle(seg,
//                      *(const sglVec3d*)vtx0,
//                      *(const sglVec3d*)vtx1,
//                      *(const sglVec3d*)vtx2,
//                      result);
// }

//----------------------------------------------------------------------------
template <class T>
void sglIntersect<T>::intersectTriangle3s(
   sglIntersect<T> &segset,
   sglSegment<T> &seg,
   const void *vtx0, const void *vtx1, const void *vtx2,
   sglBound::IntersectResultEnum &result)
{
   const short *ptr = (const short*)vtx0;
   sglVec3<T> v0(ptr[0], ptr[1], ptr[2]);
   ptr = (const short*)vtx1;
   sglVec3<T> v1(ptr[0], ptr[1], ptr[2]);
   ptr = (const short*)vtx2;
   sglVec3<T> v2(ptr[0], ptr[1], ptr[2]);

   segset.intersectTriangle(seg, v0, v1, v2, result);
}

//----------------------------------------------------------------------------
template <class T>
void sglIntersect<T>::intersectTriangle3i(
   sglIntersect<T> &segset,
   sglSegment<T> &seg,
   const void *vtx0, const void *vtx1, const void *vtx2,
   sglBound::IntersectResultEnum &result)
{
   const int *ptr = (const int*)vtx0;
   sglVec3<T> v0(ptr[0], ptr[1], ptr[2]);
   ptr = (const int*)vtx1;
   sglVec3<T> v1(ptr[0], ptr[1], ptr[2]);
   ptr = (const int*)vtx2;
   sglVec3<T> v2(ptr[0], ptr[1], ptr[2]);

   segset.intersectTriangle(seg, v0, v1, v2, result);
}

//----------------------------------------------------------------------------
template <class T>
void sglIntersect<T>::intersectTriangle3f(
   sglIntersect<T> &segset,
   sglSegment<T> &seg,
   const void *vtx0, const void *vtx1, const void *vtx2,
   sglBound::IntersectResultEnum &result)
{
   const float *ptr = (const float*)vtx0;
   sglVec3<T> v0(ptr[0], ptr[1], ptr[2]);
   ptr = (const float*)vtx1;
   sglVec3<T> v1(ptr[0], ptr[1], ptr[2]);
   ptr = (const float*)vtx2;
   sglVec3<T> v2(ptr[0], ptr[1], ptr[2]);

   segset.intersectTriangle(seg, v0, v1, v2, result);
}

//----------------------------------------------------------------------------
#if !defined(WIN32)
//HACK This is an optimization that causes linker errors with the MS compiler.
template <>
void sglIntersect<float>::intersectTriangle3f(
   sglIntersect<float> &segset,
   sglSegment<float> &seg,
   const void *vtx0, const void *vtx1, const void *vtx2,
   sglBound::IntersectResultEnum &result)
{
   segset.intersectTriangle(seg,
                            *(const sglVec3f*)vtx0,
                            *(const sglVec3f*)vtx1,
                            *(const sglVec3f*)vtx2,
                            result);
}
#endif

//----------------------------------------------------------------------------
template <class T>
void sglIntersect<T>::intersectTriangle3d(
   sglIntersect<T> &segset,
   sglSegment<T> &seg,
   const void *vtx0, const void *vtx1, const void *vtx2,
   sglBound::IntersectResultEnum &result)
{
   const double *ptr = (const double*)vtx0;
   sglVec3<T> v0(ptr[0], ptr[1], ptr[2]);
   ptr = (const double*)vtx1;
   sglVec3<T> v1(ptr[0], ptr[1], ptr[2]);
   ptr = (const double*)vtx2;
   sglVec3<T> v2(ptr[0], ptr[1], ptr[2]);

   segset.intersectTriangle(seg, v0, v1, v2, result);
}

//----------------------------------------------------------------------------
#if !defined(WIN32)
//HACK This is an optimization that causes linker errors with the MS compiler.
template <>
void sglIntersect<double>::intersectTriangle3d(
   sglIntersect<double> &segset,
   sglSegment<double> &seg,
   const void *vtx0, const void *vtx1, const void *vtx2,
   sglBound::IntersectResultEnum &result)
{
   segset.intersectTriangle(seg,
                            *(const sglVec3d*)vtx0,
                            *(const sglVec3d*)vtx1,
                            *(const sglVec3d*)vtx2,
                            result);
}
#endif

//----------------------------------------------------------------------------
// template <class T> template <class S>
// void sglIntersect<T>::intersectTriangle4(
//    sglSegment<T> &seg,
//    const void *vtx0, const void *vtx1, const void *vtx2,
//    sglBound::IntersectResultEnum &result)
// {
//    const S *ptr = (const S*)vtx0;
//    S w = ptr[3];
//    sglVec3<T> v0(ptr[0]/w, ptr[1]/w, ptr[2]/w);
//    ptr = (const S*)vtx1;
//    w = ptr[3];
//    sglVec3<T> v1(ptr[0]/w, ptr[1]/w, ptr[2]/w);
//    ptr = (const S*)vtx2;
//    w = ptr[3];
//    sglVec3<T> v2(ptr[0]/w, ptr[1]/w, ptr[2]/w);

//    intersectTriangle(seg, v0, v1, v2, result);
// }

//----------------------------------------------------------------------------
template <class T>
void sglIntersect<T>::intersectTriangle4s(
   sglIntersect<T> &segset,
   sglSegment<T> &seg,
   const void *vtx0, const void *vtx1, const void *vtx2,
   sglBound::IntersectResultEnum &result)
{
   const short *ptr = (const short*)vtx0;
   short w = ptr[3];
   sglVec3<T> v0(ptr[0]/w, ptr[1]/w, ptr[2]/w);
   ptr = (const short*)vtx1;
   w = ptr[3];
   sglVec3<T> v1(ptr[0]/w, ptr[1]/w, ptr[2]/w);
   ptr = (const short*)vtx2;
   w = ptr[3];
   sglVec3<T> v2(ptr[0]/w, ptr[1]/w, ptr[2]/w);

   segset.intersectTriangle(seg, v0, v1, v2, result);
}

//----------------------------------------------------------------------------
template <class T>
void sglIntersect<T>::intersectTriangle4i(
   sglIntersect<T> &segset,
   sglSegment<T> &seg,
   const void *vtx0, const void *vtx1, const void *vtx2,
   sglBound::IntersectResultEnum &result)
{
   const int *ptr = (const int*)vtx0;
   int w = ptr[3];
   sglVec3<T> v0(ptr[0]/w, ptr[1]/w, ptr[2]/w);
   ptr = (const int*)vtx1;
   w = ptr[3];
   sglVec3<T> v1(ptr[0]/w, ptr[1]/w, ptr[2]/w);
   ptr = (const int*)vtx2;
   w = ptr[3];
   sglVec3<T> v2(ptr[0]/w, ptr[1]/w, ptr[2]/w);

   segset.intersectTriangle(seg, v0, v1, v2, result);
}

//----------------------------------------------------------------------------
template <class T>
void sglIntersect<T>::intersectTriangle4f(
   sglIntersect<T> &segset,
   sglSegment<T> &seg,
   const void *vtx0, const void *vtx1, const void *vtx2,
   sglBound::IntersectResultEnum &result)
{
   const float *ptr = (const float*)vtx0;
   float w = ptr[3];
   sglVec3<T> v0(ptr[0]/w, ptr[1]/w, ptr[2]/w);
   ptr = (const float*)vtx1;
   w = ptr[3];
   sglVec3<T> v1(ptr[0]/w, ptr[1]/w, ptr[2]/w);
   ptr = (const float*)vtx2;
   w = ptr[3];
   sglVec3<T> v2(ptr[0]/w, ptr[1]/w, ptr[2]/w);

   segset.intersectTriangle(seg, v0, v1, v2, result);
}

//----------------------------------------------------------------------------
template <class T>
void sglIntersect<T>::intersectTriangle4d(
   sglIntersect<T> &segset,
   sglSegment<T> &seg,
   const void *vtx0, const void *vtx1, const void *vtx2,
   sglBound::IntersectResultEnum &result)
{
   const double *ptr = (const double*)vtx0;
   double w = ptr[3];
   sglVec3<T> v0(ptr[0]/w, ptr[1]/w, ptr[2]/w);
   ptr = (const double*)vtx1;
   w = ptr[3];
   sglVec3<T> v1(ptr[0]/w, ptr[1]/w, ptr[2]/w);
   ptr = (const double*)vtx2;
   w = ptr[3];
   sglVec3<T> v2(ptr[0]/w, ptr[1]/w, ptr[2]/w);

   segset.intersectTriangle(seg, v0, v1, v2, result);
}


//----------------------------------------------------------------------------
template <class T>
sglBound::IntersectResultEnum
sglIntersect<T>::intersectIndexedTriangleSet(
   const sglBoxBound &bbox,
   const sglVertexPtr &vertices,
   const sglIndexPtr &index,
   unsigned int num_prims)
{
   // check against the bounding box
   sglSegment<T> seg(getSegment());
   sglBound::IntersectResultEnum result = bbox.intersect(seg);
   if (result == sglBound::eOUTSIDE)
   {
      return result;
   }

   // else result == INSIDE | INTERSECT
   // the bounding box test trims the segment so we have to restore it
   seg = getSegment();
   result = sglBound::eOUTSIDE;

   const IntersectTriangleFunc isect_func =
      getIntersectTriangleFunc(vertices.getTypeIndex());

   const sglMultiTypePtr::iterator vbegin = vertices.begin();
   const int vinc = vertices.getInc();
   sglIndexPtr::iterator viit = index.begin();
   const int viinc = index.getInc();

   const sglIndexPtr::iterator end = viit + viinc * num_prims * 3;
   while (viit != end)
   {
      const sglMultiTypePtr::iterator it1 = vbegin + vinc * *viit;
      viit += viinc;
      const sglMultiTypePtr::iterator it2 = vbegin + vinc * *viit;
      viit += viinc;
      const sglMultiTypePtr::iterator it3 = vbegin + vinc * *viit;
      viit += viinc;

      isect_func(*this, seg, it1, it2, it3, result);
   }

   if (result == sglBound::eINTERSECT)
   {
      setSegment(seg);
   }

   return result;
}


//----------------------------------------------------------------------------
template <class T>
sglBound::IntersectResultEnum
sglIntersect<T>::intersectIndexedTriangleStripSet(
   const sglBoxBound &bbox,
   const sglVertexPtr &vertices,
   const sglIndexPtr &index,
   const sglIndexPtr &lengths,
   unsigned int num_prims)
{
   // check against the bounding box
   sglSegment<T> seg(getSegment());
   sglBound::IntersectResultEnum result = bbox.intersect(seg);
   if (result == sglBound::eOUTSIDE)
   {
      return result;
   }

   // else result == INSIDE | INTERSECT
   // the bounding box test trims the segment so we have to restore it
   seg = getSegment();
   result = sglBound::eOUTSIDE;

   const IntersectTriangleFunc isect_func =
      getIntersectTriangleFunc(vertices.getTypeIndex());

   const sglMultiTypePtr::iterator vbegin = vertices.begin();
   const int vinc = vertices.getInc();
   sglIndexPtr::iterator viit = index.begin();
   const int viinc = index.getInc();

   sglIndexPtr::iterator lit = lengths.begin();
   const int linc = lengths.getInc();
   const sglIndexPtr::iterator lend = lit + linc * num_prims;
   for (; lit != lend; lit += linc)
   {
      const sglIndexType length = *lit;
      if (length >= 3)
      {
         // do it two at a time
         const sglIndexPtr::iterator end = viit + viinc * (length&(~0x1));

         sglMultiTypePtr::iterator it0 = vbegin + vinc * *viit;
         viit += viinc;
         sglMultiTypePtr::iterator it1 = vbegin + vinc * *viit;
         viit += viinc;

         while (viit != end)
         {
            const sglMultiTypePtr::iterator it2 = vbegin + vinc * *viit;
            viit += viinc;
            const sglMultiTypePtr::iterator it3 = vbegin + vinc * *viit;
            viit += viinc;

            isect_func(*this, seg, it0, it1, it2, result);
            isect_func(*this, seg, it2, it1, it3, result);

            it0 = it2;
            it1 = it3;
         }
         if (length & 0x01) //odd
         {
            const sglMultiTypePtr::iterator it2 = vbegin + vinc * *viit;
            viit += viinc;
            isect_func(*this, seg, it0, it1, it2, result);
         }
      }
      else
      {
         viit += viinc * length;
      }
   }

   if (result == sglBound::eINTERSECT)
   {
      setSegment(seg);
   }

   return result;
}


//----------------------------------------------------------------------------
template <class T>
sglBound::IntersectResultEnum
sglIntersect<T>::intersectIndexedTriangleFanSet(
   const sglBoxBound &bbox,
   const sglVertexPtr &vertices,
   const sglIndexPtr &index,
   const sglIndexPtr &lengths,
   unsigned int num_prims)
{
   // check against the bounding box
   sglSegment<T> seg(getSegment());
   sglBound::IntersectResultEnum result = bbox.intersect(seg);
   if (result == sglBound::eOUTSIDE)
   {
      return result;
   }

   // else result == INSIDE | INTERSECT
   // the bounding box test trims the segment so we have to restore it
   seg = getSegment();
   result = sglBound::eOUTSIDE;

   const IntersectTriangleFunc isect_func =
      getIntersectTriangleFunc(vertices.getTypeIndex());

   const sglMultiTypePtr::iterator vbegin = vertices.begin();
   const int vinc = vertices.getInc();
   sglIndexPtr::iterator viit = index.begin();
   const int viinc = index.getInc();

   sglIndexPtr::iterator lit = lengths.begin();
   const int linc = lengths.getInc();
   const sglIndexPtr::iterator lend = lit + linc * num_prims;
   for (; lit != lend; lit += linc)
   {
      const sglIndexType length = *lit;
      if (length >= 3)
      {
         const sglIndexPtr::iterator end = viit + viinc * length;

         const sglMultiTypePtr::iterator it0 = vbegin + vinc * *viit;
         viit += viinc;
         sglMultiTypePtr::iterator it1 = vbegin + vinc * *viit;
         viit += viinc;

         while (viit != end)
         {
            const sglMultiTypePtr::iterator it2 = vbegin + vinc * *viit;
            viit += viinc;

            isect_func(*this, seg, it0, it1, it2, result);

            it1 = it2;
         }
      }
      else
      {
         viit += viinc * length;
      }
   }

   if (result == sglBound::eINTERSECT)
   {
      setSegment(seg);
   }

   return result;
}


//----------------------------------------------------------------------------
template <class T>
sglBound::IntersectResultEnum
sglIntersect<T>::intersectIndexedQuadSet(
   const sglBoxBound &bbox,
   const sglVertexPtr &vertices,
   const sglIndexPtr &index,
   unsigned int num_prims)
{
   // check against the bounding box
   sglSegment<T> seg(getSegment());
   sglBound::IntersectResultEnum result = bbox.intersect(seg);
   if (result == sglBound::eOUTSIDE)
   {
      return result;
   }

   // else result == INSIDE | INTERSECT
   // the bounding box test trims the segment so we have to restore it
   seg = getSegment();
   result = sglBound::eOUTSIDE;

   const IntersectTriangleFunc isect_func =
      getIntersectTriangleFunc(vertices.getTypeIndex());

   const sglMultiTypePtr::iterator vbegin = vertices.begin();
   const int vinc = vertices.getInc();
   sglIndexPtr::iterator viit = index.begin();
   const int viinc = index.getInc();

   const sglIndexPtr::iterator end = viit + viinc * num_prims * 4;
   while (viit != end)
   {
      const sglMultiTypePtr::iterator it1 = vbegin + vinc * *viit;
      viit += viinc;
      const sglMultiTypePtr::iterator it2 = vbegin + vinc * *viit;
      viit += viinc;
      const sglMultiTypePtr::iterator it3 = vbegin + vinc * *viit;
      viit += viinc;
      const sglMultiTypePtr::iterator it4 = vbegin + vinc * *viit;
      viit += viinc;

      isect_func(*this, seg, it1, it2, it3, result);
      isect_func(*this, seg, it1, it3, it4, result);
   }

   if (result == sglBound::eINTERSECT)
   {
      setSegment(seg);
   }

   return result;
}


//----------------------------------------------------------------------------
template <class T>
sglBound::IntersectResultEnum
sglIntersect<T>::intersectIndexedQuadStripSet(
   const sglBoxBound &bbox,
   const sglVertexPtr &vertices,
   const sglIndexPtr &index,
   const sglIndexPtr &lengths,
   unsigned int num_prims)
{
   // check against the bounding box
   sglSegment<T> seg(getSegment());
   sglBound::IntersectResultEnum result = bbox.intersect(seg);
   if (result == sglBound::eOUTSIDE)
   {
      return result;
   }

   // else result == INSIDE | INTERSECT
   // the bounding box test trims the segment so we have to restore it
   seg = getSegment();
   result = sglBound::eOUTSIDE;

   const IntersectTriangleFunc isect_func =
      getIntersectTriangleFunc(vertices.getTypeIndex());

   const sglMultiTypePtr::iterator vbegin = vertices.begin();
   const int vinc = vertices.getInc();
   sglIndexPtr::iterator viit = index.begin();
   const int viinc = index.getInc();

   sglIndexPtr::iterator lit = lengths.begin();
   const int linc = lengths.getInc();
   const sglIndexPtr::iterator lend = lit + linc * num_prims;
   for (; lit != lend; lit += linc)
   {
      const sglIndexType length = *lit;
      if (length >= 4)
      {
         // do it two at a time
         const sglIndexPtr::iterator end = viit + viinc * (length&(~0x1));

         sglMultiTypePtr::iterator it0 = vbegin + vinc * *viit;
         viit += viinc;
         sglMultiTypePtr::iterator it1 = vbegin + vinc * *viit;
         viit += viinc;

         while (viit != end)
         {
            const sglMultiTypePtr::iterator it2 = vbegin + vinc * *viit;
            viit += viinc;
            const sglMultiTypePtr::iterator it3 = vbegin + vinc * *viit;
            viit += viinc;

            isect_func(*this, seg, it0, it1, it2, result);
            isect_func(*this, seg, it2, it1, it3, result);

            it0 = it2;
            it1 = it3;
         }
         if (length & 0x01) //odd
         {
            viit += viinc;
         }
      }
      else
      {
         viit += viinc * length;
      }
   }

   if (result == sglBound::eINTERSECT)
   {
      setSegment(seg);
   }

   return result;
}


//----------------------------------------------------------------------------
template <class T>
typename sglIntersect<T>::IntersectTriangleFunc
sglIntersect<T>::s_isect_tri_func_table[] = {
   NULL,
   NULL,
   &sglIntersect<T>::intersectTriangle2s,
   NULL,
   &sglIntersect<T>::intersectTriangle2i,
   NULL,
   &sglIntersect<T>::intersectTriangle2f,
   &sglIntersect<T>::intersectTriangle2d,
   NULL,
   NULL,
   &sglIntersect<T>::intersectTriangle3s,
   NULL,
   &sglIntersect<T>::intersectTriangle3i,
   NULL,
   &sglIntersect<T>::intersectTriangle3f,
   &sglIntersect<T>::intersectTriangle3d,
   NULL,
   NULL,
   &sglIntersect<T>::intersectTriangle4s,
   NULL,
   &sglIntersect<T>::intersectTriangle4i,
   NULL,
   &sglIntersect<T>::intersectTriangle4f,
   &sglIntersect<T>::intersectTriangle4d
};

template class SGL_DLL_API sglIntersect<double>;
template class SGL_DLL_API sglIntersect<float>;

