/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglIntersect.hpp (was sglSegSet.hpp)
 *  Created: 18 April 1999
 *  Summary: stores the intersection traversal state
 *****************************************************************************/

#ifndef __SGL_INTERSECT_HPP
#define __SGL_INTERSECT_HPP

#include <sgl.h>
#include <sglSegment.hpp>
#include <sglBound.hpp>
#include <sglProbe.hpp>
#include <sglGeoSetDataPtr.hpp>
#include <sglBoxBound.hpp>
#include <deque>

class sglGeode;
class sglDrawable;
class sglNode;

//----------------------------------------------------------------------------

template <class T>
class SGL_DLL_API sglIntersect : public sglProbeTemplate<T>
{
public: // Sub types
   class SGL_DLL_API IntersectResult : public sglProbeTemplate<T>::ProbeResult
   {
   public:
      sglBound::IntersectResultEnum m_result;
      sglSegment<T>                 m_segment;
      sglVec3f                      m_normal;
      deque<sglNode*>               m_path;      // path to picked geode

   public:
      IntersectResult() :
            sglProbeTemplate<T>::ProbeResult(),
                                 m_result(sglBound::eOUTSIDE),
                                 m_segment(sglVec3<T>(0,0,0),
                                           sglVec3<T>(0,0,0)),
                                 m_normal(0.f, 0.f, 0.f),
                                 m_path()
         {}
      virtual ~IntersectResult() { m_path.clear(); }

      virtual void reset()
         {
            m_result = sglBound::eOUTSIDE;
            m_segment.setPos(sglVec3<T>(0,0,0));
            m_segment.setDir(sglVec3<T>(0,0,0));
            m_segment.setLen(0);
            m_normal.set(0.f, 0.f, 0.f);
            m_path.clear();
            resetInternal();
         }
   };

public:
   sglIntersect();
   ~sglIntersect();

   virtual void reset();

   void setSegment(const sglSegment<T> &segment)
      { m_result.m_segment = segment; }
   const sglSegment<T>& getSegment() const
      { return m_result.m_segment; }

   void setResultEnum(sglBound::IntersectResultEnum result)
      { m_result.m_result = result; }  // Cute huh?
   sglBound::IntersectResultEnum getResultEnum() const
      { return m_result.m_result; }

   /** Returns the path to the intersected node if ePATH flag is set.
       NOTE: The nodes returned are NOT reference counted.
             If the path is to be kept for any length of time, the
             user is responsible for calling ref() on the path nodes.
   */
   deque<sglNode*>& getPath() { return m_result.m_path; }


   void setGeode(sglGeode *geode)   { m_result.m_geode = geode; }
   const sglGeode *getGeode() const { return m_result.m_geode; }

   void setGeometry(sglDrawable *geometry) { m_result.m_geometry = geometry; }
   const sglDrawable *getGeometry() const  { return m_result.m_geometry; }

   /** The modelview matrix of the geode/geometry intersected with (only when
       the eTRANSFORM flag is set).
   */
   void setTransformMatrix(const sglMat4<T> &mat) { m_result.m_xform = mat; }
   const sglMat4<T> &getTransformMatrix() const   { return m_result.m_xform; }



   /** The normal of the point on the triangle intersected with (only when the
       eNORMAL and ePRIM flags are set).
   */
   void setNormal(const sglVec3f &normal) { m_result.m_normal = normal; }
   const sglVec3f &getNormal() const      { return m_result.m_normal; }


   template <class S>
   void translate(S x, S y, S z)   { m_result.m_segment.translate(x, y, z); }

   template <class S>
   void scale(S scale)             { m_result.m_segment.scale(scale); }

   template <class S>
   void mul(const sglMat4<S> &mat) { m_result.m_segment.mul(mat); }

   void intersectTriangle(sglSegment<T>  &seg,
                          const sglVec3<T> &vertex0,
                          const sglVec3<T> &vertex1,
                          const sglVec3<T> &vertex2,
                          sglBound::IntersectResultEnum &result);

   const IntersectResult &getResult() const { return m_result; }

   // -------- push/pop transform functions (used internally) ---------
   /// scale
   template <class S>
   void pushScale(S scale,
                  S inv_scale,
                  sglSegment<T> &old_segment,
                  sglMat4<T>    &old_matrix,
                  sglFrustum<T> &old_view_frustum)
      {
         if (m_mode & (eTRANSFORM|eUNSCALE_VIEW|eUNPROJECT_VIEW|
                       eLOD_VIEW|eBILLBOARD_VIEW))
         {
            // save the original view matrix
            old_matrix = m_modelview_matrix;

            // compute the new view matrix
            m_modelview_matrix.scale(old_matrix, scale, scale, scale);

            // compute a new view frustum
            if (m_mode & eUNPROJECT_VIEW) // |eUNSCALE_VIEW
            {
               old_view_frustum = m_current_view_frustum;
               m_current_view_frustum.scale(old_view_frustum, inv_scale);
            }
         }

         // store the original segment so that i don't have to inverse
         // transform it during the pop
         old_segment = m_result.m_segment;

         // scale the original segment
         m_result.m_segment.scale(inv_scale);
      }

   /// translate
   template <class S>
   void pushTranslate(S tx, S ty, S tz,
                      sglSegment<T> &old_segment,
                      sglMat4<T>    &old_matrix,
                      sglFrustum<T> &old_view_frustum)
      {
         if (m_mode & (eTRANSFORM|eUNSCALE_VIEW|eUNPROJECT_VIEW|
                       eLOD_VIEW|eBILLBOARD_VIEW))
         {
            // save the original view matrix
            old_matrix = m_modelview_matrix;

            // compute the new view matrix
            m_modelview_matrix.translate(old_matrix, tx, ty, tz);

            // compute a new view frustum
            if (m_mode & eUNPROJECT_VIEW)
            {
               old_view_frustum = m_current_view_frustum;
               m_current_view_frustum.translate(old_view_frustum,
                                                -tx, -ty, -tz);
            }
         }

         // store the original segment so that i don't have to inverse
         // translate it during pop;
         old_segment = m_result.m_segment;

         // translate the original segment
         m_result.m_segment.translate(-tx, -ty, -tz);
      }

   /// transform
   template <class S>
   void pushTransform(const sglMat4<S> &xform,
                      const sglMat4<S> &inv_xform,
                      sglSegment<T> &old_segment,
                      sglMat4<T>    &old_matrix,
                      sglFrustum<T> &old_view_frustum)
      {
         if (m_mode & (eTRANSFORM|eUNSCALE_VIEW|eUNPROJECT_VIEW|
                       eLOD_VIEW|eBILLBOARD_VIEW))
         {
            // save the original view matrix
            old_matrix = m_modelview_matrix;

            // compute the new view matrix
            m_modelview_matrix.mul(xform, old_matrix);

            // compute a new view frustum
            if (m_mode & eUNPROJECT_VIEW)
            {
               old_view_frustum = m_current_view_frustum;
               m_current_view_frustum.mul(old_view_frustum, inv_xform);
            }
         }

         // store the original segment so that i don't have to inverse
         // transform it
         old_segment = m_result.m_segment;

         // transform the original segment
         m_result.m_segment.mul(inv_xform);
      }

   // pop the transform (restore the old values).
   void pop(sglSegment<T> &old_segment,
            sglMat4<T>    &old_matrix,
            sglFrustum<T> &old_view_frustum)
      {
         m_result.m_segment = old_segment;

         if (m_mode & (eTRANSFORM|eUNSCALE_VIEW|eUNPROJECT_VIEW|
                       eLOD_VIEW|eBILLBOARD_VIEW))
         {
            m_modelview_matrix = old_matrix;

            // restore the old view frustum
            if (m_mode & eUNPROJECT_VIEW) // |eUNSCALE_VIEW)
            {
               m_current_view_frustum = old_view_frustum;
            }
         }
      }

   typedef void (*IntersectTriangleFunc)(
      sglIntersect<T> &segset,
      sglSegment<T> &seg,
      const void *vtx0, const void *vtx1, const void *vtx2,
      sglBound::IntersectResultEnum &result);

   // Get the triangle intersection function for the given vertex data type.
   static IntersectTriangleFunc getIntersectTriangleFunc(int vertex_type)
      { return s_isect_tri_func_table[vertex_type]; }

   sglBound::IntersectResultEnum intersectIndexedTriangleSet(
      const sglBoxBound &bbox,
      const sglVertexPtr &vertices,
      const sglIndexPtr &index,
      unsigned int num_prims);

   sglBound::IntersectResultEnum intersectIndexedTriangleStripSet(
      const sglBoxBound &bbox,
      const sglVertexPtr &vertices,
      const sglIndexPtr &index,
      const sglIndexPtr &lengths,
      unsigned int num_prims);

   sglBound::IntersectResultEnum intersectIndexedTriangleFanSet(
      const sglBoxBound &bbox,
      const sglVertexPtr &vertices,
      const sglIndexPtr &index,
      const sglIndexPtr &lengths,
      unsigned int num_prims);

   sglBound::IntersectResultEnum intersectIndexedQuadSet(
      const sglBoxBound &bbox,
      const sglVertexPtr &vertices,
      const sglIndexPtr &index,
      unsigned int num_prims);

   sglBound::IntersectResultEnum intersectIndexedQuadStripSet(
      const sglBoxBound &bbox,
      const sglVertexPtr &vertices,
      const sglIndexPtr &index,
      const sglIndexPtr &lengths,
      unsigned int num_prims);

private: // not implemented
   sglIntersect(const sglIntersect &);
   sglIntersect &operator=(const sglIntersect &);

//    template <class S> void intersectTriangle2(
//       sglSegment<T> &seg,
//       const void *vtx0, const void *vtx1, const void *vtx2,
//       sglBound::IntersectResultEnum &result);
//    template <class S> void intersectTriangle3(
//       sglSegment<T> &seg,
//       const void *vtx0, const void *vtx1, const void *vtx2,
//       sglBound::IntersectResultEnum &result);
//    template <class S> void intersectTriangle4(
//       sglSegment<T> &seg,
//       const void *vtx0, const void *vtx1, const void *vtx2,
//       sglBound::IntersectResultEnum &result);

   static void intersectTriangle2s(
      sglIntersect<T> &segset,
      sglSegment<T> &seg,
      const void *vtx0, const void *vtx1, const void *vtx2,
      sglBound::IntersectResultEnum &result);
   static void intersectTriangle2i(
      sglIntersect<T> &segset,
      sglSegment<T> &seg,
      const void *vtx0, const void *vtx1, const void *vtx2,
      sglBound::IntersectResultEnum &result);
   static void intersectTriangle2f(
      sglIntersect<T> &segset,
      sglSegment<T> &seg,
      const void *vtx0, const void *vtx1, const void *vtx2,
      sglBound::IntersectResultEnum &result);
   static void intersectTriangle2d(
      sglIntersect<T> &segset,
      sglSegment<T> &seg,
      const void *vtx0, const void *vtx1, const void *vtx2,
      sglBound::IntersectResultEnum &result);
   static void intersectTriangle3s(
      sglIntersect<T> &segset,
      sglSegment<T> &seg,
      const void *vtx0, const void *vtx1, const void *vtx2,
      sglBound::IntersectResultEnum &result);
   static void intersectTriangle3i(
      sglIntersect<T> &segset,
      sglSegment<T> &seg,
      const void *vtx0, const void *vtx1, const void *vtx2,
      sglBound::IntersectResultEnum &result);
   static void intersectTriangle3f(
      sglIntersect<T> &segset,
      sglSegment<T> &seg,
      const void *vtx0, const void *vtx1, const void *vtx2,
      sglBound::IntersectResultEnum &result);
   static void intersectTriangle3d(
      sglIntersect<T> &segset,
      sglSegment<T> &seg,
      const void *vtx0, const void *vtx1, const void *vtx2,
      sglBound::IntersectResultEnum &result);
   static void intersectTriangle4s(
      sglIntersect<T> &segset,
      sglSegment<T> &seg,
      const void *vtx0, const void *vtx1, const void *vtx2,
      sglBound::IntersectResultEnum &result);
   static void intersectTriangle4i(
      sglIntersect<T> &segset,
      sglSegment<T> &seg,
      const void *vtx0, const void *vtx1, const void *vtx2,
      sglBound::IntersectResultEnum &result);
   static void intersectTriangle4f(
      sglIntersect<T> &segset,
      sglSegment<T> &seg,
      const void *vtx0, const void *vtx1, const void *vtx2,
      sglBound::IntersectResultEnum &result);
   static void intersectTriangle4d(
      sglIntersect<T> &segset,
      sglSegment<T> &seg,
      const void *vtx0, const void *vtx1, const void *vtx2,
      sglBound::IntersectResultEnum &result);

private:
   static IntersectTriangleFunc s_isect_tri_func_table[24];

   IntersectResult  m_result;
   deque<sglNode*>  m_path;      // current path
};

//----------------------------------------------------------------------------
// convenience types:
//    template class 'should' only be float or double

typedef sglIntersect<float>  sglIntersectf;
typedef sglIntersect<double> sglIntersectd;

#endif
