/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglKeyFrameAnimation.cpp
 *   Author: Scott McMillan
 *  Created: 29 May 2001
 *  Summary: key frame animation node
 *****************************************************************************/

#include <sglKeyFrameAnimation.hpp>
#include <sglCull.hpp>

//----------------------------------------------------------------------------
sglKeyFrameAnimation::sglKeyFrameAnimation()
      : sglAnimation(),
        m_start_frame(0),
        m_pause_frame(0),
        m_total_frames(0),
        m_total_frames2(0),
        m_forward_frames(0),
        m_return_frames(0)
{
}

//----------------------------------------------------------------------------
sglKeyFrameAnimation::~sglKeyFrameAnimation()
{
}

//----------------------------------------------------------------------------
bool sglKeyFrameAnimation::addChild(sglNode *node)
{
   if (sglGroup::addChild(node))
   {
      m_frame_counts.push_back(0);
      computeValues();
      return true;
   }
   else
      return false;
}

//----------------------------------------------------------------------------
bool sglKeyFrameAnimation::insertChild(unsigned int index, sglNode *child)
{
   // HACK duplicates code in sglGroup::insertChild
   if (child)
   {
      if (index < getNumChildren())
      {
         m_frame_counts.insert(m_frame_counts.begin() + index, 0);
         computeValues();
      }
      else if (index == getNumChildren())
      {
         m_frame_counts.push_back(0);
         computeValues();
      }
   }

   return sglGroup::insertChild(index, child);
}

//----------------------------------------------------------------------------
bool sglKeyFrameAnimation::removeChild(sglNode *child)
{
   // HACK duplicates code in sglGroup::removeChild
   if (child)
   {
      int index = m_child_list.size()-1;
      while(index >= 0 && m_child_list[index] != child)
         --index;

      if (index >= 0)
      {
         m_frame_counts.erase(m_frame_counts.begin() + index);
         computeValues();
      }
   }

   return sglGroup::removeChild(child);
}

//----------------------------------------------------------------------------
sglNode *sglKeyFrameAnimation::removeChild(unsigned int index)
{
   // HACK duplicates some code in sglGroup::removeChild()
   if (index < getNumChildren())
   {
      m_frame_counts.erase(m_frame_counts.begin() + index);
      computeValues();
   }

   return sglGroup::removeChild(index);
}

//----------------------------------------------------------------------------
bool sglKeyFrameAnimation::setDuration(sglNode *node, unsigned int num_frames)
{
   bool success = false;

   // replace duration for FIRST occurence of node
   int index = searchChild(node);
   if (index != -1)
   {
      m_frame_counts[index] = num_frames;
      computeValues();

      success = true;
   }

   return success;
}

//----------------------------------------------------------------------------
unsigned int sglKeyFrameAnimation::getDuration(sglNode *node) const
{
   unsigned int num_frames = 0;
   int index = searchChild(node);
   if (index != -1)
   {
      num_frames = m_frame_counts[index];
   }

   return num_frames;
}

//----------------------------------------------------------------------------
bool sglKeyFrameAnimation::setDuration(unsigned int index,
                                       unsigned int num_frames)
{
   bool success = false;

   if (index < m_frame_counts.size())
   {
      m_frame_counts[index] = num_frames;
      computeValues();

      success = true;
   }

   return success;
}

//----------------------------------------------------------------------------
unsigned int sglKeyFrameAnimation::getDuration(unsigned int index) const
{
   unsigned int num_frames = 0;

   if (index < m_frame_counts.size())
   {
      num_frames = m_frame_counts[index];
   }
   return num_frames;
}

//----------------------------------------------------------------------------
void sglKeyFrameAnimation::computeValues()
{
   unsigned int ix;
   m_total_frames = 0;
   m_forward_frames = 0;
   m_return_frames = 0;
   if (m_range_flag)
   {
      for (ix=m_first; ix<m_last+1; ++ix)
      {
         if (ix < m_frame_counts.size())
            m_total_frames += m_frame_counts[ix];
      }

      for (ix=m_first; ix<m_last; ++ix)
      {
         if (ix < m_frame_counts.size())
            m_forward_frames += m_frame_counts[ix];
      }
      for (ix=m_first+1; ix<m_last+1; ++ix)
      {
         if (ix < m_frame_counts.size())
            m_return_frames += m_frame_counts[ix];
      }
   }
   else if (!m_frame_counts.empty())
   {
      unsigned int fcsize = m_frame_counts.size();
      for (ix=0; ix<fcsize; ++ix)
      {
         m_total_frames += m_frame_counts[ix];
      }

      for (ix=0; ix<fcsize-1; ++ix)
      {
         m_forward_frames += m_frame_counts[ix];
      }
      for (ix=1; ix<fcsize; ++ix)
      {
         m_return_frames += m_frame_counts[ix];
      }
   }

   if (m_num_cycles > 0)
   {
      m_total_frames2 = m_total_frames +
         (m_num_cycles - 1)*(m_forward_frames + m_return_frames);
      if (m_reverse_flag)
      {
         m_total_frames2 += m_return_frames;
      }
      else
      {
         m_total_frames2 += m_forward_frames;
      }
   }
}

//----------------------------------------------------------------------------
void sglKeyFrameAnimation::preDraw(const PreDrawStruct &trav_state)
{
   if (m_state_dirty_flag)
   {
      switch (m_state)
      {
         case eSTART:
            m_start_frame = trav_state.m_frame_count;
            if (m_frame_offset < m_start_frame)
                m_start_frame -= m_frame_offset;
            break;

         case ePAUSE:
            m_pause_frame = trav_state.m_frame_count;
            break;

         case eRESUME:
            // advance the start time
            m_start_frame += (trav_state.m_frame_count - m_pause_frame);
            m_state = eSTART;  // necessary?  yes
            m_pause_frame = 0;
            break;

         case eSTOP:
            m_start_frame = m_pause_frame = 0;
            break;
      }
      m_state_dirty_flag = false;
   }

   sglGroup::preDraw(trav_state);
}

//----------------------------------------------------------------------------
sglNode *sglKeyFrameAnimation::computeChild(int frame) const
{
   float atime = (float)frame*m_speed;
   if (m_mode == eCYCLE)
   {
      if (m_num_cycles == 0 || atime < m_total_frames*m_num_cycles)
         atime = atime - ((int)(atime/m_total_frames))*m_total_frames;
      else
         return NULL;
   }
   else // m_mode == eSWING
   {
      unsigned int cycle_frames = m_forward_frames + m_return_frames;

      if (m_num_cycles == 0 || atime < m_total_frames2)
         atime = atime - ((int)(atime/cycle_frames))*cycle_frames;
      else
         return NULL;

      // back swing?
      if (atime >= m_total_frames)
      {
         atime = 2.f*m_total_frames - atime;
         if (m_range_flag && m_first != m_last)
         {
            atime -= m_frame_counts[m_last];
         }
         else
         {
            atime -= m_frame_counts[m_child_list.size()-1];
         }
      }
   }

   unsigned int index;
   if (!m_reverse_flag)
   {
      if (!m_range_flag)
      {
         index = 0;
      }
      else
      {
         index = m_first;
      }

      atime -= m_frame_counts[index];
      while (atime > 0)
      {
         ++index;
         atime -= m_frame_counts[index];
      }
   }
   else
   {
      if (!m_range_flag)
      {
         index = m_child_list.size() - 1;
      }
      else
      {
         index = m_last;
      }

      atime -= m_frame_counts[index];
      while (atime > 0)
      {
         --index;
         atime -= m_frame_counts[index];
      }
   }

   return m_child_list[index];
}

//----------------------------------------------------------------------------
sglNode *sglKeyFrameAnimation::selectChild(unsigned int current_frame) const
{
   // the following are to be checked by the functions that call this one so
   // that this one does not have to do it again;
   //if (m_state == eSTOP || m_total_frames == 0 || m_speed == 0.f)
   //{
   //   return NULL;
   //}
   sglNode* child = NULL;
   if (m_state == ePAUSE)
   {
      child = computeChild((int)(m_pause_frame - m_start_frame));
   }
   else  // mode is eRESUME or eSTART
   {
      child = computeChild((int)(current_frame - m_start_frame));
   }

   return child;
}

//----------------------------------------------------------------------------
sglBound::IntersectResultEnum sglKeyFrameAnimation::intersect(
   sglIntersectf &isector) const
{
   if ((isector.getMode()&sglProbe::eANIMATION_NONE) ||
       m_state == eSTOP || m_total_frames == 0 || m_speed == 0.f ||
       m_child_list.empty())
   {
      return sglBound::eOUTSIDE;
   }
   else if (isector.getMode()&sglProbe::eANIMATION_ALL)
   {
      return sglGroup::intersect(isector);
   }
   else // default: current child selected by animation only
   {
      sglSegmentf seg(isector.getSegment());
      if (m_sphere.intersect(seg) == sglBound::eOUTSIDE)
      {
         return sglBound::eOUTSIDE;
      }

      sglBound::IntersectResultEnum result = sglBound::eOUTSIDE;
      sglNode *child = selectChild(isector.getFrameCount());
      if (child)
      {
         result = child->intersect(isector);
      }

      if (result == sglBound::eINTERSECT &&
          isector.getMode() & sglProbe::ePATH)
      {
         isector.getPath().push_front((sglNode*)this);
      }
      return result;
   }
}

//----------------------------------------------------------------------------
sglBound::IntersectResultEnum sglKeyFrameAnimation::intersect(
   sglIntersectd &isector) const
{
   if ((isector.getMode()&sglProbe::eANIMATION_NONE) ||
       m_state == eSTOP || m_total_frames == 0 || m_speed == 0.f ||
       m_child_list.empty())
   {
      return sglBound::eOUTSIDE;
   }
   else if (isector.getMode()&sglProbe::eANIMATION_ALL)
   {
      return sglGroup::intersect(isector);
   }
   else // default: current child selected by switch only
   {
      sglSegmentd seg(isector.getSegment());
      if (m_sphere.intersect(seg) == sglBound::eOUTSIDE)
      {
         return sglBound::eOUTSIDE;
      }

      sglBound::IntersectResultEnum result = sglBound::eOUTSIDE;
      sglNode *child = selectChild(isector.getFrameCount());
      if (child)
      {
         result = child->intersect(isector);
      }

      if (result == sglBound::eINTERSECT &&
          isector.getMode() & sglProbe::ePATH)
      {
         isector.getPath().push_front((sglNode*)this);
      }
      return result;
   }
}

//----------------------------------------------------------------------------
void sglKeyFrameAnimation::pick(sglPickf &pick_state,
                                unsigned int cull_flags) const
{
   if ((pick_state.getMode()&sglProbe::eANIMATION_NONE) ||
       m_state == eSTOP || m_total_frames == 0 || m_speed == 0.f ||
       m_child_list.empty())
   {
      return;  // select nothing
   }
   else if (pick_state.getMode()&sglProbe::eANIMATION_ALL)
   {
      sglGroup::pick(pick_state, cull_flags);
   }
   else // default: current child selected by animation only
   {
      if (cull_flags && (m_sphere.intersect(pick_state.getPickFrustum(),
                                            cull_flags) == sglBound::eOUTSIDE))
      {
         return; // select nothing
      }

      sglNode *child = selectChild(pick_state.getFrameCount());

      if (child)
      {
         // if path is requested store it ahead of time (different from
         // intersect), in case something beneath here is picked successfully.
         if (pick_state.getMode() & sglProbe::ePATH)
         {
            pick_state.getPath().push_back((sglNode*)this);
         }

         child->pick(pick_state, cull_flags);

         if (pick_state.getMode() & sglProbe::ePATH)
         {
            pick_state.getPath().pop_back();
         }
      }
   }
}

//----------------------------------------------------------------------------
void sglKeyFrameAnimation::pick(sglPickd &pick_state,
                                unsigned int cull_flags) const
{
   if ((pick_state.getMode()&sglProbe::eANIMATION_NONE) ||
       m_state == eSTOP || m_total_frames == 0 || m_speed == 0.f ||
       m_child_list.empty())
   {
      return;  // select nothing
   }
   else if (pick_state.getMode()&sglProbe::eANIMATION_ALL)
   {
      sglGroup::pick(pick_state, cull_flags);
   }
   else // default: current child selected by switch only
   {
      if (cull_flags && (m_sphere.intersect(pick_state.getPickFrustum(),
                                            cull_flags) == sglBound::eOUTSIDE))
      {
         return; // select nothing
      }

      sglNode *child = selectChild(pick_state.getFrameCount());

      if (child)
      {
         // if path is requested store it ahead of time (different from
         // intersect), in case something beneath here is picked successfully.
         if (pick_state.getMode() & sglProbe::ePATH)
         {
            pick_state.getPath().push_back((sglNode*)this);
         }

         child->pick(pick_state, cull_flags);

         if (pick_state.getMode() & sglProbe::ePATH)
         {
            pick_state.getPath().pop_back();
         }
      }
   }
}

//----------------------------------------------------------------------------
void sglKeyFrameAnimation::cull(sglCull<float> &trav_state,
                                unsigned int cull_flags) const
{
   if (m_state == eSTOP || m_total_frames == 0 || m_speed == 0.f ||
       m_child_list.empty() ||
       (cull_flags && (m_sphere.intersect(trav_state.getFrustum(),
                                          cull_flags) == sglBound::eOUTSIDE)))
   {
      return; // draw nothing
   }

   sglNode *child = selectChild(trav_state.getFrameCount());

   // select child based on range
   if (child)
   {
      child->cull(trav_state, cull_flags);
   }
}

//----------------------------------------------------------------------------
void sglKeyFrameAnimation::cull(sglCull<double> &trav_state,
                                unsigned int cull_flags) const
{
   if (m_state == eSTOP || m_total_frames == 0 || m_speed == 0.f ||
       m_child_list.empty() ||
       (cull_flags && (m_sphere.intersect(trav_state.getFrustum(),
                                          cull_flags) == sglBound::eOUTSIDE)))
   {
      return; // draw nothing
   }

   sglNode *child = selectChild(trav_state.getFrameCount());

   // select child based on range
   if (child)
   {
      child->cull(trav_state, cull_flags);
   }
}

//----------------------------------------------------------------------------
sglNode *sglKeyFrameAnimation::clone(unsigned int mode) const
{
   sglKeyFrameAnimation *root = new sglKeyFrameAnimation;
   sglKeyFrameAnimation::copyTo(root, mode);
   return root;
}

//----------------------------------------------------------------------------
void sglKeyFrameAnimation::copyTo(sglKeyFrameAnimation *dst,
                                  unsigned int mode) const
{
   sglAnimation::copyTo(dst, mode);

   // transfer stuff
   for (unsigned int index=0; index<getNumChildren(); ++index)
   {
      dst->setDuration(index, getDuration(index));
   }
}
//----------------------------------------------------------------------------
void sglKeyFrameAnimation::printInfo(ostream &ostrm,
                                     const char *indent_string) const
{
   sglAnimation::printInfo(ostrm, indent_string);
   ostrm << indent_string << "frame durations = [";
   for (unsigned int i=0; i<m_frame_counts.size(); ++i)
   {
      ostrm << " " << m_frame_counts[i];
   }
   ostrm << " ]" << endl;
}
