/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglKeyFrameAnimation.hpp
 *  Created: 29 May 2001
 *  Summary: animation based on frame.
 *****************************************************************************/

#ifndef __SGL_KEY_FRAME_ANIMATION_HPP
#define __SGL_KEY_FRAME_ANIMATION_HPP

#include <sgl.h>
#include <sglAnimation.hpp>

//----------------------------------------------------------------------------

/**

A subclass of sglAnimation, this node selects one child as a function of the
current frame count (not time) and the animation mode.  As each child is added
to this node the default number of frames it is to be displayed is set to
zero.  This value can be changed by calling one of the setDuration(..) member
functions and identifying the child either by pointer or index and specifying
the number of frames.

<b>Example:</b>

<pre>
   sglKeyFrameAnimation *animation = new sglKeyFrameAnimation;
   for (unsigned int ix=0; ix<10; ++ix)
   {
      sglGeode *geode = new sglGeode;
      animation->addChild(geode);
      animation->setDuration(ix, 10);  // display for 10 frames
   }
   animation->setMode(sglAnimation::eSWING);  // back and forth
   animation->setRange(3, 8);                 // display 3...8
   animation->setSpeed(2.5f);                 // 2.5 times faster (4 fr. each)
   animation->setNumCycles(0);                // continuous

   // set the animation to start on the next frame (after preDraw()).
   animation->setState(sglAnimation::eSTART);
</pre>

 */

//----------------------------------------------------------------------------

class SGL_DLL_API sglKeyFrameAnimation : public sglAnimation
{
public:
   /// default constructor
   sglKeyFrameAnimation();
   /// virtual destructor
   virtual ~sglKeyFrameAnimation();

   // ======== members from sglGroup that need to be augmented =========
   virtual bool addChild(sglNode *node);
   virtual bool insertChild(unsigned int index, sglNode *node);
   virtual bool removeChild(sglNode *node);
   virtual sglNode *removeChild(unsigned int index);
   // ==================================================================

   /** Set a frame offset when starting the animation.  Only positive offsets
       are allowed, which will start the animation further forward into the
       sequence.  This will only take effect the next time the state is changed
       to START.
       @param offset_frame Offset in number of frames
   */
   void setOffsetFrame(unsigned int frame_offset)
        { m_frame_offset = frame_offset; }

   /** Query the current frame offset.
       @return The current frame offset.
   */
   unsigned int getOffsetFrame() const { return m_frame_offset; }

   /** Sets the duration, in number of frames, of the first child in the list
       that matches the pointer.
       @param node Pointer to child to set the duration.
       @param num_frames Number of frames to display the child.  Setting to
         zero disables the display of this particular child.
       @return true if the node pointer is one of this node's children;
         otherwise false
   */
   bool setDuration(sglNode *node, unsigned int num_frames);

   /** Gets the current duration of the child specified by the node pointer.
       @param node Pointer to child to get the duration from.
       @return duration of the specified node (zero if the node was not a
       child).
   */
   unsigned int getDuration(sglNode *node) const;

   /** Sets the duration, in number of frames, of the child specified by the
       index.
       @param index Index of child to set the duration.
       @param num_frames Number of frames to display the child.  Setting to
         zero disables the display of this particular child.
       @return true if the index is valid; otherwise false
   */
   bool setDuration(unsigned int index, unsigned int num_frames);

   /** Gets the current duration of the child corresponding to the index.
       @param index Index of child to set the duration.
       @return duration of the specified node (zero if the index was out of
         range.
   */
   unsigned int getDuration(unsigned int index) const;

   // see sglNode::intersect(f) for details.
   virtual sglBound::IntersectResultEnum intersect(
      sglIntersectf &isector) const;

   // see sglNode::intersect(d) for details.
   virtual sglBound::IntersectResultEnum intersect(
      sglIntersectd &isector) const;

   // see sglNode::pick(f) for details.
   virtual void pick(sglPickf &pick_state, unsigned int cull_flags) const;

   // see sglNode::pick(d) for details.
   virtual void pick(sglPickd &pick_state, unsigned int cull_flags) const;

   // See sglNode::preDraw() for details
   virtual void preDraw(const PreDrawStruct &trav_state);

   // see sglNode::clone() for details.
   virtual sglNode *clone(unsigned int mode) const;

   // see sglObject::printInfo() for details.
   virtual void printInfo(ostream &ostrm, const char *indent_string) const;

protected:
   // sglNode::cull(f) for details
   virtual void cull(sglCull<float> &trav_state,
                     unsigned int cull_flags) const;

   // sglNode::cull(d) for details
   virtual void cull(sglCull<double> &trav_state,
                     unsigned int cull_flags) const;

   virtual void computeValues();
   sglNode* computeChild(int frame) const;
   sglNode* selectChild(unsigned int current_frame) const;

   void copyTo(sglKeyFrameAnimation *dst, unsigned int mode) const;

private:     // not implemented
   sglKeyFrameAnimation(const sglKeyFrameAnimation &);
   sglKeyFrameAnimation &operator=(const sglKeyFrameAnimation &);

protected:
   vector<unsigned int> m_frame_counts;

   unsigned int m_start_frame;      // index of frame this node was started on
   unsigned int m_pause_frame;      // index of frame this node was paused on
   unsigned int m_frame_offset;     // initial frame offset when starting the
                                    //   animation

   unsigned int m_total_frames;     // for internal use during eCYCLE

   unsigned int m_total_frames2;
   unsigned int m_forward_frames;   // for internal use during eSWING
   unsigned int m_return_frames;
};

#endif
