/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglLOD.cpp
 *  Created: 14 November 1998
 *  Summary:
 *****************************************************************************/

#include <sgl.h>
#include <sglVector.hpp>
#include <sglMatrix.hpp>

#include <sglLOD.hpp>
#include <sglCull.hpp>

float sglLOD::s_range_scale = 1.0f;

//----------------------------------------------------------------------------
sglLOD::sglLOD()
      : sglGroup(),
        m_center(0.0f, 0.0f, 0.0f),
        m_lod_state(NULL)
{
}

//----------------------------------------------------------------------------
sglLOD::~sglLOD()
{
   if (m_lod_state)
   {
       m_lod_state->deref();
       if (m_lod_state->getRefCount() == 0)
       {
           delete m_lod_state;
       }
       m_lod_state = NULL;
   }
}

//----------------------------------------------------------------------------
void sglLOD::setRange(unsigned int index, float range)
{
   if (range >= 0.0)
   {
      // expand range vector if needed
      for (unsigned int i=m_ranges.size(); i<=index; i++)
         m_ranges.push_back(-1.0);

      m_ranges[index] = range;
   }
}

//----------------------------------------------------------------------------
float sglLOD::getRange(unsigned int index) const
{
   if (index < m_ranges.size())
      return m_ranges[index];
   else
      return -1.0;
}

//----------------------------------------------------------------------------
void sglLOD::setLODState(sglLODState *lod_state)
{
   if (m_lod_state != lod_state)
   {
      if (m_lod_state)
      {
         m_lod_state->deref();
         if (m_lod_state->getRefCount() == 0)
         {
            delete m_lod_state;
         }
         m_lod_state = NULL;
      }

      if (lod_state)
      {
         m_lod_state = lod_state;
         m_lod_state->ref();
      }
   }
}

//----------------------------------------------------------------------------
sglBound::IntersectResultEnum sglLOD::intersect(sglIntersectf &segset) const
{
   if (segset.getMode()&sglProbe::eLOD_NONE || m_child_list.empty())
   {
      return sglBound::eOUTSIDE;
   }
   else if (segset.getMode()&sglProbe::eLOD_ALL)
   {
      return sglGroup::intersect(segset);
   }
   else if (segset.getMode()&sglProbe::eLOD_VIEW)
   {
      sglSegmentf seg(segset.getSegment());
      sglBound::IntersectResultEnum result = m_sphere.intersect(seg);
      if (result == sglBound::eOUTSIDE)
      {
         return result;
      }

      float range = computeRange(segset.getLODScale(),
                                 segset.getLODOffset(),
                                 segset.getModelViewMatrix());

      // select child based on range
      result = sglBound::eOUTSIDE;
      for (unsigned int i=m_ranges.size()-1; i>0; i--)
      {
         if ((range >= m_ranges[i-1]) && (range < m_ranges[i]))
         {
            if (i-1 < getNumChildren())
            {
               if (m_child_list[i-1]->intersect(segset) ==
                   sglBound::eINTERSECT)
               {
                  result = sglBound::eINTERSECT;
               }
            }
            break;
         }
      }

      if (result == sglBound::eINTERSECT &&
          segset.getMode() & sglProbe::ePATH)
      {
         segset.getPath().push_front((sglNode*)this);
      }
      return result;
   }
   else // default: intersect against range 0 only
   {
      sglSegmentf seg(segset.getSegment());
      sglBound::IntersectResultEnum result = m_sphere.intersect(seg);
      if (result == sglBound::eOUTSIDE)
      {
         return result;
      }

      result = sglBound::eOUTSIDE;
      if (m_child_list[0]->intersect(segset) == sglBound::eINTERSECT)
      {
         result = sglBound::eINTERSECT;
      }

      if (result == sglBound::eINTERSECT &&
          segset.getMode() & sglProbe::ePATH)
      {
         segset.getPath().push_front((sglNode*)this);
      }
      return result;
   }
}

//----------------------------------------------------------------------------
sglBound::IntersectResultEnum sglLOD::intersect(sglIntersectd &segset) const
{
   if (segset.getMode()&sglProbe::eLOD_NONE || m_child_list.empty())
   {
      return sglBound::eOUTSIDE;
   }
   else if (segset.getMode()&sglProbe::eLOD_ALL)
   {
      return sglGroup::intersect(segset);
   }
   else if (segset.getMode()&sglProbe::eLOD_VIEW)
   {
      sglSegmentd seg(segset.getSegment());
      sglBound::IntersectResultEnum result = m_sphere.intersect(seg);
      if (result == sglBound::eOUTSIDE)
      {
         return result;
      }

      float range = computeRange(segset.getLODScale(),
                                 segset.getLODOffset(),
                                 segset.getModelViewMatrix());

      // select child based on range
      result = sglBound::eOUTSIDE;
      for (unsigned int i=m_ranges.size()-1; i>0; i--)
      {
         if ((range >= m_ranges[i-1]) && (range < m_ranges[i]))
         {
            if (i-1 < getNumChildren())
            {
               if (m_child_list[i-1]->intersect(segset) ==
                   sglBound::eINTERSECT)
               {
                  result = sglBound::eINTERSECT;
               }
            }
            break;
         }
      }

      if (result == sglBound::eINTERSECT &&
          segset.getMode() & sglProbe::ePATH)
      {
         segset.getPath().push_front((sglNode*)this);
      }
      return result;
   }
   else // default: intersect against range 0 only
   {
      sglSegmentd seg(segset.getSegment());
      sglBound::IntersectResultEnum result = m_sphere.intersect(seg);
      if (result == sglBound::eOUTSIDE)
      {
         return result;
      }

      result = sglBound::eOUTSIDE;
      if (m_child_list[0]->intersect(segset) == sglBound::eINTERSECT)
      {
         result = sglBound::eINTERSECT;
      }

      if (result == sglBound::eINTERSECT &&
          segset.getMode() & sglProbe::ePATH)
      {
         segset.getPath().push_front((sglNode*)this);
      }
      return result;
   }
}

//----------------------------------------------------------------------------
void sglLOD::pick(sglPickf &pick_state, unsigned int cull_flags) const
{
   if (pick_state.getMode()&sglProbe::eLOD_NONE || m_child_list.empty())
   {
      return; // select nothing
   }
   else if (pick_state.getMode()&sglProbe::eLOD_ALL)
   {
      sglGroup::pick(pick_state, cull_flags);
      return;
   }
   else if (pick_state.getMode()&sglProbe::eLOD_VIEW)
   {
      // may want to do the range check first if more efficient???

      if (cull_flags && (m_sphere.intersect(pick_state.getPickFrustum(),
                                            cull_flags) == sglBound::eOUTSIDE))
      {
         return; // select nothing
      }

      float range = computeRange(pick_state.getLODScale(),
                                 pick_state.getLODOffset(),
                                 pick_state.getModelViewMatrix());

      // select child based on range
      for (unsigned int i=m_ranges.size()-1; i>0; i--)
      {
         if ((range >= m_ranges[i-1]) && (range < m_ranges[i]))
         {
            if (i-1 < getNumChildren())
            {
               // if path is requested store it ahead of time (different from
               // intersect), in case something beneath here is picked
               // successfully.
               if (pick_state.getMode() & sglProbe::ePATH)
               {
                  pick_state.getPath().push_back((sglNode*)this);
               }

               m_child_list[i-1]->pick(pick_state, cull_flags);

               if (pick_state.getMode() & sglProbe::ePATH)
               {
                  pick_state.getPath().pop_back();
               }
            }
            break;
         }
      }
   }
   else // default: pick against range 0 only
   {
      if (cull_flags && (m_sphere.intersect(pick_state.getPickFrustum(),
                                            cull_flags) == sglBound::eOUTSIDE))
      {
         return; // select nothing
      }

      // if path is requested store it ahead of time (different from
      // intersect), in case something beneath here is picked
      // successfully.
      if (pick_state.getMode() & sglProbe::ePATH)
      {
         pick_state.getPath().push_back((sglNode*)this);
      }

      m_child_list[0]->pick(pick_state, cull_flags);

      if (pick_state.getMode() & sglProbe::ePATH)
      {
         pick_state.getPath().pop_back();
      }
   }
}

//----------------------------------------------------------------------------
void sglLOD::pick(sglPickd &pick_state, unsigned int cull_flags) const
{
   if (pick_state.getMode()&sglProbe::eLOD_NONE || m_child_list.empty())
   {
      return;  // select nothing
   }
   else if (pick_state.getMode()&sglProbe::eLOD_ALL)
   {
      sglGroup::pick(pick_state, cull_flags);
      return;
   }
   else if (pick_state.getMode()&sglProbe::eLOD_VIEW)
   {
      // may want to do the range check first if more efficient???

      if (cull_flags && (m_sphere.intersect(pick_state.getPickFrustum(),
                                            cull_flags) == sglBound::eOUTSIDE))
      {
         return; // select nothing
      }

      float range = computeRange(pick_state.getLODScale(),
                                 pick_state.getLODOffset(),
                                 pick_state.getModelViewMatrix());

      // select child based on range
      for (unsigned int i=m_ranges.size()-1; i>0; i--)
      {
         if ((range >= m_ranges[i-1]) && (range < m_ranges[i]))
         {
            if (i-1 < getNumChildren())
            {
               // if path is requested store it ahead of time (different from
               // intersect), in case something beneath here is picked
               // successfully.
               if (pick_state.getMode() & sglProbe::ePATH)
               {
                  pick_state.getPath().push_back((sglNode*)this);
               }

               m_child_list[i-1]->pick(pick_state, cull_flags);

               if (pick_state.getMode() & sglProbe::ePATH)
               {
                  pick_state.getPath().pop_back();
               }
            }
            break;
         }
      }
   }
   else // default: pick against range 0 only
   {
      if (cull_flags && (m_sphere.intersect(pick_state.getPickFrustum(),
                                            cull_flags) == sglBound::eOUTSIDE))
      {
         return; // select nothing
      }

      // if path is requested store it ahead of time (different from
      // intersect), in case something beneath here is picked
      // successfully.
      if (pick_state.getMode() & sglProbe::ePATH)
      {
         pick_state.getPath().push_back((sglNode*)this);
      }

      m_child_list[0]->pick(pick_state, cull_flags);

      if (pick_state.getMode() & sglProbe::ePATH)
      {
         pick_state.getPath().pop_back();
      }
   }
}

//----------------------------------------------------------------------------
void sglLOD::cull(sglCull<float> &trav_state,
                  unsigned int cull_flags) const
{
   // may want to do the range check first if more efficient???

   if (m_child_list.empty() ||
       (cull_flags &&
        (m_sphere.intersect(trav_state.getFrustum(),
                            cull_flags) == sglBound::eOUTSIDE)))
   {
      return; // draw nothing
   }

   float range = computeRange(trav_state.getLODScale(),
                              trav_state.getLODOffset(),
                              trav_state.getModelViewMatrix());

   // select child based on range
   for (unsigned int i=m_ranges.size()-1; i>0; i--)
   {
      if ((range >= m_ranges[i-1]) && (range < m_ranges[i]))
      {
         if (i-1 < getNumChildren())
         {
            m_child_list[i-1]->cull(trav_state, cull_flags);
         }
         break;
      }
   }
}

//----------------------------------------------------------------------------
void sglLOD::cull(sglCull<double> &trav_state,
                  unsigned int cull_flags) const
{
   // may want to do the range check first if more efficient???

   if (m_child_list.empty() ||
       (cull_flags &&
        (m_sphere.intersect(trav_state.getFrustum(),
                            cull_flags) == sglBound::eOUTSIDE)))
   {
      return; // draw nothing
   }

   float range = computeRange(trav_state.getLODScale(),
                              trav_state.getLODOffset(),
                              trav_state.getModelViewMatrix());

   // select child based on range
   for (unsigned int i=m_ranges.size()-1; i>0; i--)
   {
      if ((range >= m_ranges[i-1]) && (range < m_ranges[i]))
      {
         if (i-1 < getNumChildren())
         {
            m_child_list[i-1]->cull(trav_state, cull_flags);
         }
         break;
      }
   }
}

//----------------------------------------------------------------------------
float sglLOD::computeRange(float trav_lod_scale,
                           float trav_lod_offset,
                           const sglMat4f &view_matrix) const
{
   sglVec3d offset;
   for (unsigned int j=0; j<3; j++)
   {
      offset[j] =
         view_matrix[0][j]*m_center[0] +
         view_matrix[1][j]*m_center[1] +
         view_matrix[2][j]*m_center[2] +
         view_matrix[3][j];
   }

   // HACK - not really sure where the FOV-based LOD scale (modified by any
   // sglLODStateNodes along the way) is supposed to be applied.  Multiplying
   // the range scale portion for now.  Not really sure about the offset either

   float range;
   if (m_lod_state)
   {
      range = trav_lod_scale *
         (s_range_scale * (m_lod_state->getRangeScale() *
                           sqrt(offset[0]*offset[0] +
                                offset[1]*offset[1] +
                                offset[2]*offset[2]) +
                           m_lod_state->getRangeOffset())) +
         trav_lod_offset;
   }
   else
   {
      range = trav_lod_scale *
         (s_range_scale * sqrt(offset[0]*offset[0] +
                               offset[1]*offset[1] +
                               offset[2]*offset[2])) +
         trav_lod_offset;
   }

   return range;
}

//----------------------------------------------------------------------------
float sglLOD::computeRange(float trav_lod_scale,
                           float trav_lod_offset,
                           const sglMat4d &view_matrix) const
{
   sglVec3d offset;
   for (unsigned int j=0; j<3; j++)
   {
      offset[j] =
         view_matrix[0][j]*m_center[0] +
         view_matrix[1][j]*m_center[1] +
         view_matrix[2][j]*m_center[2] +
         view_matrix[3][j];
   }

   // HACK - not really sure where the FOV-based LOD scale (modified by any
   // sglLODStateNodes along the way) is supposed to be applied.  Multiplying
   // the range scale portion for now.  Not really sure about the offset either

   float range;
   if (m_lod_state)
   {
      range = trav_lod_scale *
         (s_range_scale * (m_lod_state->getRangeScale() *
                           sqrt(offset[0]*offset[0] +
                                offset[1]*offset[1] +
                                offset[2]*offset[2]) +
                           m_lod_state->getRangeOffset())) +
         trav_lod_offset;
   }
   else
   {
      range = trav_lod_scale *
         (s_range_scale * sqrt(offset[0]*offset[0] +
                               offset[1]*offset[1] +
                               offset[2]*offset[2])) +
         trav_lod_offset;
   }

   return range;
}

//----------------------------------------------------------------------------
sglNode *sglLOD::clone(unsigned int mode) const
{
   sglLOD *root = new sglLOD;
   sglLOD::copyTo(root, mode);
   return root;
}

//----------------------------------------------------------------------------
void sglLOD::copyTo(sglLOD *dst, unsigned int mode) const
{
   sglGroup::copyTo(dst, mode);

   // transfer stuff
   for (unsigned int index=0; index<m_ranges.size(); ++index)
   {
      dst->setRange(index, m_ranges[index]);
   }
   dst->setCenter(m_center);
   dst->setLODState(m_lod_state);
}

//----------------------------------------------------------------------------
void sglLOD::printInfo(ostream &ostrm, const char *indent_string) const
{
   sglGroup::printInfo(ostrm, indent_string);
   ostrm << indent_string << "sglLOD: " << m_ranges.size()
         << " ranges" << endl;
   ostrm << indent_string << "\t center: "
         << m_center[0] << ' ' << m_center[1] << ' ' << m_center[2] << endl;
   for (unsigned int i=0; i<m_ranges.size(); i++)
   {
      ostrm << indent_string
            << "\t Range " << i << ": " << m_ranges[i] << endl;
   }
   if (m_lod_state)
   {
       m_lod_state->printInfo(ostrm, indent_string);
   }
}
