/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglLOD.hpp
 *  Created: 14 November 1998
 *  Summary: select (or not) the children to be traversed based on range
 *****************************************************************************/

#ifndef __SGL_LOD_HPP
#define __SGL_LOD_HPP

#include <sgl.h>
#include <vector>

#include <sglVector.hpp>
#include <sglGroup.hpp>
#include <sglLODState.hpp>

//----------------------------------------------------------------------------

/**

This subclass of sglGroup selects the child to be traversed based on a function
of its range from the eye point (the "origin" of the frustum) and what is
defined as the center of this node.  This node is useful if you have multiple
versions of a single model that are successively more refined.  The highest
fidelity model should be added as child 0 and the lowest fidelity model should
be added as the last child.  The result is that the closer you are to the
"center" of this node the higher the resolution of the selected child.

After the children have been added, the user specifies a list of ranges
and specifies the center point which is a reference point relative to the
origin of the local coordinate system of this node used to compute the range to
the viewpoint.  This range value can also be scaled and biased to change which
child is selected.  As with sglSwitch, the order that the children are added to
the node is significant and tied to the ranges.

<b>Ranges</b>

The ranges are specified one element at a time along with its corresponding
index in the range vector.  The vector of ranges should be positive and
monotonically increasing with the indices; however, no error checking is
performed on this.  Likewise if indices are skipped the missing ranges will be
initialized to -1.0 (an invalid range).  If the calculated range falls between
two values in the vector of ranges, a single child is selected for further
traversal according to the following table:

<pre>
         Range Value, x        Child selected:
 ---------------------------   ---------------
               x < range[0]         none
   range[0] <= x < range[1]       child[0]
   range[1] <= x < range[2]       child[1]
             :                       :
 range[n-1] <= x < range[n]       child[n-1]
   range[n] <= x                    none
</pre>

Note that if there are n+1 range values, then there should be n children.
If there are more than n children, the extra ones will never be selected.  If
there are fewer than n children, then some of the ranges do not come into
play.

<b>Scale and Bias</b>

The range value, x, that is computed by this node during traversal can
be modified in a number of different ways:

-# Associate an sglLODState object with the sglLOD node.
This adds an additional scale and offset (bias):
<pre>
   new_range1 = state_scale*(dist from eyepoint and center) + state_offset
</pre>
<p>
-# Set a static scale factor for the sglLOD node.  This has the following
effect on all range calculations:
<pre>
   new_range2 = static_scale*new_range1
</pre>
<p>
-# Set a scale and offset on the cull or isect traversal state.  This has
the following effect:
<pre>
   new_range3 = traversal_scale*new_range2 + traversal_offset
</pre>
For the cull traversal, the traversal scale and offset are computed in
the scene graph's root (sglScene) node before the traversal commences.
It does this by making a call to sglScene::computeLODScale with the size
of the window in pixels and the field of view.  In this function the
scale is computed in an attempt select higher resolution children as the
window becomes larger or the FOV gets smaller (a zoom in effect).
Note: currently the trav_offset is always set to zero.
<p>
-# Add an sglLODStateNode to the scene graph (see its page for an
explanation of this node).

<b>Special behaviour for intersect and pick traversals</b>

When using the intersect or pick traversal on a scene graph containing sglLOD
nodes a number of different behaviours can be dictated by the user when it
encounters an sglLOD node.  By default, the intersection will only be performed
for the highest resolution child (child[0]).  The other behaviours
are listed as follows for various sglProbe modes:

-# ISECT_LOD_ALL - to intersect/pick against all of the children as if an
sglGroup node.
-# ISECT_LOD_NONE - to halt the traversal of this branch.
-# ISECT_LOD_VIEW - perform intersection/picking against the same node
that the view frustum culling would select (user must specify the view matrix).

<b>Example:</b>

<pre>
   sglLOD *lod_node = new sglLOD;
   lod_node->setLODState(new sglLODState);

   sglGeode *geode1 = new sglGeode;
   sglGeode *geode2 = new sglGeode;
   sglGeode *geode3 = new sglGeode;

   lod_node->addChild(geode1);     // finest resolution
   lod_node->addChild(geode2);
   lod_node->addChild(geode3);     // coarsest resolution

                                   // select nothing less than 1.0
   lod_node->setRange(0, 1.0f);    // select geode1 between 1.0 and 5.0
   lod_node->setRange(1, 5.0f);    // select geode2 between 5.0 and 10.0
   lod_node->setRange(2, 10.0f);   // select geode3 between 10.0 and 50.0
   lod_node->setRange(3, 50.0f);   // select nothing greater than 50.0

   // double computed range for all LODs: has effect of selecting coarser
   // resolution children.
   sglLOD::setScaleFactor(2.0f);

   // halve the computed range and add a bias for this LOD only: has effect
   // of selecting finer resolution children
   sglLODState *lod_state = lod_node->getLODState();
   if (lod_state)
   {
      lod_state->setRangeScale(0.5f);
      lod_state->setRangeOffset(-2.0f);
   }
</pre>

*/

class SGL_DLL_API sglLOD : public sglGroup
{
public:
   /// default constructor
   sglLOD();
   /// virtual destructor
   virtual ~sglLOD();

   /** Get the number of elements in the range vector.
       @returns size of the range vector.
   */
   unsigned int getNumRanges() const { return m_ranges.size(); }

   /** Set a range value.
       @param index The index into the range vector.
       @param range The index-th range value.  It should be positive and
       greater than the range value in the index-1 position.
   */
   void  setRange(unsigned int index, float range);

   /** Get a range value.
       @param index The index in the range vector.
       @returns range value at the position specified by index.  If index is
       out of range then -1.0 is returned (note: -1 is returned for unspecified
       range values as well).
   */
   float getRange(unsigned int index) const;

   /** Set the reference position for computing a range to.  This defaults to
       the origin of this node's local coordinate system.
       @param center The reference position.
   */
   void setCenter(const sglVec3f &center) { m_center = center; }

   /** Get the current reference position.
       @returns The current reference position.
    */
   const sglVec3f &getCenter() const { return m_center; }

   /** Set a global scale factor used to affect transitions of groups of ALL
       LOD nodes.
       @param scale The scale factor (should be positive).
   */
   static void setScaleFactor(float scale) { s_range_scale = scale; }

   /** Get the current global scale factor.
       @returns The current scale factor.
   */
   static float getScaleFactor() { return s_range_scale; }

   /** Specify a scale and bias to affect transitions of this LOD node.  By
       using the same sglLODState object for more than one sglLOD node, a group
       of sglLOD's range calculation can be affected simultaneously.  These
       objects are reference counted with this function.  New objects are
       ref()-ed while previous objects are deref()-ed.  If an sglLODState
       object's reference count goes to zero when it is replaced by this
       function, it will be deleted.
       @param lod_scale The new sglLODState object.
       @see sglLODState
   */
   void setLODState(sglLODState *lod_state);

   /** Get a pointer to the current sglLODState object.
       @return Pointer to the current sglLODState object.  NULL if no object
       has be assigned.
   */
   sglLODState *getLODState() const { return m_lod_state; }

   // see sglNode::intersect(f) for details.
   virtual sglBound::IntersectResultEnum intersect(
      sglIntersectf &isector) const;

   // see sglNode::intersect(d) for details.
   virtual sglBound::IntersectResultEnum intersect(
      sglIntersectd &isector) const;

   // see sglNode::pick(f) for details.
   virtual void pick(sglPickf &pick_state, unsigned int cull_flags) const;

   // see sglNode::pick(d) for details.
   virtual void pick(sglPickd &pick_state, unsigned int cull_flags) const;

   // See sglNode::clone() for details.
   virtual sglNode *clone(unsigned int mode) const;

   // See sglObject::printInfo() for details.
   virtual void printInfo(ostream &ostrm, const char *indent_string) const;

protected:
   // See sglNode::cull(f) for details
   virtual void cull(sglCull<float> &trav_state,
                     unsigned int cull_flags) const;

   // See sglNode::cull(d) for details
   virtual void cull(sglCull<double> &trav_state,
                     unsigned int cull_flags) const;

   // If a user wants to subclass this node and replace the equation used for
   // determining the range used in child selection, the following function
   // should be reimplemented in the subclass (one each for single and double
   // precision traversals.
   virtual float computeRange(float trav_lod_scale,
                              float trav_lod_offset,
                              const sglMat4f &view_matrix) const;
   virtual float computeRange(float trav_lod_scale,
                              float trav_lod_offset,
                              const sglMat4d &view_matrix) const;

   void copyTo(sglLOD *dst, unsigned int mode) const;

private: // not implemented
   sglLOD(const sglLOD &);
   sglLOD &operator=(const sglLOD &);

protected:
   static float s_range_scale;

   sglVec3f m_center;
   vector<float> m_ranges;

   sglLODState *m_lod_state;
};

#endif
