/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglLODState.hpp
 *  Created: 7 April 1999
 *  Summary: Used to modify the transition behaviour of groups of sglLOD nodes
 *****************************************************************************/

#ifndef __SGL_LOD_STATE_HPP
#define __SGL_LOD_STATE_HPP

#include <sgl.h>
#include <sglObject.hpp>

//----------------------------------------------------------------------------

/**

This class contains scale and offset (bias) values that affect the range
calculations of one or more sglLOD nodes to which it is assigned.  This can be
used to modify the selection behaviour of groups of sglLOD nodes.  See the
sglLOD class documentation for the effect on the range calculations for
level-of-detail selection.

<b>Example:</b>
<pre>
   sglLOD *lod_node1 = new sglLOD;
   sglLOD *lod_node2 = new sglLOD;

   // halve the computed range and add a bias for this LOD only: has effect
   // of selecting finer resolution children at farther distances
   sglLODState *lod_state = new sglLODState;
   lod_state->setRangeScale(0.5f);    // multiplicative factor
   lod_state->setRangeOffset(-20.0f); // additive factor

   lod_node1->setLODState(lod_state);
   lod_node2->setLODState(lod_state);
</pre>

*/

class SGL_DLL_API sglLODState : public sglObject
{
public:
   /** Default constructor.  The scale and offset default to 1 and 0
       respectively (that is, no effect).
   */
   sglLODState();

   /// virtual destructor
   virtual ~sglLODState();

   /** Set the scale factor for sglLOD range calculations.
       @param val The scale value (any value is valid).  This defaults to 1.
   */
   void setRangeScale(float val) { m_range_scale = val; }

   /** Get the current scale factor.
       @return scale value.
   */
   float getRangeScale() const { return m_range_scale; }

   /** Set the offset (bias) factor for sglLOD range calculations.
       @param val The bias value (any value is valid).  This defaults to 0.
   */
   void setRangeOffset(float val) { m_range_offset = val; }

   /** Get the current bias value.
       @return bias value.
   */
   float getRangeOffset() const { return m_range_offset; }

   // See sglObject::printInfo() for details.
   virtual void printInfo(ostream &ostrm, const char *indent_string) const;

private: // not implemented
   sglLODState(const sglLODState &);
   sglLODState &operator=(const sglLODState &);

protected:
   float m_range_scale;
   float m_range_offset;
};

#endif
