/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglStateNode.hpp
 *  Created: 3 February 2000
 *  Summary: use a stateNode mask to select/deselect the entire subtree
 *****************************************************************************/

#ifndef __SGL_LODSTATENODE_HPP
#define __SGL_LODSTATENODE_HPP

#include <sgl.h>
#include <sglGroup.hpp>
#include <sglLODState.hpp>

//----------------------------------------------------------------------------

/**

The sglTraverser class (base class for cull, pick and intersect traversers)
contain LOD scale and offset (bias) quantities that will affect the child
selection computation of the sglLOD node.  This node affects those values when
the traverser encounters it.  The net effect is to scale and bias the range
calculations for any sglLOD nodes in the entire subtree below it.
The scale and offset values stored in this node affect the traversals' scale
and offset values as follows:

<pre>
   new_traversal_scale = old_traversal_scale*lod_state_node_scale

   new_traversal_offset= old_traversal_offset*lod_state_node_scale + lod_state_node_offset
</pre>

Note: the affect of nested sglLODStateNodes is cummulative.

After computation of the new scale and offset, traversal then proceeds as
if this were an sglGroup node.

<b>Example:</b>

<pre>
   sglLODStateNode *lod_state_node = new sglLODStateNode;
   sglLOD *lod_node = new sglLOD;

   lod_state_node->addChild(lod_node);

   // scales and offsets the range calculations for all sglLOD nodes that
   // appear below (as descendants of) it in the scene graph
   lod_state_node->setRangeScale(2.0f);
   lod_state_node->setRangeOffset(-20.0f);
</pre>

 */

class SGL_DLL_API sglLODStateNode : public sglGroup
{
public:
   /// default constructor
   sglLODStateNode();
   /// virtual destructor
   virtual ~sglLODStateNode();

   /** Set the scale factor for subsequent LOD range calculations.
       @param val The scale value (any value is valid).
   */
   void setRangeScale(float val) { m_lod_state.setRangeScale(val); }

   /** Get the current scale factor.
       @return scale value.
   */
   float getRangeScale() const   { return m_lod_state.getRangeScale(); }

   /** Set the offset (bias) factor for subsequent LOD range calculations.
       @param val The bias value (any value is valid).
   */
   void setRangeOffset(float val) { m_lod_state.setRangeOffset(val); }

   /** Get the currect bias value.
       @return bias value.
   */
   float getRangeOffset() const   { return m_lod_state.getRangeOffset(); }

   // see sglNode::intersect(f) for details.
   virtual sglBound::IntersectResultEnum intersect(
      sglIntersectf &isector) const;

   // see sglNode::intersect(d) for details.
   virtual sglBound::IntersectResultEnum intersect(
      sglIntersectd &isector) const;

   // see sglNode::pick(f) for details.
   virtual void pick(sglPickf &pick_state, unsigned int cull_flags) const;

   // see sglNode::pick(d) for details.
   virtual void pick(sglPickd &pick_state, unsigned int cull_flags) const;

   // See sglNode::clone() for details.
   virtual sglNode *clone(unsigned int mode) const;

   // See sglObject::printInfo() for details.
   virtual void printInfo(ostream &ostrm, const char *indent_string) const;

protected:
   // See sglNode::cull(f) for details
   virtual void cull(sglCull<float> &trav_state,
                     unsigned int cull_flags) const;

   // See sglNode::cull(d) for details
   virtual void cull(sglCull<double> &trav_state,
                     unsigned int cull_flags) const;

   void copyTo(sglLODStateNode *dst, unsigned int mode) const;

private: // not implemented
   sglLODStateNode(const sglLODStateNode &);
   sglLODStateNode &operator=(const sglLODStateNode &);

protected:
   sglLODState m_lod_state;
};

#endif
