/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglLayer.cpp
 *  Created: 5 January 2001
 *  Summary:
 *****************************************************************************/

#include <sglLayer.hpp>
#include <sglCull.hpp>

//----------------------------------------------------------------------------
class sglStencilLayerBegin : public sglDrawItem
{
public:
   sglStencilLayerBegin() {}
   virtual ~sglStencilLayerBegin() {}

   void setMask(GLint mask) { m_stencil_mask = mask; }

   virtual void draw(sglCurrState *) const
      {
         glPushAttrib(GL_STENCIL_BUFFER_BIT);
         glEnable(GL_STENCIL_TEST);
         glStencilMask(m_stencil_mask);
         glStencilFunc(GL_ALWAYS, m_stencil_mask, m_stencil_mask);
         glStencilOp(GL_ZERO, GL_ZERO, GL_REPLACE);
      }
   virtual void statsDraw(sglCurrState *curr_state,
                          sglStats&) const
      {
         draw(curr_state);
      }
   virtual const sglFullState* getFirstState() const { return NULL; }

private:
   GLint m_stencil_mask;
};

//----------------------------------------------------------------------------
class sglStencilLayerMiddle : public sglDrawItem
{
public:
   sglStencilLayerMiddle() {}
   virtual ~sglStencilLayerMiddle() {}

   void setMask(GLint mask) { m_stencil_mask = mask; }

   virtual void draw(sglCurrState *) const
      {
         glStencilFunc(GL_EQUAL, m_stencil_mask, m_stencil_mask);
         glStencilOp(GL_KEEP, GL_KEEP, GL_KEEP);
      }
   virtual void statsDraw(sglCurrState *curr_state, sglStats &) const
      {
         draw(curr_state);
      }
   virtual const sglFullState* getFirstState() const { return NULL; }

private:
   GLint m_stencil_mask;
};

//----------------------------------------------------------------------------
class sglStencilLayerEnd : public sglDrawItem
{
public:
   sglStencilLayerEnd() {}
   virtual ~sglStencilLayerEnd() {}

   void setMask(GLint mask) { m_stencil_mask = mask; }

   virtual void draw(sglCurrState *) const
      {
         glPopAttrib();
      }
   virtual void statsDraw(sglCurrState *curr_state, sglStats &) const
      {
         draw(curr_state);
      }
   virtual const sglFullState* getFirstState() const { return NULL; }

private:
   GLint m_stencil_mask;
};

//----------------------------------------------------------------------------

sglDepthMask *sglLayer::s_depth_mask = new sglDepthMask(false);
sglDepthTest *sglLayer::s_depth_test = new sglDepthTest(sglDepthTest::eALWAYS);
sglColorMask *sglLayer::s_color_mask = new sglColorMask(false, false,
                                                        false, false);


//----------------------------------------------------------------------------
sglLayer::sglLayer(LayerMethodEnum method) : sglGroup(),
   m_method(eMULTIPASS),
   m_stencil_mask(0x1),
   m_stencil_begin(NULL),
   m_stencil_middle(NULL),
   m_stencil_end(NULL)
{
   setMethod(method);
}

//----------------------------------------------------------------------------
sglLayer::~sglLayer()
{
   delete m_stencil_begin;
   delete m_stencil_middle;
   delete m_stencil_end;
}

//----------------------------------------------------------------------------
void sglLayer::setMethod(LayerMethodEnum method)
{
   if (method == eSTENCIL && sgl::s_num_stencil_bits == 0)
   {
      method = eMULTIPASS;
   }

   if (m_method != method)
   {
      m_method = method;

      switch (m_method)
      {
         case eSTENCIL:
            m_stencil_begin = new sglStencilLayerBegin();
            m_stencil_middle = new sglStencilLayerMiddle();
            m_stencil_end = new sglStencilLayerEnd();
            m_stencil_begin->setMask(m_stencil_mask);
            m_stencil_middle->setMask(m_stencil_mask);
            m_stencil_end->setMask(m_stencil_mask);
            break;
         case eMULTIPASS:
            delete m_stencil_begin;
            delete m_stencil_middle;
            delete m_stencil_end;
            m_stencil_begin = NULL;
            m_stencil_middle = NULL;
            m_stencil_end = NULL;
            break;
         default:
            sglPrintWarning() << "Unknown layer method: " << (int)m_method
                              << endl;
      }
   }
}

//----------------------------------------------------------------------------
void sglLayer::setStencilMask(GLuint mask)
{
   m_stencil_mask = mask;
   if (m_stencil_begin) m_stencil_begin->setMask(m_stencil_mask);
   if (m_stencil_middle) m_stencil_middle->setMask(m_stencil_mask);
   if (m_stencil_end) m_stencil_end->setMask(m_stencil_mask);
}

//----------------------------------------------------------------------------
void sglLayer::cull(sglCull<float> &trav_state,
                    unsigned int cull_flags) const
{
   if (cull_flags && (m_sphere.intersect(trav_state.getFrustum(),
                                         cull_flags) == sglBound::eOUTSIDE))
   {
      return; // draw nothing
   }

   if (m_child_list.size() > 1)
   {
      sglDrawablePool *pool = trav_state.pushDrawablePool();

      switch (m_method)
      {
         case eSTENCIL:
         {
            pool->insert(m_stencil_begin);
            vector<sglNode*>::const_iterator it = m_child_list.begin();
            (*it)->cull(trav_state, cull_flags);
            pool->insert(m_stencil_middle);
            trav_state.pushOverrideStatelet(s_depth_mask);
            trav_state.pushOverrideStatelet(s_depth_test);
            for (; it != m_child_list.end(); ++it)
            {
               (*it)->cull(trav_state, cull_flags);
            }
            pool->insert(m_stencil_end);
            trav_state.popOverrideStatelet();
            trav_state.popOverrideStatelet();
            break;
         }
         case eMULTIPASS:
         {
            trav_state.pushOverrideStatelet(s_depth_mask);
            for (vector<sglNode*>::const_iterator it = m_child_list.begin();
                 it != m_child_list.end(); ++it)
            {
               (*it)->cull(trav_state, cull_flags);
            }
            trav_state.popOverrideStatelet();
            trav_state.pushOverrideStatelet(s_color_mask);
            (*m_child_list.begin())->cull(trav_state, cull_flags);
            trav_state.popOverrideStatelet();
            break;
         }
         default:
            sglPrintWarning() << "Unknown layer method: " << (int)m_method
                              << endl;
      }

      trav_state.popDrawablePool();
   }
   else
   {
      for (vector<sglNode*>::const_iterator it = m_child_list.begin();
           it != m_child_list.end(); ++it)
      {
         (*it)->cull(trav_state, cull_flags);
      }
   }
}

//----------------------------------------------------------------------------
void sglLayer::cull(sglCull<double> &trav_state,
                    unsigned int cull_flags) const
{
   if (cull_flags && (m_sphere.intersect(trav_state.getFrustum(),
                                         cull_flags) == sglBound::eOUTSIDE))
   {
      return; // draw nothing
   }

   if (m_child_list.size() > 1)
   {
      sglDrawablePool *pool = trav_state.pushDrawablePool();

      switch (m_method)
      {
         case eSTENCIL:
         {
            pool->insert(m_stencil_begin);
            vector<sglNode*>::const_iterator it = m_child_list.begin();
            (*it)->cull(trav_state, cull_flags);
            pool->insert(m_stencil_middle);
            trav_state.pushOverrideStatelet(s_depth_mask);
            trav_state.pushOverrideStatelet(s_depth_test);
            for (; it != m_child_list.end(); ++it)
            {
               (*it)->cull(trav_state, cull_flags);
            }
            pool->insert(m_stencil_end);
            trav_state.popOverrideStatelet();
            trav_state.popOverrideStatelet();
            break;
         }
         case eMULTIPASS:
         {
            trav_state.pushOverrideStatelet(s_depth_mask);
            for (vector<sglNode*>::const_iterator it = m_child_list.begin();
                 it != m_child_list.end(); ++it)
            {
               (*it)->cull(trav_state, cull_flags);
            }
            trav_state.popOverrideStatelet();
            trav_state.pushOverrideStatelet(s_color_mask);
            (*m_child_list.begin())->cull(trav_state, cull_flags);
            trav_state.popOverrideStatelet();
            break;
         }
         default:
            sglPrintWarning() << "Unknown layer method: " << (int)m_method
                              << endl;
      }

      trav_state.popDrawablePool();
   }
   else
   {
      for (vector<sglNode*>::const_iterator it = m_child_list.begin();
           it != m_child_list.end(); ++it)
      {
         (*it)->cull(trav_state, cull_flags);
      }
   }
}

//----------------------------------------------------------------------------
sglNode *sglLayer::clone(unsigned int mode) const
{
   sglLayer *root = new sglLayer(m_method);
   sglLayer::copyTo(root, mode);
   return root;
}

//----------------------------------------------------------------------------
void sglLayer::copyTo(sglLayer *dst, unsigned int mode) const
{
   sglGroup::copyTo(dst, mode);

   // transfer stuff
   //dst->setMethod(m_method);  // done in clone()
   dst->setStencilMask(m_stencil_mask);
}

//----------------------------------------------------------------------------
void sglLayer::printInfo(ostream &ostrm, const char *indent_string) const
{
   sglGroup::printInfo(ostrm, indent_string);
   ostrm << indent_string << "sglLayer: method = " << (int)m_method
         << ", mask = " << m_stencil_mask << endl;
}
