/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglLayer.hpp
 *  Created: 5 January 2001
 *  Summary: Group that aids in rendering coplaner polygons without
 *           visual artifacts caused by the depth buffer.
 *****************************************************************************/

#ifndef __SGL_LAYER_HPP
#define __SGL_LAYER_HPP

#include <sgl.h>
#include <sglGroup.hpp>


class sglStencilLayerBegin;
class sglStencilLayerMiddle;
class sglStencilLayerEnd;
class sglDepthMask;
class sglDepthTest;
class sglColorMask;

//----------------------------------------------------------------------------

/**

An sglLayer node is a specialized group that aids in rendering coplaner
polygons without visual artifacts caused by z-fighting (or flimmering) in the
depth buffer.  The children of an sglLayer are assigned a visual priority based
on their order in the group.  The first child in the group (at index 0), often
called the base layer, defines depth values that will be written into the depth
buffer.  Every child after that is then rendered in order on top of the base
layer.

In order for the sglLayer to work properly, there are several restrictions:

-# The layers should be coplanar.  Every pixel of child n+1 will
     be rendered in front of child n, regardless of their depth values.
-# The geometry for child layers 1..n, must be completely contained
     within boundary of the geometry for the base layer (child 0).
-# The use of an sglLayer node will disable state sorting of the
     geometry contained in the group.

There are two methods that can be used for rendering the layers:

- <b>eMULTIPASS</b>  where each layer is rendered in order, but multiple
                     passes are required of the base layer.
- <b>eSTENCIL</b>    where only one pass is required of each layer, but
                     the stencil buffer is required.  This method
                     may perform better than LAYER_MULTIPASS, but
                     is only available on some frame buffers.

Note: Use of the sglPolygonOffset statelet can provide much better performance
than an sglLayer node without any of the above restrictions, at the cost of
some visual quality.  It is important to remember though, that sglPolygonOffset
only works with polygons, whereas an sglLayer may also contain lines and
points.  Another alternative is to manipulate the OpenGL depth range before
drawing layers using the sglDepthRange statelet, but the values needed to get
the desired affect can vary from platform to platform and will affect the
values written into the depth buffer.

<b>Example:</b>
<pre>
   sglLayer *layer = new sglLayer(sglLayer::eSTENCIL);

   for (int ix=4; ix>=0; --ix)
   {
      sglGeode *geode = new sglGeode();
         :
         :
      layer->addChild(geode);
   }
</pre>

\todo Add support for intersect traversal.
\todo Add support for the pick traversal.

 */

class SGL_DLL_API sglLayer : public sglGroup
{
public:
   /// The two different layering modes
   enum LayerMethodEnum
   {
      /// the multipass rendering approach.
      eMULTIPASS,
      /// the single pass approach requiring stencil planes.
      eSTENCIL
   };

public:
   /** Constructor
       @param method Sets the method to be used: eMULTIPASS or eSTENCIL
       (defaults to eMULTIPASS).
   */
   sglLayer(LayerMethodEnum method = eMULTIPASS);
   /// virtual destructor
   virtual ~sglLayer();

   /** Set the method to use for rendering the layers.
       @param method The method which is one of eMULTIPASS or eSTENCIL.
   */
   void setMethod(LayerMethodEnum method);

   /** Get the current method for rendering the layers.
       @param The current method which is one of eMULTIPASS or eSTENCIL.
   */
   LayerMethodEnum getMethod() const { return m_method; }

   /** Set the mask to use with the stencil buffer (for eSTENCIL method only).
       @param mask The mask defaults to 0x1.  The bit(s) set should not exceed
       the stencil buffer depth.
   */
   void setStencilMask(GLuint mask);

   /** Get the mask to use with the stencil buffer.
       @return the current stencil mask.
   */
   GLuint getStencilMask() { return m_stencil_mask; }

   // See sglNode::clone() for details.
   virtual sglNode *clone(unsigned int mode) const;

   // See sglObject::printInfo() for details.
   virtual void printInfo(ostream &ostrm, const char *indent_string) const;

protected:
   // See sglNode::cull(f) for details
   virtual void cull(sglCull<float> &trav_state,
                     unsigned int cull_flags) const;

   // See sglNode::cull(d) for details
   virtual void cull(sglCull<double> &trav_state,
                     unsigned int cull_flags) const;

   void copyTo(sglLayer *dst, unsigned int mode) const;

private:  // not implemented
   sglLayer(const sglLayer &);
   sglLayer &operator=(const sglLayer &);

private:
   LayerMethodEnum          m_method;
   GLuint                   m_stencil_mask;

   sglStencilLayerBegin    *m_stencil_begin;
   sglStencilLayerMiddle   *m_stencil_middle;
   sglStencilLayerEnd      *m_stencil_end;

   static sglDepthMask     *s_depth_mask;
   static sglDepthTest     *s_depth_test;
   static sglColorMask     *s_color_mask;
};

#endif
