/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglLight.hpp
 *   Author: Scott McMillan
 *  Created: 28 November 2000
 *  Summary: abstract base class for all types of SGL lights.  Derived from
 *           sglNode, it is a leaf node of the scene graph, equivalent to
 *           sglGeode or sglText.
 *****************************************************************************/

#ifndef __SGL_LIGHT_HPP
#define __SGL_LIGHT_HPP

#include <sgl.h>
#include <sglNode.hpp>

//----------------------------------------------------------------------------

/**

sglLight is the abstract base class for all scene graph lights.  Derived from
sglNode, it is a leaf node of the scene graph.  Every light will have
associated with it a bounding sphere, the size of which is determined
by the subclasses.  When that bounding sphere intersects the view frustum
during cull, that light will be placed on the list of lights to enable before
the draw() function is called.  The static variable sgl::s_max_lights is set
during sgl::initialize() to the maximum number of lights supported by OpenGL.
The sglLightProcessor class is used is determine which lights will be used if
more than sgl::s_max_lights are collected during cull.

When constructed the light will be enabled by default, and will have no ambient
component (black), and white diffuse and specular components.  The user can
disable the light by calling the member function, sglLight::setOn(false).  The
various components of light can be set with calls to setAmbient, setDiffuse,
and setSpecular member functions.

Some global OpenGL lighting concepts such as global ambient, local light model,
and two-sided lighting have not been implmented in SGL yet (the user must
currently call these OpenGL commands directly).  They might be implemented as
static functions in this class.

\todo Implement functions to set global ambient.
\todo Implement toggle for local light model.
\todo Implement toggle for two sided lighting.

*/

class SGL_DLL_API sglLight : public sglNode
{
public:
   /// default constructor
   sglLight();
   /// virtual destructor
   virtual ~sglLight();

   /** Enable/disable the light.
       @param on False to disable the light (default is true/on).
   */
   void setOn(bool on);  // true = on; false = off

   /** Get the current state (on or off) of this light.
       @return True if light is enabled, otherwise false.
   */
   bool getOn() const { return m_on_flag; }

   /** Set the ambient color component of the light.
       @param color Four component color vector.  Default is (0, 0, 0, 1)
   */
   void setAmbient(const sglVec4f &color) { m_ambient = color; }

   /** Set the ambient color component of the light.  Default is (0, 0, 0, 1)
       @param r Red component (valid values are from 0 to 1)
       @param g Green component (valid values are from 0 to 1)
       @param b Blue component (valid values are from 0 to 1)
       @param a Alpha component (valid values are from 0 to 1)
   */
   void setAmbient(float r, float g, float b, float a)
      { m_ambient.set(r,g,b,a); }

   /** Get the current ambient component.
       @return Reference to the 4 component color vector.
   */
   const sglVec4f &getAmbient() const { return m_ambient; }

   /** Set the diffuse color component of the light.
       @param color Four component color vector.  Default is (1, 1, 1, 1)
   */
   void setDiffuse(const sglVec4f &color) { m_diffuse = color; }

   /** Set the diffuse color component of the light.  Default is (1, 1, 1, 1)
       @param r Red component (valid values are from 0 to 1)
       @param g Green component (valid values are from 0 to 1)
       @param b Blue component (valid values are from 0 to 1)
       @param a Alpha component (valid values are from 0 to 1)
   */
   void setDiffuse(float r, float g, float b, float a)
      { m_diffuse.set(r,g,b,a); }

   /** Get the current diffuse component.
       @return Reference to the 4 component color vector.
   */
   const sglVec4f &getDiffuse() const { return m_diffuse; }

   /** Set the specular color component of the light.
       @param color Four component color vector.  Default is (1, 1, 1, 1)
   */
   void setSpecular(const sglVec4f &color) { m_specular = color; }

   /** Set the specular color component of the light.  Default is (1, 1, 1, 1)
       @param r Red component (valid values are from 0 to 1)
       @param g Green component (valid values are from 0 to 1)
       @param b Blue component (valid values are from 0 to 1)
       @param a Alpha component (valid values are from 0 to 1)
   */
   void setSpecular(float r, float g, float b, float a)
      { m_specular.set(r,g,b,a); }

   /** Get the current specular component.
       @return Reference to the 4 component color vector.
   */
   const sglVec4f &getSpecular() const { return m_specular; }

   /** This function is called during sglCull::draw() function to issue the
       OpenGL light commands needed to activate this light before rendering.
       @param gl_light_number One of GL_LIGHT0 .. GL_LIGHT<sgl::s_max_lights-1>
   */
   virtual void setup(GLenum gl_light_number) const;

   // See sglObject::printInfo() for details.
   virtual void printInfo(ostream &ostrm, const char *indent_string) const;

   // I am not sure what to do with these yet.  I can leave them here, and they
   // affect all scene graphs at once, or I can put them as a traversal mode
   // or I can leave it up to the user (external to SGL).
   //
   //static void setGlobalAmbient(const sglVec4f &color);
   //static const sglVec4f &getGlobalAmbient() { return s_global_ambient; }
   //static void enableLocal(bool on);
   //static bool getLocal() { return s_local_flag; }
   //static void enableTwoSided(bool on);
   //static bool getTwoSided() { return s_two_sided; }

protected:
   virtual void cull(sglCull<float> &trav_state,
                     unsigned int cull_flags) const;
   virtual void cull(sglCull<double> &trav_state,
                     unsigned int cull_flags) const;

   void copyTo(sglLight *dst, unsigned int mode) const;

private: // methods - not implemented
   sglLight(const sglLight &);
   sglLight &operator=(const sglLight &);

protected: // variables
   bool     m_on_flag;

   sglVec4f m_ambient;
   sglVec4f m_diffuse;
   sglVec4f m_specular;
};

#endif
