/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglLightProcessor.cpp
 *   Author: Tom Stimson
 *  Created: 20 April 2002
 *****************************************************************************/

#include <sglLightProcessor.hpp>
#include <sglLight.hpp>


//----------------------------------------------------------------------------
void sglDefaultLightProcessor::addLight(
   const sglMat4f& matrix, const sglLight* light)
{
   if (light)
   {
      m_light_list.push_back(LightStruct(matrix, light));
   }
}


//----------------------------------------------------------------------------
void sglDefaultLightProcessor::apply()
{
    GLint num_lights = m_light_list.size();
    if (num_lights)
    {
        // BRAINDEAD - apply the first max lights if num exceeds max
        if (num_lights > sgl::s_max_lights)
            num_lights = sgl::s_max_lights;
        for (GLint light_num = 0; light_num < num_lights; ++light_num)
        {
            sglLoadMatrix(m_light_list[light_num].first);
            (m_light_list[light_num].second)->setup(
                (GLenum)(GL_LIGHT0 + light_num));
            glEnable((GLenum)(GL_LIGHT0 + light_num));
        }
    }
}


//----------------------------------------------------------------------------
inline float computeLightBrightness(const sglMat4f&, const sglLight *light)
{
   sglVec4f rgba = light->getAmbient();
   rgba += light->getDiffuse();

   // assuming brightness == Y(luminance) in the YIQ colorspace.
   return 0.299f * rgba[0] + 0.587 * rgba[1] + 0.114 * rgba[2];
}


//----------------------------------------------------------------------------
inline float computeLightEffect(const sglMat4f& matrix, const sglLight *light)
{
   const sglSphereBound<float> &bound = light->getPrevBound();
   sglVec4f pos;
   mul(bound.getCenter(), 1.f, matrix, pos);
   pos[3] = 0;
   float distance = pos.length();
   return computeLightBrightness(matrix, light) * bound.getRadius() / distance;
}


//----------------------------------------------------------------------------
void sglSortedLightProcessor::addLight(
   const sglMat4f& matrix, const sglLight* light)
{
   if (light)
   {
      const sglSphereBound<float> &bound = light->getPrevBound();
      if (bound.isInfinite())
      {
         float brightness = computeLightBrightness(matrix, light);
         m_inf_lights.insert(
            pair<float, LightStruct>(brightness, LightStruct(matrix, light)));
      }
      else
      {
         float effect = computeLightEffect(matrix, light);
         m_positional_lights.insert(
            pair<float, LightStruct>(effect, LightStruct(matrix, light)));
      }
   }
}


//----------------------------------------------------------------------------
void sglSortedLightProcessor::apply()
{
   GLint i = 0;
   GLint max = sgl::s_max_lights;
   MapType::const_iterator it;
   MapType::const_iterator end = m_inf_lights.end();

   for (it = m_inf_lights.begin(); it != end && i < max; ++it, ++i)
   {
      sglLoadMatrix(it->second.first);
      GLenum light_num = (GLenum)(GL_LIGHT0 + i);
      it->second.second->setup(light_num);
      glEnable(light_num);
   }
   end = m_positional_lights.end();
   for (it = m_positional_lights.begin(); it != end && i < max; ++it, ++i)
   {
      sglLoadMatrix(it->second.first);
      GLenum light_num = (GLenum)(GL_LIGHT0 + i);
      it->second.second->setup(light_num);
      glEnable(light_num);
   }
}


//----------------------------------------------------------------------------
sglCountSortedLightProcessor::sglCountSortedLightProcessor(
   unsigned int num_bins)
   : m_count(new unsigned int[num_bins]),
     m_num_bins(num_bins),
     m_last_min_brightness(0.0f),
     m_last_max_brightness(2.0f),
     m_last_min_effect(0.0f),
     m_last_max_effect(4.0f)
{
}


//----------------------------------------------------------------------------
sglCountSortedLightProcessor::~sglCountSortedLightProcessor()
{
   delete[] m_count;
}


//----------------------------------------------------------------------------
void sglCountSortedLightProcessor::addLight(
   const sglMat4f& matrix, const sglLight* light)
{
   if (light)
   {
      if (light->getPrevBound().isInfinite())
         m_inf_lights.push_back(LightStruct(matrix, light));
      else
         m_positional_lights.push_back(LightStruct(matrix, light));
   }
}


//----------------------------------------------------------------------------
void sglCountSortedLightProcessor::sortLights(
   vector<LightStruct> &lights,
   ValFunc val_func,
   GLenum first_light,
   GLint max_lights,
   float &last_min_val,
   float &last_max_val)
{
   float min_val = FLT_MAX;
   float max_val = -1;

   float last_val_range = last_max_val - last_min_val;

   unsigned int max_bin = m_num_bins-1;
   unsigned int max_bin_less_one = max_bin-1;
   float max_bin_less_two = (float)(max_bin-2);

   vector<LightStruct>::iterator it;
   vector<LightStruct>::iterator end = lights.end();
   for (it = lights.begin(); it != end; ++it)
   {
      float val = val_func(it->m_matrix, it->m_light);
      it->m_val = val;

      if (val < min_val)
         min_val = val;
      if (val > max_val)
         max_val = val;

      unsigned int bin = 0;

      if (val < last_max_val)
      {
         if (val <= last_min_val)
            bin = max_bin;
         else
         {
//            cerr << "  val: " << val << endl;
            bin = (unsigned int)(max_bin_less_two * (val-last_min_val)
                                 / last_val_range);
            bin = max_bin_less_one - bin;
         }
      }

      it->m_bin = bin;
      ++m_count[bin];
   }
   last_min_val = min_val;
   last_max_val = max_val;

//    cerr << "count: ";
//    for (unsigned int j = 0; j < m_num_bins; ++j)
//    {
//       cerr << m_count[j] << " ";
//    }
//    cerr << endl;

   for (unsigned int i = 1; i < m_num_bins; ++i)
      m_count[i] += m_count[i-1];

   end = lights.begin();
   do
   {
      --it;
      int indx = --m_count[it->m_bin];
      if (indx < max_lights)
      {
         sglLoadMatrix(it->m_matrix);
         GLenum light_num = (GLenum)(first_light + indx);
         it->m_light->setup(light_num);
         glEnable(light_num);
      }
   } while (it != end);

//    cerr << "last_min_val = " << last_min_val << endl;
//    cerr << "last_avg_val = " << last_avg_val << endl;
}


//----------------------------------------------------------------------------
void sglCountSortedLightProcessor::apply()
{
   if (m_inf_lights.empty() && m_positional_lights.empty())
      return;

   for (unsigned int i = 0; i < m_num_bins; ++i)
      m_count[i] = 0;

   GLint max_lights = sgl::s_max_lights;

   if ((GLint)m_inf_lights.size() >= max_lights)
   {
      sortLights(m_inf_lights, computeLightBrightness, GL_LIGHT0,
                 max_lights, m_last_min_brightness, m_last_max_brightness);
   }
   else
   {
      GLenum first_light = GL_LIGHT0;
      vector<LightStruct>::iterator it;
      vector<LightStruct>::iterator end = m_inf_lights.end();
      for (it = m_inf_lights.begin(); it != end; ++it)
      {
         sglLoadMatrix(it->m_matrix);
         it->m_light->setup(first_light);
         glEnable(first_light);
         ++first_light;
      }
      max_lights -= m_inf_lights.size();

      if ((GLint)m_positional_lights.size() < max_lights)
      {
         end = m_positional_lights.end();
         for (it = m_positional_lights.begin(); it != end; ++it)
         {
            sglLoadMatrix(it->m_matrix);
            it->m_light->setup(first_light);
            glEnable(first_light);
            ++first_light;
         }
      }
      else
      {
         sortLights(m_positional_lights, computeLightEffect, first_light,
                    max_lights, m_last_min_effect, m_last_max_effect);
      }
   }
}
