/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglLightProcessor.hpp
 *   Author: Tom Stimson
 *  Created: 20 April 2002
 *  Summary: Classes to sort and apply lights during cull and draw
 *****************************************************************************/

#ifndef __SGL_LIGHT_PROCESSOR_HPP
#define __SGL_LIGHT_PROCESSOR_HPP

#include <sgl.h>
#include <sglObject.hpp>
#include <sglMatrix.hpp>

#include <map>

class sglLight;

/** The sglLightProcessor is an abstract base class that defines an
    interface that the sglCull traverser uses for processing lights
    during cull and draw.  Since OpenGL has a limit to the number of
    lights that can be used at any given time, it is the job of the
    sglLightProcessor to decide which lights get used if there are more
    than the allowed number of lights in view.
*/
class SGL_DLL_API sglLightProcessor : public sglObject
{
public:
   /// Constructor.
   sglLightProcessor() {}

   /// Destructor.
   virtual ~sglLightProcessor() {}

   /// Resets the processor for the next frame (e.g. clears list of lights).
   virtual void reset() = 0;

   /** Add a light for the current frame.
       @param matrix modelview matrix to be used when applying this light
       @param light pointer to the new light
   */
   virtual void addLight(const sglMat4f& matrix, const sglLight* light) = 0;

   /** Query the number of lights collected while culling the current frame.
       @returns the number of lights collected while culling the current frame.
   */
   virtual unsigned int getNumLights() const = 0;

   /** Applys the lights collected for this frame.  Note:  this will apply
       at most sgl::s_max_lights.
   */
   virtual void apply() = 0;
};


/** The sglDefaultLightProcessor is the default light processor used
    by sglCull.  It is the quickest, stupidest implementation of an
    sglLightProcessor.  It applies the first sgl::s_max_lights it
    receives, in the order it received them.  If the scene has many
    lights, this method will almost surely _not_ choose the lights that
    would affect the scene the most.
*/
class SGL_DLL_API sglDefaultLightProcessor : public sglLightProcessor
{
public:
   /// Constructor.
   sglDefaultLightProcessor() {}

   /// Destructor.
   virtual ~sglDefaultLightProcessor() {}

   // See sglLightProcessor::reset() for details.
   virtual void reset() { m_light_list.clear(); }

   // See sglLightProcessor::addLight() for details.
   virtual void addLight(const sglMat4f& matrix, const sglLight* light);

   // See sglLightProcessor::getNumLights() for details.
   virtual unsigned int getNumLights() const { return m_light_list.size(); }

   // See sglLightProcessor::apply() for details.
   virtual void apply();

private:
   typedef pair<sglMat4f, const sglLight*> LightStruct;
   vector<LightStruct>  m_light_list;
};


/** The sglSortedLightProcessor uses an O(n*log(n)) sort to determine
    which lights are most likely to affect the scene the most.  All
    directional lights are preferred over positional lights.  The key
    values used for sorting the directional lights are the brightness
    of the lights, where brightness is computed as the Y (luminance)
    component of the YIQ colorspace.  The sort key for positional lights
    is a combination of the brightness, bounding volume radius, and
    distance from the eyepoint.  Since this processor uses an O(n*log(n))
    sort, it can be significantly slower than the sglDefaultLightProcessor
    class for large numbers of lights.
*/
class SGL_DLL_API sglSortedLightProcessor : public sglLightProcessor
{
public:
   /// Constructor.
   sglSortedLightProcessor() {}

   /// Destructor.
   virtual ~sglSortedLightProcessor() {}

   // See sglLightProcessor::reset() for details.
   virtual void reset()
      {
         m_inf_lights.clear();
         m_positional_lights.clear();
      }

   // See sglLightProcessor::addLight() for details.
   virtual void addLight(const sglMat4f& matrix, const sglLight* light);

   // See sglLightProcessor::getNumLights() for details.
   virtual unsigned int getNumLights() const
      { return m_inf_lights.size() + m_positional_lights.size(); }

   // See sglLightProcessor::apply() for details.
   virtual void apply();

private:
   typedef pair<sglMat4f, const sglLight*> LightStruct;
   typedef multimap<float, LightStruct, greater<float> > MapType;
   MapType m_inf_lights;
   MapType m_positional_lights;
};


/** The sglCountSortedLightProcessor sacrifices some accuracy in the sort
    for speed comparible to the sglDefaultLightProcessor class.  This
    class uses the counting sort, O(n+k), where the sort keys are truncated
    to k (num_bins) integer values.
 */
class SGL_DLL_API sglCountSortedLightProcessor : public sglLightProcessor
{
public:
   /** Constructor.
       @params num_bins The number of integer sort values to use when
               sorting the lights.
   */
   sglCountSortedLightProcessor(unsigned int num_bins = 10);

   /// Destructor.
   virtual ~sglCountSortedLightProcessor();

   // See sglLightProcessor::reset() for details.
   virtual void reset()
      {
         m_inf_lights.clear();
         m_positional_lights.clear();
      }

   // See sglLightProcessor::addLight() for details.
   virtual void addLight(const sglMat4f& matrix, const sglLight* light);

   // See sglLightProcessor::getNumLights() for details.
   virtual unsigned int getNumLights() const
      { return m_inf_lights.size() + m_positional_lights.size(); }

   // See sglLightProcessor::apply() for details.
   virtual void apply();

private:
   typedef float (*ValFunc)(const sglMat4f&, const sglLight*);

   struct LightStruct
   {
      LightStruct(const sglMat4f &matrix, const sglLight* light)
         : m_matrix(matrix), m_light(light) {}

      sglMat4f         m_matrix;
      const sglLight  *m_light;
      float            m_val;
      unsigned int     m_bin;
   };

   void sortLights(
      vector<LightStruct> &lights,
      ValFunc val_func,
      GLenum first_light,
      GLint max_lights,
      float &last_min_val,
      float &last_max_val);

private:
   vector<LightStruct>   m_inf_lights;
   vector<LightStruct>   m_positional_lights;
   unsigned int         *m_count;
   unsigned int          m_num_bins;
   float                 m_last_min_brightness;
   float                 m_last_max_brightness;
   float                 m_last_min_effect;
   float                 m_last_max_effect;
};

#endif
