/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglMat2.hpp
 *  Created: 28 February 1999
 *  Summary: 2x2 templated matrix class
 *****************************************************************************/

#ifndef __SGL_MAT2_HPP
#define __SGL_MAT2_HPP

#include <sgl.h>
#include <string.h> // memcpy
#include <sglVector.hpp>
#include <sglVec2.hpp>

// ===========================================================================

template <class T>
class sglMat2
{
public:
   // constructors
   sglMat2() {};

   sglMat2(T m00, T m01, T m10, T m11) { set(m00, m01, m10, m11); }

   template <class S>
   sglMat2(const sglMat2<S>& mat)
      {
         set(mat[0][0], mat[0][1], mat[1][0], mat[1][1]);
      }

   template <class S>
   sglMat2(S mat[2][2])
      {
         set(mat[0][0], mat[0][1], mat[1][0], mat[1][1]);
      }

   // destructor
   ~sglMat2() {};  // not virtual, do not subclass

   inline T *getPtr() const { return (T *)m_matrix; }

   // indexing operator
   inline sglVec2<T>& operator[](unsigned int i)
      { 
         return *(sglVec2<T> *) m_matrix[i]; 
      }
   inline const sglVec2<T>& operator[](unsigned int i) const
      {
         return *(sglVec2<T> *) m_matrix[i];
      }

   // set/get
   template <class S>
   inline void set(S m00, S m01, S m10, S m11)
      {
         m_matrix[0][0] = (T)m00; m_matrix[0][1] = (T)m01;
         m_matrix[1][0] = (T)m10; m_matrix[1][1] = (T)m11;
      }

   template <class S>
   inline void setRow(const sglVec2<S>& v, unsigned int i)
      {
         m_matrix[i][0] = (T)v[0]; m_matrix[i][1] = (T)v[1];
      }

   template <class S>
   inline void setCol(const sglVec2<S>& v, unsigned int i)
      {
         m_matrix[0][i] = (T)v[0]; m_matrix[1][i] = (T)v[1];
      }

   template <class S>
   inline void getRow(sglVec2<S>& v, unsigned int i)
      {
        v[0] = (S)m_matrix[i][0]; v[1] = (S)m_matrix[i][1];
      }

   template <class S>
   inline void getCol(sglVec2<S>& v, unsigned int i)
      {
        v[0] = (S)m_matrix[0][i]; v[1] = (S)m_matrix[1][i];
      }

   // assignment operator
   template <class S>
   inline sglMat2 &operator=(const sglMat2<S>& mat)
      {
         set(mat[0][0], mat[0][1], mat[1][0], mat[1][1]);
         return *this;
      }

   // boolean operators
   template <class S>
   inline bool almostEqual(const sglMat2<S>& mat, double eps) const
      {
         return ((*this)[0].almostEqual(mat[0], eps) &&
                 (*this)[1].almostEqual(mat[1], eps));
      }

   template <class S>
   inline bool operator==(const sglMat2<S>& mat) const
      {
         return (*this)[0] == mat[0] && (*this)[1] == mat[1];
      }

   template <class S>
   inline bool operator!=(const sglMat2<S>& mat) const
      {
         return (*this)[0] != mat[0] || (*this)[1] != mat[1];
      }

   // math operators
   template <class S>
   inline sglMat2& operator+=(const sglMat2<S>& mat)
      {
         (*this)[0] += mat[0], (*this)[1] += mat[1]; return *this;
      }
   template <class S>
   inline sglMat2& operator-=(const sglMat2<S>& mat)
      {
         (*this)[0] -= mat[0], (*this)[1] -= mat[1]; return *this;
      }
   template <class S>
   inline sglMat2& operator*=(const sglMat2<S>& mat)
      {
         mul(*this, mat); return *this;
      }
   inline sglMat2& operator*=(const T s)
      {
         scale(*this, s, s); return *this;
      }

   template <class S>
   inline const sglMat2<T> operator+(const sglMat2<S> &rhs) const
      {
         return sglMat2(m_matrix[0][0] + rhs[0][0],
                        m_matrix[0][1] + rhs[0][1],
                        m_matrix[1][0] + rhs[1][0],
                        m_matrix[1][1] + rhs[1][1]);
      }
   template <class S>
   inline const sglMat2<T> operator-(const sglMat2<S> &rhs) const
      {
         return sglMat2(m_matrix[0][0] - rhs[0][0],
                        m_matrix[0][1] - rhs[0][1],
                        m_matrix[1][0] - rhs[1][0],
                        m_matrix[1][1] - rhs[1][1]);
      }
   template <class S>
   inline const sglMat2<T> operator*(const sglMat2<S> &rhs) const
      {
         return sglMat2(m_matrix[0][0]*rhs[0][0] + m_matrix[0][1]*rhs[1][0],
                        m_matrix[0][0]*rhs[0][1] + m_matrix[0][1]*rhs[1][1],
                        m_matrix[1][0]*rhs[0][0] + m_matrix[1][1]*rhs[1][0],
                        m_matrix[1][0]*rhs[0][1] + m_matrix[1][1]*rhs[1][1]);
      }
   inline const sglMat2<T> operator*(T scalar) const
      {
         return sglMat2(m_matrix[0][0]*scalar,
                        m_matrix[0][1]*scalar,
                        m_matrix[1][0]*scalar,
                        m_matrix[1][1]*scalar);
      }

   // explicit math operations (for backward compatibility)
   template <class S1, class S2>
   inline void add(const sglMat2<S1>& mat1, const sglMat2<S2>& mat2)
      {
         (*this)[0].add(mat1[0],mat2[0]);
         (*this)[1].add(mat1[1],mat2[1]);
      }
   template <class S1, class S2>
   inline void sub(const sglMat2<S1>& mat1, const sglMat2<S2>& mat2)
      {
         (*this)[0].sub(mat1[0],mat2[0]);
         (*this)[1].sub(mat1[1],mat2[1]);
      }
   template <class S>
   inline void scale(const sglMat2<S>& mat, T sx, T sy)
      {
         (*this)[0].scale(mat[0],sx);
         (*this)[1].scale(mat[1],sy);
      }
   // matrix multiply
   template <class S1, class S2>
   inline void mul(const sglMat2<S1>& mat1, const sglMat2<S2>& mat2)
      {
         T t[2][2];
         t[0][0] = mat1[0][0]*mat2[0][0] + mat1[0][1]*mat2[1][0];
         t[0][1] = mat1[0][0]*mat2[0][1] + mat1[0][1]*mat2[1][1];
         t[1][0] = mat1[1][0]*mat2[0][0] + mat1[1][1]*mat2[1][0];
         t[1][1] = mat1[1][0]*mat2[0][1] + mat1[1][1]*mat2[1][1];
         memcpy(m_matrix, t, sizeof(m_matrix));
      }

   // methods to build a matrix
   inline void buildIdentity()
      {
         set(1, 0,  0, 1);
      }

   inline void buildRotation(double angle) // in radians
      {
         T s = (T) sin(angle);
         T c = (T) cos(angle);
         m_matrix[0][0] =  c; m_matrix[0][1] = s;
         m_matrix[1][0] = -s; m_matrix[1][1] = c;
      }

   inline void buildScale(T sx, T sy)
      {
         m_matrix[0][0] = sx;  m_matrix[0][1] = 0;
         m_matrix[1][0] = 0;   m_matrix[1][1] = sy;
      }


   // other operations
   inline void transpose()
      {
         // then upper triangle - use tmp -- transposing self
         T tmp = m_matrix[0][1];
         m_matrix[0][1] = m_matrix[1][0];
         m_matrix[1][0] = tmp;
      }

   template <class S>
   inline void transpose(const sglMat2<S>& mat)
      {
         m_matrix[0][0] = mat[0][0]; m_matrix[1][1] = mat[1][1];

         // then upper triangle - use tmp just in case transposing self
         T tmp = mat[0][1]; m_matrix[0][1] = mat[1][0]; m_matrix[1][0] = tmp;
      }

   // cofactors, determinants, inverse
   inline T cofactor(unsigned int r, unsigned int c) const
      {
         return (((r+c)&1) ? -1:1) * m_matrix[(r ? 0:1)][(c ? 0:1)];
      }

   inline T det() const
      {
         return m_matrix[0][0]*cofactor(0,0) + m_matrix[0][1]*cofactor(0,1);
      }

   template <class S>
   inline void adjoint(const sglMat2<S>& mat)
      {
         T t[2][2];
         t[0][0] = mat.cofactor(0,0);
         t[0][1] = mat.cofactor(1,0);
         t[1][0] = mat.cofactor(0,1);
         t[1][1] = mat.cofactor(1,1);
         set(t[0][0], t[0][1], t[1][0], t[1][1]);
      }
   // adjoint divided by deteminant
   template <class S>
   inline bool inverse(const sglMat2<S>& mat)
      {
         T d = mat.det();
         if (SGL_ABS(d) < FLT_MIN) return false;
         adjoint(mat);
         *this *= (T)(1.0/d);
         return true;
      }

   // stream operators
#if defined(TEMPLATE_OSTREAM_HACK)
   template <class S>
   friend ostream& operator<<(ostream& ostrm, const sglMat2<S>& mat);

   template <class S>
   friend istream& operator>>(istream& istrm, sglMat2<S>& mat);
#else
   friend ostream& operator<<(ostream& ostrm, const sglMat2<T>& mat)
        {
            ostrm << '(';
            for (unsigned int i=0; i<2; i++)
                ostrm << '(' << mat[i][0] << ',' << mat[i][1] << ')';
            return (ostrm << ')');
        }

   friend istream& operator>>(istream& istrm, sglMat2<T>& mat)
        {
            return (istrm >> mat[0][0] >> mat[0][1] >> mat[1][0] >> mat[1][1]);
        }
#endif

private:
   T m_matrix[2][2];
};

//----------------------------------------------------------------------------
#if defined(TEMPLATE_OSTREAM_HACK)
template <class T>
inline ostream& operator<<(ostream& ostrm, const sglMat2<T>& mat)
{
    return (ostrm << '(' << mat[0] << ", " << mat[1] << ')');
}

template <class T>
inline istream& operator>>(istream& istrm, sglMat2<T>& mat)
{
    return (istrm >> mat[0][0] >> mat[0][1] >> mat[1][0] >> mat[1][1]);
}
#endif
//----------------------------------------------------------------------------


//----------------------------------------------------------------------------
// vector/matrix2x2 operations
//----------------------------------------------------------------------------

// ---------------------------------------------------------------------------
// v' = v' * m (or v = m' * v)
// ---------------------------------------------------------------------------
template <class S, class T>
inline void mul(sglVec2<T>& vec, const sglMat2<S>& mat)
{
   T x    = (T)(vec[0]*mat[0][0] + vec[1]*mat[1][0]);
   vec[1] = (T)(vec[0]*mat[0][1] + vec[1]*mat[1][1]);
   vec[0] = x;
}


// ---------------------------------------------------------------------------
// dest' = v' * m (or dest = m' * v)
// ---------------------------------------------------------------------------
template <class R, class S, class T>
inline void mul(const sglVec2<R>& vec, const sglMat2<S>& mat,
                sglVec2<T>& dest)
{
   T x     = (T)(vec[0]*mat[0][0] + vec[1]*mat[1][0]);
   dest[1] = (T)(vec[0]*mat[0][1] + vec[1]*mat[1][1]);
   dest[0] = x;
}


// ---------------------------------------------------------------------------
// v = m * v (or v' = v' * m')
// ---------------------------------------------------------------------------
template <class S, class T>
inline void mulTranspose(sglVec2<T>& vec, const sglMat2<S>& mat)
{
   T x    = (T)(vec[0]*mat[0][0] + vec[1]*mat[0][1]);
   vec[1] = (T)(vec[0]*mat[1][0] + vec[1]*mat[1][1]);
   vec[0] = x;
}

// ---------------------------------------------------------------------------
// dest = m * v (or dest' = v' * m')
// ---------------------------------------------------------------------------
template <class R, class S, class T>
inline void mulTranspose(const sglVec2<R>& vec, const sglMat2<S>& mat,
                         sglVec2<T>& dest)
{
   T x     = (T)(vec[0]*mat[0][0] + vec[1]*mat[0][1]);
   dest[1] = (T)(vec[0]*mat[1][0] + vec[1]*mat[1][1]);
   dest[0] = x;
}


#endif
