/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglMat3.hpp
 *  Created: 28 February 1999
 *  Summary: 3x3 templated matrix class
 *****************************************************************************/

#ifndef __SGL_MAT3_HPP
#define __SGL_MAT3_HPP

#include <sgl.h>
#include <string.h> // memcpy
#include <sglVector.hpp>
#include <sglVec2.hpp>
#include <sglVec3.hpp>

// ---------------------------------------------------------------------------
template <class T>
class sglMat3
{
public:
   // constructors
   sglMat3() {};

   sglMat3(T m00, T m01, T m02, T m10, T m11, T m12, T m20, T m21, T m22)
      {
         set(m00, m01, m02, m10, m11, m12, m20, m21, m22);
      }

   template <class S>
   sglMat3(const sglMat3<S>& mat)
      {
         set(mat[0][0], mat[0][1], mat[0][2],
             mat[1][0], mat[1][1], mat[1][2],
             mat[2][0], mat[2][1], mat[2][2]);
      }

   template <class S>
   sglMat3(S mat[3][3])
      {
         set(mat[0][0], mat[0][1], mat[0][2],
             mat[1][0], mat[1][1], mat[1][2],
             mat[2][0], mat[2][1], mat[2][2]);
      }

   // destructor
   ~sglMat3() {};  // not virtual, do not subclass

   inline T *getPtr() const { return (T *)m_matrix; }

   // indexing operators
   inline sglVec3<T>& operator[](unsigned int i)
      {
         return *(sglVec3<T> *) m_matrix[i];
      }
   inline const sglVec3<T>& operator[](unsigned int i) const
      {
         return *(sglVec3<T> *) m_matrix[i];
      }

   // set/get
   template <class S>
   void set(S m00, S m01, S m02, S m10, S m11, S m12, S m20, S m21, S m22)
      {
         m_matrix[0][0] = (T)m00; m_matrix[0][1] = (T)m01;
         m_matrix[0][2] = (T)m02;
         m_matrix[1][0] = (T)m10; m_matrix[1][1] = (T)m11;
         m_matrix[1][2] = (T)m12;
         m_matrix[2][0] = (T)m20; m_matrix[2][1] = (T)m21;
         m_matrix[2][2] = (T)m22;
      }

   template <class S>
   inline void setRow(const sglVec3<S>& vec, unsigned int i)
      {
         m_matrix[i][0] = (T)vec[0]; m_matrix[i][1] = (T)vec[1];
         m_matrix[i][2] = (T)vec[2];
      }

   template <class S>
   inline void setCol(const sglVec3<S>& vec, unsigned int i)
      {
         m_matrix[0][i] = (T)vec[0]; m_matrix[1][i] = (T)vec[1];
         m_matrix[2][i] = (T)vec[2];
      }

   template <class S>
   inline void getRow(sglVec3<S>& vec, unsigned int i)
      {
         vec[0] = (S)m_matrix[i][0]; vec[1] = (S)m_matrix[i][1];
         vec[2] = (S)m_matrix[i][2];
      }

   template <class S>
   inline void getCol(sglVec3<S>& vec, unsigned int i)
      {
         vec[0] = (S)m_matrix[0][i]; vec[1] = (S)m_matrix[1][i];
         vec[2] = (S)m_matrix[2][i];
      }

   // assignment operators
   template <class S>
   inline sglMat3& operator=(const sglMat3<S>& mat)
      {
         set(mat[0][0], mat[0][1], mat[0][2],
             mat[1][0], mat[1][1], mat[1][2],
             mat[2][0], mat[2][1], mat[2][2]);
         return *this;
      }

   // boolean operators
   template <class S>
   inline bool almostEqual(const sglMat3<S>& mat, double eps) const
      {
         return ((*this)[0].almostEqual(mat[0], eps) &&
                 (*this)[1].almostEqual(mat[1], eps) &&
                 (*this)[2].almostEqual(mat[2], eps));
      }

   template <class S>
   inline bool operator==(const sglMat3<S>& mat) const
      {
         return ((*this)[0] == mat[0] &&
                 (*this)[1] == mat[1] &&
                 (*this)[2] == mat[2]);
      }

   template <class S>
   inline bool operator!=(const sglMat3<S>& mat) const
      {
         return ((*this)[0] != mat[0] ||
                 (*this)[1] != mat[1] ||
                 (*this)[2] != mat[2]);
      }

   // math operators
   template <class S>
   inline sglMat3& operator+=(const sglMat3<S>& mat)
      {
         (*this)[0] += mat[0]; (*this)[1] += mat[1]; (*this)[2] += mat[2];
         return *this;
      }
   template <class S>
   inline sglMat3& operator-=(const sglMat3<S>& mat)
      {
         (*this)[0] -= mat[0]; (*this)[1] -= mat[1]; (*this)[2] -= mat[2];
         return *this;
      }
   template <class S>
   inline sglMat3& operator*=(const sglMat3<S>& mat)
      {
         mul(*this, mat); return *this;
      }
   inline sglMat3& operator*=(const T s)
      {
         scale(*this, s, s, s); return *this;
      }

   template <class S>
   inline const sglMat3<T> operator+(const sglMat3<S> &rhs) const
      {
         return sglMat3(m_matrix[0][0] + rhs[0][0],
                        m_matrix[0][1] + rhs[0][1],
                        m_matrix[0][2] + rhs[0][2],
                        m_matrix[1][0] + rhs[1][0],
                        m_matrix[1][1] + rhs[1][1],
                        m_matrix[1][2] + rhs[1][2],
                        m_matrix[2][0] + rhs[2][0],
                        m_matrix[2][1] + rhs[2][1],
                        m_matrix[2][2] + rhs[2][2]);
      }
   template <class S>
   inline const sglMat3<T> operator-(const sglMat3<S> &rhs) const
      {
         return sglMat3(m_matrix[0][0] - rhs[0][0],
                        m_matrix[0][1] - rhs[0][1],
                        m_matrix[0][2] - rhs[0][2],
                        m_matrix[1][0] - rhs[1][0],
                        m_matrix[1][1] - rhs[1][1],
                        m_matrix[1][2] - rhs[1][2],
                        m_matrix[2][0] - rhs[2][0],
                        m_matrix[2][1] - rhs[2][1],
                        m_matrix[2][2] - rhs[2][2]);
      }
   template <class S>
   inline const sglMat3<T> operator*(const sglMat3<S> &rhs) const
      {
         return sglMat3(m_matrix[0][0]*rhs[0][0] +
                        m_matrix[0][1]*rhs[1][0] +
                        m_matrix[0][2]*rhs[2][0],
                        m_matrix[0][0]*rhs[0][1] +
                        m_matrix[0][1]*rhs[1][1] +
                        m_matrix[0][2]*rhs[2][1],
                        m_matrix[0][0]*rhs[0][2] +
                        m_matrix[0][1]*rhs[1][2] +
                        m_matrix[0][2]*rhs[2][2],
                        m_matrix[1][0]*rhs[0][0] +
                        m_matrix[1][1]*rhs[1][0] +
                        m_matrix[1][2]*rhs[2][0],
                        m_matrix[1][0]*rhs[0][1] +
                        m_matrix[1][1]*rhs[1][1] +
                        m_matrix[1][2]*rhs[2][1],
                        m_matrix[1][0]*rhs[0][2] +
                        m_matrix[1][1]*rhs[1][2] +
                        m_matrix[1][2]*rhs[2][2],
                        m_matrix[2][0]*rhs[0][0] +
                        m_matrix[2][1]*rhs[1][0] +
                        m_matrix[2][2]*rhs[2][0],
                        m_matrix[2][0]*rhs[0][1] +
                        m_matrix[2][1]*rhs[1][1] +
                        m_matrix[2][2]*rhs[2][1],
                        m_matrix[2][0]*rhs[0][2] +
                        m_matrix[2][1]*rhs[1][2] +
                        m_matrix[2][2]*rhs[2][2]);
      }
   inline const sglMat3<T> operator*(T scalar) const
      {
         return sglMat3(m_matrix[0][0]*scalar,
                        m_matrix[0][1]*scalar,
                        m_matrix[0][2]*scalar,
                        m_matrix[1][0]*scalar,
                        m_matrix[1][1]*scalar,
                        m_matrix[1][2]*scalar,
                        m_matrix[2][0]*scalar,
                        m_matrix[2][1]*scalar,
                        m_matrix[2][2]*scalar);
      }

   // explicit math operations (for backward compatibility)
   template <class S1, class S2>
   inline void add(const sglMat3<S1>& mat1, const sglMat3<S2>& mat2)
      {
         (*this)[0].add(mat1[0],mat2[0]);
         (*this)[1].add(mat1[1],mat2[1]);
         (*this)[2].add(mat1[2],mat2[2]);
      }
   template <class S1, class S2>
   inline void sub(const sglMat3<S1>& mat1, const sglMat3<S2>& mat2)
      {
         (*this)[0].sub(mat1[0],mat2[0]);
         (*this)[1].sub(mat1[1],mat2[1]);
         (*this)[2].sub(mat1[2],mat2[2]);
      }
   template <class S>
   inline void scale(const sglMat3<S>& mat, T sx, T sy, T sz)
      {
         (*this)[0].scale(mat[0],sx);
         (*this)[1].scale(mat[1],sy);
         (*this)[2].scale(mat[2],sz);
      }
   // matrix multiply
   template <class S1, class S2>
   void mul(const sglMat3<S1>& mat1, const sglMat3<S2>& mat2)
      {
         T t[3][3];
         for (unsigned int i=0; i<3; i++)
            for (unsigned int j=0; j<3; j++)
               t[i][j] = (mat1[i][0]*mat2[0][j] +
                          mat1[i][1]*mat2[1][j] +
                          mat1[i][2]*mat2[2][j]);
         memcpy(m_matrix, t, sizeof(m_matrix));
      }

   // methods to build a matrix
   inline void buildIdentity()
      {
         set(1, 0, 0,  0, 1, 0,  0, 0, 1);
      }

   inline void buildScale(T sx, T sy, T sz)
      {
         m_matrix[0][0] = sx; m_matrix[0][1] = 0;  m_matrix[0][2] = 0;
         m_matrix[1][0] = 0;  m_matrix[1][1] = sy; m_matrix[1][2] = 0;
         m_matrix[2][0] = 0;  m_matrix[2][1] = 0;  m_matrix[2][2] = sz;
      }

   void buildRotation(float x, float y, float z, double angle)  // radians
      {
         double mag = x*x + y*y + z*z;

         if ((angle == 0.0) || (mag == 0.0))
         {
            buildIdentity();
            return;
         }

         mag = sqrt(mag);
         x /= mag;  y /= mag;  z /= mag;

         double s = sin(angle);
         double c = cos(angle);
         double a = 1.0 - c;

         double xx = x*x, yy = y*y, zz = z*z;
         double xy = x*y, yz = y*z, zx = z*x;
         double sx = x*s, sy = y*s, sz = z*s;

         m_matrix[0][0] = (T)(a*xx + c);
         m_matrix[0][1] = (T)(a*xy + sz);
         m_matrix[0][2] = (T)(a*zx - sy);

         m_matrix[1][0] = (T)(a*xy - sz);
         m_matrix[1][1] = (T)(a*yy + c);
         m_matrix[1][2] = (T)(a*yz + sx);

         m_matrix[2][0] = (T)(a*zx + sy);
         m_matrix[2][1] = (T)(a*yz - sx);
         m_matrix[2][2] = (T)(a*zz + c);
      }

   /** Set the rotation by specifying euler angles.
       @param h Heading angle is the rotation about the z-axis.
       @param p Pitch angle is the rotation about the once-rotated x-axis.
       @param r Roll angle is the rotation about the twice-rotated y-axis.
   */
   void buildRotation(double h, double p, double r)  // radians
      {
         T sr = sin( p );
         T cr = cos( p );

         T sh = sin( r );
         T ch = cos( r );

         T sp = sin( h );
         T cp = cos( h );

         m_matrix[0][0] = -sp*sr*sh + cp*ch;
         m_matrix[0][1] =  cp*sr*sh + sp*ch;
         m_matrix[0][2] = -cr*sh;

         m_matrix[1][0] = -sp*cr;
         m_matrix[1][1] =  cp*cr;
         m_matrix[1][2] =  sr;

         m_matrix[2][0] =  sp*sr*ch + cp*sh;
         m_matrix[2][1] = -cp*sr*ch + sp*sh;
         m_matrix[2][2] =  cr*ch;
      }


   // other operations
   inline void transpose(const sglMat3& mat)
      {
         // diagonal first
         m_matrix[0][0] = mat[0][0];
         m_matrix[1][1] = mat[1][1];
         m_matrix[2][2] = mat[2][2];
         // then upper triangle
         T tmp = mat[0][1]; m_matrix[0][1] = mat[1][0]; m_matrix[1][0] = tmp;
           tmp = mat[0][2]; m_matrix[0][2] = mat[2][0]; m_matrix[2][0] = tmp;
           tmp = mat[1][2]; m_matrix[1][2] = mat[2][1]; m_matrix[2][1] = tmp;
      }

   // cofactors, determinants, inverse
   inline T cofactor(unsigned int r,unsigned int c) const
      {
         const unsigned int index[3][2] = { {1,2}, {0,2}, {0,1} };
         unsigned int i0 = index[r][0], i1 = index[r][1];
         unsigned int j0 = index[c][0], j1 = index[c][1];
         return (((r+c)&1) ? -1: 1)*(m_matrix[i0][j0]*m_matrix[i1][j1] -
                                     m_matrix[i0][j1]*m_matrix[i1][j0]);
      }

   inline T det() const
      {
         return (m_matrix[0][0] * cofactor(0,0) +
                 m_matrix[0][1] * cofactor(0,1) +
                 m_matrix[0][2] * cofactor(0,2));
      }

   template <class S>
   void adjoint(const sglMat3<S>& mat)
      {
         T t[3][3];
         t[0][0] = mat.cofactor(0,0);
         t[0][1] = mat.cofactor(1,0);
         t[0][2] = mat.cofactor(2,0);
         t[1][0] = mat.cofactor(0,1);
         t[1][1] = mat.cofactor(1,1);
         t[1][2] = mat.cofactor(2,1);
         t[2][0] = mat.cofactor(0,2);
         t[2][1] = mat.cofactor(1,2);
         t[2][2] = mat.cofactor(2,2);
         set(t[0][0], t[0][1], t[0][2],
             t[1][0], t[1][1], t[1][2],
             t[2][0], t[2][1], t[2][2]);  // memcpy does not work here
      }

   // try a adjoint divided by deteminant
   template <class S>
   inline bool inverse(const sglMat3<S>& mat)
      {
         T d = mat.det();
         if (SGL_ABS(d) < FLT_MIN) return false;
         adjoint(mat);
         *this *= (T)(1.0/d);
         return true;
      }

   // stream operators
#if defined(TEMPLATE_OSTREAM_HACK)
   template <class S>
   friend ostream& operator<<(ostream& ostrm, const sglMat3<S>& mat);

   template <class S>
   friend istream& operator>>(istream& istrm, sglMat3<S>& mat);
#else
   friend ostream& operator<<(ostream& ostrm, const sglMat3<T>& mat)
        {
            ostrm << '(';
            for (unsigned int i=0; i<3; i++)
                ostrm << '(' << mat[i][0] << ',' << mat[i][1] << ','
                      << mat[i][2] << ')';
            return (ostrm << ')');
        }

   friend istream& operator>>(istream& istrm, sglMat3<T>& mat)
        {
            for (unsigned int i=0;i<3;i++)
                for (unsigned int j=0;j<3;j++)
                    istrm >> mat[i][j];
            return istrm;
        }
#endif

private:
   //sglVec3<T> m_matrix[3];
   T m_matrix[3][3];
};

//----------------------------------------------------------------------------
#if defined(TEMPLATE_OSTREAM_HACK)
template <class T>
inline ostream& operator<<(ostream& ostrm, const sglMat3<T>& mat)
{
    return (ostrm << '(' << mat[0] << ", " << mat[1]
            << ", " << mat[2] << ')');
}

template <class T>
inline istream& operator>>(istream& istrm, sglMat3<T>& mat)
{
    for (unsigned int i=0;i<3;i++)
        for (unsigned int j=0;j<3;j++)
            istrm >> mat[i][j];
    return istrm;
}
#endif
//----------------------------------------------------------------------------

//----------------------------------------------------------------------------
// vector/matrix3x3 operations
//----------------------------------------------------------------------------

// ---------------------------------------------------------------------------
// this' = this' * m (or this = m' * this)
// ---------------------------------------------------------------------------
template <class S, class T>
inline void mul(sglVec3<T>& vec, const sglMat3<S>& mat)
{
   T x    = (T)(vec[0]*mat[0][0] + vec[1]*mat[1][0] + vec[2]*mat[2][0]);
   T y    = (T)(vec[0]*mat[0][1] + vec[1]*mat[1][1] + vec[2]*mat[2][1]);
   vec[2] = (T)(vec[0]*mat[0][2] + vec[1]*mat[1][2] + vec[2]*mat[2][2]);
   vec[0] = x; vec[1] = y;
}

// ---------------------------------------------------------------------------
//  dest' = v' * m (or dest = m' * v)
// ---------------------------------------------------------------------------
template <class R, class S, class T>
inline void mul(const sglVec3<R> &vec, const sglMat3<S>& mat,
                sglVec3<T>& dest)
{
   T x     = (T)(vec[0]*mat[0][0] + vec[1]*mat[1][0] + vec[2]*mat[2][0]);
   T y     = (T)(vec[0]*mat[0][1] + vec[1]*mat[1][1] + vec[2]*mat[2][1]);
   dest[2] = (T)(vec[0]*mat[0][2] + vec[1]*mat[1][2] + vec[2]*mat[2][2]);
   dest[0] = x; dest[1] = y;
}

// ---------------------------------------------------------------------------
//  dest' = [v[0] v[1] z] * m (or dest = m' * v)
// ---------------------------------------------------------------------------
template <class R, class S, class T>
inline void mul(const sglVec2<R>& vec, R z, const sglMat3<S>& mat,
                sglVec3<T>& dest)
{
   dest[0] = (T)(vec[0]*mat[0][0] + vec[1]*mat[1][0] + z*mat[2][0]);
   dest[1] = (T)(vec[0]*mat[0][1] + vec[1]*mat[1][1] + z*mat[2][1]);
   dest[2] = (T)(vec[0]*mat[0][2] + vec[1]*mat[1][2] + z*mat[2][2]);
}

// ---------------------------------------------------------------------------
// v' = v' * m' (or v = m * v)
// ---------------------------------------------------------------------------
template <class S, class T>
inline void mulTranspose(sglVec3<T>& vec, const sglMat3<S>& mat)
{
   T x    = (T)(vec[0]*mat[0][0] + vec[1]*mat[0][1] + vec[2]*mat[0][2]);
   T y    = (T)(vec[0]*mat[1][0] + vec[1]*mat[1][1] + vec[2]*mat[1][2]);
   vec[2] = (T)(vec[0]*mat[2][0] + vec[1]*mat[2][1] + vec[2]*mat[2][2]);
   vec[0] = x; vec[1] = y;
}

// ---------------------------------------------------------------------------
//  dest' = v' * m' (or dest = m * v)
// ---------------------------------------------------------------------------
template <class R, class S, class T>
inline void mulTranspose(const sglVec3<R> &vec, const sglMat3<S>& mat,
                         sglVec3<T> &dest)
{
   T x     = (T)(vec[0]*mat[0][0] + vec[1]*mat[0][1] + vec[2]*mat[0][2]);
   T y     = (T)(vec[0]*mat[1][0] + vec[1]*mat[1][1] + vec[2]*mat[1][2]);
   dest[2] = (T)(vec[0]*mat[2][0] + vec[1]*mat[2][1] + vec[2]*mat[2][2]);
   dest[0] = x; dest[1] = y;
}

// ---------------------------------------------------------------------------
//  dest' = v' * m' (or dest = m * v)
// ---------------------------------------------------------------------------
template <class R, class S, class T>
inline void mulTranspose(const sglVec2<R> &vec, R z, const sglMat3<S>& mat,
                         sglVec3<T>& dest)
{
   dest[0] = (T)(vec[0]*mat[0][0] + vec[1]*mat[0][1] + z*mat[0][2]);
   dest[1] = (T)(vec[0]*mat[1][0] + vec[1]*mat[1][1] + z*mat[1][2]);
   dest[2] = (T)(vec[0]*mat[2][0] + vec[1]*mat[2][1] + z*mat[2][2]);
}

#endif
