/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglMaterial.cpp
 *  Created: 30 September 1997
 *  Summary: container to hold OpenGL material parameters
 *****************************************************************************/

#include <sgl.h>
#include <sglObject.hpp>
#include <sglMaterial.hpp>

inline float threshold(float x)
{
   if (x > 1.0) return 1.0;
   if (x < 0.0) return 0.0;
   return x;
}

//----------------------------------------------------------------------------
//    Summary: default constructor initializes to default OpenGL material
// Parameters: 
//    Returns: 
//----------------------------------------------------------------------------
sglMaterial::sglMaterial()
      : sglStatelet(eMATERIAL),
        m_front_ambient(0.2f, 0.2f, 0.2f, 1.0f),
        m_front_diffuse(0.8f, 0.8f, 0.8f, 1.0f),
        m_front_specular(0.0f, 0.0f, 0.0f, 1.0f),
        m_front_emission(0.0f, 0.0f, 0.0f, 1.0f),
        m_back_ambient(0.2f, 0.2f, 0.2f, 1.0f),
        m_back_diffuse(0.8f, 0.8f, 0.8f, 1.0f),
        m_back_specular(0.0f, 0.0f, 0.0f, 1.0f),
        m_back_emission(0.0f, 0.0f, 0.0f, 1.0f),
        m_front_shininess(0.0f),
        m_back_shininess(0.0f),
        m_color_material_mode(eOFF),
        m_color_material_side(eFRONT_AND_BACK)
{
   m_index = UINT_MAX;
}

//----------------------------------------------------------------------------
sglMaterial::~sglMaterial()
{
}

//----------------------------------------------------------------------------
void sglMaterial::setAmbient(float r, float g, float b, SideEnum side)
{
   if (side == eFRONT || side == eFRONT_AND_BACK)
   {
      m_front_ambient[0] = threshold(r);
      m_front_ambient[1] = threshold(g);
      m_front_ambient[2] = threshold(b);
   }
   if (side == eBACK || side == eFRONT_AND_BACK)
   {
      m_back_ambient[0] = threshold(r);
      m_back_ambient[1] = threshold(g);
      m_back_ambient[2] = threshold(b);
   }
}

//----------------------------------------------------------------------------
void sglMaterial::getFrontAmbient(float &r, float &g, float &b) const
{
   r = m_front_ambient[0];
   g = m_front_ambient[1];
   b = m_front_ambient[2];
}

//----------------------------------------------------------------------------
void sglMaterial::getBackAmbient(float &r, float &g, float &b) const
{
   r = m_back_ambient[0];
   g = m_back_ambient[1];
   b = m_back_ambient[2];
}

//----------------------------------------------------------------------------
void sglMaterial::setDiffuse(float r, float g, float b, SideEnum side)
{
   if (side == eFRONT || side == eFRONT_AND_BACK)
   {
      m_front_diffuse[0] = threshold(r);
      m_front_diffuse[1] = threshold(g);
      m_front_diffuse[2] = threshold(b);
   }
   if (side == eBACK || side == eFRONT_AND_BACK)
   {
      m_back_diffuse[0] = threshold(r);
      m_back_diffuse[1] = threshold(g);
      m_back_diffuse[2] = threshold(b);
   }
}

//----------------------------------------------------------------------------
void sglMaterial::getFrontDiffuse(float &r, float &g, float &b) const
{
   r = m_front_diffuse[0];
   g = m_front_diffuse[1];
   b = m_front_diffuse[2];
}

//----------------------------------------------------------------------------
void sglMaterial::getBackDiffuse(float &r, float &g, float &b) const
{
   r = m_back_diffuse[0];
   g = m_back_diffuse[1];
   b = m_back_diffuse[2];
}

//----------------------------------------------------------------------------
void sglMaterial::setSpecular(float r, float g, float b, SideEnum side)
{
   if (side == eFRONT || side == eFRONT_AND_BACK)
   {
      m_front_specular[0] = threshold(r);
      m_front_specular[1] = threshold(g);
      m_front_specular[2] = threshold(b);
   }
   if (side == eBACK || side == eFRONT_AND_BACK)
   {
      m_back_specular[0] = threshold(r);
      m_back_specular[1] = threshold(g);
      m_back_specular[2] = threshold(b);
   }
}

//----------------------------------------------------------------------------
void sglMaterial::getFrontSpecular(float &r, float &g, float &b) const
{
   r = m_front_specular[0];
   g = m_front_specular[1];
   b = m_front_specular[2];
}

//----------------------------------------------------------------------------
void sglMaterial::getBackSpecular(float &r, float &g, float &b) const
{
   r = m_back_specular[0];
   g = m_back_specular[1];
   b = m_back_specular[2];
}

//----------------------------------------------------------------------------
void sglMaterial::setEmission(float r, float g, float b, SideEnum side)
{
   if (side == eFRONT || side == eFRONT_AND_BACK)
   {
      m_front_emission[0] = threshold(r);
      m_front_emission[1] = threshold(g);
      m_front_emission[2] = threshold(b);
   }
   if (side == eBACK || side == eFRONT_AND_BACK)
   {
      m_back_emission[0] = threshold(r);
      m_back_emission[1] = threshold(g);
      m_back_emission[2] = threshold(b);
   }
}

//----------------------------------------------------------------------------
void sglMaterial::getFrontEmission(float &r, float &g, float &b) const
{
   r = m_front_emission[0];
   g = m_front_emission[1];
   b = m_front_emission[2];
}

//----------------------------------------------------------------------------
void sglMaterial::getBackEmission(float &r, float &g, float &b) const
{
   r = m_back_emission[0];
   g = m_back_emission[1];
   b = m_back_emission[2];
}

//----------------------------------------------------------------------------
void sglMaterial::setAlpha(float alpha, SideEnum side)
{
   // On page 197 of the OpenGL Reference Manual (Red Book) for Version 1.1,
   // it states:  "Regardless of what alpha values are supplied for other
   // parameters, the alpha value at any particular vertex is the diffuse
   // material alpha value."
   if (side == eFRONT || side == eFRONT_AND_BACK)
      m_front_diffuse[3] = threshold(alpha);
   if (side == eBACK || side == eFRONT_AND_BACK)
      m_back_diffuse[3] = threshold(alpha);
}

//----------------------------------------------------------------------------
void sglMaterial::setShininess(float shininess, SideEnum side)
{
   if (side == eFRONT || side == eFRONT_AND_BACK)
   {
      m_front_shininess = shininess;
      if (m_front_shininess < 0.0) m_front_shininess = 0.0;
      if (m_front_shininess > 128.0) m_front_shininess = 128.0;
   }
   if (side == eBACK || side == eFRONT_AND_BACK)
   {
      m_back_shininess = shininess;
      if (m_back_shininess < 0.0) m_back_shininess = 0.0;
      if (m_back_shininess > 128.0) m_back_shininess = 128.0;
   }
}

//----------------------------------------------------------------------------
void sglMaterial::setColorMaterial(SideEnum side, ColorModeEnum mode)
{
   m_color_material_mode = mode;
   m_color_material_side = side;
}

//----------------------------------------------------------------------------
void sglMaterial::apply(sglCurrState*) const
{
   if (m_color_material_mode != eOFF)
   {
      glColorMaterial((GLenum)m_color_material_side,
                      (GLenum)m_color_material_mode);
      glEnable(GL_COLOR_MATERIAL);
   }
   else
   {
      glDisable(GL_COLOR_MATERIAL);
   }

   // set material properties
   glMaterialfv(GL_FRONT, GL_EMISSION,  m_front_emission.getPtr());
   glMaterialfv(GL_FRONT, GL_AMBIENT,   m_front_ambient.getPtr());
   glMaterialfv(GL_FRONT, GL_DIFFUSE,   m_front_diffuse.getPtr());
   glMaterialfv(GL_FRONT, GL_SPECULAR,  m_front_specular.getPtr());
   glMaterialf (GL_FRONT, GL_SHININESS, m_front_shininess);
   glMaterialfv(GL_BACK, GL_EMISSION,   m_back_emission.getPtr());
   glMaterialfv(GL_BACK, GL_AMBIENT,    m_back_ambient.getPtr());
   glMaterialfv(GL_BACK, GL_DIFFUSE,    m_back_diffuse.getPtr());
   glMaterialfv(GL_BACK, GL_SPECULAR,   m_back_specular.getPtr());
   glMaterialf (GL_BACK, GL_SHININESS,  m_back_shininess);
}

//----------------------------------------------------------------------------
void sglMaterial::printInfo(ostream &ostrm, const char *indent_string) const
{
   sglStatelet::printInfo(ostrm, indent_string);
   ostrm << indent_string << "front ambient color:  "
         << m_front_ambient << endl;
   ostrm << indent_string << "front diffuse color:  "
         << m_front_diffuse << endl;
   ostrm << indent_string << "front specular color: "
         << m_front_specular << endl;
   ostrm << indent_string << "front emission color: "
         << m_front_emission << endl;
   ostrm << indent_string << "front shininess: " << m_front_shininess << endl;
   ostrm << indent_string << "front alpha: " << m_front_diffuse[3] << endl;
   ostrm << indent_string << "back  ambient color:  "
         << m_back_ambient << endl;
   ostrm << indent_string << "back  diffuse color:  "
         << m_back_diffuse << endl;
   ostrm << indent_string << "back  specular color: "
         << m_back_specular << endl;
   ostrm << indent_string << "back  emission color: "
         << m_back_emission << endl;
   ostrm << indent_string << "back  shininess: " << m_back_shininess << endl;
   ostrm << indent_string << "back  alpha: " << m_back_diffuse[3] << endl;
   ostrm << indent_string << "color mode: ";
   switch (m_color_material_mode)
   {
      case eOFF:
         ostrm << "disabled" << endl;
         break;
      case eAMBIENT:
         ostrm << "AMBIENT" << endl;
         break;
      case eDIFFUSE:
         ostrm << "DIFFUSE" << endl;
         break;
      case eAMBIENT_AND_DIFFUSE:
         ostrm << "AMBIENT_AND_DIFFUSE" << endl;
         break;
      case eSPECULAR:
         ostrm << "SPECULAR" << endl;
         break;
      case eEMISSION:
         ostrm << "EMISSION" << endl;
         break;
   }
   ostrm << indent_string << "color mode side: ";
   switch (m_color_material_side)
   {
      case eFRONT:
         ostrm << "FRONT" << endl;
         break;
      case eBACK:
         ostrm << "BACK" << endl;
         break;
      case eFRONT_AND_BACK:
         ostrm << "FRONT_AND_BACK" << endl;
         break;
   }
}

