/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglMaterial.hpp
 *  Created: 30 September 1997
 *  Summary: container for OpenGL material parameters and drawing code
 *****************************************************************************/

#ifndef __SGL_MATERIAL_HPP
#define __SGL_MATERIAL_HPP

#include <sgl.h>
#include <sglVector.hpp>
#include <sglStatelet.hpp>

//============================================================================

/**

This statelet encapsulates the glMaterial*(..) as well as glColorMaterial() and
glEnable(GL_COLOR_MATERIAL) functions that specifies material properties of an
object for use in lighting calculations (when the sglLighting statelet is
enabled).  This statelet contains the constant color parameters for both front
and back faces including ambient, diffuse, specular and emissive color
components.  It also manages alpha (for transparency) and shininess parameters.

Using color material enables one of the color components to track the vertex
color of the geometry (if present) rather than a constant color specified by
the components described above.  To do this, the
sglMaterial::setColorMaterial(..) member function is called with one of the
SideEnum and ColorModeEnum tokens.

The default constructor initializes the object to match the OpenGL defaults for
materials as follows (for both front and back):

 - ambient = (0.2f, 0.2f, 0.2f)
 - diffuse = (0.8f, 0.8f, 0.8f)
 - specular = (0.0f, 0.0f, 0.0f)
 - emission = (0.0f, 0.0f, 0.0f)
 - alpha = 1
 - shininess = 0

Note that (1) the alpha value is handled separately in this class and
corresponds to the alpha component of the diffuse color component in OpenGL,
(2) the color material is disabled by default, and (3) this class does not
support the color index modes.

<b>Example:</b>

<pre>
   sglMaterial *statelet = new sglMaterial;
   statelet->setAmbient(0.f, 0.f, 0.f);                       // FRONT_AND_BACK
   statelet->setAmbient(0.f, 0.f, 0.f, sglMaterial::eBACK);  // BACK only
   statelet->setDiffuse(1.f, 1.f, 1.f);
   statelet->setSpecular(1.f, 1.f, 1.f);
   statelet->setAlpha(1.0f);
   statelet->setShininess(64.0f);

   // overrides the call to setAmbient and setDiffuse for the front side
   statelet->setColorMaterial(sglMaterial::eFRONT,
                              sglMaterial::eAMBIENT_AND_DIFFUSE);
</pre>

 */

class SGL_DLL_API sglMaterial : public sglStatelet
{
public:
   /// Tokens to define which face a parameter is to be applied to
   enum SideEnum
   {
      /// the front face (as defined by the sglFrontFace statelet)
      eFRONT           = GL_FRONT,

      /// the back face
      eBACK            = GL_BACK,

      /// both sides
      eFRONT_AND_BACK  = GL_FRONT_AND_BACK
   };

   /// Color component tokens used for color material specification
   enum ColorModeEnum
   {
      /// disable color material
      eOFF                  = 0,

      /// map color vertices to the ambient color component
      eAMBIENT              = GL_AMBIENT,

      /// map color vertices to the diffuse color component
      eDIFFUSE              = GL_DIFFUSE,

      /// map color vertices to the specular color component
      eSPECULAR             = GL_SPECULAR,

      /// map color vertices to the ambient and diffuse color components
      eAMBIENT_AND_DIFFUSE  = GL_AMBIENT_AND_DIFFUSE,

      /// map color vertices to the emissive color component
      eEMISSION             = GL_EMISSION
   };

public:
   /** Default constructor.  Initializes to the default OpenGL material
       parameters.
   */
   sglMaterial();

   /// virtual destructor
   virtual ~sglMaterial();

   /** Set the ambient color components
       @param r The red color component.
       @param g The green color component.
       @param b The blue color component.
       @param side One of eFRONT, eBACK, eFRONT_AND_BACK.  The default is
                   eFRONT_AND_BACK
   */
   void setAmbient(float  r, float  g, float  b,
                   SideEnum side = eFRONT_AND_BACK);

   /** Query the ambient color components for the front face.
       @param r On return, contains the red color component.
       @param g On return, contains the green color component.
       @param b On return, contains the blue color component.
    */
   void getFrontAmbient(float &r, float &g, float &b) const;

   /** Query the ambient color components for the back face.
       @param r On return, contains the red color component.
       @param g On return, contains the green color component.
       @param b On return, contains the blue color component.
    */
   void getBackAmbient(float &r, float &g, float &b) const;

   /** Set the diffuse color components
       @param r The red color component.
       @param g The green color component.
       @param b The blue color component.
       @param side One of eFRONT, eBACK, eFRONT_AND_BACK.  The default is
                   eFRONT_AND_BACK
   */
   void setDiffuse(float  r, float  g, float  b,
                   SideEnum side = eFRONT_AND_BACK);

   /** Query the diffuse color components for the front face.
       @param r On return, contains the red color component.
       @param g On return, contains the green color component.
       @param b On return, contains the blue color component.
    */
   void getFrontDiffuse(float &r, float &g, float &b) const;

   /** Query the diffuse color components for the back face.
       @param r On return, contains the red color component.
       @param g On return, contains the green color component.
       @param b On return, contains the blue color component.
    */
   void getBackDiffuse(float &r, float &g, float &b) const;

   /** Set the specular color components
       @param r The red color component.
       @param g The green color component.
       @param b The blue color component.
       @param side One of eFRONT, eBACK, eFRONT_AND_BACK.  The default is
                   eFRONT_AND_BACK
   */
   void setSpecular(float  r, float  g, float  b,
                    SideEnum side = eFRONT_AND_BACK);

   /** Query the specular color components for the front face.
       @param r On return, contains the red color component.
       @param g On return, contains the green color component.
       @param b On return, contains the blue color component.
    */
   void getFrontSpecular(float &r, float &g, float &b) const;

   /** Query the specular color components for the back face.
       @param r On return, contains the red color component.
       @param g On return, contains the green color component.
       @param b On return, contains the blue color component.
    */
   void getBackSpecular(float &r, float &g, float &b) const;

   /** Set the emissive color components
       @param r The red color component.
       @param g The green color component.
       @param b The blue color component.
       @param side One of eFRONT, eBACK, eFRONT_AND_BACK.  The default is
                   eFRONT_AND_BACK
   */
   void setEmission(float  r, float  g, float  b,
                    SideEnum side = eFRONT_AND_BACK);

   /** Query the emissive color components for the front face.
       @param r On return, contains the red color component.
       @param g On return, contains the green color component.
       @param b On return, contains the blue color component.
    */
   void getFrontEmission(float &r, float &g, float &b) const;

   /** Query the emissive color components for the back face.
       @param r On return, contains the red color component.
       @param g On return, contains the green color component.
       @param b On return, contains the blue color component.
    */
   void getBackEmission(float &r, float &g, float &b) const;

   /** Set the alpha component (opacity) of the material.
       @param alpha Opacity in the range from zero (transparent) to one
                    (opaque).  The default is 1.0
       @param side One of eFRONT, eBACK, eFRONT_AND_BACK.  The default is
                   eFRONT_AND_BACK
   */
   void setAlpha(float alpha, SideEnum side = eFRONT_AND_BACK);

   /** Query the alpha value for the front face.
       @return The front face alpha value.
   */
   float getFrontAlpha() const { return m_front_diffuse[3]; }

   /** Query the alpha value for the back face.
       @return The back face alpha value.
   */
   float getBackAlpha() const { return m_back_diffuse[3]; }

   /** Set the shininess factor, the specular exponent (used with specular
       color component to define the specular highlights), for either side.
       @param shininess A value from 0 (matte) to 128 (mirror).  Values out
                   side this range will be clamped.  The default is zero.
       @param side One of eFRONT, eBACK, eFRONT_AND_BACK.  The default is
                   eFRONT_AND_BACK
   */
   void setShininess(float shininess, SideEnum side = eFRONT_AND_BACK);

   /** Query the shininess (specular exponent) of the front face material.
       @return The shininess factor of the front face.
   */
   float getFrontShininess() const { return m_front_shininess; }

   /** Query the shininess (specular exponent) of the back face material.
       @return The shininess factor of the back face.
   */
   float getBackShininess() const { return m_back_shininess; }

   /** Specify the color material mode for a given side.
       @param side One of eFRONT, eBACK, eFRONT_AND_BACK.  The default is
           eFRONT_AND_BACK
       @param mode One of ColorModeEnum tokens.  If mode is eOFF this is
           disabled; otherwise, the color value of the material property set by
           the mode parameter is specified via the color values in the
           geometry and not by the parameters in this instance.  Note, it
           is not possible to specify a different color mode for front and
           back sides.
       default: eFRONT_AND_BACK and eOFF
   */
   void setColorMaterial(SideEnum side, ColorModeEnum mode);

   /** Query the current color material state.
       @param side On return, contains the current side that the color material
                   will be applied to.
       @param mode On return, contains the current color material color mode.
                   This value is only valid if side is not eOFF.
   */
   void getColorMaterial(SideEnum &side, ColorModeEnum &mode) const
      { mode = m_color_material_mode; side = m_color_material_side; }

   // See sglStatelet::apply() for details.
   void apply(sglCurrState *curr_state) const;

   // See sglStatelet::getSorted() for details.
   virtual bool getSorted() const { return true; }

   // See sglObject::printInfo() for details.
   virtual void printInfo(ostream &ostrm, const char *indent_string) const;

private:
   // copy constructor forbidden
   sglMaterial(sglMaterial &);
   // assignment operator forbidden
   sglMaterial &operator=(const sglMaterial &);

private:
   // material color properties
   sglVec4f m_front_ambient;
   sglVec4f m_front_diffuse;
   sglVec4f m_front_specular;
   sglVec4f m_front_emission;
   sglVec4f m_back_ambient;
   sglVec4f m_back_diffuse;
   sglVec4f m_back_specular;
   sglVec4f m_back_emission;

   // exponent material property
   float m_front_shininess;
   float m_back_shininess;

   // color material mode
   ColorModeEnum m_color_material_mode;
   SideEnum m_color_material_side;
};

#endif
