/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglMonoIndexedGeoSet.hpp
 *  Created: 28 August 1999
 *  Summary: GeoSet where vertex data is indexed via a single index list.
 *****************************************************************************/

#ifndef __SGL_MONO_INDEXED_GEOSET_HPP
#define __SGL_MONO_INDEXED_GEOSET_HPP

#include <sgl.h>
#include <sglGeoSet.hpp>

#include <vector>

//============================================================================

/**

sglMonoIndexedGeoSet is the abstract base class for all geometry types (strip
or otherwise) that use a single index list for all attributes.  This class is a
variation on the sglIndexedGeoSet class (which allowed a different index list
for each attribute).  This implies that any attribute array that is be bound as
sglGeoSet::ePER_VERTEX (the same as the coordinate array) use the same index
array.  The advantage of this class, is the possible efficiency gained through
the use of vertex arrays (when available).  Note that color and normal
attribute arrays may still be bound as sglGeoSet::eOVERALL or
sglGeoSet::ePER_PRIM but they will not be indexed as they must be in the
sglIndexedGeoSet class.

{\bf Concrete subclasses: sglMonoIndexedPointSet, sglMonoIndexedLineSet,
sglMonoIndexedTriangleSet, sglMonoIndexedQuadSet}

These are the concrete subclasses of sglMonoIndexedGeoSet class that implement
point, line, triangle, and quad primitive types.

{\bf Example:}
\begin{verbatim}
   sglVec3f *vertex = new sglVec3f[6];
   vertex[0].set( 0.0f, 0.0f, 0.0f);
   vertex[1].set( 0.0f, 1.0f, 0.0f);
   vertex[2].set( 1.0f, 0.0f, 0.0f);
   vertex[3].set( 1.0f,-1.0f, 0.0f);
   vertex[4].set(-1.0f,-1.0f, 0.0f);
   vertex[5].set(-1.0f, 0.0f, 0.0f);

   sglVec4f *color = new sglVec4f[5];
   color[0].set( 1.0f, 0.0f, 0.0f, 1.0f);
   color[1].set( 0.0f, 1.0f, 0.0f, 1.0f);
   color[2].set( 0.0f, 0.0f, 1.0f, 1.0f);
   color[3].set( 0.0f, 0.0f, 1.0f, 1.0f);
   color[4].set( 0.0f, 1.0f, 0.0f, 1.0f);

   unsigned int *vindex = new unsigned int[10];
   vindex[0] = 0; vindex[1] = 1;   // indices for a five pointed star (sorta)
   vindex[2] = 0; vindex[3] = 2;
   vindex[4] = 0; vindex[5] = 3;
   vindex[6] = 0; vindex[7] = 4;
   vindex[8] = 0; vindex[9] = 5;

   unsigned int *cindex = new unsigned int[5];
   cindex[0] = 0;
   cindex[1] = 1;
   cindex[2] = 2;
   cindex[3] = 2;
   cindex[4] = 1;

   sglMonoIndexedLineSet *line_gset = new sglMonoIndexedPointSet;
   line_gset->setNumPrims(5);
   line_gset->setCoordList(vertex);
   line_gset->setIndexList(vindex);
   line_gset->setColorList(sglGeoSet::ePER_PRIM, color);  // doesn't use index list
\end{verbatim} */

class SGL_DLL_API sglMonoIndexedGeoSet : public sglGeoSet
{
public:
   /// default constructor
   sglMonoIndexedGeoSet();
   /// virtual destructor
   virtual ~sglMonoIndexedGeoSet();

   /// Set the index list for ePER_VERTEX attributes.  Deprecated.
   void setIndexList(sglIndexType *alist)
      { setIndexList(new sglNoRefDataPtr(alist)); }

   /** Set the array of indices for ePER_VERTEX attributes.
       This will increment the reference count
       on the given data pointer, and decrement the reference count of
       the previous data pointer.  If the previous data pointer's refernce
       count goes to zero, it will be deleted.
       The indices must be of type sglIndexType.
       @param data Pointer to the list of indices.
       @param offset Number of bytes in the array before the first index.
       @param stride Number of bytes from the start of one index to the start
                     of the next index.  If zero, the indices are assumed
                     to be tightly packed (i.e. stride == sizeof(type)).
                     Stride must be non-negative.
   */
   void setIndexList(sglDataPtr *data,
                     int offset = 0,
                     int stride = 0);

   /// Get the index list for ePER_VERTEX attributes.
   sglDataPtr *getIndexList() const {return m_index.getDataPtr();}

   /// Get the number of bytes in the array before the first index.
   int getIndexOffset() const { return m_index.getOffset(); }

   /// Get the number of bytes from the start of one index to the next.
   int getIndexStride() const { return m_index.getStride(); }

   /** If the set of indices above sparsely indexes the vertex data,
       then it might be best to disable prefetching.  This will prevent
       unnecessarily processing unused vertices.
   */
   void setDisablePrefetch(bool val);
   ///
   bool getDisablePrefetch() {return m_disable_prefetch;}

protected:
   virtual bool computeBounds();
   virtual void computeNoBounds();
   virtual void updateDrawFunc();

private: // not implemented
   sglMonoIndexedGeoSet(const sglMonoIndexedGeoSet &);
   sglMonoIndexedGeoSet &operator=(const sglMonoIndexedGeoSet &);

   void drawVT(const vector<sglTexCoords> &tex_coords) const;
   void drawVNoT(const vector<sglTexCoords> &tex_coords) const;
   void drawVNpT(const vector<sglTexCoords> &tex_coords) const;
   void drawVNvT(const vector<sglTexCoords> &tex_coords) const;
   void drawVCoT(const vector<sglTexCoords> &tex_coords) const;
   void drawVCoNoT(const vector<sglTexCoords> &tex_coords) const;
   void drawVCoNpT(const vector<sglTexCoords> &tex_coords) const;
   void drawVCoNvT(const vector<sglTexCoords> &tex_coords) const;
   void drawVCpT(const vector<sglTexCoords> &tex_coords) const;
   void drawVCpNoT(const vector<sglTexCoords> &tex_coords) const;
   void drawVCpNpT(const vector<sglTexCoords> &tex_coords) const;
   void drawVCpNvT(const vector<sglTexCoords> &tex_coords) const;
   void drawVCvT(const vector<sglTexCoords> &tex_coords) const;
   void drawVCvNoT(const vector<sglTexCoords> &tex_coords) const;
   void drawVCvNpT(const vector<sglTexCoords> &tex_coords) const;
   void drawVCvNvT(const vector<sglTexCoords> &tex_coords) const;

#ifndef SGL_NO_VERTEX_ARRAYS
   void drawVTm(const vector<sglTexCoords> &tex_coords) const;
   void drawVNoTm(const vector<sglTexCoords> &tex_coords) const;
   void drawVNpTm(const vector<sglTexCoords> &tex_coords) const;
   void drawVNvTm(const vector<sglTexCoords> &tex_coords) const;
   void drawVCoTm(const vector<sglTexCoords> &tex_coords) const;
   void drawVCoNoTm(const vector<sglTexCoords> &tex_coords) const;
   void drawVCoNpTm(const vector<sglTexCoords> &tex_coords) const;
   void drawVCoNvTm(const vector<sglTexCoords> &tex_coords) const;
   void drawVCpTm(const vector<sglTexCoords> &tex_coords) const;
   void drawVCpNoTm(const vector<sglTexCoords> &tex_coords) const;
   void drawVCpNpTm(const vector<sglTexCoords> &tex_coords) const;
   void drawVCpNvTm(const vector<sglTexCoords> &tex_coords) const;
   void drawVCvTm(const vector<sglTexCoords> &tex_coords) const;
   void drawVCvNoTm(const vector<sglTexCoords> &tex_coords) const;
   void drawVCvNpTm(const vector<sglTexCoords> &tex_coords) const;
   void drawVCvNvTm(const vector<sglTexCoords> &tex_coords) const;

   void drawPrefetchVT(const vector<sglTexCoords> &tex_coords) const;
   void drawPrefetchVNoT(const vector<sglTexCoords> &tex_coords) const;
   void drawPrefetchVNvT(const vector<sglTexCoords> &tex_coords) const;
   void drawPrefetchVCoT(const vector<sglTexCoords> &tex_coords) const;
   void drawPrefetchVCoNoT(const vector<sglTexCoords> &tex_coords) const;
   void drawPrefetchVCoNvT(const vector<sglTexCoords> &tex_coords) const;
   void drawPrefetchVCvT(const vector<sglTexCoords> &tex_coords) const;
   void drawPrefetchVCvNoT(const vector<sglTexCoords> &tex_coords) const;
   void drawPrefetchVCvNvT(const vector<sglTexCoords> &tex_coords) const;

   void drawPrefetchVTm(const vector<sglTexCoords> &tex_coords) const;
   void drawPrefetchVNoTm(const vector<sglTexCoords> &tex_coords) const;
   void drawPrefetchVNvTm(const vector<sglTexCoords> &tex_coords) const;
   void drawPrefetchVCoTm(const vector<sglTexCoords> &tex_coords) const;
   void drawPrefetchVCoNoTm(const vector<sglTexCoords> &tex_coords) const;
   void drawPrefetchVCoNvTm(const vector<sglTexCoords> &tex_coords) const;
   void drawPrefetchVCvTm(const vector<sglTexCoords> &tex_coords) const;
   void drawPrefetchVCvNoTm(const vector<sglTexCoords> &tex_coords) const;
   void drawPrefetchVCvNvTm(const vector<sglTexCoords> &tex_coords) const;
#endif

   class Initializer;
   friend class Initializer;
   static void setupFuncTable();

protected:
   sglIndexPtr     m_index;    // List of indexes
   bool            m_disable_prefetch;

#ifndef SGL_NO_VERTEX_ARRAYS
   vector<sglIndexType>  m_index_min;
   vector<sglIndexType>  m_index_max;
#endif

private:
   static DrawFunc s_func_table[(ePREFETCH|eMULTI_TEXTURE|eCOLORS|eNORMALS)+1];
   static Initializer sInitializer;
};

#endif
