/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglMonoIndexedGeoSet.hpp
 *  Created: 28 August 1999
 *  Summary: GeoSet where vertex data is indexed via a single index list.
 *****************************************************************************/

#ifndef __SGL_MONO_INDEXED_GEO_STRIP_SET_HPP
#define __SGL_MONO_INDEXED_GEO_STRIP_SET_HPP

#include <sgl.h>
#include <sglMonoIndexedGeoSet.hpp>

//============================================================================

/**
This abstract class extends the sglMonoIndexedGeoSet class to include the
lengths array needed to completely specify the variable length primitive types.

{\bf Concrete subclasses: sglMonoIndexedLineStripSet,
sglMonoIndexedTriangleFanSet, sglMonoIndexedTriangleStripSet,
sglMonoIndexedQuadStripSet, sglMonoIndexedPolygonSet}

These are the concrete subclasses of sglMonoIndexedGeoStripSet that implement
the line strip, triangle fan, triangle strip, quad strip, and polygon
primitives.

{\bf Example:}
\begin{verbatim}
   unsigned int *lengths = new unsigned int[1];
   lengths[0] = 5;

   sglVec3f *vertex = new sglVec3f[5];
   vertex[0].set( 0.0f, 1.0f, 0.0f);
   vertex[1].set( 1.0f, 0.0f, 0.0f);
   vertex[2].set( 1.0f,-1.0f, 0.0f);
   vertex[3].set(-1.0f,-1.0f, 0.0f);
   vertex[4].set(-1.0f, 0.0f, 0.0f);

   sglVec4f *color = new sglVec4f[5];
   color[0].set( 1.0f, 0.0f, 0.0f, 1.0f);
   color[1].set( 0.0f, 1.0f, 0.0f, 1.0f);
   color[2].set( 1.0f, 0.0f, 0.0f, 1.0f);
   color[3].set( 0.0f, 1.0f, 0.0f, 1.0f);
   color[4].set( 1.0f, 0.0f, 0.0f, 1.0f);

   unsigned int *index = new unsigned int[5];
   index[0] = 0;          // indices for a pentagon (irregular)
   index[1] = 1;
   index[2] = 2;
   index[3] = 3;
   index[4] = 4;

   sglMonoIndexedPolygonSet *poly_set = new sglMonoIndexedPolygonSet;
   poly_set->setNumPrims(1);
   poly_set->setPrimLengths(lengths);
   poly_set->setCoordList(vertex);
   poly_set->setColorList(sglGeoSet::ePER_VERTEX, color);  // uses index array
   poly_set->setIndexList(index);
\end{verbatim}
 */

class SGL_DLL_API sglMonoIndexedGeoStripSet : public sglMonoIndexedGeoSet
{
public:
   /// default constructor
   sglMonoIndexedGeoStripSet();
   /// virtual destructor
   virtual ~sglMonoIndexedGeoStripSet();

   /// Not very useful for variable sized primitives.
   virtual unsigned int getPrimSize() const {return 0;}

   /// Get the lengths array.
   sglDataPtr *getPrimLengths() const {return m_lengths.getDataPtr();}

   /// Set the array of lengths.  Deprecated.
   void setPrimLengths(sglIndexType *alist)
      { setPrimLengths(new sglNoRefDataPtr(alist)); }

   /** Set the array of lengths.  This will increment the reference count
       on the given data pointer, and decrement the reference count of
       the previous data pointer.  If the previous data pointer's refernce
       count goes to zero, it will be deleted.
       The lengths must be of type sglIndexType.
       @param data Pointer to the list of lengths.
       @param offset Number of bytes in the array before the first length.
       @param stride Number of bytes from the start of one length to the start
                     of the next length.  If zero, the lengths are assumed
                     to be tightly packed (i.e. stride == sizeof(type)).
                     Stride must be non-negative.
   */
   void setPrimLengths(sglDataPtr *data,
                       int offset = 0,
                       int stride = 0);

   /// Get the number of bytes in the array before the first length.
   int getPrimLengthsOffset() const { return m_lengths.getOffset(); }

   /// Get the number of bytes from the start of one length to the next.
   int getPrimLengthsStride() const { return m_lengths.getStride(); }

   /// Add this drawable's stats to the given sglStats.
   virtual void addStats(sglStats &stats) const;

protected:
   virtual bool computeBounds();
   virtual void updateDrawFunc();

private: // not implemented
   sglMonoIndexedGeoStripSet(const sglMonoIndexedGeoStripSet &);
   sglMonoIndexedGeoStripSet &operator=(const sglMonoIndexedGeoStripSet &);

   void drawVT(const vector<sglTexCoords> &tex_coords) const;
   void drawVNoT(const vector<sglTexCoords> &tex_coords) const;
   void drawVNpT(const vector<sglTexCoords> &tex_coords) const;
   void drawVNvT(const vector<sglTexCoords> &tex_coords) const;
   void drawVCoT(const vector<sglTexCoords> &tex_coords) const;
   void drawVCoNoT(const vector<sglTexCoords> &tex_coords) const;
   void drawVCoNpT(const vector<sglTexCoords> &tex_coords) const;
   void drawVCoNvT(const vector<sglTexCoords> &tex_coords) const;
   void drawVCpT(const vector<sglTexCoords> &tex_coords) const;
   void drawVCpNoT(const vector<sglTexCoords> &tex_coords) const;
   void drawVCpNpT(const vector<sglTexCoords> &tex_coords) const;
   void drawVCpNvT(const vector<sglTexCoords> &tex_coords) const;
   void drawVCvT(const vector<sglTexCoords> &tex_coords) const;
   void drawVCvNoT(const vector<sglTexCoords> &tex_coords) const;
   void drawVCvNpT(const vector<sglTexCoords> &tex_coords) const;
   void drawVCvNvT(const vector<sglTexCoords> &tex_coords) const;

#ifndef SGL_NO_VERTEX_ARRAYS
   void drawVTm(const vector<sglTexCoords> &tex_coords) const;
   void drawVNoTm(const vector<sglTexCoords> &tex_coords) const;
   void drawVNpTm(const vector<sglTexCoords> &tex_coords) const;
   void drawVNvTm(const vector<sglTexCoords> &tex_coords) const;
   void drawVCoTm(const vector<sglTexCoords> &tex_coords) const;
   void drawVCoNoTm(const vector<sglTexCoords> &tex_coords) const;
   void drawVCoNpTm(const vector<sglTexCoords> &tex_coords) const;
   void drawVCoNvTm(const vector<sglTexCoords> &tex_coords) const;
   void drawVCpTm(const vector<sglTexCoords> &tex_coords) const;
   void drawVCpNoTm(const vector<sglTexCoords> &tex_coords) const;
   void drawVCpNpTm(const vector<sglTexCoords> &tex_coords) const;
   void drawVCpNvTm(const vector<sglTexCoords> &tex_coords) const;
   void drawVCvTm(const vector<sglTexCoords> &tex_coords) const;
   void drawVCvNoTm(const vector<sglTexCoords> &tex_coords) const;
   void drawVCvNpTm(const vector<sglTexCoords> &tex_coords) const;
   void drawVCvNvTm(const vector<sglTexCoords> &tex_coords) const;

   void drawPrefetchVT(const vector<sglTexCoords> &tex_coords) const;
   void drawPrefetchVNoT(const vector<sglTexCoords> &tex_coords) const;
   void drawPrefetchVNpT(const vector<sglTexCoords> &tex_coords) const;
   void drawPrefetchVNvT(const vector<sglTexCoords> &tex_coords) const;
   void drawPrefetchVCoT(const vector<sglTexCoords> &tex_coords) const;
   void drawPrefetchVCoNoT(const vector<sglTexCoords> &tex_coords) const;
   void drawPrefetchVCoNpT(const vector<sglTexCoords> &tex_coords) const;
   void drawPrefetchVCoNvT(const vector<sglTexCoords> &tex_coords) const;
   void drawPrefetchVCpT(const vector<sglTexCoords> &tex_coords) const;
   void drawPrefetchVCpNoT(const vector<sglTexCoords> &tex_coords) const;
   void drawPrefetchVCpNpT(const vector<sglTexCoords> &tex_coords) const;
   void drawPrefetchVCpNvT(const vector<sglTexCoords> &tex_coords) const;
   void drawPrefetchVCvT(const vector<sglTexCoords> &tex_coords) const;
   void drawPrefetchVCvNoT(const vector<sglTexCoords> &tex_coords) const;
   void drawPrefetchVCvNpT(const vector<sglTexCoords> &tex_coords) const;
   void drawPrefetchVCvNvT(const vector<sglTexCoords> &tex_coords) const;

   void drawPrefetchVTm(const vector<sglTexCoords> &tex_coords) const;
   void drawPrefetchVNoTm(const vector<sglTexCoords> &tex_coords) const;
   void drawPrefetchVNpTm(const vector<sglTexCoords> &tex_coords) const;
   void drawPrefetchVNvTm(const vector<sglTexCoords> &tex_coords) const;
   void drawPrefetchVCoTm(const vector<sglTexCoords> &tex_coords) const;
   void drawPrefetchVCoNoTm(const vector<sglTexCoords> &tex_coords) const;
   void drawPrefetchVCoNpTm(const vector<sglTexCoords> &tex_coords) const;
   void drawPrefetchVCoNvTm(const vector<sglTexCoords> &tex_coords) const;
   void drawPrefetchVCpTm(const vector<sglTexCoords> &tex_coords) const;
   void drawPrefetchVCpNoTm(const vector<sglTexCoords> &tex_coords) const;
   void drawPrefetchVCpNpTm(const vector<sglTexCoords> &tex_coords) const;
   void drawPrefetchVCpNvTm(const vector<sglTexCoords> &tex_coords) const;
   void drawPrefetchVCvTm(const vector<sglTexCoords> &tex_coords) const;
   void drawPrefetchVCvNoTm(const vector<sglTexCoords> &tex_coords) const;
   void drawPrefetchVCvNpTm(const vector<sglTexCoords> &tex_coords) const;
   void drawPrefetchVCvNvTm(const vector<sglTexCoords> &tex_coords) const;
#endif

   class Initializer;
   friend class Initializer;
   static void setupFuncTable();

protected:
   sglIndexPtr m_lengths; // List of strip lengths

private:
   static DrawFunc s_func_table[(ePREFETCH|eMULTI_TEXTURE|eCOLORS|eNORMALS)+1];
   static Initializer sInitializer;
};

#endif
