/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglNode.cpp
 *  Created: 30 December 1997
 *  Summary: Base class for all scene graph node types
 *****************************************************************************/

#include <sgl.h>
#include <sglNode.hpp>
#include <sglGroup.hpp>
#include <sglCull.hpp>       // is this necessary?
#include <sglGeode.hpp>      // is this necessary?
#include <sglOStream.hpp>

#include <algorithm>

//----------------------------------------------------------------------------
sglNode::sglNode() : sglObject(),
   m_sphere(),
   m_bound_mode(sglBound::eDYNAMIC),
   m_dirty(eBOUND)
{
}

//----------------------------------------------------------------------------
sglNode::~sglNode()
{
   // pop any parents off the list
   while (m_parent_list.size())
   {
      sglGroup *parent = m_parent_list.back();
      parent->removeChild(this); // this takes it out of the list (I hope)
   }
}

//----------------------------------------------------------------------------
sglGroup *sglNode::getParent(unsigned int i) const
{
   if (i < getNumParents())
   {
      return m_parent_list[i];
   }
   return NULL;
}

//----------------------------------------------------------------------------
void sglNode::addParent(sglGroup *parent)
{
   if (parent)
   {
      vector<sglGroup*>::iterator elt = find(m_parent_list.begin(),
                                             m_parent_list.end(),
                                             parent);
      if (elt == m_parent_list.end())
      {
         m_parent_list.push_back(parent);
         parent->dirty(m_dirty|eBOUND);
      }
      else
      {
         sglPrintWarning() << "sglNode::addParent error: "
                           << "trying to add a duplicate parent.\n";
      }
   }
   else
   {
      sglPrintWarning() << "sglNode::addParent error: "
                        << "trying to add a NULL parent.\n";
   }
}

//----------------------------------------------------------------------------
void sglNode::removeParent(sglGroup *parent)
{
   if (parent)
   {
      vector<sglGroup*>::iterator elt = find(m_parent_list.begin(),
                                             m_parent_list.end(),
                                             parent);
      if (elt != m_parent_list.end())
      {
         elt = m_parent_list.erase(elt);
         parent->dirty(eBOUND);
      }
      else
      {
         sglPrintWarning() << "sglNode::removeParent error: "
                           << "could not find referenced parent.\n";
      }
   }
   else
   {
      sglPrintWarning() << "sglNode::removeParent error: "
                        << "trying to remove a NULL parent.\n";
   }
}

//----------------------------------------------------------------------------
void sglNode::setBound(const sglSphereBound<float> *sphere,
                       sglBound::ModeEnum mode)
{
   // compute up the tree even if mode was just set to static.
   // Warning, this will compute upwards whether computeBounds is
   // successful or not.  This may or may not be the desired behaviour.
   dirty(eBOUND);

   // if sphere is NULL compute the bounds for this node
   if (sphere == NULL)
   {
      if (mode == sglBound::eSTATIC)
      {
         m_bound_mode = sglBound::eDYNAMIC;  // to get recursive computation
         computeBoundsRecursiveDown(); // compute bound immediately for node
         m_dirty &= ~eBOUND;
      }
   }
   else
   {
      m_sphere = *sphere;
      m_dirty &= ~eBOUND;
   }

   m_bound_mode = mode;
}

//----------------------------------------------------------------------------
void sglNode::dirty(unsigned char mask)
{
   if ((m_dirty|mask) != m_dirty)
   {
      m_dirty |= mask;
      for (vector<sglGroup*>::iterator ii = m_parent_list.begin();
           ii != m_parent_list.end(); ++ii)
      {
         (*ii)->dirty(mask);
      }
   }
}

//----------------------------------------------------------------------------
void sglNode::preDraw(const PreDrawStruct &)
{
   if (m_bound_mode == sglBound::eSTATIC)
   {
      m_dirty &= ~eBOUND;
   }
   else if (m_dirty&eBOUND)  // implies bound is dynamic
   {
      computeBound();
   }

   m_dirty &= ~(eDLIST|eANIMATION);

   sglPrintDebug() << "sglNode::preDraw()" << endl;
}

//----------------------------------------------------------------------------
const sglSphereBound<float> &sglNode::getBound()
{
   computeBoundsRecursiveDown();
   return m_sphere;
}

//----------------------------------------------------------------------------
void sglNode::computeBoundsRecursiveDown()
{
   if ((m_dirty&eBOUND) && m_bound_mode != sglBound::eSTATIC)
   {
      computeBound();
   }
}

//----------------------------------------------------------------------------
void sglNode::copyTo(sglNode *dst, unsigned int mode) const
{
   sglObject::copyTo(dst, mode);
   if (m_bound_mode == sglBound::eSTATIC)
   {
      dst->setBound(&m_sphere, m_bound_mode);
   }
}

//----------------------------------------------------------------------------
void sglNode::printInfo(ostream &ostrm, const char *indent_string) const
{
   sglObject::printInfo(ostrm, indent_string);

   ostrm << indent_string << "sglNode: " << getNumParents() << " parents"
         << endl;
   ostrm << indent_string << "Bounds computation mode: ";
   switch (m_bound_mode)
   {
      case sglBound::eSTATIC:
         ostrm << "eSTATIC" << endl;
         break;
      case sglBound::eDYNAMIC:
         ostrm << "eDYNAMIC" << endl;
         break;
      default:
         ostrm << "unknown (" << (int)m_bound_mode << ")" << endl;
         break;
   }

   ostrm << indent_string << "Bounding Sphere: " << m_sphere << endl;

   if (getNumParents() > getRefCount())
   {
      ostrm << indent_string << "**** Reference Count Error ("
            << getNumParents() << "," << getRefCount() << ") ****" << endl;
   }

   ostrm << indent_string << "    parent_list capacity = "
         << m_parent_list.capacity() << endl;
   for (unsigned int i=0; i<getNumParents(); i++)
   {
      ostrm << indent_string << "    parent " << i << ": " << m_parent_list[i]
            << ", name = " << m_parent_list[i]->getName() << endl;
   }
}
