/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglNode.hpp
 *  Created: 30 December 1997
 *****************************************************************************/

#ifndef __SGL_NODE_HPP
#define __SGL_NODE_HPP

#include <vector>
#include <deque>

#include <sgl.h>
#include <sglTimespec.hpp>
#include <sglIntersect.hpp>
#include <sglPick.hpp>
#include <sglSphereBound.hpp>
#include <sglObject.hpp>
#include <sglFrustum.hpp>
#include <sglMatrix.hpp>

// foward reference this stuff
template <class T> class sglCull;

class sglGroup;
class sglGeode;
class sglDrawable;
class sglGeoSet;
class sglStatelet;

//----------------------------------------------------------------------------

/**

The abstract base class for all scene graph node types including
sglGroup, sglGeode, sglViewPlatform, and sglLight nodes.
It provides management functionality needed to make the traversal functions
more efficient as described in the following sections.

<b>Parent Lists</b>

Whenever a node is added as a child to a group node, not only does the
child get placed in a list of children in the group, but also the parent
node gets placed in a list of parents in the child node.  This latter
step <i>does not</i> increase the reference count of the parent.
sglNodes provide the functionality to manage the parent list of the nodes.
For reasons of efficiency, a group will not appear more than once in a
child's parent list even if that child is added more than once to the same
sglGroup.

<b>Bounding Spheres</b>

The cull traversal compares the spatial bounds of each node with the view
frustum specified by the user to determine which nodes are completely out
of the frustum, and which are either partially of wholly inside the view
frustum.  sglNodes contain the definition and management functions
of bounding spheres.  Under default conditions, an sglNode's bounding
sphere will be automatically computed to be large enough to contain the
bounding spheres of all of its children.

For reasons of efficiency, these spheres <i>do not<i> get updated immediately
upon addition of a new child.  Rather, a bounds dirty flag is propagated
up the tree from the point of insertion to the root of the tree (along
multiple paths in the case where nodes have more then one parent).
Before traversing the tree, the user must call preDraw to compute all the
new bounding spheres for those nodes with the bounds dirty flag set.

The user can override the default behaviour (autocomputation of the
bounding sphere) by calling the setBound function with sglBound::eSTATIC
token for the mode.  This is useful for nodes whose child list may
change often, but the user knows the smallest sphere that will contain
all of the possibilities.  Setting the static bounds prevents the
upward bounds computation from proceding up the tree past this point (from
the parents perspective, the bounding sphere of a child with static bound
doesn't change).

<b>Traversal Functions</b>

All the traversal functions make their first appearance in the class hierarchy
in the sglNode class.  Currently single and double precision versions
of the segment intersection (intersect), frustum intersection (pick), and
view frustum culling (cull) traversals, and the preDraw functions are currently
defined.  See the documentation on Traversals for more information.

<b>Example:</b>

<pre>
   sglGroup *node = new sglGroup;
   unsigned int num_parents = node->getNumParents();  // returns 0

   sglSphereBoundf sph;                       // single precision
   sph.setCenter(1.0f, 1.3f, -3.2f);
   sph.setRadius(13.3f);
   node->setBound(&sph, sglBound::eSTATIC);   // set a static bound

   node->setBound(NULL, sglBound::eSTATIC);   // set static bound but compute
                                              // bounds based on children's
                                              // bounds (see sglGroup).

   node->setBound(NULL, sglBound::eDYNAMIC);  // set dynamic bounds computation
</pre>

*/

//----------------------------------------------------------------------------

class SGL_DLL_API sglNode : public sglObject
{
public:
   /// default constructor
   sglNode();
   /// destructor
   virtual ~sglNode();

   /** Get the number of nodes (parents) that reference this one.
       @return number of parents
   */
   unsigned int getNumParents() const { return m_parent_list.size(); }

   /** Get a pointer to the i-th parent.
       @param index index of parent node.  If index is out of range then NULL
       is returned.
       @return a pointer to the parent node
   */
   sglGroup* getParent(unsigned int index) const;

   /** Set the bounding sphere and its update behaviour.
       @param sphere A pointer to a bounding sphere.  If the pointer is NULL,
           the bounding sphere is computed (immediately if mode is
           sglBound::eSTATIC) based on the scene graph below this node;
           otherwise, it is copied from the one passed in.
       @param mode If equal to sglBound::eDYNAMIC, then the bounding sphere
           will be automatically computed by preDraw() or the traversers, and
           updated when the scene graph below this node changes.  If it is set
           to sglBound::eSTATIC, the the bounding sphere will not be recomputed
           even if the underlying graph changes.
   */
   void setBound(const sglSphereBound<float> *sphere, sglBound::ModeEnum mode);

   /** Get the current bounding sphere.  If the bound mode is set to
       sglBound::eDYNAMIC and the bounding sphere is dirty, it will be computed
       immediately before returning the refernce
       @return reference to the current bounding sphere.
   */
   const sglSphereBound<float> &getBound();

   /** Get the current bound recomputation mode.
       @return the current bound mode: either sglBound::eSTATIC or
           sglBound::eDYNAMIC.
   */
   sglBound::ModeEnum getBoundMode() const {return m_bound_mode;}

   /// Returns the bounding volume as it was last computed.
   const sglSphereBound<float> &getPrevBound() const { return m_sphere; }

   /// preDraw traversal state struct
   struct PreDrawStruct
   {
      unsigned int m_frame_count;
      sglTimespec  m_frame_time;
   };

   /** This is the pre-draw traversal function which should be called before
       intersecting or culling the scene graph.  This "cleans" the tree by
       computing the bound boxes and spheres that have been marked as dirty,
       building display lists for geometry that has changed, binding
       textures that need to be, updating the animation nodes whose state has
       changed.  This should be called with a valid graphics context current.
       @param trav_state A struct containing state information needed by the
         preDraw traversal (currently frame time and count for animation
         nodes).
   */
   virtual void preDraw(const PreDrawStruct &trav_state);

   /** Compute a single precision transform to parent and append (via
       multiplication) the matrix passed in.  This is part of an upward
       traversal, which most commonly begins with an sglViewPlatform and used
       for computing the position and orientation of said platform.  This is
       the default implementation for all nodes that do not affect this matrix
       and therefore this function does nothing.
       @param curr_mat The current view matrix for nodes below this one.
   */
   virtual void computeTxToParent(sglMat4f &) const {}

   /** Compute a double precision transform to parent and append (via
       multiplication) the matrix passed in.  This is part of an upward
       traversal, which most commonly begins with an sglViewPlatform and used
       for computing the position and orientation of said platform.  This is
       the default implementation for all nodes that do not affect this matrix
       and therefore this function does nothing.
       @param curr_mat The current view matrix for nodes below this one.
   */
   virtual void computeTxToParent(sglMat4d &) const {}

   /** The single precision intersection traversal function which returns the
       closest object (bounding volume and/or triangle) that intersects with
       the given intersect segment.
       @param isector A reference to a single precision intersection state
   */
   virtual sglBound::IntersectResultEnum intersect(sglIntersectf &) const
      {
         return sglBound::eOUTSIDE;
      }

   /** The double precision intersection traversal function which returns the
       closest object (bounding volume and/or triangle) that intersects with
       the given intersect segment.
       @param isector A reference to a double precision intersection state
   */
   virtual sglBound::IntersectResultEnum intersect(sglIntersectd &) const
      {
         return sglBound::eOUTSIDE;
      }

   /** The single precision pick traversal function which returns all objects
       that fall within the pick frustum.
       @param pick_state A reference to a single precision pick state.
       @param cull_flags Bit flags that indicate which planes of the pick
           frustum (polytope) still need to be tested to determine if bounding
           spheres and boxes lie within the frustum.  Cull-free picking can be
           accomplished by setting this to 0.
   */
   virtual void pick(sglPickf &, unsigned int) const {}

   /** The double precision pick traversal function which returns all objects
       that fall within the pick frustum.
       @param pick_state A reference to a double precision pick state.
       @param cull_flags Bit flags that indicate which planes of the pick
           frustum (polytope) still need to be tested to determine if bounding
           spheres and boxes lie within the frustum.  Cull-free picking can be
           accomplished by setting this to 0.
   */
   virtual void pick(sglPickd &, unsigned int) const {}

   /** The single precision cull traversal function that culls out subgraphs
       that do not lie in the view frustum (stored in the sglCull parameter).
       Subclasses must implement this function.  The entry point for
       user-friendly culling is in the sglScene class.
       @param trav_state The single precision traversal state that collects all
           the state and geometry information that passes the cull.
       @param cull_flags Bit flags that indicate which planes of the view
           frustum (polytope) still need to be tested to determine if bounding
           spheres and boxes lie within the frustum.  Cull-free drawing can be
           accomplished with cull_flags = 0.
   */
   virtual void cull(sglCull<float> &trav_state,
                     unsigned int cull_flags) const = 0;

   /** The double precision cull traversal function that culls out subgraphs
       that do not lie in the view frustum (stored in the sglCull parameter).
       Subclasses must implement this function.  The entry point for
       user-friendly culling is in the sglScene class.
       @param trav_state The single precision traversal state that collects all
           the state and geometry information that passes the cull.
       @param cull_flags Bit flags that indicate which planes of the view
           frustum (polytope) still need to be tested to determine if bounding
           spheres and boxes lie within the frustum.  Cull-free drawing can be
           accomplished with cull_flags = 0.
    */
   virtual void cull(sglCull<double> &trav_state,
                     unsigned int cull_flags) const = 0;

   /** Make a copy of the scenegraph rooted at this node.
       @param mode Bit masks to control the behaviour of the clone.  These are
         OR-ed together from the mode values in sglObject::CloneModeEnum.
       @return Pointer to root of cloned scene graph.
   */
   virtual sglNode *clone(unsigned int mode) const = 0;

   // See sglObject::printInfo() for details.
   virtual void printInfo(ostream &ostrm, const char *indent_string) const;

protected:
   virtual void computeBound() = 0; // this node only - no traversal
   virtual void computeBoundsRecursiveDown(); // downward trav

   // bit or the dirty bits with the given mask
   virtual void dirty(unsigned char mask);

   // used by the clone traversal
   void copyTo(sglNode *dst, unsigned int mode) const;

   friend class sglDrawable;

private:
   friend class sglGroup;
   void addParent(sglGroup *parent);
   void removeParent(sglGroup *parent);

   // not implemented
   sglNode(const sglNode &);
   sglNode &operator=(const sglNode &);

protected:
   sglSphereBound<float> m_sphere;
   sglBound::ModeEnum m_bound_mode;

   enum DirtyEnum { eBOUND = 0x01, eDLIST = 0x02, eANIMATION = 0x04 };
   unsigned char m_dirty;

private:
   vector<sglGroup*> m_parent_list;
};

#endif
