/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglOStream.cpp
 *   Author: Tom Stimson
 *  Created: 8 April 2000
 *****************************************************************************/

#ifdef WIN32
# include <windows.h>
#endif

#include <sglOStream.hpp>


//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
// This class provides a bit bucket to be used with ostreams.
// Nothing put into an sglNullStreamBuf ever comes back out.
class sglNullStreamBuf : public streambuf
{
public:
   sglNullStreamBuf();
   ~sglNullStreamBuf();

protected:
   int overflow(int c);
   int sync();

private:
   // Not implemented.  Do not use.
   sglNullStreamBuf(const sglNullStreamBuf &);
   sglNullStreamBuf &operator=(const sglNullStreamBuf &);

private:
   char    m_buffer[256];
};


//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
sglOStream::PrintLevelEnum sglOStream::s_threshold = sglOStream::eWARNING;

// NOTE s_stream is NOT initialized to cout/cerr to avoid order of 
//      static initialization problems.  Instead, s_stream will
//      be initialized by sgl::initialize().
ostream*               sglOStream::s_stream    = NULL;

sglOStream::callback   sglOStream::s_pre_func  = NULL;
sglOStream::callback   sglOStream::s_post_func = NULL;

sglNullStreamBuf      *sglOStream::s_null_buff = new sglNullStreamBuf();
ostream                sglOStream::s_null_ostream(sglOStream::s_null_buff);


//----------------------------------------------------------------------------
void sglOStream::setStream(ostream* stream)
{
   s_stream = stream;
}


//----------------------------------------------------------------------------
ostream* sglOStream::getStream()
{
   return s_stream;
}


//----------------------------------------------------------------------------
void sglOStream::setThreshold(PrintLevelEnum threshold)
{
   s_threshold = threshold;
}


//----------------------------------------------------------------------------
sglOStream::PrintLevelEnum sglOStream::getThreshold()
{
   return s_threshold;
}


//----------------------------------------------------------------------------
void sglOStream::setPreFunc(callback func)
{
   s_pre_func = func;
}

//----------------------------------------------------------------------------
sglOStream::callback sglOStream::getPreFunc()
{
   return s_post_func;
}

//----------------------------------------------------------------------------
void sglOStream::setPostFunc(callback func)
{
   s_post_func = func;
}

//----------------------------------------------------------------------------
sglOStream::callback sglOStream::getPostFunc()
{
   return s_post_func;
}

#ifdef WIN32
//----------------------------------------------------------------------------
string sglOStream::formatWIN32Error(int error_number)
{
   LPVOID lpMsgBuf;
   DWORD len = FormatMessage(
      FORMAT_MESSAGE_ALLOCATE_BUFFER|FORMAT_MESSAGE_FROM_SYSTEM,
      NULL, error_number, MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
      (LPTSTR)&lpMsgBuf, 0, NULL);
   if (len)
   {
      string description((char *)lpMsgBuf, len-2);
      LocalFree( lpMsgBuf );

      return description;
   }
   else
   {
      return "Unkown error";
   }
}
#endif



//----------------------------------------------------------------------------
sglNullStreamBuf::sglNullStreamBuf()
{
   setp(m_buffer, m_buffer + 256);
}

//----------------------------------------------------------------------------
sglNullStreamBuf::~sglNullStreamBuf()
{
}
  
//----------------------------------------------------------------------------
int sglNullStreamBuf::overflow(int c)
{
   if (c != EOF)
   {
      if (pptr() == epptr())
      {
         setp(m_buffer, m_buffer + 256);
      }
      return sputc(c);
   }
   return 0;
}

//----------------------------------------------------------------------------
int sglNullStreamBuf::sync()
{
   *(pptr()) = 0;

   setp(m_buffer, m_buffer + 256);
   return 0;
}
