/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglObject.hpp
 *  Created: 30 December 1997
 *****************************************************************************/

#ifndef __SGL_OBJECT_HPP
#define __SGL_OBJECT_HPP

#include <sgl.h>
#include <sglUserData.hpp>
#include <string>

//----------------------------------------------------------------------------

/**
This is the abstract base class for (nearly?) all objects that appear in the
scene graph including the group and leaf nodes, the geometry nodes, and
statelets.  Through this class the user can assign a name (character string),
associate application-specific user data and query/modify the reference count
of these objects.

<b>Reference Counting</b>

All objects in the scene graph are (or should be) reference counted.
Whenever a node is added as a child to another (see sglGroup), the node's
reference count is incremented.  Likewise, the reference count is
decremented when the node is removed as a child to another
node.  If the reference count goes to zero it is safe to destruct
the object as it is no longer referenced by anything in the scene graph.
The scene graph does not automatically destruct a node when its reference
count goes to zero (particularly when calling sglGroup::removeChild() or
sglGeode::removeGeometry()).  The exception to this rule is the destruction
of statelets if their reference count goes to zero during a call to
sglDrawable::removeState().

Another circumstance in which SGL will automatically destruct an
sglObject is when an sglObject references other sglObjects (as in the cases of
child sglNodes of an sglGroup, child sglDrawable's of an sglGeode, or
sglStatelets of an sglDrawable) and is destructed, the reference counts of
those objects it references is decremented.  If at anytime those reference
counts goes to zero, then those objects are destructed as well.  The
process continues recursively to all objects in the scene graph below the
object destructed by the user.

The user may also explicitly increment and decrement the reference count with
ref() and deref() member functions.  For example, the ref() function can be
called when the object is an integral part (member) of another class or even
allocated off the stack to prevent improper destruction of the object.
Caution: if deref() is attempted for an sglNode that has a reference count of
zero, causes an assertion failure.

NOTE: not all data in the scene graph is currently reference counted and
special care must be taken by the user to manage it.  This is most important
for the attribute data referenced by sglDrawable nodes (vertices, normals,
color vertices, and index arrays and their states (texture coordinates).

<b>Application-specific data</b>

An application that uses SGL can associate application specific data with any
of the classes derived from sglObject.  The only caveat is that the
application's user data must be encapsulated in a class derived from
sglUserData.

<b>The printInfo function</b>

All subclasses of sglObject should also implement the printInfo function
that first calls its base class' version of this function and then outputs
an informational data about the node's current state.  This function
is used in a couple of sglu functions for outputting all or portions of
a scene graph (good for debugging).

<b>Example:</b>

<pre>
   sglGroup *node = new sglGroup; // a subclass of sglObject (via sglNode)
   node->setName("Foo");          // sets name
   char *name = node->getName();  // returns pointer to name
   node->printInfo(cout, "  ");   // output node info to cout w/ 2 space indent

   node->ref();                   // increment the reference count
   unsigned int count = node->getRefCount();  // returns 1
   node->deref();                 // decrement the reference count
   node->deref();                 // ERROR: causes assertion failure
</pre>

<i>
Proposal: subclass from an sglMemory class to manage/track memory allocation?
</i>

*/

//----------------------------------------------------------------------------

class SGL_DLL_API sglObject /* : public sglMemory */
{
public:
   /** Mode bits for the clone traversal function in sub classes
    */
   enum CloneModeEnum
   {
      /// clone just this node
      eCLONE_ONE_NODE        = 0,

      /// clone all sglNodes (not drawables)
      eCLONE_NODES_RECURSIVE = 0x01,

      /// clone the user data (references only)
      eCLONE_USER_DATA       = 0x02

//    eCLONE_CAMERAS = 0x04,
//    eCLONE_GEOMETRY = 0x08,        // clone drawables, too (not vertex data)
//    eCLONE_GEOMETRY_DATA = 0x10,   // clones drawables per vertex data too.
//    eCLONE_STATE = 0x20,           // clone the sglStatelets
//    eCLONE_TEXTURE_IMAGE = 0x40    // copy the texture image data
   };

public:
   /// default constructor
   sglObject();
   /// destructor: pure virtual to ensure abstractness
   virtual ~sglObject() = 0;

   /// Increment the reference count.
   void ref()   { m_ref_count++; }

   /** Decrement the reference count.  An assertion failure will occur if this
       count goes negative.
   */
   void deref() { assert(m_ref_count != 0); --m_ref_count; }

   /** Get the current reference count.
       @return current reference count
   */
   unsigned int getRefCount() const { return m_ref_count; }

   /** Set the name of this object (the string is copied into this object).
       @param name reference to a string containing the name.
   */
   void setName(const string &name) { m_name = name; }

   /** Get the name of this object
       @return const reference to the string
   */
   const string &getName() const { return m_name; }

   /** Set the user data.  This data will NOT be destructed with this
       sglObject.
       @param data pointer to the user data object.  Pass NULL to clear the
       data.
   */
   void setUserData(sglUserData *data) { m_user_data = data; }

   /** Get the user data pointer
       @return pointer to the user data object
   */
   sglUserData* getUserData() const    { return m_user_data; }

   /** Output the state of this node to the specified ostream.
       @param ostrm the ostream to which the output is sent
       @param indent_string the string (usually spaces) that is output at the
                            beginning of every line of output
   */
   virtual void printInfo(ostream &ostrm, const char *indent_string) const;

protected:
   // used by the clone traversal
   void copyTo(sglObject *dst, unsigned int mode) const;

private: // not implemented
   sglObject(const sglObject &);
   sglObject &operator=(const sglObject &);

protected:
   unsigned int m_ref_count;
   sglUserData *m_user_data;
   string m_name;
};

//----------------------------------------------------------------------------

inline ostream& operator<<(ostream& ostrm, const sglObject& object)
{
   object.printInfo(ostrm, " ");
   return ostrm;
}

#endif
