/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglOrthographicCamera.hpp
 *   Author: Scott McMillan
 *  Created: 19 December 2000
 *  Summary: Orthographic camera class
 *****************************************************************************/

#ifndef __SGL_ORTHOGRAPHIC_CAMERA_HPP
#define __SGL_ORTHOGRAPHIC_CAMERA_HPP

#include <sglCamera.hpp>

/**

This subclass of sglCamera is used to specify an orthographic projection
parameters and view frustum.  The parameters of the
sglOrthographicCamera::setOrtho(..) member function are the same as those for
OpenGL's glOrtho command.  This class is usually assigned to an sglViewPlatform
node to place and orient the camera in the 3D scene.

<b>Example:</b>

<pre>
   sglOrthographicCamera *view_camera = new sglOrthographicCamera;
   view_camera->setOrtho(-10.0, 10.0, -8.0, 8.0, 1.0, 200.0);

   // OR (sets near/far to -1.0, 1.0)
   view_camera->setOrtho(-10.0, 10.0, -8.0, 8.0);
</pre>
 */

class SGL_DLL_API sglOrthographicCamera : public sglCamera
{
public:
   /// default constructor
   sglOrthographicCamera();

   /// virtual destructor
   virtual ~sglOrthographicCamera();

   /** Set the orthographic projection parameters.  These are the same as the
       arguments to the glOrtho command.  The lower left and upper right
       positions on the near clip plane are defined as (left, bottom, -near)
       and (right, bottom, -near).  The eyepoint position is at the origin.
       Default values of -1 and 1 for near_clip and far_clip allow this
       function to behave like gluOrtho2D if the user omits these two
       parameters.
       @param left Left edge of the frustum.  The default is -10.
       @param right Right edge of the frustum.  The default is 10.
       @param bottom Bottom edge of the frustum.  The default is -10.
       @param top Top edge of the frustum.  The default is 10.
       @param near_clip The near clip plane distance.  Defaults to 0.1.
       @param far_clip The far clip plane distance.  Defaults to 1000.0.
   */
   void setOrtho(double  left, double  right, double  bottom, double  top,
                 double near_clip = -1.0, double far_clip = 1.0);

   /** Query the current orthographic parameters.  The user passes in
       references to doubles to get the values.
   */
   void getOrtho(double &left, double &right, double &bottom, double &top,
                 double &near_clip, double &far_clip) const;

   // See sglCamera::applyProjection() for details.
   virtual void applyProjection() const;

   // See sglObject::printInfo() for details.
   virtual void printInfo(ostream &ostrm, const char *indent_string) const;

protected:
   sglOrthographicCamera(const sglOrthographicCamera &);
   sglOrthographicCamera &operator=(const sglOrthographicCamera &);

   void buildFrustum();

protected:
   double m_left;
   double m_right;
   double m_bottom;
   double m_top;
};

#endif
