/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglOverrideStateNode.hpp
 *   Author: Tom Stimson
 *  Created: 21 July 2000
 *  Summary: override the state of all nodes below this one
 *****************************************************************************/

#ifndef __SGL_OVERRIDESTATENODE_HPP
#define __SGL_OVERRIDESTATENODE_HPP

#include <sgl.h>
#include <sglGroup.hpp>

/**

An sglOverrideStateNode is a container of sglStatelets that override the state
specified in node below it in the scene graph.  These statelets take precedence
over those in sglOverrideStateNodes lower in the scene graph, but not those
higher in the graph.  These statelets also have higher priority than those in
sglDefaultStateNodes and sglDrawable nodes.

<b>Example:</b>

<pre>
   vector<sglStatelet*> override_statelets;
   override_statelets.push_back(new sglLighting(false));

   // sets the override lighting mode to OFF for all nodes below this one.
   sglOverrideStateNode *override_node = new sglOverrideStateNode;
   override_node->setStatelets(override_statelets);

   override_node->addChild(...);
</pre>

 */

class SGL_DLL_API sglOverrideStateNode : public sglGroup
{
public:
   /// default constructor
   sglOverrideStateNode();
   /// virtual destructor
   virtual ~sglOverrideStateNode();

   /** Set the current set of default statelets from the vector provided.  The
       new statelets' reference counts are incremented, while the previous
       statelets are decremented, and if their reference counts go to zero they
       are deleted.
       @param statelets Vector of new statelets (that are dynamically
         allocated).
   */
   void setStatelets(const vector<sglStatelet*> &statelets);

   /** Get a reference to the current set of override statelets.
       @return A reference to the internal list of statelets.
   */
   const vector<sglStatelet*> &getStatelets() const;

   // See sglNode::clone() for details.
   virtual sglNode *clone(unsigned int mode) const;

   // See sglObject::printInfo() for details.
   virtual void printInfo(ostream &ostrm, const char *indent_string) const;

protected:
   // See sglNode::cull(f) for details
   virtual void cull(sglCull<float> &trav_state,
                     unsigned int cull_flags) const;

   // See sglNode::cull(d) for details
   virtual void cull(sglCull<double> &trav_state,
                     unsigned int cull_flags) const;

   void copyTo(sglOverrideStateNode *dst, unsigned int mode) const;

private:   // not implemented
   sglOverrideStateNode(const sglOverrideStateNode &);
   sglOverrideStateNode &operator=(const sglOverrideStateNode &);


protected:
   vector<sglStatelet*>  m_statelets;
};

#endif
