/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglPerspectiveCamera.cpp
 *   Author: Scott McMillan
 *  Created: 19 December 2000
 *****************************************************************************/

#include <sglPerspectiveCamera.hpp>
#include <sglOStream.hpp>

//----------------------------------------------------------------------------
sglPerspectiveCamera::sglPerspectiveCamera() :
      sglCamera(),
      m_simple_flag(false),
      m_left(0.0),
      m_right(0.0),
      m_bottom(0.0),
      m_top(0.0),
      m_fovx(M_PI/3.0)
{
   double fovy     = M_PI*0.25;
   double aspect   = 640./480.;

   m_top    = m_near * tan( fovy * 0.5 );
   m_bottom =-m_top;
   m_left   = m_bottom * aspect;
   m_right  = m_top * aspect;

   if (!m_frustum.buildPerspective(m_left, m_right, m_bottom, m_top,
                                   m_near, m_far))
   {
      sglPrintWarning() << "sglPerspectiveCamera::sglPerspectiveCamera error: "
                        << "sglFrustum::buildPerspective failed" << endl;
   }
}

//----------------------------------------------------------------------------
sglPerspectiveCamera::~sglPerspectiveCamera()
{
}

//----------------------------------------------------------------------------
bool sglPerspectiveCamera::setFOV(double fovy, double aspect,
                                  double near_clip, double far_clip)
{
   if (fovy > 0.0 && fovy <= M_PI && aspect > 0.0 &&
       near_clip > 0.0 && far_clip > 0.0)
   {
      m_simple_flag = false;

      if (near_clip < far_clip)
      {
         m_near = near_clip;
         m_far  = far_clip;
      }
      else
      {
         m_near = far_clip;
         m_far  = near_clip;
      }

      m_top    = m_near * tan( fovy * 0.5 );
      m_bottom =-m_top;
      m_left   = m_bottom * aspect;
      m_right  = m_top * aspect;

      // m_fovx = atan2(m_right, m_near) - atan2(m_left, m_near);
      m_fovx = fovy*aspect;

      if (!m_frustum.buildPerspective(m_left, m_right, m_bottom, m_top,
                                      m_near, m_far))
      {
          sglPrintWarning() << "sglPerspectiveCamera error: "
                            << "sglFrustum::buildPerspective failed" << endl;
          return false;
      }
      return true;
   }

   return false;
}

//----------------------------------------------------------------------------
bool sglPerspectiveCamera::getFOV(double &fovy, double &aspect,
                                  double &near_clip, double &far_clip) const
{
   fovy = m_frustum.getFOVY();
   aspect = m_frustum.getAspect();
   near_clip = m_near;
   far_clip = m_far;

   return !m_simple_flag;
}

//----------------------------------------------------------------------------
void sglPerspectiveCamera::setSimple(double left, double right,
                                     double bottom, double top,
                                     double near_clip, double far_clip)
{
   m_simple_flag = true;

   m_left = left;
   m_right = right;
   m_bottom = bottom;
   m_top = top;
   m_near = near_clip;
   m_far = far_clip;

   m_fovx = atan2(m_right, m_near) - atan2(m_left, m_near);

   if (!m_frustum.buildPerspective(m_left, m_right, m_bottom, m_top,
                                   m_near, m_far))
   {
      sglPrintWarning() << "sglPerspectiveCamera error: "
                        << "sglFrustum::buildPerspective failed" << endl;
   }
}

//----------------------------------------------------------------------------
void sglPerspectiveCamera::getSimple(double &left, double &right,
                                     double &bottom, double &top,
                                     double &near_clip, double &far_clip) const
{
   left = m_left;
   right = m_right;
   bottom = m_bottom;
   top = m_top;
   near_clip = m_near;
   far_clip = m_far;
}

//----------------------------------------------------------------------------
void sglPerspectiveCamera::applyProjection() const
{
   glMatrixMode(GL_PROJECTION);
   glLoadIdentity();
   glFrustum(m_left, m_right, m_bottom, m_top, m_near, m_far);
}

//----------------------------------------------------------------------------
void sglPerspectiveCamera::printInfo(ostream &ostrm,
                                     const char *indent_string) const
{
   sglCamera::printInfo(ostrm, indent_string);
   if (m_simple_flag)
   {
      ostrm << indent_string << "right/left = " << m_right << "/" << m_left
            << endl;
      ostrm << indent_string << "bottom/top = " << m_bottom << "/" << m_top
            << endl;
   }
   else
   {
      ostrm << indent_string << "fovy = " << m_frustum.getFOVY() << endl;
      ostrm << indent_string << "aspect = " << m_frustum.getAspect() << endl;
   }
}
