/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglPerspectiveCamera.hpp
 *   Author: Scott McMillan
 *  Created: 19 December 2000
 *  Summary: Perspective camera class
 *****************************************************************************/

#ifndef __SGL_PERSPECTIVE_CAMERA_HPP
#define __SGL_PERSPECTIVE_CAMERA_HPP

#include <sglCamera.hpp>

/**

This subclass of sglCamera provides the interface for specifying perspective
frustums/projections.  The sglPerspectiveCamera::setFOV(..)  method takes the
same arguments as the gluPerspective function, except that the fovy argument
must be given in radians instead of degrees.  The user may also use this class'
sglPerspectiveCamera::setSimple(..) function to setup the projection just like
that provided by OpenGL's glFrustum command.  This is useful for specifying
off-axis projections that cannot be specifed with FOV and aspect alone.  Note
that when this function has been used to specify the projection, then
sglPerspective::getFOV(..) will return false to indicate that some of the
values were computed (estimated?) internal because setSimple (and not setFOV)
was used to define projection.

<b>Example:</b>

<pre>
   sglPerspectiveCamera *view_camera = new sglPerspectiveCamera;
   view_camera->setFOV(0.25*M_PI, (double)win_x/(double)win_y, 1.0, 200.0);

   // OR
   view_camera->setSimple(0.0, 0.4, 0.0, 0.3, 1.0, 200.0); // off-axis
</pre>

 */

class SGL_DLL_API sglPerspectiveCamera : public sglCamera
{
public:
   /// default constructor
   sglPerspectiveCamera();

   /// virtual destructor
   virtual ~sglPerspectiveCamera();

   /** This function is used to specify a symmetric perspective frusta
       (projections).  The arguments match those of the gluPerspective function
       in OpenGL.  If any of the parameters are out of range then the function
       returns false, and none of the parameters are set.  If near_clip is
       greater than the far_clip then the parameters are swapped.
       @param fovy Vertical field of view in radian in the range (0..PI).  The
         default is PI/4.
       @param aspect The width to height aspect ratio and must be greater
         than zero.  The default is 640/480.
       @param near_clip The near clip distance and must be greater than
         zero.  The default is 0.1.
       @param far_clip The far clip distance and should be greater than the
         near clip distance.  The default is 1000.
       @return false if one or more arguments were out of range.
   */
   bool setFOV(double fovy, double aspect, double near_clip, double far_clip);

   /** This function gets the parameter values, but returns false if setSimple
       was last called to indicate that the fovy parameter returned is computed
       internally based on the parameters to setSimple (and not that specified
       in setFOV) and does not have the implied symmetry.
   */
   bool getFOV(double &fovy, double &aspect,
               double &near_clip, double &far_clip) const;

   /// Returns the vertical field of view in radians.
   double getFOVY() const { return m_frustum.getFOVY(); }
   /// Returns the horizontal field of view in radians.
   double getFOVX() const { return m_fovx; }

   /** The most basic method for setting frusta (parameters corresponds to
       those in glFrustum) which can also be used to set off-axis projections.
       The lower left and upper right locations on the near clip plane is
       defined as (left, bottom, -near) and (right, top, -near) with the camera
       positioned at the origin.
       @param left Left edge of the frustum.
       @param right Right edge of the frustum.
       @param bottom Bottom edge of the frustum.
       @param top Top edge of the frustum.
       @param near_clip The near clip plane distance.
       @param far_clip The far clip plane distance.
   */
   void setSimple(double  left, double  right, double  bottom, double  top,
                  double near_clip, double far_clip);

   /** Query the parameters for the current perspective frustum.
    */
   void getSimple(double &left, double &right, double &bottom, double &top,
                  double &near_clip, double &far_clip) const;

   // See sglCamera::applyProjection() for details.
   virtual void applyProjection() const;

   // See sglObject::printInfo() for details.
   virtual void printInfo(ostream &ostrm, const char *indent_string) const;

protected:
   sglPerspectiveCamera(const sglPerspectiveCamera &);
   sglPerspectiveCamera &operator=(const sglPerspectiveCamera &);

protected:
   bool m_simple_flag;

   double m_left;
   double m_right;
   double m_bottom;
   double m_top;

   double m_fovx;   // radians (computed internally)
};

#endif
