/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglPositionalLight.cpp
 *   Author: Scott McMillan
 *  Created: 28 November 2000
 *****************************************************************************/

#include <sglPositionalLight.hpp>
#include <sglOStream.hpp>

//----------------------------------------------------------------------------
sglPositionalLight::sglPositionalLight() : sglLight(),
   m_position(0.0f, 0.0f, 0.0f, 1.0f),
   m_attenuation_constant(1.0f),
   m_attenuation_linear(0.0f),
   m_attenuation_quadratic(0.0f),
   m_falloff_inverse(5.0)
{
}

//----------------------------------------------------------------------------
sglPositionalLight::~sglPositionalLight()
{
}

//----------------------------------------------------------------------------
void sglPositionalLight::setPosition(const sglVec3f &pos)
{
   m_position.set(pos[0], pos[1], pos[2], 1.0f);
}

//----------------------------------------------------------------------------
void sglPositionalLight::getPosition(sglVec3f &pos) const
{
   pos.set(m_position[0], m_position[1], m_position[2]);
}

//----------------------------------------------------------------------------
void sglPositionalLight::setAttenuation(float constant,
                                        float linear,
                                        float quadratic)
{
   bool change_flag = false;
   if (constant >= 0.0f && constant != m_attenuation_constant)
   {
      m_attenuation_constant = constant;
      change_flag = true;
   }

   if (linear >= 0.0f && linear != m_attenuation_linear)
   {
      m_attenuation_linear = linear;
      change_flag = true;
   }

   if (quadratic >= 0.0f && quadratic != m_attenuation_quadratic)
   {
      m_attenuation_quadratic = quadratic;
      change_flag = true;
   }

   if (change_flag && m_bound_mode == sglBound::eDYNAMIC)
   {
      dirty(eBOUND);
   }
}

//----------------------------------------------------------------------------
void sglPositionalLight::getAttenuation(float &constant,
                                        float &linear,
                                        float &quadratic) const
{
   constant  = m_attenuation_constant;
   linear    = m_attenuation_linear;
   quadratic = m_attenuation_quadratic;
}

//----------------------------------------------------------------------------
void sglPositionalLight::setFalloff(float falloff)
{
   if (falloff > 0.0f && falloff != 1.0f/m_falloff_inverse)
   {
      m_falloff_inverse = 1.0f/falloff;
      dirty(eBOUND);
   }
}

//----------------------------------------------------------------------------
void sglPositionalLight::computeBound()
{
   m_sphere.setCenter(m_position[0], m_position[1], m_position[2]);

   // compute a reasonable sphere based on attentuation parameters
   float radius = -1;

   if (m_attenuation_linear != 0.0f || m_attenuation_quadratic != 0.0f)
   {
      // attenuation equation is atten = 1/(kc + d*kl + d*d*kq)
      // need to solve the following quadratic:
      //    0 = kq*d*d + kl*d + (kc - 1.0/falloff)
      // where atten is the desired percentage of light at the falloff
      // e.g. if you want bounding sphere to falloff at 1% of total light
      // then falloff is set to 0.01

      // check discriminant for a valid solution
      double cterm = m_attenuation_constant - m_falloff_inverse;
      double discrim = m_attenuation_linear*m_attenuation_linear -
         4.0*m_attenuation_quadratic*cterm;

      if (discrim > 0.0)
      {
         double r1 = (-m_attenuation_linear - sqrt(discrim))/
            (2*m_attenuation_quadratic);
         double r2 = (-m_attenuation_linear + sqrt(discrim))/
            (2*m_attenuation_quadratic);

         if (r1 > r2 && r1 > 0.0)
            radius = r1;
         else if (r2 > 0.0)
            radius = r2;
      }
   }

   sglPrintDebug() << "sglPositionalLight radius = " << radius << endl;
   m_sphere.setRadius(radius);
   m_dirty &= ~eBOUND;
}

//----------------------------------------------------------------------------
void sglPositionalLight::setup(GLenum gl_light_number) const
{
   sglLight::setup(gl_light_number);

   glLightfv(gl_light_number, GL_POSITION, m_position.getPtr());

   glLightf(gl_light_number, GL_CONSTANT_ATTENUATION,
            m_attenuation_constant);
   glLightf(gl_light_number, GL_LINEAR_ATTENUATION, m_attenuation_linear);
   glLightf(gl_light_number, GL_QUADRATIC_ATTENUATION,
            m_attenuation_quadratic);
}

//----------------------------------------------------------------------------
sglNode *sglPositionalLight::clone(unsigned int mode) const
{
   sglPositionalLight *root = new sglPositionalLight;
   sglPositionalLight::copyTo(root, mode);
   return root;
}

//----------------------------------------------------------------------------
void sglPositionalLight::copyTo(sglPositionalLight *dst,
                                unsigned int mode) const
{
   sglLight::copyTo(dst, mode);

   // transfer stuff
   sglVec3f pos;
   getPosition(pos);
   dst->setPosition(pos);

   dst->setAttenuation(m_attenuation_constant, m_attenuation_linear,
                       m_attenuation_quadratic);
   dst->setFalloff(1.f/m_falloff_inverse);
}

//----------------------------------------------------------------------------
void sglPositionalLight::printInfo(ostream &ostrm,
                                   const char *indent_string) const
{
   sglLight::printInfo(ostrm, indent_string);
   ostrm << indent_string << "position: " << m_position[0] << ", "
         << m_position[1] << ", " << m_position[2] << endl;
   ostrm << indent_string << "attenuation (c,l,q): "
         << m_attenuation_constant << ", " << m_attenuation_linear
         << ", " << m_attenuation_quadratic << endl;
   ostrm << indent_string << "falloff: " << 1.f/m_falloff_inverse << endl;
}
