/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglPositionalLight.hpp
 *   Author: Scott McMillan
 *  Created: 9 December 2000
 *  Summary: encapsulates a positional (local?) light source
 *****************************************************************************/

#ifndef __SGL_POSITIONAL_LIGHT_HPP
#define __SGL_POSITIONAL_LIGHT_HPP

#include <sgl.h>
#include <sglLight.hpp>

//----------------------------------------------------------------------------

/**

sglPositionalLight is a subclass of sglLight that is given a position in 3D
space which the user can specify with call to
sglPositionalLight::setPosition().  The position specified is relative to this
node's local coordinate system and is affected by matrix transformation nodes
in the scene graph above this node.

Associated with a positional light are a set of attenuation parameters
that cause the light to falloff with distance (controled using the
sglPositionalLight::setAttenuation(..) function).  The size of the bounding
sphere for the light is computed with a radius that equals the
distance to a certain amount of falloff in the light.  The percentage of
falloff to use in this calculation is configurable by the user using the
setFalloff() function.

<b>Example:</b>

<pre>
   sglPositionalLight *plite = new sglPositionalLight;
   plite->setPosition(sglVec3f(0., 0., 8.0f));
   plite->setDiffuse(0.f, 1.f, 0.f, 1.f);           // green light
   plite->setAmbient(0.25f, 0.25f, 0.25f, 1.f);     // gray ambient

   // enable automatic bounding sphere computation to 20% of light
   plite->setFalloff(0.2f);
   plite->setAttenuation(1.0f, 0.0f, 0.01f);
</pre>
 */

class SGL_DLL_API sglPositionalLight : public sglLight
{
public:
   /// default constructor
   sglPositionalLight();
   /// virtual destructor
   virtual ~sglPositionalLight();

   /** Set the light position.  Default is (0, 0, 0)
       @param pos Vector containing the position of the light relative to the
       local coordinate system.
   */
   void setPosition(const sglVec3f &pos);

   /** Get the current light position.
       @param pos Reference to vector to be set to the current position.
   */
   void getPosition(sglVec3f &pos) const;

   /** Set the OpenGL attenuation parameters where the attenuation factor is
       equal to 1/(constant + linear*distance + quadratic*distance*distance).
       Only non-negative values will be accepted.
       @param constant Constant attenuation factor (default is 1).
       @param linear Linear attenuation factor (default is 0).
       @param quadratic Quadratic attenuation factor (default is 0).
   */
   void setAttenuation(float constant, float linear, float quadratic);

   /** Query the current attenuation factors.
       @param constant Reference to variable that will be set with the constant
         attenuation factor.
       @param linear Reference to variable that will be set with the linear
         attenuation factor.
       @param quadratic Reference to variable that will be set with the
         quadratic attenuation factor.
   */
   void getAttenuation(float &constant, float &linear, float &quadratic) const;

   /** Set the falloff threshold for computing a dynamic bounding sphere for
       this light.  The radius of the sphere will equal the distance to where
       the light falls off to the value specified here.  This will only go into
       effect if eDYNAMIC bounding sphere computation is selected.
       @param falloff A percentage greater than 0.  The default is 0.2 (20
         percent).
   */
   void setFalloff(float falloff);

   /** Query the current falloff percentage.
       @return The current falloff threshold.
   */
   float getFalloff() const { return 1.0f/m_falloff_inverse; }

   // See sglLight::setup() for details.
   virtual void setup(GLenum gl_light_number) const;

   // See sglNode::clone(..) for details.
   virtual sglNode *clone(unsigned int mode) const;

   // See sglObject::printInfo() for details.
   virtual void printInfo(ostream &ostrm, const char *indent_string) const;

protected:
   virtual void computeBound();

   void copyTo(sglPositionalLight *dst, unsigned int mode) const;

private: // methods - not implemented
   sglPositionalLight(const sglPositionalLight &);
   sglPositionalLight &operator=(const sglPositionalLight &);

protected: // variables
   sglVec4f m_position;

   float    m_attenuation_constant;
   float    m_attenuation_linear;
   float    m_attenuation_quadratic;

   float    m_falloff_inverse;
};

#endif
