/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglProbe.hpp
 *   Author: Scott McMillan
 *  Created: 18 April 2001
 *  Summary: common base class for sglIntersect and sglPick classes only
 *           defines the mode enums
 *****************************************************************************/

#ifndef __SGL_PROBE_HPP
#define __SGL_PROBE_HPP

#include <sgl.h>
#include <sglTraverser.hpp>
#include <sglMat4.hpp>
#include <sglFrustum.hpp>

class sglGeode;
class sglDrawable;

//----------------------------------------------------------------------------

/**

Dummy class to hold all the probe mode bit masks without having a template

*/

class SGL_DLL_API sglProbe : public sglTraverser
{
public:
   enum ModeEnum
   {
      ePRIMITIVE      = 0x00000001, // intersect w/ prims (incomplete)
      eGEOMETRY       = 0x00000002, // intersect w/ bounding boxes
      eGEODE          = 0x00000004, // intersect w/ bounding spheres (default)
      eCULL_BACK      = 0x00000008, // cull back faces (default)
      eCULL_FRONT     = 0x00000010, // cull front faces
      eNORMAL         = 0x00000020, // return the normal of the tri isected w/
      eTRANSFORM      = 0x00000040, // return the transform to the geode
      ePATH           = 0x00000080, // return the path to intersected node

      eSWITCH_ALL     = 0x00000100, // default is curr switch value
      eSWITCH_NONE    = 0x00000200, //

      eLOD_ALL        = 0x00000400, // default is range 0
      eLOD_NONE       = 0x00000800, //
      eLOD_VIEW       = 0x00001000, //

      // the following requires the current modelview matrix
      eBILLBOARD_VIEW = 0x00002000, // default is no billboards

      eANIMATION_ALL  = 0x00004000, // default is curr animation child
      eANIMATION_NONE = 0x00008000, //

      eLAYER_NONE     = 0x00010000, // default is all children
      eLAYER_BASE     = 0x00020000, // sglLayer::intersect()  not implemented
      eLAYER_DECAL    = 0x00040000, // sglLayer::pick()       not implemented

      // the following require the current modelview matrix and the view
      // frustum, and view port size in pixels.
      eUNSCALE_VIEW   = 0x00080000, // default is no unscale nodes
      eUNPROJECT_VIEW = 0x00100000  // default is no unproject nodes
   };

   sglProbe() : sglTraverser() {}
   virtual ~sglProbe() {}
};

//----------------------------------------------------------------------------

/**

Base traversal class for intersect and pick traversals

*/

template <class T>
class SGL_DLL_API sglProbeTemplate : public sglProbe
{
public: // Sub types
   class SGL_DLL_API ProbeResult
   {
   public:
      const sglGeode    *m_geode;     // sglGeode above the picked bbox
      const sglDrawable *m_geometry;  // sglDrawable within the picked bbox
      sglMat4<T>         m_xform;     // homogeneous transform to picked geode

   public:
      ProbeResult() : m_geode(NULL), m_geometry(NULL)
         { m_xform.buildIdentity(); }
      virtual ~ProbeResult() {}

      virtual void reset() = 0;

   protected:
      // called by the subclasses reset() function.  This is a HACK due to
      // Windoze stupidity
      void resetInternal()
         {
            m_geode = NULL;
            m_geometry = NULL;
            m_xform.buildIdentity();
         }
   };

public:  // Methods
   sglProbeTemplate() : sglProbe(), m_mode(eGEODE|eCULL_BACK) {}
   virtual ~sglProbeTemplate() {}

   virtual void reset()
      {
         m_modelview_matrix.buildIdentity(); // = pick_matrix;
      }

   void setMode(unsigned int mode) { m_mode = mode; }
   unsigned int getMode() const    { return m_mode; }

   ///
   void setViewportSize(unsigned int x, unsigned int y)
      { m_viewport_x = x; m_viewport_y = y; }
   ///
   void getViewportSize(unsigned int &x, unsigned int &y)
      { x = m_viewport_x; y = m_viewport_y; }

   /// Used only to initialize before a traversal (NOT DURING).
   void setViewFrustum(const sglFrustum<T> &frustum)
      {
         m_orig_view_frustum = frustum;
         m_current_view_frustum = frustum;
      }

   /** Used internally by the sglUnScale::pick|intersect(..) functions to
       retrieve the untransformed frustum.
   */
   sglFrustum<T> &getOrigViewFrustum() { return m_orig_view_frustum; }
   ///
   sglFrustum<T> &getViewFrustum() { return m_current_view_frustum; }

   /** Used to determine which lod to intersect for eLOD_VIEW,
       or to transform billboards for eBILLBOARD_VIEW.
   */
   sglMat4<T>& getModelViewMatrix() { return m_modelview_matrix; }

   /** Used by transform group nodes to set the new modelview matrix
    */
   void setModelViewMatrix(const sglMat4<T> &mat) { m_modelview_matrix = mat; }


private: // Methods - NOT IMPLEMENTED
   sglProbeTemplate(const sglProbeTemplate &);
   sglProbeTemplate &operator=(const sglProbeTemplate &);

protected: // Variables
   unsigned int     m_mode;

   unsigned int     m_viewport_x;
   unsigned int     m_viewport_y;
   sglFrustum<T>    m_orig_view_frustum;
   sglFrustum<T>    m_current_view_frustum;

   sglMat4<T>       m_modelview_matrix;  // current transform matrix
};

#endif
