/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglScale.cpp
 *  Created: 7 August 1999
 *  Summary: Uniform scale of the model view matrix
 *****************************************************************************/

#include <sglScale.hpp>
#include <sglCull.hpp>

//----------------------------------------------------------------------------
template <class T>
void sglScale<T>::setScale(T scale)
{
   if (scale != m_scale && scale != 0.)
   {
      m_scale = scale;
      m_inv_scale = 1./scale;
      dirty(eBOUND);
   }
}

//----------------------------------------------------------------------------
template <class T>
void sglScale<T>::computeBound()
{
   sglGroup::computeBound();
   if (!m_sphere.isEmpty() && !m_sphere.isInfinite())
   {
      m_sphere.setCenter(m_sphere.getCenter()*m_scale); // center);
      m_sphere.setRadius(m_sphere.getRadius()*m_scale);
   }
}

//----------------------------------------------------------------------------
template <class T>
sglBound::IntersectResultEnum sglScale<T>::intersect(
   sglIntersectf &segset) const
{
   if (m_scale == (T)1.0)
   {
      return sglGroup::intersect(segset);
   }

   if (m_child_list.empty())
   {
      return sglBound::eOUTSIDE;
   }

   sglSegmentf seg(segset.getSegment()); // copy the current segment

   // test against the current bounding sphere
   sglBound::IntersectResultEnum result = m_sphere.intersect(seg);
   if (result == sglBound::eOUTSIDE)
   {
      return result;
   }

   // Push a scale onto the stack.
   sglMat4f old_matrix;
   sglFrustumf old_view_frustum;
   segset.pushScale(m_scale, m_inv_scale, seg, old_matrix, old_view_frustum);

   // traverse the children
   result = sglBound::eOUTSIDE;
   for (vector<sglNode*>::const_iterator elt = m_child_list.begin();
        elt != m_child_list.end(); ++elt)
   {
      if ((*elt)->intersect(segset) == sglBound::eINTERSECT)
      {
         result = sglBound::eINTERSECT;
      }
   }

   // restore the original pos (leave len and dir intact) and unscale
   seg.setLen(segset.getSegment().getLen() * m_scale);

   // pop the scale off the stack
   segset.pop(seg, old_matrix, old_view_frustum);

   // push this node onto the path
   if (result == sglBound::eINTERSECT && segset.getMode() & sglProbe::ePATH)
   {
      segset.getPath().push_front((sglNode*)this);
   }

   return result;
}

//----------------------------------------------------------------------------
template <class T>
sglBound::IntersectResultEnum sglScale<T>::intersect(
   sglIntersectd &segset) const
{
   if (m_scale == (T)1.0)
   {
      return sglGroup::intersect(segset);
   }

   if (m_child_list.empty())
   {
      return sglBound::eOUTSIDE;
   }

   sglSegmentd seg(segset.getSegment()); // copy the current segment

   // test against the current bounding sphere
   sglBound::IntersectResultEnum result = m_sphere.intersect(seg);
   if (result == sglBound::eOUTSIDE)
   {
      return result;
   }

   // Push a scale onto the stack.
   sglMat4d old_matrix;
   sglFrustumd old_view_frustum;
   segset.pushScale(m_scale, m_inv_scale, seg, old_matrix, old_view_frustum);

   // traverse the children
   result = sglBound::eOUTSIDE;
   for (vector<sglNode*>::const_iterator elt = m_child_list.begin();
        elt != m_child_list.end(); ++elt)
   {
      if ((*elt)->intersect(segset) == sglBound::eINTERSECT)
      {
         result = sglBound::eINTERSECT;
      }
   }

   // restore the original pos (leave len and dir intact)
   seg.setLen(segset.getSegment().getLen() * m_scale);

   // pop the scale off the stack
   segset.pop(seg, old_matrix, old_view_frustum);

   // push this node onto the path
   if (result == sglBound::eINTERSECT && segset.getMode() & sglProbe::ePATH)
   {
      segset.getPath().push_front((sglNode*)this);
   }

   return result;
}

//----------------------------------------------------------------------------
template <class T>
void sglScale<T>::pick(sglPickf &pick_state, unsigned int cull_flags) const
{
   if (m_scale == (T)1.0)
   {
      sglGroup::pick(pick_state, cull_flags);
   }
   else if (m_child_list.empty() ||
            (cull_flags &&
             (m_sphere.intersect(pick_state.getPickFrustum(),
                                 cull_flags) == sglBound::eOUTSIDE)))
   {
      return; // select nothing
   }
   else
   {
      sglMat4f old_matrix;
      sglFrustumf old_frustum, old_view_frustum;
      pick_state.pushScale(m_scale, m_inv_scale,
                           old_frustum, old_matrix, old_view_frustum);

      // if path is requested store it ahead of time (different from
      // intersect), in case something beneath here is picked successfully.
      if (pick_state.getMode() & sglProbe::ePATH)
      {
         pick_state.getPath().push_back((sglNode*)this);
      }

      // --------------------------------------------------

      for (vector<sglNode*>::const_iterator elt = m_child_list.begin();
           elt != m_child_list.end(); ++elt)
      {
         (*elt)->pick(pick_state, cull_flags);
      }

      // --------------------------------------------------
      // restore old state
      pick_state.pop(old_frustum, old_matrix, old_view_frustum);

      if (pick_state.getMode() & sglProbe::ePATH)
      {
         pick_state.getPath().pop_back();
      }
   }
}

//----------------------------------------------------------------------------
template <class T>
void sglScale<T>::pick(sglPickd &pick_state, unsigned int cull_flags) const
{
   if (m_scale == (T)1.0)
   {
      sglGroup::pick(pick_state, cull_flags);
   }
   else if (m_child_list.empty() ||
            (cull_flags &&
             (m_sphere.intersect(pick_state.getPickFrustum(),
                                 cull_flags) == sglBound::eOUTSIDE)))
   {
      return; // select nothing
   }
   else
   {
      sglMat4d old_matrix;
      sglFrustumd old_frustum, old_view_frustum;
      pick_state.pushScale(m_scale, m_inv_scale,
                           old_frustum, old_matrix, old_view_frustum);

      // if path is requested store it ahead of time (different from
      // intersect), in case something beneath here is picked successfully.
      if (pick_state.getMode() & sglProbe::ePATH)
      {
         pick_state.getPath().push_back((sglNode*)this);
      }

      // --------------------------------------------------

      for (vector<sglNode*>::const_iterator elt = m_child_list.begin();
           elt != m_child_list.end(); ++elt)
      {
         (*elt)->pick(pick_state, cull_flags);
      }

      // --------------------------------------------------
      // restore old state
      pick_state.pop(old_frustum, old_matrix, old_view_frustum);

      if (pick_state.getMode() & sglProbe::ePATH)
      {
         pick_state.getPath().pop_back();
      }
   }
}

//----------------------------------------------------------------------------
template <class T>
void sglScale<T>::cull(sglCull<float> &trav_state,
                       unsigned int cull_flags) const
{
   if (m_scale == (T)1.0)
   {
      sglGroup::cull(trav_state, cull_flags);
   }
   else if (m_child_list.empty() ||
            (cull_flags &&
             (m_sphere.intersect(trav_state.getFrustum(),
                                 cull_flags) == sglBound::eOUTSIDE)))
   {
      return; // draw nothing
   }
   else
   {
      // compute a new matrix
      sglMat4f new_mat;
      new_mat.scale(trav_state.getModelViewMatrix(),
                    m_scale, m_scale, m_scale);
      trav_state.pushModelViewMatrix(new_mat, true, false);

      // compute a new frustum
      sglFrustumf old_frustum = trav_state.getFrustum();
      trav_state.getFrustum().scale(old_frustum, m_inv_scale);

      // --------------------------------------------------

      for (vector<sglNode*>::const_iterator elt = m_child_list.begin();
           elt != m_child_list.end(); ++elt)
      {
         (*elt)->cull(trav_state, cull_flags);
      }

      // --------------------------------------------------
      // restore old state
      trav_state.popModelViewMatrix();
      trav_state.getFrustum() = old_frustum;
   }
}

//----------------------------------------------------------------------------
template <class T>
void sglScale<T>::cull(sglCull<double> &trav_state,
                       unsigned int cull_flags) const
{
   if (m_scale == (T)1.0)
   {
      sglGroup::cull(trav_state, cull_flags);
   }
   else if (m_child_list.empty() ||
            (cull_flags &&
             (m_sphere.intersect(trav_state.getFrustum(),
                                 cull_flags) == sglBound::eOUTSIDE)))
   {
      return; // draw nothing
   }
   else
   {
      // compute a new matrix
      sglMat4d new_mat;
      new_mat.scale(trav_state.getModelViewMatrix(),
                    m_scale, m_scale, m_scale);
      trav_state.pushModelViewMatrix(new_mat, true, false);

      // compute a new frustum
      sglFrustumd old_frustum = trav_state.getFrustum();
      trav_state.getFrustum().scale(old_frustum, m_inv_scale);

      // --------------------------------------------------

      for (vector<sglNode*>::const_iterator elt = m_child_list.begin();
           elt != m_child_list.end(); ++elt)
      {
         (*elt)->cull(trav_state, cull_flags);
      }

      // --------------------------------------------------
      // restore old state
      trav_state.popModelViewMatrix();
      trav_state.getFrustum() = old_frustum;
   }
}

//----------------------------------------------------------------------------
template <class T>
void sglScale<T>::computeTxToParent(sglMat4f &curr_mat) const
{
   curr_mat.scale(curr_mat, m_inv_scale, m_inv_scale, m_inv_scale);
}

//----------------------------------------------------------------------------
template <class T>
void sglScale<T>::computeTxToParent(sglMat4d &curr_mat) const
{
   curr_mat.scale(curr_mat, m_inv_scale, m_inv_scale, m_inv_scale);
}

//----------------------------------------------------------------------------
template <class T>
sglNode *sglScale<T>::clone(unsigned int mode) const
{
   sglScale<T> *root = new sglScale<T>;
   sglScale<T>::copyTo(root, mode);
   return root;
}

//----------------------------------------------------------------------------
template <class T>
void sglScale<T>::copyTo(sglScale<T> *dst, unsigned int mode) const
{
   sglGroup::copyTo(dst, mode);

   // transfer stuff
   dst->setScale(m_scale);
}

//----------------------------------------------------------------------------
template <class T>
void sglScale<T>::printInfo(ostream &ostrm,
                            const char *indent_string) const
{
    sglGroup::printInfo(ostrm, indent_string);

    ostrm << indent_string << "sglScale: " << m_scale << endl;
}


template class SGL_DLL_API sglScale<double>;
template class SGL_DLL_API sglScale<float>;
