/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglScale.hpp
 *  Created: 7 August 1999
 *  Summary: Uniform scale of the model view matrix
 *****************************************************************************/

#ifndef __SGL_SCALE_HPP
#define __SGL_SCALE_HPP

#include <sgl.h>
#include <sglMatrix.hpp>
#include <sglGroup.hpp>

//----------------------------------------------------------------------------

/**

The sglScale node scales the traversal's model view matrix effectively changing
the size of the translations and geometry in the subtree below this one.  This
node specifies a uniform scale (the same in all dimensions) and is more
efficient to use that a general sglTransform matrix.  sglScalef and sglScaled
are typedefs for single and double precision (templated) versions.

<b>Example:</b>

<pre>
   sglScaled *dp_scale_node = new sglScaled;
   dp_translate_node->setScale(25.0000001);
</pre>

*/

template <class T>
class SGL_DLL_API sglScale : public sglGroup
{
public:
   /// default constructor intializes to unit scale.
   sglScale() : sglGroup(), m_scale(1.), m_inv_scale(1.) {}
   /// virtual destructor
   virtual ~sglScale() {}

   /** Set the scale factor.
       @param scale The scale factor.  A zero scale will not be accepted.
   */
   void setScale(T scale);

   /** Get the current scale factor.
       @return The current scale.
    */
   T getScale() const { return m_scale; }

   // See sglNode::intersect(f) for details.
   virtual sglBound::IntersectResultEnum intersect(
      sglIntersectf &isector) const;

   // See sglNode::intersect(d) for details.
   virtual sglBound::IntersectResultEnum intersect(
      sglIntersectd &isector) const;

   // See sglNode::pick(f) for details.
   virtual void pick(sglPickf &pick_state, unsigned int cull_flags) const;

   // See sglNode::pick(d) for details.
   virtual void pick(sglPickd &pick_state, unsigned int cull_flags) const;

   // See sglNode::computeTxToParent(f) for details.
   virtual void computeTxToParent(sglMat4f &curr_mat) const;

   // See sglNode::computeTxToParent(d) for details.
   virtual void computeTxToParent(sglMat4d &curr_mat) const;

   // See sglNode::clone() for details.
   virtual sglNode *clone(unsigned int mode) const;

   // See sglObject::printInfo() for details.
   virtual void printInfo(ostream &ostrm, const char *indent_string) const;

protected:
   // compute the bounding sphere
   virtual void computeBound();

   // See sglNode::cull(f) for details
   virtual void cull(sglCull<float> &trav_state,
                     unsigned int cull_flags) const;

   // See sglNode::cull(d) for details
   virtual void cull(sglCull<double> &trav_state,
                     unsigned int cull_flags) const;

   void copyTo(sglScale *dst, unsigned int mode) const;

private:   // not implemented
   sglScale(const sglScale &);
   sglScale &operator=(const sglScale &);

protected:
   T m_scale;
   T m_inv_scale;
};

//----------------------------------------------------------------------------
// convenience types:
//    template class 'should' only be float or double

template <class T> class sglScale;

///
typedef sglScale<float>  sglScalef;
///
typedef sglScale<double> sglScaled;

#endif
