/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglScene.cpp
 *  Created: 2 January 1999
 *  Summary: root of scenegraph, initializes and begins the traversal
 *****************************************************************************/

#include <sglVector.hpp>
#include <sglMatrix.hpp>
#include <sglPolytope.hpp>
#include <sglScene.hpp>
#include <sglOStream.hpp>

#include <sglStatelet.hpp>
#include <sglFog.hpp>
#include <sglMaterial.hpp>
#include <sglTexEnv.hpp>
#include <sglTexGen.hpp>
#include <sglTexture2D.hpp>

#include <sglViewPlatform.hpp>
#include <sglPerspectiveCamera.hpp>

//----------------------------------------------------------------------------
sglScene::sglScene()
      : sglGroup(),
        m_frame_count(0),
        m_frame_time()
{
   createDefaultStatelets(m_default_statelets);
}

//----------------------------------------------------------------------------
sglScene::~sglScene()
{
   for (vector<const sglStatelet*>::iterator ii = m_default_statelets.begin();
        ii != m_default_statelets.end(); ++ii)
   {
      if ((*ii)->getRefCount() != 0)
      {
         sglPrintWarning() << "~sglScene() warning: default statelet (type = "
                           << (*ii)->getType() << "), ref count = "
                           << (*ii)->getRefCount() << endl;
      }
      else
      {
         delete *ii;
      }
   }
}

//----------------------------------------------------------------------------
float sglScene::computeLODScale(float fov_x, float fov_y,
                                unsigned int viewport_x,
                                unsigned int viewport_y)
{
   float horiz = (viewport_x*0.5f)/tan(fov_x*0.5f);
   float vert  = (viewport_y*0.5f)/tan(fov_y*0.5f);
   float max = (horiz > vert) ? horiz : vert;

   // not sure if fov should always be fov_x or the following:
   float fov = (horiz > vert) ? fov_x : fov_y;

   float lod_aspect_2 = (512/cTanPi8)/max;

   sglPrintDebug() << "sglScene::computeLODScale() "
                   << "fox x,y: " << fov_x*180./M_PI << ", "
                   << fov_y*180./M_PI << ", viewport x,y: "
                   << viewport_x << ", " << viewport_y << endl;
   sglPrintDebug() << "sglScene::computeLODScale() "
                   << "horiz/vert: " << horiz << "/" << vert << ": " << max
                   << "; fov " << fov << ", lod_aspect_2: "
                   << lod_aspect_2 << endl;
   sglPrintDebug() << "scale: " << (fov/M_PI_4)*lod_aspect_2 << endl;

   return (fov/M_PI_4)*lod_aspect_2;
}

//----------------------------------------------------------------------------
void sglScene::createDefaultStatelets(vector<const sglStatelet*> &def)
{
   if (!sgl::s_initialized)
   {
      sglPrintWarning() << "sglScene::createDefaultStatelets() error: "
                        << "called before sgl::initialize().  "
                        << "The cull traversal will not work if there is more"
                        << " than one texture unit" << endl;
   }

   def.push_back(new sglAlphaTest());
   def.push_back(new sglAntiAliasLine());
   def.push_back(new sglAntiAliasPoint());
   def.push_back(new sglAntiAliasPolygon());
   def.push_back(new sglBlending());
   def.push_back(new sglColorMask());
   def.push_back(new sglCullMode());
   def.push_back(new sglDepthMask());
   def.push_back(new sglDepthRange());
   def.push_back(new sglDepthTest());
   def.push_back(new sglDithering());
   def.push_back(new sglFog());
   def.push_back(new sglFrontFace());
   def.push_back(new sglHighlight());
   def.push_back(new sglLighting());
   def.push_back(new sglLineStipple());
   def.push_back(new sglLineWidth());
   def.push_back(new sglLogicOp());
   def.push_back(new sglMaterial());
   def.push_back(new sglOcclusionCull());
   def.push_back(new sglPointSize());
   def.push_back(new sglPointParameters());
   def.push_back(new sglPolyBackStyle());
   def.push_back(new sglPolyFrontStyle());
   def.push_back(new sglPolygonOffset());
   def.push_back(new sglPolygonStipple());
   def.push_back(new sglShadeModel());
   def.push_back(new sglTransparency());

   // first (default) texture unit.
   def.push_back(new sglTexEnv());
   def.push_back(new sglTexGen());
   def.push_back(new sglTexture2D());
   def.push_back(new sglTextureMatrix());

   // additional texture units.
   if (sgl::s_num_texture_units > 1)
   {
      sglTexEnv *tex_env = new sglTexEnv();
      sglTexGen *tex_gen = new sglTexGen();
      sglTexture2D *tex = new sglTexture2D();
      sglTextureMatrix *tex_matrix = new sglTextureMatrix();

      for (GLint i=1; i<sgl::s_num_texture_units; ++i)
      {
         def.push_back(new sglMultiTextureStatelet(tex_env, i));
         def.push_back(new sglMultiTextureStatelet(tex_gen, i));
         def.push_back(new sglMultiTextureStatelet(tex, i));
         def.push_back(new sglMultiTextureStatelet(tex_matrix, i));
      }
   }

   if (def.size() != sgl::s_num_statelet_types-1) // minus transform matrix
   {
      sglPrintWarning()
          << "sglScene::createDefaultStatelets() error: "
          << "default_statelets.size() != sgl::s_num_statelet_types" << endl;
   }
}

//----------------------------------------------------------------------------
void sglScene::preDraw()
{
   sglTimespec frame_time;
   sglTimespec::getSysTime(frame_time);
   preDraw(frame_time, m_frame_count+1);
}

//----------------------------------------------------------------------------
void sglScene::preDraw(const sglTimespec &frame_time, unsigned int frame_count)
{
   sglPrintDebug() << "sglScene::preDraw(ft, fc)" << endl;

   sglTexture::bindDirtyTextures();

   PreDrawStruct trav_state;
   trav_state.m_frame_count = m_frame_count = frame_count;
   trav_state.m_frame_time  = m_frame_time  = frame_time;

   // start the preDraw traversal
   if (m_dirty)
      sglGroup::preDraw(trav_state);

   // clean up stuff
   sgl::deleteUnusedDisplayLists();
   sgl::deleteUnusedTextureObjs();
}

//----------------------------------------------------------------------------
// culled draw functions
//----------------------------------------------------------------------------
void sglScene::drawCulled(float fov_x, float fov_y,
                          unsigned int viewport_x, unsigned int viewport_y,
                          const sglFrustumf &orig_frustum,
                          const sglMat4f &view_matrix,
                          const deque<const sglStatelet*> &override_statelets,
                          unsigned int discriminator,
                          unsigned int state_mask) const
{
   // create the traversal state
   sglCullf trav_state;

   cull(trav_state, fov_x, fov_y, viewport_x, viewport_y, orig_frustum,
        view_matrix, override_statelets, discriminator, state_mask);
   trav_state.draw();
}

//----------------------------------------------------------------------------
void sglScene::cull(sglCullf &trav_state,
                    unsigned int viewport_x, unsigned int viewport_y,
                    const sglViewPlatform &view_platform,
                    const deque<const sglStatelet*> &override_statelets,
                    unsigned int discriminator,
                    unsigned int state_mask) const
{
   // HACK - default for ortho cameras, this isn't the desired behaviour.
   float fov_y = 0.25f*(float)M_PI;
   float fov_x = 0.25f*(float)M_PI;

   sglPerspectiveCamera *camera = dynamic_cast<sglPerspectiveCamera*>(
      &(view_platform.getCamera()));
   if (camera)
   {
      fov_x = camera->getFOVX();
      fov_y = camera->getFOVY();
      sglPrintDebug() << "f: fov x, y: " << fov_x*180.0/M_PI << ", "
                      << fov_y*180.0/M_PI << endl;
   }

   float lod_scale = computeLODScale(fov_x, fov_y, viewport_x, viewport_y);
   sglPrintDebug() << "sglScene::cull() lod_scale = " << lod_scale << endl;

   // HACK what do I do if the root of the scene graph containing this
   // view platform node is not this sglScene node??
   sglMat4f view_matrix;
   view_platform.computeViewMatrix(view_matrix);

   sglFrustumf frustum = view_platform.getCamera().getFrustum();

   // create and initialize the traversal state
   trav_state.initialize(lod_scale, discriminator, state_mask,
                         view_matrix, frustum,
                         viewport_x, viewport_y,
                         m_default_statelets,
                         override_statelets,
                         m_frame_count,
                         m_frame_time);

   unsigned int cull_flags = (1 << frustum.getNumPlanes()) - 1;
   cull(trav_state, cull_flags);
}

//----------------------------------------------------------------------------
void sglScene::cull(sglCullf &trav_state,
                    float fov_x, float fov_y,
                    unsigned int viewport_x, unsigned int viewport_y,
                    const sglFrustumf &orig_frustum,
                    const sglMat4f &view_matrix,
                    const deque<const sglStatelet*> &override_statelets,
                    unsigned int discriminator,
                    unsigned int state_mask) const
{
   float lod_scale = computeLODScale(fov_x, fov_y, viewport_x, viewport_y);
   sglPrintDebug() << "sglScene::cull() lod_scale = " << lod_scale << endl;

   // create and initialize the traversal state
   trav_state.initialize(lod_scale, discriminator, state_mask,
                         view_matrix, orig_frustum,
                         viewport_x, viewport_y,
                         m_default_statelets,
                         override_statelets,
                         m_frame_count,
                         m_frame_time);

   unsigned int cull_flags = (1 << orig_frustum.getNumPlanes()) - 1;
   cull(trav_state, cull_flags);
}

//----------------------------------------------------------------------------
void sglScene::drawCulled(float fov_x, float fov_y,
                          unsigned int viewport_x, unsigned int viewport_y,
                          const sglFrustumd &orig_frustum,
                          const sglMat4d &view_matrix,
                          const deque<const sglStatelet*> &override_statelets,
                          unsigned int discriminator,
                          unsigned int state_mask) const
{
   // create the traversal state
   sglCulld trav_state;

   cull(trav_state, fov_x, fov_y, viewport_x, viewport_y, orig_frustum,
        view_matrix, override_statelets, discriminator, state_mask);
   trav_state.draw();
}

//----------------------------------------------------------------------------
void sglScene::cull(sglCulld &trav_state,
                    unsigned int viewport_x, unsigned int viewport_y,
                    const sglViewPlatform &view_platform,
                    const deque<const sglStatelet*> &override_statelets,
                    unsigned int discriminator,
                    unsigned int state_mask) const
{
   // HACK - default for ortho cameras, this isn't the desired behaviour.
   float fov_y = 0.25f*(float)M_PI;
   float fov_x = 0.25f*(float)M_PI;

   sglPerspectiveCamera *camera = dynamic_cast<sglPerspectiveCamera*>(
      &(view_platform.getCamera()));
   if (camera)
   {
      fov_x = (float) camera->getFOVX();
      fov_y = (float) camera->getFOVY();
      sglPrintDebug() << "d: fov x, y: " << fov_x*180.0/M_PI << ", "
                      << fov_y*180.0/M_PI << endl;
   }

   float lod_scale = computeLODScale(fov_x, fov_y, viewport_x, viewport_y);
   sglPrintDebug() << "sglScene::cull() lod_scale = " << lod_scale << endl;

   // HACK what do I do if the root of the scene graph containing this
   // view platform node is not this sglScene node??
   sglMat4d view_matrix;
   view_platform.computeViewMatrix(view_matrix);

   // initialize the traversal state
   trav_state.initialize(lod_scale, discriminator, state_mask,
                         view_matrix,
                         view_platform.getCamera().getFrustum(),
                         viewport_x, viewport_y,
                         m_default_statelets,
                         override_statelets,
                         m_frame_count,
                         m_frame_time);

   unsigned int cull_flags =
      (1 << view_platform.getCamera().getFrustum().getNumPlanes()) - 1;
   cull(trav_state, cull_flags);
}

//----------------------------------------------------------------------------
void sglScene::cull(sglCulld &trav_state,
                    float fov_x, float fov_y,
                    unsigned int viewport_x, unsigned int viewport_y,
                    const sglFrustumd &orig_frustum,
                    const sglMat4d &view_matrix,
                    const deque<const sglStatelet*> &override_statelets,
                    unsigned int discriminator,
                    unsigned int state_mask) const
{
   float lod_scale = computeLODScale(fov_x, fov_y, viewport_x, viewport_y);
   sglPrintDebug() << "sglScene::cull() lod_scale = " << lod_scale << endl;

   // initialize the traversal state
   trav_state.initialize(lod_scale, discriminator, state_mask,
                         view_matrix, orig_frustum,
                         viewport_x, viewport_y,
                         m_default_statelets,
                         override_statelets,
                         m_frame_count,
                         m_frame_time);

   unsigned int cull_flags = (1 << orig_frustum.getNumPlanes()) - 1;
   cull(trav_state, cull_flags);
}

//----------------------------------------------------------------------------
sglNode *sglScene::clone(unsigned int mode) const
{
   sglScene *root = new sglScene;
   sglScene::copyTo(root, mode);
   return root;
}

//----------------------------------------------------------------------------
void sglScene::copyTo(sglScene *dst, unsigned int mode) const
{
   sglGroup::copyTo(dst, mode);
   dst->m_frame_time  = m_frame_time;
   dst->m_frame_count = m_frame_count;
}

//----------------------------------------------------------------------------
void sglScene::printInfo(ostream &ostrm, const char *indent_string) const
{
   sglGroup::printInfo(ostrm, indent_string);
   ostrm << indent_string << "frame time:  " << m_frame_time  << endl;
   ostrm << indent_string << "frame count: " << m_frame_count << endl;
}
