/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglScene.hpp
 *  Created: 2 January 1999
 *  Summary: Root node in a scene graph initializes and begins the traversal
 *****************************************************************************/

#ifndef __SGL_SCENE_HPP
#define __SGL_SCENE_HPP

#include <sgl.h>
#include <sglGroup.hpp>
#include <sglPolytope.hpp>
#include <sglCull.hpp>

class sglViewPlatform;

//----------------------------------------------------------------------------

/**

Under normal circumstances, the sglScene node is the root of the scene graph.
It contains cull functions that initialize the cull traversal.  One version of
this function takes arguments that define the view frustum, window size (in
pixels), the initial view matrix, a list (actually an STL deque) of override
statelets, discriminator mask, and state mask.  Another version of this
function uses the sglViewPlatform class to define the FOVs, view frustum, and
view matrix (all other parameters are the same).

The computeLODScale helper function determines a scale to be applied to the
level-of-detail (sglLOD) range calculations.  It is a function of the size of
the window in pixels and the field of view (FOV).  This function does not need
to be called if you call the cull entry points in this class.  See sglLOD for
an explanation for this function.

The createDefaultStatelets helper function fills the provided vector with
dynamically allocated statelets set to their default values.  The is used
internally with the cull traversal, but has been exposed for SGL users as
well.

<b>Example:</b>

<pre>
   // Must come before sglScene and sglCull classes are constructed
   sgl::initialize();

    // Add all geometry below this node.
   sglScene *root = new sglScene;

        :
        :

   // Reuse the same traversal state node from frame to frame for better
   // performance.
   sglCullf trav_state;

   while (!quit_flag)
   {
      // Do application specific work
      doAppStuff();

      // Adjust internal data-structures (display lists, bounding volumes,
      // etc...) for the parts of the scene graph that may have changed
      // since last frame.
      root->preDraw();

      // Collect the geometry intersecting the given frustum &amp; view_matrix.
      root->cull(trav_state, fov_x, fov_y, win_x, win_y,
                 frustum, view_matrix, override_statelets,
                 disc_mask, state_mask);

      // Render the visible geometry.
      trav_state.draw();
   }
</pre>

<b>Example using the sglViewPlatform:</b>

<pre>
   // Must come before sglScene and sglCull classes are constructed
   sgl::initialize();

   // Add all geometry below this node.
   sglScene *root = new sglScene;

   // used to move the camera around
   sglTransformf view_trans = new sglTransformf;
   root->addChild(view_trans);

   // set up the view platform and camera
   sglPerspectiveCamera *view_camera = new sglPerspectiveCamera;
   view_camera->setFOV(M_PI*0.25, (double)win_x/(double)win_y, 1.0, 200.0);

   sglViewPlatform *view_platform = new sglViewPlatform(*view_camera);
   view_platform->setEnableZUp(true);
   view_trans->addChild(view_platform);

        :
        :

   // Reuse the same traversal state node from frame to frame for better
   // performance.
   sglCullf trav_state;

   while (!quit_flag)
   {
      // Do application specific work
      doAppStuff();

      // Adjust internal data-structures (display lists, bounding volumes,
      // etc...) for the parts of the scene graph that may have changed
      // since last frame.
      root->preDraw();

      // Collect the geometry intersecting the given frustum &amp; view_matrix.
      root->cull(trav_state,
                 win_x, win_y,
                 *view_platform,
                 override_statelets, disc_mask, state_mask);

      // setup OpenGL projection matrix
      (view_platform->getCamera()).applyProjection();

      // Render the visible geometry.
      trav_state.draw();
   }
</pre>

 */

class SGL_DLL_API sglScene : public sglGroup
{
public:
   /// Default constructor.  Must be called only after sgl::initialize().
   sglScene();
   /// virtual destructor
   virtual ~sglScene();

   /** Helper function to compute an LOD scale value based on the current field
       of view, and window size.  This is used to select higher detail children
       of sglLOD nodes in larger windows and/or smaller fields of view.
       @param fov_x Horizontal field of view in radians
       @param fov_y Vertical field of view in radians
       @param viewport_x Horizontal dimension of viewport in pixels
       @param viewport_y Vertical dimension of viewport in pixels
       @return An LOD scale factor.
   */
   static float computeLODScale(float fov_x, float fov_y,
                                unsigned int viewport_x,
                                unsigned int viewport_y);

   /** Helper function to fill a vector with all of the default statelet nodes
       in the correct order (used by the state sorting mechanism).
       @param def A vector which will be filled with the statelets.  It is
         assumed to be empty before calling this function.  All statelets
         placed in this vector are dynamically allocated (with new) and it is
         the responsibility of the user of the function to delete the statelets
         when no longer needed.
   */
   static void createDefaultStatelets(vector<const sglStatelet*> &def);

   /** This functions must be called with a valid OpenGL graphics context
       current.  Must be called once before each "frame" to increment the frame
       count and frame time.  If necessary it will also traverse the scene
       graph and "clean" it (clear the dirty bits) by creating any needed
       display lists, binding any textures(?), recomputing any bounding boxes,
       and reconfiguring animation nodes.
   */
   void preDraw();

   /** An alternative to the preDraw() function in which the user may specify
       the current frame time and frame count.  This functions must be called
       with a valid OpenGL graphics context current.  Must be called once
       before each "frame" to increment the frame count and frame time.  If
       necessary it will also traverse the scene graph and "clean" it (clear
       the dirty bits) by creating any needed display lists, binding any
       textures(?), recomputing any bounding boxes, and reconfiguring animation
       nodes.
       @param frame_time The time for the next cull/draw traversal.
       @param frame_count The count for the next cull/draw traversal.
   */
   void preDraw(const sglTimespec &frame_time, unsigned int frame_count);

   /// To prevent ambiguity on SGI's MIPSPro 7.3.x compilers.  See sglNode.
   virtual void preDraw(const PreDrawStruct &trav) { sglGroup::preDraw(trav); }

   /** Get the current frame count (used in sglKeyFrameAnimation nodes).
       @return The current frame count.
   */
   unsigned int getFrameCount() const { return m_frame_count; }

   /** Get the current frame time (used in sglTimedAnimation nodes).
       @return The current frame time in an sglTimespec
   */
   const sglTimespec &getFrameTime() const { return m_frame_time; }

   // ----------------- single precision cull functions --------------------

   /** An entry point for the single precision cull traversal to perform
       view frustum culling (before rendering occurs).
       @param trav_state The object that will collect all the state and
         geometry information that will be rendered.
       @param fov_x The horizontal field of view in radians.
       @param fov_y The vertical field of view in radians.
       @param viewport_x The horizontal size of the viewport in pixels.
       @param viewport_y The vertical size of the viewport in pixels.
       @param orig_frustum The sglFrustum representing the view frustum defined
         relative to the current eye point.
       @param view_matrix The sglMat4 that is the inverse of the matrix
         representing the current position and orientation of the eye point
         relative to the rendering origin.  The origin of the orig_frustum is
         at the position defined by this matrix, and the frustum lies along the
         Z-axis, the X-axis points to the right, and the Y-axis points up.
       @param override_statelets An STL deque (double-ended queue?) that
         contains any override statelets the user may want to apply during the
         current frame.
       @param discriminator_mask A 32-bit (unsigned int) bit mask representing
         the discriminator mask (affects the behaviour of sglDiscriminator
         nodes).
       @param state_mask A 32-bit (unsigned int) bit mask representing the
         state mask (affects the selection of states in sglDrawable nodes).
   */
   void cull(sglCullf &trav_state,
             float fov_x, float fov_y,
             unsigned int viewport_x, unsigned int viewport_y,
             const sglFrustumf &orig_frustum,
             const sglMat4f &view_matrix,
             const deque<const sglStatelet*> &override_statelets,
             unsigned int discriminator_mask = 0,
             unsigned int state_mask = 0) const;

   /** An second entry point for the single precision cull traversal to perform
       view frustum culling (before rendering occurs) using an sglViewPlatform.
       @param trav_state The object that will collect all the state and
         geometry information that will be rendered.
       @param viewport_x The horizontal size of the viewport in pixels.
       @param viewport_y The vertical size of the viewport in pixels.
       @param view_platform The sglViewPlatform (which should be contained in
         the scene graph rooted by this sglScene node) that defines the view
         frustum, view matrix, and fields of view (if perspective).
       @param override_statelets An STL deque (double-ended queue?) that
         contains any override statelets the user may want to apply during the
         current frame.
       @param discriminator_mask A 32-bit (unsigned int) bit mask representing
         the discriminator mask (affects the behaviour of sglDiscriminator
         nodes).
       @param state_mask A 32-bit (unsigned int) bit mask representing the
         state mask (affects the selection of states in sglDrawable nodes).
   */
   void cull(sglCullf &trav_state,
             unsigned int viewport_x, unsigned int viewport_y,
             const sglViewPlatform &view_platform,
             const deque<const sglStatelet*> &override_statelets,
             unsigned int discriminator_mask = 0,
             unsigned int state_mask = 0) const;

   /** A deprecated entry point for the single precision cull traversal to
       perform view frustum culling AND rendering.  This function provides its
       own sglCull object for culling and drawing and destructs it before
       returning.
       @param fov_x The horizontal field of view in radians.
       @param fov_y The vertical field of view in radians.
       @param viewport_x The horizontal size of the viewport in pixels.
       @param viewport_y The vertical size of the viewport in pixels.
       @param orig_frustum The sglFrustum representing the view frustum defined
         relative to the current eye point.
       @param view_matrix The sglMat4 that is the inverse of the matrix
         representing the current position and orientation of the eye point
         relative to the rendering origin.  The origin of the orig_frustum is
         at the position defined by this matrix, and the frustum lies along the
         Z-axis, the X-axis points to the right, and the Y-axis points up.
       @param override_statelets An STL deque (double-ended queue?) that
         contains any override statelets the user may want to apply during the
         current frame.
       @param discriminator_mask A 32-bit (unsigned int) bit mask representing
         the discriminator mask (affects the behaviour of sglDiscriminator
         nodes).
       @param state_mask A 32-bit (unsigned int) bit mask representing the
         state mask (affects the selection of states in sglDrawable nodes).
       @deprecated
   */
   void drawCulled(float fov_x, float fov_y,
                   unsigned int viewport_x, unsigned int viewport_y,
                   const sglFrustumf &orig_frustum,
                   const sglMat4f &view_matrix,
                   const deque<const sglStatelet*> &override_statelets,
                   unsigned int discriminator_mask = 0,
                   unsigned int state_mask = 0) const;

   // ----------------- double precision cull functions --------------------

   /** An entry point for the double precision cull traversal to perform
       view frustum culling (before rendering occurs).  See single precision
       version for an explanation of the parameters.
   */
   void cull(sglCulld &trav_state,
             float fov_x, float fov_y,
             unsigned int viewport_x, unsigned int viewport_y,
             const sglFrustumd &orig_frustum,
             const sglMat4d &view_matrix,
             const deque<const sglStatelet*> &override_statelets,
             unsigned int discriminator_mask = 0,
             unsigned int state_mask = 0) const;

   /** An entry point for the double precision cull traversal to perform
       view frustum culling (before rendering occurs).  See single precision
       version for an explanation of the parameters.
   */
   void cull(sglCulld &trav_state,
             unsigned int viewport_x, unsigned int viewport_y,
             const sglViewPlatform &view_platform,
             const deque<const sglStatelet*> &override_statelets,
             unsigned int discriminator_mask = 0,
             unsigned int state_mask = 0) const;

   /** A deprecated entry point for the double precision cull traversal to
       perform view frustum culling (before rendering occurs).  See single
       precision version for an explanation of the parameters.
   */
   void drawCulled(float fov_x, float fov_y,
                   unsigned int viewport_x, unsigned int viewport_y,
                   const sglFrustumd &orig_frustum,
                   const sglMat4d &view_matrix,
                   const deque<const sglStatelet*> &override_statelets,
                   unsigned int discriminator_mask = 0,
                   unsigned int state_mask = 0) const;

   // See sglNode::clone() for details.
   virtual sglNode *clone(unsigned int mode) const;

   // See sglObject::printInfo() for details.
   virtual void printInfo(ostream &ostrm, const char *indent_string) const;

protected:
   // uses sglGroup's versions of these - these are here to avoid MIPSPro
   // compiler warnings:

   // see sglNode::cull(f) for details
   virtual void cull(sglCull<float> &trav_state,
                     unsigned int cull_flags) const
      {
         sglGroup::cull(trav_state, cull_flags);
      }

   // see sglNode::cull(d) for details
   virtual void cull(sglCull<double> &trav_state,
                     unsigned int cull_flags) const
      {
         sglGroup::cull(trav_state, cull_flags);
      }

   void copyTo(sglScene *dst, unsigned int mode) const;

private:
   sglScene(const sglScene &);            // not implemented
   sglScene &operator=(const sglScene&);  // not implemented

private:
   vector<const sglStatelet*>  m_default_statelets;

   unsigned int m_frame_count;
   sglTimespec  m_frame_time;
};

#endif
