/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglSphereBound.hpp
 *   Author: Scott McMillan
 *  Created: 24 March 2001 (extracted from sglSphereBound.cpp)
 *****************************************************************************/

#include <sglSphereBound.hpp>
#include <math.h>

//----------------------------------------------------------------------------
template <class T>
void sglSphereBound<T>::computeFromBox(const sglBoxBound &box)
{
   // compute sphere when leaving node
   if (box.isEmpty())
   {
      makeEmpty();
   }
   else
   {
      sglVec3f minp(box.getMin());
      sglVec3f maxp(box.getMax());

      m_center.set(0.5*(minp[0] + maxp[0]),
                   0.5*(minp[1] + maxp[1]),
                   0.5*(minp[2] + maxp[2]));

      T dx = m_center[0] - minp[0];
      T dy = m_center[1] - minp[1];
      T dz = m_center[2] - minp[2];

      m_radius = (T)sqrt(dx*dx + dy*dy + dz*dz);

      // just in case box encloses a single point;
      if (m_radius == 0.0) m_radius += FLT_MIN;
   }
}

//----------------------------------------------------------------------------
template <class T>
void sglSphereBound<T>::expandToInclude(const sglSphereBound<T> &sphere)
{
   // infinite spheres include everything
   if (isInfinite())
   {
      return;
   }

   if (sphere.isInfinite())
   {
      m_radius = -1.0;
      return;
   }

   if (!sphere.isEmpty())
   {
      double dx = m_center[0] - sphere.m_center[0];
      double dy = m_center[1] - sphere.m_center[1];
      double dz = m_center[2] - sphere.m_center[2];

      double dist = sqrt(dx*dx + dy*dy + dz*dz);

      // if new sphere contains this one
      if (sphere.m_radius > (dist + m_radius))
      {
         *this = sphere;
      }
      // if this sphere contains the new one
      else if (m_radius > (dist + sphere.m_radius))
      {
         ; // no update required
      }
      // update this sphere if sphere extends outside this radius
      else
      {
         if (dist > 1.0e-6)
         {
            double new_radius = 0.5f*(m_radius + dist + sphere.m_radius);

            double delta = (new_radius - sphere.m_radius);
            m_center[0] = sphere.m_center[0] + delta*dx/dist;
            m_center[1] = sphere.m_center[1] + delta*dy/dist;
            m_center[2] = sphere.m_center[2] + delta*dz/dist;

            m_radius = new_radius*1.0000001;  // roundoff fudge factor
         }
         else  // difference in centers too small to worry about
         {
            if (m_radius > sphere.m_radius)
            {
               m_radius += (float)1.0e-6;
            }
            else
            {
               m_radius = sphere.m_radius + 1.0e-6;
            }
         }
      }
   }
}

//----------------------------------------------------------------------------
template <class T>
inline void sglSphereBound<T>::drawImmediate() const
{
   unsigned int i;
   const unsigned int NUM_POINTS = 8;
   double x[NUM_POINTS], y[NUM_POINTS];

   glBegin(GL_LINE_LOOP);
   for (i=0; i<NUM_POINTS; i++)
   {
      x[i] = cos(i*M_PI*2.0/NUM_POINTS);
      y[i] = sin(i*M_PI*2.0/NUM_POINTS);
      glVertex3f(m_radius*x[i]+m_center[0],
                 m_radius*y[i]+m_center[1],
                 0.f+m_center[2]);
   }
   glEnd();

   glBegin(GL_LINE_LOOP);
   for (i=0; i<NUM_POINTS; i++)
      glVertex3f(m_radius*x[i]+m_center[0],
                 0.f+m_center[1],
                 m_radius*y[i]+m_center[2]);
   glEnd();

   glBegin(GL_LINE_LOOP);
   for (i=0; i<NUM_POINTS; i++)
      glVertex3f(0.f+m_center[0],
                 m_radius*y[i]+m_center[1],
                 m_radius*x[i]+m_center[2]);
   glEnd();
}

template class SGL_DLL_API sglSphereBound<double>;
template class SGL_DLL_API sglSphereBound<float>;
