/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglSphericalTransform.cpp
 *   Author: Scott McMillan
 *  Created: 16 October 2001
 *****************************************************************************/

#include <sglSphericalTransform.hpp>

//----------------------------------------------------------------------------
template <class T>
sglSphericalTransform<T>::sglSphericalTransform()
      : sglTransform<T>(),
        m_mode(eZ_UP),
        m_pos_coi((T)0., (T)0., (T)0.),
        m_delta_pos((T)0, (T)0)
{
   loadIdentity();

   setAzimuth(0.0);
   setElevation(0.0);
   setRadius(0.0);
}

//----------------------------------------------------------------------------
template <class T>
void sglSphericalTransform<T>::reset()
{
   loadIdentity();

   m_pos_coi.set((T)0., (T)0., (T)0.);
   setAzimuth(0.0);
   setElevation(0.0);
   setRadius(0.0);
   m_delta_pos.set((T)0., (T)0.);
}

//----------------------------------------------------------------------------
template <class T>
void sglSphericalTransform<T>::setAzimuth(T new_azimuth)
{
   m_azimuth = new_azimuth;

   // allow it to wrap around
   if (m_azimuth > M_PI)
      m_azimuth -= (T)M_PI*2.0f;
   else if (m_azimuth < -M_PI)
      m_azimuth += (T)M_PI*2.0f;

   m_sin_az = (T)sin(m_azimuth);
   m_cos_az = (T)cos(m_azimuth);
}

//----------------------------------------------------------------------------
template <class T>
void sglSphericalTransform<T>::setElevation(T new_elevation)
{
   m_elevation = new_elevation;

   if (m_elevation > (T)M_PI/2.0)  // looking straight down from overhead
      m_elevation = (T)M_PI/2.0f;
   else if (m_elevation < (T)-M_PI/2.0)   // straight up from underneath
      m_elevation = -(T)M_PI/2.0f;

   m_sin_el = (T)sin(m_elevation);
   m_cos_el = (T)cos(m_elevation);
}

//----------------------------------------------------------------------------
template <class T>
void sglSphericalTransform<T>::setRadius(T new_radius)
{
   m_radius = new_radius;
   if (m_radius < 0.0f)           // don't let it go negative
      m_radius = 0.0f;
}

//----------------------------------------------------------------------------
template <class T>
void sglSphericalTransform<T>::computeMatrix()
{
   sglMat4<T> mat;

   switch (m_mode)
   {
      case eZ_UP:
         mat[0][0] = m_cos_az;
         mat[0][1] = m_sin_az;
         mat[0][2] = 0.0f;
         mat[0][3] = 0.0f;

         mat[1][0] =  m_sin_az*m_sin_el;
         mat[1][1] = -m_cos_az*m_sin_el;
         mat[1][2] =  m_cos_el;
         mat[1][3] = 0.0f;

         mat[2][0] =  m_sin_az*m_cos_el;
         mat[2][1] = -m_cos_az*m_cos_el;
         mat[2][2] = -m_sin_el;
         mat[2][3] = 0.0f;
         break;

      case eY_UP:
         mat[0][0] = m_cos_az;
         mat[0][1] = 0.0f;
         mat[0][2] =-m_sin_az;
         mat[0][3] = 0.0f;

         mat[1][0] = m_sin_az*m_sin_el;
         mat[1][1] = m_cos_el;
         mat[1][2] = m_cos_az*m_sin_el;
         mat[1][3] = 0.0f;

         mat[2][0] = m_sin_az*m_cos_el;
         mat[2][1] =-m_sin_el;
         mat[2][2] = m_cos_az*m_cos_el;
         mat[2][3] = 0.0f;
         break;

      case eX_UP:
         mat[0][0] = 0.0f;
         mat[0][1] =-m_cos_az;
         mat[0][2] =-m_sin_az;
         mat[0][3] = 0.0f;

         mat[1][0] = m_cos_el;
         mat[1][1] =-m_sin_az*m_sin_el;
         mat[1][2] = m_cos_az*m_sin_el;
         mat[1][3] = 0.0f;

         mat[2][0] =-m_sin_el;
         mat[2][1] =-m_sin_az*m_cos_el;
         mat[2][2] = m_cos_az*m_cos_el;
         mat[2][3] = 0.0f;
         break;
   }

   for (unsigned int i=0; i<3; ++i)
   {
      m_pos_coi[i] += m_delta_pos[0]*mat[0][i] + m_delta_pos[1]*mat[1][i];
      mat[3][i] = m_pos_coi[i] + m_radius*mat[2][i];
   }
   mat[3][3] = 1.0f;

   m_delta_pos[0] = m_delta_pos[1] = 0.0;

   setMatrix(mat);
}

//----------------------------------------------------------------------------
template <class T>
sglNode *sglSphericalTransform<T>::clone(unsigned int mode) const
{
   sglSphericalTransform<T> *root = new sglSphericalTransform<T>;
   sglSphericalTransform<T>::copyTo(root, mode);
   return root;
}

//----------------------------------------------------------------------------
template <class T>
void sglSphericalTransform<T>::copyTo(sglSphericalTransform<T> *dst,
                                      unsigned int mode) const
{
   sglTransform<T>::copyTo(dst, mode);

   // transfer stuff
   dst->setMode(m_mode);
   dst->setAzimuth(m_azimuth);
   dst->setElevation(m_elevation);
   dst->setRadius(m_radius);
   dst->setCOI(m_pos_coi);
}

//----------------------------------------------------------------------------
template <class T>
void sglSphericalTransform<T>::printInfo(ostream &ostrm,
                                         const char *indent_spaces) const
{
   sglTransform<T>::printInfo(ostrm, indent_spaces);

   switch (m_mode)
   {
      case eX_UP:
         ostrm << indent_spaces << "mode:      eX_UP" << endl;
         break;
      case eY_UP:
         ostrm << indent_spaces << "mode:      eY_UP" << endl;
         break;
      case eZ_UP:
         ostrm << indent_spaces << "mode:      eZ_UP" << endl;
         break;
   }
   ostrm << indent_spaces << "azimuth:   " << m_azimuth << endl;
   ostrm << indent_spaces << "elevation: " << m_elevation << endl;
   ostrm << indent_spaces << "radius:    " << m_radius << endl;
   ostrm << indent_spaces << "COI pos:   " << m_pos_coi << endl;
}

template class SGL_DLL_API sglSphericalTransform<double>;
template class SGL_DLL_API sglSphericalTransform<float>;
