/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglSphericalTransform.hpp
 *  Created: 15 January 2001 (was sgluSphericalTransform)
 *  Summary: spherical transform subclass of sglTransform
 *****************************************************************************/

#ifndef __SGL_SPHERICAL_TRANSFORM_HPP
#define __SGL_SPHERICAL_TRANSFORM_HPP

#include <sgl.h>
#include <sglTransform.hpp>

/**

This is a convenience subclass of sglTransform that restricted the motion to 3
DOF spherical motion consisting of rotations of azimuth (heading) and
elevation (pitch), and translations of radius.  An additional 3 DOF
translation of the center of interest (COI) can move this "sphere" anywhere in
3-space.  It rotates and translates the coordinate system while always keeping
the z-axis of the new coordinate system pointing directly away from the origin
of the previous system (if OpenGL's view is defined by the new coordinate
system it will always look at the center of interest.

The user can select between three modes where any of the basic axes of the
local coordinate system is considered "up".  The azimuth angle rotates about
this axes and can range from -pi to pi (values outside this range are wrapped
around).  The pitch angle rotates about the "right" vector and can range from
-pi/2 to pi/2.  Values outside this range will be clamped to this range.  A
radius of zero places the transformed coordinate system at the COI and positive
values move it away from this point (a negative radius is clamped to zero).

After any changes to the variables, the user must call
sglSphericalTransform::computeMatrix() to build the new transformation matrix
and store it in the base class.

<b>Example:</b>

<pre>
    sglSphericalTransformf *tx_node = new sglSphericalTransformf;
    tx_node->setMode(sglSphericalTransform::eY_UP); // y up
    tx_node->setCOI(-10.f, 0.f, 0.f);               // move out along -X axis
    tx_node->setAzimuth(M_PI*0.5f);                 // rotate "right"
    tx_node->setElevation(M_PI*0.25f);              // rotate "up"
    tx_node->setRadius(1000.f);                     // move out
    tx_node->computeMatrix();                       // compute tx matrix.

    tx_node->addChild(...);
</pre>

*/

template <class T>
class SGL_DLL_API sglSphericalTransform : public sglTransform<T>
{
public:
   /// Used to specify which vector is "up"
   enum ModeEnum { eX_UP, eY_UP, eZ_UP };

public:
   /// default constructor: initializes all values to zero (equals identity)
   sglSphericalTransform();
   /// virtual destructor
   virtual ~sglSphericalTransform() {}

   /** Set the "up" axis.
       @param mode One of the three ModeEnums (for each axis).  The default is
       eZ_UP.
   */
   void setMode(ModeEnum mode) { m_mode = mode; }

   /** Query the current "up" axis.
       @return The ModeEnum corresponding to the current up axis.
   */
   ModeEnum getMode() const { return m_mode; }

   /// Reset this transform back to identity (all zeroes).
   virtual void reset();

   /** Add a delta to the current azimuth (and wrap around if necessary).
       @param delta_azimuth The angle to add to the current azimuth (radians).
   */
   void updateAzimuth(T delta_azimuth)
      {
         setAzimuth(m_azimuth + delta_azimuth);
      }

   /** Add a delta to the current elevation (and clamp if necessary).
       @param delta_elevation The angle to add to the current elevation
           (radians).
   */
   void updateElevation(T delta_elevation)
      {
         setElevation(m_elevation + delta_elevation);
      }

   /** Add a delta to the current radius (and clamp if necessary).
       @param delta_radius The angle to add to the current radius.
   */
   void updateRadius(T delta_radius)
      {
         setRadius(m_radius + delta_radius);
      }

   /** Add a 3D delta vector to the current center of interest position.
       @param delta_coi The change vector.
   */
   void updateCOI(const sglVec3<T> &delta_coi)
      {
         m_pos_coi += delta_coi;
      }

   /** Add a 3D delta to the current center of interest position.
       @param x X-component of the change vector.
       @param y Y-component of the change vector.
       @param z Z-component of the change vector.
   */
   void updateCOI(T delta_x, T delta_y, T delta_z)
      {
         m_pos_coi += sglVec3<T>(delta_x, delta_y, delta_z);
      }

   /** Query the current azimuth angle.
       @return The current azimuth in radians.
   */
   T getAzimuth() const { return m_azimuth; }

   /** Set the current azimuth angle.
       @param new_azimuth The new azimuth angle in radians.  Values that are
          out of range will be wrapped around.
   */
   void setAzimuth(T new_azimuth);

   /** Query the current elevation angle.
       @return The current elevation in radians.
   */
   T getElevation() const { return m_elevation; }

   /** Set the current elevation angle.
       @param new_elevation The new elevation angle in radians.  Values that
         are out of range will be clamped in the range -pi/2 to pi/2.
   */
   void setElevation(T new_elevation);

   /** Query the current radius.
       @return The current radius.
   */
   T getRadius() const { return m_radius; }

   /** Set the current radius.
       @param new_radius The new radius. Values less than zero will be clamped
         to zero.
   */
   void setRadius(T new_radius);

   /** Query the current center of interest position.
       @return The current COI (as a sglVec3).
   */
   const sglVec3<T> &getCOI() const { return m_pos_coi; }

   /** Set the current COI as an sglVec3.
       @param pos_coi The new COI.
   */
   void setCOI(const sglVec3<T> &pos_coi) { m_pos_coi = pos_coi; }

   /** Set the current COI as three scalars.
       @param x X-component of the new COI.
       @param y Y-component of the new COI.
       @param z Z-component of the new COI.
   */
   void setCOI(T x, T y, T z) { m_pos_coi.set(x, y, z); }

   /** Must be called after any changes so that the new transformation matrix
       is computed.
   */
   void computeMatrix();

   // See sglNode::clone() for details.
   virtual sglNode *clone(unsigned int mode) const;

   // See sglObject::printInfo() for details.
   virtual void printInfo(ostream &ostrm, const char *indent_spaces) const;

private:  // not implemented
   sglSphericalTransform(const sglSphericalTransform&);
   sglSphericalTransform &operator=(const sglSphericalTransform&);

   void copyTo(sglSphericalTransform *dst, unsigned int mode) const;

protected:
   ModeEnum   m_mode;     // defines which is the polar axis (z by default)

   sglVec3<T> m_pos_coi;  // lookat position for center of interest

   T m_radius;      // distance from center of interest
   T m_elevation, m_sin_el, m_cos_el;  // elevation angle in degrees
   T m_azimuth,   m_sin_az, m_cos_az;  // azimuth angle in degrees

   sglVec2<T>   m_delta_pos;  // HACK - set by subclass
};

//----------------------------------------------------------------------------
// convenience types:
//    template class 'should' only be float or double

///
typedef sglSphericalTransform<float>  sglSphericalTransformf;
///
typedef sglSphericalTransform<double> sglSphericalTransformd;

#endif
