/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglSpotLight.hpp
 *   Author: Scott McMillan
 *  Created: 9 December 2000
 *  Summary: a positional light subclass with spot characteristics
 *****************************************************************************/

#ifndef __SGL_SPOT_LIGHT_HPP
#define __SGL_SPOT_LIGHT_HPP

#include <sgl.h>
#include <sglPositionalLight.hpp>

/**

The sglSpotLight class is a subclass of sglPositionalLight that allows the user
to project a cone of light onto the geometry.  In addition to the sglLight and
sglPositionalLight functionality, the direction of the spotlight (relative to
the local coordinate system) can be set with a call to
sglSpotLight::setDirection(); the angle of the cone is set with a call to
sglSpotlight::setCutoff() where the argument is between 0 and 0.5*M_PI radians
(90 degrees); and the intensity is set with a call to
sglSpotLight::setExponent() where the argument must be greater than or equal to
zero.

The following example, configures a spotlight to be positioned along the +Z
axis and shine back along the -Z axis.  The width of the cone is 45 degrees
(actually 90 degrees total), and the exponent (intensity) is set to 3.0:

<b>Example:</b>

<pre>
   sglSpotLight *spot = new sglSpotLight;
   spot->setDiffuse(0.f, 0.f, 1.f, 1.f);        // blue light
   spot->setPosition(sglVec3f(0., 0., 7.0f));   // move out along +Z axis
   spot->setDirection(sglVec3f(0., 0., -1.0f)); // point along -Z axis
   spot->setCutoff(M_PI*0.25f);                 // 45 degrees
   spot->setExponent(3.0f);
</pre>
 */

class SGL_DLL_API sglSpotLight : public sglPositionalLight
{
public:
   /// default constructor
   sglSpotLight();
   /// virtual destuctor
   virtual ~sglSpotLight();

   /** Set the direction of the spot light.  The default is (0, 0, -1).
       @param dir The direction.  It does not need to be normalized, but its
       length must be non-zero.
   */
   void setDirection(const sglVec3f &dir);

   /** Query the current direction of the spot light.
       @return The current direction normalized.
   */
   const sglVec3f &getDirection() const { return m_direction; }

   /** Set the half-width of the spot light
       @param cutoff Angle in radians of half the "spread" of the spot.  Valid
       values fall between 0 and PI/2, inclusive.
   */
   void setCutoff(float cutoff);  // in radians

   /** Query the current cutoff angle.
       @return Cutoff angle
   */
   float getCutoff() const { return m_cutoff; }

   /** Control the concentration of the spot light.
       @param exponent The higher the value the more focused the light.  Only
       non-negative values are valid.  The default value is 0.
   */
   void setExponent(float exponent);

   /** Query the current exponent value.
       @return The current exponent.
   */
   float getExponent() const { return m_exponent; }

   // See sglLight::setup() for details.
   virtual void setup(GLenum gl_light_number) const;

   // See sglNode::clone(..) for details.
   virtual sglNode *clone(unsigned int mode) const;

   // See sglObject::printInfo() for details.
   virtual void printInfo(ostream &ostrm, const char *indent_string) const;

protected:
   void copyTo(sglSpotLight *dst, unsigned int mode) const;

private: // methods - not implemented
   sglSpotLight(const sglLight &);
   sglSpotLight &operator=(const sglSpotLight &);

protected: // variables
   sglVec3f m_direction;
   float    m_exponent;
   float    m_cutoff;
};

#endif
