/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglStatelet.cpp
 *   Author: Tom Stimson
 *  Created: 25 June 2000
 *****************************************************************************/

#include <sgl.h>
#include <sglStatelet.hpp>
#include <sglCurrState.hpp>
#include <string.h>

//----------------------------------------------------------------------------
void sglStatelet::printInfo(ostream &ostrm, const char *indent_string) const
{
   sglObject::printInfo(ostrm, indent_string);
   ostrm << indent_string << "sglStatelet: getType() = " << getType() << endl;
   ostrm << indent_string << "sglStatelet: m_index = " << m_index << endl;
}

//----------------------------------------------------------------------------
void sglBooleanStatelet::printInfo(ostream &ostrm,
                                   const char *indent_string) const
{
   sglStatelet::printInfo(ostrm, indent_string);
   ostrm << indent_string << "sglBooleanStatelet: m_on = "
         << (m_on ? "true" : "false") << endl;
}

//----------------------------------------------------------------------------
void sglAlphaTest::setAlphaFunc(AlphaFuncEnum func, GLclampf ref)
{
   m_alpha_func = func;

   if (ref < 0.0f)
      ref = 0.0f;
   else if (ref > 1.0f)
      ref = 1.0f;

   m_alpha_ref = ref;

   setIndex();
}

//----------------------------------------------------------------------------
void sglAlphaTest::apply(sglCurrState*) const
{
   if (m_alpha_func != eALWAYS)
   {
      glAlphaFunc((GLenum)m_alpha_func, m_alpha_ref);
      glEnable(GL_ALPHA_TEST);
   }
   else
   {
      glDisable(GL_ALPHA_TEST);
   }
}

//----------------------------------------------------------------------------
void sglAlphaTest::printInfo(ostream &ostrm,
                             const char *indent_string) const
{
   sglStatelet::printInfo(ostrm, indent_string);
   switch (m_alpha_func)
   {
      case eLESS:
         ostrm << indent_string << "alpha func: eLESS"; break;
      case eALWAYS:
         ostrm << indent_string << "alpha func: eALWAYS"; break;
      case eLEQUAL:
         ostrm << indent_string << "alpha func: eLEQUAL"; break;
      case eEQUAL:
         ostrm << indent_string << "alpha func: eEQUAL"; break;
      case eGEQUAL:
         ostrm << indent_string << "alpha func: eGEQUAL"; break;
      case eGREATER:
         ostrm << indent_string << "alpha func: eGREATER"; break;
      case eNOTEQUAL:
         ostrm << indent_string << "alpha func: eNOTEQUAL"; break;
      case eNEVER:
         ostrm << indent_string << "alpha func: eNEVER"; break;
      default:
         ostrm << indent_string << "alpha func: unknown"; break;
   }
   ostrm << ", alpha_ref = " << m_alpha_ref << endl;
}

//----------------------------------------------------------------------------
void sglAlphaTest::setIndex()
{
   switch (m_alpha_func)
   {
      case eALWAYS:
         m_index = 0;
         break;
      case eNEVER:
         m_index = 1;
         break;
      default:
         m_index = UINT_MAX;
         break;
   }
}

//----------------------------------------------------------------------------
void sglAntiAliasLine::apply(sglCurrState*) const
{
   // done by sglCurrState
}

//----------------------------------------------------------------------------
void sglAntiAliasPoint::apply(sglCurrState*) const
{
   // done by sglCurrState
}

//----------------------------------------------------------------------------
void sglAntiAliasPolygon::apply(sglCurrState*) const
{
   // done by sglCurrState
}

//----------------------------------------------------------------------------
void sglBlending::apply(sglCurrState *curr_state) const
{
   glBlendFunc((GLenum)m_src_func, (GLenum)m_dst_func);
   curr_state->checkBlend();
}

//----------------------------------------------------------------------------
void sglBlending::printFuncName(ostream &ostrm, BlendFuncEnum func) const
{
   switch (func)
   {
      case eZERO:
         ostrm << "eZERO" << endl;
         break;
      case eONE:
         ostrm << "eONE" << endl;
         break;
      case eSRC_COLOR:
         ostrm << "eSRC_COLOR" << endl;
         break;
      case eONE_MINUS_SRC_COLOR:
         ostrm << "eONE_MINUS_SRC_COLOR" << endl;
         break;
      case eDST_COLOR:
         ostrm << "eDST_COLOR" << endl;
         break;
      case eONE_MINUS_DST_COLOR:
         ostrm << "eONE_MINUS_DST_COLOR" << endl;
         break;
      case eSRC_ALPHA:
         ostrm << "eSRC_ALPHA" << endl;
         break;
      case eONE_MINUS_SRC_ALPHA:
         ostrm << "eONE_MINUS_SRC_ALPHA" << endl;
         break;
      case eDST_ALPHA:
         ostrm << "eDST_ALPHA" << endl;
         break;
      case eONE_MINUS_DST_ALPHA:
         ostrm << "eONE_MINUS_DST_ALPHA" << endl;
         break;
      case eSRC_ALPHA_SATURATE:
         ostrm << "eSRC_ALPHA_SATURATE" << endl;
         break;
   }
}

//----------------------------------------------------------------------------
void sglBlending::printInfo(ostream &ostrm, const char *indent_string) const
{
   sglStatelet::printInfo(ostrm, indent_string);
   ostrm << indent_string << "blending " << (m_on ? "on" : "off") << endl;
   ostrm << indent_string << "blend src func = ";
   printFuncName(ostrm, m_src_func);
   ostrm << indent_string << "blend dst func = ";
   printFuncName(ostrm, m_dst_func);
}

//----------------------------------------------------------------------------
void sglColorMask::setMask(bool red, bool green, bool blue, bool alpha)
{
   m_index = 0;
   if (red)
   {
      m_red = GL_TRUE;
   }
   else
   {
      m_red = GL_FALSE;
      m_index |= 0x1;
   }
   if (green)
   {
      m_green = GL_TRUE;
   }
   else
   {
      m_green = GL_FALSE;
      m_index |= 0x2;
   }
   if (blue)
   {
      m_blue = GL_TRUE;
   }
   else
   {
      m_blue = GL_FALSE;
      m_index |= 0x4;
   }
   if (alpha)
   {
      m_alpha = GL_TRUE;
   }
   else
   {
      m_alpha = GL_FALSE;
      m_index |= 0x8;
   }
}

//----------------------------------------------------------------------------
void sglColorMask::getMask(
   bool &red, bool &green, bool &blue, bool &alpha) const
{
   red = (m_red == GL_TRUE);
   blue = (m_blue == GL_TRUE);
   green = (m_green == GL_TRUE);
   alpha = (m_alpha == GL_TRUE);
}

//----------------------------------------------------------------------------
void sglColorMask::apply(sglCurrState*) const
{
   glColorMask(m_red, m_green, m_blue, m_alpha);
}

//----------------------------------------------------------------------------
void sglColorMask::printInfo(ostream &ostrm, const char *indent_string) const
{
   sglStatelet::printInfo(ostrm, indent_string);
   ostrm << indent_string << "color mask = ("
         << (m_red ? "true" : "false") << ", "
         << (m_green ? "true" : "false") << ", "
         << (m_blue ? "true" : "false") << ", "
         << (m_alpha ? "true" : "false") << ")" << endl;
}

//----------------------------------------------------------------------------
void sglCullMode::apply(sglCurrState*) const
{
   switch (m_cull_mode)
   {
      case eBACK:
         glCullFace(GL_BACK);
         glEnable(GL_CULL_FACE);
         break;
      case eOFF:
         glDisable(GL_CULL_FACE);
         break;
      case eFRONT:
         glCullFace(GL_FRONT);
         glEnable(GL_CULL_FACE);
         break;
      case eFRONT_AND_BACK:
         glCullFace(GL_FRONT_AND_BACK);
         glEnable(GL_CULL_FACE);
         break;
   }
}

//----------------------------------------------------------------------------
void sglCullMode::printInfo(ostream &ostrm, const char *indent_string) const
{
   sglStatelet::printInfo(ostrm, indent_string);
   ostrm << indent_string << "cull mode = ";

   switch (m_cull_mode)
   {
      case eBACK:
         ostrm << "eBACK" << endl;
         break;
      case eOFF:
         ostrm << "eOFF" << endl;
         break;
      case eFRONT:
         ostrm << "eFRONT" << endl;
         break;
      case eFRONT_AND_BACK:
         ostrm << "eFRONT_AND_BACK" << endl;
         break;
   }
}

//----------------------------------------------------------------------------
void sglDepthMask::apply(sglCurrState *curr_state) const
{
   curr_state->checkDepthMask();
}

//----------------------------------------------------------------------------
void sglDepthRange::setDepthRange(double near_val, double far_val)
{
   if (near_val < 0.0) near_val = 0.0;
   else if (near_val > 1.0) near_val = 1.0;

   if (far_val < 0.0) far_val = 0.0;
   else if (far_val > 1.0) far_val = 1.0;

   m_near = near_val;
   m_far  = far_val;

   setIndex();
}

//----------------------------------------------------------------------------
void sglDepthRange::apply(sglCurrState*) const
{
   glDepthRange(m_near, m_far);
}

//----------------------------------------------------------------------------
void sglDepthRange::printInfo(ostream &ostrm,
                              const char *indent_string) const
{
   sglStatelet::printInfo(ostrm, indent_string);
   ostrm << indent_string << "depth range: " << m_near << ", "
         << m_far << endl;
}

//----------------------------------------------------------------------------
void sglDepthTest::apply(sglCurrState*) const
{
   if (m_depth_func != eALWAYS)
   {
      glDepthFunc((GLenum)m_depth_func);
      glEnable(GL_DEPTH_TEST);
   }
   else
   {
      glDisable(GL_DEPTH_TEST);
   }
}

//----------------------------------------------------------------------------
void sglDepthTest::printInfo(ostream &ostrm, const char *indent_string) const
{
   sglStatelet::printInfo(ostrm, indent_string);

   switch (m_depth_func)
   {
      case eLESS:
         ostrm << indent_string << "depth func: eLESS" << endl; break;
      case eALWAYS:
         ostrm << indent_string << "depth func: eALWAYS" << endl; break;
      case eLEQUAL:
         ostrm << indent_string << "depth func: eLEQUAL" << endl; break;
      case eEQUAL:
         ostrm << indent_string << "depth func: eEQUAL" << endl; break;
      case eGEQUAL:
         ostrm << indent_string << "depth func: eGEQUAL" << endl; break;
      case eGREATER:
         ostrm << indent_string << "depth func: eGREATER" << endl; break;
      case eNOTEQUAL:
         ostrm << indent_string << "depth func: eNOTEQUAL" << endl; break;
      case eNEVER:
         ostrm << indent_string << "depth func: eNEVER" << endl; break;
      default:
         ostrm << indent_string << "depth func: unknown" << endl; break;
   }
}

//----------------------------------------------------------------------------
void sglDepthTest::setIndex()
{
   switch (m_depth_func)
   {
      case eLESS:
         m_index = 0; break;
      case eALWAYS:
         m_index = 1; break;
      case eLEQUAL:
         m_index = 2; break;
      case eEQUAL:
         m_index = 3; break;
      case eGEQUAL:
         m_index = 4; break;
      case eGREATER:
         m_index = 5; break;
      case eNOTEQUAL:
         m_index = 6; break;
      case eNEVER:
         m_index = 7; break;
   }
}

//----------------------------------------------------------------------------
void sglDithering::apply(sglCurrState*) const
{
   if (m_on)
   {
      glEnable(GL_DITHER);
   }
   else
   {
      glDisable(GL_DITHER);
   }
}

//----------------------------------------------------------------------------
void sglFrontFace::apply(sglCurrState*) const
{
   glFrontFace((GLenum)m_winding);
}

//----------------------------------------------------------------------------
void sglFrontFace::printInfo(ostream &ostrm, const char *indent_string) const
{
   sglStatelet::printInfo(ostrm, indent_string);

   if (m_winding == eCCW)
      ostrm << indent_string << "front face winding: eCCW" << endl;
   else
      ostrm << indent_string << "front face winding: eCW" << endl;
}

//----------------------------------------------------------------------------
void sglHighlight::apply(sglCurrState*) const
{
   // done by sglCurrState
}

//----------------------------------------------------------------------------
void sglLighting::apply(sglCurrState*) const
{
   if (m_on)
   {
      glEnable(GL_LIGHTING);
   }
   else
   {
      glDisable(GL_LIGHTING);
   }
}

//----------------------------------------------------------------------------
void sglLineStipple::setLineStipple(GLushort pattern, int factor)
{
   m_pattern = pattern;
   if (factor < 1) factor = 1;
   if (factor > 255) factor = 255;
   m_factor = factor;
   setIndex();
}

//----------------------------------------------------------------------------
void sglLineStipple::apply(sglCurrState*) const
{
   if (m_pattern != 0xffff)
   {
      glLineStipple(m_factor, m_pattern);
      glEnable(GL_LINE_STIPPLE);
   }
   else
   {
      glDisable(GL_LINE_STIPPLE);
   }
}

//----------------------------------------------------------------------------
void sglLineStipple::printInfo(ostream &ostrm, const char *indent_string) const
{
   sglStatelet::printInfo(ostrm, indent_string);
   ostrm << indent_string << "factor = " << m_factor
         << ", pattern = 0x" << hex << m_pattern << dec << endl;
}


//----------------------------------------------------------------------------
void sglLineWidth::setLineWidth(float size)
{
   if (size > 0.0f)
      m_line_width = size;
   else
      m_line_width = 1.0f;
   setIndex();
}

//----------------------------------------------------------------------------
void sglLineWidth::apply(sglCurrState*) const
{
   glLineWidth(m_line_width);
}

//----------------------------------------------------------------------------
void sglLineWidth::printInfo(ostream &ostrm, const char *indent_string) const
{
   sglStatelet::printInfo(ostrm, indent_string);
   ostrm << indent_string << "line width = " << m_line_width << endl;
}

//----------------------------------------------------------------------------
void sglLogicOp::apply(sglCurrState*) const
{
   if (m_index)
   {
      glLogicOp((GLenum)m_logic_op);
      glEnable(GL_COLOR_LOGIC_OP);
   }
   else
   {
      glDisable(GL_COLOR_LOGIC_OP);
   }
}

//----------------------------------------------------------------------------
void sglLogicOp::printInfo(ostream &ostrm, const char *indent_string) const
{
   sglStatelet::printInfo(ostrm, indent_string);

   switch (m_logic_op)
   {
      case eCOPY:
         ostrm << indent_string << "logic op: eCOPY" << endl; break;
      case eCLEAR:
         ostrm << indent_string << "logic op: eCLEAR" << endl; break;
      case eSET:
         ostrm << indent_string << "logic op: eSET" << endl; break;
      case eCOPY_INVERTED:
         ostrm << indent_string << "logic op: eCOPY_INVERTED" << endl; break;
      case eNOOP:
         ostrm << indent_string << "logic op: eNOOP" << endl; break;
      case eINVERT:
         ostrm << indent_string << "logic op: eINVERT" << endl; break;
      case eAND:
         ostrm << indent_string << "logic op: eAND" << endl; break;
      case eNAND:
         ostrm << indent_string << "logic op: eNAND" << endl; break;
      case eOR:
         ostrm << indent_string << "logic op: eOR" << endl; break;
      case eNOR:
         ostrm << indent_string << "logic op: eNOR" << endl; break;
      case eXOR:
         ostrm << indent_string << "logic op: eXOR" << endl; break;
      case eEQUIV:
         ostrm << indent_string << "logic op: eEQUIV" << endl; break;
      case eAND_REVERSE:
         ostrm << indent_string << "logic op: eAND_REVERSE" << endl; break;
      case eAND_INVERTED:
         ostrm << indent_string << "logic op: eAND_INVERTED" << endl; break;
      case eOR_REVERSE:
         ostrm << indent_string << "logic op: eOR_REVERSE" << endl; break;
      case eOR_INVERTED:
         ostrm << indent_string << "logic op: eOR_INVERTED" << endl; break;
      default:
         ostrm << indent_string << "logic op: unknown" << endl; break;
   }
}

//----------------------------------------------------------------------------
void sglLogicOp::setIndex()
{
   switch (m_logic_op)
   {
      case eCOPY:
         m_index = 0; break;
      case eCLEAR:
         m_index = 1; break;
      case eSET:
         m_index = 2; break;
      case eCOPY_INVERTED:
         m_index = 3; break;
      case eNOOP:
         m_index = 4; break;
      case eINVERT:
         m_index = 5; break;
      case eAND:
         m_index = 6; break;
      case eNAND:
         m_index = 7; break;
      case eOR:
         m_index = 8; break;
      case eNOR:
         m_index = 9; break;
      case eXOR:
         m_index = 10; break;
      case eEQUIV:
         m_index = 11; break;
      case eAND_REVERSE:
         m_index = 12; break;
      case eAND_INVERTED:
         m_index = 13; break;
      case eOR_REVERSE:
         m_index = 14; break;
      case eOR_INVERTED:
         m_index = 15; break;
   }
}

//----------------------------------------------------------------------------
void sglOcclusionCull::apply(sglCurrState*) const
{
   // done by sglCull
}

//----------------------------------------------------------------------------
void sglOcclusionCull::printInfo(ostream &ostrm,
                                 const char *indent_string) const
{
   sglStatelet::printInfo(ostrm, indent_string);
   ostrm << indent_string << "sglOcclusionCull: m_on = "
         << (m_on ? "true" : "false") << endl;
}

//----------------------------------------------------------------------------
void sglPointSize::setPointSize(float size)
{
   if (size > 0.0f)
      m_point_size = size;
   else
      m_point_size = 1.0f;
   setIndex();
}

//----------------------------------------------------------------------------
void sglPointSize::apply(sglCurrState*) const
{
   glPointSize(m_point_size);
}

//----------------------------------------------------------------------------
void sglPointSize::printInfo(ostream &ostrm, const char *indent_string) const
{
   sglStatelet::printInfo(ostrm, indent_string);
   ostrm << indent_string << "point size = " << m_point_size << endl;
}

//----------------------------------------------------------------------------
void sglPointParameters::apply( sglCurrState* ) const
{
   if( sgl::s_point_parameters_capable )
   {
      float v[3];

      v[0] = m_dist_atten_A;
      v[1] = m_dist_atten_B;
      v[2] = m_dist_atten_C;

      sgl::pointParameterf( GL_POINT_SIZE_MIN_EXT, m_min_size );
      sgl::pointParameterf( GL_POINT_SIZE_MAX_EXT, m_max_size );

      sgl::pointParameterf( GL_POINT_FADE_THRESHOLD_SIZE_EXT, m_fade_threshold);

      sgl::pointParameterfv( GL_DISTANCE_ATTENUATION_EXT, v );
   }
}

//----------------------------------------------------------------------------
void sglPointParameters::printInfo(ostream &ostrm, const char *indent_string)
                                                                          const
{
   sglStatelet::printInfo(ostrm, indent_string);
   ostrm << indent_string << "point param A = " << m_dist_atten_A << endl;
   ostrm << indent_string << "point param B = " << m_dist_atten_B << endl;
   ostrm << indent_string << "point param C = " << m_dist_atten_C << endl;
   ostrm << indent_string << "fade threshold = " << m_fade_threshold << endl;
   ostrm << indent_string << "point min size = " << m_min_size << endl;
   ostrm << indent_string << "point max size = " << m_max_size << endl;
}

//----------------------------------------------------------------------------
void sglPolyStyle::printInfo(ostream &ostrm, const char *indent_string) const
{
   sglStatelet::printInfo(ostrm, indent_string);
   ostrm << indent_string << "style = ";
   switch (m_style)
   {
      case eFILLED:
         ostrm << "eFILLED" << endl;
         break;
      case eLINE:
         ostrm << "eLINE" << endl;
         break;
      case ePOINT:
         ostrm << "ePOINT" << endl;
         break;
   }
}

//----------------------------------------------------------------------------
void sglPolyBackStyle::apply(sglCurrState*) const
{
   switch (m_style)
   {
      case eFILLED:
         glPolygonMode(GL_BACK, GL_FILL);
         break;
      case eLINE:
         glPolygonMode(GL_BACK, GL_LINE);
         break;
      case ePOINT:
         glPolygonMode(GL_BACK, GL_POINT);
         break;
   }
}

//----------------------------------------------------------------------------
void sglPolyFrontStyle::apply(sglCurrState*) const
{
   switch (m_style)
   {
      case eFILLED:
         glPolygonMode(GL_FRONT, GL_FILL);
         break;
      case eLINE:
         glPolygonMode(GL_FRONT, GL_LINE);
         break;
      case ePOINT:
         glPolygonMode(GL_FRONT, GL_POINT);
         break;
   }
}

//----------------------------------------------------------------------------
void sglPolygonOffset::setPolygonOffset(float  factor, float  units)
{
   m_factor = factor;
   m_units  = units;
   m_polygon_offset_flag = (factor != 0.0f ||
                            units  != 0.0f);
   setIndex();
}

//----------------------------------------------------------------------------
void sglPolygonOffset::apply(sglCurrState*) const
{
   if (m_polygon_offset_flag)
   {
      glPolygonOffset(m_factor, m_units);
      glEnable(GL_POLYGON_OFFSET_POINT);
      glEnable(GL_POLYGON_OFFSET_LINE);
      glEnable(GL_POLYGON_OFFSET_FILL);
   }
   else
   {
      glDisable(GL_POLYGON_OFFSET_POINT);
      glDisable(GL_POLYGON_OFFSET_LINE);
      glDisable(GL_POLYGON_OFFSET_FILL);
   }
}

//----------------------------------------------------------------------------
void sglPolygonOffset::printInfo(ostream &ostrm,
                                 const char *indent_string) const
{
   sglStatelet::printInfo(ostrm, indent_string);
   if (!m_polygon_offset_flag)
      ostrm << indent_string << "polygon offset: disabled" << endl;
   else
   {
      ostrm << indent_string << "polygon offset: factor = " << m_factor
            << ", units = " << m_units << endl;
   }
}

//----------------------------------------------------------------------------
void sglPolygonStipple::setPolygonStipple(const GLubyte *pattern)
{
   if (pattern)
   {
      memcpy(m_pattern, pattern, 128);
      m_polygon_stipple_flag = true;
   }
   else
      m_polygon_stipple_flag = false;

   setIndex();
}

//----------------------------------------------------------------------------
void sglPolygonStipple::apply(sglCurrState*) const
{
   if (m_polygon_stipple_flag)
   {
      glPixelStorei(GL_UNPACK_ALIGNMENT, 4);  // 1 is equally valid
      glPixelStorei(GL_UNPACK_ROW_LENGTH, 0);
      glPixelStorei(GL_UNPACK_SKIP_ROWS, 0);
      glPixelStorei(GL_UNPACK_SKIP_PIXELS, 0);

      glPolygonStipple(m_pattern);
      glEnable(GL_POLYGON_STIPPLE);
   }
   else
   {
      glDisable(GL_POLYGON_STIPPLE);
   }
}

//----------------------------------------------------------------------------
void sglPolygonStipple::printInfo(ostream &ostrm,
                                  const char *indent_string) const
{
   sglStatelet::printInfo(ostrm, indent_string);
   if (!m_polygon_stipple_flag)
      ostrm << indent_string << "polygon stipple: disabled" << endl;
   else
   {
      ostrm << indent_string << "polygon stipple: pattern = " << endl;
      const GLubyte *ptr = m_pattern;
      for (unsigned int iy=0; iy<32; ++iy)
      {
         ostrm << indent_string << "   ";
         for (unsigned int ix=0; ix<4; ++ix)
         {
            if (*ptr < 16)
               ostrm << "0x0" << hex << (short)*ptr++ << " ";
            else
               ostrm << "0x"  << hex << (short)*ptr++ << " ";
         }
         ostrm << endl;
      }
   }
}

//----------------------------------------------------------------------------
void sglShadeModel::apply(sglCurrState*) const
{
   glShadeModel((GLenum)m_model);
}

//----------------------------------------------------------------------------
void sglShadeModel::printInfo(ostream &ostrm,
                              const char *indent_string) const
{
   sglStatelet::printInfo(ostrm, indent_string);

   if (m_model == eFLAT)
      ostrm << indent_string << "shade model: eFLAT" << endl;
   else
      ostrm << indent_string << "shade model: eSMOOTH" << endl;
}

//----------------------------------------------------------------------------
sglMultiTextureStatelet::~sglMultiTextureStatelet()
{
   if (m_statelet)
   {
      m_statelet->deref();
      if (m_statelet->getRefCount() == 0)
      {
         delete m_statelet;
      }
   }
}

//----------------------------------------------------------------------------
void sglMultiTextureStatelet::setStatelet(sglTexturingStatelet *statelet)
{
   if (m_statelet)
      m_statelet->deref();
   if (statelet)
      statelet->ref();
   m_statelet = statelet;
}

//----------------------------------------------------------------------------
void sglMultiTextureStatelet::apply(sglCurrState *curr_state) const
{
   m_statelet->applyToUnit(curr_state, m_texture_unit);
}

//----------------------------------------------------------------------------
unsigned int sglMultiTextureStatelet::getType() const
{
   if (m_statelet && m_texture_unit < (unsigned int)sgl::s_num_texture_units)
   {
      // HACK HACK HACK This could be called by multiple threads at the
      // same time.  We know that getType() is called before getIndex, and
      // are assuming that writing the same value to an address will not
      // screw up a reader.
      m_index = m_statelet->getIndex();
      return m_statelet->getType() + (m_texture_unit * 4);
   }
   else
      return UINT_MAX;
}

//----------------------------------------------------------------------------
void sglTextureMatrix::setTextureMatrix(sglMat4f *tmat)
{
   if (tmat == NULL)
   {
      m_texture_matrix_flag = false;
      m_texture_matrix.buildIdentity();
   }
   else
   {
      m_texture_matrix_flag = !tmat->almostEqual(sglIdentMat4f, sglFltEpsilon);
      m_texture_matrix = *tmat;
   }

   setIndex();
}

//----------------------------------------------------------------------------
void sglTextureMatrix::applyToUnit(sglCurrState*, unsigned int unit) const
{
   if (sgl::s_num_texture_units > 1)
   {
      sgl::setActiveTextureUnit(GL_TEXTURE0 + unit);
   }

   if (m_texture_matrix_flag)
   {
      glMatrixMode(GL_TEXTURE);
      sglLoadMatrix(m_texture_matrix);
      glMatrixMode(GL_MODELVIEW);
   }
   else
   {
      glMatrixMode(GL_TEXTURE);
      glLoadIdentity();
      glMatrixMode(GL_MODELVIEW);
   }
}

//----------------------------------------------------------------------------
void sglTextureMatrix::printInfo(ostream &ostrm,
                                 const char *indent_string) const
{
   sglStatelet::printInfo(ostrm, indent_string);
   ostrm << indent_string << "matrix = " << m_texture_matrix << endl;
}

//----------------------------------------------------------------------------
void sglTransparency::apply(sglCurrState *curr_state) const
{
   curr_state->checkBlend();
   curr_state->checkDepthMask();
}
