/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglStatelet.hpp
 *   Author: Tom Stimson
 *  Created: 25 June 2000
 *  Summary: elements of OpenGL state that are managed by SGL
 *****************************************************************************/

#ifndef __SGL_STATELET_HPP
#define __SGL_STATELET_HPP

#include <sgl.h>
#include <sglObject.hpp>
#include <sglMatrix.hpp>

class sglCurrState;

//============================================================================

/**

This is the base class for all "elements" of OpenGL state that SGL
manages/sorts.  Examples of subclasses include sglTexture2D, sglMaterial,
sglFog, sglLineAntiAlias, etc.  An entire set of sglStatelet subclasses
determine the "state" of any given sglDrawable.  This "state" is determined
from a number of different sources in a specific priority order as follows:

-# The sglScene::createDefaultStatelets() creates a complete set of "baseline"
statelets using the default constructor of each.  These are generated at the
beginning of the cull traversal and are the lowest priority.
<p>
-# Stateletes contained in any sglDefaultStateNode's encountered in the scene
graph traversal overwrite these values in the order they are encountered;
therefore, sglDefaultStateNode's lower in the tree (farther from the root) take
precedence over those higher in the tree.
<p>
-# Statelets encountered in the sglDrawable (using the sglDrawable::setState()
function) overwrite the full state last and thus take precedence over all the
default states and the sglDefaultStateNode's.
<p>
-# Statelets in sglOverrideStateNode's have priority over all the default
and drawable statelets.  Statelets in sglOverrideStateNode's higher in the
scene graph (closer to the root) have priority over those in nodes lower in the
scene graph (opposite that of sglDefaultStateNode).
<p>
-# Statelets in the override_statelets parameter passed
into the sglScene::cull() functions when starting the cull traversal have the
highest priority.

Each element of OpenGL state has been implemented in a separate class to also
aid in state sorting.  State sorting occurs as the cull traversal encounters
sglDrawables that have passed the cull test and before they are rendered.  By
performing state sorting, sglDrawables with statelets that are the same can be
grouped together to avoid issuing duplicate OpenGL state commands.  For
example, two sglDrawables that use the same texture can be grouped together,
and a single set of OpenGL texture commands will be issued to the rendering
pipeline before both drawables are rendered.  This reduction in OpenGL state
commands can improve the performance of an application.  The statelets can be
sorted by SGL in any user defined order by passing a vector of
sglStatelet::StateEnum's into sglCull::setSortOrder().  For instance, if you
want sort the nodes to be rendered by texture first and then within that group
by material the following code might be used:

<pre>
   sglCullf trav_state;
     :
   vector<sglStatelet::StateEnum> sort_order;
   sort_order.push_back(sglStatelet::eTEXTURE);
   sort_order.push_back(sglStatelet::eMATERIAL);

   trav_state.setSortOrder(sort_order);
</pre>

Note, there is no "best" sort order. The best sort order is dependent on
the graphics hardware/driver and the contents of your scene.  See the
testTraversal program for examples of two scene graphs that work best with
completely different sort orders.

Arbitrarily long sort orders can be specified but longer sort orders require
more computation and quickly suffer from diminishing returns after the first
two or three.  Also note that some sorts are more expensive than others.
Obviously, the more attributes that you sort on the more time it is going to
take.  You should also be aware that statelets that can take on an unlimited
number of values (e.g. material, texture, etc) have a worst case O(n*logn) sort
time.  Statelets with a fixed number of values (e.g., any boolean statelets,
polygon style, etc.) have a worst case O(n) sort time.


<b>Implementation Notes:</b>

This base class implements a number of the functions that the internal state
sorting mechanism uses: getSorted, getSortKey, and getIndex.  Each subclass is
responsible for setting the m_index member variable in the sglStatelet base
class.  If the statelet can only take on a few distinct modes (e.g. disable or
enable line anti-aliasing) then m_index should be set to a unique small
unsigned integer corresponding to each unique mode.  This value should start at
zero and cover a contiguous set of values.  If the statelet can take on a
nearly infinite number of modes (e.g. RGB floating point color value), then
m_index should be set to UINT_MAX.  The state sorting algorithm calls the
getIndex() member function to get this index, and if it is set to UINT_MAX it
will sort this particular statelet by its pointer rather than the index number.

The default implementation of getSorted() is to return false indicating that
the statelet should be sorted by using m_index to index into an array of
possible statelets (which take on the state corresponding to that index).  If
m_index can be set to UINT_MAX, then that subclass must override the
getSorted() function to return true.  In this case the state sorting mechanism
will call the getSortKey() function and sort on its return value.

Note that a given subclass can set m_index to small unsigned integers for a
subset of its more common states and UINT_MAX for other states (see
sglAlphaTest for an example).  In this case the getSorted() function should
return true and the getSortKey() will return a pointer to the statelet if
m_index is UINT_MAX or a pseudo-pointer equal to m_index otherwise to make it
appear that statelets of this type with the same m_index will be treated as the
same state for sorting purposes.


<b>Subclassing sglStatelet:</b>

The following is the procedure for creating new subclasses of sglStatelet:

-# Add an enum to the sglStatelet::StateEnum list corresponding to the statelet
created.  The naming convention for the enum: if the class is sglCullMode the
enum label is eCULL_MODE (an 'e' prefix, and underscore separated words).  This
can be added to the enum list anywhere before the texture related statelets.
<p>
-# Implement the default constructor that builds the statelet with the default
mode for that portion of OpenGL state.  Generally this constructor should
initialize the object so that apply() will issue OpenGL calls equal to OpenGL's
default for them.  This and all other constructors must call the sglStatelet
constructor in the initialization list with the correct StateEnum as the
parameter.
<p>
-# Add code to set sglStatelet::m_index variable appropriately (see discussion
above).
<p>
-# Implement the virtual apply() method with all of the OpenGL commands needed
to set up the state.
<p>
-# Implement a printInfo() function that calls the base class printInfo()
function and then outputs its own state.
<p>
-# If necessary, implement the getSorted() virtual function to return true
(only if m_index can be set to UINT_MAX).
<p>
-# Modify sglScene::createDefaultStatelets to push a heap allocated copy of the
statelet (using the default constructor) onto the the full list of statelets.
IMPORTANT: it must be added to the vector in the order they appear in the
sglStatelet::StateEnum list below.

\todo Add sglScribe statelet (a la Performer) where wire frame is rendered on
top of the filled geometry.

\todo Add sglNormalsDisplay statelet (a la Performer) where normals are
rendered along with the geometry.

 */

class SGL_DLL_API sglStatelet : public sglObject
{
public:
   /** The complete list of state enums in alphabetical order with the
       exception that all texture/multitexture states must come last.  These
       enum values will be used as indexes into arrays of statelets
       (sglFullState).
   */
   enum StateEnum
   {
      eALPHA_TEST,
      eANTI_ALIAS_LINE,
      eANTI_ALIAS_POINT,
      eANTI_ALIAS_POLYGON,
      eBLENDING,
      eCOLOR_MASK,
      eCULL_MODE,
      eDEPTH_MASK,
      eDEPTH_RANGE,
      eDEPTH_TEST,
      eDITHERING,
      eFOG,                    // see sglFog.hpp
      eFRONT_FACE,
      eHIGHLIGHT,
      eLIGHTING,
      eLINE_STIPPLE,
      eLINE_WIDTH,
      eLOGIC_OP,
      eMATERIAL,               // see sglMaterial.hpp
      eOCCLUSION_CULL,
      ePOINT_SIZE,
      ePOINT_PARAMETERS,
      ePOLY_BACK_STYLE,
      ePOLY_FRONT_STYLE,
      ePOLYGON_OFFSET,
      ePOLYGON_STIPPLE,
      eSHADE_MODEL,
      eTRANSFORM_MATRIX,       // see sglCull.hpp
      eTRANSPARENCY,

      // DO NOT ADD ANY NEW STATLETS BEYOND THIS POINT.

      eTEX_ENV,                // see sglTexEnv.hpp
      eTEX_GEN,                // see sglTexGen.hpp
      eTEXTURE,                // see sglTexture.hpp
      eTEXTURE_MATRIX,
      eNUM_BASIC_STATES

      // eTEX_ENV_UNIT_2
      // eTEX_GEN_UNIT_2
      // eTEXTURE_UNIT_2
      // eTEXTURE_MATRIX_UNIT_2
      // eTEX_ENV_UNIT_3
      // eTEX_GEN_UNIT_3
      // eTEXTURE_UNIT_3
      // eTEXTURE_MATRIX_UNIT_3
      // etc...
   };

public:
   /** virtual destructor */
   virtual ~sglStatelet() {}

   /** Get an identifier for the type of this statelet.
       @return The StateEnum corresponding to the type of statelet.  For
          texturing statelets beyond the first texture unit integers beyond the
          last StateEnum value are returned.
   */
   virtual unsigned int getType() const { return m_type; }

   /** Apply this state change to the current rendering state.
       @param curr_state Required for statelets that affect one another.  For
         example the sglTransparency sets up depth mask and blending modes
         which interact with the sglBlending and sglDepthMask statelets, and
         the curr_state object is used to resolve these internal issues.
   */
   virtual void apply(sglCurrState* curr_state) const = 0;

   /** Query the sorting mode for this statelet.
       @return true if this type of statelet should be sorted by pointer
         rather than using m_index to index into an array.  The defualt
         implementation is the latter, but if a statelet can take on an
         inordinate number of states (e.g. a floating point parameter) then
         true should be returned.
   */
   virtual bool getSorted() const { return false; }

   /** If getSorted() returns true, then this function is used to retrieve
       the key (pointer) to sort with.  This function is non-virtual so that
       access to it can be inlined in the cull traversal for speed.
       @return Pointer to the statelet or a pseudo pointer equal to m_index.
   */
   inline const void* getSortKey() const
      {
         if (m_index == UINT_MAX)
            return this;
         else
            return (const void*)m_index;
      }

   /** If getSorted() returns false, then this function is used to retrieve
       an index that can be used as a lookup key in a table for state sorting.
       This function is non-virtual so that access to it can be inlined in the
       cull traversal for speed.
       @return Index corresponding to the state of the statelet.
   */
   unsigned int getIndex() const { return m_index; }

   /** The equality operator
       @param rhs The object to which to compare this statelet.
       @return true if the rhs is the same as this.
   */
   bool operator==(const sglStatelet &rhs) const
      {
         return (this == &rhs) || (m_type == rhs.m_type &&
                                   m_index == rhs.m_index &&
                                   m_index != UINT_MAX);
      }

   // See sglObject::printInfo() for details.
   virtual void printInfo(ostream &ostrm, const char *indent_string) const;

protected:
   // constructor
   sglStatelet(StateEnum type)
      : m_index(0), m_type(type) {}

private:
   sglStatelet();  // not implemented
   sglStatelet(const sglStatelet &); // not implemented
   sglStatelet &operator=(const sglStatelet &); // not implemented

protected:
   // If getSorted() == false
   //    m_index is an array index where every index fully defines that state
   // If getSorted() == true
   //    0:        default state; sort key = 0
   //    UINT_MAX: sort key = this pointer
   mutable unsigned int m_index;   // should be set by subclass

private:
   StateEnum    m_type;
};


//============================================================================

/**

This is the abstract base class of all sglStatelets that use OpenGL's
glEnable()/glDisable() calls.  The boolean passed into the enable() function
determines if the statelet in question is enabled (true) or disabled (false).
Subclasses must still implement the apply() (that issue the actual OpenGL
command) and printInfo() member functions.

 */

class SGL_DLL_API sglBooleanStatelet : public sglStatelet
{
public:
   /** Constructor.
       @param type The StateEnum type passed in by the subclass constructors.
       @param on The initial value of the boolean.
   */
   sglBooleanStatelet(StateEnum type, bool on)
      : sglStatelet(type), m_on(on) { setIndex(); }

   /** virtual desctructor. */
   virtual ~sglBooleanStatelet() {}

   /** Enable or disable this statelet
       @param on True to enable, or false to disable.
   */
   void enable(bool on) { m_on = on; setIndex(); }

   /** Query the current state of this statelet.
       @return true if enabled, or false if disabled.
   */
   bool isEnabled() const { return m_on; }

   // See sglObject::printInfo() for details.
   virtual void printInfo(ostream &ostrm, const char *indent_string) const;

private:
   void setIndex() { m_index = (m_on ? 1 : 0); }

   sglBooleanStatelet(); // not implemented
   sglBooleanStatelet(const sglBooleanStatelet &); // not implemented
   sglBooleanStatelet &operator=(const sglBooleanStatelet &);// not implemented

protected:
   bool m_on;
};


//============================================================================

/**

Statelet to control the alpha function test (glAlphaFunc), and the enable or
disable [glEnable(GL_ALPHA_TEST)].  Constructor arguments allow you to
enable/disable the test, choose the comparison test (e.g. GL_GREATER), and set
the reference value.  These can also be set with the setAlphaFunc(..) member
functions after construction.  The alpha function is disabled when the
AlphaFuncEnum equal to sglAlphaTest::eALWAYS is chosen.

<b>Example:</b>
<pre>
   sglAlphaTest *statelet = new sglAlphaTest;   // default eALWAYS
   statelet->setAlphaFunc(sglAlphaTest::eLEQUAL, 0.5f);
</pre>

 */

class SGL_DLL_API sglAlphaTest : public sglStatelet
{
public:
   /** The list of possible alpha functions. */
   enum AlphaFuncEnum
   {
      eLESS     = GL_LESS,
      eALWAYS   = GL_ALWAYS,    // turns alpha test off
      eLEQUAL   = GL_LEQUAL,
      eEQUAL    = GL_EQUAL,
      eGEQUAL   = GL_GEQUAL,
      eGREATER  = GL_GREATER,
      eNOTEQUAL = GL_NOTEQUAL,
      eNEVER    = GL_NEVER      // what's the point of this?
   };

public:
   /** Constructor (also default constructor).
       @param func One of the AlphaFuncEnum values.  Defaults to eALWAYS.
       @param ref The floating point reference value for comparison.  It is
          clamped to the range [0..1], and defaults to 0.
   */
   sglAlphaTest(AlphaFuncEnum func = eALWAYS, float ref = 0.0f)
         : sglStatelet(eALPHA_TEST), m_alpha_func(eALWAYS), m_alpha_ref(0.f)
      {
         setAlphaFunc(func, ref);
      }

   /// virtual destructor
   virtual ~sglAlphaTest() {}

   /** Set the alpha test function and reference value.
       @param func One of the AlphaFuncEnum values.
       @param ref The floating point reference value for comparison.  It is
          clamped to the range [0..1].
   */
   void setAlphaFunc(AlphaFuncEnum  func, float  ref);

   /** Query the current alpha test function and reference value.
       @param func On return contains one of the enums corresponding to the
         test.
       @param ref On return contains the current reference value from zero to
         one.
   */
   void getAlphaFunc(AlphaFuncEnum &func, float &ref) const
      { func = m_alpha_func;  ref = m_alpha_ref; }

   // See sglStatelet::apply() for details.
   void apply(sglCurrState *curr_state) const;

   // See sglStatelet::getSorted() for details.
   virtual bool getSorted() const { return true; }

   // See sglObject::printInfo() for details.
   virtual void printInfo(ostream &ostrm, const char *indent_string) const;

private:
   sglAlphaTest(const sglAlphaTest &); // not implemented
   sglAlphaTest &operator=(const sglAlphaTest &); // not implemented

   void setIndex();

private:
   AlphaFuncEnum  m_alpha_func;
   float          m_alpha_ref;
};


//============================================================================

/**

Enables/disables the OpenGL state needed to anti-alias lines.

<b>Example:</b>
<pre>
   sglAntiAliasLine *statelet = new sglAntiAliasLine(true);  // enable it
   statelet->enable(false);                                  // disable it
</pre>
 */

class SGL_DLL_API sglAntiAliasLine : public sglBooleanStatelet
{
public:
   /** Constructor (also default constructor)
       @param on Enables line anti-aliasing if true, otherwise it is disabled
                 (the default).
   */
   sglAntiAliasLine(bool on = false)
      : sglBooleanStatelet(eANTI_ALIAS_LINE, on) {}

   /// virtual destructor.
   virtual ~sglAntiAliasLine() {}

   // See sglStatelet::apply() for details.
   void apply(sglCurrState *curr_state) const;

private:
   sglAntiAliasLine(const sglAntiAliasLine &); // not implemented
   sglAntiAliasLine &operator=(const sglAntiAliasLine &); // not implemented
};


//============================================================================

/**

Enables/disables the OpenGL state needed to anti-alias points.

<b>Example:</b>
<pre>
   sglAntiAliasPoint *statelet = new sglAntiAliasPoint(true);  // enable it
   statelet->enable(false);                                    // disable it
</pre>
 */

class SGL_DLL_API sglAntiAliasPoint : public sglBooleanStatelet
{
public:
   /** Constructor (also default constructor).
       @param on Enables point anti-aliasing if true, otherwise it disables
                 (the default).
   */
   sglAntiAliasPoint(bool on = false)
      : sglBooleanStatelet(eANTI_ALIAS_POINT, on) {}

   /// virtual destructor
   virtual ~sglAntiAliasPoint() {}

   // See sglStatelet::apply() for details.
   void apply(sglCurrState *curr_state) const;

private:
   sglAntiAliasPoint(const sglAntiAliasPoint &); // not implemented
   sglAntiAliasPoint &operator=(const sglAntiAliasPoint &); // not implemented
};


//============================================================================

/**

Enables/disables the OpenGL state needed to anti-alias polygons.  The method
of polygon antialiasing is determined by the capabilities of the version of
OpenGL (either true polygon anti aliasing or the two pass "trick" where
wireframe is drawn in the second pass with anti-aliased lines).

<b>Example:</b>
<pre>
   sglAntiAliasPoint *statelet = new sglAntiAliasPoint(true);  // enable it
   statelet->enable(false);                                    // disable it
</pre>
 */

class SGL_DLL_API sglAntiAliasPolygon : public sglBooleanStatelet
{
public:
   /** Constructor (also default constructor).
       @param on Enables polygon anti-aliasing if true, otherwise it disables
         (the default).
   */
   sglAntiAliasPolygon(bool on = false)
      : sglBooleanStatelet(eANTI_ALIAS_POLYGON, on) {}

   /// virtual destructor
   virtual ~sglAntiAliasPolygon() {}

   // See sglStatelet::apply() for details.
   void apply(sglCurrState *curr_state) const;

private:
   sglAntiAliasPolygon(const sglAntiAliasPolygon &); // not implemented
   sglAntiAliasPolygon &operator=(const sglAntiAliasPolygon &); // not impl'ed
};

//============================================================================

/**

This statelet controls the OpenGL blending state by issuing the glBlendFunc()
function and the glEnable(GL_BLEND_FUNC).  Even when this statelet is
disabled by calling sglBlending::enable(false), the blend function defined
here is used by the sglTransparency and most of the sglAntiAlias* statelets to
perform their blending.  Because of this the default blending function is
SRC_ALPHA, ONE_MINUS_SRC_ALPHA.  If this blending mode is not desired for
some of these other statelets then an sglBlending statelet can be included
which defines the desired blend function (without ever enabling it).

<b>Example:</b>
<pre>
   sglBlending *blending = new sglBlending;  // defaults to disabled
   blending->setBlendFunc(sglBlending::eSRC_ALPHA,
                          sglBlending::eONE_MINUS_SRC_ALPHA);
   blending->enable(true);
</pre>

\todo Add support for GL_EXT_blend_color.

 */

class SGL_DLL_API sglBlending : public sglStatelet
{
public:
   /** The list of possible blend functions. */
   enum BlendFuncEnum
   {
      eZERO                = GL_ZERO,
      eONE                 = GL_ONE,
      eSRC_COLOR           = GL_SRC_COLOR,
      eONE_MINUS_SRC_COLOR = GL_ONE_MINUS_SRC_COLOR,
      eDST_COLOR           = GL_DST_COLOR,
      eONE_MINUS_DST_COLOR = GL_ONE_MINUS_DST_COLOR,
      eSRC_ALPHA           = GL_SRC_ALPHA,
      eONE_MINUS_SRC_ALPHA = GL_ONE_MINUS_SRC_ALPHA,
      eDST_ALPHA           = GL_DST_ALPHA,
      eONE_MINUS_DST_ALPHA = GL_ONE_MINUS_DST_ALPHA,
      eSRC_ALPHA_SATURATE  = GL_SRC_ALPHA_SATURATE
   };

public:
   /** Constructor (also default constructor).
       @param src_func Enum indicating how the source values are combined in the
         blend function (default is sglBlending::eSRC_ALPHA).
       @param dst_func Enum indicating how the destination values are combined
         in the blend function (default is sglBlending::eONE_MINUS_SRC_ALPHA).
       @param on True to enable blending otherwise false (default is false).
   */
   sglBlending(BlendFuncEnum src_func = eSRC_ALPHA,
               BlendFuncEnum dst_func = eONE_MINUS_SRC_ALPHA,
               bool on = false)
         : sglStatelet(eBLENDING),
           m_src_func(src_func), m_dst_func(dst_func), m_on(on)
      { setIndex(); }

   /// virtual destructor
   virtual ~sglBlending() {}

   /** Enable or disable blending.
       @param on True to enable and false to disable.
   */
   void enable(bool on) { m_on = on; setIndex(); }

   /** Query whether or not blending is enabled.
       @return True if enabled and false if disabled.
   */
   bool isEnabled() const  { return m_on; }

   /** Set the blend function factors for source and destination values.
       @param src_func The factor used on the source values (one of the
         BlendFuncEnum tokens).
       @param dst_func The factor used on the destination values.
   */
   void setBlendFunc(BlendFuncEnum src_func, BlendFuncEnum dst_func)
      { m_src_func = src_func; m_dst_func = dst_func; setIndex(); }

   /** Query the current blend function factors.
       @param src_func On return will contain the factor enum for source.
       @param dst_func On return will contain the factor enum for destination.
   */
   void getBlendFunc(BlendFuncEnum &src_func, BlendFuncEnum &dst_func) const
      { src_func = m_src_func; dst_func = m_dst_func; }

   // See sglStatelet::apply() for details.
   void apply(sglCurrState *curr_state) const;

   // See sglStatelet::getSorted() for details.
   virtual bool getSorted() const { return true; }

   // See sglObject::printInfo() for details.
   virtual void printInfo(ostream &ostrm, const char *indent_string) const;

private:
   sglBlending(const sglBlending &); // not implemented
   sglBlending &operator=(const sglBlending &); // not implemented

   void setIndex()
      {
         m_index = ((m_on == false && m_src_func == eSRC_ALPHA &&
                     m_dst_func == eONE_MINUS_SRC_ALPHA) ? 0 : UINT_MAX);
      }

   void printFuncName(ostream &ostrm, BlendFuncEnum func) const;

private:
   BlendFuncEnum m_src_func;
   BlendFuncEnum m_dst_func;
   bool          m_on;
};


//============================================================================

/**

This statelet encapsulates OpenGL's glColorMask() function which controls which
color components actually get written into the color buffer.

<b>Example:</b>
<pre>
   sglColorMask *cmask = new sglColorMask;     // all enabled by default
   cmask->setMask(true, false, false, false);  // only enable red components
</pre>

 */

class SGL_DLL_API sglColorMask : public sglStatelet
{
public:
   /** Constructor (also the default constructor).  All component masks are set
       to true (enabled) by default.
       @param red True to enable writing of red component.
       @param green True to enable writing of green component.
       @param blue True to enable writing of blue component.
       @param alpha True to enable wrting of alpha component.
   */
   sglColorMask(bool red = true, bool green = true,
                bool blue = true, bool alpha = true)
      : sglStatelet(eCOLOR_MASK) { setMask(red, green, blue, alpha); }
   /// virtual destructor
   virtual ~sglColorMask() {}

   /** Set all the color component masks.
       @param red True to enable writing of red component.
       @param green True to enable writing of green component.
       @param blue True to enable writing of blue component.
       @param alpha True to enable wrting of alpha component.
   */
   void setMask(bool red, bool green, bool blue, bool alpha);

   /** Query all the color component masks.
       @param red On return, contains true if red enabled, otherwise false.
       @param green On return, contains true if green enabled, otherwise false.
       @param blue On return, contains true if blue enabled, otherwise false.
       @param alpha On return, contains true if alpha enabled, otherwise false.
   */
   void getMask(bool &red, bool &green, bool &blue, bool &alpha) const;

   // See sglStatelet::apply() for details.
   void apply(sglCurrState *curr_state) const;

   // See sglObject::printInfo() for details.
   virtual void printInfo(ostream &ostrm, const char *indent_string) const;

private:
   sglColorMask(const sglColorMask &); // not implemented
   sglColorMask &operator=(const sglColorMask &); // not implemented

private:
   GLboolean m_red;
   GLboolean m_green;
   GLboolean m_blue;
   GLboolean m_alpha;
};


//============================================================================

/**

This statelet encapsulates OpenGL's glCullFace() and glEnable(GL_CULL_FACE)
functionality to determine which sides of polygons are culled (discarded)
before they are processed by the rendering pipeline.  The back faces (as
determined by polygon winding and the sglFrontFace statelet) are culled by
default, but the user can also cull the front faces, both sides, or neither
side by specifying one of the sglCullMode::CullModeEnum values.

See also: sglFrontFace

<b>Example:</b>
<pre>
   sglCullMode *cullMode = new sglCullMode;       // defaults to eBACK
   cull_mode->setCullMode(sglCullMode::eFRONT);
</pre>
 */

class SGL_DLL_API sglCullMode : public sglStatelet
{
public:
   /// The cull mode enums
   enum CullModeEnum
   {
      /// cull back faces
      eBACK = 0,
      /// don't cull any faces
      eOFF,
      /// cull the front faces
      eFRONT,
      /// cull both faces
      eFRONT_AND_BACK
   };

public:
   /** Constructor (also the default constructor).
       @param mode One of the CullModeEnum values to specify which face(s) to
         cull (defaults to sglCullMode::eBACK.
   */
   sglCullMode(CullModeEnum mode = eBACK)
      : sglStatelet(eCULL_MODE), m_cull_mode(mode) { setIndex(); }

   /// virtual destructor
   virtual ~sglCullMode() {}

   /** Set the cull mode.
       @param mode One of the CullModeEnum values to specify which face(s) to
         cull.
   */
   void setCullMode(CullModeEnum mode) { m_cull_mode = mode; setIndex(); }

   /** Query the current cull mode.
       @return One of the CullModeEnum's corresponding to the current cull
         mode.
   */
   CullModeEnum getCullMode() const    { return m_cull_mode; }

   // See sglStatelet::apply() for details.
   void apply(sglCurrState *curr_state) const;

   // See sglObject::printInfo() for details.
   virtual void printInfo(ostream &ostrm, const char *indent_string) const;

private:
   sglCullMode(const sglCullMode &); // not implemented
   sglCullMode &operator=(const sglCullMode &); // not implemented

private:
   void setIndex()
      {
         m_index = m_cull_mode;
      }

private:
   CullModeEnum m_cull_mode;
};


//============================================================================

/**

This statelet encapsulates OpenGL's glDepthMask() function which controls
whether or not depth values actually get written into the depth buffer.
It is a subclass of sglBooleanStatelet which uses the enable() member function
to control this mask.

<b>Example:</b>
<pre>
   sglDepthMask *dmask = new sglDepthMask;     // enabled by default
   dmask->enable(false);                       // disable depth writes

    // OR

   sglDepthMask *dmask = new sglDepthMask(false);
</pre>

 */

class SGL_DLL_API sglDepthMask : public sglBooleanStatelet
{
public:
   /** Constructor (default constructor too).
       @param on True to enable depth buffer writes (default), false to
         disable.
   */
   sglDepthMask(bool on = true)
      : sglBooleanStatelet(eDEPTH_MASK, on) {}

   /// virtual destructor
   virtual ~sglDepthMask() {}

   // See sglStatelet::apply() for details.
   void apply(sglCurrState *curr_state) const;

private:
   sglDepthMask(const sglDepthMask &); // not implemented
   sglDepthMask &operator=(const sglDepthMask &); // not implemented
};


//============================================================================

/**

This statelet encapsulates OpenGL's glDepthRange() function which defines an
encoding for z coordinates that is preformed during the viewport
transformation.  The near_val and far_val parameters represent adjustments to
the minimum and maximum values that can be stored in the depth buffer.  The
default to 0 and 1 respectively and are clamped to [0, 1].  This can be used
for a number of effects including polygon offset like effects as in the
following example:

<pre>
   double epsilon = 1.0e-4;

   // bias the depth values closer to zero (near clip plane).
   sglDepthRange *drange = new sglDepthRange(0.0, 1.0 - epsilon);

   // bias the depth values closer to one (far clip plane).
   drange->setDepthRange(epsilon, 1.0);
</pre>
 */

class SGL_DLL_API sglDepthRange : public sglStatelet
{
public:
   /** Constructor (also the default constructor).
       @param near_val The depth value corresponding to the near clip plane.
         Values are clamped to [0,1]. The default is 0.0.
       @param far_val The depth value corresponding to the far clip plane.
         Values are clamped to [0,1]. The default is 1.0.
   */
   sglDepthRange(double near_val = 0.0, double far_val = 1.0)
         : sglStatelet(eDEPTH_RANGE)
      { setDepthRange(near_val, far_val); }

   /// virtual destructor
   virtual ~sglDepthRange() {}

   /** Set a new depth range
       @param near_val The depth value corresponding to the near clip plane.
         Values are clamped to [0,1]. The default is 0.0.
       @param far_val The depth value corresponding to the far clip plane.
         Values are clamped to [0,1]. The default is 1.0.
   */
   void setDepthRange(double  near_val, double  far_val);

   /** Query the current depth range.
       @param near_val The depth value corresponding to the near clip plane.
       @param far_val The depth value corresponding to the far clip plane.
   */
   void getDepthRange(double &near_val, double &far_val) const
      {
         near_val = m_near;
         far_val  = m_far;
      }

   // See sglStatelet::apply() for details.
   void apply(sglCurrState *curr_state) const;

   // See sglStatelet::getSorted() for details.
   virtual bool getSorted() const { return true; }

   // See sglObject::printInfo() for details.
   virtual void printInfo(ostream &ostrm, const char *indent_string) const;

private:
   sglDepthRange(const sglDepthRange &); // not implemented
   sglDepthRange &operator=(const sglDepthRange &); // not implemented

   void setIndex()
      {
         m_index = ((m_near == 0.0 && m_far == 1.0) ? 0 : UINT_MAX);
      }

private:
   double m_near, m_far;
};


//============================================================================

/**

This statelet encapsulates the OpenGL's glDepthFunc() and
glEnable(GL_DEPTH_TEST) functions which controls how new rendering fragments
are tested against the depth buffer to determine if they are visible.
The user only needs to specify the depth function which is the test used to
compare incoming fragments with existing depth buffer values.  If the user
passes in sglDepthTest::eALWAYS, this is equivalent to disabling depth test.

<b>Example:</b>
<pre>
   sglDepthTest *dtest = new sglDepthTest(sglDepthTest::eALWAYS); // disable
   dtest->setDepthFunc(sglDepthFunc::eLEQUAL);
</pre>

 */

class SGL_DLL_API sglDepthTest : public sglStatelet
{
public:
   /// The various depth functions
   enum DepthFuncEnum
   {
      /// The default test: pass if fragment depths are less (closer).
      eLESS     = GL_LESS,
      /// Disables the depth test (everything passes)
      eALWAYS   = GL_ALWAYS,
      ///
      eLEQUAL   = GL_LEQUAL,
      ///
      eEQUAL    = GL_EQUAL,
      ///
      eGEQUAL   = GL_GEQUAL,
      ///
      eGREATER  = GL_GREATER,
      ///
      eNOTEQUAL = GL_NOTEQUAL,
      /// Whats the point of this?
      eNEVER    = GL_NEVER
   };

public:
   /** Constructor (also the default constructor).
       @param func One of the DepthFuncEnum tokens.  The default is eLESS.
   */
   sglDepthTest(DepthFuncEnum func = eLESS)
         : sglStatelet(eDEPTH_TEST), m_depth_func(func)
      {
         setIndex();
      }

   /// virtual destructor
   virtual ~sglDepthTest() {}

   /** Set a depth function.
       @param func One of the DepthFuncEnum tokens.
   */
   void setDepthFunc(DepthFuncEnum func);

   /** Query the current depth function.
       @return One of the DepthFuncEnum tokens.
   */
   DepthFuncEnum getDepthFunc() const { return m_depth_func; }

   // See sglStatelet::apply() for details.
   void apply(sglCurrState *curr_state) const;

   // See sglObject::printInfo() for details.
   virtual void printInfo(ostream &ostrm, const char *indent_string) const;

private:
   sglDepthTest(const sglDepthTest &); // not implemented
   sglDepthTest &operator=(const sglDepthTest &); // not implemented

   void setIndex();

private:
   DepthFuncEnum m_depth_func;
};


//============================================================================

/**

This statelet encapsulates OpenGL's glEnable(GL_DITHER) functionality, which
determines whether or not dithering is desired to increase the number of
apparent colors on graphics hardware.  This is enabled by default (as it is in
OpenGL.

<b>Example:</b>
<pre>
   sglDithering *dither = new sglDithering(false);     // disable it
   dither->enable(true);                               // enable it
</pre>
 */

class SGL_DLL_API sglDithering : public sglBooleanStatelet
{
public:
   /** Constructor (also default constructor).
       @param on True to enable dithering (the default); false to disable.
   */
   sglDithering(bool on = true)
      : sglBooleanStatelet(eDITHERING, on) {}

   /// virtual destructor
   virtual ~sglDithering() {}

   // See sglStatelet::apply() for details.
   void apply(sglCurrState *curr_state) const;

private:
   sglDithering(const sglDithering &); // not implemented
   sglDithering &operator=(const sglDithering &); // not implemented
};

//============================================================================

/**

This statelet encapsulates OpenGL's glFrontFace() function which controls how
front-facing polygons are determined.  By default, the mode is eCCW which
corresponds to a counter-clockwise orientation of the ordered vertices of a
projected polygon in window coordinates.  The other side of the polygon can be
specified as the "front" face if the mode is switched to eCW (clockwise).  The
presence of this statelet affects the sglCullMode operation.

<b>Example:</b>
<pre>
   sglFrontFace *cmask = new sglFrontFace;     // eCCW by default
   cmask->setFrontFace(sglFrontFace::eCW);     // switch front face
</pre>


 */

class SGL_DLL_API sglFrontFace : public sglStatelet
{
public:
   /// The front face mode tokens
   enum FrontFaceEnum
   {
      /// counterclockwise (the default)
      eCCW = GL_CCW,

      /// clockwise
      eCW = GL_CW
   };

public:
   /** Constructor (also the default constructor).
       @param winding One of the two FrontFaceEnum tokens.  Defaults to eCCW.
   */
   sglFrontFace(FrontFaceEnum winding = eCCW)
      : sglStatelet(eFRONT_FACE), m_winding(winding) { setIndex(); }

   /// virtual destructor
   virtual ~sglFrontFace() {}

   /** Set the polygon winding mode.
       @param winding One of the two FrontFaceEnum tokens.
   */
   void setFrontFace(FrontFaceEnum winding) { m_winding = winding; setIndex(); }

   /** Query the current winding mode.
       @return One of the two FrontFaceEnum tokens.
   */
   FrontFaceEnum getFrontFace() const { return m_winding; }

   // See sglStatelet::apply() for details.
   void apply(sglCurrState *curr_state) const;

   // See sglObject::printInfo() for details.
   virtual void printInfo(ostream &ostrm, const char *indent_string) const;

private:
   void setIndex() { m_index = ((m_winding == eCCW) ? 0 : 1); }

   sglFrontFace(const sglFrontFace &); // not implemented
   sglFrontFace &operator=(const sglFrontFace &);// not implemented

protected:
   FrontFaceEnum m_winding;
};

//============================================================================

/**

This statelet controls whether or not the affected sglDrawable(s) will have
bounding boxes rendered around the geometry.  This is strictly an SGL feature
and does not encapsulate any specific component of OpenGL state.

<b>Example:</b>
<pre>
   sglHighlight *hilite = new sglHighlight;     // disabled by default
   hilite->enable(true);  // only enable red components
</pre>

 */

class SGL_DLL_API sglHighlight : public sglBooleanStatelet
{
public:
   /** Constructor (and the default constructor).
       @param on True to enable, and false to disable this feature.
   */
   sglHighlight(bool on = false)
      : sglBooleanStatelet(eHIGHLIGHT, on) {}

   /// virtual destructor
   virtual ~sglHighlight() {}

   // See sglStatelet::apply() for details.
   void apply(sglCurrState *curr_state) const;

private:
   sglHighlight(const sglHighlight &); // not implemented
   sglHighlight &operator=(const sglHighlight &); // not implemented
};

//============================================================================

/**

This statelet encapsulates OpenGL's glEnable(GL_LIGHTING) function which
controls whether or not lighting calculations are performed on subsequent
geometry (sglDrawable) in determining the final fragment colors.  This is
disabled by default.

<b>Example:</b>
<pre>
   sglLighting *lighting = new sglLighting(true);   // enable lighting
   lighting->enable(false);                         // disable lighting
</pre>

 */

class SGL_DLL_API sglLighting : public sglBooleanStatelet
{
public:
   /** Constructor (and the default constructor).
       @param on True to enable lighting, false (the default) to disable.
   */
   sglLighting(bool on = false)
      : sglBooleanStatelet(eLIGHTING, on) {}

   /// virtual destructor
   virtual ~sglLighting() {}

   // See sglStatelet::apply() for details.
   void apply(sglCurrState *curr_state) const;

private:
   sglLighting(const sglLighting &); // not implemented
   sglLighting &operator=(const sglLighting &); // not implemented
};


//============================================================================

/**

This statelet encapsulates OpenGL's glLineStipple() and
glEnable(GL_LINE_STIPPLE) functions which controls the stippling pattern for
line primitives.  This is disabled by default (by setting the pattern is
0xffff).  The factor scales the stipple pattern and can take on a value in the
range [1..255].  Values outside this range will be clamped.

<b>Example:</b>
<pre>
   sglLineStipple *lstipple = new sglLineStipple(0xf0f0, 1);  // enable - - -
   lstipple->setLineStipple(0x5555, 1);                       // enable .......
   lstipple->setLineStipple(0xffff, 1);                       // disable
</pre>

 */

class SGL_DLL_API sglLineStipple : public sglStatelet
{
public:
   /** Constructor (also the default constructor).
       @param pattern The bit pattern (16 bits) defining the basic pattern
          where each 1 bit defines an on pixel and each 0 bit defines an off
          pixel.  The default is 0xffff which is all on and disables this
          statelet.
       @param factor The scale factor for the pattern.  The default is 1 (no
          scale), and larger scales factors lengthen the pattern in integer
          multiples.  This value will be clamped to the range [1..255].
   */
   sglLineStipple(GLushort pattern = 0xffff, int factor = 1)
      : sglStatelet(eLINE_STIPPLE)
      { setLineStipple(pattern, factor); }

   /// virtual destructor
   virtual ~sglLineStipple() {}

   /** Set the line stipple pattern and scale factor.
       @param pattern The bit pattern (16 bits) defining the basic pattern
          where each 1 bit defines an on pixel and each 0 bit defines an off
          pixel.  A pattern of 0xffff (all bits on) and disables this
          statelet.
       @param factor The scale factor for the pattern.  Scale factors larger
          larger than one lengthen the pattern in integer multiples.  This
          value will be clamped to the range [1..255].
   */
   void setLineStipple(GLushort  pattern, int  factor);

   /** Query the current stipple pattern and scale factor.
       @param pattern On return, contains the current bit pattern.
       @param factor On return, contains the current scale factor.
   */
   void getLineStipple(GLushort &pattern, int &factor) const
      {
         pattern = m_pattern;
         factor  = m_factor;
      }

   // See sglStatelet::apply() for details.
   void apply(sglCurrState *curr_state) const;

   // See sglStatelet::getSorted() for details.
   virtual bool getSorted() const { return true; }

   // See sglObject::printInfo() for details.
   virtual void printInfo(ostream &ostrm, const char *indent_string) const;

private:
   sglLineStipple(const sglLineStipple &); // not implemented
   sglLineStipple &operator=(const sglLineStipple &); // not implemented

   void setIndex()
      {
         m_index = ((m_pattern == 0xffff) ? 0 : UINT_MAX);
      }

private:
   GLushort m_pattern;
   int      m_factor;
};

//============================================================================

/**

This statelet encapsulates OpenGL's glLineWidth() function which controls the
width, in pixels, for rendered lines.  The default value is one pixel.

<b>Example:</b>
<pre>
   sglLineWidth *line_width = new sglLineWidth;   // the default of 1.
   line_width->setLineWidth(2.5f);                // 2.5 times large
</pre>

 */

class SGL_DLL_API sglLineWidth : public sglStatelet
{
public:
   /** Constructor (and the default constructor).
       @param line_width The width of lines in pixels.  This must be greater
         than zero, and is 1.0 by default.  Values less than or equal to zero
         will result in the default width of 1.0.
   */
   sglLineWidth(float line_width = 1.0f)
      : sglStatelet(eLINE_WIDTH)
      { setLineWidth(line_width); }

   /// virtual destructor
   virtual ~sglLineWidth() {}

   /** Set the line width.
       @param line_width The width of lines in pixels.  This must be greater
         than zero.  Values less than zero will be set to 1.0.
   */
   void setLineWidth(float line_width);

   /** Query the current line width
       @return The current line width.
   */
   float getLineWidth() const { return m_line_width; }

   // See sglStatelet::apply() for details.
   void apply(sglCurrState *curr_state) const;

   // See sglStatelet::getSorted() for details.
   virtual bool getSorted() const { return true; }

   // See sglObject::printInfo() for details.
   virtual void printInfo(ostream &ostrm, const char *indent_string) const;

private:
   sglLineWidth(const sglLineWidth &); // not implemented
   sglLineWidth &operator=(const sglLineWidth &); // not implemented

   void setIndex()
      {
         m_index = ((m_line_width == 1.0f) ? 0 : UINT_MAX);
      }

private:
   float m_line_width;
};


//============================================================================

/**

This statelet encapsulates OpenGL's glLogicOp() and glEnable(GL_COLOR_LOGIC_OP)
functions (note that glEnable(GL_INDEX_LOGIC_OP) is not supported by SGL) which
sets the logical operation to be performed given the incoming fragment and the
pixel currently stored in the color buffer.  The default operation is
sglLogicOp::eCOPY which overwrites the color buffer (destination) with the
incoming fragment (source).  The complete list of logical operations is
contained in the LogicOpEnum list of this class.

<b>Example:</b>
<pre>
   sglLogicOp *logic_op = new sglLogicOp;     // defaults to eCOPY
   logic_op->setLogicOp(sglLogicOp::eEQUIV);  // ~(source XOR dest)
</pre>

 */

class SGL_DLL_API sglLogicOp : public sglStatelet
{
public:
   /// The list of valid logic op tokens.
   enum LogicOpEnum
   {
      /// source (the default)
      eCOPY          = GL_COPY,
      /// zero
      eCLEAR         = GL_CLEAR,
      /// one
      eSET           = GL_SET,
      /// ~source
      eCOPY_INVERTED = GL_COPY_INVERTED,
      /// dest
      eNOOP          = GL_NOOP,
      /// ~dest
      eINVERT        = GL_INVERT,
      /// source & dest
      eAND           = GL_AND,
      /// ~(source & dest)
      eNAND          = GL_NAND,
      /// source | dest
      eOR            = GL_OR,
      /// ~(source | dest)
      eNOR           = GL_NOR,
      /// source ^ dest
      eXOR           = GL_XOR,
      /// ~(source ^ dest)
      eEQUIV         = GL_EQUIV,
      /// source & ~dest
      eAND_REVERSE   = GL_AND_REVERSE,
      /// ~source & dest
      eAND_INVERTED  = GL_AND_INVERTED,
      /// source | ~dest
      eOR_REVERSE    = GL_OR_REVERSE,
      /// ~source | dest
      eOR_INVERTED   = GL_OR_INVERTED
   };

public:
   /** Constructor (also the default constructor).
       @param op The logical operation (one of the LogicOpEnum tokens).  The
         default is eCOPY.
   */
   sglLogicOp(LogicOpEnum op = eCOPY)
      : sglStatelet(eLOGIC_OP), m_logic_op(op) { setIndex(); }

   /// virtual destructor
   virtual ~sglLogicOp() {}

   /** Set the logical operation.
       @param op The logical operation (one of the LogicOpEnum tokens).
   */
   void setLogicOp(LogicOpEnum op) { m_logic_op = op; setIndex(); }

   /** Query the current logical operation.
       @return One of the LogicOpEnum tokens corresponding to the current
         logical operation.
   */
   LogicOpEnum getLogicOp() const { return m_logic_op; }

   // See sglStatelet::apply() for details.
   void apply(sglCurrState *curr_state) const;

   // See sglObject::printInfo() for details.
   virtual void printInfo(ostream &ostrm, const char *indent_string) const;

private:
   void setIndex();

   sglLogicOp(const sglLogicOp &); // not implemented
   sglLogicOp &operator=(const sglLogicOp &);// not implemented

protected:
   LogicOpEnum m_logic_op;
};


//============================================================================

/**

This statelet works with either the GL_EXT_occlusion_test or
GL_HP_occlusion_test extensions and encapsulates the
glEnable(GL_OCCLUSION_TEST_HP) and glGetBooleanv(GL_OCCLUSION_RESULT_HP)
functions to perform occlusion testing on geometry's bounding boxes to
determine visibility before passing geometry onto the OpenGL rendering
engines.  Geometry that passes the cull test and have the occlusion statelet
enabled will be placed in a separate "occlusion bin" outside of the normal
state sorting mechanism and will be rendered before the geometry in the
transparency bin but after all other state sorted geometry.  To increase the
chances that occlusions will be detected this bin is sorted and rendered from
front to back.

If occlusion culling is enabled on hardware that does not support it the flag
will be ignored and occlusion sorting will not occur.

Note: occlusion culling is expensive and should only be enabled for even more
expensive geometry that is likely to be occluded.

What happens when transparency and occlusion are both enabled?

<b>Example:</b>
<pre>
   sglOcclusionCull *occlude = new sglOcclusionCull;   // disabled by default
   occlude->enable(true);                              // enable it
</pre>

 */

class SGL_DLL_API sglOcclusionCull : public sglStatelet
{
public:
   /** Constructor (also the default constructor)
       @param on True to enable and false (the default) to disable occlusion
         cull.
   */
   sglOcclusionCull(bool on = false)
      : sglStatelet(eOCCLUSION_CULL) { enable(on); }

   /// virtual constructor
   virtual ~sglOcclusionCull() {}

   /** Enable/disable occlusion cull (if supported by the rendering hardware).
       @param on True to enable and false to disable occlusion culling.
   */
   void enable(bool on)
      {
         m_on = (on && sgl::s_occlusion_capable);
         setIndex();
      }

   /** Query the flag for occlusion culling.
       @return True if enabled, or false if disabled.
   */
   bool isEnabled() const  { return m_on; }

   // See sglStatelet::apply() for details.
   void apply(sglCurrState *curr_state) const;

   // See sglObject::printInfo() for details.
   virtual void printInfo(ostream &ostrm, const char *indent_string) const;

private:
   sglOcclusionCull(const sglOcclusionCull &); // not implemented
   sglOcclusionCull &operator=(const sglOcclusionCull &); // not implemented

private:
   void setIndex() { m_index = (m_on ? 1 : 0); }

private:
   bool m_on;
};

//============================================================================

/**

This statelet encapsulates OpenGL's glPointSize() function which controls the
width, in pixels, for rendered points.  The default value is one pixel.

<b>Example:</b>
<pre>
   sglPointSize *point_size = new sglPointSize;   // the default of 1.
   point_size->setPointSize(2.5f);                // 2.5 times large
</pre>

*/

class SGL_DLL_API sglPointSize : public sglStatelet
{
public:
   /** Constructor (also the default constructor)
       @param point_size The width of points in pixels.  This must be greater
         than zero, and is 1.0 by default.  Values less than or equal to zero
         will be set to 1.0.
   */
   sglPointSize(float point_size = 1.0f)
      : sglStatelet(ePOINT_SIZE)
      { setPointSize(point_size); }

   /// virtual destructor
   virtual ~sglPointSize() {}

   /** Set the point size
       @param point_size The width of points in pixels.  This must be greater
         than zero.  Values less than or equal to zero will be set to 1.0.
   */
   void setPointSize(float  point_size);

   /** Query the current point size
       @return The current point size
   */
   float getPointSize() const
      {
         return m_point_size;
      }

   // See sglStatelet::apply() for details.
   void apply(sglCurrState *curr_state) const;

   // See sglStatelet::getSorted() for details.
   virtual bool getSorted() const { return true; }

   // See sglObject::printInfo() for details.
   virtual void printInfo(ostream &ostrm, const char *indent_string) const;

private:
   sglPointSize(const sglPointSize &); // not implemented
   sglPointSize &operator=(const sglPointSize &); // not implemented

   void setIndex()
      {
         m_index = ((m_point_size == 1.0f) ? 0 : UINT_MAX);
      }

private:
   float m_point_size;
};

//============================================================================

/**

This statelet works with either the GL_EXT_point_parameters extension to
OpenGL (or the GL_SGIS_point_parameters in earlier versions) and encapsulates
the glPointParameterfEXT() and glPointParameterfvEXT() functions to specify
point parameters for point primitives or other geometries in point polygon mode.

With this extensions points are rendered with a given size (as specified by
glPointSize() via the sglPointSize class) which can be attenuated with distance
according to the following equation (similar to light attenuation):
<pre>
    dist_atten(dist) = 1 / ( A + B * dist + C * dist^2 )
</pre>
where A, B, and C, constant, linear, and quadratic constants and default
to (1, 0, 0) respectively.  Below a certain minimum point size the alpha
component of the point's color will also start to be attenuated to continue to
attenuate the point below the minimum size.

With this statelet the minimum and maximum size of the points can be set, the
attenuation constants, and the fade threshold.  This can be used to particles
or tiny light sources commonly referred to as "light points."

<b>Example:</b>
<pre>
   sglPointParameters *pt_param = new sglPointParameters;
   pt_param->setPointParameters(1.0f, 0.34f, 0.8f,    // set new attenuation
                                sgl::s_point_default_fade_threshold,
                                sgl::s_point_default_size_min,
                                sgl::s_point_default_size_max);
</pre>

*/

class SGL_DLL_API sglPointParameters : public sglStatelet
{
public:
   /** Constructor (also the default constructor).
       @param dist_atten_A Constant attenuation factor, defaults to 1.0.
       @param dist_atten_B Linear attenuation factor, defaults to 0.0.
       @param dist_atten_C Quadratic attenuation factor, defaults to 0.0.
       @param fade_threshold When multisampled (anti-aliased) points are
         enabled, this parameter defines the point at which size (diameter) is
         no longer attenuated (above this value) and alpha component is (below
         this value).
       @param min_size The minimum derived point size (after attenuation).
       @param max_size The maximum derived point size (after attenuation).
   */
   sglPointParameters(float dist_atten_A = 1.0f,
                      float dist_atten_B = 0.0f,
                      float dist_atten_C = 0.0f,
                      float fade_threshold =
                                       sgl::s_point_default_fade_threshold,
                      float min_size = sgl::s_point_default_size_min,
                      float max_size = sgl::s_point_default_size_max )
         : sglStatelet(ePOINT_PARAMETERS)
      {
         setPointParameters( dist_atten_A, dist_atten_B, dist_atten_C,
                             fade_threshold,
                             min_size, max_size );
      }

   /// virtual destructor
   virtual ~sglPointParameters() {}

   /** Set all the point parameters at once.
       @param dist_atten_A Constant attenuation factor
       @param dist_atten_B Linear attenuation factor
       @param dist_atten_C Quadratic attenuation factor
       @param fade_threshold When multisampled (anti-aliased) points are
         enabled, this parameter defines the point at which size (diameter) is
         no longer attenuated (above this value) and alpha component is (below
         this value).
       @param min_size The minimum derived point size (after attenuation).
       @param max_size The maximum derived point size (after attenuation).
    */
   void setPointParameters(float dist_atten_A,
                           float dist_atten_B,
                           float dist_atten_C,
                           float fade_threshold =
                                            sgl::s_point_default_fade_threshold,
                           float min_size = sgl::s_point_default_size_min,
                           float max_size = sgl::s_point_default_size_max )
      {
         // HACK do we need to range check any of these?
         m_min_size = min_size;
         m_max_size = max_size;
         m_fade_threshold = fade_threshold;
         m_dist_atten_A = dist_atten_A;
         m_dist_atten_B = dist_atten_B;
         m_dist_atten_C = dist_atten_C;

         setIndex();
      }

   /** Query all the point parameters.
       @param dist_atten_A On return, constant attenuation factor
       @param dist_atten_B On return, linear attenuation factor
       @param dist_atten_C On return, quadratic attenuation factor
       @param fade_threshold On return, contains the fade threshold size.
       @param min_size On return, the minimum derived point size.
       @param max_size On return, the maximum derived point size.
    */
   void getPointParameters(float &dist_atten_A,
                           float &dist_atten_B,
                           float &dist_atten_C,
                           float &fade_threshold,
                           float &min_size,
                           float &max_size)
      {
         min_size = m_min_size;
         max_size = m_max_size;
         fade_threshold = m_fade_threshold;
         dist_atten_A = m_dist_atten_A;
         dist_atten_B = m_dist_atten_B;
         dist_atten_C = m_dist_atten_C;
      }

   // See sglStatelet::apply() for details.
   void apply(sglCurrState *curr_state) const;

   // See sglStatelet::getSorted() for details.
   virtual bool getSorted() const { return true; }

   // See sglObject::printInfo() for details.
   virtual void printInfo(ostream &ostrm, const char *indent_string) const;

private:
   sglPointParameters(const sglPointParameters &); // not implemented
   sglPointParameters &operator=(const sglPointParameters &); // not implemented

   void setIndex()
   {
      if (m_dist_atten_A != 1.0f ||
          m_dist_atten_B != 0.0f ||
          m_dist_atten_C != 0.0f ||
          m_fade_threshold != sgl::s_point_default_fade_threshold ||
          m_min_size != sgl::s_point_default_size_min ||
          m_max_size != sgl::s_point_default_size_max )
      {
         m_index = UINT_MAX;
      }
      else
      {
         m_index = 0;
      }
   }

private:
   float m_min_size;
   float m_max_size;
   float m_dist_atten_A;
   float m_dist_atten_B;
   float m_dist_atten_C;
   float m_fade_threshold;
};

//============================================================================

/**

This statelet encapsulates OpenGL's glPolygonMode() function which controls the
drawing mode for polygons as filled, wireframe (LINE), or points.  This is the
abstract base class for the sglPolyFrontStyle and sglPolyBackStyle subclasses
that control which side of the polygon is affected.  By default both front and
back faces are drawn filled.

 */

class SGL_DLL_API sglPolyStyle : public sglStatelet
{
public:
   /// Tokens for the three styles.
   enum StyleEnum
   {
      /// filled, the default.
      eFILLED = 0,

      /// wireframe.
      eLINE,

      /// points (at each vertex).
      ePOINT
   };

public:
   /** Set the polygon style for the side specified by the subclass.
       @param style One of the StyleEnum tokens
   */
   void setStyle(StyleEnum style) { m_style = style; setIndex(); }

   /** Query the current polygon style
       @return The StyleEnum token corresponding to the current polygon mode.
   */
   StyleEnum getStyle() const { return m_style; }

   // See sglObject::printInfo() for details.
   virtual void printInfo(ostream &ostrm, const char *indent_string) const;

protected:
   sglPolyStyle(StateEnum type, StyleEnum style)
      : sglStatelet(type), m_style(style) { setIndex(); }
   virtual ~sglPolyStyle() {}

private:
   sglPolyStyle(); // not implemented
   sglPolyStyle(const sglPolyStyle &); // not implemented
   sglPolyStyle &operator=(const sglPolyStyle &); // not implemented

   void setIndex() { m_index = m_style; }

protected:
   StyleEnum m_style;
};

//============================================================================

/**

This statelet encapsulates OpenGL's glPolygonMode(GL_BACK,..) function which
controls the drawing style for the back faces of polygons.  The polygons can be
rendered as either filled (the default), wireframe (LINE), or points.

<b>Example:</b>
<pre>
   sglPolyBackStyle *pbs = new sglPolyBackStyle;  // filled by default
   pbs->setStyle(sglPolyStyle::eLINE);            // set back face to wireframe
</pre>

 */

class SGL_DLL_API sglPolyBackStyle : public sglPolyStyle
{
public:
   /** Constructor (also the default constructor).
       @param style The drawing mode which is one of the three tokens from
         sglPolyStyle::StyleEnum.  The default is eFILLED.
   */
   sglPolyBackStyle(StyleEnum style = eFILLED)
      : sglPolyStyle(ePOLY_BACK_STYLE, style) {}

   /// virtual destructor
   virtual ~sglPolyBackStyle() {}

   // See sglStatelet::apply() for details.
   void apply(sglCurrState *curr_state) const;

private:
   sglPolyBackStyle(const sglPolyBackStyle &); // not implemented
   sglPolyBackStyle &operator=(const sglPolyBackStyle &); // not implemented
};


//============================================================================

/**

This statelet encapsulates OpenGL's glPolygonMode(GL_FRONT,..) function which
controls the drawing style for the front faces of polygons.  The polygons can
be rendered as either filled (the default), wireframe (LINE), or points.

<b>Example:</b>
<pre>
   sglPolyFrontStyle *pfs = new sglPolyFrontStyle;  // filled by default
   pfs->setStyle(sglPolyStyle::eLINE);          // set front face to wireframe
</pre>

 */

class SGL_DLL_API sglPolyFrontStyle : public sglPolyStyle
{
public:
   /** Constructor (also the default constructor).
       @param style The drawing mode which is one of the three tokens from
         sglPolyStyle::StyleEnum.  The default is eFILLED.
   */
   sglPolyFrontStyle(StyleEnum style = eFILLED)
      : sglPolyStyle(ePOLY_FRONT_STYLE, style) {}

   /// virtual destructor
   virtual ~sglPolyFrontStyle() {}

   // See sglStatelet::apply() for details.
   void apply(sglCurrState *curr_state) const;

private:
   sglPolyFrontStyle(const sglPolyFrontStyle &); // not implemented
   sglPolyFrontStyle &operator=(const sglPolyFrontStyle &); // not implemented
};

//============================================================================

/**

This statelet encapsulates OpenGL's glPolygonOffset() and
glEnable(GL_POLYGON_OFFSET_*) functions.  When enabled the depth value of each
fragment is added to a calculated offset according to following formula:
<pre>
   offset = m * factor + r * units
</pre>
where (according to the OpenGL Red Book) "m is the maximum depth slope of the
polygon and r is the smallest value guaranteed to produce a resolvable
difference in the window coordinate depth values."  The factor and units are
provided by this class.  When factor and units are set to zero, that disables
the polygon offset.  When either is non-zero then polygon offset will be
enabled for POINTS, LINES, and FILLed polygon modes.

Note that this statelet enables polygon offset for polygons in fill, line, and
point mode.

<b>Example:</b>
<pre>
   sglPolygonOffset *po = new sglPolygonOffset;  // disabled by default
   po->setPolygonOffset(1.0, 1.0);               // pushes polygons farther.
</pre>

 */

class SGL_DLL_API sglPolygonOffset : public sglStatelet
{
public:
   /** Constructor (also the default constructor).  If both parameters are set
       to zero then polygon offset is disabled.
       @param factor The multiplicative factor for the maximum depth slope of
         the polygon.  This can take on any floating point value, and defaults
         to zero.
       @param units The multiplicative factor the resolution of the depth
         buffer (essentially an offset value).  Defaults to zero.
   */
   sglPolygonOffset(float factor = 0, float  units = 0)
      : sglStatelet(ePOLYGON_OFFSET)
      { setPolygonOffset(factor, units); }

   /// virtual destructor
   virtual ~sglPolygonOffset() {}

   /** Set the polygon offset parameters.  If both parameters are set
       to zero then polygon offset is disabled.
       @param factor The multiplicative factor for the maximum depth slope of
         the polygon.  This can take on any floating point value.
       @param units The multiplicative factor the resolution of the depth
         buffer (essentially an offset value).
   */
   void setPolygonOffset(float  factor, float  units);

   /** Query the polygon offset parameters.
       @param factor On return, contains the polygon offset factor.
       @param units On return, contains the polygon offset units value.
    */
   void getPolygonOffset(float &factor, float &units) const
      {
         factor = m_factor;
         units = m_units;
      }

   /** Query whether or not polygon offset is enable (it compares the two
       parameters to zero as a convenience to the user).
       @return true if either parameter is non-zero; otherwise, false.
   */
   bool getPolygonOffsetFlag() const { return m_polygon_offset_flag; }

   // See sglStatelet::apply() for details.
   void apply(sglCurrState *curr_state) const;

   // See sglStatelet::getSorted() for details.
   virtual bool getSorted() const { return true; }

   // See sglObject::printInfo() for details.
   virtual void printInfo(ostream &ostrm, const char *indent_string) const;

private:
   sglPolygonOffset(const sglPolygonOffset &); // not implemented
   sglPolygonOffset &operator=(const sglPolygonOffset &); // not implemented

   void setIndex()
      {
         m_index = (m_polygon_offset_flag ? UINT_MAX : 0);
      }

private:
   bool  m_polygon_offset_flag;
   float m_factor;
   float m_units;
};


//============================================================================

/**

This statelet encapsulates OpenGL's glPolygonStipple() and
glEnable(GL_POLYGON_STIPPLE) commands which defines the 32x32 bit stipple
pattern where each bit corresponds to whether or not the corresponding pixel in
the polygon is rendered (1) or not (0).  The 32x32 bit pattern is defined with
an 128 byte memory block starting in the lower left of the bitmap scanning to
the right and upwards (see the OpenGL Red Book for details).  A pointer to this
memory is passed to the setPolygonStipple function to set the pattern (or pass
it into the constructor).  This memory is then copied into the class's internal
memory.  Passing NULL in for the pattern parameter disables polygon stipple.

WARNING: this statelet issues glPixelStorei(GL_UNPACK_*, ) commands for
ALIGNMENT (4), ROW_LENGTH (0), SKIP_ROWS (0), SKIP_PIXELS (0), and may
interfere with user settings in the application outside of SGL.

<b>Example:</b>
<pre>
   GLubtye pattern[128] = { 0x10, 0x18, 0x18, 0x08,      // first (bottom) row
                                    :
                            0x10, 0x18, 0x18, 0x08 };    // last (top) row

   sglPolygonStipple *pstipple = new sglPolygonStipple;  // disabled by default
   pstipple->setPolygonStipple(pattern);                 // set the pattern
</pre>

 */

class SGL_DLL_API sglPolygonStipple : public sglStatelet
{
public:
   /** Constructor (also the default constructor).
       @param pattern Pointer to 128 packed bytes that define the stipple
          pattern.  Passing in NULL (the default) disables this statelet.
   */
   sglPolygonStipple(const GLubyte *pattern = NULL) //assuming 128 bytes packed
         : sglStatelet(ePOLYGON_STIPPLE)
      { setPolygonStipple(pattern); }

   /// virtual destructor
   virtual ~sglPolygonStipple() {}

   /** Set the polygon stipple pattern.
       @param pattern Pointer to 128 packed bytes that define the stipple
          pattern.  Passing in NULL disables this statelet.
   */
   void setPolygonStipple(const GLubyte *pattern);

   /** Query the current polygon stipple pattern.
       @return A pointer to the stipple pattern (if enabled) or NULL (if
          disabled).  Note that this class passes back a pointer to internal
          data, and the user should copy this data to memory that they "own" to
          ensure that this pointer is not invalidated if the statelet is
          destructed.
   */
   const GLubyte *getPolygonStipple() const
      {
         if (m_polygon_stipple_flag)
            return m_pattern;
         else
            return NULL;
      }

   /** Query whether or not this statelet is enabled.
       @return true if a non-NULL pointer has been supplied for the pattern;
          otherwise false.
   */
   bool getPolygonStippleFlag() const { return m_polygon_stipple_flag; }

   // See sglStatelet::apply() for details.
   void apply(sglCurrState *curr_state) const;

   // See sglStatelet::getSorted() for details.
   virtual bool getSorted() const { return true; }

   // See sglObject::printInfo() for details.
   virtual void printInfo(ostream &ostrm, const char *indent_string) const;

private:
   sglPolygonStipple(const sglPolygonStipple &); // not implemented
   sglPolygonStipple &operator=(const sglPolygonStipple &); // not implemented

   void setIndex()
      {
         m_index = (m_polygon_stipple_flag ? UINT_MAX : 0);
      }

private:
   bool    m_polygon_stipple_flag;
   GLubyte m_pattern[128];
};


//============================================================================

/**

This statelet encapsulates OpenGL's glShadeModel() function which sets the
shading model to either flat or smooth (the default).

<b>Example:</b>
<pre>
   sglShadeModel *shade_model = new sglShadeModel;    // default: smooth
   shade_model->setShadeModel(sglShadeModel::eFLAT);  // flat shading
</pre>

 */

class SGL_DLL_API sglShadeModel : public sglStatelet
{
public:
   /// The shade model type tokens.
   enum ShadeModelEnum
   {
      /// flat shading
      eFLAT = GL_FLAT,

      /// smooth (or Gouraud) shading
      eSMOOTH = GL_SMOOTH
   };

public:
   /** Constructor (also the default constructor)
       @param model One of the ShadeModelEnum tokens: eFLAT for flat shading,
         or eSMOOTH (the default) for Gouraud shading.
   */
   sglShadeModel(ShadeModelEnum model = eSMOOTH)
         : sglStatelet(eSHADE_MODEL), m_model(model) { setIndex(); }

   /// virtual destructor
   virtual ~sglShadeModel() {}

   /** Set the shading model.
       @param model One of the ShadeModelEnum tokens: eFLAT for flat shading,
         or eSMOOTH for Gouraud shading.
   */
   void setShadeModel(ShadeModelEnum model) { m_model = model; setIndex(); }

   /** Query the current shading model.
       @return One of the ShadeModelEnum tokens corresponding to the current
          shading model.
   */
   ShadeModelEnum getShadeModel() const { return m_model; }

   // See sglStatelet::apply() for details.
   void apply(sglCurrState *curr_state) const;

   // See sglObject::printInfo() for details.
   virtual void printInfo(ostream &ostrm, const char *indent_string) const;

private:
   void setIndex() { m_index = ((m_model == eFLAT) ? 1 : 0); }

   sglShadeModel(const sglShadeModel &); // not implemented
   sglShadeModel &operator=(const sglShadeModel &);// not implemented

protected:
   ShadeModelEnum m_model;
};

//============================================================================

/**

SGL supports use of more than one texture unit for OpenGL version 1.3 or later,
or earlier versions that implemented the GL_ARB_multitexture extension.  To
accomplish this all statelets that issue OpenGL commands to texture units are
subclassed from this one.

This class adds a pure virtual applyToTextureUnit() memeber function which the
subclasses must implement to apply their OpenGL texture elements to a given
texture unit.  For systems that do not have more than on texture units, then
these applies will revert to simpler OpenGL calls with an implied texture unit
of zero (that is, there is no call to set the active texture unit).  The
apply() method is implemented to direct all texture calls to the first texture
unit by default.

 */

class SGL_DLL_API sglTexturingStatelet : public sglStatelet
{
public:
   /** Constructor
       @param The sglStatelet::StateEnum corresponding to the subclass's
       statelet type.
   */
   sglTexturingStatelet(StateEnum type) : sglStatelet(type) {}

   /// virtual destructor
   virtual ~sglTexturingStatelet() {}

   // See sglStatelet::apply() for details.
   void apply(sglCurrState *curr_state) const { applyToUnit(curr_state, 0); }

   /** Apply this state change to the current rendering state for the given
       texture unit.
       @param curr_state Required for statelets that affect one another.  For
         example the sglTransparency sets up depth mask and blending modes
         which interact with the sglBlending and sglDepthMask statelets, and
         the curr_state object is used to resolve these internal issues.
       @param unit Index for the texture unit (0 is the first).
   */
   virtual void applyToUnit(sglCurrState *curr_state,
                            unsigned int unit) const = 0;
};

//============================================================================

/**

This statelet manages the dispatch of the various sglTexturingStatelet
subclasses (sglTexture, sglTexEnv, sglTexGen, and sglTextureMatrix) to the
different texture units.  Because this is handled differently from the
other (non-texture related) statelets the default constructor is not necessary
nor available.

<b>Example:</b>

The following example shows how to set up two texture units with different
textures.  Note how the sglTexCoords class is used to specify different texture
coordinates for the different units:

<pre>
   vector<sglStatelet*> state;
   vector<sglTexCoords> tex_coords;

   // setup for unit 0 (a 2D texture)
   sglTexture *tex0 = new sglTexture2D;
   tex0->setImage(...);
   sglTexEnv *tenv0 = new sglTexEnv;
   tenv0->setMode(...);
   sglTextureMatrix *tmat0 = new sglTextureMatrix(...);

   sglVec2f *tcoords0 = new sglVec2f[...];
   tex_coords.push_back(sglTexCoords((float*)tcoords0, 2));

   state->push_back(new sglMultiTextureStatelet(tex0, 0));
   state->push_back(new sglMultiTextureStatelet(tenv0, 0));
   state->push_back(new sglMultiTextureStatelet(tmat0, 0));

   // setup for unit 1 (a 1D texture)
   sglTexture *tex1 = new sglTexture1D;
   tex1->setImage(...);

   float *tcoords1 = new float[...];
   tex_coords.push_back(sglTexCoords(tcoords1, 1));

   state->push_back(new sglMultiTextureStatelet(tex1, 1));
   state->push_back(new sglMultiTextureStatelet(tgen0, 1));

   // some subclass of sglDrawable...
   geometry->setState(0, state, tex_coords);
</pre>

 */

class SGL_DLL_API sglMultiTextureStatelet : public sglStatelet
{
public:
   /** Constructor.
       @param statelet Pointer to the managed sglTexturingStatelet.
       @param unit Index of the texture unit to which to apply the statelet.
         This must be in the range [0 .. sgl::s_num_texture_units - 1] and
         defaults to zero.
   */
   sglMultiTextureStatelet(sglTexturingStatelet *statelet,
                           unsigned int unit = 0)
         : sglStatelet(eNUM_BASIC_STATES),
           m_statelet(NULL),
           m_texture_unit(unit)
      { setStatelet(statelet); }

   /** Virtual destructor.  This will dereference the contained statelet and if
       its reference count goes to zero it will be destructed as well.
   */
   virtual ~sglMultiTextureStatelet();

   /** Set the texture unit to which the contained sglTexturingStatelet object
       is applied.
       @param unit Index of the texture unit.  This should be in the range of
         [0 .. sgl::s_num_texture_units - 1].  (Values out if this range will
         result it being applied to unit 0?)
   */
   void setTextureUnit(unsigned int unit) { m_texture_unit = unit; }

   /** Query the current texture unit.
       @return The current texture unit in the range
         [0 .. sgl::s_num_texture_units - 1].
   */
   unsigned int getTextureUnit() const { return m_texture_unit; }

   /** Set the current sglTexturingStatelet object.  This will reference the
       new one (increment the reference count) and dereference any previous
       statelet.  If the previous statelet's reference count goes to zero it
       will NOT be deleted.
       @param statelet Pointer to the texturing statelet to be managed.
   */
   void setStatelet(sglTexturingStatelet *statelet);

   /** Query the current sglTexturingStateletObject.
       @return Pointer to the current statelet.
   */
   sglTexturingStatelet* getStatelet() const { return m_statelet; }

   // See sglStatelet::apply() for details.
   void apply(sglCurrState *curr_state) const;

   // See sglStatelet::getSorted() for details.
   virtual bool getSorted() const { return m_statelet->getSorted(); }

protected:
   // See sglStatelet::getType() for details.
   unsigned int getType() const;

private: // not implemented
   sglMultiTextureStatelet();
   sglMultiTextureStatelet(const sglMultiTextureStatelet &);
   sglMultiTextureStatelet &operator=(const sglMultiTextureStatelet &);

private:
   sglTexturingStatelet *m_statelet;
   unsigned int          m_texture_unit;
};

//============================================================================

/**

In SGL, the modelview matrix is affected by sglBillboard, sglUnProject,
sglUnScale, sglTransform, sglTranslate, and sglScale group nodes, and is
accumulated as the scene graph is traversed during the cull operation.
Currently, there is no analagous process for the texture matrix.  Instead, a
texture matrix can be set via the statelet mechanism using the sglTextureMatrix
class.  This statelet encapsulates the OpenGL's manipulation of the texture
matrix [glMatrixMode(GL_TEXTURE)] for the various texture units.
The matrix can be specified during construction by passing a pointer to an
sglMat4f that has already been set to the desired value (the value is copied
into the statelet).  Subsequent changes to the matrix can be done by calling
the sglTextureMatrix::setTextureMatrix(..).  The statelet can be disabled by
passing a NULL pointer when setting the matrix, or applying a matrix that is
identity.

<b>Example:</b>

<pre>
   sglMat4 mat;
   mat.buildRotation(0.f, 0.f, 1.f, M_PI*0.25f);

   sglTextureMatrix *tmat = new sglTextureMatrix(&mat);

   mat.buildTranslation(1.0, -2.0, 1.5);
   statelet->setTextureMatrix(&mat);
</pre>

 */

class SGL_DLL_API sglTextureMatrix : public sglTexturingStatelet
{
public:
   /** Constructor (also the default constructor).
       @param tmat Pointer to an sglMat4f object containing the value of the
         texture matrix.  If non-NULL, this matrix will get copied into the
         statelet.  If NULL, then the matrix resets to identity (the
         default).
   */
   sglTextureMatrix(sglMat4f *tmat = NULL)
      : sglTexturingStatelet(eTEXTURE_MATRIX) { setTextureMatrix(tmat); }

   /// Virtual destructor.
   virtual ~sglTextureMatrix() {}

   /** Set the current texture matrix value.  Texturing must be enabled (above)
       for this matrix to have an effect.
       @param tmat Pointer to an sglMat4f object containing the value of the
         texture matrix.  If non-NULL, this matrix will get copied into the
         statelet.  If NULL, then the matrix resets to identity.
   */
   void setTextureMatrix(sglMat4f *tmat);

   /** Query the current texture matrix stored in this statelet.
       @return Reference to the internally stored matrix.
   */
   const sglMat4f &getTextureMatrix() const { return m_texture_matrix; }

   // See sglTexturingStatelet::applyToUnit() for details.
   void applyToUnit(sglCurrState*, unsigned int) const;

   // See sglStatelet::getSorted() for details.
   virtual bool getSorted() const { return true; }

   // See sglObject::printInfo() for details.
   virtual void printInfo(ostream &ostrm, const char *indent_string) const;

private:
   sglTextureMatrix(const sglTextureMatrix &); // not implemented
   sglTextureMatrix &operator=(const sglTextureMatrix &); // not implemented

   void setIndex()
      {
         m_index = (m_texture_matrix_flag ? UINT_MAX : 0);
      }

private:
   bool m_texture_matrix_flag;
   sglMat4f m_texture_matrix;
};


//============================================================================

/**

This statelet controls how geometry with transparency are rendered in SGL.  If
this parameter in this statelet is true (enabled), then any subsequent geometry
that passes the cull test will placed in the "transparency bin" to be sorted
and rendered from back to front according to the distance to the centers of
their respective bounding boxes.  All geometry in this bin is rendered after
all non-transparent geometry with blending enabled using the blending function
from the sglBlending statelet (even if that statelet is disabled it will use
the current blend function stored within).  The geometry in the "transparent
bin" will also be rendered with the depth writing disabled (even if an
sglDepthMask is currently enabling it).

<b>Example:</b>
<pre>
   sglTransparency *transparent = new sglTransparency; // disabled by default
   transparent->enable(true);                          // enable it.
</pre>

 */

class SGL_DLL_API sglTransparency : public sglBooleanStatelet
{
public:
   /** Constructor (and the default constructor).
       @param on True to enable, false (default) to disable.
   */
   sglTransparency(bool on = false)
      : sglBooleanStatelet(eTRANSPARENCY, on) {}

   /// virtual destructor
   virtual ~sglTransparency() {}

   // See sglStatelet::apply() for details.
   void apply(sglCurrState *curr_state) const;

private:
   sglTransparency(const sglTransparency &); // not implemented
   sglTransparency &operator=(const sglTransparency &); // not implemented
};

#endif
