/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglStats.cpp
 *   Author: Tom Stimson
 *  Created: 2 April 2000
 *****************************************************************************/

#include <sglStats.hpp>
#include <sglTexture.hpp>

#include <algorithm>

const char* sglStats::s_geometry_type_names[] = {
   "Points",
   "Lines",
   "LineStrips",
   "Triangles",
   "Quads",
   "TriangleStrips",
   "TriangleFans",
   "QuadStrips",
   "Polygons",
   "Indexed Points",
   "Indexed Lines",
   "Indexed LineStrips",
   "Indexed Triangles",
   "Indexed Quads",
   "Indexed TriangleStrips",
   "Indexed TriangleFans",
   "Indexed QuadStrips",
   "Indexed Polygons",
   "Mono-Indexed Points",
   "Mono-Indexed Lines",
   "Mono-Indexed LineStrips",
   "Mono-Indexed Triangles",
   "Mono-Indexed Quads",
   "Mono-Indexed TriangleStrips",
   "Mono-Indexed TriangleFans",
   "Mono-Indexed QuadStrips",
   "Mono-Indexed Polygons",
   "Interleaved Points",
   "Interleaved Lines",
   "Interleaved LineStrips",
   "Interleaved LineLoops",
   "Interleaved Triangles",
   "Interleaved Quads",
   "Interleaved TriangleStrips",
   "Interleaved TriangleFans",
   "Interleaved QuadStrips",
   "Interleaved Polygons",
   "Bezier Curves",
   "Bezier Surfaces"
};


//----------------------------------------------------------------------------

sglStats::sglStats(const sglStats &rhs)
{
   *this = rhs;
}

//----------------------------------------------------------------------------

sglStats& sglStats::operator=(const sglStats &rhs)
{
   reset();
   *this += rhs;
   return *this;
}

//----------------------------------------------------------------------------

void sglStats::reset()
{
   m_transforms = 0;
   m_state_changes = 0;
   m_total_gsets = 0;
   m_dlisted_gsets = 0;
   m_transparency_enabled_gsets = 0;
   m_occlusion_enabled_gsets = 0;
   m_occluded_gsets = 0;
   m_lit_gsets = 0;
   m_textured_gsets = 0;
   m_num_textures = 0;
   m_ideal_tex_mem = 0;
   m_likely_tex_mem = 0;
   m_text = 0;
   for (unsigned int i=0; i<eNUM_PRIM_TYPES; ++i)
   {
      m_gsets[i] = 0;
      m_prims[i] = 0;
      m_vertices[i] = 0;
      m_triangles[i] = 0;
   }

   m_num_lights_used = 0;
   m_num_lights_desired = 0;

   m_frames = 0;
   m_textures.clear();
}

//----------------------------------------------------------------------------

sglStats& sglStats::operator+=(const sglStats &rhs)
{
   m_transforms += rhs.m_transforms;
   m_state_changes += rhs.m_state_changes;
   m_total_gsets += rhs.m_total_gsets;
   m_dlisted_gsets += rhs.m_dlisted_gsets;
   m_transparency_enabled_gsets += rhs.m_transparency_enabled_gsets;
   m_occlusion_enabled_gsets += rhs.m_occlusion_enabled_gsets;
   m_occluded_gsets += rhs.m_occluded_gsets;
   m_lit_gsets += rhs.m_lit_gsets;
   m_textured_gsets += rhs.m_textured_gsets;
   m_num_textures += rhs.m_num_textures;
   m_ideal_tex_mem += rhs.m_ideal_tex_mem;
   m_likely_tex_mem += rhs.m_likely_tex_mem;
   m_text += rhs.m_text;
   for (unsigned int i=0; i<eNUM_PRIM_TYPES; ++i)
   {
      m_gsets[i] += rhs.m_gsets[i];
      m_prims[i] += rhs.m_prims[i];
      m_vertices[i] += rhs.m_vertices[i];
      m_triangles[i] += rhs.m_triangles[i];
   }

   m_num_lights_used += rhs.m_num_lights_used;
   m_num_lights_desired += rhs.m_num_lights_desired;

   m_frames += rhs.m_frames;
   m_textures.clear();

   return *this;
}

//----------------------------------------------------------------------------

void sglStats::average()
{
   float factor = 1.0f / m_frames;
   m_transforms *= factor;
   m_state_changes *= factor;
   m_total_gsets *= factor;
   m_dlisted_gsets *= factor;
   m_transparency_enabled_gsets *= factor;
   m_occlusion_enabled_gsets *= factor;
   m_occluded_gsets *= factor;
   m_lit_gsets *= factor;
   m_textured_gsets *= factor;
   m_num_textures *= factor;
   m_ideal_tex_mem *= factor;
   m_likely_tex_mem *= factor;
   m_text *= factor;
   for (unsigned int i=0; i<eNUM_PRIM_TYPES; ++i)
   {
      m_gsets[i] *= factor;
      m_prims[i] *= factor;
      m_vertices[i] *= factor;
      m_triangles[i] *= factor;
   }

   m_num_lights_used *= factor;
   m_num_lights_desired *= factor;

   m_frames = 1;
   m_textures.clear();
}

//----------------------------------------------------------------------------

void sglStats::addTexture(const sglTexture *tex)
{
   vector<const sglTexture*>::iterator it =
      find(m_textures.begin(), m_textures.end(), tex);

   if (it == m_textures.end())
   {
      m_textures.push_back(tex);
      ++m_num_textures;
      m_ideal_tex_mem += tex->getIdealInternalSize();
      m_likely_tex_mem += tex->getLikelyInternalSize();
   }
}

//----------------------------------------------------------------------------

void sglStats::printInfo(ostream &ostrm, const char *indent_string) const
{
   ostrm << indent_string << "# frames = " << m_frames << endl;
   ostrm << indent_string << "# lights (used/desired) = " << m_num_lights_used
         << "/" << m_num_lights_desired << endl;
   ostrm << indent_string << "# transforms = " << m_transforms << endl;
   ostrm << indent_string << "# state changes = " << m_state_changes << endl;
   ostrm << indent_string << "# total gsets = " << m_total_gsets << endl;
   ostrm << indent_string << "# dlisted gsets = " << m_dlisted_gsets << endl;
   ostrm << indent_string << "# transparency enabled gsets = "
         << m_transparency_enabled_gsets << endl;
   ostrm << indent_string << "# occlusion enabled gsets = "
         << m_occlusion_enabled_gsets << endl;
   ostrm << indent_string << "# occluded gsets = "
         << m_occluded_gsets << endl;
   ostrm << indent_string << "# lit gsets = " << m_lit_gsets << endl;
   ostrm << indent_string << "# textured gsets = "
         << m_textured_gsets << endl;
   ostrm << indent_string << "# textures = " << m_num_textures << endl;
   ostrm << indent_string << "ideal texture memory (bytes)  = "
         << m_ideal_tex_mem << endl;
   ostrm << indent_string << "likely texture memory (bytes) = "
         << m_likely_tex_mem << endl;
   ostrm << indent_string << "# text strings = " << m_text << endl;
   ostrm << indent_string << "gsets\tprims\tverts\ttris\ttype" << endl;
   ostrm << indent_string << "-----\t-----\t-----\t----\t----" << endl;
   for (unsigned int i=0; i<eNUM_PRIM_TYPES; ++i)
   {
      if (m_gsets[i] || m_prims[i] || m_vertices[i])
      {
         ostrm << indent_string << m_gsets[i]
               << "\t" << m_prims[i]
               << "\t" << m_vertices[i]
               << "\t" << m_triangles[i]
               << "\t" << s_geometry_type_names[i] << endl;
      }
   }
}
