/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglSwitch.cpp
 *  Created: 14 November 1998
 *  Summary: selects which child (or none or all children) to be traversed
 *****************************************************************************/

#include <sgl.h>
#include <sglSwitch.hpp>
#include <sglCull.hpp>
#include <sglOStream.hpp>

//----------------------------------------------------------------------------
sglSwitch::sglSwitch() : sglGroup(),
   m_switch_val(sglSwitch::eON)   // defaults to a sglGroup
{
}

//----------------------------------------------------------------------------
sglSwitch::~sglSwitch()
{
}

//----------------------------------------------------------------------------
sglBound::IntersectResultEnum sglSwitch::intersect(sglIntersectf &segset) const
{
   if ((segset.getMode()&sglProbe::eSWITCH_NONE) ||
       (m_switch_val == sglSwitch::eOFF) ||
       m_child_list.empty())
   {
      return sglBound::eOUTSIDE;
   }
   else if ((segset.getMode()&sglProbe::eSWITCH_ALL) ||
            (m_switch_val == sglSwitch::eON))
   {
      return sglGroup::intersect(segset);
   }
   else // default: current child selected by switch only
   {
      sglSegmentf seg(segset.getSegment());
      sglBound::IntersectResultEnum result = m_sphere.intersect(seg);
      if (result == sglBound::eOUTSIDE)
      {
         return result;
      }

      result = sglBound::eOUTSIDE;
      if ((m_switch_val >=0) && (m_switch_val < (int) getNumChildren()))
      {
         if (m_child_list[m_switch_val]->intersect(segset) ==
             sglBound::eINTERSECT)
         {
            result = sglBound::eINTERSECT;
         }
      }
      else
      {
         sglPrintWarning()
            << "sglSwitch::intersect() error: invalid switch value "
            << m_switch_val << endl;
      }

      if (result == sglBound::eINTERSECT &&
          segset.getMode() & sglProbe::ePATH)
      {
         segset.getPath().push_front((sglNode*)this);
      }
      return result;
   }
}

//----------------------------------------------------------------------------
sglBound::IntersectResultEnum sglSwitch::intersect(sglIntersectd &segset) const
{
   if ((segset.getMode()&sglProbe::eSWITCH_NONE) ||
       (m_switch_val == sglSwitch::eOFF) ||
       m_child_list.empty())
   {
      return sglBound::eOUTSIDE;
   }
   else if ((segset.getMode()&sglProbe::eSWITCH_ALL) ||
            (m_switch_val == sglSwitch::eON))
   {
      return sglGroup::intersect(segset);
   }
   else // default: current child selected by switch only
   {
      sglSegmentd seg(segset.getSegment());
      sglBound::IntersectResultEnum result = m_sphere.intersect(seg);
      if (result == sglBound::eOUTSIDE)
      {
         return result;
      }

      result = sglBound::eOUTSIDE;
      if ((m_switch_val >=0) && (m_switch_val < (int) getNumChildren()))
      {
         if (m_child_list[m_switch_val]->intersect(segset) ==
             sglBound::eINTERSECT)
         {
            result = sglBound::eINTERSECT;
         }
      }
      else
      {
         sglPrintWarning()
            << "sglSwitch::intersect() error: invalid switch value "
            << m_switch_val << endl;
      }

      if (result == sglBound::eINTERSECT &&
          segset.getMode() & sglProbe::ePATH)
      {
         segset.getPath().push_front((sglNode*)this);
      }
      return result;
   }
}

//----------------------------------------------------------------------------
void sglSwitch::pick(sglPickf &pick_state, unsigned int cull_flags) const
{
   if ((pick_state.getMode()&sglProbe::eSWITCH_NONE) ||
       (m_switch_val == sglSwitch::eOFF) ||
       m_child_list.empty())
   {
      return;  // select nothing
   }
   else if ((pick_state.getMode()&sglProbe::eSWITCH_ALL) ||
            (m_switch_val == sglSwitch::eON))
   {
      sglGroup::pick(pick_state, cull_flags);
   }
   else // default: current child selected by switch only
   {
      if (cull_flags && (m_sphere.intersect(pick_state.getPickFrustum(),
                                            cull_flags) == sglBound::eOUTSIDE))
      {
         return; // select nothing
      }

      if ((m_switch_val >=0) && (m_switch_val < (int) getNumChildren()))
      {
         // if path is requested store it ahead of time (different from
         // intersect), in case something beneath here is picked successfully.
         if (pick_state.getMode() & sglProbe::ePATH)
         {
            pick_state.getPath().push_back((sglNode*)this);
         }

         m_child_list[m_switch_val]->pick(pick_state, cull_flags);

         if (pick_state.getMode() & sglProbe::ePATH)
         {
            pick_state.getPath().pop_back();
         }
      }
      else
      {
         sglPrintWarning() << "sglSwitch::pick() error: invalid switch value "
                           << m_switch_val << endl;
      }
   }
}

//----------------------------------------------------------------------------
void sglSwitch::pick(sglPickd &pick_state, unsigned int cull_flags) const
{
   if ((pick_state.getMode()&sglProbe::eSWITCH_NONE) ||
       (m_switch_val == sglSwitch::eOFF) ||
       m_child_list.empty())
   {
      return;  // select nothing
   }
   else if ((pick_state.getMode()&sglProbe::eSWITCH_ALL) ||
            (m_switch_val == sglSwitch::eON))
   {
      sglGroup::pick(pick_state, cull_flags);
   }
   else // default: current child selected by switch only
   {
      if (cull_flags && (m_sphere.intersect(pick_state.getPickFrustum(),
                                            cull_flags) == sglBound::eOUTSIDE))
      {
         return; // select nothing
      }

      if ((m_switch_val >=0) && (m_switch_val < (int) getNumChildren()))
      {
         // if path is requested store it ahead of time (different from
         // intersect), in case something beneath here is picked successfully.
         if (pick_state.getMode() & sglProbe::ePATH)
         {
            pick_state.getPath().push_back((sglNode*)this);
         }

         m_child_list[m_switch_val]->pick(pick_state, cull_flags);

         if (pick_state.getMode() & sglProbe::ePATH)
         {
            pick_state.getPath().pop_back();
         }
      }
      else
      {
         sglPrintWarning() << "sglSwitch::pick() error: invalid switch value "
                           << m_switch_val << endl;
      }
   }
}

//----------------------------------------------------------------------------
void sglSwitch::cull(sglCull<float> &trav_state,
                     unsigned int cull_flags, int switch_val) const
{
   if (switch_val == sglSwitch::eOFF || m_child_list.empty())
   {
      return; // draw nothing
   }
   else if (switch_val == sglSwitch::eON)
   {
      sglGroup::cull(trav_state, cull_flags);
   }
   else
   {
      if (cull_flags && (m_sphere.intersect(trav_state.getFrustum(),
                                            cull_flags) == sglBound::eOUTSIDE))
      {
         return; // draw nothing
      }

      if ((switch_val >=0) && (switch_val < (int) getNumChildren()))
      {
         m_child_list[switch_val]->cull(trav_state, cull_flags);
      }
      else
      {
         sglPrintWarning() << "sglSwitch::cull() error: invalid switch value "
                           << switch_val << endl;
      }
   }
}

//----------------------------------------------------------------------------
void sglSwitch::cull(sglCull<double> &trav_state,
                     unsigned int cull_flags, int switch_val) const
{
   if (switch_val == sglSwitch::eOFF || m_child_list.empty())
   {
      return; // draw nothing
   }
   else if (switch_val == sglSwitch::eON)
   {
      sglGroup::cull(trav_state, cull_flags);
   }
   else
   {
      if (cull_flags && (m_sphere.intersect(trav_state.getFrustum(),
                                            cull_flags) == sglBound::eOUTSIDE))
      {
         return; // draw nothing
      }

      if ((switch_val >=0) && (switch_val < (int) getNumChildren()))
      {
         m_child_list[switch_val]->cull(trav_state, cull_flags);
      }
      else
      {
         sglPrintWarning() << "sglSwitch::cull() error: invalid switch value "
                           << switch_val << endl;
      }
   }
}

//----------------------------------------------------------------------------
sglNode *sglSwitch::clone(unsigned int mode) const
{
   sglSwitch *root = new sglSwitch;
   sglSwitch::copyTo(root, mode);
   return root;
}

//----------------------------------------------------------------------------
void sglSwitch::copyTo(sglSwitch *dst, unsigned int mode) const
{
   sglGroup::copyTo(dst, mode);

   // transfer stuff
   dst->setVal(m_switch_val);
}

//----------------------------------------------------------------------------
void sglSwitch::printInfo(ostream &ostrm, const char *indent_string) const
{
    sglGroup::printInfo(ostrm, indent_string);

    ostrm << indent_string << "sglSwitch val = " << m_switch_val << endl;
}
