/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglSwitch.hpp
 *  Created: 14 November 1998
 *  Summary: select (or not) the children to be traversed
 *****************************************************************************/

#ifndef __SGL_SWITCH_HPP
#define __SGL_SWITCH_HPP

#include <sgl.h>
#include <sglGroup.hpp>

//----------------------------------------------------------------------------

/**

This node will select (or not) the child or children to be traversed.
For this subclass of sglGroup, the order that the children are added to
the sglGroup is significant.  The value of the switch determines which
(if any) children get traversed according to the following table:

<pre>
   Switch Value       Traversal Behaviour
   ---------------    ----------------------------------------------
   sglSwitch::eOFF    do not traversal any children.
   sglSwitch::eON     traversal all the children (as if an sglGroup)
   0 .. MAX_INT       traverse only one child whose index equals the
                          switch value
</pre>

The switch value is a signed integer (because the eOFF and eON tokens are
negative values).  That means it can be used to select one of up to
MAX_INT children.  If the switch value is out of range of the number
of children currently referenced by this node, then the traversal stops
at this point in the branch, a warning message is printed, and the traversal
function returns.

<b>Special behaviour for cull traversals</b>

The protected cull functions have been rewritten as thin wrappers around
reentrant cull functions that take the switch value as parameter.  The default
implementation in this class is to pass the switch value directly to these
reentrant functions.  This node can be subclassed, however, and the cull
functions can pass different values in as the switch value parameter to create
more complex behaviours.

<b>Special behaviour for intersect and pick traversals</b>

In addition to this behaviour, the intersect and pick traversal behaviour for
this node can be changed by adding sglProbe::eSWITCH_ALL or
sglProbe::eSWITCH_NONE to its mode.  The former behaves as if sglSwitch::eON
has been set on this node (all nodes are traversed), and the latter corresponds
to sglSwitch::eOFF (no nodes traversed).

<b>Example</b>

<pre>
   sglGeode  *geode1 = new sglGeode;
   sglGeode  *geode2 = new sglGeode;
   sglSwitch *switch_node = new sglSwitch;
   switch_node->addChild(geode1);
   switch_node->addChild(geode2);

   switch_node->setVal(sglSwitch::eOFF);  // select neither
   switch_node->setVal(sglSwitch::eON);   // select both  (the default)
   switch_node->setVal(0);                // select geode1
   switch_node->setVal(1);                // select geode2
   switch_node->setVal(2);                // ERROR: will select neither
</pre>

\todo Should the pick and intersect functions be implemented as wrappers to
reentrant functions like cull?

 */

//----------------------------------------------------------------------------

class SGL_DLL_API sglSwitch : public sglGroup
{
public:
   /// Some useful switch value constants scoped to this class
   enum
   {
      /// Select no children
      eOFF = -2,

      /// Select all children
      eON  = -1
   };

public:
   /// default constructor
   sglSwitch();
   /// virtual destructor
   virtual ~sglSwitch();

   /** Set the switch value (defaults to eON).
       @param val The switch value which can be one of eOff, eOn, 0..MAX_INT.
   */
   void setVal(int val) {m_switch_val = val;}

   /** Get the current switch value.
       @returns current switch value
   */
   int  getVal() const  {return m_switch_val;}

   // See sglNode::intersect(f) for details.
   virtual sglBound::IntersectResultEnum intersect(
      sglIntersectf &isector) const;

   // See sglNode::intersect(d) for details.
   virtual sglBound::IntersectResultEnum intersect(
      sglIntersectd &isector) const;

   // See sglNode::pick(f) for details.
   virtual void pick(sglPickf &pick_state, unsigned int cull_flags) const;

   // See sglNode::pick(d) for details.
   virtual void pick(sglPickd &pick_state, unsigned int cull_flags) const;

   // See sglNode::clone() for details.
   virtual sglNode *clone(unsigned int mode) const;

   // See sglObject::printInfo() for details.
   virtual void printInfo(ostream &ostrm, const char *indent_string) const;

protected:
   // See sglNode::cull(f) for details
   virtual void cull(sglCull<float> &trav_state,
                     unsigned int cull_flags) const
      {
         cull(trav_state, cull_flags, m_switch_val);
      }

   // See sglNode::cull(d) for details
   virtual void cull(sglCull<double> &trav_state,
                     unsigned int cull_flags) const
      {
         cull(trav_state, cull_flags, m_switch_val);
      }

   // reentrant versions of the draw{Culled} functions for multithreading
   void cull(sglCull<float> &trav_state,
             unsigned int cull_flags, int switch_val) const;
   void cull(sglCull<double> &trav_state,
             unsigned int cull_flags, int switch_val) const;

   void copyTo(sglSwitch *dst, unsigned int mode) const;

private:
   sglSwitch(const sglSwitch &);            // not implemented
   sglSwitch &operator=(const sglSwitch &); // not implemented

protected:
   int m_switch_val;
};

#endif
