/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglTexCombinerParams.cpp
 *   Author: Tom Stimson
 *  Created: 31 December 2001
 *****************************************************************************/

#include <sglTexCombinerParams.hpp>

//----------------------------------------------------------------------------
sglTexCombinerParams::sglTexCombinerParams()
   : sglObject(),
     m_rgb_func(RGB::eMODULATE),
     m_alpha_func(Alpha::eMODULATE),
     m_rgb_scale(1.0f),
     m_alpha_scale(1.0f),
     m_rgb_num_args(2),
     m_alpha_num_args(2)
{
   m_rgb_src[0] = m_alpha_src[0] = eTEXTURE;
   m_rgb_src[1] = m_alpha_src[1] = ePREVIOUS;
   m_rgb_src[2] = m_alpha_src[2] = eCONSTANT;
   m_rgb_op[0] = m_rgb_op[1] = RGB::eSRC_COLOR;
   m_rgb_op[2] = RGB::eSRC_ALPHA;
   m_alpha_op[0] = m_alpha_op[1] = m_alpha_op[2] = Alpha::eSRC_ALPHA;
}

//----------------------------------------------------------------------------
sglTexCombinerParams::~sglTexCombinerParams()
{
}

//----------------------------------------------------------------------------
sglTexCombinerParams::sglTexCombinerParams(const sglTexCombinerParams &rhs)
{
   *this = rhs;
}

//----------------------------------------------------------------------------
sglTexCombinerParams &sglTexCombinerParams::operator=(
   const sglTexCombinerParams &rhs)
{
   if (this != &rhs)
   {
      m_rgb_func = rhs.m_rgb_func;
      m_alpha_func = rhs.m_alpha_func;
      m_rgb_scale = rhs.m_rgb_scale;
      m_alpha_scale = rhs.m_alpha_scale;
      m_rgb_num_args = rhs.m_rgb_num_args;
      m_alpha_num_args = rhs.m_alpha_num_args;
      for (unsigned int i=0; i<3; ++i)
      {
         m_rgb_src[i] = rhs.m_rgb_src[i];
         m_alpha_src[i] = rhs.m_alpha_src[i];
         m_rgb_op[i] = rhs.m_rgb_op[i];
         m_alpha_op[i] = rhs.m_alpha_op[i];
      }
   }
   return *this;
}

//----------------------------------------------------------------------------
void sglTexCombinerParams::setRGBFunc(RGB::FuncEnum func)
{
   if ((func == RGB::eDOT3_RGB || func == RGB::eDOT3_RGBA) &&
       !sgl::s_texture_env_dot3_capable)
   {
      // Reset to default of modulate, if dot3 not supported
      m_rgb_func = RGB::eMODULATE;
   }
   else
   {
      m_rgb_func = func;
   }

   if (m_rgb_func == RGB::eREPLACE)
      m_rgb_num_args = 1;
   else if (m_rgb_func == RGB::eINTERPOLATE)
      m_rgb_num_args = 3;
   else
      m_rgb_num_args = 2;
}

//----------------------------------------------------------------------------
void sglTexCombinerParams::setAlphaFunc(Alpha::FuncEnum func)
{
   m_alpha_func = func;

   if (m_alpha_func == Alpha::eREPLACE)
      m_alpha_num_args = 1;
   else if (m_alpha_func == Alpha::eINTERPOLATE)
      m_alpha_num_args = 3;
   else
      m_alpha_num_args = 2;
}

//----------------------------------------------------------------------------
void sglTexCombinerParams::setRGBScale(float scale)
{
   if (scale <= 1.0f)
      m_rgb_scale = 1.0f;
   else if (scale >= 4.0f)
      m_rgb_scale = 4.0f;
   else
      m_rgb_scale = 2.0f;
}

//----------------------------------------------------------------------------
void sglTexCombinerParams::setAlphaScale(float scale)
{
   if (scale <= 1.0f)
      m_alpha_scale = 1.0f;
   else if (scale >= 4.0f)
      m_alpha_scale = 4.0f;
   else
      m_alpha_scale = 2.0f;
}

//----------------------------------------------------------------------------
void sglTexCombinerParams::setRGBParameter(
   ArgEnum arg, SourceEnum src, RGB::OperandEnum op)
{
   m_rgb_src[arg] = src;
   m_rgb_op[arg] = op;
}

//----------------------------------------------------------------------------
void sglTexCombinerParams::getRGBParameter(
   ArgEnum arg, SourceEnum &src, RGB::OperandEnum &op) const
{
   src = m_rgb_src[arg];
   op = m_rgb_op[arg];
}

//----------------------------------------------------------------------------
void sglTexCombinerParams::setAlphaParameter(
   ArgEnum arg, SourceEnum src, Alpha::OperandEnum op)
{
   m_alpha_src[arg] = src;
   m_alpha_op[arg] = op;
}

//----------------------------------------------------------------------------
void sglTexCombinerParams::getAlphaParameter(
   ArgEnum arg, SourceEnum &src, Alpha::OperandEnum &op) const
{
   src = m_alpha_src[arg];
   op = m_alpha_op[arg];
}

//----------------------------------------------------------------------------
void sglTexCombinerParams::apply() const
{
   glTexEnvf(GL_TEXTURE_ENV, GL_COMBINE_RGB, m_rgb_func);
   glTexEnvf(GL_TEXTURE_ENV, GL_RGB_SCALE, m_rgb_scale);
   for (int i=0; i<m_rgb_num_args; ++i)
   {
      glTexEnvf(GL_TEXTURE_ENV, GL_SOURCE0_RGB+i, m_rgb_src[i]);
      glTexEnvf(GL_TEXTURE_ENV, GL_OPERAND0_RGB+i, m_rgb_op[i]);
   }

   glTexEnvf(GL_TEXTURE_ENV, GL_COMBINE_ALPHA, m_alpha_func);
   glTexEnvf(GL_TEXTURE_ENV, GL_ALPHA_SCALE, m_alpha_scale);
   for (int j=0; j<m_alpha_num_args; ++j)
   {
      glTexEnvf(GL_TEXTURE_ENV, GL_SOURCE0_ALPHA+j, m_alpha_src[j]);
      glTexEnvf(GL_TEXTURE_ENV, GL_OPERAND0_ALPHA+j, m_alpha_op[j]);
   }
}

//----------------------------------------------------------------------------
void sglTexCombinerParams::printInfo(
   ostream &ostrm, const char *indent_string) const
{
   ostrm << indent_string << "rgb func: ";
   switch (m_rgb_func)
   {
      case RGB::eREPLACE:
         ostrm << "eREPLACE" << endl;
         break;
      case RGB::eMODULATE:
         ostrm << "eMODULATE" << endl;
         break;
      case RGB::eADD:
         ostrm << "eADD" << endl;
         break;
      case RGB::eADD_SIGNED:
         ostrm << "eADD_SIGNED" << endl;
         break;
      case RGB::eINTERPOLATE:
         ostrm << "eINTERPOLATE" << endl;
         break;
      case RGB::eSUBTRACT:
         ostrm << "eSUBTRACT" << endl;
         break;
      case RGB::eDOT3_RGB:
         ostrm << "eDOT3_RGB" << endl;
         break;
      case RGB::eDOT3_RGBA:
         ostrm << "eDOT3_RGBA" << endl;
         break;
   }
   ostrm << indent_string << "alpha func: ";
   switch (m_alpha_func)
   {
      case Alpha::eREPLACE:
         ostrm << "eREPLACE" << endl;
         break;
      case Alpha::eMODULATE:
         ostrm << "eMODULATE" << endl;
         break;
      case Alpha::eADD:
         ostrm << "eADD" << endl;
         break;
      case Alpha::eADD_SIGNED:
         ostrm << "eADD_SIGNED" << endl;
         break;
      case Alpha::eINTERPOLATE:
         ostrm << "eINTERPOLATE" << endl;
         break;
      case Alpha::eSUBTRACT:
         ostrm << "eSUBTRACT" << endl;
         break;
   }
   ostrm << indent_string << "rgb scale: " << m_rgb_scale << endl;
   ostrm << indent_string << "alpha  scale: " << m_alpha_scale << endl;
   ostrm << indent_string << "rgb num args: " << m_rgb_num_args << endl;
   ostrm << indent_string << "alpha  num args: " << m_alpha_num_args << endl;
   for (unsigned int i=0; i<3; ++i)
   {
      ostrm << indent_string << "RGB ARG" << i << ": ";
      switch (m_rgb_src[i])
      {
         case eTEXTURE:
            ostrm << "eTEXTURE";
            break;
         case eCONSTANT:
            ostrm << "eCONSTANT";
            break;
         case ePRIMARY_COLOR:
            ostrm << "ePRIMARY_COLOR";
            break;
         case ePREVIOUS:
            ostrm << "ePREVIOUS";
            break;
      }
      ostrm << ", ";
      switch (m_rgb_op[i])
      {
         case RGB::eSRC_COLOR:
            ostrm << "eSRC_COLOR" << endl;
            break;
         case RGB::eONE_MINUS_SRC_COLOR:
            ostrm << "eONE_MINUS_SRC_COLOR" << endl;
            break;
         case RGB::eSRC_ALPHA:
            ostrm << "eSRC_ALPHA" << endl;
            break;
         case RGB::eONE_MINUS_SRC_ALPHA:
            ostrm << "eONE_MINUS_SRC_ALPHA" << endl;
            break;
      }
   }
   for (unsigned int j=0; j<3; ++j)
   {
      ostrm << indent_string << "Alpha ARG" << j << ": ";
      switch (m_alpha_src[j])
      {
         case eTEXTURE:
            ostrm << "eTEXTURE";
            break;
         case eCONSTANT:
            ostrm << "eCONSTANT";
            break;
         case ePRIMARY_COLOR:
            ostrm << "ePRIMARY_COLOR";
            break;
         case ePREVIOUS:
            ostrm << "ePREVIOUS";
            break;
      }
      ostrm << ", ";
      switch (m_alpha_op[j])
      {
         case Alpha::eSRC_ALPHA:
            ostrm << "eSRC_ALPHA" << endl;
            break;
         case Alpha::eONE_MINUS_SRC_ALPHA:
            ostrm << "eONE_MINUS_SRC_ALPHA" << endl;
            break;
      }
   }
}
