/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglTexCombinerParams.hpp
 *   Author: Tom Stimson
 *  Created: 31 December 2001
 *  Summary: Paramters to control the texture environment mode eCOMBINE.
 *****************************************************************************/

#ifndef __SGL_TEXCOMBINERPARAMS_HPP
#define __SGL_TEXCOMBINERPARAMS_HPP

#include <sgl.h>
#include <sglObject.hpp>

class sglTexEnv;

//============================================================================

/**

This object contains all the information required to define the
functions used by the sglTexEnv::eCOMBINE mode.

<b>Example 1:</b>  To emulate the eBLEND mode,
<pre>
   sglTexEnv *tenv = new sglTexEnv;
   tenv->setBlendColor(1.0f, 0.7f, 0.0f, 1.0f);
   tenv->setMode(sglTexEnv::eCOMBINE);
   combiner_params = new sglTexCombinerParams();
   combiner_params->setRGBFunc(sglTexCombinerParams::RGB::eINTERPOLATE);
   combiner_params->setRGBParameter(sglTexCombinerParams::eARG0,
                                    sglTexCombinerParams::eCONSTANT,
                                    sglTexCombinerParams::RGB::eSRC_COLOR);
   combiner_params->setRGBParameter(sglTexCombinerParams::eARG1,
                                    sglTexCombinerParams::ePREVIOUS,
                                    sglTexCombinerParams::RGB::eSRC_COLOR);
   combiner_params->setRGBParameter(sglTexCombinerParams::eARG2,
                                    sglTexCombinerParams::eTEXTURE,
                                    sglTexCombinerParams::RGB::eSRC_COLOR);
   combiner_params->setAlphaFunc(sglTexCombinerParams::Alpha::eMODULATE);
   combiner_params->setAlphaParameter(
      sglTexCombinerParams::eARG0,
      sglTexCombinerParams::ePREVIOUS,
      sglTexCombinerParams::Alpha::eSRC_ALPHA);
   combiner_params->setAlphaParameter(
      sglTexCombinerParams::eARG1,
      sglTexCombinerParams::eTEXTURE,
      sglTexCombinerParams::Alpha::eSRC_ALPHA);
   tenv->setCombinerParams(combiner_params);
</pre>

<b>Example 2:</b>  To emulate the eDECAL mode,
<pre>
   sglTexEnv *tenv = new sglTexEnv;
   tenv->setMode(sglTexEnv::eCOMBINE);
   combiner_params = new sglTexCombinerParams();
   combiner_params->setRGBFunc(sglTexCombinerParams::RGB::eINTERPOLATE);
   combiner_params->setRGBParameter(sglTexCombinerParams::eARG0,
                                    sglTexCombinerParams::eTEXTURE,
                                    sglTexCombinerParams::RGB::eSRC_COLOR);
   combiner_params->setRGBParameter(sglTexCombinerParams::eARG1,
                                    sglTexCombinerParams::ePREVIOUS,
                                    sglTexCombinerParams::RGB::eSRC_COLOR);
   combiner_params->setRGBParameter(sglTexCombinerParams::eARG2,
                                    sglTexCombinerParams::eTEXTURE,
                                    sglTexCombinerParams::RGB::eSRC_ALPHA);
   combiner_params->setAlphaFunc(sglTexCombinerParams::Alpha::eREPLACE);
   combiner_params->setAlphaParameter(
      sglTexCombinerParams::eARG0,
      sglTexCombinerParams::ePREVIOUS,
      sglTexCombinerParams::Alpha::eSRC_ALPHA);
   tenv->setCombinerParams(combiner_params);
</pre>

*/

class SGL_DLL_API sglTexCombinerParams : public sglObject
{
public:
   class SGL_DLL_API RGB
   {
   public:
      /// Tokens for the various RGB texture functions.
      enum FuncEnum
      {
         /// Arg0
         eREPLACE      = GL_REPLACE,

         /// Arg0 * Arg1
         eMODULATE     = GL_MODULATE,

         /// Arg0 + Arg1
         eADD          = GL_ADD,

         /// Arg0 + Arg1 - 0.5
         eADD_SIGNED   = GL_ADD_SIGNED,

         /// Arg0 * Arg2 + Arg1 * (1 - Arg2)
         eINTERPOLATE  = GL_INTERPOLATE,

         /// Arg0 - Arg1
         eSUBTRACT     = GL_SUBTRACT,

         /** 4 X ((Arg0r - 0.5) * (Arg1r - 0.5) +
                  (Arg0g - 0.5) * (Arg1g - 0.5) +
                  (Arg0b - 0.5) * (Arg1b - 0.5))
         */
         eDOT3_RGB     = GL_DOT3_RGB,

         /** 4 X ((Arg0r - 0.5) * (Arg1r - 0.5) +
                  (Arg0g - 0.5) * (Arg1g - 0.5) +
                  (Arg0b - 0.5) * (Arg1b - 0.5))
             Note that the result generated from the Alpha function is
             ignored for eDOT3_RGBA.
         */
         eDOT3_RGBA    = GL_DOT3_RGBA
      };

      enum OperandEnum
      {
         eSRC_COLOR           = GL_SRC_COLOR,
         eONE_MINUS_SRC_COLOR = GL_ONE_MINUS_SRC_COLOR,
         eSRC_ALPHA           = GL_SRC_ALPHA,
         eONE_MINUS_SRC_ALPHA = GL_ONE_MINUS_SRC_ALPHA
      };
   };

   class SGL_DLL_API Alpha
   {
   public:
      /// Tokens for the various Alpha texture functions.
      enum FuncEnum
      {
         /// Arg0
         eREPLACE      = GL_REPLACE,

         /// Arg0 * Arg1
         eMODULATE     = GL_MODULATE,

         /// Arg0 + Arg1
         eADD          = GL_ADD,

         /// Arg0 + Arg1 - 0.5
         eADD_SIGNED   = GL_ADD_SIGNED,

         /// Arg0 * Arg2 + Arg1 * (1 - Arg2)
         eINTERPOLATE  = GL_INTERPOLATE,

         /// Arg0 - Arg1
         eSUBTRACT     = GL_SUBTRACT
      };

      enum OperandEnum
      {
         eSRC_ALPHA           = GL_SRC_ALPHA,
         eONE_MINUS_SRC_ALPHA = GL_ONE_MINUS_SRC_ALPHA
      };
   };

   enum ArgEnum
   {
      eARG0 = 0,
      eARG1,
      eARG2
   };

   /// Tokens to specify the source for each argument of the texture functions.
   enum SourceEnum
   {
      /// Input color for the given Arg comes from the current texture.
      eTEXTURE        = GL_TEXTURE,

      /// Input color for the given Arg comes from the TexEnv blend color.
      eCONSTANT       = GL_CONSTANT,

      /// Input color for the given Arg comes from the base polygon.
      ePRIMARY_COLOR  = GL_PRIMARY_COLOR,

      /** Input color for the given Arg comes from the output of the previous
          texture unit.  For the first texture unit, ePREVIOUS behaves like
          ePRIMARY_COLOR.
      */
      ePREVIOUS       = GL_PREVIOUS
   };

public:
   /** Default constructor.
   */
   sglTexCombinerParams();

   /// Virtual destructor
   virtual ~sglTexCombinerParams();

   /// Copy constructor.
   sglTexCombinerParams(const sglTexCombinerParams &rhs);

   /// Assignment operator.
   sglTexCombinerParams &operator=(const sglTexCombinerParams &rhs);

   /** Set the texture function for the RGB components.
       @param func One of the RGB::FuncEnum tokens.  If eDOT3_RGB or
         eDOT3_RGBA is specified but is not supported through the extensions,
         then eMODULATE will be used instead.
         The default value is eMODULATE.
   */
   void setRGBFunc(RGB::FuncEnum func);

   /** Query the current texture function for the RGB components.
       @return The RGB::FuncEnum token corresponding to the current texture
         function.
   */
   RGB::FuncEnum getRGBFunc() const { return m_rgb_func; }

   /** Set the texture function for the Alpha component.
       @param func One of the Alpha::FuncEnum tokens.
         The default value is eMODULATE.
   */
   void setAlphaFunc(Alpha::FuncEnum func);

   /** Query the current texture function for the Alpha component.
       @return The Alpha::FuncEnum token corresponding to the current texture
         function.
   */
   Alpha::FuncEnum getAlphaFunc() const { return m_alpha_func; }

   /** The result of the RGB texture function may be scaled by 1.0, 2.0 or 4.0.
       @param scale The scale factor to use on the RGB components.
         The default value is 1.0f.
   */
   void setRGBScale(float scale);

   /** Query the current RGB scale factor.
       @return The current RGB scale factor.
   */
   float getRGBScale() const { return m_rgb_scale; }

   /** The result of the Alpha texture function may be scaled
       by 1.0, 2.0 or 4.0.
       @param scale The scale factor to use on the Alpha components.
   */
   void setAlphaScale(float scale);

   /** Query the current Alpha scale factor.
       @return The current Alpha scale factor.
         The default value is 1.0f.
   */
   float getAlphaScale() const { return m_alpha_scale; }

   /** Change one of the arguments of the RGB texture function.
       The default values:
         ARG0 = eTEXTURE, eSRC_COLOR;
         ARG1 = ePREVIOUS, eSRC_COLOR;
         ARG2 = eCONSTANT, eSRC_ALPHA;
       @param arg The argument to be changed.
       @param src The source of the color values.
       @param op The component of the source.
   */
   void setRGBParameter(ArgEnum arg, SourceEnum src, RGB::OperandEnum op);

   /** Query one of the arguments of the RGB texture function.
       @param arg The argument to be queried.
       @param src On return set to the source of the color values.
       @param op One return set to the component of the source.
   */
   void getRGBParameter(
      ArgEnum arg, SourceEnum &src, RGB::OperandEnum &op) const;

   /** Change one of the arguments of the Alpha texture function.
       The default values:
         ARG0 = eTEXTURE, eSRC_ALPHA;
         ARG1 = ePREVIOUS, eSRC_ALPHA;
         ARG2 = eCONSTANT, eSRC_ALPHA;
       @param arg The argument to be changed.
       @param src The source of the color values.
       @param op The component of the source.
   */
   void setAlphaParameter(ArgEnum arg, SourceEnum src, Alpha::OperandEnum op);

   /** Query one of the arguments of the Alpha texture function.
       @param arg The argument to be queried.
       @param src On return set to the source of the color values.
       @param op One return set to the component of the source.
   */
   void getAlphaParameter(
      ArgEnum arg, SourceEnum &src, Alpha::OperandEnum &op) const;

   // See sglObject::printInfo() for details.
   virtual void printInfo(ostream &ostrm, const char *indent_string) const;

private:
   friend class sglTexEnv;
   void apply() const;

private:
   RGB::FuncEnum       m_rgb_func;
   Alpha::FuncEnum     m_alpha_func;
   float               m_rgb_scale;
   float               m_alpha_scale;
   int                 m_rgb_num_args;
   SourceEnum          m_rgb_src[3];
   RGB::OperandEnum    m_rgb_op[3];
   int                 m_alpha_num_args;
   SourceEnum          m_alpha_src[3];
   Alpha::OperandEnum  m_alpha_op[3];
};

#endif
