/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglTexEnv.cpp
 *  Created: 2 May 1999
 *  Summary: create, modify and query texture environment
 *****************************************************************************/

#include <sglTexEnv.hpp>
#include <sglTexCombinerParams.hpp>

//----------------------------------------------------------------------------
sglTexEnv::sglTexEnv()
      : sglTexturingStatelet(eTEX_ENV),
        m_mode(sglTexEnv::eMODULATE),
        m_blend_color(0.0f,0.0f,0.0f,0.0f),
        m_lod_bias(0.0f),
        m_combiner_params(NULL)
{
   setIndex();
}

//----------------------------------------------------------------------------
sglTexEnv::~sglTexEnv()
{
   if (m_combiner_params)
   {
      m_combiner_params->deref();
      if (m_combiner_params->getRefCount() == 0)
      {
         delete m_combiner_params;
      }
   }
}

//----------------------------------------------------------------------------
void sglTexEnv::setMode(ModeEnum mode)
{
   if(mode == eADD && !sgl::s_texture_env_add_capable)
   {
      // Reset to default of modulate, if add not supported
      mode = eMODULATE;
   }
   if (mode == eCOMBINE && !sgl::s_texture_env_combine_capable)
   {
      // Reset to default of modulate, if combine not supported
      mode = eMODULATE;
   }

   m_mode = mode;
   setIndex();
}

//----------------------------------------------------------------------------
void sglTexEnv::setBlendColor(float r, float g, float b, float a)
{
   m_blend_color.set(r, g, b, a);
   setIndex();
}

//----------------------------------------------------------------------------
void sglTexEnv::setCombinerParams(sglTexCombinerParams *params)
{
   m_combiner_params = params;
}

//----------------------------------------------------------------------------
void sglTexEnv::setLODBias(float bias)
{
   if (sgl::s_texture_lod_bias_capable)
   {
      if( bias > sgl::s_max_texture_lod_bias )
         m_lod_bias = sgl::s_max_texture_lod_bias;
      else if( bias < -sgl::s_max_texture_lod_bias )
         m_lod_bias = -sgl::s_max_texture_lod_bias;
      else
         m_lod_bias = bias;

      setIndex();
   }
}

//----------------------------------------------------------------------------
void sglTexEnv::getBlendColor(float &r, float &g, float &b, float &a) const
{
   r = m_blend_color[0];
   g = m_blend_color[1];
   b = m_blend_color[2];
   a = m_blend_color[3];
}

//----------------------------------------------------------------------------
void sglTexEnv::applyToUnit(sglCurrState*, unsigned int unit) const
{
   if (sgl::s_num_texture_units > 1)
   {
      sgl::setActiveTextureUnit(GL_TEXTURE0 + unit);
   }

   if (m_mode == sglTexEnv::eCOMBINE)
   {
      if (m_combiner_params)
      {
         glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE);
         glTexEnvfv(GL_TEXTURE_ENV, GL_TEXTURE_ENV_COLOR,
                    m_blend_color.getPtr());
         m_combiner_params->apply();
      }
      else
         glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
   }
   else
   {
      glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, m_mode);

      if (m_mode == sglTexEnv::eBLEND)
      {
         glTexEnvfv(GL_TEXTURE_ENV, GL_TEXTURE_ENV_COLOR,
                    m_blend_color.getPtr());
      }
   }

   if( sgl::s_texture_lod_bias_capable )
   {
      glTexEnvf(GL_TEXTURE_FILTER_CONTROL_EXT, GL_TEXTURE_LOD_BIAS_EXT,
                m_lod_bias);
   }
}

//----------------------------------------------------------------------------
void sglTexEnv::setIndex()
{
   if (m_lod_bias != 0.0f)
   {
      m_index = UINT_MAX;
   }
   else if (m_mode == eMODULATE)
   {
      m_index = 0;
   }
   else if (m_mode == eDECAL)
   {
      m_index = 1;
   }
   else if (m_mode == eREPLACE)
   {
      m_index = 2;
   }
   else if (m_mode == eADD)
   {
      m_index = 3;
   }
   else  // eBLEND, eCOMBINE
   {
      m_index = UINT_MAX;
   }

}

//----------------------------------------------------------------------------
void sglTexEnv::printInfo(ostream &ostrm, const char *indent_string) const
{
   sglStatelet::printInfo(ostrm, indent_string);
   ostrm << indent_string << "mode: ";
   switch (m_mode)
   {
      case eDECAL:
         ostrm << "eDECAL" << endl;
         break;
      case eREPLACE:
         ostrm << "eREPLACE" << endl;
         break;
      case eMODULATE:
         ostrm << "eMODULATE" << endl;
         break;
      case eADD:
         ostrm << "eADD" << endl;
         break;
      case eBLEND:
         ostrm << "eBLEND" << endl;
         ostrm << indent_string << "blend color: " << m_blend_color << endl;
         break;
      case eCOMBINE:
         ostrm << "eCOMBINE" << endl;
         ostrm << indent_string << "blend color: " << m_blend_color << endl;
         if (m_combiner_params)
         {
            ostrm << indent_string << "combiner params: " << endl;
            m_combiner_params->printInfo(ostrm, indent_string);
            ostrm << endl;
         }
         else
            ostrm << indent_string << "combiner params: NULL" << endl;
         break;
   }

   if( sgl::s_texture_lod_bias_capable )
   {
      ostrm << indent_string << "LOD bias: " << m_lod_bias << endl;
   }
}
