/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglTexEnv.hpp
 *  Created: 2 May 1999
 *  Summary: create, modify and query texture environment
 *****************************************************************************/

#ifndef __SGL_TEXENV_HPP
#define __SGL_TEXENV_HPP

#include <sgl.h>
#include <sglStatelet.hpp>
#include <sglVector.hpp>

class sglTexCombinerParams;

//============================================================================

/**

This statelet encapsulates OpenGL's glTexEnv*() functions which sets the
texturing function for any subsequent textured geometry that is rendered.
These functions affect how the texture interacts with the underlying colors in
the polygons to which it is being applied.  This statelet also encapsulates an
ever expanding set of extensions (if available) that add new texturing
functions:

- GL_ARB_texture_env_add (or GL_EXT_texture_env_add).
- GL_EXT_texture_lod_bias.
- GL_ARB_texture_env_dot3
- GL_ARB_texture_env_combine

<b>Example:</b>
<pre>
   sglTexEnv *tenv = new sglTexEnv;
   tenv->setBlendColor(1.0f, 0.7f, 0.0f, 1.0f);
   tenv->setMode(sglTexEnv::eBLEND);

   if (sgl::s_texture_lod_bias_capable)
   {
      tenv->setLODBias(sgl::s_max_texture_lod_bias);  // maximum
   }
</pre>

 */

class SGL_DLL_API sglTexEnv : public sglTexturingStatelet
{
public:
   /// Tokens for the various texture environment modes.
   enum ModeEnum
   {
      /// Uses texture color.  Only works with RGB and RGBA textures.
      eDECAL    = GL_DECAL,

      /** Uses texture color, uses face alpha if texture format doesn't include
          alpha.  Works with all internal texture formats.
      */
      eREPLACE  = GL_REPLACE,

      /// Modulates texture color with face color.
      eMODULATE = GL_MODULATE,

      /** Adds the face color to the texture color (if extension is available).
          See sgl::s_texture_env_add_capable.
      */
      eADD      = GL_ADD,

      /** Blends the face with a blend color, and uses the face's color value
          somewhat like an alpha value to perform the blending.
      */
      eBLEND    = GL_BLEND,

      /** The texel color is computed from a combination of multiple inputs
          (if extension is available).
          See sgl::s_texture_env_combine_capable.
      */
      eCOMBINE  = GL_COMBINE
   };

public:
   /** Default constructor.  The mode defaults to eMODULATE (the OpenGL
       default).
   */
   sglTexEnv();

   /// Virtual destructor
   virtual ~sglTexEnv();

   /** Set the texture function.
       @param mode One of the ModeEnum tokens.  If eADD or eCombine is
         specified but is not supported through the extensions, then
         eMODULATE will be used instead.
         The default value is eMODULATE.
   */
   void setMode(ModeEnum mode);

   /** Query the current texture function.
       @return The ModeEnum token corresponding to the current texture
         function.
   */
   ModeEnum getMode() const { return m_mode; }

   /** Set the blend color to be used if the texture function is eBLEND,
       or eCOMBINE with an eCONSTANT source parameter.
       @param r The red component of the blend color.
       @param g The green component of the blend color.
       @param b The blue component of the blend color.
       @param a The alpha component of the blend color.
   */
   void setBlendColor(float r, float g, float b, float a);

   /** Query the current blend color.
       @param r On return, contains the red component of the blend color.
       @param g On return, contains the green component of the blend color.
       @param b On return, contains the blue component of the blend color.
       @param a On return, contains the alpha component of the blend color.
   */
   void getBlendColor(float &r, float &g, float &b, float &a) const;

   /** Set the parameters to use for the eCOMBINE mode.  If params is
       set to NULL and eCOMBINE has been selected, eMODULATE will be used
       instead.
       @param params The combiner parameters.
   */
   void setCombinerParams(sglTexCombinerParams *params);

   /** Get the parameters used for the eCOMBINE mode.
       @erturn The combiner parameters.
   */
   sglTexCombinerParams* getCombinerParams() const
      { return m_combiner_params; }

   /** Set the texture LOD bias which can affect which mipmap levels are chosen
       This is only available if the GL_EXT_texture_lod_bias extension is
       available; otherwise this factor is ignored.
       @param bias The bias value in the range [-sgl::s_max_texture_lod_bias
         .. sgl::s_max_texture_lod_bias].  Values outside this range will be
         clamped.
   */
   void setLODBias(float bias);

   /** Query the current LOD bias factor.
       @return The current LOD bias factor.
   */
   float getLODBias() const { return m_lod_bias; }

   // See sglTexturingStatelet::applyToUnit() for details.
   void applyToUnit(sglCurrState *curr_state, unsigned int unit) const;

   // See sglStatelet::getSorted() for details.
   virtual bool getSorted() const { return true; }

   // See sglObject::printInfo() for details.
   virtual void printInfo(ostream &ostrm, const char *indent_string) const;

private:
   sglTexEnv(const sglTexEnv &);  // not implemented
   sglTexEnv &operator=(const sglTexEnv &); // not implemented

   void setIndex();

private:
   ModeEnum m_mode;
   sglVec4f m_blend_color;
   float m_lod_bias;
   sglTexCombinerParams *m_combiner_params;
};

#endif
