/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglTexGen.cpp
 *  Created: 2 May 1999
 *  Summary: create, modify, and query texture coordinate generator
 *****************************************************************************/

#include <sglTexGen.hpp>

const GLenum tex_coord[4] = {GL_S, GL_T, GL_R, GL_Q};
const GLenum texgen_coord[4] = {GL_TEXTURE_GEN_S,
                                GL_TEXTURE_GEN_T,
                                GL_TEXTURE_GEN_R,
                                GL_TEXTURE_GEN_Q};

//----------------------------------------------------------------------------
sglTexGen::sglTexGen()
      : sglTexturingStatelet(eTEX_GEN)
{
   m_mode[0] = sglTexGen::eOFF;
   m_mode[1] = sglTexGen::eOFF;
   m_mode[2] = sglTexGen::eOFF;
   m_mode[3] = sglTexGen::eOFF;

   m_plane[0].set(1.0f, 0.0f, 0.0f, 0.0f);
   m_plane[1].set(0.0f, 1.0f, 0.0f, 0.0f);
   m_plane[2].set(0.0f, 0.0f, 1.0f, 0.0f);
   m_plane[3].set(0.0f, 0.0f, 0.0f, 1.0f);

   setIndex();
}

//----------------------------------------------------------------------------
sglTexGen::~sglTexGen()
{
}

//----------------------------------------------------------------------------
void sglTexGen::setMode(TexCoordEnum coord, ModeEnum mode)
{
   if ((mode == eREFLECTION_MAP || mode == eNORMAL_MAP) &&
       !sgl::s_texture_cube_map_capable)
   {
      // Reset to default of eOFF, if cube map not supported
      mode = eOFF;
   }

   switch (coord)
   {
      case sglTexGen::eS:
         m_mode[0] = mode;
         break;

      case sglTexGen::eT:
         m_mode[1] = mode;
         break;

      case sglTexGen::eR:
         m_mode[2] = mode;
         break;

      case sglTexGen::eQ:
         m_mode[3] = mode;
         break;
   }

   setIndex();
}

//----------------------------------------------------------------------------
sglTexGen::ModeEnum sglTexGen::getMode(TexCoordEnum coord) const
{
   ModeEnum mode = sglTexGen::eOFF;

   switch (coord)
   {
      case sglTexGen::eS:
         mode = m_mode[0];
         break;

      case sglTexGen::eT:
         mode = m_mode[1];
         break;

      case sglTexGen::eR:
         mode = m_mode[2];
         break;

      case sglTexGen::eQ:
         mode = m_mode[3];
         break;
   }
   return mode;
}

//----------------------------------------------------------------------------
void sglTexGen::setPlane(TexCoordEnum coord,
                         float x, float y, float z, float d)
{
   switch (coord)
   {
      case sglTexGen::eS:
         m_plane[0].set(x,y,z,d);
         break;

      case sglTexGen::eT:
         m_plane[1].set(x,y,z,d);
         break;

      case sglTexGen::eR:
         m_plane[2].set(x,y,z,d);
         break;

      case sglTexGen::eQ:
         m_plane[3].set(x,y,z,d);
         break;
   }

   setIndex();
}

//----------------------------------------------------------------------------
void sglTexGen::getPlane(TexCoordEnum coord,
                         float &x, float &y, float &z, float &d) const
{
   switch (coord)
   {
      case sglTexGen::eS:
         x = m_plane[0][0];
         y = m_plane[0][1];
         z = m_plane[0][2];
         d = m_plane[0][3];
         break;

      case sglTexGen::eT:
         x = m_plane[1][0];
         y = m_plane[1][1];
         z = m_plane[1][2];
         d = m_plane[1][3];
         break;

      case sglTexGen::eR:
         x = m_plane[2][0];
         y = m_plane[2][1];
         z = m_plane[2][2];
         d = m_plane[2][3];
         break;

      case sglTexGen::eQ:
         x = m_plane[3][0];
         y = m_plane[3][1];
         z = m_plane[3][2];
         d = m_plane[3][3];
         break;
   }
}

//----------------------------------------------------------------------------
void sglTexGen::applyToUnit(sglCurrState*, unsigned int unit) const
{
   if (sgl::s_num_texture_units > 1)
   {
      sgl::setActiveTextureUnit(GL_TEXTURE0 + unit);
   }

   for (unsigned int i=0; i<4; i++)
   {
      switch (m_mode[i])
      {
         case sglTexGen::eOFF:
            glDisable(texgen_coord[i]);
            break;

         case sglTexGen::eOBJECT_LINEAR:
            glTexGeni(tex_coord[i], GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR);
            glTexGenfv(tex_coord[i], GL_OBJECT_PLANE, m_plane[i].getPtr());
            glEnable(texgen_coord[i]);
            break;

         case sglTexGen::eEYE_LINEAR:
            glTexGeni(tex_coord[i], GL_TEXTURE_GEN_MODE, GL_EYE_LINEAR);
            glTexGenfv(tex_coord[i], GL_EYE_PLANE, m_plane[i].getPtr());
            glEnable(texgen_coord[i]);
            break;

         case sglTexGen::eSPHERE_MAP:
            glTexGeni(tex_coord[i], GL_TEXTURE_GEN_MODE, GL_SPHERE_MAP);
            glEnable(texgen_coord[i]);
            break;

         case sglTexGen::eREFLECTION_MAP:
            glTexGeni(tex_coord[i],
                      GL_TEXTURE_GEN_MODE, GL_REFLECTION_MAP_ARB);
            glEnable(texgen_coord[i]);
            break;

         case sglTexGen::eNORMAL_MAP:
            glTexGeni(tex_coord[i], GL_TEXTURE_GEN_MODE, GL_NORMAL_MAP_ARB);
            glEnable(texgen_coord[i]);
            break;
      }
   }
}

//----------------------------------------------------------------------------
void sglTexGen::setIndex()
{
   if (m_mode[0] == sglTexGen::eOFF &&
       m_mode[1] == sglTexGen::eOFF &&
       m_mode[2] == sglTexGen::eOFF &&
       m_mode[3] == sglTexGen::eOFF)
   {
      m_index = 0;
   }
   else
   {
      m_index = UINT_MAX;
   }
}


//----------------------------------------------------------------------------
void sglTexGen::printInfo(ostream &ostrm, const char *indent_string) const
{
   sglStatelet::printInfo(ostrm, indent_string);
   for (unsigned int i=0; i<4; i++)
   {
      if (i == 0)
         ostrm << indent_string << "S coord: ";
      else if (i == 1)
         ostrm << indent_string << "T coord: ";
      else if (i == 2)
         ostrm << indent_string << "R coord: ";
      else if (i == 3)
         ostrm << indent_string << "Q coord: ";

      switch (m_mode[i])
      {
         case eOFF:
            ostrm << "disabled" << endl;
            break;
         case eOBJECT_LINEAR:
            ostrm << "eOBJECT_LINEAR" << endl;
            ostrm << indent_string << "    plane: " << m_plane[i] << endl;
            break;
         case eEYE_LINEAR:
            ostrm << "eEYE_LINEAR" << endl;
            ostrm << indent_string << "    plane: " << m_plane[i] << endl;
            break;
         case eSPHERE_MAP:
            ostrm << "eSPHERE_MAP" << endl;
            break;
         case eREFLECTION_MAP:
            ostrm << "eREFLECTION_MAP" << endl;
            break;
         case eNORMAL_MAP:
            ostrm << "eNORMAL_MAP" << endl;
            break;
      }
   }
}
