/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglTexGen.hpp
 *  Created: 2 May 1999
 *  Summary: create, modify, and query texture coordinate generator
 *****************************************************************************/

#ifndef __SGL_TEXGEN_HPP
#define __SGL_TEXGEN_HPP

#include <sgl.h>
#include <sglStatelet.hpp>
#include <sglVector.hpp>

//============================================================================

/**

This statelet encapsulates OpenGL's glTexGen*() functions which specify
functions for automatically generating texture coordinates to create effects
such as reflections or making contours on the models.  A separate function for
each of the four texture cooredinates (S, T, R, and Q) can be specified as one
of three different functions: object linear, eye linear, and sphere map.
If the GL_ARB_texture_cube_map extension is supported, this class also
supports the reflection map and normal map modes.

When object linear or eye linear mode is used for any of the coordinates then
the generated coordinate is a linear combination of the vertex (x, y, z, w) a
reference plane can be defined for each coordinate to define the coefficients
of the linear combination.  In the case of the object linear mode these planes
are defined with respect to the object's coordinate system; therefore, the
texture coordinates will be fixed with respect to the object.  In the case of
eye linear mode, the planes are defined with respect to the view point
coordinate system so they will move with the view point.  Planes are only used
by the object and eye linear modes.

Texture coordinate generation can be disabled by calling setMode for all the
texture coordinate generation with sglTexGen::eOFF for the mode parameter.

<b>Example</b>
<pre>
   sglTexGen *tgen = new sglTexGen;  // default mode is OFF

   tgen->setMode(sglTexGen::S, sglTexGen::OBJECT_LINEAR);
   tgen->setPlane(sglTexGen::S, 1.0f, 0.0f, 0.0f, 0.0f);
   tgen->setMode(sglTexGen::T, sglTexGen::OBJECT_LINEAR);
   tgen->setPlane(sglTexGen::T, 0.0f, 1.0f, 0.0f, 0.0f);

   // OR sphere map which ignores plane equations.

   tgen->setMode(sglTexGen::eS, sglTexGen::eSPHERE_MAP);
   tgen->setMode(sglTexGen::eT, sglTexGen::eSPHERE_MAP);
</pre>

 */

class SGL_DLL_API sglTexGen : public sglTexturingStatelet
{
public:
   /// Tokens that wrap OpenGLs texture coordinate specification
   enum TexCoordEnum
   {
      /// The "x" coordinate
      eS = GL_S,

      /// The "y" coordinate
      eT = GL_T,

      /// The "z" coordinate
      eR = GL_R,

      /// The "w" coordinate
      eQ = GL_Q
   };

   /// The tokens for the various texture generation functions
   enum ModeEnum
   {
      /// disable texture generation for this coordinate
      eOFF = -1,

      /// Use object linear function (fixed to object)
      eOBJECT_LINEAR = GL_OBJECT_LINEAR,

      /// Use eye linear function (fixed wrt viewpoint)
      eEYE_LINEAR = GL_EYE_LINEAR,

      /// Use environment mapping
      eSPHERE_MAP = GL_SPHERE_MAP,

      /// reflection map (if cube map extension supported).
      eREFLECTION_MAP = GL_REFLECTION_MAP_ARB,

      /// normal map (if cube map extension supported).
      eNORMAL_MAP = GL_NORMAL_MAP_ARB
   };

public:
   /// Default constructor.  All texture coordinate generation is disabled.
   sglTexGen();

   /// Virtual destructor.
   virtual ~sglTexGen();

   /** Set the texture coordinate generation functions for a specified
       coordinate
       @param coord One of the TexCoordEnum tokens (eS, eT, eR, or eQ).
       @param mode The function for the specified coord parameter.  The default
          is OFF for all four coordinates.
   */
   void setMode(TexCoordEnum coord, ModeEnum mode);

   /** Query the current generation function for the specifed coordinate.
       @param coord One of the TexCoordEnum tokens (eS, eT, eR, or eQ).
       @return The current mode token corresponding to the function for the
         coordinate specified by coord.
   */
   ModeEnum getMode(TexCoordEnum coord) const;

   /** Define the planes for either object linear or eye linear mode.  The
       default planes are perpendicular to their respective dimensions and
       intersecting the origin.  Note a single set of four planes are stored
       in this object and are used in either object or eye linear modes.
       Switching modes will retain the previous planes.
       @param coord One of the TexCoordEnum tokens (eS, eT, eR, or eQ).
       @param x X component of the plane equation.
       @param y Y component of the plane equation.
       @param z Z component of the plane equation.
       @param w W component of the plane equation (displacement?).
   */
   void setPlane(TexCoordEnum coord,
                 float x, float y, float z, float w);

   /** Query the current plane for the given coordinate.
       @param coord One of the TexCoordEnum tokens (eS, eT, eR, or eQ).
       @param x On return, contains X component of the plane equation.
       @param y On return, contains Y component of the plane equation.
       @param z On return, contains Z component of the plane equation.
       @param w On return, contains W component of the plane equation.
    */
   void getPlane(TexCoordEnum coord,
                 float &x, float &y, float &z, float &w) const;

   // See sglTexturingStatelet::applyToUnit() for details.
   void applyToUnit(sglCurrState*, unsigned int) const;

   // See sglStatelet::getSorted() for details.
   virtual bool getSorted() const { return true; }

   // See sglObject::printInfo() for details.
   virtual void printInfo(ostream &ostrm, const char *indent_string) const;

private:
   sglTexGen(const sglTexGen &); // not implemented
   sglTexGen &operator=(const sglTexGen &); // not implemented

   void setIndex();

private:
   ModeEnum m_mode[4];
   sglVec4f m_plane[4];
};

#endif
