/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglTextSet.cpp
 *  Created: 11 September 2001
 *  Summary: render 2D bitmapped text in a 3D location in the scene
 *****************************************************************************/

#include <sglTextSet.hpp>

//----------------------------------------------------------------------------
sglTextSet::sglTextSet()
      : sglDrawable(),
        m_position(0.f, 0.f, 0.f),
        m_color(0.f, 0.f, 0.f, 1.f),
        m_text(""),
        m_font(NULL),
        m_transparency_statelet(false)
{
   m_transparency_statelet.ref();
   m_statelets.push_back(&m_transparency_statelet);

   setColor(sglVec4f(1.0f, 1.0f, 1.0f, 1.0f));

   setState(0, m_statelets);

   m_draw_func = reinterpret_cast<DrawFunc>(&sglTextSet::drawGeometry);
}

//----------------------------------------------------------------------------
sglTextSet::~sglTextSet()
{
   vector<sglStatelet*> tmp;
   setState(0, tmp);
   m_transparency_statelet.deref();
}

//----------------------------------------------------------------------------
void sglTextSet::setString(const string &str)
{
   bool compute_bounds_flag = false;
   if ((!m_text.empty() && str.empty()) ||
       (m_text.empty() && !str.empty()))
   {
      compute_bounds_flag = true;
   }

   m_text = str;

   if (compute_bounds_flag)
      dirty(eBOUND|eDLIST, true);
   else
      dirty(eDLIST, true);
}

//----------------------------------------------------------------------------
void sglTextSet::setFont(sglBitmapFont *font)
{
   if (m_font != font)
   {
      if (m_font)
      {
         m_font->deref();
         if (m_font->getRefCount() == 0)
            delete m_font;
      }
      if (font)
         font->ref();
      m_font = font;
      dirty(eDLIST, true);
   }
}

//----------------------------------------------------------------------------
void sglTextSet::setColor(const sglVec4<float> &color)
{
   if (color[0] != m_color[0] || color[1] != m_color[1] ||
       color[2] != m_color[2] || color[3] != m_color[3])
   {
      m_color = color;
      m_transparency_statelet.enable(1.0f - color[3] >= sglFltEpsilon);
      dirty(eDLIST, true);
   }
}

//----------------------------------------------------------------------------
void sglTextSet::setColor(float r, float g, float b, float a)
{
   if (r != m_color[0] || g != m_color[1] || b != m_color[2] ||
       a != m_color[3])
   {
      m_color.set(r, g, b, a);
      m_transparency_statelet.enable(1.0f - a >= sglFltEpsilon);
      dirty(eDLIST, true);
   }
}

//----------------------------------------------------------------------------
void sglTextSet::setPosition(const sglVec3f &pos)
{
   if (pos[0] != m_position[0] || pos[1] != m_position[1] ||
       pos[2] != m_position[2])
   {
      m_position = pos;
      dirty(eBOUND|eDLIST, true);
   }
}

//----------------------------------------------------------------------------
void sglTextSet::setPosition(float x, float y, float z)
{
   if (x != m_position[0] || y != m_position[1] || z != m_position[2])
   {
      m_position.set(x, y, z);
      dirty(eBOUND|eDLIST, true);
   }
}

//----------------------------------------------------------------------------
bool sglTextSet::computeBounds()
{
   if (m_text.empty())
   {
      m_bbox.makeEmpty();
   }
   else
   {
      m_bbox.setMin(-0.1f + m_position[0],
                    -0.1f + m_position[1],
                    -0.1f + m_position[2]);
      m_bbox.setMax( 0.1f + m_position[0],
                     0.1f + m_position[1],
                     0.1f + m_position[2]);
   }

   return true;
}

//----------------------------------------------------------------------------
void sglTextSet::drawGeometry(const vector<sglTexCoords> &) const
{
   if (!m_text.empty())
   {
      glColor4fv(m_color.getPtr());
      //glRasterPos3i(0,0,0);
      glRasterPos3f(m_position[0], m_position[1], m_position[2]);

      if (m_font)
         m_font->rasterString(m_text);
      else if (sglBitmapFont::getDefaultFont())
         sglBitmapFont::getDefaultFont()->rasterString(m_text);
   }
}

//----------------------------------------------------------------------------
void sglTextSet::printInfo(ostream &ostrm, const char *indent_string) const
{
   sglDrawable::printInfo(ostrm, indent_string);

   ostrm << indent_string << "text = '" << m_text << "'" << endl;
   ostrm << indent_string << "position = " << m_position << endl;
   ostrm << indent_string << "color = " << m_color << endl;
}
