/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglTextSet.hpp
 *  Created: 11 September 2001
 *  Summary: render 2D bitmapped text in a 3D location in the scene
 *****************************************************************************/

#ifndef __SGL_TEXTSET_HPP
#define __SGL_TEXTSET_HPP

#include <sgl.h>
#include <sglDrawable.hpp>
#include <sglBitmapFont.hpp>

//============================================================================

/**

sglTextSet is a subclass of sglDrawable that renders 2D (bitmapped, constant
size in screen space) text in a 3D location in the 3D scene.

sglTextSet supports the rendering (rasterization really) of a 2D bitmap type
font at a position relative to the origin of the parent sglGeode's local
coordinate system in the 3D scene.  The user specifies a single line of text
(determining the raster position of the text after a carriage return for
multiline text is an task on the TODO list), the color of the text, the font
(via an sglBitmapFont object), and the position.  Note that a default font is
provided which is platform-specific if the user does not specify one.  See
sglBitMapFont (and its subclasses) for more information.  This node also
manages its own default statelet vector that includes an sglTransparency
statelet which is set to true when the alpha component of the color is not 1.0.
The user can override this state vector by calling setState(0, ...) on this
drawable.

As a subclass of sglDrawable, the user also has the option of enabling or
disabling the generation of an OpenGL display list for this drawable.  The rule
of thumb is that if the string is not going to change, it is more efficient to
enable the display list generation (this is the default).  The string will
change often, it may be more efficient to explicitly disable this feature.

{\bf Example:}

\begin{verbatim}
   sglTextSet *text_node = new sglTextSet;
   text_node->setColor(1.0f, 1.0f, 0.0f, 0.3f);  // translucent
   text_node->setPosition(10.0f, 0.0f, 0.0f);    // offset along the X axis
   text_node->setString("This text is in the scene graph");
   text_node->enableDisplayList(false);          // disable display listing

   sglGeode *geode = new sglGeode;
   geode->addGeometry(text_node);
\end{verbatim}

 */

class SGL_DLL_API sglTextSet : public sglDrawable
{
public:
   /// Constructor.
   sglTextSet();

   /// Destructor.
   virtual ~sglTextSet();

   // --------------- Functions specific to this drawable ----------------
   /** Set the string to render.
       @param str the string to render
   */
   void setString(const string &str);

   /** Get the string to render.
       @return the string to render
   */
   const string &getString() const { return m_text; }

   /** Set the font to use.
       @param font The font to use.  If NULL, the default bitmap font will be
       used.
   */
   void setFont(sglBitmapFont *font);

   /** Get the font to use.
       @return the font to use
   */
   sglBitmapFont *getFont() const { return m_font; }


   /** Set the text color.
       @param color the text color; default (1,1,1,1)
   */
   void setColor(const sglVec4f &color);

   /** Set the text color (default is white opaque).
       @param r Red color component.
       @param g Green color component.
       @param b Blue color component.
       @param a Alpha color component.
   */
   void setColor(float r, float g, float b, float a);

   /** Get the text color.
       @return the text color
   */
   const sglVec4f &getColor() const { return m_color; }

   /** Set the offset vector.
       @param pos The offset vector (default: 0, 0, 0).
   */
   void setPosition(const sglVec3f &pos);

   /** Set the offset vector.
       @param x X component of the offset vector.
       @param y Y component of the offset vector.
       @param z Z component of the offset vector.
   */
   void setPosition(float x, float y, float z);

   /** Get the current offset vector.
       @return The current offset vector.
   */
   const sglVec3f &getPosition() const { return m_position; }

   // ---------------- Functions required by sglDrawable -----------------
   // Add this drawable's stats to the given debug_stats.
   virtual void addStats(sglStats &stats) const { stats.addText(1); }

   virtual void drawGeometry(const vector<sglTexCoords> &) const;
   virtual void printInfo(ostream &ostrm, const char *indent_string) const;

protected:
   virtual bool computeBounds();

private: // not implemented
   sglTextSet(const sglTextSet &);
   sglTextSet &operator=(const sglTextSet &);

private:
   sglVec3f             m_position;
   sglVec4f             m_color;
   string               m_text;

   sglBitmapFont       *m_font;

   sglTransparency      m_transparency_statelet;
   vector<sglStatelet*> m_statelets;
};

#endif
