/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglTexture.cpp
 *  Created: 2 October 1997
 *  Summary: container for texture image data and corresponding functions
 *****************************************************************************/

#include <sgl.h>
#include <string.h>
#include <sglObject.hpp>
#include <sglTexture.hpp>
#include <sglOStream.hpp>
#include <sglLock.hpp>

set<sglTexture *> sglTexture::s_dirty_textures;
bool              sglTexture::s_auto_bind = true;
sglLock           sglTexture::s_lock;


//----------------------------------------------------------------------------
sglTexture::sglTexture(GLenum target) : sglTexturingStatelet(eTEXTURE),
   m_target(target),
   m_dirty_flag(0),
   m_external_format(eRGBA),
   m_image(NULL),
   m_own_image(false),
   m_border(0),
   m_border_color(1.0, 1.0, 1.0, 1.0),
   m_mag_filter(eLINEAR),
   m_min_filter(eLINEAR),
   m_max_anisotropy(1.0f),
   m_type(eUNSIGNED_BYTE),
   m_internal_format(eINT_RGBA),
   m_internal_format_flag(false),
   m_texture_object(0),
   m_valid_flag(false),
   m_auto_mipmap_flag(eMIP_NONE),
   m_compression_flag(eCOMPRESS_NONE),
   m_texture_lod_active(false),
   m_texture_lod_min(-1000.0f),
   m_texture_lod_max(1000.0f),
   m_texture_lod_base_level(0),
   m_texture_lod_max_level(1000),
   m_bound_mipmaps(false),
   m_pre_bind_callback(NULL),
   m_post_bind_callback(NULL),
   m_internal_size(0),
   m_internal_size_dirty(false)
{
}

//----------------------------------------------------------------------------
sglTexture::~sglTexture()
{
   sgl::releaseTextureObj(m_texture_object);

   if (m_auto_mipmap_flag == eMIP_SW)
   {
      for (vector<MipmapStruct>::iterator it = m_mipmaps.begin();
           it != m_mipmaps.end(); ++it)
      {
         free(it->m_image);
      }
   }

   if (m_own_image)
      free(m_image);
}

//----------------------------------------------------------------------------
void sglTexture::setInternalFormat(InternalFormatEnum format)
{
   m_internal_format_flag = true;

   if (format != m_internal_format)
   {
      dirty(eDIRTY_INTERNAL_FORMAT|eDIRTY_MIPMAP);
      m_internal_format = format;
   }
}

//----------------------------------------------------------------------------
void sglTexture::bindDirtyTextures()
{
   s_lock.acquire();
   set<sglTexture*>::iterator end = s_dirty_textures.end();
   for (set<sglTexture*>::iterator it = s_dirty_textures.begin();
        it != end; ++it)
   {
      (*it)->bind();
   }
   s_dirty_textures.clear();
   s_lock.release();
}

//----------------------------------------------------------------------------
void sglTexture::setAutoBind(bool on)
{
   s_lock.acquire();
   s_auto_bind = on;
   s_lock.release();
}

//----------------------------------------------------------------------------
bool sglTexture::getAutoBind()
{
   s_lock.acquire();
   bool on = s_auto_bind;
   s_lock.release();
   return on;
}

//----------------------------------------------------------------------------
unsigned int sglTexture::getBitsPerTexel(InternalFormatEnum format)
{
   switch (format)
   {
      case eINT_ALPHA4:
      case eINT_LUMINANCE4:
      case eINT_INTENSITY4:
         return 4;

      case eINT_ALPHA: // Assume 8 bits
      case eINT_ALPHA8:
      case eINT_LUMINANCE: // Assume 8 bits
      case eINT_LUMINANCE8:
      case eINT_LUMINANCE4_ALPHA4:
      case eINT_LUMINANCE6_ALPHA2:
      case eINT_INTENSITY: // Assume 8 bits
      case eINT_INTENSITY8:
      case eINT_R3_G3_B2:
      case eINT_RGBA2:
         return 8;

      case eINT_ALPHA12:
      case eINT_LUMINANCE12:
      case eINT_INTENSITY12:
      case eINT_RGB4:
         return 12;

      case eINT_ALPHA16:
      case eINT_LUMINANCE16:
      case eINT_LUMINANCE_ALPHA: // Assume 16 bits
      case eINT_LUMINANCE8_ALPHA8:
      case eINT_LUMINANCE12_ALPHA4:
      case eINT_INTENSITY16:
      case eINT_RGBA4:
      case eINT_RGB5_A1:
         return 16;

      case eINT_LUMINANCE12_ALPHA12:
      case eINT_RGB: // Assume 24 bits
      case eINT_RGB8:
         return 24;

      case eINT_LUMINANCE16_ALPHA16:
      case eINT_RGBA: // Assume 32 bits
      case eINT_RGBA8:
      case eINT_RGB10_A2:
         return 32;

      case eINT_RGB5:
         return 15;

      case eINT_RGB10:
         return 30;

      case eINT_RGB12:
         return 36;

      case eINT_RGB16:
      case eINT_RGBA12:
         return 48;

      case eINT_RGBA16:
         return 64;

      default:
         sglPrintWarning()
            << "sglTexture::getBitsPerTexel(): "
            << "Unknown texture internal format = 0x"
            << hex << (int)format << dec << endl;
         return 32;
   }
}

//----------------------------------------------------------------------------
void sglTexture::setBorderColor(const sglVec4f &color)
{
   if (!(m_border_color == color))
   {
      m_border_color = color;
      dirty(eDIRTY_BORDER_COLOR);
   }
}

//----------------------------------------------------------------------------
void sglTexture::setMinFilter(FilterEnum filter)
{
   if (filter != m_min_filter)
   {
      m_min_filter = filter;
      dirty(eDIRTY_MIN_FILTER);
   }
}

//----------------------------------------------------------------------------
void sglTexture::setMagFilter(FilterEnum filter)
{
   if (filter != m_mag_filter)

   {
      if (filter == eLINEAR || filter == eNEAREST)
      {
         m_mag_filter = filter;
         dirty(eDIRTY_MAG_FILTER);
      }
   }
}

//----------------------------------------------------------------------------
void sglTexture::setMaxAnisotropy(float max)
{
   if (max < 1.0f) max = 1.0f;

   if (max != m_max_anisotropy)
   {
      m_max_anisotropy = max;
      dirty(eDIRTY_ANISOTROPY);
   }
}

//----------------------------------------------------------------------------
void sglTexture::setAutoMipmapFlag(bool on)
{
   unsigned int mask = eDIRTY_MIPMAP;

   if (on)
   {
      if (m_auto_mipmap_flag)
         // flag already set
         return;

      m_mipmaps.clear();

      // check for hardware support
      if (sgl::s_auto_mipmap_capable)
      {
         m_auto_mipmap_flag = eMIP_HW;

         // force base level to re-bind
         mask |= eDIRTY_IMAGE;
      }
      else
      {
         m_auto_mipmap_flag = eMIP_SW;
         buildAutoMipmaps(eDIRTY_IMAGE);
      }
   }
   else
   {
      if (m_auto_mipmap_flag == eMIP_NONE)
         // flag already off
         return;

      if (m_auto_mipmap_flag == eMIP_SW)
      {
         // clear old mips
         for (vector<MipmapStruct>::iterator it = m_mipmaps.begin();
              it != m_mipmaps.end(); ++it)
         {
            free(it->m_image);
         }
         m_mipmaps.clear();
      }

      // nothing to be done here if hardware mip

      m_auto_mipmap_flag = eMIP_NONE;
   }

   dirty(mask);
}

//----------------------------------------------------------------------------
void sglTexture::setCompressionLevel( CompressionLevelEnum level )
{
   // check for hardware support
   if( sgl::s_texture_compression_capable )
   {
      m_compression_flag = level;

      // if level is eCOMPRESS_NONE, no problem, as m_compression_enum is
      // ignored.

      //
      // S3TC extension says that ARB_texture_compression is required, so
      // use generic formats for those we don't know about in S3 format.
      //

      switch( m_external_format )
      {
      case eALPHA:
         m_compression_enum = 0x84e9; // COMP_ALPHA_ARB
         break;

      case eLUMINANCE:
         m_compression_enum = 0x84ea; // COMP_LUMINANCE_ARB
         break;

      case eLUMINANCE_ALPHA:
         m_compression_enum = 0x84eb; // COMP_LUMINANCE_ALPHA_ARB
         break;

      case eRED:
      case eGREEN:
      case eBLUE:
         m_compression_enum = 0x84ec; // COMP_INTENSITY_ARB
         break;

      case eRGB:
         if (sgl::s_texture_compression_vendor == sgl::eCOMPRESSION_VENDOR_ARB)
            m_compression_enum = 0x84ed; // COMP_RGB_ARB
         else
            // s3tc only has 1 "RGB" format, no real quality settings
            m_compression_enum = 0x83f0; // RGB_S3TC_DXT1
         break;

      case eRGBA:
         if (sgl::s_texture_compression_vendor == sgl::eCOMPRESSION_VENDOR_ARB)
            m_compression_enum = 0x84ee; // COMP_RGBA_ARB
         else
         {
            // assume S3 for now
            switch( level )
            {
            case eCOMPRESS_MIN:
               m_compression_enum = 0x83f2; // RGBA_S3TC_DXT3 (4 bit alpha)
               break;
            case eCOMPRESS_MED:
               m_compression_enum = 0x83f3; // RGBA_S3TC_DXT5 (interp  alpha)
               break;
            case eCOMPRESS_MAX:
               m_compression_enum = 0x83f1; // RGBA_S3TC_DXT1 (1 bit alpha)
               break;
            case eCOMPRESS_NONE: // to keep compiler happy
               m_compression_enum = 0;
               break;
            }
         }
         break;
      }

      // may need more?
      dirty(eDIRTY_INTERNAL_FORMAT|eDIRTY_MIPMAP);
   }
   else
      m_compression_flag = eCOMPRESS_NONE;
}

//----------------------------------------------------------------------------
void sglTexture::setLODMinMax( float minLOD, float maxLOD )
{
   if( sgl::s_texture_lod_capable )
   {
      m_texture_lod_active = true;

      m_texture_lod_min = minLOD;
      m_texture_lod_max = maxLOD;

      dirty(eDIRTY_LOD);
   }
}

//----------------------------------------------------------------------------
void sglTexture::setLODBaseLevel( GLuint baseLevel )
{
   if( sgl::s_texture_lod_capable )
   {
      m_texture_lod_active = true;

      m_texture_lod_base_level = baseLevel;

      dirty(eDIRTY_LOD);
   }
}

//----------------------------------------------------------------------------
void sglTexture::setLODMaxLevel( GLuint maxLevel )
{
   if( sgl::s_texture_lod_capable )
   {
      m_texture_lod_active = true;

      m_texture_lod_max_level = maxLevel;

      dirty(eDIRTY_LOD);
   }
}

//----------------------------------------------------------------------------
void sglTexture::unbind()
{
   if (m_texture_object)
   {
      sgl::releaseTextureObj(m_texture_object);
      m_texture_object = 0;
      dirty(0xffffffff);
   }
}

//----------------------------------------------------------------------------
void sglTexture::reset()
{
   dirty(eDIRTY_RESET);
}

//----------------------------------------------------------------------------
void sglTexture::dirty(unsigned int mask)
{
   m_dirty_flag |= mask;

   s_lock.acquire();
   if (s_auto_bind)
      s_dirty_textures.insert(this);
   s_lock.release();
}

//----------------------------------------------------------------------------
void sglTexture::applyToUnit(sglCurrState*, unsigned int unit) const
{
   if (sgl::s_num_texture_units > 1)
   {
      sgl::setActiveTextureUnit(GL_TEXTURE0 + unit);
   }

   if (m_valid_flag)
   {
      // NOTE: bind() should be called before calling apply().
      if (m_dirty_flag)
      {
         if (m_texture_object == 0)
         {
            glDisable(GL_TEXTURE_1D);
            glDisable(GL_TEXTURE_2D);
            if (sgl::s_texture_cube_map_capable)
                glDisable(GL_TEXTURE_CUBE_MAP_ARB);
            return;
         }
         else
         {
            sglPrintWarning()
               << "Warning: sglTexture::apply() called before"
               << " sglTexture::bind()"
               << endl;
         }
      }

      if (sgl::s_texture_cube_map_capable)
      {
          switch (m_target)
          {
              case GL_TEXTURE_1D:
                  glDisable(GL_TEXTURE_2D);
                  glDisable(GL_TEXTURE_CUBE_MAP_ARB);
                  break;

              case GL_TEXTURE_2D:
                  glDisable(GL_TEXTURE_1D);
                  glDisable(GL_TEXTURE_CUBE_MAP_ARB);
                  break;

              case GL_TEXTURE_CUBE_MAP_ARB:
                  glDisable(GL_TEXTURE_1D);
                  glDisable(GL_TEXTURE_2D);
                  break;

              default:
                  sglPrintWarning()
                      << "Warning: Unsupported texture target type: "
                      << (int)m_target << endl;
                  break;
          }
      }
      else
      {
          switch (m_target)
          {
              case GL_TEXTURE_1D:
                  glDisable(GL_TEXTURE_2D);
                  break;

              case GL_TEXTURE_2D:
                  glDisable(GL_TEXTURE_1D);
                  break;

              default:
                  sglPrintWarning()
                      << "Warning: Unsupported texture target type: "
                      << (int)m_target << endl;
                  break;
          }
      }

      glBindTexture(m_target, m_texture_object);
      glEnable(m_target);
   }
   else
   {
      glDisable(GL_TEXTURE_1D);
      glDisable(GL_TEXTURE_2D);
      if (sgl::s_texture_cube_map_capable) glDisable(GL_TEXTURE_CUBE_MAP_ARB);
   }
}

//----------------------------------------------------------------------------
void sglTexture::setPreBindCallback(callback func)
{
   if (func != m_pre_bind_callback)
   {
      m_pre_bind_callback = func;
      dirty(eDIRTY_RESET); // Is this too much?
   }
}

//----------------------------------------------------------------------------
void sglTexture::setPostBindCallback(callback func)
{
   if (func != m_post_bind_callback)
   {
      m_post_bind_callback = func;
      dirty(eDIRTY_RESET); // Is this too much?
   }
}

//----------------------------------------------------------------------------
unsigned int sglTexture::getNumComponents(ExternalFormatEnum external_format)
{
   switch (external_format)
   {
      case eRGBA:
         return 4;
      case eRGB:
         return 3;
      case eLUMINANCE_ALPHA:
         return 2;
      default:
         return 1;
   }
}

//----------------------------------------------------------------------------
void sglTexture::buildAutoMipmaps(unsigned int mask)
{
   vector<void*> mipmaps;

   for (vector<MipmapStruct>::iterator it = m_mipmaps.begin();
        it != m_mipmaps.end(); ++it)
   {
      mipmaps.push_back(it->m_image);
   }

   if (buildMipmaps(mipmaps))
   {
      unsigned int size = mipmaps.size();
      unsigned int old_size = m_mipmaps.size();
      unsigned int common = (size < old_size ? size : old_size);
      unsigned int n;
      for (n = 0; n < common; ++n)
      {
         m_mipmaps[n].m_image = mipmaps[n];
         m_mipmaps[n].m_dirty |= mask;
      }
      for (n = common; n < old_size; ++n)
      {
         m_mipmaps.pop_back();
      }
      for (n = common; n < size; ++n)
      {
         m_mipmaps.push_back(MipmapStruct(mipmaps[n], eDIRTY_IMAGE));
      }
   }
   else
   {
      m_mipmaps.clear();
      for (vector<void*>::iterator dd = mipmaps.begin();
           dd != mipmaps.end(); ++dd)
      {
         if (*dd) free(*dd);
      }
   }
}

//----------------------------------------------------------------------------
void sglTexture::printInfo(ostream &ostrm, const char *indent_string) const
{
    sglStatelet::printInfo(ostrm, indent_string);

    ostrm << indent_string << "dirty flag = " << m_dirty_flag << endl;
    ostrm << indent_string << "external format = " << (int)m_external_format
          << endl;
    ostrm << indent_string << "image ptr = 0x" << hex << m_image << dec <<endl;
    ostrm << indent_string << "own image flag = " << m_own_image << endl;
    ostrm << indent_string << "no. of mipmaps = " << m_mipmaps.size() << endl;
    ostrm << indent_string << "border = " << m_border << endl;
    ostrm << indent_string << "border color = " << m_border_color << endl;
    ostrm << indent_string << "min/mag filter = " << (int)m_min_filter << "/"
          << (int)m_mag_filter << endl;
    ostrm << indent_string << "anisotropy = " << m_max_anisotropy << endl;
    ostrm << indent_string << "type = " << (int)m_type << endl;
    ostrm << indent_string << "internal format = " << (int)m_internal_format
          << endl;
    ostrm << indent_string << "internal format flag = "
          << m_internal_format_flag << endl;
    ostrm << indent_string << "texture object = " << m_texture_object << endl;
    ostrm << indent_string << "valid flag = " << m_valid_flag << endl;
    ostrm << indent_string << "auto mipmap flag = " << (int)m_auto_mipmap_flag
          << endl;
    ostrm << indent_string << "compression flag = " << (int)m_compression_flag
          << endl;

    ostrm << indent_string << "texture LOD flag = " << m_texture_lod_active
          << endl;
    if (m_texture_lod_active)
    {
        ostrm << indent_string << "   LOD min/max = " << m_texture_lod_min
              << "/" << m_texture_lod_max << endl;
        ostrm << indent_string << "   LOD level base/max = "
              << m_texture_lod_base_level << "/" << m_texture_lod_max_level
              << endl;
    }
    ostrm << indent_string << "mipmaps bound flag = " << m_bound_mipmaps
          << endl;
    ostrm << indent_string << "pre bind callback = 0x" << hex
          << m_pre_bind_callback << dec << endl;
    ostrm << indent_string << "post bind callback = 0x" << hex
          << m_post_bind_callback << dec << endl;
    ostrm << indent_string << "internal size dirty flag = "
          << m_internal_size_dirty << endl;
    ostrm << indent_string << "internal size = " << m_internal_size << endl;
}
