/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglTexture.hpp
 *  Created: 29 September 1997
 *  Summary: base class container for image data used for texturing (1D,2D,3D)
 *****************************************************************************/

#ifndef __SGL_TEXTURE_HPP
#define __SGL_TEXTURE_HPP

#include <sgl.h>
#include <vector>
#include <string>
#include <set>
#include <sglStatelet.hpp>
#include <sglVector.hpp>

//============================================================================

/** base class container for image data used for texturing (1D, 2D, and someday
    3D)
*/

class SGL_DLL_API sglTexture : public sglTexturingStatelet
{
public:
   /// Texture wrap modes.
   enum WrapEnum
   {
      eREPEAT,
      eCLAMP,
      eCLAMP_TO_EDGE    // same as eCLAMP when not supported
   };

   /// Filters used for minifying or magnifying the texture.
   enum FilterEnum
   {
      eNEAREST                = GL_NEAREST,
      eLINEAR                 = GL_LINEAR,
      eNEAREST_MIPMAP_NEAREST = GL_NEAREST_MIPMAP_NEAREST,
      eNEAREST_MIPMAP_LINEAR  = GL_NEAREST_MIPMAP_LINEAR,
      eLINEAR_MIPMAP_NEAREST  = GL_LINEAR_MIPMAP_NEAREST,
      eLINEAR_MIPMAP_LINEAR   = GL_LINEAR_MIPMAP_LINEAR
   };

   /// Valid formats for images used to create textures.
   enum ExternalFormatEnum
   {
      eRGB              = GL_RGB,
      eRGBA             = GL_RGBA,
      eRED              = GL_RED,
      eGREEN            = GL_GREEN,
      eBLUE             = GL_BLUE,
      eALPHA            = GL_ALPHA,
      eLUMINANCE        = GL_LUMINANCE,
      eLUMINANCE_ALPHA  = GL_LUMINANCE_ALPHA
   };

   /// Valid data types for images used to create textures.
   enum TypeEnum
   {
      eBYTE             = GL_BYTE,
      eUNSIGNED_BYTE    = GL_UNSIGNED_BYTE,
      eSHORT            = GL_SHORT,
      eFLOAT            = GL_FLOAT,
      eUNSIGNED_SHORT   = GL_UNSIGNED_SHORT,
      eINT              = GL_INT,
      eUNSIGNED_INT     = GL_UNSIGNED_INT,
      eBITMAP           = GL_BITMAP
   };

   /// Possible storage formats for textures in internal display memory.
   enum InternalFormatEnum
   {
      eINT_ALPHA4              = GL_ALPHA4,
      eINT_LUMINANCE4          = GL_LUMINANCE4,
      eINT_INTENSITY4          = GL_INTENSITY4,
      eINT_ALPHA               = GL_ALPHA,
      eINT_ALPHA8              = GL_ALPHA8,
      eINT_LUMINANCE           = GL_LUMINANCE,
      eINT_LUMINANCE8          = GL_LUMINANCE8,
      eINT_LUMINANCE4_ALPHA4   = GL_LUMINANCE4_ALPHA4,
      eINT_LUMINANCE6_ALPHA2   = GL_LUMINANCE6_ALPHA2,
      eINT_INTENSITY           = GL_INTENSITY,
      eINT_INTENSITY8          = GL_INTENSITY8,
      eINT_R3_G3_B2            = GL_R3_G3_B2,
      eINT_RGBA2               = GL_RGBA2,
      eINT_ALPHA12             = GL_ALPHA12,
      eINT_LUMINANCE12         = GL_LUMINANCE12,
      eINT_INTENSITY12         = GL_INTENSITY12,
      eINT_RGB4                = GL_RGB4,
      eINT_ALPHA16             = GL_ALPHA16,
      eINT_LUMINANCE16         = GL_LUMINANCE16,
      eINT_LUMINANCE_ALPHA     = GL_LUMINANCE_ALPHA,
      eINT_LUMINANCE8_ALPHA8   = GL_LUMINANCE8_ALPHA8,
      eINT_LUMINANCE12_ALPHA4  = GL_LUMINANCE12_ALPHA4,
      eINT_INTENSITY16         = GL_INTENSITY16,
      eINT_RGBA4               = GL_RGBA4,
      eINT_RGB5_A1             = GL_RGB5_A1,
      eINT_LUMINANCE12_ALPHA12 = GL_LUMINANCE12_ALPHA12,
      eINT_RGB                 = GL_RGB,
      eINT_RGB8                = GL_RGB8,
      eINT_LUMINANCE16_ALPHA16 = GL_LUMINANCE16_ALPHA16,
      eINT_RGBA                = GL_RGBA,
      eINT_RGBA8               = GL_RGBA8,
      eINT_RGB10_A2            = GL_RGB10_A2,
      eINT_RGB5                = GL_RGB5,
      eINT_RGB10               = GL_RGB10,
      eINT_RGB12               = GL_RGB12,
      eINT_RGB16               = GL_RGB16,
      eINT_RGBA12              = GL_RGBA12,
      eINT_RGBA16              = GL_RGBA16
   };

   enum CompressionLevelEnum
   {
      eCOMPRESS_NONE = 0,     // highest quality
      eCOMPRESS_MIN = 1,
      eCOMPRESS_MED = 2,
      eCOMPRESS_MAX = 3       // lowest quality, highest compression
   };

public:
   /** Destructor. */
   virtual ~sglTexture();

   /** Bind all textures that are currently dirty.  Note that if
       (getAutoBind() == false), this method does nothing.  This
       method is called by sglScene::preDraw().
   */
   static void bindDirtyTextures();

   /** Enables/Disables the bindDirtyTextures() method.
       default: true (dirty textures will automatically be bound
                by bindDirtyTextures())
   */
   static void setAutoBind(bool on);

   /** Returns true if bindDirtyTextures() is enabled. */
   static bool getAutoBind();

   /** Returns the ideal number of bits per texel used for the given
       internal texture format.
   */
   static unsigned int getBitsPerTexel(InternalFormatEnum format);

   /** Set the mipmap images.  Returns false if number of images is not
       correct or if the AutoMipmap feature is on.  Assumes that the
       images are correct in size and format.
   */
   virtual bool setMipmapLevels(const vector<void *> &images) = 0;

   /** Helper function to replace an entire image via subtexload funtionality.
       A previous image must first have been set with the setImage or
       setMipmapLevels methods.  The new_image data must be the same size
       as the original image, and the pointer to the old image data is lost.
       The level parameter specifies which mipmap level to replace.
   */
   virtual bool replaceImage(void *new_image, GLint level = 0) = 0;

   /** Request textures to be stored with the given format in internal
       texture memory.  Note: this only a request and is not guaranteed.
       default: automatically determined during call to setImage
   */
   void setInternalFormat(InternalFormatEnum format);

   /** Returns the requested internal texture format. */
   InternalFormatEnum getInternalFormat() const { return m_internal_format; }

   /** Returns the current external texture format. */
   ExternalFormatEnum getExternalFormat() const { return m_external_format; }

   /** Sets the requested compression format */
   void setCompressionLevel( CompressionLevelEnum fact );

   /** returns the requested compression format */
   CompressionLevelEnum getCompressionLevel() { return m_compression_flag; }

   /** Texture Level Of Detail */

   /** Sets the min/max lod to use for this texture */
   void setLODMinMax( float minLOD, float maxLOD );
   void getLODMinMax( float &minLOD, float &maxLOD ) const
      { minLOD = m_texture_lod_min; maxLOD = m_texture_lod_max; }

   /** sets the "base" level to BASE instead of zero */
   void setLODBaseLevel( GLuint base );
   GLuint getLODBaseLevel() const { return m_texture_lod_base_level; }

   /** sets the "max" level to maxLevel instead of funtion of "base" size */
   void setLODMaxLevel( GLuint maxLevel );
   GLuint getLODMaxLevel() const { return m_texture_lod_max_level; }

   /** Returns the number of bytes of texture memory this texture will
       use if the requested internal format was honored.
   */
   virtual unsigned int getIdealInternalSize() const = 0;

   /** Returns the number of bytes of texture memory this texture is
       most likely using.
       Must be called by a thread with a current rendering context.
   */
   virtual unsigned int getLikelyInternalSize() const = 0;

   /** Set the color to use past the edge of a texture when CLAMP is used.
       default: (1.0, 1.0, 1.0, 1.0)
   */
   void setBorderColor(const sglVec4f &color);

   /** Get the color to use past the edge of a texture when CLAMP is used. */
   void getBorderColor(sglVec4f &color) const {color = m_border_color;}

   /** Set the filter to use for texture minification.
       default: eLINEAR
   */
   void setMinFilter(FilterEnum filter);

   /** Returns the filter to use for texture minification. */
   FilterEnum getMinFilter() const { return m_min_filter; }

   /** Set the filter to use for texture magnification.
       default: eLINEAR
   */
   void setMagFilter(FilterEnum filter);

   /** Returns the filter to use for texture magnification. */
   FilterEnum getMagFilter() const { return m_mag_filter; }

   /** Get/set the maximum degree of anisotropy to account for in
       texture filtering (if supported).  A value of 1.0f disables
       anisotropic filtering.
       default: 1.0f
   */
   void setMaxAnisotropy(float max);
   ///
   float getMaxAnisotropy() const { return m_max_anisotropy; }

   /** Get the maximum max-anisotropy supported on this system.
       Returns 1.0f if anisotropic filtering is not supported.
   */
   static float getMaxSupportedMaxAnisotropy()
      { return sgl::s_max_texture_max_anisotropy; }

   /** Enable/Disable the automatic generation of mipmaps.
       default: false
   */
   void setAutoMipmapFlag(bool on);

   /** Returns true if automatic mipmap generation is on. */
   bool getAutoMipmapFlag() { return (m_auto_mipmap_flag)?true:false; }

   /** Return status of this object (as a valid texture).
       if (isValid() == false) texturing will be disabled
   */
   inline bool isValid() const {return m_valid_flag;}

   /** Use the GL_PROXY_TEXTURE mechanism to check if the texture
       is capable of being loaded into texture memory.
       Returns false if the texture is too large.
       Must be called by a thread with a current rendering context.
   */
   virtual bool checkProxy() = 0;

   /** Copy texture (and changes) into texture memory.
       If (getAutoBind() == false), this method must be called before
       before the texture can be applied.
       Must be called by a thread with a current rendering context.
   */
   virtual void bind() = 0;

   /** Removes the texture from texture memory. */
   void unbind();

   /** Dirties the texture, such that during the next bind the texture
       will be completely recreated in texture memory.  This may be useful
       if the texture had mipmaps defined, but they are no longer needed.
   */
   void reset();

   /** Returns true if the texture is ready to be rendered. */
   inline bool isBound() const {return (m_dirty_flag == 0); }

   /** Apply this state change to the current rendering state.
       NOTE: bind() should be called before calling apply().
   */
   void applyToUnit(sglCurrState*, unsigned int) const;

   /** Returns true if this type of statelet should be sorted by pointer
       rather than indexed into an array.
   */
   virtual bool getSorted() const { return true; }

   /** Callback function prototype. */
   typedef void (*callback)(sglTexture*);

   /** Set the callback function to call before this texture is bound.
       This could be used to setup state like glPixelTransfer.
       default: NULL
   */
   void setPreBindCallback(callback func);

   /** Returns the callback function to call before this texture is bound. */
   callback getPreBindCallback() const { return m_pre_bind_callback; }

   /** Set the callback function to call after this texture is bound.
       This should be used to revert any state changes made by
       the function registered in setPreBindCallback.
       default: NULL
   */
   void setPostBindCallback(callback func);

   /** Returns the callback function to call after this texture is bound. */
   callback getPostBindCallback() const { return m_post_bind_callback; }

   /** Returns the number of color components in the given external format. */
   static unsigned int getNumComponents(ExternalFormatEnum external_format);

   /** Generates a set of mipmaps from the level 0 image in this texture.
       Existing images in the given mipmaps vector will be resized (via
       realloc) and excess images will be free()'d.
       Returns false if the mipmaps could not be built.
   */
   virtual bool buildMipmaps(vector<void*> &mipmaps) = 0;

   // See sglObject::printInfo() for details.
   virtual void printInfo(ostream &ostrm, const char *indent_string) const;

protected:
   /** Constructor: target = one of GL_TEXTURE_1D or GL_TEXTURE_2D */
   sglTexture(GLenum target);

   void dirty(unsigned int mask);

   void buildAutoMipmaps(unsigned int mask);

   // flags indicating which aspects of the texture need to be updated on the
   // bind
   enum DirtyEnum
   {
        eDIRTY_IMAGE           = 0x0001,
        eDIRTY_SUBIMAGE        = 0x0002,
        eDIRTY_INTERNAL_FORMAT = 0x0004,
        eDIRTY_WRAP_S_MODE     = 0x0008,
        eDIRTY_WRAP_T_MODE     = 0x0010,
        eDIRTY_BORDER_COLOR    = 0x0020,
        eDIRTY_MIN_FILTER      = 0x0040,
        eDIRTY_MAG_FILTER      = 0x0080,
        eDIRTY_MIPMAP          = 0x0100,
        eDIRTY_ANISOTROPY      = 0x0200,
        eDIRTY_REPLACE         = 0x0400,
        eDIRTY_LOD             = 0x0800,
        eDIRTY_RESET           = 0x8000
   };

   // flags for hardware mip
   enum MipStateEnum
   {
      eMIP_NONE = 0,
      eMIP_SW   = 1,
      eMIP_HW   = 2
   };

   struct MipmapStruct
   {
      MipmapStruct(void *image, unsigned int mask)
         : m_image(image), m_dirty(mask) {}

      void          *m_image;
      unsigned int   m_dirty;
   };

private:
   sglTexture();                                // NOT IMPLEMENTED
   sglTexture(const sglTexture&);               // NOT IMPLEMENTED
   sglTexture& operator=(const sglTexture &);   // NOT IMPLEMENTED

protected:
   static set<sglTexture *> s_dirty_textures;
   static bool              s_auto_bind;
   static sglLock           s_lock;

   const GLenum             m_target;        // GL_TEXTURE_*D
   unsigned int             m_dirty_flag;
   ExternalFormatEnum       m_external_format;
   void*                    m_image;         // image data
   bool                     m_own_image;
   vector<MipmapStruct>     m_mipmaps;
   GLint                    m_border;

   sglVec4f                 m_border_color;
   FilterEnum               m_mag_filter;
   FilterEnum               m_min_filter;
   float                    m_max_anisotropy;

   TypeEnum                 m_type;
   InternalFormatEnum       m_internal_format;
   bool                     m_internal_format_flag;

   GLuint                   m_texture_object;
   bool                     m_valid_flag;
   MipStateEnum             m_auto_mipmap_flag;

   CompressionLevelEnum     m_compression_flag;
   GLenum                   m_compression_enum; // actual enum to bind call

   // texture lod
   bool                     m_texture_lod_active; // opt: using lod?
   float                    m_texture_lod_min;
   float                    m_texture_lod_max;
   GLuint                   m_texture_lod_base_level;
   GLuint                   m_texture_lod_max_level;

   // true if mipmaps have been bound since the m_texture_object was created
   bool                     m_bound_mipmaps;

   callback                 m_pre_bind_callback;
   callback                 m_post_bind_callback;

   mutable unsigned int     m_internal_size;
   mutable bool             m_internal_size_dirty;
};

#endif
