/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglTexture1D.cpp
 *  Created: 11 March 2000
 *  Summary: container for 1D texture image data and corresponding functions
 *****************************************************************************/

#include <sgl.h>
#include <string.h>
#include <sglObject.hpp>
#include <sglTexture1D.hpp>
#include <sglOStream.hpp>
#include <sglLock.hpp>

//----------------------------------------------------------------------------
sglTexture1D::sglTexture1D() : sglTexture(GL_TEXTURE_1D),
   m_width(0),
   m_repeat_flag(GL_REPEAT),
   m_pow_s(0)
{
}

//----------------------------------------------------------------------------
sglTexture1D::~sglTexture1D()
{
   // Remove this texture from the dirty list _before_ allowing the
   // object to become partially destructed (ie this can't be done in
   // the base class, or bind could get called during destruction).
   s_lock.acquire();
   s_dirty_textures.erase(this);
   s_lock.release();
}

//----------------------------------------------------------------------------
void sglTexture1D::getImage(void *&image,
                            ExternalFormatEnum &format, TypeEnum &type,
                            unsigned int &num_s, bool &border) const
{
   image = NULL;
   if (m_valid_flag)
   {
      image = m_image;
      format = m_external_format;
      type = m_type;
      num_s = m_width;
      border = (m_border ? true : false);
   }
}

//----------------------------------------------------------------------------
bool sglTexture1D::setImage(void *image,
                            ExternalFormatEnum format, TypeEnum type,
                            unsigned int num_s, bool border)
{
   unsigned int borderless_s = (border ? num_s - 2 : num_s);
   unsigned int pow_s = 0;
   for (unsigned int tmp_s = borderless_s; tmp_s > 1; ++pow_s) {tmp_s >>= 1;}

   if (borderless_s == 0 || format == 0 || image == NULL)
   {
      m_valid_flag = false;
      m_index = 0;

      dirty(eDIRTY_IMAGE);
   }

   if ((GLint)borderless_s > sgl::s_max_texture_size
       || (1u<<pow_s) != borderless_s)
   {
      unsigned int old_s = num_s;

      if ((GLint)borderless_s > sgl::s_max_texture_size)
      {
         sglPrintWarning() << "sglTexture2D::setImage() size ("
                           << num_s
                           << ") > GL_MAX_TEXTURE_SIZE; auto resizing"
                           << endl;

         num_s = sgl::s_max_texture_size;
         borderless_s = (border ? num_s - 2 : num_s);
         pow_s = 0;
         unsigned int tmp;
         for (tmp = borderless_s; tmp > 1; ++pow_s) {tmp >>= 1;}
      }

      if ((1u<<pow_s) != borderless_s)
      {
         sglPrintWarning() << "sglTexture2D::setImage() size ("
                           << num_s
                           << ") not power of 2; auto resizing"
                           << endl;

         num_s = (1u<<pow_s) + (border ? 2 : 0);
      }

      if (!m_own_image)
         m_image = NULL;

      resizeImage(format, type, old_s, border, num_s, image, m_image);

      m_own_image = true;
   }
   else
   {
      if (m_own_image && image != m_image)
      {
         free(m_image);
         m_own_image = false;
      }
      // note: a pointer to any preexisting image is lost
      m_image = image;
   }

   m_width           = num_s;
   m_external_format = format;
   m_type            = type;
   m_border          = (border ? 1 : 0);
   m_pow_s           = pow_s;

   // find some reasonable default
   if (!m_internal_format_flag)
   {
      switch (type)
      {
         case eSHORT:
         case eUNSIGNED_SHORT:
         case eINT:
         case eUNSIGNED_INT:
         case eFLOAT:
            switch (format)
            {
               case eRGBA:
                  m_internal_format = eINT_RGBA16;
                  break;
               case eALPHA:
                  m_internal_format = eINT_ALPHA16;
                  break;
               case eLUMINANCE:
                  m_internal_format = eINT_LUMINANCE16;
                  break;
               case eLUMINANCE_ALPHA:
                  m_internal_format = eINT_LUMINANCE16_ALPHA16;
                  break;
               default:
                  m_internal_format = eINT_RGB16;
            }
            break;
         default:
            switch (format)
            {
               case eRGBA:
                  m_internal_format = eINT_RGBA;
                  break;
               case eALPHA:
                  m_internal_format = eINT_ALPHA;
                  break;
               case eLUMINANCE:
                  m_internal_format = eINT_LUMINANCE;
                  break;
               case eLUMINANCE_ALPHA:
                  m_internal_format = eINT_LUMINANCE_ALPHA;
                  break;
               default:
                  m_internal_format = eINT_RGB;
            }
      }
   }

   if (m_auto_mipmap_flag)
      buildAutoMipmaps(eDIRTY_IMAGE);
   else
      m_mipmaps.clear();

   m_valid_flag = true;
   m_index = UINT_MAX;

   // not sure this is the correct set
   dirty(eDIRTY_IMAGE | eDIRTY_INTERNAL_FORMAT |
         eDIRTY_WRAP_S_MODE |
         eDIRTY_MIN_FILTER | eDIRTY_MAG_FILTER |
         eDIRTY_MIPMAP);

   return m_valid_flag;
}

//----------------------------------------------------------------------------
bool sglTexture1D::setMipmapLevels(const vector<void *> &images)
{
   if (m_auto_mipmap_flag)
      return false;

   if (images.empty() && !m_mipmaps.empty())
   {
      m_mipmaps.clear();
      dirty(eDIRTY_RESET);
      return true;
   }

   if (images.size() != (unsigned int)m_pow_s)
      return false;

   if (m_mipmaps.empty())
   {
      for (vector<void*>::const_iterator it = images.begin();
           it != images.end(); ++it)
      {
         m_mipmaps.push_back(MipmapStruct(*it, eDIRTY_IMAGE));
      }
   }
   else
   {
      vector<MipmapStruct>::iterator elt = m_mipmaps.begin();
      for (vector<void*>::const_iterator it = images.begin();
           it != images.end(); ++it, ++elt)
      {
         elt->m_image = *it;
         elt->m_dirty |= eDIRTY_REPLACE;
      }
   }
   dirty(eDIRTY_MIPMAP);

   return true;
}

//----------------------------------------------------------------------------
unsigned int sglTexture1D::getLikelyInternalSize() const
{
   if (m_internal_size_dirty)
   {
      s_lock.acquire();
      if (m_internal_size_dirty)
      {
         glBindTexture(m_target, m_texture_object);

         GLint red, green, blue, alpha, lumin, intensity;
         glGetTexLevelParameteriv(
            GL_TEXTURE_2D, 0, GL_TEXTURE_RED_SIZE, &red);
         glGetTexLevelParameteriv(
            GL_TEXTURE_2D, 0, GL_TEXTURE_GREEN_SIZE, &green);
         glGetTexLevelParameteriv(
            GL_TEXTURE_2D, 0, GL_TEXTURE_BLUE_SIZE, &blue);
         glGetTexLevelParameteriv(
            GL_TEXTURE_2D, 0, GL_TEXTURE_ALPHA_SIZE, &alpha);
         glGetTexLevelParameteriv(
            GL_TEXTURE_2D, 0, GL_TEXTURE_LUMINANCE_SIZE, &lumin);
         glGetTexLevelParameteriv(
            GL_TEXTURE_2D, 0, GL_TEXTURE_INTENSITY_SIZE, &intensity);

//          sglPrint() << "red size = " << red << endl;
//          sglPrint() << "green size = " << green << endl;
//          sglPrint() << "blue size = " << blue << endl;
//          sglPrint() << "alpha size = " << alpha << endl;
//          sglPrint() << "lumin size = " << lumin << endl;
//          sglPrint() << "intensity size = " << intensity << endl;

         m_internal_size = m_width;
         if (m_bound_mipmaps || !m_mipmaps.empty())
            m_internal_size = 2*(m_internal_size + m_border*(m_pow_s - 1)) - 1;

         m_internal_size *= (red + green + blue + alpha + lumin + intensity);
         m_internal_size /= 8;

         m_internal_size_dirty = false;
      }
      s_lock.release();
   }
   return m_internal_size;
}

//----------------------------------------------------------------------------
unsigned int sglTexture1D::getIdealInternalSize() const
{
   unsigned int size = m_width;

   if (m_bound_mipmaps || !m_mipmaps.empty())
      size = 2*(size + m_border*(m_pow_s - 1)) - 1;

   size *= getBitsPerTexel(m_internal_format);

   return size/8;
}

//----------------------------------------------------------------------------
bool sglTexture1D::replaceImage(void *new_image, GLint level)
{
   if (!m_valid_flag || new_image == NULL || level < 0 || level > m_pow_s)
      return false;

   if (level == 0)
   {
      if (m_own_image && m_image != new_image)
      {
         free(m_image);
         m_own_image = false;
      }

      m_image = new_image;
      dirty(eDIRTY_REPLACE|eDIRTY_MIPMAP);

      if (m_auto_mipmap_flag)
         buildAutoMipmaps(eDIRTY_REPLACE);
   }
   else
   {
      if (m_auto_mipmap_flag || m_mipmaps.empty())
         return false;

      m_mipmaps[level-1].m_image = new_image;
      m_mipmaps[level-1].m_dirty |= eDIRTY_REPLACE;
      dirty(eDIRTY_MIPMAP);
   }
   return true;
}

//----------------------------------------------------------------------------
bool sglTexture1D::subloadImage(void *sub_image,
                                GLint level,
                                GLint offset_x,
                                GLint width)
{
   if (!m_valid_flag || sub_image == NULL
       || width <= 0 || offset_x < 0 || level < 0
       || level > m_pow_s)
      return false;

   if (level > 0 && !m_auto_mipmap_flag && m_mipmaps.empty())
      return false;

   m_subload_list.push_back(new SubloadImageStruct(sub_image, level,
                                                   offset_x,
                                                   width));
   dirty(eDIRTY_SUBIMAGE);
   return true;
}

//----------------------------------------------------------------------------
void sglTexture1D::setWrapMode(WrapEnum s_repeat_flag)
{
   GLint wrap_param = GL_REPEAT;
   switch (s_repeat_flag)
   {
      case eREPEAT:
         wrap_param = GL_REPEAT;
         break;
      case eCLAMP:
         wrap_param = GL_CLAMP;
         break;
      case eCLAMP_TO_EDGE:
         wrap_param = sgl::s_clamp_to_edge;
         break;
   }

   if (m_repeat_flag != wrap_param)
   {
      m_repeat_flag = wrap_param;
      dirty(eDIRTY_WRAP_S_MODE);
   }
}

//----------------------------------------------------------------------------
void sglTexture1D::getWrapMode(WrapEnum &s_repeat_flag) const
{
   if (m_repeat_flag == GL_REPEAT)
      s_repeat_flag = eREPEAT;
   else if (m_repeat_flag == GL_CLAMP)
      s_repeat_flag = eCLAMP;
   else if (m_repeat_flag == sgl::s_clamp_to_edge)
      s_repeat_flag = eCLAMP_TO_EDGE;
   else
      sglPrintWarning()
         << "sglTexture1D::getWrapMode: m_repeat_flag is invalid!" << endl;
}

//----------------------------------------------------------------------------
bool sglTexture1D::checkProxy()
{
   // test for size restrictions
   glTexImage1D(GL_PROXY_TEXTURE_1D, 0, m_internal_format,
                m_width, m_border,
                (GLenum)m_external_format, (GLenum)m_type, NULL);
   GLint width;
   glGetTexLevelParameteriv(GL_PROXY_TEXTURE_1D, 0,
                            GL_TEXTURE_WIDTH, &width);
   if (width == 0)
   {
      return false;
   }
   return true;
}

//----------------------------------------------------------------------------
//    Summary: this should only be called by the drawing thread with the
//             proper context current.
void sglTexture1D::bind()
{
   if (m_valid_flag && m_dirty_flag)
   {
      if ((m_dirty_flag & eDIRTY_RESET) && m_texture_object)
      {
         // delete any preexisting texture
         glDeleteTextures(1, &m_texture_object);
         m_texture_object = 0;
         m_dirty_flag = (unsigned int)~eDIRTY_RESET;
      }

      // create the texture object (delete any preexisting texture)
      if (!m_texture_object)
      {
         m_texture_object = sgl::getTextureObj();
         m_bound_mipmaps = false;

         if (m_texture_object == 0)
            return;
      }

      glBindTexture(GL_TEXTURE_1D, m_texture_object);

      if (m_dirty_flag & eDIRTY_WRAP_S_MODE)
      {
         glTexParameterf(GL_TEXTURE_1D,
                         GL_TEXTURE_WRAP_S, m_repeat_flag);
      }

      if (m_dirty_flag & eDIRTY_BORDER_COLOR)
      {
         glTexParameterfv(GL_TEXTURE_1D, GL_TEXTURE_BORDER_COLOR,
                          m_border_color.getPtr());
      }

      if (m_dirty_flag & eDIRTY_MAG_FILTER)
      {
         glTexParameteri(GL_TEXTURE_1D,
                         GL_TEXTURE_MAG_FILTER, m_mag_filter);
      }

      if (m_dirty_flag & eDIRTY_MIN_FILTER)
      {
         glTexParameteri(GL_TEXTURE_1D,
                         GL_TEXTURE_MIN_FILTER, m_min_filter);
      }

      if ((m_dirty_flag & eDIRTY_ANISOTROPY) &&
          sgl::s_max_texture_max_anisotropy != 1.0f)
      {
         glTexParameterf(GL_TEXTURE_1D,
                         (GLenum)0x84FE, // TEXTURE_MAX_ANISOTROPY_EXT
                         m_max_anisotropy);
      }

      // Should we allow the user to set these?
      glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
      glPixelStorei(GL_UNPACK_ROW_LENGTH, 0);
      glPixelStorei(GL_UNPACK_SKIP_ROWS, 0);
      glPixelStorei(GL_UNPACK_SKIP_PIXELS, 0);

      if (m_pre_bind_callback)
         (*m_pre_bind_callback)(this);

      if ((m_dirty_flag & eDIRTY_IMAGE) ||
          (m_dirty_flag & eDIRTY_INTERNAL_FORMAT))
      {
         //cerr << "glTexImage1D(0)" << endl;

         glTexImage1D(GL_TEXTURE_1D,
                      0, m_internal_format,
                      m_width, m_border,
                      (GLenum)m_external_format, (GLenum)m_type, m_image);
      }
      else if (m_dirty_flag & eDIRTY_REPLACE)
      {
         //cerr << "glTexSubImage1D(0)" << endl;
         glTexSubImage1D(GL_TEXTURE_1D,
                         0, 0, m_width,
                         (GLenum)m_external_format, (GLenum)m_type,
                         m_image);
      }

      if ((m_dirty_flag & eDIRTY_MIPMAP) && !m_mipmaps.empty())
      {
         GLint level = 1;
         for (vector<MipmapStruct>::iterator it = m_mipmaps.begin();
              it != m_mipmaps.end(); ++it, ++level)
         {
            GLint width = (level >= m_pow_s ? 1 : (1<<(m_pow_s-level)))
               + m_border*2;

            if ((m_dirty_flag & eDIRTY_INTERNAL_FORMAT) ||
                (it->m_dirty & eDIRTY_IMAGE))
            {
               //cerr << "glTexImage1D(" << level << ")" << endl;
               glTexImage1D(GL_TEXTURE_1D,
                            level, m_internal_format,
                            width, m_border,
                            (GLenum)m_external_format, (GLenum)m_type,
                            it->m_image);
            }
            else if (it->m_dirty & eDIRTY_REPLACE)
            {
               //cerr << "glTexSubImage1D(" << level << ")" << endl;
               glTexSubImage1D(GL_TEXTURE_1D,
                               level, 0, width,
                               (GLenum)m_external_format, (GLenum)m_type,
                               it->m_image);
            }

            it->m_dirty = 0;
         }
         m_bound_mipmaps = true;
      }

      if (m_dirty_flag & eDIRTY_SUBIMAGE)
      {
         while (m_subload_list.size())
         {
            SubloadImageStruct *subimage = m_subload_list.back();

            glTexSubImage1D(GL_TEXTURE_1D,
                            subimage->m_level,
                            subimage->m_offset_x,
                            subimage->m_width,
                            (GLenum)m_external_format, (GLenum)m_type,
                            subimage->m_sub_image);
            m_subload_list.pop_back();
            delete subimage;
         }
      }

      m_dirty_flag = 0;

      if (m_post_bind_callback)
         (*m_post_bind_callback)(this);
   }
}

//----------------------------------------------------------------------------
bool sglTexture1D::copyTexSubImage(GLint level,
                                   GLint offset_x,
                                   GLint win_x, GLint win_y,
                                   GLint width)
{
   if (!m_valid_flag
       || width <= 0
       || offset_x < 0 || level < 0
       || level > m_pow_s)
      return false;

   if (level > 0 && !m_auto_mipmap_flag && m_mipmaps.empty())
      return false;

   bind();
   apply(NULL);

   glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
   glPixelStorei(GL_UNPACK_ROW_LENGTH, 0);
   glPixelStorei(GL_UNPACK_SKIP_ROWS, 0);
   glPixelStorei(GL_UNPACK_SKIP_PIXELS, 0);

   glCopyTexSubImage1D(GL_TEXTURE_1D, level, offset_x,
                       win_x, win_y, width);
   //cerr << "glCopyTexSubImage1D" << endl;
   return true;
}


//----------------------------------------------------------------------------
template <class T>
void shrinkImage(T *base, T **mipmapptr, unsigned int &width,
                 unsigned int num_components, GLint border)
{
   unsigned int new_width = width;

   if (border)
      new_width = ((width-2)/2)+2;
   else
      new_width >>= 1;

//    cerr << "new_width = " << new_width << endl;

   const unsigned int new_size = new_width*num_components;
   const unsigned int border_size = border*num_components;

//    cerr << "new_size = " << new_size << endl;
//    cerr << "border_size = " << border_size << endl;

   T *mipmap = (T*)realloc(*mipmapptr, new_size*sizeof(T));
   *mipmapptr = mipmap;

   if (border)
   {
      // copy the borders
      T* s = base+(width-1)*num_components;
      T* d = mipmap+(new_width-1)*num_components;
      for (unsigned int i=0; i<num_components; ++i)
      {
         mipmap[i] = base[i];
         d[i] = s[i];
      }
   }

   // copy the center
   T *last = mipmap + new_size - border_size;
   T *src = base + border_size;
   T *dst = mipmap + border_size;
   while (dst != last)
   {
      T *src2 = src+num_components;
      for (unsigned int i=0; i<num_components; ++i)
         dst[i] = (T)(((double)src[i] + (double)src2[i]) * 0.5);
      dst += num_components;
      src += 2*num_components;
   }

   width = new_width;
//    cerr << "done" << endl;
}

//----------------------------------------------------------------------------
bool sglTexture1D::halveImage(sglTexture::ExternalFormatEnum format,
                              sglTexture::TypeEnum type,
                              unsigned int &num_s, bool border,
                              void *old_image, void *&new_image)
{
   unsigned int num_components = getNumComponents(format);

   switch (type)
   {
      case eBYTE:
         shrinkImage(
            (signed char*)old_image, (signed char**)&new_image,
            num_s, num_components, border);
         break;
      case eUNSIGNED_BYTE:
         shrinkImage(
            (unsigned char*)old_image, (unsigned char**)&new_image,
            num_s, num_components, border);
         break;
      case eSHORT:
         shrinkImage(
            (short*)old_image, (short**)&new_image,
            num_s, num_components, border);
         break;
      case eUNSIGNED_SHORT:
         shrinkImage(
            (unsigned short*)old_image, (unsigned short**)&new_image,
            num_s, num_components, border);
         break;
      case eINT:
         shrinkImage(
            (unsigned short*)old_image, (unsigned short**)&new_image,
            num_s, num_components, border);
         break;
      case eUNSIGNED_INT:
         shrinkImage(
            (unsigned int*)old_image, (unsigned int**)&new_image,
            num_s, num_components, border);
         break;
      case eFLOAT:
         shrinkImage(
            (float*)old_image, (float**)&new_image,
            num_s, num_components, border);
         break;
      case eBITMAP:
//          shrinkBitmapImage(old_image, new_image,
//                            num_s, num_components, border);
         sglPrintWarning()
            << "sglTexture1D::halveImage() currently doesn't support bitmap textures"
            << endl;
         return false;
      default:
         sglPrintWarning()
            << "sglTexture1D::halveImage() unkown data type = "
            << (int)type << endl;
         return false;
   }
   return true;
}

//----------------------------------------------------------------------------
bool sglTexture1D::buildMipmaps(vector<void*> &mipmaps)
{
   unsigned int num_images = (unsigned int)m_pow_s;
   unsigned int prev_num = mipmaps.size();

   for (unsigned int j=prev_num; j<num_images; ++j)
   {
      mipmaps.push_back(NULL);
   }
   for (unsigned int k=num_images; k<prev_num; ++k)
   {
      free(mipmaps.back());
      mipmaps.pop_back();
   }

   void *base = m_image;
   unsigned int width = m_width;
   for (unsigned int i=0; i<num_images; ++i)
   {
      if (!halveImage(m_external_format, m_type, width,
                      (m_border ? true : false), base, mipmaps[i]))
         return false;

      base = mipmaps[i];
   }

   return true;
}


//----------------------------------------------------------------------------
template <class T>
void sglTexture1DResizeImage(T *old_image, T **new_image,
                             unsigned int num_components,
                             unsigned int old_width,
                             unsigned int border,
                             unsigned int new_width)
{
   const unsigned int new_size = new_width*num_components;

   *new_image = (T*)realloc(*new_image, new_size*sizeof(T));

   if (border)
   {
      // copy the borders
      T* s = old_image+(old_width-1)*num_components;
      T* d = (*new_image)+(new_width-1)*num_components;
      for (unsigned int i=0; i<num_components; ++i)
      {
         (*new_image)[i] = old_image[i];
         d[i] = s[i];
      }

      old_width -= 2;
      new_width -= 2;
   }

   double xinc, xinc2;
   if (old_width <= new_width)
   {
      xinc = (double)old_width - 1.0;
      if (new_width > 1)
         xinc /= (double)new_width - 1.0;
      xinc2 = 1.0;
   }
   else
   {
      xinc = (double)old_width / (double)new_width;
      xinc2 = xinc;
   }

   T* nptr = *new_image + border*num_components;
   for (unsigned int x = 0; x < new_width; ++x)
   {
      double old_x_min = (double)x * xinc;
      double old_x_max = old_x_min + xinc2;

      // Note this assumes that images have at most four components.
      double val[4];
      val[0] = val[1] = val[2] = val[3] = 0;

      for (unsigned int old_x = (unsigned int)old_x_min;
           (double)old_x < old_x_max; ++old_x)
      {
         double x_factor = 1.0;
         if (old_x < old_x_min)
            x_factor = 1.0 - (old_x_min - old_x);
         else if (old_x == (unsigned int)old_x_max)
            x_factor = old_x_max - old_x;

         T* optr = old_image + (old_x + border) * num_components;
         for (unsigned int i=0; i<num_components; ++i)
         {
            val[i] += *optr * x_factor;
            ++optr;
         }
      }

      double tmp = 1.0 / (old_x_max - old_x_min);
      for (unsigned int i=0; i<num_components; ++i)
      {
         *nptr = (T)(val[i] * tmp);
         ++nptr;
      }
   }
}

//----------------------------------------------------------------------------
bool sglTexture1D::resizeImage(sglTexture::ExternalFormatEnum format,
                               sglTexture::TypeEnum type,
                               unsigned int old_s,
                               bool border,
                               unsigned int new_s,
                               void *old_image, void *&new_image)
{
   unsigned int num_components = getNumComponents(format);

   switch (type)
   {
      case sglTexture::eBYTE:
         sglTexture1DResizeImage((signed char*)old_image,
                                 (signed char**)&new_image,
                                 num_components, old_s,
                                 (unsigned int)(border ? 1 : 0),
                                 new_s);
         break;
      case sglTexture::eUNSIGNED_BYTE:
         sglTexture1DResizeImage((unsigned char*)old_image,
                                 (unsigned char**)&new_image,
                                 num_components, old_s,
                                 (unsigned int)(border ? 1 : 0),
                                 new_s);
         break;
      case sglTexture::eSHORT:
         sglTexture1DResizeImage((signed short*)old_image,
                                 (signed short**)&new_image,
                                 num_components, old_s,
                                 (unsigned int)(border ? 1 : 0),
                                 new_s);
         break;
      case sglTexture::eUNSIGNED_SHORT:
         sglTexture1DResizeImage((unsigned short*)old_image,
                                 (unsigned short**)&new_image,
                                 num_components, old_s,
                                 (unsigned int)(border ? 1 : 0),
                                 new_s);
         break;
      case sglTexture::eINT:
         sglTexture1DResizeImage((signed int*)old_image,
                                 (signed int**)&new_image,
                                 num_components, old_s,
                                 (unsigned int)(border ? 1 : 0),
                                 new_s);
         break;
      case sglTexture::eUNSIGNED_INT:
         sglTexture1DResizeImage((unsigned int*)old_image,
                                 (unsigned int**)&new_image,
                                 num_components, old_s,
                                 (unsigned int)(border ? 1 : 0),
                                 new_s);
         break;
      case sglTexture::eFLOAT:
         sglTexture1DResizeImage((float*)old_image,
                                 (float**)&new_image,
                                 num_components, old_s,
                                 (unsigned int)(border ? 1 : 0),
                                 new_s);
         break;
      case sglTexture::eBITMAP:
         sglPrintWarning()
            << "sglTexture1D::resizeImage()"
            << " currently doesn't support bitmap images"
            << endl;
         return false;
      default:
         sglPrintWarning()
            << "sglTexture1D::resizeImage() unkown data type = "
            << (int)type << endl;
         return false;
   }
   return true;
}

//----------------------------------------------------------------------------
void sglTexture1D::printInfo(ostream &ostrm, const char *indent_string) const
{
    sglTexture::printInfo(ostrm, indent_string);

    ostrm << indent_string << "width = " << m_width << endl;
    ostrm << indent_string << "wrap mode = " << m_repeat_flag << endl;
    ostrm << indent_string << "subload list size = " << m_subload_list.size()
          << endl;
}
