/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglTexture1D.hpp
 *  Created: 11 March 2000
 *  Summary: container for 1D image data used for texturing
 *****************************************************************************/

#ifndef __SGL_TEXTURE_1D_HPP
#define __SGL_TEXTURE_1D_HPP

#include <sgl.h>
#include <vector>
#include <sglObject.hpp>
#include <sglVector.hpp>
#include <sglTexture.hpp>

//============================================================================

/** container for 1D image data used for texturing
 */

class SGL_DLL_API sglTexture1D : public sglTexture
{
public:
   ///
   sglTexture1D();
   ///
   virtual ~sglTexture1D();

   /// Set the image to be textured (level zero).
   //   format = format of each texel in the image
   //   type   = data type of texel data
   //   num_s  = width of the image
   //   border = true if the image has a one pixel border
   // The num_s parameter should have the form 2^m +2b, where m is a
   //   non-negative integer and b is zero or one (depending on border).
   //   If num_s (minus the border) is not a power of two, or if the
   //   image is too large, the image will be automatically resized.
   // Returns false if the parameters are out of range.
   // Note: this will not free the memory from the previous image.
   // Note: this will clear the mipmap levels if not automatic.
   bool setImage(void  *image, ExternalFormatEnum format, TypeEnum type,
                 unsigned int  num_s, bool border = false);
   ///
   void getImage(void *&image, ExternalFormatEnum &format, TypeEnum &type,
                 unsigned int &num_s, bool &border) const;

   /** Set the mipmap images.  Returns false if number of images is not
       correct or if the AutoMipmap feature is on.  Assumes that the
       images are correct in size and format.
   */
   bool setMipmapLevels(const vector<void *> &images);

   /// Replace all or part of a contiguous subregion of the current image.
   //   sub_image = the new texels
   //   level     = mipmap level of image
   //   offset_x  = texel offset in the original image
   //   width     = width of the replacement image
   // Returns false if the parameters are out of range.
   // Note: sub-textures once loaded are forgotten and will not
   //   be restored if the base image parameters are changed.
   bool subloadImage(void *sub_image,
                     GLint level,
                     GLint offset_x,
                     GLint width);

   /** Helper function to replace an entire image via subtexload funtionality.
       The new_image data must be the same size as the original image set by
       setImage, and the pointer to the old image data is lost.
       Returns false if the parameters are out of range.
   */
   virtual bool replaceImage(void *new_image, GLint level = 0);

   /// Get the size of the level zero image.
   inline void getSize(unsigned int &width) const { width  = m_width; }

   /** Returns the number of bytes of texture memory this texture will
       use if the requested internal format was honored.
   */
   virtual unsigned int getIdealInternalSize() const;
   /** Returns the number of bytes of texture memory this texture is
       most likely using.
       Must be called by a thread with a current rendering context.
   */
   virtual unsigned int getLikelyInternalSize() const;

   /// Get/set whether the texture repeats or is stretched by tex coords > 1.
   void setWrapMode(WrapEnum s_repeat_flag);
   ///
   void getWrapMode(WrapEnum &s_repeat_flag) const;

   /** Use the GL_PROXY_TEXTURE mechanism to check if the texture
       is capable of being loaded into texture memory.
       Returns false if the texture is too large.
       Must be called by a thread with a current rendering context.
   */
   virtual bool checkProxy();

   /** Copy texture (and changes) into texture memory.
       If (getAutoBind() == false), this method must be called before
       before the texture can be applied.
       Must be called by a thread with a current rendering context.
   */
   virtual void bind();

   /// Replace all or part of a contiguous subregion of the current image.
   // The replacement texels are taken from the framebuffer.
   //   level     = mipmap level of image
   //   offset_x  = texel offset in the original image
   //   win_x     = origin of the replacement image in the frame buffer
   //   win_y     = origin of the replacement image in the frame buffer
   //   width     = width of the replacement image
   // Returns false if the parameters are out of range.
   // Note: sub-textures once loaded are forgotten and will not
   //   be restored if the base image parameters are changed.
   // Must be called by a thread with a current rendering context.
   bool copyTexSubImage(GLint level,
                        GLint offset_x,
                        GLint win_x, GLint win_y,
                        GLint width);

   /** Generates a set of mipmaps from the level 0 image in this texture.
       Existing images in the given mipmaps vector will be resized (via
       realloc) and excess images will be free()'d.
       Returns false if the mipmaps could not be built.
   */
   virtual bool buildMipmaps(vector<void*> &mipmaps);

   /** Creates a new image (using malloc) from another image, but with
       a different size.  Note that the new and old sizes are NOT constrained
       to be a power of two.  Also note that this method is slow.
       @param format the format of the image data
       @param type the type of each pixel
       @param old_s the width of the old image
       @param border true if the image has a one pixel border
       @param new_s the width of the new image
       @param old_image the old image
       @param new_image a pointer to the pointer to the new image;
                    this pointer will be realloced to the appropriate size
       @return false on failure
   */
   static bool resizeImage(sglTexture::ExternalFormatEnum format,
                           sglTexture::TypeEnum type,
                           unsigned int old_s,
                           bool border,
                           unsigned int new_s,
                           void *old_image, void *&new_image);

   /** Creates a new image one half the size (num_s/2) of another image.
       This method is much faster than resizeImage().
       Note that num_s must be divisible by two.
       @param format the format of the image data
       @param type the type of each pixel
       @param num_s should be called with the width of the old image,
                    on return this is set the the width of the new image
       @param border true if the image has a one pixel border
       @param old_image the old image
       @param new_image a pointer to a pointer to the new image;
                    this pointer will be realloced to the appropriate size
       @return false on failure
   */
   static bool halveImage(sglTexture::ExternalFormatEnum format,
                          sglTexture::TypeEnum type,
                          unsigned int &num_s,
                          bool border,
                          void *old_image, void *&new_image);

   // See sglObject::printInfo() for details.
   virtual void printInfo(ostream &ostrm, const char *indent_string) const;

private:
   sglTexture1D(const sglTexture1D&);               // NOT IMPLEMENTED
   sglTexture1D& operator=(const sglTexture1D &);   // NOT IMPLEMENTED

   // used internally for sub tex loading
   struct SubloadImageStruct
   {
      SubloadImageStruct(void* sub_image,
                         GLint level, GLint offset_x, GLint width)
            : m_sub_image(sub_image), m_level(level),
              m_offset_x(offset_x), m_width(width) {};

      void *m_sub_image;
      GLint m_level;
      GLint m_offset_x;
      GLint m_width;
   };

private:
   unsigned int m_width;
   GLint m_repeat_flag;
   GLint m_pow_s;

   vector <SubloadImageStruct*> m_subload_list;
};

#endif
