/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglTexture2D.cpp
 *  Created: 2 October 1997
 *  Summary: container for texture image data and corresponding functions
 *****************************************************************************/

#include <sgl.h>
#include <string.h>
#include <sglObject.hpp>
#include <sglTexture2D.hpp>
#include <sglOStream.hpp>
#include <sglLock.hpp>

//----------------------------------------------------------------------------
sglTexture2D::sglTexture2D() : sglTexture(GL_TEXTURE_2D),
   m_width(0),
   m_height(0),
   m_pow_s(0),
   m_pow_t(0)
{
   m_repeat_flag[0] = m_repeat_flag[1] = GL_REPEAT;
}

//----------------------------------------------------------------------------
sglTexture2D::~sglTexture2D()
{
   // Remove this texture from the dirty list _before_ allowing the
   // object to become partially destructed (ie this can't be done in
   // the base class, or bind could get called during destruction).
   s_lock.acquire();
   s_dirty_textures.erase(this);
   s_lock.release();
}

//----------------------------------------------------------------------------
void sglTexture2D::getImage(void *&image,
                            ExternalFormatEnum &format, TypeEnum &type,
                            unsigned int &num_s, unsigned int &num_t,
                            bool &border) const
{
   image = NULL;
   if (m_valid_flag)
   {
      image = m_image;
      format = m_external_format;
      type = m_type;
      num_s = m_width;
      num_t = m_height;
      border = (m_border ? true : false);
   }
}

//----------------------------------------------------------------------------
bool sglTexture2D::setImage(void *image,
                            ExternalFormatEnum format, TypeEnum type,
                            unsigned int num_s, unsigned int num_t,
                            bool border)
{
   unsigned int borderless_s = (border ? num_s - 2 : num_s);
   unsigned int borderless_t = (border ? num_t - 2 : num_t);
   unsigned int pow_s = 0;
   unsigned int pow_t = 0;
   for (unsigned int tmp_s = borderless_s; tmp_s > 1; ++pow_s) {tmp_s >>= 1;}
   for (unsigned int tmp_t = borderless_t; tmp_t > 1; ++pow_t) {tmp_t >>= 1;}

   if (borderless_s == 0 || borderless_t == 0 || format == 0 || image == NULL)
   {
      m_valid_flag = false;
      m_index = 0;

      dirty(eDIRTY_IMAGE);
      return false;
   }

   if ((GLint)borderless_s > sgl::s_max_texture_size
       || (GLint)borderless_t > sgl::s_max_texture_size
       || (1u<<pow_s) != borderless_s || (1u<<pow_t) != borderless_t)
   {
      unsigned int old_s = num_s;
      unsigned int old_t = num_t;

      if ((GLint)borderless_s > sgl::s_max_texture_size
          || (GLint)borderless_t > sgl::s_max_texture_size)
      {
         sglPrintWarning() << "sglTexture2D::setImage() size ("
                           << num_s << ", " << num_t
                           << ") > GL_MAX_TEXTURE_SIZE; auto resizing"
                           << endl;

         num_s = sgl::s_max_texture_size;
         num_t = sgl::s_max_texture_size;
         borderless_s = (border ? num_s - 2 : num_s);
         borderless_t = (border ? num_t - 2 : num_t);
         pow_s = 0;
         pow_t = 0;
         unsigned int tmp;
         for (tmp = borderless_s; tmp > 1; ++pow_s) {tmp >>= 1;}
         for (tmp = borderless_t; tmp > 1; ++pow_t) {tmp >>= 1;}
      }

      if ((1u<<pow_s) != borderless_s || (1u<<pow_t) != borderless_t)
      {
         sglPrintWarning() << "sglTexture2D::setImage() size ("
                           << num_s << ", " << num_t
                           << ") not powers of 2; auto resizing"
                           << endl;

         num_s = (1u<<pow_s) + (border ? 2 : 0);
         num_t = (1u<<pow_t) + (border ? 2 : 0);
      }

      if (!m_own_image)
         m_image = NULL;

      resizeImage(format, type, old_s, old_t, border,
                  num_s, num_t, image, m_image);

      m_own_image = true;
   }
   else
   {
      if (m_own_image && image != m_image)
      {
         free(m_image);
         m_own_image = false;
      }
      // note: a pointer to any preexisting image is lost
      m_image = image;
   }

   m_width           = num_s;
   m_height          = num_t;
   m_external_format = format;
   m_type            = type;
   m_border          = (border ? 1 : 0);
   m_pow_s           = pow_s;
   m_pow_t           = pow_t;

   // find some reasonable default
   if (!m_internal_format_flag)
   {
      switch (type)
      {
         case eSHORT:
         case eUNSIGNED_SHORT:
         case eINT:
         case eUNSIGNED_INT:
         case eFLOAT:
            switch (format)
            {
               case eRGBA:
                  m_internal_format = eINT_RGBA16;
                  break;
               case eALPHA:
                  m_internal_format = eINT_ALPHA16;
                  break;
               case eLUMINANCE:
                  m_internal_format = eINT_LUMINANCE16;
                  break;
               case eLUMINANCE_ALPHA:
                  m_internal_format = eINT_LUMINANCE16_ALPHA16;
                  break;
               default:
                  m_internal_format = eINT_RGB16;
            }
            break;
         default:
            switch (format)
            {
               case eRGBA:
                  m_internal_format = eINT_RGBA;
                  break;
               case eALPHA:
                  m_internal_format = eINT_ALPHA;
                  break;
               case eLUMINANCE:
                  m_internal_format = eINT_LUMINANCE;
                  break;
               case eLUMINANCE_ALPHA:
                  m_internal_format = eINT_LUMINANCE_ALPHA;
                  break;
               default:
                  m_internal_format = eINT_RGB;
            }
      }
   }

   if (m_auto_mipmap_flag == eMIP_SW)
      buildAutoMipmaps(eDIRTY_IMAGE);
   else
      m_mipmaps.clear();

   m_valid_flag = true;
   m_index = UINT_MAX;

      // not sure this is the correct set
   dirty(eDIRTY_IMAGE | eDIRTY_INTERNAL_FORMAT |
         eDIRTY_WRAP_S_MODE | eDIRTY_WRAP_T_MODE |
         eDIRTY_MIN_FILTER | eDIRTY_MAG_FILTER |
         eDIRTY_MIPMAP | eDIRTY_ANISOTROPY);

   return true;
}

//----------------------------------------------------------------------------
bool sglTexture2D::setMipmapLevels(const vector<void *> &images)
{
   if (m_auto_mipmap_flag)
      return false;

   if (images.empty() && !m_mipmaps.empty())
   {
      m_mipmaps.clear();
      dirty(eDIRTY_RESET);
      return true;
   }

   if (m_pow_s > m_pow_t)
   {
      if (images.size() != (unsigned int)m_pow_s)
         return false;
   }
   else
   {
      if (images.size() != (unsigned int)m_pow_t)
         return false;
   }

   if (m_mipmaps.empty())
   {
      for (vector<void*>::const_iterator it = images.begin();
           it != images.end(); ++it)
      {
         m_mipmaps.push_back(MipmapStruct(*it, eDIRTY_IMAGE));
      }
   }
   else
   {
      vector<MipmapStruct>::iterator elt = m_mipmaps.begin();
      for (vector<void*>::const_iterator it = images.begin();
           it != images.end(); ++it, ++elt)
      {
         elt->m_image = *it;
         elt->m_dirty |= eDIRTY_REPLACE;
      }
   }
   dirty(eDIRTY_MIPMAP);

   return true;
}

//----------------------------------------------------------------------------
unsigned int sglTexture2D::getLikelyInternalSize() const
{
   if (m_internal_size_dirty)
   {
      s_lock.acquire();
      if (m_internal_size_dirty)
      {
         glBindTexture(m_target, m_texture_object);

         GLint red, green, blue, alpha, lumin, intensity;
         glGetTexLevelParameteriv(
            GL_TEXTURE_2D, 0, GL_TEXTURE_RED_SIZE, &red);
         glGetTexLevelParameteriv(
            GL_TEXTURE_2D, 0, GL_TEXTURE_GREEN_SIZE, &green);
         glGetTexLevelParameteriv(
            GL_TEXTURE_2D, 0, GL_TEXTURE_BLUE_SIZE, &blue);
         glGetTexLevelParameteriv(
            GL_TEXTURE_2D, 0, GL_TEXTURE_ALPHA_SIZE, &alpha);
         glGetTexLevelParameteriv(
            GL_TEXTURE_2D, 0, GL_TEXTURE_LUMINANCE_SIZE, &lumin);
         glGetTexLevelParameteriv(
            GL_TEXTURE_2D, 0, GL_TEXTURE_INTENSITY_SIZE, &intensity);

//          sglPrint() << "red size = " << red << endl;
//          sglPrint() << "green size = " << green << endl;
//          sglPrint() << "blue size = " << blue << endl;
//          sglPrint() << "alpha size = " << alpha << endl;
//          sglPrint() << "lumin size = " << lumin << endl;
//          sglPrint() << "intensity size = " << intensity << endl;

         m_internal_size = getNumTexels();
         m_internal_size *= (red + green + blue + alpha + lumin + intensity);
         m_internal_size /= 8;

         m_internal_size_dirty = false;
      }
      s_lock.release();
   }
   return m_internal_size;
}

//----------------------------------------------------------------------------
unsigned int sglTexture2D::getIdealInternalSize() const
{
   unsigned int size = getNumTexels();
   size *= getBitsPerTexel(m_internal_format);
   return size/8;
}

//----------------------------------------------------------------------------
unsigned int sglTexture2D::getNumTexels() const
{
   unsigned int size = 0;

   if (m_bound_mipmaps || !m_mipmaps.empty())
   {
      unsigned int num_s = m_width;
      unsigned int num_t = m_height;

      if (m_border)
      {
         unsigned int num_levels = (m_pow_s > m_pow_t ? m_pow_s : m_pow_t) + 1;
         num_s -= 2;
         num_t -= 2;

         // add the number of border texels
         size = 4*(num_levels + num_s + num_t - 1) + 2*abs(m_pow_s - m_pow_t);
      }

      // add the number of interior texels
      if (num_s == num_t)
         size += (num_s * num_t * 4) / 3;
      else
      {
         while (num_s > 1 || num_t > 1)
         {
            size += num_s * num_t;
            if (num_s > 1) num_s >>= 1;
            if (num_t > 1) num_t >>= 1;
         }
         size += 1;
      }
   }
   else
      size = m_width * m_height;

   return size;
}

//----------------------------------------------------------------------------
bool sglTexture2D::replaceImage(void *new_image, GLint level)
{
   if (!m_valid_flag || new_image == NULL || level < 0 ||
       (level > m_pow_s && level > m_pow_t))
      return false;

   if (level == 0)
   {
      if (m_own_image && m_image != new_image)
      {
         free(m_image);
         m_own_image = false;
      }

      m_image = new_image;
      dirty(eDIRTY_REPLACE|eDIRTY_MIPMAP);

      if (m_auto_mipmap_flag == eMIP_SW)
         buildAutoMipmaps(eDIRTY_REPLACE);
   }
   else
   {
      if (m_auto_mipmap_flag || m_mipmaps.empty())
         return false;

      m_mipmaps[level-1].m_image = new_image;
      m_mipmaps[level-1].m_dirty |= eDIRTY_REPLACE;
      dirty(eDIRTY_MIPMAP);
   }
   return true;
}

//----------------------------------------------------------------------------
bool sglTexture2D::subloadImage(void *sub_image,
                                GLint level,
                                GLint offset_x, GLint offset_y,
                                GLint width, GLint height)
{
   if (!m_valid_flag || sub_image == NULL
       || width <= 0 || height <= 0
       || offset_x < 0 || offset_y < 0 || level < 0
       || (level > m_pow_s && level > m_pow_t))
      return false;

   if (level > 0 && (m_auto_mipmap_flag == eMIP_NONE) && m_mipmaps.empty())
      return false;

   m_subload_list.push_back(new SubloadImageStruct(sub_image, level,
                                                   offset_x, offset_y,
                                                   width, height));
   dirty(eDIRTY_SUBIMAGE);
   return true;
}

//----------------------------------------------------------------------------
void sglTexture2D::setWrapMode(WrapEnum s_repeat_flag, WrapEnum t_repeat_flag)
{
   GLint wrap_param = GL_REPEAT;
   switch (s_repeat_flag)
   {
      case eREPEAT:
         wrap_param = GL_REPEAT;
         break;
      case eCLAMP:
         wrap_param = GL_CLAMP;
         break;
      case eCLAMP_TO_EDGE:
         wrap_param = sgl::s_clamp_to_edge;
         break;
   }

   if (m_repeat_flag[0] != wrap_param)
   {
      m_repeat_flag[0] = wrap_param;
      dirty(eDIRTY_WRAP_S_MODE);
   }

   switch (t_repeat_flag)
   {
      case eREPEAT:
         wrap_param = GL_REPEAT;
         break;
      case eCLAMP:
         wrap_param = GL_CLAMP;
         break;
      case eCLAMP_TO_EDGE:
         wrap_param = sgl::s_clamp_to_edge;
         break;
   }

   if (m_repeat_flag[1] != wrap_param)
   {
      m_repeat_flag[1] = wrap_param;
      dirty(eDIRTY_WRAP_T_MODE);
   }
}

//----------------------------------------------------------------------------
void sglTexture2D::getWrapMode(WrapEnum &s_repeat_flag,
                               WrapEnum &t_repeat_flag) const
{
   if (m_repeat_flag[0] == GL_REPEAT)
      s_repeat_flag = eREPEAT;
   else if (m_repeat_flag[0] == GL_CLAMP)
      s_repeat_flag = eCLAMP;
   else if (m_repeat_flag[0] == sgl::s_clamp_to_edge)
      s_repeat_flag = eCLAMP_TO_EDGE;
   else
      sglPrintWarning()
         << "sglTexture2D::getWrapMode: m_repeat_flag[0] is invalid!"
         << endl;

   if (m_repeat_flag[1] == GL_REPEAT)
      t_repeat_flag = eREPEAT;
   else if (m_repeat_flag[1] == GL_CLAMP)
      t_repeat_flag = eCLAMP;
   else if (m_repeat_flag[1] == sgl::s_clamp_to_edge)
      t_repeat_flag = eCLAMP_TO_EDGE;
   else
      sglPrintWarning()
         << "sglTexture2D::getWrapMode: m_repeat_flag[1] is invalid!"
         << endl;
}

//----------------------------------------------------------------------------
bool sglTexture2D::checkProxy()
{
   // test for size restrictions
   glTexImage2D(GL_PROXY_TEXTURE_2D, 0,
               (m_compression_flag)?m_compression_enum:m_internal_format,
                m_width, m_height, m_border,
                (GLenum)m_external_format, (GLenum)m_type, NULL);
   GLint width;
   glGetTexLevelParameteriv(GL_PROXY_TEXTURE_2D, 0,
                            GL_TEXTURE_WIDTH, &width);
   if (width == 0)
   {
      return false;
   }
   return true;
}

//----------------------------------------------------------------------------
//    Summary: this should only be called by the drawing thread with the
//             proper context current.
void sglTexture2D::bind()
{
   if (m_valid_flag && m_dirty_flag)
   {
      if ((m_dirty_flag & eDIRTY_RESET) && m_texture_object)
      {
         // delete any preexisting texture
         glDeleteTextures(1, &m_texture_object);
         m_texture_object = 0;
         m_dirty_flag = (unsigned int)~eDIRTY_RESET;
      }

      // create the texture object
      if (!m_texture_object)
      {
         m_texture_object = sgl::getTextureObj();
         m_bound_mipmaps = false;

         if (m_texture_object == 0)
            return;
      }

      glBindTexture(GL_TEXTURE_2D, m_texture_object);

      if (m_dirty_flag & eDIRTY_WRAP_S_MODE)
      {
         glTexParameterf(GL_TEXTURE_2D,
                         GL_TEXTURE_WRAP_S, m_repeat_flag[0]);
      }

      if (m_dirty_flag & eDIRTY_WRAP_T_MODE)
      {
         glTexParameterf(GL_TEXTURE_2D,
                         GL_TEXTURE_WRAP_T, m_repeat_flag[1]);
      }

      if (m_dirty_flag & eDIRTY_BORDER_COLOR)
      {
         glTexParameterfv(GL_TEXTURE_2D, GL_TEXTURE_BORDER_COLOR,
                          m_border_color.getPtr());
      }

      if (m_dirty_flag & eDIRTY_MAG_FILTER)
      {
         glTexParameteri(GL_TEXTURE_2D,
                         GL_TEXTURE_MAG_FILTER, m_mag_filter);
      }

      if (m_dirty_flag & eDIRTY_MIN_FILTER)
      {
         glTexParameteri(GL_TEXTURE_2D,
                         GL_TEXTURE_MIN_FILTER, m_min_filter);
      }

      if ((m_dirty_flag & eDIRTY_ANISOTROPY) &&
          sgl::s_max_texture_max_anisotropy != 1.0f)
      {
         glTexParameterf(GL_TEXTURE_2D,
                         (GLenum)0x84FE, // TEXTURE_MAX_ANISOTROPY_EXT
                         m_max_anisotropy);
      }

      if ((m_dirty_flag & eDIRTY_LOD) && m_texture_lod_active )
      {
         glTexParameterf(GL_TEXTURE_2D,
                         GL_TEXTURE_MIN_LOD_SGIS,
                         m_texture_lod_min);

         glTexParameterf(GL_TEXTURE_2D,
                         GL_TEXTURE_MAX_LOD_SGIS,
                         m_texture_lod_max);

         glTexParameteri(GL_TEXTURE_2D,
                         GL_TEXTURE_BASE_LEVEL_SGIS,
                         m_texture_lod_base_level);

         glTexParameteri(GL_TEXTURE_2D,
                         GL_TEXTURE_MAX_LEVEL_SGIS,
                         m_texture_lod_max_level);
      }

      // Should we allow the user to set these?
      glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
      glPixelStorei(GL_UNPACK_ROW_LENGTH, 0);
      glPixelStorei(GL_UNPACK_SKIP_ROWS, 0);
      glPixelStorei(GL_UNPACK_SKIP_PIXELS, 0);

      if (m_pre_bind_callback)
         (*m_pre_bind_callback)(this);

      if ((m_dirty_flag & eDIRTY_MIPMAP) && sgl::s_auto_mipmap_capable)
      {
         if (m_auto_mipmap_flag == eMIP_HW)
         {
            glTexParameteri(
               GL_TEXTURE_2D, sgl::s_auto_mipmap_enable_flag, GL_TRUE);

            // set to show there are bound mip maps
            m_bound_mipmaps = true;
         }
         else
         {
            glTexParameteri(
               GL_TEXTURE_2D, sgl::s_auto_mipmap_enable_flag, GL_FALSE);
         }
      }

      if ((m_dirty_flag & eDIRTY_IMAGE) ||
          (m_dirty_flag & eDIRTY_INTERNAL_FORMAT))
      {
         //cerr << "glTexImage2D(0)" << endl;
         m_internal_size_dirty = true;

         glTexImage2D(GL_TEXTURE_2D, 0,
                  (m_compression_flag)?m_compression_enum:m_internal_format,
                      m_width, m_height, m_border,
                      (GLenum)m_external_format, (GLenum)m_type, m_image);
      }
      else if (m_dirty_flag & eDIRTY_REPLACE)
      {
         //cerr << "glTexSubImage2D(0)" << endl;
         glTexSubImage2D(GL_TEXTURE_2D,
                         0, 0, 0, m_width, m_height,
                         (GLenum)m_external_format, (GLenum)m_type,
                         m_image);
      }

      if ((m_dirty_flag & eDIRTY_MIPMAP) &&
          m_auto_mipmap_flag != eMIP_HW &&
          !m_mipmaps.empty())
      {
         GLint level = 1;
         for (vector<MipmapStruct>::iterator it = m_mipmaps.begin();
              it != m_mipmaps.end(); ++it, ++level)
         {
            GLint width = (level >= m_pow_s ? 1 : (1<<(m_pow_s-level)))
               + m_border*2;
            GLint height = (level >= m_pow_t ? 1 : (1<<(m_pow_t-level)))
               + m_border*2;

            if ((m_dirty_flag & eDIRTY_INTERNAL_FORMAT) ||
                (it->m_dirty & eDIRTY_IMAGE))
            {
               //cerr << "glTexImage2D(" << level << ")" << endl;
               m_internal_size_dirty = true;

               glTexImage2D(
                  GL_TEXTURE_2D, level,
                  (m_compression_flag)?m_compression_enum:m_internal_format,
                  width, height, m_border,
                  (GLenum)m_external_format, (GLenum)m_type,
                  it->m_image);
            }
            else if (it->m_dirty & eDIRTY_REPLACE)
            {
               //cerr << "glTexSubImage2D(" << level << ")" << endl;
               glTexSubImage2D(GL_TEXTURE_2D,
                               level, 0, 0, width, height,
                               (GLenum)m_external_format, (GLenum)m_type,
                               it->m_image);
            }

            it->m_dirty = 0;
         }

         // set to show there are bound mip maps
         m_bound_mipmaps = true;
      }

      if (m_dirty_flag & eDIRTY_SUBIMAGE)
      {
         while (m_subload_list.size())
         {
            SubloadImageStruct *subimage = m_subload_list.back();

            glTexSubImage2D(GL_TEXTURE_2D,
                            subimage->m_level,
                            subimage->m_offset_x,
                            subimage->m_offset_y,
                            subimage->m_width,
                            subimage->m_height,
                            (GLenum)m_external_format, (GLenum)m_type,
                            subimage->m_sub_image);
            m_subload_list.pop_back();
            delete subimage;
         }
      }

      m_dirty_flag = 0;

      if (m_post_bind_callback)
         (*m_post_bind_callback)(this);
   }
}

//----------------------------------------------------------------------------
bool sglTexture2D::copyTexSubImage(GLint level,
                                   GLint offset_x, GLint offset_y,
                                   GLint win_x, GLint win_y,
                                   GLint width, GLint height)
{
   if (!m_valid_flag
       || width <= 0 || height <= 0
       || offset_x < 0 || offset_y < 0 || level < 0
       || (level > m_pow_s && level > m_pow_t))
      return false;

   if (level > 0 && (m_auto_mipmap_flag == eMIP_NONE) && m_mipmaps.empty())
      return false;

   bind();
   apply(NULL);

   glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
   glPixelStorei(GL_UNPACK_ROW_LENGTH, 0);
   glPixelStorei(GL_UNPACK_SKIP_ROWS, 0);
   glPixelStorei(GL_UNPACK_SKIP_PIXELS, 0);

   glCopyTexSubImage2D(GL_TEXTURE_2D, level, offset_x, offset_y,
                       win_x, win_y, width, height);
   //cerr << "glCopyTexSubImage2D" << endl;
   return true;
}

//----------------------------------------------------------------------------
template <class T>
void copy_texel(T *dst, T *src, unsigned int num_components)
{
   for (unsigned int i=0; i<num_components; ++i)
      dst[i] = src[i];
}

//----------------------------------------------------------------------------
template <class T>
void blend_texel(T *dst, T *src1, T *src2,
                 unsigned int num_components)
{
   for (unsigned int i=0; i<num_components; ++i)
      dst[i] = (T)(((double)src1[i] + (double)src2[i]) * 0.5);
}

//----------------------------------------------------------------------------
template <class T>
void blend_texel(T *dst, T *src1, T *src2, T *src3, T *src4,
                 unsigned int num_components)
{
   for (unsigned int i=0; i<num_components; ++i)
      dst[i] = (T)(((double)src1[i] + (double)src2[i] +
                    (double)src3[i] + (double)src4[i]) * 0.25);
}

//----------------------------------------------------------------------------
template <class T>
void sglTexture2DShrinkImageWithBorder(
   T *base, T **mipmapptr, unsigned int &width,
   unsigned int &height, unsigned int num_components)
{
   unsigned int new_width = width;
   unsigned int new_height = height;

   if (width != 3)
      new_width = ((width-2)/2)+2;

   if (height != 3)
      new_height = ((height-2)/2)+2;

//    cerr << "new_width = " << new_width << endl;
//    cerr << "new_height = " << new_height << endl;

   const unsigned int new_row_size = new_width*num_components;
   const unsigned int row_size = width*num_components;

//    cerr << "new_row_size = " << new_row_size << endl;
//    cerr << "row_size = " << row_size << endl;

   T *mipmap = (T*)realloc(*mipmapptr, new_height*new_row_size*sizeof(T));
   *mipmapptr = mipmap;

   // copy the corners
   copy_texel(mipmap, base, num_components);
   copy_texel(mipmap+(new_width-1)*num_components,
              base+(width-1)*num_components,
              num_components);
   copy_texel(mipmap+(new_height-1)*new_row_size,
              base+(height-1)*row_size,
              num_components);
   copy_texel(mipmap+new_height*new_row_size-num_components,
              base+height*row_size-num_components,
              num_components);

//    cerr << "after corners" << endl;

   if (height == 3)
   {
      // copy left border
      copy_texel(mipmap+new_row_size, base+row_size, num_components);
      // copy right border
      copy_texel(mipmap+2*new_row_size-num_components,
                 base+2*row_size-num_components,
                 num_components);
   }
   else
   {
      // copy left border
      T *last = mipmap + (new_height-1)*new_row_size;
      T *src = base+row_size;
      T *dst = mipmap+new_row_size;
      while (dst != last)
      {
         blend_texel(dst, src, src+row_size, num_components);
         dst += new_row_size;
         src += 2*row_size;
      }
      // copy right border
      last = mipmap + new_height*new_row_size - num_components;
      src = base + 2*row_size - num_components;
      dst = mipmap + 2*new_row_size - num_components;
      while (dst != last)
      {
         blend_texel(dst, src, src+row_size, num_components);
         dst += new_row_size;
         src += 2*row_size;
      }
   }

//    cerr << "after left/right" << endl;

   if (width == 3)
   {
      // copy top border
      copy_texel(mipmap+num_components, base+num_components, num_components);
      // copy bottom border
      copy_texel(mipmap+(new_row_size*(new_height-1))+num_components,
                 base+(row_size*(height-1))+num_components,
                 num_components);
   }
   else
   {
      // copy top border
      T *last = mipmap + new_row_size - num_components;
      T *src = base + num_components;
      T *dst = mipmap + num_components;
      while (dst != last)
      {
         blend_texel(dst, src, src+num_components, num_components);
         dst += num_components;
         src += 2*num_components;
      }
      // copy bottom border
      last = mipmap + new_height*new_row_size - num_components;
      src = base + (height-1)*row_size + num_components;
      dst = mipmap + (new_height-1)*new_row_size + num_components;
      while (dst != last)
      {
         blend_texel(dst, src, src+num_components, num_components);
         dst += num_components;
         src += 2*num_components;
      }
   }

//    cerr << "after top/bottom" << endl;

   // copy the center
   if (width == 3)
   {
      T *last = mipmap + (new_height-1)*new_row_size + num_components;
      T *src = base + row_size + num_components;
      T *dst = mipmap + new_row_size + num_components;
      while (dst != last)
      {
         blend_texel(dst, src, src+row_size, num_components);
         dst += new_row_size;
         src += 2*row_size;
      }
   }
   else if (height == 3)
   {
      T *last = mipmap + 2*new_row_size - num_components;
      T *src = base + row_size + num_components;
      T *dst = mipmap + new_row_size + num_components;
      while (dst != last)
      {
         blend_texel(dst, src, src+num_components, num_components);
         dst += num_components;
         src += 2*num_components;
      }
   }
   else
   {
      T *src = base + row_size + num_components;
      T *dst = mipmap + new_row_size + num_components;
      for (unsigned int y = 1; y < new_height-1; ++y)
      {
         for (unsigned int x = 1; x < new_width-1; ++x)
         {
            blend_texel(dst, src, src+num_components,
                        src+row_size, src+row_size+num_components,
                        num_components);
            dst += num_components;
            src += 2*num_components;
         }
         dst += 2*num_components;
         src += 2*num_components + row_size;
      }
   }

   width = new_width;
   height = new_height;
//    cerr << "done" << endl;
}

//----------------------------------------------------------------------------
template <class T>
void sglTexture2DShrinkImage(T *base, T **mipmapptr, unsigned int &width,
                             unsigned int &height, unsigned int num_components)
{
   unsigned int new_width = width;
   unsigned int new_height = height;

   if (width != 1)
      new_width >>= 1;

   if (height != 1)
      new_height >>= 1;

//    cerr << "new_width = " << new_width << endl;
//    cerr << "new_height = " << new_height << endl;

   const unsigned int new_row_size = new_width*num_components;
   const unsigned int row_size = width*num_components;
   const unsigned int new_size = new_height*new_row_size;

//    cerr << "new_row_size = " << new_row_size << endl;
//    cerr << "row_size = " << row_size << endl;
//    cerr << "new_size = " << new_size << endl;

   T *mipmap = (T*)realloc(*mipmapptr, new_size*sizeof(T));
   *mipmapptr = mipmap;

   if (width == 1 || height == 1)
   {
      T *last = mipmap + new_size;
      T *src = base;
      T *dst = mipmap;
      while (dst != last)
      {
         blend_texel(dst, src, src+num_components, num_components);
         dst += num_components;
         src += 2*num_components;
      }
   }
   else
   {
      T *src = base;
      T *dst = mipmap;
      for (unsigned int y = 0; y < new_height; ++y)
      {
         for (unsigned int x = 0; x < new_width; ++x)
         {
            blend_texel(dst, src, src+num_components,
                        src+row_size, src+row_size+num_components,
                        num_components);
            dst += num_components;
            src += 2*num_components;
         }
         src += row_size;
      }
   }

   width = new_width;
   height = new_height;
//    cerr << "done" << endl;
}

//----------------------------------------------------------------------------
bool sglTexture2D::halveImage(sglTexture::ExternalFormatEnum format,
                              sglTexture::TypeEnum type,
                              unsigned int &num_s, unsigned int &num_t,
                              bool border,
                              void *old_image, void *&new_image)
{
   unsigned int num_components = getNumComponents(format);

   switch (type)
   {
      case eBYTE:
         if (border)
            sglTexture2DShrinkImageWithBorder(
               (signed char*)old_image, (signed char**)&new_image,
               num_s, num_t, num_components);
         else
            sglTexture2DShrinkImage(
               (signed char*)old_image, (signed char**)&new_image,
               num_s, num_t, num_components);
         break;
      case eUNSIGNED_BYTE:
         if (border)
            sglTexture2DShrinkImageWithBorder(
               (unsigned char*)old_image, (unsigned char**)&new_image,
               num_s, num_t, num_components);
         else
            sglTexture2DShrinkImage(
               (unsigned char*)old_image, (unsigned char**)&new_image,
               num_s, num_t, num_components);
         break;
      case eSHORT:
         if (border)
            sglTexture2DShrinkImageWithBorder(
               (short*)old_image, (short**)&new_image,
               num_s, num_t, num_components);
         else
            sglTexture2DShrinkImage(
               (short*)old_image, (short**)&new_image,
               num_s, num_t, num_components);
         break;
      case eUNSIGNED_SHORT:
         if (border)
            sglTexture2DShrinkImageWithBorder(
               (unsigned short*)old_image, (unsigned short**)&new_image,
               num_s, num_t, num_components);
         else
            sglTexture2DShrinkImage(
               (unsigned short*)old_image, (unsigned short**)&new_image,
               num_s, num_t, num_components);
         break;
      case eINT:
         if (border)
            sglTexture2DShrinkImageWithBorder(
               (int*)old_image, (int**)&new_image,
               num_s, num_t, num_components);
         else
            sglTexture2DShrinkImage(
               (unsigned short*)old_image, (unsigned short**)&new_image,
               num_s, num_t, num_components);
         break;
      case eUNSIGNED_INT:
         if (border)
            sglTexture2DShrinkImageWithBorder(
               (unsigned int*)old_image, (unsigned int**)&new_image,
               num_s, num_t, num_components);
         else
            sglTexture2DShrinkImage(
               (unsigned int*)old_image, (unsigned int**)&new_image,
               num_s, num_t, num_components);
         break;
      case eFLOAT:
         if (border)
            sglTexture2DShrinkImageWithBorder(
               (float*)old_image, (float**)&new_image,
               num_s, num_t, num_components);
         else
            sglTexture2DShrinkImage(
               (float*)old_image, (float**)&new_image,
               num_s, num_t, num_components);
         break;
      case eBITMAP:
//          sglTexture2DShrinkBitmapImage(old_image, new_image,
//                                        num_s, num_t, num_components);
         sglPrintWarning()
            << "sglTexture2D::halveImage() currently doesn't support bitmap textures"
            << endl;
         return false;
      default:
         sglPrintWarning()
            << "sglTexture2D::halveImage() unkown data type = "
            << (int)type << endl;
         return false;
   }
   return true;
}

//----------------------------------------------------------------------------
bool sglTexture2D::buildMipmaps(vector<void*> &mipmaps)
{
   unsigned int num_images;
   if (m_pow_s > m_pow_t)
      num_images = (unsigned int)m_pow_s;
   else
      num_images = (unsigned int)m_pow_t;

   unsigned int prev_num = mipmaps.size();
   for (unsigned int j=prev_num; j<num_images; ++j)
   {
      mipmaps.push_back(NULL);
   }
   for (unsigned int k=num_images; k<prev_num; ++k)
   {
      free(mipmaps.back());
      mipmaps.pop_back();
   }

   void *base = m_image;
   unsigned int width = m_width;
   unsigned int height = m_height;
   for (unsigned int i=0; i<num_images; ++i)
   {
      if (!halveImage(m_external_format, m_type, width, height,
                      (m_border ? true : false), base, mipmaps[i]))
         return false;

      base = mipmaps[i];
   }

   return true;
}


//----------------------------------------------------------------------------
template <class T>
void sglTexture2DResizeImage(T *old_image, T **new_image,
                             unsigned int num_components,
                             unsigned int old_width, unsigned int old_height,
                             unsigned int border,
                             unsigned int new_width, unsigned int new_height)
{
   const unsigned int new_size = new_height*new_width*num_components;
   const unsigned int old_row_size = old_width*num_components;
   const unsigned int new_row_size = new_width*num_components;

   *new_image = (T*)realloc(*new_image, new_size*sizeof(T));
   T* nimage = *new_image;

   if (border)
   {
      // copy the corners
      copy_texel(*new_image, old_image, num_components);
      copy_texel(*new_image+(new_width-1)*num_components,
                 old_image+(old_width-1)*num_components,
                 num_components);
      copy_texel(*new_image+(new_height-1)*new_row_size,
                 old_image+(old_height-1)*old_row_size,
                 num_components);
      copy_texel(*new_image+new_height*new_row_size-num_components,
                 old_image+old_height*old_row_size-num_components,
                 num_components);

      old_height -= 2;
      old_width -= 2;
      new_height -= 2;
      new_width -= 2;

      double xinc, yinc, xinc2, yinc2;
      if (old_height <= new_height)
      {
         yinc = (double)old_height - 1.0;
         if (new_height > 1)
            yinc /= (double)new_height - 1.0;
         yinc2 = 1.0;
      }
      else
      {
         yinc = (double)old_height / (double)new_height;
         yinc2 = yinc;
      }

      if (old_width <= new_width)
      {
         xinc = (double)old_width - 1.0;
         if (new_width > 1)
            xinc /= (double)new_width - 1.0;
         xinc2 = 1.0;
      }
      else
      {
         xinc = (double)old_width / (double)new_width;
         xinc2 = xinc;
      }

      // filter the left & right sides
      T* nptr1 = nimage + new_row_size;
      T* nptr2 = nimage + new_row_size + new_row_size - num_components;
      for (unsigned int y = 0; y < new_height; ++y)
      {
         double old_y_min = (double)y * yinc;
         double old_y_max = old_y_min + yinc2;

         // Note this assumes that images have at most four components.
         double val1[4], val2[4];
         val1[0] = val1[1] = val1[2] = val1[3] = 0;
         val2[0] = val2[1] = val2[2] = val2[3] = 0;

         for (unsigned int old_y = (unsigned int)old_y_min;
              (double)old_y < old_y_max; ++old_y)
         {
            double y_factor = 1.0;
            if (old_y < old_y_min)
               y_factor = 1.0 - (old_y_min - old_y);
            else if (old_y == (unsigned int)old_y_max)
               y_factor = old_y_max - old_y;

            unsigned int orowindex = old_row_size * (old_y+1);

            T* optr1 = old_image + orowindex;
            T* optr2 = old_image + orowindex + (old_width+1) * num_components;
            for (unsigned int i=0; i<num_components; ++i)
            {
               val1[i] += *optr1 * y_factor;
               ++optr1;
               val2[i] += *optr2 * y_factor;
               ++optr2;
            }
         }

         double tmp = 1.0 / (old_y_max - old_y_min);
         for (unsigned int i=0; i<num_components; ++i)
         {
            nptr1[i] = (T)(val1[i] * tmp);
            nptr2[i] = (T)(val2[i] * tmp);
         }
         nptr1 += new_row_size;
         nptr2 += new_row_size;
      }

      // filter the top & bottom sides
      nptr1 = nimage + num_components;
      nptr2 = nimage + (new_height+1)*new_row_size + num_components;
      for (unsigned int x = 0; x < new_width; ++x)
      {
         double old_x_min = (double)x * xinc;
         double old_x_max = old_x_min + xinc2;

         // Note this assumes that images have at most four components.
         double val1[4], val2[4];
         val1[0] = val1[1] = val1[2] = val1[3] = 0;
         val2[0] = val2[1] = val2[2] = val2[3] = 0;

         for (unsigned int old_x = (unsigned int)old_x_min;
              (double)old_x < old_x_max; ++old_x)
         {
            double x_factor = 1.0;
            if (old_x < old_x_min)
               x_factor = 1.0 - (old_x_min - old_x);
            else if (old_x == (unsigned int)old_x_max)
               x_factor = old_x_max - old_x;

            unsigned int ocolindex = (old_x+1) * num_components;
            T* optr1 = old_image + ocolindex;
            T* optr2 = old_image + ocolindex + (old_height+1)*old_row_size;
            for (unsigned int i=0; i<num_components; ++i)
            {
               val1[i] += *optr1 * x_factor;
               ++optr1;
               val2[i] += *optr2 * x_factor;
               ++optr2;
            }
         }

         double tmp = 1.0 / (old_x_max - old_x_min);
         for (unsigned int i=0; i<num_components; ++i)
         {
            *nptr1 = (T)(val1[i] * tmp);
            *nptr2 = (T)(val2[i] * tmp);
            ++nptr1;
            ++nptr2;
         }
      }
   }

   // filter the interior
   if (old_height <= new_height && old_width <= new_width)
   {
      // Note that the else case would work, but this is faster.
      double xinc = (double)old_width - 1.0;
      if (new_width > 1)
         xinc /= (double)new_width - 1.0;
      double yinc = (double)old_height - 1.0;
      if (new_height > 1)
         yinc /= (double)new_height - 1.0;

      for (unsigned int y = 0; y < new_height; ++y)
      {
         double old_y = (double)y * yinc;
         unsigned int y1 = (unsigned int)old_y;
         unsigned int y2 = y1 + 1;

         if (y2 == old_height)
            --y2;

         double yfactor2 = old_y - (double)y1;
         double yfactor1 = 1.0 - yfactor2;

         unsigned int orowindex1 = old_row_size * (y1 + border);
         unsigned int orowindex2 = old_row_size * (y2 + border);
         unsigned int rowindex = new_row_size * (y + border);

         T* nptr = nimage + rowindex + border*num_components;
         T* optr1 = old_image + orowindex1 + border*num_components;
         T* optr2 = old_image + orowindex2 + border*num_components;

         for (unsigned int x = 0; x < new_width; ++x)
         {
            double old_x = (double)x * xinc;
            unsigned int x1 = (unsigned int)old_x;
            unsigned int x2 = x1 + 1;

            if (x2 == old_width)
               --x2;

            double xfactor2 = old_x - (double)x1;
            double xfactor1 = 1.0 - xfactor2;

            unsigned int ocolindex1 = num_components * x1;
            unsigned int ocolindex2 = num_components * x2;

            T* optr11 = optr1 + ocolindex1;
            T* optr12 = optr1 + ocolindex2;
            T* optr21 = optr2 + ocolindex1;
            T* optr22 = optr2 + ocolindex2;
            double factor11 = yfactor1 * xfactor1;
            double factor12 = yfactor1 * xfactor2;
            double factor21 = yfactor2 * xfactor1;
            double factor22 = yfactor2 * xfactor2;

            for (unsigned int i=0; i<num_components; ++i)
            {
               double val = (double)*optr11 * factor11;
               val += (double)*optr12 * factor12;
               val += (double)*optr21 * factor21;
               val += (double)*optr22 * factor22;
               ++optr11; ++optr12; ++optr21; ++optr22;

               *nptr = (T)val;
               ++nptr;
            }
         }
      }
   }
   else
   {
      double xinc, yinc, xinc2, yinc2;
      if (old_height <= new_height)
      {
         yinc = (double)old_height - 1.0;
         if (new_height > 1)
            yinc /= (double)new_height - 1.0;
         yinc2 = 1.0;
      }
      else
      {
         yinc = (double)old_height / (double)new_height;
         yinc2 = yinc;
      }

      if (old_width <= new_width)
      {
         xinc = (double)old_width - 1.0;
         if (new_width > 1)
            xinc /= (double)new_width - 1.0;
         xinc2 = 1.0;
      }
      else
      {
         xinc = (double)old_width / (double)new_width;
         xinc2 = xinc;
      }

      for (unsigned int y = 0; y < new_height; ++y)
      {
         double old_y_min = (double)y * yinc;
         double old_y_max = old_y_min + yinc2;

         unsigned int rowindex = new_row_size * (y + border);
         T* nptr = nimage + rowindex + border*num_components;

         for (unsigned int x = 0; x < new_width; ++x)
         {
            double old_x_min = (double)x * xinc;
            double old_x_max = old_x_min + xinc2;

            // Note this assumes that images have at most four components.
            double val[4];
            val[0] = val[1] = val[2] = val[3] = 0;

            for (unsigned int old_y = (unsigned int)old_y_min;
                 (double)old_y < old_y_max; ++old_y)
            {
               double y_factor = 1.0;
               if (old_y < old_y_min)
                  y_factor = 1.0 - (old_y_min - old_y);
               else if (old_y == (unsigned int)old_y_max)
                  y_factor = old_y_max - old_y;

               unsigned int orowindex = old_row_size * (old_y + border);

               for (unsigned int old_x = (unsigned int)old_x_min;
                    (double)old_x < old_x_max; ++old_x)
               {
                  double x_factor = 1.0;
                  if (old_x < old_x_min)
                     x_factor = 1.0 - (old_x_min - old_x);
                  else if (old_x == (unsigned int)old_x_max)
                     x_factor = old_x_max - old_x;

                  unsigned int ocolindex = (old_x + border) * num_components;
                  double factor = x_factor * y_factor;
                  T* optr = old_image + orowindex + ocolindex;
                  for (unsigned int i=0; i<num_components; ++i)
                  {
                     val[i] += *optr * factor;
                     ++optr;
                  }
               }
            }

            double tmp = 1.0 /
               ((old_x_max - old_x_min) * (old_y_max - old_y_min));
            for (unsigned int i=0; i<num_components; ++i)
            {
               *nptr = (T)(val[i] * tmp);
               ++nptr;
            }
         }
      }
   }
}

//----------------------------------------------------------------------------
bool sglTexture2D::resizeImage(sglTexture::ExternalFormatEnum format,
                               sglTexture::TypeEnum type,
                               unsigned int old_s, unsigned int old_t,
                               bool border,
                               unsigned int new_s, unsigned int new_t,
                               void *old_image, void *&new_image)
{
   unsigned int num_components = getNumComponents(format);

   switch (type)
   {
      case sglTexture::eBYTE:
         sglTexture2DResizeImage((signed char*)old_image,
                                 (signed char**)&new_image,
                                 num_components, old_s, old_t,
                                 (unsigned int)(border ? 1 : 0),
                                 new_s, new_t);
         break;
      case sglTexture::eUNSIGNED_BYTE:
         sglTexture2DResizeImage((unsigned char*)old_image,
                                 (unsigned char**)&new_image,
                                 num_components, old_s, old_t,
                                 (unsigned int)(border ? 1 : 0),
                                 new_s, new_t);
         break;
      case sglTexture::eSHORT:
         sglTexture2DResizeImage((signed short*)old_image,
                                 (signed short**)&new_image,
                                 num_components, old_s, old_t,
                                 (unsigned int)(border ? 1 : 0),
                                 new_s, new_t);
         break;
      case sglTexture::eUNSIGNED_SHORT:
         sglTexture2DResizeImage((unsigned short*)old_image,
                                 (unsigned short**)&new_image,
                                 num_components, old_s, old_t,
                                 (unsigned int)(border ? 1 : 0),
                                 new_s, new_t);
         break;
      case sglTexture::eINT:
         sglTexture2DResizeImage((signed int*)old_image,
                                 (signed int**)&new_image,
                                 num_components, old_s, old_t,
                                 (unsigned int)(border ? 1 : 0),
                                 new_s, new_t);
         break;
      case sglTexture::eUNSIGNED_INT:
         sglTexture2DResizeImage((unsigned int*)old_image,
                                 (unsigned int**)&new_image,
                                 num_components, old_s, old_t,
                                 (unsigned int)(border ? 1 : 0),
                                 new_s, new_t);
         break;
      case sglTexture::eFLOAT:
         sglTexture2DResizeImage((float*)old_image,
                                 (float**)&new_image,
                                 num_components, old_s, old_t,
                                 (unsigned int)(border ? 1 : 0),
                                 new_s, new_t);
         break;
      case sglTexture::eBITMAP:
         sglPrintWarning()
            << "sglTexture2D::resizeImage()"
            << " currently doesn't support bitmap images"
            << endl;
         return false;
      default:
         sglPrintWarning()
            << "sglTexture2D::resizeImage() unkown data type = "
            << (int)type << endl;
         return false;
   }
   return true;
}

//----------------------------------------------------------------------------
void sglTexture2D::printInfo(ostream &ostrm, const char *indent_string) const
{
    sglTexture::printInfo(ostrm, indent_string);

    ostrm << indent_string << "width/height = " << m_width << "/"
          << m_height << endl;
    ostrm << indent_string << "wrap mode (s/t) = " << m_repeat_flag[0] << "/"
          << m_repeat_flag[1] << endl;
    ostrm << indent_string << "subload list size = " << m_subload_list.size()
          << endl;
}
