/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglTextureCubeMap.cpp
 *  Created: 30 November 2001
 *  Summary: container for cube map texture image data and corresponding
 *           functions.
 *****************************************************************************/

#include <sgl.h>
#include <string.h>
#include <sglObject.hpp>
#include <sglTextureCubeMap.hpp>
#include <sglTexture2D.hpp>
#include <sglOStream.hpp>
#include <sglLock.hpp>


const GLenum cube_face[6] = {GL_TEXTURE_CUBE_MAP_POSITIVE_X_ARB,
                            GL_TEXTURE_CUBE_MAP_NEGATIVE_X_ARB,
                            GL_TEXTURE_CUBE_MAP_POSITIVE_Y_ARB,
                            GL_TEXTURE_CUBE_MAP_NEGATIVE_Y_ARB,
                            GL_TEXTURE_CUBE_MAP_POSITIVE_Z_ARB,
                            GL_TEXTURE_CUBE_MAP_NEGATIVE_Z_ARB};

//----------------------------------------------------------------------------
sglTextureCubeMap::sglTextureCubeMap() : sglTexture(GL_TEXTURE_CUBE_MAP_ARB),
   m_size(0),
   m_max_anisotropy(1.0f),
   m_power(0)
{
   m_images[0] = m_images[1] = m_images[2] = m_images[3] = m_images[4]
      = m_images[5] = NULL;
   m_repeat_flag[0] = m_repeat_flag[1] = GL_REPEAT;
}

//----------------------------------------------------------------------------
sglTextureCubeMap::~sglTextureCubeMap()
{
   // Remove this texture from the dirty list _before_ allowing the
   // object to become partially destructed (ie this can't be done in
   // the base class, or bind could get called during destruction).
   s_lock.acquire();
   s_dirty_textures.erase(this);
   s_lock.release();

   if (m_own_image)
   {
      for (unsigned int ix = 0; ix < 6; ++ix)
      {
         free(m_images[ix]);
      }
   }
}

//----------------------------------------------------------------------------
bool sglTextureCubeMap::setImage(void *px_image,
                                 void *nx_image,
                                 void *py_image,
                                 void *ny_image,
                                 void *pz_image,
                                 void *nz_image,
                                 ExternalFormatEnum format,
                                 TypeEnum type,
                                 unsigned int size,
                                 bool border)
{
   void *image[6] = {px_image, nx_image,
                     py_image, ny_image,
                     pz_image, nz_image};

   unsigned int borderless = (border ? size - 2 : size);

   unsigned int power = 0;
   for (unsigned int tmp = borderless; tmp > 1; ++power)
   {
      tmp >>= 1;
   }

   if (borderless == 0 || format == 0 ||
       px_image == NULL || nx_image == NULL ||
       py_image == NULL || ny_image == NULL ||
       pz_image == NULL || nz_image == NULL)
   {
      m_valid_flag = false;
      m_index = 0;

      dirty(eDIRTY_IMAGE);
      return false;
   }

   if ((GLint)borderless > sgl::s_max_cube_map_texture_size ||
       (1u<<power) != borderless)
   {
      unsigned int old_size = size;

      if ((GLint)borderless > sgl::s_max_cube_map_texture_size)
      {
         sglPrintWarning()
            << "sglTextureCubeMap::setImage() size (" << size
            << ") > GL_MAX_CUBE_MAP_TEXTURE_SIZE_ARB; auto resizing" << endl;

         size = sgl::s_max_texture_size;
         borderless = (border ? size - 2 : size);
         power = 0;
         unsigned int tmp;
         for (tmp = borderless; tmp > 1; ++power) {tmp >>= 1;}
      }

      if ((1u<<power) != borderless)
      {
         sglPrintWarning() << "sglTextureCubeMap::setImage() size (" << size
                           << ") not a power of 2; auto resizing"
                           << endl;

         size = (1u<<power) + (border ? 2 : 0);
      }

      for (unsigned int ix = 0; ix < 6; ++ix)
      {
         if (!m_own_image)
            m_images[ix] = NULL;

         sglTexture2D::resizeImage(format, type, old_size, old_size, border,
                                   size, size, image[ix], m_images[ix]);
      }
      m_own_image = true;
   }
   else
   {
      for (unsigned int ix=0; ix<6; ++ix)
      {
         if (m_own_image && image[ix] != m_images[ix])
         {
            free(m_images[ix]);

         }
         // note: a pointer to any preexisting image is lost
         m_images[ix] = image[ix];
      }
      m_own_image = false;
   }

   m_size            = size;
   m_external_format = format;
   m_type            = type;
   m_border          = (border ? 1 : 0);
   m_power           = power;

   // find some reasonable default
   if (!m_internal_format_flag)
   {
      switch (type)
      {
         case eSHORT:
         case eUNSIGNED_SHORT:
         case eINT:
         case eUNSIGNED_INT:
         case eFLOAT:
            switch (format)
            {
               case eRGBA:
                  m_internal_format = eINT_RGBA16;
                  break;
               case eALPHA:
                  m_internal_format = eINT_ALPHA16;
                  break;
               case eLUMINANCE:
                  m_internal_format = eINT_LUMINANCE16;
                  break;
               case eLUMINANCE_ALPHA:
                  m_internal_format = eINT_LUMINANCE16_ALPHA16;
                  break;
               default:
                  m_internal_format = eINT_RGB16;
            }
            break;
         default:
            switch (format)
            {
               case eRGBA:
                  m_internal_format = eINT_RGBA;
                  break;
               case eALPHA:
                  m_internal_format = eINT_ALPHA;
                  break;
               case eLUMINANCE:
                  m_internal_format = eINT_LUMINANCE;
                  break;
               case eLUMINANCE_ALPHA:
                  m_internal_format = eINT_LUMINANCE_ALPHA;
                  break;
               default:
                  m_internal_format = eINT_RGB;
            }
      }
   }

   if (m_auto_mipmap_flag == eMIP_SW)
      buildAutoMipmaps(eDIRTY_IMAGE);
   else
      m_mipmaps.clear();

   m_valid_flag = true;
   m_index = UINT_MAX;

      // not sure this is the correct set
   dirty(eDIRTY_IMAGE | eDIRTY_INTERNAL_FORMAT |
         eDIRTY_WRAP_S_MODE | eDIRTY_WRAP_T_MODE |
         eDIRTY_MIN_FILTER | eDIRTY_MAG_FILTER |
         eDIRTY_MIPMAP | eDIRTY_ANISOTROPY);

   return true;
}

//----------------------------------------------------------------------------
void sglTextureCubeMap::getImage(void *&px_image,
                                 void *&nx_image,
                                 void *&py_image,
                                 void *&ny_image,
                                 void *&pz_image,
                                 void *&nz_image,
                                 ExternalFormatEnum &format,
                                 TypeEnum &type,
                                 unsigned int &size,
                                 bool &border) const
{
   px_image = py_image = pz_image =
      nx_image = ny_image = nz_image = NULL;

   if (m_valid_flag)
   {
      px_image = m_images[0];
      nx_image = m_images[1];
      py_image = m_images[2];
      ny_image = m_images[3];
      pz_image = m_images[4];
      nz_image = m_images[5];

      format = m_external_format;
      type = m_type;
      size = m_size;
      border = (m_border ? true : false);
   }
}

//----------------------------------------------------------------------------
//  bool sglTextureCubeMap::setMipmapLevels(const vector<void *> &images)
//  {
//     if (m_auto_mipmap_flag)
//        return false;

//     if (images.empty() && !m_mipmaps.empty())
//     {
//        m_mipmaps.clear();
//        dirty(eDIRTY_RESET);
//        return true;
//     }

//     if (m_pow_s > m_pow_t)
//     {
//        if (images.size() != (unsigned int)m_pow_s)
//           return false;
//     }
//     else
//     {
//        if (images.size() != (unsigned int)m_pow_t)
//           return false;
//     }

//     if (m_mipmaps.empty())
//     {
//        for (vector<void*>::const_iterator it = images.begin();
//             it != images.end(); ++it)
//        {
//           m_mipmaps.push_back(MipmapStruct(*it, eDIRTY_IMAGE));
//        }
//     }
//     else
//     {
//        vector<MipmapStruct>::iterator elt = m_mipmaps.begin();
//        for (vector<void*>::const_iterator it = images.begin();
//             it != images.end(); ++it, ++elt)
//        {
//           elt->m_image = *it;
//           elt->m_dirty |= eDIRTY_REPLACE;
//        }
//     }
//     dirty(eDIRTY_MIPMAP);

//     return true;
//  }

//----------------------------------------------------------------------------
unsigned int sglTextureCubeMap::getIdealInternalSize() const
{
   unsigned int size = getNumTexels();
   size *= getBitsPerTexel(m_internal_format);
   return 6*size/8;
}

//----------------------------------------------------------------------------
unsigned int sglTextureCubeMap::getLikelyInternalSize() const
{
   if (m_internal_size_dirty)
   {
      s_lock.acquire();
      if (m_internal_size_dirty)
      {
         glBindTexture(m_target, m_texture_object);

         m_internal_size = 0;
         for (unsigned int ix = 0; ix < 6; ++ix)
         {
            GLint red, green, blue, alpha, lumin, intensity;
            glGetTexLevelParameteriv(
               cube_face[ix], 0, GL_TEXTURE_RED_SIZE, &red);
            glGetTexLevelParameteriv(
               cube_face[ix], 0, GL_TEXTURE_GREEN_SIZE, &green);
            glGetTexLevelParameteriv(
               cube_face[ix], 0, GL_TEXTURE_BLUE_SIZE, &blue);
            glGetTexLevelParameteriv(
               cube_face[ix], 0, GL_TEXTURE_ALPHA_SIZE, &alpha);
            glGetTexLevelParameteriv(
               cube_face[ix], 0, GL_TEXTURE_LUMINANCE_SIZE, &lumin);
            glGetTexLevelParameteriv(
               cube_face[ix], 0, GL_TEXTURE_INTENSITY_SIZE, &intensity);

//          sglPrint() << "red size = " << red << endl;
//          sglPrint() << "green size = " << green << endl;
//          sglPrint() << "blue size = " << blue << endl;
//          sglPrint() << "alpha size = " << alpha << endl;
//          sglPrint() << "lumin size = " << lumin << endl;
//          sglPrint() << "intensity size = " << intensity << endl;

            unsigned int size = getNumTexels();
            size *= (red + green + blue + alpha + lumin + intensity);
            size /= 8;

            m_internal_size += size;
         }
         m_internal_size_dirty = false;
      }
      s_lock.release();
   }
   return m_internal_size;
}

//----------------------------------------------------------------------------
unsigned int sglTextureCubeMap::getNumTexels() const
{
   unsigned int num_texels = 0;

   if (m_bound_mipmaps || !m_mipmaps.empty())
   {
      unsigned int size = m_size;

      if (m_border)
      {
         unsigned int num_levels = m_power + 1;
         size -= 2;

         // add the number of border texels
         num_texels = 4*(num_levels + 2*size - 1);
      }

      // add the number of interior texels
      num_texels += (size * size * 4) / 3;
   }
   else
      num_texels = m_size*m_size;

   return num_texels;
}

//----------------------------------------------------------------------------
void sglTextureCubeMap::setWrapMode(WrapEnum s_repeat_flag,
                                    WrapEnum t_repeat_flag)
{
   GLint wrap_param = GL_REPEAT;
   switch (s_repeat_flag)
   {
      case eREPEAT:
         wrap_param = GL_REPEAT;
         break;
      case eCLAMP:
         wrap_param = GL_CLAMP;
         break;
      case eCLAMP_TO_EDGE:
         wrap_param = sgl::s_clamp_to_edge;
         break;
   }

   if (m_repeat_flag[0] != wrap_param)
   {
      m_repeat_flag[0] = wrap_param;
      dirty(eDIRTY_WRAP_S_MODE);
   }

   switch (t_repeat_flag)
   {
      case eREPEAT:
         wrap_param = GL_REPEAT;
         break;
      case eCLAMP:
         wrap_param = GL_CLAMP;
         break;
      case eCLAMP_TO_EDGE:
         wrap_param = sgl::s_clamp_to_edge;
         break;
   }

   if (m_repeat_flag[1] != wrap_param)
   {
      m_repeat_flag[1] = wrap_param;
      dirty(eDIRTY_WRAP_T_MODE);
   }
}

//----------------------------------------------------------------------------
void sglTextureCubeMap::getWrapMode(WrapEnum &s_repeat_flag,
                                    WrapEnum &t_repeat_flag) const
{
   if (m_repeat_flag[0] == GL_REPEAT)
      s_repeat_flag = eREPEAT;
   else if (m_repeat_flag[0] == GL_CLAMP)
      s_repeat_flag = eCLAMP;
   else if (m_repeat_flag[0] == sgl::s_clamp_to_edge)
      s_repeat_flag = eCLAMP_TO_EDGE;
   else
      sglPrintWarning()
         << "sglTextureCubeMap::getWrapMode: m_repeat_flag[0] is invalid!"
         << endl;

   if (m_repeat_flag[1] == GL_REPEAT)
      t_repeat_flag = eREPEAT;
   else if (m_repeat_flag[1] == GL_CLAMP)
      t_repeat_flag = eCLAMP;
   else if (m_repeat_flag[1] == sgl::s_clamp_to_edge)
      t_repeat_flag = eCLAMP_TO_EDGE;
   else
      sglPrintWarning()
         << "sglTextureCubeMap::getWrapMode: m_repeat_flag[1] is invalid!"
         << endl;
}

//----------------------------------------------------------------------------
void sglTextureCubeMap::setMaxAnisotropy(float max)
{
   if (max < 1.0f) max = 1.0f;

   if (max != m_max_anisotropy)
   {
      m_max_anisotropy = max;
      dirty(eDIRTY_ANISOTROPY);
   }
}

//----------------------------------------------------------------------------
bool sglTextureCubeMap::checkProxy()
{
   // test for size restrictions
   glTexImage2D(GL_PROXY_TEXTURE_CUBE_MAP_ARB, 0,
                (m_compression_flag)?m_compression_enum:m_internal_format,
                m_size, m_size, m_border,
                (GLenum)m_external_format, (GLenum)m_type, NULL);
   GLint width;
   glGetTexLevelParameteriv(GL_PROXY_TEXTURE_CUBE_MAP_ARB, 0,
                            GL_TEXTURE_WIDTH, &width);
   if (width == 0)
   {
      return false;
   }
   return true;
}

//----------------------------------------------------------------------------
//    Summary: this should only be called by the drawing thread with the
//             proper context current.
void sglTextureCubeMap::bind()
{
   if (m_valid_flag && m_dirty_flag)
   {
      if ((m_dirty_flag & eDIRTY_RESET) && m_texture_object)
      {
         // delete any preexisting texture
         glDeleteTextures(1, &m_texture_object);
         m_texture_object = 0;
         m_dirty_flag = (unsigned int)~eDIRTY_RESET;
      }

      // create the texture object
      if (!m_texture_object)
      {
         m_texture_object = sgl::getTextureObj();
         m_bound_mipmaps = false;

         if (m_texture_object == 0)
            return;
      }

      glBindTexture(GL_TEXTURE_CUBE_MAP_ARB, m_texture_object);

      if (m_dirty_flag & eDIRTY_WRAP_S_MODE)
      {
         glTexParameterf(GL_TEXTURE_CUBE_MAP_ARB,
                         GL_TEXTURE_WRAP_S, m_repeat_flag[0]);
      }

      if (m_dirty_flag & eDIRTY_WRAP_T_MODE)
      {
         glTexParameterf(GL_TEXTURE_CUBE_MAP_ARB,
                         GL_TEXTURE_WRAP_T, m_repeat_flag[1]);
      }

      if (m_dirty_flag & eDIRTY_BORDER_COLOR)
      {
         glTexParameterfv(GL_TEXTURE_CUBE_MAP_ARB, GL_TEXTURE_BORDER_COLOR,
                          m_border_color.getPtr());
      }

      if (m_dirty_flag & eDIRTY_MAG_FILTER)
      {
         glTexParameteri(GL_TEXTURE_CUBE_MAP_ARB,
                         GL_TEXTURE_MAG_FILTER, m_mag_filter);
      }

      if (m_dirty_flag & eDIRTY_MIN_FILTER)
      {
         glTexParameteri(GL_TEXTURE_CUBE_MAP_ARB,
                         GL_TEXTURE_MIN_FILTER, m_min_filter);
      }

      if ((m_dirty_flag & eDIRTY_ANISOTROPY) &&
          sgl::s_max_texture_max_anisotropy != 1.0f)
      {
         glTexParameterf(GL_TEXTURE_CUBE_MAP_ARB,
                         (GLenum)0x84FE, // TEXTURE_MAX_ANISOTROPY_EXT
                         m_max_anisotropy);
      }

      if ((m_dirty_flag & eDIRTY_LOD) && m_texture_lod_active )
      {
         glTexParameterf(GL_TEXTURE_CUBE_MAP_ARB,
                         GL_TEXTURE_MIN_LOD_SGIS,
                         m_texture_lod_min);

         glTexParameterf(GL_TEXTURE_CUBE_MAP_ARB,
                         GL_TEXTURE_MAX_LOD_SGIS,
                         m_texture_lod_max);

         glTexParameteri(GL_TEXTURE_CUBE_MAP_ARB,
                         GL_TEXTURE_BASE_LEVEL_SGIS,
                         m_texture_lod_base_level);

         glTexParameteri(GL_TEXTURE_CUBE_MAP_ARB,
                         GL_TEXTURE_MAX_LEVEL_SGIS,
                         m_texture_lod_max_level);
      }

      // Should we allow the user to set these?
      glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
      glPixelStorei(GL_UNPACK_ROW_LENGTH, 0);
      glPixelStorei(GL_UNPACK_SKIP_ROWS, 0);
      glPixelStorei(GL_UNPACK_SKIP_PIXELS, 0);

      if (m_pre_bind_callback)
         (*m_pre_bind_callback)(this);

      if ((m_dirty_flag & eDIRTY_MIPMAP) && sgl::s_auto_mipmap_capable)
      {
         if (m_auto_mipmap_flag == eMIP_HW)
         {
            glTexParameteri(
               GL_TEXTURE_CUBE_MAP_ARB,
               sgl::s_auto_mipmap_enable_flag, GL_TRUE);

            // set to show there are bound mip maps
            m_bound_mipmaps = true;
         }
         else
         {
            glTexParameteri(
               GL_TEXTURE_CUBE_MAP_ARB,
               sgl::s_auto_mipmap_enable_flag, GL_FALSE);
         }
      }

      if ((m_dirty_flag & eDIRTY_IMAGE) ||
          (m_dirty_flag & eDIRTY_INTERNAL_FORMAT))
      {
         //cerr << "glTexImage2D(0)" << endl;
         m_internal_size_dirty = true;
         for (unsigned int ix = 0; ix < 6; ++ix)
         {
            glTexImage2D(
               cube_face[ix], 0,
               (m_compression_flag)?m_compression_enum:m_internal_format,
               m_size, m_size, m_border,
               (GLenum)m_external_format, (GLenum)m_type, m_images[ix]);
         }
      }
      else if (m_dirty_flag & eDIRTY_REPLACE)
      {
         //cerr << "glTexSubImage2D(0)" << endl;
         for (unsigned int ix = 0; ix < 6; ++ix)
         {
            glTexSubImage2D(GL_TEXTURE_CUBE_MAP_ARB,
                            0, 0, 0, m_size, m_size,
                            (GLenum)m_external_format, (GLenum)m_type,
                            m_images[ix]);
         }
      }

      if ((m_dirty_flag & eDIRTY_MIPMAP) &&
          m_auto_mipmap_flag != eMIP_HW &&
          !m_mipmaps.empty())
      {
         for (unsigned int ix = 0; ix < 6; ++ix)
         {
            GLint level = 1;
            for (vector<MipmapStruct>::iterator it = m_mipmaps.begin();
                 it != m_mipmaps.end(); ++it, ++level)
            {
               GLint size = (level >= m_power ? 1 : (1<<(m_power-level)))
                  + m_border*2;

               if ((m_dirty_flag & eDIRTY_INTERNAL_FORMAT) ||
                   (it->m_dirty & eDIRTY_IMAGE))
               {
                  //cerr << "glTexImage2D(" << level << ")" << endl;
                  m_internal_size_dirty = true;

                  glTexImage2D(
                     cube_face[ix], level,
                     (m_compression_flag)?m_compression_enum:m_internal_format,
                     size, size, m_border,
                     (GLenum)m_external_format, (GLenum)m_type,
                     it->m_image);
               }
               else if (it->m_dirty & eDIRTY_REPLACE)
               {
                  //cerr << "glTexSubImage2D(" << level << ")" << endl;
                  glTexSubImage2D(cube_face[ix],
                                  level, 0, 0, size, size,
                                  (GLenum)m_external_format, (GLenum)m_type,
                                  it->m_image);
               }

               it->m_dirty = 0;
            }
         }
         // set to show there are bound mip maps
         m_bound_mipmaps = true;
      }

      m_dirty_flag = 0;

      if (m_post_bind_callback)
         (*m_post_bind_callback)(this);
   }
}
