/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglTextureCubeMap.hpp
 *  Created: 30 November 2001
 *  Summary: container for 2D image data used for cube map texturing
 *****************************************************************************/

#ifndef __SGL_TEXTURE_CUBE_MAP_HPP
#define __SGL_TEXTURE_CUBE_MAP_HPP

#include <sgl.h>
#include <vector>
#include <sglObject.hpp>
#include <sglVector.hpp>
#include <sglTexture.hpp>

//============================================================================

/**

This statelet works with the GL_ARB_texture_cube_map, GL_EXT_texture_cube_map,
or GL_NV_texture_cube_map extensions (they are all equivalent).  Please see the
official OpenGL extension documentation available at opengl.org for most of the
details.  This statelet is a container for the 6 images used for cube map
texturing.  As near as I can tell the images have the following orientations:

<pre>
positive X (px):  z <----+
                         |
                   right |
                         V
                         y

negative X (nx):  +----> z
                  |
                  | left
                  V
                  y

positive Z (pz):  +----> x
                  |
                  | front
                  V
                  y

negative Z (pz):  x <----+
                         |
                    back |
                         V
                         y

positive Y (py):  z
                  ^
                  | bottom
                  |
                  +----> x

negative Y (ny):  +----> x
                  |
                  | top
                  V
                  z
</pre>

Note that all six images must be the same size (square) and have the same
type and external format.

<b>
Example:
</b>
<pre>
   // all six images must be the same size (and square) format and type
   sglTextureCubeMap *cube_map = new sglTextureCubeMap;
   if (!cube_map->setImage(image[0], image[1], image[2], image[3],
                           image[4], image[5],
                           format, type, size, false))
   {
      cerr << "Warning: sglTextureCubeMap::setImage() failed" << endl;
   }

   sglTexGen *tgen = new sglTexGen;
   tgen->setMode(sglTexGen::eS, sglTexGen::eREFLECTION_MAP);
   tgen->setMode(sglTexGen::eT, sglTexGen::eREFLECTION_MAP);
   tgen->setMode(sglTexGen::eR, sglTexGen::eREFLECTION_MAP);

   // used to account for view position.
   tmat = new sglTextureMatrix(NULL);

   vector<sglStatelet*> state;
   state.push_back(cube_map);
   state.push_back(tgen);
   state.push_back(tmat);
</pre>

 */

class SGL_DLL_API sglTextureCubeMap : public sglTexture
{
public:
   /// Default constructor.  Creates an empty, invalid texture.
   sglTextureCubeMap();

   /// Virtual destructor.
   virtual ~sglTextureCubeMap();

   /** Set the six images in the cube map.  All must be same square size and in
       the same format.
       Note: this will not free the memory from the previous image.
       Note: this will clear the mipmap levels if not automatic.
       @param px_image The positive X image data.
       @param nx_image The negative X image data.
       @param py_image The positive Y image data.
       @param ny_image The negative Y image data.
       @param pz_image The positive Z image data.
       @param nz_image The negative Z image data.
       @param format Format of each texel in the images
       @param type Data type of texel data
       @param size Width an height of the image data (square only).  Must be
         equal to the equation, 2^m +2b, where m is a
         non-negative integer (which can be different for s and t) and
         b is zero or one (depending on border).  ??? If num_s or num_t
         (minus the border) are not powers of two, or if the image is
         too large, the image will be automatically resized ???
       @param border True if the image has a one pixel border
       @return false if the parameters are out of range.
   */
   bool setImage(void *px_image,
                 void *nx_image,
                 void *py_image,
                 void *ny_image,
                 void *pz_image,
                 void *nz_image,
                 ExternalFormatEnum format,
                 TypeEnum type,
                 unsigned int size,
                 bool border = false);

   /** Query the current cube map data.  If the valid flag is false the
       image pointers will be set to NULL.
       @param px_image The positive X image data.
       @param nx_image The negative X image data.
       @param py_image The positive Y image data.
       @param ny_image The negative Y image data.
       @param pz_image The positive Z image data.
       @param nz_image The negative Z image data.
       @param format Format of each texel in the images
       @param type Data type of texel data
       @param size Width an height of the image data (square only).
       @param border True if the image has a one pixel border
   */
   void getImage(void *&px_image,
                 void *&nx_image,
                 void *&py_image,
                 void *&ny_image,
                 void *&pz_image,
                 void *&nz_image,
                 ExternalFormatEnum &format,
                 TypeEnum &type,
                 unsigned int &size,
                 bool &border) const;

   /** Set the mipmap images.  This feature is currently disabled for cube
       maps.
   */
   bool setMipmapLevels(const vector<void *> &) { return false; }

   /** Query the current size of the level zero images.
       @return the size (width and height are the same value.
   */
   inline unsigned int getSize() const { return m_size; }

   /** Returns the number of bytes of texture memory this texture will
       use if the requested internal format was honored.
   */
   virtual unsigned int getIdealInternalSize() const;
   /** Returns the number of bytes of texture memory this texture is
       most likely using.
       Must be called by a thread with a current rendering context.
   */
   virtual unsigned int getLikelyInternalSize() const;

   /** Set whether the texture repeats or is stretched by tex coords > 1.
       @param s_repeat_flag One of eREPEAT, eCLAMP, or eCLAMP_TO_EDGE (when
         supported) for the s dimension
       @param t_repeat_flag Same choices as above.
   */
   void setWrapMode(WrapEnum s_repeat_flag, WrapEnum t_repeat_flag);

   /** Query the current texture wrap mode.
       @param s_repeat_flag On return, contains the WrapEnum token for the s
          dimension.
       @param t_repeat_flag On return, contains the WrapEnum token for the t
          dimension.
   */
   void getWrapMode(WrapEnum &s_repeat_flag, WrapEnum &t_repeat_flag) const;

   /** Get/set the maximum degree of anisotropy to account for in
       texture filtering (if supported).  A value of 1.0f disables
       anisotropic filtering.
       default: 1.0f
   */
   void setMaxAnisotropy(float max);
   ///
   float getMaxAnisotropy() const { return m_max_anisotropy; }

   /** Get the maximum max-anisotropy supported on this system.
       Returns 1.0f if anisotropic filtering is not supported.
   */
   static float getMaxSupportedMaxAnisotropy()
      { return sgl::s_max_texture_max_anisotropy; }

   /** Use the GL_PROXY_TEXTURE_CUBE_MAP_ARB mechanism to check if the texture
       is capable of being loaded into texture memory.
       Returns false if the texture is too large.
       Must be called by a thread with a current rendering context.
   */
   virtual bool checkProxy();

   /** Copy texture (and changes) into texture memory.
       If (getAutoBind() == false), this method must be called before
       before the texture can be applied.
       Must be called by a thread with a current rendering context.
   */
   virtual void bind();

   virtual bool replaceImage(void *, GLint) { return false; }
   virtual bool buildMipmaps(vector<void*> &) { return false; }

private:
   sglTextureCubeMap(const sglTextureCubeMap&);              // NOT IMPLEMENTED
   sglTextureCubeMap& operator=(const sglTextureCubeMap &);  // NOT IMPLEMENTED

   // helper function for get*InternalSize
   unsigned int getNumTexels() const;

private:
   void *m_images[6];

   unsigned int m_size;       // must be square
   GLint m_repeat_flag[2];
   float m_max_anisotropy;
   GLint m_power;             // m_pow_s, m_pow_t
};

#endif
