/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglTexturedFont.hpp
 *  Created: 18 June 2001
 *  Summary:
 *****************************************************************************/

#if defined(WIN32)
# include <windows.h>
# include <wingdi.h>
#endif

#include <sglTextureFont.hpp>
#include <sglTexture2D.hpp>
#include <sglBitmapFont.hpp>
#include <sglOStream.hpp>

#if defined( MAC_OSX )
#elif !defined(WIN32)
# include <GL/glx.h>
#endif

//----------------------------------------------------------------------------
sglTextureFont::sglTextureFont(const sglBitmapFont &font,
                               bool use_invalid_char,
                               char invalid_char)
   : m_texture(new sglTexture2D()),
     m_image(NULL),
     m_height(font.getHeight()),
     m_max_width(0),
     m_invalid_char(invalid_char),
     m_invalid_char_flag(use_invalid_char)
{
   m_texture->ref();

   for (unsigned char c=0; c<128; ++c)
   {
      unsigned int width = font.getCharWidth(c);
      if (width > 0)
      {
         m_valid_chars[c] = true;
         m_char_widths[c] = (float)width/(float)m_height;

         if (width > m_max_width)
            m_max_width = width;
      }
      else
      {
         m_valid_chars[c] = false;
      }
   }
   if (m_max_width == 0)
      return;

   unsigned int tex_width, tex_height;
   computeTextureSize(tex_width, tex_height);
   generateTexture(font, tex_width, tex_height);
}


//----------------------------------------------------------------------------
sglTextureFont::sglTextureFont(const sglBitmapFont &font,
                               const vector<char> &valid_chars,
                               bool use_invalid_char,
                               char invalid_char)
   : m_texture(new sglTexture2D()),
     m_image(NULL),
     m_height(font.getHeight()),
     m_max_width(0),
     m_invalid_char(invalid_char),
     m_invalid_char_flag(use_invalid_char)
{
   m_texture->ref();

   for (vector<char>::const_iterator it = valid_chars.begin();
        it != valid_chars.end(); ++it)
   {
      char c = *it;

      unsigned int width = font.getCharWidth(c);
      if (width > 0)
      {
         m_valid_chars[c] = true;
         m_char_widths[c] = (float)width/(float)m_height;

         if (width > m_max_width)
            m_max_width = width;
      }
   }
   if (m_max_width == 0)
      return;

   unsigned int tex_width, tex_height;
   computeTextureSize(tex_width, tex_height);
   generateTexture(font, tex_width, tex_height);
}


//----------------------------------------------------------------------------
sglTextureFont::~sglTextureFont()
{
   m_texture->deref();
   if (m_texture->getRefCount() == 0)
      delete m_texture;
   if (m_image)
      free(m_image);
}


//----------------------------------------------------------------------------
void sglTextureFont::computeTextureSize(
   unsigned int &tex_width,
   unsigned int &tex_height)
{
//   unsigned int max_size = 128;
   unsigned int max_size = (unsigned int)sgl::s_max_texture_size;
   tex_height = 0;
   tex_width = 0;

   tex_height = m_height;
   for (map<char, float>::iterator it = m_char_widths.begin();
        it != m_char_widths.end(); ++it)
   {
      unsigned int width =
         (unsigned int)((it->second * (float)m_height) + 0.5f);
      if (tex_width + width > max_size)
      {
         // Add a one pixel buffer to prevent linear filter bleed.
         tex_height += m_height + 1;
         tex_width = 0;
      }
      else
         // Add a one pixel buffer to prevent linear filter bleed.
         tex_width += width + 1;
   }
   if (tex_height == m_height)
      tex_width = sgl::nextPower2(tex_width);
   else
      tex_width = max_size;
   tex_height = sgl::nextPower2(tex_height);

   sglPrintDebug() << "sglTextureFont tex_width = " << tex_width << endl;
   sglPrintDebug() << "sglTextureFont tex_height = " << tex_height << endl;
}

//----------------------------------------------------------------------------
void sglTextureFont::generateTexture(
   const sglBitmapFont &font,
   unsigned int tex_width,
   unsigned int tex_height)
{
#if defined(WIN32)
   HDC old_hdc = wglGetCurrentDC();

   HBITMAP bitmap = CreateCompatibleBitmap(old_hdc, tex_width, tex_height);
   if (bitmap == NULL)
   {
      sglPrintWarning()
         << "sglTextureFont: CreateCompatibleBitmap() failed: "
         << sglOStream::formatWIN32Error(GetLastError()) << endl;
   }
   BITMAP bmInfo;
   if (GetObject(bitmap, sizeof(BITMAP), &bmInfo) == 0)
   {
      sglPrintWarning()
         << "sglTextureFont: GetObject() failed: "
         << sglOStream::formatWIN32Error(GetLastError()) << endl;
   }
   sglPrintDebug() << "sglTextureFont bits per pixel = "
                   << bmInfo.bmBitsPixel << endl;

   HDC hdc_mem = CreateCompatibleDC(old_hdc);
   if (hdc_mem == NULL)
   {
      sglPrintWarning()
         << "sglTextureFont: CreateCompatibleDC() failed: "
         << sglOStream::formatWIN32Error(GetLastError()) << endl;
   }
   SelectObject(hdc_mem, bitmap);
   HGLRC old_context = wglGetCurrentContext();

   PIXELFORMATDESCRIPTOR pfd;
   memset(&pfd, 0, sizeof(PIXELFORMATDESCRIPTOR));
   pfd.nSize = sizeof(PIXELFORMATDESCRIPTOR);
   pfd.nVersion = 1;
   pfd.dwFlags =
       PFD_DRAW_TO_BITMAP | PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER_DONTCARE;
   pfd.iPixelType = PFD_TYPE_RGBA;
   pfd.cColorBits = bmInfo.bmBitsPixel;
   pfd.cDepthBits = 32;
   pfd.iLayerType = PFD_MAIN_PLANE;

   int pixel_format = ChoosePixelFormat(hdc_mem, &pfd);
   if (pixel_format == 0)
   {
      sglPrintWarning()
         << "sglTextureFont: ChoosePixelFormat() failed: "
         << sglOStream::formatWIN32Error(GetLastError()) << endl;
   }
   if (SetPixelFormat(hdc_mem, pixel_format, &pfd) != TRUE)
   {
      int erno = GetLastError();
      sglPrintWarning()
         << "sglTextureFont: SetPixelFormat() failed: "
         << sglOStream::formatWIN32Error(GetLastError()) << endl;
   }

   HGLRC context = wglCreateContext(hdc_mem);
   if (context == NULL)
   {
      sglPrintWarning()
         << "sglTextureFont: wglCreateContext() failed: "
         << sglOStream::formatWIN32Error(GetLastError()) << endl;
   }
   if (wglMakeCurrent(hdc_mem, context) != TRUE)
   {
      sglPrintWarning() << "sglTextureFont: wglMakeCurrent() failed" << endl;
   }

#elif defined( MAC_OSX )
#else
   Display* dpy = glXGetCurrentDisplay();
   GLXContext old_context = glXGetCurrentContext();
   GLXDrawable old_drawable = glXGetCurrentDrawable();

   Window root;
   int win_x, win_y;
   unsigned int win_width, win_height, win_border_width, win_depth;
   if (XGetGeometry(dpy, old_drawable, &root, &win_x, &win_y, &win_width,
                    &win_height, &win_border_width, &win_depth) == 0)
   {
      sglPrintWarning()
         << "sglTextureFont: XGetGeometry() failed on current drawable."
         << endl;
   }
   sglPrintDebug() << "sglTextureFont depth = " << win_depth << endl;

   Pixmap pixmap = XCreatePixmap(dpy, old_drawable,
                                 tex_width, tex_height, win_depth);
   if (XGetGeometry(dpy, pixmap, &root, &win_x, &win_y, &win_width,
                    &win_height, &win_border_width, &win_depth) == 0)
   {
      sglPrintWarning() << "sglTextureFont: XGetGeometry() failed on pixmap."
                        << endl;
   }
   sglPrintDebug() << "sglTextureFont width=" << win_width
                   << ", height=" << win_height << endl;

   int attribs[] = { GLX_RGBA,
                     GLX_RED_SIZE, 1,
                     GLX_GREEN_SIZE, 1,
                     GLX_BLUE_SIZE, 1,
                     None};
   XVisualInfo *vis = glXChooseVisual(dpy, DefaultScreen(dpy), attribs);
   if (vis == NULL)
   {
      int attribs2[] = { GLX_RGBA,
                         GLX_DOUBLEBUFFER,
                         GLX_RED_SIZE, 1,
                         GLX_GREEN_SIZE, 1,
                         GLX_BLUE_SIZE, 1,
                         None};
      vis = glXChooseVisual(dpy, DefaultScreen(dpy), attribs2);
      if (vis == NULL)
      {
         sglPrintWarning()
            << "sglTextureFont: glXChooseVisual() failed" << endl;
      }
   }
   GLXPixmap glxpixmap = glXCreateGLXPixmap(dpy, vis, pixmap);
   GLXContext context = glXCreateContext(dpy, vis, NULL, False);
   if (glXMakeCurrent(dpy, glxpixmap, context) == False)
   {
      sglPrintWarning() << "sglTextureFont: glXMakeCurrent() failed" << endl;
   }
#endif

   sglBitmapFont *tmp_font = font.clone();

   glPushAttrib(GL_ALL_ATTRIB_BITS);
   glDrawBuffer(GL_FRONT);
   glReadBuffer(GL_FRONT);
   glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
   glPixelStorei(GL_PACK_ALIGNMENT, 1);
   glPixelStorei(GL_UNPACK_ROW_LENGTH, 0);
   glPixelStorei(GL_UNPACK_SKIP_ROWS, 0);
   glPixelStorei(GL_UNPACK_SKIP_PIXELS, 0);

   glDisable(GL_SCISSOR_TEST);
   glClearColor(0, 0, 0, 0);
   glDepthMask(GL_FALSE);
   glDisable(GL_DEPTH_TEST);
   glMatrixMode(GL_PROJECTION);
   glPushMatrix();
   glMatrixMode(GL_MODELVIEW);
   glPushMatrix();
   glDisable(GL_TEXTURE_1D);
   glDisable(GL_TEXTURE_2D);
   glDisable(GL_CULL_FACE);

   glClear(GL_COLOR_BUFFER_BIT);
   glColor4f(1.0f, 1.0f, 1.0f, 1.0f);

   unsigned int tex_size = tex_width*tex_height;
   m_image = malloc(tex_size*2);
   unsigned char *tmp_image = (unsigned char*)malloc(tex_size);

   unsigned int descent = tmp_font->getDescent();

   unsigned int tmp_y = 0;
   unsigned int tmp_x = 0;
   float h1 = 0;
   float h2 = (float)m_height/(float)tex_height;
   for (map<char, float>::iterator it = m_char_widths.begin();
        it != m_char_widths.end(); ++it)
   {
      unsigned int width =
         (unsigned int)((it->second * (float)m_height) + 0.5f);

      if (tmp_x + width > tex_width)
      {
         // Add a one pixel buffer to prevent linear filter bleed.
         tmp_y += m_height + 1;

         h1 = (float)tmp_y/(float)tex_height;
         h2 = (float)(tmp_y+m_height)/(float)tex_height;

         tmp_x = 0;
      }

      glViewport(tmp_x, tmp_y+descent, width, m_height);
      glMatrixMode(GL_PROJECTION);
      glLoadIdentity();
      glOrtho(0, width, 0, m_height, -1, 1);
      glMatrixMode(GL_MODELVIEW);

      glLoadIdentity();
      glRasterPos3i(0,0,0);
      tmp_font->rasterString(string(&it->first, 1));

      float w1 = (float)tmp_x/(float)tex_width;
      tmp_x += width;
      float w2 = (float)tmp_x/(float)tex_width;
      m_tex_coords[it->first][0].set(w1, h1);
      m_tex_coords[it->first][1].set(w2, h1);
      m_tex_coords[it->first][2].set(w2, h2);
      m_tex_coords[it->first][3].set(w1, h2);

      // Add a one pixel buffer to prevent linear filter bleed.
      ++tmp_x;
   }

   glFlush();
   glLoadIdentity();
   glRasterPos3i(0,0,0);
   glReadPixels(0, 0, tex_width, tex_height,
                GL_LUMINANCE, GL_UNSIGNED_BYTE, tmp_image);

   glMatrixMode(GL_PROJECTION);
   glPopMatrix();
   glMatrixMode(GL_MODELVIEW);
   glPopMatrix();

   glPopAttrib();

   unsigned char *ptr = (unsigned char*)m_image;
   for (unsigned int x=0; x<tex_size; ++x)
   {
      *ptr = 255;
//      *ptr = (tmp_image[x] ? 255 : 0);
      ++ptr;
      *ptr = tmp_image[x];
//      *ptr = 255;
      ++ptr;
   }
   free(tmp_image);

//    ptr = (unsigned char*)m_image;
//    for (unsigned int a=0; a<tex_height; a+=m_height)
//       for (unsigned int b=0; b<tex_width; b+=2)
//       {
//          ptr[(a*tex_width + b)*2] = 255;
//       }

   delete tmp_font;

#if defined(WIN32)
   wglMakeCurrent(old_hdc, old_context);
   wglDeleteContext(context);
   DeleteDC(hdc_mem);
   DeleteObject(bitmap);
#elif defined( MAC_OSX )
#else
   if (glXMakeCurrent(dpy, old_drawable, old_context) == False)
   {
      sglPrintWarning() << "sglTextureFont: glXMakeCurrent() failed" << endl;
   }
   glXDestroyContext(dpy, context);
   glXDestroyGLXPixmap(dpy, glxpixmap);
   XFreePixmap(dpy, pixmap);
#endif

   m_texture->setImage(m_image, sglTexture::eLUMINANCE_ALPHA,
                       sglTexture::eUNSIGNED_BYTE,
                       tex_width, tex_height, false);
   m_texture->setWrapMode(sglTexture::eCLAMP_TO_EDGE,
                          sglTexture::eCLAMP_TO_EDGE);
   m_texture->setMinFilter(sglTexture::eLINEAR);
   m_texture->setMagFilter(sglTexture::eLINEAR);
}


//----------------------------------------------------------------------------
unsigned int sglTextureFont::getStringInfo(const string &str,
                                           sglVec3f *coord,
                                           sglVec2f *tex_coords,
                                           float scale,
                                           const sglVec3f &offset,
                                           bool vertical) const
{
   float x = offset[0];
   float y = offset[1];
   float z = offset[2];
   unsigned int num_quads = 0;
   for (unsigned int i=0; i<str.size(); ++i)
   {
      char c = str[i];
      bool good = true;

      map<char, bool>::const_iterator it = m_valid_chars.find(c);
      if (it == m_valid_chars.end() || !it->second)
      {
         if (m_invalid_char_flag)
            c = m_invalid_char;
         else
            good = false;
      }

      if (good)
      {
         unsigned int j = num_quads*4;

         map<char, TexCorners>::const_iterator ii = m_tex_coords.find(c);
         tex_coords[j]   = ii->second[0];
         tex_coords[j+1] = ii->second[1];
         tex_coords[j+2] = ii->second[2];
         tex_coords[j+3] = ii->second[3];

         map<char, float>::const_iterator jj = m_char_widths.find(c);
         float x2 = x + jj->second*scale;
         float y2 = y + scale;
         if (vertical)
         {
            y2 = y;
            y = y - scale;
         }

         coord[j  ].set(x,  y,  z);
         coord[j+1].set(x2, y,  z);
         coord[j+2].set(x2, y2, z);
         coord[j+3].set(x,  y2, z);

         if (!vertical)
            x = x2;

         ++num_quads;
      }
   }
   return num_quads;
}


//----------------------------------------------------------------------------
unsigned int sglTextureFont::getHeight() const
{
   return m_height;
}


//----------------------------------------------------------------------------
unsigned int sglTextureFont::getCharWidth(char c) const
{
   map<char, bool>::const_iterator it = m_valid_chars.find(c);
   if (it == m_valid_chars.end() || !it->second)
   {
      if (m_invalid_char_flag)
         c = m_invalid_char;
      else
         return 0;
   }
   map<char, float>::const_iterator ii = m_char_widths.find(c);
   return (unsigned int)((ii->second * (float)m_height) + 0.5f);
}


//----------------------------------------------------------------------------
unsigned int sglTextureFont::getStringWidth(const string &str) const
{
   unsigned int width = 0;
   for (string::const_iterator it = str.begin(); it != str.end(); ++it)
   {
      width += getCharWidth(*it);
   }
   return width;
}
