/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglTexturedFont.hpp
 *  Created: 18 June 2001
 *  Summary: Class that represents a font with a texture.
 *****************************************************************************/

#ifndef __SGL_TEXTUREFONT_HPP
#define __SGL_TEXTUREFONT_HPP

#include <sgl.h>
#include <sglObject.hpp>
#include <sglVector.hpp>

#include <map>
#include <string>

class sglTexture2D;
class sglBitmapFont;

/** The sglTextureFont class represents a font with a texture.  This
    allows the user of the texture to generate text (by texture mapping
    quads) in 3-space.  The texture defaults to eLINEAR min filter and
    eLINEAR mag filter.  The user may wish to change these defaults
*/
class SGL_DLL_API sglTextureFont : public sglObject
{
public:
   /** Constructs a complete texture font from a bitmap font.  All characters
       in the range of 0-127 will be considered valid, if they have
       non-zero width.
       @param font A bitmap font to use as a source for generating the texture.
       @param use_invalid_char If true, any character that is not represented
              by the font, will be rendered as the given invalid_char.  If
              false, invalid characters will be treated as non-existant (zero
              width).
       @param invalid_char The character to replace invalid characters with
              if use_invalid_char is true.
   */
   sglTextureFont(const sglBitmapFont &font,
                  bool use_invalid_char,
                  char invalid_char);

   /** Constructs a partial texture font from a bitmap font.  Only the
       specified characters will be considered valid.  Careful use of this
       constructor may generate substantially smaller textures than the
       other constructor.
       @param font A bitmap font to use as a source for generating the texture.
       @param valid_chars The set of characters to be treated as valid.
       @param use_invalid_char If true, any character that is not represented
              by the font, will be rendered as the given invalid_char.  If
              false, invalid characters will be treated as non-existant (zero
              width).
       @param invalid_char The character to replace invalid characters with
              if use_invalid_char is true.
   */
   sglTextureFont(const sglBitmapFont &font,
                  const vector<char> &valid_chars,
                  bool use_invalid_char,
                  char invalid_char);

   /// destructor
   ~sglTextureFont();

   /** Queries all the information required to render a text string with
       this font.
       @param str The string to be rendered.
       @param coord Must point to an array of at least 4*str.size() elements.
              On return will contain coordinate values for each valid
              character, offset from (0,0,0).  The coordinates are in
              XZ plane.
       @param tex_coords Must point to an array of at least 4*str.size()
              elements.  On return will contain texture coordinate values
              for each valid character.
       @param scale The desired height of the string in world space.
       @param vertical If true, the text will run in the negative z direction,
              otherwize in the positive x direction.
   */
   unsigned int getStringInfo(const string &str,
                              sglVec3f *coord,
                              sglVec2f *tex_coords,
                              float scale,
                              const sglVec3f &offset,
                              bool vertical = false) const;

   /** Gets the texture.
       @return The fonts texture.
   */
   sglTexture2D* getTexture() const { return m_texture; }

   /** Gets the height of the font in texels.
       @return the height of the font in texels
   */
   unsigned int getHeight() const;

   /** Gets the width of the given character in texels.
       @param c the character whose width will be returned
       @return the width of the given character in texels
   */
   unsigned int getCharWidth(char c) const;

   /** Gets the width in texels of the widest character in this font.
       @return the width in texels of the widest character in this font
   */
   unsigned int getCharWidthMax() const { return m_max_width; }

   /** Gets the width in texels of the given string.
       @param str the string whose width will be returned
       @return the width in texels of the given string
   */
   unsigned int getStringWidth(const string &str) const;

private:
   void computeTextureSize(
      unsigned int &tex_width,
      unsigned int &tex_height);

   void generateTexture(
      const sglBitmapFont &font,
      unsigned int tex_width,
      unsigned int tex_height);

   // not implemented
   sglTextureFont();
   sglTextureFont(const sglTextureFont &);
   sglTextureFont& operator=(const sglTextureFont &);

private:
   struct TexCorners
   {
      sglVec2f& operator[](unsigned int indx) { return m_corners[indx]; }
      const sglVec2f& operator[](unsigned int indx) const
         { return m_corners[indx]; }

      sglVec2f m_corners[4];
   };

private:
   sglTexture2D           *m_texture;
   void                   *m_image;

   unsigned int            m_height;
   unsigned int            m_max_width;

   char                    m_invalid_char;
   bool                    m_invalid_char_flag;

   map<char, bool>         m_valid_chars;
   map<char, TexCorners>   m_tex_coords;
   map<char, float>        m_char_widths;
};

#endif
