/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglTexturedTextSet.cpp (was sglTexturedText)
 *  Created: 18 June 2001
 *  Summary: render 2D text in a 3D location in the scene using a texture font
 *****************************************************************************/

#include <sglTexturedTextSet.hpp>
#include <sglTextureFont.hpp>
//#include <sglQuadSet.hpp>

#include <algorithm>

//----------------------------------------------------------------------------
// class sglTexturedTextSet
//----------------------------------------------------------------------------
sglTexturedTextSet::sglTexturedTextSet(sglTextureFont *font)
      : sglDrawable(),
        m_text(""),
        m_height(1.0f),
        m_vertical(false),
        m_font(NULL),
        m_position(0.f, 0.f, 0.f),
        m_num_prims(0),
        m_coords(NULL),
        m_color(1.f, 1.f, 1.f, 1.f),
        m_normal(0.f, 0.f, 1.f),
        m_transparency_statelet(true)
{
   m_transparency_statelet.ref();
   m_statelets.push_back(&m_transparency_statelet);

   setFont(font);

   m_draw_func = reinterpret_cast<DrawFunc>(&sglTexturedTextSet::drawGeometry);
}

//----------------------------------------------------------------------------
sglTexturedTextSet::~sglTexturedTextSet()
{
   vector<sglStatelet*>::iterator it =
      find(m_statelets.begin(), m_statelets.end(), &m_transparency_statelet);
   if (it != m_statelets.end())
   {
      m_statelets.erase(it);
      setState(0, m_statelets);
   }
   m_transparency_statelet.deref();

   if (m_font)
   {
       m_font->deref();
       if (m_font->getRefCount() == 0)
           delete m_font;
   }

   delete[] m_coords;
}

//----------------------------------------------------------------------------
void sglTexturedTextSet::setString(const string &str)
{
   if (str != m_text)
   {
      m_text = str;
      updateArrays();
   }
}

//----------------------------------------------------------------------------
void sglTexturedTextSet::setCharHeight(float height)
{
   m_height = height;
   updateArrays();
}

//----------------------------------------------------------------------------
float sglTexturedTextSet::getHeight() const
{
   if (m_vertical)
      return m_height * m_text.size();
   else
      return m_height;
}

//----------------------------------------------------------------------------
float sglTexturedTextSet::getWidth() const
{
   float width = 0;

   if (m_font)
   {
      if (m_vertical)
      {
         unsigned int size = m_text.size();
         for (unsigned int i=0; i<size; ++i)
         {
            float tmp = m_font->getCharWidth(m_text[i]);
            if (tmp > width)
               width = tmp;
         }
      }
      else
      {
         width = (float)m_font->getStringWidth(m_text)
            * m_height / (float)m_font->getHeight();
      }
   }
   return width;
}

//----------------------------------------------------------------------------
void sglTexturedTextSet::setVertical(bool on)
{
   if (on != m_vertical)
   {
      m_vertical = on;
      updateArrays();
   }
}

//----------------------------------------------------------------------------
void sglTexturedTextSet::setFont(sglTextureFont *font)
{
   if (m_font != font)
   {
      if (m_font)
      {
         m_font->deref();
         if (m_font->getRefCount() == 0)
            delete m_font;

         m_statelets.pop_back();
      }
      if (font)
      {
         font->ref();
         m_statelets.push_back((sglStatelet*)font->getTexture());
      }
      m_font = font;

      updateArrays();
   }
}

//----------------------------------------------------------------------------
void sglTexturedTextSet::setColor(const sglVec4f &color)
{
   if (color[0] != m_color[0] || color[1] != m_color[1] ||
       color[2] != m_color[2] || color[3] != m_color[3])
   {
      m_color = color;
      dirty(eDLIST, true);
   }
}

//----------------------------------------------------------------------------
void sglTexturedTextSet::setColor(float r, float g, float b, float a)
{
   if (r != m_color[0] || g != m_color[1] || b != m_color[2] ||
       a != m_color[3])
   {
      m_color.set(r, g, b, a);
      dirty(eDLIST, true);
   }
}

//----------------------------------------------------------------------------
void sglTexturedTextSet::getColor(sglVec4f &color) const
{
   color = m_color;
}

//----------------------------------------------------------------------------
void sglTexturedTextSet::setPosition(const sglVec3f &pos)
{
   if (pos[0] != m_position[0] || pos[1] != m_position[1] ||
       pos[2] != m_position[2])
   {
      m_position = pos;
      updateArrays();
   }
}

//----------------------------------------------------------------------------
void sglTexturedTextSet::setPosition(float x, float y, float z)
{
   if (x != m_position[0] || y != m_position[1] || z != m_position[2])
   {
      m_position.set(x, y, z);
      updateArrays();
   }
}

//----------------------------------------------------------------------------
void sglTexturedTextSet::updateArrays()
{
   delete[] m_coords;
   sglVec2f *tex_coords = NULL;

   m_num_prims = m_text.size();
   if (m_num_prims && m_font)
   {
      m_coords = new sglVec3f[m_num_prims*4];
      tex_coords = new sglVec2f[m_num_prims*4];
      m_num_prims = m_font->getStringInfo(m_text, m_coords, tex_coords,
                                          m_height, m_position, m_vertical);
   }
   else
   {
      m_num_prims = 0;
      m_coords = NULL;
   }

   sglTexCoords tmp(new sglDeleteableDataPtr<sglVec2f>(tex_coords));
   setState(0, m_statelets, tmp);
   dirty(eBOUND|eDLIST, true);
}

//----------------------------------------------------------------------------
void sglTexturedTextSet::addStats(sglStats &stats) const
{
   //stats.addText(1);
   stats.addGSets(sglStats::eQUADS, 1);
   stats.addPrims(sglStats::eQUADS, m_num_prims);
   stats.addVertices(sglStats::eQUADS, m_num_prims*4);
   stats.addTriangles(sglStats::eQUADS, (float)(m_num_prims*2));
}

//----------------------------------------------------------------------------
bool sglTexturedTextSet::computeBounds()
{
   m_dirty &= ~eBOUND;

   if (m_num_prims && m_coords)
   {
      m_bbox.makeEmpty();

      const sglVec3f *ptr = m_coords;
      const sglVec3f *end = ptr + m_num_prims * 4;
      for (; ptr != end; ++ptr)
         m_bbox.expandToInclude(*ptr);

      return true;
   }
   else
   {
      m_bbox.makeEmpty();
      return false;
   }
}

//----------------------------------------------------------------------------
void sglTexturedTextSet::drawGeometry(
   const vector<sglTexCoords> &tex_coords_list) const
{
   if (tex_coords_list.empty())
      return;

   glNormal3fv(m_normal.getPtr());
   glColor4fv(m_color.getPtr());

#ifndef SGL_NO_VERTEX_ARRAYS
   if (sgl::s_num_texture_units > 1)
      sgl::setClientActiveTextureUnit(GL_TEXTURE0);

   glEnableClientState(GL_VERTEX_ARRAY);
   glEnableClientState(GL_TEXTURE_COORD_ARRAY);
   glVertexPointer(3, GL_FLOAT, 0, m_coords);
   sglTexCoords::applyArraySet(tex_coords_list);
   glDrawArrays(GL_QUADS, 0, m_num_prims * 4);
   glDisableClientState(GL_VERTEX_ARRAY);
   glDisableClientState(GL_TEXTURE_COORD_ARRAY);

#else
   const sglTexCoords *tex_coords = &tex_coords_list[0];

   unsigned int index = 0;
   glBegin(GL_QUADS);
   for (unsigned int k=0; k<m_num_prims; k++)
   {
      for (unsigned int j=0; j<4; ++j)
      {
         tex_coords->apply(index);
         glVertex3fv(m_coords[index++].getPtr());
      }
   }
   glEnd();
#endif
}

//----------------------------------------------------------------------------
sglBound::IntersectResultEnum sglTexturedTextSet::intersectPrim(
   sglIntersectf &segset)
{
   sglSegmentf seg(segset.getSegment());

   // check against the bounding box
   sglBound::IntersectResultEnum result = m_bbox.intersect(seg);
   if (result == sglBound::eOUTSIDE)
   {
      return result;
   }

   // else result == INSIDE | INTERSECT
   // the bounding box test trims the segment so we have to restore it
   seg = segset.getSegment();

   result = sglBound::eOUTSIDE;

   unsigned int index = 0;

   // perform the intersect on each triangle
   for (unsigned int k=0; k<m_num_prims; k++)
   {
      index += 2;
      int index0, index1;

      for (unsigned int j=2; j<4; j++, index++)
      {
         // compute the vertex indices for the front face:
         index1 = index - 1;
         index0 = index - 2;

         segset.intersectTriangle(seg,
                                  m_coords[index0],
                                  m_coords[index1],
                                  m_coords[index],
                                  result);
      }
   }

   if (result == sglBound::eINTERSECT)
   {
      segset.setSegment(seg);
   }

   return result;
}

//----------------------------------------------------------------------------
sglBound::IntersectResultEnum sglTexturedTextSet::intersectPrim(
   sglIntersectd &segset)
{
   sglSegmentd seg(segset.getSegment());

   // check against the bounding box
   sglBound::IntersectResultEnum result = m_bbox.intersect(seg);
   if (result == sglBound::eOUTSIDE)
   {
      return result;
   }

   // else result == INSIDE | INTERSECT
   // the bounding box test trims the segment so we have to restore it
   seg = segset.getSegment();

   result = sglBound::eOUTSIDE;

   unsigned int index = 0;

   // perform the intersect on each triangle
   for (unsigned int k=0; k<m_num_prims; k++)
   {
      index += 2;
      int index0, index1;

      for (unsigned int j=2; j<4; j++, index++)
      {
         // compute the vertex indices for the front face:
         index1 = index - 1;
         index0 = index - 2;

         segset.intersectTriangle(seg,
                                  m_coords[index0],
                                  m_coords[index1],
                                  m_coords[index],
                                  result);
      }
   }

   if (result == sglBound::eINTERSECT)
   {
      segset.setSegment(seg);
   }

   return result;
}

//----------------------------------------------------------------------------
void sglTexturedTextSet::printInfo(ostream &ostrm,
                                   const char *indent_string) const
{
   sglDrawable::printInfo(ostrm, indent_string);

   ostrm << indent_string << "text        = '" << m_text << "'" << endl;
   ostrm << indent_string << "height      = " << m_height << endl;
   ostrm << indent_string << "orientation = "
         << (m_vertical ? "vertical" : "horizontal") << endl;
   ostrm << indent_string << "color       = " << m_color << endl;
}
