/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglTexturedTextSet.hpp (was sglTexturedText)
 *  Created: 18 June 2001
 *  Summary: render text in a 3D location in the scene using a texture font
 *****************************************************************************/

#ifndef __SGL_TEXTUREDTEXTSET_HPP
#define __SGL_TEXTUREDTEXTSET_HPP

#include <sgl.h>

#include <sglDrawable.hpp>
#include <sglStatelet.hpp>
#include <sglVector.hpp>
#include <string>

class sglTextureFont;

//============================================================================

/** sglTexturedTextSet is a special subclass of sglDrawable that contains
    geometry to render a text string in world space in the scene.  The text is
    rendered in the XY plane.
*/
class SGL_DLL_API sglTexturedTextSet : public sglDrawable
{
public:
   /** Constructor.
       @param font The texture font to use.
   */
   sglTexturedTextSet(sglTextureFont *font);

   /// Destructor.
   virtual ~sglTexturedTextSet();

   /** Set the string to render.
       @param str the string to render
   */
   void setString(const string &str);

   /** Get the string to render.
       @return the string to render
   */
   const string& getString() const {return m_text;}

   /** Set the font to use.
       @param font The font to use.  If NULL, nothing will render.
   */
   void setFont(sglTextureFont *font);

   /** Get the font to use.
       @return the font to use
   */
   sglTextureFont* getFont() const { return m_font; }

   /** Set the text color.
       @param color the text color; default (1,1,1,1)
   */
   void setColor(const sglVec4f &color);

   /** Set the text color (default is white opaque).
       @param r Red color component.
       @param g Green color component.
       @param b Blue color component.
       @param a Alpha color component.
   */
   void setColor(float r, float g, float b, float a);

   /** Get the text color.
       @return the text color
   */
   void getColor(sglVec4f &color) const;

   /** Set the offset vector.
       @param pos The offset vector (default: 0, 0, 0).
   */
   void setPosition(const sglVec3f &pos);

   /** Set the offset vector.
       @param x X component of the offset vector.
       @param y Y component of the offset vector.
       @param z Z component of the offset vector.
   */
   void setPosition(float x, float y, float z);

   /** Get the current offset vector.
       @return The current offset vector.
   */
   const sglVec3f &getPosition() const { return m_position; }

   /** Set the height (in world space) of each character in the string.
       @param height the height of each character
   */
   void setCharHeight(float height);

   /** Get the height of the text string in world space.
       @return the height of the text string in world space
   */
   float getHeight() const;

   /** Gets the width of the text string in world space.
       @return the width of the text string in world space
   */
   float getWidth() const;

   /** Controls whether the string flows left to right or top to bottom.
       @param on If true, text flows top down.
   */
   void setVertical(bool on);

   /** Gets the direction of text.
       @return True if text flows top down.
   */
   bool getVertical() const { return m_vertical; }

   // ---------------- Functions required by sglDrawable -----------------
   // Add this drawable's stats to the given debug_stats.
   virtual void addStats(sglStats &stats) const;

   virtual void drawGeometry(const vector<sglTexCoords> &tex_coords) const;

   ///
   virtual sglBound::IntersectResultEnum intersectPrim(sglIntersectf &segset);
   ///
   virtual sglBound::IntersectResultEnum intersectPrim(sglIntersectd &segset);

   /** Output the state of this node to the specified stream.  See
       sglObject::printInfo() for details.
   */
   virtual void printInfo(ostream &ostrm, const char *indent_string) const;

protected:
   void updateArrays();

   // ---------------- Functions required by sglDrawable -----------------
   virtual bool computeBounds();

private:  // not implemented
   sglTexturedTextSet();
   sglTexturedTextSet(const sglTexturedTextSet &);
   sglTexturedTextSet& operator=(const sglTexturedTextSet &);

private:
   string                 m_text;
   float                  m_height;
   bool                   m_vertical;

   sglTextureFont        *m_font;
   sglVec3f               m_position;

   unsigned int           m_num_prims;
   sglVec3f              *m_coords;
   sglVec4f               m_color;
   sglVec3f               m_normal;

   sglTransparency        m_transparency_statelet;
   vector<sglStatelet*>   m_statelets;
};

#endif
