/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglTimedAnimation.cpp
 *   Author: Scott McMillan
 *  Created: 29 May 2001
 *  Summary: key frame animation node
 *****************************************************************************/

#include <sglTimedAnimation.hpp>
#include <sglCull.hpp>

//----------------------------------------------------------------------------
sglTimedAnimation::sglTimedAnimation()
      : sglAnimation(),
        m_start_time(),
        m_pause_time(),
        m_pause_diff(0.),
        m_total_time(0.f),
        m_total_time2(0.f),
        m_forward_time(0.f),
        m_return_time(0.f)
{
}

//----------------------------------------------------------------------------
sglTimedAnimation::~sglTimedAnimation()
{
}

//----------------------------------------------------------------------------
bool sglTimedAnimation::addChild(sglNode *node)
{
   if (sglGroup::addChild(node))
   {
      m_durations.push_back(0.f);
      computeValues();
      return true;
   }
   else
      return false;
}

//----------------------------------------------------------------------------
bool sglTimedAnimation::insertChild(unsigned int index, sglNode *child)
{
   // HACK duplicates code in sglGroup::insertChild
   if (child)
   {
      if (index < getNumChildren())
      {
         m_durations.insert(m_durations.begin() + index, 0.f);
         computeValues();
      }
      else if (index == getNumChildren())
      {
         m_durations.push_back(0.f);
         computeValues();
      }
   }

   return sglGroup::insertChild(index, child);
}

//----------------------------------------------------------------------------
bool sglTimedAnimation::removeChild(sglNode *child)
{
   // HACK duplicates code in sglGroup::removeChild
   if (child)
   {
      int index = m_child_list.size()-1;
      while(index >= 0 && m_child_list[index] != child)
         --index;

      if (index >= 0)
      {
         m_durations.erase(m_durations.begin() + index);
         computeValues();
      }
   }

   return sglGroup::removeChild(child);
}

//----------------------------------------------------------------------------
sglNode *sglTimedAnimation::removeChild(unsigned int index)
{
   // HACK duplicates some code in sglGroup::removeChild()
   if (index < getNumChildren())
   {
      m_durations.erase(m_durations.begin() + index);
      computeValues();
   }

   return sglGroup::removeChild(index);
}

//----------------------------------------------------------------------------
void sglTimedAnimation::setOffsetTime(const sglTimespec &offset_time)
{
    if (offset_time.m_sec >= 0)
    {
        m_offset_time = offset_time;
    }
}

//----------------------------------------------------------------------------
bool sglTimedAnimation::setDuration(sglNode *node, float sec)
{
   bool success = false;

   if (sec > 0.f)
   {
      // replace duration for FIRST occurence of node
      int index = searchChild(node);
      if (index != -1)
      {
         m_durations[index] = sec;
         computeValues();
         success = true;
      }
   }

   return success;
}

//----------------------------------------------------------------------------
float sglTimedAnimation::getDuration(sglNode *node) const
{
   float sec = 0.f;
   int index = searchChild(node);
   if (index != -1)
   {
      sec = m_durations[index];
   }

   return sec;
}

//----------------------------------------------------------------------------
bool sglTimedAnimation::setDuration(unsigned int index, float sec)
{
   bool success = false;

   if (index < m_durations.size() && sec > 0.f)
   {
      m_durations[index] = sec;
      computeValues();
      success = true;
   }

   return success;
}

//----------------------------------------------------------------------------
float sglTimedAnimation::getDuration(unsigned int index) const
{
   float sec = 0.f;

   if (index < m_durations.size())
   {
      sec = m_durations[index];
   }
   return sec;
}

//----------------------------------------------------------------------------
void sglTimedAnimation::computeValues()
{
   unsigned int ix;
   m_total_time = 0.f;
   m_forward_time = 0.f;
   m_return_time = 0.f;
   if (m_range_flag)
   {
      for (ix=m_first; ix<m_last+1; ++ix)
      {
         m_total_time += m_durations[ix];
      }
      for (ix=m_first; ix<m_last; ++ix)
      {
         m_forward_time += m_durations[ix];
      }
      for (ix=m_first+1; ix<m_last+1; ++ix)
      {
         m_return_time += m_durations[ix];
      }
   }
   else if (!m_durations.empty())
   {
      unsigned int dsize = m_durations.size();
      for (ix=0; ix<dsize; ++ix)
      {
         m_total_time += m_durations[ix];
      }
      for (ix=0; ix<dsize-1; ++ix)
      {
         m_forward_time += m_durations[ix];
      }
      for (ix=1; ix<dsize; ++ix)
      {
         m_return_time += m_durations[ix];
      }
   }

   if (m_num_cycles > 0)
   {
      m_total_time2 = m_total_time +
         (m_num_cycles - 1)*(m_forward_time + m_return_time);
      if (m_reverse_flag)
      {
         m_total_time2 += m_return_time;
      }
      else
      {
         m_total_time2 += m_forward_time;
      }
   }
}

//----------------------------------------------------------------------------
void sglTimedAnimation::preDraw(const PreDrawStruct &trav_state)
{
   if (m_state_dirty_flag)
   {
      switch (m_state)
      {
         case eSTART:
            m_start_time = trav_state.m_frame_time;
            if (m_offset_time < m_start_time)
                m_start_time -= m_offset_time;
            break;

         case ePAUSE:
            m_pause_time = trav_state.m_frame_time;
            m_pause_diff =
               sglTimespec::convertToDouble(m_pause_time - m_start_time);
            break;

         case eRESUME:
            // advance the start time
            // m_start_time = m_start_time + (current_time - m_pause_time)
            m_start_time += (trav_state.m_frame_time - m_pause_time);
            m_state = eSTART;  // necessary? yes
            m_pause_time.m_sec = m_pause_time.m_nsec = 0;
            break;

         case eSTOP:
            m_start_time.m_sec  = m_pause_time.m_sec = 0;
            m_start_time.m_nsec = m_pause_time.m_nsec = 0;
            break;
      }
      m_state_dirty_flag = false;
   }

   sglGroup::preDraw(trav_state);
}

//----------------------------------------------------------------------------
sglNode *sglTimedAnimation::computeChild(double atime) const
{
   atime *= m_speed;

   if (m_mode == eCYCLE)
   {
      if (m_num_cycles == 0 || atime < m_total_time*m_num_cycles)
      {
         atime = atime - ((int)(atime/m_total_time))*m_total_time;
      }
      else
      {
         return NULL;
      }
   }
   else // m_mode == eSWING
   {
      double cycle_time = m_forward_time + m_return_time;

      if (m_num_cycles == 0 || atime < m_total_time2)
         atime = atime - ((int)(atime/cycle_time))*cycle_time;
      else
         return NULL;

      // back swing?
      if (atime >= m_total_time)
      {
         atime = 2*m_total_time - atime;
         if (m_range_flag && m_first != m_last)
         {
            atime -= m_durations[m_last];
         }
         else
         {
            atime -= m_durations[m_durations.size()-1];
         }
      }
   }

   unsigned int index;
   if (!m_reverse_flag)
   {
      if (!m_range_flag)
      {
         index = 0;
      }
      else
      {
         index = m_first;
      }

      atime -= m_durations[index];
      while (atime > 0)
      {
         ++index;
         atime -= m_durations[index];
      }
   }
   else
   {
      if (!m_range_flag)
      {
         index = m_child_list.size() - 1;
      }
      else
      {
         index = m_last;
      }

      atime -= m_durations[index];
      while (atime > 0)
      {
         --index;
         atime -= m_durations[index];
      }
   }

   return m_child_list[index];
}

//----------------------------------------------------------------------------
sglNode *sglTimedAnimation::selectChild(const sglTimespec &current_time) const
{
   // the following are to be checked by the functions that call this one so
   // that this one does not have to do it again;
   //if (m_state == eSTOP || m_total_time == 0.f || m_speed == 0.f)
   //{
   //   return NULL;
   //}

   sglNode* child = NULL;
   if (m_state == ePAUSE)
   {
      child = computeChild(m_pause_diff);
   }
   else  // mode is eRESUME or eSTART
   {
      child = computeChild(sglTimespec::convertToDouble(current_time -
                                                        m_start_time));
   }

   return child;
}

//----------------------------------------------------------------------------
sglBound::IntersectResultEnum sglTimedAnimation::intersect(
   sglIntersectf &isector) const
{
   if ((isector.getMode()&sglProbe::eANIMATION_NONE) ||
       m_state == eSTOP || m_total_time == 0.f || m_speed == 0.f ||
       m_child_list.empty())
   {
      return sglBound::eOUTSIDE;
   }
   else if (isector.getMode()&sglProbe::eANIMATION_ALL)
   {
      return sglGroup::intersect(isector);
   }
   else // default: current child selected by switch only
   {
      sglSegmentf seg(isector.getSegment());
      if (m_sphere.intersect(seg) == sglBound::eOUTSIDE)
      {
         return sglBound::eOUTSIDE;
      }

      sglBound::IntersectResultEnum result = sglBound::eOUTSIDE;
      sglNode *child = selectChild(isector.getFrameCount());
      if (child)
      {
         result = child->intersect(isector);
      }

      if (result == sglBound::eINTERSECT &&
          isector.getMode() & sglProbe::ePATH)
      {
         isector.getPath().push_front((sglNode*)this);
      }
      return result;
   }
}

//----------------------------------------------------------------------------
sglBound::IntersectResultEnum sglTimedAnimation::intersect(
   sglIntersectd &isector) const
{
   if ((isector.getMode()&sglProbe::eANIMATION_NONE) ||
       m_state == eSTOP || m_total_time == 0.f || m_speed == 0.f ||
       m_child_list.empty())
   {
      return sglBound::eOUTSIDE;
   }
   else if (isector.getMode()&sglProbe::eANIMATION_ALL)
   {
      return sglGroup::intersect(isector);
   }
   else // default: current child selected by switch only
   {
      sglSegmentd seg(isector.getSegment());
      if (m_sphere.intersect(seg) == sglBound::eOUTSIDE)
      {
         return sglBound::eOUTSIDE;
      }

      sglBound::IntersectResultEnum result = sglBound::eOUTSIDE;
      sglNode *child = selectChild(isector.getFrameCount());
      if (child)
      {
         result = child->intersect(isector);
      }

      if (result == sglBound::eINTERSECT &&
          isector.getMode() & sglProbe::ePATH)
      {
         isector.getPath().push_front((sglNode*)this);
      }
      return result;
   }
}

//----------------------------------------------------------------------------
void sglTimedAnimation::pick(sglPickf &pick_state,
                             unsigned int cull_flags) const
{
   if ((pick_state.getMode()&sglProbe::eANIMATION_NONE) ||
       m_state == eSTOP || m_total_time == 0.f || m_speed == 0.f ||
       m_child_list.empty())
   {
      return;  // select nothing
   }
   else if (pick_state.getMode()&sglProbe::eANIMATION_ALL)
   {
      sglGroup::pick(pick_state, cull_flags);
   }
   else // default: current child selected by switch only
   {
      if (cull_flags && (m_sphere.intersect(pick_state.getPickFrustum(),
                                            cull_flags) == sglBound::eOUTSIDE))
      {
         return; // select nothing
      }

      sglNode *child = selectChild(pick_state.getFrameCount());

      if (child)
      {
         // if path is requested store it ahead of time (different from
         // intersect), in case something beneath here is picked successfully.
         if (pick_state.getMode() & sglProbe::ePATH)
         {
            pick_state.getPath().push_back((sglNode*)this);
         }

         child->pick(pick_state, cull_flags);

         if (pick_state.getMode() & sglProbe::ePATH)
         {
            pick_state.getPath().pop_back();
         }
      }
   }
}

//----------------------------------------------------------------------------
void sglTimedAnimation::pick(sglPickd &pick_state,
                             unsigned int cull_flags) const
{
   if ((pick_state.getMode()&sglProbe::eANIMATION_NONE) ||
       m_state == eSTOP || m_total_time == 0.f || m_speed == 0.f ||
       m_child_list.empty())
   {
      return;  // select nothing
   }
   else if (pick_state.getMode()&sglProbe::eANIMATION_ALL)
   {
      sglGroup::pick(pick_state, cull_flags);
   }
   else // default: current child selected by switch only
   {
      if (cull_flags && (m_sphere.intersect(pick_state.getPickFrustum(),
                                            cull_flags) == sglBound::eOUTSIDE))
      {
         return; // select nothing
      }

      sglNode *child = selectChild(pick_state.getFrameCount());

      if (child)
      {
         // if path is requested store it ahead of time (different from
         // intersect), in case something beneath here is picked successfully.
         if (pick_state.getMode() & sglProbe::ePATH)
         {
            pick_state.getPath().push_back((sglNode*)this);
         }

         child->pick(pick_state, cull_flags);

         if (pick_state.getMode() & sglProbe::ePATH)
         {
            pick_state.getPath().pop_back();
         }
      }
   }
}

//----------------------------------------------------------------------------
void sglTimedAnimation::cull(sglCull<float> &trav_state,
                             unsigned int cull_flags) const
{
   if (m_state == eSTOP || m_total_time == 0.f || m_speed == 0.f ||
       m_child_list.empty() ||
       (cull_flags && (m_sphere.intersect(trav_state.getFrustum(),
                                          cull_flags) == sglBound::eOUTSIDE)))
   {
      return; // draw nothing
   }

   sglNode *child = selectChild(trav_state.getFrameTime());

   // select child based on range
   if (child)
   {
      child->cull(trav_state, cull_flags);
   }
}

//----------------------------------------------------------------------------
void sglTimedAnimation::cull(sglCull<double> &trav_state,
                             unsigned int cull_flags) const
{
   if (m_state == eSTOP || m_total_time == 0.f || m_speed == 0.f ||
       m_child_list.empty() ||
       (cull_flags && (m_sphere.intersect(trav_state.getFrustum(),
                                          cull_flags) == sglBound::eOUTSIDE)))
   {
      return; // draw nothing
   }

   sglNode *child = selectChild(trav_state.getFrameTime());

   // select child based on range
   if (child)
   {
      child->cull(trav_state, cull_flags);
   }
}

//----------------------------------------------------------------------------
sglNode *sglTimedAnimation::clone(unsigned int mode) const
{
   sglTimedAnimation *root = new sglTimedAnimation;
   sglTimedAnimation::copyTo(root, mode);
   return root;
}

//----------------------------------------------------------------------------
void sglTimedAnimation::copyTo(sglTimedAnimation *dst, unsigned int mode) const
{
   sglAnimation::copyTo(dst, mode);

   // transfer stuff
   for (unsigned int index=0; index<getNumChildren(); ++index)
   {
      dst->setDuration(index, getDuration(index));
   }
}

//----------------------------------------------------------------------------
void sglTimedAnimation::printInfo(ostream &ostrm,
                                  const char *indent_string) const
{
   sglAnimation::printInfo(ostrm, indent_string);
   ostrm << indent_string << "frame durations = [";
   for (unsigned int i=0; i<m_durations.size(); ++i)
   {
      ostrm << " " << m_durations[i];
   }
   ostrm << " ]" << endl;
}
