/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglTimedAnimation.hpp
 *  Created: 29 May 2001
 *  Summary: animation based on frame.
 *****************************************************************************/

#ifndef __SGL_TIMED_ANIMATION_HPP
#define __SGL_TIMED_ANIMATION_HPP

#include <sgl.h>
#include <sglTimespec.hpp>
#include <sglAnimation.hpp>

//----------------------------------------------------------------------------

/**

A subclass of sglAnimation, this node selects one child as a function of the
current frame time (not count) and the animation mode.  As each child is added
to this node the default number of seconds it is to be displayed is set to
zero.  This value can be changed by calling one of the setDuration(..) member
functions and identifying the child either by pointer or index and specifying
the number of seconds.

<b>Example:</b>

<pre>
   sglTimedAnimation *animation = new sglTimedAnimation;
   for (unsigned int ix=0; ix<10; ++ix)
   {
      sglGeode *geode = new sglGeode;
      animation->addChild(geode);
      animation->setDuration(ix, 0.6f);  // display for 0.6 seconds
   }
   animation->setMode(sglAnimation::eCYCLE); // forward only
   animation->setRange();                    // display all
   animation->setSpeed(0.75f);               // .75 of normal speed
                                             //    (0.8 seconds each)
   animation->setNumCycles(1);               // one-shot

   // set the animation to start on the next frame (after preDraw()).
   animation->setState(sglAnimation::eSTART);
</pre>

 */

//----------------------------------------------------------------------------

class SGL_DLL_API sglTimedAnimation : public sglAnimation
{
public:
   /// default constructor
   sglTimedAnimation();
   /// virtual destructor
   virtual ~sglTimedAnimation();

   // ======== members from sglGroup that need to be augmented =========
   virtual bool addChild(sglNode *node);
   virtual bool insertChild(unsigned int index, sglNode *node);
   virtual bool removeChild(sglNode *node);
   virtual sglNode *removeChild(unsigned int index);
   // ==================================================================

   /** Set an offset time when starting the animation.  Only positive times are
       allowed, which will start the animation further forward into the
       sequence.  This will only take effect the next time the state is changed
       to START.
       @param offset_time Timespec containing the offset (in secs, nsecs)
   */
   void setOffsetTime(const sglTimespec &offset_time);

   /** Query the current time offset.
       @return The current time offset in an sglTimespec object.
   */
   const sglTimespec &getOffsetTime() const { return m_offset_time; }

   /** Sets the duration, in number of seconds, of the first child in the list
       that matches the pointer.
       @param node Pointer to child to set the duration.
       @param sec  Number of seconds to display the child.  Setting to
         zero disables the display of this particular child.
       @return true if the node pointer is one of this node's children;
         otherwise false
   */
   bool setDuration(sglNode *node, float sec);

   /** Gets the current duration of the child specified by the node pointer.
       @param node Pointer to child to get the duration from.
       @return duration of the specified node (zero if the node was not a
       child).
   */
   float getDuration(sglNode *node) const;

   /** Sets the duration, in number of seconds, of the child specified by the
       index.
       @param index Index of child to set the duration.
       @param sec   Number of seconds to display the child.  Setting to
         zero disables the display of this particular child.
       @return true if the index is valid; otherwise false
   */
   bool setDuration(unsigned int index, float sec);

   /** Gets the current duration of the child corresponding to the index.
       @param index Index of child to set the duration.
       @return duration of the specified node (zero if the index was out of
         range.
   */
   float getDuration(unsigned int index) const;

   // see sglNode::intersect(f) for details.
   virtual sglBound::IntersectResultEnum intersect(
      sglIntersectf &isector) const;

   // see sglNode::intersect(d) for details.
   virtual sglBound::IntersectResultEnum intersect(
      sglIntersectd &isector) const;

   // see sglNode::pick(f) for details.
   virtual void pick(sglPickf &pick_state, unsigned int cull_flags) const;

   // see sglNode::pick(d) for details.
   virtual void pick(sglPickd &pick_state, unsigned int cull_flags) const;

   // See sglNode::preDraw() for details
   virtual void preDraw(const PreDrawStruct &trav_state);

   // see sglNode::clone() for details.
   virtual sglNode *clone(unsigned int mode) const;

   // see sglObject::printInfo() for details.
   virtual void printInfo(ostream &ostrm, const char *indent_string) const;

protected:
   // sglNode::cull(f) for details
   virtual void cull(sglCull<float> &trav_state,
                     unsigned int cull_flags) const;

   // sglNode::cull(d) for details
   virtual void cull(sglCull<double> &trav_state,
                     unsigned int cull_flags) const;

   virtual void computeValues();
   sglNode* computeChild(double atime) const;
   sglNode* selectChild(const sglTimespec &current_time) const;

   void copyTo(sglTimedAnimation *dst, unsigned int mode) const;

private:     // not implemented
   sglTimedAnimation(const sglTimedAnimation &);
   sglTimedAnimation &operator=(const sglTimedAnimation &);

protected:
   vector<float> m_durations;

   sglTimespec   m_start_time;      // time this node was started on
   sglTimespec   m_pause_time;      // time this node was paused on
   sglTimespec   m_offset_time;     // initial time offset when starting the
                                    //   animation
   double        m_pause_diff;

   float         m_total_time;

   float         m_total_time2;
   float         m_forward_time;    // for internal use during eSWING
   float         m_return_time;
};

#endif
