/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglTransform.cpp
 *  Created: 14 November 1998
 *  Summary: transform the model view matrix before traversing the children
 *****************************************************************************/

#include <sglTransform.hpp>
#include <sglCull.hpp>

//----------------------------------------------------------------------------
template <class T>
void sglTransform<T>::checkForDeformation()
{
   T x = (m_matrix[0][0]*m_matrix[0][0] +
          m_matrix[0][1]*m_matrix[0][1] +
          m_matrix[0][2]*m_matrix[0][2]);
   T y = (m_matrix[1][0]*m_matrix[1][0] +
          m_matrix[1][1]*m_matrix[1][1] +
          m_matrix[1][2]*m_matrix[1][2]);
   T z = (m_matrix[2][0]*m_matrix[2][0] +
          m_matrix[2][1]*m_matrix[2][1] +
          m_matrix[2][2]*m_matrix[2][2]);
   if (sglAlmostEqual(x, (T)1.0) &&
       sglAlmostEqual(y, (T)1.0) &&
       sglAlmostEqual(z, (T)1.0))
   {
      m_rescale = false;
      m_normalize = false;
   }
   else if (x != 0 && sglRelativelyEqual(x, y) && sglRelativelyEqual(x, z))
   {
      m_rescale = true;
      m_normalize = false;
      //cout << "rescale = " << m_rescale << endl;
   }
   else
   {
      m_rescale = false;
      m_normalize = true;
      //cout << "normalize = " << m_normalize << endl;
      //cout << "x=" << x << ", y=" << y << ", z=" << z << endl;
      //cout << "mat = " << m_matrix << endl;
   }
}

//----------------------------------------------------------------------------
bool checkForIdentity(const sglMat4d &matrix)
{
   return matrix.almostEqual(sglIdentMat4d, sglDblEpsilon);
}

//----------------------------------------------------------------------------
bool checkForIdentity(const sglMat4f &matrix)
{
   return matrix.almostEqual(sglIdentMat4f, sglFltEpsilon);
}

//----------------------------------------------------------------------------
template <class T>
void sglTransform<T>::checkForIdentity()
{
   m_is_ident = ::checkForIdentity(m_matrix);
   checkForDeformation();
}

//----------------------------------------------------------------------------
template <class T>
void sglTransform<T>::loadIdentity()
{
   m_matrix.buildIdentity();
   m_inv_mat.buildIdentity();
   m_is_ident = true;
   m_rescale = false;
   m_normalize = false;

   dirty(eBOUND);
}

//----------------------------------------------------------------------------
template <class T>
void sglTransform<T>::setTranslation(T x, T y, T z)
{
   m_matrix[3][0] = x;
   m_matrix[3][1] = y;
   m_matrix[3][2] = z;
   m_matrix[3][3] = 1.0;

   m_inv_mat.inverseAffine(m_matrix);
   checkForIdentity();
   dirty(eBOUND);
}

//----------------------------------------------------------------------------
template <class T>
void sglTransform<T>::setRotation(T angle, T x, T y, T z)
{
   /* This function contributed to Mesa by Erich Boleyn (erich@uruk.org) */
   T mag, s, c;
   T xx, yy, zz, xy, yz, zx, xs, ys, zs, one_c;

   mag = x*x + y*y + z*z;

   if ((angle == 0.0) || (mag == 0.0))
   {
      /* generate an identity matrix and return */
      for (int i=0; i<3; i++)
      {
         for (int j=0; j<3; j++)
         {
            m_matrix[i][j] = 0.0;
         }
         m_matrix[i][i] = 1.0;
      }
      return;
   }

   mag = sqrt( mag );
   x /= mag;
   y /= mag;
   z /= mag;

   /*
    *     Arbitrary axis rotation matrix.
    *
    *  This is composed of 5 matrices, Rz, Ry, T, Ry', Rz', multiplied
    *  like so:  Rz * Ry * T * Ry' * Rz'.  T is the final rotation
    *  (which is about the X-axis), and the two composite transforms
    *  Ry' * Rz' and Rz * Ry are (respectively) the rotations necessary
    *  from the arbitrary axis to the X-axis then back.  They are
    *  all elementary rotations.
    *
    *  Rz' is a rotation about the Z-axis, to bring the axis vector
    *  into the x-z plane.  Then Ry' is applied, rotating about the
    *  Y-axis to bring the axis vector parallel with the X-axis.  The
    *  rotation about the X-axis is then performed.  Ry and Rz are
    *  simply the respective inverse transforms to bring the arbitrary
    *  axis back to it's original orientation.  The first transforms
    *  Rz' and Ry' are considered inverses, since the data from the
    *  arbitrary axis gives you info on how to get to it, not how
    *  to get away from it, and an inverse must be applied.
    *
    *  The basic calculation used is to recognize that the arbitrary
    *  axis vector (x, y, z), since it is of unit length, actually
    *  represents the sines and cosines of the angles to rotate the
    *  X-axis to the same orientation, with theta being the angle about
    *  Z and phi the angle about Y (in the order described above)
    *  as follows:
    *
    *  cos ( theta ) = x / sqrt ( 1 - z^2 )
    *  sin ( theta ) = y / sqrt ( 1 - z^2 )
    *
    *  cos ( phi ) = sqrt ( 1 - z^2 )
    *  sin ( phi ) = z
    *
    *  Note that cos ( phi ) can further be inserted to the above
    *  formulas:
    *
    *  cos ( theta ) = x / cos ( phi )
    *  sin ( theta ) = y / sin ( phi )
    *
    *  ...etc.  Because of those relations and the standard trigonometric
    *  relations, it is pssible to reduce the transforms down to what
    *  is used below.  It may be that any primary axis chosen will give the
    *  same results (modulo a sign convention) using thie method.
    *
    *  Particularly nice is to notice that all divisions that might
    *  have caused trouble when parallel to certain planes or
    *  axis go away with care paid to reducing the expressions.
    *  After checking, it does perform correctly under all cases, since
    *  in all the cases of division where the denominator would have
    *  been zero, the numerator would have been zero as well, giving
    *  the expected result.
    */

   s = sin( angle );
   c = cos( angle );

   xx = x * x;
   yy = y * y;
   zz = z * z;
   xy = x * y;
   yz = y * z;
   zx = z * x;
   xs = x * s;
   ys = y * s;
   zs = z * s;
   one_c = 1.0F - c;

   m_matrix[0][0] = (one_c * xx) + c;
   m_matrix[0][1] = (one_c * xy) + zs;
   m_matrix[0][2] = (one_c * zx) - ys;

   m_matrix[1][0] = (one_c * xy) - zs;
   m_matrix[1][1] = (one_c * yy) + c;
   m_matrix[1][2] = (one_c * yz) + xs;

   m_matrix[2][0] = (one_c * zx) + ys;
   m_matrix[2][1] = (one_c * yz) - xs;
   m_matrix[2][2] = (one_c * zz) + c;

   m_inv_mat.inverseAffine(m_matrix);
   checkForIdentity();
   dirty(eBOUND);
}

//----------------------------------------------------------------------------
template <class T>
void sglTransform<T>::setRotation(T h, T p, T r)
{
   T sr = sin( p );
   T cr = cos( p );

   T sh = sin( r );
   T ch = cos( r );

   T sp = sin( h );
   T cp = cos( h );

   m_matrix[0][0] = -sp*sr*sh + cp*ch;
   m_matrix[0][1] =  cp*sr*sh + sp*ch;
   m_matrix[0][2] = -cr*sh;

   m_matrix[1][0] = -sp*cr;
   m_matrix[1][1] =  cp*cr;
   m_matrix[1][2] =  sr;

   m_matrix[2][0] =  sp*sr*ch + cp*sh;
   m_matrix[2][1] = -cp*sr*ch + sp*sh;
   m_matrix[2][2] =  cr*ch;

   m_inv_mat.inverseAffine(m_matrix);
   checkForIdentity();
   dirty(eBOUND);
}

//----------------------------------------------------------------------------
template <class T>
void sglTransform<T>::setMatrix(const sglMat4<T> &m)
{
   m_matrix = m;
   m_inv_mat.inverseAffine(m_matrix);
   checkForIdentity();
   dirty(eBOUND);
}

//----------------------------------------------------------------------------
template <class T>
void sglTransform<T>::computeBound()
{
   sglGroup::computeBound();
   if (!m_sphere.isEmpty() && !m_sphere.isInfinite() && !m_is_ident)
   {
      m_sphere.mul(m_matrix);
      //sglVec4f tmp;
      //::mul(m_sphere.getCenter(), (double)1.0, m_matrix, tmp);
      //m_sphere.setCenter(tmp[0], tmp[1], tmp[2]);

      // STILL NEED TO DO THE SCALE for the radius
   }
}

//----------------------------------------------------------------------------
template <class T>
sglBound::IntersectResultEnum sglTransform<T>::intersect(
   sglIntersectf &segset) const
{
   if (m_is_ident)
   {
      return sglGroup::intersect(segset);
   }

   if (m_child_list.empty())
   {
      return sglBound::eOUTSIDE;
   }

   sglSegmentf seg(segset.getSegment()); // copy the current segment

   // test against the current bounding sphere
   sglBound::IntersectResultEnum result = m_sphere.intersect(seg);
   if (result == sglBound::eOUTSIDE)
   {
      return result;
   }

   // Push a translate onto the stack.
   sglMat4f old_matrix;
   sglFrustumf old_view_frustum;
   segset.pushTransform(m_matrix, m_inv_mat,
                        seg, old_matrix, old_view_frustum);

   // traverse the children
   result = sglBound::eOUTSIDE;
   for (vector<sglNode*>::const_iterator elt = m_child_list.begin();
        elt != m_child_list.end(); ++elt)
   {
      if ((*elt)->intersect(segset) == sglBound::eINTERSECT)
      {
         result = sglBound::eINTERSECT;
      }
   }

   // restore the original pos and dir (leave len intact)
   seg.setLen(segset.getSegment().getLen());

   // pop the translate off the stack
   segset.pop(seg, old_matrix, old_view_frustum);

   if (result == sglBound::eINTERSECT && segset.getMode() & sglProbe::ePATH)
   {
      segset.getPath().push_front((sglNode*)this);
   }

   return result;
}

//----------------------------------------------------------------------------
template <class T>
sglBound::IntersectResultEnum sglTransform<T>::intersect(
   sglIntersectd &segset) const
{
   if (m_is_ident)
   {
      return sglGroup::intersect(segset);
   }

   if (m_child_list.empty())
   {
      return sglBound::eOUTSIDE;
   }

   sglSegmentd seg(segset.getSegment()); // copy the current segment

   // test against the current bounding sphere
   sglBound::IntersectResultEnum result = m_sphere.intersect(seg);
   if (result == sglBound::eOUTSIDE)
   {
      return result;
   }

   // Push a translate onto the stack.
   sglMat4d old_matrix;
   sglFrustumd old_view_frustum;
   segset.pushTransform(m_matrix, m_inv_mat,
                        seg, old_matrix, old_view_frustum);

   // traverse the children
   result = sglBound::eOUTSIDE;
   for (vector<sglNode*>::const_iterator elt = m_child_list.begin();
        elt != m_child_list.end(); ++elt)
   {
      if ((*elt)->intersect(segset) == sglBound::eINTERSECT)
      {
         result = sglBound::eINTERSECT;
      }
   }

   // restore the original pos and dir (leave len intact)
   seg.setLen(segset.getSegment().getLen());

   // pop the translate off the stack
   segset.pop(seg, old_matrix, old_view_frustum);

   if (result == sglBound::eINTERSECT && segset.getMode() & sglProbe::ePATH)
   {
      segset.getPath().push_front((sglNode*)this);
   }

   return result;
}

//----------------------------------------------------------------------------
template <class T>
void sglTransform<T>::pick(sglPickf &pick_state, unsigned int cull_flags) const
{
   if (m_is_ident)
   {
      sglGroup::pick(pick_state, cull_flags);
   }
   else if (m_child_list.empty() ||
            (cull_flags &&
             (m_sphere.intersect(pick_state.getPickFrustum(),
                                 cull_flags) == sglBound::eOUTSIDE)))
   {
      return; // select nothing
   }
   else
   {
      sglMat4f old_matrix;
      sglFrustumf old_frustum, old_view_frustum;
      pick_state.pushTransform(m_matrix, m_inv_mat, old_frustum, old_matrix,
                               old_view_frustum);

      // if path is requested store it ahead of time (different from
      // intersect), in case something beneath here is picked successfully.
      if (pick_state.getMode() & sglProbe::ePATH)
      {
         pick_state.getPath().push_back((sglNode*)this);
      }

      // --------------------------------------------------

      for (vector<sglNode*>::const_iterator elt = m_child_list.begin();
           elt != m_child_list.end(); ++elt)
      {
         (*elt)->pick(pick_state, cull_flags);
      }

      // --------------------------------------------------
      // restore old state
      pick_state.pop(old_frustum, old_matrix, old_view_frustum);

      if (pick_state.getMode() & sglProbe::ePATH)
      {
         pick_state.getPath().pop_back();
      }
   }
}

//----------------------------------------------------------------------------
template <class T>
void sglTransform<T>::pick(sglPickd &pick_state, unsigned int cull_flags) const
{
   if (m_is_ident)
   {
      sglGroup::pick(pick_state, cull_flags);
   }
   else if (m_child_list.empty() ||
            (cull_flags &&
             (m_sphere.intersect(pick_state.getPickFrustum(),
                                 cull_flags) == sglBound::eOUTSIDE)))
   {
      return; // select nothing
   }
   else
   {
      sglMat4d old_matrix;
      sglFrustumd old_frustum, old_view_frustum;
      pick_state.pushTransform(m_matrix, m_inv_mat, old_frustum, old_matrix,
                               old_view_frustum);

      // if path is requested store it ahead of time (different from
      // intersect), in case something beneath here is picked successfully.
      if (pick_state.getMode() & sglProbe::ePATH)
      {
         pick_state.getPath().push_back((sglNode*)this);
      }

      // --------------------------------------------------

      for (vector<sglNode*>::const_iterator elt = m_child_list.begin();
           elt != m_child_list.end(); ++elt)
      {
         (*elt)->pick(pick_state, cull_flags);
      }

      // --------------------------------------------------
      // restore old state
      pick_state.pop(old_frustum, old_matrix, old_view_frustum);

      if (pick_state.getMode() & sglProbe::ePATH)
      {
         pick_state.getPath().pop_back();
      }
   }
}

//----------------------------------------------------------------------------
template <class T>
void sglTransform<T>::cull(sglCull<float> &trav_state,
                           unsigned int cull_flags) const
{
   if (m_is_ident)
   {
      sglGroup::cull(trav_state, cull_flags);
   }
   else if (m_child_list.empty() ||
            (cull_flags &&
             (m_sphere.intersect(trav_state.getFrustum(),
                                 cull_flags) == sglBound::eOUTSIDE)))
   {
      return; // draw nothing
   }
   else
   {
      // compute a new matrix
      sglMat4f new_mat;
      new_mat.mul(m_matrix, trav_state.getModelViewMatrix());
      trav_state.pushModelViewMatrix(new_mat, m_rescale, m_normalize);

      // compute a new frustum
      sglFrustumf old_frustum = trav_state.getFrustum();
      trav_state.getFrustum().mul(old_frustum, m_inv_mat);

      // --------------------------------------------------

      for (vector<sglNode*>::const_iterator elt = m_child_list.begin();
           elt != m_child_list.end(); ++elt)
      {
         (*elt)->cull(trav_state, cull_flags);
      }

      // --------------------------------------------------
      // restore old state
      trav_state.popModelViewMatrix();
      trav_state.getFrustum() = old_frustum;
   }
}

//----------------------------------------------------------------------------
template <class T>
void sglTransform<T>::cull(sglCull<double> &trav_state,
                           unsigned int cull_flags) const
{
   if (m_is_ident)
   {
      sglGroup::cull(trav_state, cull_flags);
   }
   else if (m_child_list.empty() ||
            (cull_flags &&
             (m_sphere.intersect(trav_state.getFrustum(),
                                 cull_flags) == sglBound::eOUTSIDE)))
   {
      return; // draw nothing
   }
   else
   {
      // compute a new matrix
      sglMat4d new_mat;
      new_mat.mul(m_matrix, trav_state.getModelViewMatrix());
      trav_state.pushModelViewMatrix(new_mat, m_rescale, m_normalize);

      // compute a new frustum
      sglFrustumd old_frustum = trav_state.getFrustum();
      trav_state.getFrustum().mul(old_frustum, m_inv_mat);

      // --------------------------------------------------

      for (vector<sglNode*>::const_iterator elt = m_child_list.begin();
           elt != m_child_list.end(); ++elt)
      {
         (*elt)->cull(trav_state, cull_flags);
      }

      // --------------------------------------------------
      // restore old state
      trav_state.popModelViewMatrix();
      trav_state.getFrustum() = old_frustum;
   }
}

//----------------------------------------------------------------------------
template <class T>
void sglTransform<T>::computeTxToParent(sglMat4f &curr_mat) const
{
   curr_mat.mul(m_inv_mat, curr_mat);
}

//----------------------------------------------------------------------------
template <class T>
void sglTransform<T>::computeTxToParent(sglMat4d &curr_mat) const
{
   curr_mat.mul(m_inv_mat, curr_mat);
}

//----------------------------------------------------------------------------
template <class T>
sglNode *sglTransform<T>::clone(unsigned int mode) const
{
   sglTransform<T> *root = new sglTransform<T>;
   sglTransform<T>::copyTo(root, mode);
   return root;
}

//----------------------------------------------------------------------------
template <class T>
void sglTransform<T>::copyTo(sglTransform<T> *dst, unsigned int mode) const
{
   sglGroup::copyTo(dst, mode);

   // transfer stuff
   dst->setMatrix(m_matrix);
}

//----------------------------------------------------------------------------
template <class T>
void sglTransform<T>::printInfo(ostream &ostrm,
                                const char *indent_string) const
{
    sglGroup::printInfo(ostrm, indent_string);

    ostrm << indent_string << "sglTransform<T>: " << endl;
    for (int i=0; i<4; i++)
    {
       ostrm << indent_string << "\t"
             << m_matrix[i][0] << ' ' << m_matrix[i][1] << ' '
             << m_matrix[i][2] << ' ' << m_matrix[i][3] << endl;
    }
}

template class SGL_DLL_API sglTransform<double>;
template class SGL_DLL_API sglTransform<float>;
