/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglTransform.hpp
 *  Created: 14 November 1998
 *  Summary: transform the model view matrix before traversing the children
 *****************************************************************************/

#ifndef __SGL_TRANSFORM_HPP
#define __SGL_TRANSFORM_HPP

#include <sgl.h>
#include <sglMatrix.hpp>
#include <sglGroup.hpp>

//----------------------------------------------------------------------------

/**

With the matrix defined within this node, the model view matrix maintained
within the various traversals is transformed (multiplied) before visiting the
children.  This node contains functions to define affine(?) 3D transformations
that contain translation, rotation, and uniform scale.  By changing the matrix
within this node, you can change the position, orientation and size of all the
children below.  Currently the only way to specify the scale is by sending a
matrix with scale into the setMatrix() member function.  Note that even
non-affine transformations can be specified using the setMatrix() member
function, but I don't currently guarantee correct results (in particular the
intersect traversal does not perform correctly in the presence of non-uniform
scale).

Note that the sglTransform class is a templated class which can allow this to
be instantiated with either single or double precision.  Two typedefs,
sglTransformf and sglTransformd, are added for convenience.  The behaviour when
using other templated types is undefined.

<b>Example:</b>

<pre>
   sglTransformf *dcs_node = new sglTransformf;  // or new sglTransform<float>;
   dcs_node->setTranslation(0.0f, 0.0f, 25.0f);
   dcs_node->setRotation(M_PI, 1.0f, 0.0f, 0.0f);
</pre>

\todo Add support for non-uniform scale (currently intersect traversal does not
work in its presence).

 */

template <class T>
class SGL_DLL_API sglTransform : public sglGroup
{
public:
   /// default constructor - creates an identity transform
   sglTransform() { loadIdentity(); }
   /// virtual destructor
   virtual ~sglTransform() {}

   /// Reset this node to an identity transform.
   void loadIdentity();

   /** Set the translation part of the transform (affects the 4th row of the
       matrix only).
       @param x Translation along the x axis.
       @param y Translation along the y axis.
       @param z Translation along the z axis.
   */
   void setTranslation(T x, T y, T z);

   /** Set the rotation portion of the transform (affects the upper left 3x3
       portion of the matrix only) by specifying and angle and axis.  The axis
       does not have to be normalized.
       @param angle The angle in radians.
       @param x X component of the axis.
       @param y Y component of the axis.
       @param z Z component of the axis.
   */
   void setRotation(T angle, T x, T y, T z);  // radians

   /** Set the rotation portion of the transform (affects the upper left 3x3
       portion of the matrix only) by specifying euler angles.
       @param h Heading angle is the rotation about the z-axis.
       @param p Pitch angle is the rotation about the once-rotated x-axis.
       @param r Roll angle is the rotation about the twice-rotated y-axis.
   */
   void setRotation(T h, T p, T r);  // radians

   /** Set the transform matrix directly.  There is currently no error checking
       to see if non-supported matrix types are set here.  It should be
       non-singular though.
       @param m Homogenous transformation matrix in OpenGL format where the
         translation vector appears in the fourth row.
   */
   void setMatrix(const sglMat4<T> &m);

   /** Get the current transform matrix.
       @return Current matrix.
   */
   const sglMat4<T> &getMatrix() const { return m_matrix; }

   // See sglNode::intersect(f) for details.
   virtual sglBound::IntersectResultEnum intersect(
      sglIntersectf &isector) const;

   // See sglNode::intersect(d) for details.
   virtual sglBound::IntersectResultEnum intersect(
      sglIntersectd &isector) const;

   // See sglNode::pick(f) for details.
   virtual void pick(sglPickf &pick_state, unsigned int cull_flags) const;

   // See sglNode::pick(d) for details.
   virtual void pick(sglPickd &pick_state, unsigned int cull_flags) const;

   // See sglNode::computeTxToParent() for details.
   virtual void computeTxToParent(sglMat4f &curr_mat) const;

   // See sglNode::computeTxToParent() for details.
   virtual void computeTxToParent(sglMat4d &curr_mat) const;

   // See sglNode::clone() for details.
   virtual sglNode *clone(unsigned int mode) const;

   // See sglObject::printInfo() for details.
   virtual void printInfo(ostream &ostrm, const char *indent_string) const;

protected:
   // compute the bounding sphere
   virtual void computeBound();

   // See sglNode::cull(f) for details
   virtual void cull(sglCull<float> &trav_state,
                     unsigned int cull_flags) const;

   // See sglNode::cull(d) for details
   virtual void cull(sglCull<double> &trav_state,
                     unsigned int cull_flags) const;

   void checkForIdentity();
   void checkForDeformation();

   void copyTo(sglTransform *dst, unsigned int mode) const;

private:  // not implemented
   sglTransform(const sglTransform &);
   sglTransform& operator=(const sglTransform &);

protected:
   sglMat4<T> m_matrix;
   sglMat4<T> m_inv_mat;
   bool       m_is_ident;
   bool       m_rescale;
   bool       m_normalize;
};

//----------------------------------------------------------------------------
// convenience types:
//    template class 'should' only be float or double

///
typedef sglTransform<float>  sglTransformf;
///
typedef sglTransform<double> sglTransformd;

#endif
