/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglTranslate.cpp
 *  Created: 14 November 1998
 *  Summary: (only) translate the model view matrix before traversing children
 *****************************************************************************/

#include <sglTranslate.hpp>
#include <sglCull.hpp>

//----------------------------------------------------------------------------
template <class T>
void sglTranslate<T>::setTranslation(T x, T y, T z)
{
   m_no_translation_flag = false;

   if (x == (T)0 && y == (T)0 && z == (T)0)
      m_no_translation_flag = true;

   m_translation.set(x, y, z);
   dirty(eBOUND);
}

//----------------------------------------------------------------------------
template <class T>
void sglTranslate<T>::getTranslation(T &x, T &y, T &z) const
{
   x = m_translation[0];
   y = m_translation[1];
   z = m_translation[2];
}

//----------------------------------------------------------------------------
template <class T>
void sglTranslate<T>::computeBound()
{
   sglGroup::computeBound();
   if (!m_sphere.isEmpty() && !m_sphere.isInfinite())
   {
      sglVec3d center = m_sphere.getCenter();
      center += m_translation;

      m_sphere.setCenter(center);
   }
}

//----------------------------------------------------------------------------
template <class T>
sglBound::IntersectResultEnum sglTranslate<T>::intersect(
   sglIntersectf &segset) const
{
   if (m_no_translation_flag)
   {
      return sglGroup::intersect(segset);
   }

   if (m_child_list.empty())
   {
      return sglBound::eOUTSIDE;
   }

   sglSegmentf seg(segset.getSegment()); // copy the current segment

   // test against the current bounding sphere
   sglBound::IntersectResultEnum result = m_sphere.intersect(seg);
   if (result == sglBound::eOUTSIDE)
   {
      return result;
   }

   // Push a translate onto the stack.
   sglMat4f old_matrix;
   sglFrustumf old_view_frustum;
   segset.pushTranslate(m_translation[0], m_translation[1], m_translation[2],
                        seg, old_matrix, old_view_frustum);

   // traverse the children
   result = sglBound::eOUTSIDE;
   for (vector<sglNode*>::const_iterator elt = m_child_list.begin();
        elt != m_child_list.end(); ++elt)
   {
      if ((*elt)->intersect(segset) == sglBound::eINTERSECT)
      {
         result = sglBound::eINTERSECT;
      }
   }

   // restore the original pos (leave len and dir intact)
   seg.setLen(segset.getSegment().getLen());

   // pop the translate off the stack
   segset.pop(seg, old_matrix, old_view_frustum);

   if (result == sglBound::eINTERSECT && segset.getMode() & sglProbe::ePATH)
   {
      segset.getPath().push_front((sglNode*)this);
   }

   return result;
}

//----------------------------------------------------------------------------
template <class T>
sglBound::IntersectResultEnum sglTranslate<T>::intersect(
   sglIntersectd &segset) const
{
   if (m_no_translation_flag)
   {
      return sglGroup::intersect(segset);
   }

   if (m_child_list.empty())
   {
      return sglBound::eOUTSIDE;
   }

   sglSegmentd seg(segset.getSegment()); // copy the current segment

   // test against the current bounding sphere
   sglBound::IntersectResultEnum result = m_sphere.intersect(seg);
   if (result == sglBound::eOUTSIDE)
   {
      return result;
   }

   // Push a translate onto the stack.
   sglMat4d old_matrix;
   sglFrustumd old_view_frustum;
   segset.pushTranslate(m_translation[0], m_translation[1], m_translation[2],
                        seg, old_matrix, old_view_frustum);

   // traverse the children
   result = sglBound::eOUTSIDE;
   for (vector<sglNode*>::const_iterator elt = m_child_list.begin();
        elt != m_child_list.end(); ++elt)
   {
      if ((*elt)->intersect(segset) == sglBound::eINTERSECT)
      {
         result = sglBound::eINTERSECT;
      }
   }

   // restore the original pos (leave len and dir intact)
   seg.setLen(segset.getSegment().getLen());

   // pop the translate off the stack
   segset.pop(seg, old_matrix, old_view_frustum);

   if (result == sglBound::eINTERSECT && segset.getMode() & sglProbe::ePATH)
   {
      segset.getPath().push_front((sglNode*)this);
   }

   return result;
}

//----------------------------------------------------------------------------
template <class T>
void sglTranslate<T>::pick(sglPickf &pick_state, unsigned int cull_flags) const
{
   if (m_no_translation_flag)
   {
      sglGroup::pick(pick_state, cull_flags);
   }
   else if (m_child_list.empty() ||
            (cull_flags &&
             (m_sphere.intersect(pick_state.getPickFrustum(),
                                 cull_flags) == sglBound::eOUTSIDE)))
   {
      return; // select nothing
   }
   else
   {
      sglMat4f old_matrix;
      sglFrustumf old_frustum, old_view_frustum;
      pick_state.pushTranslate(m_translation[0],
                               m_translation[1],
                               m_translation[2],
                               old_frustum, old_matrix, old_view_frustum);

      // if path is requested store it ahead of time (different from
      // intersect), in case something beneath here is picked successfully.
      if (pick_state.getMode() & sglProbe::ePATH)
      {
         pick_state.getPath().push_back((sglNode*)this);
      }

      // --------------------------------------------------

      for (vector<sglNode*>::const_iterator elt = m_child_list.begin();
           elt != m_child_list.end(); ++elt)
      {
         (*elt)->pick(pick_state, cull_flags);
      }

      // --------------------------------------------------
      // restore old state
      pick_state.pop(old_frustum, old_matrix, old_view_frustum);

      if (pick_state.getMode() & sglProbe::ePATH)
      {
         pick_state.getPath().pop_back();
      }
   }
}

//----------------------------------------------------------------------------
template <class T>
void sglTranslate<T>::pick(sglPickd &pick_state, unsigned int cull_flags) const
{
   if (m_no_translation_flag)
   {
      sglGroup::pick(pick_state, cull_flags);
   }
   else if (m_child_list.empty() ||
            (cull_flags &&
             (m_sphere.intersect(pick_state.getPickFrustum(),
                                 cull_flags) == sglBound::eOUTSIDE)))
   {
      return; // select nothing
   }
   else
   {
      sglMat4d old_matrix;
      sglFrustumd old_frustum, old_view_frustum;
      pick_state.pushTranslate(m_translation[0],
                               m_translation[1],
                               m_translation[2],
                               old_frustum, old_matrix, old_view_frustum);

      // if path is requested store it ahead of time (different from
      // intersect), in case something beneath here is picked successfully.
      if (pick_state.getMode() & sglProbe::ePATH)
      {
         pick_state.getPath().push_back((sglNode*)this);
      }

      // --------------------------------------------------

      for (vector<sglNode*>::const_iterator elt = m_child_list.begin();
           elt != m_child_list.end(); ++elt)
      {
         (*elt)->pick(pick_state, cull_flags);
      }

      // --------------------------------------------------
      // restore old state
      pick_state.pop(old_frustum, old_matrix, old_view_frustum);

      if (pick_state.getMode() & sglProbe::ePATH)
      {
         pick_state.getPath().pop_back();
      }
   }
}

//----------------------------------------------------------------------------
template <class T>
void sglTranslate<T>::cull(sglCull<float> &trav_state,
                           unsigned int cull_flags) const
{
   if (m_no_translation_flag)
   {
      sglGroup::cull(trav_state, cull_flags);
   }
   else if (m_child_list.empty() ||
            (cull_flags &&
             (m_sphere.intersect(trav_state.getFrustum(),
                                 cull_flags) == sglBound::eOUTSIDE)))
   {
      return; // draw nothing
   }
   else
   {
      // compute a new matrix
      sglMat4f new_mat;
      new_mat.translate(trav_state.getModelViewMatrix(),
                        m_translation[0], m_translation[1], m_translation[2]);
      trav_state.pushModelViewMatrix(new_mat, false, false);

   // compute a new frustum
      sglFrustumf old_frustum = trav_state.getFrustum();
      trav_state.getFrustum().translate(old_frustum,
                                        -m_translation[0],
                                        -m_translation[1],
                                        -m_translation[2]);

      // --------------------------------------------------

      for (vector<sglNode*>::const_iterator elt = m_child_list.begin();
           elt != m_child_list.end(); ++elt)
      {
         (*elt)->cull(trav_state, cull_flags);
      }

      // --------------------------------------------------
      // restore old state
      trav_state.popModelViewMatrix();
      trav_state.getFrustum() = old_frustum;
   }
}

//----------------------------------------------------------------------------
template <class T>
void sglTranslate<T>::cull(sglCull<double> &trav_state,
                           unsigned int cull_flags) const
{
   if (m_no_translation_flag)
   {
      sglGroup::cull(trav_state, cull_flags);
      return;
   }
   else if (m_child_list.empty() ||
            (cull_flags &&
             (m_sphere.intersect(trav_state.getFrustum(),
                                 cull_flags) == sglBound::eOUTSIDE)))
   {
      return; // draw nothing
   }
   else
   {
      // compute a new matrix
      sglMat4d new_mat;
      new_mat.translate(trav_state.getModelViewMatrix(),
                        m_translation[0], m_translation[1], m_translation[2]);
      trav_state.pushModelViewMatrix(new_mat, false, false);

   // compute a new frustum
      sglFrustumd old_frustum = trav_state.getFrustum();
      trav_state.getFrustum().translate(old_frustum,
                                        -m_translation[0],
                                        -m_translation[1],
                                        -m_translation[2]);

      // --------------------------------------------------

      for (vector<sglNode*>::const_iterator elt = m_child_list.begin();
           elt != m_child_list.end(); ++elt)
      {
         (*elt)->cull(trav_state, cull_flags);
      }

      // --------------------------------------------------
      // restore old state
      trav_state.popModelViewMatrix();
      trav_state.getFrustum() = old_frustum;
   }
}

//----------------------------------------------------------------------------
template <class T>
void sglTranslate<T>::computeTxToParent(sglMat4f &curr_mat) const
{
   curr_mat.translate(curr_mat,
                      (float)-m_translation[0],
                      (float)-m_translation[1],
                      (float)-m_translation[2]);
}

//----------------------------------------------------------------------------
template <class T>
void sglTranslate<T>::computeTxToParent(sglMat4d &curr_mat) const
{
   curr_mat.translate(curr_mat,
                      (double)-m_translation[0],
                      (double)-m_translation[1],
                      (double)-m_translation[2]);
}

//----------------------------------------------------------------------------
template <class T>
sglNode *sglTranslate<T>::clone(unsigned int mode) const
{
   sglTranslate<T> *root = new sglTranslate<T>;
   sglTranslate<T>::copyTo(root, mode);
   return root;
}

//----------------------------------------------------------------------------
template <class T>
void sglTranslate<T>::copyTo(sglTranslate<T> *dst, unsigned int mode) const
{
   sglGroup::copyTo(dst, mode);

   // transfer stuff
   dst->setTranslation(m_translation[0], m_translation[1], m_translation[2]);
}

//----------------------------------------------------------------------------
template <class T>
void sglTranslate<T>::printInfo(ostream &ostrm,
                                       const char *indent_string) const
{
    sglGroup::printInfo(ostrm, indent_string);

    ostrm << indent_string << "sglTranslate: " << endl;
    ostrm << indent_string << "\t" << m_translation[0] << ' '
          << m_translation[1] << ' ' << m_translation[2] << endl;
}


template class SGL_DLL_API sglTranslate<double>;
template class SGL_DLL_API sglTranslate<float>;
