/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglTranslate.hpp
 *  Created: 14 November 1998
 *  Summary: (only) translate the model view matrix before traversing children
 *****************************************************************************/

#ifndef __SGL_TRANSLATE_HPP
#define __SGL_TRANSLATE_HPP

#include <sgl.h>
#include <sglMatrix.hpp>
#include <sglGroup.hpp>

//----------------------------------------------------------------------------

/**

Like the sglTransform node, the sglTranslate node also modifies the model view
matrix.  When only a change in position is needed, it is more efficient to use
an sglTranslate node instead of a more complex sglTransform node.  As with the
sglTransform nodes this can also be instantiated as single or double precision:
sglTranslated or sglTranslatef.

<b>Example:</b>

<pre>
   sglTranslated *dp_translate_node = new sglTranslated;
   dp_translate_node->setTranslation(0.0, 0.0, 25.0000001);
   dp_translate_node->addChild(...);
</pre>

 */

template <class T>
class SGL_DLL_API sglTranslate : public sglGroup
{
public:
   /// default constructor initializes to no translation.
   sglTranslate()
         : sglGroup(), m_no_translation_flag(true), m_translation(0.,0.,0.) {}

   /// virtual destructor
   virtual ~sglTranslate() {}


   /** Set the translation part of the transform (affects the 4th row of the
       matrix only).
       @param x Translation along the x axis.
       @param y Translation along the y axis.
       @param z Translation along the z axis.
   */
   void setTranslation(T x, T y, T z);

   /** Get the current translation
       @param x Translation along the x axis.
       @param y Translation along the y axis.
       @param z Translation along the z axis.
   */
   void getTranslation(T &x, T &y, T &z) const;

   // See sglNode::intersect(f) for details.
   virtual sglBound::IntersectResultEnum intersect(
      sglIntersectf &isector) const;

   // See sglNode::intersect(d) for details.
   virtual sglBound::IntersectResultEnum intersect(
      sglIntersectd &isector) const;

   // See sglNode::pick(f) for details.
   virtual void pick(sglPickf &pick_state, unsigned int cull_flags) const;

   // See sglNode::pick(d) for details.
   virtual void pick(sglPickd &pick_state, unsigned int cull_flags) const;

   // See sglNode::computeTxToParent(f) for details.
   virtual void computeTxToParent(sglMat4f &curr_mat) const;

   // See sglNode::computeTxToParent(d) for details.
   virtual void computeTxToParent(sglMat4d &curr_mat) const;

   // See sglNode::clone() for details.
   virtual sglNode *clone(unsigned int mode) const;

   // See sglObject::printInfo() for details.
   virtual void printInfo(ostream &ostrm, const char *indent_string) const;

protected:
   // compute the bounding sphere
   virtual void computeBound();

   // See sglNode::cull(f) for details
   virtual void cull(sglCull<float> &trav_state,
                     unsigned int cull_flags) const;

   // See sglNode::cull(d) for details
   virtual void cull(sglCull<double> &trav_state,
                     unsigned int cull_flags) const;

   void copyTo(sglTranslate *dst, unsigned int mode) const;

private:   // not implemented
   sglTranslate(const sglTranslate &);
   sglTranslate& operator=(const sglTranslate &);

protected:
   bool       m_no_translation_flag;
   sglVec3<T> m_translation;
};

//----------------------------------------------------------------------------
// convenience types:
//    template class 'should' only be float or double

template <class T> class sglTranslate;

///
typedef sglTranslate<float>  sglTranslatef;
///
typedef sglTranslate<double> sglTranslated;

#endif
