/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglUnProject.cpp
 *   Author: Tom Stimson
 *  Created: 1 October 2000
 *  Summary: Translates & rotates to the near clip plane.
 *****************************************************************************/

#include <sgl.h>
#include <sglUnProject.hpp>
#include <sglCull.hpp>

//----------------------------------------------------------------------------
sglUnProject::sglUnProject(bool fixedPixelScale)
   : sglGroup(),
     m_pixel_scale(fixedPixelScale)
{
}

//----------------------------------------------------------------------------
sglUnProject::~sglUnProject()
{
}

//----------------------------------------------------------------------------
sglBound::IntersectResultEnum sglUnProject::intersect(sglIntersectf &) const
{
   // There is not enough information in the segset to compute
   // the intersection (eg near clip distance).
   return sglBound::eOUTSIDE;
}

//----------------------------------------------------------------------------
sglBound::IntersectResultEnum sglUnProject::intersect(sglIntersectd &) const
{
   // There is not enough information in the segset to compute
   // the intersection (eg near clip distance).
   return sglBound::eOUTSIDE;
}

//----------------------------------------------------------------------------
void sglUnProject::pick(sglPickf &, unsigned int) const
{
}

//----------------------------------------------------------------------------
void sglUnProject::pick(sglPickd &, unsigned int) const
{
}

//----------------------------------------------------------------------------
inline void sglUnProject::cull(sglCullf &trav_state,
                               unsigned int cull_flags) const
{
   // compute a new matrix
   sglMat4f new_mat;
   const sglMat4f &modelview = trav_state.getModelViewMatrix();

   // keep size constant regardless of the fov
   float dot = sglVec3f::dot(
      trav_state.getOrigFrustum().getPlane(
         sglFrustum<float>::eUP).getNormal(),
      trav_state.getOrigFrustum().getPlane(
         sglFrustum<float>::eDOWN).getNormal());
   float near_clip = trav_state.getOrigFrustum().getPlane(
      sglFrustum<float>::eFRONT).getDistance();
   double scale = 1.0;

   new_mat[3][0] = modelview[3][0];
   new_mat[3][1] = modelview[3][1];

   // Check if the frustum is ortho or perspective.
   // Note: the dot variable is too impercise to compare with sglFltEpsilon.
   if (SGL_ALMOST_EQUAL(dot, -1.0f, 0.0001f))
   {
      scale =
         trav_state.getOrigFrustum().getPlane(
            sglFrustum<float>::eUP).getDistance() +
         trav_state.getOrigFrustum().getPlane(
            sglFrustum<float>::eDOWN).getDistance();
   }
   else
   {
      double fov_y = M_PI - acos(dot);
      scale = tan(fov_y*0.5) * 2.0;
      if (modelview[3][2])
      {
         float val = near_clip/modelview[3][2];
         new_mat[3][0] *= val;
         new_mat[3][1] *= val;
      }
   }

   if (m_pixel_scale)
   {
      unsigned int viewport_x, viewport_y;
      trav_state.getViewportSize(viewport_x, viewport_y);
      scale /= viewport_y;
   }

   double factor[3];
   for (int iz=0; iz<3; iz++)
   {
      factor[iz] = sqrt(modelview[iz][0]*modelview[iz][0] +
                        modelview[iz][1]*modelview[iz][1] +
                        modelview[iz][2]*modelview[iz][2]) * scale;
   }
   for (int ix=0; ix<3; ix++)
   {
      for (int iy=0; iy<3; iy++)
         new_mat[ix][iy] = 0.0;
      new_mat[ix][ix] = factor[ix];
      new_mat[ix][3] = modelview[ix][3];
   }
   new_mat[3][2] = near_clip - sglFltEpsilon;
   new_mat[3][3] = modelview[3][3];

   // compute a new frustum
   sglFrustumf old_frustum = trav_state.getFrustum();
   sglMat4f inv_mat;
   inv_mat.inverseAffine(new_mat);
   trav_state.getFrustum().mul(trav_state.getOrigFrustum(), inv_mat);

   if (cull_flags && (m_sphere.intersect(trav_state.getFrustum(),
                                         cull_flags) == sglBound::eOUTSIDE))
   {
      trav_state.getFrustum() = old_frustum;
      return; // draw nothing
   }
   trav_state.pushModelViewMatrix(new_mat, false, false);

   // --------------------------------------------------

   for (vector<sglNode*>::const_iterator elt = m_child_list.begin();
        elt != m_child_list.end(); ++elt)
   {
      (*elt)->cull(trav_state, cull_flags);
   }

   // --------------------------------------------------
   // restore old state
   trav_state.popModelViewMatrix();
   trav_state.getFrustum() = old_frustum;
}

//----------------------------------------------------------------------------
inline void sglUnProject::cull(sglCulld &trav_state,
                               unsigned int cull_flags) const
{
   // compute a new matrix
   sglMat4d new_mat;
   const sglMat4d &modelview = trav_state.getModelViewMatrix();

   // keep size constant regardless of the fov
   double dot = sglVec3d::dot(
      trav_state.getOrigFrustum().getPlane(
         sglFrustum<double>::eUP).getNormal(),
      trav_state.getOrigFrustum().getPlane(
         sglFrustum<double>::eDOWN).getNormal());
   double near_clip = trav_state.getOrigFrustum().getPlane(
      sglFrustum<double>::eFRONT).getDistance();
   double scale = 1.0;

   new_mat[3][0] = modelview[3][0];
   new_mat[3][1] = modelview[3][1];

   // Check if the frustum is ortho or perspective.
   // Note: the dot variable is too impercise to compare with sglFltEpsilon.
   if (SGL_ALMOST_EQUAL(dot, -1.0, 0.0001))
   {
      scale =
         trav_state.getOrigFrustum().getPlane(
            sglFrustum<double>::eUP).getDistance() +
         trav_state.getOrigFrustum().getPlane(
            sglFrustum<double>::eDOWN).getDistance();
   }
   else
   {
      double fov_y = M_PI - acos(dot);
      scale = tan(fov_y*0.5) * 2.0;
      if (modelview[3][2])
      {
         double val = near_clip/modelview[3][2];
         new_mat[3][0] *= val;
         new_mat[3][1] *= val;
      }
   }

   if (m_pixel_scale)
   {
      unsigned int viewport_x, viewport_y;
      trav_state.getViewportSize(viewport_x, viewport_y);
      scale /= viewport_y;
   }

   double factor[3];
   for (int iz=0; iz<3; iz++)
   {
      factor[iz] = sqrt(modelview[iz][0]*modelview[iz][0] +
                        modelview[iz][1]*modelview[iz][1] +
                        modelview[iz][2]*modelview[iz][2]) * scale;
   }
   for (int ix=0; ix<3; ix++)
   {
      for (int iy=0; iy<3; iy++)
         new_mat[ix][iy] = 0.0;
      new_mat[ix][ix] = factor[ix];
      new_mat[ix][3] = modelview[ix][3];
   }
   new_mat[3][2] = near_clip - sglFltEpsilon;
   new_mat[3][3] = modelview[3][3];

   // compute a new frustum
   sglFrustumd old_frustum = trav_state.getFrustum();
   sglMat4d inv_mat;
   inv_mat.inverseAffine(new_mat);
   trav_state.getFrustum().mul(trav_state.getOrigFrustum(), inv_mat);

   if (cull_flags && (m_sphere.intersect(trav_state.getFrustum(),
                                         cull_flags) == sglBound::eOUTSIDE))
   {
      trav_state.getFrustum() = old_frustum;
      return; // draw nothing
   }
   trav_state.pushModelViewMatrix(new_mat, false, false);

   // --------------------------------------------------

   for (vector<sglNode*>::const_iterator elt = m_child_list.begin();
        elt != m_child_list.end(); ++elt)
   {
      (*elt)->cull(trav_state, cull_flags);
   }

   // --------------------------------------------------
   // restore old state
   trav_state.popModelViewMatrix();
   trav_state.getFrustum() = old_frustum;
}

//----------------------------------------------------------------------------
sglNode *sglUnProject::clone(unsigned int mode) const
{
   sglUnProject *root = new sglUnProject;
   sglUnProject::copyTo(root, mode);
   return root;
}

//----------------------------------------------------------------------------
void sglUnProject::copyTo(sglUnProject *dst, unsigned int mode) const
{
   sglGroup::copyTo(dst, mode);

   // transfer stuff
   dst->setFixedPixelScale(m_pixel_scale);
}

//----------------------------------------------------------------------------
void sglUnProject::printInfo(ostream &ostrm, const char *indent_string) const
{
   sglGroup::printInfo(ostrm, indent_string);
   ostrm << indent_string << "fixed pixel flag = "
         << (m_pixel_scale ? "true" : "false") << endl;
}
