/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglUnProject.hpp
 *   Author: Tom Stimson
 *  Created: 1 October 2000
 *  Summary: Translates & rotates to the near clip plane.
 *****************************************************************************/

#ifndef __SGL_UNPROJECT_HPP
#define __SGL_UNPROJECT_HPP

#include <sgl.h>
#include <sglGroup.hpp>

/**

This node modifies the model view matrix maintained in the various traversals
such that the child of this node are transformed to appear on the near clip
plane.  For example, this node could be used to create target indicators on a
game HUD (heads up display).

There are two modes for this node:  with and without fixed pixel scale:

-# with fixed pixel scale, the following coordinate system results for the
   children of this node:
   <pre>
      x - left to right, a unit length of one = one pixel
      y - bottom to top, a unit length of one = one pixel
      z - 0 (on the near clip plane) ... (negatives extending into the scene)
   </pre>
   <p>
-# without fixed pixel scale:
   <pre>
      x - left to right, viewport width = 1.0 * aspect ratio \\
      y - bottom to top, viewport height = 1.0 \\
      z - 0 (on the near clip plane) ... (negatives extending into the scene)
   </pre>

Note: Due to the way culling is performed (with pre-computed bounding volumes),
this node may exhibit non-intuitive culling behavior.  The dynamically generate
bounding-volume for this node will reflect the size of its children in
world-space, not the transformed screen space.  Thus, by default, the bounding
volume is likely to be much smaller than the drawn object.  It is recommended
that you set a static bound of appropriate size for this sub-tree.

<b>Example:</b>

<pre>
   sglGeode *geode = new sglGeode;
   sglUnProject *unproject_node = new sglUnProject(true); // fixed pixel scale
   unproject_node->addChild(geode);
</pre>

\todo Implement pick and intersect functions.

  */

class SGL_DLL_API sglUnProject : public sglGroup
{
public:
   /** Constructor
       @param fixedPixelScale boolean flag that sets whether fixed pixel scale
         mode is to be used or not (defaults to false).
   */
   sglUnProject(bool fixedPixelScale = false);

   /// virtual destructor
   virtual ~sglUnProject();

   /** Set whether geometry is scaled in units of pixels.
       @param on If true then fixed pixel scale is to be used.
   */
   void setFixedPixelScale(bool on) { m_pixel_scale = on; }

   /** Query for fixed pixel scale mode.
       @return true if fixed pixel scale mode is being used; otherwise, false.
   */
   bool getFixedPixelScale() const { return m_pixel_scale; }

   // See sglNode::intersect(f) for details.
   virtual sglBound::IntersectResultEnum intersect(
      sglIntersectf &isector) const;

   // See sglNode::intersect(d) for details.
   virtual sglBound::IntersectResultEnum intersect(
      sglIntersectd &isector) const;

   // See sglNode::pick(f) for details.
   virtual void pick(sglPickf &pick_state, unsigned int cull_flags) const;

   // See sglNode::pick(d) for details.
   virtual void pick(sglPickd &pick_state, unsigned int cull_flags) const;

   // See sglNode::clone() for details.
   virtual sglNode *clone(unsigned int mode) const;

   // See sglObject::printInfo() for details.
   virtual void printInfo(ostream &ostrm, const char *indent_string) const;

protected:
   // See sglNode::cull(f) for details
   virtual void cull(sglCull<float> &trav_state,
                     unsigned int cull_flags) const;

   // See sglNode::cull(d) for details
   virtual void cull(sglCull<double> &trav_state,
                     unsigned int cull_flags) const;

   void copyTo(sglUnProject *dst, unsigned int mode) const;

private:   // not implemented
   sglUnProject(const sglUnProject &);
   sglUnProject &operator=(const sglUnProject &);

private:
   bool m_pixel_scale;
};

#endif
