/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglUnScale.cpp
 *   Author: Tom Stimson
 *  Created: 1 October 2000
 *  Summary: Dynamically scales to maintain a constant visible size.
 *****************************************************************************/

#include <sgl.h>
#include <sglUnScale.hpp>
#include <sglCull.hpp>

//----------------------------------------------------------------------------
sglUnScale::sglUnScale(bool fixedPixelScale)
   : sglGroup(),
     m_pixel_scale(fixedPixelScale)
{
}

//----------------------------------------------------------------------------
sglUnScale::~sglUnScale()
{
}

//----------------------------------------------------------------------------
sglBound::IntersectResultEnum sglUnScale::intersect(
   sglIntersectf &segset) const
{
   // Segments will most likely fail to intersect with the bounding
   // volumes of ancestor nodes before this method is called unless the a
   // static infinite bounding sphere is defined for this node.
   if (!m_child_list.empty() && segset.getMode()&sglProbe::eUNSCALE_VIEW)
   {
      sglSegmentf seg(segset.getSegment()); // copy the current segment

      // testing against the current bounding sphere
      sglBound::IntersectResultEnum result = m_sphere.intersect(seg);
      if (result == sglBound::eOUTSIDE)
      {
         return result;
      }

      float scale = computeScale(segset);

      // Push a scale onto the stack.
      sglMat4f old_matrix;
      sglFrustumf old_view_frustum;
      segset.pushScale(scale, 1.f/scale, seg, old_matrix, old_view_frustum);

      // traverse the children
      result = sglBound::eOUTSIDE;
      for (vector<sglNode*>::const_iterator elt = m_child_list.begin();
           elt != m_child_list.end(); ++elt)
      {
         if ((*elt)->intersect(segset) == sglBound::eINTERSECT)
         {
            result = sglBound::eINTERSECT;
         }
      }

      // restore the original pos (leave len and dir intact) and unscale
      seg.setLen(segset.getSegment().getLen() * scale);

      // pop the scale off the stack
      segset.pop(seg, old_matrix, old_view_frustum);

      if (result == sglBound::eINTERSECT && segset.getMode() & sglProbe::ePATH)
      {
         segset.getPath().push_front((sglNode*)this);
      }

      return result;
   }
   else
   {
      return sglBound::eOUTSIDE;
   }
}

//----------------------------------------------------------------------------
sglBound::IntersectResultEnum sglUnScale::intersect(
   sglIntersectd &segset) const
{
   // Segments will most likely fail to intersect with the bounding
   // volumes of ancestor nodes before this method is called unless the a
   // static infinite bounding sphere is defined for this node.
   if (!m_child_list.empty() && segset.getMode()&sglProbe::eUNSCALE_VIEW)
   {
      sglSegmentd seg(segset.getSegment()); // copy the current segment

      // testing against the current bounding sphere
      sglBound::IntersectResultEnum result = m_sphere.intersect(seg);
      if (result == sglBound::eOUTSIDE)
      {
         return result;
      }

      double scale = computeScale(segset);

      // Push a scale onto the stack.
      sglMat4d old_matrix;
      sglFrustumd old_view_frustum;
      segset.pushScale(scale, 1./scale, seg, old_matrix, old_view_frustum);

      // traverse the children
      result = sglBound::eOUTSIDE;
      for (vector<sglNode*>::const_iterator elt = m_child_list.begin();
           elt != m_child_list.end(); ++elt)
      {
         if ((*elt)->intersect(segset) == sglBound::eINTERSECT)
         {
            result = sglBound::eINTERSECT;
         }
      }

      // restore the original pos (leave len and dir intact) and unscale
      seg.setLen(segset.getSegment().getLen() * scale);

      // pop the scale off the stack
      segset.pop(seg, old_matrix, old_view_frustum);

      if (result == sglBound::eINTERSECT &&
          segset.getMode() & sglProbe::ePATH)
      {
         segset.getPath().push_front((sglNode*)this);
      }

      return result;
   }
   else
   {
      return sglBound::eOUTSIDE;
   }
}

//----------------------------------------------------------------------------
void sglUnScale::pick(sglPickf &pick_state, unsigned int cull_flags) const
{
   if (!(pick_state.getMode()&sglProbe::eUNSCALE_VIEW) ||
       m_child_list.empty() ||
       (cull_flags &&
        (m_sphere.intersect(pick_state.getPickFrustum(),
                            cull_flags) == sglBound::eOUTSIDE)))
   {
      return; // select nothing
   }
   else
   {
      float scale = computeScale(pick_state);
      sglMat4f old_matrix;
      sglFrustumf old_frustum, old_view_frustum;
      pick_state.pushScale(scale, 1.f/scale,
                           old_frustum, old_matrix, old_view_frustum);

      // if path is requested store it ahead of time (different from
      // intersect), in case something beneath here is picked successfully.
      if (pick_state.getMode() & sglProbe::ePATH)
      {
         pick_state.getPath().push_back((sglNode*)this);
      }

      // --------------------------------------------------

      for (vector<sglNode*>::const_iterator elt = m_child_list.begin();
           elt != m_child_list.end(); ++elt)
      {
         (*elt)->pick(pick_state, cull_flags);
      }

      // --------------------------------------------------
      // restore old state
      pick_state.pop(old_frustum, old_matrix, old_view_frustum);

      if (pick_state.getMode() & sglProbe::ePATH)
      {
         pick_state.getPath().pop_back();
      }
   }
}

//----------------------------------------------------------------------------
void sglUnScale::pick(sglPickd &pick_state, unsigned int cull_flags) const
{
   if (!(pick_state.getMode()&sglProbe::eUNSCALE_VIEW) ||
       m_child_list.empty() ||
       (cull_flags &&
        (m_sphere.intersect(pick_state.getPickFrustum(),
                            cull_flags) == sglBound::eOUTSIDE)))
   {
      return; // select nothing
   }
   else
   {
      double scale = computeScale(pick_state);
      sglMat4d old_matrix;
      sglFrustumd old_frustum, old_view_frustum;
      pick_state.pushScale(scale, 1./scale,
                           old_frustum, old_matrix, old_view_frustum);

      // if path is requested store it ahead of time (different from
      // intersect), in case something beneath here is picked successfully.
      if (pick_state.getMode() & sglProbe::ePATH)
      {
         pick_state.getPath().push_back((sglNode*)this);
      }

      // --------------------------------------------------

      for (vector<sglNode*>::const_iterator elt = m_child_list.begin();
           elt != m_child_list.end(); ++elt)
      {
         (*elt)->pick(pick_state, cull_flags);
      }

      // --------------------------------------------------
      // restore old state
      pick_state.pop(old_frustum, old_matrix, old_view_frustum);

      if (pick_state.getMode() & sglProbe::ePATH)
      {
         pick_state.getPath().pop_back();
      }
   }
}

//----------------------------------------------------------------------------
inline void sglUnScale::cull(sglCullf &trav_state,
                             unsigned int cull_flags) const
{
   float scale = computeScale(trav_state);

   // compute a new matrix
   sglMat4f new_mat;
   new_mat.scale(trav_state.getModelViewMatrix(), scale, scale, scale);

   // compute a new frustum
   sglFrustumf old_frustum = trav_state.getFrustum();
   trav_state.getFrustum().scale(old_frustum, 1.0f/scale);

   if (cull_flags && (m_sphere.intersect(trav_state.getFrustum(),
                                         cull_flags) == sglBound::eOUTSIDE))
   {
      trav_state.getFrustum() = old_frustum;
      return; // draw nothing
   }
   trav_state.pushModelViewMatrix(new_mat, true, false);

   // --------------------------------------------------

   for (vector<sglNode*>::const_iterator elt = m_child_list.begin();
        elt != m_child_list.end(); ++elt)
   {
      (*elt)->cull(trav_state, cull_flags);
   }

   // --------------------------------------------------
   // restore old state
   trav_state.popModelViewMatrix();
   trav_state.getFrustum() = old_frustum;
}

//----------------------------------------------------------------------------
inline void sglUnScale::cull(sglCulld &trav_state,
                             unsigned int cull_flags) const
{
   double scale = computeScale(trav_state);

   // compute a new matrix
   sglMat4d new_mat;
   new_mat.scale(trav_state.getModelViewMatrix(), scale, scale, scale);

   // compute a new frustum
   sglFrustumd old_frustum = trav_state.getFrustum();
   trav_state.getFrustum().scale(old_frustum, 1.0/scale);

   if (cull_flags && (m_sphere.intersect(trav_state.getFrustum(),
                                         cull_flags) == sglBound::eOUTSIDE))
   {
      trav_state.getFrustum() = old_frustum;
      return; // draw nothing
   }
   trav_state.pushModelViewMatrix(new_mat, true, false);

   // --------------------------------------------------

   for (vector<sglNode*>::const_iterator elt = m_child_list.begin();
        elt != m_child_list.end(); ++elt)
   {
      (*elt)->cull(trav_state, cull_flags);
   }

   // --------------------------------------------------
   // restore old state
   trav_state.popModelViewMatrix();
   trav_state.getFrustum() = old_frustum;
}

//----------------------------------------------------------------------------
sglNode *sglUnScale::clone(unsigned int mode) const
{
   sglUnScale *root = new sglUnScale;
   sglUnScale::copyTo(root, mode);
   return root;
}

//----------------------------------------------------------------------------
void sglUnScale::copyTo(sglUnScale *dst, unsigned int mode) const
{
   sglGroup::copyTo(dst, mode);

   // transfer stuff
   dst->setFixedPixelScale(m_pixel_scale);
}

//----------------------------------------------------------------------------
void sglUnScale::printInfo(ostream &ostrm, const char *indent_string) const
{
   sglGroup::printInfo(ostrm, indent_string);
   ostrm << indent_string << "fixed pixel flag = "
         << (m_pixel_scale ? "true" : "false") << endl;
}
