/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglUnScale.hpp
 *   Author: Tom Stimson
 *  Created: 1 October 2000
 *  Summary: Dynamically scales to maintain a constant visible size.
 *****************************************************************************/

#ifndef __SGL_UNSCALE_HPP
#define __SGL_UNSCALE_HPP

#include <sgl.h>
#include <sglGroup.hpp>

/**

This node modifies the traversal's view matrix such that geometry is scaled to
appear the same size no matter how far away it is viewed.  If
getFixedPixelScale() is true, a unit length of one is equal to one pixel;
otherwise, a unit length of one is equal to the height of the viewport.

Note: Due to the way culling is performed (with pre-computed bounding volumes),
this node may exhibit non-intuitive culling behavior.  The dynamically generate
bounding-volume for this node will reflect the size of its children in
world-space, not the transformed screen space.  Thus, by default, the bounding
volume is likely to be much smaller than the drawn object.  It is recommended
that you set a static bound of appropriate size for this node (and its
children).

<b>Example:</b>

<pre>
   sglGeode geode = new sglGeode;
   sglUnScale *unscale = new sglUnScale; // default: fixed pixel scale is false
   unscale->addChild(geode);
</pre>

 */

class SGL_DLL_API sglUnScale : public sglGroup
{
public:
   /** Constructor
       @param fixedPixelScale boolean flag that sets whether fixed pixel scale
         mode is to be used or not (defaults to false).
   */
   sglUnScale(bool fixedPixelScale = false);

   /// virtual destructor
   virtual ~sglUnScale();

   /** Set whether geometry is scaled in units of pixels.
       @param on If true then fixed pixel scale is to be used.
   */
   void setFixedPixelScale(bool on) { m_pixel_scale = on; }

   /** Query the fixed pixel scale mode.
       @return true if fixed pixel scale mode is being used; otherwise, false.
   */
   bool getFixedPixelScale() const { return m_pixel_scale; }

   // See sglNode::intersect(f) for details.
   virtual sglBound::IntersectResultEnum intersect(
      sglIntersectf &isector) const;

   // See sglNode::intersect(d) for details.
   virtual sglBound::IntersectResultEnum intersect(
      sglIntersectd &isector) const;

   // See sglNode::pick(f) for details.
   virtual void pick(sglPickf &pick_state, unsigned int cull_flags) const;

   // See sglNode::pick(d) for details.
   virtual void pick(sglPickd &pick_state, unsigned int cull_flags) const;

   // See sglNode::clone() for details.
   virtual sglNode *clone(unsigned int mode) const;

   // See sglObject::printInfo() for details.
   virtual void printInfo(ostream &ostrm, const char *indent_string) const;

protected:
   // See sglNode::cull(f) for details
   virtual void cull(sglCull<float> &trav_state,
                     unsigned int cull_flags) const;

   // See sglNode::cull(d) for details
   virtual void cull(sglCull<double> &trav_state,
                     unsigned int cull_flags) const;

   void copyTo(sglUnScale *dst, unsigned int mode) const;

private:
   sglUnScale(const sglUnScale &);              // not implemented
   sglUnScale &operator=(const sglUnScale &);   // not implemented

   template <class T>
   T computeScale(sglProbeTemplate<T> &probe) const
        {
            T scale = 1.0f;

            if (probe.getOrigViewFrustum().getOrthoFlag())
            {
                // HACK precompute this in sglProbeTemplate.
                scale =
                    probe.getOrigViewFrustum().getPlane(
                        sglFrustum<T>::eUP).getDistance() +
                    probe.getOrigViewFrustum().getPlane(
                        sglFrustum<T>::eDOWN).getDistance();
            }
            else
            {
                scale = -(probe.getModelViewMatrix())[3][2] *
                    tan((probe.getOrigViewFrustum().getFOVY())*0.5) * 2.0;
            }

            if (m_pixel_scale)
            {
                unsigned int viewport_x, viewport_y;
                probe.getViewportSize(viewport_x, viewport_y);
                scale /= (T)viewport_y;
            }

            return scale;
        }

   template <class T>
   T computeScale(sglCull<T> &trav_state) const
        {
            T scale = 1.0f;

            if (trav_state.getOrigFrustum().getOrthoFlag())
            {
                // HACK precompute this in sglProbeTemplate.
                scale =
                    trav_state.getOrigFrustum().getPlane(
                        sglFrustum<T>::eUP).getDistance() +
                    trav_state.getOrigFrustum().getPlane(
                        sglFrustum<T>::eDOWN).getDistance();
            }
            else
            {
                scale = -(trav_state.getModelViewMatrix())[3][2] *
                    tan((trav_state.getOrigFrustum().getFOVY())*0.5) * 2.0;
            }

            if (m_pixel_scale)
            {
                unsigned int viewport_x, viewport_y;
                trav_state.getViewportSize(viewport_x, viewport_y);
                scale /= (T)viewport_y;
            }

            return scale;
        }

private:
   bool m_pixel_scale;
};

#endif
