/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglUnScaleRange.cpp
 *   Author: Tom Stimson
 *  Created: 1 October 2000
 *  Summary: Dynamically scales to maintain a constant visible size.
 *****************************************************************************/

#include <sgl.h>
#include <sglUnScaleRange.hpp>
#include <sglCull.hpp>

//----------------------------------------------------------------------------
sglUnScaleRange::sglUnScaleRange()
   : sglGroup(),
     m_reference_size(1.f),
     m_min_pixel_size(0),
     m_max_pixel_size(0),
     m_inverse_mode(false)
{
}

//----------------------------------------------------------------------------
sglUnScaleRange::~sglUnScaleRange()
{
}

//----------------------------------------------------------------------------
void sglUnScaleRange::setScaleParameters(float reference_size,
                                         unsigned int min_pixel_size,
                                         unsigned int max_pixel_size)
{
    if (reference_size > 0.f)
        m_reference_size = reference_size;
    else
        m_reference_size = 1.f;

    if (min_pixel_size == 0 || max_pixel_size == 0)
    {
        m_min_pixel_size = min_pixel_size;
        m_max_pixel_size = max_pixel_size;
    }
    else
    {
        if (min_pixel_size > max_pixel_size)
        {
            m_min_pixel_size = max_pixel_size;
            m_max_pixel_size = min_pixel_size;
        }
        else
        {
            m_min_pixel_size = min_pixel_size;
            m_max_pixel_size = max_pixel_size;
        }
    }
}

//----------------------------------------------------------------------------
sglBound::IntersectResultEnum sglUnScaleRange::intersect(
   sglIntersectf &segset) const
{
   if (m_min_pixel_size == 0 && m_max_pixel_size == 0)
      return sglGroup::intersect(segset);

   // Segments will most likely fail to intersect with the bounding
   // volumes of ancestor nodes before this method is called unless the a
   // static infinite bounding sphere is defined for this node.
   if (!m_child_list.empty() && segset.getMode()&sglProbe::eUNSCALE_VIEW)
   {
      sglSegmentf seg(segset.getSegment()); // copy the current segment

      // testing against the current bounding sphere
      sglBound::IntersectResultEnum result = m_sphere.intersect(seg);
      if (result == sglBound::eOUTSIDE)
      {
         return result;
      }

      float scale = computeScale(segset);

      // Push a scale onto the stack.
      sglMat4f old_matrix;
      sglFrustumf old_view_frustum;
      segset.pushScale(scale, 1.f/scale, seg, old_matrix, old_view_frustum);

      // traverse the children
      result = sglBound::eOUTSIDE;
      for (vector<sglNode*>::const_iterator elt = m_child_list.begin();
           elt != m_child_list.end(); ++elt)
      {
         if ((*elt)->intersect(segset) == sglBound::eINTERSECT)
         {
            result = sglBound::eINTERSECT;
         }
      }

      // restore the original pos (leave len and dir intact) and unscale
      seg.setLen(segset.getSegment().getLen() * scale);

      // pop the scale off the stack
      segset.pop(seg, old_matrix, old_view_frustum);

      if (result == sglBound::eINTERSECT && segset.getMode() & sglProbe::ePATH)
      {
         segset.getPath().push_front((sglNode*)this);
      }

      return result;
   }
   else
   {
      return sglBound::eOUTSIDE;
   }
}

//----------------------------------------------------------------------------
sglBound::IntersectResultEnum sglUnScaleRange::intersect(
   sglIntersectd &segset) const
{
   if (m_min_pixel_size == 0 && m_max_pixel_size == 0)
      return sglGroup::intersect(segset);

   // Segments will most likely fail to intersect with the bounding
   // volumes of ancestor nodes before this method is called unless the a
   // static infinite bounding sphere is defined for this node.
   if (!m_child_list.empty() && segset.getMode()&sglProbe::eUNSCALE_VIEW)
   {
      sglSegmentd seg(segset.getSegment()); // copy the current segment

      // testing against the current bounding sphere
      sglBound::IntersectResultEnum result = m_sphere.intersect(seg);
      if (result == sglBound::eOUTSIDE)
      {
         return result;
      }

      double scale = computeScale(segset);

      // Push a scale onto the stack.
      sglMat4d old_matrix;
      sglFrustumd old_view_frustum;
      segset.pushScale(scale, 1./scale, seg, old_matrix, old_view_frustum);

      // traverse the children
      result = sglBound::eOUTSIDE;
      for (vector<sglNode*>::const_iterator elt = m_child_list.begin();
           elt != m_child_list.end(); ++elt)
      {
         if ((*elt)->intersect(segset) == sglBound::eINTERSECT)
         {
            result = sglBound::eINTERSECT;
         }
      }

      // restore the original pos (leave len and dir intact) and unscale
      seg.setLen(segset.getSegment().getLen() * scale);

      // pop the scale off the stack
      segset.pop(seg, old_matrix, old_view_frustum);

      if (result == sglBound::eINTERSECT &&
          segset.getMode() & sglProbe::ePATH)
      {
         segset.getPath().push_front((sglNode*)this);
      }

      return result;
   }
   else
   {
      return sglBound::eOUTSIDE;
   }
}

//----------------------------------------------------------------------------
void sglUnScaleRange::pick(sglPickf &pick_state, unsigned int cull_flags) const
{
   if (m_min_pixel_size == 0 && m_max_pixel_size == 0)
   {
      sglGroup::pick(pick_state, cull_flags);
      return;
   }

   if (!(pick_state.getMode()&sglProbe::eUNSCALE_VIEW) ||
       m_child_list.empty() ||
       (cull_flags &&
        (m_sphere.intersect(pick_state.getPickFrustum(),
                            cull_flags) == sglBound::eOUTSIDE)))
   {
      return; // select nothing
   }
   else
   {
      float scale = computeScale(pick_state);
      sglMat4f old_matrix;
      sglFrustumf old_frustum, old_view_frustum;
      pick_state.pushScale(scale, 1.f/scale,
                           old_frustum, old_matrix, old_view_frustum);

      // if path is requested store it ahead of time (different from
      // intersect), in case something beneath here is picked successfully.
      if (pick_state.getMode() & sglProbe::ePATH)
      {
         pick_state.getPath().push_back((sglNode*)this);
      }

      // --------------------------------------------------

      for (vector<sglNode*>::const_iterator elt = m_child_list.begin();
           elt != m_child_list.end(); ++elt)
      {
         (*elt)->pick(pick_state, cull_flags);
      }

      // --------------------------------------------------
      // restore old state
      pick_state.pop(old_frustum, old_matrix, old_view_frustum);

      if (pick_state.getMode() & sglProbe::ePATH)
      {
         pick_state.getPath().pop_back();
      }
   }
}

//----------------------------------------------------------------------------
void sglUnScaleRange::pick(sglPickd &pick_state, unsigned int cull_flags) const
{
   if (m_min_pixel_size == 0 && m_max_pixel_size == 0)
   {
      sglGroup::pick(pick_state, cull_flags);
      return;
   }

   if (!(pick_state.getMode()&sglProbe::eUNSCALE_VIEW) ||
       m_child_list.empty() ||
       (cull_flags &&
        (m_sphere.intersect(pick_state.getPickFrustum(),
                            cull_flags) == sglBound::eOUTSIDE)))
   {
      return; // select nothing
   }
   else
   {
      double scale = computeScale(pick_state);
      sglMat4d old_matrix;
      sglFrustumd old_frustum, old_view_frustum;
      pick_state.pushScale(scale, 1./scale,
                           old_frustum, old_matrix, old_view_frustum);

      // if path is requested store it ahead of time (different from
      // intersect), in case something beneath here is picked successfully.
      if (pick_state.getMode() & sglProbe::ePATH)
      {
         pick_state.getPath().push_back((sglNode*)this);
      }

      // --------------------------------------------------

      for (vector<sglNode*>::const_iterator elt = m_child_list.begin();
           elt != m_child_list.end(); ++elt)
      {
         (*elt)->pick(pick_state, cull_flags);
      }

      // --------------------------------------------------
      // restore old state
      pick_state.pop(old_frustum, old_matrix, old_view_frustum);

      if (pick_state.getMode() & sglProbe::ePATH)
      {
         pick_state.getPath().pop_back();
      }
   }
}

//----------------------------------------------------------------------------
void sglUnScaleRange::cull(sglCullf &trav_state, unsigned int cull_flags) const
{
   float scale = 1.f;

   if ((m_min_pixel_size == 0 && m_max_pixel_size == 0) ||
       (scale = computeScale(trav_state)) == 1.f)
   {
      sglGroup::cull(trav_state, cull_flags);
      return;
   }

   // compute a new matrix
   sglMat4f new_mat;
   new_mat.scale(trav_state.getModelViewMatrix(), scale, scale, scale);

   // compute a new frustum
   sglFrustumf old_frustum = trav_state.getFrustum();
   trav_state.getFrustum().scale(old_frustum, 1.0f/scale);

   if (cull_flags && (m_sphere.intersect(trav_state.getFrustum(),
                                         cull_flags) == sglBound::eOUTSIDE))
   {
      trav_state.getFrustum() = old_frustum;
      return; // draw nothing
   }
   trav_state.pushModelViewMatrix(new_mat, true, false);

   // --------------------------------------------------

   for (vector<sglNode*>::const_iterator elt = m_child_list.begin();
        elt != m_child_list.end(); ++elt)
   {
      (*elt)->cull(trav_state, cull_flags);
   }

   // --------------------------------------------------
   // restore old state
   trav_state.popModelViewMatrix();
   trav_state.getFrustum() = old_frustum;
}

//----------------------------------------------------------------------------
void sglUnScaleRange::cull(sglCulld &trav_state, unsigned int cull_flags) const
{
   double scale = 1.f;

   if ((m_min_pixel_size == 0 && m_max_pixel_size == 0) ||
       (scale = computeScale(trav_state)) == 1.f)
   {
      sglGroup::cull(trav_state, cull_flags);
      return;
   }

   // compute a new matrix
   sglMat4d new_mat;
   new_mat.scale(trav_state.getModelViewMatrix(), scale, scale, scale);

   // compute a new frustum
   sglFrustumd old_frustum = trav_state.getFrustum();
   trav_state.getFrustum().scale(old_frustum, 1.0/scale);

   if (cull_flags && (m_sphere.intersect(trav_state.getFrustum(),
                                         cull_flags) == sglBound::eOUTSIDE))
   {
      trav_state.getFrustum() = old_frustum;
      return; // draw nothing
   }
   trav_state.pushModelViewMatrix(new_mat, true, false);

   // --------------------------------------------------

   for (vector<sglNode*>::const_iterator elt = m_child_list.begin();
        elt != m_child_list.end(); ++elt)
   {
      (*elt)->cull(trav_state, cull_flags);
   }

   // --------------------------------------------------
   // restore old state
   trav_state.popModelViewMatrix();
   trav_state.getFrustum() = old_frustum;
}

//----------------------------------------------------------------------------
sglNode *sglUnScaleRange::clone(unsigned int mode) const
{
   sglUnScaleRange *root = new sglUnScaleRange;
   sglUnScaleRange::copyTo(root, mode);
   return root;
}

//----------------------------------------------------------------------------
void sglUnScaleRange::copyTo(sglUnScaleRange *dst, unsigned int mode) const
{
   sglGroup::copyTo(dst, mode);

   // transfer stuff
   dst->setScaleParameters(m_reference_size,
                           m_min_pixel_size, m_max_pixel_size);
   dst->setInverseMode(m_inverse_mode);
}

//----------------------------------------------------------------------------
void sglUnScaleRange::printInfo(ostream &ostrm,
                                const char *indent_string) const
{
   sglGroup::printInfo(ostrm, indent_string);
   ostrm << indent_string << "reference size : " << m_reference_size << endl;
   ostrm << indent_string << "min/max pixel size: " << m_min_pixel_size << "/"
         << m_max_pixel_size << endl;
   ostrm << indent_string << "inverse mode: "
         << (m_inverse_mode ? "enabled" : "disabled") << endl;

}
