/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglUnScaleRange.hpp
 *   Author: Scott McMillan
 *  Created: 6 May 2002
 *  Summary: Dynamically scales to maintain a constant minimum pixel size
 *           beyond a certain range, and a constant maximum pixel size inside
 *           a certain range (the user only specifies the min and max sizes).
 *****************************************************************************/

#ifndef __SGL_UNSCALE_RANGE_HPP
#define __SGL_UNSCALE_RANGE_HPP

#include <sgl.h>
#include <sglGroup.hpp>

/**

This node modifies the traversal's view matrix such that geometry below it in
the scene graph is scaled in a complex fashion to appear the same minimum pixel
size when it is greater than a "certain" distance away, and the same maximum
pixel size when it is closer than a second "certain" distance away.  To
determine the distances the user supplies a reference size (in scene graph
length units) that is to be mapped to the minimum and maximum pixel sizes.  In
between these two distances it acts as if there is no scale factor.

Setting the minimum pixel size to 0 disables this limit so that objects can
shrink without limit as they get farther away.  Likewise, specifying the
maximum pixel size to 0 allows that objects to grow larger without limit as
they approach the view point.  Setting both to negative values results in an
sglGroup node behaviour.

Note: Due to the way culling is performed (with pre-computed bounding volumes),
this node may exhibit non-intuitive culling behavior.  The dynamically generate
bounding-volume for this node will reflect the size of its children in
world-space, not the transformed screen space.  Thus, by default, the bounding
volume is likely to be much smaller than the drawn object.  It is recommended
that you set a static bound of appropriate size for this node (and its
children).

<b>Example:</b>

<pre>
   sglGeode geode = new sglGeode;
   sglUnScaleRange *unscale = new sglUnScaleRange;

   // scale the children so that an object with dimension 4 will not shrink
   // less than 10 pixels on that dimension or grow larger than 40 pixels.
   unscale->setScaleParameters(4.f, 10, 40);

   unscale->addChild(geode);
</pre>

 */

class SGL_DLL_API sglUnScaleRange : public sglGroup
{
public:
   /// Default constructor
   sglUnScaleRange();

   /// virtual destructor
   virtual ~sglUnScaleRange();

   /** Set the scaling parameters.
       @param reference_size How tall the geometry beneath this node is.
          Values not greater than zero will be set to one.
       @param min_pixel_size The minimum size (in pixels) the geometry
          (or more precisely, reference_size) is to be rendered at.
          A zero value disables this limit.
       @param max_pixel_size The maximum size (in pixels) the geometry
          (or more precisely, reference_size) is to be rendered at.
          A zero value disables this limit.
   */
   void setScaleParameters(float reference_size,
                           unsigned int min_pixel_size,
                           unsigned int max_pixel_size);

   /** Query the minimum pixel size.  It will return zero if this limit has
       been disabled.
       @return The minimum pixel size.
   */
   unsigned int getMinPixelSize() const { return m_min_pixel_size; }

   /** Query the maximum pixel size.  It will return zero if this limit has
       been disabled.
       @return The maximum pixel size.
   */
   unsigned int getMaxPixelSize() const { return m_max_pixel_size; }

   /** Query the reference size.
       @return The reference size.
   */
   float getReferenceSize() const { return m_reference_size; }

   /** Set whether or not to apply the inverse of the scale computed.  This is
       useful for sequential nodes of this type if you want to undo the
       previous one.  The default is disabled.
       @param on If true apply the inverse instead.
   */
   void setInverseMode(bool on) { m_inverse_mode = on; }

   /** Query whether or not the inverse of the scale computed is applied.
       @return True if this is enabled.
   */
   bool getInverseMode() const { return m_inverse_mode; }

   // See sglNode::intersect(f) for details.
   virtual sglBound::IntersectResultEnum intersect(
      sglIntersectf &isector) const;

   // See sglNode::intersect(d) for details.
   virtual sglBound::IntersectResultEnum intersect(
      sglIntersectd &isector) const;

   // See sglNode::pick(f) for details.
   virtual void pick(sglPickf &pick_state, unsigned int cull_flags) const;

   // See sglNode::pick(d) for details.
   virtual void pick(sglPickd &pick_state, unsigned int cull_flags) const;

   // See sglNode::clone() for details.
   virtual sglNode *clone(unsigned int mode) const;

   // See sglObject::printInfo() for details.
   virtual void printInfo(ostream &ostrm, const char *indent_string) const;

protected:
   // See sglNode::cull(f) for details
   virtual void cull(sglCull<float> &trav_state,
                     unsigned int cull_flags) const;

   // See sglNode::cull(d) for details
   virtual void cull(sglCull<double> &trav_state,
                     unsigned int cull_flags) const;

   void copyTo(sglUnScaleRange *dst, unsigned int mode) const;

private:
   sglUnScaleRange(const sglUnScaleRange &);              // not implemented
   sglUnScaleRange &operator=(const sglUnScaleRange &);   // not implemented

   template <class T>
   T computeScale(sglProbeTemplate<T> &probe) const
        {
            T scale = (T)1;

            if (probe.getOrigViewFrustum().getOrthoFlag())
            {
                // HACK precompute this in sglProbeTemplate.
                scale =
                    probe.getOrigViewFrustum().getPlane(
                        sglFrustum<T>::eUP).getDistance() +
                    probe.getOrigViewFrustum().getPlane(
                        sglFrustum<T>::eDOWN).getDistance();
            }
            else
            {
                scale = -(probe.getModelViewMatrix())[3][2] *
                    tan((probe.getOrigViewFrustum().getFOVY())*0.5) * 2.0;
            }

            // do I need to check for zero scale?

            // compute size of reference size in pixels, the apparent size
            unsigned int viewport_x, viewport_y;
            probe.getViewportSize(viewport_x, viewport_y);
            T apparent_size = viewport_y*m_reference_size/scale;

            if (apparent_size < m_min_pixel_size && m_min_pixel_size != 0)
                scale = (T)m_min_pixel_size/apparent_size;
            else if (apparent_size > m_max_pixel_size && m_max_pixel_size != 0)
                scale = (T)m_max_pixel_size/apparent_size;
            else
                scale = (T)1;

            if (!m_inverse_mode)
                return scale;
            else
                return (T)1/scale;
        }

   template <class T>
   T computeScale(sglCull<T> &trav_state) const
        {
            T scale = 1.0f;

            if (trav_state.getOrigFrustum().getOrthoFlag())
            {
                // HACK precompute this in sglProbeTemplate.
                scale =
                    trav_state.getOrigFrustum().getPlane(
                        sglFrustum<T>::eUP).getDistance() +
                    trav_state.getOrigFrustum().getPlane(
                        sglFrustum<T>::eDOWN).getDistance();
            }
            else
            {
                scale = -(trav_state.getModelViewMatrix())[3][2] *
                    tan((trav_state.getOrigFrustum().getFOVY())*0.5) * 2.0;
            }

            // do I need to check for zero scale?

            // compute size of reference size in pixels, the apparent size
            unsigned int viewport_x, viewport_y;
            trav_state.getViewportSize(viewport_x, viewport_y);
            T apparent_size = viewport_y*m_reference_size/scale;

            if (apparent_size < m_min_pixel_size && m_min_pixel_size != 0)
                scale = (T)m_min_pixel_size/apparent_size;
            else if (apparent_size > m_max_pixel_size && m_max_pixel_size != 0)
                scale = (T)m_max_pixel_size/apparent_size;
            else
                scale = (T)1;

            if (!m_inverse_mode)
                return scale;
            else
                return (T)1/scale;
        }

private:
   float        m_reference_size;
   unsigned int m_min_pixel_size;
   unsigned int m_max_pixel_size;
   bool         m_inverse_mode;
};

#endif
