/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglVec4.hpp
 *  Created: 8 March 1999
 *  Summary: 4D templated vector class
 *****************************************************************************/

#ifndef __SGL_VEC4_HPP
#define __SGL_VEC4_HPP

#include <sgl.h>

template <class T> class sglVec3;
template <class T> class sglMat4;

// ---------------------------------------------------------------------------
template <class T>
class sglVec4
{
public:
    // contructors
   sglVec4() {}; 

   sglVec4(T x, T y, T z, T w)
      {
         m_vector[0] = x; m_vector[1] = y;
         m_vector[2] = z; m_vector[3] = w;
      }

   template <class S>
   sglVec4(const sglVec4<S>& vec)
      {
         m_vector[0] = (T)vec[0]; m_vector[1] = (T)vec[1];
         m_vector[2] = (T)vec[2]; m_vector[3] = (T)vec[3];
      }

   // destructor
   ~sglVec4() {};  // not virtual, do not subclass

   inline T *getPtr() const { return (T *)m_vector; }

   // indexing operators
   inline       T& operator[](unsigned int i)       { return m_vector[i]; }
   inline const T& operator[](unsigned int i) const { return m_vector[i]; }

   // set from any other type
   template <class S>
   inline void set(S x, S y, S z, S w)
      {
         m_vector[0] = (T)x; m_vector[1] = (T)y;
         m_vector[2] = (T)z; m_vector[3] = (T)w;
      }  

   // from 3 to 4 D
   template <class S>
   inline void set(sglVec3<S>& vec, S s = (S)1)
      {
         m_vector[0] = (T)vec[0]; m_vector[1] = (T)vec[1];
         m_vector[2] = (T)vec[2]; m_vector[3] = (T)s;
      }

   // assignment operators
   template <class S>
   inline sglVec4 &operator=(const sglVec4<S>& vec)
      {
         m_vector[0] = (T)vec[0]; m_vector[1] = (T)vec[1];
         m_vector[2] = (T)vec[2]; m_vector[3] = (T)vec[3];
         return *this;
      }

   // boolean operators
   template <class S>
   inline bool operator==(const sglVec4<S>& vec) const
      {
         return (m_vector[0] == (T)vec[0] && m_vector[1] == (T)vec[1] &&
                 m_vector[2] == (T)vec[2] && m_vector[3] == (T)vec[3]);
      }

   template <class S>
   inline bool operator!=(const sglVec4<S>& vec) const
      {
         return (m_vector[0] != (T)vec[0] || m_vector[1] != (T)vec[1] || 
                 m_vector[2] != (T)vec[2] || m_vector[3] != (T)vec[3]);
      }

   template <class S>
   inline bool operator<(const sglVec4<S>& vec) const
      {
         return (dot(*this) < vec.dot(vec));
      }
   template <class S>
   inline bool operator>(const sglVec4<S>& vec) const
      {
         return (dot(*this) > vec.dot(vec));
      }

   // math operators
   template <class S>
   inline sglVec4& operator+=(const sglVec4<S>& vec)
      {
         m_vector[0] += (T)vec[0]; m_vector[1] += (T)vec[1];
         m_vector[2] += (T)vec[2]; m_vector[3] += (T)vec[3];
         return *this;
      }
   template <class S>
   inline sglVec4& operator-=(const sglVec4<S>& vec)
      {
         m_vector[0] -= (T)vec[0]; m_vector[1] -= (T)vec[1];
         m_vector[2] -= (T)vec[2]; m_vector[3] -= (T)vec[3];
         return *this;
      }
   inline sglVec4& operator*=(const T s)
      {
         m_vector[0]*=s; m_vector[1]*=s; m_vector[2]*=s; m_vector[3]*=s;
         return *this;
      }

   template <class S>
   inline const sglVec4<T> operator+(const sglVec4<S> &rhs) const
      {
         return sglVec4(m_vector[0]+(T)rhs[0],
                        m_vector[1]+(T)rhs[1],
                        m_vector[2]+(T)rhs[2],
                        m_vector[3]+(T)rhs[3]);
      }
   template <class S>
   inline const sglVec4<T> operator-(const sglVec4<S> &rhs) const
      {
         return sglVec4(m_vector[0]-(T)rhs[0],
                        m_vector[1]-(T)rhs[1],
                        m_vector[2]-(T)rhs[2],
                        m_vector[3]-(T)rhs[3]);
      }
   inline const sglVec4<T> operator*(T scalar) const
      {
         return sglVec4(m_vector[0]*scalar,
                        m_vector[1]*scalar,
                        m_vector[2]*scalar,
                        m_vector[3]*scalar);
      }

   // explicit math operations (for backward compatibility)
   template <class S1, class S2>
   inline void add(const sglVec4<S1>& vec1, const sglVec4<S2>& vec2)
      {
         m_vector[0] = (T)vec1[0] + (T)vec2[0];
         m_vector[1] = (T)vec1[1] + (T)vec2[1];
         m_vector[2] = (T)vec1[2] + (T)vec2[2];
         m_vector[3] = (T)vec1[3] + (T)vec2[3];
      }
   template <class S1, class S2>
   inline void sub(const sglVec4<S1>& vec1, const sglVec4<S2>& vec2)
      {
         m_vector[0] = (T)vec1[0] - (T)vec2[0];
         m_vector[1] = (T)vec1[1] - (T)vec2[1];
         m_vector[2] = (T)vec1[2] - (T)vec2[2];
         m_vector[3] = (T)vec1[3] - (T)vec2[3]; 
      }
   template <class S>
   inline void scale(const sglVec4<S>& vec, T s)
      {
         m_vector[0] = (T)vec[0]*s; m_vector[1] = (T)vec[1]*s;
         m_vector[2] = (T)vec[2]*s; m_vector[3] = (T)vec[3]*s;
      }

   // other ops
   template <class S>
   inline bool almostEqual(const sglVec4<S>& vec, double eps) const
      {
         return (SGL_ALMOST_EQUAL(m_vector[0], vec[0], eps) && 
                 SGL_ALMOST_EQUAL(m_vector[1], vec[1], eps) &&
                 SGL_ALMOST_EQUAL(m_vector[2], vec[2], eps) &&
                 SGL_ALMOST_EQUAL(m_vector[3], vec[3], eps));
      }

   // reducing methods which result is a T
   template <class S>
   inline T dot(const sglVec4<S>& vec) const
      {
         return (m_vector[0]*(T)vec[0] + m_vector[1]*(T)vec[1] +
                 m_vector[2]*(T)vec[2] + m_vector[3]*(T)vec[3]);
      }

   template <class S, class R>
   inline static R dot(const sglVec4<R>& vec1, const sglVec4<S>& vec2)
      {
         return (vec1[0]*(R)vec2[0] + vec1[1]*(R)vec2[1] +
                 vec1[2]*(R)vec2[2] + vec1[3]*(R)vec2[3]);
      }

   inline T length() const
      {
         return (T) sqrt(dot(*this));
      }

   template <class S>
   inline T distance(const sglVec4<S>& vec) const
      {
         return (T) sqrt(((T)vec[0] - m_vector[0])*((T)vec[0] - m_vector[0]) + 
                         ((T)vec[1] - m_vector[1])*((T)vec[1] - m_vector[1]) + 
                         ((T)vec[2] - m_vector[2])*((T)vec[2] - m_vector[2]) + 
                         ((T)vec[3] - m_vector[3])*((T)vec[3] - m_vector[3]));
      }

   inline T normalize()
      {
         T len = length();
         if (len) *this *= (1.0/len);
         return len;
      }

   template <class S>
   inline T normalize(const sglVec4<S>& vec)
      {
         T len = vec.length();
         if (len) scale(vec, 1.0/len);
         return len;
      }

   // stream operators
#if defined(TEMPLATE_OSTREAM_HACK)
   template <class S> 
   friend ostream& operator<<(ostream& ostrm, const sglVec4<S>& vec);

   template <class S> 
   friend istream& operator>>(istream& istrm, sglVec4<S>& vec);
#else
   friend ostream& operator<<(ostream& ostrm, const sglVec4<T>& vec)
        {
            return (ostrm << "(" << vec[0] << ", " << vec[1] << ", "
                    << vec[2] << ", " << vec[3] << ")");
        }

   friend istream& operator>>(istream& istrm, sglVec4<T>& vec)
        {
            return (istrm >> vec[0] >> vec[1] >> vec[2] >> vec[3]);
        }
#endif
private:
   T m_vector[4];
};

//----------------------------------------------------------------------------
#if defined(TEMPLATE_OSTREAM_HACK)
template <class T>
inline ostream& operator<<(ostream& ostrm, const sglVec4<T>& vec)
{
    return (ostrm << "(" << vec[0] << ", " << vec[1] << ", "
            << vec[2] << ", " << vec[3] << ")");
}

template <class T>
inline istream& operator>>(istream& istrm, sglVec4<T>& vec)
{
    return (istrm >> vec[0] >> vec[1] >> vec[2] >> vec[3]);
}
#endif
//----------------------------------------------------------------------------

#endif
