/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglView.cpp
 *   Author: Scott McMillan
 *  Created: 6 November 2000
 *****************************************************************************/

#include <sglView.hpp>
#include <sglCull.hpp>
#include <sglScene.hpp>
#include <sglGeode.hpp>
#include <sglViewPlatform.hpp>
#include <sglPerspectiveCamera.hpp>
#include <sglFrustum.hpp>
#include <sglLineSet.hpp>
#include <sglBitmapFont.hpp>

#include <stdio.h>

//----------------------------------------------------------------------------
sglView::sglView()
      : m_clear_color(0.0f, 0.0f, 0.0f, 1.0f),
        m_clear_depth(1.0),
        m_origin_x(0),
        m_origin_y(0),
        m_width(640),
        m_height(480),

        m_root(NULL),
        m_view_platform(NULL),

        m_dptrav_flag(false),
        m_stats_flag(false),
        m_trav_state(new sglCullf),
        m_trav_state_dp(NULL),
        m_stats_trav_state(NULL),
        m_stats_trav_state_dp(NULL),
        m_override_statelets(),
        m_discriminator_mask(~0x0u),
        m_state_mask(0x0),

        m_draw_callback(NULL),
        m_draw_callback_data(NULL),

        m_last_ts(),
        m_ts1(),
        m_ts2(),
        m_ts3(),
        m_cull_time(0.0),
        m_draw_time(0.0),
        m_frame_count(0),
        m_stats_interval(2.0),

        m_frame_rate_flag(false),
        m_frame_rate(0.0)
{
   if (!sgl::s_initialized)
   {
      sglPrintWarning() << "sglView::sglView() error: "
                        << "called before sgl::initialize().  "
                        << "The cull traversal will not work if there is more"
                        << " than one texture unit" << endl;
   }

   sglTimespec::getSysTime(m_last_ts);
}

//----------------------------------------------------------------------------
sglView::~sglView()
{
   if (m_root)
   {
      m_root->deref();  // should I delete the root if refCount == 0?
   }

   delete m_trav_state;
   delete m_trav_state_dp;
   delete m_stats_trav_state;
   delete m_stats_trav_state_dp;
}

//----------------------------------------------------------------------------
void sglView::setClearColor(const sglVec4f &color)
{
   //if (m_clear_color != color)
   {
      m_clear_color = color;
   }
}

//----------------------------------------------------------------------------
void sglView::setClearColor(float r, float g, float b, float a)
{
   setClearColor(sglVec4f(r, g, b, a));
}

//----------------------------------------------------------------------------
void sglView::setClearDepth(double depth)
{
   if (depth > 1.0) depth = 1.0;
   else if (depth < 0.0) depth = 0.0;

   m_clear_depth = depth;
}

//----------------------------------------------------------------------------
void sglView::setViewport(int origin_x, int origin_y,
                          unsigned int width, unsigned int height)
{
   if (m_origin_x != origin_x || m_origin_y != origin_y ||
       m_width != width || m_height != height)
   {
      m_origin_x = origin_x;
      m_origin_y = origin_y;
      m_width    = width;
      m_height   = height;
   }
}

//----------------------------------------------------------------------------
void sglView::getViewport(int &origin_x, int &origin_y,
                          unsigned int &width, unsigned int &height)
{
   origin_x = m_origin_x;
   origin_y = m_origin_y;
   width    = m_width;
   height   = m_height;
}


//----------------------------------------------------------------------------
void sglView::setScene(sglScene *root)
{
   if (root != m_root)
   {
      if (m_root)
      {
         m_root->deref();
      }

      m_root = root;

      if (m_root)
      {
         m_root->ref();
      }
   }
}

//----------------------------------------------------------------------------
void sglView::setViewPlatform(sglViewPlatform *view_platform)
{
   m_view_platform = view_platform;
}

//----------------------------------------------------------------------------
void sglView::setScene(sglScene *root, sglViewPlatform *view_platform)
{
   setScene(root);
   setViewPlatform(view_platform);
}

//----------------------------------------------------------------------------
void sglView::setTraversalType(bool dp_flag, bool stats_flag)
{
   sglLightProcessor *light_processor = getLightProcessor();

   m_dptrav_flag = dp_flag;
   m_stats_flag  = stats_flag;

   if (m_dptrav_flag)
   {
      if (m_stats_flag)
      {
         if (m_stats_trav_state_dp == NULL)
         {
            m_stats_trav_state_dp = new sglStatsCulld;
            m_stats_trav_state_dp->setLightProcessor(light_processor);
         }
      }
      else
      {
         if (m_trav_state_dp == NULL)
         {
            m_trav_state_dp = new sglCulld;
            m_trav_state_dp->setLightProcessor(light_processor);
         }
      }
   }
   else
   {
      if (m_stats_flag)
      {
         if (m_stats_trav_state == NULL)
         {
            m_stats_trav_state = new sglStatsCullf;
            m_stats_trav_state->setLightProcessor(light_processor);
         }
      }
      else
      {
         if (m_trav_state == NULL)
         {
            m_trav_state = new sglCullf;
            m_trav_state->setLightProcessor(light_processor);
         }
      }
   }
}

//----------------------------------------------------------------------------
void sglView::preDraw()
{
    if (m_root && m_view_platform)
    {
        m_root->preDraw();
    }
}

//----------------------------------------------------------------------------
void sglView::preDraw(const sglTimespec &frame_time, unsigned int frame_count)
{
    if (m_root && m_view_platform)
    {
        m_root->preDraw(frame_time, frame_count);
    }
}

//----------------------------------------------------------------------------
void sglView::cull()
{
   if (m_root && m_view_platform)
   {
      m_frame_count++;

      // HACK - default for ortho cameras, this isn't the desired behaviour.
      float fov_y = 0.25f*(float)M_PI;
      float fov_x = 0.25f*(float)M_PI;

      sglPerspectiveCamera *camera = dynamic_cast<sglPerspectiveCamera*>(
         &(m_view_platform->getCamera()));
      if (camera)
      {
         fov_x = camera->getFOVX();
         fov_y = camera->getFOVY();
         sglPrintDebug() << "f: fov x, y: " << fov_x*180.0/M_PI << ", "
                         << fov_y*180.0/M_PI << endl;
      }

      sglTimespec::getSysTime(m_ts1);

      if (m_dptrav_flag)
      {
         // HACK what do I do if the root of the scene graph containing this
         // view platform node is not this sglScene node??
         m_view_platform->computeViewMatrix(m_view_matrix_d);

         if (m_stats_flag)
         {
            m_root->cull(*m_stats_trav_state_dp,
                         fov_x, fov_y,
                         m_width, m_height,
                         (m_view_platform->getCamera()).getFrustum(),
                         m_view_matrix_d,
                         m_override_statelets,
                         m_discriminator_mask,
                         m_state_mask);
         }
         else
         {
            m_root->cull(*m_trav_state_dp,
                         fov_x, fov_y,
                         m_width, m_height,
                         (m_view_platform->getCamera()).getFrustum(),
                         m_view_matrix_d,
                         m_override_statelets,
                         m_discriminator_mask,
                         m_state_mask);
         }
      }
      else
      {
         // HACK what do I do if the root of the scene graph containing this
         // view platform node is not this sglScene node??
         m_view_platform->computeViewMatrix(m_view_matrix_f);

         sglFrustumf frustum = (m_view_platform->getCamera()).getFrustum();

         if (m_stats_flag)
         {
            m_root->cull(*m_stats_trav_state,
                         fov_x, fov_y,
                         m_width, m_height,
                         frustum, m_view_matrix_f,
                         m_override_statelets,
                         m_discriminator_mask,
                         m_state_mask);
         }
         else
         {
            m_root->cull(*m_trav_state,
                         fov_x, fov_y,
                         m_width, m_height,
                         frustum, m_view_matrix_f,
                         m_override_statelets,
                         m_discriminator_mask,
                         m_state_mask);
         }
      }

      sglTimespec::getSysTime(m_ts2);
      m_cull_time += sglTimespec::convertToDouble(m_ts2 - m_ts1);
   }
}

//----------------------------------------------------------------------------
void sglView::draw()
{
   if (m_root && m_view_platform)
   {
      sglTimespec::getSysTime(m_ts2);

      glViewport(m_origin_x, m_origin_y, m_width, m_height);
      glScissor(m_origin_x, m_origin_y, m_width, m_height);
      glEnable(GL_SCISSOR_TEST);

      glClearColor(m_clear_color[0], m_clear_color[1], m_clear_color[2],
                   m_clear_color[3]);
      glClearDepth(m_clear_depth);
      m_view_platform->getCamera().applyProjection();

      if (m_draw_callback)
      {
         glMatrixMode(GL_MODELVIEW);
         if (m_dptrav_flag)
            sglLoadMatrix(m_view_matrix_d);
         else
            sglLoadMatrix(m_view_matrix_f);

         // HACK - still need to apply lights here.

         (*m_draw_callback)(m_draw_callback_data);
      }
      else
      {
         glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
         //glClear(GL_DEPTH_BUFFER_BIT); // This periodically dumps core on
                                         // Mesa during a resize event
      }

      if (m_dptrav_flag)
      {
         if (m_stats_flag)
         {
            m_stats_trav_state_dp->draw();
         }
         else
         {
            m_trav_state_dp->draw();
         }
      }
      else
      {
         if (m_stats_flag)
         {
            m_stats_trav_state->draw();
         }
         else
         {
            m_trav_state->draw();
         }
      }

      sglTimespec::getSysTime(m_ts3);

      m_draw_time += sglTimespec::convertToDouble(m_ts3 - m_ts2);
      double elapsed_time = sglTimespec::convertToDouble(m_ts3 - m_last_ts);

      if (elapsed_time > m_stats_interval)
      {
         m_frame_rate = (double) m_frame_count/elapsed_time;

         if (m_stats_flag)
         {
            if (!m_dptrav_flag &&
                m_stats_trav_state->getStats().getFrames() > 0)
            {
               m_cull_time /=
                  (double)m_stats_trav_state->getStats().getFrames();
               m_draw_time /=
                  (double)m_stats_trav_state->getStats().getFrames();
               cerr << "avg. cull time: " << m_cull_time << endl;
               cerr << "avg. draw time: " << m_draw_time << endl;
               m_stats_trav_state->getStats().average();
               m_stats_trav_state->getStats().printInfo(cerr, "");
               m_stats_trav_state->getStats().reset();
               m_cull_time = 0;
               m_draw_time = 0;
            }
            else if (m_dptrav_flag &&
                     m_stats_trav_state_dp->getStats().getFrames() > 0)
            {
               m_cull_time /=
                  (double)m_stats_trav_state_dp->getStats().getFrames();
               m_draw_time /=
                  (double)m_stats_trav_state_dp->getStats().getFrames();
               cerr << "avg. cull time: " << m_cull_time << endl;
               cerr << "avg. draw time: " << m_draw_time << endl;
               m_stats_trav_state_dp->getStats().average();
               m_stats_trav_state_dp->getStats().printInfo(cerr, "");
               m_stats_trav_state_dp->getStats().reset();
               m_cull_time = 0;
               m_draw_time = 0;
            }
         }

         m_frame_count = 0;
         m_last_ts = m_ts3;
      }

      if (m_frame_rate_flag)
      {
         outputFramerate();
      }

      GLenum glerror = glGetError();
      if (glerror)
      {
         cerr << "glGetError() = 0x" << hex << (int)glerror << dec << endl;
      }
   }
}

//----------------------------------------------------------------------------
void sglView::outputFramerate()
{
   glDisable(GL_DEPTH_TEST);
   glMatrixMode(GL_PROJECTION);
   glPushMatrix();
   glLoadIdentity();
   glOrtho(0, m_width, 0, m_height, -1, 1);

   glMatrixMode(GL_MODELVIEW);
   glLoadIdentity();

   glEnable(GL_BLEND);
   glColor3f(1.0f, 1.0f, 0.5f);
   char fr[80];
   sprintf(fr, "framerate = %.2f", m_frame_rate);
   glRasterPos3i(1,1,0);
   sglBitmapFont::getDefaultFont()->rasterString(fr);
   glDisable(GL_BLEND);

   glMatrixMode(GL_PROJECTION);
   glPopMatrix();
   glMatrixMode(GL_MODELVIEW);
   glEnable(GL_DEPTH_TEST);
}

//----------------------------------------------------------------------------
void sglView::setLightProcessor(sglLightProcessor *light_processor)
{
   if (m_stats_trav_state_dp)
      m_stats_trav_state_dp->setLightProcessor(light_processor);

   if (m_trav_state_dp)
      m_trav_state_dp->setLightProcessor(light_processor);

   if (m_stats_trav_state)
      m_stats_trav_state->setLightProcessor(light_processor);

   if (m_trav_state)
      m_trav_state->setLightProcessor(light_processor);
}

//----------------------------------------------------------------------------
sglLightProcessor *sglView::getLightProcessor() const
{
   if (m_dptrav_flag)
   {
      if (m_stats_flag)
      {
         return m_stats_trav_state_dp->getLightProcessor();
      }
      else
      {
         return m_trav_state_dp->getLightProcessor();
      }
   }
   else
   {
      if (m_stats_flag)
      {
         return m_stats_trav_state->getLightProcessor();
      }
      else
      {
         return m_trav_state->getLightProcessor();
      }
   }
}
