/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglViewPlatform.hpp
 *   Author: Scott McMillan
 *  Created: 14 December 2000
 *****************************************************************************/

#ifndef __SGL_VIEW_PLATFORM_HPP
#define __SGL_VIEW_PLATFORM_HPP

#include <sgl.h>
#include <sglNode.hpp>
#include <sglCamera.hpp>

/**

sglViewPlatform is a placeholder in the scene graph that can be used to define
the position and orientation of the camera.  This class must be constructed
with a reference to an sglCamera subclass to define the view frustum.  This
camera can subsequently be changed with a call to setCamera().  The
sglViewPlatform::computeViewMatrix() member function (overloaded for single or
double precision) traverses the scene graph from this node upward to the root
of the tree while computing the matrix that is needed by the sglScene::cull()
functions.  If multiple paths to the root of the tree are encountered
(i.e. multiple parents), an arbitrary selection is made (generally the path
using the parents with the 0 index).  An overloaded version of
sglScene::cull(..) can take an sglViewPlatform reference which, in turn, will
call the computeViewMatrix(..) function.

By default, the view matrix is computed assuming the OpenGL convention that the
camera looks down the -Z axis with the +X axis to the right and the +Y axis up.
A convenience mode for Performer users (and others?) can be enabled by calling
sglViewPlatform::enableZUp(true).  This will cause the computeViewMatrix
function to return a matrix that has the camera looking down the +Y axis, with
+Z up and +X to the right.

The following example attaches the view platform to an sglTransform node which
will be used to move the camera around.  Also shown in this example is the
method for defining the camera projection (view frustum) parameters,
associating them with the view platform, and using it in the cull traversal:

<b>Example:</b>

<pre>
   sglCullf trav_state;

   // used to move the camera
   sgluSphericalTransformf view_trans = new sgluSphericalTransformf;

   sglPerspectiveCamera *view_camera = new sglPerspectiveCamera;
   view_camera->setFOV(M_PI*0.25, (double)win_x/(double)win_y, 1.0, 200.0);

   sglViewPlatform *view_platform = new sglViewPlatform(*view_camera);
   view_platform->setEnableZUp(true);
   view_trans->addChild(view_platform);
        :
   root->cull(trav_state,                          //  sglScene::cull(..)
              ...,
              *view_platform,
              ...);

   (view_platform->getCamera()).applyProjection(); // setup projection matrix

   trav_state.draw();                              // draw the geometry
</pre>
*/

class SGL_DLL_API sglViewPlatform : public sglNode
{
public:
   /** Constructor
       @param camera Reference to an sglCamera
   */
   sglViewPlatform(sglCamera &camera);

   /// virtual destructor.  Will dereference but not delete camera object.
   virtual ~sglViewPlatform();

   /** This function enables the "Performer/PLIB" compatibility mode
       that assumes the camera looks along the y-axis (x-axis to the right)
       Otherwise it assumes the default OpenGL mode where the camera looks
       along the negative z-axis.
       @param on If true then enable the compatibility mode.
   */
   void setEnableZUp(bool on) { m_rotx_flag = on; }

   /** Query the current compatibility mode.
       @return True if compatibility mode is on.
   */
   bool getEnableZUp() const  { return m_rotx_flag; }

   /** Compute a single precision transform to root of tree this node is
       connected to. Since this node is the first in the progression these
       functions initialize the matrix.
       @param curr_mat Reference to the matrix that will contain the result on
       return from this function.
   */
   void computeViewMatrix(sglMat4f &curr_mat) const;

   /** Compute a double precision transform to root of tree this node is
       connected to. Since this node is the first in the progression these
       functions initialize the matrix.
       @param curr_mat Reference to the matrix that will contain the result on
       return from this function.
   */
   void computeViewMatrix(sglMat4d &curr_mat) const;

   /** Query the current sglCamera.
       @return A reference to the sglCamera object being used by this class.
   */
   sglCamera &getCamera() const { return *m_camera; }

   /** Switch the camera being used by this node.  The old camera object is
       dereferenced but if the reference count goes to zero it will not be
       deleted.  The reference count of the new camera will be incremented.
       @param camera Reference to an sglCamera (subclass) object.
   */
   void setCamera(sglCamera &camera);

   // See sglNode::clone(..) for details.
   virtual sglNode *clone(unsigned int mode) const;

   // See sglObject::printInfo() for details.
   virtual void printInfo(ostream &ostrm, const char *indent_string) const;

protected:
   void computeBound();

   // See sglNode::cull(f) for details
   virtual void cull(sglCull<float> &,  unsigned int) const {}
   // See sglNode::cull(d) for details
   virtual void cull(sglCull<double> &, unsigned int) const {}

   void copyTo(sglViewPlatform *dst, unsigned int mode) const;

private:   // not implemented
   sglViewPlatform();
   sglViewPlatform(const sglViewPlatform &);
   sglViewPlatform &operator=(const sglViewPlatform &);

private:
   bool m_rotx_flag;
   sglCamera *m_camera;
};

#endif
