/*****************************************************************************
 * SGL: A SceneGraph Library, Version 1.0
 * Copyright (C) 1997-2001  Scott McMillan
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglLoadModel3ds.cpp
 *   Author: Scott McMillan
 *  Created: 5 July 2000
 *  Summary:
 *
 *  This was written to be a part of Stephen J Baker's
 *  PLIB (http://plib.sourceforge.net)
 *
 *  Originally written by Per Liedman (liedman@home.se)
 *  Last updated: 2000-09-08
 *
 * KNOWN ISSUES:

 * Some models (one of the test cases) gets turned "inside out" - obviously
   the triangle winding has been reversed so that all triangles that should be
   culled is show and vice versa. I don't know where this information can
   be found in the file.

 * Models that uses double-sided materials sometimes look strange. My approach
   to double sided materials is to add a copy of each triangle with the
   winding reversed and also negate the normals...this seems to *almost* work.

 * The transformation matrix in the 3DS files is still a mystery to me -
   whatever I do with it seems to make things worse than ignoring it, although
   that doesn't work perfect either. I've taken a look at other loaders, and
   it seems that no one really knows what to do with them.
 *****************************************************************************/

#include <sglLoadModel3ds.hpp>
#include <sglImageLoader.hpp>

//#include <sglOStream.hpp>
#include <sglGroup.hpp>
#include <sglTransform.hpp>
#include <sglGeode.hpp>
#include <sglStatelet.hpp>
#include <sglMaterial.hpp>
#include <sglTexture2D.hpp>
#include <sglTriangleSet.hpp>

#include <stdio.h>
#include <string.h>

extern "C"
{
   SGL3DS_DLL_API
   void *loadModel3DS(const string &filepath, unsigned int)
   {
      sglLoadModel3ds ldr;
      sglNode *model = ldr.load(filepath);
      return model;
   }
}

class sgl3dsTexture : public sglTexture2D
{
public:
   sgl3dsTexture() {};
   ~sgl3dsTexture()
      {
         void *image;
         ExternalFormatEnum format;
         TypeEnum type;
         unsigned int num_s, num_t;
         bool border;
         getImage(image, format, type, num_s, num_t, border);
         if (image)
            free(image);
      }
};


const unsigned int SGL_3DS_CHUNK_HEADER_SIZE = (2 + 4);

/* this is the minimum value of the dot product for to faces if their normals
   should be smoothed, if they don't use smooth groups. */
const float sgl_smooth_threshold = 0.8f;

// 3ds chunk identifiers
enum {
   LIB3DS_NULL_CHUNK             = 0x0000,
   SGL_3DS_CHUNK_VERSION         = 0x0002,  // M3D_VERSION
   LIB3DS_M3D_VERSION            = 0x0005,
   LIB3DS_SMAGIC                 = 0x2D2D,
   LIB3DS_LMAGIC                 = 0x2D3D,
   SGL_3DS_CHUNK_OBJMESH         = 0x3D3D,  // MDATA
   LIB3DS_MESH_VERSION           = 0x3D3E,
   LIB3DS_MLIBMAGIC              = 0x3DAA,
   LIB3DS_MATMAGIC               = 0x3DFF,
   SGL_3DS_CHUNK_MAIN            = 0x4D4D,  // M3DMAGIC
   SGL_3DS_CHUNK_MATERIAL        = 0xAFFF,  // MAT_ENTRY
   LIB3DS_CMAGIC                 = 0xC23D,

   SGL_3DS_CHUNK_RGB1            = 0x0010,  // COLOR_F      3 floats of RGB
   SGL_3DS_CHUNK_RGB2            = 0x0011,  // COLOR_24     3 bytes of RGB
   SGL_3DS_CHUNK_RGB3            = 0x0012,  // LIN_COLOR_24 3 bytes of RGB (gamma corrected)
   LIB3DS_LIN_COLOR_F            = 0x0013,  //              3 floats of RGB (gamma corrected)

   SGL_3DS_CHUNK_AMOUNT          = 0x0030,  // INT_PERCENTAGE
   LIB3DS_FLOAT_PERCENTAGE       = 0x0031,

   SGL_3DS_CHUNK_ONEUNIT         = 0x0100,  // MASTER_SCALE

   // Shadow
   LIB3DS_LO_SHADOW_BIAS         = 0x1400,
   LIB3DS_HI_SHADOW_BIAS         = 0x1410,
   LIB3DS_SHADOW_MAP_SIZE        = 0x1420,
   LIB3DS_SHADOW_SAMPLES         = 0x1430,
   LIB3DS_SHADOW_RANGE           = 0x1440,
   LIB3DS_SHADOW_FILTER          = 0x1450,
   LIB3DS_RAY_BIAS               = 0x1460,
   LIB3DS_O_CONSTS               = 0x1500,

   // Background?
   LIB3DS_BIT_MAP                = 0x1100,
   LIB3DS_USE_BIT_MAP            = 0x1101,
   SGL_3DS_CHUNK_BKGCOLOR        = 0x1200,  // SOLID_BGND
   LIB3DS_USE_SOLID_BGND         = 0x1201,
   LIB3DS_V_GRADIENT             = 0x1300,
   LIB3DS_USE_V_GRADIENT         = 0x1301,

   // Global? ambient light
   SGL_3DS_CHUNK_AMBCOLOR        = 0x2100,  // AMBIENT_LIGHT

   // Fog
   LIB3DS_FOG                    = 0x2200,
   LIB3DS_USE_FOG                = 0x2201,
   LIB3DS_FOG_BGND               = 0x2210,
   LIB3DS_DISTANCE_CUE           = 0x2300,
   LIB3DS_USE_DISTANCE_CUE       = 0x2301,
   LIB3DS_LAYER_FOG              = 0x2302,
   LIB3DS_USE_LAYER_FOG          = 0x2303,
   LIB3DS_DCUE_BGND              = 0x2310,

   // View
   LIB3DS_DEFAULT_VIEW           = 0x3000,
   LIB3DS_VIEW_TOP               = 0x3010,
   LIB3DS_VIEW_BOTTOM            = 0x3020,
   LIB3DS_VIEW_LEFT              = 0x3030,
   LIB3DS_VIEW_RIGHT             = 0x3040,
   LIB3DS_VIEW_FRONT             = 0x3050,
   LIB3DS_VIEW_BACK              = 0x3060,
   LIB3DS_VIEW_USER              = 0x3070,
   LIB3DS_VIEW_CAMERA            = 0x3080,
   LIB3DS_VIEW_WINDOW            = 0x3090,

   // Objects
   SGL_3DS_CHUNK_OBJBLOCK        = 0x4000,  // NAMED_OBJECT
   LIB3DS_OBJ_HIDDEN             = 0x4010,
   LIB3DS_OBJ_VIS_LOFTER         = 0x4011,
   LIB3DS_OBJ_DOESNT_CAST        = 0x4012,
   LIB3DS_OBJ_MATTE              = 0x4013,
   LIB3DS_OBJ_FAST               = 0x4014,
   LIB3DS_OBJ_PROCEDURAL         = 0x4015,
   LIB3DS_OBJ_FROZEN             = 0x4016,
   LIB3DS_OBJ_DONT_RECVSHADOW    = 0x4017,

   // Triangle meshes
   SGL_3DS_CHUNK_TRIMESH         = 0x4100,  // N_TRI_OBJECT
   SGL_3DS_CHUNK_VERTLIST        = 0x4110,  // POINT_ARRAY
   LIB3DS_POINT_FLAG_ARRAY       = 0x4111,
   SGL_3DS_CHUNK_FACELIST        = 0x4120,  // FACE_ARRAY
   SGL_3DS_CHUNK_FACEMAT         = 0x4130,  // MSH_MAT_GROUP
   SGL_3DS_CHUNK_MAPLIST         = 0x4140,  // TEX_VERTS
   SGL_3DS_CHUNK_SMOOLIST        = 0x4150,  // SMOOTH_GROUP
   SGL_3DS_CHUNK_TRMATRIX        = 0x4160,  // MESH_MATRIX
   LIB3DS_MESH_COLOR             = 0x4165,
   LIB3DS_MESH_TEXTURE_INFO      = 0x4170,
   LIB3DS_MSH_BOXMAP             = 0x4190,

   // Lights
   SGL_3DS_CHUNK_LIGHT           = 0x4600,  // N_DIRECT_LIGHT
   SGL_3DS_CHUNK_SPOTLIGHT       = 0x4610,  // DL_SPOTLIGHT
   LIB3DS_DL_OFF                 = 0x4620,
   LIB3DS_DL_ATTENUATE           = 0x4625,
   LIB3DS_DL_RAYSHAD             = 0x4627,
   LIB3DS_DL_SHADOWED            = 0x4630,
   LIB3DS_DL_LOCAL_SHADOW2       = 0x4641,
   LIB3DS_DL_SEE_CONE            = 0x4650,
   LIB3DS_DL_SPOT_RECTANGULAR    = 0x4651,
   LIB3DS_DL_SPOT_OVERSHOOT      = 0x4652,
   LIB3DS_DL_SPOT_PROJECTOR      = 0x4653,
   LIB3DS_DL_EXCLUDE             = 0x4654,
   LIB3DS_DL_SPOT_ROLL           = 0x4656,
   LIB3DS_DL_SPOT_ASPECT         = 0x4657,
   LIB3DS_DL_RAY_BIAS            = 0x4658,
   LIB3DS_DL_INNER_RANGE         = 0x4659,
   LIB3DS_DL_OUTER_RANGE         = 0x465A,
   LIB3DS_DL_MULTIPLIER          = 0x465B,

   // ???
   SGL_3DS_CHUNK_CAMERA          = 0x4700,  // N_CAMERA
   LIB3DS_CAM_SEE_CONE           = 0x4710,
   LIB3DS_CAM_RANGES             = 0x4720,

   // ???
   LIB3DS_POLY_2D                = 0x5000,
   LIB3DS_SHAPE_OK               = 0x5010,
   LIB3DS_SHAPE_NOT_OK           = 0x5011,
   LIB3DS_SHAPE_HOOK             = 0x5020,

   // ???
   LIB3DS_PATH_3D                = 0x6000,
   LIB3DS_PATH_MATRIX            = 0x6005,
   LIB3DS_SHAPE_2D               = 0x6010,
   LIB3DS_M_SCALE                = 0x6020,
   LIB3DS_M_TWIST                = 0x6030,
   LIB3DS_M_TEETER               = 0x6040,
   LIB3DS_M_FIT                  = 0x6050,
   LIB3DS_M_BEVEL                = 0x6060,
   LIB3DS_XZ_CURVE               = 0x6070,
   LIB3DS_YZ_CURVE               = 0x6080,
   LIB3DS_INTERPCT               = 0x6090,
   LIB3DS_DEFORM_LIMIT           = 0x60A0,

   LIB3DS_USE_CONTOUR            = 0x6100,
   LIB3DS_USE_TWEEN              = 0x6110,
   LIB3DS_USE_SCALE              = 0x6120,
   LIB3DS_USE_TWIST              = 0x6130,
   LIB3DS_USE_TEETER             = 0x6140,
   LIB3DS_USE_FIT                = 0x6150,
   LIB3DS_USE_BEVEL              = 0x6160,

   // Viewport
   LIB3DS_VIEWPORT_LAYOUT_OLD    = 0x7000,
   LIB3DS_VIEWPORT_DATA_OLD      = 0x7010,
   LIB3DS_VIEWPORT_LAYOUT        = 0x7001,
   LIB3DS_VIEWPORT_DATA          = 0x7011,
   LIB3DS_VIEWPORT_DATA_3        = 0x7012,
   LIB3DS_VIEWPORT_SIZE          = 0x7020,
   LIB3DS_NETWORK_VIEW           = 0x7030,

   // Materials
   SGL_3DS_CHUNK_MATNAME         = 0xA000,  // MAT_NAME
   SGL_3DS_CHUNK_AMBIENT         = 0xA010,  // MAT_AMBIENT
   SGL_3DS_CHUNK_DIFFUSE         = 0xA020,  // MAT_DIFFUSE
   SGL_3DS_CHUNK_SPECULAR        = 0xA030,  // MAT_SPECULAR
   SGL_3DS_CHUNK_SHININESS       = 0xA040,  // MAT_SHININESS
   SGL_3DS_CHUNK_SHINE_STRENGTH  = 0xA041,  // MAT_SHIN2PCT
   SGL_3DS_CHUNK_TRANSPARENCY    = 0xA050,  // MAT_TRANSPARENCY
   SGL_3DS_CHUNK_TRANSP_FALLOFF  = 0xA052,  // MAT_XPFALL
   SGL_3DS_CHUNK_UNKNOWN_DOUBLE1 = 0xA053,  // MAT_REFBLUR
   LIB3DS_MAT_SELF_ILLUM         = 0xA080,
   SGL_3DS_CHUNK_DOUBLESIDED     = 0xA081,  // MAT_TWO_SIDE
   LIB3DS_MAT_DECAL              = 0xA082,
   LIB3DS_MAT_ADDITIVE           = 0xA083,
   LIB3DS_MAT_WIRE               = 0xA085,
   LIB3DS_MAT_WIRE_SIZE          = 0xA087,
   LIB3DS_MAT_FACEMAP            = 0xA088,
   LIB3DS_MAT_PHONGSOFT          = 0xA08C,
   LIB3DS_MAT_WIREABS            = 0xA08E,

   LIB3DS_MAT_SHADING            = 0xA100,

   // Textures
   SGL_3DS_CHUNK_TEXTURE         = 0xA200,  // MAT_TEXMAP
   LIB3DS_MAT_SPECMAP            = 0xA204,
   LIB3DS_MAT_OPACMAP            = 0xA210,
   SGL_3DS_CHUNK_REFLECTIONMAP   = 0xA220,  // MAT_REFLMAP       SPECULATION
   SGL_3DS_CHUNK_BUMPMAP         = 0xA230,  // MAT_BUMPMAP
   LIB3DS_MAT_USE_XPFALL         = 0xA240,
   LIB3DS_MAT_USE_REFBLUR        = 0xA250,

   SGL_3DS_CHUNK_MAPFILENAME     = 0xA300,  // MAT_MAPNAME
   LIB3DS_MAT_ACUBIC             = 0xA310,
   LIB3DS_MAT_SXP_TEXT_DATA      = 0xA320,
   LIB3DS_MAT_SXP_TEXT2_DATA     = 0xA321,
   LIB3DS_MAT_SXP_OPAC_DATA      = 0xA322,
   LIB3DS_MAT_SXP_BUMP_DATA      = 0xA324,
   LIB3DS_MAT_SXP_SPEC_DATA      = 0xA325,
   LIB3DS_MAT_SXP_SHIN_DATA      = 0xA326,
   LIB3DS_MAT_SXP_SELFI_DATA     = 0xA328,
   LIB3DS_MAT_SXP_TEXTMASK_DATA  = 0xA32A,
   LIB3DS_MAT_SXP_TEXT2MASK_DATA = 0xA32C,
   LIB3DS_MAT_SXP_OPACMASK_DATA  = 0xA32E,
   LIB3DS_MAT_SXP_BUMPMASK_DATA  = 0xA330,
   LIB3DS_MAT_SXP_SPECMASK_DATA  = 0xA332,
   LIB3DS_MAT_SXP_SHINMASK_DATA  = 0xA334,
   LIB3DS_MAT_SXP_SELFIMASK_DATA = 0xA336,
   LIB3DS_MAT_SXP_REFLMASK_DATA  = 0xA338,
   LIB3DS_MAT_TEX2MAP            = 0xA33A,
   LIB3DS_MAT_SHINMAP            = 0xA33C,
   LIB3DS_MAT_SELFIMAP           = 0xA33D,
   LIB3DS_MAT_TEXMASK            = 0xA33E,
   LIB3DS_MAT_TEX2MASK           = 0xA340,
   LIB3DS_MAT_OPACMASK           = 0xA342,
   LIB3DS_MAT_BUMPMASK           = 0xA344,
   LIB3DS_MAT_SHINMASK           = 0xA346,
   LIB3DS_MAT_SPECMASK           = 0xA348,
   LIB3DS_MAT_SELFIMASK          = 0xA34A,
   LIB3DS_MAT_REFLMASK           = 0xA34C,

   SGL_3DS_CHUNK_MAPOPTIONS      = 0xA351,  // MAT_MAP_TILING
   LIB3DS_MAT_MAP_TEXBLUR        = 0xA353,
   SGL_3DS_CHUNK_MAP_VSCALE      = 0xA354,  // MAT_MAP_USCALE    REVERSED?
   SGL_3DS_CHUNK_MAP_USCALE      = 0xA356,  // MAT_MAP_VSCALE
   SGL_3DS_CHUNK_MAP_UOFFSET     = 0xA358,  // MAT_MAP_UOFFSET
   SGL_3DS_CHUNK_MAP_VOFFSET     = 0xA35A,  // MAT_MAP_VOFFSET
   LIB3DS_MAT_MAP_ANG            = 0xA35C,

   LIB3DS_MAT_MAP_COL1           = 0xA360,
   LIB3DS_MAT_MAP_COL2           = 0xA362,
   LIB3DS_MAT_MAP_RCOL           = 0xA364,
   LIB3DS_MAT_MAP_GCOL           = 0xA366,
   LIB3DS_MAT_MAP_BCOL           = 0xA368,

   // Keyframe
   SGL_3DS_CHUNK_KEYFRAMER       = 0xB000,  // KFDATA
   LIB3DS_AMBIENT_NODE_TAG       = 0xB001,
   LIB3DS_OBJECT_NODE_TAG        = 0xB002,
   LIB3DS_CAMERA_NODE_TAG        = 0xB003,
   LIB3DS_TARGET_NODE_TAG        = 0xB004,
   LIB3DS_LIGHT_NODE_TAG         = 0xB005,
   LIB3DS_L_TARGET_NODE_TAG      = 0xB006,
   LIB3DS_SPOTLIGHT_NODE_TAG     = 0xB007,
   SGL_3DS_CHUNK_FRAMES          = 0xB008,   // KFSEG
   LIB3DS_KFCURTIME              = 0xB009,
   LIB3DS_KFHDR                  = 0xB00A,

   // Nodes?
   LIB3DS_NODE_HDR               = 0xB010,
   LIB3DS_INSTANCE_NAME          = 0xB011,
   LIB3DS_PIVOT                  = 0xB013,
   LIB3DS_BOUNDBOX               = 0xB014,
   LIB3DS_MORPH_SMOOTH           = 0xB015,
   LIB3DS_POS_TRACK_TAG          = 0xB020,
   LIB3DS_ROT_TRACK_TAG          = 0xB021,
   LIB3DS_SCL_TRACK_TAG          = 0xB022,
   LIB3DS_FOV_TRACK_TAG          = 0xB023,
   LIB3DS_ROLL_TRACK_TAG         = 0xB024,
   LIB3DS_COL_TRACK_TAG          = 0xB025,
   LIB3DS_MORPH_TRACK_TAG        = 0xB026,
   LIB3DS_HOT_TRACK_TAG          = 0xB027,
   LIB3DS_FALL_TRACK_TAG         = 0xB028,
   LIB3DS_HIDE_TRACK_TAG         = 0xB029,
   LIB3DS_NODE_ID                = 0xB030
};

//----------------------------------------------------------------------------

static sglLoadModel3ds::Chunk FaceListDataChunks[] =
{
   { SGL_3DS_CHUNK_SMOOLIST, NULL, sglLoadModel3ds::parseSmoothList },
   { SGL_3DS_CHUNK_FACEMAT,  NULL, sglLoadModel3ds::identifyFaceMaterials },
   { 0, NULL, NULL }
};

static sglLoadModel3ds::Chunk FaceListChunks[] =
{
   { SGL_3DS_CHUNK_FACEMAT, NULL, sglLoadModel3ds::parseFaceMaterials },
   { 0, NULL, NULL }
};

static sglLoadModel3ds::Chunk TriMeshDataChunks[] =
{
   { SGL_3DS_CHUNK_VERTLIST, NULL, sglLoadModel3ds::parseVertList },
   { SGL_3DS_CHUNK_MAPLIST, NULL, sglLoadModel3ds::parseMapList   },
   { 0, NULL, NULL }
};  /* these chunks have to be known *before* we call parseFaceList
       (see parseTriMesh for more info) */

static sglLoadModel3ds::Chunk TriMeshChunks[] =
{
   { SGL_3DS_CHUNK_FACELIST, FaceListChunks, sglLoadModel3ds::parseFaceList },
   { SGL_3DS_CHUNK_TRMATRIX, NULL, sglLoadModel3ds::parseTraMatrix          },
   { 0, NULL, NULL }
};

static sglLoadModel3ds::Chunk ObjBlockChunks[] =
{
   { SGL_3DS_CHUNK_TRIMESH, TriMeshChunks, sglLoadModel3ds::parseTriMesh },
   { 0, NULL, NULL }
};

static sglLoadModel3ds::Chunk TextureChunks[] =
{
   { SGL_3DS_CHUNK_MAPFILENAME, NULL, sglLoadModel3ds::parseMapName      },
   { SGL_3DS_CHUNK_MAP_USCALE,  NULL, sglLoadModel3ds::parseUScale       },
   { SGL_3DS_CHUNK_MAP_VSCALE,  NULL, sglLoadModel3ds::parseVScale       },
   { SGL_3DS_CHUNK_MAP_UOFFSET, NULL, sglLoadModel3ds::parseUOffset      },
   { SGL_3DS_CHUNK_MAP_VOFFSET, NULL, sglLoadModel3ds::parseVOffset      },
   { SGL_3DS_CHUNK_MAPOPTIONS,  NULL, sglLoadModel3ds::parseMapOptions   },
   { 0, NULL, NULL }
};

static sglLoadModel3ds::Chunk MaterialColorChunks[] =
{
   { SGL_3DS_CHUNK_RGB1, NULL, sglLoadModel3ds::parseRGB1                },
   { SGL_3DS_CHUNK_RGB2, NULL, sglLoadModel3ds::parseRGB2                },
   { SGL_3DS_CHUNK_RGB3, NULL, sglLoadModel3ds::parseRGB2                },
   { 0, NULL, NULL }
};

static sglLoadModel3ds::Chunk MaterialChunks[] =
{
   { SGL_3DS_CHUNK_MATNAME, NULL,                sglLoadModel3ds::parseMaterialName  },
   { SGL_3DS_CHUNK_AMBIENT, MaterialColorChunks, sglLoadModel3ds::parseAmbient       },
   { SGL_3DS_CHUNK_DIFFUSE, MaterialColorChunks, sglLoadModel3ds::parseDiffuse       },
   { SGL_3DS_CHUNK_SPECULAR,MaterialColorChunks, sglLoadModel3ds::parseSpecular      },
   { SGL_3DS_CHUNK_SHININESS,      NULL,         sglLoadModel3ds::parseShininess     },
   { SGL_3DS_CHUNK_TRANSPARENCY,   NULL,         sglLoadModel3ds::parseTransparency  },
   { SGL_3DS_CHUNK_TRANSP_FALLOFF, NULL,         sglLoadModel3ds::parse8Bytes},
   { SGL_3DS_CHUNK_UNKNOWN_DOUBLE1,NULL,         sglLoadModel3ds::parse8Bytes},
   { SGL_3DS_CHUNK_TEXTURE,        TextureChunks,NULL               },
   { SGL_3DS_CHUNK_REFLECTIONMAP,  NULL,         sglLoadModel3ds::parseNBytes},
   { SGL_3DS_CHUNK_DOUBLESIDED,    NULL,         sglLoadModel3ds::parseDoublesided   },
   { 0, NULL, NULL }
};

static sglLoadModel3ds::Chunk ObjMeshChunks[] =
{
   { SGL_3DS_CHUNK_MATERIAL, MaterialChunks, sglLoadModel3ds::parseMaterial  },
   { SGL_3DS_CHUNK_OBJBLOCK, ObjBlockChunks, sglLoadModel3ds::parseOBJBlock  },
   { SGL_3DS_CHUNK_ONEUNIT,  NULL,           sglLoadModel3ds::parseOneUnit   },
   { 0, NULL, NULL }
};

static sglLoadModel3ds::Chunk MainChunks[] =
{
   { SGL_3DS_CHUNK_OBJMESH, ObjMeshChunks, NULL    },
   { SGL_3DS_CHUNK_VERSION, NULL,          sglLoadModel3ds::parseVersion },
   { 0, NULL, NULL }
};

static sglLoadModel3ds::Chunk TopChunk[] =
{
   { SGL_3DS_CHUNK_MAIN, MainChunks, NULL          },
   { 0, NULL, NULL }
};

//----------------------------------------------------------------------------

class sgl3dsMaterial
{
public:
   sgl3dsMaterial() :
         m_material_name(""),
         m_material(NULL),
         m_tex_name(""),
         m_texture(NULL),
         m_double_sided_flag(false),
         m_has_texture_flag(false)
      {
         m_texture_scale[0] = m_texture_scale[1] = 1.0f;
         m_texture_offset[0] = m_texture_offset[1] = 0.0f;
      }

   ~sgl3dsMaterial()
      {
         if (m_material && m_material->getRefCount() == 0)
            delete m_material;
         if (m_texture && m_texture->getRefCount() == 0)
            delete m_texture;
      }

   vector<sglStatelet *> m_statelets;

   string        m_material_name;
   sglMaterial  *m_material;

   string        m_tex_name;
   sglTexture2D *m_texture;

   bool          m_double_sided_flag;
   bool          m_has_texture_flag;

   sglVec2f      m_texture_scale;
   sglVec2f      m_texture_offset;
};


//----------------------------------------------------------------------------

static float get_float(FILE *fd)
{
   float f;
   fread( &f, 4, 1, fd );

#ifdef SGL_BIG_ENDIAN
   sglEndianSwap((unsigned int*)&f);
#endif

   return f;
}

static unsigned int get_dword(FILE *fd)
{
   unsigned int d;
   fread( &d, 4, 1, fd );

#ifdef SGL_BIG_ENDIAN
   sglEndianSwap(&d);
#endif

   return d;
}

static unsigned short get_word(FILE *fd)
{
   unsigned short w;
   fread( &w, 2, 1, fd );

#ifdef SGL_BIG_ENDIAN
   sglEndianSwap(&w);
#endif

   return w;
}

static unsigned char get_byte(FILE *fd)
{
   unsigned char b;
   fread( &b, 1, 1, fd );
   return b;
}

void get_string(FILE *fd, string &s)
{
   char read;

   while ( (read = fgetc(fd)) != 0 )
   {
      s += read;
   }
}

//----------------------------------------------------------------------------
inline void computeFaceNormal(const sglVec3f &v0,
                              const sglVec3f &v1,
                              const sglVec3f &v2,
                              sglVec3f &normal)
{
    sglVec3f a, b;

    a.sub(v1, v0);
    b.sub(v2, v0);

    normal.cross(a, b);
    normal.normalize();
}

//===========================================================================

sglLoadModel3ds::sglLoadModel3ds() : //sglLoadModel(),
      m_current_material(NULL),
      m_double_sided_flag(false),
      //m_num_objects(0),
      //m_num_materials(0),
      //m_num_textures(0),
      m_top_object(NULL),
      m_vertex_index(NULL),
      //m_normal_index(NULL),
      //m_num_vertices(0),
      m_num_faces(0),
      m_smooth_list(NULL),
      m_current_transform(NULL),
      m_vertex_list(NULL),
      m_face_normals(NULL),
      m_vertex_normals(NULL),
      m_texcrd_list(NULL),
      m_smooth_found(false),
      m_facemat_found(false),
      m_color_mode(GL_DIFFUSE)
{
}

// --------------------------------------------------------------------------
sglLoadModel3ds::~sglLoadModel3ds()
{
}

// --------------------------------------------------------------------------
bool sglLoadModel3ds::parseChunks(FILE *fd,
                                  unsigned int length,
                                  Chunk *chunk_list)
{
   bool parse_ok = true;
   unsigned short id;
   unsigned int p = 0;
   unsigned int sub_length;
   Chunk *t;

   while (parse_ok && p < length)
   {
      id = get_word(fd);
      sub_length = get_dword(fd);

      if (p + sub_length > length)
      {
         sglPrintWarning() << "Illegal chunk 0x" << hex << id << dec
                           << " of length " << sub_length
                           << ". Chunk is longer than parent chunk." << endl;
         return false;
      }

      p += sub_length;
      sub_length -= SGL_3DS_CHUNK_HEADER_SIZE;

      for (t = chunk_list; t->id != 0 && t->id != id; t++);

      if (t->id == id)
      {
         // this is a known chunk

         // do chunk-specific parsing if available
         unsigned long cp = ftell(fd);

         if (t->parse_func)
            parse_ok = t->parse_func(*this, fd, sub_length );

         // if chunk can have subchunks, parse these
         if (t->subchunks && parse_ok)
         {
            parse_ok = parseChunks( fd, sub_length - (ftell(fd)-cp),
                                    t->subchunks );
         }
      }
      else
      {
         // unknown chunk
         sglPrintInfo() << "Skipping unknown chunk 0x" << hex << id << dec
                        << " of length " << sub_length << "." << endl;

#if 0
//#ifdef SGL_3DS_DEBUG
         for (unsigned int i=0; i<sub_length; i++)
         {
            if ((i%4) == 0)
            {
               //sglPrintDebug() << "|0x";
            }
            unsigned char d = get_byte(fd);
            if (d < 16)
            {
               //sglPrintDebug() << "0";
            }
            //sglPrintDebug() << hex << (int)d << dec;
         }
         //sglPrintDebug() << endl;
#else
         fseek(fd, (long)sub_length, SEEK_CUR);
#endif
      }
   }

   return parse_ok;
}

// --------------------------------------------------------------------------
sglNode* sglLoadModel3ds::load(const string &filename)
{
   // need to do something about a default path list here

   FILE *loader_fd = fopen(filename.c_str(), "rb");

   if ( loader_fd == NULL )
   {
      sglPrintWarning() << "ssgLoad3ds: Failed to open '" << filename
                        << "' for reading" << endl;
      return NULL;
   }

   // Get file length and load into buffer
   fseek(loader_fd, 0, SEEK_END);
   unsigned int size = (unsigned int)ftell(loader_fd);
   rewind(loader_fd);

   //m_num_objects = m_num_materials = m_num_textures = 0;
   m_vertex_list = NULL;
   m_texcrd_list = NULL;
   m_face_normals = NULL;
   m_vertex_index = NULL;
   //m_normal_index = NULL;
   m_top_object = new sglGroup();

   m_double_sided_flag = false;

   parseChunks( loader_fd, size, TopChunk );

   fclose ( loader_fd ) ;

   while (m_material_list.size())
   {
      sgl3dsMaterial *mat = m_material_list.back();
      m_material_list.pop_back();
      delete mat;
   }

   freeTrimesh();

   return m_top_object;
}

// --------------------------------------------------------------------------
// MATERIAL PARSERS

// load a texture (called a map)
bool sglLoadModel3ds::parseMapName( sglLoadModel3ds &ldr,
                                    FILE *fd, unsigned int /*length*/ )
{
   // DO SOMETHING ABOUT PREPENDING TEX PATH NAME
   string tex_filename;
   get_string(fd, tex_filename);
   //sglPrintDebug() << "Texture filename: " << tex_filename << endl;

   // find a material with the same texture
   // since texture path is in common, just store/compare texture filename
   sglTexture2D *tex = NULL;
   for (vector<sgl3dsMaterial*>::iterator mat = ldr.m_material_list.begin();
        mat != ldr.m_material_list.end();
        ++mat)
   {
      if ( (*mat)->m_tex_name.length() && (*mat)->m_tex_name == tex_filename )
      {
         tex = (*mat)->m_texture;
         break;
      }
   }

   if ( tex == NULL )
   {
      // didn't find a texture so try to create a new one
      if (tex_filename.size())
      {
         string foundname;
         if (sglDataLoader::findFile(tex_filename, NULL, foundname))
         {
            // load texture here
            sglTexture::ExternalFormatEnum format;
            sglTexture::TypeEnum type;
            unsigned int num_s, num_t;
            void *image_data = sglImageLoader::loadImage(
               foundname, num_s, num_t, format, type);

            if (image_data)
            {
               tex = new sgl3dsTexture();
               tex->setImage(image_data, format, type, num_s, num_t, false);
               ldr.m_current_material->m_texture = tex;
               ldr.m_current_material->m_statelets.push_back(tex);
            }
         }
         if (tex == NULL)
         {
            sglPrintWarning() << "sglLoadModel3ds::parseMapName warning: "
                              << "unable to load texture." << endl;
         }
      }
   }
   else
   {
      ldr.m_current_material->m_texture = tex;
   }

   if (ldr.m_current_material->m_texture)
   {
      ldr.m_current_material->m_tex_name = tex_filename;
      ldr.m_current_material->m_has_texture_flag = true;
   }
   return true;
}

// --------------------------------------------------------------------------
bool sglLoadModel3ds::parseMapOptions( sglLoadModel3ds &ldr,
                                       FILE *fd, unsigned int /*length*/ )
{
   unsigned short value = get_word(fd);
   // bit 4: 0=tile (default), 1=do not tile (a single bit for both u and v)
   bool tile = ((value & 0x10) == 0);

   //GLuint handle = materials[num_materials-1]->getTextureHandle();
   sglTexture2D *tex = ldr.m_current_material->m_texture;

   if (tex)
   {
      tex->setWrapMode(tile? sglTexture::eREPEAT : sglTexture::eCLAMP_TO_EDGE,
                       tile? sglTexture::eREPEAT : sglTexture::eCLAMP_TO_EDGE);

      // IF I DO CLAMP TO EDGE HERE, DO I NEED TO DO DIFFUSE BORDER COLOR
      // BELOW?

      if (ldr.m_current_material->m_material)
      {
         float r, g, b;
         ldr.m_current_material->m_material->getFrontDiffuse(r, g, b);
         sglVec4f diff_rgb(r, g, b, 1.0f);
         tex->setBorderColor(diff_rgb);
      }
   }
   else
   {
      sglPrintWarning() << "sglLoadModel3ds::parseMapOptions error: "
                        << "m_current_material does not have a texture"
                        << endl;
   }
   return true;
}

// --------------------------------------------------------------------------
bool sglLoadModel3ds::parseUScale( sglLoadModel3ds &ldr,
                                   FILE *fd, unsigned int /*length*/ )
{
   ldr.m_current_material->m_texture_scale[1] = get_float(fd);
   return true;
}

// --------------------------------------------------------------------------
bool sglLoadModel3ds::parseVScale( sglLoadModel3ds &ldr,
                                   FILE *fd, unsigned int /*length*/ )
{
   ldr.m_current_material->m_texture_scale[0] = get_float(fd);
   return true;
}

// --------------------------------------------------------------------------
bool sglLoadModel3ds::parseUOffset( sglLoadModel3ds &ldr,
                                    FILE *fd, unsigned int /*length*/ )
{
   ldr.m_current_material->m_texture_offset[1] = get_float(fd);
   return true;
}

// --------------------------------------------------------------------------
bool sglLoadModel3ds::parseVOffset( sglLoadModel3ds &ldr,
                                    FILE *fd, unsigned int /*length*/ )
{
   ldr.m_current_material->m_texture_offset[0] = get_float(fd);
   return true;
}

// --------------------------------------------------------------------------
bool sglLoadModel3ds::parseMaterial( sglLoadModel3ds &ldr,
                                     FILE* /*fd*/, unsigned int /*length*/ )
{
   //materials[num_materials] = m_current_material = new ssgSimpleState();
   //mat_flag[num_materials] = 0;
   //num_materials++;
   ldr.m_current_material = new sgl3dsMaterial;
   ldr.m_material_list.push_back(ldr.m_current_material);

   // necessary??
   ldr.m_current_material->m_statelets.push_back(new sglLighting(true));
   ldr.m_current_material->m_statelets.push_back(
      new sglCullMode(sglCullMode::eBACK));

   sglMaterial *mat = ldr.m_current_material->m_material = new sglMaterial;
   ldr.m_current_material->m_statelets.push_back(mat);

   //sglPrintDebug() << "New material found." << endl;

   return true;
}

// --------------------------------------------------------------------------
bool sglLoadModel3ds::parseMaterialName( sglLoadModel3ds &ldr,
                                         FILE *fd, unsigned int /*length*/ )
{
   get_string(fd, ldr.m_current_material->m_material_name);
   ldr.m_current_material->m_material->setName(
      ldr.m_current_material->m_material_name.c_str());
   //sglPrintDebug() << "Material name: "
   //                << ldr.m_current_material->m_material_name << endl;
   return true;
}

// --------------------------------------------------------------------------
bool sglLoadModel3ds::parseRGB1( sglLoadModel3ds &ldr,
                                 FILE *fd, unsigned int /*length*/ )
{
   float r = get_float(fd);
   float g = get_float(fd);
   float b = get_float(fd);

   //sglPrintDebug() << "Color: " << r << ", " << g << ", " << b << endl;

   switch(ldr.m_color_mode)
   {
      case GL_SPECULAR:
         ldr.m_current_material->m_material->setSpecular(r, g, b);
         break;
      case GL_AMBIENT:
         ldr.m_current_material->m_material->setAmbient(r, g, b);
         break;
      case GL_DIFFUSE:
      default:
         ldr.m_current_material->m_material->setDiffuse(r, g, b);
         break;
   }

   return true;
}

// --------------------------------------------------------------------------
bool sglLoadModel3ds::parseRGB2( sglLoadModel3ds &ldr,
                                 FILE *fd, unsigned int /*length*/ )
{
   unsigned char r = get_byte(fd);
   unsigned char g = get_byte(fd);
   unsigned char b = get_byte(fd);

//     sglPrintDebug() << "Color: " << hex << (short)r << (short)g << (short)b
//                     << dec << endl;
//     sglPrintDebug() << "       " << (float)r/255.f << ", " << (float)g/255.f
//                     << ", " << (float)b/255.f << endl;

   switch(ldr.m_color_mode)
   {
      case GL_SPECULAR:
         ldr.m_current_material->m_material->setSpecular((float)r/255.f,
                                                         (float)g/255.f,
                                                         (float)b/255.f);
         break;
      case GL_AMBIENT:
         ldr.m_current_material->m_material->setAmbient((float)r/255.f,
                                                        (float)g/255.f,
                                                        (float)b/255.f);
         break;
      case GL_DIFFUSE:
      default:
         ldr.m_current_material->m_material->setDiffuse((float)r/255.f,
                                                        (float)g/255.f,
                                                        (float)b/255.f);
         break;
   }

   return true;
}

// --------------------------------------------------------------------------
bool sglLoadModel3ds::parseAmbient( sglLoadModel3ds &ldr,
                                    FILE* /*fd*/, unsigned int /*length*/ )
{
   ldr.m_color_mode = GL_AMBIENT;
   //sglPrintDebug() << "Ambient color" << endl;
   return true;
}

// --------------------------------------------------------------------------
bool sglLoadModel3ds::parseDiffuse( sglLoadModel3ds &ldr,
                                    FILE* /*fd*/, unsigned int /*length*/ )
{
   ldr.m_color_mode = GL_DIFFUSE;
   //sglPrintDebug() << "Diffuse color" << endl;
   return true;
}

// --------------------------------------------------------------------------
bool sglLoadModel3ds::parseSpecular( sglLoadModel3ds &ldr,
                                     FILE* /*fd*/, unsigned int /*length*/ )
{
   ldr.m_color_mode = GL_SPECULAR;
   //sglPrintDebug() << "Specular color" << endl;
   return true;
}

// --------------------------------------------------------------------------
bool sglLoadModel3ds::parseShininess( sglLoadModel3ds &ldr,
                                      FILE *fd, unsigned int /*length*/ )
{
   // this chunk contains a percentage chunk,
   // so just read that chunk's header
   get_word(fd); get_dword(fd);

   float shininess = (float)get_word(fd)*128.0f/100.0f;
   ldr.m_current_material->m_material->setShininess(shininess);
   //sglPrintDebug() << "Shininess: " << shininess << endl;
   return true;
}

// --------------------------------------------------------------------------
bool sglLoadModel3ds::parseTransparency( sglLoadModel3ds &ldr,
                                         FILE *fd, unsigned int /*length*/ )
{
   // this chunk contains a percentage chunk,
   // so just read that chunk's header
//   unsigned short s1 = get_word(fd);
//   unsigned short s2 = get_dword(fd);
   //sglPrintDebug() << "Unknown 4 bytes: " << hex << s1 << "|" << s2 << dec << endl;

   float alpha = 1.0f - (float)get_word(fd) / 100.0f;
   //sglPrintDebug() << "Alpha: " << alpha << endl;
   if (alpha <= 0.99f)
   {
      ldr.m_current_material->m_material->setAlpha(alpha);
      ldr.m_current_material->m_statelets.push_back(new sglTransparency(true));
   }
//   else
//   {
//      ldr.m_current_material->m_material->setAlpha(1.0f);
//      ldr.m_current_material->m_statelets.push_back(
//                                                   new sglTransparency(false));
//   }

   return true;
}

// --------------------------------------------------------------------------
bool sglLoadModel3ds::parse8Bytes( sglLoadModel3ds &/*ldr*/,
                                   FILE *fd, unsigned int /*length*/ )
{
//   unsigned int f1 = get_dword(fd);
//   unsigned int f2 = get_dword(fd);
   //sglPrintDebug() << "Unknown 8 bytes: " << f1 << ", " << f2
   //                << " -> " << hex << f1 << "|" << f2 << dec << endl;
   return true;
}

// --------------------------------------------------------------------------
bool sglLoadModel3ds::parseNBytes( sglLoadModel3ds &/*ldr*/,
                                   FILE *fd, unsigned int length )
{
   //sglPrintDebug() << "I think I have found a reflection texture chunk: "
   //                << endl;

   for (unsigned int i=0; i<length; i++)
   {
      if ((i%4) == 0)
      {
         //sglPrintDebug() << "|0x";
      }
      unsigned char d = get_byte(fd);
      if (d < 16)
      {
         //sglPrintDebug() << "0";
      }
      //sglPrintDebug() << hex << (int)d << dec;
   }
   //sglPrintDebug() << endl;

   return true;
}

// --------------------------------------------------------------------------
bool sglLoadModel3ds::parseDoublesided( sglLoadModel3ds &ldr,
                                        FILE* /*fd*/, unsigned int /*length*/ )
{
   ldr.m_current_material->m_double_sided_flag = true;

   //sglPrintDebug() << "Warning: '" << ldr.m_current_material->m_material_name
   //                << "' is doublesided." << endl;

   return true;
}

// --------------------------------------------------------------------------
// TRIMESH PARSERS
// --------------------------------------------------------------------------

void sglLoadModel3ds::freeTrimesh()
{
   if (m_vertex_list)
      delete [] m_vertex_list;

   if (m_face_normals)
      delete [] m_face_normals;

   if (m_vertex_normals)
      delete [] m_vertex_normals;

   if (m_texcrd_list)
      delete [] m_texcrd_list;

   if (m_smooth_list)
      delete [] m_smooth_list;

   if (m_vertex_index)
      delete [] m_vertex_index;

   m_vertex_list = NULL;
   m_face_normals = NULL;
   m_vertex_normals = NULL;
   m_texcrd_list = NULL;
   m_smooth_list = NULL;
   m_vertex_index = NULL;
}

// --------------------------------------------------------------------------
bool sglLoadModel3ds::parseTriMesh( sglLoadModel3ds &ldr,
                                    FILE *fd, unsigned int length )
{
   ldr.m_current_transform = new sglTransformf();

   ldr.freeTrimesh();

   ldr.m_top_object->addChild(ldr.m_current_transform);

   /* Before we parse SGL_3DS_CHUNK_FACEMAT, we have to know vertices and
      texture coordinates. To ensure this, we make a special pass of the
      Trimesh chunks, just extracting this information.  This is kind of a
      kludge, but it was the easiest way to solve this problem */
   //sglPrintDebug() << "Prescanning sub_chunks for vertices" << endl;
   long p = ftell(fd);
   bool success =  ldr.parseChunks( fd, length, TriMeshDataChunks );
   fseek(fd, p, SEEK_SET);
   //sglPrintDebug() << "Done prescanning" << endl;
   return success;
}

// --------------------------------------------------------------------------
bool sglLoadModel3ds::parseVertList( sglLoadModel3ds &ldr,
                                     FILE *fd, unsigned int /*length*/ )
{
   unsigned short num_vertices = get_word(fd);
   ldr.m_vertex_list = new sglVec3f[num_vertices];

   for (unsigned int i = 0; i < num_vertices; i++)
   {
      ldr.m_vertex_list[i][0] = get_float(fd);
      ldr.m_vertex_list[i][1] = get_float(fd);
      ldr.m_vertex_list[i][2] = get_float(fd);
   }

   return true;
}

// --------------------------------------------------------------------------
bool sglLoadModel3ds::parseSmoothList( sglLoadModel3ds &ldr,
                                       FILE *fd, unsigned int /*length*/ )
{
   ldr.m_smooth_found = true;

   ldr.m_smooth_list = new unsigned int[ldr.m_num_faces];

   for (unsigned int i = 0; i < ldr.m_num_faces; i++)
      ldr.m_smooth_list[i] = get_dword(fd);

   return true;
}

// --------------------------------------------------------------------------
void sglLoadModel3ds::smoothNormals(bool use_smooth_list)
{
   for (unsigned int i = 0; i < m_num_faces; i++)
   {
      unsigned int v1 = i * 3;
      unsigned int v2 = i * 3 + 1;
      unsigned int v3 = i * 3 + 2;

      m_vertex_normals[v1].set(0.f,0.f,0.f);
      m_vertex_normals[v2].set(0.f,0.f,0.f);
      m_vertex_normals[v3].set(0.f,0.f,0.f);

      for (unsigned int j = 0; j < m_num_faces; j++)
      {
         bool should_smooth;

         if (use_smooth_list)
         {
            should_smooth = (m_smooth_list[i] & m_smooth_list[j]) &&
               (m_face_normals[i].dot(m_face_normals[j]) >= -0.5f);
         }
         else
         {
            should_smooth = ((m_face_normals[i].dot(m_face_normals[j])) >
                             sgl_smooth_threshold);
         }

         if (should_smooth)
         {
            // now we have to know if the faces shares vertices
            for (unsigned int a = v1; a < v1 + 3; a++)
            {
               for (unsigned int b = j * 3; b < j * 3 + 3; b++)
               {
                  if (m_vertex_index[a] == m_vertex_index[b])
                  {
                     m_vertex_normals[a] += m_face_normals[j];
                  }
               }
            }
         }
      }

      m_vertex_normals[v1].normalize();
      m_vertex_normals[v2].normalize();
      m_vertex_normals[v3].normalize();
   }
}

//----------------------------------------------------------------------------
bool sglLoadModel3ds::identifyFaceMaterials(sglLoadModel3ds &ldr,
                                            FILE *fd, unsigned int length)
{
   ldr.m_facemat_found = true;
   //sglPrintDebug() << "Face materials found" << endl;
   fseek(fd, (long)length, SEEK_CUR );
   return true;
}

// --------------------------------------------------------------------------
bool sglLoadModel3ds::parseFaceList( sglLoadModel3ds &ldr,
                                     FILE *fd, unsigned int length )
{
   ldr.m_num_faces = get_word(fd);

   ldr.m_vertex_index   = new unsigned int[ldr.m_num_faces*3];
   ldr.m_face_normals   = new sglVec3f[ldr.m_num_faces];

   ldr.m_vertex_normals = new sglVec3f[ldr.m_num_faces * 3    ];

   for (unsigned int i = 0; i < ldr.m_num_faces; i++)
   {
      ldr.m_vertex_index[i*3    ] = get_word(fd);
      ldr.m_vertex_index[i*3 + 1] = get_word(fd);
      ldr.m_vertex_index[i*3 + 2] = get_word(fd);
      unsigned short flags  = get_word(fd);

      if (flags & 7 == 0)     // Triangle vertices order should be swapped
      {
         unsigned int tmp            = ldr.m_vertex_index[i*3 + 1];
         ldr.m_vertex_index[i*3 + 1] = ldr.m_vertex_index[i*3 + 2];
         ldr.m_vertex_index[i*3 + 2] = tmp;
      }

      computeFaceNormal(ldr.m_vertex_list[ldr.m_vertex_index[i*3    ]],
                        ldr.m_vertex_list[ldr.m_vertex_index[i*3 + 1]],
                        ldr.m_vertex_list[ldr.m_vertex_index[i*3 + 2]],
                        ldr.m_face_normals[i]);
   }

   /* this is a special "hack" for the face list chunk
      because we HAVE to know the smooth list (if there is one)
      before building the ssgVtxTable objects, this parsing has to
      be done first...*ugh*/

   ldr.m_smooth_found = false;
   ldr.m_facemat_found = false;

   //sglPrintDebug() << "Prescanning sub-chunks for smooth lists" << endl;
   long p = ftell(fd);

   ldr.parseChunks( fd, length - (2 + 8*ldr.m_num_faces),
                    FaceListDataChunks );
   fseek(fd, p, SEEK_SET);
   //sglPrintDebug() << "Done prescanning." << endl;

   /* now apply correct smoothing. If smooth list has been found,
      use it, otherwise use threshold value. */
   ldr.smoothNormals( ldr.m_smooth_found );

   if (!ldr.m_facemat_found && ldr.m_material_list.size())
   {
      // no FACEMAT chunk found. Adding default faces of materials
      unsigned short *face_indices = new unsigned short[ldr.m_num_faces];
      for (unsigned int i=0; i<ldr.m_num_faces; ++i)
         face_indices[i] = i;

      addLeaf(ldr, ldr.m_material_list[0], ldr.m_num_faces, face_indices);

      delete [] face_indices; // CORRECT????
   }

   return true;
}

// --------------------------------------------------------------------------
bool sglLoadModel3ds::parseMapList( sglLoadModel3ds &ldr,
                                    FILE *fd, unsigned int /*length*/ )
{
   unsigned short num_v = get_word(fd);
   ldr.m_texcrd_list = new sglVec2f[num_v];

   for (unsigned int i = 0; i < num_v; i++)
   {
      ldr.m_texcrd_list[i][0] = get_float(fd);
      ldr.m_texcrd_list[i][1] = get_float(fd);
   }

   return true;
}

// --------------------------------------------------------------------------
bool sglLoadModel3ds::parseTraMatrix( sglLoadModel3ds &ldr,
                                      FILE *fd, unsigned int /*length*/ )
{
   sglMat4f mat;
   mat.buildIdentity();

   /* Strange things seems to be going on with the
      local coordinate system in 3ds - I have commented
      this out, but things seems to work better without
      it (which is odd). */

   int i, j;

   //sglPrintDebug() << "Parsing transformation matrix: "
   //                << ldr.m_current_transform <<endl;

   for (i = 0; i < 4; i++)
   {
      for (j = 0; j < 3; j++)
      {
         mat[i][j] = get_float(fd);
         //sglPrintDebug() << mat[i][j] << ", ";
      }
      //double norm = sqrt(mat[i][0]*mat[i][0] +
      //                   mat[i][1]*mat[i][1] +
      //                   mat[i][2]*mat[i][2]);
      //sglPrintDebug() << norm << endl;
   }

   mat[3][3] = 1.0f; // redundant?
   // sgTransposeNegateMat4(m)

   //sglMat4f mat2;
   //mat2.inverseAffine(mat);
   //ldr.m_current_transform->setMatrix(mat2);

   return true;
}

// --------------------------------------------------------------------------
void sglLoadModel3ds::addLeaf(sglLoadModel3ds &ldr,
                              sgl3dsMaterial *material, int listed_faces,
                              unsigned short *face_indices)
{
   bool is_ds = material->m_double_sided_flag;
   bool has_texture = material->m_has_texture_flag;
   int num_allocate = listed_faces * 3;
   if (is_ds)
      num_allocate *= 2;

   sglVec3f *vertices = new sglVec3f[ num_allocate ];
   sglVec3f *normals  = new sglVec3f[ num_allocate ];
   sglVec2f *texcrds  = NULL;

   if (has_texture)
   {
      if (ldr.m_texcrd_list == NULL)
      {
         sglPrintWarning() << "sglLoadModel3ds warning: "
                           << "Texture coords seems to be missing." << endl;
      }
      else
      {
         texcrds = new sglVec2f[ num_allocate ];
      }
   }

   int vertex_count = 0;
   int normal_count = 0;
   int texcrd_count = 0;

   for (int i = 0; i < listed_faces; i++)
   {
      unsigned short faceindex = face_indices[i];
      int v1 = faceindex * 3;
      int v2 = faceindex * 3 + 1;
      int v3 = faceindex * 3 + 2;

      vertices[vertex_count++] = ldr.m_vertex_list[ ldr.m_vertex_index[v1] ];
      vertices[vertex_count++] = ldr.m_vertex_list[ ldr.m_vertex_index[v2] ];
      vertices[vertex_count++] = ldr.m_vertex_list[ ldr.m_vertex_index[v3] ];

      normals[normal_count++] = ldr.m_vertex_normals[ v1 ];
      normals[normal_count++] = ldr.m_vertex_normals[ v2 ];
      normals[normal_count++] = ldr.m_vertex_normals[ v3 ];

      if (has_texture && ldr.m_texcrd_list != NULL)
      {
         int num_texcrds = 3;

         texcrds[texcrd_count  ] = ldr.m_texcrd_list[ ldr.m_vertex_index[v1] ];
         texcrds[texcrd_count+1] = ldr.m_texcrd_list[ ldr.m_vertex_index[v2] ];
         texcrds[texcrd_count+2] = ldr.m_texcrd_list[ ldr.m_vertex_index[v3] ];

         if (is_ds)
         {
            num_texcrds = 6;
            texcrds[texcrd_count+3]=ldr.m_texcrd_list[ldr.m_vertex_index[v1] ];
            texcrds[texcrd_count+4]=ldr.m_texcrd_list[ldr.m_vertex_index[v3] ];
            texcrds[texcrd_count+5]=ldr.m_texcrd_list[ldr.m_vertex_index[v2] ];
         }

         for (int j = 0; j < num_texcrds; j++)
         {
            texcrds[texcrd_count + j][0] *= material->m_texture_scale[0];
            texcrds[texcrd_count + j][1] *= material->m_texture_scale[1];
            texcrds[texcrd_count + j][0] += material->m_texture_offset[0];
            texcrds[texcrd_count + j][1] += material->m_texture_offset[1];
         }

         texcrd_count += num_texcrds;
      }

      if (is_ds)
      {
         vertices[vertex_count++] = ldr.m_vertex_list[ldr.m_vertex_index[v1] ];
         vertices[vertex_count++] = ldr.m_vertex_list[ldr.m_vertex_index[v3] ];
         vertices[vertex_count++] = ldr.m_vertex_list[ldr.m_vertex_index[v2] ];

         sglVec3f n;
         n = ldr.m_vertex_normals[v1];
         n *= -1.f;
         normals[normal_count++] = n;
         n = ldr.m_vertex_normals[v3];
         n *= -1.f;
         normals[normal_count++] = n;
         n = ldr.m_vertex_normals[v2];
         n *= -1.f;
         normals[normal_count++] = n;
      }
   }

   // DO SOMETHING WITH THIS
//     sglPrintDebug() << "vertex count = " << vertex_count
//                     << ", normal count = " << normal_count
//                     << ", texcrd_count = " << texcrd_count << endl;

   sglTriangleSet *gset = new sglTriangleSet;
   gset->setNumPrims(vertex_count/3);
   gset->setCoordList(new sglDeleteableDataPtr<sglVec3f>(vertices));
   if (normal_count)
      gset->setNormalList(sglGeoSet::ePER_VERTEX,
                          new sglDeleteableDataPtr<sglVec3f>(normals));

   //getState(material);

   if (texcrd_count)
      gset->setState(0, material->m_statelets,
                     new sglDeleteableDataPtr<sglVec2f>(texcrds));
   else
      gset->setState(0, material->m_statelets);

   gset->setName((material->m_material_name).c_str());

   sglGeode *geode = new sglGeode;
   geode->addGeometry(gset);

   ldr.m_current_transform->addChild(geode);
}

// --------------------------------------------------------------------------
bool sglLoadModel3ds::parseFaceMaterials(sglLoadModel3ds &ldr,
                                         FILE *fd, unsigned int /*length*/)
{
   string material_name;
   get_string(fd, material_name);

   sgl3dsMaterial *material = NULL;

   // find the material
   for (vector<sgl3dsMaterial*>::iterator elt = ldr.m_material_list.begin();
        elt != ldr.m_material_list.end();
        ++elt)
   {
      if (material_name == (*elt)->m_material_name)
      {
         material = *elt;
         break;
      }
   }

   // if we didn't find a material ERROR
   if (material == NULL)
   {
      sglPrintWarning() << "sglLoadModel3ds::parseFaceMaterials error: "
                        << " undefined reference to material "
                        << material_name << " found." << endl;
      return false;
   }

   unsigned short listed_faces = get_word(fd);

   unsigned short *face_indices = new unsigned short[listed_faces];
   for (unsigned int i=0; i<listed_faces; ++i)
      face_indices[i] = get_word(fd);

   addLeaf(ldr, material, listed_faces, face_indices);
   delete [] face_indices;

   return true;
}

// --------------------------------------------------------------------------
// OBJBLOCK PARSER
// --------------------------------------------------------------------------
bool sglLoadModel3ds::parseOBJBlock( sglLoadModel3ds &/*ldr*/,
                                     FILE *fd, unsigned int /*length*/ )
{
   string name;
   get_string(fd, name);
   //sglPrintDebug() << "Object block: " << name << endl;

   return true;
}

bool sglLoadModel3ds::parseOneUnit( sglLoadModel3ds &/*ldr*/,
                                    FILE *fd, unsigned int /*length*/ )
{
   //float one_unit = get_float(fd);
   //sglPrintDebug() << "One unit: " << one_unit << endl;

   return true;
}

//----------------------------------------------------------------------------
bool sglLoadModel3ds::parseVersion(sglLoadModel3ds &/*ldr*/,
                                   FILE *fd, unsigned int /*length*/ )
{
   //unsigned int version = get_dword(fd);
   //sglPrintDebug() << "3DS version: " << version << endl;

   return true;
}
