/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglCone.cpp
 *   Author: Scott McMillan
 *  Created: 23 January 2001
 *****************************************************************************/

#include <sglCone.hpp>
#include <sglLineSet.hpp>

//----------------------------------------------------------------------------
sglCone::sglCone(unsigned int num_sections, float height, float radius)
      : sglTriangleStripSet(),
        m_num_sections(3),
        m_height(0.f),
        m_radius(0.f)
{
   if (num_sections > 3) m_num_sections = num_sections;
   if (height >  0.f)  m_height = height;
   if (radius > 0.f)   m_radius = radius;

   makeCone(num_sections, m_height, m_radius);
}

//----------------------------------------------------------------------------
sglCone::~sglCone()
{
}

//----------------------------------------------------------------------------
void sglCone::makeCone(unsigned int num_sections, float height, float radius)
{
   num_sections++; // add one to "close" the strips;

   sglIndexType *length = new sglIndexType[2];
   length[0] = length[1] = 2*num_sections;

   sglVec3f *vertex = new sglVec3f[4*num_sections];
   sglVec3f *normal = new sglVec3f[4*num_sections];

   double psi = atan2(radius, height);
   float cp = (float)cos(psi);
   float sp = (float)sin(psi);

   unsigned int index = 0;

   for (unsigned int ix = 0; ix < num_sections; ++ix)
   {
      double theta = (double)ix*M_PI*2.0/((double)num_sections - 1.0);
      float ct = (float) cos(theta);
      float st = (float) sin(theta);

      // the base
      vertex[index].set(0.f, 0.f, 0.f);
      normal[index].set(0.f, 0.f, -1.f);

      // the side
      vertex[index+2*num_sections].set(radius*st, radius*ct, 0.f);
      normal[index+2*num_sections].set(st*cp, ct*cp, sp);

      // the base
      vertex[index+1].set(radius*st, radius*ct, 0.f);
      normal[index+1].set(0.f, 0.f, -1.f);

      // the side
      vertex[index+2*num_sections+1].set(0.f, 0.f, height);
      normal[index+2*num_sections+1].set(st*cp, ct*cp, sp);
      index += 2;
   }

   setNumPrims(2);
   setPrimLengths(new sglDeleteableDataPtr<sglIndexType>(length));
   setCoordList(new sglDeleteableDataPtr<sglVec3f>(vertex));
   setNormalList(sglGeoSet::ePER_VERTEX,
                 new sglDeleteableDataPtr<sglVec3f>(normal));
}

//----------------------------------------------------------------------------
void sglCone::printInfo(ostream &ostrm, const char *indent_string) const
{
   sglGeoSet::printInfo(ostrm, indent_string);
   ostrm << indent_string << "num_sections: " << m_num_sections << endl;
   ostrm << indent_string << "height: " << m_height << endl;
   ostrm << indent_string << "radius: " << m_radius << endl;
}

