/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglCylinder.cpp
 *   Author: John K. Hammond
 *  Created: 1 October 2001
 *****************************************************************************/

#include <sglCylinder.hpp>

//----------------------------------------------------------------------------
sglCylinder::sglCylinder(unsigned int num_slices, unsigned int num_stacks,
                         float height, float base_radius, float top_radius,
                         bool end_caps_flag)
      : sglTriangleStripSet(),
        m_num_slices(3),
        m_num_stacks(1),
        m_base_radius(1.0f),
        m_top_radius(1.0f),
        m_height(1.f),
        m_end_caps_flag(end_caps_flag)
{
   if (num_slices > 3)      m_num_slices = num_slices;
   if (num_stacks > 1)      m_num_stacks = num_stacks;
   if (base_radius > 0.0f)  m_base_radius = base_radius;
   if (top_radius >= 0.0f)  m_top_radius = top_radius;
   if (height > 0.0f)       m_height  = height;

   makeCylinder();
}

//----------------------------------------------------------------------------
sglCylinder::~sglCylinder()
{
}

//----------------------------------------------------------------------------
void sglCylinder::makeCylinder()
{
   unsigned int num_caps = 0;
   if (m_end_caps_flag)
   {
      ++num_caps;    // for the bottom cap
      if (m_top_radius > 0.0f)
      {
         ++num_caps; // for the top cap
      }
   }

   // allocate the arrays
   unsigned int num_vertices =
      (m_num_stacks*(m_num_slices + 1))*2 + num_caps*m_num_slices;

   sglIndexType *length  = new sglIndexType[m_num_stacks + num_caps];
   sglVec3f *vertex  = new sglVec3f[num_vertices];
   sglVec3f *normal  = new sglVec3f[num_vertices];

   for (unsigned int ii=0; ii < m_num_stacks; ii++)
      length[ii] = m_num_slices*2 + 2;
   for (unsigned int ij=0; ij < num_caps; ij++)
      length[m_num_stacks + ij] = m_num_slices;

   // build the side
   float da = 2.0*M_PI/(float)m_num_slices;
   float dr = (m_top_radius - m_base_radius)/(float)m_num_stacks;
   float dz = m_height/(float)m_num_stacks;
   float dt = 1.0 / m_num_stacks;

   float nz = (m_base_radius - m_top_radius)/m_height; // Z component of normal

   unsigned int index = 0;

   for (unsigned int j = 0; j < m_num_stacks; j++)
   {
      float z = j*dz;
      float t = j*dt;
      float r = m_base_radius + j*dr;

      for (unsigned int i = 0; i <= m_num_slices; i++)
      {
         double angle = (i == m_num_slices) ? 0.0 : ((double) i*da);

         float x = sin(angle);
         float y = cos(angle);

         normal[index].set (x, y, nz);
         normal[index].normalize ();
         vertex[index].set ((float) (x * r), (float) (y * r), (float) z);
         index++;

         normal[index].set (x, y, nz);
         normal[index].normalize ();
         vertex[index].set (x * (r + dr), y * (r + dr), z + dz);
         index++;
      } // slices
      r += dr;
      t += dt;
      z += dz;
   } // stacks

   if (num_caps)
   {
      unsigned int v1 = m_num_slices/2 - 1;
      unsigned int v2 = m_num_slices/2;

      // create the bottom cap
      for (unsigned int i = 0; i < m_num_slices/2; i++)
      {
         // i = m_num_slices/2 - 1, m_num_slices - 2, ...
         double angle = (double)(v1-i)*da;
         float x = sin(angle);
         float y = cos(angle);
         normal[index].set(0.f, 0.f, -1.f);
         vertex[index].set(x*m_base_radius, y*m_base_radius, 0.f);
         ++index;

         // i = m_num_slices/2, m_num_slices/2 + 1, ...
         angle = (double)(v2+i)*da;
         x = sin(angle);
         y = cos(angle);
         normal[index].set(0.f, 0.f, -1.f);
         vertex[index].set(x*m_base_radius, y*m_base_radius, 0.f);
         ++index;
      }
      if (m_num_slices > 2*(m_num_slices/2))
      {
         float x = sin((m_num_slices - 1)*da);
         float y = cos((m_num_slices - 1)*da);
         normal[index].set(0.f, 0.f, -1.f);
         vertex[index].set(x*m_base_radius, y*m_base_radius, 0.f);
         ++index;
      }

      if (num_caps > 1)
      {
         unsigned int v1 = m_num_slices/2 - 1;
         unsigned int v2 = m_num_slices/2;

         // create the top cap
         for (unsigned int i = 0; i < m_num_slices/2; i++)
         {
            // i = m_num_slices/2, m_num_slices/2 + 1, ...
            double angle = (double)(v2+i)*da;
            float x = sin(angle);
            float y = cos(angle);
            normal[index].set(0.f, 0.f, 1.f);
            vertex[index].set(x*m_top_radius, y*m_top_radius, m_height);
            ++index;

            // i = m_num_slices/2 - 1, m_num_slices - 2, ...
            angle = (double)(v1-i)*da;
            x = sin(angle);
            y = cos(angle);
            normal[index].set(0.f, 0.f, 1.f);
            vertex[index].set(x*m_top_radius, y*m_top_radius, m_height);
            ++index;
         }
         if (m_num_slices > 2*(m_num_slices/2))
         {
            float x = sin((m_num_slices - 1)*da);
            float y = cos((m_num_slices - 1)*da);
            normal[index].set(0.f, 0.f, 1.f);
            vertex[index].set(x*m_top_radius, y*m_top_radius, m_height);
            ++index;
         }
      }
   }

   setNumPrims(m_num_stacks + num_caps);
   setPrimLengths(new sglDeleteableDataPtr<sglIndexType>(length));
   setCoordList(new sglDeleteableDataPtr<sglVec3f>(vertex));
   setNormalList(sglGeoSet::ePER_VERTEX,
                 new sglDeleteableDataPtr<sglVec3f>(normal));
}

//----------------------------------------------------------------------------
void sglCylinder::printInfo(ostream &ostrm, const char *indent_string) const
{
   sglGeoSet::printInfo(ostrm, indent_string);
   ostrm << indent_string << "num_slices: " << m_num_slices << endl;
   ostrm << indent_string << "num_stacks: " << m_num_stacks << endl;
   ostrm << indent_string << "height: " << m_height << endl;
   ostrm << indent_string << "top_radius: " << m_top_radius << endl;
   ostrm << indent_string << "base_radius: " << m_base_radius << endl;
   ostrm << indent_string << "end_caps: "
         << (m_end_caps_flag ? "on" : "off") << endl;
}
