/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglCylinder.hpp
 *   Author: John K. Hammond
 *  Created: 1 October 2001
 *  Summary: an leaf sglNode containing the geometry for "cylinders".
 *****************************************************************************/

#ifndef SGL_CYLINDER_HPP
#define SGL_CYLINDER_HPP

#include <sgldb.h>
#include <sglTriangleStripSet.hpp>

/**

Create a geometry for a "cylinder."  The cylinder is aligned to the Z axis
where the base starts at z=0 and height runs along the positive Z
direction.  This is not a strict cylinder because the base radius can be
different from the top radius (a top_radius == 0, results in a cone).  The
number of slices corresponds to the number of divisions around the
circumference, and the number stacks corresponds to the number of divisions
along the height dimension (normally 1).  A boolean in the constructor argument
is used to specify whether or not end caps are to be created (without them
this object is a tube).

{\bf Example:}

\begin{verbatim}
   vector<sglStatelet*> state;
   state.push_back(new sglLighting(true));
   state.push_back(new sglTransparency(true));
   state.push_back(new sglCullMode(sglCullMode::eOFF));
   sglMaterial *material = new sglMaterial();
   material->setAmbient(0.0f, 0.5f, 1.0f);
   material->setDiffuse(0.0f, 0.5f, 1.0f);
   material->setAlpha(0.5f);
   state.push_back(material);

   sglCylinder *cylinder = new sglCylinder(20, 1, 4.0f, 2.0f, 0.5f, true);
   cylinder->setState(0, state, NULL);

   sglGeode *geode = new sglGeode;
   geode->addGeometry(cylinder);
\end{verbatim}

*/

//----------------------------------------------------------------------------

class SGLDB_DLL_API sglCylinder : public sglTriangleStripSet
{
public:
   /** Constructor.
       @param num_slices Number of divisions in the tesselation around the
              circumference; must be greater than 2.
       @param num_stacks Number of divisions in the tesselation along the
              length; must be greater than 0
       @param height Length of the cylinder; must be greater than 0.
       @param base_radius Radius at the base; must be greater than 0.
       @param top_radius Radius at the top; must be non-negative.
       @param end_caps_flag If true, the end caps will be created.
   */
   sglCylinder(unsigned int num_slices, unsigned int num_stacks,
               float height = 1.0f,
               float base_radius = 1.0f, float top_radius = 1.0f,
               bool end_caps_flag = true);

   /// virtual destructor
   virtual ~sglCylinder();

   /** Output the state of this node to the specified stream.  See
       sglObject::printInfo() for details.
   */
   virtual void printInfo(ostream &ostrm, const char *indent_string) const;

protected:

private:
   sglCylinder();                               // not implemented
   sglCylinder(const sglCylinder &);            // not implemented
   sglCylinder &operator=(const sglCylinder &); // not implemented

   void makeCylinder();

private:
   unsigned int  m_num_slices, m_num_stacks;
   float         m_base_radius;
   float         m_top_radius;
   float         m_height;
   bool          m_end_caps_flag;
};

#endif
