/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglDataLoader.cpp
 *   Author: Scott McMillan
 *  Created: 12 February 2001
 *****************************************************************************/

#ifdef WIN32
# include <windows.h>
#endif

#include <sglDataLoader.hpp>
#include <sglOStream.hpp>
#include <sglImageLoader.hpp>
#include <sglModelLoader.hpp>

#include <stdio.h>
#include <ctype.h>

#ifndef WIN32
# include <sys/types.h>
# include <dirent.h>
# include <unistd.h>
# include <errno.h>
#endif


//----------------------------------------------------------------------------
// model and image paths

vector<string> sglDataLoader::s_data_path;

#ifdef WIN32
string sglDataLoader::s_dir_separators("\\/");
#else
string sglDataLoader::s_dir_separators("/");
#endif

map<string, string> sglDataLoader::s_extension_dll_map;
map<string, bool> sglDataLoader::s_dll_loaded_map;
sglLock sglDataLoader::s_lock;

//----------------------------------------------------------------------------
class sglDataLoaderStaticInitializer
{
public:
   sglDataLoaderStaticInitializer()
      {
         vector<string> path;
         if (getenv("SGL_DATA_PATH"))
         {
            sglDataLoader::splitString(getenv("SGL_DATA_PATH"), path);
         }
         sglDataLoader::setDataPath(path);

         // HACK These should be auto-configured somehow.
         sglDataLoader::registerDLL("obj",  "sglobj");
         sglDataLoader::registerDLL("3ds",  "sgl3ds");
         sglDataLoader::registerDLL("flt",  "sglflt");
         sglDataLoader::registerDLL("md3",  "sglmd3");
         sglDataLoader::registerDLL("txp",  "sgltxp");
         sglDataLoader::registerDLL("tiff", "sgltiff");
         sglDataLoader::registerDLL("tif",  "sgltiff");
         sglDataLoader::registerDLL("jpeg", "sgljpeg");
         sglDataLoader::registerDLL("jpg",  "sgljpeg");
         sglDataLoader::registerDLL("png",  "sglpng");
         sglDataLoader::registerDLL("wrl",  "sglwrl");
      }
   ~sglDataLoaderStaticInitializer() {}
};
static sglDataLoaderStaticInitializer initializer;


//----------------------------------------------------------------------------
sglDataLoader::sglDataLoader()
{
}

//----------------------------------------------------------------------------
sglDataLoader::~sglDataLoader()
{
}

//----------------------------------------------------------------------------
void sglDataLoader::registerDLL(const string &extension,
                                const string &dll_name)
{
   string ext = extension;
   toLower(ext);

   s_lock.acquire();
   s_extension_dll_map[ext] = dll_name;
   s_lock.release();
}

//----------------------------------------------------------------------------
void sglDataLoader::getDLL(const string &extension, string &dll_name)
{
   string ext = extension;
   toLower(ext);

   s_lock.acquire();
   map<string, string>::const_iterator elt =
      s_extension_dll_map.find(ext);

   if (elt != s_extension_dll_map.end())
      dll_name = elt->second;
   else
      dll_name = "";
   s_lock.release();
}

//----------------------------------------------------------------------------
bool sglDataLoader::loadDLL(const string &dll_name)
{
   bool ret = false;

   s_lock.acquire();
   map<string, bool>::const_iterator elt = s_dll_loaded_map.find(dll_name);
   if (elt != s_dll_loaded_map.end())
   {
      ret = elt->second;
      s_lock.release();
      return ret;
   }

   void *dll = sgl::getDLLHandle(dll_name.c_str());
   if (!dll)
   {
      sglPrintWarning() << "Error loading dll: " << dll_name << endl;
   }
   else
   {
      sglPrintInfo() << "Loaded DLL: " << dll_name << endl;
      map<string, string>::const_iterator elt2 = s_extension_dll_map.begin();
      for (; elt2 != s_extension_dll_map.end(); ++elt2)
      {
         if (elt2->second == dll_name)
         {
            string ext = elt2->first;
            toUpper(ext);

            string func_name = string("loadImage") + ext;
            void *func = sgl::getFuncPtr(dll, func_name.c_str());
            if (func)
               sglImageLoader::registerImageLoader(
                  elt2->first, (sglImageLoader::loaderFunc)func);

            func_name = string("writeImage") + ext;
            func = sgl::getFuncPtr(dll, func_name.c_str());
            if (func)
               sglImageLoader::registerImageWriter(
                  elt2->first, (sglImageLoader::writerFunc)func);

            func_name = string("loadModel") + ext;
            func = sgl::getFuncPtr(dll, func_name.c_str());
            if (func)
               sglModelLoader::registerModelLoader(
                  elt2->first, (sglModelLoader::loaderFunc)func);
         }
      }
      ret = true;
   }

   s_dll_loaded_map[dll_name] = ret;
   s_lock.release();
   return ret;
}

//----------------------------------------------------------------------------
void sglDataLoader::splitString(
   const string &str, vector<string> &vec, const string &delimiters)
{
   vec.clear();
   string::size_type start = 0;
   string::size_type next;
   while ((next = str.find_first_of(delimiters, start)) != string::npos)
   {
      vec.push_back(str.substr(start, next-start));
      start = next+1;
   }
   if (start != str.length())
      vec.push_back(str.substr(start, next-start));
}

//----------------------------------------------------------------------------
void sglDataLoader::setDataPath(const vector<string> &path)
{
   s_lock.acquire();
   if (!path.empty())
      s_data_path = path;
   else
   {
      s_data_path.clear();
      s_data_path.push_back(string("."));
   }
   s_lock.release();
}

//----------------------------------------------------------------------------
void sglDataLoader::addToDataPath(const string &path)
{
   s_lock.acquire();
   s_data_path.push_back(path);
   s_lock.release();
}

//----------------------------------------------------------------------------
void sglDataLoader::getDataPath(vector<string> &path)
{
   s_lock.acquire();
   path = s_data_path;
   s_lock.release();
}

//----------------------------------------------------------------------------
// strips any path information from filename
string sglDataLoader::extractFilenameFromPath(const string &filepath)
{
   string::size_type pos = filepath.find_last_of(s_dir_separators);
   if (pos == string::npos)
      return filepath;
   else if (pos == filepath.length()-1)
      return string("");
   else
      return filepath.substr(pos+1);
}

//----------------------------------------------------------------------------
// strips any path information from filename
string sglDataLoader::extractDirectoryFromPath(const string &filepath)
{
   string::size_type pos = filepath.find_last_of(s_dir_separators);
   if (pos == string::npos)
      return string("");
   else if (pos == filepath.length()-1)
      return filepath;
   else
      return filepath.substr(0, pos);
}

//----------------------------------------------------------------------------
// extracts extension of filename
string sglDataLoader::extractExtensionFromFilename(const string &filename)
{
   string::size_type pos = filename.rfind('.');
   if (pos == string::npos || pos == filename.length()-1)
      return string("");
   else
      return filename.substr(pos+1);
}

//----------------------------------------------------------------------------
bool sglDataLoader::compareExtension(const string &filename, const string &ext)
{
   if (strCaseCmp(extractExtensionFromFilename(filename).c_str(),
                  ext.c_str()) == 0)
      return true;
   else
      return false;
}

//----------------------------------------------------------------------------
bool sglDataLoader::checkFile(const string &filename)
{
   FILE *fh = fopen(filename.c_str(), "rb");
   if (fh)
   {
      fclose(fh);
      return true;
   }
   return false;
}

//----------------------------------------------------------------------------
bool sglDataLoader::searchDir(
   const string &directory, const string &filename, string &foundname)
{
#if defined(WIN32)
   string search_string = directory + "/*";
   BOOL ret = true;
   WIN32_FIND_DATA find_data;
   HANDLE hFind = FindFirstFile(search_string.c_str(), &find_data);
   while (hFind != INVALID_HANDLE_VALUE && ret)
   {
      if (strCaseCmp(find_data.cFileName, filename.c_str()) == 0)
      {
         foundname = find_data.cFileName;
         FindClose(hFind);
         return true;
      }
      ret = FindNextFile(hFind, &find_data);
   }
   FindClose(hFind);
   return false;
#elif defined(vxw)
   // HACK This isn't thread safe.
   DIR *dir = opendir((char *)directory.c_str());
   if (dir)
   {
      dirent *dp;
      while ((dp = readdir(dir)) != NULL)
      {
         if (strCaseCmp(dp->d_name, filename.c_str()) == 0)
         {
            foundname = dp->d_name;
            closedir(dir);
            return true;
         }
      }
      closedir(dir);
   }
   return false;
#else
   long namemax = pathconf(directory.c_str(), _PC_NAME_MAX);
   if (namemax <= 0)
   {
      if (errno != 0 && errno != EINVAL)
         return false;
      else
         // _PC_NAME_MAX is not supported by the underlying filesystem.
         // Use a large default number and hope for the best.
         namemax = 10240;
   }

   DIR *dir = opendir(directory.c_str());
   if (dir)
   {
      dirent *dp = (dirent*)malloc(sizeof(dirent) + namemax + 1);
      if (dp)
      {
         dirent *dpp;
         while (readdir_r(dir, dp, &dpp) == 0 && dpp == dp)
         {
            if (strCaseCmp(dp->d_name, filename.c_str()) == 0)
            {
               foundname = dp->d_name;
               closedir(dir);
               free(dp);
               return true;
            }
         }
         free(dp);
      }
      closedir(dir);
   }
   return false;
#endif
}

//----------------------------------------------------------------------------
bool sglDataLoader::findFile(const string &filepath,
                             const string *alternate_path,
                             string &foundname)
{
   // Try filename "as is" (including any path).
   foundname = filepath;
   if (checkFile(foundname))
      return true;

   string fname = extractFilenameFromPath(filepath);
   string pname(filepath, 0, filepath.length() - fname.length());
   string tmp_path;
   string tmp_name;

   // Try case insensitive in given path.
   if (searchDir(pname, fname, tmp_name))
   {
      foundname = pname + "/" + tmp_name;
      return true;
   }

   if (alternate_path && !alternate_path->empty())
   {
      // strip trailing slashes
      string alt = *alternate_path;
      while (alt.find_last_of(s_dir_separators) == alt.length()-1)
         alt.resize(alt.length()-1);

      // Try case sensitive in alternate path.
      foundname = alt + "/" + fname;
      if (checkFile(foundname))
         return true;

      // Try case insensitive in alternate path.
      if (searchDir(alt, fname, tmp_name))
      {
         foundname = alt + "/" + tmp_name;
         return true;
      }

      tmp_path = alt + "/" + pname;

      // Try case sensitive in alternative path + given path.
      foundname = tmp_path + "/" + fname;
      if (checkFile(foundname))
         return true;

      // Try case insensitive in alternative path + given path.
      if (searchDir(tmp_path, fname, tmp_name))
      {
         foundname = tmp_path + "/" + tmp_name;
         return true;
      }
   }

   // Try the search paths.
   vector<string> path;
   getDataPath(path);
   for (vector<string>::const_iterator it = path.begin();
        it != path.end(); ++it)
   {
      // Try case sensitive.
      foundname = *it + "/" + fname;
      if (checkFile(foundname))
         return true;

      // Try case insensitive.
      if (searchDir(*it, fname, tmp_name))
      {
         foundname = *it + "/" + tmp_name;
         return true;
      }

      tmp_path = *it + "/" + pname;

      // Try case sensitive.
      foundname = tmp_path + "/" + fname;
      if (checkFile(foundname))
         return true;

      // Try case insensitive.
      if (searchDir(tmp_path, fname, tmp_name))
      {
         foundname = tmp_path + "/" + tmp_name;
         return true;
      }
   }

   foundname = "";
   return false;
}

//----------------------------------------------------------------------------
void sglDataLoader::toUpper(string &str)
{
   string::iterator end = str.end();
   for (string::iterator elt = str.begin(); elt != end; ++elt)
   {
      *elt = toupper(*elt);
   }
}

//----------------------------------------------------------------------------
void sglDataLoader::toLower(string &str)
{
   string::iterator end = str.end();
   for (string::iterator elt = str.begin(); elt != end; ++elt)
   {
      *elt = tolower(*elt);
   }
}

//----------------------------------------------------------------------------
int sglDataLoader::strCaseCmp(const char *s1, const char *s2 )
{
    int l1 = strlen ( s1 ) ;
    int l2 = strlen ( s2 ) ;

    //if ( l1 != l2 ) return FALSE ;
    int min_len = (l1<l2 ? l1 : l2);

    for (int i = 0; i < min_len; i++)
    {
        int c1 = s1[i];
        if (c1 >= 'A' && c1 <= 'Z')
            c1 += ('a' - 'A');

        int c2 = s2[i];
        if (c2 >= 'A' && c2 <= 'Z')
            c2 += ('a' - 'A');

        if (c1 == c2)
            continue;

        else if (c1 < c2) return -1;
        else              return 1;
    }

    if (l1 == l2)
        return 0;
    else if (l1 < l2)
        return -1;
    else // (l1 > l2)
        return 1;
}
