/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglImageLoader.cpp
 *   Author: Scott McMillan
 *  Created: 15 February 2001
 *****************************************************************************/

#include <sglImageLoader.hpp>
#include <sglOStream.hpp>
#include <sglLoadImage_sgi.hpp>
#include <sglLoadImage_bmp.hpp>
#include <sglLoadImage_tga.hpp>

map<string, sglImageLoader::loaderFunc> sglImageLoader::s_extension_loader_map;
map<string, sglImageLoader::writerFunc> sglImageLoader::s_extension_writer_map;
sglLock sglImageLoader::s_lock;

//----------------------------------------------------------------------------
class sglImageLoaderStaticInitializer
{
public:
   sglImageLoaderStaticInitializer()
      {
         // These are built-in.
         sglImageLoader::registerImageLoader("bw", loadImageSGI);
         sglImageLoader::registerImageLoader("int", loadImageSGI);
         sglImageLoader::registerImageLoader("inta", loadImageSGI);
         sglImageLoader::registerImageLoader("rgb", loadImageSGI);
         sglImageLoader::registerImageLoader("rgba", loadImageSGI);
         sglImageLoader::registerImageLoader("sgi", loadImageSGI);
//         sglImageLoader::registerImageLoader("img", loadImageSGI);
         sglImageLoader::registerImageLoader("bmp", loadImageBMP);
         sglImageLoader::registerImageLoader("tga", loadImageTGA);

         sglImageLoader::registerImageWriter("bw", writeImageSGI);
         sglImageLoader::registerImageWriter("int", writeImageSGI);
         sglImageLoader::registerImageWriter("inta", writeImageSGI);
         sglImageLoader::registerImageWriter("rgb", writeImageSGI);
         sglImageLoader::registerImageWriter("rgba", writeImageSGI);
         sglImageLoader::registerImageWriter("sgi", writeImageSGI);
//         sglImageLoader::registerImageWriter("img", writeImageSGI);
      }
   ~sglImageLoaderStaticInitializer() {}
};
static sglImageLoaderStaticInitializer initializer;


//----------------------------------------------------------------------------
void* sglImageLoader::loadImage(const string &filepath,
                                unsigned int &width,
                                unsigned int &height,
                                sglTexture::ExternalFormatEnum &format,
                                sglTexture::TypeEnum &type)
{
   string ext = extractExtensionFromFilename(filepath);
   toLower(ext);
   sglImageLoader::loaderFunc func = getImageLoader(ext);
   if (!func)
   {
      string dll;
      getDLL(ext, dll);
      if (!dll.empty())
      {
         loadDLL(dll);
         func = getImageLoader(ext);
      }
   }
   if (func)
      return (*func)(filepath, width, height, format, type);
   else if (ext != "rgb")
   {
      // If we can't load the file due to a missing loader, see if we
      // can load a .rgb file of the same name.
      func = getImageLoader("rgb");
      if (func)
      {
         // The .rgb file may not be in the same directory as the original
         // file, so do a file search.
         string fallback =
            filepath.substr(0, filepath.length()-ext.length()) + "rgb";
         string foundname;
         if (findFile(fallback, NULL, foundname))
         {
            return (*func)(foundname, width, height, format, type);
         }
      }
   }

   sglPrintWarning() << "Error loading file " << filepath
                     << "; no reader found for files of this format." << endl;
   return NULL;
}

//----------------------------------------------------------------------------
bool sglImageLoader::writeImage(const string &filepath,
                                const void *image,
                                unsigned int width,
                                unsigned int height,
                                sglTexture::ExternalFormatEnum format,
                                sglTexture::TypeEnum type)
{
   string ext = extractExtensionFromFilename(filepath);
   toLower(ext);
   sglImageLoader::writerFunc func = getImageWriter(ext);
   if (!func)
   {
      string dll;
      getDLL(ext, dll);
      if (!dll.empty())
      {
         loadDLL(dll);
         func = getImageWriter(ext);
      }
   }
   if (func)
      return (*func)(filepath, image, width, height, format, type);

   sglPrintWarning() << "Error writing file " << filepath
                     << "; no writer found for files of this format." << endl;
   return false;
}

//----------------------------------------------------------------------------
void sglImageLoader::registerImageLoader(
   const string &extension,
   sglImageLoader::loaderFunc image_loader_func)
{
   string ext = extension;
   toLower(ext);

   s_lock.acquire();
   if (image_loader_func)
   {
      //cerr << "Image loader for " << ext << " registered" << endl;
      s_extension_loader_map[ext] = image_loader_func;
   }
   else  // remove the registration???
   {
      map<string, sglImageLoader::loaderFunc>::iterator elt =
         s_extension_loader_map.find(ext);
      if (elt != s_extension_loader_map.end())
          s_extension_loader_map.erase(elt);
   }
   s_lock.release();
}

//----------------------------------------------------------------------------
void sglImageLoader::registerImageWriter(
   const string &extension,
   sglImageLoader::writerFunc image_writer_func)
{
   string ext = extension;
   toLower(ext);

   s_lock.acquire();
   if (image_writer_func)
   {
      //cerr << "Image writer for " << ext << " registered" << endl;
      s_extension_writer_map[ext] = image_writer_func;
   }
   else  // remove the registration???
   {
      map<string, sglImageLoader::writerFunc>::iterator elt =
         s_extension_writer_map.find(ext);
      if (elt != s_extension_writer_map.end())
          s_extension_writer_map.erase(elt);
   }
   s_lock.release();
}

//----------------------------------------------------------------------------
sglImageLoader::loaderFunc sglImageLoader::getImageLoader(
   const string &extension)
{
   loaderFunc ret = NULL;

   string ext = extension;
   toLower(ext);

   s_lock.acquire();
   map<string, sglImageLoader::loaderFunc>::iterator elt =
      s_extension_loader_map.find(ext);
   if (elt != s_extension_loader_map.end())
      ret = elt->second;
   s_lock.release();

   return ret;
}

//----------------------------------------------------------------------------
sglImageLoader::writerFunc sglImageLoader::getImageWriter(
   const string &extension)
{
   writerFunc ret = NULL;

   string ext = extension;
   toLower(ext);

   s_lock.acquire();
   map<string, sglImageLoader::writerFunc>::iterator elt =
      s_extension_writer_map.find(ext);
   if (elt != s_extension_writer_map.end())
      ret = elt->second;
   s_lock.release();

   return ret;
}

//----------------------------------------------------------------------------
template <class T>
T *sglImageLoaderReformatImage(
   sglTexture::ExternalFormatEnum old_format,
   unsigned int num_s, unsigned int num_t,
   sglTexture::ExternalFormatEnum new_format,
   const T *old_image)
{
   unsigned int new_components = sglTexture::getNumComponents(new_format);
   unsigned int old_components = sglTexture::getNumComponents(old_format);
   T* new_image = (T*)malloc(num_s * num_t * new_components * sizeof(T));

   int new_alpha = -1;
   if (new_format == sglTexture::eALPHA) new_alpha = 0;
   else if (new_format == sglTexture::eRGBA) new_alpha = 3;
   else if (new_format == sglTexture::eLUMINANCE_ALPHA) new_alpha = 1;

   if (new_alpha != -1)
   {
      int old_alpha = -1;
      if (old_format == sglTexture::eALPHA) old_alpha = 0;
      else if (old_format == sglTexture::eRGBA) old_alpha = 3;
      else if (old_format == sglTexture::eLUMINANCE_ALPHA) old_alpha = 1;

      if (old_alpha == -1)
      {
         if (new_format != sglTexture::eALPHA)
         {
            T *tmp = new_image + new_alpha;
            for (unsigned int y = 0; y < num_t; ++y)
               for (unsigned int x = 0; x < num_s; ++x)
               {
                  *tmp = (T)~0;
                  tmp += new_components;
               }
         }
      }
      else
      {
         T *tmp1 = new_image + new_alpha;
         const T *tmp2 = old_image + old_alpha;
         for (unsigned int y = 0; y < num_t; ++y)
            for (unsigned int x = 0; x < num_s; ++x)
            {
               *tmp1 = *tmp2;
               tmp1 += new_components;
               tmp2 += old_components;
            }
         if (new_format == sglTexture::eALPHA)
            return new_image;
      }
   }

   if (new_format == sglTexture::eRGB || new_format == sglTexture::eRGBA)
   {
      if (old_format == sglTexture::eRGB || old_format == sglTexture::eRGBA)
      {
         T *tmp1 = new_image;
         const T *tmp2 = old_image;
         for (unsigned int y = 0; y < num_t; ++y)
            for (unsigned int x = 0; x < num_s; ++x)
            {
               tmp1[0] = tmp2[0];
               tmp1[1] = tmp2[1];
               tmp1[2] = tmp2[2];
               tmp1 += new_components;
               tmp2 += old_components;
            }
      }
      else if (old_format == sglTexture::eLUMINANCE ||
               old_format == sglTexture::eLUMINANCE_ALPHA)
      {
         T *tmp1 = new_image;
         const T *tmp2 = old_image;
         for (unsigned int y = 0; y < num_t; ++y)
            for (unsigned int x = 0; x < num_s; ++x)
            {
               tmp1[0] = *tmp2;
               tmp1[1] = *tmp2;
               tmp1[2] = *tmp2;
               tmp1 += new_components;
               tmp2 += old_components;
            }
      }
      else if (old_format == sglTexture::eRED)
      {
         T *tmp1 = new_image;
         const T *tmp2 = old_image;
         for (unsigned int y = 0; y < num_t; ++y)
            for (unsigned int x = 0; x < num_s; ++x)
            {
               tmp1[0] = *tmp2;
               tmp1[1] = (T)0;
               tmp1[2] = (T)0;
               tmp1 += new_components;
               tmp2 += old_components;
            }
      }
      else if (old_format == sglTexture::eGREEN)
      {
         T *tmp1 = new_image;
         const T *tmp2 = old_image;
         for (unsigned int y = 0; y < num_t; ++y)
            for (unsigned int x = 0; x < num_s; ++x)
            {
               tmp1[0] = (T)0;
               tmp1[1] = *tmp2;
               tmp1[2] = (T)0;
               tmp1 += new_components;
               tmp2 += old_components;
            }
      }
      else if (old_format == sglTexture::eBLUE)
      {
         T *tmp1 = new_image;
         const T *tmp2 = old_image;
         for (unsigned int y = 0; y < num_t; ++y)
            for (unsigned int x = 0; x < num_s; ++x)
            {
               tmp1[0] = (T)0;
               tmp1[1] = (T)0;
               tmp1[2] = *tmp2;
               tmp1 += new_components;
               tmp2 += old_components;
            }
      }
      else if (old_format == sglTexture::eALPHA)
      {
         T *tmp1 = new_image;
         const T *tmp2 = old_image;
         for (unsigned int y = 0; y < num_t; ++y)
            for (unsigned int x = 0; x < num_s; ++x)
            {
               tmp1[0] = (T)0;
               tmp1[1] = (T)0;
               tmp1[2] = (T)0;
               tmp1 += new_components;
               tmp2 += old_components;
            }
      }
   }
   else if (old_format == sglTexture::eRGB || old_format == sglTexture::eRGBA)
   {
      if (new_format == sglTexture::eLUMINANCE ||
          new_format == sglTexture::eLUMINANCE_ALPHA ||
          new_format == sglTexture::eALPHA)
      {
         T *tmp1 = new_image;
         const T *tmp2 = old_image;
         for (unsigned int y = 0; y < num_t; ++y)
            for (unsigned int x = 0; x < num_s; ++x)
            {
#if 1
               // Standard RGB to YIQ conversion (Foley et al).
               *tmp1 = (T)(0.299f * (float)tmp2[0] +
                           0.587f * (float)tmp2[1] +
                           0.114f * (float)tmp2[2]);
#else
               *tmp1 = (T)(((float)tmp2[0] +
                            (float)tmp2[1] +
                            (float)tmp2[2]) / 3.0f);
#endif
               tmp1 += new_components;
               tmp2 += old_components;
            }
      }
      else
      {
         unsigned int component = 0;
         if (new_format == sglTexture::eGREEN) component = 1;
         else if (new_format == sglTexture::eBLUE) component = 2;

         T *tmp1 = new_image;
         const T *tmp2 = old_image;
         for (unsigned int y = 0; y < num_t; ++y)
            for (unsigned int x = 0; x < num_s; ++x)
            {
               *tmp1 = tmp2[component];
               tmp1 += new_components;
               tmp2 += old_components;
            }
      }
   }
   else
   {
      T *tmp1 = new_image;
      const T *tmp2 = old_image;
      for (unsigned int y = 0; y < num_t; ++y)
         for (unsigned int x = 0; x < num_s; ++x)
         {
            *tmp1 = *tmp2;
            tmp1 += new_components;
            tmp2 += old_components;
         }
   }

   return new_image;
}

//----------------------------------------------------------------------------
void *sglImageLoader::reformatImage(
   sglTexture::ExternalFormatEnum old_format,
   sglTexture::TypeEnum type,
   unsigned int num_s, unsigned int num_t,
   sglTexture::ExternalFormatEnum new_format,
   const void *old_image)
{
   switch (type)
   {
      case sglTexture::eBYTE:
         return sglImageLoaderReformatImage(
            old_format, num_s, num_t, new_format,
            (const signed char*)old_image);
      case sglTexture::eUNSIGNED_BYTE:
         return sglImageLoaderReformatImage(
            old_format, num_s, num_t, new_format,
            (const unsigned char*)old_image);
      case sglTexture::eSHORT:
         return sglImageLoaderReformatImage(
            old_format, num_s, num_t, new_format,
            (const signed short*)old_image);
      case sglTexture::eUNSIGNED_SHORT:
         return sglImageLoaderReformatImage(
            old_format, num_s, num_t, new_format,
            (const unsigned short*)old_image);
      case sglTexture::eINT:
         return sglImageLoaderReformatImage(
            old_format, num_s, num_t, new_format,
            (const signed int*)old_image);
      case sglTexture::eUNSIGNED_INT:
         return sglImageLoaderReformatImage(
            old_format, num_s, num_t, new_format,
            (const unsigned int*)old_image);
      case sglTexture::eFLOAT:
         return sglImageLoaderReformatImage(
            old_format, num_s, num_t, new_format,
            (const float*)old_image);
      case sglTexture::eBITMAP:
         sglPrintWarning()
            << "sglImageLoader::reformatImage()"
            << " currently doesn't support bitmap images"
            << endl;
         break;
      default:
         sglPrintWarning()
            << "sglImageLoader::reformatImage() unkown data type = "
            << (int)type << endl;
         break;
   }
   return NULL;
}

//----------------------------------------------------------------------------
template <class T>
void sglImageLoaderScaleComponent(
   T* image, sglTexture::ExternalFormatEnum format,
   unsigned int num_s, unsigned int num_t,
   float scale, unsigned int component)
{
   unsigned int num_components = sglTexture::getNumComponents(format);
   T *tmp = image + component;
   for (unsigned int y = 0; y < num_t; ++y)
      for (unsigned int x = 0; x < num_s; ++x)
      {
         *tmp = (T)((float)*tmp * scale);
         tmp += num_components;
      }
}

//----------------------------------------------------------------------------
void sglImageLoader::scaleImageComponent(
   void* image, sglTexture::ExternalFormatEnum format,
   sglTexture::TypeEnum type,
   unsigned int num_s, unsigned int num_t,
   float scale, unsigned int component)
{
   switch (type)
   {
      case sglTexture::eBYTE:
         sglImageLoaderScaleComponent(
            (signed char*)image, format, num_s, num_t, scale, component);
         break;
      case sglTexture::eUNSIGNED_BYTE:
         sglImageLoaderScaleComponent(
            (unsigned char*)image, format, num_s, num_t, scale, component);
         break;
      case sglTexture::eSHORT:
         sglImageLoaderScaleComponent(
            (signed short*)image, format, num_s, num_t, scale, component);
         break;
      case sglTexture::eUNSIGNED_SHORT:
         sglImageLoaderScaleComponent(
            (unsigned short*)image, format, num_s, num_t, scale, component);
         break;
      case sglTexture::eINT:
         sglImageLoaderScaleComponent(
            (signed int*)image, format, num_s, num_t, scale, component);
         break;
      case sglTexture::eUNSIGNED_INT:
         sglImageLoaderScaleComponent(
            (unsigned int*)image, format, num_s, num_t, scale, component);
         break;
      case sglTexture::eFLOAT:
         sglImageLoaderScaleComponent(
            (float*)image, format, num_s, num_t, scale, component);
         break;
      case sglTexture::eBITMAP:
         sglPrintWarning()
            << "sglImageLoader::scaleComponent()"
            << " currently doesn't support bitmap images"
            << endl;
         break;
      default:
         sglPrintWarning()
            << "sglImageLoader::scaleComponent() unkown data type = "
            << (int)type << endl;
         break;
   }
}

//----------------------------------------------------------------------------
template <class T, class S>
void sglImageLoaderChangeImageType(
   sglTexture::ExternalFormatEnum format,
   unsigned int num_s, unsigned int num_t,
   double old_min, double old_max, double new_min, double new_max,
   const T* old_image, S** new_image)
{
   unsigned int num_components = sglTexture::getNumComponents(format);
   unsigned int num_pixels = num_s * num_t * num_components;
   *new_image = (S*)malloc(num_pixels * sizeof(S));

   old_max -= old_min;
   new_max -= new_min;

   S* tmp = *new_image;
   S* stop = *new_image + num_pixels;
   while (tmp != stop)
   {
      double val = (double)*old_image;
      val -= old_min;
      val /= old_max;
      val *= new_max;
      val += new_min;
      *tmp = (S)val;
      ++old_image;
      ++tmp;
   }
}

//----------------------------------------------------------------------------
template <class T>
void *sglImageLoaderChangeImageType(
   sglTexture::ExternalFormatEnum format,
   unsigned int num_s, unsigned int num_t,
   sglTexture::TypeEnum new_type,
   double old_min, double old_max,
   const T* old_image)
{
   void *new_image = NULL;
   switch (new_type)
   {
      case sglTexture::eBYTE:
         sglImageLoaderChangeImageType(
            format, num_s, num_t,
            old_min, old_max, (double)SCHAR_MIN, (double)SCHAR_MAX,
            old_image, (signed char**)&new_image);
         break;
      case sglTexture::eUNSIGNED_BYTE:
         sglImageLoaderChangeImageType(
            format, num_s, num_t,
            old_min, old_max, (double)0, (double)UCHAR_MAX,
            old_image, (unsigned char**)&new_image);
         break;
      case sglTexture::eSHORT:
         sglImageLoaderChangeImageType(
            format, num_s, num_t,
            old_min, old_max, (double)SHRT_MIN, (double)SHRT_MAX,
            old_image, (signed short**)&new_image);
         break;
      case sglTexture::eUNSIGNED_SHORT:
         sglImageLoaderChangeImageType(
            format, num_s, num_t,
            old_min, old_max, (double)0, (double)USHRT_MAX,
            old_image, (unsigned short**)&new_image);
         break;
      case sglTexture::eINT:
         sglImageLoaderChangeImageType(
            format, num_s, num_t,
            old_min, old_max, (double)INT_MIN, (double)INT_MAX,
            old_image, (signed int**)&new_image);
         break;
      case sglTexture::eUNSIGNED_INT:
         sglImageLoaderChangeImageType(
            format, num_s, num_t,
            old_min, old_max, (double)0, (double)UINT_MAX,
            old_image, (unsigned int**)&new_image);
         break;
      case sglTexture::eFLOAT:
         sglImageLoaderChangeImageType(
            format, num_s, num_t,
            old_min, old_max, 0.0f, 1.0f,
            old_image, (float**)&new_image);
         break;
      case sglTexture::eBITMAP:
         sglPrintWarning()
            << "sglImageLoader::changeImageType()"
            << " currently doesn't support bitmap images"
            << endl;
         break;
      default:
         sglPrintWarning()
            << "sglImageLoader::changeImageType() unkown data type = "
            << (int)new_type << endl;
         break;
   }
   return new_image;
}

//----------------------------------------------------------------------------
void *sglImageLoader::changeImageType(
   sglTexture::ExternalFormatEnum format,
   sglTexture::TypeEnum old_type,
   unsigned int num_s, unsigned int num_t,
   sglTexture::TypeEnum new_type,
   const void *old_image)
{
   switch (old_type)
   {
      case sglTexture::eBYTE:
         return sglImageLoaderChangeImageType(
            format, num_s, num_t, new_type,
            (double)SCHAR_MIN, (double)SCHAR_MAX,
            (const signed char*)old_image);
      case sglTexture::eUNSIGNED_BYTE:
         return sglImageLoaderChangeImageType(
            format, num_s, num_t, new_type,
            (double)0, (double)UCHAR_MAX,
            (const unsigned char*)old_image);
      case sglTexture::eSHORT:
         return sglImageLoaderChangeImageType(
            format, num_s, num_t, new_type,
            (double)SHRT_MIN, (double)SHRT_MAX,
            (const signed short*)old_image);
      case sglTexture::eUNSIGNED_SHORT:
         return sglImageLoaderChangeImageType(
            format, num_s, num_t, new_type,
            (double)0, (double)USHRT_MAX,
            (const unsigned short*)old_image);
      case sglTexture::eINT:
         return sglImageLoaderChangeImageType(
            format, num_s, num_t, new_type,
            (double)INT_MIN, (double)INT_MAX,
            (const signed int*)old_image);
      case sglTexture::eUNSIGNED_INT:
         return sglImageLoaderChangeImageType(
            format, num_s, num_t, new_type,
            (double)0, (double)UINT_MAX,
            (const unsigned int*)old_image);
      case sglTexture::eFLOAT:
         return sglImageLoaderChangeImageType(
            format, num_s, num_t, new_type,
            0.0, 1.0, (const float*)old_image);
      case sglTexture::eBITMAP:
         sglPrintWarning()
            << "sglImageLoader::changeImageType()"
            << " currently doesn't support bitmap images"
            << endl;
         break;
      default:
         sglPrintWarning()
            << "sglImageLoader::changeImageType() unkown data type = "
            << (int)old_type << endl;
         break;
   }
   return NULL;
}
