/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglImageLoader.hpp
 *   Author: Scott McMillan
 *  Created: 15 February 2001
 *  Summary: Abstract base class for all 3D model loaders.
 *****************************************************************************/

#ifndef SGL_IMAGE_LOADER
#define SGL_IMAGE_LOADER

#include <sgldb.h>
#include <sglDataLoader.hpp>
#include <sglTexture.hpp>
#include <map>


/** Abstract base class for 3D models.
 */

class SGLDB_DLL_API sglImageLoader : public sglDataLoader
{
public:
   /** The prototype for loader functions. */
   typedef void *(*loaderFunc)(const string &filepath,
                               unsigned int &width,
                               unsigned int &height,
                               sglTexture::ExternalFormatEnum &format,
                               sglTexture::TypeEnum &type);

   /** The prototype for writer functions. */
   typedef bool (*writerFunc)(const string &filepath,
                              const void *image,
                              unsigned int width,
                              unsigned int height,
                              sglTexture::ExternalFormatEnum format,
                              sglTexture::TypeEnum type);

public:
   /** Destructor. */
   virtual ~sglImageLoader() = 0;

   /** Load the given image file using whatever loader has been registered
       for files of that type.
       @param filepath the full pathname of the file to load
       @param width on return contains the width of the image in pixels
       @param height on return contains the height of the image in pixels
       @param format on return contains the format of the image
       @param type on return contains the data type of the image
       @return NULL on failure, otherwise a pointer to the malloc'ed image
              data (which the user is responsible for free()ing)
   */
   static void* loadImage(const string &filepath,
                          unsigned int &width,
                          unsigned int &height,
                          sglTexture::ExternalFormatEnum &format,
                          sglTexture::TypeEnum &type);

   /** Write the given image to a file using whatever loader has been
       registered for files of that type.
       @param filepath the full pathname of the file to write
       @param image the pixel data to write
       @param width the width of the image in pixels
       @param height the height of the image in pixels
       @param format the format of the image
       @param type the data type of the image
       @return false on failure
   */
   static bool writeImage(const string &filepath,
                          const void *image,
                          unsigned int width,
                          unsigned int height,
                          sglTexture::ExternalFormatEnum format,
                          sglTexture::TypeEnum type);

   /** Register a loader function pointer and the extension it supports.  Note
       that a single function can support multiple extensions.  Registering an
       extension a second time will replace the existing entry.
       @param extension extension string that the function supports.
       @param image_loader_func pointer to function supporting the extension
   */
   static void registerImageLoader(const string &extension,
                                   loaderFunc image_loader_func);

   /** Register a writer function pointer and the extension it supports.  Note
       that a single function can support multiple extensions.  Registering an
       extension a second time will replace the existing entry.
       @param extension extension string that the function supports.
       @param image_writer_func pointer to function supporting the extension
   */
   static void registerImageWriter(const string &extension,
                                   writerFunc image_writer_func);

   /** Get the registered image loader function pointer based on extension.
       @param extension extension to search on.
       @return function pointer, or NULL if not found
   */
   static loaderFunc getImageLoader(const string &extension);

   /** Get the registered image writer function pointer based on extension.
       @param extension extension to search on.
       @return function pointer, or NULL if not found
   */
   static writerFunc getImageWriter(const string &extension);

   /** Create a new image with a new format from an old image.
       @param old_format the format of the old image
       @param type the type of the pixel data
       @param num_s the width of the image
       @param num_t the height of the image
       @param new_format the format of the image to be created
       @param old_image the original image
       @return NULL on failure, otherwise a pointer to the malloc'ed image
              data (which the user is responsible for free()ing)
    */
   static void *reformatImage(
      sglTexture::ExternalFormatEnum old_format,
      sglTexture::TypeEnum type,
      unsigned int num_s, unsigned int num_t,
      sglTexture::ExternalFormatEnum new_format,
      const void *old_image);

   /** Scale the nth component of each pixel in the given image by the
       given scale factor.
       @param image the image to modify
       @param format the format of the image data
       @param type the type of each pixel
       @param num_s the width of the image
       @param num_t the height of the image
       @param scale the factor to multiply by
       @param component the component to scale
   */
   static void scaleImageComponent(
      void* image, sglTexture::ExternalFormatEnum format,
      sglTexture::TypeEnum type,
      unsigned int num_s, unsigned int num_t,
      float scale, unsigned int component);

   /** Create a new image with a new type from an old image.
       @param the format of the image
       @param old_type the type of the pixel data in the old image
       @param num_s the width of the image
       @param num_t the height of the image
       @param new_type the type of the pixel data in the new image
       @param old_image the original image
       @return NULL on failure, otherwise a pointer to the malloc'ed image
              data (which the user is responsible for free()ing)
   */
   static void *changeImageType(
      sglTexture::ExternalFormatEnum format,
      sglTexture::TypeEnum old_type,
      unsigned int num_s, unsigned int num_t,
      sglTexture::TypeEnum new_type,
      const void *old_image);

protected:
   sglImageLoader();
   sglImageLoader(const sglImageLoader &);
   sglImageLoader &operator=(const sglImageLoader &);

private:  // Variables
   static map<string, loaderFunc> s_extension_loader_map;
   static map<string, writerFunc> s_extension_writer_map;
   static sglLock s_lock;
};

#endif
