/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglLoadFile_scm.cpp
 *  Created: 12 July 1999
 *  Summary: simple model format loader for dynamechs 2.1 formats
 *****************************************************************************/

#include <sglLoadFile_scm.hpp>
#include <sglGeode.hpp>
#include <sglStatelet.hpp>
#include <sglMaterial.hpp>
#include <sglTriangleStripSet.hpp>
#include <sglOStream.hpp>




//----------------------------------------------------------------------------
inline void compute_face_normal(const sglVec3f &v0,
                                const sglVec3f &v1,
                                const sglVec3f &v2,
                                sglVec3f &normal)
{
    sglVec3f a, b;

    a.sub(v1, v0);
    b.sub(v2, v0);

    normal.cross(a, b);
    normal.normalize();
}

//----------------------------------------------------------------------------

sglNode *sglLoadFile_scm(const string &model_filename,
                         unsigned int)
{
   ifstream data_ptr;

   data_ptr.open(model_filename.c_str());
   if (!data_ptr)
   {
      sglPrintWarning() << "sglLoadFile_scm: Error unable to open data file: "
                        << model_filename << endl;
      return NULL;
   }

   sglMaterial *material = new sglMaterial();

   GLfloat color[4] = {0.0, 0.0, 0.0, 1.0};

   data_ptr >> color[0] >> color[1] >> color[2];  // emission color
   material->setEmission(color[0], color[1], color[2]);

   data_ptr >> color[0] >> color[1] >> color[2];  // ambient color
   material->setAmbient(color[0], color[1], color[2]);

   data_ptr >> color[0] >> color[1] >> color[2];  // diffuse color
   material->setDiffuse(color[0], color[1], color[2]);

   data_ptr >> color[0] >> color[1] >> color[2];  // specular color
   material->setSpecular(color[0], color[1], color[2]);

   GLfloat shininess;
   data_ptr >> shininess;
   material->setShininess(shininess);

   GLfloat alpha;
   data_ptr >> alpha;
   material->setAlpha(alpha);

   vector<sglStatelet*> statelets;

   sglTransparency* tr = new sglTransparency();
   tr->enable(alpha < 1.0);
   statelets.push_back(tr);

   statelets.push_back(material);

   sglLighting* li = new sglLighting();
   li->enable(true);
   statelets.push_back(li);

   unsigned int num_faces, num_vertices_tot;

   data_ptr >> num_faces;
   sglVec3f *normal = new sglVec3f[num_faces];
   unsigned int *length = new unsigned int[num_faces];

   data_ptr >> num_vertices_tot;
   sglVec3f *vertex = new sglVec3f[num_vertices_tot];

   unsigned int index = 0;
   for (unsigned int i=0; i<num_faces; i++)
   {
      data_ptr >> length[i]; //num_vertices;

      for (unsigned int j=0; j<length[i]; j++)
      {
         data_ptr >> vertex[index][0] >> vertex[index][1] >> vertex[index][2];
         index++;
      }

      if (length[i] < 3)
      {
         sglPrintWarning()
            << "sglLoadFile_scm() error: too few vertices in face" << endl;
         normal[i].set(0.0, 0.0, 1.0);
      }
      else
      {
         unsigned int ix = index - length[i];
         compute_face_normal(vertex[ix + 2],
                             vertex[ix],
                             vertex[ix + 1],
                             normal[i]);
      }
   }

   sglTriangleStripSet *gset = new sglTriangleStripSet();
   gset->setNumPrims(num_faces);
   gset->setPrimLengths(new sglDeleteableDataPtr<unsigned int>(length));
   gset->setCoordList(new sglDeleteableDataPtr<sglVec3f>(vertex));
   gset->setNormalList(sglGeoSet::ePER_PRIM,
                       new sglDeleteableDataPtr<sglVec3f>(normal));
   gset->setState(0, statelets);

   sglGeode *geode = new sglGeode();
   geode->addGeometry(gset);

   data_ptr.close();
   return geode;
}
