/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglLoadFile_xan.cpp
 *  Created: 12 July 1999
 *  Summary: simple model format loader for dynamechs 2.1 formats
 *****************************************************************************/

#include <sglLoadFile_xan.hpp>
#include <sglGeode.hpp>
#include <sglStatelet.hpp>
#include <sglMaterial.hpp>
#include <sglIndexedPolygonSet.hpp>
#include <sglOStream.hpp>


//----------------------------------------------------------------------------
inline void compute_face_normal(const sglVec3f &v0,
                                const sglVec3f &v1,
                                const sglVec3f &v2,
                                sglVec3f &normal)
{
    sglVec3f a, b;

    a.sub(v1, v0);
    b.sub(v2, v0);

    normal.cross(a, b);
    normal.normalize();
}

//----------------------------------------------------------------------------

sglNode *sglLoadFile_xan(const string &model_filename,
                         unsigned int)
{
   ifstream data_ptr;

   data_ptr.open(model_filename.c_str());
   if (!data_ptr)
   {
      sglPrintWarning() << "sglLoadFile_xan: Error unable to open data file: "
                        << model_filename << endl;
      return NULL;
   }

   sglMaterial *material = new sglMaterial();

   GLfloat color[4] = {0.0, 0.0, 0.0, 1.0};

   data_ptr >> color[0] >> color[1] >> color[2];  // emission color
   material->setEmission(color[0], color[1], color[2]);

   data_ptr >> color[0] >> color[1] >> color[2];  // ambient color
   material->setAmbient(color[0], color[1], color[2]);

   data_ptr >> color[0] >> color[1] >> color[2];  // diffuse color
   material->setDiffuse(color[0], color[1], color[2]);

   data_ptr >> color[0] >> color[1] >> color[2];  // specular color
   material->setSpecular(color[0], color[1], color[2]);

   GLfloat shininess;
   data_ptr >> shininess;
   material->setShininess(shininess);

   GLfloat alpha;
   data_ptr >> alpha;
   material->setAlpha(alpha);

   vector<sglStatelet*> statelets;

   sglTransparency* tr = new sglTransparency();
   tr->enable(alpha < 1.0);
   statelets.push_back(tr);

   statelets.push_back(material);

   sglLighting* li = new sglLighting();
   li->enable(true);
   statelets.push_back(li);

   sglVec3f scale;
   data_ptr >> scale[0] >> scale[1] >> scale[2];

   unsigned int  num_vertices, num_faces;
   data_ptr >> num_vertices >> num_faces;
   sglVec3f *vertex = new sglVec3f[num_vertices];
   sglVec3f *normal = new sglVec3f[num_faces];
   unsigned int *length = new unsigned int[num_faces];

   for (unsigned int ix=0; ix<num_vertices; ix++)
   {
      data_ptr >> vertex[ix][0] >> vertex[ix][1] >> vertex[ix][2];
      vertex[ix][0] *= scale[0];
      vertex[ix][1] *= scale[1];
      vertex[ix][2] *= scale[2];
   }

   unsigned int num_indices = 0;
   for (unsigned int iy=0; iy<num_faces; iy++)
   {
      data_ptr >> length[iy]; //num_vertices;
      num_indices += length[iy];
   }

   unsigned int *nindex = new unsigned int[num_faces];
   unsigned int *vindex = new unsigned int[num_indices];

   unsigned int index = 0;
   for (unsigned int iz=0; iz<num_faces; iz++)
   {
      nindex[iz] = iz;

      for (unsigned int j=0; j<length[iz]; j++)
      {
         data_ptr >> vindex[index + length[iz] - j - 1];
         //index++;
      }
      index += length[iz];

      if (length[iz] < 3)
      {
         sglPrintWarning()
            << "sglLoadFile_xan() error: too few vertices in face" << endl;
         normal[iz].set(0.0, 0.0, 1.0);
      }
      else
      {
         unsigned int ix = index - length[iz];
         compute_face_normal(vertex[vindex[ix]],
                             vertex[vindex[ix + 1]],
                             vertex[vindex[ix + 2]],
                             normal[iz]);
      }
   }

   sglIndexedPolygonSet *gset = new sglIndexedPolygonSet();
   gset->setNumPrims(num_faces);
   gset->setPrimLengths(new sglDeleteableDataPtr<unsigned int>(length));
   gset->setCoordList(new sglDeleteableDataPtr<sglVec3f>(vertex));
   gset->setCoordIndexList(new sglDeleteableDataPtr<unsigned int>(vindex));
   gset->setNormalList(sglGeoSet::ePER_PRIM,
                       new sglDeleteableDataPtr<sglVec3f>(normal));
   gset->setNormalIndexList(new sglDeleteableDataPtr<unsigned int>(nindex));
   gset->setState(0, statelets);

   sglGeode *geode = new sglGeode();
   geode->addGeometry(gset);

   data_ptr.close();
   return geode;
}
