/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglLoadImage_bmp.cpp
 *   Author: Tom Stimson
 *  Created: 7 April 2001
 *  Summary: read BMP format image files
 *****************************************************************************/

#include <sglLoadImage_bmp.hpp>
#include <sglOStream.hpp>
#include <stdio.h>

/*****************************************************************************/
struct sglBitmapArrayHeader
{
   unsigned int     size;
   unsigned int     next_offset;
   unsigned short   x_display;
   unsigned short   y_display;
};

/*****************************************************************************/
struct sglBMPFileHeader
{
   int              size;
   unsigned short   reserved1;
   unsigned short   reserved2;
   int              offset_bits;
};

/*****************************************************************************/
struct sglBMPInfoHeader
{
   int              width;
   int              height;
   short            planes;
   short            bit_count;
   int              compression;
   int              image_size;
   int              x_pixels_per_meter;
   int              y_pixels_per_meter;
   int              colors_used;
   int              colors_important;
};

/*****************************************************************************/
struct sglBMPInfoHeaderOld
{
   short            width;
   short            height;
   short            planes;
   short            bit_count;
};

/*****************************************************************************/
bool readBMPHeaders(const string &filepath, FILE *file,
                    sglBMPFileHeader &header,
                    sglBMPInfoHeader &info, int &info_size)
{
   fread(&header, 1, sizeof(sglBMPFileHeader), file);

#if defined(SGL_BIG_ENDIAN)
   sglEndianSwap(&header.size);
   sglEndianSwap(&header.offset_bits);
#endif
//   sglPrint() << "file size = " << header.size << endl;
//   sglPrint() << "offset_bits = " << header.offset_bits << endl;

   fread(&info_size, 1, sizeof(int), file);
#if defined(SGL_BIG_ENDIAN)
   sglEndianSwap(&info_size);
#endif
//   sglPrint() << "info size = " << info_size << endl;

   if (info_size == 12)
   {
//      sglPrint() << "old header" << endl;
      sglBMPInfoHeaderOld old_info;
      fread(&old_info, 1, sizeof(sglBMPInfoHeaderOld), file);

#if defined(SGL_BIG_ENDIAN)
      sglEndianSwap(&old_info.width);
      sglEndianSwap(&old_info.height);
      sglEndianSwap(&old_info.planes);
      sglEndianSwap(&old_info.bit_count);
#endif

      info.width = old_info.width;
      info.height = old_info.height;
      info.planes = old_info.planes;
      info.bit_count = old_info.bit_count;

      info.compression = 0;
      info.image_size = ((info.bit_count * info.width + 31) / 32)
         * 4 * info.height;
      info.x_pixels_per_meter = 0;
      info.y_pixels_per_meter = 0;
      info.colors_used = 0;
      info.colors_important = 0;
   }
   else if (info_size >= 40)
   {
//      sglPrint() << "new header" << endl;
      fread(&info, 1, sizeof(sglBMPInfoHeader), file);

#if defined(SGL_BIG_ENDIAN)
      sglEndianSwap(&info.width);
      sglEndianSwap(&info.height);
      sglEndianSwap(&info.planes);
      sglEndianSwap(&info.bit_count);
      sglEndianSwap(&info.compression);
      sglEndianSwap(&info.image_size);
      sglEndianSwap(&info.x_pixels_per_meter);
      sglEndianSwap(&info.y_pixels_per_meter);
      sglEndianSwap(&info.colors_used);
      sglEndianSwap(&info.colors_important);
#endif
   }
   else
   {
      sglPrintWarning() << "loadImageBMP(): invalid info header size in file "
                        << filepath << endl;
      return false;
   }
   return true;
}

/*****************************************************************************/
void *loadImageBMP(const string &filepath,
                   unsigned int &width,
                   unsigned int &height,
                   sglTexture::ExternalFormatEnum &format,
                   sglTexture::TypeEnum &type)
{
   FILE *file;
   if ((file = fopen(filepath.c_str(), "rb")) == NULL)
   {
      sglPrintWarning() << "loadImageBMP(): Error opening file "
                        << filepath << endl;
      return NULL;
   }

   char file_type[2];
   fread(&file_type, 1, 2, file);

   if (file_type[0] != 'B' || (file_type[1] != 'A' && file_type[1] != 'M'))
   {
      sglPrintWarning() << "loadImageBMP(): invalid type("
                        << file_type[0] << file_type[1] << ") in file "
                        << filepath << endl;
      fclose(file);
      return NULL;
   }
   int best_res = -1;
   long best_offset = 0;
   if (file_type[1] == 'A')
   {
      // File is an archive of bitmaps.  Find the bitmap with the most detail.

      int count = 0;
      for (;;)
      {
         sglBitmapArrayHeader array;
         fread(&array, 1, sizeof(sglBitmapArrayHeader), file);
#if defined(SGL_BIG_ENDIAN)
         sglEndianSwap(&array.next_offset);
         sglEndianSwap(&array.x_display);
         sglEndianSwap(&array.y_display);
#endif
         fread(&file_type, 1, 2, file);
         long curr_offset = ftell(file);
         if (file_type[0] == 'B' && file_type[1] == 'M')
         {
            sglBMPFileHeader header;
            sglBMPInfoHeader info;
            int info_size;
            if (readBMPHeaders(filepath, file, header, info, info_size))
            {
               ++count;
//               sglPrint() << "archive bmp size= " << info.image_size << endl;
               if (info.image_size >= best_res)
               {
                  best_res = info.image_size;
                  best_offset = curr_offset;
               }
// For debugging
//                if (info.compression)
//                {
//                   sglPrint() << "COMPRESSION ********" << endl;
//                   best_offset = curr_offset;
//                   break;
//                }
//                if (info.bit_count == 1)
//                {
//                   sglPrint() << "MONOCHROME ********" << endl;
//                   best_offset = curr_offset;
//                   break;
//                }
            }
         }
         if (array.next_offset)
         {
            fseek(file, (long)(array.next_offset+2), SEEK_SET);
         }
         else
         {
            break;
         }
      }
//      sglPrint() << count << " images in bmp archive" << endl;
      if (best_res == -1)
      {
         sglPrintWarning() << "loadImageBMP(): No bitmap found in bmp archive "
                           << filepath << endl;
         fclose(file);
         return NULL;
      }
      fseek(file, best_offset, SEEK_SET);
   }

   sglBMPFileHeader header;
   sglBMPInfoHeader info;
   int info_size;

   if (!readBMPHeaders(filepath, file, header, info, info_size))
   {
      fclose(file);
      return NULL;
   }

//   sglPrint() << "width = " << info.width << endl;
//   sglPrint() << "height = " << info.height << endl;
//   sglPrint() << "planes = " << info.planes << endl;
//   sglPrint() << "bit_count = " << info.bit_count << endl;
//   sglPrint() << "compression = " << info.compression << endl;
//   sglPrint() << "image_size = " << info.image_size << endl;
//   sglPrint() << "x_pixels_per_meter = " << info.x_pixels_per_meter << endl;
//   sglPrint() << "y_pixels_per_meter = " << info.y_pixels_per_meter << endl;
//   sglPrint() << "colors_used = " << info.colors_used << endl;
//   sglPrint() << "colors_important = " << info.colors_important << endl;

   if (info.width <= 0)
   {
      sglPrintWarning() << "loadImageBMP(): invalid width in file "
                        << filepath << endl;
      fclose(file);
      return NULL;
   }
   if (info.height <= 0)
   {
      sglPrintWarning() << "loadImageBMP(): invalid height in file "
                        << filepath << endl;
      fclose(file);
      return NULL;
   }
   if (info.planes != 1)
   {
      sglPrintWarning() << "loadImageBMP(): invalid # of planes in file "
                        << filepath << endl;
      fclose(file);
      return NULL;
   }
   if (info.bit_count != 1 && info.bit_count != 4 &&
       info.bit_count != 8 && info.bit_count != 24)
   {
      sglPrintWarning() << "loadImageBMP(): invalid bit count in file "
                        << filepath << endl;
      fclose(file);
      return NULL;
   }
   if (info.compression < 0 || info.compression > 2 ||
       (info.compression == 1 && info.bit_count != 8) ||
       (info.compression == 2 && info.bit_count != 4))
   {
      sglPrintWarning() << "loadImageBMP(): invalid compression in file "
                        << filepath << endl;
      fclose(file);
      return NULL;
   }
   if (info.bit_count == 1 && info.colors_used != 0 && info.colors_used != 2)
   {
      sglPrintWarning() << "loadImageBMP(): invalid # of colors used in file "
                        << filepath << endl;
      fclose(file);
      return NULL;
   }

   if (info_size > 40)
   {
      fseek(file, info_size-40, SEEK_CUR);
   }

   unsigned int num_pixels = info.width * info.height * 3;
   void *image = malloc(num_pixels);

   char *color_map = NULL;
   if (info.bit_count != 24)
   {
      if (!info.colors_used)
         info.colors_used = 1 << info.bit_count;
      color_map = (char*)malloc(info.colors_used * 3);
      for (int i=0; i<3*info.colors_used; i+=3)
      {
         fread(&color_map[i+2], 1, 1, file);
         fread(&color_map[i+1], 1, 1, file);
         fread(&color_map[i], 1, 1, file);
         if (info_size != 12)
            fseek(file, 1, SEEK_CUR);
      }
   }

   fseek(file, header.offset_bits, SEEK_SET);
   if (info.bit_count == 1)
   {
      long pad = ((info.width + 31) / 32)*4 - ((info.width + 7)/8);
      char *iptr = (char*)image;
      for (int y=0; y<info.height; ++y)
      {
         for (int x=0; x<info.width;)
         {
            char buf;
            fread(&buf, 1, 1, file);
            for (int i=0; i<8 && x<info.width; ++x, ++i)
            {
               if (buf & 0x80)
               {
                  *iptr = color_map[3]; ++iptr;
                  *iptr = color_map[4]; ++iptr;
                  *iptr = color_map[5]; ++iptr;
               }
               else
               {
                  *iptr = color_map[0]; ++iptr;
                  *iptr = color_map[1]; ++iptr;
                  *iptr = color_map[2]; ++iptr;
               }
               buf <<= 1;
            }
         }
         fseek(file, pad, SEEK_CUR);
      }
   }
   else if (info.bit_count == 4)
   {
      if (info.compression)
      {
         int x = 0, y = 0;
         char *iptr = (char*)image;
         while (y < info.height)
         {
            unsigned char buf;
            fread(&buf, 1, 1, file);
            if (buf == 0)
            {
               fread(&buf, 1, 1, file);
               if (buf == 0)
               {
                  x = 0; ++y;
                  iptr = (char*)image + y*info.width*3 + x*3;
               }
               else if (buf == 1)
               {
                  break;
               }
               else if (buf == 2)
               {
                  fread(&buf, 1, 1, file);
                  x += buf;
                  fread(&buf, 1, 1, file);
                  y += buf;
                  iptr = (char*)image + y*info.width*3 + x*3;
               }
               else
               {
                  unsigned char num = buf;
                  for (unsigned char i = 0; i < num; ++i)
                  {
                     fread(&buf, 1, 1, file);
                     int indx = 3*((buf >> 4) & 0x0f);
                     *iptr = color_map[indx]; ++iptr;
                     *iptr = color_map[indx+1]; ++iptr;
                     *iptr = color_map[indx+2]; ++iptr;
                     ++x;
                     if (++i < num)
                     {
                        indx = 3 * (buf & 0x0f);
                        *iptr = color_map[indx]; ++iptr;
                        *iptr = color_map[indx+1]; ++iptr;
                        *iptr = color_map[indx+2]; ++iptr;
                        ++x;
                     }
                  }
                  if (num & 0x03 == 0x01 || num & 0x03 == 0x02)
                     fseek(file, 1, SEEK_CUR);
               }
            }
            else
            {
               unsigned char num = buf;
               fread(&buf, 1, 1, file);
               int indx1 = 3*((buf >> 4) & 0x0f);
               int indx2 = 3 * (buf & 0x0f);
               for (unsigned char i = 0; i < num; ++i)
               {
                  *iptr = color_map[indx1]; ++iptr;
                  *iptr = color_map[indx1+1]; ++iptr;
                  *iptr = color_map[indx1+2]; ++iptr;
                  if (++i < num)
                  {
                     *iptr = color_map[indx2]; ++iptr;
                     *iptr = color_map[indx2+1]; ++iptr;
                     *iptr = color_map[indx2+2]; ++iptr;
                  }
               }
               x += num;
            }
         }
      }
      else
      {
         long pad = ((4*info.width + 31) / 32)*4 - ((4*info.width + 7)/8);
         char *iptr = (char*)image;
         for (int y=0; y<info.height; ++y)
         {
            for (int x=0; x<info.width; ++x)
            {
               unsigned char buf;
               fread(&buf, 1, 1, file);
               int indx = 3*((buf >> 4) & 0x0f);
               *iptr = color_map[indx]; ++iptr;
               *iptr = color_map[indx+1]; ++iptr;
               *iptr = color_map[indx+2]; ++iptr;
               if (++x < info.width)
               {
                  indx = 3 * (buf & 0x0f);
                  *iptr = color_map[indx]; ++iptr;
                  *iptr = color_map[indx+1]; ++iptr;
                  *iptr = color_map[indx+2]; ++iptr;
               }
            }
            fseek(file, pad, SEEK_CUR);
         }
      }
   }
   else if (info.bit_count == 8)
   {
      if (info.compression)
      {
         int x = 0, y = 0;
         char *iptr = (char*)image;
         while (y < info.height)
         {
            unsigned char buf;
            fread(&buf, 1, 1, file);
            if (buf == 0)
            {
               fread(&buf, 1, 1, file);
               if (buf == 0)
               {
                  x = 0; ++y;
                  iptr = (char*)image + y*info.width*3 + x*3;
               }
               else if (buf == 1)
               {
                  break;
               }
               else if (buf == 2)
               {
                  fread(&buf, 1, 1, file);
                  x += buf;
                  fread(&buf, 1, 1, file);
                  y += buf;
                  iptr = (char*)image + y*info.width*3 + x*3;
               }
               else
               {
                  unsigned char num = buf;
                  for (unsigned char i = 0; i < num; ++i)
                  {
                     fread(&buf, 1, 1, file);
                     int indx = 3*buf;
                     *iptr = color_map[indx]; ++iptr;
                     *iptr = color_map[indx+1]; ++iptr;
                     *iptr = color_map[indx+2]; ++iptr;
                     ++x;
                  }
                  if (num & 0x01)
                     fseek(file, 1, SEEK_CUR);
               }
            }
            else
            {
               unsigned char num = buf;
               fread(&buf, 1, 1, file);
               int indx = 3*buf;
               for (unsigned char i = 0; i < num; ++i)
               {
                  *iptr = color_map[indx]; ++iptr;
                  *iptr = color_map[indx+1]; ++iptr;
                  *iptr = color_map[indx+2]; ++iptr;
               }
               x += num;
            }
         }
      }
      else
      {
         long pad = ((info.width + 3) / 4)*4 - info.width;
         char *iptr = (char*)image;
         for (int y=0; y<info.height; ++y)
         {
            for (int x=0; x<info.width; ++x)
            {
               unsigned char buf;
               fread(&buf, 1, 1, file);
               int indx = 3*buf;
               *iptr = color_map[indx]; ++iptr;
               *iptr = color_map[indx+1]; ++iptr;
               *iptr = color_map[indx+2]; ++iptr;
            }
            fseek(file, pad, SEEK_CUR);
         }
      }
   }
   else
   {
      long pad = ((3*info.width + 3) / 4)*4 - info.width*3;
      char *iptr = (char*)image;
      for (int y=0; y<info.height; ++y)
      {
         fread(iptr, 1, 3*info.width, file);
         for (int x=0; x<info.width; ++x)
         {
            unsigned char tmp = *iptr;
            *iptr = *(iptr+2);
            *(iptr+2) = tmp;
            iptr += 3;
         }
         fseek(file, pad, SEEK_CUR);
      }
   }
//   sglPrint() << filepath << endl;
   type = sglTexture::eUNSIGNED_BYTE;
   format = sglTexture::eRGB;
   width = info.width;
   height = info.height;

   if (color_map)
      free(color_map);

   fclose(file);

   return image;
}
