/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglLoadImage_sgi.cpp
 *   Author: Tom Stimson
 *  Created: 26 March 2001
 *  Summary: read SGI format image files
 *****************************************************************************/

#include <sglLoadImage_sgi.hpp>
#include <sglImageLoader.hpp>
#include <sglOStream.hpp>
#include <stdio.h>

/*****************************************************************************/
struct sglSGIImageHeader {
   short          magic;
   unsigned char  storage;
   unsigned char  bytes_per_channel;
   unsigned short dimension;
   unsigned short xsize, ysize, zsize;
   unsigned int   min, max;
   unsigned int   waste_bytes1;
   char           name[80];
   unsigned int   colormap;
   char           waste_bytes2[404];
};

/*****************************************************************************/
template <class T>
void readSGIImage(sglSGIImageHeader &header, FILE *file,
                  unsigned short zsize, T *image)
{
   int row_size = header.xsize * header.bytes_per_channel;
   T* tmp = (T*)malloc(row_size);

   for (unsigned short z = 0; z < zsize; ++z)
   {
      for (unsigned short y = 0; y < header.ysize; ++y)
      {
         fread(tmp, 1, row_size, file);
         T *iPtr = tmp;
         T *oPtr = image + y*header.xsize*zsize + z;
         for (int i = 0; i < header.xsize; ++i)
         {
            *oPtr = *iPtr++;
            oPtr += zsize;
         }
      }
   }
   free(tmp);
}

/*****************************************************************************/
template <class T>
void writeSGIImage(FILE *file, const T *image, unsigned short xsize,
                   unsigned short ysize, unsigned short zsize,
                   unsigned char bytes_per_channel)
{
   int row_size = xsize * bytes_per_channel;
   T* tmp = (T*)malloc(row_size);

   for (unsigned short z = 0; z < zsize; ++z)
   {
      for (unsigned short y = 0; y < ysize; ++y)
      {
         const T *iPtr = image + y*xsize*zsize + z;
         T *oPtr = tmp;
         for (int i = 0; i < xsize; ++i)
         {
            *oPtr = *iPtr;
            oPtr++;
            iPtr += zsize;
         }
         fwrite(tmp, 1, row_size, file);
      }
   }
   free(tmp);
}

/*****************************************************************************/
template <class T>
void readRLESGIImage(sglSGIImageHeader &header, FILE *file,
                     unsigned short zsize, T *image)
{
   unsigned int tablen = header.ysize * header.zsize * sizeof(int);
   int *row_start = (int*)malloc(tablen);
   int *row_length = (int*)malloc(tablen);
   fread(row_start, 1, tablen, file);
   fread(row_length, 1, tablen, file);

   T* tmp = (T*)malloc(header.xsize * header.bytes_per_channel * 2);

   int row = 0;
   for (unsigned short z = 0; z < zsize; ++z)
   {
      for (unsigned short y = 0; y < header.ysize; ++y)
      {
#if defined(SGL_LITTLE_ENDIAN)
         sglEndianSwap(&row_start[row]);
         sglEndianSwap(&row_length[row]);
#endif
//         sglPrint() << "row[" << z << "][" << y << "]("
//                    << row << ") start=" << dec << row_start[row]
//                    << ", length=" << row_length[row] << endl;

         fseek(file, row_start[row], SEEK_SET);
         fread(tmp, 1, row_length[row], file);

         T *iPtr = tmp;
         T *oPtr = image + y*header.xsize*zsize + z;
         T count;
         do
         {
            T pixel = *iPtr++;
            count = pixel & 0x7F;
            if (count)
            {
               if (pixel & 0x80)
               {
                  while (count--)
                  {
                     *oPtr = *iPtr++;
                     oPtr += zsize;
                  }
               }
               else
               {
                  pixel = *iPtr++;
                  while (count--)
                  {
                     *oPtr = pixel;
                     oPtr += zsize;
                  }
               }
            }
//            sglPrint() << (iPtr-tmp) << endl;
         } while (count);
         ++row;
      }
   }
   free(row_start);
   free(row_length);
   free(tmp);
}

/*****************************************************************************/
template <class T>
T* addRLERun(T* oPtr, unsigned char count, const T* iPtr,
             unsigned short zsize, bool same)
{
   *oPtr = count;
   if (!same)
      *oPtr |= 0x80;
   ++oPtr;
   if (same)
      count = 1;
   for (unsigned char i = 0; i < count; ++i)
   {
      oPtr[i] = iPtr[i*zsize];
   }
   oPtr += count;
   return oPtr;
}

/*****************************************************************************/
template <class T>
void writeRLESGIImage(FILE *file, const T *image, unsigned short xsize,
                      unsigned short ysize, unsigned short zsize,
                      unsigned char bytes_per_channel)
{
   unsigned int tablen = ysize * zsize * sizeof(int);
   int *row_start = (int*)malloc(tablen);
   int *row_length = (int*)malloc(tablen);
   unsigned int row_size = xsize * zsize;

   T* tmp = (T*)malloc(xsize * ysize * zsize * bytes_per_channel * 2);

   int row = 0;
   T *oPtr = tmp;
   for (unsigned short z = 0; z < zsize; ++z)
   {
      for (unsigned short y = 0; y < ysize; ++y)
      {
         T *oStart = oPtr;
         const T *iPtr = image + y*row_size + z;

         unsigned int start = 0;
         bool same = (iPtr[0] == iPtr[zsize]);
         for (unsigned int x = zsize*2; x < row_size; x += zsize)
         {
            unsigned int run_len = (x-start)/zsize;
            if (run_len == 127 || (same && iPtr[x] != iPtr[start]))
            {
               oPtr = addRLERun(oPtr, run_len, iPtr + start, zsize, same);
               start = x;

               if (x+zsize < row_size)
               {
                  same = (iPtr[x] == iPtr[x+zsize]);
                  x += zsize;
               }
               else
                  same = false;
            }
            else if (!same && iPtr[x] == iPtr[x-zsize])
            {
               oPtr = addRLERun(oPtr, run_len-1, iPtr + start, zsize, same);
               start = x-zsize;
               same = true;
            }
         }
         oPtr = addRLERun(oPtr, (row_size-start)/zsize,
                          iPtr + start, zsize, same);

         *oPtr++ = 0;
         int row_len = oPtr-oStart;
         row_length[row] = row_len;

         // reuse preexisting rows if possible
         unsigned short i = 0;
         for (; i < row; ++i)
         {
            if (row_length[i] == row_len && 
                memcmp(oStart, &tmp[row_start[i]], row_len) == 0)
            {
               row_start[row] = row_start[i];
               oPtr = oStart;
               break;
            }
         }
         if (i == row)
            row_start[row] = oStart-tmp;

         ++row;
      }
   }
   for (int j = 0; j < row; ++j)
   {
      row_start[j] *= sizeof(T);
      row_start[j] += sizeof(sglSGIImageHeader) + 2*tablen;
      row_length[j] *= sizeof(T);
#if defined(SGL_LITTLE_ENDIAN)
      sglEndianSwap(&row_start[j]);
      sglEndianSwap(&row_length[j]);
#endif
   }

   fwrite(row_start, 1, tablen, file);
   fwrite(row_length, 1, tablen, file);
   fwrite(tmp, 1, (oPtr-tmp)*sizeof(T), file);

   free(row_start);
   free(row_length);
   free(tmp);
}

/*****************************************************************************/
void *loadImageSGI(const string &filepath,
                   unsigned int &width,
                   unsigned int &height,
                   sglTexture::ExternalFormatEnum &format,
                   sglTexture::TypeEnum &type)
{
   FILE *file;
   if ((file = fopen(filepath.c_str(), "rb")) == NULL)
   {
      sglPrintWarning() << "loadImageSGI(): Error opening file "
                        << filepath << endl;
      return NULL;
   }

   sglSGIImageHeader header;
   fread(&header, 1, sizeof(sglSGIImageHeader), file);

#if defined(SGL_LITTLE_ENDIAN)
   sglEndianSwap(&header.magic);
   sglEndianSwap(&header.dimension);
   sglEndianSwap(&header.xsize);
   sglEndianSwap(&header.ysize);
   sglEndianSwap(&header.zsize);
   sglEndianSwap(&header.min);
   sglEndianSwap(&header.max);
   sglEndianSwap(&header.colormap);
#endif

   //sglPrintDebug() << "file: " << filepath << endl;
   //sglPrintDebug() << "dimension: " << header.dimension << endl;
   //sglPrintDebug() << "xsize: " << header.xsize << endl;
   //sglPrintDebug() << "ysize: " << header.ysize << endl;
   //sglPrintDebug() << "zsize: " << header.zsize << endl;
   //sglPrintDebug() << "min: " << header.min << endl;
   //sglPrintDebug() << "max: " << header.max << endl;
   //sglPrintDebug() << "colormap: " << header.colormap << endl;
   //sglPrintDebug() << "bytes_per_channel: "
   //                << (int)header.bytes_per_channel << endl;
   //sglPrintDebug() << "storage: " << hex << (int)header.storage << dec << endl;

   if (header.magic != 474)
   {
      sglPrintWarning() << "loadImageSGI(): invalid magic number in file "
                        << filepath << endl;
      fclose(file);
      return NULL;
   }
   if (header.bytes_per_channel != 1 &&
       header.bytes_per_channel != 2)
   {
      sglPrintWarning() << "loadImageSGI(): invalid bytes_per_channel in file "
                        << filepath << endl;
      fclose(file);
      return NULL;
   }
   if (header.dimension == 1 && (header.ysize != 1 || header.zsize != 1))
   {
      sglPrintWarning() << "loadImageSGI(): invalid dimension (1) for a "
                        << header.ysize << " row & " << header.zsize
                        << " channel image: " << filepath << endl;
      if (header.zsize == 1)
         header.dimension = 2;
      else
         header.dimension = 3;
   }
   if (header.dimension == 2 && header.zsize != 1)
   {
      sglPrintWarning() << "loadImageSGI(): invalid dimension (2) for a "
                        << header.zsize
                        << " channel image: " << filepath << endl;
      header.dimension = 3;
   }
   if (header.storage != 0x01 && header.storage != 0)
   {
      sglPrintWarning() << "loadImageSGI(): invalid storage format in file "
                        << filepath << endl;
      fclose(file);
      return NULL;
   }

   unsigned short zsize = header.zsize;
   if (zsize > 4)
   {
      zsize = 4;
   }
   switch (zsize)
   {
      case 1:
         format = sglTexture::eLUMINANCE;
         break;
      case 2:
         format = sglTexture::eLUMINANCE_ALPHA;
         break;
      case 3:
         format = sglTexture::eRGB;
         break;
      case 4:
         format = sglTexture::eRGBA;
         break;
      default:
         sglPrintWarning() << "loadImageSGI(): invalid zsize in file "
                           << filepath << endl;
         fclose(file);
         return NULL;
   }
   switch (header.colormap)
   {
      case 0:
         // normal
         break;
      case 1:
         // dithered
         sglPrintWarning()
            << "loadImageSGI(): obsolete dithered format not supported "
            << filepath << endl;
         fclose(file);
         return NULL;
      case 2:
         // color-indexed
         sglPrintWarning()
            << "loadImageSGI(): obsolete color-indexed format not supported "
            << filepath << endl;
         fclose(file);
         return NULL;
      case 3:
         // colormap
         sglPrintWarning()
            << "loadImageSGI(): colormap format not supported "
            << filepath << endl;
         fclose(file);
         return NULL;
      default:
         sglPrintWarning()
            << "loadImageSGI(): unknown color format in file "
            << filepath << endl;
         fclose(file);
         return NULL;
   }

   unsigned int num_pixels = header.xsize * header.ysize * zsize;
   void *image = malloc(num_pixels * header.bytes_per_channel);

   //sglPrintDebug() << "zsize: " << header.zsize << endl;
   //sglPrintDebug() << "format: " << format << endl;
   //sglPrintDebug() << "num_pixels: " << num_pixels << endl;

   if (header.bytes_per_channel == 1)
   {
      type = sglTexture::eUNSIGNED_BYTE;

      if (header.storage == 0x01)
         readRLESGIImage(header, file, zsize, (char*)image);
      else
         readSGIImage(header, file, zsize, (char*)image);
   }
   else
   {
      //HACK Should this be UNSIGNED or not?
      type = sglTexture::eUNSIGNED_SHORT;

      if (header.storage == 0x01)
         readRLESGIImage(header, file, zsize, (short*)image);
      else
         readSGIImage(header, file, zsize, (short*)image);

#if defined(SGL_LITTLE_ENDIAN)
      short *tmp = (short*)image;
      for (unsigned int i = 0; i < num_pixels; ++i)
         sglEndianSwap(tmp++);
#endif
   }

   width = header.xsize;
   height = header.ysize;
   fclose(file);

   return image;
}


/*****************************************************************************/
bool writeImageSGI(const string &filepath,
                   const void *image,
                   unsigned int width,
                   unsigned int height,
                   sglTexture::ExternalFormatEnum format,
                   sglTexture::TypeEnum type)
{
   if (image == NULL)
      return false;

   FILE *file;
   if ((file = fopen(filepath.c_str(), "wb")) == NULL)
   {
      sglPrintWarning() << "loadImageSGI(): Error opening file "
                        << filepath << endl;
      return false;
   }

   void *tmp_image = NULL;

   sglSGIImageHeader header;
   memset(&header, 0, sizeof(header));
   header.magic = 474;

   bool rle = false;
   if (width > 1)
      rle = true;

   header.storage = (rle ? 0x01 : 0);

   sglTexture::TypeEnum new_type = type;
   switch (type)
   {
      case sglTexture::eSHORT:
      case sglTexture::eUNSIGNED_SHORT:
      case sglTexture::eINT:
      case sglTexture::eUNSIGNED_INT:
      case sglTexture::eFLOAT:
         header.bytes_per_channel = 2;
         //HACK Should this be UNSIGNED or not?
         new_type = sglTexture::eUNSIGNED_SHORT;
         break;
      default:
         header.bytes_per_channel = 1;
         new_type = sglTexture::eUNSIGNED_BYTE;
         break;
   }

   if (type != new_type)
   {
      tmp_image = sglImageLoader::changeImageType(
         format, type, width, height, new_type, image);
      type = new_type;
   }

   if (format == sglTexture::eLUMINANCE)
      header.dimension = 2;
   else
      header.dimension = 3;

   header.xsize = width;
   header.ysize = height;

   sglTexture::ExternalFormatEnum new_format = format;
   switch (format)
   {
      case sglTexture::eALPHA:
      case sglTexture::eRGBA:
         header.zsize = 4;
         new_format = sglTexture::eRGBA;
         break;
      case sglTexture::eLUMINANCE_ALPHA:
         header.zsize = 2;
         break;
      case sglTexture::eLUMINANCE:
         header.zsize = 1;
         break;
      default:
         header.zsize = 3;
         new_format = sglTexture::eRGB;
         break;
   }

   if (new_format != format)
   {
      void *tmp_image2 = sglImageLoader::reformatImage(
         format, type, width, height, new_format,
         (tmp_image ? (const void *)tmp_image : image));
      if (tmp_image)
         free(tmp_image);
      tmp_image = tmp_image2;
      format = new_format;
   }

   header.min = 0;
   header.max = 255;
   header.colormap = 0;

   unsigned short xsize = header.xsize;
   unsigned short ysize = header.ysize;
   unsigned short zsize = header.zsize;
   unsigned char bytes_per_channel = header.bytes_per_channel;

#if defined(SGL_LITTLE_ENDIAN)
   sglEndianSwap(&header.magic);
   sglEndianSwap(&header.dimension);
   sglEndianSwap(&header.xsize);
   sglEndianSwap(&header.ysize);
   sglEndianSwap(&header.zsize);
   sglEndianSwap(&header.min);
   sglEndianSwap(&header.max);
   sglEndianSwap(&header.colormap);
#endif

   fwrite(&header, 1, sizeof(sglSGIImageHeader), file);

   if (header.bytes_per_channel == 1)
   {
      const char *out_image =
         (tmp_image ? (const char*)tmp_image : (const char*)image);
      if (rle)
         writeRLESGIImage(file, out_image,
                          xsize, ysize, zsize, bytes_per_channel);
      else
         writeSGIImage(file, out_image,
                       xsize, ysize, zsize, bytes_per_channel);
   }
   else
   {
#if defined(SGL_LITTLE_ENDIAN)
      unsigned int num_pixels = xsize * ysize * zsize;
      if (tmp_image == NULL)
      {
         tmp_image = malloc(num_pixels*2);
         memcpy(tmp_image, image, num_pixels*2);
      }
      short *tmp = (short*)tmp_image;
      for (unsigned int i = 0; i < num_pixels; ++i)
         sglEndianSwap(tmp++);
#endif

      const short *out_image =
         (tmp_image ? (const short*)tmp_image : (const short*)image);
      if (rle)
         writeRLESGIImage(file, out_image,
                          xsize, ysize, zsize, bytes_per_channel);
      else
         writeSGIImage(file, out_image,
                       xsize, ysize, zsize, bytes_per_channel);
   }

   if (tmp_image)
      free(tmp_image);

   fclose(file);

   return true;
}
