/* vi:set ts=3 sw=3 et: */
/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglLoadImage_tga.cpp
 *   Author: Michael M. Morrison
 *  Created: March 2002
 *  Summary: read certain TGA format image files
 *           based on sglLoadFile_bmp, and format specs from Jan Horn
 *****************************************************************************/

#include <sglLoadImage_tga.hpp>
#include <sglOStream.hpp>
#include <stdio.h>

/*****************************************************************************/
struct sglTGAFileHeader
{
   unsigned char   fileType;
   unsigned char   colorMapType;
   unsigned char	 imageType;
   unsigned char	 colorMapSpec[4];
   unsigned short  origX, origY;
   unsigned short  width, height;
   unsigned char   bpp;
   unsigned char   info;
};


/*****************************************************************************/
bool readTGAHeader( FILE *file, sglTGAFileHeader &header )
{

   fread(&header, 1, sizeof(sglTGAFileHeader), file);

#if defined(SGL_BIG_ENDIAN)
   sglEndianSwap(&header.origX);
   sglEndianSwap(&header.origY);
   sglEndianSwap(&header.width);
   sglEndianSwap(&header.height);
#endif

   return true;
}

/*****************************************************************************/
void *loadImageTGA(const string &filepath,
                   unsigned int &width,
                   unsigned int &height,
                   sglTexture::ExternalFormatEnum &format,
                   sglTexture::TypeEnum &type)
{
   FILE *file;
   if ((file = fopen(filepath.c_str(), "rb")) == NULL)
   {
      sglPrintWarning() << "loadImageTGA(): Error opening file "
                        << filepath << endl;
      return NULL;
   }

   sglTGAFileHeader header;

   if (!readTGAHeader( file, header ))
   {
      fclose(file);
      return NULL;
   }

   sglPrint() << "width = " << header.width << endl;
   sglPrint() << "height = " << header.height << endl;

   if (header.width <= 0)
   {
      sglPrintWarning() << "loadImageTGA(): invalid width in file "
                        << filepath << endl;
      fclose(file);
      return NULL;
   }

   if (header.height <= 0)
   {
      sglPrintWarning() << "loadImageTGA(): invalid height in file "
                        << filepath << endl;
      fclose(file);
      return NULL;
   }

   if (header.imageType != 2)
   {
      sglPrintWarning() << "loadImageTGA(): compressed images not supported "
                        << filepath << endl;
      fclose(file);
      return NULL;
   }

   if (header.colorMapType != 0)
   {
      sglPrintWarning() << "loadImageTGA(): color mapped images not supported "
                        << filepath << endl;
      fclose(file);
      return NULL;
   }

   if (header.bpp != 24 && header.bpp != 32)
   {
      sglPrintWarning() << "loadImageTGA(): 24 & 32 bit depths supported "
                        << filepath << endl;
      fclose(file);
      return NULL;
   }

   unsigned int num_bytes;

   if( header.bpp == 32 )
      num_bytes = header.width * header.height * 4;
   else
      num_bytes = header.width * header.height * 3;

   unsigned char *image = (unsigned char *) malloc(num_bytes);

   if( image )
   {
      if( fread( image, num_bytes, 1, file ) != 1 )
      {
         free( image );
         return 0;
      }

      // swap image to RGB(a) instead of BGR(a)
      
      if( header.bpp == 24 )
      {
         for( int i=0; i < (header.width * header.height) ; i++ )
         {
            char tmp;

            tmp = image[i*3+0];
            image[i*3+0] = image[i*3+2];
            image[i*3+2] = tmp;
         }
      }
      else
      {
         for( int i=0; i < (header.width * header.height); i++ )
         {
            char tmp;

            tmp = image[i*4+0];
            image[i*4+0] = image[i*4+2];
            image[i*4+2] = tmp;
         }
      }
   }

   type = sglTexture::eUNSIGNED_BYTE;

   if( header.bpp == 32 )
      format = sglTexture::eRGBA;
   else
      format = sglTexture::eRGB;

   width = header.width;
   height = header.height;

   fclose(file);

   return (void *)image;
}
