/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglModelLoader.cpp
 *   Author: Scott McMillan
 *  Created: 15 February 2001
 *****************************************************************************/

#include <sglModelLoader.hpp>
#include <sglOStream.hpp>

#include <sglLoadFile_scm.hpp>
#include <sglLoadFile_xan.hpp>
#include <sglLoadFile_cmb.hpp>

map<string, void *> sglModelLoader::s_callback_userdata_map;
map<string, string> sglModelLoader::s_option_map;
map<string, sglModelLoader::loaderFunc> sglModelLoader::s_extension_func_map;
map<string, sglModelLoader::callbackFunction> 
                                        sglModelLoader::s_callback_func_map;
sglLock sglModelLoader::s_lock;

//----------------------------------------------------------------------------
class sglModelLoaderStaticInitializer
{
public:
   sglModelLoaderStaticInitializer()
      {
         // These are built-in.
         sglModelLoader::registerModelLoader("scm", sglLoadFile_scm);
         sglModelLoader::registerModelLoader("xan", sglLoadFile_xan);
         sglModelLoader::registerModelLoader("cmb", sglLoadFile_cmb);
      }
   ~sglModelLoaderStaticInitializer() {}
};
static sglModelLoaderStaticInitializer initializer;


//----------------------------------------------------------------------------
sglNode* sglModelLoader::loadModel(const string &filepath, unsigned int mode)
{
   string ext = extractExtensionFromFilename(filepath);
   toLower(ext);
   loaderFunc func = getModelLoader(ext);
   if (!func)
   {
      string dll;
      getDLL(ext, dll);
      if (!dll.empty())
      {
         loadDLL(dll);
         func = getModelLoader(ext);
      }
   }
   if (func)
      return (*func)(filepath, mode);

   sglPrintWarning() << "Error loading file " << filepath
                     << "; no reader found for files of this format." << endl;
   return NULL;
}

//----------------------------------------------------------------------------
void sglModelLoader::registerModelLoader(
   const string &extension,
   sglModelLoader::loaderFunc model_loader_func)
{
   string ext = extension;
   toLower(ext);

   s_lock.acquire();
   if (model_loader_func)
   {
      //cerr << "Model loader for " << ext << " registered" << endl;
      s_extension_func_map[ext] = model_loader_func;
   }
   else  // remove the registration???
   {
      map<string, loaderFunc>::iterator elt =
         s_extension_func_map.find(ext);
      if (elt != s_extension_func_map.end())
          s_extension_func_map.erase(elt);
   }
   s_lock.release();
}

//----------------------------------------------------------------------------
sglModelLoader::loaderFunc sglModelLoader::getModelLoader(
   const string &extension)
{
   loaderFunc ret = NULL;

   string ext = extension;
   toLower(ext);

   s_lock.acquire();
   map<string, loaderFunc>::iterator elt =
      s_extension_func_map.find(ext);
   if (elt != s_extension_func_map.end())
      ret = elt->second;
   s_lock.release();

   return ret;
}

//----------------------------------------------------------------------------
void sglModelLoader::registerCallback(
   const string &extension,
   sglModelLoader::callbackFunction callback_func,
	 void * userData )
{
   string ext = extension;
   toLower(ext);

   s_lock.acquire();
   if (callback_func)
   {
      //cerr << "callback func for " << ext << " registered" << endl;
      s_callback_func_map[ext] = callback_func;
      s_callback_userdata_map[ext] = userData;
   }
   else  // remove the registration
   {
      map<string, callbackFunction>::iterator elt =
         s_callback_func_map.find(ext);
      if (elt != s_callback_func_map.end())
			{
          s_callback_func_map.erase(elt);
			}
      map<string, void *>::iterator eltu =
         s_callback_userdata_map.find(ext);
      if (eltu != s_callback_userdata_map.end())
			{
          s_callback_userdata_map.erase(eltu);
			}
   }
   s_lock.release();
}

//----------------------------------------------------------------------------
sglModelLoader::callbackFunction sglModelLoader::getCallback(
   const string &extension,
	 void **userData )
{
   callbackFunction ret = NULL;

	 *userData = 0;

   string ext = extension;
   toLower(ext);

   s_lock.acquire();
   map<string, callbackFunction>::iterator elt =
      s_callback_func_map.find(ext);
   if (elt != s_callback_func_map.end())
      ret = elt->second;

   map<string, void *>::iterator eltu =
      s_callback_userdata_map.find(ext);
   if (eltu != s_callback_userdata_map.end())
      *userData = eltu->second;
   s_lock.release();

   return ret;
}

//----------------------------------------------------------------------------
void sglModelLoader::registerOptionString( const string &extension, 
                                           const string &options )
{
   string ext = extension;
   toLower(ext);

   s_lock.acquire();

   s_option_map[ext] = options;

   s_lock.release();
}

//----------------------------------------------------------------------------
string sglModelLoader::getOptionString(
   const string &extension)
{
   string ret("NONE");

   string ext = extension;
   toLower(ext);

   s_lock.acquire();

   map<string, string>::iterator elt = s_option_map.find(ext);
   if (elt != s_option_map.end())
      ret = elt->second;

   s_lock.release();

   return ret;
}
