/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglModelLoader.hpp
 *   Author: Scott McMillan
 *  Created: 15 February 2001
 *  Summary: Abstract base class for all 3D model loaders.
 *****************************************************************************/

#ifndef SGL_MODEL_LOADER
#define SGL_MODEL_LOADER

#include <sgldb.h>
#include <sglDataLoader.hpp>
#include <map>

class sglNode;



/** Abstract base class for 3D models.
 */

class SGLDB_DLL_API sglModelLoader : public sglDataLoader
{
public:
   /** Valid loader mode bit flags. */
   enum ModeEnum
   {
      eNONE                    = 0x00,  // default
      eSTRIPIFY                = 0x01,
      eCOMPUTE_FACETED_NORMALS = 0x02,
      eCOMPUTE_SMOOTH_NORMALS  = 0x04
   };

   /** The prototype for loader functions. */
   typedef sglNode *(*loaderFunc)(const string &, unsigned int);

   /** The prototype for callback functions. */
   typedef bool (*callbackFunction)( sglNode * node, void * loaderNode, 
                                     string &function, void * userData );

public:
   /** Destructor. */
   virtual ~sglModelLoader() = 0;

   /** Load the given model file using whatever loader has been registered
       for files of that type.
       @param filepath the full pathname of the file to load
       @return NULL on failure, otherwise a pointer to the root node for
              the model subtree (note, all memory allocated for this model
              will be automatically freed when the root note is destructed)
    */
   static sglNode* loadModel(const string &filepath, unsigned int mode);

   /** Register a model function pointer and the extension it supports.  Note
       that a single function can support multiple extensions.  Registering an
       extension a second time will replace the existing entry.
       @param extension extension string that the function supports.
       @param model_loader_func pointer to function supporting the extension
    */
   static void registerModelLoader(const string &extension,
                                   loaderFunc model_loader_func);

   /** Get the registered model loader function pointer based on extension.
       @param extension extension to search on.
       @return function pointer, or NULL if not found
    */
   static loaderFunc getModelLoader(const string &extension);

   /** Register a callback function pointer and the extension it is for.
        Applications should use this function to register callbacks for the
        extensions they are interested in.
       Registering an extension a second time will replace the existing entry.
       @param extension extension string that the function supports.
       @param callback_func pointer to function supporting the extension.
       @param userData user data to pass to the callback when executed.
    */
   static void registerCallback(const string &extension,
                                   callbackFunction callback_func,
																	 void *userData );

   /** Get the registered callback function pointer based on extension.  Model
        loaders should use this function to retrieve the callback for their
        extension.
       @param extension extension to search on.
       @param ud pointer to void pointer to receive the userData
       @return function pointer, or NULL if not found
    */
   static callbackFunction getCallback(const string &extension, void **ud );

   /** Register a loader option string, and the extension it is for.  The
        application should use this function to register an option string for
        individual model loaders.
       Registering an extension a second time will replace the existing entry.
       @param extension extension that the option string is for.
       @param options the loader option string
    */
   static void registerOptionString(const string &extension,
                                    const string &options );

   /** Get the option string for an extension.  Model loaders should use this
        function to determine if an option string has been registered for
        their extension.
       @param extension extension to search on.
       @return the option string for this extension, or a string containing
               NONE if no options have been registered.
    */
   static string getOptionString(const string &extension);

protected:
   sglModelLoader();
   sglModelLoader(const sglModelLoader &);
   sglModelLoader &operator=(const sglModelLoader &);

private:  // Variables
   static map<string, loaderFunc> s_extension_func_map;
   static map<string, callbackFunction> s_callback_func_map;
   static map<string, void *> s_callback_userdata_map;
   static map<string, string> s_option_map;
   static sglLock s_lock;
};

#endif
