/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 * SGL's FLT loader: Copyright (C) 2001  Michael M. Morrison
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: ObjOptimizer.cpp
 *  Created: November 2001
 *  Author : Michael M. Morrison
 *  Summary: Sorts state, and stripifies at the OpenFLT Object level.
 *****************************************************************************/

#ifndef _OBJ_OPTIMIZER_H_
#define _OBJ_OPTIMIZER_H_

#ifdef _MSC_VER
#pragma warning( disable : 4786 )
#endif

#include <map>
#include <list>
#include <utility>
#include <vector>
#include <string>
#include "sglTriStripper.hpp"

using namespace std;

class OOStateAttr {
	
public:
	OOStateAttr() : stateFlags(0), textureIndex(0), materialIndex(0), 
																							primaryColorIndex(0) {}
	~OOStateAttr() {}

	enum StateFlags {
		HAS_PRIMARY_COLOR		= 0x01,
		HAS_TEXTURE					= 0x02,
		HAS_MATERIAL 				= 0x04,
		HAS_VNORMALS				= 0x08,
		HAS_VCOLORS					= 0x10
	};

	enum {
		NONE
	};

	void setStateFlags( int sf ) { stateFlags = sf; } 
	int getStateFlags( void ) { return stateFlags; }

	void addTexture( unsigned int indx ) {
		if( indx != NONE ) {
			stateFlags |= HAS_TEXTURE;
		}
		textureIndex = indx;
	}

	void addMaterial( unsigned int indx ) { 
		if( indx != NONE ) {
			stateFlags |= HAS_MATERIAL;
		}
		materialIndex = indx;
	}
	
	void addPrimaryColor( unsigned int indx ) { 
		if( indx != NONE ) {
			stateFlags |= HAS_PRIMARY_COLOR;
		}
		primaryColorIndex = indx;
	}

	void addStateFlags( int sf ) { stateFlags |= sf; }

	unsigned int getMaterial( void )     { return materialIndex; }
	unsigned int getTexture( void )      { return textureIndex; }
	unsigned int getPrimaryColor( void ) { return primaryColorIndex; }

	bool isEqual( int sf, 
								unsigned int primaryColor,
								unsigned int texture,
								unsigned int material )
	{
		if( stateFlags == sf &&
				primaryColor == primaryColorIndex &&
				texture == textureIndex &&
				material == materialIndex )
			return true;
		else
			return false;
	}

private:

	// had to be 'int' for bitfield enum
	int stateFlags;

	unsigned int textureIndex;
	unsigned int materialIndex;
	unsigned int primaryColorIndex;

	// per-vertex attributes are stored elsewhere.

};

class ObjectOptimizer {

public:
	ObjectOptimizer( char * objName );
	~ObjectOptimizer();

	void addPrimitive(	int stateFlags, 
											vector<unsigned int> & indexList,
											unsigned int primaryColor,
											unsigned int texture,
											unsigned int material );

	// optimize the geometry and stripify
	void optimize( void );

	// vertex list
	void vertexListOptimize( void );
	void vertexListBegin( void );
	void vertexListEnd( void );
	void vertexListEntry( unsigned int, float, float, float );

	typedef map<OOStateAttr *, 
						sglTriStripper<sglVertStructV> * >::const_iterator mapIter;
	typedef map<OOStateAttr *, sglTriStripper<sglVertStructV> * > optMapType;
	typedef optMapType::const_iterator optMapIter;

	class SortVertex3D {

		public:
			SortVertex3D( unsigned int idx, float x, float y, float z ) {
				mIndx = idx;
				mX = x;
				mY = y;
				mZ = z;
				mDist = x*x+y*y+z*z;
			}

			unsigned int mIndx;
			float mX, mY, mZ, mDist;
		
	};

private:
	string * mName;

	// vertexlist optimization maps

	// state bins
	optMapType mStateBin;

	// optimized output

	vector<unsigned int> mRemappedIndices;
	map<unsigned int, unsigned int> mReverseIndexMap;

	OOStateAttr * findStateAttr(  int sf,
																unsigned int primaryColor,
																unsigned int texture,
																unsigned int material );

	OOStateAttr * addStateAttr( int sf,
																unsigned int primaryColor,
																unsigned int texture,
																unsigned int material );

public:

	// getters
	const optMapType * getStateBins() { return &mStateBin; }
	const vector<unsigned int> * getRemappedIndices() { 
		return &mRemappedIndices;
	}
	const map<unsigned int, unsigned int> * getReverseIndexMapping() {
		return &mReverseIndexMap;
	}

private:
	// Copy constructor forbidden
	ObjectOptimizer(const ObjectOptimizer &);
	// Assignment operator forbidden
	ObjectOptimizer &operator=(const ObjectOptimizer &);

};

#endif /* _OBJ_OPTIMIZER_H_ */
