/* vi:set ts=3 sw=3 expandtab: */
/* I do not believe that the "original" vi has 'expandtab' (TABs=spaces) */
/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 * SGL's FLT loader: Copyright (C) 2001  Michael M. Morrison
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglLoadFile_flt.cpp
 *  Created: June 2001
 *  Author : Michael M. Morrison
 *  Summary: Wrapper class/frontend to the .flt library.  Based on
 *           sglLoadFile_obj.cpp.
 *****************************************************************************/

#include <sglLoadFile_flt.hpp>
#include <sglImageLoader.hpp>

#include <sglGroup.hpp>
#include <sglTransform.hpp>
#include <sglTranslate.hpp>
#include <sglLOD.hpp>
#include <sglIndexedBitMaskSwitch.hpp>
#include <sglKeyFrameAnimation.hpp>
#include <sglGeode.hpp>

#include <sglGeoSet.hpp>
#include <sglQuadSet.hpp>
#include <sglLineSet.hpp>
#include <sglPointSet.hpp>
#include <sglPolygonSet.hpp>
#include <sglTriangleSet.hpp>
#include <sglIndexedTriangleStripSet.hpp>
#include <sglIndexedTriangleSet.hpp>

#include <sglMonoIndexedTriangleStripSet.hpp>
#include <sglMonoIndexedTriangleSet.hpp>

#include <sglStatelet.hpp>
#include <sglMaterial.hpp>
#include <sglTexture2D.hpp>
#include <sglTexEnv.hpp>
#include <sglTexGen.hpp>

#undef USE_BSP
#ifdef USE_BSP
#include <sglBSP.hpp>
#endif

#include <sglBillboard.hpp>

#include <math.h>
#include <sys/stat.h>
#include <flt.h>

#undef DEBUG_STRIPPER

//----------------------------------------------------------------------------

extern "C"
{
   SGLFLT_DLL_API
   void *loadModelFLT(const string &filepath,
                      unsigned int mode)
   {
      sglLoadFile_flt ldr;
      sglNode *model = ldr.loadFile(filepath, mode);
      return model;
   }
}

//============================================================================
// class sglLoadFile_flt
//============================================================================

map<string, sglTexture2D*> sglLoadFile_flt::s_tex_map;
map<string, CombinerEntry*> sglLoadFile_flt::s_combiner_map;

//----------------------------------------------------------------------------
sglLoadFile_flt::sglLoadFile_flt()
{
   //
   // tunable params  see sglLoadFile_flt.hpp for details
   //
   m_compute_all_normals = false;
   m_compute_normals_if_absent = false;

   m_use_colormaterial = false;
   m_compress_textures = false;

   m_verbose = false;

   m_use_combiner_file = false;

   m_retain_comments = true;

   m_enable_stripper = true;
   
   m_use_attr_files = true;

   m_use_bsp = 0;

   char * ptr;
   if( (ptr = (char *)getenv( "FLT_ENABLE_STRIPPER" )) && *ptr == '0' )
      m_enable_stripper = false;

   if( (ptr = (char *)getenv( "FLT_USE_ATTR_FILES" )) && *ptr == '0' )
      m_use_attr_files = false;

   if( (ptr = (char *)getenv( "FLT_BSP_MODE" )) && *ptr == '1' )
      m_use_bsp = 1;

   m_enable_lighting = false;
   m_enable_shading = false;

   m_enable_line_aa = false;

   // must pick one or the other below, if both off, use sglTransparency,
   m_use_blendfunc_transparency = false;
   m_use_alphafunc_transparency = true;
   m_alpha_threshold = 0.3f;

   m_free_texture_images = false; // don't turn this on yet.

   m_default_min_filter = sglTexture::eLINEAR_MIPMAP_LINEAR;
   m_default_mag_filter = sglTexture::eLINEAR;
   m_default_compression_level = sglTexture::eCOMPRESS_MAX;

   m_min_texture_dimension = 1;

   m_display_stats = true;
   m_num_points = 
   m_num_lines = 
   m_num_tris = 
   m_num_quads =
   m_num_polys = 
   m_num_strip_tris = 
   m_num_strip_quads =
   m_num_strip_polys = 
   m_avg_strip_len = 
   m_max_strip_len = 
   m_num_lines = 
   m_num_verts =
   m_num_strips =
   m_num_materials =
   m_num_textures =
   m_opaque_texels =
   m_trans_texels = 0;

   m_min_strip_len = 99999;
}

//----------------------------------------------------------------------------
sglLoadFile_flt::~sglLoadFile_flt()
{
   // free all resources
   if( m_flt ) 
      fltFileFree( m_flt );
}

//----------------------------------------------------------------------------
static void 
postTextureBindCallback( sglTexture * tex )
{
   void *image = 0;
   sglTexture::ExternalFormatEnum format; 
   sglTexture::TypeEnum type;
   unsigned int num_s, num_t;
   bool border;

   //
   // NOTE: Unsafe now, since image would be free'd twice
   //
   ((sglTexture2D *)tex)->getImage( image, format, type, num_s, num_t, border );

   if( image )
   {
      free( image );
   }
}

//----------------------------------------------------------------------------
sglTexture2D *
sglLoadFile_flt::lookupTexture( int indx, CombinerEntry ** ce )
{
   sglTexture2D * tex = 0;
   bool found = false;

   if( (m_use_combiner_file == false) || (ce == 0) )
   {
      // first, check "found" list
      map<int, sglTexture2D *>::iterator fIter = 
                                             m_textures_found_map.find( indx );

      if ( fIter != m_textures_found_map.end() )
      {
         return fIter->second;
      }

      // next, check "not found" list
      map<int, sglTexture2D *>::iterator nfIter = 
                                         m_textures_not_found_map.find( indx );

      if ( nfIter != m_textures_not_found_map.end() )
      {
         return 0;
      }
   }

   // set defaults
   if( ce )
      *ce = 0;

   FltTexture * t = fltLookupTexture(m_flt, indx);

   if (t)
   {
      char fname[256];
      char searchName[256];

      strcpy( searchName, t->ID );

      // first, check combiner file
      if( m_use_combiner_file )
      {
         if( !s_combiner_map.empty() )
         {
            map<string, CombinerEntry*>::iterator cIter;
            char * ptr;

            if( (ptr = strrchr( searchName, '/'  )) ||
                (ptr = strrchr( searchName, '\\' )) )
               ptr++;
            else
               ptr = t->ID;

            cIter = s_combiner_map.find( ptr );

            if( cIter != s_combiner_map.end() )
            {
               *ce = cIter->second;

               // found in combiner file, switch file name
               strcpy( searchName, (*ce)->combinedTexFile->c_str() );
            }
         }
      }

      if (fltFindFile(m_flt, searchName, fname))
      {
         // may fail
         tex = createTexture(fname);
         if( tex )
            found = true;
      }
      else
      {
         string foundname;
         if (sglDataLoader::findFile(searchName, &m_alternate_path, foundname))
         {
            char *tid = (char*)malloc(foundname.size()+1);
            tid[0] = '\0';
            strcat(tid, foundname.c_str());

            if (fltFindFile(m_flt, tid, fname))
            {
               tex = createTexture(fname);
               if( tex )
                  found = true;
            }
            free(tid);
         }
         else
         {
            sglPrintWarning() << "sglLoadFile_flt::buildScene error: "
                                 << "Unable to find texture: '"
                                 << searchName << endl;
         }
      }
   }
   else
   {
      sglPrintWarning() << "sglLoadFile_flt::buildScene error: "
      << "unable to find texture pattern "
      << hex << indx << dec
      << " in hierarchy." << endl;
   }

   if( found == false )
   {
      m_textures_not_found_map[ indx ] = 0; 
      return 0;
   }
   else
   {
      m_textures_found_map[ indx ] = tex;
      return tex;
   }
}

//----------------------------------------------------------------------------
void sglLoadFile_flt::buildInstanceDefinition(
                                 FltNode * node, 
                                 FltFile * flt,
                                 sglVec3f * offset )
{
   FltInstanceDefinition * idef = (FltInstanceDefinition *)node;
   sglGroup * group = new sglGroup();

   if( m_verbose )
      sglPrintInfo() << "sglLoadFile_flt::buildScene "
                     << "creating instace definition: "
                     << idef->instance << endl;

   for (unsigned int i=0;i<idef->node.numChildren;i++)
      buildScene(group, idef->node.child[i], flt, offset);

   m_instance[ idef->instance ] = group;
}

//----------------------------------------------------------------------------
sglNode * sglLoadFile_flt::createInstance( int indx )
{
   sglNode * node = 0;

   map<int, sglNode *>::iterator elt = m_instance.find( indx );

   if (elt == m_instance.end())
   {
      sglPrintWarning() << "sglLoadFile_flt:buildScene: instance " <<
                              indx << " not found." << endl;
      return 0;
   }
   else
   {
      node = elt->second;

      if( node )
         return node->clone( sglObject::eCLONE_NODES_RECURSIVE );
      else
         return 0;
   }
}

//----------------------------------------------------------------------------
sglTexture2D *sglLoadFile_flt::createTexture(const string &filename)
{
   sglTexture2D * tex = NULL;

   // lookup texture to see if we have already loaded it
   map<string, sglTexture2D*>::iterator elt = s_tex_map.find(filename);

   if (elt == s_tex_map.end())
   {
      sglTexture::ExternalFormatEnum format;
      sglTexture::TypeEnum type;
      unsigned int num_s, num_t;
      void * image = NULL;

      image = sglImageLoader::loadImage(filename, num_s, num_t, format, type);

      if (image && num_s && num_t)
      {
         // check attributes (thanks to Yury Demendeev)
         FltTxAttributes* attr = NULL;
         sglTexture::WrapEnum s_repeat = sglTexture::eREPEAT;
         sglTexture::WrapEnum t_repeat = sglTexture::eREPEAT;
         sglTexture::FilterEnum min_filter = m_default_min_filter;
         sglTexture::FilterEnum mag_filter = m_default_mag_filter;

         attr = fltLoadAttributes( (filename + ".attr").c_str() );

         if( attr ) 
         {
            // not sure if this one is necessary
            if(attr->repetitionType < 2)
            {
               s_repeat = (attr->repetitionType == 0) ? sglTexture::eREPEAT :
                                                         sglTexture::eCLAMP;
               t_repeat = s_repeat;
            }

            if(attr->repetitionU < 2)
               s_repeat = (attr->repetitionU == 0) ? sglTexture::eREPEAT :
                                                      sglTexture::eCLAMP;

            if(attr->repetitionV < 2)
               t_repeat = (attr->repetitionV == 0) ? sglTexture::eREPEAT :
                                                      sglTexture::eCLAMP;

            switch(attr->minificationFilter)
            {
               case 0:
                  min_filter = sglTexture::eNEAREST;
                  break;

               case 1:
                  min_filter = sglTexture::eLINEAR;
                  break;

               case 2:
               case 3:
                  min_filter = sglTexture::eNEAREST_MIPMAP_NEAREST;
                  break;

               case 4:
                  min_filter = sglTexture::eNEAREST_MIPMAP_LINEAR;
                  break;

               case 5:
                  min_filter = sglTexture::eLINEAR_MIPMAP_NEAREST;
                  break;

               default:
                  min_filter = sglTexture::eLINEAR_MIPMAP_LINEAR;
                  break;
            }

            switch(attr->magnificationFilter)
            {
               case 0:
                  mag_filter = sglTexture::eNEAREST;
                  break;

               default:
                  mag_filter = sglTexture::eLINEAR;
                  break;
            }

            fltFreeAttributes(attr);
            attr = NULL;
         }

         // ignore "small" textures
         if( num_s <= (unsigned)m_min_texture_dimension ||
             num_t <= (unsigned)m_min_texture_dimension )
            return 0;

         tex = new sglTexture2D;

         tex->setAutoMipmapFlag(true);

         tex->setWrapMode( s_repeat, t_repeat );
         tex->setMinFilter( min_filter );
         tex->setMagFilter( mag_filter );

         // set max anisotropic filter
         //tex->setMaxAnisotropy( tex->getMaxSupportedMaxAnisotropy() );

         tex->setImage(image, format, type, num_s, num_t, false);

         if( m_compress_textures ) 
         {
            tex->setCompressionLevel( m_default_compression_level );
         }

         if( m_free_texture_images )
            tex->setPostBindCallback( postTextureBindCallback );

         if( m_verbose )
            sglPrintInfo()<<"sglLoadFile_flt::createTexture: Texture loaded: '"
                       << filename << "' (" << num_s << "x" << num_t << ")"
                       << endl;

         m_num_textures++;
         if( format == sglTexture::eRGB )
            m_opaque_texels += (num_s * num_t);
         else
            m_trans_texels += (num_s * num_t);
      }
      else
         return 0;

      s_tex_map[filename] = tex;

   }
   else
   {
      //sglPrintInfo() << "sglLoadFile_flt::createTexture: Texture reused: '"
      //               << filename << "'" << endl;
      tex = elt->second;
   }
   return tex;
}

//----------------------------------------------------------------------------
// taken from sgl3ds.cpp //
static inline void 
computeFaceNormal( const sglVec3f &v0,
                   const sglVec3f &v1,
                   const sglVec3f &v2,
                   sglVec3f &normal )
{
   sglVec3f a, b;

   a.sub(v1, v0);
   b.sub(v2, v0);

   normal.cross(a, b);
   normal.normalize();
}


//----------------------------------------------------------------------------
void
sglLoadFile_flt::parseGeometry( sglGroup * parent, ObjectOptimizer *o, 
                                                FltNode *node, FltFile *flt,
                                                sglVec3f * offset )
{
   if (node->type != FLTRECORD_FACE)
   {
      unsigned int i;
      for(i=0;i<node->numChildren;i++)
         parseGeometry( parent, o, node->child[i], flt, offset );

      return;
   }

   FltFace * face = (FltFace *)node;

   // skip hidden faces
   if( face->miscFlags & FLTFACEMF_HIDDEN )
      return;

   // check for various state we don't handle here yet
   if( face->billboardFlags == FLTFACEBB_AXIALROTATE ||
       face->billboardFlags == FLTFACEBB_POINTROTATE ||
       face->drawType == FLTFACEDT_DRAWWIREFRAME || 
       face->drawType == FLTFACEDT_DRAWWIREFRAMECLOSE ||
       face->drawType == FLTFACEDT_DRAWSOLIDNOBACKFACE )
   {
      // use regular func for this
      if( parent )
         buildScene( parent, node, flt, offset );

      return;
   }

   vector<unsigned int>vec;

#ifdef DEBUG_STRIPPER
   printf("Face: %s\t", fltSafeNodeName( node ) );
#endif

   for (uint32 j=0;j<node->numChildren;j++)
   {
      FltVertexList * vlist = (FltVertexList *)node->child[j];

      if (vlist->numVerts == 3)
      {
#ifdef DEBUG_STRIPPER
         printf("(Triangle)\n");
#endif
         m_num_strip_tris++;
      }
      else if( vlist->numVerts == 4 )
      {
#ifdef DEBUG_STRIPPER
         printf("(Quad)\n");
#endif
         m_num_strip_quads++;
      }
      else
      {
#ifdef DEBUG_STRIPPER
         printf("(Polygon)\n");
#endif
         m_num_strip_polys++;
      }

      for( unsigned int i=0;i<vlist->numVerts;i++) {
#ifdef DEBUG_STRIPPER
         printf("\t[%d]:[%d]: %d (%f %f %f)\n", j, i, vlist->indexList[i], 
                  flt->vertPalette->verts[vlist->indexList[i]]->x,
                  flt->vertPalette->verts[vlist->indexList[i]]->y,
                  flt->vertPalette->verts[vlist->indexList[i]]->z );
#endif
         vec.push_back( vlist->indexList[i] );
      }

      //
      // May rework interface here
      //

      int flags = 0;
      unsigned int texture = OOStateAttr::NONE, 
                   primaryColor = OOStateAttr::NONE, 
                   material = OOStateAttr::NONE;

      // check per-vertex attributes
      if (  (face->lightMode == FLTFACELM_FCVN || 
             face->lightMode == FLTFACELM_VCVN) && m_enable_lighting )
         flags |= OOStateAttr::HAS_VNORMALS;

      if (  (face->lightMode == FLTFACELM_VCNOTILLUM || 
             face->lightMode == FLTFACELM_VCVN) && m_enable_shading )
         flags |= OOStateAttr::HAS_VCOLORS;

      // check per-poly attrs
      if (face->texturePatternIndex != -1)
      {
         flags |= OOStateAttr::HAS_TEXTURE;
         texture = face->texturePatternIndex;
      }

      if (face->primaryColorIndex != 0xffffffff)
      {
         flags |= OOStateAttr::HAS_PRIMARY_COLOR;
         primaryColor = face->primaryColorIndex;
      }

      if ( face->textureWhite )
      {
         // reset/override primary color
         flags |= OOStateAttr::HAS_PRIMARY_COLOR;

         // large int to end up with "white"
         primaryColor = 0xfffffffe;
      }

      if( (face->materialIndex != -1) && m_enable_lighting )
      {
         flags |= OOStateAttr::HAS_MATERIAL;
         material = face->materialIndex;
      }

      o->addPrimitive( flags, vec, primaryColor, texture, material );

      vec.clear();
   }
}

//----------------------------------------------------------------------------
sglGeode *
sglLoadFile_flt::buildGeometry( sglGroup * parent, FltNode * node, 
                                          FltFile * flt, sglVec3f * offset )
{
   sglGeode * geode = new sglGeode;
   ObjectOptimizer * opt = new ObjectOptimizer( fltSafeNodeName( node ) );
   const ObjectOptimizer::optMapType * stateBins;
   ObjectOptimizer::optMapIter om;

   bool vnormalsUsed = false, vcolorsUsed = false;

   const map<unsigned int, unsigned int> * reverseIdx;
   const vector<unsigned int> * remappedIndices;
   map<unsigned int, unsigned int>::const_iterator ridxIter;

   // build up all geometry under this object
   parseGeometry( parent, opt, node, flt, offset );

#ifdef DEBUG_STRIPPER
   printf("Before Optimizer\n");
   fflush(stdout);
#endif

   // optimize it
   opt->optimize();

#ifdef DEBUG_STRIPPER
   printf("After Optimizer\n");
#endif

   // retrieve maps
   stateBins       = opt->getStateBins();
   reverseIdx      = opt->getReverseIndexMapping();
   remappedIndices = opt->getRemappedIndices();

   // set group name
   geode->setName( fltSafeNodeName( node ) );

   // these are all global for the object
   sglVec3f * vertexList = new sglVec3f[ remappedIndices->size() ];
   sglVec2f * uvList = new sglVec2f[ remappedIndices->size() ];
   // these may not be used in all cases
   sglVec3f * normalList = new sglVec3f[ remappedIndices->size() ];
   sglVec4f * colorList  = new sglVec4f[ remappedIndices->size() ];

   sglVec3f Origin( 0.0f, 0.0f, 0.0f );

   if( offset )
      Origin += *offset;

#ifdef DEBUG_STRIPPER
   printf("Remapped indices for object: %s\n", fltSafeNodeName( node ) );
#endif

   // create mono indexed lists
   for( unsigned int i=0; i<remappedIndices->size(); i++ )
   {
      float r = 1.0f, g = 1.0f, b = 1.0f, a = 1.0f;
      FltVertex *v = flt->vertPalette->verts[ (*remappedIndices)[i] ];

      vertexList[i].set( v->x, v->y, v->z );
      vertexList[i] -= Origin;

      uvList[i].set( v->u, v->v );

      normalList[i].set( v->i, v->j, v->k );

      // MMM: need to check for packed color as well
      // if this fails, it is OK
      fltLookupColor( flt, v->colorIndex, &r, &g, &b, &a );

      colorList[i].set( r, g, b, a );

#ifdef DEBUG_STRIPPER
      printf("\t[%d]: %d (%f %f %f)\n", i, (*remappedIndices)[i],
                                          v->x, v->y, v->z );
#endif
   }

   // one global texenv for this geode
   sglTexEnv  *tev = new sglTexEnv();

   // set all to modulate for now
   tev->setMode(sglTexEnv::eMODULATE);

   // create geometry
   for( om = stateBins->begin(); om != stateBins->end(); ++om )
   {
      OOStateAttr * attr = om->first;
      int stateFlags = attr->getStateFlags();
      sglTriStripper<sglVertStructV> * ts = om->second;
      unsigned int numStrips = ts->getNumStrips();
      sglTexture2D * tex = 0;
      sglMaterial * material = 0;
      sglVec4f * color = 0;
      vector<sglStatelet *>state;

#ifdef DEBUG_STRIPPER
      printf("***** New Texture: %d ***** \n", textureIndex );
#endif

      sglMonoIndexedTriangleStripSet * strip = 
                                    new sglMonoIndexedTriangleStripSet();

      strip->setCoordList( vertexList );

      // turn on display lists
      strip->enableDisplayList( true );

      // 
      // primary color
      //
      if( stateFlags & OOStateAttr::HAS_PRIMARY_COLOR )
      {
         float r = 1.0f, g = 1.0f, b = 1.0f, a = 1.0f;
         color = new sglVec4f();

         // MMM: need to check for packed color as well
         // if this fails, it is OK
         fltLookupColor( flt, attr->getPrimaryColor(), &r, &g, &b, &a );

         color->set( r, g, b, a );
      }

      // 
      // Material
      // 
      if (stateFlags & OOStateAttr::HAS_MATERIAL)
      {
         FltMaterial * fmat;

         if( (fmat = fltLookupMaterial( flt, attr->getMaterial() )) )
         {
            material = new sglMaterial();

            // set material properties

            if( m_use_colormaterial )
            {
               material->setColorMaterial( sglMaterial::eFRONT,
                                           sglMaterial::eAMBIENT_AND_DIFFUSE );
            }
            else
            {
               material->setAmbient(   fmat->ambientRed,
                                       fmat->ambientGreen,
                                       fmat->ambientBlue );

               material->setDiffuse(   fmat->diffuseRed,
                                       fmat->diffuseGreen,
                                       fmat->diffuseBlue );
            }

            material->setSpecular(  fmat->specularRed,
                                    fmat->specularGreen,
                                    fmat->specularBlue );

            material->setEmission(  fmat->emissiveRed,
                                    fmat->emissiveGreen,
                                    fmat->emissiveBlue );

            material->setShininess( fmat->shininess );
            material->setAlpha( fmat->alpha );

            state.push_back( material );

            m_num_materials++;
         }
         else
         {
            sglPrintWarning() << "sglLoadFile_flt::buildGeometry error: "
                              << "unable to find material index "
                              << hex << attr->getMaterial() << dec
                              << " in flt file." << endl;
         }
      }

      if( stateFlags & OOStateAttr::HAS_TEXTURE )
         tex = lookupTexture( attr->getTexture(), 0 );

      // set state if texture found
      if( tex )
      {
         if ( textureContainsAlpha( tex ) )
         {
            if( m_use_alphafunc_transparency )
               state.push_back(new 
                  sglAlphaTest(sglAlphaTest::eGREATER, m_alpha_threshold ) );
            else if( m_use_blendfunc_transparency )
               state.push_back(new sglBlending(sglBlending::eSRC_ALPHA,
                                       sglBlending::eONE_MINUS_SRC_ALPHA,true));
            else
               state.push_back(new sglTransparency( true ) );
         }

         state.push_back(tex);
         state.push_back(tev);
      }

      // do strips first
      if( numStrips )
      {
         strip->setNumPrims( numStrips ); 
         unsigned int * primLengths = new unsigned int[ numStrips ];
         unsigned int * indexList = 0;
         unsigned int indexListSize = 0;
         unsigned int numIndices = 0, i;

         // first, figure out how many indices there are total
         for(i=0; i<numStrips; i++) 
         {
            vector<sglVertStructV> * v = ts->getStrip( i );

            primLengths[i] = v->size();
            numIndices += primLengths[i];
         }

         //
         // alloc indices
         //
         indexList = new unsigned int[ numIndices ];

         for(i=0; i<numStrips; i++)
         {
            vector<sglVertStructV> * v = ts->getStrip( i );

   #ifdef DEBUG_STRIPPER
         printf("\tNew Strip: %d\n", i);
   #endif

            for( unsigned int j=0; j<v->size(); j++ ) {
               unsigned int curIdx = (*v)[j].vi;

               ridxIter = reverseIdx->find( curIdx );

               // error check??
               indexList[indexListSize] = ridxIter->second;

   #ifdef DEBUG_STRIPPER
         printf("\t[%d]:[%d] %d -> %d\n", indexListSize, j, curIdx,
                                                         ridxIter->second );
   #endif
               indexListSize++;
            }

            // stats
            m_num_strips++;

            if( v->size() < (unsigned)m_min_strip_len )
               m_min_strip_len = v->size();
            if( v->size() > (unsigned)m_max_strip_len )
               m_max_strip_len = v->size();

            m_avg_strip_len += v->size();
         }

         strip->setPrimLengths( primLengths );
         strip->setIndexList( indexList );

         if( stateFlags & OOStateAttr::HAS_VCOLORS )
         {
            strip->setColorList(sglGeoSet::ePER_VERTEX, colorList);
            vcolorsUsed = true;
         }
         else if( color )
            strip->setColorList(sglGeoSet::eOVERALL, color);

         if( stateFlags & OOStateAttr::HAS_VNORMALS )
         {
            strip->setNormalList(sglGeoSet::ePER_VERTEX, normalList);
            vnormalsUsed = true;
         }

         // set state
         strip->setState(0, state, uvList);

         // add geometry
         geode->addGeometry( strip );
      }

      // now extract isolated triangles
      unsigned int numTris = ts->getNumTriangles();

      if( numTris )
      {
         sglMonoIndexedTriangleSet * tset = new sglMonoIndexedTriangleSet(); 
         unsigned int * ilist = new unsigned int[ numTris * 3 ];

         tset->setCoordList( vertexList );
         tset->setNumPrims( numTris ); 

         ts->copyTriangleIndexLists( ilist, 
                                     (unsigned int*)0, 
                                     (unsigned int*)0, 
                                     (unsigned int*)0 );

         // remap indices
         for( unsigned int j=0; j<numTris*3; j++ ) {

            // now remap the index
            ridxIter = reverseIdx->find( ilist[j] );

            // error check??
            ilist[j] = ridxIter->second;
         }

         tset->setIndexList( ilist );

         // turn on display lists
         tset->enableDisplayList( true );

         if( stateFlags & OOStateAttr::HAS_VCOLORS )
         {
            tset->setColorList(sglGeoSet::ePER_VERTEX, colorList);
            vcolorsUsed = true;
         }
         else if( color )
            tset->setColorList(sglGeoSet::eOVERALL, color);

         if( stateFlags & OOStateAttr::HAS_VNORMALS )
         {
            tset->setNormalList(sglGeoSet::ePER_VERTEX, normalList);
            vnormalsUsed = true;
         }

         // set state
         tset->setState(0, state, uvList);

         geode->addGeometry( tset );
      }
   }

   if( !vnormalsUsed )
      delete [] normalList;

   if( !vcolorsUsed )
      delete [] colorList;

   // dump the optimizer
   delete opt;

   // return our hard work
   return geode;
}

//----------------------------------------------------------------------------
void sglLoadFile_flt::buildScene(sglGroup * parent,
                                 FltNode * node, 
                                 FltFile * flt,
                                 sglVec3f * offset )
{
   uint32 i;
   sglFltUserData * fltUserData = 0;

   if (node == NULL)
   {
      sglPrintWarning() << "sglLoadFile_flt::buildScene error: "
                        << "NULL node in scene graph." << endl;
      return;
   }

   // create instance definitions
   if( node->type == FLTRECORD_INSTANCEDEFINITION )
   {
      buildInstanceDefinition( node, flt, offset );

      return;
   }

   // search for comments
   if (m_retain_comments)
   {
      FltNode * find_node;

      // search for a comment field
      find_node = node->attr;

      while (find_node)
      {
         if (find_node->type == FLTRECORD_COMMENT)
         {
            FltComment * comment = (FltComment *)find_node;
            fltUserData = new sglFltUserData();

            fltUserData->setComment( comment->text );

            if( m_verbose )
               sglPrintInfo() << "Adding Comment: \""
                              << comment->text
                              << "\"" << endl;

            break;
         }

         find_node = find_node->next;
      }
   }

   // check for known node types
   if (node->type == FLTRECORD_EXTERNALREFERENCE)
   {
      FltExternalReference * fext = (FltExternalReference *)node;
      sglTransformf *ext_group = new sglTransformf;
      char testPath[MAX_PATHLEN];
      char fname[MAX_PATHLEN];
      FltNode * find_node;

      // quick check for older software that did not append ".flt" to 
      // the DB name.
      testPath[0] = 0;
      strcat( testPath, fext->path );

      if(!strstr( fext->path, ".flt" )) 
         strcat( testPath, ".flt" );

      // try to find file
      if (fltFindFile(flt, testPath, fname))
      {
         if( m_verbose )
            sglPrintInfo() << "sglLoadFile_flt::buildScene "
                           << "loading external reference: '"
                           << fname << "'" << endl;

#if 0
         // NOTE: Could have used loadModelFLT(), but desire to
         // add caching later.
         FltFile * ext_ref = NULL;

         // open, parse, close file
         ext_ref = fltOpen(fname);
         if (!ext_ref)
         {
            sglPrintWarning() << "sglLoadFile_flt::buildScene error: "
                              << "Unable to open external reference: '"
                              << fname << "'" << endl;
            delete ext_group;
            return;
         }

         fltParse(ext_ref, 0);

         fltClose(ext_ref);

         buildScene(ext_group, (FltNode *)ext_ref->header, ext_ref);
#else
         // NOTE: Now required to use loadModelFLT(), because of texture
         //       index cache.

         string filepath(fname);
         sglNode * e;

         e = (sglNode *)loadModelFLT(filepath, 0 );
         if(!e)
         {
            sglPrintWarning() << "sglLoadFile_flt::buildScene error: "
                              << "Unable to open external reference: '"
                              << fname << "'" << endl;
            delete ext_group;
            return;
         }

         ext_group->addChild( e );
#endif

         // set name to be the external ref file name
         ext_group->setName(fext->path);

         if (fltUserData)
            ext_group->setUserData( fltUserData );

         // in case no transform found
         ext_group->loadIdentity();

         // get the transformation matrix, if it exists in attributes
         find_node = fext->node.attr;

         while (find_node)
         {
            if (find_node->type == FLTRECORD_MATRIX)
            {
               sglMat4f mat4;
               FltMatrix * fmat = (FltMatrix *)find_node;

               mat4.set(fmat->matrix[0], fmat->matrix[1],
                        fmat->matrix[2], fmat->matrix[3],
                        fmat->matrix[4], fmat->matrix[5],
                        fmat->matrix[6], fmat->matrix[7],
                        fmat->matrix[8], fmat->matrix[9],
                        fmat->matrix[10], fmat->matrix[11],
                        fmat->matrix[12], fmat->matrix[13],
                        fmat->matrix[14], fmat->matrix[15]);

               ext_group->setMatrix(mat4);
               break;
            }

            find_node = find_node->next;
         }

         if (parent)
            parent->addChild(ext_group);
      }
      else
      {
         delete ext_group;
         sglPrintWarning() << "External reference: '" << fext->path
                           << "' not found" << endl;
      }

      return;
   }

  // check for rogue transform mats
  if (node->attr)
  {
     // get the transformation matrix, if it exists in attributes
     //
     sglTransformf* transform = NULL;
     sglTranslatef* translate = NULL;
     FltNode* find_node = node->attr;
                      
     while (find_node)
     {
         if (find_node->type == FLTRECORD_MATRIX)
         {
            sglMat4f mat4;
            FltMatrix * fmat = (FltMatrix *)find_node;
            bool isValid = true;

mat4.set(fmat->matrix[0], fmat->matrix[1], fmat->matrix[2], fmat->matrix[3],
      fmat->matrix[4], fmat->matrix[5], fmat->matrix[6], fmat->matrix[7],
      fmat->matrix[8], fmat->matrix[9], fmat->matrix[10], fmat->matrix[11],
      fmat->matrix[12], fmat->matrix[13], fmat->matrix[14], fmat->matrix[15]);

            // 
            // naive check for translate only or scale
            //
            if( fmat->matrix[0] == fmat->matrix[5] &&
                  fmat->matrix[5] == fmat->matrix[10] )
            {
               if( fmat->matrix[0] == 1.0f )
               {
                  // translate only
                  // create the translate
                  translate = new sglTranslatef;
                  translate->setTranslation( fmat->matrix[12],
                                             fmat->matrix[13],
                                             fmat->matrix[14] );

#if 0
                  sglPrintInfo() << "Adding Translate: " << 
                                       fmat->matrix[12] << " " <<
                                       fmat->matrix[13] << " " <<
                                       fmat->matrix[14] << endl;
#endif

               }
               else
                  // naive check to ignore uniform scale
                  isValid = false;
            }
            else
            {
               // create the transform
               transform = new sglTransformf;
               transform->setMatrix(mat4);

#if 0
               sglPrintInfo() << "Adding Matrix: " << mat4 << endl;
#endif

               break;
            }

            if( !isValid )
               break;

         }

         find_node = find_node->next;
      }

      if( parent )
      {
         if( transform )
         {
            parent->addChild(transform);
            parent = transform;
         }
         else if( translate )
         {
            parent->addChild(translate);
            parent = translate;
         }
      }
      else
      {
         if( transform )
            delete transform;
         if( translate )
            delete translate;
      }

   }

   if (node->type == FLTRECORD_INSTANCEREFERENCE)
   {
      FltInstanceReference * iref = (FltInstanceReference *)node;
      sglNode * instNode = 0;

      instNode = createInstance( iref->instance );

      if( parent && instNode )
         parent->addChild( instNode );

      return;
   }

   if (node->type == FLTRECORD_FACE)
   {
      sglGeode * geode = new sglGeode;
      sglVec4f * color = 0;
      FltFace * face = (FltFace *)node;
      sglTexture2D * tex = 0;
      sglMaterial * material = 0;
      bool isBillboard = false;
      CombinerEntry * ce = 0;

      // skip hidden faces
      if( face->miscFlags & FLTFACEMF_HIDDEN )
      {
         delete geode;
         return;
      }

      geode->setName( fltSafeNodeName( node ) );
      if (fltUserData)
         geode->setUserData(fltUserData);

      // check for textures
      if (face->texturePatternIndex != -1)
      {
         // may fail, which is fine.
         tex = lookupTexture( face->texturePatternIndex, &ce );
      }

      // 
      // Materials
      // 
      if( (face->materialIndex != -1) && m_enable_lighting )
      {
         FltMaterial * fmat;

         if( (fmat = fltLookupMaterial( flt, face->materialIndex )) )
         {
            material = new sglMaterial;

            // set material properties

            if( m_use_colormaterial )
            {
               material->setColorMaterial( sglMaterial::eFRONT,
                                           sglMaterial::eAMBIENT_AND_DIFFUSE );
            }
            else
            {
               material->setAmbient(   fmat->ambientRed,
                                       fmat->ambientGreen,
                                       fmat->ambientBlue );

               material->setDiffuse(   fmat->diffuseRed,
                                       fmat->diffuseGreen,
                                       fmat->diffuseBlue );
            }

            material->setSpecular(  fmat->specularRed,
                                    fmat->specularGreen,
                                    fmat->specularBlue );

            material->setEmission(  fmat->emissiveRed,
                                    fmat->emissiveGreen,
                                    fmat->emissiveBlue );

            material->setShininess( fmat->shininess );
            material->setAlpha( fmat->alpha );

            m_num_materials++;
         }
         else
         {
            sglPrintWarning() << "sglLoadFile_flt::buildScene error: "
                              << "unable to find material index "
                              << hex << face->materialIndex << dec
                              << " in hierarchy." << endl;
         }
      }

      // 
      // Primary Face Color
      //
      if (face->primaryColorIndex != 0xffffffff || face->textureWhite)
      {
         real32 r, g, b, a;
         uint32 idx;

         // MMM: need to check for packed color as well
         // primarily in 14.x formats
         if (face->primaryColorIndex == 0)
            idx = face->colorNameIndex;
         else
            idx = face->primaryColorIndex;

         // reset if face is supposed to be white
         if( face->textureWhite )
            idx = 0xfffffffe;

         if (fltLookupColor(flt, idx, &r, &g, &b, &a))
         {
            color = new sglVec4f;

            color->set(r, g, b, a);
         }
      }

      // 
      // Billboard Flags
      //
      if( face->billboardFlags == FLTFACEBB_AXIALROTATE ||
                        face->billboardFlags == FLTFACEBB_POINTROTATE ) {
         sglBillboard * bb = new sglBillboard;
         sglVec3d pos( 0.0, 0.0, 0.0 );

         isBillboard = true;

         if( face->billboardFlags == FLTFACEBB_AXIALROTATE )
            bb->setMode( sglBillboard::eAXIAL_ROT );
         else 
         {
            // POINT ROTATE
            bb->setMode( sglBillboard::ePOINT_ROT_EYE );
            bb->setPos( pos );
         }

         if( parent )
            parent->addChild( bb );

         // reset parent to be the billboard for the geometry
         parent = bb;
      }

      for (uint32 j=0;j<node->numChildren;j++)
      {
         FltVertexList * vlist = (FltVertexList *)node->child[j];
         sglGeoSet * geom = 0;
         sglGeoSet * geomBack = 0;
         bool twoSided = false;
         vector<sglStatelet*> state;

         #define HAS_VNORMALS(x) ((x)->lightMode == FLTFACELM_FCVN || \
                            (x)->lightMode == FLTFACELM_VCVN )
         #define HAS_VCOLORS(x) ((x)->lightMode == FLTFACELM_VCNOTILLUM || \
                            (x)->lightMode == FLTFACELM_VCVN )

         // check for two-sided prims
         if (face->drawType == FLTFACEDT_DRAWSOLIDNOBACKFACE)
            twoSided = true;

         // check for lines
         if( face->drawType == FLTFACEDT_DRAWWIREFRAME || 
             face->drawType == FLTFACEDT_DRAWWIREFRAMECLOSE || 
             vlist->numVerts == 2 )
         {
            geom = new sglLineSet;
            m_num_lines++;

            if( m_enable_line_aa )
               state.push_back( new sglAntiAliasLine( true ) );
         }
         else if (vlist->numVerts == 4)
         {
            geom = new sglQuadSet;
            m_num_quads++;
            if( twoSided ) {
               geomBack = new sglQuadSet;
               m_num_quads++;
            }
         }
         else if( vlist->numVerts == 3 )
         {
            geom = new sglTriangleSet;
            m_num_tris++;
            if( twoSided ) {
               geomBack = new sglTriangleSet;
               m_num_tris++;
            }
         }
         else if( vlist->numVerts == 1 )
         {
            geom = new sglPointSet;
            m_num_points++;
         }
         else
         {
            geom = new sglPolygonSet;
            unsigned int *primLength = new unsigned int;

            *primLength = vlist->numVerts;
            ((sglPolygonSet *)geom)->setPrimLengths(primLength);

            m_num_polys++;

            if( twoSided )
            {
               geomBack = new sglPolygonSet;
               unsigned int *bprimLength = new unsigned int;

               *bprimLength = vlist->numVerts;
               ((sglPolygonSet *)geomBack)->setPrimLengths(bprimLength);

               m_num_polys++;
            }
         }

         m_num_verts += vlist->numVerts;

         geom->setNumPrims(1);

         if( twoSided )
         {
            m_num_verts += vlist->numVerts;
            geomBack->setNumPrims(1);
         }

         //
         // Coords
         //
         sglVec3f *vertex = new sglVec3f[ vlist->numVerts ];
         sglVec3f *vertexBack = 0;

         // configure origin offset
         sglVec3f Origin(0.0f, 0.0f, 0.0f);

         if( offset )
            Origin += *offset;

         if( twoSided )
            vertexBack = new sglVec3f[ vlist->numVerts ];

         if (isBillboard)
         {
            // HACK to swap Y,Z to conform to SGL's billboards
            for(i=0;i<vlist->numVerts;i++)
            {
               vertex[i].set((float) vlist->list[i]->x,
                             (float) vlist->list[i]->z,
                             (float) vlist->list[i]->y);

               // may be wrong
               vertex[i] -= Origin;
            }
         }
         else
         {
            for(i=0;i<vlist->numVerts;i++)
            {
               vertex[i].set((float) vlist->list[i]->x,
                             (float) vlist->list[i]->y,
                             (float) vlist->list[i]->z);
               vertex[i] -= Origin;

               if( twoSided )
               {
                  // reverse the winding
                  vertexBack[i].set(
                                 (float) vlist->list[vlist->numVerts-1-i]->x,
                                 (float) vlist->list[vlist->numVerts-1-i]->y,
                                 (float) vlist->list[vlist->numVerts-1-i]->z );
                  vertexBack[i] -= Origin;
               }
            }
         }

         geom->setCoordList(vertex);

         if( twoSided )
            geomBack->setCoordList( vertexBack );

         //
         // Normals
         //
         if( m_enable_lighting )
         {
            if( m_compute_all_normals ||
                (m_compute_normals_if_absent && !HAS_VNORMALS(face)) )
            {
               sglVec3f *normal = new sglVec3f;
               sglVec3f n;

               computeFaceNormal( vertex[0], vertex[1], vertex[2], n );

               *normal = n;
               geom->setNormalList( sglGeoSet::eOVERALL, normal );

               if( twoSided )
               {
                  normal = new sglVec3f;

                  computeFaceNormal( vertex[2], vertex[1], vertex[0], n );
                  *normal = n;

                  geomBack->setNormalList( sglGeoSet::eOVERALL, normal );
               }
            }
            else if( HAS_VNORMALS( face ) )
            {
               sglVec3f *normal = new sglVec3f[ vlist->numVerts ];

               for(i=0;i<vlist->numVerts;i++)
                  normal[i].set( (float) vlist->list[i]->i,
                                 (float) vlist->list[i]->j,
                                 (float) vlist->list[i]->k );

               geom->setNormalList( sglGeoSet::ePER_VERTEX, normal );

               if( twoSided )
               {
                  normal = new sglVec3f[ vlist->numVerts ];

                  for(i=0;i<vlist->numVerts;i++)
                     normal[i].set( 
                              (float) -vlist->list[vlist->numVerts-1-i]->i,
                              (float) -vlist->list[vlist->numVerts-1-i]->j,
                              (float) -vlist->list[vlist->numVerts-1-i]->k );

                  geomBack->setNormalList( sglGeoSet::ePER_VERTEX, normal );
               }
            }
         }

         //
         // Colors
         //
         if( HAS_VCOLORS( face ) && m_enable_shading )
         {
            real32 r, g, b, a;
            uint32 idx;

            if( color )
               delete color;

            sglVec4f *colors = new sglVec4f[ vlist->numVerts ];

            for(i=0;i<vlist->numVerts;i++)
            {
               // MMM: need to check for packed color as well
               idx = vlist->list[i]->colorIndex;

               if (fltLookupColor(flt, idx, &r, &g, &b, &a))
                  colors[i].set( (float) r, (float) g, (float) b, (float) a );
               else
                  colors[i].set( 1.0, 1.0, 1.0, 1.0 );
            }

            geom->setColorList( sglGeoSet::ePER_VERTEX, colors );

            if( twoSided )
            {
               colors = new sglVec4f[ vlist->numVerts ];

               for(i=0;i<vlist->numVerts;i++)
               {
                  // MMM: need to check for packed color as well
                  idx = vlist->list[vlist->numVerts-1-i]->colorIndex;

                  if (fltLookupColor(flt, idx, &r, &g, &b, &a))
                     colors[i].set( (float) r, (float) g, (float) b, (float) a);
                  else
                     colors[i].set( 1.0, 1.0, 1.0, 1.0 );
               }

               geomBack->setColorList( sglGeoSet::ePER_VERTEX, colors );
            }
         }
         else
         {
            if (color)
            {
               geom->setColorList(sglGeoSet::eOVERALL, color);

               if( twoSided )
                  geomBack->setColorList(sglGeoSet::eOVERALL, color);
            }
         }

         sglVec2f * uv = 0;
         sglVec2f * uvBack = 0;

         //
         // Texture
         //
         if (tex)
         {
            uv = new sglVec2f[ vlist->numVerts ];

            if( twoSided )
               uvBack = new sglVec2f[ vlist->numVerts ];

            // check if it was a combiner entry
            if( ce )
            {
               if( ce->rotated )
               {
                  // assume rotated CCW for now
                  for(i=0;i<vlist->numVerts;i++)
                     uv[i].set( (float) (( (-(vlist->list[i]->v-0.5)+0.5) * 
                                                               ce->w) + ce->u),
                                (float) (( ((vlist->list[i]->u-0.5)+0.5) * 
                                                               ce->h) + ce->u));
               }
               else
               {
                  for(i=0;i<vlist->numVerts;i++)
                     uv[i].set( (float) (vlist->list[i]->u * ce->w) + ce->u,
                                (float) (vlist->list[i]->v * ce->h) + ce->v );
               }
            }
            else
            {
               for(i=0;i<vlist->numVerts;i++)
                  uv[i].set( (float) vlist->list[i]->u,
                              (float) vlist->list[i]->v );

               if( twoSided )
               {
                  for(i=0;i<vlist->numVerts;i++)
                     uvBack[i].set( (float) vlist->list[vlist->numVerts-1-i]->u,
                                 (float) vlist->list[vlist->numVerts-1-i]->v );
               }
            }

            sglTexEnv  *tev = new sglTexEnv();

            state.push_back(tex);
            state.push_back(tev);

            if ( textureContainsAlpha( tex ) )
            {
               if( m_use_alphafunc_transparency )
                  state.push_back(new 
                     sglAlphaTest(sglAlphaTest::eGREATER, m_alpha_threshold ) );
               else if( m_use_blendfunc_transparency )
                  state.push_back(new sglBlending(sglBlending::eSRC_ALPHA,
                                       sglBlending::eONE_MINUS_SRC_ALPHA,true));
               else
                  state.push_back(new sglTransparency( true ) );
            }
            else
               // possible optimization here?
               tev->setMode(sglTexEnv::eMODULATE);
         }

         if( material )
            state.push_back( material );

         // this didn't work on the models I tested with
         // check for "closed" line sets
         //if( face->drawType == FLTFACEDT_DRAWWIREFRAMECLOSE )
         //  state.push_back( new sglPolyFrontStyle( sglPolyStyle::eLINE ) );

         geom->setState(0, state, uv);

         if( twoSided )
            geomBack->setState( 0, state, uvBack );

         geode->addGeometry(geom);

         if( twoSided )
            geode->addGeometry( geomBack );
      }

      if (parent)
         parent->addChild(geode);

      return;

   }
   else if (node->type == FLTRECORD_LOD)
   {
      FltLOD * fltLOD = (FltLOD *)node;
      sglLOD * lod = new sglLOD;
      sglVec3f v3f;
      uint32 i;

      lod->setName( fltSafeNodeName( node ) );
      if (fltUserData)
         lod->setUserData( fltUserData );

      //
      // insert a "dummy" group because of differences in FLT LODs vs SGL LODs
      // eventually, need to combine like ranges, but do this for now.
      //
      sglGroup * dummyGroup = new sglGroup;
      dummyGroup->setName("DUMMYLODGROUP");
      lod->addChild(dummyGroup);

      // traverse children
      for (i=0;i<node->numChildren;i++)
         buildScene(dummyGroup, node->child[i], flt, offset);

      if( fltFinite64( fltLOD->centerX ) &&
            fltFinite64( fltLOD->centerY ) &&
            fltFinite64( fltLOD->centerZ ) )
      {
         v3f.set(fltLOD->centerX, fltLOD->centerY, fltLOD->centerZ);
      }
      else 
      {
         sglPrintWarning() << "sglLoadFile_flt::LOD error: "
                           << "Infinite Center Value in : " 
                           << lod->getName() << endl;

         sglSphereBoundf bsph = dummyGroup->getBound();

         v3f.set( bsph.getCenter()[0], 
                  bsph.getCenter()[1],
                  bsph.getCenter()[2] );

#if 0
         sglPrintWarning() << "New Center: " << v3f << endl;
#endif
      }

      lod->setCenter(v3f);
      lod->setRange(0, fltLOD->switchOutDistance);
      lod->setRange(1, fltLOD->switchInDistance);

      if (parent)
         parent->addChild(lod);

      return;
   }
   else if (node->type == FLTRECORD_SWITCH)
   {
      FltSwitch * fltSwitch = (FltSwitch *)node;
      sglIndexedBitMaskSwitch * sw = new sglIndexedBitMaskSwitch;
      uint32 i;

      // set switch info
      sw->setName( fltSafeNodeName( node ) );
      if (fltUserData)
         sw->setUserData( fltUserData );

      // does not account for bitmasks > than 32 bits.
      for(i=0;i<fltSwitch->numMasks;i++)
         sw->addMask( fltSwitch->masks[i] );

      sw->setIndex( fltSwitch->currentMask );

      for (i=0;i<node->numChildren;i++)
      {
         //
         // Create a temporary dummy group to maintain switch
         // structure.
         //
         sglGroup * tgroup = new sglGroup;
         tgroup->setName( "_SG_" );

         buildScene(tgroup, node->child[i], flt, offset);

         // maintain group if 0 (jic) or > 1 children
         if( tgroup->getNumChildren() == 0 || 
            tgroup->getNumChildren() > 1 )
         {
            sw->addChild( tgroup );
         }
         else
         {
            // add the single child, and delete the temporary group
            sw->addChild( tgroup->getChild(0) );
            delete tgroup;
         }
      }

      if (parent)
         parent->addChild(sw);

      return;

   }
   else if (node->type == FLTRECORD_DOF)
   {
      FltDOF * fltDOF = (FltDOF *)node;
      sglTransformf * dof = new sglTransformf;
      sglTranslatef * dofP = new sglTranslatef;
      sglVec3f * doffset = new sglVec3f;

      dof->setName( fltSafeNodeName( node ) );

      if (fltUserData)
         dof->setUserData( fltUserData );

      dof->setRotation(fltDOF->localCurYaw * M_PI/180,
                       fltDOF->localCurPitch * M_PI/180,
                       fltDOF->localCurRoll * M_PI/180);

      dof->setTranslation(fltDOF->localCurX,
                          fltDOF->localCurY,
                          fltDOF->localCurZ);

      // The Local Coordinate Frame Origin, X Axis Pt,
      // and XY Axis Pt define the local coordinate
      // frame.  Lets check to see if this coordinate
      // frame is well defined.
      
      sglVec3f origin(fltDOF->localOriginX,
                           fltDOF->localOriginY,
                           fltDOF->localOriginZ);

      sglVec3f xAxisPt(fltDOF->localPointX,
                           fltDOF->localPointY,
                           fltDOF->localPointZ);			   

      sglVec3f xyPlanePt(fltDOF->localPlanePointX,
                           fltDOF->localPlanePointY,
                           fltDOF->localPlanePointZ);			   

      sglVec3f xAxis = xAxisPt - origin;
      xAxis.normalize();
      
      sglVec3f xyPlane = xyPlanePt - origin;
      xyPlane.normalize();
      
      sglVec3f zAxis;
      zAxis.cross(xAxis,xyPlane);
      zAxis.normalize();
      
      if (sglAlmostEqual(xAxis.length(), 0.0f) &&
          sglAlmostEqual(zAxis.length(), 0.0f)) 
      {
          // Undefined Coordinate System
	  // Set some reasonable values...
          xAxis[0] = 1.0f;
	  zAxis[2] = 1.0f;
          origin.set(0.0,0.0,0.0);	  
      }	  

      dofP->setTranslation(origin[0],origin[1],origin[2]);

#if 0
      cout << "DOF Node: " << fltSafeNodeName(node) << " " <<
            fltDOF->localOriginX << " " <<
            fltDOF->localOriginY << " " <<
            fltDOF->localOriginZ << endl;
#endif

      doffset->set(origin[0],origin[1],origin[2]);

      // take into account any previous offset
      if( offset )
         doffset->add( *doffset, *offset );

      for (i=0;i<node->numChildren;i++)
         buildScene(dof, node->child[i], flt, doffset);

      delete doffset;

      dofP->addChild( dof );

      if (parent)
         parent->addChild( dofP );

      return;
   }
   else if (node->type == FLTRECORD_OBJECT)
   {
      if( m_enable_stripper == false )
      {
         // do more processing here later
         sglGroup * group = new sglGroup;

         group->setName( fltSafeNodeName( node ) );
         if (fltUserData)
            group->setUserData( fltUserData );

         for (i=0;i<node->numChildren;i++)
            buildScene(group, node->child[i], flt, offset);

         if (parent)
            parent->addChild(group);
      }
      else 
      {
         sglGeode * geode;

         geode = buildGeometry( parent, node, flt, offset );

         if( parent && geode )
            parent->addChild( geode );
      }

      return;
   }
   else if (node->type == FLTRECORD_GROUP)
   {
      FltGroup * fgroup = (FltGroup *)node;

      //
      // Check for animation types
      //

      // swing & forward are both set when it is a swing
      if (fgroup->flags & FLTGROUP_SWING_ANIM)
      {
         sglKeyFrameAnimation * key = new sglKeyFrameAnimation;

         key->setMode(sglAnimation::eSWING);
         key->setName( fltSafeNodeName( node ) );
         if (fltUserData)
            key->setUserData( fltUserData );

         for (i=0;i<node->numChildren;i++)
         {
            //
            // Create a temporary dummy group to maintain animation
            // structure.
            //
            sglGroup * tgroup = new sglGroup;
            tgroup->setName( "_AG_" );

            buildScene(tgroup, node->child[i], flt, offset);

            // maintain group if 0 (jic) or > 1 children
            if( tgroup->getNumChildren() == 0 || 
               tgroup->getNumChildren() > 1 )
            {
               key->addChild( tgroup );
            }
            else
            {
               // add the single child, and delete the temporary group
               key->addChild( tgroup->getChild(0) );
               delete tgroup;
            }

            key->setDuration(i, 1);
         }

         key->setRange(0, node->numChildren-1);
         key->setSpeed(1.0f);
         key->setNumCycles(0);
         key->setState(sglAnimation::eSTART);

         if (parent)
            parent->addChild(key);

      }
      else if (fgroup->flags & FLTGROUP_FORWARD_ANIM)
      {
         sglKeyFrameAnimation * key = new sglKeyFrameAnimation;

         key->setMode(sglAnimation::eCYCLE);
         key->setName( fltSafeNodeName( node ) );
         if (fltUserData)
            key->setUserData( fltUserData );

         for (i=0;i<node->numChildren;i++)
         {
            //
            // Create a temporary dummy group to maintain animation
            // structure.
            //
            sglGroup * tgroup = new sglGroup;
            tgroup->setName( "_AG_" );

            buildScene(tgroup, node->child[i], flt, offset);

            // maintain group if 0 (jic) or > 1 children
            if( tgroup->getNumChildren() == 0 || 
               tgroup->getNumChildren() > 1 )
            {
               key->addChild( tgroup );
            }
            else
            {
               // add the single child, and delete the temporary group
               key->addChild( tgroup->getChild(0) );
               delete tgroup;
            }

            key->setDuration(i, 1);
         }

         key->setRange(0, node->numChildren-1);
         key->setSpeed(1.0f);
         key->setNumCycles(0);
         key->setState(sglAnimation::eSTART);

         if (parent)
            parent->addChild(key);

      }
      else
      {
         //
         // regular group
         //
         sglGroup * group = new sglGroup;

         group->setName( fltSafeNodeName( node ) );
         if (fltUserData)
            group->setUserData( fltUserData );

         for (i=0;i<node->numChildren;i++)
            buildScene(group, node->child[i], flt, offset);

         if (parent)
            parent->addChild(group);
      }

      return;
   }
#ifdef USE_BSP
   else if ( node->type == FLTRECORD_BSP && node->numChildren == 3 &&
            m_use_bsp > 0 )
   {
      FltBSP * fbsp = (FltBSP *)node;

      sglBSP * bsp = new sglBSP();
      sglGroup * leftNode = new sglGroup(), * rightNode = new sglGroup();

      bsp->setName( fltSafeNodeName( node ) );
      if (fltUserData)
         bsp->setUserData( fltUserData );

      // convert to OpenGL coord system for later processing
      if( 1 )
         bsp->setPlaneEq( fbsp->coefA, -fbsp->coefC, fbsp->coefB, fbsp->coefD );
      else
         bsp->setPlaneEq( fbsp->coefA, fbsp->coefB, fbsp->coefC, fbsp->coefD );

      cout << "PlaneEQ: " <<
                        fbsp->coefA << " " <<
                        fbsp->coefB << " " <<
                        fbsp->coefC << " " <<
                        fbsp->coefD << " " <<
                        node->numChildren << " children" << endl;

      // should be 3 children ( left, poly, right )
      // do more error checking here!
      buildScene(leftNode, node->child[0], flt, offset);
      buildScene(rightNode, node->child[2], flt, offset);

      bsp->enable( true );

      if (parent)
         parent->addChild(bsp);

      return;
   }
#endif

   // dump it if we got here.
   if (fltUserData)
      delete fltUserData;

   // descend hier for everything else (flt_node not needed)
   for (i=0;i<node->numChildren;i++)
      buildScene(parent, node->child[i], flt, offset);

   return;
}

//----------------------------------------------------------------------------
sglNode *sglLoadFile_flt::loadFile(const string &filename,
                                   unsigned int /* mode */)
{
   sglGroup *group = NULL;

   // store the path for possible future use
   m_alternate_path = sglDataLoader::extractDirectoryFromPath(filename);

   if( s_combiner_map.empty() )
   {
      FILE * f;

      //
      // Combiner file format:
      //
      // texFileName combinedTexFileName newU newV newW newH rotated[y|n]
      //
      //  where newU newV newW newH are normalized in the combined texmap
      //  obviously, "wrap" cannot be used if textures are combined.
      //

      if(( f = fopen("combinedTex.map", "r") ))
      {
         char texName[256], cTexName[256], rot[10];
         float u, v, w, h;
         CombinerEntry * entry = 0;

         while( 1 ) 
         {
            if( fscanf( f, "%s %s %f %f %f %f %s\n", texName, cTexName, 
                                       &u, &v, &w, &h, rot ) != 5 )
               break;

            entry = (CombinerEntry *) calloc( 1, sizeof( CombinerEntry ) );

            if( entry )
            {
               entry->u = u;
               entry->v = v;
               entry->w = w;
               entry->h = h;
               entry->rotated = (tolower(*rot)=='y')?true:false;
               entry->combinedTexFile = new string( cTexName );

               s_combiner_map[texName] = entry;
               m_use_combiner_file = true;
            }
         }

         fclose( f );
      }
   }
   else
      m_use_combiner_file = true;

   // open, parse, close file
   m_flt = fltOpen(filename.c_str());

   // try using sglDataLoader's search capabilities
   if (!m_flt)
   {
      string foundname;
      if (sglDataLoader::findFile(filename, NULL, foundname))
      {
         m_flt = fltOpen(foundname.c_str());
         if (m_flt)
         {
            m_alternate_path =
               sglDataLoader::extractDirectoryFromPath(foundname);
         }
      }
   }

   if (m_flt)
   {
      fltParse(m_flt, 0);
      fltClose(m_flt);

      group = new sglGroup;
      buildScene(group, (FltNode *)m_flt->header, m_flt, 0);
      if (group->getNumChildren() == 0)
      {
         delete group;
         group = NULL;
      }

      if( m_display_stats )
      {
         float tex_mb = m_trans_texels * 4 + m_opaque_texels * 3;

         tex_mb /= (1024.0f*1024.0f);

         sglPrintInfo() << "-- Statistics for: " << filename << " ---" << endl;
         sglPrintInfo() << "   Num Points     : " << m_num_points << endl;
         sglPrintInfo() << "   Num Lines      : " << m_num_lines << endl;
         sglPrintInfo() << "   Num Tris       : " << m_num_tris << endl;
         sglPrintInfo() << "   Num Quads      : " << m_num_quads << endl;
         sglPrintInfo() << "   Num Polygons   : " << m_num_polys << endl;
         sglPrintInfo() << "   Total Prims    : " << 
                  m_num_points+m_num_lines+m_num_polys+m_num_tris+m_num_quads 
                  << endl;
         sglPrintInfo() << "   Num Verts      : " << m_num_verts << endl;

         if( m_enable_stripper )
         {
         sglPrintInfo() << "   Num S.Tris     : " << m_num_strip_tris << endl;
         sglPrintInfo() << "   Num S.Quads    : " << m_num_strip_quads << endl;
         sglPrintInfo() << "   Num S.Polygons : " << m_num_strip_polys << endl;
         sglPrintInfo() << "   Num Strips     : " << m_num_strips << endl;
         if( m_num_strips )
            sglPrintInfo() << "    Length Stats  : " 
                                                << "Min: " << m_min_strip_len 
                                                << " Max: " << m_max_strip_len
                              << " Avg: " << (m_avg_strip_len/m_num_strips)
                                                << endl;
         }

         sglPrintInfo() << "   Num Materials  : " << m_num_materials << endl;
         sglPrintInfo() << "   Num Textures   : " << m_num_textures << endl;
         sglPrintInfo() << "   Texture Size MB: " 
                        << tex_mb << " (32bpp) "
                        << tex_mb/2.0f << " (16bpp) " << endl;
         sglPrintInfo() << "    Mipped        : " 
                        << tex_mb * 1.333f << " (32bpp) "
                        << (tex_mb/2.0f) * 1.333f << " (16bpp) " << endl;

         tex_mb = (m_opaque_texels / 16) * 8;
         tex_mb += ( (m_trans_texels / 16) * 16 );
         tex_mb *= 1.3333f; // mips

         tex_mb /= (1024.0f*1024.0f);
         sglPrintInfo() << "   S3TC Cmp+Mipped: " << tex_mb << endl; 
      }
   }

   return group;
}
