/* vi:set ts=3 sw=3 expandtab: */
/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 * SGL's FLT loader Copyright (C) 2001  Michael M. Morrison  All Rights
 *   Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglLoadFile_flt.hpp
 *  Created: June 2001
 *  Author : Michael M. Morrison
 *  Summary: Wrapper class for the .flt loader library.
 *           Based on sglLoadFile_obj.
 *****************************************************************************/

#ifndef __SGL_LOAD_FILE_FLT_HPP
#define __SGL_LOAD_FILE_FLT_HPP

#include <sglflt.h>
#include <sglModelLoader.hpp>
#include <sglUserData.hpp>
#include <sglTexture.hpp>

#include <map>

#include "ObjOptimizer.hpp"

class sglTexture2D;
class sglNode;
class sglGroup;
class sglGeode;
typedef struct _FltNode FltNode;
typedef struct _FltFile FltFile;

typedef struct _CombinerEntry {
   float u, v;                  // remapped U,V coords
   float w, h;                  // remapped width, height
   bool rotated;
   string * combinedTexFile;    // combined tex file
} CombinerEntry;

//============================================================================

class SGLFLT_DLL_API sglLoadFile_flt
{
public:
   /// default constructor
   sglLoadFile_flt();
   /// virtual destructor
   virtual ~sglLoadFile_flt();

   /// load a file, build a scene graph, and get the sglGeode* pointing to it
   sglNode *loadFile(const string &filename, unsigned int mode);

private:
   /// Copy constructor forbidden
   sglLoadFile_flt(const sglLoadFile_flt &);
   /// Assignment operator forbidden
   sglLoadFile_flt &operator=(const sglLoadFile_flt &);

   sglTexture2D *lookupTexture( int indx, CombinerEntry ** e );
   sglTexture2D *createTexture(const string &filename);
   void buildScene(sglGroup *parent, FltNode *node, FltFile *flt, 
                                                      sglVec3f *o );
   sglGeode * buildGeometry(sglGroup * g, FltNode *node, FltFile *flt, 
                                                            sglVec3f * v );
   void parseGeometry(sglGroup *, ObjectOptimizer *o, FltNode *n, FltFile *f,
                                                            sglVec3f * v );

   void buildInstanceDefinition( FltNode * node, FltFile * flt,
                                 sglVec3f * offset );

   sglNode * createInstance( int indx );

   bool textureContainsAlpha( sglTexture * tex )
   {
      // check for alpha
      if( tex->getExternalFormat() == sglTexture::eRGBA ||
          tex->getExternalFormat() == sglTexture::eALPHA ||
          tex->getExternalFormat() == sglTexture::eLUMINANCE_ALPHA )
         return true;
      else
         return false;
   }

private:
   static sglLock s_lock;
   static map<string, sglTexture2D*> s_tex_map;

   bool   m_use_combiner_file;   // whether to use externally combined file
	 static map<string, CombinerEntry *> s_combiner_map; 

   map<int, sglTexture2D*> m_textures_found_map;
   map<int, sglTexture2D*> m_textures_not_found_map;

   map<int, sglNode*> m_instance;

   FltFile * m_flt;

   string m_alternate_path;

   bool   m_verbose;             // supply verbose output
   bool   m_display_stats;       // display statistics output
   bool   m_compute_all_normals; // compute all new normals from face
   bool   m_compute_normals_if_absent; // compute normals if they don't exist
   bool   m_use_colormaterial;   // use colormaterial for amb and diff.
   bool   m_enable_lighting;     // read and store materials and normals
   bool   m_enable_shading;      // read and store per-vertex colors
   bool   m_enable_line_aa;      // enable line anti-alias

   bool   m_enable_stripper;     // enable triangle stripper

   bool   m_use_attr_files;      // enable use of attribute files

   int    m_use_bsp;             // enabling the use of bsp structures

   // default minification and magnification filters
   sglTexture::FilterEnum m_default_min_filter;
   sglTexture::FilterEnum m_default_mag_filter;


   bool   m_compress_textures;   // compress textures during load
   sglTexture::CompressionLevelEnum m_default_compression_level;

   int    m_min_texture_dimension; // textures with a dimension smaller:ignored

   bool   m_retain_comments;     // whether to retain comments in userdata

   bool   m_free_texture_images; // free texture images after bind

   bool   m_use_blendfunc_transparency; // use glBlendFunc vs glAlphaFunc
   bool   m_use_alphafunc_transparency; // use glAlphaFunc vs glBlendFunc
   float  m_alpha_threshold;     // alpha transparency threshold for a-test 

   // statistics
   int    m_num_points;          // number of points
   int    m_num_lines;           // number of lines
   int    m_num_tris;            // number of triangles in db
   int    m_num_quads;           // number of quads in db
   int    m_num_polys;           // number of misc polygons in db
   int    m_num_verts;           // total verts in db
   int    m_num_strip_quads;     // number of stripped quads in db
   int    m_num_strip_polys;     // number of misc stripped polygons in db
   int    m_num_strip_tris;      // total tripped tris 
   int    m_num_strips;          // total number of strips created
   int    m_num_materials;       // total materials
   int    m_num_textures;        // total number of textures
   int    m_opaque_texels;       // # opaque texels
   int    m_trans_texels;        // # translucent texels (+alpha)

   int    m_min_strip_len;       // strip stats
   int    m_max_strip_len;
   int    m_avg_strip_len;
};

//----------------------------------------------------------------------------
// unfortunate connector with C library
class SGLFLT_DLL_API sglFltUserData: public sglUserData
{
public:
   sglFltUserData() {}
   ~sglFltUserData() {}

   bool     hasComment() { return (m_comment)?true:false; }
   string * getComment() const { return m_comment; }
   void     setComment( char * str ) { m_comment = new string( str ); }

private:
   string * m_comment;
   // add others..

};


#endif /* __SGL_LOAD_FILE_FLT_HPP */
