/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglLoadImage_jpeg.cpp
 *   Author: Tom Stimson
 *  Created: January 2002
 *  Summary: read jpeg format image files using external library
 *****************************************************************************/

#include <sglLoadImage_jpeg.hpp>
#include <sglOStream.hpp>
#include <sglImageLoader.hpp>
#include <stdio.h>

// Define SGL_JPEG_CXX at compile time if your thirdpart JPEG library was
// compiled with a C++ compiler with all its name mangling.

#if defined(SGL_JPEG_CXX)
#include <jpeglib.h>
#else
extern "C"
{
#include <jpeglib.h>
}
#endif
#include <setjmp.h>

/*****************************************************************************/
struct my_error_mgr {
  struct jpeg_error_mgr pub;

  jmp_buf setjmp_buffer;
};

typedef struct my_error_mgr * my_error_ptr;

/*****************************************************************************/
void my_error_exit (j_common_ptr cinfo)
{
   /* cinfo->err really points to a my_error_mgr struct, so coerce pointer */
   my_error_ptr myerr = (my_error_ptr) cinfo->err;

   /* Return control to the setjmp point */
   longjmp(myerr->setjmp_buffer, 1);
}

/*****************************************************************************/
void *loadImageJPG(const string &filepath,
                   unsigned int &width,
                   unsigned int &height,
                   sglTexture::ExternalFormatEnum &format,
                   sglTexture::TypeEnum &type)
{
   return loadImageJPEG(filepath, width, height, format, type);
}

/*****************************************************************************/
void *loadImageJPEG(const string &filepath,
                    unsigned int &width,
                    unsigned int &height,
                    sglTexture::ExternalFormatEnum &format,
                    sglTexture::TypeEnum &type)
{
//   sglPrint() << "loadImageJPEG()" << endl;

   JSAMPLE *image = NULL;
   FILE *file;
   if ((file = fopen(filepath.c_str(), "rb")) == NULL)
   {
      sglPrintWarning() << "loadImageJPEG(): Error opening file "
                        << filepath << endl;
      return NULL;
   }

   jpeg_decompress_struct cinfo;
   my_error_mgr jerr;

   cinfo.err = jpeg_std_error(&jerr.pub);
   jerr.pub.error_exit = my_error_exit;

   /* Establish the setjmp return context for my_error_exit to use. */
   if (setjmp(jerr.setjmp_buffer))
   {
      /* If we get here, the JPEG code has signaled an error.
       * We need to clean up the JPEG object, close the input file, and return.
       */
      char msg[JMSG_LENGTH_MAX];
      (*cinfo.err->format_message)((j_common_ptr)&cinfo, msg);
      sglPrintWarning() << "Error reading file "
                        << filepath << "; " << msg << endl;

      jpeg_destroy_decompress(&cinfo);
      fclose(file);
      if (image) free(image);
      return NULL;
   }

   jpeg_create_decompress(&cinfo);
   jpeg_stdio_src(&cinfo, file);

   jpeg_read_header(&cinfo, TRUE);
   jpeg_start_decompress(&cinfo);

   width = cinfo.output_width;
   height = cinfo.output_height;
   unsigned int size = width * height;

//   sglPrint() << "width = " << width << endl;
//   sglPrint() << "height = " << height << endl;

   if (cinfo.out_color_components == 1)
   {
      format = sglTexture::eLUMINANCE;
//      sglPrint() << "format = LUMINANCE" << endl;
   }
   else if (cinfo.out_color_components == 3)
   {
      if (cinfo.output_components == 3)
      {
         format = sglTexture::eRGB;
//         sglPrint() << "format = RGB" << endl;
         size *= 3;
      }
      else
      {
         sglPrintWarning() << "Unable to handle quantized colors in file: "
                           << filepath << endl;
         jpeg_destroy_decompress(&cinfo);
         fclose(file);
         return NULL;
      }
   }
   else
   {
      sglPrintWarning() << "Don't know what to do with "
                        << cinfo.out_color_components
                        << " component images in file: "
                        << filepath << endl;
      jpeg_destroy_decompress(&cinfo);
      fclose(file);
      return NULL;
   }

#if BITS_IN_JSAMPLE == 8
   type = sglTexture::eUNSIGNED_BYTE;
#elif BITS_IN_JSAMPLE == 12
   type = sglTexture::eUNSIGNED_SHORT;
   size *= sizeof(unsigned short);
#else
# error "Invalid JSAMPLE size"
#endif

//   sglPrint() << "size = " << size << endl;

   unsigned int line_length = cinfo.output_components * width;

   image = (JSAMPLE*)malloc(size);
   JSAMPLE *tmp = image + line_length*(height-1);
   while (cinfo.output_scanline < height)
   {
//      sglPrint() << ".";
      jpeg_read_scanlines(&cinfo, &tmp, 1);
      tmp -= line_length;
   }
//   sglPrint() << endl << "done" << endl;

#if BITS_IN_JSAMPLE == 12
   for (unsigned int i=0; i<width*height*cinfo.output_components; ++i)
   {
      image[i] = (float)image[i] * (float)USHRT_MAX / (float)MAXJSAMPLE;
   }
#endif

   jpeg_finish_decompress(&cinfo);
   jpeg_destroy_decompress(&cinfo);

   fclose(file);

   return image;
}


/*****************************************************************************/
bool writeImageJPG(const string &filepath,
                   const void *image,
                   unsigned int width,
                   unsigned int height,
                   sglTexture::ExternalFormatEnum format,
                   sglTexture::TypeEnum type)
{
   return writeImageJPEG(filepath, image, width, height, format, type);
}

/*****************************************************************************/
bool writeImageJPEG(const string &filepath,
                    const void *image,
                    unsigned int width,
                    unsigned int height,
                    sglTexture::ExternalFormatEnum format,
                    sglTexture::TypeEnum type)
{
//   sglPrint() << "writeImageJPEG()" << endl;

   if (image == NULL)
      return false;

   FILE *file;
   if ((file = fopen(filepath.c_str(), "wb")) == NULL)
   {
      sglPrintWarning() << "writeImageJPEG(): Error opening file "
                        << filepath << endl;
      return false;
   }

   void *tmp_image = NULL;

#if BITS_IN_JSAMPLE == 8
# define new_type sglTexture::eUNSIGNED_BYTE
#elif BITS_IN_JSAMPLE == 12
# define new_type sglTexture::eUNSIGNED_SHORT
#else
# error "Invalid JSAMPLE size"
#endif

   if (type != new_type)
   {
      tmp_image = sglImageLoader::changeImageType(
         format, type, width, height, new_type, image);
      type = new_type;
   }

   sglTexture::ExternalFormatEnum new_format = format;
   switch (format)
   {
      case sglTexture::eALPHA:
      case sglTexture::eLUMINANCE_ALPHA:
      case sglTexture::eLUMINANCE:
         new_format = sglTexture::eLUMINANCE;
         break;
      default:
         new_format = sglTexture::eRGB;
         break;
   }

   if (new_format != format)
   {
      void *tmp_image2 = sglImageLoader::reformatImage(
         format, type, width, height, new_format,
         (tmp_image ? (const void *)tmp_image : image));
      if (tmp_image)
         free(tmp_image);
      tmp_image = tmp_image2;
      format = new_format;
   }

//   sglPrint() << "1" << endl;

   jpeg_compress_struct cinfo;
   my_error_mgr jerr;

   cinfo.err = jpeg_std_error(&jerr.pub);
   jerr.pub.error_exit = my_error_exit;

   /* Establish the setjmp return context for my_error_exit to use. */
   if (setjmp(jerr.setjmp_buffer))
   {
      /* If we get here, the JPEG code has signaled an error.
       * We need to clean up the JPEG object, close the input file, and return.
       */
      char msg[JMSG_LENGTH_MAX];
      (*cinfo.err->format_message)((j_common_ptr)&cinfo, msg);
      sglPrintWarning() << "Error reading file "
                        << filepath << "; " << msg << endl;

      jpeg_destroy_compress(&cinfo);
      fclose(file);
      if (tmp_image)
         free(tmp_image);
      return false;
   }

//   sglPrint() << "2" << endl;

   jpeg_create_compress(&cinfo);
   jpeg_stdio_dest(&cinfo, file);

   cinfo.image_width = width;
   cinfo.image_height = height;
   if (format == sglTexture::eLUMINANCE)
   {
      cinfo.input_components = 1;
      cinfo.in_color_space = JCS_GRAYSCALE;
   }
   else
   {
      cinfo.input_components = 3;
      cinfo.in_color_space = JCS_RGB;
   }

//   sglPrint() << "3" << endl;

   jpeg_set_defaults(&cinfo);
   jpeg_start_compress(&cinfo, TRUE);

#if BITS_IN_JSAMPLE == 12
   if (tmp_image == NULL)
   {
      tmp_image = malloc(width*height*cinfo.input_components);
      memcpy(tmp_image, image, width*height*cinfo.input_components);
   }
   for (unsigned int i=0; i<width*height*cinfo.input_components; ++i)
   {
      tmp_image[i] =
         (float)tmp_image[i] * (float)MAXJSAMPLE / (float)USHRT_MAX;
   }
#endif

   unsigned int line_length = cinfo.input_components * width;
   JSAMPLE *tmp = (tmp_image ? (JSAMPLE*)tmp_image : (JSAMPLE*)image)
      + line_length*(height-1);

//   sglPrint() << "4" << endl;
   while (cinfo.next_scanline < height)
   {
//      sglPrint() << ".";
      jpeg_write_scanlines(&cinfo, &tmp, 1);
      tmp -= line_length;
   }
//   sglPrint() << endl << "done" << endl;

   jpeg_finish_compress(&cinfo);
   jpeg_destroy_compress(&cinfo);

   if (tmp_image)
      free(tmp_image);

   fclose(file);

   return true;
}
