/* vi:set ts=3 sw=3 expandtab: */
/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 * SGL's MD3 loader Copyright (C) 2002  Michael M. Morrison  All Rights
 *   Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: md3file.c
 *  Created: March 2002
 *  Author : Michael M. Morrison
 *  Summary: Quake 3 MD3 file format loader.
 *****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "md3file.h"

#ifdef _TEST_

static void
printMD3Header( MD3Header * h )
{
   printf("+MD3Header ----\n");
   printf("\tID: %.4s\n", h->ID );
   printf("\tVersion: %d\n", h->Version );
   printf("\tfileName: %s\n", h->fileName );
   printf("\tnumBoneFrames: %d\n", h->numBoneFrames );
   printf("\tnumTags: %d\n", h->numTags );
   printf("\tnumMeshes: %d\n",   h->numMeshes ); 
   printf("\tnumTextures: %d\n",   h->maxTextures );
   printf("\theaderSize: %d\n",   h->headerSize );
   printf("\tstartTag: %d\n",   h->startTag ); 
   printf("\tendTag: %d\n",   h->endTag );  
   printf("\tfileSize: %d\n",   h->fileSize );
}

static void
printMD3MeshHeader( MD3MeshHeader * h )
{
   printf("+MD3MeshHeader ----\n");
   printf("\tID: %.4s\n", h->ID );
   printf("\tname: %s\n", h->name );
   printf("\tnumMeshFrames: %d\n", h->numMeshFrames );
   printf("\tnumTextures: %d\n", h->numTextures );
   printf("\tnumVerts: %d\n", h->numVerts );
   printf("\tnumTris: %d\n",   h->numTris ); 
   printf("\tstartTri: %d\n",   h->startTri );
   printf("\theaderSize: %d\n",   h->headerSize );
   printf("\tstartTxVec: %d\n",   h->startTxVec ); 
   printf("\tstartVertex: %d\n",   h->startVertex );  
   printf("\tmeshSize: %d\n",   h->meshSize );
}

static void
printMD3Texture( MD3Texture * h )
{
   printf("+MD3Texture ----\n");
   printf("\tname: %s\n", h->name );
}

static void
printMD3Triangle( MD3Triangle * h )
{
   printf("+MD3Triangle ----\n");
   printf("\tindices: %d %d %d\n", h->indices[0], h->indices[1], h->indices[2]);
}

static void
printMD3UVCoord( MD3UVCoord * h )
{
   printf("+MD3UVCoord ----\n");
   printf("\tuv: %f %f\n", h->uv[0], h->uv[1] );
}

static void
printMD3Vertex( MD3Vertex * h )
{
   printf("+MD3Vertex ----\n");
   printf("\tcoord: %d %d %d\n", h->coord[0], h->coord[1], h->coord[2] );
   printf("\tenvTex: %d %d\n", h->envTex[0], h->envTex[1] );
}

static void
printMD3Mesh( MD3Mesh * m )
{
   int i;

   printMD3MeshHeader( &m->header );

   for(i=0;i<m->header.numTextures;i++)
         printMD3Texture( &m->texture[i] );

   for(i=0;i<m->header.numTris;i++)
         printMD3Triangle( &m->indices[i] );

   for(i=0;i<m->header.numVerts;i++)
         printMD3UVCoord( &m->uvCoord[i] );

   for(i=0;i<m->header.numVerts;i++)
         printMD3Vertex( &m->vertex[i] );
}

#endif /* _TEST_ */

static void
freeMD3( MD3File * f )
{
   int i;

   if( f )
   {
      if( f->tag )
         free( f->tag );

      if( f->boneFrame )
         free( f->boneFrame );

      if( f->mesh )
      {
         for(i=0;i<f->header.numMeshes;i++)
         {
            if( f->mesh[i].texture )
               free( f->mesh[i].texture );

            if( f->mesh[i].indices )
               free( f->mesh[i].indices );

            if( f->mesh[i].uvCoord )
               free( f->mesh[i].uvCoord );

            if( f->mesh[i].vertex )
               free( f->mesh[i].vertex );
         }

         free( f->mesh );
      }

      free( f );
   }
}

static MD3File *
fail( MD3File * md3 )
{
   if( md3->file )
      fclose( md3->file );

   freeMD3( md3 );

   return 0;
}

MD3File * 
md3Open( const char * name )
{
   MD3File * md3;

   if(!( md3 = (MD3File *) calloc( 1, sizeof( MD3File ) ) ))
      return fail( md3 );

   if(!( md3->file = fopen( name, "rb" ) ) )
      return fail( md3 );

   return md3;
}

int
md3Parse( MD3File * md3 )
{
   int result = 0;

   // read header
   result = fread( &md3->header, sizeof( MD3Header ), 1, md3->file );

   if( result != 1 )
      return (int)fail( md3 );

   // check for tags
   if( md3->header.startTag != md3->header.endTag && md3->header.numTags > 0 )
   {
      int numTags = md3->header.numTags * md3->header.numBoneFrames;

      if( numTags )
      {
         md3->tag = (MD3Tag *) calloc( 1, numTags * sizeof( MD3Tag ) );

         if( !md3->tag )
            return (int)fail( md3 );

         result = fread( md3->tag, sizeof( MD3Tag ), numTags, md3->file ); 

         if( result != numTags )
            return (int)fail( md3 );
      }
   }

   // check for boneFrames
   if( md3->header.numBoneFrames > 0 )
   {
      md3->boneFrame = (MD3BoneFrame *) calloc( 1, md3->header.numBoneFrames * 
                                                sizeof( MD3BoneFrame ) );

      if( !md3->boneFrame )
         return (int)fail( md3 );

      result = fread( md3->boneFrame, sizeof( MD3BoneFrame ), 
                                       md3->header.numBoneFrames, md3->file );

      if( result != md3->header.numBoneFrames )
         return (int)fail( md3 );
   }

   // check for meshes
   if( md3->header.numMeshes > 0 )
   {
      int i;

      md3->mesh = (MD3Mesh *) calloc( 1, md3->header.numMeshes * 
                                                sizeof( MD3Mesh ) );

      if( !md3->mesh )
         return (int)fail( md3 );

      for( i=0; i<md3->header.numMeshes; i++ )
      {
         if( 1 != fread( &md3->mesh[i].header, sizeof( MD3MeshHeader ), 1, 
                                                                  md3->file )) 
            return (int)fail( md3 );

         if( md3->mesh[i].header.numTextures )
         {
            if(!( md3->mesh[i].texture = (MD3Texture *) calloc( 
                     md3->mesh[i].header.numTextures, sizeof( MD3Texture ) ) ))
               return (int)fail( md3 );

            if( (unsigned)md3->mesh[i].header.numTextures != 
                  fread( md3->mesh[i].texture, sizeof( MD3Texture ), 
                                 md3->mesh[i].header.numTextures, md3->file )) 
               return (int)fail( md3 );
         }

         if( md3->mesh[i].header.numTris )
         {
            if(!( md3->mesh[i].indices = (MD3Triangle *) calloc( 
                                 md3->mesh[i].header.numTris, 
                                 sizeof( MD3Triangle ) ) ))
               return (int)fail( md3 );

            if( (unsigned)md3->mesh[i].header.numTris != 
                  fread( md3->mesh[i].indices, sizeof( MD3Triangle ), 
                                    md3->mesh[i].header.numTris, md3->file )) 
               return (int)fail( md3 );
         }

         if( md3->mesh[i].header.numVerts )
         {
            //
            // Texture Coords
            //
            if(!( md3->mesh[i].uvCoord = (MD3UVCoord *) calloc( 
                        md3->mesh[i].header.numVerts, sizeof( MD3UVCoord ) ) ))
               return (int)fail( md3 );

            if( (unsigned)md3->mesh[i].header.numVerts != 
                  fread( md3->mesh[i].uvCoord, sizeof( MD3UVCoord ), 
                                    md3->mesh[i].header.numVerts, md3->file )) 
               return (int)fail( md3 );

            //
            // Verts
            // NOTE: Read Frame 0 only for now
            //
            //
            if(!( md3->mesh[i].vertex = (MD3Vertex *) calloc( 
                        md3->mesh[i].header.numVerts, sizeof( MD3Vertex ) ) ))
               return (int)fail( md3 );

            if( (unsigned)md3->mesh[i].header.numVerts != 
                  fread( md3->mesh[i].vertex, sizeof( MD3Vertex ), 
                                    md3->mesh[i].header.numVerts, md3->file )) 
               return (int)fail( md3 );

            // skip other frames, if they exist

            if( md3->mesh[i].header.numMeshFrames > 1 )
            {
               result = fseek( md3->file, sizeof( MD3Vertex ) * 
                        md3->mesh[i].header.numVerts * 
                        ( md3->mesh[i].header.numMeshFrames - 1 ), SEEK_CUR );

               if( result != 0 )
                  return (int)fail( md3 );
            }
         }
      }
   }

   return 1;
}

void
md3Close( MD3File * f )
{
   if( f )
   {
      if( f->file )
         fclose( f->file );

      f->file = 0;
   }
}

void
md3Free( MD3File * f )
{
   freeMD3( f );
}

#ifdef _TEST_

int
main( int argc, char **argv )
{
   MD3File * f;
   int i;

   f = md3Open( argv[1] );

   if( f ) {
      md3Parse( f );
      md3Close( f );

      printf("success\n");
      printMD3Header( &f->header );

      for(i=0;i<f->header.numMeshes;i++)
         printMD3Mesh( &f->mesh[i] );

      md3Free( f );
   } else
      printf("Fail\n");
}

#endif
