/* vi:set ts=3 sw=3 expandtab: */
/* I do not believe that the "original" vi has 'expandtab' (TABs=spaces) */
/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 * SGL's md3 loader: Copyright (C) 2002  Michael M. Morrison
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglLoadFile_md3.cpp
 *  Created: March 2002
 *  Author : Michael M. Morrison
 *  Summary: Loader for Quake 3 MD3 files.
 *****************************************************************************/

#include <sglLoadFile_md3.hpp>
#include <sglImageLoader.hpp>

#include <sglGroup.hpp>
#include <sglTransform.hpp>
#include <sglLOD.hpp>
#include <sglIndexedBitMaskSwitch.hpp>
#include <sglKeyFrameAnimation.hpp>
#include <sglGeode.hpp>

#include <sglGeoSet.hpp>
#include <sglQuadSet.hpp>
#include <sglLineSet.hpp>
#include <sglPointSet.hpp>
#include <sglPolygonSet.hpp>
#include <sglTriangleSet.hpp>
#include <sglIndexedTriangleStripSet.hpp>
#include <sglIndexedTriangleSet.hpp>

#include <sglMonoIndexedTriangleStripSet.hpp>
#include <sglMonoIndexedTriangleSet.hpp>

#include <sglStatelet.hpp>
#include <sglMaterial.hpp>
#include <sglTexture2D.hpp>
#include <sglTexEnv.hpp>
#include <sglTexGen.hpp>

#include <sglBillboard.hpp>

#include <math.h>

//----------------------------------------------------------------------------

extern "C"
{
   SGLMD3_DLL_API
   void *loadModelMD3(const string &filepath,
                      unsigned int mode)
   {
      sglPrint() << "loadModelMD3()" << endl;
      sglLoadFile_md3 ldr;
      sglNode *model = ldr.loadFile(filepath, mode);
      return model;
   }
}

//============================================================================
// class sglLoadFile_md3
//============================================================================

map<string, sglTexture2D*> sglLoadFile_md3::s_tex_map;

//----------------------------------------------------------------------------
sglLoadFile_md3::sglLoadFile_md3()
{
   sglPrint() << "sglLoadFile_md3::sglLoadFile_md3()" << endl;
   m_verbose = true;
   m_compress_textures = false;

   m_default_min_filter = sglTexture::eLINEAR_MIPMAP_LINEAR;
   m_default_mag_filter = sglTexture::eLINEAR;
   m_default_compression_level = sglTexture::eCOMPRESS_MAX;
}

//----------------------------------------------------------------------------
sglLoadFile_md3::~sglLoadFile_md3()
{
   // free all resources
   if( m_md3 )
      md3Free( m_md3 );
}

//----------------------------------------------------------------------------
sglTexture2D *
sglLoadFile_md3::lookupTexture( const char * fileName )
{
   sglTexture2D * tex = 0;
   string searchName( fileName );
   string foundName;

   if (sglDataLoader::findFile(searchName, &m_alternate_path, foundName))
   {
      tex = createTexture( foundName );
   }
   else
   {
      sglPrintWarning() << "sglLoadFile_md3::buildScene error: "
                           << "Unable to find texture: '"
                           << searchName << endl;
   }

   return tex;
}

//----------------------------------------------------------------------------
sglTexture2D *sglLoadFile_md3::createTexture(const string &filename)
{
   sglTexture::ExternalFormatEnum format;
   sglTexture::TypeEnum type;
   sglTexture2D * tex = NULL;
   unsigned int num_s, num_t;
   void * image = NULL;

   map<string, sglTexture2D*>::iterator elt = s_tex_map.find(filename);

   if (elt == s_tex_map.end())
   {
      image = sglImageLoader::loadImage(filename, num_s, num_t, format, type);

      if (image && num_s && num_t)
      {
         tex = new sglTexture2D;

         tex->setAutoMipmapFlag(true);

         tex->setWrapMode(sglTexture::eREPEAT, sglTexture::eREPEAT);
         tex->setMinFilter( m_default_min_filter );
         tex->setMagFilter( m_default_mag_filter );

         // set max anisotropic filter
         //tex->setMaxAnisotropy( tex->getMaxSupportedMaxAnisotropy() );

         tex->setImage(image, format, type, num_s, num_t, false);

         if( m_compress_textures )
         {
            tex->setCompressionLevel( m_default_compression_level );
         }

         if( m_verbose )
            sglPrintInfo()<<"sglLoadFile_md3::createTexture: Texture loaded: '"
                       << filename << "' (" << num_s << "x" << num_t << ")"
                       << endl;
      }
      else
         return 0;

      s_tex_map[filename] = tex;

   }
   else
   {
      if( m_verbose )
         sglPrintInfo() << "sglLoadFile_md3::createTexture: Texture reused: '"
                     << filename << "'" << endl;
      tex = elt->second;
   }

   return tex;
}

//----------------------------------------------------------------------------
void sglLoadFile_md3::buildScene( sglGroup * parent )
{
   int i;

   // name the file
   parent->setName( m_md3->header.fileName );

   //
   // Create one geode per mesh, and name them
   // appropriately.
   //

   for( i=0; i < m_md3->header.numMeshes; i++ )
   {
      MD3Mesh * mesh = &m_md3->mesh[i];
      sglGeode * geode = new sglGeode();
      sglMonoIndexedTriangleSet * geom = new sglMonoIndexedTriangleSet();

      //
      // name geode
      //
      geode->setName( mesh->header.name );

      //
      // copy indices
      //
      unsigned int * ilist = new unsigned int[ mesh->header.numTris * 3 ];
      int j;
      for( j=0; j<mesh->header.numTris; j++ )
      {
         memcpy( &ilist[j*3], &mesh->indices[j], sizeof( unsigned int ) * 3 );
      }

      geom->setIndexList( ilist );

      //
      // Set CoordList
      //

      sglVec3f * vertexList = new sglVec3f[ mesh->header.numVerts ];

      for( j=0; j<mesh->header.numVerts; j++ )
      {
         vertexList[j].set( (float)(mesh->vertex[j].coord[0])/100.0f,
                            (float)(mesh->vertex[j].coord[1])/100.0f,
                            (float)(mesh->vertex[j].coord[2])/100.0f );
      }

      geom->setCoordList( new sglDeleteableDataPtr<sglVec3f>(vertexList) );

      //
      // Set Number of Primitives
      //
      geom->setNumPrims( mesh->header.numTris );

      // turn on display lists
      geom->enableDisplayList( true );

      //
      // Set up Statelets
      //

      vector<sglStatelet *>state;
      sglVec2f * uvCoord = 0;

      sglFrontFace * face = new sglFrontFace();
      face->setFrontFace( sglFrontFace::eCW );

      state.push_back( face );

      //
      // Texture
      //

      if( mesh->header.numTextures )
      {
         sglTexture2D * tex = lookupTexture( mesh->texture[0].name );

         if( tex )
         {
            //
            // Get UV Coords
            //
            uvCoord = new sglVec2f[ mesh->header.numVerts ];

            for( int j=0; j < mesh->header.numVerts; j++ )
            {
               uvCoord[j].set( mesh->uvCoord[j].uv[0],
                               mesh->uvCoord[j].uv[1] );
            }

            //
            // Set up texture modes
            //
            sglTexEnv * tev = new sglTexEnv();

            tev->setMode( sglTexEnv::eMODULATE );

            state.push_back( tex );
            state.push_back( tev );
         }
      }

      geom->setState( 0, state, new sglDeleteableDataPtr<sglVec2f>(uvCoord) );

      geode->addGeometry( geom );
      parent->addChild( geode );
   }
}

//----------------------------------------------------------------------------
sglNode *sglLoadFile_md3::loadFile(const string &filename,
                                   unsigned int /* mode */)
{
   sglPrint() << "sglLoadFile_md3::loadFile()" << endl;
   sglGroup *group = NULL;

   // store the path for possible future use
   m_alternate_path = sglDataLoader::extractDirectoryFromPath(filename);

   // open, parse, close file
   m_md3 = md3Open( filename.c_str() );

   // try using sglDataLoader's search capabilities
   if (!m_md3)
   {
      string foundname;
      if (sglDataLoader::findFile(filename, NULL, foundname))
      {
         m_md3 = md3Open( foundname.c_str() );
         if (m_md3)
         {
            m_alternate_path =
               sglDataLoader::extractDirectoryFromPath(foundname);
         }
      }
   }

   if (m_md3)
   {
      md3Parse( m_md3 );
      md3Close( m_md3 );

      group = new sglGroup;

      buildScene( group );

      if (group->getNumChildren() == 0)
      {
         delete group;
         group = NULL;
      }
   }

   return group;
}
