/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglLoadFile_obj.cpp
 *  Created: 7 August 1997
 *  Summary: Wrapper class/frontend to the Obj-loader which moves the data
 *           into a SGL sub-graph
 *****************************************************************************/

#include <sglobj.h>
#include <sglGeode.hpp>
#include <sglGeoSet.hpp>
#include <sglTexture2D.hpp>
#include <sglTexEnv.hpp>
#include <sglTexGen.hpp>

#include <sglObjModel.hpp>
#include <sglTriStripper.hpp>
#include <sglObjMtl.hpp>

#include <sglLoadFile_obj.hpp>


extern "C"
{
   SGLOBJ_DLL_API
   void *loadModelOBJ(const string &filepath,
                      unsigned int mode)
   {
      sglLoadFile_obj ldr;
      sglNode *model = ldr.loadFile(filepath, mode);
      return model;
   }
}



//============================================================================
// class sglLoadFile_obj
//============================================================================

//----------------------------------------------------------------------------
sglLoadFile_obj::sglLoadFile_obj()
{
}

//----------------------------------------------------------------------------
sglLoadFile_obj::~sglLoadFile_obj()
{
   if (m_vlist && m_vlist->getRefCount() == 0)
      delete m_vlist;
   if (m_nlist && m_nlist->getRefCount() == 0)
      delete m_nlist;
   if (m_tlist && m_tlist->getRefCount() == 0)
      delete m_tlist;
}

//----------------------------------------------------------------------------
sglGeode *sglLoadFile_obj::loadFile(const string &filename,
                                    unsigned int mode)
{
   unsigned int i;
   sglGeode *geode = new sglGeode;

   sglObjModel obj_model(filename, mode);

   // convert the attribute data from the sglObjModel to the pfVec* lists
   // allocated out of shared memory.
   geometryStruct geom;
   obj_model.getGeometry(&geom);

   m_vlist = new sglFreeableDataPtr(geom.vlist);
   m_nlist = new sglFreeableDataPtr(geom.nlist);
   m_tlist = new sglFreeableDataPtr(geom.tlist);

   // Now get the state and index list info from the sglTriStrippers
   unsigned int num_mg = obj_model.getNumMtlGroups();
   for (i=0; i<num_mg; i++)
   {
      // get the material (if any)
      sglObjMtlGroup *group = obj_model.getMtlGroup(i);
      if (group->stripperVNT->getNumTriangles() ||
          group->stripperVNT->getNumStrips())
      {
         getTriangles(group->mtl, geode, group->stripperVNT, true, true);
      }
      if (group->stripperVN->getNumTriangles() ||
          group->stripperVN->getNumStrips())
      {
         getTriangles(group->mtl, geode, group->stripperVN, true, false);
      }
      if (group->stripperVT->getNumTriangles() ||
          group->stripperVT->getNumStrips())
      {
         getTriangles(group->mtl, geode, group->stripperVT, false, true);
      }
      if (group->stripperV->getNumTriangles() ||
          group->stripperV->getNumStrips())
      {
         getTriangles(group->mtl, geode, group->stripperV, false, false);
      }
   }

   return geode;
}

//----------------------------------------------------------------------------
void sglLoadFile_obj::makeStatelets(sglObjMtl *obj_mtl,
                                    bool hasNormals,
                                    bool hasTexCoords,
                                    vector<sglStatelet*> &statelets)
{
   float su = 1.0f;
   float sv = 1.0f;

   if (obj_mtl)
   {
      // get the material information for this material group
      statelets.push_back(obj_mtl->getMaterial());

      sglLighting* li = new sglLighting();
      li->enable(hasNormals);
      statelets.push_back(li);

      if (obj_mtl->getTexture(&su, &sv) && hasTexCoords)
      {
         statelets.push_back(obj_mtl->getTexture(&su, &sv));

         if (!hasNormals)
         {
            sglTexEnv *tenv = new sglTexEnv();
            tenv->setMode(sglTexEnv::eDECAL);
            statelets.push_back(tenv);
         }

         if (su != 1.f || sv != 1.f)
         {
            sglMat4f mat;
            mat.buildScale(su, sv, 1.f);

            sglTextureMatrix* tm = new sglTextureMatrix();
            tm->setTextureMatrix(&mat);
            statelets.push_back(tm);
         }
      }
      else if (obj_mtl->getReflTexture(&su, &sv))
      {
         statelets.push_back(obj_mtl->getReflTexture(&su, &sv));

         sglTexGen *tgen = new sglTexGen();
         tgen->setMode(sglTexGen::eS, sglTexGen::eSPHERE_MAP);
         tgen->setMode(sglTexGen::eT, sglTexGen::eSPHERE_MAP);
         statelets.push_back(tgen);

         sglTexEnv *tenv = new sglTexEnv();
         tenv->setMode(sglTexEnv::eDECAL);
         statelets.push_back(tenv);

         if (su != 1.f || sv != 1.f)
         {
            sglMat4f mat;
            mat.buildScale(su, sv, 1.f);

            sglTextureMatrix* tm = new sglTextureMatrix();
            tm->setTextureMatrix(&mat);
            statelets.push_back(tm);
         }
      }
   }
}
