/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglLoadFile_obj.hpp
 *  Created: 7 August 1997
 *  Summary: Wrapper class/frontend to the Obj-loader which moves the data
 *           into a scene graph (morphable) structure.
 *****************************************************************************/

#ifndef __SGL_LOAD_FILE_OBJ_HPP
#define __SGL_LOAD_FILE_OBJ_HPP

#include <sglobj.h>
#include <sglGeode.hpp>
#include <sglGeoSet.hpp>
#include <sglStatelet.hpp>
#include <sglMaterial.hpp>
#include <sglIndexedTriangleStripSet.hpp>
#include <sglIndexedTriangleSet.hpp>
#include <sglModelLoader.hpp>

template <class T> class SGLOBJ_DLL_API sglTriStripper;
class sglObjMtl;

//============================================================================

class SGLOBJ_DLL_API sglLoadFile_obj
{
public:
   /// default constructor
   sglLoadFile_obj();
   /// virtual destructor
   virtual ~sglLoadFile_obj();

   /// load a file, build a scene graph, and get the sglGeode* pointing to it
   sglGeode *loadFile(const string &filename, unsigned int mode);

private:
   /// Copy constructor forbidden
   sglLoadFile_obj(const sglLoadFile_obj &);
   /// Assignment operator forbidden
   sglLoadFile_obj &operator=(const sglLoadFile_obj &);

   /// build statelets with all the info in the material group
   void makeStatelets(sglObjMtl *obj_mg,
                      bool hasNormals,
                      bool hasTexCoords,
                      vector<sglStatelet*> &statelets);

   /// build a GeoSet with all the isolated triangles in a material group
   template <class T>
   sglGeoSet *getIsolatedTriangles(T *tri_stripper,
                                   bool hasNormals,
                                   bool hasTexCoords,
                                   sglMaterial *material,
                                   const vector<sglStatelet*> &statelets)
      {
         unsigned int num_tris = tri_stripper->getNumTriangles();
         sglIndexedTriangleSet *gset = new sglIndexedTriangleSet;
         gset->setNumPrims(num_tris);

         // always get the vertices
         unsigned int *index_vlist =
            (unsigned int *) malloc(sizeof(unsigned int)*num_tris*3);

         // get the normals and texcoords if present
         unsigned int *index_nlist = NULL, *index_tlist = NULL;

         if (hasNormals)
         {
            index_nlist = (unsigned int *)
               malloc(sizeof(unsigned int)*num_tris*3);
         }
         if (hasTexCoords)
         {
            index_tlist = (unsigned int *)
               malloc(sizeof(unsigned int)*num_tris*3);
         }

         tri_stripper->copyTriangleIndexLists(
            index_vlist, index_nlist, index_tlist, NULL);

         gset->setCoordList(m_vlist);
         gset->setCoordIndexList(new sglFreeableDataPtr(index_vlist));

         if  (hasNormals)
         {
            gset->setNormalList(sglGeoSet::ePER_VERTEX, m_nlist);
            gset->setNormalIndexList(new sglFreeableDataPtr(index_nlist));
         }
         else
         {
            sglVec4f *color = (sglVec4f*) malloc(sizeof(sglVec4f));
            material->getFrontDiffuse(color[0][0], color[0][1], color[0][2]);
            color[0][3] = material->getFrontAlpha();
            gset->setColorList(sglGeoSet::eOVERALL,
                               new sglFreeableDataPtr(color));

            unsigned int *index_clist =
               (unsigned int *)malloc(sizeof(unsigned int));
            index_clist[0] = 0;
            gset->setColorIndexList(new sglFreeableDataPtr(index_clist));
         }

         if (hasTexCoords)
         {
             gset->setState(0, statelets, m_tlist);
             gset->setTexCoordIndexList(new sglFreeableDataPtr(index_tlist));
         }
         else
             gset->setState(0, statelets);

         return gset;
      }

   /// build a GeoSet with all the indexed tristrips in a material group
   template <class T>
   sglGeoSet *getTriStrips(T *tri_stripper,
                           bool hasNormals,
                           bool hasTexCoords,
                           sglMaterial *material,
                           const vector<sglStatelet*> &statelets)
      {
         unsigned int j;

         unsigned int num_tstrips = tri_stripper->getNumStrips();

         sglIndexedTriangleStripSet *gset = new sglIndexedTriangleStripSet;
         gset->setNumPrims(num_tstrips);

         unsigned int *lengths =
            (unsigned int *) malloc(sizeof(unsigned int)*num_tstrips);

         // count the total number of vertices in all the strips
         unsigned int total_indices = 0;

         for (j=0; j<num_tstrips; j++)
         {
            lengths[j] = tri_stripper->getStrip(j)->size();
            total_indices += lengths[j];
         }
         gset->setPrimLengths(new sglFreeableDataPtr(lengths));

         // allocate index lists that are large enough
         // always get the vertices
         unsigned int *index_vlist =
            (unsigned int *) malloc(sizeof(unsigned int)*total_indices);

         // get the normals and texcoords if present
         unsigned int *index_nlist = 0, *index_tlist = 0;

         if (hasNormals)
         {
            index_nlist =  (unsigned int *)
               malloc(sizeof(unsigned int)*total_indices);
         }
         if (hasTexCoords)
         {
            index_tlist =  (unsigned int *)
               malloc(sizeof(unsigned int)*total_indices);
         }

         unsigned int index = 0;
         for (j=0; j<num_tstrips; j++)
         {
            tri_stripper->copyStrip(j, index_vlist+index,
                                    index_nlist+index,
                                    index_tlist+index,
                                    NULL);
            index += lengths[j];
         }

         gset->setCoordList(m_vlist);
         gset->setCoordIndexList(new sglFreeableDataPtr(index_vlist));
         if  (hasNormals)
         {
            gset->setNormalList(sglGeoSet::ePER_VERTEX, m_nlist);
            gset->setNormalIndexList(new sglFreeableDataPtr(index_nlist));
         }
         else
         {
            sglVec4f *color = (sglVec4f*) malloc(sizeof(sglVec4f));
            material->getFrontDiffuse(color[0][0], color[0][1], color[0][2]);
            color[0][3] = material->getFrontAlpha();
            gset->setColorList(sglGeoSet::eOVERALL,
                               new sglFreeableDataPtr(color));

            unsigned int *index_clist =
               (unsigned int *)malloc(sizeof(unsigned int));
            index_clist[0] = 0;
            gset->setColorIndexList(new sglFreeableDataPtr(index_clist));
         }

         if (hasTexCoords)
         {
             gset->setState(0, statelets, m_tlist);
             gset->setTexCoordIndexList(new sglFreeableDataPtr(index_tlist));
         }
         else
             gset->setState(0, statelets);


         return (sglGeoSet *) gset;
      }

   template <class T>
   void getTriangles(sglObjMtl *obj_mtl,
                     sglGeode* geode,
                     T stripper,
                     bool hasNormals,
                     bool hasTexCoords)
      {
         vector<sglStatelet*> statelets;
         makeStatelets(obj_mtl, hasNormals, hasTexCoords, statelets);

         if (stripper->getNumTriangles())
         {
            sglGeoSet *gset = getIsolatedTriangles(stripper, hasNormals,
                                                   hasTexCoords,
                                                   obj_mtl->getMaterial(),
                                                   statelets);
            geode->addGeometry(gset);
         }

         if (stripper->getNumStrips())
         {
            sglGeoSet *gset = getTriStrips(stripper, hasNormals,
                                           hasTexCoords,
                                           obj_mtl->getMaterial(),
                                           statelets);
            geode->addGeometry(gset);
         }

      }

private:
   sglDataPtr *m_vlist;
   sglDataPtr *m_nlist;
   sglDataPtr *m_tlist;
};


#endif
