/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglObjDataList.cpp
 *  Created: 23 July 1997
 *  Summary: Class to manage the vertex, normal, and texture coordinate lists.
 *           I also make sure none are repeated (i.e., it collapses the list
 *           of duplicates).
 *****************************************************************************/

#include <sglobj.h>
#include <stdio.h>
#include <sglObjDataList.hpp>
#include <sglOStream.hpp>

const float EPS = 1.0e-4f;
const int HASHTABLESIZE = 9991;
const unsigned int CHUNK = 1024;

//----------------------------------------------------------------------------
sglObjDataList::sglObjDataList() :
      m_num_entries(0),
      m_num_entries_available(0),
      m_data(NULL),
      m_num_ii(0),
      m_num_ii_available(0),
      m_ii(NULL),
      m_hash_table(NULL)
{
   m_hash_table = new hashStruct*[HASHTABLESIZE];
   for (int i=0; i < HASHTABLESIZE; i++)
      m_hash_table[i] = NULL;
}

//----------------------------------------------------------------------------
sglObjDataList::~sglObjDataList()
{
   for (int i=0; i < HASHTABLESIZE; i++)
   {
      hashStruct *elt=m_hash_table[i];
      while (elt)
      {
         hashStruct *next = elt->hnext;
         delete elt;
         elt = next;
      }
   }
   delete[] m_hash_table;

   if (m_ii) free(m_ii);
}

//----------------------------------------------------------------------------
void sglObjDataList::add(sglVec3f &data)
{
   // expand indirect indexing list, if necessary
   if (m_ii == NULL)
   {
      m_num_ii_available = CHUNK;
      m_ii = (int *)malloc(sizeof(int)*m_num_ii_available);
   }
   else if (m_num_ii >= m_num_ii_available)
   {
      m_num_ii_available *= 2;
      m_ii = (int *)realloc(m_ii, sizeof(int)*m_num_ii_available);
   }

   // store index of "new" data in indirect index list
   m_ii[m_num_ii++] = addList(data);
}

//----------------------------------------------------------------------------
int sglObjDataList::addWithoutIndirection(sglVec3f &data)
{
   return addList(data);
}

//----------------------------------------------------------------------------
int sglObjDataList::addList(sglVec3f &data)
{
   int index = findHash(data);

   if (index == -1)  // index not found, therefore add it to the list
   {
      if (m_data == NULL)
      {
         m_num_entries_available = CHUNK;
         m_data = (sglVec3f *)malloc(sizeof(sglVec3f)*m_num_entries_available);
      }
      else if (m_num_entries >= m_num_entries_available)
      {
         m_num_entries_available *= 2;
         m_data = (sglVec3f *)realloc(m_data,
                                      sizeof(sglVec3f)*m_num_entries_available);
      }

      // copy into the list
      m_data[m_num_entries][0] = data[0];
      m_data[m_num_entries][1] = data[1];
      m_data[m_num_entries][2] = data[2];

      // add to hash table:
      // can't send a pointer to the element because subsequent reallocs
      // could move the location of the whole list
      addHash(m_num_entries, data);

      // set index and then advance texture coordinate count
      index = m_num_entries++;
   }

   sglPrintDebug() << "\tdata index = " << index << endl;

   return index;
}

//----------------------------------------------------------------------------
sglVec3f& sglObjDataList::getData(unsigned int direct_index)
{
   if (direct_index >= m_num_entries)
   {
      sglPrintWarning() << "sglObjDataList::getData error: index out of range"
                        << endl;
      exit(1);
   }

   return m_data[direct_index];
}

//----------------------------------------------------------------------------
sglVec3f* sglObjDataList::getDataPtr()
{
   return m_data;
}

//----------------------------------------------------------------------------
inline int sglObjDataList::hash(int val[3])
{
   return (((val[0] + val[1] + val[2]) & 0x7fffffff) % HASHTABLESIZE);
}

//----------------------------------------------------------------------------
void sglObjDataList::addHash(int index, sglVec3f &data)
{
   int pos = hash((int*)data.getPtr());

   hashStruct *elt = new hashStruct;
   elt->index = index;

   // add new element to the beginning of the entry list
   elt->hnext = m_hash_table[pos];
   m_hash_table[pos] = elt;
}

//----------------------------------------------------------------------------
int sglObjDataList::findHash(sglVec3f &data)
{
   int pos = hash((int*)data.getPtr());

   for (hashStruct *elt=m_hash_table[pos]; elt; elt=elt->hnext)
   {
      if ((fabs(data[0] - m_data[elt->index][0]) < EPS) &&
          (fabs(data[1] - m_data[elt->index][1]) < EPS) &&
          (fabs(data[2] - m_data[elt->index][2]) < EPS))
      {
         return elt->index;
      }
   }

   return -1;
}

//----------------------------------------------------------------------------
int sglObjDataList::getIndex(int indirect_index)
{
   if (m_num_ii == 0)
   {
      sglPrintWarning() << "Error: cannot get index without adding any data!"
                        << endl;
      exit(1);
   }

   // a normal index
   if (indirect_index > 0)
   {
      if (indirect_index <= m_num_ii)
      {
         return m_ii[indirect_index - 1];
      }
      else
      {
         sglPrintWarning() << "Error: index out of range" << endl;
         exit(1);
      }
   }
      
   // for those icky relative (negative) index references
   else if (indirect_index < 0)
   {
      if ((m_num_ii + indirect_index) >= 0)
      {
         return m_ii[m_num_ii + indirect_index];
      }
      else
      {
         sglPrintWarning() << "Error: negative index is too large." << endl;
         exit(1);
      }
   }

   // a zero index
   else
   {
      sglPrintWarning() << "Error: zero index is not valid in OBJ format."
                        << endl;
      exit(1);
   }

   // it never gets here
   return -1;
}


//----------------------------------------------------------------------------
//    Summary: cleanup excess memory allocation and return handle to list and
//             size
sglVec3f *sglObjDataList::getList(unsigned int &size)
{
   if (m_num_entries == 0)
   {
      size = 0;
      return NULL;
   }

   if (m_num_entries < m_num_entries_available)
   {
      sglPrintDebug() << "Freeing excess data memory "
                      << m_num_entries_available
                      << " - "
                      << m_num_entries
                      << endl;

      sglPrintDebug() << "\t" << m_data << " -> ";
      m_data = (sglVec3f *) realloc(m_data, m_num_entries*sizeof(sglVec3f));
      sglPrintDebug() << m_data << endl;

      m_num_entries_available = m_num_entries;
   }

   size = m_num_entries;
   return m_data;
}


//=========================================================================

//=========================================================================

//=========================================================================

//=========================================================================



//----------------------------------------------------------------------------
sglObjData2List::sglObjData2List() :
      m_num_entries(0),
      m_num_entries_available(0),
      m_data(NULL),
      m_num_ii(0),
      m_num_ii_available(0),
      m_ii(NULL),
      m_hash_table(NULL)
{
   m_hash_table = new hashStruct*[HASHTABLESIZE];
   for (int i=0; i < HASHTABLESIZE; i++)
      m_hash_table[i] = NULL;
}

//----------------------------------------------------------------------------
sglObjData2List::~sglObjData2List()
{
   for (int i=0; i < HASHTABLESIZE; i++)
   {
      hashStruct *elt=m_hash_table[i];
      while (elt)
      {
         hashStruct *next = elt->hnext;
         delete elt;
         elt = next;
      }
   }
   delete[] m_hash_table;

   if (m_ii) free(m_ii);
}

//----------------------------------------------------------------------------
void sglObjData2List::add(sglVec2f &data)
{
   // expand indirect indexing list, if necessary
   if (m_ii == NULL)
   {
      m_num_ii_available = CHUNK;
      m_ii = (int *)malloc(sizeof(int)*m_num_ii_available);
   }
   else if (m_num_ii >= m_num_ii_available)
   {
      m_num_ii_available *= 2;
      m_ii = (int *)realloc(m_ii, sizeof(int)*m_num_ii_available);
   }

   // store index of "new" data in indirect index list
   m_ii[m_num_ii++] = addList(data);
}

//----------------------------------------------------------------------------
int sglObjData2List::addWithoutIndirection(sglVec2f &data)
{
   return addList(data);
}

//----------------------------------------------------------------------------
//    Summary: add data if unique, always return the index of its list
//             storage location.
int sglObjData2List::addList(sglVec2f &data)
{
   int index = findHash(data);

   if (index == -1)  // index not found, therefore add it to the list
   {
      if (m_data == NULL)
      {
         m_num_entries_available = CHUNK;
         m_data = (sglVec2f *)malloc(sizeof(sglVec2f)*m_num_entries_available);
      }
      else if (m_num_entries >= m_num_entries_available)
      {
         m_num_entries_available *= 2;
         m_data = (sglVec2f *)realloc(m_data,
                                      sizeof(sglVec2f)*m_num_entries_available);
      }

      // copy into the list
      m_data[m_num_entries][0] = data[0];
      m_data[m_num_entries][1] = data[1];

      // add to hash table:
      // can't send a pointer to the element because subsequent reallocs
      // could move the location of the whole list
      addHash(m_num_entries, data);

      // set index and then advance texture coordinate count
      index = m_num_entries++;
   }

   sglPrintDebug() << "\tdata index = " << index << endl;

   return index;
}

//----------------------------------------------------------------------------
sglVec2f& sglObjData2List::getData(unsigned int direct_index)
{
   if (direct_index >= m_num_entries)
   {
      sglPrintWarning() << "sglObjData2List::getData error: index out of range"
                        << endl;
      exit(1);
   }

   return m_data[direct_index];
}

//----------------------------------------------------------------------------
sglVec2f* sglObjData2List::getDataPtr()
{
   return m_data;
}

//----------------------------------------------------------------------------
inline int sglObjData2List::hash(int val[2])
{
   return (((val[0] + val[1]) & 0x7fffffff) % HASHTABLESIZE);
}

//----------------------------------------------------------------------------
void sglObjData2List::addHash(int index, sglVec2f &data)
{
   int pos = hash((int*)data.getPtr());

   hashStruct *elt = new hashStruct;
   elt->index = index;

   // add new element to the beginning of the entry list
   elt->hnext = m_hash_table[pos];
   m_hash_table[pos] = elt;
}

//----------------------------------------------------------------------------
int sglObjData2List::findHash(sglVec2f &data)
{
   int pos = hash((int*)data.getPtr());

   for (hashStruct *elt=m_hash_table[pos]; elt; elt=elt->hnext)
   {
      if ((fabs(data[0] - m_data[elt->index][0]) < EPS) &&
          (fabs(data[1] - m_data[elt->index][1]) < EPS))
      {
         return elt->index;
      }
   }

   return -1;
}

//----------------------------------------------------------------------------
int sglObjData2List::getIndex(int indirect_index)
{
   if (m_num_ii == 0)
   {
      sglPrintWarning() << "Error: cannot get index without adding any data!"
                        << endl;
      exit(1);
   }

   // a normal index
   if (indirect_index > 0)
   {
      if (indirect_index <= m_num_ii)
      {
         return m_ii[indirect_index - 1];
      }
      else
      {
         sglPrintWarning() << "Error: index out of range" << endl;
         exit(1);
      }
   }
      
   // for those icky relative (negative) index references
   else if (indirect_index < 0)
   {
      if ((m_num_ii + indirect_index) >= 0)
      {
         return m_ii[m_num_ii + indirect_index];
      }
      else
      {
         sglPrintWarning() << "Error: negative index is too large." << endl;
         exit(1);
      }
   }

   // a zero index
   else
   {
      sglPrintWarning() << "Error: zero index is not valid in OBJ format."
                        << endl;
      exit(1);
   }

   // it never gets here
   return -1;
}


//----------------------------------------------------------------------------
//    Summary: cleanup excess memory allocation and return handle to list and
//             size
sglVec2f *sglObjData2List::getList(unsigned int &size)
{
   if (m_num_entries == 0)
   {
      size = 0;
      return NULL;
   }

   if (m_num_entries < m_num_entries_available)
   {
      sglPrintDebug() << "Freeing excess data memory "
                      << m_num_entries_available
                      << " - "
                      << m_num_entries
                      << endl;

      sglPrintDebug() << "\t" << m_data << " -> ";
      m_data = (sglVec2f *) realloc(m_data, m_num_entries*sizeof(sglVec2f));
      sglPrintDebug() << m_data << endl;

      m_num_entries_available = m_num_entries;
   }

   size = m_num_entries;
   return m_data;
}
