/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglObjDataList.hpp
 *  Created: 23 July 1997
 *  Summary: Class to manage the vertex, normal, and texture coordinate lists.
 *           I also makes sure none are repeated (i.e., it collapses the list
 *           of duplicates).
 *****************************************************************************/

#ifndef __SGL_OBJ_DATA_LIST_HPP
#define __SGL_OBJ_DATA_LIST_HPP

#include <sglobj.h>
#include <sglVector.hpp>

//----------------------------------------------------------------------------

//struct hashStruct;

/// element of data hash table
struct SGLOBJ_DLL_API hashStruct
{
   /// ptr to next hash element with the same hash value
   struct hashStruct *hnext;

   /// actual index of vertex/normal/texcoord in the array
   int index;
};
   
//----------------------------------------------------------------------------

class SGLOBJ_DLL_API sglObjDataList
{
public:
   /// default constructor
   sglObjDataList();
   /// destructor
   virtual ~sglObjDataList();

   /// add a data element to the list and make sure it is not duplicated
   void add(sglVec3f &data);

   /** add a data element to the list but bypass the indirection and return the
       index immediately.  Used to add extra computed normal to the list.
   */
   int addWithoutIndirection(sglVec3f &data);

   /** get the data pointed to by direct reference.  NOTE if indirect indices
       are used a call to getIndex is needed to provide the direct reference.
   */
   sglVec3f& getData(unsigned int direct_index);

   /// returns a pointer to the entire list of data.
   sglVec3f* getDataPtr();

   /// dereference indirect indices, also handles the negative indices
   int getIndex(int indirect_index);

   /// get a handle to the data list and number of elements (size)
   sglVec3f *getList(unsigned int &size);

private:
   /// compute the hash value for the data
   inline int hash(int val[3]);

   /** add data to list and hash table if unique, and return index into
       list; otherwise return index to copy found.
    */
   int  addList(sglVec3f &data);
   /// add data entry to hash table
   void addHash(int index, sglVec3f &data);
   /// return index into m_data array if data is found there
   int findHash(sglVec3f &data);

   // not implemented
   sglObjDataList(const sglObjDataList &);
   sglObjDataList &operator=(const sglObjDataList &);

private:
   /// number of datas stored
   unsigned int m_num_entries;
   /// buffer size of datas
   unsigned int m_num_entries_available;
   /// data list - packed array of sglVec3f's
   sglVec3f *m_data;

   /// number of indirect indexes, also equals number of coords added
   int m_num_ii;
   /// buffer size of indirect indexes
   int m_num_ii_available;
   /// list of indirect indexes
   int *m_ii;

   /// hash table of datas for fast look up
   hashStruct **m_hash_table;
};
   
//----------------------------------------------------------------------------
// These should be a single templated class (with sglVec2f/sglVec3f types)

class SGLOBJ_DLL_API sglObjData2List
{
public:
   /// default constructor
   sglObjData2List();
   /// destructor
   virtual ~sglObjData2List();

   /// add a data element to the list and make sure it is not duplicated
   void add(sglVec2f &data);

   /** add a data element to the list but bypass the indirection and return the
       index immediately.  Used to add extra computed normal to the list.
   */
   int addWithoutIndirection(sglVec2f &data);

   /** get the data pointed to by direct reference.  NOTE if indirect indices
       are used a call to getIndex is needed to provide the direct reference.
   */
   sglVec2f& getData(unsigned int direct_index);

   /// returns a pointer to the entire list of data.
   sglVec2f* getDataPtr();

   /// dereference indirect indices, also handles the negative indices
   int getIndex(int indirect_index);

   /// get a handle to the data list and number of elements (size)
   sglVec2f *getList(unsigned int &size);

private:
   /// compute the hash value for the data
   inline int hash(int val[3]);

   /** add data to list and hash table if unique, and return index into
       list; otherwise return index to copy found.
    */
   int  addList(sglVec2f &data);
   /// add data entry to hash table
   void addHash(int index, sglVec2f &data);
   /// return index into m_data array if data is found there
   int findHash(sglVec2f &data);

   // not implemented
   sglObjData2List(const sglObjData2List &);
   sglObjData2List &operator=(const sglObjData2List &);

private:
   /// number of datas stored
   unsigned int m_num_entries;
   /// buffer size of datas
   unsigned int m_num_entries_available;
   /// data list - packed array of sglVec2f's
   sglVec2f *m_data;

   /// number of indirect indexes, also equals number of coords added
   int m_num_ii;
   /// buffer size of indirect indexes
   int m_num_ii_available;
   /// list of indirect indexes
   int *m_ii;

   /// hash table of datas for fast look up
   hashStruct **m_hash_table;
};


#endif
