/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglObjModel.cpp
 *  Created: 24 June 1997
 *  Summary: 
 *****************************************************************************/

#include <sglobj.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <sglDataLoader.hpp>
#include <sglObjModel.hpp>
#include <sglObjMtlMgr.hpp>
#include <sglObjMtl.hpp>
#include <sglObjDataList.hpp>

const int BUFFER_SIZE = 8192;

//============================================================================
// class sglObjModel
//============================================================================

//----------------------------------------------------------------------------
sglObjModel::sglObjModel(const string &filename,
                         unsigned int mode) :
      m_num_coords(0),
      m_coords(NULL),
      m_num_normals(0),
      m_normals(NULL),
      m_num_texcoords(0),
      m_texcoords(NULL),
      m_mode(mode)
{
   m_mtl_mgr = new sglObjMtlMgr();

   if (parseObjFile(filename))
   {
      if (m_mode&sglModelLoader::eSTRIPIFY)
      {
         for (unsigned int i=0; i<m_groups.size(); i++)
         {
            if (m_groups[i]->stripperV->getNumTriangles())
               m_groups[i]->stripperV->triStripify();
            if (m_groups[i]->stripperVN->getNumTriangles())
               m_groups[i]->stripperVN->triStripify();
            if (m_groups[i]->stripperVT->getNumTriangles())
               m_groups[i]->stripperVT->triStripify();
            if (m_groups[i]->stripperVNT->getNumTriangles())
               m_groups[i]->stripperVNT->triStripify();
         }
      }
   }
}

//----------------------------------------------------------------------------
sglObjModel::~sglObjModel()
{
   // delete the material groups
   while (m_groups.size())
   {
      sglObjMtlGroup *group = m_groups.back();
      m_groups.pop_back();
      delete group->stripperV;
      delete group->stripperVN;
      delete group->stripperVT;
      delete group->stripperVNT;
      delete group;
   }

   delete m_mtl_mgr;

   // delete the coordinate arrays
   if (m_coords)    free(m_coords);
   if (m_normals)   free(m_normals);
   if (m_texcoords) free(m_texcoords);
}

//----------------------------------------------------------------------------
bool sglObjModel::parseObjFile(const string &filename)
{
   FILE *objFile;

   char  buffer[BUFFER_SIZE];
   char  token[BUFFER_SIZE];
   char *next  = NULL;
   char *backslash = NULL;

   int   width  = 0;

   unsigned int   num_triangles = 0;
   unsigned int   num_faces = 0;
   unsigned int   num_groups = 0;

   sglObjMtl *current_mtl = NULL;
   sglObjMtlGroup *current_group = NULL;

   // temporary storage for vertex coordinates (culls duplicates)
   sglObjDataList vlist;
   // temporary storage for normal coordinates (culls duplicates)
   sglObjDataList nlist;
   // temporary storage for texture coordinates (culls duplicates)
   sglObjData2List tlist;

   /* open Wavefront ".obj" file */
   if ((objFile = fopen(filename.c_str(),"r")) == NULL)
   {
      sglPrintWarning() << "sglObjModel::parseOBJFile Error: Unable to open "
                        << filename << endl;
      return false;
   }

   unsigned int line_count = 0;

   /* read Wavefront ".obj" file */
   while (fgets(buffer, BUFFER_SIZE, objFile) != NULL)
   {
      line_count++;

      /* concatenate continuation lines */
      while ((backslash = strchr(buffer, '\\')) != NULL)
      {
         /* replace backslash with space (thanks to Ken Sakai) */
         *backslash++ = ' ';
         *backslash   = '\0';

         /* keep reading */
         if (fgets(backslash, BUFFER_SIZE - strlen(buffer), objFile) == NULL)
            break;
      }

      /* find first non-"space" character in line */
      for (next = buffer; *next != '\0' && isspace(*next); next++)
         /* EMPTY */ {};

      /* skip blank lines and comments ('$' is comment in "cow.obj") */
      if (*next == '\0' || *next == '#' || *next == '!' || *next == '$')
         continue;

      /* extract token */
      sscanf(next, "%s%n", token, &width);
      next += width;

      /* identify token */

      //---------------------------------------------------------------
      if (!sglDataLoader::strCaseCmp(token, "v"))
      {
         // read vertex and insert into list
         sglVec3f vertex;
         sscanf(next, "%f %f %f", &vertex[0], &vertex[1], &vertex[2]);

         vlist.add(vertex);

         sglPrintDebug() << line_count << ": Got vertex "
                         << m_num_coords << endl;
      }

      //---------------------------------------------------------------
      else if (!sglDataLoader::strCaseCmp(token, "vn"))
      {
         // read vertex normal and insert into list
         sglVec3f normal;
         sscanf(next, "%f %f %f", &normal[0], &normal[1], &normal[2]);

         nlist.add(normal);

         sglPrintDebug() << line_count << ": Got normal" << endl;
      }

      //---------------------------------------------------------------
      else if (!sglDataLoader::strCaseCmp(token, "vt"))
      {
         // read texture coordinate and insert into the list
         sglVec2f texcoord;
         float dummy;
         sscanf(next, "%f %f %f", &texcoord[0], &texcoord[1], &dummy);

         tlist.add(texcoord);

         sglPrintDebug() << line_count << ": Got texcoord" << endl;
      }

      //---------------------------------------------------------------
      else if (!sglDataLoader::strCaseCmp(token, "g"))
      {
         // groups are basically ignored
         ++num_groups;
      }

      //---------------------------------------------------------------
      else if (!sglDataLoader::strCaseCmp(token, "f") ||
               !sglDataLoader::strCaseCmp(token, "fo"))
      {
         bool  texture_valid_flag = true;
         bool  normals_valid_flag = true;
         vector<sglVertStructVNT> tmpList;

         char *slash;
         char  vertexData[256];

         // if no material has been specified up to this point, allocate and
         // use the default material.
         if (current_group == NULL)
         {
            if (current_mtl == NULL)
            {
               // get the default material if no material has been specified
               // yet.
               current_mtl = m_mtl_mgr->getMtl("\0");
            }

            // allocate a material group if none found and assign the material
            current_group = findStripper(current_mtl);
         }

         /* parse vertex data from input buffer */
         while (sscanf(next, "%s%n", vertexData, &width) == 1)
         {
            /* advance next pointer past data packet ("n/n/n") */
            next += width;

            /* get vertex coordinate index */
            unsigned int vi = (int)strtol(vertexData, NULL, 10);

            /* get texture coordinate index */
            unsigned int ti = 0;
            if ((slash = strchr(vertexData, '/')) == NULL ||
                (ti = (int)strtol(slash+1, NULL, 10)) == 0)
               texture_valid_flag = false;

            /* get vertex normal index */
            unsigned int ni = 0;
            if (slash == NULL || (slash = strchr(slash+1, '/')) == NULL ||
                (ni = (int)strtol(slash+1, NULL, 10)) == 0)
               normals_valid_flag = false;

            if (normals_valid_flag == false)
               ni = UINT_MAX;

            // dereference indices to get index into new lists
            vi = vlist.getIndex(vi);

            if (texture_valid_flag)
               ti = tlist.getIndex(ti);

            if (normals_valid_flag)
               ni = nlist.getIndex(ni);

            sglVertStructVNT tmp;
            tmp.vi = vi;
            tmp.ni = ni;
            tmp.ti = ti;
            tmpList.push_back(tmp);
         }

         /* setup vertex position  information*/
         sglPrintDebug() << line_count << ": Gotta face:";
         for (unsigned int i=0; i<tmpList.size(); i++)
            sglPrintDebug() << " " << tmpList[i].vi;
         sglPrintDebug() << endl;

         if (tmpList.size() > 2)
         {
            if ((normals_valid_flag || 
                 m_mode&(sglModelLoader::eCOMPUTE_SMOOTH_NORMALS |
                         sglModelLoader::eCOMPUTE_FACETED_NORMALS))
                && texture_valid_flag)
            {
               current_group->stripperVNT->addPolyFace(tmpList);
            }
            else if (normals_valid_flag || 
                     m_mode&(sglModelLoader::eCOMPUTE_SMOOTH_NORMALS |
                             sglModelLoader::eCOMPUTE_FACETED_NORMALS))
            {
               vector<sglVertStructVN> tmpList2;
               for (vector<sglVertStructVNT>::iterator k = tmpList.begin();
                    k != tmpList.end(); ++k)
               {
                  sglVertStructVN tmp;
                  tmp.vi = k->vi;
                  tmp.ni = k->ni;
                  tmpList2.push_back(tmp);
               }
               current_group->stripperVN->addPolyFace(tmpList2);
            }
            else if (texture_valid_flag)
            {
               vector<sglVertStructVT> tmpList2;
               for (vector<sglVertStructVNT>::iterator k = tmpList.begin();
                    k != tmpList.end(); ++k)
               {
                  sglVertStructVT tmp;
                  tmp.vi = k->vi;
                  tmp.ti = k->ti;
                  tmpList2.push_back(tmp);
               }
               current_group->stripperVT->addPolyFace(tmpList2);
            }
            else
            {
               vector<sglVertStructV> tmpList2;
               for (vector<sglVertStructVNT>::iterator k = tmpList.begin();
                    k != tmpList.end(); ++k)
               {
                  sglVertStructV tmp;
                  tmp.vi = k->vi;
                  tmpList2.push_back(tmp);
               }
               current_group->stripperV->addPolyFace(tmpList2);
            }
            num_triangles += tmpList.size() - 2;
            num_faces++;
         }
         else
         {
            sglPrintInfo() << "sglObjModel::parseObjFile Warning: "
                           << "degenerate face encountered - omitting"
                           << endl;
         }
      }

      //---------------------------------------------------------------
      else if (!sglDataLoader::strCaseCmp(token, "usemtl"))
      {
         char mtl_name[256];
         sscanf(next, "%s", mtl_name);
         sglPrintDebug() << line_count
                         << ": Switching materials: " << mtl_name << endl;
         current_mtl = m_mtl_mgr->getMtl(mtl_name);
         current_group = findStripper(current_mtl);
      }

      //---------------------------------------------------------------
      else if (!sglDataLoader::strCaseCmp(token, "mtllib"))
      {
         char lib_name[256];
         sscanf(next, "%s", lib_name);
         sglPrintDebug() << line_count << ": Need to load mtllib: "
                         << lib_name << endl;

         string alt = sglDataLoader::extractDirectoryFromPath(filename);
         string foundname;
         if (sglDataLoader::findFile(lib_name, &alt, foundname))
            m_mtl_mgr->addMtlLib(foundname.c_str());
      }

      //---------------------------------------------------------------
      else if ( !sglDataLoader::strCaseCmp(token, "bevel")  ||
                !sglDataLoader::strCaseCmp(token, "bmat")   ||
                !sglDataLoader::strCaseCmp(token, "bsp")    ||
                !sglDataLoader::strCaseCmp(token, "bzp")    ||
                !sglDataLoader::strCaseCmp(token, "c_interp") ||
                !sglDataLoader::strCaseCmp(token, "cdc")    ||
                !sglDataLoader::strCaseCmp(token, "con")    ||
                !sglDataLoader::strCaseCmp(token, "cstype") ||
                !sglDataLoader::strCaseCmp(token, "ctech")  ||
                !sglDataLoader::strCaseCmp(token, "curv")   ||
                !sglDataLoader::strCaseCmp(token, "curv2")  ||
                !sglDataLoader::strCaseCmp(token, "d_interp") ||
                !sglDataLoader::strCaseCmp(token, "deg")    ||
                !sglDataLoader::strCaseCmp(token, "end")    ||
                !sglDataLoader::strCaseCmp(token, "hole")   ||
                !sglDataLoader::strCaseCmp(token, "l")      ||
                !sglDataLoader::strCaseCmp(token, "lod")    ||
                !sglDataLoader::strCaseCmp(token, "maplib") ||
                !sglDataLoader::strCaseCmp(token, "mg")     ||
                !sglDataLoader::strCaseCmp(token, "o")      ||
                !sglDataLoader::strCaseCmp(token, "p")      ||
                !sglDataLoader::strCaseCmp(token, "param")  ||
                !sglDataLoader::strCaseCmp(token, "parm")   ||
                !sglDataLoader::strCaseCmp(token, "res")    ||
                !sglDataLoader::strCaseCmp(token, "s")      ||
                !sglDataLoader::strCaseCmp(token, "scrv")   ||
                !sglDataLoader::strCaseCmp(token, "shadow_obj") ||
                !sglDataLoader::strCaseCmp(token, "sp")     ||
                !sglDataLoader::strCaseCmp(token, "stech")  ||
                !sglDataLoader::strCaseCmp(token, "step")   ||
                !sglDataLoader::strCaseCmp(token, "surf")   ||
                !sglDataLoader::strCaseCmp(token, "trace_obj") ||
                !sglDataLoader::strCaseCmp(token, "trim")   ||
                !sglDataLoader::strCaseCmp(token, "usemap") ||
                !sglDataLoader::strCaseCmp(token, "vp"))
      {
         // skipped - do nothing
      }

      //---------------------------------------------------------------
      else
      {
         string tmp = ": unrecognized: ";
         tmp += buffer;
         sglPrintInfo() << line_count << tmp << endl;
      }
   }

   /* close Wavefront ".obj" file */
   fclose(objFile);

   // Compute normals, if necessary.
   if (m_mode&sglModelLoader::eCOMPUTE_SMOOTH_NORMALS)
   {
      for (unsigned int i=0; i<m_groups.size(); i++)
      {
         m_groups[i]->stripperV->computeSmoothNormals(
             vlist.getDataPtr(), addNormal, &nlist);
         m_groups[i]->stripperVN->computeSmoothNormals(
             vlist.getDataPtr(), addNormal, &nlist);
         m_groups[i]->stripperVT->computeSmoothNormals(
             vlist.getDataPtr(), addNormal, &nlist);
         m_groups[i]->stripperVNT->computeSmoothNormals(
             vlist.getDataPtr(), addNormal, &nlist);
      }
   }
   else if (m_mode&sglModelLoader::eCOMPUTE_FACETED_NORMALS)
   {
      for (unsigned int i=0; i<m_groups.size(); i++)
      {
         m_groups[i]->stripperV->computeFacetedNormals(
             vlist.getDataPtr(), addNormal, &nlist);
         m_groups[i]->stripperVN->computeFacetedNormals(
             vlist.getDataPtr(), addNormal, &nlist);
         m_groups[i]->stripperVT->computeFacetedNormals(
             vlist.getDataPtr(), addNormal, &nlist);
         m_groups[i]->stripperVNT->computeFacetedNormals(
             vlist.getDataPtr(), addNormal, &nlist);
      }
   }

   // clean up data lists, get a handle to them, and size
   m_coords    = vlist.getList(m_num_coords);
   m_normals   = nlist.getList(m_num_normals);
   m_texcoords = tlist.getList(m_num_texcoords);

   // output sadistics
   sglPrintInfo() <<  "sglObjModel::parseObjFile: " << filename << endl;
   sglPrintInfo() <<  "  Input Data:" << endl;
   if (num_groups != 0)
      sglPrintInfo() <<  "    Groups processed:    "
                     << num_groups << endl;
   if (m_num_coords != 0)
      sglPrintInfo() <<  "    Vertex coordinates:  "
                     << m_num_coords << endl;
   if (m_num_normals != 0)
      sglPrintInfo() <<  "    Vertex normals:      "
                     << m_num_normals << endl;
   if (m_num_texcoords != 0)
      sglPrintInfo() <<  "    Texture coordinates: "
                     << m_num_texcoords << endl;
   if (num_faces != 0)
      sglPrintInfo() <<  "    Input polygons:      "
                     << num_faces << endl;
   if (num_triangles != 0)
      sglPrintInfo() <<  "    Total triangles:     "
                     << num_triangles << endl;

   return true;
}

//----------------------------------------------------------------------------
unsigned int sglObjModel::addNormal(void* normlist, sglVec3f& normal)
{
   sglObjDataList *vlist = (sglObjDataList*)normlist;
   return vlist->addWithoutIndirection(normal);
}

//----------------------------------------------------------------------------
unsigned int sglObjModel::searchNormals(const sglVec3f &normal)
{
   for (unsigned int i=0; i<m_num_normals; i++)
   {
      // do the dot product test (nearly equal to 1 is same normal)
      // assuming of course that the normals have been normalized.
      if (fabs(1.0 - (m_normals[i][0]*normal[0] +
                      m_normals[i][1]*normal[1] +
                      m_normals[i][2]*normal[2])) < 1.0e-3)
      {
         return i;
      }
   }
   return m_num_normals;
}

//----------------------------------------------------------------------------
sglObjMtlGroup *sglObjModel::findStripper(sglObjMtl *mtl)
{
   for (vector<sglObjMtlGroup*>::iterator i = m_groups.begin();
        i != m_groups.end(); ++i)
   {
      if ((*i)->mtl == mtl)
      {
         return *i;
      }
   }

   sglObjMtlGroup *group = new sglObjMtlGroup;
   group->stripperV = new sglTriStripper<sglVertStructV>();
   group->stripperVN = new sglTriStripper<sglVertStructVN>();
   group->stripperVT = new sglTriStripper<sglVertStructVT>();
   group->stripperVNT = new sglTriStripper<sglVertStructVNT>();

   group->mtl = mtl;
   m_groups.push_back(group);

   return group;
}

//----------------------------------------------------------------------------
void sglObjModel::getGeometry(geometryStruct *geom)
{
   geom->num_coords = m_num_coords;
   if (m_num_coords)
   {
      geom->vlist = m_coords;
      m_coords = NULL;
   }

   geom->num_normals = m_num_normals;
   if (m_num_normals)
   {
      geom->nlist = m_normals;
      m_normals = NULL;
   }

   geom->num_texcoords = m_num_texcoords;
   if (m_num_texcoords)
   {
      geom->tlist = m_texcoords;
      m_texcoords = NULL;
   }
}

//----------------------------------------------------------------------------
sglObjMtlGroup *sglObjModel::getMtlGroup(unsigned int index)
{
   return m_groups[index];
}
