/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglObjModel.hpp
 *  Created: 23 June 1997 (tea time)
 *  Summary: 
 *****************************************************************************/

#ifndef __SGL_OBJ_MODEL_HPP
#define __SGL_OBJ_MODEL_HPP

#include <sglobj.h>
#include <vector>

#include <sglVector.hpp>
#include <sglTriStripper.hpp>
#include <sglLoadFile_obj.hpp>
#include <sglModelLoader.hpp>

// forward references
class sglObjMtlMgr;
class sglObjMtl;

//============================================================================

struct SGLOBJ_DLL_API geometryStruct
{
   geometryStruct() 
      : num_coords(0), vlist(NULL), num_normals(0),
        nlist(NULL), num_texcoords(0), tlist(NULL) {}

   unsigned int num_coords;
   sglVec3f *vlist;
   unsigned int num_normals;
   sglVec3f *nlist;
   unsigned int num_texcoords;
   sglVec2f *tlist;
};

//============================================================================

struct SGLOBJ_DLL_API sglObjMtlGroup
{
   sglObjMtl* mtl;
   sglTriStripper<sglVertStructV>* stripperV;
   sglTriStripper<sglVertStructVN>* stripperVN;
   sglTriStripper<sglVertStructVT>* stripperVT;
   sglTriStripper<sglVertStructVNT>* stripperVNT;
};

//============================================================================
/** Top level object - instantiated to load an Wavefront/OBJ file.

  The constructor parses the material and vertex information, and then creates
  a number of instances of sglTriStripper that store each material and the
  associated indexed face data.

  */

class SGLOBJ_DLL_API sglObjModel
{
public:
   // Constructor - calls parseObjFile directly.
   sglObjModel(const string &filename, unsigned int mode);
   // destructor
   virtual ~sglObjModel();

   // return sizes and pointers to all the geometry
   void getGeometry(geometryStruct *geometry);

   // retrieve the number of material groups
   unsigned int getNumMtlGroups() { return m_groups.size(); }

   // retrieve handle to a material group by index
   sglObjMtlGroup *getMtlGroup(unsigned int index);

private:
   // Default constructor forbidden
   sglObjModel();
   // Copy constructor forbidden
   sglObjModel(const sglObjModel &);
   // Assignment operator forbidden
   sglObjModel &operator=(const sglObjModel &);

   // OBJ file parser - heavily borrowed from Performer obj loader source.
   bool parseObjFile(const string &filename);

   // return index of normal
   unsigned int searchNormals(const sglVec3f &normal);

   /** find a tristripper in the current list using the given material,
       if not present, allocate a new sglTriStripper and assign this material
    */
   sglObjMtlGroup *findStripper(sglObjMtl *mtl);

   /// used in conjunction with normal calculation by the tri-stripper.
   static unsigned int addNormal(void* normlist, sglVec3f& normal);

private:
   // Remember we should define one default material for those things that do
   // not have an associated material.

   // material manager
   sglObjMtlMgr *m_mtl_mgr;

   // Pointers to the different material groups.
   vector<sglObjMtlGroup*> m_groups;

   // total number of vertices parsed.
   unsigned int  m_num_coords;
   // array of 3D vertices
   sglVec3f *m_coords;

   // total number of normals parsed.
   unsigned int  m_num_normals;
   // array of 3D normalize normals
   sglVec3f *m_normals;

   // total number of texture coordinates parsed.
   unsigned int m_num_texcoords;
   // array of 2D texture coordinates
   sglVec2f *m_texcoords;

   // an internal model to control behaviour
   unsigned int m_mode;
};

#endif
