/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglObjMtlMgr.cpp
 *  Created: 30 September 1997
 *  Summary: The material manager that maintains a list of mtllib files, and
 *           searches them in order when queried for named materials
 *****************************************************************************/

#include <sglobj.h>
#include <string>

#include <stdio.h>
#include <ctype.h>

#include <sglObjMtlMgr.hpp>
#include <sglObjMtl.hpp>
#include <sglLoadFile_obj.hpp>
#include <sglOStream.hpp>
#include <sglDataLoader.hpp>

//----------------------------------------------------------------------------
sglObjMtlMgr::sglObjMtlMgr()
{

}

//----------------------------------------------------------------------------
sglObjMtlMgr::~sglObjMtlMgr()
{
   while (m_mtl_list.size())
   {
      sglObjMtl *mtl = m_mtl_list.back();
      m_mtl_list.pop_back();
      delete mtl;
   }
   
}

//----------------------------------------------------------------------------
bool sglObjMtlMgr::addMtlLib(const char *filename)
{
   // determine if name is already in the list
   for (unsigned int i=0; i<m_mtllib_list.size(); i++)
   {
      // if name is in the list return without doing anything
      if (strcmp(filename, m_mtllib_list[i].c_str()) == 0)
      {
         return true;
      }
   }

   // if we get here the filename needs to be added to the list
   m_mtllib_list.push_back(string(filename));
   return true;
}

//----------------------------------------------------------------------------
sglObjMtl *sglObjMtlMgr::getMtl(const char *mtl_name)
{
   unsigned int i;

   // check to see if this mtl has already been loaded
   for (i=0; i<m_mtl_list.size(); i++)
   {
      if (strcmp(mtl_name, m_mtl_list[i]->getName()) == 0)
      {
         return m_mtl_list[i];
      }
   }

   // Not loaded previously... Now search the mtllibs in order
   sglObjMtl *mtl = NULL;
   for (i=0; i<m_mtllib_list.size() && mtl==NULL; i++)
   {
      mtl = parseMtlLib(m_mtllib_list[i], mtl_name);
   }

   // add the mtl to the list and return it (if found)
   if (mtl)
   {
      addMtl(mtl);
      return mtl;
   }

   sglPrintInfo() << "sglObjMtlMgr::getMtl warning: " << mtl_name
                  << " mtl not found.  Using default." << endl;

   // not found in any of the mtllibs so find the default mtl
   for (i=0; i<m_mtl_list.size(); i++)
   {
      if (strcmp("", m_mtl_list[i]->getName()) == 0)
      {
         return m_mtl_list[i];
      }
   }

   // default mtl not found so create one
   mtl = new sglObjMtl();
   if (mtl)
   {
      addMtl(mtl);
   }
   
   return mtl;
}

//----------------------------------------------------------------------------
bool sglObjMtlMgr::addMtl(sglObjMtl *mtl)
{
   if (mtl == NULL)
      return false;

   // check to see if this mtl is already in the list
   for (unsigned int i=0; i<m_mtl_list.size(); i++)
   {
      if (mtl == m_mtl_list[i])
      {
         return true;
      }
   }

   // if we get here the mtl needs to be added to the list
   m_mtl_list.push_back(mtl);
   return true;
}

//----------------------------------------------------------------------------
sglObjMtl *sglObjMtlMgr::parseMtlLib(const string &filename,
                                     const char *mtl_name)
{
   const int BUFFER_SIZE = 8192;

   FILE *mtl_fp;
   char  buffer[BUFFER_SIZE];
   char  token[BUFFER_SIZE];
   char *next;
   char *backslash;
   int   width = 0;
   float alpha;
   float shininess;
   float r,g,b;

   sglObjMtl *mtl = NULL;

   /* open file */
   if ((mtl_fp = fopen(filename.c_str(),"r")) == NULL)
   {
      sglPrintWarning() << "Warning: can't open mtl library "
                        << filename << endl;
      return NULL;
   }

   sglPrintDebug() << "Parsing the mtl file: " << filename << endl;

   int line_count = 0;

   /* read Wavefront ".mtl" file */
   while (fgets(buffer, BUFFER_SIZE, mtl_fp) != NULL)
   {
      line_count++;

      /* concatenate continuation lines */
      while ((backslash = strchr(buffer, '\\')) != NULL)
         if (fgets(backslash, BUFFER_SIZE - strlen(buffer), mtl_fp) == NULL)
            break;

      /* find first non-"space" character in line */
      for (next = buffer; *next != '\0' && isspace(*next); next++)
         /* EMPTY */ {};

      /* skip blank lines and comments ('$' is comment in "cow.obj") */
      if (*next == '\0' || *next == '#' || *next == '!' || *next == '$')
         continue;

      /* extract token */
      sscanf(next, "%s%n", token, &width);
      next += width;

      /* identify token */
      if (!sglDataLoader::strCaseCmp(token, "newmtl"))
      {
         // if the mtl has already been loaded and we just got a newmtl
         // token it is time to bail....THIS IS A SUCCESSFUL EXIT POINT
         if (mtl)
         {
            fclose(mtl_fp);
            return mtl;
         }

         char name[256];
         sscanf(next, "%s", name);

         // check for name of mtl requested if it matches allocate the
         // mtl.
         if (strcmp(mtl_name, name) == 0)
         {
            mtl = new sglObjMtl();
            mtl->setName(name);

            sglPrintDebug() << line_count << ": mtl " << name
                            << " found in " << filename << endl;
         }
      }
      else if (mtl)
      {
         if (!sglDataLoader::strCaseCmp(token, "Ka"))
         {
            sscanf(next, "%f %f %f", &r, &g, &b);
            (mtl->getMaterial())->setAmbient(r,g,b);

            sglPrintDebug() << line_count << ": Got ambient color" << endl;
         }
         else if (!sglDataLoader::strCaseCmp(token, "Kd"))
         {
            sscanf(next, "%f %f %f", &r, &g, &b);
            (mtl->getMaterial())->setDiffuse(r,g,b);

            sglPrintDebug() << line_count << ": Got diffuse color" << endl;
         }
         else if (!sglDataLoader::strCaseCmp(token, "Ks"))
         {
            sscanf(next, "%f %f %f", &r, &g, &b);
            (mtl->getMaterial())->setSpecular(r,g,b);

            sglPrintDebug() << line_count << ": Got specular color" << endl;
         }
         else if (!sglDataLoader::strCaseCmp(token, "Tr"))
         {
            sscanf(next, "%f", &alpha);
            (mtl->getMaterial())->setAlpha(1.0f - alpha);

            sglPrintDebug() << line_count << ": Got alpha" << endl;
         }
         else if (!sglDataLoader::strCaseCmp(token, "Ns"))
         {
            sscanf(next, "%f", &shininess);
            (mtl->getMaterial())->setShininess(shininess);

            sglPrintDebug() << line_count << ": Got Shininess" << endl;
         }
         else if (!sglDataLoader::strCaseCmp(token, "map_Kd"))
         {
            // textures
            string tex_filename;
            float su, sv;

            sglPrintDebug() << line_count
                            << ": Gotta texture specification" << endl;

            parseTexture(next, tex_filename, su, sv);
            string alt = sglDataLoader::extractDirectoryFromPath(filename);
            string foundname;
            if (sglDataLoader::findFile(tex_filename, &alt, foundname))
            {
               mtl->loadTexture(foundname, su, sv);
            }
         }
         else if (!sglDataLoader::strCaseCmp(token, "refl"))
         {
            // reflective textures
            string tex_filename;
            float su, sv;

            sglPrintDebug() << line_count
                            << ": Gotta reflection texture specification"
                            << endl;

            parseTexture(next, tex_filename, su, sv);
            string alt = sglDataLoader::extractDirectoryFromPath(filename);
            string foundname;
            if (sglDataLoader::findFile(tex_filename, &alt, foundname))
            {
               mtl->loadReflTexture(foundname, su, sv);
            }
         }
         else if (!sglDataLoader::strCaseCmp(token, "Ni")     ||
                  !sglDataLoader::strCaseCmp(token, "Tf")     ||
                  !sglDataLoader::strCaseCmp(token, "bump")   ||
                  !sglDataLoader::strCaseCmp(token, "d")      ||
                  !sglDataLoader::strCaseCmp(token, "decal")  ||
                  !sglDataLoader::strCaseCmp(token, "illum")  ||
                  !sglDataLoader::strCaseCmp(token, "map_Ka") ||
                  !sglDataLoader::strCaseCmp(token, "map_Ks") ||
                  !sglDataLoader::strCaseCmp(token, "map_Ns") ||
                  !sglDataLoader::strCaseCmp(token, "map_d")  ||
                  !sglDataLoader::strCaseCmp(token, "vp")     ||
                  !sglDataLoader::strCaseCmp(token, "sharpness"))
         {
            // skipped - do nothing
         }
         else
         {
            string tmp = ": unrecognized: ";
            tmp += buffer;
            sglPrintInfo() << line_count << tmp << endl;
         }
      }
   }

   /* close Wavefront ".mtl" file */
   fclose(mtl_fp);

   // if it was the last newmtl entry this closes it out properly
   return mtl;
}


//----------------------------------------------------------------------------
void sglObjMtlMgr::parseTexture(char *next, string &filename,
                                float &su, float &sv)
{
   int   width;
   char buffer[256];
   char reflect[80] = "none";
   float dummy;

   /* set default texture scale factors */
   su = 1.0f;
   sv = 1.0f;

   /* parse texture name */
   sscanf(next, "%s%n", buffer, &width);

   /* parse texture option strings */
   for (;;)
   {
      next += width;

      if (strcmp(buffer, "-s") == 0)
         sscanf(next, "%f %f %*f %s%n",
                &su, &sv, buffer, &width);
      else if (strcmp(buffer, "-mm") == 0)
         sscanf(next, "%f %f %s%n",
                &dummy, &dummy, buffer, &width);
      else if (strcmp(buffer, "-t") == 0)
         sscanf(next, "%f %f %*f %s%n",
                &dummy, &dummy, buffer, &width);
      else if (strcmp(buffer, "-type") == 0)
         sscanf(next, "%s %s%n",
                reflect, buffer, &width);
      else
         break;
   }

   filename = buffer;

   sglPrintDebug() << "\tsglObjModel::parseTexture()" << endl;
   sglPrintDebug() << "\tTexture name: " << filename << endl;
   sglPrintDebug() << "\tsu=" << su << ", sv=" << sv
                   << ", reflection=" << reflect << endl;
}
