/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglLoadImage_png.cpp
 *   Author: Tom Stimson
 *  Created: February 2002
 *  Summary: read png format image files using external library
 *****************************************************************************/

#include <sglLoadImage_png.hpp>
#include <sglOStream.hpp>
#include <sglImageLoader.hpp>
#include <stdio.h>

#include <png.h>
#include <setjmp.h>

/*****************************************************************************/
void *loadImagePNG(const string &filepath,
                    unsigned int &width,
                    unsigned int &height,
                    sglTexture::ExternalFormatEnum &format,
                    sglTexture::TypeEnum &type)
{
//   sglPrint() << "loadImagePNG()" << endl;

   FILE *fp;
   if ((fp = fopen(filepath.c_str(), "rb")) == NULL)
   {
      sglPrintWarning() << "loadImagePNG(): Error opening file "
                        << filepath << endl;
      return NULL;
   }

   png_uint_32 png_width, png_height;
   int bit_depth, color_type, interlace_type;

   png_structp png_ptr = png_create_read_struct(
      PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);

   if (png_ptr == NULL)
   {
      sglPrintWarning() << "Error creating png read struct" << endl;
      fclose(fp);
      return NULL;
   }

   png_infop info_ptr = png_create_info_struct(png_ptr);
   if (info_ptr == NULL)
   {
      sglPrintWarning() << "Error creating png info struct" << endl;
      png_destroy_read_struct(&png_ptr, (png_infopp)NULL, (png_infopp)NULL);
      fclose(fp);
      return NULL;
   }

   if (setjmp(png_ptr->jmpbuf))
   {
      /* Free all of the memory associated with the png_ptr and info_ptr */
      png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp)NULL);
      fclose(fp);

      sglPrintWarning() << "Error reading file "
                        << filepath << "; " << endl;

      /* If we get here, we had a problem reading the file */
      return NULL;
   }

   /* Set up the input control if you are using standard C streams */
   png_init_io(png_ptr, fp);

   /* The call to png_read_info() gives us all of the information from the
    * PNG file before the first IDAT (image data chunk).  REQUIRED
    */
   png_read_info(png_ptr, info_ptr);

   png_get_IHDR(png_ptr, info_ptr, &png_width, &png_height, &bit_depth,
                &color_type, &interlace_type, NULL, NULL);
   width = png_width;
   height = png_height;

   /* Extract multiple pixels with bit depths of 1, 2, and 4 from a single
    * byte into separate bytes (useful for paletted and grayscale images).
    */
   png_set_packing(png_ptr);

   /* Expand paletted colors into true RGB triplets */
   if (color_type == PNG_COLOR_TYPE_PALETTE)
      png_set_expand(png_ptr);

   /* Expand grayscale images to the full 8 bits from 1, 2, or 4 bits/pixel */
   if (color_type == PNG_COLOR_TYPE_GRAY && bit_depth < 8)
      png_set_expand(png_ptr);

   /* Expand paletted or RGB images with transparency to full alpha channels
    * so the data will be available as RGBA quartets.
    */
   if (png_get_valid(png_ptr, info_ptr, PNG_INFO_tRNS))
      png_set_expand(png_ptr);

   /* Tell libpng to handle the gamma conversion for you.  The final call
    * is a good guess for PC generated images, but it should be configurable
    * by the user at run time by the user.  It is strongly suggested that
    * your application support gamma correction.
    */
   double screen_gamma = 2.0;
   int intent;
   if (png_get_sRGB(png_ptr, info_ptr, &intent))
      png_set_gamma(png_ptr, screen_gamma, 0.45455);
   else
   {
      double image_gamma;
      if (png_get_gAMA(png_ptr, info_ptr, &image_gamma))
         png_set_gamma(png_ptr, screen_gamma, image_gamma);
      else
         png_set_gamma(png_ptr, screen_gamma, 0.45455);
   }

   /* Optional call to gamma correct and add the background to the palette
    * and update info structure.  REQUIRED if you are expecting libpng to
    * update the palette for you (ie you selected such a transform above).
    */
   png_read_update_info(png_ptr, info_ptr);

   /* Allocate the memory to hold the image using the fields of info_ptr. */
   unsigned int size = width * height;

   if (info_ptr->bit_depth == 8)
   {
      type = sglTexture::eUNSIGNED_BYTE;
   }
   else if (info_ptr->bit_depth == 16)
   {
      type = sglTexture::eUNSIGNED_SHORT;
      size *= sizeof(unsigned short);
   }
   else
   {
      sglPrintWarning() << "Invalid bit depth (" << info_ptr->bit_depth
                        << ") in file: " << filepath << endl;
      png_destroy_read_struct(&png_ptr, (png_infopp)NULL, (png_infopp)NULL);
      fclose(fp);
      return NULL;
   }

   switch (info_ptr->color_type)
   {
      case PNG_COLOR_TYPE_GRAY:
         format = sglTexture::eLUMINANCE;
         break;
      case PNG_COLOR_TYPE_GRAY_ALPHA:
         format = sglTexture::eLUMINANCE_ALPHA;
         size *= 2;
         break;
      case PNG_COLOR_TYPE_RGB:
         format = sglTexture::eRGB;
         size *= 3;
         break;
      case PNG_COLOR_TYPE_RGB_ALPHA:
         format = sglTexture::eRGBA;
         size *= 4;
         break;
      default:
         sglPrintWarning() << "Unknown color type (" << info_ptr->color_type
                           << ") in file: " << filepath << endl;
         png_destroy_read_struct(&png_ptr, (png_infopp)NULL, (png_infopp)NULL);
         fclose(fp);
         return NULL;
   }

   if (height*info_ptr->rowbytes != size)
   {
      sglPrintWarning() << "Invalid rowbytes (" << info_ptr->rowbytes
                        << ") in file: " << filepath << endl;
      png_destroy_read_struct(&png_ptr, (png_infopp)NULL, (png_infopp)NULL);
      fclose(fp);
      return NULL;
   }

   png_bytep image = (png_bytep)malloc(size);

   png_bytep *row_ptrs = (png_bytep*)malloc(height*sizeof(png_bytep));
   for (unsigned int i=0; i<height; ++i)
   {
      row_ptrs[i] = &image[(height-i-1)*info_ptr->rowbytes];
   }

   /* read the image */
   png_read_image(png_ptr, row_ptrs);

   free(row_ptrs);

   /* clean up after the read, and free any memory allocated - REQUIRED */
   png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp)NULL);

   /* close the file */
   fclose(fp);

   return image;
}


/*****************************************************************************/
bool writeImagePNG(const string &filepath,
                    const void *image,
                    unsigned int width,
                    unsigned int height,
                    sglTexture::ExternalFormatEnum format,
                    sglTexture::TypeEnum type)
{
//   sglPrint() << "writeImagePNG()" << endl;

   if (image == NULL)
      return false;

   FILE *fp;
   if ((fp = fopen(filepath.c_str(), "wb")) == NULL)
   {
      sglPrintWarning() << "writeImagePNG(): Error opening file "
                        << filepath << endl;
      return false;
   }

   /* Create and initialize the png_struct with the desired error handler
    * functions.  If you want to use the default stderr and longjump method,
    * you can supply NULL for the last three parameters.  We also check that
    * the library version is compatible with the one used at compile time,
    * in case we are using dynamically linked libraries.  REQUIRED.
    */
   png_structp png_ptr = png_create_write_struct(
      PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);

   if (png_ptr == NULL)
   {
      sglPrintWarning() << "Error creating png write struct" << endl;
      fclose(fp);
      return false;
   }

   /* Allocate/initialize the image information data.  REQUIRED */
   png_infop info_ptr = png_create_info_struct(png_ptr);
   if (info_ptr == NULL)
   {
      sglPrintWarning() << "Error creating png info struct" << endl;
      png_destroy_write_struct(&png_ptr,  (png_infopp)NULL);
      fclose(fp);
      return false;
   }

   void *tmp_image = NULL;

   /* Set error handling.  REQUIRED if you aren't supplying your own
    * error handling functions in the png_create_write_struct() call.
    */
   if (setjmp(png_ptr->jmpbuf))
   {
      png_destroy_write_struct(&png_ptr, &info_ptr);
      fclose(fp);
      if (tmp_image) free(tmp_image);

      sglPrintWarning() << "Error writing file "
                        << filepath << "; " << endl;

      return false;
   }

   /* set up the output control if you are using standard C streams */
   png_init_io(png_ptr, fp);


   sglTexture::TypeEnum new_type = sglTexture::eUNSIGNED_SHORT;
   unsigned int bit_depth = 16;
   switch (type)
   {
      case sglTexture::eBYTE:
      case sglTexture::eUNSIGNED_BYTE:
      case sglTexture::eBITMAP:
         new_type = sglTexture::eUNSIGNED_BYTE;
         bit_depth = 8;
         break;

      default:
         break;
   }

   if (type != new_type)
   {
      tmp_image = sglImageLoader::changeImageType(
         format, type, width, height, new_type, image);
      type = new_type;
   }

   unsigned int bits_per_texel = bit_depth;
   sglTexture::ExternalFormatEnum new_format;
   unsigned int color_type = PNG_COLOR_TYPE_RGB;
   switch (format)
   {
      case sglTexture::eLUMINANCE:
         new_format = sglTexture::eLUMINANCE;
         color_type = PNG_COLOR_TYPE_GRAY;
         break;
      case sglTexture::eALPHA:
      case sglTexture::eLUMINANCE_ALPHA:
         new_format = sglTexture::eLUMINANCE_ALPHA;
         color_type = PNG_COLOR_TYPE_GRAY_ALPHA;
         bits_per_texel *= 2;
         break;
      case sglTexture::eRGBA:
         new_format = sglTexture::eRGBA;
         color_type = PNG_COLOR_TYPE_RGB_ALPHA;
         bits_per_texel *= 4;
         break;
      default:
         new_format = sglTexture::eRGB;
         bits_per_texel *= 3;
         break;
   }

   if (new_format != format)
   {
      void *tmp_image2 = sglImageLoader::reformatImage(
         format, type, width, height, new_format,
         (tmp_image ? (const void *)tmp_image : image));
      if (tmp_image)
         free(tmp_image);
      tmp_image = tmp_image2;
      format = new_format;
   }

//   sglPrint() << "1" << endl;

   /* Set the image information here.  Width and height are up to 2^31,
    * bit_depth is one of 1, 2, 4, 8, or 16, but valid values also depend on
    * the color_type selected. color_type is one of PNG_COLOR_TYPE_GRAY,
    * PNG_COLOR_TYPE_GRAY_ALPHA, PNG_COLOR_TYPE_PALETTE, PNG_COLOR_TYPE_RGB,
    * or PNG_COLOR_TYPE_RGB_ALPHA.  interlace is either PNG_INTERLACE_NONE or
    * PNG_INTERLACE_ADAM7, and the compression_type and filter_type MUST
    * currently be PNG_COMPRESSION_TYPE_BASE and PNG_FILTER_TYPE_BASE. REQUIRED
    */
   png_set_IHDR(png_ptr, info_ptr, width, height, bit_depth, color_type,
                PNG_INTERLACE_NONE, PNG_COMPRESSION_TYPE_BASE,
                PNG_FILTER_TYPE_BASE);

   /* Write the file header information.  REQUIRED */
   png_write_info(png_ptr, info_ptr);

   unsigned int row_bytes = (width * bits_per_texel) / 8;
   png_byte *img_ptr =
      (tmp_image == NULL ? (png_byte*)image : (png_byte*)tmp_image);

   png_bytep *row_ptrs = (png_bytep*)malloc(height*sizeof(png_bytep));
   for (unsigned int i=0; i<height; ++i)
   {
      row_ptrs[i] = &img_ptr[(height-i-1)*row_bytes];
   }

   png_write_image(png_ptr, row_ptrs);

   /* It is REQUIRED to call this to finish writing the rest of the file */
   png_write_end(png_ptr, info_ptr);

   free(row_ptrs);

   if (tmp_image)
      free(tmp_image);

   /* clean up after the write, and free any memory allocated */
   png_destroy_write_struct(&png_ptr, &info_ptr);

   fclose(fp);

   return true;
}
