/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglLoadImage_tiff.cpp
 *   Author: Michael M. Morrison
 *  Created: September 2001
 *  Summary: read TIFF format image files using external library
 *****************************************************************************/

#include <tiffio.h>

#include <sglLoadImage_tiff.hpp>
#include <sglOStream.hpp>
#include <stdio.h>

/*****************************************************************************/
void *loadImageTIF(const string &filepath,
                   unsigned int &width,
                   unsigned int &height,
                   sglTexture::ExternalFormatEnum &format,
                   sglTexture::TypeEnum &type)
{
   return loadImageTIFF(filepath, width, height, format, type);
}

/*****************************************************************************/
void *loadImageTIFF(const string &filepath,
                    unsigned int &width,
                    unsigned int &height,
                    sglTexture::ExternalFormatEnum &format,
                    sglTexture::TypeEnum &type)
{
   TIFF *file;
   uint32 w, h;
   uint16 spp;
   void * image;
   uint32 * tmpImage;
   unsigned char * imgPtr;

   if(!( file = TIFFOpen( filepath.c_str(), "r" ) )) 
   {
      sglPrintWarning() << "loadImageTIFF(): Error opening file "
                        << filepath << endl;
      return NULL;
   }

   TIFFGetField( file, TIFFTAG_IMAGEWIDTH, &w );
   TIFFGetField( file, TIFFTAG_IMAGELENGTH, &h );
   TIFFGetField( file, TIFFTAG_SAMPLESPERPIXEL, &spp );

   // check spp (since it is appropriately converted)
   if( spp <= 3 )
      spp = 3;
   else
      spp = 4;

   // always 4 bytes per pixel for this routine
   if(!( tmpImage = (uint32 *)_TIFFmalloc((tsize_t)(w * h * sizeof(uint32)))))
   {
      sglPrintWarning() << "loadImageTIFF(): Unable to _TIFFmalloc: "
                        << filepath << endl;
      return NULL;
   }

   // read image with the library
   if(!( TIFFReadRGBAImage( file, w, h, tmpImage, 0 ) )) 
   {
      sglPrintWarning() << "loadImageTIFF(): Error reading file "
                        << filepath << endl;
      return NULL;
   }

   // convert component format
   type = sglTexture::eUNSIGNED_BYTE;

   if(!( image = malloc( w * h * spp ) ))
   {
      sglPrintWarning() << "loadImageTIFF(): Unable to malloc for: "
                        << filepath << endl;
      _TIFFfree( tmpImage );
      return NULL;
   }

   imgPtr = (unsigned char *)image;

   for(unsigned int i=0;i<h;i++)
   {
      for(unsigned int j=0;j<w;j++)
      {
         uint32 pixel = tmpImage[ (i*w) + j ];

         *imgPtr++ = (unsigned char)TIFFGetR( pixel );
         *imgPtr++ = (unsigned char)TIFFGetG( pixel );
         *imgPtr++ = (unsigned char)TIFFGetB( pixel );
         if( spp == 4 )
            *imgPtr++ = (unsigned char)TIFFGetA( pixel );
      }
   }

   _TIFFfree( tmpImage );
   TIFFClose( file );

   width = (unsigned int)w;
   height = (unsigned int)h;

   if( spp == 3 )
      format = sglTexture::eRGB;
   else
      format = sglTexture::eRGBA;

   return image;
}
