/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglTrPageArchive.hpp
 *   Author: Bryan Walsh
 *  Created: 5 October 2001
 *  Summary: Implementation of sglTrPageArchive object.  This class parses a
 *           TerraPage tile into SGL's scene graph structure.
 *
 * This loader is based on/modified from Terrain Experts Performer Loader,
 * and was ported to SGL by Bryan Walsh / bryanw@earthlink.net
 *
 * That loader is redistributed under the terms listed on Terrain Experts
 * website (www.terrex.com/www/pages/technology/technologypage.htm)
 *
 * "TerraPage is provided as an Open Source format for use by anyone...
 * We supply the TerraPage C++ source code free of charge.  Anyone
 * can use it and redistribute it as needed (including our competitors).
 * We do, however, ask that you keep the TERREX copyrights intact."
 *
 * Copyright Terrain Experts Inc. 1999.
 * All Rights Reserved.
 *
 *****************************************************************************/

#include <sglTrPageArchive.hpp>
#include <sglTrPageParser.hpp>

#include <sgl.h>
#include <sglGroup.hpp>
#include <sglTexture2D.hpp>
#include <sglMaterial.hpp>
#include <sglTexEnv.hpp>
#include <sglStatelet.hpp>
#include <sglOStream.hpp>

#include <sglImageLoader.hpp>
#include <sglModelLoader.hpp>

#include <trpage_geom.h>
#include <trpage_read.h>
#include <trpage_write.h>
#include <trpage_scene.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

//----------------------------------------------------------------------------
sglTrPageArchive::sglTrPageArchive()
      : trpgr_Archive()

{
   // Set up the callbacks for parsing tiles
   parse = new sglTrPageParser();
}

//----------------------------------------------------------------------------
sglTrPageArchive::~sglTrPageArchive()
{
   // Note: Delete stuff
   delete parse;
   void *image = NULL;
   sglTexture::ExternalFormatEnum format;
   sglTexture::TypeEnum type;
   unsigned int num_s, num_t;
   bool border;
   for (vector<sglTexture2D*>::iterator ii = m_textures.begin();
        ii != m_textures.end(); ++ii)
   {
      if (*ii)
      {
         (*ii)->getImage(image, format, type, num_s, num_t, border);
         if (image)
            free(image);
      }
   }
   m_textures.clear();

}

//----------------------------------------------------------------------------
// Open a TerraPage archive for read into Performer. Read the header as well.
//
bool sglTrPageArchive::OpenFile(const string &file)
{
   m_alternate_path = sglDataLoader::extractDirectoryFromPath(file);
   string name = sglDataLoader::extractFilenameFromPath(file);

   SetDirectory(m_alternate_path.c_str());
   if (!trpgr_Archive::OpenFile(name.c_str()))
   {
      sglPrintWarning() << "sglTrPageArchive::OpenFile() error: "
                        << "couldn't open archive: " << file << endl;
      return false;
   }
   if (!ReadHeader())
   {
      sglPrintWarning() << "sglTrPageArchive::OpenFile() error: "
                        << "couldn't read header for archive: " << file
                        << endl;
      return false;
   }

   // Set the max group ID here because it won't change
   const trpgHeader *head = GetHeader();
   int maxID;
   head->GetMaxGroupID(maxID);
   parse->SetMaxGroupID(maxID);

   return true;
}

//----------------------------------------------------------------------------
// Create texture and state structures for the materials.
//
void sglTrPageArchive::LoadMaterials()
{
   int i;

   // Read the textures
   int numTex;
   texTable.GetNumTextures(numTex);

   for (i=0;i<numTex;i++)
   {
      trpgTexture *tex;
      tex = texTable.GetTextureRef(i);
      char texName[1024];  texName[0] = 0;
      tex->GetName(texName,1023);
      // Create a texture by name.
      sglTexture2D *pfTex = new sglTexture2D();
      // Load Texture and Create Texture State
      sglTexture::ExternalFormatEnum format;
      sglTexture::TypeEnum type;
      unsigned int width, height;
      void *image;
      string filename = sglDataLoader::extractFilenameFromPath(texName);
      string path(getDir());
      string theFile = path + '/'+ filename ;
      image = sglImageLoader::loadImage(theFile,width,height,format,type);

      if (image)
      {
         pfTex->setImage(image, format,type,width, height, false);
         pfTex->setWrapMode(sglTexture2D::eREPEAT, sglTexture2D::eREPEAT);
         m_textures.push_back(pfTex);
      }
   }

   sglPrintInfo() << "sglTrPageArchive::LoadMaterials(): loaded "
                  << numTex << " textures." << endl;

   // Create the materials
   int numMat;
   tmpGeoState pfGeo;
   materialTable.GetNumMaterial(numMat);
   for (i=0;i<numMat;i++)
   {
      trpgMaterial *mat;
      mat = materialTable.GetMaterialRef(0,i);
      // Set texture
      int numMatTex;
      mat->GetNumTexture(numMatTex);
      if (numMatTex)
      {
         // Note: Only dealing with the first texture right now
         int texId;
         trpgTextureEnv texEnv;
         mat->GetTexture(0,texId,texEnv);

         // Set up texture environment
         sglTexEnv *perfTexEnv;
         perfTexEnv = new sglTexEnv();
         perfTexEnv->setMode(sglTexEnv::eMODULATE);

         // Create material
         float64 alpha;
         mat->GetAlpha(alpha);

         sglMaterial *perfMtl = new sglMaterial();
         trpgColor color;
         mat->GetAmbient(color);
         perfMtl->setAmbient((float)color.red,
                             (float)color.green,
                             (float)color.blue);
         mat->GetDiffuse(color);
         perfMtl->setDiffuse((float)color.red,
                             (float)color.green,
                             (float)color.blue);
         mat->GetSpecular(color);
         perfMtl->setSpecular((float)color.red,
                              (float)color.green,
                              (float)color.blue);
         mat->GetEmission(color);
         perfMtl->setEmission((float)color.red,
                              (float)color.green,
                              (float)color.blue);
         float64 dbl;
         mat->GetShininess(dbl);
         perfMtl->setShininess((float)dbl);
         mat->GetAlpha(dbl);
         perfMtl->setAlpha((float)dbl);
         pfGeo.material = perfMtl;

         //pfLightModel *twoSideLM = new pfLightModel();
         //twoSideLM->setTwoSide(1);

         // Create GeoState
         int alphaFunc;
         mat->GetAlphaFunc(alphaFunc);

         pfGeo.tex = m_textures[texId];

         pfGeo.transparency = new sglTransparency(true);
         sglTexture::TypeEnum type;
         sglTexture::ExternalFormatEnum format;
         unsigned int width, height;
         void *image = NULL;
         bool border;
         pfGeo.tex->getImage(image, format, type, width, height, border);
         if (sglTexture::getNumComponents(format) != 4)
         {
            pfGeo.transparency->enable(false);
         }

         pfGeo.texEnv = perfTexEnv;
         pfGeo.lighting = new sglLighting(true);
         //pfGeo->setMode(PFSTATE_ENTEXTURE,PF_ON);
         //pfGeo->setMode(PFSTATE_ENLIGHTING,PF_ON);
         //pfGeo->setAttr(PFSTATE_TEXENV,perfTexEnv);
         //pfGeo->setAttr(PFSTATE_FRONTMTL,perfMtl);
         //pfGeo->setAttr(PFSTATE_BACKMTL,perfMtl);

         // Culling state is per geostate
         int cullMode;
         mat->GetCullMode(cullMode);
         switch (cullMode)
         {
            case trpgMaterial::Front:
               pfGeo.cullMode = new sglCullMode(sglCullMode::eBACK);
               break;
            case trpgMaterial::Back:
               pfGeo.cullMode = new sglCullMode(sglCullMode::eFRONT);
               break;
            case trpgMaterial::FrontAndBack:
               pfGeo.cullMode = new sglCullMode(sglCullMode::eOFF);
               break;
         }
      }
      m_gstates.push_back(pfGeo);
   }

   sglPrintInfo() << "sglTrPageArchive::LoadMaterials(): loaded " << numMat
                  << " materials." << endl;
}

//----------------------------------------------------------------------------
//
// Load all the models and keep them around.
//
bool sglTrPageArchive::LoadModels()
{
   int i;
   int numModel;
   modelTable.GetNumModels(numModel);

   // Iterate over the models
   for (i=0; i<numModel; i++)
   {
      trpgModel *mod = modelTable.GetModelRef(i);
      int type;
      mod->GetType(type);

      // Only dealing with external models currently
      if (type == trpgModel::External)
      {
         char name[1024];
         mod->GetName(name,1023);

         // Load the model.  It's probably not TerraPage
         // pfNode *pfMod = pfdLoadFile(name);
         string found_name;
         sglNode *pfMod = NULL;

         if (sglDataLoader::findFile(name, &m_alternate_path, found_name))
         {
            pfMod = sglModelLoader::loadModel(found_name, 0);
         }
         else
         {
            sglPrintWarning() << "sglTrPageArchive::LoadModels() error: "
                              << "couldn't find file: "
                              << name << ", alternate path = "
                              << m_alternate_path << endl;
         }

         // Do this even if it's NULL
         m_models.push_back(pfMod);
         if (!pfMod)
         {
            sglPrintWarning() << "sglTrPageArchive::LoadModels() error: "
                              << "failed to load model: "
                              << found_name << endl;
         }
      }
   }
   return true;
}

//----------------------------------------------------------------------------
sglGroup *sglTrPageArchive::LoadTile(int x,int y,int lod,int &parentID)
{
   // Set up the memory buffer
   // Note: Move this into sglTrPageArchive
   trpgMemReadBuffer buf(GetEndian());

   // Read the tile data in, but don't parse it
   if (!ReadTile(x,y,lod,buf))
      return NULL;

   // Call the parser
   sglGroup *tile = parse->ParseScene(buf,m_gstates,m_models);
   if (tile)
   {
      parentID = parse->GetParentID();
      // Note: This is where you would page in textures and models
   }

   // That's it
   return tile;
}

//----------------------------------------------------------------------------
sglGroup *sglTrPageArchive::LoadAllTiles()
{
   // Size information comes out of the header
   const trpgHeader *head = GetHeader();

   // Create one group as the top
   sglGroup *topGroup = new sglGroup();

   int32 numLod;
   head->GetNumLods(numLod);
   // Iterate over the LODs.  Lower res LODs must be loaded
   //  first, otherwise there's nothing to hook the higher res
   //  LODs into.
   trpg2iPoint tileSize;

   // The group list is used to map parent IDs to pfGroup nodes
   vector<sglGroup *> *groupList = parse->GetGroupList();

   for (int nl=0;nl<numLod;nl++)
   {
      // The number of tiles in X and Y changes per LOD
      head->GetLodSize(nl,tileSize);
      for (int x=0; x < tileSize.x; x++)
      {
         for (int y=0; y < tileSize.y; y++)
         {
            int parentID;
            sglGroup *tile = LoadTile(x,y,nl,parentID);
            if (!tile)
            {
               sglPrintWarning() << "sglTrPageArchive::LoadAllTiles error: "
                                 << "failed to load tile ("
                                 << x << "," << y << "," << nl << ")"
                                 << endl;
            }
            else
            {
               if (parentID == -1)
               {
                  // Get added to the top level node
                  topGroup->addChild(tile);
               }
               else
               {
                  // Added below some other node
                  (*groupList)[parentID]->addChild(tile);
               }
            }
         }
      }
   }
   return topGroup;
}
