/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglTrPageParser.cpp
 *   Author: Bryan Walsh
 *  Created: 5 October 2001
 *  Summary: Implementation of sglTrPageParser object.
 *
 * This loader is based on/modified from Terrain Experts Performer Loader,
 * and was ported to SGL by Bryan Walsh / bryanw at earthlink dot net
 *
 * That loader is redistributed under the terms listed on Terrain Experts
 * website (www.terrex.com/www/pages/technology/technologypage.htm)
 *
 * "TerraPage is provided as an Open Source format for use by anyone...
 * We supply the TerraPage C++ source code free of charge.  Anyone
 * can use it and redistribute it as needed (including our competitors).
 * We do, however, ask that you keep the TERREX copyrights intact."
 *
 * Copyright Terrain Experts Inc. 1999.
 * All Rights Reserved.
 *
 *****************************************************************************/

#include <trpage_sys.h>
#include <sgl.h>
#include <sglGroup.hpp>
#include <sglMaterial.hpp>
#include <sglTexture2D.hpp>
#include <sglTexEnv.hpp>
#include <sglLOD.hpp>
#include <sglGeode.hpp>
#include <sglBillboard.hpp>
#include <sglMatrix.hpp>
#include <sglTransform.hpp>
#include <sglTriangleSet.hpp>
#include <sglTriangleFanSet.hpp>
#include <sglTriangleStripSet.hpp>

#include <sglTrPageParser.hpp>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>


//
// TerraPage Performer scene parser helpers
//

//----------------------------------------------------------------------------
//
// Geometry Reader Class
//
geomRead::geomRead(sglTrPageParser *in_parse)
{
   parse = in_parse;
   m_verts = NULL;
   m_texCoord  = NULL;
   m_norms = NULL;
   m_primLens = NULL;
}

//----------------------------------------------------------------------------
geomRead::~geomRead()
{
   for (vector<sglVec3f*>::iterator i3 = m_vec3_malloc_list.begin();
        i3 != m_vec3_malloc_list.end(); ++i3)
   {
      if (*i3) delete [] (*i3);
   }
   m_vec3_malloc_list.clear();

   for (vector<sglVec2f*>::iterator i2 = m_vec2_malloc_list.begin();
        i2 != m_vec2_malloc_list.end(); ++i2)
   {
      if (*i2) delete [] (*i2);
   }
   m_vec2_malloc_list.clear();

   for (vector<unsigned int*>::iterator iu = m_uint_malloc_list.begin();
        iu != m_uint_malloc_list.end(); ++iu)
   {
      if (*iu) delete [] (*iu);
   }
   m_uint_malloc_list.clear();
}

//----------------------------------------------------------------------------
void* geomRead::Parse(trpgToken /*tok*/,trpgReadBuffer &buf)
{
   trpgGeometry geom;
   int i;
   if (!geom.Read(buf)) return NULL;

   // Get the necessary info out of the geom
   trpgGeometry::PrimType primType;
   int numPrims;
   int numVert;
   int numNorm;
   int matId;
   geom.GetPrimType(primType);
   geom.GetNumPrims(numPrims);
   geom.GetNumVertex(numVert);
   numVert /= 3;
   geom.GetMaterial(0,matId);
   geom.GetNumNormal(numNorm);
   numNorm /= 3;
   m_verts = new sglVec3f[numVert];
   m_vec3_malloc_list.push_back(m_verts);
   // Get vertices
   geom.GetVertices((float32 *)m_verts);

   // Turn the trpgGeometry into something Performer can understand
   sglGeoSet *gset = NULL;

   // Get texture coordinates
   trpgTexData td;
   m_texCoord = NULL;
   if (geom.GetTexCoordSet(0,&td))
   {
      m_texCoord = new sglVec2f[numVert];
      m_vec2_malloc_list.push_back(m_texCoord);
      for (i=0;i<numVert;i++)
      {
         m_texCoord[i][0] = td.floatData[2*i+0];
         m_texCoord[i][1] = td.floatData[2*i+1];
      }
   }

   // Get normals
   // Note: only dealing with vertex normals
   m_norms = NULL;
   if (numNorm == numVert)
   {
      m_norms = new sglVec3f[numVert];
      m_vec3_malloc_list.push_back(m_norms);
      geom.GetNormals((float32 *)m_norms);
   }

   sglGeode *geode = new sglGeode();
   // Set up the primitive type
   switch (primType)
   {
      case trpgGeometry::Triangles:
      {
         gset = new sglTriangleSet();
      }
      break;
      case trpgGeometry::TriStrips:
      {
         // Need primitive lengths too
         m_primLens = new unsigned int[numPrims];
         m_uint_malloc_list.push_back(m_primLens);
         geom.GetPrimLengths((int*)m_primLens);

         // Define GeoSet
         gset = new sglTriangleStripSet();
         static_cast<sglTriangleStripSet*>(gset)->setPrimLengths(m_primLens);
      }
      break;
      case trpgGeometry::TriFans:
      {
         // Need primitive lengths too
         m_primLens = new unsigned int[numPrims];
         m_uint_malloc_list.push_back(m_primLens);

         geom.GetPrimLengths((int*)m_primLens);

         // Need to flip the fans
         int ind = 0;
         for (i=0;i<numPrims;i++)
         {
            int start=ind+1;
            int end=m_primLens[i]+ind-1;
            // Swap from start+1 to end
            sglVec3f swap,swapN;
            sglVec2f swapT;
            int numSwap = (end-start+1)/2;
            // Swap vertices, texture coords & normals
            for (int j=0;j<numSwap;j++)
            {
               swapT = m_texCoord[start];
               swap = m_verts[start];
               swapN = m_norms[start];
               m_verts[start] = m_verts[end];
               m_texCoord[start] = m_texCoord[end];
               m_norms[start] = m_norms[end];
               m_verts[end] = swap;
               m_texCoord[end] = swapT;
               m_norms[end] = swapN;
               start++;
               end--;
            }
            ind += m_primLens[i];
         }

         // Define GeoSet
         gset = new sglTriangleFanSet();
         static_cast<sglTriangleFanSet*>(gset)->setPrimLengths(m_primLens);
      }
      break;
      default:

         break;
   };


   // Add it to the current parent group
   sglGroup *top = parse->GetCurrTop();
   if (gset)
   {
      gset->setCoordList(m_verts);
      gset->setNumPrims(numPrims);
      if (m_norms)
         gset->setNormalList(sglGeoSet::ePER_VERTEX,m_norms);
      // Note: Should check number of materials first
      // Note: Should be combining multiple geosets
      tmpGeoState pfGeo = (*parse->GetMaterials())[matId];
      vector<sglStatelet*> statelets;

      statelets.push_back(pfGeo.cullMode);
      statelets.push_back(pfGeo.lighting);
      statelets.push_back(pfGeo.material);
      if (m_texCoord)
      {
         statelets.push_back(pfGeo.tex);
         statelets.push_back(pfGeo.texEnv);
         statelets.push_back(pfGeo.transparency);
         gset->setState(0, statelets, m_texCoord);
      }
      else
      {
         gset->setState(0, statelets);
      }
      geode->addGeometry(gset);
      top->addChild(geode);
   }
   return (void *) 1;
}

//
// Group Reader Class
//

//----------------------------------------------------------------------------
groupRead::groupRead(sglTrPageParser *in_parse)
{
   parse = in_parse;
}

//----------------------------------------------------------------------------
void* groupRead::Parse(trpgToken /*tok*/,trpgReadBuffer &buf)
{
   trpgGroup group;
   if (!group.Read(buf))
      return NULL;
   // Create a new Performer group
   sglGroup *perfGroup = new sglGroup();
   // Dump this group into the hierarchy
   sglGroup *top = parse->GetCurrTop();
   if (top)
      top->addChild(perfGroup);
   int32 id;
   group.GetID(id);
   parse->AddToGroupList(id,perfGroup);
   return (void *) perfGroup;
}

//----------------------------------------------------------------------------
//
// Attach Reader Class
//
attachRead::attachRead(sglTrPageParser *in_parse)
{
   parse = in_parse;
}

//----------------------------------------------------------------------------
void* attachRead::Parse(trpgToken /*tok*/,trpgReadBuffer &buf)
{
   trpgAttach group;
   if (!group.Read(buf))
      return NULL;
   // Create a new Performer group
   sglGroup *perfGroup = new sglGroup();
   // Dump this group into the hierarchy
   sglGroup *top = parse->GetCurrTop();
   if (top)
      top->addChild(perfGroup);
   int32 id;
   group.GetID(id);
   parse->AddToGroupList(id,perfGroup);
   // This sets the parent ID for the current tile too
   int32 parentID;
   group.GetParentID(parentID);
   parse->SetParentID(parentID);
   return (void *) perfGroup;
}

//----------------------------------------------------------------------------
//
// Billboard Reader Class
//
billboardRead::billboardRead(sglTrPageParser *in_parse)
{
   parse = in_parse;
}

//----------------------------------------------------------------------------
void* billboardRead::Parse(trpgToken /*tok*/,trpgReadBuffer &buf)
{
   trpgBillboard bill;
   if (!bill.Read(buf))
      return NULL;
   // Create a new Performer group
   sglGroup *perfGroup = new sglGroup();

   //sglBillboard *perfGroup = new sglBillboard();
   //perfGroup->setMode(sglBillboard::ePOINT_ROT_EYE);
   //sglVec3f axis(0.f, 0.f, 1.f);
   //perfGroup->setAxis(axis);
   //perfGroup->setMode(sglBillboard::eAXIAL_ROT);

   // Dump this group into the hierarchy
   sglGroup *top = parse->GetCurrTop();
   if (top)
      top->addChild(perfGroup);
   int32 id;
   bill.GetID(id);
   parse->AddToGroupList(id,perfGroup);
   return (void *) perfGroup;
}

//----------------------------------------------------------------------------
//
// LOD Reader Class
//
lodRead::lodRead (sglTrPageParser *in_parse)
{
   parse = in_parse;
}

//----------------------------------------------------------------------------
void* lodRead::Parse(trpgToken /*tok*/,trpgReadBuffer &buf)
{
   trpgLod lod;
   if (!lod.Read(buf))
      return NULL;
   // Pull out the LOD data we'll need
   trpg3dPoint center;
   lod.GetCenter(center);
   double in,out,width;
   lod.GetLOD(in,out,width);
   double minRange = MIN(in,out);
   double maxRange = MAX(in,out);

   // Create a new Performer LOD
   sglLOD *perfLod = new sglLOD();
   sglVec3f perfCenter;
   perfCenter[0] = center.x;  perfCenter[1] = center.y;  perfCenter[2] = center.z;
   perfLod->setCenter(perfCenter);
   perfLod->setRange(0,minRange);
   perfLod->setRange(1,maxRange);
// fprintf(stderr,"lodRead: center = (%f,%f,%f); minRange = %f;  maxRange = %f\n",
//    perfCenter[0],perfCenter[1],perfCenter[2],minRange,maxRange);
   // Note: Do something with "width"

   // Our LODs are binary so we need to add a group under this LOD and attach stuff
   //  to that instead of the LOD
   sglGroup *perfLodG = new sglGroup();
   perfLod->addChild(perfLodG);

   // Dump this group into the hierarchy
   sglGroup *top = parse->GetCurrTop();
   if (top)
      top->addChild(perfLod);
   int32 id;
   lod.GetID(id);
   // Add the sub-group to the group list, not the LOD
   parse->AddToGroupList(id,perfLodG);
   return (void *) perfLodG;
}

//----------------------------------------------------------------------------
//
// Model Reference Reader Class
//
modelRefRead::modelRefRead(sglTrPageParser *in_parse)
{
   parse = in_parse;
}

//----------------------------------------------------------------------------
void *modelRefRead::Parse(trpgToken /*tok*/,trpgReadBuffer &buf)
{
   trpgModelRef model;
   if (!model.Read(buf))
         return NULL;
   // Get the matrix and pfNode for the model
   int modelID;
   model.GetModel(modelID);
   float64 mat[16];
   model.GetMatrix(mat);
   sglMat4f perfMat(
      (float)mat[0], (float)mat[1], (float)mat[2], (float)mat[3],
      (float)mat[4], (float)mat[5], (float)mat[6], (float)mat[7],
      (float)mat[8], (float)mat[9], (float)mat[10],(float)mat[11],
      (float)mat[12],(float)mat[13],(float)mat[14],(float)mat[15]
   );

   // Note: Array check before you do this
   sglNode *perfModel = (*parse->GetModels())[modelID];
   // Create the SCS and position the model
   if (perfModel) {
      sglTransformf *scs = new sglTransformf();
      scs->setMatrix(perfMat);
      scs->addChild(perfModel);
      // Add the SCS to the hierarchy
      sglGroup *top = parse->GetCurrTop();
      if (top)
         top->addChild(scs);
   }
   return (void *) 1;
}

//----------------------------------------------------------------------------
//
// Tile Header Reader Class
//
tileHeaderRead::tileHeaderRead(sglTrPageParser *in_parse)
{
   parse = in_parse;
}

//----------------------------------------------------------------------------
void* tileHeaderRead::Parse(trpgToken /*tok*/,trpgReadBuffer &buf)
{
   trpgTileHeader *tileHead = parse->GetTileHeaderRef();
   if (!tileHead->Read(buf))
      return NULL;
   return (void *) 1;
}


/* ********************************** */

//----------------------------------------------------------------------------
// Constructor for scene graph parser
sglTrPageParser::sglTrPageParser()
{
   currTop = NULL;
   top = NULL;

   // Register the readers
   AddCallback(TRPG_GEOMETRY,new geomRead(this));
   AddCallback(TRPG_GROUP,new groupRead(this));
   AddCallback(TRPG_ATTACH,new attachRead(this));
   AddCallback(TRPG_BILLBOARD,new billboardRead(this));
   AddCallback(TRPG_LOD,new lodRead(this));
   AddCallback(TRPG_MODELREF,new modelRefRead(this));
   AddCallback(TRPGTILEHEADER,new tileHeaderRead(this));
}

//----------------------------------------------------------------------------
// Destructor for scene graph parser
sglTrPageParser::~sglTrPageParser()
{
}

//----------------------------------------------------------------------------
// Return a reference to the Tile Header
// Doesn't do much if you haven't just read a tile
trpgTileHeader *sglTrPageParser::GetTileHeaderRef()
{
   return &tileHead;
}

//----------------------------------------------------------------------------
// Parse a buffer and return a (chunk of) Performer
//  scene graph.
sglGroup *sglTrPageParser::ParseScene(trpgReadBuffer &buf,vector<tmpGeoState> &in_mat,vector<sglNode *> &in_model)
{

   top = currTop = new sglGroup();
   materials = &in_mat;
   models = &in_model;
   parentID = -1;

   // All the setup is handled in the constructor.
   // Just parse and return the top
   if (!Parse(buf))
   {
      fprintf(stderr,"trpgFPParser::ParseScene failed to parse tile.\n");
      delete top;
      return NULL;
   }

   sglGroup *ret = top;
   top = currTop = NULL;
   return ret;
}

//----------------------------------------------------------------------------
// Start Children
// This is called when the parser hits a push.
// We'll want to make the node it's handing us the "top" node
bool sglTrPageParser::StartChildren(void *in_node)
{
   sglGroup *node = (sglGroup *)in_node;

   currTop = node;

   return true;
}

//----------------------------------------------------------------------------
// This is called when the parser hits a pop.
// We'll want to look on the parent list (in trpgSceneParser)
// for the parent above the current one.
// If there isn't one, we'll just stick things in our top group.
bool sglTrPageParser::EndChildren(void * /* in_node */)
{
   // Get the parent above the current one
   int pos = parents.size()-2;
   if (pos < 0)
   {
      // Nothing above the current one.  Fall back on our top group
      currTop = top;
   }
   else
   {
      currTop = (sglGroup *)parents[pos];
   }
   return true;
}

//----------------------------------------------------------------------------
// Return the current top node
sglGroup *sglTrPageParser::GetCurrTop()
{
   if (currTop)
   {
      return currTop;
   }
   else
   {
      return top;
   }
}

//----------------------------------------------------------------------------
// Add the given pfGroup to the group list at position ID
bool sglTrPageParser::AddToGroupList(int ID,sglGroup *group)
{
   // Note: check bounds
   groupList[ID] = group;

   return true;
}

//----------------------------------------------------------------------------
// Initialize the group list
void sglTrPageParser::SetMaxGroupID(int maxGroupID)
{
   fprintf(stderr,"trpgFPParser: max group ID = %d.\n",maxGroupID);
   // Initialize the group list with -1's
   groupList.resize(0);
   // Note: Fix this
   for (int i=0;i<maxGroupID;i++)
      groupList.push_back(NULL);
}
