/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sglu.cpp
 *  Created: 14 November 1998
 *  Summary: Miscellaneous utility functions
 *****************************************************************************/

#include <sglu.h>

#include <sglNode.hpp>
#include <sglGroup.hpp>
#include <sglGeode.hpp>
#include <sglGeoSet.hpp>
#include <sglTexture.hpp>

#include <string>

//----------------------------------------------------------------------------
void sgluOutputStates(sglDrawable *node,
                      ostream &ostrm,
                      unsigned int indent_spaces)
{
   string indent_string(indent_spaces, ' ');
   for (unsigned int i=0; i<node->getNumStates(); ++i)
   {
      ostrm << indent_string << "------------------------------------" << endl;
      ostrm << indent_string << "state #" << i << ": " << endl;
      node->getState(i)->printInfo(ostrm, indent_string.c_str());
   }
}

//----------------------------------------------------------------------------
void sgluOutputGeometry(sglDrawable *node,
                        ostream &ostrm,
                        unsigned int indent_spaces)
{
   string indent_string(indent_spaces, ' ');
   ostrm << indent_string << "------------------------------------" << endl;
   node->printInfo(ostrm, indent_string.c_str());
   sgluOutputStates(node, ostrm, indent_spaces+3);
   ostrm << indent_string << "------------------------------------" << endl;
}

//----------------------------------------------------------------------------
void sgluOutputTree(sglNode *root, ostream &ostrm, unsigned int indent_spaces)
{
   string indent_string(indent_spaces, ' ');
   ostrm << indent_string << "------------------------------------" << endl;
   root->printInfo(ostrm, indent_string.c_str());
   ostrm << indent_string << "------------------------------------" << endl;

   sglGroup *group_node = dynamic_cast<sglGroup*>(root);
   if (group_node)
   {
      ostrm << indent_string << "{" << endl;
      for (unsigned int i=0; i<group_node->getNumChildren(); i++)
      {
         sgluOutputTree(group_node->getChild(i), ostrm, indent_spaces+3);
      }
      ostrm << indent_string << "}" << endl;
   }
   else
   {
      sglGeode *geode = dynamic_cast<sglGeode*>(root);
      if (geode)
      {
         ostrm << indent_string << "{" << endl;
         for (unsigned int i=0; i<geode->getNumGeometries(); i++)
         {
            sgluOutputGeometry(geode->getGeometry(i), ostrm, indent_spaces+3);
         }
         ostrm << indent_string << "}" << endl;
      }
   }
}

//----------------------------------------------------------------------------
//    Summary: This really does not work when num parents > 0 (multiple paths)
void sgluOutputPath(sglNode *end_point,
                    ostream &ostrm,
                    unsigned int indent_spaces)
{
   if (end_point->getNumParents())
   {
      sglNode *parent = end_point->getParent(0);
      sgluOutputPath(parent, ostrm, indent_spaces);
   }

   string indent_string(indent_spaces, ' ');
   ostrm << indent_string << "------------------------------------" << endl;
   end_point->printInfo(ostrm, indent_string.c_str());
   ostrm << indent_string << "------------------------------------" << endl;
}


//----------------------------------------------------------------------------
void sgluBindTextures(sglNode *root)
{
   sglGroup *group_node = dynamic_cast<sglGroup*>(root);
   if (group_node)
   {
      for (unsigned int i=0; i<group_node->getNumChildren(); i++)
      {
         sgluBindTextures(group_node->getChild(i));
      }
   }
   else
   {
      sglGeode *geode = dynamic_cast<sglGeode*>(root);
      if (geode)
      {
         for (unsigned int i=0; i<geode->getNumGeometries(); i++)
         {
            sglDrawable *gset = geode->getGeometry(i);
            for (unsigned int k=0; k<gset->getNumStates(); k++)
            {
               const vector<sglStatelet*> &statelets =
                  gset->getState(k)->getStatelets();

               vector<sglStatelet*>::const_iterator end = statelets.end();
               for (vector<sglStatelet*>::const_iterator j = statelets.begin();
                    j != end; ++j)
               {
                  if ((*j)->getType() == sglStatelet::eTEXTURE)
                  {
                     sglTexture *tex = static_cast<sglTexture*>(*j);
                     if (tex->isValid() && !tex->isBound())
                     {
                        tex->bind();
                     }
                  }
               }
            }
         }
      }
   }
}

//----------------------------------------------------------------------------
sglObject *sgluFindObject(const string &name, sglObject *root)
{
   if (name == root->getName())
   {
       return root;
   }

   // Is it a group?
   sglGroup *group_node = dynamic_cast<sglGroup*>(root);
   if (group_node)
   {
       for (unsigned int i=0; i<group_node->getNumChildren(); i++)
       {
           return sgluFindObject(name, group_node->getChild(i));
       }
   }

   // Is it a geode?
   sglGeode *geode = dynamic_cast<sglGeode*>(root);
   if (geode)
   {
       for (unsigned int i=0; i<geode->getNumGeometries(); i++)
       {
           return sgluFindObject(name, geode->getGeometry(i));
       }
   }

   // Don't currently search cameras attached to view platforms

   // Is it a drawable?
   sglDrawable *gset = dynamic_cast<sglDrawable*>(root);
   if (gset)
   {
       for (unsigned int k=0; k<gset->getNumStates(); k++)
       {
           const vector<sglStatelet*> &statelets =
               gset->getState(k)->getStatelets();

           vector<sglStatelet*>::const_iterator end = statelets.end();
           for (vector<sglStatelet*>::const_iterator it = statelets.begin();
                it != end; ++it)
           {
               if (name == (*it)->getName())
               {
                   return (*it);
               }
           }
       }
   }

   return NULL;
}
