/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: sgluFlyCamera.cpp
 *  Created: 22 March 1997
 *  Summary: was dmGLFlyCamera.cpp
 *****************************************************************************/

#if defined(WIN32)
#include <windows.h>
#endif

#include <GL/gl.h>
#include <GL/glu.h>
#include <sgluFlyCamera.hpp>
#include <sglOStream.hpp>

//============================================================================
// class sgluFlyCamera : public sgluCamera
//============================================================================

//----------------------------------------------------------------------------
sgluFlyCamera::sgluFlyCamera() : sgluCamera()
{
   reset();
}

//----------------------------------------------------------------------------
//    Summary:
// Parameters:
//    Returns:
//----------------------------------------------------------------------------
sgluFlyCamera::~sgluFlyCamera()
{
   sglPrintDebug() << "destructing sgluFlyCamera" << endl;
}

//----------------------------------------------------------------------------
void sgluFlyCamera::reset()
{
   sgluCamera::reset();

   m_pos_ff[0] = m_pos_ff[1] = m_pos_ff[2] = 0.0;
   setAzimuth(0.0);
   setElevation(0.0);
   m_trans_scale = 1.0;
}

//----------------------------------------------------------------------------
ostream& operator<<(ostream &ostrm, const sgluFlyCamera &cam)
{
   ostrm << "sgluFlyCamera state:" << endl;
   ostrm << "Azimuth:    " << cam.m_azimuth << endl;
   ostrm << "Elevation:  " << cam.m_elevation << endl;
   ostrm << "Camera pos: " << cam.m_pos_ff[0] << ", "
         << cam.m_pos_ff[1] << ", "
         << cam.m_pos_ff[2] << endl;
   return ostrm;
}

//----------------------------------------------------------------------------
void sgluFlyCamera::updateAzimuth(float delta_azimuth)
{
   setAzimuth(m_azimuth + delta_azimuth);
}

//----------------------------------------------------------------------------
void sgluFlyCamera::setAzimuth(float new_azimuth)
{
   m_azimuth = new_azimuth;

   // allow it to wrap around
   while (m_azimuth > M_PI)
   {
      m_azimuth -= (float)M_PI*2.0f;
   }
   while (m_azimuth < -M_PI)
   {
      m_azimuth += (float)M_PI*2.0f;
   }

   m_sin_az = sin(m_azimuth);
   m_cos_az = cos(m_azimuth);
}

//----------------------------------------------------------------------------
void sgluFlyCamera::updateElevation(float delta_elevation)
{
   setElevation(m_elevation + delta_elevation);
}

//----------------------------------------------------------------------------
void sgluFlyCamera::setElevation(float elevation)
{
   m_elevation = elevation;

   // allow it to wrap around
   while (m_elevation > M_PI)
   {
      m_elevation -= (float)M_PI*2.0f;
   }
   while (m_elevation < -M_PI)
   {
      m_elevation += (float)M_PI*2.0f;
   }

   m_sin_el = sin(m_elevation);
   m_cos_el = cos(m_elevation);
}

//----------------------------------------------------------------------------
void sgluFlyCamera::updateFlyPos(float delta_pos[3])
{
   m_pos_ff[0] += delta_pos[0];
   m_pos_ff[1] += delta_pos[1];
   m_pos_ff[2] += delta_pos[2];
}

void sgluFlyCamera::updateFlyPos(float delta_x, float delta_y, float delta_z)
{
   m_pos_ff[0] += delta_x;
   m_pos_ff[1] += delta_y;
   m_pos_ff[2] += delta_z;
}

//----------------------------------------------------------------------------
void sgluFlyCamera::update(sgluMouse *mouse)
{
   static float delta_pos[3];

   if (mouse->in_window_flag)
   {
      if (mouse->button_flags & (sgluMouse::eLEFT_DOWN|
                                 sgluMouse::eMIDDLE_DOWN|
                                 sgluMouse::eRIGHT_DOWN))
      {
         if (mouse->button_flags & sgluMouse::eLEFT_DOWN)
         {
            // mouse x position controls azimuth
            // mouse y position controls elevation
            updateAzimuth((M_PI/90.0)*mouse->xchan);
            updateElevation((M_PI/90.0)*mouse->ychan);
         }

         delta_pos[0] = delta_pos[1] = delta_pos[2] = 0.0;

         if (mouse->button_flags & sgluMouse::eMIDDLE_DOWN)
         {
            delta_pos[0] = 2.0*m_trans_scale*mouse->xchan;
            delta_pos[2] = 2.0*m_trans_scale*mouse->ychan;
         }

         if (mouse->button_flags & sgluMouse::eRIGHT_DOWN)
         {
            delta_pos[1] = -2.0*fabs(mouse->ychan)*mouse->ychan;
         }

         m_pos_ff[0] += delta_pos[0]*m_cos_az
                      - delta_pos[1]*m_sin_az*m_cos_el
                      + delta_pos[2]*m_sin_az*m_sin_el;

         m_pos_ff[1] += delta_pos[0]*m_sin_az
                      + delta_pos[1]*m_cos_az*m_cos_el
                      - delta_pos[2]*m_cos_az*m_sin_el;

         m_pos_ff[2] += delta_pos[1]*m_sin_el
                      + delta_pos[2]*m_cos_el;
      }
   }
}

//----------------------------------------------------------------------------
void sgluFlyCamera::applyView()
{
   glMatrixMode(GL_MODELVIEW);
   glLoadIdentity();

   m_look_at[0] =  m_sin_az*m_cos_el;
   m_look_at[1] =  m_sin_el;
   m_look_at[2] = -m_cos_az*m_cos_el;

   m_up[0] = -m_cos_az;
   m_up[1] = 0.0;
   m_up[2] = -m_sin_az;

   gluLookAt(m_pos_ff[0],
             m_pos_ff[1],
             m_pos_ff[2],
             m_pos_ff[0] + m_look_at[0],
             m_pos_ff[1] + m_look_at[1],
             m_pos_ff[2] + m_look_at[2],
             m_up[0], m_up[1], m_up[2]);
}
