/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: globeTest.cpp
 *  Created: 12 February 1999
 *  Summary: Test LARGE model of globe (test double/single precision traversal)
 *****************************************************************************/

#include <GL/glut.h>
#include <sgl.h>
#include <sglu.h>

#include <sgluMouse.hpp>
#include <sgluSphericalTransform.hpp>

#include <sglVector.hpp>
#include <sglMatrix.hpp>

#include <sglGroup.hpp>
#include <sglScene.hpp>
#include <sglTranslate.hpp>
#include <sglGeode.hpp>
#include <sglIndexedTriangleStripSet.hpp>
#include <sglDefaultStateNode.hpp>

#include <sglViewPlatform.hpp>
#include <sglPerspectiveCamera.hpp>
#include <sglView.hpp>
#include <sglAxes.hpp>

sglView                   *view;
sglScene                  *root;

sgluMouse                 *mouse;
sgluSphericalTransform<double> *view_trans;
sglViewPlatform           *view_platform;
sglPerspectiveCamera      *view_camera;

deque<const sglStatelet*>  override_statelets;

sglAntiAliasLine aa_line;
sglAntiAliasPoint aa_point;
sglAntiAliasPolygon aa_poly;
sglPolyFrontStyle pfs;
sglPolyBackStyle pbs;

unsigned int win_x = 640, win_y = 480;

float translationScaleFactor = 4.0e3;
double near_clip = 20000.;
double far_clip = 2.0e8;
double fov = M_PI/3.0;
float radius = 3.0e7;

//----------------------------------------------------------------------------
void displayFunc()
{
   view->cullAndDraw();
   glutSwapBuffers();
}

//----------------------------------------------------------------------------
void reshapeFunc(int w, int h)
{
   view->setViewport(0, 0, w, h);

   mouse->win_size_x = win_x = w;
   mouse->win_size_y = win_y = h;

   double aspect = (double)w/(double)h;

   view_camera->setFOV(fov, aspect, near_clip, far_clip);
}

//----------------------------------------------------------------------------
void keyboardFunc(unsigned char key, int, int)
{
   switch (key)
   {
      case 27:
         glutDestroyWindow(glutGetWindow());
         exit(1);
         break;

      case 'a':
         aa_point.enable(!aa_point.isEnabled());
         cout << "point aa override = " << aa_point.isEnabled() << endl;
         break;

      case 'A':
         aa_line.enable(!aa_line.isEnabled());
         cout << "line aa override = " << aa_line.isEnabled() << endl;
         break;

      case 'z':
         aa_poly.enable(!aa_poly.isEnabled());
         cout << "polygon aa override = " << aa_poly.isEnabled() << endl;
         break;

      case 'F':
         if (pfs.getStyle() == sglPolyStyle::eFILLED)
            pfs.setStyle(sglPolyStyle::ePOINT);
         else if (pfs.getStyle() == sglPolyStyle::ePOINT)
            pfs.setStyle(sglPolyStyle::eLINE);
         else if (pfs.getStyle() == sglPolyStyle::eLINE)
            pfs.setStyle(sglPolyStyle::eFILLED);
         break;

      case 'B':
         if (pbs.getStyle() == sglPolyStyle::eFILLED)
            pbs.setStyle(sglPolyStyle::ePOINT);
         else if (pbs.getStyle() == sglPolyStyle::ePOINT)
            pbs.setStyle(sglPolyStyle::eLINE);
         else if (pbs.getStyle() == sglPolyStyle::eLINE)
            pbs.setStyle(sglPolyStyle::eFILLED);
         break;

      case 't':
         if (translationScaleFactor > 0.1f) translationScaleFactor *= 0.1f;
         view_trans->setTranslationScale(translationScaleFactor);
         cout << "translationScaleFactor = "
              << translationScaleFactor << endl;
         break;

      case 'T':
         translationScaleFactor *= 10.0f;
         view_trans->setTranslationScale(translationScaleFactor);
         cout << "translationScaleFactor = "
              << translationScaleFactor << endl;
         break;

      case 'f':
         view->enableFrameRateOutput(!(view->getFrameRateOutputFlag()));
         break;

      case 'd':
         view->setTraversalType(!(view->getDPTraversalFlag()),
                               view->getStatsFlag());
         cout << "Setting to " << (view->getDPTraversalFlag() ? "DP" : "SP")
              << " traversal." << endl;
         break;

      case '$':
         view->setTraversalType(view->getDPTraversalFlag(),
                               !(view->getStatsFlag()));
         cout << "Setting stats traversal "
              << (view->getStatsFlag() ? "on" : "off") << endl;
         break;

      case 'o':
         sgluOutputTree(root, cout, 3);
         break;

      case '/':
      case '?':
         cerr << "a - toggle point anti-aliasing" << endl;
         cerr << "A - toggle line anti-aliasing" << endl;
         cerr << "z - toggle polygon anti-aliasing" << endl;
         cerr << "F - cycle through front face polygon styles" << endl;
         cerr << "B - cycle through back face polygon styles" << endl;
         cerr << "t/T - inc/dec translation scale factor" << endl;
         cerr << "f - toggle frame rate statistics" << endl;
         cerr << "d - toggle between SP and DP traversal" << endl;
         cerr << "$ - toggle scene graph statistics reporting" << endl;
         cerr << "o - output scenegraph" << endl;
         cerr << "? - print help (this list)" << endl;
         cerr << "ECSAPE - exit" << endl;
         break;
   }
}

//----------------------------------------------------------------------------
void specialFunc(int key, int, int)
{
   switch (key)
   {
      case GLUT_KEY_LEFT:
         break;
      case GLUT_KEY_RIGHT:
         break;
      case GLUT_KEY_UP:
         break;
      case GLUT_KEY_DOWN:
         break;
   }
}

//----------------------------------------------------------------------------
void idleFunc()
{
   view_trans->update(mouse);
   displayFunc();
}

//----------------------------------------------------------------------------
inline void compute_face_normal(sglVec3f& v0, sglVec3f& v1, sglVec3f& v2,
                                sglVec3f& normal)
{
   sglVec3f a, b;

   a.sub(v1, v0);
   b.sub(v2, v0);
   normal.cross(a,b);
   normal.normalize();
}

//----------------------------------------------------------------------------
sglGeode *makeTile(char *filename)
{
   int ix, iy;
   int xsize = 0;
   int ysize = 0;

   // Load vertex data;
   ifstream file_ptr(filename);
   if (!file_ptr)
   {
      cerr << "Unable to open terrain data file: " << filename << endl;
      exit(1);
   }

   file_ptr >> xsize >> ysize;
   unsigned int vector_size = (xsize+1)*(ysize+1);

   sglVec3f *vertex = new sglVec3f[vector_size];
   int index = 0;

   for (ix=0; ix<=xsize; ix++)
   {
      for (iy=0; iy<=ysize; iy++)
      {
         file_ptr >> vertex[index][0] >> vertex[index][1] >> vertex[index][2];
         index++;
      }
   }
   file_ptr.close();

   // generate some reasonable normals
   sglVec3f *normal = new sglVec3f[vector_size];
   index = 0;

   for (ix=0; ix<=xsize; ix++)
   {
      for (iy=0; iy<=ysize; iy++)
      {
         if ((ix < xsize) && (iy < ysize))
         {
            compute_face_normal(vertex[index],
                                vertex[index + ysize + 1],
                                vertex[index + 1],
                                normal[index]);
         }
         else if ((ix == xsize) && (iy != ysize))
         {
            //cerr << "1: " << index << ' ' << ix << ',' << iy << ' '
            //     << index+1 << ',' << index-ysize << endl;
            compute_face_normal(vertex[index],
                                vertex[index + 1],
                                vertex[index - ysize -1],
                                normal[index]);
         }
         else if ((ix != xsize) && (iy == ysize))
         {
            //cerr << "2: " << index << ' ' << ix << ',' << iy << ' '
            //     << index-1 << ',' << index+ysize << endl;
            compute_face_normal(vertex[index],
                                vertex[index - 1],
                                vertex[index + ysize + 1],
                                normal[index]);
         }
         else
         {
            //cerr << "3: " << index << ' ' << ix << ',' << iy << ' '
            //     << index-ysize << ',' << index-1 << endl;
            compute_face_normal(vertex[index],
                                vertex[index - ysize -1],
                                vertex[index - 1],
                                normal[index]);
            //cerr << normal[index][0] << ' '
            //     << normal[index][1] << ' '
            //     << normal[index][2] << endl;

         }
         index++;
      }
   }

   // create the GeoSet
   int prim_size = 0;
   unsigned int *length = new unsigned int[xsize];
   for (ix=0; ix<xsize; ix++)
   {
      length[ix] = 2*(ysize+1);
      prim_size += length[ix];
   }

   index = 0;
   unsigned int *vindex = new unsigned int[prim_size];
   for (ix=0; ix<xsize; ix++)
   {
      for (iy=0; iy<=ysize; iy++)
      {
         vindex[index++] = ix*(ysize + 1) + iy;
         vindex[index++] = (ix + 1)*(ysize + 1) + iy;
      }
   }

   sglIndexedTriangleStripSet *gset = new sglIndexedTriangleStripSet();

   gset->setCoordList(vertex);
   gset->setNormalList(sglGeoSet::ePER_VERTEX, normal);
   gset->setNumPrims(xsize);
   gset->setPrimLengths(length);
   gset->setCoordIndexList(vindex);
   gset->setNormalIndexList(vindex);

   cerr << "Tile set: " << gset << endl;

   sglGeode *geode = new sglGeode();
   geode->addGeometry(gset);

   return geode;
}

//----------------------------------------------------------------------------
sglGroup *makeWorld(unsigned int numLonDiv, unsigned int numLatDiv)
{
   const unsigned int numDiv = 5;
   float deltaLat = M_PI/numLatDiv;
   float deltaLon = 2.0*M_PI/numLonDiv;

   //cerr << "deltas: " << deltaLat << ' ' << deltaLon << endl;

   sglGroup *group = new sglGroup();
   for (unsigned int i=0; i<numLatDiv; i+=2)
   {
      float latStart = i*deltaLat - M_PI/2.0;
      for (unsigned int j=0; j<numLonDiv; j+=2)
      {
         float lonStart = j*deltaLon - M_PI;

         sglVec3d localOrigin, tmp;
         float latMid = latStart + 0.5f*deltaLat;
         float lonMid = lonStart + 0.5f*deltaLon;
         localOrigin.set(radius*cos(lonMid)*cos(latMid),
                         radius*sin(lonMid)*cos(latMid),
                         radius*sin(latMid));
         //cerr << localOrigin << endl;

         // -------------------------------------------------
         // make vertex list
         sglVec3f *vertex = new sglVec3f[(numDiv+1)*(numDiv+1)];
         //cerr << (numDiv+1)*(numDiv+1) << " vertices" << endl;

         for (unsigned int iy=0; iy<=numDiv; iy++)
         {
            float lat = latStart + iy*deltaLat/numDiv;
            unsigned int indexRow = iy*(numDiv+1);
            for (unsigned int ix=0; ix<=numDiv; ix++)
            {
               float lon = lonStart + ix*deltaLon/numDiv;
               //cerr << indexRow+ix << ": " << lat << ", " << lon << endl;
               tmp.set(radius*cos(lon)*cos(lat) - localOrigin[0],
                       radius*sin(lon)*cos(lat) - localOrigin[1],
                       radius*sin(lat) - localOrigin[2]);
               vertex[indexRow+ix].set(tmp[0], tmp[1], tmp[2]);
               //cerr << vertex[indexRow+ix] << endl;
            }
            //cerr << endl;
         }
         //cerr << endl;

         // create the GeoSet
         unsigned int prim_size = 0;
         unsigned int *length = new unsigned int[numDiv];
         for (unsigned int i=0; i<numDiv; i++)
         {
            length[i] = 2*(numDiv+1);
            prim_size += length[i];
         }

         unsigned int index = 0;
         unsigned int *vindex = new unsigned int[prim_size];
         for (unsigned int ix=0; ix<numDiv; ix++)
         {
            for (unsigned int iy=0; iy<=numDiv; iy++)
            {
               vindex[index++] = (ix + 1)*(numDiv + 1) + iy;
               vindex[index++] = ix*(numDiv + 1) + iy;
            }
         }

         sglIndexedTriangleStripSet *gset =
            new sglIndexedTriangleStripSet();

         gset->setCoordList(vertex);
         gset->setCoordIndexList(vindex);
         gset->setNumPrims(numDiv);
         gset->setPrimLengths(length);

         sglVec4f *color2 = new sglVec4f[1];
         color2[0][0] = color2[0][3] = color2[0][1] = color2[0][2] = 1.0;
         unsigned int *cindex = new unsigned int[1];
         cindex[0] = 0;
         gset->setColorIndexList(cindex);
         gset->setColorList(sglGeoSet::eOVERALL, color2);
         // -------------------------------------------------

         sglGeode *geode = new sglGeode();
         geode->addGeometry(gset);
         sglTranslated *trans = new sglTranslated;
         trans->setTranslation(localOrigin[0], localOrigin[1], localOrigin[2]);
         trans->addChild(geode);

         group->addChild(trans);
      }
   }
   return group;
}

//----------------------------------------------------------------------------
void buildScenegraph()
{
   root = new sglScene();

   vector<sglStatelet*> default_statelets;
   default_statelets.push_back(new sglCullMode(sglCullMode::eOFF));

   sglDefaultStateNode *def = new sglDefaultStateNode();
   def->setStatelets(default_statelets);

   root->addChild(def);

   def->addChild(makeWorld(24, 12));

   // the camera (frustum)
   double aspect = (double)win_x/(double)win_y;

   view_camera = new sglPerspectiveCamera;
   view_camera->setFOV(fov, aspect, near_clip, far_clip);

   // the view platform
   view_platform = new sglViewPlatform(*view_camera);
   //view_platform->setEnableZup(true);

   // the view transform
   view_trans = new sgluSphericalTransform<double>();
   view_trans->setMode(sgluSphericalTransform<double>::eZ_UP);
   view_trans->setRadius(2.f*radius);
   view_trans->setAzimuth((float)M_PI/6.0f);
   view_trans->setElevation((float)-M_PI*0.25f);
   view_trans->setCOI(0.0, 0.0, 0.0);
   translationScaleFactor = 0.01f*radius;
   view_trans->setTranslationScale(translationScaleFactor);
   view_trans->computeMatrix();
   view_trans->addChild(view_platform);
   root->addChild(view_trans);

   sglGeode *geode = new sglGeode;
   geode->addGeometry(new sglAxes(2.0e6f));
   root->addChild(geode);

   root->preDraw();

   view = new sglView;
   view->setScene(root, view_platform);
   view->setViewport(0, 0, win_x, win_y);

   pfs.setStyle(sglPolyStyle::eLINE);
   override_statelets.push_back(&pfs);

   pbs.setStyle(sglPolyStyle::ePOINT);
   override_statelets.push_back(&pbs);

   override_statelets.push_back(&aa_point);
   override_statelets.push_back(&aa_line);
   override_statelets.push_back(&aa_poly);
   view->setOverrideStatelets(override_statelets);
}


//----------------------------------------------------------------------------
int main(int argc, char** argv)
{
   glutInit(&argc, argv);

   glutInitWindowSize(700, 512);
   glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);
   glutCreateWindow("DP Culling Testbed");

   sgl::initialize();
   mouse = sgluMouse::initialize();
   buildScenegraph();

   glutReshapeFunc(reshapeFunc);
   glutKeyboardFunc(keyboardFunc);
   glutSpecialFunc(specialFunc);
   glutDisplayFunc(displayFunc);
   glutIdleFunc(idleFunc);

   glutMainLoop();
   return 0;             /* ANSI C requires main to return int. */
}
