/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: multipassStencil.cpp
 *   Author: Scott McMillan
 *  Created: 20 May 2000
 *  Summary: test multipass rendering techniques using the stencil buffer, this
 *           algorithm is derived from Mark Kilgard's Real-time Shadow library.
 *****************************************************************************/

#include <GL/glut.h>
#include <sgl.h>
#include <sglu.h>
#include <stdio.h>

#include <sglTimespec.hpp>
#include <sgluMouse.hpp>
#include <sgluSphericalTransform.hpp>
#include <sglViewPlatform.hpp>
#include <sglPerspectiveCamera.hpp>

#include <sglScene.hpp>
#include <sglDiscriminator.hpp>
#include <sglTransform.hpp>
#include <sglTranslate.hpp>

#include <sglGeode.hpp>
#include <sglBitmapFont.hpp>

#include <sglGeoSet.hpp>
#include <sglTriangleStripSet.hpp>
#include <sglTriangleFanSet.hpp>
#include <sglQuadSet.hpp>
#include <sglLineSet.hpp>

#include <sglMaterial.hpp>
#include <sglTexture2D.hpp>
#include <sglTexEnv.hpp>

#include <sglCone.hpp>
#include <sglCube.hpp>
#include <sglAxes.hpp>

sgluMouse                     *mouse;
sglScene                      *root;

sgluSphericalTransform<float> *view_trans;
sglViewPlatform               *view_platform;
sglPerspectiveCamera          *view_camera;

sglMat4f                       view_matrix;
sglFrustumf                    frustum;
sglCullf                      *trav_state;
deque<const sglStatelet*>      override_statelets;
deque<const sglStatelet*>      stencil_override_statelets;

float far_clip = 200.0f;

GLfloat light_pos[4] = {0.0, -7.0, 1.0, 1.0};
GLfloat light_color0[4] = {1.0, 1.0, 1.0, 1.0};

double angle = 0.0;
sglTransformf *rot_node[3];
GLfloat axis[3][3] =  {{1.7f, 0.5f, -0.0f},
                       {0.0f, 1.0f, -0.3f},
                       {-0.6f, 0.0f, -2.0f}};

GLfloat light_offset[3][4] = {{ 3.0,  0.0,  0.0},
                              { 0.0,  3.0,  0.0},
                              { 0.0,  0.0,  3.0}};

GLfloat light_color[3][4] = {{0.75, 0.0, 0.0, 1.0},
                             {0.0, 0.75, 0.0, 1.0},
                             {0.0, 0.0, 0.75, 1.0}};

const unsigned int STENCIL_VOLUME       = 0x01; // discriminator mask
const unsigned int STENCIL_VOLUME_FRONT = 0x01; // state mask
const unsigned int STENCIL_VOLUME_BACK  = 0x02; // state mask

unsigned int win_x = 640, win_y = 480;
double aspect = 4./3.;
double fov_x = (float)M_PI/4.0f;
double fov_y = fov_x/aspect;

bool paused_flag = false;
sglTimespec tv, last_tv;
int timer_count = 0;
float frame_rate = 0.f;

enum
{
    M_NO_SHADOW, M_FRONT_VOLUME, M_BACK_VOLUME, M_SHADOW
};

int render_mode = M_NO_SHADOW;

//----------------------------------------------------------------------------
void outputFramerate()
{
   glDisable(GL_DEPTH_TEST);
   glMatrixMode(GL_PROJECTION);
   glPushMatrix();
   glLoadIdentity();
   glOrtho(0, mouse->win_size_x, 0, mouse->win_size_y, -1, 1);

   glMatrixMode(GL_MODELVIEW);
   glLoadIdentity();

   glEnable(GL_BLEND);
   glColor3f(0.3f,1.0f,0.0f);
   char fr[80];
   sprintf(fr, "framerate = %.2f", frame_rate);
   glRasterPos3i(1,1,0);
   sglBitmapFont::getDefaultFont()->rasterString(fr);
   glDisable(GL_BLEND);

   glMatrixMode(GL_PROJECTION);
   glPopMatrix();
   glMatrixMode(GL_MODELVIEW);
   glEnable(GL_DEPTH_TEST);

   glColor3f(1.,1.,1.);
}

//----------------------------------------------------------------------------
void displayFunc()
{
   glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);

   (view_platform->getCamera()).applyProjection();

   glMatrixMode (GL_MODELVIEW);
   glPushMatrix ();

   root->preDraw();

   float fov_y = (float)view_camera->getFOVY();
   float fov_x = (float)view_camera->getFOVX();
   view_platform->computeViewMatrix(view_matrix);

   frustum = view_camera->getFrustum();

   // Place lights relative to view
   glLoadIdentity();
   //sglLoadMatrix(view_matrix);  // place them relative to scene

   glLightfv(GL_LIGHT0, GL_POSITION, light_pos);
   glLightfv(GL_LIGHT0, GL_DIFFUSE, light_color0);
   glLightfv(GL_LIGHT0, GL_SPECULAR, light_color0);

   glLightfv(GL_LIGHT1, GL_POSITION, light_pos);
   glLightfv(GL_LIGHT1, GL_DIFFUSE, light_color[0]);
   glLightfv(GL_LIGHT1, GL_SPECULAR, light_color[0]);

   root->cull(*trav_state, fov_x, fov_y, win_x, win_y,
              frustum, view_matrix, override_statelets,
              ~STENCIL_VOLUME, 0);

   glEnable(GL_LIGHT0);
   glDisable(GL_LIGHT1);

   // ===============================================================
   switch (render_mode)
   {
       case M_NO_SHADOW:
           trav_state->draw();
           break;

       case M_FRONT_VOLUME:
           trav_state->draw();
           root->drawCulled(fov_x, fov_y, win_x, win_y,
                            frustum, view_matrix, override_statelets,
                            STENCIL_VOLUME, STENCIL_VOLUME_FRONT);
           break;

       case M_BACK_VOLUME:
           trav_state->draw();
           root->drawCulled(fov_x, fov_y, win_x, win_y,
                            frustum, view_matrix, override_statelets,
                            STENCIL_VOLUME, STENCIL_VOLUME_BACK);
           break;

       case M_SHADOW:
           // 1st scene pass
           trav_state->draw();

           // ===============================================================

           // 1st shadow volume pass:  Enable stencil to increment the stencil
           // value of pixels that pass the depth test when drawing the front
           // facing polygons of the shadow volume.  Do not update the depth
           // buffer while rendering the shadow volume.
           glEnable(GL_STENCIL_TEST);
           glStencilFunc(GL_ALWAYS, 0, 0);
           glStencilOp(GL_KEEP, GL_KEEP, sgl::s_stencil_inc_wrap);

           root->drawCulled(fov_x, fov_y, win_x, win_y,
                            frustum, view_matrix, stencil_override_statelets,
                            STENCIL_VOLUME, STENCIL_VOLUME_FRONT);

           // 2nd shadow volume pass: Now, draw the back facing polygons of the
           // shadow volume except decrement pixels that pass the depth test.
           // Again, do not update the depth buffer.
           glStencilOp(GL_KEEP, GL_KEEP, sgl::s_stencil_dec_wrap);

           root->drawCulled(fov_x, fov_y, win_x, win_y,
                            frustum, view_matrix, stencil_override_statelets,
                            STENCIL_VOLUME, STENCIL_VOLUME_BACK);

           // Now, pixels that lie within the shadow volume are tagged with a
           // one stencil value.  Empty shadowed regions of the shadow volume
           // get incremented, then decremented, to resolve to a net zero
           // stencil value.

           // 2nd scene pass (render shadowed region): Re-enable update of the
           // depth and color buffer (use GL_LEQUAL for depth buffer so we can
           // over-write depth values again with color.  Switch back to
           // backface culling and disable the light source.  Only update
           // pixels with a stencil value of one (shadowed).
           glStencilOp(GL_KEEP, GL_KEEP, GL_KEEP);
           glStencilFunc(GL_NOTEQUAL, 0, ~0u);

           glDisable(GL_LIGHT0);
           glEnable(GL_LIGHT1);

           trav_state->draw();

           // Put state back to sane modes.
           glDisable(GL_STENCIL_TEST);

           break;
   }
   // ===============================================================

   glPopMatrix ();
   outputFramerate();

   glFlush ();
   glutSwapBuffers();

   GLenum glerror = glGetError();
   if (glerror)
   {
      cerr << "glGetError() = " << (int)glerror << endl;
   }
}

//----------------------------------------------------------------------------
void reshapeFunc(int w, int h)
{
   glViewport (0, 0, w, h);

   mouse->win_size_x = win_x = w;
   mouse->win_size_y = win_y = h;

   aspect = (double)w/(double)h;
   fov_x = (float)M_PI/4.0f;
   fov_y = fov_x/aspect;

   view_camera->setFOV(fov_y, aspect, 1.0, far_clip);
}

//----------------------------------------------------------------------------
void keyboardFunc(unsigned char key, int, int)
{
   switch (key)
   {
      case 27:
         glutDestroyWindow(glutGetWindow());
         exit(1);
         break;

       case 's':
           render_mode = M_SHADOW;
           cout << "rendering scene with stencil" << endl;
           break;

       case 'n':
           render_mode = M_NO_SHADOW;
           cout << "rendering scene without stencil" << endl;
           break;

       case 'f':
           render_mode = M_FRONT_VOLUME;
           cout << "rendering scene with front face stencil volume" << endl;
           break;

       case 'b':
           render_mode = M_BACK_VOLUME;
           cout << "rendering scene with back face stencil volume" << endl;
           break;

      case 'p':
         paused_flag = !paused_flag;
         break;
   }
}


//----------------------------------------------------------------------------
void specialFunc(int key, int, int)
{
   switch (key)
   {
      case GLUT_KEY_LEFT:
         break;
      case GLUT_KEY_RIGHT:
         break;
      case GLUT_KEY_UP:
         break;
      case GLUT_KEY_DOWN:
         break;
   }
}


//----------------------------------------------------------------------------
void idleFunc()
{
   view_trans->update(mouse);

   if (!paused_flag)
   {
      angle += 0.02;

      for (int ix=0; ix<3; ix++)
      {
         if (rot_node[ix])
            rot_node[ix]->setRotation(axis[ix][0]*angle,
                                      axis[ix][1]*angle,
                                      axis[ix][2]*angle);
      }
   }

   glutPostRedisplay();

   // compute render rate
   timer_count++;
   sglTimespec::getSysTime(tv);
   double elapsed_time = sglTimespec::convertToDouble(tv - last_tv);

   if (elapsed_time > 2.5)
   {
      frame_rate =  (float) timer_count/elapsed_time;
      timer_count = 0;

      last_tv = tv;
   }
}

//----------------------------------------------------------------------------
#define CHECK_IMAGE_SIZE 64
static GLubyte check_image[CHECK_IMAGE_SIZE][CHECK_IMAGE_SIZE][4];

void makeCheckImage(void)
{
    int i,j;
    GLubyte c;
    for (i=0; i<CHECK_IMAGE_SIZE; i++)
    {
        for (j=0; j<CHECK_IMAGE_SIZE; j++)
        {
            c = 255;
            if ((i&0x4) != (j&0x4)) c = 0;
            check_image[i][j][0] = 255;
            check_image[i][j][1] = 255;
            check_image[i][j][2] = 255;
            check_image[i][j][3] = c;
        }
    }
}

//----------------------------------------------------------------------------

sglGeode *makeRoom()
{
   sglVec4f *color  = new sglVec4f[1]; // overall
   color[0].set(0.7, 0.7, 1.0, 1.0);

   sglMaterial *mat = new sglMaterial;
   mat->setAmbient(color[0][0], color[0][1], color[0][2]);
   mat->setDiffuse(color[0][0], color[0][1], color[0][2]);

   // floor is textured
   sglVec3f *vertex = new sglVec3f[4];
   sglVec2f *tcoord = new sglVec2f[4]; // per vertex
   sglVec3f *normal = new sglVec3f[1]; // per prim

   normal[0].set(0.0f, 0.0f, 1.0f);
   vertex[0].set(-10.0f, 10.0f, -10.0f); tcoord[0].set( 1.0f,  1.0f);
   vertex[1].set(-10.0f,-10.0f, -10.0f); tcoord[1].set( 1.0f,  0.0f);
   vertex[2].set( 10.0f,-10.0f, -10.0f); tcoord[2].set( 0.0f,  0.0f);
   vertex[3].set( 10.0f, 10.0f, -10.0f); tcoord[3].set( 0.0f,  1.0f);

   sglQuadSet *gset_floor = new sglQuadSet();
   gset_floor->setCoordList(vertex);
   gset_floor->setNormalList(sglGeoSet::ePER_PRIM, normal);
   gset_floor->setColorList(sglGeoSet::eOVERALL, color);
   gset_floor->setNumPrims(1);

   makeCheckImage();
   sglTexture2D *tex = new sglTexture2D;
   tex->setImage(&check_image[0][0][0], sglTexture::eRGBA,
                 sglTexture::eUNSIGNED_BYTE,
                 CHECK_IMAGE_SIZE, CHECK_IMAGE_SIZE, false);

   tex->setAutoMipmapFlag(true);
   tex->setMinFilter(sglTexture::eNEAREST);
   tex->setMagFilter(sglTexture::eNEAREST);
   tex->setWrapMode(sglTexture::eCLAMP_TO_EDGE, sglTexture::eCLAMP_TO_EDGE);

   tex->bind();

   sglTexEnv  *tev = new sglTexEnv();
   tev->setMode(sglTexEnv::eMODULATE);

   vector<sglStatelet*> tex_state;
   tex_state.push_back(new sglLighting(true));
   tex_state.push_back(new sglAlphaTest(sglAlphaTest::eGREATER,0.0f));
   tex_state.push_back(tex);
   tex_state.push_back(tev);
   tex_state.push_back(mat);

   gset_floor->setState(0, tex_state, tcoord);

   // -----------------------------------------------------------------
   // other walls are not textured

   vertex = new sglVec3f[5*4];
   normal = new sglVec3f[5];        // per prim

   // ceiling
   normal[0].set(0.0f, 0.0f, -1.0f);
   vertex[0].set( 10.0f,-10.0f, 10.0f);
   vertex[1].set(-10.0f,-10.0f, 10.0f);
   vertex[2].set(-10.0f, 10.0f, 10.0f);
   vertex[3].set( 10.0f, 10.0f, 10.0f);

   // wall 1
   normal[1].set(0.0f, 1.0f, 0.0f);
   vertex[4].set( 10.0f, -10.0f,  10.0f);
   vertex[5].set( 10.0f, -10.0f, -10.0f);
   vertex[6].set(-10.0f, -10.0f, -10.0f);
   vertex[7].set(-10.0f, -10.0f,  10.0f);

   // wall 2
   normal[2].set(0.0f, -1.0f, 0.0f);
   vertex[8].set(-10.0f, 10.0f,  10.0f);
   vertex[9].set(-10.0f, 10.0f, -10.0f);
   vertex[10].set( 10.0f, 10.0f, -10.0f);
   vertex[11].set( 10.0f, 10.0f,  10.0f);

   // wall 3
   normal[3].set(1.0f, 0.0f, 0.0f);
   vertex[12].set(-10.0f, -10.0f,  10.0f);
   vertex[13].set(-10.0f, -10.0f, -10.0f);
   vertex[14].set(-10.0f,  10.0f, -10.0f);
   vertex[15].set(-10.0f,  10.0f,  10.0f);

   // wall 4
   normal[4].set(-1.0f, 0.0f, 0.0f);
   vertex[16].set(10.0f,  10.0f,  10.0f);
   vertex[17].set(10.0f,  10.0f, -10.0f);
   vertex[18].set(10.0f, -10.0f, -10.0f);
   vertex[19].set(10.0f, -10.0f,  10.0f);

   sglQuadSet *gset = new sglQuadSet();
   gset->setCoordList(vertex);
   gset->setNormalList(sglGeoSet::ePER_PRIM, normal);
   gset->setColorList(sglGeoSet::eOVERALL, color);
   gset->setNumPrims(5);

   vector<sglStatelet*> lit_state;
   lit_state.push_back(new sglLighting(true));
   lit_state.push_back(mat);
   gset->setState(0, lit_state);

   sglGeode *geode = new sglGeode();
   geode->addGeometry(gset_floor);
   geode->addGeometry(gset);

   return geode;
}

//----------------------------------------------------------------------------
sglNode *makeCone()
{
   vector<sglStatelet*> state;
   state.push_back(new sglLighting(true));
   state.push_back(new sglTransparency(true));
   state.push_back(new sglCullMode(sglCullMode::eOFF));
   sglMaterial *material = new sglMaterial();
   material->setAmbient(0.0f, 0.5f, 1.0f);
   material->setDiffuse(0.0f, 0.5f, 1.0f);
   material->setAlpha(0.5f);
   state.push_back(material);

   sglCone *cone = new sglCone(20, 4.0, 2.0f);
   cone->setState(0, state);

   sglGeode *geode = new sglGeode;
   geode->addGeometry(cone);
   return geode;
}

//----------------------------------------------------------------------------
const double sensor_fov = M_PI/16.0;

sglNode *makeFrusta()
{
   float v = 20.0f*tan(sensor_fov);
   sglDiscriminator *group = new sglDiscriminator();
   group->setMask(~0x0);

   vector<sglStatelet*> state_front;
   state_front.push_back(new sglCullMode(sglCullMode::eBACK));
   vector<sglStatelet*> state_back;
   state_back.push_back(new sglCullMode(sglCullMode::eFRONT));

   for (int ix = 0; ix < 3; ix++)
   {
      sglTranslatef *trans_node = new sglTranslatef;
      trans_node->setTranslation(light_offset[ix][0],
                                 light_offset[ix][1],
                                 light_offset[ix][2]);
      rot_node[ix] = new sglTransformf();

      sglVec4f *color  = new sglVec4f[1];
      color[0].set(light_color[ix][0], light_color[ix][1], light_color[ix][2],
                   1.0f);

      sglVec3f *vertex = new sglVec3f[8];
      vertex[0].set(0,0,0);
      vertex[1].set( v, v,20.0f);
      vertex[2].set(0,0,0);
      vertex[3].set(v,-v,20.0f);
      vertex[4].set(0,0,0);
      vertex[5].set(-v, v,20.0f);
      vertex[6].set(0,0,0);
      vertex[7].set(-v,-v,20.0f);

      sglLineSet *gset = new sglLineSet;
      gset->setCoordList(vertex);
      gset->setColorList(sglGeoSet::eOVERALL, color);
      gset->setNumPrims(4);

      sglGeode *geode = new sglGeode;
      geode->addGeometry(gset);

      sglDiscriminator *disc_node = new sglDiscriminator;
      disc_node->setMask(~STENCIL_VOLUME);
      disc_node->addChild(geode);

      rot_node[ix]->addChild(disc_node);

      //-----------------------------
      unsigned int *length = new unsigned int[1];
      length[0] = 6;

      vertex = new sglVec3f[6];
      vertex[0].set(0,0,0);
      vertex[1].set( v,  v, 20.0f);
      vertex[2].set( v, -v, 20.0f);
      vertex[3].set(-v, -v, 20.0f);
      vertex[4].set(-v,  v, 20.0f);
      vertex[5].set( v,  v, 20.0f);

      sglTriangleFanSet *gset1 = new sglTriangleFanSet();
      gset1->setNumPrims(1);
      gset1->setPrimLengths(length);
      gset1->setCoordList(vertex);
      gset1->setState(STENCIL_VOLUME_BACK, state_back);
      gset1->setState(STENCIL_VOLUME_FRONT, state_front);

      geode = new sglGeode;
      geode->addGeometry(gset1);

      disc_node = new sglDiscriminator;
      disc_node->setMask(STENCIL_VOLUME);
      disc_node->addChild(geode);

      rot_node[ix]->addChild(disc_node);


      trans_node->addChild(rot_node[ix]);
      group->addChild(trans_node);
   }


   return group;
}

//----------------------------------------------------------------------------
void buildScenegraph()
{
   override_statelets.push_back(new sglDepthTest(sglDepthTest::eLEQUAL));

   stencil_override_statelets.push_back(new sglLighting(false));
   stencil_override_statelets.push_back(new sglColorMask(false, false,
                                                         false, false));
   stencil_override_statelets.push_back(new sglDepthMask(false));

   // make scene and default state:
   sglDiscriminator *disc_node = new sglDiscriminator;
   disc_node->setMask(~STENCIL_VOLUME);
   disc_node->addChild(makeRoom());

   sglTranslatef *trans_node = new sglTranslatef;
   trans_node->setTranslation(-4, 0, -4);
   trans_node->addChild(makeCone());
   disc_node->addChild(trans_node);

   // make a cube with material properties
   trans_node = new sglTranslatef;
   trans_node->setTranslation(3, -1, -5);

   sglCube *cube = new sglCube(3.0f, 3.0f, 3.0f);
   vector<sglStatelet*> state;
   state.push_back(new sglLighting(true));
   sglMaterial *material = new sglMaterial();
   material->setAmbient(1.0f, 0.5f, 0.0f);
   material->setDiffuse(1.0f, 0.5f, 0.0f);
   state.push_back(material);
   cube->setState(0, state);
   sglGeode *geode = new sglGeode;
   geode->addGeometry(cube);
   trans_node->addChild(geode);

   disc_node->addChild(trans_node);

   root = new sglScene;
   root->addChild(disc_node);
   root->addChild(makeFrusta());

   // the camera (frustum)
   double aspect = (double)win_x/(double)win_y;
   view_camera = new sglPerspectiveCamera;
   view_camera->setFOV(M_PI/4.0, aspect, 0.1, 200.0);

   // the view platform
   view_platform = new sglViewPlatform(*view_camera);

   // the view transform
   view_trans = new sgluSphericalTransform<float>();
   view_trans->setMode(sgluSphericalTransform<float>::eZ_UP);
   view_trans->setRadius(33.0);
   view_trans->setCOI(0.0, 0.0, 0.0);
   view_trans->setTranslationScale(0.02f);
   view_trans->computeMatrix();
   view_trans->addChild(view_platform);
   root->addChild(view_trans);

   geode = new sglGeode;
   geode->addGeometry(new sglAxes(2.0f));
   root->addChild(geode);

   root->preDraw();

   trav_state = new sglCullf;
}

//----------------------------------------------------------------------------
int main(int argc, char** argv)
{
   glutInitDisplayString("stencil>=1 rgb double depth samples");
   glutInit(&argc, argv);
   glutInitWindowSize(480, 480);

   glutCreateWindow("SGL Multipass Stencil Demo");

   int stencil_bits = glutGet((GLenum)GLUT_WINDOW_STENCIL_SIZE);
   cout << "Bits of stencil = " << stencil_bits << endl;

   sgl::initialize();
   mouse = sgluMouse::initialize();
   buildScenegraph();

   glutReshapeFunc(reshapeFunc);
   glutKeyboardFunc(keyboardFunc);
   glutSpecialFunc(specialFunc);
   glutDisplayFunc(displayFunc);
   glutIdleFunc(idleFunc);

   sglTimespec::getSysTime(last_tv);
   cerr << "RENDERER = " << ((char *) glGetString(GL_RENDERER)) << endl;
   glutMainLoop();
   return 0;             /* ANSI C requires main to return int. */
}
