/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: multipassTexture.cpp
 *   Author: Scott McMillan
 *  Created: 26 October 2000
 *  Summary: Illustrate how multipass rendering can be done with SGL get a
 *           shiny effect (using reflection mapping) on textured objects.
 *****************************************************************************/

#include <GL/glut.h>
#include <sgl.h>
#include <sglu.h>
#include <stdio.h>

#include <sglTimespec.hpp>
#include <sgluMouse.hpp>
#include <sgluSphericalTransform.hpp>
#include <sglViewPlatform.hpp>
#include <sglPerspectiveCamera.hpp>

#include <sglScene.hpp>
#include <sglDiscriminator.hpp>
#include <sglOrderedGroup.hpp>
#include <sglTransform.hpp>
#include <sglTranslate.hpp>
#include <sglGeode.hpp>

#include <sglMonoIndexedTriangleStripSet.hpp>

#include <sglStatelet.hpp>
#include <sglMaterial.hpp>
#include <sglTexture2D.hpp>
#include <sglTexEnv.hpp>
#include <sglTexGen.hpp>

#include <sgldb.h>
#include <sglImageLoader.hpp>

#include <sglBitmapFont.hpp>
#include <sglSphere.hpp>
#include <sglAxes.hpp>

sgluMouse                     *mouse;
sglScene                      *root;

sgluSphericalTransform<float> *view_trans;
sglViewPlatform               *view_platform;
sglPerspectiveCamera          *view_camera;

sglMat4f                       view_matrix;
sglFrustumf                    frustum;
deque<const sglStatelet*>      override_statelets;

sglTransformf *rot_node1;
sglTransformf *rot_node2;

enum {eBASE, eREFLECTION, eBOTH, eORDERED_GROUP};
int render_mode = eBOTH;

GLfloat light_pos[4] = {0.0, -7.0, 1.0, 1.0};
GLfloat light_color0[4] = {1.0, 1.0, 1.0, 1.0};

double angle = 0.0;

const unsigned int BASE_TEXTURE       = 0x00; // state mask
const unsigned int REFLECTION_TEXTURE = 0x01; // state mask

unsigned int win_x = 640, win_y = 480;
double aspect = 4./3.;
double fov_x = M_PI/4.0;
double fov_y = fov_x/aspect;
double far_clip = 20.0;

bool paused_flag = false;
sglTimespec tv, last_tv, prev_tv;
int timer_count = 0;
float frame_rate = 0.f;

//----------------------------------------------------------------------------
void outputFramerate()
{
   glDisable(GL_DEPTH_TEST);
   glMatrixMode(GL_PROJECTION);
   glPushMatrix();
   glLoadIdentity();
   glOrtho(0, mouse->win_size_x, 0, mouse->win_size_y, -1, 1);

   glMatrixMode(GL_MODELVIEW);
   glLoadIdentity();

   glEnable(GL_BLEND);
   char fr[80];
   sprintf(fr, "framerate = %.2f", frame_rate);
   glColor3f(0.3f,1.0f,0.0f);
   glRasterPos2i(1,1);
   sglBitmapFont::getDefaultFont()->rasterString(fr);

   glRasterPos2i(1, 15);
   switch (render_mode)
   {
      case eBASE:
         sglBitmapFont::getDefaultFont()->rasterString("Base Texture");
         break;

      case eREFLECTION:
         sglBitmapFont::getDefaultFont()->rasterString("Reflection Texture");
         break;

      case eBOTH:
         sglBitmapFont::getDefaultFont()->rasterString("Two SGL passes");
         break;

      case eORDERED_GROUP:
         sglBitmapFont::getDefaultFont()->rasterString("Ordered Group");
         break;
   }

   glDisable(GL_BLEND);

   glMatrixMode(GL_PROJECTION);
   glPopMatrix();
   glMatrixMode(GL_MODELVIEW);
   glEnable(GL_DEPTH_TEST);
   glColor3f(1.,1.,1.);
}

//----------------------------------------------------------------------------
void displayFunc()
{
   glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);

   (view_platform->getCamera()).applyProjection();

   glMatrixMode (GL_MODELVIEW);
   glPushMatrix ();

   root->preDraw();

   float fov_y = (float)view_camera->getFOVY();
   float fov_x = (float)view_camera->getFOVX();
   view_platform->computeViewMatrix(view_matrix);

   frustum = view_camera->getFrustum();

   // Place lights relative to scene (not view)
   sglLoadMatrix(view_matrix);
   glLightfv(GL_LIGHT0, GL_POSITION, light_pos);
   glLightfv(GL_LIGHT0, GL_DIFFUSE, light_color0);
   glLightfv(GL_LIGHT0, GL_SPECULAR, light_color0);

   glEnable(GL_LIGHT0);

   switch (render_mode)
   {
      case eBASE:
         root->drawCulled(fov_x, fov_y, win_x, win_y,
                          frustum, view_matrix, override_statelets,
                          0x01, BASE_TEXTURE);
         break;

      case eREFLECTION:
         root->drawCulled(fov_x, fov_y, win_x, win_y,
                          frustum, view_matrix, override_statelets,
                          0x01, REFLECTION_TEXTURE);
         break;

      case eBOTH:
         root->drawCulled(fov_x, fov_y, win_x, win_y,
                          frustum, view_matrix, override_statelets,
                          0x01, BASE_TEXTURE);

         root->drawCulled(fov_x, fov_y, win_x, win_y,
                          frustum, view_matrix, override_statelets,
                          0x01, REFLECTION_TEXTURE);
         break;

      case eORDERED_GROUP:
         root->drawCulled(fov_x, fov_y, win_x, win_y,
                          frustum, view_matrix, override_statelets,
                          0x02, 0x00);

      default:
           break;
   }

   // ===============================================================
   glPopMatrix ();
   outputFramerate();

   glFlush ();
   glutSwapBuffers();

   GLenum glerror = glGetError();
   if (glerror)
   {
      cerr << "glGetError() = " << (int)glerror << endl;
   }
}

//----------------------------------------------------------------------------
void reshapeFunc(int w, int h)
{
   glViewport (0, 0, w, h);

   mouse->win_size_x = win_x = w;
   mouse->win_size_y = win_y = h;

   aspect = (double)w/(double)h;
   fov_x = (float)M_PI/4.0f;
   fov_y = fov_x/aspect;

   view_camera->setFOV(fov_y, aspect, 1.0, far_clip);
}

//----------------------------------------------------------------------------
void keyboardFunc(unsigned char key, int, int)
{
   switch (key)
   {
      case 27:
         glutDestroyWindow(glutGetWindow());
         exit(1);
         break;

      case 'r':
         cout << "render mode: ";
         if (render_mode == eBASE)
         {
            render_mode = eREFLECTION;
            cout << "eREFLECTION" << endl;
         }
         else if (render_mode == eREFLECTION)
         {
            render_mode = eBOTH;
            cout << "eBOTH" << endl;
         }
         else if (render_mode == eBOTH)
         {
            render_mode = eORDERED_GROUP;
            cout << "eORDERED_GROUP" << endl;
         }
         else if (render_mode == eORDERED_GROUP)
         {
            render_mode = eBASE;
            cout << "eBASE" << endl;
         }
         break;

      case 'o':
         sgluOutputTree(root, cout, 3);
         break;

      case 'p':
         paused_flag = !paused_flag;
         break;

      case '?':
      case '/':
         cerr << "r - cycle through rendering modes" << endl;
         cerr << "o - output scenegraph" << endl;
         cerr << "? - print help (this list)" << endl;
         cerr << "ECSAPE - exit" << endl;
         break;
   }
}


//----------------------------------------------------------------------------
void specialFunc(int key, int, int)
{
   switch (key)
   {
      case GLUT_KEY_LEFT:
         break;
      case GLUT_KEY_RIGHT:
         break;
      case GLUT_KEY_UP:
         break;
      case GLUT_KEY_DOWN:
         break;
   }
}


//----------------------------------------------------------------------------
void idleFunc()
{
   view_trans->update(mouse);

   if (!paused_flag)
   {
      sglTimespec::getSysTime(tv);
      double elapsed_time = sglTimespec::convertToDouble(tv - prev_tv);
      angle += 0.5*elapsed_time;
      if (angle > 2.0*M_PI)
         angle -= 2.0*M_PI;

      rot_node1->setRotation(-angle, 0.0, 0.0);
      rot_node2->setRotation(2.0*angle, 0.0, 0.0);

      prev_tv = tv;
   }

   glutPostRedisplay();

   // compute render rate
   timer_count++;
   sglTimespec::getSysTime(tv);
   double elapsed_time = sglTimespec::convertToDouble(tv - last_tv);

   if (elapsed_time > 2.5)
   {
      frame_rate =  (float) timer_count/elapsed_time;
      timer_count = 0;

      last_tv = tv;
   }
}

//----------------------------------------------------------------------------
#define CHECK_IMAGE_SIZE 64
static GLubyte check_image[CHECK_IMAGE_SIZE][CHECK_IMAGE_SIZE][4];

sglTexture2D *makeCheckTexture(GLubyte alpha)
{
   sglTexture2D *tex = new sglTexture2D;

   for (int i=0; i<CHECK_IMAGE_SIZE; i++)
   {
      for (int j=0; j<CHECK_IMAGE_SIZE; j++)
      {
         GLubyte c = 255;
         if ((i&0x4) != (j&0x4)) c = 0;
         check_image[i][j][0] = c;
         check_image[i][j][1] = 255;
         check_image[i][j][2] = 64;
         check_image[i][j][3] = alpha;
      }
   }

   tex->setImage(&check_image[0][0][0],
                 sglTexture::eRGBA, sglTexture::eUNSIGNED_BYTE,
                 CHECK_IMAGE_SIZE, CHECK_IMAGE_SIZE, false);

   tex->setAutoMipmapFlag(true);
   tex->setMinFilter(sglTexture::eNEAREST);
   tex->setMagFilter(sglTexture::eNEAREST);
   tex->setWrapMode(sglTexture::eCLAMP_TO_EDGE, sglTexture::eCLAMP_TO_EDGE);

   tex->bind();
   return tex;
}

//----------------------------------------------------------------------------
sglNode *makeSpheres(float radius,
                     unsigned int num_slices, unsigned int num_stacks,
                     const char *refl_tex_fname)
{
   if (num_slices < 3 || num_stacks < 2 || radius <= 0.0)
      return NULL;

   sglSphere *sphere = new sglSphere(num_slices, num_stacks, radius);
   sglSphere *sphereA = new sglSphere(num_slices, num_stacks, radius);
   sglSphere *sphereB = new sglSphere(num_slices, num_stacks, radius);

   ++num_slices;
   ++num_stacks;

   sglVec2f *tcoords  = new sglVec2f[num_slices*num_stacks];
   sglVec4f *color    = new sglVec4f[1]; // overall
   color[0].set(1.0, 1.0, 1.0, 1.0);

   unsigned int index = 0;
   for (unsigned int iy=0; iy<num_stacks; ++iy)
   {
      for (unsigned int ix=0; ix<num_slices; ++ix)
      {
         tcoords[index].set((float)ix/(float)(num_slices - 1),
                            (float)iy/(float)(num_stacks - 1));
         ++index;
      }
   }

   sphere->setColorList(sglGeoSet::eOVERALL, color);
   sphereA->setColorList(sglGeoSet::eOVERALL, color);
   sphereA->setColorList(sglGeoSet::eOVERALL, color);

   // ----------- make the state for the base texture ------------------
   sglTexture2D *tex = makeCheckTexture(0xff);
   sglTexEnv  *tev = new sglTexEnv();
   tev->setMode(sglTexEnv::eMODULATE);

   vector<sglStatelet*> tex_state;
   tex_state.push_back(new sglLighting(true));
   tex_state.push_back(tex);
   tex_state.push_back(tev);

   sphere->setState(BASE_TEXTURE, tex_state, tcoords);
   sphereA->setState(0, tex_state, tcoords);

   // ---------- make the state for the reflection texture -------------
   unsigned int width = 0;
   unsigned int height = 0;
   sglTexture::ExternalFormatEnum format;
   sglTexture::TypeEnum type;
   void *refl_image = sglImageLoader::loadImage(
      refl_tex_fname, width, height, format, type);
   if (format != sglTexture::eRGBA ||
       format != sglTexture::eLUMINANCE_ALPHA)
   {
      void *old_image = refl_image;
      refl_image = sglImageLoader::reformatImage(
         format, type, width, height, sglTexture::eRGBA, old_image);
      format = sglTexture::eRGBA;
      free(old_image);
   }
   sglImageLoader::scaleImageComponent(
      refl_image, format, type, width, height,
      0.25f, sglTexture::getNumComponents(format)-1);

   if (refl_image && width>0 && height>0)
   {
      sglTexture2D *tex = new sglTexture2D;
      tex->setImage(refl_image, format, type, width, height, false);
      sglTexGen  *tgen = new sglTexGen();
      tgen->setMode(sglTexGen::eS, sglTexGen::eSPHERE_MAP);
      tgen->setMode(sglTexGen::eT, sglTexGen::eSPHERE_MAP);

      vector<sglStatelet*> tex_state;
      tex_state.push_back(tex);
      tex_state.push_back(tgen);
      tex_state.push_back(new sglTransparency(true));
      tex_state.push_back(new sglDepthTest(sglDepthTest::eLEQUAL));

      sphere->setState(REFLECTION_TEXTURE, tex_state);
      sphereB->setState(0, tex_state);
   }

   sglGeode *geode = new sglGeode;
   geode->addGeometry(sphere);

   sglDiscriminator *disc1 = new sglDiscriminator;
   disc1->setMask(0x01);
   disc1->addChild(geode);

   sglOrderedGroup *order = new sglOrderedGroup;
   geode = new sglGeode;
   geode->addGeometry(sphereA);
   order->addChild(geode);
   geode = new sglGeode;
   geode->addGeometry(sphereB);
   order->addChild(geode);

   sglDiscriminator *disc2 = new sglDiscriminator;
   disc2->setMask(0x02);
   disc2->addChild(order);

   sglGroup *group = new sglGroup;
   group->addChild(disc1);  // for the two pass method
   group->addChild(disc2);  // for the ordered group method
   return group;
}

//----------------------------------------------------------------------------
void buildScenegraph(const char *refl_tex_fname)
{
   // make scene and default state:
   sglNode *spheres = makeSpheres(2.0f, 24, 24, refl_tex_fname);

   rot_node2 = new sglTransformf;
   rot_node2->addChild(spheres);

   sglTranslatef *trans_node = new sglTranslatef;
   trans_node->setTranslation(3.0f, 0.0, 0.0);
   trans_node->addChild(rot_node2);

   rot_node1 = new sglTransformf;
   rot_node1->addChild(trans_node);

   root = new sglScene();
   root->addChild(rot_node1);

   // the camera (frustum)
   double aspect = (double)win_x/(double)win_y;
   view_camera = new sglPerspectiveCamera;
   view_camera->setFOV(fov_y, aspect, 0.1, 200.0);

   // the view platform
   view_platform = new sglViewPlatform(*view_camera);

   // the view transform
   view_trans = new sgluSphericalTransform<float>();
   view_trans->setMode(sgluSphericalTransform<float>::eZ_UP);
   view_trans->setRadius(12.0);
   view_trans->setCOI(0.0, 0.0, 0.0);
   view_trans->setTranslationScale(0.01f);
   view_trans->computeMatrix();
   view_trans->addChild(view_platform);
   root->addChild(view_trans);

   sglGeode *geode = new sglGeode;
   geode->addGeometry(new sglAxes(0.5f));
   root->addChild(geode);

   root->preDraw();
}

//----------------------------------------------------------------------------
int main(int argc, char** argv)
{
   glutInitDisplayString("rgb double depth");
   glutInit(&argc, argv);

   char *image_fname = NULL;
   if (argc < 2)
   {
      cerr << "Usage error: " << argv[0] << " <reflection image filename>"
           << endl;
      exit(1);
   }
   else
   {
      image_fname = argv[1];
   }

   glutInitWindowSize(480, 320);
   glutCreateWindow("SGL Multipass Texture Demo");

   sgl::initialize();
   mouse = sgluMouse::initialize();
   buildScenegraph(image_fname);

   glutReshapeFunc(reshapeFunc);
   glutKeyboardFunc(keyboardFunc);
   glutSpecialFunc(specialFunc);
   glutDisplayFunc(displayFunc);
   glutIdleFunc(idleFunc);

   sglTimespec::getSysTime(last_tv);
   prev_tv = last_tv;

   glutMainLoop();
   return 0;             /* ANSI C requires main to return int. */
}
