/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: testBillboard.cpp
 *  Created: 6 March 2000
 *  Summary: GLUT-based sglBillboard testbed
 *****************************************************************************/

#include <GL/glut.h>
#include <sgl.h>
#include <sglu.h>

#include <sgluMouse.hpp>
#include <sgluSphericalTransform.hpp>

#include <sglGroup.hpp>
#include <sglScene.hpp>
#include <sglTranslate.hpp>
#include <sglBillboard.hpp>
#include <sglGeode.hpp>
#include <sglGeoSet.hpp>
#include <sglLineStripSet.hpp>
#include <sglLineSet.hpp>

#include <sglViewPlatform.hpp>
#include <sglPerspectiveCamera.hpp>
#include <sglView.hpp>
#include <sglAxes.hpp>

sglView                   *view;
sglScene                  *root;

sgluMouse                 *mouse;
sgluSphericalTransform<float> *view_trans;
sglViewPlatform           *view_platform;
sglPerspectiveCamera      *view_camera;

deque<const sglStatelet*>  override_statelets;

sglBillboard              *billboard_node;
bool                       billboard_axis_flag = true;

unsigned int win_x = 640, win_y = 480;

//----------------------------------------------------------------------------
void displayFunc()
{
   view->cullAndDraw();

   glutSwapBuffers();
}

//----------------------------------------------------------------------------
void reshapeFunc(int w, int h)
{
   view->setViewport(0, 0, w, h);

   mouse->win_size_x = win_x = w;
   mouse->win_size_y = win_y = h;

   double aspect = (double)w/(double)h;
   double fov_y = M_PI/3.0;

   view_camera->setFOV(fov_y, aspect, 1.0, 200.0);
}

//----------------------------------------------------------------------------
void keyboardFunc(unsigned char key, int, int)
{
   switch (key)
   {
      case 27:
         glutDestroyWindow(glutGetWindow());
         exit(1);
         break;

      case 'b':
         billboard_axis_flag = !billboard_axis_flag;
         if (billboard_axis_flag)
         {
            billboard_node->setMode(sglBillboard::eAXIAL_ROT);
            cout << "billboard: axis" << endl;
         }
         else
         {
            billboard_node->setMode(sglBillboard::ePOINT_ROT_EYE);
            cout << "billboard: point" << endl;
         }
         break;

      case 'f':
         view->enableFrameRateOutput(!(view->getFrameRateOutputFlag()));
         break;

      case 'd':
         view->setTraversalType(!(view->getDPTraversalFlag()),
                               view->getStatsFlag());
         break;

      case '$':
         view->setTraversalType(view->getDPTraversalFlag(),
                               !(view->getStatsFlag()));
         break;

      case 'o':
         sgluOutputTree(root, cout, 3);
         break;

      case '/':
      case '?':
         cerr << "b - toggle between billboard modes" << endl;
         cerr << "f - toggle frame rate statistics" << endl;
         cerr << "d - toggle between SP and DP traversal" << endl;
         cerr << "$ - toggle scene graph statistics reporting" << endl;
         cerr << "o - output scenegraph" << endl;
         cerr << "? - print help (this list)" << endl;
         cerr << "ECSAPE - exit" << endl;
         break;
   }
}


//----------------------------------------------------------------------------
void specialFunc(int key, int, int)
{
   switch (key)
   {
      case GLUT_KEY_LEFT:
         break;
      case GLUT_KEY_RIGHT:
         break;
      case GLUT_KEY_UP:
         break;
      case GLUT_KEY_DOWN:
         break;
   }
}


//----------------------------------------------------------------------------
void idleFunc()
{
   view_trans->update(mouse);
   displayFunc();
}

//----------------------------------------------------------------------------
sglGeode *makeLineSet()
{
   sglVec3f center(0.f, 0.f, 0.f);

   unsigned int num_points = 15;
   sglVec3f *vertices = new sglVec3f[num_points+1];
   for (unsigned int j=0; j<=num_points; j++)
   {
      vertices[j][0] = 10.0*sin((2*M_PI*j)/num_points);
      vertices[j][1] = 10.0*cos((2*M_PI*j)/num_points);
      vertices[j][2] = 0.0;
   }

   unsigned int *length = new unsigned int[1];
   length[0] = num_points+1;

   sglLineStripSet *gset = new sglLineStripSet();
   gset->setCoordList(vertices);
   gset->setNumPrims(1);
   gset->setPrimLengths(length);

   sglGeode *geode = new sglGeode();
   geode->addGeometry(gset);
   return geode;
}

//----------------------------------------------------------------------------
void buildScenegraph()
{
   root = new sglScene;

   sglTranslatef *scs_node = new sglTranslatef();
   scs_node->setTranslation(0.0,0.0,3.0);
   root->addChild(scs_node);

   billboard_node = new sglBillboard();
   scs_node->addChild(billboard_node);
   billboard_node->addChild(makeLineSet());

   sglVec3d pos(3.0, 0.0, 0.0);
   billboard_node->setPos(pos);
   billboard_node->setMode(sglBillboard::ePOINT_ROT_EYE);

   sglVec3d axis(1.0, 1.0, 0.0);
   billboard_node->setAxis(axis);
   billboard_node->setMode(sglBillboard::eAXIAL_ROT);

   sglGeode *geode = new sglGeode;
   geode->addGeometry(new sglAxes);
   billboard_node->addChild(geode);

   // ================================================================

   // the camera (frustum)
   double aspect = (double)win_x/(double)win_y;
   double fov_y = (double)M_PI/(3.0f*aspect);

   view_camera = new sglPerspectiveCamera;
   view_camera->setFOV(fov_y, aspect, 1.0, 200.0);

   // the view platform
   view_platform = new sglViewPlatform(*view_camera);
   //view_platform->setEnableZup(true);

   // the view transform
   view_trans = new sgluSphericalTransform<float>();
   view_trans->setMode(sgluSphericalTransform<float>::eZ_UP);
   view_trans->setRadius(30.0);
   view_trans->setCOI(0.0, 0.0, 0.0);
   view_trans->setTranslationScale(0.02f);
   view_trans->computeMatrix();
   view_trans->addChild(view_platform);
   root->addChild(view_trans);

   geode = new sglGeode;
   geode->addGeometry(new sglAxes(2.0f));
   root->addChild(geode);

   root->preDraw();

   view = new sglView;
   view->setScene(root, view_platform);
   view->setViewport(0, 0, win_x, win_y);

   override_statelets.push_back(new sglCullMode(sglCullMode::eOFF));
   override_statelets.push_back(new sglPolyFrontStyle(sglPolyStyle::eLINE));
   override_statelets.push_back(new sglPolyBackStyle(sglPolyStyle::eLINE));
   view->setOverrideStatelets(override_statelets);
}

//----------------------------------------------------------------------------
int main(int argc, char** argv)
{
   glutInit(&argc, argv);

   glutInitWindowSize(640, 480);
   glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);
   glutCreateWindow("sglfly");

   sgl::initialize();
   mouse = sgluMouse::initialize();
   buildScenegraph();

   glutReshapeFunc(reshapeFunc);
   glutKeyboardFunc(keyboardFunc);
   glutSpecialFunc(specialFunc);
   glutDisplayFunc(displayFunc);
   glutIdleFunc(idleFunc);

   glutMainLoop();
   return 0;             /* ANSI C requires main to return int. */
}
