/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: testClone.cpp
 *   Author: Scott McMillan
 *  Created: 16 October 2001
 *  Summary: Test the clone traversal
 *****************************************************************************/

#include <GL/glut.h>
#include <sgl.h>
#include <sglu.h>

#include <sglObject.hpp>
#include <sglNode.hpp>
#include <sglGroup.hpp>
#include <sglKeyFrameAnimation.hpp>
#include <sglTimedAnimation.hpp>
#include <sglBillboard.hpp>
#include <sglCallback.hpp>
#include <sglDiscriminator.hpp>
#include <sglLayer.hpp>
#include <sglDistanceSwitch.hpp>
#include <sglLOD.hpp>
#include <sglLODStateNode.hpp>
#include <sglScene.hpp>
#include <sglSwitch.hpp>
#include <sglBitMaskSwitch.hpp>
#include <sglIndexedBitMaskSwitch.hpp>
#include <sglTransform.hpp>
#include <sglSphericalTransform.hpp>
#include <sglTranslate.hpp>
#include <sglScale.hpp>
#include <sglDefaultStateNode.hpp>
#include <sglOverrideStateNode.hpp>
#include <sglUnProject.hpp>
#include <sglUnScale.hpp>
#include <sglGeode.hpp>
#include <sglLight.hpp>
#include <sglDirectionalLight.hpp>
#include <sglPositionalLight.hpp>
#include <sglSpotLight.hpp>
#include <sglViewPlatform.hpp>
#include <sglPerspectiveCamera.hpp>

//----------------------------------------------------------------------------
void callback(void *)
{
   cerr << "callback function" << endl;
}

//----------------------------------------------------------------------------
sglScene *buildSceneGraph()
{
   sglScene *root = new sglScene;

   // test name clone
   root->setName("root");

   // test duplicate children
   {
      sglGroup *group1 = new sglGroup;
      group1->setName("group1");
      sglSphereBoundf bsph;
      bsph.setCenter(1.f, 2.2f, -3.3f);
      bsph.setRadius(2.3f);
      root->setBound(&bsph, sglBound::eSTATIC);
      root->addChild(group1);
      root->addChild(group1);
   }

   // test animation nodes
   {
      sglKeyFrameAnimation *kfa = new sglKeyFrameAnimation;
      kfa->setName("keyframe anim");
      for (unsigned int ix=0; ix<4; ++ix)
      {
         sglGeode *geode = new sglGeode;
         geode->setName(string(ix+1,'I'));
         kfa->addChild(geode);
         kfa->setDuration(ix, ix);
      }
      kfa->setRange(1, 2);
      kfa->setMode(sglAnimation::eSWING);  // back and forth
      kfa->setSpeed(2.5f);                 // 2.5 times faster (4 fr. each)
      kfa->setNumCycles(0);                // continuous
      kfa->setState(sglAnimation::eSTART);
      root->addChild(kfa);
   }

   {
      sglTimedAnimation *ta = new sglTimedAnimation;
      ta->setName("timed anim");
      for (unsigned int ix=0; ix<3; ++ix)
      {
         sglGeode *geode = new sglGeode;
         geode->setName(string(ix+1,'X'));
         ta->addChild(geode);
         ta->setDuration(ix, 0.6f);  // display for 0.6 seconds
      }
      ta->setMode(sglAnimation::eCYCLE); // forward only
      ta->setRange();                    // display all
      ta->setSpeed(0.75f);               // .75 of normal speed
                                         //    (0.8 seconds each)
      ta->setNumCycles(1);               // one-shot
      ta->setState(sglAnimation::eSTART);
      root->addChild(ta);
   }

   // billboard
   {
      sglBillboard *bb = new sglBillboard;
      bb->setName("billboard");
      sglVec3d pos(3.0, 0.0, 0.0);    // origin at a point along the x axis
      bb->setPos(pos);
      sglVec3d axis(1.0, 1.0, 0.0);   // will be normalized automatically
      bb->setAxis(axis);  // rotate around the (1, 1, 0) axis
      bb->setMode(sglBillboard::eAXIAL_ROT);
      root->addChild(bb);
   }

   // callback
   {
      sglCallback *cb = new sglCallback;
      cb->setName("callback");
      cb->setCallbackFunc(callback);
      cb->setCallbackData(cb);
      root->addChild(cb);
   }

   // discrim
   {
      sglDiscriminator *disc = new sglDiscriminator;
      disc->setName("discrim");
      disc->setMask(0x02);
      root->addChild(disc);
   }

   // layer
   {
      sglLayer *layer = new sglLayer(sglLayer::eSTENCIL);
      layer->setName("layer");
      layer->setStencilMask(0x02);
      for (int ix=4; ix>=0; --ix)
      {
         sglGeode *geode = new sglGeode();
         layer->addChild(geode);
      }
      root->addChild(layer);
   }

   // LODStateNode
   {
      sglLODStateNode *lodsn = new sglLODStateNode;
      lodsn->setName("lodstatenode");
      lodsn->setRangeScale(1.5f);
      lodsn->setRangeOffset(-2.f);
      root->addChild(lodsn);
   }

   // LOD
   {
      sglLOD *lod_node = new sglLOD;
      lod_node->setName("lod");
      lod_node->setLODState(new sglLODState);

      sglGeode *geode1 = new sglGeode;
      sglGeode *geode2 = new sglGeode;
      sglGeode *geode3 = new sglGeode;

      lod_node->addChild(geode1);     // finest resolution
      lod_node->addChild(geode2);
      lod_node->addChild(geode3);     // coarsest resolution

                                      // select nothing less than 1.0
      lod_node->setRange(0, 1.0f);    // select geode1 between 1.0 and 5.0
      lod_node->setRange(1, 5.0f);    // select geode2 between 5.0 and 10.0
      lod_node->setRange(2, 10.0f);   // select geode3 between 10.0 and 50.0
      lod_node->setRange(3, 50.0f);   // select nothing greater than 50.0


      // halve the computed range and add a bias for this LOD only: has effect
      // of selecting finer resolution children
      sglLODState *lod_state = lod_node->getLODState();
      if (lod_state)
      {
         lod_state->setRangeScale(0.5f);
         lod_state->setRangeOffset(-2.0f);
      }
      lod_node->addChild(lod_node);
   }



   // distance switch
   {
      sglDistanceSwitch *switch_node = new sglDistanceSwitch;
      sglGeode  *geode1 = new sglGeode;
      switch_node->addChild(geode1);
      switch_node->setRange(69.69f);
      switch_node->setCenter(sglVec3f(1.f, 2.f, 3.f));
      switch_node->setName("distance switch");
      root->addChild(switch_node);
   }

   // switch
   {
      sglSwitch *switch_node = new sglSwitch;
      sglGeode  *geode1 = new sglGeode;
      switch_node->addChild(geode1);
      switch_node->setVal(sglSwitch::eOFF);  // select neither
      switch_node->setName("switch");
      root->addChild(switch_node);
   }

   // bit mask switch
   {
      sglBitMaskSwitch *bm = new sglBitMaskSwitch;
      bm->setName("bitswitch");
      sglGeode *geode0 = new sglGeode;
      sglGeode *geode1 = new sglGeode;

      bm->addChild(geode0);
      bm->addChild(geode1);
      bm->setMask(0x3);     // select both
      root->addChild(bm);
   }

   // indexed bit mask switch
   {
      sglIndexedBitMaskSwitch *bm = new sglIndexedBitMaskSwitch;
      bm->setName("indexedbitswitch");
      sglGeode *geode0 = new sglGeode;
      sglGeode *geode1 = new sglGeode;

      bm->addChild(geode0);
      bm->addChild(geode1);
      bm->addMask(0x1);     // select first geode in mask 0
      bm->addMask(0x2);     // select second geode in mask 1
      bm->addMask(0x3);     // select both in mask 2
      bm->setIndex(2);      // select mask 0
      root->addChild(bm);
   }

   // transform
   {
      sglTransformf *tx = new sglTransformf;
      tx->setName("transformf");
      tx->setTranslation(7.0f, 0.450f, 25.0f);
      tx->setRotation((float)(M_PI/3.), 1.0f, 1.0f, 0.0f);
      root->addChild(tx);
   }

   // translate
   {
      sglTranslated *trans = new sglTranslated;
      trans->setName("translated");
      trans->setTranslation(1.00001, 1.00000001, 25.0000000001);
      root->addChild(trans);
   }

   // scale
   {
      sglScalef *scale = new sglScalef;
      scale->setName("scale");
      scale->setScale(6.6666666f);
      root->addChild(scale);
   }

   // default state node
   {
      sglDefaultStateNode *def = new sglDefaultStateNode;
      def->setName("def state");
      vector<sglStatelet*> state;
      state.push_back(new sglLighting(true));
      def->setStatelets(state);
      root->addChild(def);
   }

   // override state node
   {
      sglOverrideStateNode *over = new sglOverrideStateNode;
      over->setName("override state");
      vector<sglStatelet*> state;
      state.push_back(new sglTransparency(true));
      over->setStatelets(state);
      root->addChild(over);
   }

   // unproject
   {
      sglUnProject *unproj = new sglUnProject;
      unproj->setName("unproject");
      unproj->setFixedPixelScale(true);
      root->addChild(unproj);
   }

   // unscale
   {
      sglUnScale *unscale = new sglUnScale;
      unscale->setName("unscale");
      unscale->setFixedPixelScale(true);
      root->addChild(unscale);
   }

   // lights
   {
      sglDirectionalLight *dlite = new sglDirectionalLight;
      dlite->setName("dlite");
      dlite->setOn(false);
      dlite->setAmbient(0.5f, 1.0f, 1.0f, 1.0f);
      dlite->setDiffuse(1.0f, 0.5f, 0.5f, 1.0f);    // reddish light
      dlite->setSpecular(0.5f, 0.5f, 1.0f, 1.0f);
      dlite->setDirection(sglVec3f(0., 0., 1.));    // shines along -Z axis
      root->addChild(dlite);
   }

   {
      sglPositionalLight *plite = new sglPositionalLight;
      plite->setName("plite");
      plite->setOn(false);
      plite->setAmbient(0.5f, 1.0f, 1.0f, 1.0f);
      plite->setDiffuse(1.0f, 0.5f, 0.5f, 1.0f);    // reddish light
      plite->setSpecular(0.5f, 0.5f, 1.0f, 1.0f);
      plite->setPosition(sglVec3f(0., 0., 8.0f));
      plite->setFalloff(0.2f);
      plite->setAttenuation(1.0f, 0.0f, 0.01f);
      plite->setBound(NULL, sglBound::eDYNAMIC);
      root->addChild(plite);
   }

   {
      sglSpotLight *spot = new sglSpotLight;
      spot->setName("spot");
      spot->setOn(false);
      spot->setAmbient(0.5f, 1.0f, 1.0f, 1.0f);
      spot->setDiffuse(1.0f, 0.5f, 0.5f, 1.0f);    // reddish light
      spot->setSpecular(0.5f, 0.5f, 1.0f, 1.0f);
      spot->setPosition(sglVec3f(0., 0., 7.0f));   // move out along +Z axis
      spot->setDirection(sglVec3f(0., 0., -1.0f)); // point along -Z axis
      spot->setCutoff((float)(M_PI*0.25));                 // 45 degrees
      spot->setExponent(3.0f);
      root->addChild(spot);
   }

   // view platform
   {
      sglPerspectiveCamera *camera = new sglPerspectiveCamera;
      camera->setName("persp cam");

      sglViewPlatform *view = new sglViewPlatform(*camera);
      view->setEnableZUp(true);
      view->setName("view platform");
      root->addChild(view);
   }


   return root;
}

//----------------------------------------------------------------------------
int main(int argc, char **argv)
{
   glutInit(&argc, argv);
   glutInitWindowSize(700, 512);
   glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);
   glutCreateWindow("Cloning Testbed");
   sgl::initialize();

   sglScene *root = buildSceneGraph();
   root->preDraw();

   cout << "Original scene graph: " << endl;
   sgluOutputTree(root, cout, 3);

   sglScene *clone = (sglScene*)root->clone(sglObject::eCLONE_NODES_RECURSIVE |
                                            sglObject::eCLONE_USER_DATA);

   cout << endl << "Cloned scene graph: " << endl;
   sgluOutputTree(clone, cout, 3);

   clone->preDraw();
   cout << endl << "Cloned scene graph (after preDraw)" << endl;
   sgluOutputTree(clone, cout, 3);

   return 0;
}
