/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: testCubeMap.cpp
 *  Created: 2 December 2001
 *  Summary: test the GL_ARB_texture_cube_map extension
 *****************************************************************************/

#include <GL/glut.h>
#include <sgl.h>
#include <sglu.h>

#include <sgluMouse.hpp>
#include <sgluSphericalTransform.hpp>

#include <sglScene.hpp>
#include <sglGeode.hpp>
#include <sglStatelet.hpp>
#include <sglTexture2D.hpp>
#include <sglTextureCubeMap.hpp>
#include <sglTexGen.hpp>

#include <sglViewPlatform.hpp>
#include <sglPerspectiveCamera.hpp>
#include <sglView.hpp>
#include <sglAxes.hpp>
#include <sglSphere.hpp>
#include <sglCube.hpp>
#include <sglCylinder.hpp>

#include <sglImageLoader.hpp>

sglView                   *view;
sglScene                  *root;

sgluMouse                 *mouse;
sgluSphericalTransform<float> *view_trans;
sglViewPlatform           *view_platform;
sglPerspectiveCamera      *view_camera;

deque<const sglStatelet*> override_statelets;
sglAntiAliasPoint   aa_point;
sglAntiAliasLine    aa_line;
sglAntiAliasPolygon aa_poly;
sglTextureMatrix    *tmat;


unsigned int win_x = 640, win_y = 480;

//----------------------------------------------------------------------------
void displayFunc()
{
   view->cullAndDraw();
   glutSwapBuffers();

   GLenum glerror = glGetError();
   if (glerror)
   {
      cerr << "glGetError() = 0x" << hex << (int)glerror << dec << endl;
   }
}

//----------------------------------------------------------------------------
void reshapeFunc(int w, int h)
{
   view->setViewport(0, 0, w, h);

   mouse->win_size_x = win_x = w;
   mouse->win_size_y = win_y = h;

   double aspect = (double)w/(double)h;

   view_camera->setFOV(M_PI/4.0, aspect, 0.1, 800.0);
}

//----------------------------------------------------------------------------
void keyboardFunc(unsigned char key, int, int)
{
   switch (key)
   {
      case 27:
         glutDestroyWindow(glutGetWindow());
         exit(1);
         break;

      case 'a':
         aa_point.enable(!aa_point.isEnabled());
         cout << "point aa override = " << aa_point.isEnabled() << endl;
         break;

      case 'A':
         aa_line.enable(!aa_line.isEnabled());
         cout << "line aa override = " << aa_line.isEnabled() << endl;
         break;

      case 'z':
         aa_poly.enable(!aa_poly.isEnabled());
         cout << "polygon aa override = " << aa_poly.isEnabled() << endl;
         break;

      case 'f':
         view->enableFrameRateOutput(!(view->getFrameRateOutputFlag()));
         break;

      case 'd':
         view->setTraversalType(!(view->getDPTraversalFlag()),
                                view->getStatsFlag());
         break;

      case '$':
         view->setTraversalType(view->getDPTraversalFlag(),
                                !(view->getStatsFlag()));
         break;

      case 'o':
         sgluOutputTree(root, cout, 3);
         break;

      case '/':
      case '?':
         cerr << "a - toggle point anti-aliasing" << endl;
         cerr << "A - toggle line anti-aliasing" << endl;
         cerr << "z - toggle polygon anti-aliasing" << endl;
         cerr << "f - toggle frame rate statistics" << endl;
         cerr << "d - toggle between SP and DP traversal" << endl;
         cerr << "$ - toggle scene graph statistics reporting" << endl;
         cerr << "o - output scenegraph" << endl;
         cerr << "? - print help (this list)" << endl;
         cerr << "ECSAPE - exit" << endl;
         break;
   }
}


//----------------------------------------------------------------------------
void specialFunc(int key, int, int)
{
   switch (key)
   {
      case GLUT_KEY_LEFT:
         break;
      case GLUT_KEY_RIGHT:
         break;
      case GLUT_KEY_UP:
         break;
      case GLUT_KEY_DOWN:
         break;
   }
}


//----------------------------------------------------------------------------
void idleFunc()
{
   view_trans->update(mouse);

   sglMat4f mat, mat1;
   mat1.buildIdentity();
   view_platform->computeViewMatrix(mat1);
   mat1[3][0] = mat1[3][1] = mat1[3][2] = 0.f;
   mat.transpose(mat1);

   tmat->setTextureMatrix(&mat);

   glutPostRedisplay();
}

//----------------------------------------------------------------------------
void buildScenegraph(const vector<string> &filename)
{
   // make scene and default state:
   root = new sglScene();

   // add your stuff here
   sglGeode *geode1 = new sglGeode;
   root->addChild(geode1);

   // load all size images
   unsigned int size = 0;
   sglTexture::ExternalFormatEnum format;
   sglTexture::TypeEnum type;
   void *image[6];
   sglTexture2D *tex[6];

   for (unsigned int ix=0; ix<6; ++ix)
   {
      unsigned int width, height;
      image[ix] = sglImageLoader::loadImage(filename[ix],
                                            width, height, format, type);

      tex[ix] = new sglTexture2D;
      tex[ix]->setImage(image[ix], format, type, width, height, false);
      tex[ix]->setWrapMode(sglTexture::eCLAMP_TO_EDGE,
                           sglTexture::eCLAMP_TO_EDGE);

      if (size == 0)
         size = (width > height) ? height : width;

      if (width != height || width != size)
      {
         void *new_image = NULL;
         if (!sglTexture2D::resizeImage(format, type,
                                        width, height, false,
                                        size, size,
                                        image[ix], new_image))
         {
            cerr << "Error: failed to resize image" << endl;
            exit(0);
         }

         image[ix] = new_image;
      }
   }

   // make a cube
   { // positive x
      sglVec3f *normal = new sglVec3f;
      sglVec3f *vertex = new sglVec3f[4];
      sglVec2f *tex_coord = new sglVec2f[4];

      normal[0].set(-1.f, 0.f, 0.f);

      vertex[0].set(400.f,  400.0f,  400.f);
      vertex[1].set(400.f,  400.0f, -400.f);
      vertex[2].set(400.f, -400.0f, -400.f);
      vertex[3].set(400.f, -400.0f,  400.f);

      tex_coord[0].set(0.f, 0.f);
      tex_coord[1].set(1.f, 0.f);
      tex_coord[2].set(1.f, 1.f);
      tex_coord[3].set(0.f, 1.f);

      sglQuadSet *gset = new sglQuadSet;
      gset->setNumPrims(1);
      gset->setCoordList(vertex);
      gset->setNormalList(sglGeoSet::eOVERALL, normal);

      vector<sglStatelet*> state;
      state.push_back(tex[0]);
      gset->setState(0, state, tex_coord);
      geode1->addGeometry(gset);
   }

   { // negative x
      sglVec3f *normal = new sglVec3f;
      sglVec3f *vertex = new sglVec3f[4];
      sglVec2f *tex_coord = new sglVec2f[4];

      normal[0].set(1.f, 0.f, 0.f);

      vertex[0].set(-400.f,  400.0f, -400.f);
      vertex[1].set(-400.f,  400.0f,  400.f);
      vertex[2].set(-400.f, -400.0f,  400.f);
      vertex[3].set(-400.f, -400.0f, -400.f);

      tex_coord[0].set(0.f, 0.f);
      tex_coord[1].set(1.f, 0.f);
      tex_coord[2].set(1.f, 1.f);
      tex_coord[3].set(0.f, 1.f);

      sglQuadSet *gset = new sglQuadSet;
      gset->setNumPrims(1);
      gset->setCoordList(vertex);
      gset->setNormalList(sglGeoSet::eOVERALL, normal);

      vector<sglStatelet*> state;
      state.push_back(tex[1]);
      gset->setState(0, state, tex_coord);
      geode1->addGeometry(gset);
   }

   { // positive y
      sglVec3f *normal = new sglVec3f;
      sglVec3f *vertex = new sglVec3f[4];
      sglVec2f *tex_coord = new sglVec2f[4];

      normal[0].set(0.f, -1.f, 0.f);

      vertex[0].set(-400.f,  400.0f, -400.f);
      vertex[1].set( 400.f,  400.0f, -400.f);
      vertex[2].set( 400.f,  400.0f,  400.f);
      vertex[3].set(-400.f,  400.0f,  400.f);

      tex_coord[0].set(0.f, 0.f);
      tex_coord[1].set(1.f, 0.f);
      tex_coord[2].set(1.f, 1.f);
      tex_coord[3].set(0.f, 1.f);

      sglQuadSet *gset = new sglQuadSet;
      gset->setNumPrims(1);
      gset->setCoordList(vertex);
      gset->setNormalList(sglGeoSet::eOVERALL, normal);

      vector<sglStatelet*> state;
      state.push_back(tex[2]);
      gset->setState(0, state, tex_coord);
      geode1->addGeometry(gset);
   }

   { // negative y
      sglVec3f *normal = new sglVec3f;
      sglVec3f *vertex = new sglVec3f[4];
      sglVec2f *tex_coord = new sglVec2f[4];

      normal[0].set(0.f, 1.f, 0.f);

      vertex[0].set(-400.f, -400.0f,  400.f);
      vertex[1].set( 400.f, -400.0f,  400.f);
      vertex[2].set( 400.f, -400.0f, -400.f);
      vertex[3].set(-400.f, -400.0f, -400.f);

      tex_coord[0].set(0.f, 0.f);
      tex_coord[1].set(1.f, 0.f);
      tex_coord[2].set(1.f, 1.f);
      tex_coord[3].set(0.f, 1.f);

      sglQuadSet *gset = new sglQuadSet;
      gset->setNumPrims(1);
      gset->setCoordList(vertex);
      gset->setNormalList(sglGeoSet::eOVERALL, normal);

      vector<sglStatelet*> state;
      state.push_back(tex[3]);
      gset->setState(0, state, tex_coord);
      geode1->addGeometry(gset);
   }

   { // positive z
      sglVec3f *normal = new sglVec3f;
      sglVec3f *vertex = new sglVec3f[4];
      sglVec2f *tex_coord = new sglVec2f[4];

      normal[0].set(0.f, 0.f, -1.f);

      vertex[0].set(-400.0f,  400.0f, 400.f);
      vertex[1].set( 400.0f,  400.0f, 400.f);
      vertex[2].set( 400.0f, -400.0f, 400.f);
      vertex[3].set(-400.0f, -400.0f, 400.f);

      tex_coord[0].set(0.f, 0.f);
      tex_coord[1].set(1.f, 0.f);
      tex_coord[2].set(1.f, 1.f);
      tex_coord[3].set(0.f, 1.f);

      sglQuadSet *gset = new sglQuadSet;
      gset->setNumPrims(1);
      gset->setCoordList(vertex);
      gset->setNormalList(sglGeoSet::eOVERALL, normal);

      vector<sglStatelet*> state;
      state.push_back(tex[4]);
      gset->setState(0, state, tex_coord);
      geode1->addGeometry(gset);
   }

   { // negative z
      sglVec3f *normal = new sglVec3f;
      sglVec3f *vertex = new sglVec3f[4];
      sglVec2f *tex_coord = new sglVec2f[4];

      normal[0].set(0.f, 0.f, 1.f);

      vertex[0].set( 400.0f,  400.0f, -400.f);
      vertex[1].set(-400.0f,  400.0f, -400.f);
      vertex[2].set(-400.0f, -400.0f, -400.f);
      vertex[3].set( 400.0f, -400.0f, -400.f);

      tex_coord[0].set(0.f, 0.f);
      tex_coord[1].set(1.f, 0.f);
      tex_coord[2].set(1.f, 1.f);
      tex_coord[3].set(0.f, 1.f);

      sglQuadSet *gset = new sglQuadSet;
      gset->setNumPrims(1);
      gset->setCoordList(vertex);
      gset->setNormalList(sglGeoSet::eOVERALL, normal);

      vector<sglStatelet*> state;
      state.push_back(tex[5]);
      gset->setState(0, state, tex_coord);
      geode1->addGeometry(gset);
   }

   // make the shiny thing
   sglTextureCubeMap *cube_map = new sglTextureCubeMap;
   if (!cube_map->setImage(image[0], image[1], image[2], image[3],
                           image[4], image[5],
                           format, type, size, false))
   {
      cerr << "Warning: sglTextureCubeMap::setImage() failed" << endl;
   }

   sglTexGen *tgen = new sglTexGen;
   tgen->setMode(sglTexGen::eS, sglTexGen::eREFLECTION_MAP);
   tgen->setMode(sglTexGen::eT, sglTexGen::eREFLECTION_MAP);
   tgen->setMode(sglTexGen::eR, sglTexGen::eREFLECTION_MAP);

   tmat = new sglTextureMatrix(NULL);

   vector<sglStatelet*> state;
   state.push_back(cube_map);
   state.push_back(tgen);
   state.push_back(tmat);

   sglSphere *object = new sglSphere(48, 24, 5.f);
   object->setState(0, state);
   geode1->addGeometry(object);

   //sglCube *object1 = new sglCube(5.f, 5.f, 5.f);
   //sglCylinder *object1 = new sglCylinder(48, 24, 2.5f, 5.0f, 5.0f, true);
   //object1->setState(0, state);
   //geode1->addGeometry(object1);

   // the camera (frustum)
   double aspect = (double)win_x/(double)win_y;
   view_camera = new sglPerspectiveCamera;
   view_camera->setFOV(M_PI/4.0, aspect, 0.1, 800.0);

   // the view platform
   view_platform = new sglViewPlatform(*view_camera);

   // the view transform
   view_trans = new sgluSphericalTransform<float>();
   view_trans->setMode(sgluSphericalTransform<float>::eY_UP);
   view_trans->setRadius(30.0);
   view_trans->setCOI(0.0, 0.0, 0.0);
   view_trans->setTranslationScale(0.02f);
   view_trans->computeMatrix();
   view_trans->addChild(view_platform);

   // flip the whole scene upside-down
   sglTransformf *trans = new sglTransformf;
   trans->setRotation(0.f, (float)M_PI, 0.f);
   trans->addChild(view_trans);

   root->addChild(trans);

   sglGeode *geode = new sglGeode;
   geode->addGeometry(new sglAxes(7.0f));
   root->addChild(geode);

   root->preDraw();

   view = new sglView;
   view->setScene(root, view_platform);
   view->setViewport(0, 0, win_x, win_y);

   override_statelets.push_back(&aa_point);
   override_statelets.push_back(&aa_line);
   override_statelets.push_back(&aa_poly);
   view->setOverrideStatelets(override_statelets);
}

//----------------------------------------------------------------------------
int main(int argc, char** argv)
{
   glutInit(&argc, argv);

   glutInitWindowSize(640, 480);
   glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);
   glutCreateWindow("testCubeMap");

   sgl::initialize();
   if (!sgl::s_texture_cube_map_capable)
   {
      cerr << "Cube map not supported" << endl;
      exit(0);
   }

   if (argc < 7)
   {
      cerr << "Usage: " << argv[0]
           << " <right,+x> <left,-x> <bottom,+y> <top,-y> "
           << "<front,+z> <back,-z>" << endl;
      exit(0);
   }

   vector<string> filename;
   for (unsigned int ix = 0; ix < 6; ++ix)
   {
      filename.push_back(string(""));
      if (!sglDataLoader::findFile(argv[ix+1], NULL, filename[ix]))
      {
         cerr << "File not found: " << argv[ix+1] << endl;
         exit(0);
      }
   }

   mouse = sgluMouse::initialize();
   buildScenegraph(filename);

   glutReshapeFunc(reshapeFunc);
   glutKeyboardFunc(keyboardFunc);
   glutSpecialFunc(specialFunc);
   glutDisplayFunc(displayFunc);
   glutIdleFunc(idleFunc);

   glutMainLoop();
   return 0;             /* ANSI C requires main to return int. */
}
