/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: testIsectBillboard.cpp
 *   Author: Tom Stimson
 *  Created: 17 April 2000
 *  Summary: test intersection with a lod node
 *****************************************************************************/

#include <GL/glut.h>
#include <sgl.h>
#include <sglu.h>

#include <sgluMouse.hpp>
#include <sgluSphericalTransform.hpp>

#include <sglScene.hpp>
#include <sglTransform.hpp>
#include <sglTranslate.hpp>
#include <sglBillboard.hpp>
#include <sglTextSet.hpp>
#include <sglGeode.hpp>
#include <sglTriangleSet.hpp>

#include <sglSegment.hpp>
#include <sglIntersect.hpp>
#include <sglSphereBound.hpp>

#include <sglViewPlatform.hpp>
#include <sglPerspectiveCamera.hpp>
#include <sglView.hpp>
#include <sglAxes.hpp>

sglView                   *view;
sglScene                  *root;

sgluMouse                 *mouse;
sgluSphericalTransform<float> *view_trans;
sglViewPlatform           *view_platform;
sglPerspectiveCamera      *view_camera;

deque<const sglStatelet*> override_statelets;

bool help_flag = false;

sglBillboard *billboard_node;

sglIntersectf  isector;
sglSegmentf    segment;

bool paused_flag = false;
sglVec3f pos(0.0, -10.0, 0.0);
double heading = 0.0, pitch = 0;
float len = 20.0;
float angle = 0.0;

unsigned int win_x = 640, win_y = 480;

//----------------------------------------------------------------------------
void outputHelp()
{
   glMatrixMode(GL_PROJECTION);
   glPushMatrix();
   glLoadIdentity();
   glOrtho(0, mouse->win_size_x, 0, mouse->win_size_y, -1, 1);

   glMatrixMode(GL_MODELVIEW);
   glLoadIdentity();
   glEnable(GL_BLEND);
   glColor3f(1.f, 1.f, 1.f);

   glRasterPos3i(10,85,0);
   sglBitmapFont::getDefaultFont()->rasterString(
      "/? - toggle help");
   glRasterPos3i(10,70,0);
   sglBitmapFont::getDefaultFont()->rasterString(
      "f - toggle culling front face");
   glRasterPos3i(10,55,0);
   sglBitmapFont::getDefaultFont()->rasterString(
      "b - toggle culling back face");
   glRasterPos3i(10,40,0);
   sglBitmapFont::getDefaultFont()->rasterString(
      "v - toggle billboard intersect");
   glRasterPos3i(10,25,0);
   sglBitmapFont::getDefaultFont()->rasterString(
      "p - toggle (pause) segment motion");
   glDisable(GL_BLEND);
   glMatrixMode(GL_PROJECTION);
   glPopMatrix();
   glMatrixMode(GL_MODELVIEW);
   glEnable(GL_DEPTH_TEST);
}

//----------------------------------------------------------------------------
void preDrawFunc(void *)
{
   glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
   glEnable(GL_DEPTH_TEST);

   if (!paused_flag)
   {
      heading += 0.005;
      if (heading > M_PI/4.0) heading -= M_PI/2.0;
   }

   segment.makePolar(pos, heading, pitch, len);

   isector.setSegment(segment);
   isector.setResultEnum(sglBound::eOUTSIDE);
   sglMat4f view_mat;
   view_platform->computeViewMatrix(view_mat);
   isector.setModelViewMatrix(view_mat);
   sglBound::IntersectResultEnum result = root->intersect(isector);
   glColor3f(0.0f, 1.0f, 0.0f);
   if (result == sglBound::eINTERSECT)
   {
      sglVec3f endpoint = isector.getSegment().getEndPoint();
      sglMat4f matrix = isector.getTransformMatrix();
      sglVec3f normal = isector.getNormal();
      sglVec4f n;

      mul(normal, 0.f, matrix, n);

      // undo the affects of the view matrix
      sglVec4f n1;
      matrix.inverse(view_mat);
      mul(sglVec3f(n[0], n[1], n[2]), 0.f, matrix, n1);

      sglSegmentf nseg(endpoint, sglVec3f(n1[0], n1[1], n1[2]), 4.0f);
      glColor3f(0.f, 0.f, 1.f);

      nseg.drawImmediate();

      glColor3f(1.0f, 0.4f, 0.0f);
   }
   (isector.getSegment()).drawImmediate();
}

//----------------------------------------------------------------------------
void displayFunc()
{
   view->cullAndDraw();

   if (help_flag)
   {
      outputHelp();
   }

   glutSwapBuffers();
}

//----------------------------------------------------------------------------
void reshapeFunc(int w, int h)
{
   view->setViewport(0, 0, w, h);

   mouse->win_size_x = win_x = w;
   mouse->win_size_y = win_y = h;

   double aspect = (double)w/(double)h;

   view_camera->setFOV(M_PI/4.0, aspect, 0.1, 200.0);
}

//----------------------------------------------------------------------------
void keyboardFunc(unsigned char key, int, int)
{
   switch (key)
   {
      case 27:
         glutDestroyWindow(glutGetWindow());
         exit(1);
         break;

      case 'f':
         if (isector.getMode() & sglProbe::eCULL_FRONT)
         {
            cout << "Intersecting with front facing polygons" << endl;
            isector.setMode(isector.getMode() & ~sglProbe::eCULL_FRONT);
         }
         else
         {
            cout << "Not intersecting with front facing polygons" << endl;
            isector.setMode(isector.getMode() | sglProbe::eCULL_FRONT);
         }
         break;

      case 'b':
         if (isector.getMode() & sglProbe::eCULL_BACK)
         {
            cout << "Intersecting with back facing polygons" << endl;
            isector.setMode(isector.getMode() & ~sglProbe::eCULL_BACK);
         }
         else
         {
            cout << "Not intersecting with back facing polygons" << endl;
            isector.setMode(isector.getMode() | sglProbe::eCULL_BACK);
         }
         break;

      case 'v':
         if (isector.getMode() & sglProbe::eBILLBOARD_VIEW)
         {
            cout << "Not intersecting with billboards" << endl;
            isector.setMode(isector.getMode() & ~sglProbe::eBILLBOARD_VIEW);
         }
         else
         {
            cout << "Intersecting with billboards" << endl;
            isector.setMode(isector.getMode() | sglProbe::eBILLBOARD_VIEW);
         }
         break;

      case 'p':
         paused_flag = !paused_flag;
         break;

      case '?':
      case '/':
         help_flag = !help_flag;
         cout << "help_flag = " << help_flag << endl;
         break;
   }
}


//----------------------------------------------------------------------------
void specialFunc(int key, int, int)
{
   switch (key)
   {
      case GLUT_KEY_LEFT:
         break;
      case GLUT_KEY_RIGHT:
         break;
      case GLUT_KEY_UP:
         break;
      case GLUT_KEY_DOWN:
         break;
   }
}


//----------------------------------------------------------------------------
void idleFunc()
{
   view_trans->update(mouse);
   displayFunc();
}

//----------------------------------------------------------------------------
sglGroup* makeTriangleSet()
{
   sglGroup *group = new sglGroup;
   sglTextSet  *text  = new sglTextSet;
   text->setString("TriangleSet");

   sglGeode *geode = new sglGeode();
   geode->addGeometry(text);

   sglTriangleSet *gset = new sglTriangleSet();

   sglVec3f *vertex = new sglVec3f[9];
   sglVec4f *color  = new sglVec4f[3];

   for (int j=0; j<3; j++)
   {
      vertex[j*3 + 0].set(-6.0f + j*6.0f, -6.0f, 0.f);
      vertex[j*3 + 1].set(-2.0f + j*6.0f, -6.0f, 0.f);
      vertex[j*3 + 2].set(-4.0f + j*6.0f, 6.0f, 0.f);

      color[j].set((float)j/3.0f, 1.0f - (float)j/3.0f, 0.5f, 1.0f);
   }

   gset->setCoordList(vertex);
   gset->setColorList(sglGeoSet::ePER_PRIM, color);
   gset->setNumPrims(3);

   geode->addGeometry(gset);
   group->addChild(geode);

   sglTransformd *txnode = new sglTransformd;
   txnode->setRotation(M_PI, 0., 1., 0.);
   txnode->setTranslation(0., 4., 0.);
   txnode->addChild(geode);
   group->addChild(txnode);

   return group;
}

//----------------------------------------------------------------------------
void buildScenegraph()
{
   // make scene and default state:
   root = new sglScene();

   // add your stuff here
   billboard_node = new sglBillboard();
   billboard_node->addChild(makeTriangleSet());
   billboard_node->setMode(sglBillboard::ePOINT_ROT_EYE);

   root->addChild(billboard_node);

   // ================================================================
   isector.setMode(sglProbe::eGEODE|sglProbe::eGEOMETRY|sglProbe::ePRIMITIVE|
                   sglProbe::eTRANSFORM|sglProbe::eNORMAL|
                   sglProbe::eCULL_BACK|sglProbe::eBILLBOARD_VIEW);
   isector.setDiscriminatorMask(~0xffffffff);
   // ================================================================

   // the camera (frustum)
   double aspect = (double)win_x/(double)win_y;
   view_camera = new sglPerspectiveCamera;
   view_camera->setFOV(M_PI/4.0, aspect, 0.1, 200.0);

   // the view platform
   view_platform = new sglViewPlatform(*view_camera);

   // the view transform
   view_trans = new sgluSphericalTransform<float>();
   view_trans->setMode(sgluSphericalTransform<float>::eZ_UP);
   view_trans->setRadius(30.0);
   view_trans->setCOI(0.0, 0.0, 0.0);
   view_trans->setTranslationScale(0.02f);
   view_trans->computeMatrix();
   view_trans->addChild(view_platform);
   root->addChild(view_trans);

   sglGeode *geode = new sglGeode;
   geode->addGeometry(new sglAxes(2.0f));
   root->addChild(geode);

   root->preDraw();

   view = new sglView;
   view->setScene(root, view_platform);
   view->setViewport(0, 0, win_x, win_y);
   view->setDrawCallback(preDrawFunc, NULL);

   override_statelets.push_back(new sglCullMode(sglCullMode::eOFF));
   override_statelets.push_back(new sglPolyFrontStyle(sglPolyStyle::eFILLED));
   override_statelets.push_back(new sglPolyBackStyle(sglPolyStyle::eLINE));
   override_statelets.push_back(new sglShadeModel(sglShadeModel::eFLAT));
   view->setOverrideStatelets(override_statelets);
}

//----------------------------------------------------------------------------
int main(int argc, char** argv)
{
   glutInit(&argc, argv);

   glutInitWindowSize(700, 512);
   glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);
   glutCreateWindow("testIsectBillboard");

   sgl::initialize();
   mouse = sgluMouse::initialize();
   buildScenegraph();

   glutReshapeFunc(reshapeFunc);
   glutKeyboardFunc(keyboardFunc);
   glutSpecialFunc(specialFunc);
   glutDisplayFunc(displayFunc);
   glutIdleFunc(idleFunc);

   glutMainLoop();
   return 0;             /* ANSI C requires main to return int. */
}
