/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: testIsectDiscrim.cpp
 *  Created: 19 April 1999
 *  Summary: GLUT-based sgl testbed for discriminated intersections
 *****************************************************************************/

#include <GL/glut.h>
#include <sgl.h>
#include <sglu.h>

#include <sgluMouse.hpp>
#include <sgluSphericalTransform.hpp>

#include <sglScene.hpp>
#include <sglDiscriminator.hpp>
#include <sglScale.hpp>
#include <sglGeode.hpp>
#include <sglTriangleSet.hpp>
#include <sglTriangleStripSet.hpp>

#include <sglSegment.hpp>
#include <sglIntersect.hpp>
#include <sglSphereBound.hpp>

#include <sglViewPlatform.hpp>
#include <sglPerspectiveCamera.hpp>
#include <sglView.hpp>
#include <sglAxes.hpp>

sglView                   *view;
sglScene                  *root;

sgluMouse                 *mouse;
sgluSphericalTransform<float> *view_trans;
sglViewPlatform           *view_platform;
sglPerspectiveCamera      *view_camera;

deque<const sglStatelet*> override_statelets;

sglHighlight *highlight;

sglScalef  *scale;

sglIntersectf  isector;
sglSegmentf segment;

bool paused_flag = false;
sglVec3f pos(0.0, -0.0, 0.0);
double heading = 0.0, pitch = 0;
float len = 30.0;

unsigned int win_x = 640, win_y = 480;

//----------------------------------------------------------------------------
void preDrawFunc(void *)
{
   glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
   glEnable(GL_DEPTH_TEST);

   if (!paused_flag)
   {
      heading += 0.005;
      if (heading > M_PI/2.0) heading -= M_PI;
   }

   scale->setScale(1.0f + 0.5f*sin(heading));

   segment.makePolar(pos, heading, pitch, len);

   isector.setSegment(segment);
   isector.setResultEnum(sglBound::eOUTSIDE);
   sglBound::IntersectResultEnum result = root->intersect(isector);
   glColor3f(0.0f, 1.0f, 0.0f);

   if (isector.getResultEnum() == sglBound::eINTERSECT ||
       result == sglBound::eINTERSECT) // huh?)
   {

      sglVec3f endpoint = isector.getSegment().getEndPoint();
      sglMat4f matrix = isector.getTransformMatrix();
      sglVec3f normal = isector.getNormal();
      sglVec4f n;

      mul(normal, 0.f, matrix, n);

      sglSegmentf nseg(endpoint, sglVec3f(n[0], n[1], n[2]), 4.0f);
      glColor3f(0.f, 0.f, 1.f);

      nseg.drawImmediate();

      highlight->enable(true);
      glColor3f(1.f, 0.f, 0.f);
   }
   else
   {
      highlight->enable(false);
   }
   (isector.getSegment()).drawImmediate();
}

//----------------------------------------------------------------------------
void displayFunc()
{
   view->cullAndDraw();
   glutSwapBuffers();
}

//----------------------------------------------------------------------------
void reshapeFunc(int w, int h)
{
   view->setViewport(0, 0, w, h);

   mouse->win_size_x = win_x = w;
   mouse->win_size_y = win_y = h;

   double aspect = (double)w/(double)h;

   view_camera->setFOV(M_PI/3.0, aspect, 0.1, 200.0);
}

//----------------------------------------------------------------------------
void keyboardFunc(unsigned char key, int, int)
{
   switch (key)
   {
      case 27:
         glutDestroyWindow(glutGetWindow());
         exit(1);
         break;

      case 'f':
         if (isector.getMode() & sglProbe::eCULL_FRONT)
         {
            cout << "Intersecting with front facing polygons" << endl;
            isector.setMode(isector.getMode() & ~sglProbe::eCULL_FRONT);
         }
         else
         {
            cout << "Not intersecting with front facing polygons" << endl;
            isector.setMode(isector.getMode() | sglProbe::eCULL_FRONT);
         }
         break;

      case 'b':
         if (isector.getMode() & sglProbe::eCULL_BACK)
         {
            cout << "Intersecting with back facing polygons" << endl;
            isector.setMode(isector.getMode() & ~sglProbe::eCULL_BACK);
         }
         else
         {
            cout << "Not intersecting with back facing polygons" << endl;
            isector.setMode(isector.getMode() | sglProbe::eCULL_BACK);
         }
         break;

      case 'p':
         paused_flag = !paused_flag;
         break;

      case '?':
      case '/':
         cout << "f - toggle intersection with front faces" << endl;
         cout << "b - toggle intersection with back faces" << endl;
         cout << "p - toggle (pause) segment motion" << endl;
         cout << "ESCAPE - exit program" << endl;
         break;
   }
}


//----------------------------------------------------------------------------
void specialFunc(int key, int, int)
{
   switch (key)
   {
      case GLUT_KEY_LEFT:
         break;
      case GLUT_KEY_RIGHT:
         break;
      case GLUT_KEY_UP:
         break;
      case GLUT_KEY_DOWN:
         break;
   }
}


//----------------------------------------------------------------------------
void idleFunc()
{
   view_trans->update(mouse);
   displayFunc();
}

//----------------------------------------------------------------------------
sglGroup *makeTris()
{
   sglGroup *group = new sglGroup();

   sglDiscriminator *disc = new sglDiscriminator();
   disc->setMask(0x00000001);
   group->addChild(disc);

   scale = new sglScalef;
   scale->setScale(1.0f);
   group->addChild(scale);

   sglGeode *geode = new sglGeode();
   scale->addChild(geode);

   sglGeode *trisgeode = new sglGeode();
   disc->addChild(trisgeode);

   sglTriangleSet *gset1 = new sglTriangleSet();
   gset1->setNumPrims(10);

   sglVec3f *vertices = new sglVec3f[30];
   unsigned int index = 0;

   for (int i=0; i<10; i++)
   {
      const double delta = 8.0*M_PI/180.0;
      double angle = delta + 18.0*i*M_PI/180.0;

      vertices[index++].set( 10.0*cos(angle - delta),
                             10.0*sin(angle - delta),
                            -1.0);
      vertices[index++].set( 10.0*cos(angle),
                             10.0*sin(angle),
                             2.0);
      vertices[index++].set( 10.0*cos(angle + delta),
                             10.0*sin(angle + delta),
                            -1.0);
   }

   sglVec4f *color = new sglVec4f[1];
   color[0].set(1.0f, 1.0f, 0.0f, 1.0f);

   gset1->setCoordList(vertices);
   gset1->setColorList(sglGeoSet::eOVERALL, color);
   trisgeode->addGeometry(gset1);

   sglTriangleStripSet *gset2 = new sglTriangleStripSet();
   gset2->setNumPrims(2);
   unsigned int *length = new unsigned int[2];
   length[0] = 10;
   length[1] = 10;
   gset2->setPrimLengths(length);
   sglVec3f *vertices2 = new sglVec3f[20];

   vector<sglStatelet*> state;
   highlight = new sglHighlight(false);
   state.push_back(highlight);
   gset2->setState(0, state);

   int ii;
   index = 0;
   for (ii=0; ii<5; ii++)
   {
      const double delta = 8.0*M_PI/180.0;
      double angle = delta + 18.0*ii*M_PI/180.0;

      vertices2[index++].set( 10.0*cos(angle),
                              10.0*sin(angle),
                              -2.0);
      vertices2[index++].set( 10.0*cos(angle),
                              10.0*sin(angle),
                              2.0);
   }
   for (ii=5; ii<10; ii++)
   {
      const double delta = 8.0*M_PI/180.0;
      double angle = delta + 18.0*ii*M_PI/180.0;

      vertices2[index++].set( 10.0*cos(angle),
                              10.0*sin(angle),
                              2.0);
      vertices2[index++].set( 10.0*cos(angle),
                              10.0*sin(angle),
                              -2.0);
   }
   gset2->setCoordList(vertices2);

   geode->addGeometry(gset2);
   return group;
}

//----------------------------------------------------------------------------
void buildScenegraph()
{
   // make scene and default state:
   root = new sglScene();

   // add your stuff here
   root->addChild(makeTris());

   // ================================================================
   isector.setMode(sglProbe::eGEODE|sglProbe::eGEOMETRY|
                   sglProbe::ePRIMITIVE|
                   sglProbe::eTRANSFORM|sglProbe::eNORMAL|
                   sglProbe::eCULL_BACK);
   isector.setDiscriminatorMask(~0x000000001);  // mask out the tris
   // ================================================================

   // the camera (frustum)
   double aspect = (double)win_x/(double)win_y;
   view_camera = new sglPerspectiveCamera;
   view_camera->setFOV(M_PI/3.0, aspect, 0.1, 200.0);

   // the view platform
   view_platform = new sglViewPlatform(*view_camera);

   // the view transform
   view_trans = new sgluSphericalTransform<float>();
   view_trans->setMode(sgluSphericalTransform<float>::eZ_UP);
   view_trans->setRadius(30.0);
   view_trans->setCOI(0.0, 0.0, 0.0);
   view_trans->setTranslationScale(0.02f);
   view_trans->computeMatrix();
   view_trans->addChild(view_platform);
   root->addChild(view_trans);

   sglGeode *geode = new sglGeode;
   geode->addGeometry(new sglAxes(2.0f));
   root->addChild(geode);

   root->preDraw();

   view = new sglView;
   view->setScene(root, view_platform);
   view->setViewport(0, 0, win_x, win_y);
   view->setDrawCallback(preDrawFunc, NULL);

   override_statelets.push_back(new sglCullMode(sglCullMode::eOFF));
   override_statelets.push_back(new sglPolyFrontStyle(sglPolyStyle::eFILLED));
   override_statelets.push_back(new sglPolyBackStyle(sglPolyStyle::eLINE));
   override_statelets.push_back(new sglShadeModel(sglShadeModel::eFLAT));
   view->setOverrideStatelets(override_statelets);
}

//----------------------------------------------------------------------------
int main(int argc, char** argv)
{
   glutInit(&argc, argv);

   glutInitWindowSize(700, 512);
   glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);
   glutCreateWindow("testIsectDiscrim");

   sgl::initialize();
   mouse = sgluMouse::initialize();
   buildScenegraph();

   glutReshapeFunc(reshapeFunc);
   glutKeyboardFunc(keyboardFunc);
   glutSpecialFunc(specialFunc);
   glutDisplayFunc(displayFunc);
   glutIdleFunc(idleFunc);

   glutMainLoop();
   return 0;             /* ANSI C requires main to return int. */
}
