/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: testLights.cpp
 *   Author: Scott McMillan
 *  Created: 11 December 2000
 *  Summary: test various types of light nodes
 *****************************************************************************/

#include <GL/glut.h>
#include <sgl.h>
#include <sglu.h>
#include <stdio.h>

#include <sgluMouse.hpp>
#include <sgluSphericalTransform.hpp>

#include <sglScene.hpp>
#include <sglSwitch.hpp>
#include <sglTransform.hpp>
#include <sglDirectionalLight.hpp>
#include <sglPositionalLight.hpp>
#include <sglSpotLight.hpp>
#include <sglBitmapFont.hpp>
#include <sglGeode.hpp>
#include <sglGeoSet.hpp>
#include <sglPointSet.hpp>
#include <sglLineSet.hpp>
#include <sglTriangleStripSet.hpp>
#include <sglTriangleFanSet.hpp>
#include <sglMonoIndexedTriangleStripSet.hpp>
#include <sglMaterial.hpp>

#include <sglViewPlatform.hpp>
#include <sglPerspectiveCamera.hpp>
#include <sglView.hpp>
#include <sglAxes.hpp>
#include <sglSphere.hpp>

sglView                   *view;
sglScene                  *root;

sgluMouse                 *mouse;
sgluSphericalTransform<float> *view_trans;
sglViewPlatform           *view_platform;
sglPerspectiveCamera      *view_camera;

deque<const sglStatelet*> override_statelets;
sglPolyFrontStyle pfs;
sglPolyBackStyle pbs;

sglSwitch      *switch_node;
sglTransformf  *transformf;

unsigned int    state_mask = 0;

sglDirectionalLight *dlite = NULL;
sglPositionalLight *plite = NULL;
sglSpotLight *spot = NULL;

float constant_atten = 1.0f;
float linear_atten = 0.0f;
float quadratic_atten = 0.01f;

float far_clip = 200.0f;

double angle[3] = {0.0, 0.0, 0.0};

unsigned int win_x = 640, win_y = 480;

bool paused_flag = false;

sglVec4f amb_light(0.75f, 0.75f, 0.75f, 1.0f);
sglVec4f no_light(0.30f, 0.30f, 0.30f, 1.0f);

int material_index = 0;
char *material_desc[] = {"material: default",
                         "material: diffuse only, matte",
                         "material: diffuse+specular, med. shine",
                         "material: diffuse+specular, high shine",
                         "material: color material (diffuse+ambient)",
                         "material: emissive+diffuse"};

bool global_ambient_flag = false;
bool local_light_flag = false;
bool light_color_flag = true;
int light_index = 0;
char *light_desc[] = {"light: directional (diffuse only)",
                      "light: positional (diffuse+ambient)",
                      "light: spot (diffuse)"};

//----------------------------------------------------------------------------
void outputDescription()
{
   glDisable(GL_DEPTH_TEST);
   glMatrixMode(GL_PROJECTION);
   glPushMatrix();
   glLoadIdentity();
   glOrtho(0, mouse->win_size_x, 0, mouse->win_size_y, -1, 1);

   glMatrixMode(GL_MODELVIEW);
   glLoadIdentity();

   glEnable(GL_BLEND);
   glColor3f(1.0f, 1.0f, 0.5f);
   glRasterPos2i(1, 15);
   sglBitmapFont::getDefaultFont()->rasterString(
      material_desc[material_index]);
   glRasterPos2i(1, 30);
   sglBitmapFont::getDefaultFont()->rasterString(
      light_desc[light_index]);
   glDisable(GL_BLEND);

   glMatrixMode(GL_PROJECTION);
   glPopMatrix();
   glMatrixMode(GL_MODELVIEW);
   glEnable(GL_DEPTH_TEST);
}

//----------------------------------------------------------------------------
void displayFunc()
{
   view->cullAndDraw();

   outputDescription();

   glutSwapBuffers();
}

//----------------------------------------------------------------------------
void reshapeFunc(int w, int h)
{
   view->setViewport(0, 0, w, h);

   mouse->win_size_x = win_x = w;
   mouse->win_size_y = win_y = h;

   double aspect = (double)w/(double)h;

   view_camera->setFOV(M_PI/4.0, aspect, 0.1, 200.0);
}

//----------------------------------------------------------------------------
void keyboardFunc(unsigned char key, int, int)
{
   switch (key)
   {
      case 27:
         glutDestroyWindow(glutGetWindow());
         exit(1);
         break;

      case 'F':
         if (pfs.getStyle() == sglPolyStyle::eFILLED)
         {
            pfs.setStyle(sglPolyStyle::ePOINT);
            pbs.setStyle(sglPolyStyle::ePOINT);
         }
         else if (pfs.getStyle() == sglPolyStyle::ePOINT)
         {
            pfs.setStyle(sglPolyStyle::eLINE);
            pbs.setStyle(sglPolyStyle::eLINE);
         }
         else if (pfs.getStyle() == sglPolyStyle::eLINE)
         {
            pfs.setStyle(sglPolyStyle::eFILLED);
            pbs.setStyle(sglPolyStyle::eFILLED);
         }

         cout << "polygon style = " << (int) pfs.getStyle() << endl;
         break;

      case 'a':
         global_ambient_flag = !global_ambient_flag;
         if (global_ambient_flag)
         {
            glLightModelfv(GL_LIGHT_MODEL_AMBIENT, amb_light.getPtr());
            cerr << "global ambient: enabled" << endl;
         }
         else
         {
            glLightModelfv(GL_LIGHT_MODEL_AMBIENT, no_light.getPtr());
            cerr << "global ambient: disabled" << endl;
         }
         break;

      case 'm':
         local_light_flag = !local_light_flag;
         if (local_light_flag)
         {
            glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER, GL_TRUE);
            cerr << "local light model: enabled" << endl;
         }
         else
         {
            glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER, GL_FALSE);
            cerr << "local light model: disabled" << endl;
         }
         break;

      case 'h':
         if (light_color_flag)
         {
            dlite->setDiffuse(1.0, 1.0, 1.0, 1.0);
            plite->setDiffuse(1.0, 1.0, 1.0, 1.0);
            spot->setDiffuse(1.0, 1.0, 1.0, 1.0);
         }
         else
         {
            dlite->setDiffuse(1.0, 0.5, 0.5, 1.0);
            plite->setDiffuse(0.0, 1.0, 0.0, 1.0);
            spot->setDiffuse(0.0, 0.0, 1.0, 1.0);
         }
         light_color_flag = !light_color_flag;
         break;

      case 'c':
         if (constant_atten > 0.1)
            constant_atten -= 0.1f;
         cerr << "constant_atten = " << constant_atten << endl;
         plite->setAttenuation(constant_atten, linear_atten, quadratic_atten);
         break;
      case 'C':
         constant_atten += 0.1f;
         cerr << "constant_atten = " << constant_atten << endl;
         plite->setAttenuation(constant_atten, linear_atten, quadratic_atten);
         break;

      case 'l':
         linear_atten -= 0.05f;
         if (linear_atten < 0.0f)
            linear_atten = 0.0f;
         cerr << "linear_atten = " << linear_atten << endl;
         plite->setAttenuation(constant_atten, linear_atten, quadratic_atten);
         break;
      case 'L':
         linear_atten += 0.05f;
         cerr << "linear_atten = " << linear_atten << endl;
         plite->setAttenuation(constant_atten, linear_atten, quadratic_atten);
         break;

      case 'q':
         quadratic_atten -= 0.01f;
         if (quadratic_atten < 0.0f)
            quadratic_atten = 0.0f;
         cerr << "quadratic_atten = " << quadratic_atten << endl;
         plite->setAttenuation(constant_atten, linear_atten, quadratic_atten);
         break;
      case 'Q':
         quadratic_atten += 0.01f;
         cerr << "quadratic_atten = " << quadratic_atten << endl;
         plite->setAttenuation(constant_atten, linear_atten, quadratic_atten);
         break;

      case 'p':
         paused_flag = !paused_flag;
         break;

      case 'f':
         view->enableFrameRateOutput(!(view->getFrameRateOutputFlag()));
         break;

      case 'd':
         view->setTraversalType(!(view->getDPTraversalFlag()),
                               view->getStatsFlag());
         break;

      case '$':
         view->setTraversalType(view->getDPTraversalFlag(),
                               !(view->getStatsFlag()));
         break;

      case 'o':
         sgluOutputTree(root, cout, 3);
         break;

      case '/':
      case '?':
         cerr << "F - cycle through polygon modes" << endl;
         cerr << "a - toggle global ambient" << endl;
         cerr << "m - toggle local light model" << endl;
         cerr << "h - toggle light colors" << endl;
         cerr << "c/C - dec/inc positional light constant attenuation" << endl;
         cerr << "l/L - dec/inc positional light linear attenuation" << endl;
         cerr << "q/Q - dec/inc positional light quadratic attenuation"
              << endl;
         cerr << "p - toggle light animation" << endl;
         cerr << "f - toggle frame rate statistics" << endl;
         cerr << "d - toggle between SP and DP traversal" << endl;
         cerr << "$ - toggle scene graph statistics reporting" << endl;
         cerr << "o - output scenegraph" << endl;
         cerr << "? - print help (this list)" << endl;
         cerr << "ECSAPE - exit" << endl;
         break;
   }
}


//----------------------------------------------------------------------------
void specialFunc(int key, int, int)
{
   int val = 0;
   switch (key)
   {
      case GLUT_KEY_LEFT:
         if (state_mask == 0)
         {
            material_index = 5;
            state_mask = 0x10;
         }
         else
         {
            --material_index;
            state_mask /= 2;
         }
         view->setStateMask(state_mask);
         break;

      case GLUT_KEY_RIGHT:
         ++material_index;
         if (state_mask == 0)
            state_mask = 0x1;
         else
         {
            state_mask *= 2;
            if (state_mask > 0x10)
            {
               material_index = 0;
               state_mask = 0;
            }
         }
         view->setStateMask(state_mask);
         break;

      case GLUT_KEY_UP:
         light_index++;
         val = switch_node->getVal() + 1;
         if ((unsigned int)val == switch_node->getNumChildren())
         {
            val = 0;
            light_index = 0;
         }
         switch_node->setVal(val);
         break;
      case GLUT_KEY_DOWN:
         light_index--;
         val = switch_node->getVal() - 1;
         if (val < 0)
         {
            light_index = (int)(switch_node->getNumChildren() - 1);
         }
         switch_node->setVal(light_index);
         break;
   }
}


//----------------------------------------------------------------------------
void idleFunc()
{
   view_trans->update(mouse);

   if (!paused_flag)
   {
      angle[0] += 0.02;
      angle[1] += 0.043;
      angle[2] += 0.11;

      transformf->setRotation(angle[0], angle[1], angle[2]);
   }

   displayFunc();
}

//----------------------------------------------------------------------------

sglGroup *makeRoom()
{
   // Make one wall and transform it to the other walls
   sglVec4f *color  = new sglVec4f[1]; // overall
   color[0].set(1.0, 1.0, 1.0, 1.0);

   sglVec3f *normal = new sglVec3f[1]; // per prim
   normal[0].set(0.0f, 0.0f, 1.0f);

   const unsigned int size = 24;
   sglVec3f *vertex = new sglVec3f[size*(2*size + 2)];
   unsigned int *length = new unsigned int[size];

   unsigned int index = 0;
   for (unsigned int ix = 0; ix < size; ix++)
   {
      float x1 = 20.0f*((float)ix/(float)size - 0.5f);
      float x2 = 20.0f*((float)(ix + 1)/(float)size - 0.5f);

      length[ix] = 2*size + 2;

      for (unsigned int iy = 0; iy <= size; ++iy)
      {
         float y = 20.0f*((float)iy/(float)size - 0.5f);

         vertex[index].set(x1, y, -10.0f);
         index++;
         vertex[index].set(x2, y, -10.0f);
         index++;
      }
   }

   sglTriangleStripSet *gset_floor = new sglTriangleStripSet;
   gset_floor->setNumPrims(size);
   gset_floor->setPrimLengths(length);
   gset_floor->setCoordList(vertex);
   gset_floor->setNormalList(sglGeoSet::eOVERALL, normal);
   gset_floor->setColorList(sglGeoSet::eOVERALL, color);

   sglMaterial *mat = new sglMaterial;
   mat->setAmbient(color[0][0], color[0][1], color[0][2]);
   mat->setDiffuse(color[0][0], color[0][1], color[0][2]);

   vector<sglStatelet*> tex_state;
   tex_state.push_back(new sglLighting(true));
   tex_state.push_back(mat);

   gset_floor->setState(0, tex_state);

   // -----------------------------------------------------------------
   sglGroup *group = new sglGroup;
   sglGeode *geode = new sglGeode;
   geode->addGeometry(gset_floor);
   group->addChild(geode);          // floor

   sglTransformf *trans_node = new sglTransformf;
   trans_node->setRotation((float)M_PI*0.5f, 1.0f, 0.0f, 0.0f);
   trans_node->addChild(geode);
   group->addChild(trans_node);

   trans_node = new sglTransformf;
   trans_node->setRotation((float)M_PI, 1.0f, 0.0f, 0.0f);
   trans_node->addChild(geode);
   group->addChild(trans_node);

   trans_node = new sglTransformf;
   trans_node->setRotation(-(float)M_PI*0.5f, 1.0f, 0.0f, 0.0f);
   trans_node->addChild(geode);
   group->addChild(trans_node);

   trans_node = new sglTransformf;
   trans_node->setRotation((float)M_PI*0.5f, 0.0f, 1.0f, 0.0f);
   trans_node->addChild(geode);
   group->addChild(trans_node);

   trans_node = new sglTransformf;
   trans_node->setRotation(-(float)M_PI*0.5f, 0.0f, 1.0f, 0.0f);
   trans_node->addChild(geode);
   group->addChild(trans_node);

   return group;
}

//----------------------------------------------------------------------------

sglDrawable *makeSphere(float radius,
                        unsigned int num_slices, unsigned int num_stacks)
{
   if (num_slices<3 || num_stacks <2 || radius <= 0.0)
      return NULL;

   sglSphere *sphere = new sglSphere(num_slices, num_stacks, radius);

   ++num_slices;
   ++num_stacks;

   sglVec2f *tcoords = new sglVec2f[num_slices*num_stacks];
   sglVec4f *color   = new sglVec4f[num_slices*num_stacks];

   unsigned int index = 0;
   for (unsigned int iy=0; iy<num_stacks; ++iy)
   {
      for (unsigned int ix=0; ix<num_slices; ++ix)
      {
         color[index].set(1.0f - (float)ix/((float)num_slices - 1),
                          (float)ix/((float)num_slices - 1),
                          (float)iy/((float)num_stacks - 1),
                          1.0f);
         tcoords[index].set((float)ix/(float)(num_slices - 1),
                            (float)iy/(float)(num_stacks - 1));
         ++index;
      }
   }

   sphere->setColorList(sglGeoSet::ePER_VERTEX, color);

   // ----------- make the states for different materials  ------------------
   {
      // default material
      vector<sglStatelet*> state;
      state.push_back(new sglLighting(true));
      state.push_back(new sglMaterial);
      sphere->setState(0, state); //, tcoords);
   }

   {
      // material - diffuse only
      sglMaterial *mat = new sglMaterial;
      mat->setAmbient(0.f, 0.f, 0.f);
      mat->setDiffuse(0.5f, 0.5f, 0.5f);
      mat->setSpecular(0.f, 0.f, 0.f);
      mat->setEmission(0.f, 0.f, 0.f);
      mat->setShininess(0.0f);

      vector<sglStatelet*> state;
      state.push_back(new sglLighting(true));
      state.push_back(mat);
      sphere->setState(0x1, state); //, tcoords);
   }

   {
      // material - diffuse and specular
      sglMaterial *mat = new sglMaterial;
      mat->setAmbient(0.f, 0.f, 0.f);
      mat->setDiffuse(0.5f, 0.5f, 0.5f);
      mat->setSpecular(1.f, 1.f, 1.f);
      mat->setEmission(0.f, 0.f, 0.f);
      mat->setShininess(5.0f);

      vector<sglStatelet*> state;
      state.push_back(new sglLighting(true));
      state.push_back(mat);
      sphere->setState(0x2, state); //, tcoords);
   }

   {
      // material - diffuse and specular, very shiny
      sglMaterial *mat = new sglMaterial;
      mat->setAmbient(0.f, 0.f, 0.f);
      mat->setDiffuse(0.5f, 0.5f, 0.5f);
      mat->setSpecular(1.f, 1.f, 1.f);
      mat->setEmission(0.f, 0.f, 0.f);
      mat->setShininess(100.0f);

      vector<sglStatelet*> state;
      state.push_back(new sglLighting(true));
      state.push_back(mat);
      sphere->setState(0x4, state); //, tcoords);
   }

   {
      // material - using color material
      sglMaterial *mat = new sglMaterial;
      mat->setAmbient(0.f, 0.f, 0.f);
      mat->setDiffuse(1.f, 1.f, 1.f);
      mat->setSpecular(1.f, 1.f, 1.f);
      mat->setShininess(8.0f);
      mat->setColorMaterial(sglMaterial::eFRONT_AND_BACK,
                            sglMaterial::eAMBIENT_AND_DIFFUSE);

      vector<sglStatelet*> state;
      state.push_back(new sglLighting(true));
      state.push_back(mat);
      sphere->setState(0x8, state); //, tcoords);
   }

   {
      // material - diffuse and emission
      sglMaterial *mat = new sglMaterial;
      mat->setAmbient(0.f, 0.f, 0.f);
      mat->setDiffuse(0.4f, 0.4f, 0.4f);
      mat->setSpecular(0.f, 0.f, 0.f);
      mat->setEmission(0.9f, 0.9f, 0.5f);
      mat->setShininess(0.0f);

      vector<sglStatelet*> state;
      state.push_back(new sglLighting(true));
      state.push_back(mat);
      sphere->setState(0x10, state); //, tcoords);
   }
   return sphere;
}

//----------------------------------------------------------------------------
sglGroup *makeLights()
{
   switch_node = new sglSwitch;
   switch_node->setVal(0);

   //-------------------------------------------------------------------------
   // make the directional light with a vector pointing in the right direction
   dlite = new sglDirectionalLight;
   dlite->setDiffuse(1.0f, 0.5f, 0.5f, 1.0f);
   dlite->setDirection(sglVec3f(0., 0., 1.));

   sglVec3f *coord = new sglVec3f[2];
   coord[0].set(0.f, 0.f, 0.f);
   coord[1].set(0.f, 0.f, 5.f);

   sglLineSet *gset = new sglLineSet;
   gset->setNumPrims(1);
   gset->setCoordList(coord);

   vector<sglStatelet*> state;
   state.push_back(new sglLighting(false));
   gset->setState(0, state);

   sglGeode *geode = new sglGeode;
   geode->addGeometry(gset);

   sglGroup *group = new sglGroup;
   group->addChild(dlite);
   group->addChild(geode);
   switch_node->addChild(group);

   //------------------------------------------------------------------------
   // make a positional light with a point at the correct position
   plite = new sglPositionalLight;
   plite->setPosition(sglVec3f(0., 0., 8.0f));
   plite->setDiffuse(0.f, 1.f, 0.f, 1.f);
   plite->setAmbient(0.25f, 0.25f, 0.25f, 1.f);
   plite->setFalloff(0.5f);
   plite->setAttenuation(constant_atten, linear_atten, quadratic_atten);
   plite->setBound(NULL, sglBound::eDYNAMIC);

   coord = new sglVec3f[1];
   coord[0].set(0.f, 0.f, 8.f);

   sglPointSet *gset1 = new sglPointSet;
   gset1->setNumPrims(1);
   gset1->setCoordList(coord);

   vector<sglStatelet*> state1;
   state1.push_back(new sglLighting(false));
   state1.push_back(new sglPointSize(4.0f));
   gset1->setState(0, state1);

   geode = new sglGeode;
   geode->addGeometry(gset1);

   group = new sglGroup;
   group->addChild(plite);
   group->addChild(geode);
   switch_node->addChild(group);

   //-------------------------------------------------------------------------
   // make a spot light with a vector pointing in the correct direction
   spot = new sglSpotLight;
   spot->setDiffuse(0.f, 0.f, 1.f, 1.f);
   spot->setPosition(sglVec3f(0., 0., 7.0f));
   spot->setDirection(sglVec3f(0., 0., -1.0f));
   spot->setCutoff((float)(M_PI*0.25));
   spot->setExponent(3.0f);

   coord = new sglVec3f[2];
   coord[0].set(0.f, 0.f, 7.f);
   coord[1].set(0.f, 0.f, 4.f);

   gset = new sglLineSet;
   gset->setNumPrims(1);
   gset->setCoordList(coord);

   vector<sglStatelet*> state2;
   state2.push_back(new sglLighting(false));
   gset->setState(0, state2);

   geode = new sglGeode;
   geode->addGeometry(gset);

   group = new sglGroup;
   group->addChild(spot);
   group->addChild(geode);
   switch_node->addChild(group);

   //------------------------------------------------------------------------
   transformf = new sglTransformf;
   transformf->addChild(switch_node);
   return transformf;
}

//----------------------------------------------------------------------------
void buildScenegraph()
{
   glLightModelfv(GL_LIGHT_MODEL_AMBIENT, no_light.getPtr());

   // make scene
   root = new sglScene;

   // add your stuff here.
   root->addChild(makeRoom());

   sglGeode *geode = new sglGeode;
   geode->addGeometry(makeSphere(3.0, 50, 25));
   root->addChild(geode);

   root->addChild(makeLights());

   // the camera (frustum)
   double aspect = (double)win_x/(double)win_y;
   view_camera = new sglPerspectiveCamera;
   view_camera->setFOV(M_PI/4.0, aspect, 0.1, 200.0);

   // the view platform
   view_platform = new sglViewPlatform(*view_camera);

   // the view transform
   view_trans = new sgluSphericalTransform<float>();
   view_trans->setMode(sgluSphericalTransform<float>::eZ_UP);
   view_trans->setRadius(33.0);
   view_trans->setCOI(0.0, 0.0, 0.0);
   view_trans->setTranslationScale(0.02f);
   view_trans->computeMatrix();
   view_trans->addChild(view_platform);
   root->addChild(view_trans);

   geode = new sglGeode;
   geode->addGeometry(new sglAxes(2.0f));
   root->addChild(geode);

   root->preDraw();

   view = new sglView;
   view->setScene(root, view_platform);
   view->setViewport(0, 0, win_x, win_y);

   override_statelets.push_back(&pfs);
   override_statelets.push_back(&pbs);
   view->setOverrideStatelets(override_statelets);
}

//----------------------------------------------------------------------------
int main(int argc, char** argv)
{
   glutInitDisplayString("stencil>=1 rgb double depth samples");
   glutInit(&argc, argv);
   glutInitWindowSize(480, 480);

   glutCreateWindow("sglLight/sglMaterial Test Program");

   sgl::initialize();
   mouse = sgluMouse::initialize();
   buildScenegraph();

   glutReshapeFunc(reshapeFunc);
   glutKeyboardFunc(keyboardFunc);
   glutSpecialFunc(specialFunc);
   glutDisplayFunc(displayFunc);
   glutIdleFunc(idleFunc);

   glutMainLoop();
   return 0;             /* ANSI C requires main to return int. */
}
