/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: testModelLoader.cpp
 *  Created: 20 March 1998
 *  Summary: test 3D model loaders from sgldb module
 *****************************************************************************/

#include <GL/glut.h>
#include <sgl.h>
#include <sglu.h>
#include <sgldb.h>
#include <stdio.h>

#include <sglOStream.hpp>
#include <sglNode.hpp>
#include <sglScene.hpp>
#include <sglLOD.hpp>
#include <sglGeode.hpp>
#include <sglDirectionalLight.hpp>
#include <sglTexture.hpp>
#include <sglModelLoader.hpp>

#include <sgluMouse.hpp>
#include <sgluSphericalTransform.hpp>

#include <sglViewPlatform.hpp>
#include <sglPerspectiveCamera.hpp>
#include <sglView.hpp>
#include <sglAxes.hpp>
#include <sglTexEnv.hpp>

sglView                   *view;
sglScene                  *root;

sglTexEnv                 *texEnv;

sglLighting               *lighting;
int lmode = 0;
sglShadeModel             *shading;
int smode = 0;

sgluMouse                 *mouse;
sgluSphericalTransform<float> *view_trans;
sglViewPlatform           *view_platform;
sglPerspectiveCamera      *view_camera;

deque<const sglStatelet*> override_statelets;
sglAntiAliasPoint   aa_point;
sglAntiAliasLine    aa_line;
sglAntiAliasPolygon aa_poly;

sglNode *model_root;
sglPolyFrontStyle *pfs;

bool polyfilled = 1;

bool print_stats = false;
double cull_time = 0;
double draw_time = 0;

double near_clip = 0.1;
double far_clip = 1000.0;
unsigned int win_x = 640, win_y = 480;

float trans_scale = 0.2f;
float rot_scale = 1.0f;

//----------------------------------------------------------------------------
void displayFunc()
{
   view->cullAndDraw();
   glutSwapBuffers();
}

//----------------------------------------------------------------------------
void reshapeFunc(int w, int h)
{
   view->setViewport(0, 0, w, h);

   mouse->win_size_x = win_x = w;
   mouse->win_size_y = win_y = h;

   double aspect = (double)w/(double)h;

   view_camera->setFOV(M_PI/4.0, aspect, near_clip, far_clip);
}

//----------------------------------------------------------------------------
void keyboardFunc(unsigned char key, int, int)
{
   switch (key)
   {
      case 27:
         glutDestroyWindow(glutGetWindow());
         exit(1);
         break;

      case 'F':
         if (pfs->getStyle() == sglPolyStyle::eLINE)
         {
            cout << "front style: eFILLED" << endl;
            pfs->setStyle(sglPolyStyle::eFILLED);
         }
         else
         {
            cout << "front style: LINE" << endl;
            pfs->setStyle(sglPolyStyle::eLINE);
         }
         break;

      case 'f':
         view->enableFrameRateOutput(!(view->getFrameRateOutputFlag()));
         break;

      case 'd':
         view->setTraversalType(!(view->getDPTraversalFlag()),
                               view->getStatsFlag());
         break;

      case '$':
         view->setTraversalType(view->getDPTraversalFlag(),
                               !(view->getStatsFlag()));
         break;

      case 'o':
         sgluOutputTree(root, cout, 3);
         break;

      case 's':
         sglLOD::setScaleFactor(sglLOD::getScaleFactor()/1.2f);
         break;
      case 'S':
         sglLOD::setScaleFactor(sglLOD::getScaleFactor()*1.2f);
         break;

			case 'l':
				 if( lmode == 0 )
				 {
					 lmode = 1;
					 lighting->enable( true );
					 cerr << "lighting enabled\n";
				 }
				 else
				 {
					 lmode = 0;
					 lighting->enable( false );
					 cerr << "lighting disabled\n";
				 }
				 break;

			case 'L':
				 if( smode == 0 )
				 {
					 smode = 1;
					 shading->setShadeModel( sglShadeModel::eSMOOTH );
					 cerr << "smooth shading\n";
				 }
				 else
				 {
					 smode = 0;
					 shading->setShadeModel( sglShadeModel::eFLAT );
					 cerr << "flat shading\n";
				 }
				 break;

      case 't':
         {
            float tmp = trans_scale*.1f;
            if (tmp > 0.f)
            {
               view_trans->setTranslationScale(tmp);
               trans_scale = tmp;
            }
            cerr << "translation scale = " << trans_scale << endl;
         }
         break;

      case 'T':
         trans_scale *= 10.f;
         view_trans->setTranslationScale(trans_scale);
         cerr << "translation scale = " << trans_scale << endl;
         break;

      case 'r':
         {
            float tmp = rot_scale*.1f;
            if (tmp > 0.f)
            {
               view_trans->setRotationScale(tmp);
               rot_scale = tmp;
            }
            cerr << "rotation scale = " << rot_scale << endl;
         }
         break;

      case 'R':
         rot_scale *= 10.f;
         view_trans->setRotationScale(rot_scale);
         cerr << "rotation scale = " << rot_scale << endl;
         break;

      case '/':
      case '?':
         cerr << "F - toggle front face between filled and wireframe" << endl;
         cerr << "f - toggle frame rate statistics" << endl;
         cerr << "d - toggle between SP and DP traversal" << endl;
         cerr << "$ - toggle scene graph statistics reporting" << endl;
         cerr << "l - toggle lighting on/off" << endl;
         cerr << "L - toggle shading model flat/smooth" << endl;
         cerr << "o - output scenegraph" << endl;
         cerr << "s/S - dec/inc static LOD scale factor" << endl;
         cerr << "t/T - dec/inc translation scale by order of mag." << endl;
         cerr << "r/T - dec/inc rotation scale by order of mag." << endl;
         cerr << "? - print help (this list)" << endl;
         cerr << "up/dn - increase/decrease texture LOD (if supported)" << endl;
         cerr << "ECSAPE - exit" << endl;
         break;
   }
}


//----------------------------------------------------------------------------
void specialFunc(int key, int, int)
{
   switch (key)
   {
      case GLUT_KEY_LEFT:
         break;
      case GLUT_KEY_RIGHT:
         break;

      case GLUT_KEY_UP:
				texEnv->setLODBias( texEnv->getLODBias() + 0.01 );
         break;
      case GLUT_KEY_DOWN:
				texEnv->setLODBias( texEnv->getLODBias() - 0.01 );
         break;
   }
}


//----------------------------------------------------------------------------
void idleFunc()
{
   view_trans->update(mouse);
   displayFunc();
}

//----------------------------------------------------------------------------
void buildScenegraph(const string &filename)
{
   // make scene and default state:
   root = new sglScene();

   // add your stuff here
   sglOStream::setThreshold(sglOStream::eINFO);

   // set the obj loader modes (soon to be used by all loaders?)
   //sglLoadFile_obj::setMode(OBJMODE_COMPUTE_FACETED_NORMALS);
   //sglLoadFile_obj::setMode(OBJMODE_COMPUTE_SMOOTH_NORMALS);
//   sglLoadFile_obj::setMode(OBJMODE_COMPUTE_FACETED_NORMALS|OBJMODE_STRIPIFY);
   //sglLoadFile_obj::setMode(OBJMODE_COMPUTE_SMOOTH_NORMALS|OBJMODE_STRIPIFY);

   unsigned int mode = sglModelLoader::eNONE;
   mode |= sglModelLoader::eCOMPUTE_FACETED_NORMALS;
//   mode |= sglModelLoader::eCOMPUTE_SMOOTH_NORMALS;
   mode |= sglModelLoader::eSTRIPIFY;

   // load the model
   model_root = sglModelLoader::loadModel(filename, mode);

   if (model_root == NULL)
   {
      sglPrintWarning() << "Unable to load model: " << filename << endl;
      exit(1);
   }

   root->addChild(model_root);

   // the camera (frustum)
   double aspect = (double)win_x/(double)win_y;
   view_camera = new sglPerspectiveCamera;
   view_camera->setFOV(M_PI/4.0, aspect, near_clip, far_clip);

   // the view platform
   view_platform = new sglViewPlatform(*view_camera);

   // the view transform
   view_trans = new sgluSphericalTransform<float>();
   view_trans->setMode(sgluSphericalTransform<float>::eZ_UP);
   view_trans->setRadius(30.0);
   view_trans->setCOI(0.0, 0.0, 0.0);
   view_trans->setTranslationScale(trans_scale);
   view_trans->computeMatrix();
   view_trans->addChild(view_platform);

#if 1
   sglDirectionalLight *l0 = new sglDirectionalLight;
   l0->setAmbient(1.f, 1.f, 1.f, 1.f);
   l0->setDiffuse(1.f, 1.f, 1.f, 1.f);
   l0->setSpecular(1.f, 1.f, 1.f, 1.f);
   l0->setDirection(sglVec3f(0.f, 0.f, 1.f));
   view_trans->addChild(l0);
#endif

   sglDirectionalLight *l1 = new sglDirectionalLight;
   l1->setAmbient(1.f, 1.f, 1.f, 1.f);
   l1->setDiffuse(1.f, 1.f, 1.f, 1.f);
   l1->setSpecular(1.f, 1.f, 1.f, 1.f);
   l1->setDirection(sglVec3f(-1.f, -1.f, -1.f));

   // infinite lights are constant with respect to the view point:
   view_trans->addChild(l1);

   root->addChild(view_trans);

   sglGeode *geode = new sglGeode;
   geode->addGeometry(new sglAxes(2.0f));
   root->addChild(geode);

   root->preDraw();
   sglSphereBoundf bsph = model_root->getBound();

   view = new sglView;
   view->setScene(root, view_platform);
   view->setViewport(0, 0, win_x, win_y);
   view->setClearColor(0.5f, 0.5f, 0.5f, 1.0f);

   if (!bsph.isEmpty())
   {
      far_clip = 4.0*bsph.getRadius();
      near_clip = far_clip*1.0e-4;
      view_camera->setFOV(M_PI/4.0, aspect, near_clip, far_clip);
      view_trans->setCOI(bsph.getCenter()[0],
                         bsph.getCenter()[1],
                         bsph.getCenter()[2]);
      view_trans->setRadius(bsph.getRadius());
      trans_scale = bsph.getRadius()/200.0;
      view_trans->setTranslationScale(trans_scale);
   }
#if 0
      far_clip = 1500.0;
      near_clip = 0.1;
      view_camera->setFOV(M_PI/4.0, aspect, near_clip, far_clip);
      view_trans->setCOI(0.0, 0.0, 0.0 );
      view_trans->setRadius(1500.0);
      trans_scale = 1500.0/200.0;
      view_trans->setTranslationScale(trans_scale);
#endif

   view_trans->computeMatrix();
   root->preDraw();

	 // lighting and shading
	 lighting = new sglLighting( false );
	 shading  = new sglShadeModel( sglShadeModel::eFLAT );

	 override_statelets.push_back( lighting );
	 override_statelets.push_back( shading );

   pfs = new sglPolyFrontStyle(sglPolyStyle::eFILLED);
	 texEnv = new sglTexEnv();
	 texEnv->setMode( sglTexEnv::eMODULATE );
   override_statelets.push_back(texEnv);
   override_statelets.push_back(pfs);
   //override_statelets.push_back(new sglCullMode(sglCullMode::eBACK));
   view->setOverrideStatelets(override_statelets);
}

//----------------------------------------------------------------------------
int main(int argc, char** argv)
{
   glutInit(&argc, argv);

   //=========================
   if (argc < 2)
   {
      cerr << "Usage: " << argv[0] << " <filename>." << endl;
      exit(1);
   }

   glutInitWindowSize(640, 480);
   glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);
   glutCreateWindow("testModelLoader");

   sgl::initialize();

   vector<string> data_path;
   data_path.push_back(".");
   data_path.push_back("./models");
   data_path.push_back("../models");
   sglDataLoader::setDataPath(data_path);

   string filename;
   if (sglDataLoader::findFile(argv[1], NULL, filename))
      buildScenegraph(filename);
   else
      buildScenegraph(argv[1]);
   mouse = sgluMouse::initialize();

   glutReshapeFunc(reshapeFunc);
   glutKeyboardFunc(keyboardFunc);
   glutSpecialFunc(specialFunc);
   glutDisplayFunc(displayFunc);
   glutIdleFunc(idleFunc);

   glutMainLoop();
   return 0;             /* ANSI C requires main to return int. */
}
