/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: testOcclusion.cpp
 *   Author: Tom Stimson
 *  Created: 2 April 2000
 *  Summary: Test program for occlusion culling.
 *****************************************************************************/

#include <GL/glut.h>
#include <sgl.h>
#include <sglu.h>

#include <sgluMouse.hpp>
#include <sgluSphericalTransform.hpp>

#include <sglScene.hpp>
#include <sglGeode.hpp>
#include <sglGeoSet.hpp>
#include <sglQuadSet.hpp>
#include <sglTriangleSet.hpp>
#include <sglMaterial.hpp>
#include <sglStatelet.hpp>

#include <sglViewPlatform.hpp>
#include <sglPerspectiveCamera.hpp>
#include <sglDirectionalLight.hpp>
#include <sglView.hpp>

sglView                   *view;
sglScene                  *root;

sgluMouse                 *mouse;
sgluSphericalTransform<float> *view_trans;
sglViewPlatform           *view_platform;
sglPerspectiveCamera      *view_camera;

deque<const sglStatelet*> override_statelets;
sglAntiAliasPoint   aa_point;
sglAntiAliasLine    aa_line;
sglAntiAliasPolygon aa_poly;

unsigned int win_x = 640, win_y = 480;

const unsigned int num_tris = 10000;

double total_time = -1;
double total_tris = 0;
sglTimespec tv, last_tv;

sglOcclusionCull *occlusion = NULL;
sglTransparency *transparency = NULL;

//----------------------------------------------------------------------------
void displayFunc()
{
   view->cullAndDraw();
   glutSwapBuffers();
}

//----------------------------------------------------------------------------
void reshapeFunc(int w, int h)
{
   view->setViewport(0, 0, w, h);

   mouse->win_size_x = win_x = w;
   mouse->win_size_y = win_y = h;

   double aspect = (double)w/(double)h;

   view_camera->setFOV(M_PI/3.0, aspect, 0.1, 200.0);
}

//----------------------------------------------------------------------------
void keyboardFunc(unsigned char key, int, int)
{
   switch (key)
   {
      case 27:
         glutDestroyWindow(glutGetWindow());
         cout << "avg. triangles per second = "
              << (total_tris/total_time) << endl;
         exit(1);
         break;

      case 'o':
         occlusion->enable(true);
         cout << "occlusion culling on" << endl;
         break;
      case 'O':
         occlusion->enable(false);
         cout << "occlusion culling off" << endl;
         break;

      case 't':
         transparency->enable(true);
         cout << "transparency on" << endl;
         break;
      case 'T':
         transparency->enable(false);
         cout << "transparency off" << endl;
         break;

      case 'a':
         aa_point.enable(!aa_point.isEnabled());
         cout << "point aa override = " << aa_point.isEnabled() << endl;
         break;

      case 'A':
         aa_line.enable(!aa_line.isEnabled());
         cout << "line aa override = " << aa_line.isEnabled() << endl;
         break;

      case 'z':
         aa_poly.enable(!aa_poly.isEnabled());
         cout << "polygon aa override = " << aa_poly.isEnabled() << endl;
         break;


      case 'f':
         view->enableFrameRateOutput(!(view->getFrameRateOutputFlag()));
         break;

      case 'd':
         view->setTraversalType(!(view->getDPTraversalFlag()),
                               view->getStatsFlag());
         break;

      case '$':
         view->setTraversalType(view->getDPTraversalFlag(),
                               !(view->getStatsFlag()));
         break;

      case '/':
      case '?':
         cerr << "o/O - enable/disable occlusion culling" << endl;
         cerr << "t/T - enable/disable transparency" << endl;
         cerr << "a - toggle point anti-aliasing" << endl;
         cerr << "A - toggle line anti-aliasing" << endl;
         cerr << "z - toggle polygon anti-aliasing" << endl;
         cerr << "f - toggle frame rate statistics" << endl;
         cerr << "d - toggle between SP and DP traversal" << endl;
         cerr << "$ - toggle scene graph statistics reporting" << endl;
         cerr << "? - print help (this list)" << endl;
         cerr << "ECSAPE - exit" << endl;
         break;
   }
}


//----------------------------------------------------------------------------
void specialFunc(int key, int, int)
{
   switch (key)
   {
      case GLUT_KEY_LEFT:
         break;
      case GLUT_KEY_RIGHT:
         break;
      case GLUT_KEY_UP:
         break;
      case GLUT_KEY_DOWN:
         break;
   }
}


//----------------------------------------------------------------------------
void idleFunc()
{
   view_trans->update(mouse);
   displayFunc();

   glutPostRedisplay();

   sglTimespec::getSysTime(tv);
   double elapsed_time = sglTimespec::convertToDouble(tv - last_tv);

   if (elapsed_time > 1.0)
   {
      // Ignore the first time, as it is usually an outlyer.
      if (total_time == -1)
      {
         total_time = 0;
      }
      else
      {
         total_time += elapsed_time;
         total_tris += num_tris;
      }

      last_tv = tv;
   }
}


//----------------------------------------------------------------------------
sglNode* makeTris()
{
    unsigned int num_verts = 3 * num_tris;

    sglGeode* geode = new sglGeode();
    sglTriangleSet* gset = new sglTriangleSet();
    sglVec3f* verts = new sglVec3f[num_verts];
    sglVec3f* norms = new sglVec3f[num_verts];
    sglVec4f* color = new sglVec4f[num_verts];
    for (unsigned int i=0; i<num_verts; ++i)
    {
        for (unsigned int j=0; j<3; ++j)
        {
            verts[i][j] = rand();
            verts[i][j] /= (float)RAND_MAX;
            color[i][j] = verts[i][j];
            verts[i][j] *= 20.0f;
            verts[i][j] -= 10.0f;
        }
        color[i][3] = 1.0f;
        norms[i] = verts[i];
        norms[i].normalize();
    }

    gset->setCoordList(verts);
    gset->setNormalList(sglGeoSet::ePER_VERTEX, norms);
    gset->setColorList(sglGeoSet::ePER_VERTEX, color);
    gset->setNumPrims(num_tris);

    sglMaterial *mat = new sglMaterial();
    mat->setColorMaterial(sglMaterial::eFRONT_AND_BACK, sglMaterial::eDIFFUSE);

    occlusion = new sglOcclusionCull();
    transparency = new sglTransparency();

    vector<sglStatelet*> state;
    sglLighting *li = new sglLighting(true);

    state.push_back(li);
    state.push_back(mat);
    occlusion->enable(true);
    state.push_back(occlusion);
    transparency->enable(false);
    state.push_back(transparency);

//     sglCullMode *cull_mode = new sglCullMode();
//     cull_mode->setCullMode(sglCullMode::eOFF);
//     state.push_back(cull_mode);

    gset->setState(0, state);
    geode->addGeometry(gset);

    sglQuadSet* quad = new sglQuadSet();
    verts = new sglVec3f[4];
    norms = new sglVec3f[1];
    color = new sglVec4f[1];
    verts[0] = sglVec3f(12.0f,  12.0f,  12.0f);
    verts[1] = sglVec3f(12.0f, -12.0f,  12.0f);
    verts[2] = sglVec3f(12.0f, -12.0f, -12.0f);
    verts[3] = sglVec3f(12.0f,  12.0f, -12.0f);
    norms[0] = sglVec3f(1.0f, 0, 0);
    color[0] = sglVec4f(1.0f, 0, 0, 1.0f);

    quad->setCoordList(verts);
    quad->setNormalList(sglGeoSet::eOVERALL, norms);
    quad->setColorList(sglGeoSet::eOVERALL, color);
    quad->setNumPrims(1);

    mat = new sglMaterial();
    mat->setColorMaterial(sglMaterial::eFRONT_AND_BACK, sglMaterial::eDIFFUSE);

    state.clear();
    state.push_back(li);
    state.push_back(mat);

    quad->setState(0, state);
    geode->addGeometry(quad);

    return geode;
}


//----------------------------------------------------------------------------
void buildScenegraph()
{
   // make scene and default state:
   root = new sglScene();

   // add your stuff here
   root->addChild(makeTris());

   // light source.
   sglDirectionalLight *dlite = new sglDirectionalLight;
   dlite->setAmbient(0.f, 0.f, 0.f, 1.f);
   dlite->setDiffuse(1.f, 1.f, 1.f, 1.f);
   dlite->setSpecular(1.f,1.f, 1.f, 1.f);
   dlite->setDirection(sglVec3f(1.f, 1.f, 1.f));
   dlite->setOn(true);

   root->addChild(dlite);

   // the camera (frustum)
   double aspect = (double)win_x/(double)win_y;
   view_camera = new sglPerspectiveCamera;
   view_camera->setFOV(M_PI/3.0, aspect, 0.1, 200.0);

   // the view platform
   view_platform = new sglViewPlatform(*view_camera);

   // the view transform
   view_trans = new sgluSphericalTransform<float>();
   view_trans->setMode(sgluSphericalTransform<float>::eZ_UP);
   view_trans->setRadius(30.0);
   view_trans->setCOI(0.0, 0.0, 0.0);
   view_trans->setTranslationScale(0.02f);
   view_trans->computeMatrix();
   view_trans->addChild(view_platform);
   root->addChild(view_trans);

   root->preDraw();

   view = new sglView;
   view->setScene(root, view_platform);
   view->setViewport(0, 0, win_x, win_y);

   override_statelets.push_back(&aa_point);
   override_statelets.push_back(&aa_line);
   override_statelets.push_back(&aa_poly);
   override_statelets.push_back(new sglShadeModel(sglShadeModel::eFLAT));
   view->setOverrideStatelets(override_statelets);
}

//----------------------------------------------------------------------------
int main(int argc, char** argv)
{
   glutInit(&argc, argv);

   glutInitWindowSize(700, 512);
   glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);
   glutCreateWindow("testOcclusion");

   sgl::initialize();
   if (sgl::s_occlusion_capable)
      cout << "Occlusion culling supported" << endl;
   else
      cout << "Occlusion culling NOT supported" << endl;

   mouse = sgluMouse::initialize();
   buildScenegraph();

   glutReshapeFunc(reshapeFunc);
   glutKeyboardFunc(keyboardFunc);
   glutSpecialFunc(specialFunc);
   glutDisplayFunc(displayFunc);
   glutIdleFunc(idleFunc);

   sglTimespec::getSysTime(last_tv);

   idleFunc();

   glutMainLoop();
   return 0;             /* ANSI C requires main to return int. */
}
