/*****************************************************************************
 * SGL: A Scene Graph Library
 *
 * Copyright (C) 1997-2001  Scott McMillan   All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************
 *     File: testPick.cpp
 *  Created: 21 April 2001
 *  Summary: GLUT-based sgl testbed for picking
 *****************************************************************************/

#include <GL/glut.h>
#include <sgl.h>
#include <sglu.h>

#include <sgluMouse.hpp>
#include <sgluSphericalTransform.hpp>

#include <sglScene.hpp>
#include <sglScale.hpp>
#include <sglTranslate.hpp>
#include <sglTransform.hpp>
#include <sglBillboard.hpp>
#include <sglUnScale.hpp>
#include <sglGeode.hpp>
#include <sglTriangleSet.hpp>
#include <sglTriangleStripSet.hpp>

#include <sglFrustum.hpp>
#include <sglPick.hpp>
#include <sglSphereBound.hpp>

#include <sglViewPlatform.hpp>
#include <sglPerspectiveCamera.hpp>
#include <sglView.hpp>
#include <sglAxes.hpp>
#include <sglCube.hpp>
#include <sglSphere.hpp>
#include <sglCone.hpp>
#include <sglPointSet.hpp>
#include <sglLineSet.hpp>

#include <stdio.h>

sglView                   *view;
sglScene                  *root;

sgluMouse                 *mouse;
sgluSphericalTransform<float> *view_trans;
sglViewPlatform           *view_platform;
sglPerspectiveCamera      *view_camera;

deque<const sglStatelet*> override_statelets;
sglAntiAliasPoint   aa_point;
sglAntiAliasLine    aa_line;
sglAntiAliasPolygon aa_poly;
sglHighlight        hilite;

vector<sglPickf::PickResult*> m_results_pool(4);
sglPickf    picker(&m_results_pool);

sglFrustumf pick_frustum;
sglMat4f    pick_view_matrix;
bool        pick_flag = false;
bool        button_flag = false;

sglVec3f pos(0.0, 0.0, 0.0);
double heading = 0.0, pitch = 0;
float len = 30.0;
double near_clip = 0.1;
double far_clip = 200.0;
double fov_y = M_PI/3.0;

sglVec4f white_color(1.f, 1.f, 1.f, 1.f);
sglVec4f   red_color(1.f, 0.f, 0.f, 1.f);

unsigned int win_x = 640, win_y = 480;

//----------------------------------------------------------------------------
void preDrawFunc(void *)
{
   glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
   glEnable(GL_DEPTH_TEST);

#ifdef SGL_FRUSTUM_DEBUG
   // draw the pick frustum used in the last pick
   glPushMatrix();
   glMultMatrixf(pick_view_matrix.getPtr());
   pick_frustum.drawImmediate();
   glPopMatrix();
#endif
}

//----------------------------------------------------------------------------
void displayFunc()
{
   view->cullAndDraw();
   glutSwapBuffers();
}

//----------------------------------------------------------------------------
void reshapeFunc(int w, int h)
{
   view->setViewport(0, 0, w, h);

   mouse->win_size_x = win_x = w;
   mouse->win_size_y = win_y = h;

   double aspect = (double)w/(double)h;

   view_camera->setFOV(fov_y, aspect, near_clip, far_clip);
}

//----------------------------------------------------------------------------
void pickScreen(int x, int y, unsigned int pixel_size)
{
   // clear the previous result
   {
      const vector<sglPickf::PickResult*> &results = picker.getResults();
      if (results.size())
      {
         for (vector<sglPickf::PickResult*>::const_iterator elt =
                 results.begin();
              elt != results.end();
              ++elt)
         {
            if ((*elt)->m_geometry)
            {
               sglGeoSet *source_gset =
                  dynamic_cast<sglGeoSet*>(
                     const_cast<sglDrawable*>((*elt)->m_geometry));
               if (source_gset)
                  source_gset->setColorList(sglGeoSet::eOVERALL, &white_color);
            }
         }
      }
   }

   // convert screen pixel locations to 3D location
   double xfactor_left   =  2.0*((double)(x - pixel_size))/(double)win_x - 1.0;
   double xfactor_right  =  2.0*((double)(x + pixel_size))/(double)win_x - 1.0;

   double yfactor_top    = -2.0*((double)(y - pixel_size))/(double)win_y + 1.0;
   double yfactor_bottom = -2.0*((double)(y + pixel_size))/(double)win_y + 1.0;

   cout << "pick location : " << (xfactor_left + xfactor_right)/2.0 << ", "
        << (yfactor_top + yfactor_bottom)/2.0 << endl;

   double ysize = near_clip*tan(fov_y*0.5);
   double near_bottom = yfactor_bottom*ysize;
   double near_top    = yfactor_top*ysize;

   double xsize = ysize*(double)win_x/(double)win_y;
   double near_left   = xfactor_left*xsize;
   double near_right  = xfactor_right*xsize;

   pick_frustum.buildPerspective(near_left, near_right, near_bottom, near_top,
                                 near_clip, far_clip);

   sglMat4f view_matrix;
   view_platform->computeViewMatrix(view_matrix);
   sglFrustumf frustum = view_camera->getFrustum();

   picker.reset();
   picker.setPickFrustum(view_matrix, pick_frustum);
   picker.setModelViewMatrix(view_matrix);
   picker.setViewFrustum(frustum);
   picker.setViewportSize(win_x, win_y);
   pick_view_matrix.inverseAffine(view_matrix);

   //-----------------------------------------------
   root->pick(picker, 0x3f);
   //-----------------------------------------------

   const vector<sglPickf::PickResult*> &results = picker.getResults();

   if (results.size())
   {
      cerr << "pick results: " << results.size() << endl;
      for (vector<sglPickf::PickResult*>::const_iterator elt = results.begin();
           elt != results.end();
           ++elt)
      {
         cerr << "---------------------------------------------------" << endl;
         cerr << "Geode name: "
              << (((*elt)->m_geode==NULL) ? "nil" : (*elt)->m_geode->getName())
              << ", Geometry name: "
              << (((*elt)->m_geometry==NULL) ? "nil" :
                  (*elt)->m_geometry->getName())
              << ", path size: " << ((*elt)->m_path).size() << endl;
         cerr << "Transform: " << (*elt)->m_xform << endl;

         if (((*elt)->m_path).size())
         {
            for (vector<sglNode*>::iterator nelt = ((*elt)->m_path).begin();
                 nelt != ((*elt)->m_path).end();
                 ++nelt)
            {
               cerr << "............................................." << endl;
               (*nelt)->printInfo(cerr, "   ");
            }
         }

         if ((*elt)->m_geometry)
         {
            sglGeoSet *source_gset =
               dynamic_cast<sglGeoSet*>(
                  const_cast<sglDrawable*>((*elt)->m_geometry));
            if (source_gset)
               source_gset->setColorList(sglGeoSet::eOVERALL, &red_color);

            cerr << "............................................." << endl;
            (*elt)->m_geometry->printInfo(cerr, "   ");
         }
         cerr << "............................................." << endl;
      }
      cerr << "---------------------------------------------------" << endl;
   }
   else
   {
      cerr << "pick result: empty" << endl;
   }
}

//----------------------------------------------------------------------------
void handleMouse(int button, int state, int x, int y)
{
   if (!(glutGetModifiers()&GLUT_ACTIVE_CTRL))
   {
      sgluMouse::handleMouse(button, state, x, y);
      return;
   }

   if (state == GLUT_DOWN && button == GLUT_LEFT_BUTTON && !button_flag)
   {
      button_flag = true;
      pickScreen(x, y, 2);
   }
   else if (state == GLUT_UP && button == GLUT_LEFT_BUTTON)
   {
      button_flag = false;
   }
}

//----------------------------------------------------------------------------
void keyboardFunc(unsigned char key, int, int)
{
   switch (key)
   {
      case 27:
         glutDestroyWindow(glutGetWindow());
         exit(1);
         break;

      case 'a':
         aa_point.enable(!aa_point.isEnabled());
         cout << "point aa override = " << aa_point.isEnabled() << endl;
         break;

      case 'A':
         aa_line.enable(!aa_line.isEnabled());
         cout << "line aa override = " << aa_line.isEnabled() << endl;
         break;

      case 'z':
         aa_poly.enable(!aa_poly.isEnabled());
         cout << "polygon aa override = " << aa_poly.isEnabled() << endl;
         break;

      case 'h':
         hilite.enable(!hilite.isEnabled());
         cout << "bbox highlighting = " << hilite.isEnabled() << endl;
         break;

      case 'o':
         sgluOutputTree(root, cout, 3);
         break;

      case '?':
      case '/':
         cout << "CTRL-<left mouse button> to pick" << endl;
         cout << "h - toggle bounding box highlighting" << endl;
         cout << "a - toggle point anti-aliasing" << endl;
         cout << "A - toggle line anti-aliasing" << endl;
         cout << "z - toggle polygon anti-aliasing" << endl;
         cout << "o - output scenegraph" << endl;
         cout << "? - print help (this list)" << endl;
         cout << "ESCAPE - exit program" << endl;
         break;
   }
}


//----------------------------------------------------------------------------
void specialFunc(int key, int, int)
{
   switch (key)
   {
      case GLUT_KEY_LEFT:
         break;
      case GLUT_KEY_RIGHT:
         break;
      case GLUT_KEY_UP:
         break;
      case GLUT_KEY_DOWN:
         break;
   }
}

//----------------------------------------------------------------------------
void idleFunc()
{
   view_trans->update(mouse);
   displayFunc();
}

//----------------------------------------------------------------------------
void addPoints()
{
   sglGeode *geode = new sglGeode;
   geode->setName("4_points");

   vector<sglStatelet*> state;
   sglPointSize *psize = new sglPointSize;
   psize->setPointSize(2.0f);
   state.push_back(psize);

   for (unsigned int i=0; i<4; i++)
   {
      sglPointSet *gset = new sglPointSet;
      char name[8];
      sprintf(name, "pnt %d", i);
      gset->setName(name);
      gset->setNumPrims(1);
      sglVec3f *vertices = new sglVec3f[1];
      vertices[0].set(i*4.f - 8.0f, 0.0f, 8.0f);
      gset->setCoordList(vertices);
      gset->setState(0UL, state);

      geode->addGeometry(gset);
   }

   root->addChild(geode);
}

//----------------------------------------------------------------------------
void addLines()
{
   sglGeode *geode = new sglGeode;
   geode->setName("3_lines");

   sglVec3f vertices[6];
   vertices[0].set(-2.0f, 0.0f, -8.0f);
   vertices[1].set( 2.0f, 0.0f, -8.0f);
   vertices[2].set( 4.0f,-2.0f, -8.0f);
   vertices[3].set( 4.0f, 2.0f, -8.0f);
   vertices[4].set( 6.0f, 0.0f, -6.0f);
   vertices[5].set( 6.0f, 0.0f, -10.0f);

   for (unsigned int i=0; i<3; ++i)
   {
      sglLineSet *gset = new sglLineSet;
      char name[8];
      sprintf(name, "line %d", i);
      gset->setName(name);
      gset->setNumPrims(1);
      sglVec3f *verts = new sglVec3f[2];
      verts[0] = vertices[i*2];
      verts[1] = vertices[i*2 + 1];
      gset->setCoordList(verts);

      geode->addGeometry(gset);
   }

   root->addChild(geode);
}

//----------------------------------------------------------------------------
sglGeoSet *makeQuad()
{
   sglVec3f *vertex = new sglVec3f[5];
   sglVec4f *color = new sglVec4f[1];

   color[0][0] = color[0][1] = color[0][2] = color[0][3] = 1.f;

   float size = 20.0f;
   vertex[0][0] =  0.0;
   vertex[0][1] =  size;
   vertex[0][2] =  size;

   vertex[1][0] =  0.0;
   vertex[1][1] =  size;
   vertex[1][2] = -size;

   vertex[2][0] =  0.0;
   vertex[2][1] = -size;
   vertex[2][2] = -size;

   vertex[3][0] =  0.0;
   vertex[3][1] = -size;
   vertex[3][2] =  size;

   vertex[4][0] =  0.0;
   vertex[4][1] =  size;
   vertex[4][2] =  size;

   sglQuadSet *gset = new sglQuadSet();
   gset->setCoordList(vertex);
   gset->setNumPrims(1);

   return gset;
}

//----------------------------------------------------------------------------
sglGeoSet *makeBBQuad()
{
   sglVec3f *vertex = new sglVec3f[5];
   sglVec4f *color = new sglVec4f[1];

   color[0][0] = color[0][1] = color[0][2] = color[0][3] = 1.f;

   float size = 2.0f;
   vertex[0][0] =  -size;
   vertex[0][1] =  -size;
   vertex[0][2] =  0.0;

   vertex[1][0] =  size;
   vertex[1][1] = -size;
   vertex[1][2] =  0.0;

   vertex[2][0] =  size;
   vertex[2][1] =  size;
   vertex[2][2] =  0.0;

   vertex[3][0] = -size;
   vertex[3][1] =  size;
   vertex[3][2] =  0.0;

   vertex[4][0] = -size;
   vertex[4][1] = -size;
   vertex[4][2] =  0.0;

   sglQuadSet *gset = new sglQuadSet();
   gset->setCoordList(vertex);
   gset->setNumPrims(1);

   return gset;
}

//----------------------------------------------------------------------------
void buildScenegraph()
{
   // make scene and default state:
   root = new sglScene();

   // add your stuff here
   addPoints();
   addLines();

   // The cube
   sglGeode *geode = new sglGeode;
   geode->setName("cube");
   geode->addGeometry(new sglCube(6.0f, 6.0f, 6.0f));

   sglTransformf *transformf = new sglTransformf;
   transformf->setName("cubeTransform");
   transformf->setRotation((float)(M_PI/4.0f), 1.0f, 0.32f);
   //transformf->setScale(6.0f, 6.0f, 6.0f);
   transformf->addChild(geode);

   sglTranslatef *trans = new sglTranslatef;
   trans->setName("cubeTranslate");
   trans->setTranslation(15.0f, 5.0f, -3.0f);
   trans->addChild(transformf);
   root->addChild(trans);

   // The cone
   sglCone *cone = new sglCone(10, 2.0f, 1.0f);
   geode = new sglGeode;
   geode->setName("cone");
   geode->addGeometry(cone);

   sglScalef *scalef = new sglScalef;
   scalef->setName("coneScale");
   scalef->setScale(5.0f);
   scalef->addChild(geode);

   trans = new sglTranslatef;
   trans->setName("coneTranslate");
   trans->setTranslation(0.f, 18.0f, -5.0f);
   trans->addChild(scalef);
   root->addChild(trans);

   // The sphere
   trans = new sglTranslatef;
   trans->setTranslation(-15.0f, 5.0f, 0.0f);
   geode = new sglGeode;
   geode->setName("sphere");
   geode->addGeometry(new sglSphere(12, 6, 5.0f));
   trans->addChild(geode);
   root->addChild(trans);

   // The unscaled quad
   sglUnScale *unscale = new sglUnScale(true);
   geode = new sglGeode();
   geode->setName("unscale quad");
   unscale->addChild(geode);
   geode->addGeometry(makeQuad());
   trans = new sglTranslatef();
   trans->setTranslation(-2, 0, 0);
   trans->addChild(unscale);
   root->addChild(trans);

   // The billboard quad
   sglBillboard *bb = new sglBillboard;
   geode = new sglGeode();
   geode->setName("bb quad");
   bb->addChild(geode);
   geode->addGeometry(makeBBQuad());
   trans = new sglTranslatef();
   trans->setTranslation(0, -8, 0);
   trans->addChild(bb);
   root->addChild(trans);

   // ================================================================
   picker.setMode(sglProbe::eGEODE|sglProbe::eGEOMETRY|
                  sglProbe::eTRANSFORM|sglProbe::ePATH|
                  sglProbe::eBILLBOARD_VIEW|
                  sglProbe::eUNSCALE_VIEW);
   picker.setDiscriminatorMask(~0x000000001);  // mask out the tris
   // ================================================================

   // the camera (frustum)
   double aspect = (double)win_x/(double)win_y;
   view_camera = new sglPerspectiveCamera;
   view_camera->setFOV(M_PI/3.0, aspect, 0.1, 200.0);

   // the view platform
   view_platform = new sglViewPlatform(*view_camera);

   // the view transform
   view_trans = new sgluSphericalTransform<float>();
   view_trans->setMode(sgluSphericalTransform<float>::eZ_UP);
   view_trans->setRadius(30.0);
   view_trans->setCOI(0.0, 0.0, 0.0);
   view_trans->setTranslationScale(0.02f);
   view_trans->computeMatrix();
   view_trans->addChild(view_platform);
   root->addChild(view_trans);

   geode = new sglGeode;
   geode->setName("axes");
   geode->addGeometry(new sglAxes(2.0f));
   root->addChild(geode);

   root->preDraw();

   view = new sglView;
   view->setScene(root, view_platform);
   view->setViewport(0, 0, win_x, win_y);
   view->setDrawCallback(preDrawFunc, NULL);

   override_statelets.push_back(new sglCullMode(sglCullMode::eOFF));
   override_statelets.push_back(new sglPolyBackStyle(sglPolyStyle::eLINE));
   override_statelets.push_back(new sglPolyFrontStyle(sglPolyStyle::eLINE));
   override_statelets.push_back(new sglShadeModel(sglShadeModel::eFLAT));
   override_statelets.push_back(&hilite);
   override_statelets.push_back(&aa_point);
   override_statelets.push_back(&aa_line);
   override_statelets.push_back(&aa_poly);
   view->setOverrideStatelets(override_statelets);
}

//----------------------------------------------------------------------------
int main(int argc, char** argv)
{
   glutInit(&argc, argv);

   glutInitWindowSize(700, 512);
   glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);
   glutCreateWindow("testIsectDiscrim");

   sgl::initialize();
   mouse = sgluMouse::initialize();
   buildScenegraph();

   // Override the mouse functions for picking
   glutMouseFunc(handleMouse);

   glutReshapeFunc(reshapeFunc);
   glutKeyboardFunc(keyboardFunc);
   glutSpecialFunc(specialFunc);
   glutDisplayFunc(displayFunc);
   glutIdleFunc(idleFunc);

   glutMainLoop();
   return 0;             /* ANSI C requires main to return int. */
}
